!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types and set/get functions for auxiliary denisty matrix methods
!> \par History
!>      11.2014 created [Ole Schuett]
!> \author Ole Schuett
! **************************************************************************************************
MODULE admm_dm_types
   USE cp_control_types,                ONLY: admm_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_release,&
                                              cp_dbcsr_type
   USE input_constants,                 ONLY: do_admm_basis_projection,&
                                              do_admm_purify_mcweeny
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: admm_dm_type, mcweeny_history_type
   PUBLIC :: admm_dm_create, admm_dm_release

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'admm_dm_types'

   TYPE mcweeny_history_type
      TYPE(cp_dbcsr_type)                        :: m
      INTEGER                                    :: count = -1
      TYPE(mcweeny_history_type), POINTER        :: next => Null()
   END TYPE mcweeny_history_type

   TYPE mcweeny_history_p_type
      TYPE(mcweeny_history_type), POINTER        :: p => Null()
   END TYPE mcweeny_history_p_type

   TYPE admm_dm_type
      LOGICAL                                  :: purify = .FALSE.
      INTEGER                                  :: method = -1
      TYPE(cp_dbcsr_type), POINTER             :: matrix_a => Null()
      REAL(KIND=dp)                            :: eps_filter = 1e-20_dp
      INTEGER                                  :: mcweeny_max_steps = 100
      INTEGER, DIMENSION(:, :), POINTER         :: block_map => Null()
      TYPE(mcweeny_history_p_type), &
         DIMENSION(:), POINTER                  :: mcweeny_history => Null()
   END TYPE

CONTAINS

! **************************************************************************************************
!> \brief Create a new admm_dm type
!> \param admm_dm ...
!> \param admm_control ...
!> \param nspins ...
!> \param natoms ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE admm_dm_create(admm_dm, admm_control, nspins, natoms)
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(admm_control_type), POINTER                   :: admm_control
      INTEGER, INTENT(IN)                                :: nspins, natoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'admm_dm_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, iblock, j, jatom

      ALLOCATE (admm_dm)
      ! copy settings from admm_control
      admm_dm%purify = (admm_control%purification_method == do_admm_purify_mcweeny)
      admm_dm%method = admm_control%method
      admm_dm%eps_filter = admm_control%eps_filter

      ALLOCATE (admm_dm%mcweeny_history(nspins))

      IF (admm_dm%method /= do_admm_basis_projection) THEN
         ! create block map
         ALLOCATE (admm_dm%block_map(natoms, natoms))
         admm_dm%block_map(:, :) = 0
         DO iblock = 1, SIZE(admm_control%blocks)
            DO i = 1, SIZE(admm_control%blocks(iblock)%list)
               iatom = admm_control%blocks(iblock)%list(i)
               DO j = 1, SIZE(admm_control%blocks(iblock)%list)
                  jatom = admm_control%blocks(iblock)%list(j)
                  admm_dm%block_map(iatom, jatom) = 1
               END DO
            END DO
         END DO
      ENDIF
   END SUBROUTINE admm_dm_create

! **************************************************************************************************
!> \brief Release a admm_dm type
!> \param admm_dm ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE admm_dm_release(admm_dm)
      TYPE(admm_dm_type), POINTER                        :: admm_dm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'admm_dm_release', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(admm_dm)) RETURN

      IF (ASSOCIATED(admm_dm%matrix_a)) THEN
         CALL cp_dbcsr_release(admm_dm%matrix_a)
         DEALLOCATE (admm_dm%matrix_a)
      ENDIF

      IF (ASSOCIATED(admm_dm%block_map)) &
         DEALLOCATE (admm_dm%block_map)

      DEALLOCATE (admm_dm%mcweeny_history)
      DEALLOCATE (admm_dm)

   END SUBROUTINE admm_dm_release

END MODULE admm_dm_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      06.2004 created
!> \author jgh
! **************************************************************************************************
MODULE sto_ng

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'sto_ng'

   PUBLIC :: get_sto_ng

! **************************************************************************************************
!> \brief return STO-NG parameters;
!>        INPUT: zeta (Slater exponent)
!>               n    (Expansion length)
!>               nq   (principle quantum number)
!>               lq   (angular quantum number)
!>        OUTPUT: alpha (n Gaussian exponents)
!>                coef  (n corresponding expansion coefficients)
!> \par History
!>      06.2004 created using parameters from : RF Stewart JCP 52 p431 1970
!>      01.2008 added newly generated parameters for 6s - 7i
!> \author jgh
CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param zeta ...
!> \param n ...
!> \param nq ...
!> \param lq ...
!> \param alpha ...
!> \param coef ...
! **************************************************************************************************
   SUBROUTINE get_sto_ng(zeta, n, nq, lq, alpha, coef)

      REAL(KIND=dp), INTENT(IN)                          :: zeta
      INTEGER, INTENT(IN)                                :: n, nq, lq
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: alpha, coef

      CHARACTER(len=*), PARAMETER :: routineN = 'get_sto_ng', routineP = moduleN//':'//routineN

      INTEGER                                            :: m, na, nc

      na = SIZE(alpha)
      nc = SIZE(coef)

      IF(.NOT.(n >= 1 .AND. n <= 6))CALL cp__a("aobasis/sto_ng.F",60)
      IF(.NOT.(na >= n))CALL cp__a("aobasis/sto_ng.F",61)
      IF(.NOT.(nc >= n))CALL cp__a("aobasis/sto_ng.F",62)
      IF(.NOT.(zeta > 0.0_dp))CALL cp__a("aobasis/sto_ng.F",63)
      IF(.NOT.(nq >= 1))CALL cp__a("aobasis/sto_ng.F",64)
      IF(.NOT.(lq >= 0))CALL cp__a("aobasis/sto_ng.F",65)

      m = (nq*(nq-1))/2+lq+1
      SELECT CASE (m)
      CASE (1) ! 1s
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 2.709496091e-1_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 8.518186635e-1_dp; coef(1) = 4.301284983e-1_dp
            alpha(2) = 1.516232927e-1_dp; coef(2) = 6.789135305e-1_dp
         CASE (3)
            alpha(1) = 2.227660584e+0_dp; coef(1) = 1.543289673e-1_dp
            alpha(2) = 4.057711562e-1_dp; coef(2) = 5.353281423e-1_dp
            alpha(3) = 1.098175104e-1_dp; coef(3) = 4.446345422e-1_dp
         CASE (4)
            alpha(1) = 5.216844534e+0_dp; coef(1) = 5.675242080e-2_dp
            alpha(2) = 9.546182760e-1_dp; coef(2) = 2.601413550e-1_dp
            alpha(3) = 2.652034102e-1_dp; coef(3) = 5.328461143e-1_dp
            alpha(4) = 8.801862774e-2_dp; coef(4) = 2.916254405e-1_dp
         CASE (5)
            alpha(1) = 1.130563696e+1_dp; coef(1) = 2.214055312e-2_dp
            alpha(2) = 2.071728178e+0_dp; coef(2) = 1.135411520e-1_dp
            alpha(3) = 5.786484833e-1_dp; coef(3) = 3.318161484e-1_dp
            alpha(4) = 1.975724573e-1_dp; coef(4) = 4.825700713e-1_dp
            alpha(5) = 7.445271746e-2_dp; coef(5) = 1.935721966e-1_dp
         CASE (6)
            alpha(1) = 2.310303149e+1_dp; coef(1) = 9.163596280e-3_dp
            alpha(2) = 4.235915534e+0_dp; coef(2) = 4.936149294e-2_dp
            alpha(3) = 1.185056519e+0_dp; coef(3) = 1.685383049e-1_dp
            alpha(4) = 4.070988982e-1_dp; coef(4) = 3.705627997e-1_dp
            alpha(5) = 1.580884151e-1_dp; coef(5) = 4.164915298e-1_dp
            alpha(6) = 6.510953954e-2_dp; coef(6) = 1.303340841e-1_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",99,"Illegal expansion")
         END SELECT
      CASE (2) ! 2s
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 1.012151084e-1_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 1.292278611e-1_dp; coef(1) = 7.470867124e-1_dp
            alpha(2) = 4.908584205e-2_dp; coef(2) = 2.855980556e-1_dp
         CASE (3)
            alpha(1) = 2.581578398e+0_dp; coef(1) = -5.994474934e-2_dp
            alpha(2) = 1.567622104e-1_dp; coef(2) = 5.960385398e-1_dp
            alpha(3) = 6.018332272e-2_dp; coef(3) = 4.581786291e-1_dp
         CASE (4)
            alpha(1) = 1.161525551e+1_dp; coef(1) = -1.198411747e-2_dp
            alpha(2) = 2.000243111e+0_dp; coef(2) = -5.472052539e-2_dp
            alpha(3) = 1.607280687e-1_dp; coef(3) = 5.805587176e-1_dp
            alpha(4) = 6.125744532e-2_dp; coef(4) = 4.770079976e-1_dp
         CASE (5)
            alpha(1) = 8.984956862e+0_dp; coef(1) = -1.596349096e-2_dp
            alpha(2) = 1.673710636e+0_dp; coef(2) = -5.685884883e-2_dp
            alpha(3) = 1.944726668e-1_dp; coef(3) = 3.698265599e-1_dp
            alpha(4) = 8.806345634e-2_dp; coef(4) = 5.480512593e-1_dp
            alpha(5) = 4.249068522e-2_dp; coef(5) = 1.472634893e-1_dp
         CASE (6)
            alpha(1) = 2.768496241e+1_dp; coef(1) = -4.151277819e-3_dp
            alpha(2) = 5.077140627e+0_dp; coef(2) = -2.067024148e-2_dp
            alpha(3) = 1.426786050e+0_dp; coef(3) = -5.150303337e-2_dp
            alpha(4) = 2.040335729e-1_dp; coef(4) = 3.346271174e-1_dp
            alpha(5) = 9.260298399e-2_dp; coef(5) = 5.621061301e-1_dp
            alpha(6) = 4.416183978e-2_dp; coef(6) = 1.712994697e-1_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",131,"Illegal expansion")
         END SELECT
      CASE (3) ! 2p
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 1.759666885e-1_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 4.323908358e-1_dp; coef(1) = 4.522627513e-1_dp
            alpha(2) = 1.069439065e-1_dp; coef(2) = 6.713122642e-1_dp
         CASE (3)
            alpha(1) = 9.192379002e-1_dp; coef(1) = 1.623948553e-1_dp
            alpha(2) = 2.359194503e-1_dp; coef(2) = 5.661708862e-1_dp
            alpha(3) = 8.009805746e-2_dp; coef(3) = 4.223071752e-1_dp
         CASE (4)
            alpha(1) = 1.798260992e+0_dp; coef(1) = 5.713170255e-2_dp
            alpha(2) = 4.662622228e-1_dp; coef(2) = 2.857455515e-1_dp
            alpha(3) = 1.643718620e-1_dp; coef(3) = 5.517873105e-1_dp
            alpha(4) = 6.543927065e-2_dp; coef(4) = 2.632314924e-1_dp
         CASE (5)
            alpha(1) = 3.320386533e+0_dp; coef(1) = 2.079051117e-2_dp
            alpha(2) = 8.643257633e-1_dp; coef(2) = 1.235472099e-1_dp
            alpha(3) = 3.079819284e-1_dp; coef(3) = 3.667738886e-1_dp
            alpha(4) = 1.273309895e-1_dp; coef(4) = 4.834930290e-1_dp
            alpha(5) = 5.606243164e-2_dp; coef(5) = 1.653444074e-1_dp
         CASE (6)
            alpha(1) = 5.868285913e+0_dp; coef(1) = 7.924233646e-3_dp
            alpha(2) = 1.530329631e+0_dp; coef(2) = 5.144104825e-2_dp
            alpha(3) = 5.475665231e-1_dp; coef(3) = 1.898400060e-1_dp
            alpha(4) = 2.288932733e-1_dp; coef(4) = 4.049863191e-1_dp
            alpha(5) = 1.046655969e-1_dp; coef(5) = 4.012362861e-1_dp
            alpha(6) = 4.948220127e-2_dp; coef(6) = 1.051855189e-1_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",163,"Illegal expansion")
         END SELECT
      CASE (4) ! 3s
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 5.296881757e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 6.694095822e-1_dp; coef(1) = -1.529645716e-1_dp
            alpha(2) = 5.837135094e-2_dp; coef(2) = 1.051370110e+0_dp
         CASE (3)
            alpha(1) = 5.641487709e-1_dp; coef(1) = -1.782577972e-1_dp
            alpha(2) = 6.924421391e-2_dp; coef(2) = 8.612761663e-1_dp
            alpha(3) = 3.269529097e-2_dp; coef(3) = 2.261841969e-1_dp
         CASE (4)
            alpha(1) = 1.513265591e+0_dp; coef(1) = -3.295496352e-2_dp
            alpha(2) = 4.262497508e-1_dp; coef(2) = -1.724516959e-1_dp
            alpha(3) = 7.643320863e-2_dp; coef(3) = 7.518511194e-1_dp
            alpha(4) = 3.760545063e-2_dp; coef(4) = 3.589627310e-1_dp
         CASE (5)
            alpha(1) = 4.275877914e+0_dp; coef(1) = -3.920358850e-3_dp
            alpha(2) = 1.132409433e+0_dp; coef(2) = -4.168430506e-2_dp
            alpha(3) = 4.016256968e-1_dp; coef(3) = -1.637440990e-1_dp
            alpha(4) = 7.732370620e-2_dp; coef(4) = 7.419373723e-1_dp
            alpha(5) = 3.800708627e-2_dp; coef(5) = 3.724364929e-1_dp
         CASE (6)
            alpha(1) = 3.273031938e+0_dp; coef(1) = -6.775596947e-3_dp
            alpha(2) = 9.200611311e-1_dp; coef(2) = -5.639325779e-2_dp
            alpha(3) = 3.593349765e-1_dp; coef(3) = -1.587856086e-1_dp
            alpha(4) = 8.636686991e-2_dp; coef(4) = 5.534527651e-1_dp
            alpha(5) = 4.797373812e-2_dp; coef(5) = 5.015351020e-1_dp
            alpha(6) = 2.724741144e-2_dp; coef(6) = 7.223633674e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",195,"Illegal expansion")
         END SELECT
      CASE (5) ! 3p
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 9.113614253e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 1.458620964e-1_dp; coef(1) = 5.349653144e-1_dp
            alpha(2) = 5.664210742e-2_dp; coef(2) = 5.299607212e-1_dp
         CASE (3)
            alpha(1) = 2.692880368e+0_dp; coef(1) = -1.061945788e-2_dp
            alpha(2) = 1.489359592e-1_dp; coef(2) = 5.218564264e-1_dp
            alpha(3) = 5.739585040e-2_dp; coef(3) = 5.450015143e-1_dp
         CASE (4)
            alpha(1) = 1.853180239e+0_dp; coef(1) = -1.434249391e-2_dp
            alpha(2) = 1.915075719e-1_dp; coef(2) = 2.755177589e-1_dp
            alpha(3) = 8.655487938e-2_dp; coef(3) = 5.846750879e-1_dp
            alpha(4) = 4.184253862e-2_dp; coef(4) = 2.144986514e-1_dp
         CASE (5)
            alpha(1) = 6.466803859e+0_dp; coef(1) = -2.329023747e-3_dp
            alpha(2) = 1.555914802e+0_dp; coef(2) = -1.357395221e-2_dp
            alpha(3) = 1.955925255e-1_dp; coef(3) = 2.632185383e-1_dp
            alpha(4) = 8.809647701e-2_dp; coef(4) = 5.880427024e-1_dp
            alpha(5) = 4.234835707e-2_dp; coef(5) = 2.242794445e-1_dp
         CASE (6)
            alpha(1) = 5.077973607e+0_dp; coef(1) = -3.329929840e-3_dp
            alpha(2) = 1.340786940e+0_dp; coef(2) = -1.419488340e-2_dp
            alpha(3) = 2.248434849e-1_dp; coef(3) = 1.639395770e-1_dp
            alpha(4) = 1.131741848e-1_dp; coef(4) = 4.485358256e-1_dp
            alpha(5) = 6.076408893e-2_dp; coef(5) = 3.908813050e-1_dp
            alpha(6) = 3.315424265e-2_dp; coef(6) = 7.411456232e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",227,"Illegal expansion")
         END SELECT
      CASE (6) ! 3d
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 1.302270363e-1_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 2.777427345e-1_dp; coef(1) = 4.666137923e-1_dp
            alpha(2) = 8.336507714e-2_dp; coef(2) = 6.644706516e-1_dp
         CASE (3)
            alpha(1) = 5.229112225e-1_dp; coef(1) = 1.696596060e-1_dp
            alpha(2) = 1.639595876e-1_dp; coef(2) = 5.847984817e-1_dp
            alpha(3) = 6.386630021e-2_dp; coef(3) = 4.056779523e-1_dp
         CASE (4)
            alpha(1) = 9.185846715e-1_dp; coef(1) = 5.799057705e-2_dp
            alpha(2) = 2.920461109e-1_dp; coef(2) = 3.045581349e-1_dp
            alpha(3) = 1.187568890e-1_dp; coef(3) = 5.601358038e-1_dp
            alpha(4) = 5.286755896e-2_dp; coef(4) = 2.432423313e-1_dp
         CASE (5)
            alpha(1) = 1.539033958e+0_dp; coef(1) = 2.020869128e-2_dp
            alpha(2) = 4.922090297e-1_dp; coef(2) = 1.321157923e-1_dp
            alpha(3) = 2.029756928e-1_dp; coef(3) = 3.911240346e-1_dp
            alpha(4) = 9.424112917e-2_dp; coef(4) = 4.779609701e-1_dp
            alpha(5) = 4.569058269e-2_dp; coef(5) = 1.463662294e-1_dp
         CASE (6)
            alpha(1) = 2.488296923e+0_dp; coef(1) = 7.283828112e-3_dp
            alpha(2) = 7.981487853e-1_dp; coef(2) = 5.386799363e-2_dp
            alpha(3) = 3.311327490e-1_dp; coef(3) = 2.072139149e-1_dp
            alpha(4) = 1.559114463e-1_dp; coef(4) = 4.266269092e-1_dp
            alpha(5) = 7.877734732e-2_dp; coef(5) = 3.843100204e-1_dp
            alpha(6) = 4.058484363e-2_dp; coef(6) = 8.902827546e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",259,"Illegal expansion")
         END SELECT
      CASE (7) ! 4s
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 3.264600274e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 2.441785453e-1_dp; coef(1) = -3.046656896e-1_dp
            alpha(2) = 4.051097664e-2_dp; coef(2) = 1.146877294e+0_dp
         CASE (3)
            alpha(1) = 2.267938753e-1_dp; coef(1) = -3.349048323e-1_dp
            alpha(2) = 4.448178019e-2_dp; coef(2) = 1.056744667e+0_dp
            alpha(3) = 2.195294664e-2_dp; coef(3) = 1.256661680e-1_dp
         CASE (4)
            alpha(1) = 3.242212833e-1_dp; coef(1) = -1.120682822e-1_dp
            alpha(2) = 1.663217177e-1_dp; coef(2) = -2.845426863e-1_dp
            alpha(3) = 5.081097451e-2_dp; coef(3) = 8.909873788e-1_dp
            alpha(4) = 2.829066600e-2_dp; coef(4) = 3.517811205e-1_dp
         CASE (5)
            alpha(1) = 2.980263783e+0_dp; coef(1) = 1.513948997e-3_dp
            alpha(2) = 3.792228833e-1_dp; coef(2) = -7.316801548e-2_dp
            alpha(3) = 1.789717224e-1_dp; coef(3) = -3.143703799e-1_dp
            alpha(4) = 5.002110360e-2_dp; coef(4) = 9.032615169e-1_dp
            alpha(5) = 2.789361681e-2_dp; coef(5) = 3.294210848e-1_dp
         CASE (6)
            alpha(1) = 3.232838646e+0_dp; coef(1) = 1.374817488e-3_dp
            alpha(2) = 3.605788802e-1_dp; coef(2) = -8.666390043e-2_dp
            alpha(3) = 1.717905487e-1_dp; coef(3) = -3.130627309e-1_dp
            alpha(4) = 5.277666487e-2_dp; coef(4) = 7.812787397e-1_dp
            alpha(5) = 3.163400284e-2_dp; coef(5) = 4.389247988e-1_dp
            alpha(6) = 1.874093091e-2_dp; coef(6) = 2.487178756e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",291,"Illegal expansion")
         END SELECT
      CASE (8) ! 4p
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 5.578350235e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 6.190052680e-2_dp; coef(1) = 8.743116767e-1_dp
            alpha(2) = 2.648418407e-2_dp; coef(2) = 1.513640107e-1_dp
         CASE (3)
            alpha(1) = 4.859692220e-1_dp; coef(1) = -6.147823411e-2_dp
            alpha(2) = 7.430216918e-2_dp; coef(2) = 6.604172234e-1_dp
            alpha(3) = 3.653340923e-2_dp; coef(3) = 3.932639495e-1_dp
         CASE (4)
            alpha(1) = 1.492607880e+0_dp; coef(1) = -6.035216774e-3_dp
            alpha(2) = 4.327619272e-1_dp; coef(2) = -6.013310874e-2_dp
            alpha(3) = 7.553156064e-2_dp; coef(3) = 6.451518200e-1_dp
            alpha(4) = 3.706272183e-2_dp; coef(4) = 4.117923820e-1_dp
         CASE (5)
            alpha(1) = 1.091977298e+0_dp; coef(1) = -1.143929558e-2_dp
            alpha(2) = 3.719985051e-1_dp; coef(2) = -6.322651538e-2_dp
            alpha(3) = 8.590019352e-2_dp; coef(3) = 4.398907721e-1_dp
            alpha(4) = 4.786503860e-2_dp; coef(4) = 5.245859166e-1_dp
            alpha(5) = 2.730479990e-2_dp; coef(5) = 1.017072253e-1_dp
         CASE (6)
            alpha(1) = 2.389722618e+0_dp; coef(1) = -1.665913575e-3_dp
            alpha(2) = 7.960947826e-1_dp; coef(2) = -1.657464971e-2_dp
            alpha(3) = 3.415541380e-1_dp; coef(3) = -5.958513378e-2_dp
            alpha(4) = 8.847434525e-2_dp; coef(4) = 4.053115554e-1_dp
            alpha(5) = 4.958248334e-2_dp; coef(5) = 5.433958189e-1_dp
            alpha(6) = 2.816929784e-2_dp; coef(6) = 1.204970491e-1_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",323,"Illegal expansion")
         END SELECT
      CASE (9) ! 4d
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 7.941656339e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 1.330958892e-1_dp; coef(1) = 4.932764167e-1_dp
            alpha(2) = 5.272119659e-2_dp; coef(2) = 5.918727866e-1_dp
         CASE (3)
            alpha(1) = 1.777717219e-1_dp; coef(1) = 2.308552718e-1_dp
            alpha(2) = 8.040647350e-2_dp; coef(2) = 6.042409177e-1_dp
            alpha(3) = 3.949855551e-2_dp; coef(3) = 2.595768926e-1_dp
         CASE (4)
            alpha(1) = 1.995825422e+0_dp; coef(1) = -2.816702620e-3_dp
            alpha(2) = 1.823461280e-1_dp; coef(2) = 2.177095871e-1_dp
            alpha(3) = 8.197240896e-2_dp; coef(3) = 6.058047348e-1_dp
            alpha(4) = 4.000634951e-2_dp; coef(4) = 2.717811257e-1_dp
         CASE (5)
            alpha(1) = 1.522122079e+0_dp; coef(1) = -3.673711876e-3_dp
            alpha(2) = 2.173041823e-1_dp; coef(2) = 1.167122499e-1_dp
            alpha(3) = 1.084876577e-1_dp; coef(3) = 4.216476416e-1_dp
            alpha(4) = 5.836797641e-2_dp; coef(4) = 4.547673415e-1_dp
            alpha(5) = 3.206682246e-2_dp; coef(5) = 1.037803318e-1_dp
         CASE (6)
            alpha(1) = 4.634239420e+0_dp; coef(1) = -4.749842876e-4_dp
            alpha(2) = 1.341648295e+0_dp; coef(2) = -3.566777891e-3_dp
            alpha(3) = 2.209593028e-1_dp; coef(3) = 1.108670481e-1_dp
            alpha(4) = 1.101467943e-1_dp; coef(4) = 4.159646930e-1_dp
            alpha(5) = 5.904190370e-2_dp; coef(5) = 4.621672517e-1_dp
            alpha(6) = 3.232628887e-2_dp; coef(6) = 1.081250196e-1_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",355,"Illegal expansion")
         END SELECT
      CASE (10) ! 4f
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 1.033434062e-1_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 2.006693538e-1_dp; coef(1) = 4.769346276e-1_dp
            alpha(2) = 6.865384900e-2_dp; coef(2) = 6.587383976e-1_dp
         CASE (3)
            alpha(1) = 3.483826963e-1_dp; coef(1) = 1.737856685e-1_dp
            alpha(2) = 1.249380537e-1_dp; coef(2) = 5.973380628e-1_dp
            alpha(3) = 5.349995725e-2_dp; coef(3) = 3.929395614e-1_dp
         CASE (4)
            alpha(1) = 5.691670217e-1_dp; coef(1) = 5.902730589e-2_dp
            alpha(2) = 2.074585819e-1_dp; coef(2) = 3.191828952e-1_dp
            alpha(3) = 9.298346885e-2_dp; coef(3) = 5.639423893e-1_dp
            alpha(4) = 4.473508853e-2_dp; coef(4) = 2.284796537e-1_dp
         CASE (5)
            alpha(1) = 8.925960415e-1_dp; coef(1) = 1.999839052e-2_dp
            alpha(2) = 3.277589120e-1_dp; coef(2) = 1.395427440e-1_dp
            alpha(3) = 1.492869962e-1_dp; coef(3) = 4.091508237e-1_dp
            alpha(4) = 7.506099109e-2_dp; coef(4) = 4.708252119e-1_dp
            alpha(5) = 3.892475795e-2_dp; coef(5) = 1.328082566e-1_dp
         CASE (6)
            alpha(1) = 1.357718039e+0_dp; coef(1) = 6.930234381e-3_dp
            alpha(2) = 5.004907278e-1_dp; coef(2) = 5.634263745e-2_dp
            alpha(3) = 2.296565064e-1_dp; coef(3) = 2.217065797e-1_dp
            alpha(4) = 1.173146814e-1_dp; coef(4) = 4.411388883e-1_dp
            alpha(5) = 6.350097171e-2_dp; coef(5) = 3.688112625e-1_dp
            alpha(6) = 3.474556673e-2_dp; coef(6) = 7.787514504e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",387,"Illegal expansion")
         END SELECT
      CASE (11) ! 5s
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 2.216912938e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 1.213425654e-1_dp; coef(1) = -5.114756049e-1_dp
            alpha(2) = 3.133152144e-2_dp; coef(2) = 1.307377277e+0_dp
         CASE (3)
            alpha(1) = 1.080198458e-1_dp; coef(1) = -6.617401158e-1_dp
            alpha(2) = 4.408119328e-2_dp; coef(2) = 7.467595004e-1_dp
            alpha(3) = 2.610811810e-2_dp; coef(3) = 7.146490945e-1_dp
         CASE (4)
            alpha(1) = 8.602284252e-1_dp; coef(1) = 1.103657561e-2_dp
            alpha(2) = 1.189050200e-1_dp; coef(2) = -5.606519023e-1_dp
            alpha(3) = 3.446076176e-2_dp; coef(3) = 1.179429987e+0_dp
            alpha(4) = 1.974798796e-2_dp; coef(4) = 1.734974376e-1_dp
         CASE (5)
            alpha(1) = 7.403763257e-1_dp; coef(1) = 1.375523371e-2_dp
            alpha(2) = 1.367990863e-1_dp; coef(2) = -3.097344179e-1_dp
            alpha(3) = 9.135301779e-2_dp; coef(3) = -3.199192259e-1_dp
            alpha(4) = 3.726907315e-2_dp; coef(4) = 1.084547038e+0_dp
            alpha(5) = 2.241490836e-2_dp; coef(5) = 3.345288361e-1_dp
         CASE (6)
            alpha(1) = 1.410128298e+0_dp; coef(1) = 2.695439582e-3_dp
            alpha(2) = 5.077878915e-1_dp; coef(2) = 1.850157487e-2_dp
            alpha(3) = 1.847926858e-1_dp; coef(3) = -9.588628125e-2_dp
            alpha(4) = 1.061070594e-1_dp; coef(4) = -5.200673560e-1_dp
            alpha(5) = 3.669584901e-2_dp; coef(5) = 1.087619490e+0_dp
            alpha(6) = 2.213558430e-2_dp; coef(6) = 3.103964343e-1_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",419,"Illegal expansion")
         END SELECT
      CASE (12) ! 5p
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 3.769845216e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 2.691294191e-1_dp; coef(1) = -1.034227010e-1_dp
            alpha(2) = 3.980805011e-2_dp; coef(2) = 1.033376378e+0_dp
         CASE (3)
            alpha(1) = 2.127482317e-1_dp; coef(1) = -1.389529695e-1_dp
            alpha(2) = 4.729648620e-2_dp; coef(2) = 8.076691064e-1_dp
            alpha(3) = 2.604865324e-2_dp; coef(3) = 2.726029342e-1_dp
         CASE (4)
            alpha(1) = 3.962838833e-1_dp; coef(1) = -1.801459207e-2_dp
            alpha(2) = 1.838858552e-1_dp; coef(2) = -1.360777372e-1_dp
            alpha(3) = 4.943555157e-2_dp; coef(3) = 7.533973719e-1_dp
            alpha(4) = 2.750222273e-2_dp; coef(4) = 3.409304859e-1_dp
         CASE (5)
            alpha(1) = 3.422168934e-1_dp; coef(1) = -3.113958289e-2_dp
            alpha(2) = 1.665099900e-1_dp; coef(2) = -1.374007017e-1_dp
            alpha(3) = 5.443732013e-2_dp; coef(3) = 5.573881018e-1_dp
            alpha(4) = 3.367775277e-2_dp; coef(4) = 4.855428100e-1_dp
            alpha(5) = 2.091949042e-2_dp; coef(5) = 6.605423564e-2_dp
         CASE (6)
            alpha(1) = 3.778623374e+0_dp; coef(1) = 1.163246387e-4_dp
            alpha(2) = 3.499121109e-1_dp; coef(2) = -2.920771322e-2_dp
            alpha(3) = 1.683175469e-1_dp; coef(3) = -1.381051233e-1_dp
            alpha(4) = 5.404070736e-2_dp; coef(4) = 5.706134877e-1_dp
            alpha(5) = 3.328911801e-2_dp; coef(5) = 4.768808140e-1_dp
            alpha(6) = 2.063815019e-2_dp; coef(6) = 6.021665516e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",451,"Illegal expansion")
         END SELECT
      CASE (13) ! 5d
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 5.352200793e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 6.906014388e-2_dp; coef(1) = 6.539405185e-1_dp
            alpha(2) = 3.399457777e-2_dp; coef(2) = 3.948945302e-1_dp
         CASE (3)
            alpha(1) = 4.913352950e-1_dp; coef(1) = -2.010175008e-2_dp
            alpha(2) = 7.329090601e-2_dp; coef(2) = 5.899370608e-1_dp
            alpha(3) = 3.594209290e-2_dp; coef(3) = 4.658445960e-1_dp
         CASE (4)
            alpha(1) = 4.230617826e-1_dp; coef(1) = -2.421626009e-2_dp
            alpha(2) = 8.293863702e-2_dp; coef(2) = 3.937644956e-1_dp
            alpha(3) = 4.590326388e-2_dp; coef(3) = 5.489520286e-1_dp
            alpha(4) = 2.628744797e-2_dp; coef(4) = 1.190436963e-1_dp
         CASE (5)
            alpha(1) = 9.702946470e-1_dp; coef(1) = -3.231527611e-3_dp
            alpha(2) = 3.603270196e-1_dp; coef(2) = -2.434931372e-2_dp
            alpha(3) = 8.668717752e-2_dp; coef(3) = 3.440817054e-1_dp
            alpha(4) = 4.833708379e-2_dp; coef(4) = 5.693674376e-1_dp
            alpha(5) = 2.751899341e-2_dp; coef(5) = 1.511340183e-1_dp
         CASE (6)
            alpha(1) = 8.820520428e-1_dp; coef(1) = -4.097377019e-3_dp
            alpha(2) = 3.410838409e-1_dp; coef(2) = -2.508271857e-2_dp
            alpha(3) = 9.204308840e-2_dp; coef(3) = 2.648458555e-1_dp
            alpha(4) = 5.472831774e-2_dp; coef(4) = 5.097437054e-1_dp
            alpha(5) = 3.391202830e-2_dp; coef(5) = 2.654483467e-1_dp
            alpha(6) = 2.108227374e-2_dp; coef(6) = 2.623132212e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",483,"Illegal expansion")
         END SELECT
      CASE (14) ! 5f
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 6.952785407e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 1.156094555e-1_dp; coef(1) = 4.856637346e-1_dp
            alpha(2) = 4.778940916e-2_dp; coef(2) = 6.125980914e-1_dp
         CASE (3)
            alpha(1) = 1.649233885e-1_dp; coef(1) = 1.909729355e-1_dp
            alpha(2) = 7.487066646e-2_dp; coef(2) = 6.146060459e-1_dp
            alpha(3) = 3.735787219e-2_dp; coef(3) = 3.059611271e-1_dp
         CASE (4)
            alpha(1) = 2.017831152e-1_dp; coef(1) = 9.174268830e-2_dp
            alpha(2) = 1.001952178e-1_dp; coef(2) = 4.023496947e-1_dp
            alpha(3) = 5.447006630e-2_dp; coef(3) = 4.937432100e-1_dp
            alpha(4) = 3.037569283e-2_dp; coef(4) = 1.254001522e-1_dp
         CASE (5)
            alpha(1) = 1.670735676e+0_dp; coef(1) = -7.301193568e-4_dp
            alpha(2) = 2.072477219e-1_dp; coef(2) = 8.414991343e-2_dp
            alpha(3) = 1.024709357e-1_dp; coef(3) = 3.923683153e-1_dp
            alpha(4) = 5.537913898e-2_dp; coef(4) = 5.040033146e-1_dp
            alpha(5) = 3.072866652e-2_dp; coef(5) = 1.328979300e-1_dp
         CASE (6)
            alpha(1) = 1.334096840e+0_dp; coef(1) = -9.486751531e-4_dp
            alpha(2) = 2.372312347e-1_dp; coef(2) = 4.624275998e-2_dp
            alpha(3) = 1.269485744e-1_dp; coef(3) = 2.373699784e-1_dp
            alpha(4) = 7.290318381e-2_dp; coef(4) = 4.589112231e-1_dp
            alpha(5) = 4.351355997e-2_dp; coef(5) = 3.205010548e-1_dp
            alpha(6) = 2.598071843e-2_dp; coef(6) = 5.077063693e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",515,"Illegal expansion")
         END SELECT
      CASE (15) ! 5g
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 8.565417784e-2_dp; coef(1) = 1.000000000e+0_dp
         CASE (2)
            alpha(1) = 1.554531559e-1_dp; coef(1) = 4.848298074e-1_dp
            alpha(2) = 5.854079811e-2_dp; coef(2) = 6.539381621e-1_dp
         CASE (3)
            alpha(1) = 2.545432122e-1_dp; coef(1) = 1.780980905e-1_dp
            alpha(2) = 1.006544376e-1_dp; coef(2) = 6.063757846e-1_dp
            alpha(3) = 4.624463922e-2_dp; coef(3) = 3.828552923e-1_dp
         CASE (4)
            alpha(1) = 3.945205573e-1_dp; coef(1) = 6.010484250e-2_dp
            alpha(2) = 1.588100623e-1_dp; coef(2) = 3.309738329e-1_dp
            alpha(3) = 7.646521729e-2_dp; coef(3) = 5.655207585e-1_dp
            alpha(4) = 3.898703611e-2_dp; coef(4) = 2.171122608e-1_dp
         CASE (5)
            alpha(1) = 5.895429375e-1_dp; coef(1) = 1.998085812e-2_dp
            alpha(2) = 2.393343780e-1_dp; coef(2) = 1.460384050e-1_dp
            alpha(3) = 1.172646904e-1_dp; coef(3) = 4.230565459e-1_dp
            alpha(4) = 6.254074479e-2_dp; coef(4) = 4.635699665e-1_dp
            alpha(5) = 3.411243214e-2_dp; coef(5) = 1.226411691e-1_dp
         CASE (6)
            alpha(1) = 8.574668996e-1_dp; coef(1) = 6.729778096e-3_dp
            alpha(2) = 3.497184772e-1_dp; coef(2) = 5.874145170e-2_dp
            alpha(3) = 1.727917060e-1_dp; coef(3) = 2.339955227e-1_dp
            alpha(4) = 9.373643151e-2_dp; coef(4) = 4.512983737e-1_dp
            alpha(5) = 5.340032759e-2_dp; coef(5) = 3.552053926e-1_dp
            alpha(6) = 3.057364464e-2_dp; coef(6) = 6.974153145e-2_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",547,"Illegal expansion")
         END SELECT
      CASE (16) ! 6s
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.016053956888690133_dp; coef(1) = 1.000000000000000_dp
         CASE (2)
            alpha(1) = 0.026188993287017705_dp; coef(1) = 1.6061745480037777_dp
            alpha(2) = 0.06916086082602657_dp; coef(2) = -0.8493191121412285_dp
         CASE (3)
            alpha(1) = 0.025012605890022674_dp; coef(1) = 1.4570914918085252_dp
            alpha(2) = 0.08530028544682955_dp; coef(2) = -0.7792202598134607_dp
            alpha(3) = 0.23664111527549864_dp; coef(3) = 0.08944830584494995_dp
         CASE (4)
            alpha(1) = 0.018111879224996335_dp; coef(1) = 0.2742426110211285_dp
            alpha(2) = 0.028489853269487397_dp; coef(2) = 1.3154513257751648_dp
            alpha(3) = 0.07575858742450885_dp; coef(3) = -0.8700267823821266_dp
            alpha(4) = 0.3135211502054625_dp; coef(4) = 0.04933652686369165_dp
         CASE (5)
            alpha(1) = 0.017704122931955696_dp; coef(1) = 0.23979549908360312_dp
            alpha(2) = 0.02816438780280081_dp; coef(2) = 1.339995854967376_dp
            alpha(3) = 0.07635157423691516_dp; coef(3) = -0.8631396815113626_dp
            alpha(4) = 0.29378907244008734_dp; coef(4) = 0.05070584407672884_dp
            alpha(5) = 0.7609367674925651_dp; coef(5) = 0.0017452546896168622_dp
         CASE (6)
            alpha(1) = 0.018375086251195376_dp; coef(1) = 0.3019609612694464_dp
            alpha(2) = 0.02883213591739931_dp; coef(2) = 1.3047682665168656_dp
            alpha(3) = 0.0735171952084867_dp; coef(3) = -0.8423706115013339_dp
            alpha(4) = 0.13718508837897095_dp; coef(4) = -0.08693662797569257_dp
            alpha(5) = 0.21928060661775448_dp; coef(5) = 0.08040482815742164_dp
            alpha(6) = 0.47255402578342814_dp; coef(6) = 0.011258177324556012_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",579,"Illegal expansion")
         END SELECT
      CASE (17) ! 6p
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.02720326117498455_dp; coef(1) = 1.0000000000000000_dp
         CASE (2)
            alpha(1) = 0.03068554921378584_dp; coef(1) = 1.0932524470142648_dp
            alpha(2) = 0.13327144868476606_dp; coef(2) = -0.20022333756955418_dp
         CASE (3)
            alpha(1) = 0.021306449714000038_dp; coef(1) = 0.2589896573574047_dp
            alpha(2) = 0.03558737683798757_dp; coef(2) = 0.8948520093263641_dp
            alpha(3) = 0.1159879775604416_dp; coef(3) = -0.2541710812641519_dp
         CASE (4)
            alpha(1) = 0.02094853126286551_dp; coef(1) = 0.23874968368587424_dp
            alpha(2) = 0.03524009985593678_dp; coef(2) = 0.9113069782415131_dp
            alpha(3) = 0.1172381116493136_dp; coef(3) = -0.25095169849099785_dp
            alpha(4) = 1.074757975437711_dp; coef(4) = 0.0014270633206840072_dp
         CASE (5)
            alpha(1) = 0.02191796448535917_dp; coef(1) = 0.29958347534070895_dp
            alpha(2) = 0.036406617593622265_dp; coef(2) = 0.8669095451744081_dp
            alpha(3) = 0.10823682952141411_dp; coef(3) = -0.24403207325952317_dp
            alpha(4) = 0.1912844969312091_dp; coef(4) = -0.025610699331187835_dp
            alpha(5) = 0.605311988824089_dp; coef(5) = 0.003533893449371835_dp
         CASE (6)
            alpha(1) = 0.01836987035282947_dp; coef(1) = 0.0881865874022419_dp
            alpha(2) = 0.028283616813330213_dp; coef(2) = 0.58260237820929_dp
            alpha(3) = 0.0424808315746305_dp; coef(3) = 0.5465800645782257_dp
            alpha(4) = 0.08775028399997764_dp; coef(4) = -0.20802232268176662_dp
            alpha(5) = 0.14284922158351138_dp; coef(5) = -0.1117378797434228_dp
            alpha(6) = 0.6631101746841396_dp; coef(6) = 0.0028390610573703423_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",611,"Illegal expansion")
         END SELECT
      CASE (18) ! 6d
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.038539962557578994_dp; coef(1) = 1.0000000000000000_dp
         CASE (2)
            alpha(1) = 0.018343252403866257_dp; coef(1) = 0.1002697090882244_dp
            alpha(2) = 0.04095508659742878_dp; coef(2) = 0.921707450678634_dp
         CASE (3)
            alpha(1) = 0.026815363906634464_dp; coef(1) = 0.36430995341001843_dp
            alpha(2) = 0.048285664795074575_dp; coef(2) = 0.6902756637148024_dp
            alpha(3) = 0.20398527022566643_dp; coef(3) = -0.061643555476797514_dp
         CASE (4)
            alpha(1) = 0.014812592588885804_dp; coef(1) = 0.011887725165713237_dp
            alpha(2) = 0.028604392422323988_dp; coef(2) = 0.4126202273551119_dp
            alpha(3) = 0.049571759954916264_dp; coef(3) = 0.6324841000859261_dp
            alpha(4) = 0.20042444873055187_dp; coef(4) = -0.06355840930231531_dp
         CASE (5)
            alpha(1) = 0.02104884011402785_dp; coef(1) = 0.09247565169723151_dp
            alpha(2) = 0.03387333992146401_dp; coef(2) = 0.5183469446932626_dp
            alpha(3) = 0.05464871753445727_dp; coef(3) = 0.4583625355467117_dp
            alpha(4) = 0.16559124243612725_dp; coef(4) = -0.06557358037625106_dp
            alpha(5) = 0.3388543915705088_dp; coef(5) = -0.01073326765431463_dp
         CASE (6)
            alpha(1) = 0.012538741501253952_dp; coef(1) = 0.0013891559057176176_dp
            alpha(2) = 0.022411877391156923_dp; coef(2) = 0.12323543239586941_dp
            alpha(3) = 0.03502206086139318_dp; coef(3) = 0.516335137452715_dp
            alpha(4) = 0.05542414129417389_dp; coef(4) = 0.42911284700977503_dp
            alpha(5) = 0.16410056143462523_dp; coef(5) = -0.065949692922377_dp
            alpha(6) = 0.33496788346358164_dp; coef(6) = -0.01123082945384288_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",643,"Illegal expansion")
         END SELECT
      CASE (19) ! 6f
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.04999708511585302_dp; coef(1) = 1._dp
         CASE (2)
            alpha(1) = 0.03443190219402774_dp; coef(1) = 0.5095525299458369_dp
            alpha(2) = 0.06993938754199006_dp; coef(2) = 0.556840322894986_dp
         CASE (3)
            alpha(1) = 0.022913376030207443_dp; coef(1) = 0.0844980697596352_dp
            alpha(2) = 0.040621057779302076_dp; coef(2) = 0.558907102542338_dp
            alpha(3) = 0.07551452585114922_dp; coef(3) = 0.4268730468790412_dp
         CASE (4)
            alpha(1) = 0.026723978066792376_dp; coef(1) = 0.18313299355518015_dp
            alpha(2) = 0.04675842108187079_dp; coef(2) = 0.5931630352672648_dp
            alpha(3) = 0.08424111895397314_dp; coef(3) = 0.29815135014219757_dp
            alpha(4) = 0.40182223710248033_dp; coef(4) = -0.009003129808147884_dp
         CASE (5)
            alpha(1) = 0.019820990273457767_dp; coef(1) = 0.026554164306668928_dp
            alpha(2) = 0.031943455135755784_dp; coef(2) = 0.2857152467032053_dp
            alpha(3) = 0.05180437300600356_dp; coef(3) = 0.5249457199886216_dp
            alpha(4) = 0.08839526561404966_dp; coef(4) = 0.23802584890828332_dp
            alpha(5) = 0.3861344493937384_dp; coef(5) = -0.00959390690588398_dp
         CASE (6)
            alpha(1) = 0.021979238503298942_dp; coef(1) = 0.05529549645466706_dp
            alpha(2) = 0.0353635379450325_dp; coef(2) = 0.36260443662220726_dp
            alpha(3) = 0.056802330945072296_dp; coef(3) = 0.47981873895702293_dp
            alpha(4) = 0.09436112630657385_dp; coef(4) = 0.17913120326169676_dp
            alpha(5) = 0.3258167395765562_dp; coef(5) = -0.010025342773471209_dp
            alpha(6) = 0.7758137452357111_dp; coef(6) = -0.0013665302206626205_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",675,"Illegal expansion")
         END SELECT
      CASE (20) ! 6g
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.061536068942275354_dp; coef(1) = 1._dp
         CASE (2)
            alpha(1) = 0.04332629465536664_dp; coef(1) = 0.6209033605306379_dp
            alpha(2) = 0.10013340560986961_dp; coef(2) = 0.4864065117811783_dp
         CASE (3)
            alpha(1) = 0.03454561337035713_dp; coef(1) = 0.3227633523004909_dp
            alpha(2) = 0.06729191664458245_dp; coef(2) = 0.6193784317147506_dp
            alpha(3) = 0.14454284915249543_dp; coef(3) = 0.1810539000206162_dp
         CASE (4)
            alpha(1) = 0.028983896307773004_dp; coef(1) = 0.15152805717139864_dp
            alpha(2) = 0.05135519389407211_dp; coef(2) = 0.5309701439600488_dp
            alpha(3) = 0.09369371519055944_dp; coef(3) = 0.37457949569535787_dp
            alpha(4) = 0.18939419012133174_dp; coef(4) = 0.06851383804287256_dp
         CASE (5)
            alpha(1) = 0.024699673908324754_dp; coef(1) = 0.06048210643335771_dp
            alpha(2) = 0.04083036114944879_dp; coef(2) = 0.3536959373883507_dp
            alpha(3) = 0.06763537320097322_dp; coef(3) = 0.46434612683985144_dp
            alpha(4) = 0.1170091091566765_dp; coef(4) = 0.2130170165201656_dp
            alpha(5) = 0.22066124899602302_dp; coef(5) = 0.0343912188428939_dp
         CASE (6)
            alpha(1) = 0.024944035346224053_dp; coef(1) = 0.06443914638193514_dp
            alpha(2) = 0.04138718127148023_dp; coef(2) = 0.36507904601268215_dp
            alpha(3) = 0.06886079537637349_dp; coef(3) = 0.4624056701486661_dp
            alpha(4) = 0.11972435105059721_dp; coef(4) = 0.20306165490832248_dp
            alpha(5) = 0.22639206282070937_dp; coef(5) = 0.03099725943908058_dp
            alpha(6) = 1.4892687021530484_dp; coef(6) = -0.00018343848780163463_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",707,"Illegal expansion")
         END SELECT
      CASE (21) ! 6h
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.07313314184000295_dp; coef(1) = 1._dp
         CASE (2)
            alpha(1) = 0.05113303517913134_dp; coef(1) = 0.6498709211538302_dp
            alpha(2) = 0.12607272311510218_dp; coef(2) = 0.4911257514330729_dp
         CASE (3)
            alpha(1) = 0.040851454032819305_dp; coef(1) = 0.3746467681927208_dp
            alpha(2) = 0.08415648437016819_dp; coef(2) = 0.6132065860909666_dp
            alpha(3) = 0.1974425255793456_dp; coef(3) = 0.18179709985670622_dp
         CASE (4)
            alpha(1) = 0.034679493065603105_dp; coef(1) = 0.20806488992498948_dp
            alpha(2) = 0.06498507795614539_dp; coef(2) = 0.5659152898854298_dp
            alpha(3) = 0.1276799968878006_dp; coef(3) = 0.3407357791588573_dp
            alpha(4) = 0.293954336816017_dp; coef(4) = 0.061163634708810435_dp
         CASE (5)
            alpha(1) = 0.030487493495559674_dp; coef(1) = 0.1147257294844731_dp
            alpha(2) = 0.05371457065868773_dp; coef(2) = 0.45668261354533374_dp
            alpha(3) = 0.09619925104884211_dp; coef(3) = 0.434116528713387_dp
            alpha(4) = 0.18557752006634115_dp; coef(4) = 0.15177021621261605_dp
            alpha(5) = 0.42330005591830827_dp; coef(5) = 0.02006866296723323_dp
         CASE (6)
            alpha(1) = 0.027419761875036298_dp; coef(1) = 0.06355638710922437_dp
            alpha(2) = 0.04620886209790054_dp; coef(2) = 0.34336050905161936_dp
            alpha(3) = 0.07793999480997445_dp; coef(3) = 0.45863493125119364_dp
            alpha(4) = 0.13712767813357474_dp; coef(4) = 0.24456067327385683_dp
            alpha(5) = 0.2621356941062822_dp; coef(5) = 0.061018224093793164_dp
            alpha(6) = 0.5948121608710077_dp; coef(6) = 0.006618011961621055_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",739,"Illegal expansion")
         END SELECT
      CASE (22) ! 7s
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.012170316100569041_dp; coef(1) = 1._dp

         CASE (2)
            alpha(1) = 0.02380792064880498_dp; coef(1) = 2.4733498823505897_dp
            alpha(2) = 0.0413963085083486_dp; coef(2) = -1.7500626134752972_dp
         CASE (3)
            alpha(1) = 0.020979847986630586_dp; coef(1) = 1.675236183601737_dp
            alpha(2) = 0.061178936229820555_dp; coef(2) = -1.147227387784115_dp
            alpha(3) = 0.1401881324437541_dp; coef(3) = 0.2121061484210291_dp
         CASE (4)
            alpha(1) = 0.015915368005026652_dp; coef(1) = 0.3275001969711621_dp
            alpha(2) = 0.023863997843078046_dp; coef(2) = 1.5476507021661345_dp
            alpha(3) = 0.054562166429946386_dp; coef(3) = -1.2697132490611722_dp
            alpha(4) = 0.16342646591350343_dp; coef(4) = 0.13574231777373438_dp
         CASE (5)
            alpha(1) = 0.01606210559622015_dp; coef(1) = 0.3489870695674181_dp
            alpha(2) = 0.024001717565201447_dp; coef(2) = 1.5335261688987925_dp
            alpha(3) = 0.05439364645164688_dp; coef(3) = -1.275299080068857_dp
            alpha(4) = 0.16456835599844608_dp; coef(4) = 0.13406783441420897_dp
            alpha(5) = 1.8632404725730192_dp; coef(5) = -0.0001818568135349486_dp
         CASE (6)
            alpha(1) = 0.01565231604983483_dp; coef(1) = 0.29154442199853614_dp
            alpha(2) = 0.023623292255693457_dp; coef(2) = 1.5691344169994597_dp
            alpha(3) = 0.05497538487325693_dp; coef(3) = -1.2618704056879249_dp
            alpha(4) = 0.15604786033627724_dp; coef(4) = 0.13683678315315853_dp
            alpha(5) = 0.30526699650551353_dp; coef(5) = 0.006442390493458186_dp
            alpha(6) = 0.7613613898332257_dp; coef(6) = -0.0009736214432373223_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",772,"Illegal expansion")
         END SELECT
      CASE (23) ! 7p
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.02056749089204262_dp; coef(1) = 1.0000000000000000_dp
         CASE (2)
            alpha(1) = 0.02509434622887706_dp; coef(1) = 1.1953936450048124_dp
            alpha(2) = 0.07932687651759715_dp; coef(2) = -0.33554981426999153_dp
         CASE (3)
            alpha(1) = 0.02216294295259616_dp; coef(1) = 0.8184285461616602_dp
            alpha(2) = 0.042045475539838104_dp; coef(2) = 0.8328277926936017_dp
            alpha(3) = 0.0622825034958455_dp; coef(3) = -0.7865628896072743_dp
         CASE (4)
            alpha(1) = 0.01798640692496774_dp; coef(1) = 0.2517796607641491_dp
            alpha(2) = 0.02838134507613473_dp; coef(2) = 1.0084655616939668_dp
            alpha(3) = 0.07596133909152862_dp; coef(3) = -0.40416345513837804_dp
            alpha(4) = 0.30794728288131357_dp; coef(4) = 0.011606225585276365_dp
         CASE (5)
            alpha(1) = 0.018253221396022562_dp; coef(1) = 0.27532738928860534_dp
            alpha(2) = 0.028708043693626412_dp; coef(2) = 0.9951443329321299_dp
            alpha(3) = 0.07061876309146031_dp; coef(3) = -0.2663034712401708_dp
            alpha(4) = 0.08392634608758946_dp; coef(4) = -0.1489548148570699_dp
            alpha(5) = 0.298415801799801_dp; coef(5) = 0.012451440490101195_dp
         CASE (6)
            ! Warning: optimization not complete
            alpha(1) = 0.018266628106424383_dp; coef(1) = 0.2762223367567883_dp
            alpha(2) = 0.02870452146813995_dp; coef(2) = 0.992858662003434_dp
            alpha(3) = 0.07430061134905402_dp; coef(3) = -0.4029967818464786_dp
            alpha(4) = 0.18615871115244434_dp; coef(4) = -41.116221124256434_dp
            alpha(5) = 0.1861966605860832_dp; coef(5) = 41.115741037337784_dp
            alpha(6) = 0.476774875907382_dp; coef(6) = 0.0020913658772264638_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",805,"Illegal expansion")
         END SELECT
      CASE (24) ! 7d
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.02909039015051073_dp; coef(1) = 1.0000000000000002_dp
         CASE (2)
            alpha(1) = 0.03018482426518148_dp; coef(1) = 1.025499895914138_dp
            alpha(2) = 0.15242731708213653_dp; coef(2) = -0.0801196219224754_dp
         CASE (3)
            alpha(1) = 0.021405281219812042_dp; coef(1) = 0.2999132044312777_dp
            alpha(2) = 0.03568228986173447_dp; coef(2) = 0.7809338560049855_dp
            alpha(3) = 0.11574152434647746_dp; coef(3) = -0.1242352213758718_dp
         CASE (4)
            alpha(1) = 0.021660976213634955_dp; coef(1) = 0.3177854736353611_dp
            alpha(2) = 0.03606163962267541_dp; coef(2) = 0.7675316218690252_dp
            alpha(3) = 0.10985898319545016_dp; coef(3) = -0.11712958841599931_dp
            alpha(4) = 0.15908195611925_dp; coef(4) = -0.011747385912275846_dp
         CASE (5)
            ! Warning: optimization not complete
            alpha(1) = 0.018501533387022245_dp; coef(1) = 0.11676114014037238_dp
            alpha(2) = 0.02923074108604399_dp; coef(2) = 0.6656817970408199_dp
            alpha(3) = 0.05423405317769654_dp; coef(3) = 600.6430437919528_dp
            alpha(4) = 0.054241824018008936_dp; coef(4) = -600.3759617830652_dp
            alpha(5) = 0.12494194517082237_dp; coef(5) = -0.09295829453419228_dp
         CASE (6)
            alpha(1) = 0.017457157083607015_dp; coef(1) = 0.07250557935571439_dp
            alpha(2) = 0.026593769405605233_dp; coef(2) = 0.5007763109536889_dp
            alpha(3) = 0.04008932402169768_dp; coef(3) = 0.5339957399786912_dp
            alpha(4) = 0.09560872001992007_dp; coef(4) = -0.11880032574322265_dp
            alpha(5) = 0.15425745053701811_dp; coef(5) = -0.03241452360588799_dp
            alpha(6) = 0.5964979973145561_dp; coef(6) = 0.0007012418324482315_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",838,"Illegal expansion")
         END SELECT
      CASE (25) ! 7f
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.037696619888011394_dp; coef(1) = 1._dp
         CASE (2)
            alpha(1) = 0.02399618488506292_dp; coef(1) = 0.2938876408614713_dp
            alpha(2) = 0.0441727483070652_dp; coef(2) = 0.7461850531456674_dp
         CASE (3)
            alpha(1) = 0.02649072884749045_dp; coef(1) = 0.41965629582857367_dp
            alpha(2) = 0.04781482493797306_dp; coef(2) = 0.6320266162094202_dp
            alpha(3) = 0.20807007812612838_dp; coef(3) = -0.025856724476949224_dp
         CASE (4)
            alpha(1) = 0.019516667217357033_dp; coef(1) = 0.07760665191295739_dp
            alpha(2) = 0.031536542788218214_dp; coef(2) = 0.5138599419165162_dp
            alpha(3) = 0.0518809354792556_dp; coef(3) = 0.4658306981238731_dp
            alpha(4) = 0.18978036627517764_dp; coef(4) = -0.03032718718680631_dp
         CASE (5)
            alpha(1) = 0.02085272354115486_dp; coef(1) = 0.11822768394863675_dp
            alpha(2) = 0.033574888134901455_dp; coef(2) = 0.5490677063820284_dp
            alpha(3) = 0.05432263605261335_dp; coef(3) = 0.3941968220546063_dp
            alpha(4) = 0.1670441095948162_dp; coef(4) = -0.031436205691901344_dp
            alpha(5) = 0.3442595341279306_dp; coef(5) = -0.0033909812829636394_dp
         CASE (6)
            alpha(1) = 0.01641594422436568_dp; coef(1) = 0.01709822656839514_dp
            alpha(2) = 0.024762426761735977_dp; coef(2) = 0.22140165119709196_dp
            alpha(3) = 0.03711681387997685_dp; coef(3) = 0.5157825082425954_dp
            alpha(4) = 0.057006537132452996_dp; coef(4) = 0.30947877900909615_dp
            alpha(5) = 0.15940261392199198_dp; coef(5) = -0.032377223278152595_dp
            alpha(6) = 0.3168807735151213_dp; coef(6) = -0.004680397785899489_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",870,"Illegal expansion")
         END SELECT
      CASE (26) ! 7g
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.04636088534182354_dp; coef(1) = 1.0000000000000002_dp
         CASE (2)
            alpha(1) = 0.033089231025105_dp; coef(1) = 0.5580805992890315_dp
            alpha(2) = 0.06689561975474233_dp; coef(2) = 0.5213501458304487_dp
         CASE (3)
            alpha(1) = 0.02536613332751714_dp; coef(1) = 0.1958347745511226_dp
            alpha(2) = 0.04392292490201606_dp; coef(2) = 0.6069779654592313_dp
            alpha(3) = 0.07935572931927462_dp; coef(3) = 0.28358299617643856_dp
         CASE (4)
            alpha(1) = 0.026003867919966516_dp; coef(1) = 0.2209616450163026_dp
            alpha(2) = 0.045349648935287534_dp; coef(2) = 0.6131566254139853_dp
            alpha(3) = 0.08207336886154683_dp; coef(3) = 0.25321528599323867_dp
            alpha(4) = 0.43630841028097767_dp; coef(4) = -0.0028051508587602996_dp
         CASE (5)
            alpha(1) = 0.021295076240560495_dp; coef(1) = 0.06733674728927509_dp
            alpha(2) = 0.03407297844465803_dp; coef(2) = 0.39912419436935137_dp
            alpha(3) = 0.05463343066262556_dp; coef(3) = 0.47165433606934565_dp
            alpha(4) = 0.09142718539759649_dp; coef(4) = 0.15054429419419305_dp
            alpha(5) = 0.3631735814032239_dp; coef(5) = -0.003735801628196915_dp
         CASE (6)
            alpha(1) = 0.02154419398101922_dp; coef(1) = 0.07294053032928545_dp
            alpha(2) = 0.03458506500828156_dp; coef(2) = 0.41255126079553106_dp
            alpha(3) = 0.055591221080949486_dp; coef(3) = 0.4632159545572935_dp
            alpha(4) = 0.0929128679746364_dp; coef(4) = 0.14023937661406255_dp
            alpha(5) = 0.3353651542417124_dp; coef(5) = -0.00381264707490568_dp
            alpha(6) = 0.8332537115987487_dp; coef(6) = -0.0002777669788215993_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",902,"Illegal expansion")
         END SELECT
      CASE (27) ! 7h
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.055067144574685124_dp; coef(1) = 1.0000000000000002_dp
         CASE (2)
            alpha(1) = 0.039504306298949056_dp; coef(1) = 0.6244224977499361_dp
            alpha(2) = 0.08748752023471955_dp; coef(2) = 0.48947010311740946_dp
         CASE (3)
            alpha(1) = 0.031898703524951964_dp; coef(1) = 0.329315310335007_dp
            alpha(2) = 0.060322239914691864_dp; coef(2) = 0.6230689660925297_dp
            alpha(3) = 0.12542722211419774_dp; coef(3) = 0.1792628471936351_dp
         CASE (4)
            alpha(1) = 0.027133297390849163_dp; coef(1) = 0.16105237279351875_dp
            alpha(2) = 0.04709911293196238_dp; coef(2) = 0.5444503553138231_dp
            alpha(3) = 0.08417053216258696_dp; coef(3) = 0.36701928787783233_dp
            alpha(4) = 0.16769024925679932_dp; coef(4) = 0.06200372233567246_dp
         CASE (5)
            alpha(1) = 0.0236897013320386_dp; coef(1) = 0.07357940603188007_dp
            alpha(2) = 0.0387782492773279_dp; coef(2) = 0.39203903619828495_dp
            alpha(3) = 0.0637197820910416_dp; coef(3) = 0.46247919691910666_dp
            alpha(4) = 0.10978567271801772_dp; coef(4) = 0.1832898128285397_dp
            alpha(5) = 0.2085364358880184_dp; coef(5) = 0.023617413014962847_dp
         CASE (6)
            alpha(1) = 0.020820906789523693_dp; coef(1) = 0.029194538236458435_dp
            alpha(2) = 0.032494198134763036_dp; coef(2) = 0.23574088940165333_dp
            alpha(3) = 0.05023584266268183_dp; coef(3) = 0.44360291827793874_dp
            alpha(4) = 0.0794251556241889_dp; coef(4) = 0.3155567845762703_dp
            alpha(5) = 0.13148712880820534_dp; coef(5) = 0.09872391034903429_dp
            alpha(6) = 0.2359872300875659_dp; coef(6) = 0.012260115931631366_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",934,"Illegal expansion")
         END SELECT
      CASE (28) ! 7i
         SELECT CASE (n)
         CASE (1)
            alpha(1) = 0.06380473723420682_dp; coef(1) = 0.9999999999999998_dp
         CASE (2)
            alpha(1) = 0.045457834184705205_dp; coef(1) = 0.6463784604860205_dp
            alpha(2) = 0.10559967649617007_dp; coef(2) = 0.49629977931368163_dp
         CASE (3)
            alpha(1) = 0.03666777792117483_dp; coef(1) = 0.3678119967003228_dp
            alpha(2) = 0.07224640498805254_dp; coef(2) = 0.618554233547922_dp
            alpha(3) = 0.15965614371413928_dp; coef(3) = 0.18501800416077063_dp
         CASE (4)
            alpha(1) = 0.031314605885208034_dp; coef(1) = 0.20067272567322478_dp
            alpha(2) = 0.05654362551330421_dp; coef(2) = 0.5656501384781_dp
            alpha(3) = 0.1062477818701912_dp; coef(3) = 0.3489852946890129_dp
            alpha(4) = 0.230241610829344_dp; coef(4) = 0.06217858499289526_dp
         CASE (5)
            alpha(1) = 0.02764326937219174_dp; coef(1) = 0.1083794411727515_dp
            alpha(2) = 0.047149043197080166_dp; coef(2) = 0.4503068342384467_dp
            alpha(3) = 0.08137277011103897_dp; coef(3) = 0.4431260931378365_dp
            alpha(4) = 0.1500423393404324_dp; coef(4) = 0.15687089041229385_dp
            alpha(5) = 0.32194304635741355_dp; coef(5) = 0.020215660073712404_dp
         CASE (6)
            alpha(1) = 0.024937530903194137_dp; coef(1) = 0.058696755716280226_dp
            alpha(2) = 0.04081596003413946_dp; coef(2) = 0.3330255270864295_dp
            alpha(3) = 0.06665805855772095_dp; coef(3) = 0.4640568401312629_dp
            alpha(4) = 0.11296341467059927_dp; coef(4) = 0.253753255644_dp
            alpha(5) = 0.2062966749989214_dp; coef(5) = 0.06316006418978992_dp
            alpha(6) = 0.4401320366389093_dp; coef(6) = 0.0065607991581329616_dp
         CASE DEFAULT
            CALL cp__b("aobasis/sto_ng.F",966,"Illegal expansion")
         END SELECT
      CASE DEFAULT
         CALL cp__b("aobasis/sto_ng.F",969,"Illegal STO")
      END SELECT
      ! scaling: a = zeta^2 * a
      alpha(1:n) = zeta**2*alpha(1:n)

   END SUBROUTINE get_sto_ng

END MODULE sto_ng

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides Cartesian and spherical orbital pointers and indices
!> \par History
!>      - reallocate eliminated (17.07.2002,MK)
!>      - restructured and cleaned (20.05.2004,MK)
!> \author Matthias Krack (07.06.2000)
! **************************************************************************************************
MODULE orbital_pointers

! co    : Cartesian orbital pointer for a orbital shell.
! coset : Cartesian orbital pointer for a set of orbitals.
! nco   : Number of Cartesian orbitals for the angular momentum quantum
!         number l.
! ncoset: Number of Cartesian orbitals up to the angular momentum quantum
!         number l.
! nso   : Number of spherical orbitals for the angular momentum quantum
!         number l.
! nsoset: Number of spherical orbitals up to the angular momentum quantum
!         number l.
! nsoset_pm: Number of spherical orbitals up to the angular momentum quantum
!         number l for positive m only, i.e. m >= 0

   USE kinds,                           ONLY: int_size
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'orbital_pointers'

   INTEGER, SAVE :: current_maxl = -1

   INTEGER, DIMENSION(:), ALLOCATABLE     :: nco, ncoset, nso, nsoset, nsoset_pm
   INTEGER, DIMENSION(:, :), ALLOCATABLE   :: indco, indso, indso_inv
   INTEGER, DIMENSION(:, :), ALLOCATABLE   :: so, soset
   INTEGER, DIMENSION(:, :, :), ALLOCATABLE :: co, coset

! *** Public subroutines ***

   PUBLIC :: deallocate_orbital_pointers, &
             init_orbital_pointers

! *** Public variables ***

   PUBLIC :: co, &
             coset, &
             current_maxl, &
             indco, &
             indso, &
             indso_inv, &
             nco, &
             ncoset, &
             nso, &
             nsoset, &
             nsoset_pm, &
             soset

CONTAINS

! **************************************************************************************************
!> \brief  Allocate and initialize the orbital pointers.
!> \param maxl ...
!> \date   20.05.2004
!> \author MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE create_orbital_pointers(maxl)
      INTEGER, INTENT(IN)                                :: maxl

      CHARACTER(len=*), PARAMETER :: routineN = 'create_orbital_pointers', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iso, istat, l, lx, ly, lz, m

      IF (current_maxl > -1) THEN
         CALL cp_abort(cp__l("aobasis/orbital_pointers.F",85), &
                       "Orbital pointers are already allocated. "// &
                       "Use the init routine for an update")
      END IF

      IF (maxl < 0) THEN
         CALL cp_abort(cp__l("aobasis/orbital_pointers.F",91), &
                       "A negative maximum angular momentum quantum "// &
                       "number is invalid")
      END IF

!   *** Number of Cartesian orbitals for each l ***

      ALLOCATE (nco(-1:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 100, &
                          "nco", (maxl+2)*int_size)
      END IF

      nco(-1) = 0

      DO l = 0, maxl
         nco(l) = (l+1)*(l+2)/2
      END DO

!   *** Number of Cartesian orbitals up to l ***

      ALLOCATE (ncoset(-1:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 114, &
                          "ncoset", (maxl+2)*int_size)
      END IF

      ncoset(-1) = 0

      DO l = 0, maxl
         ncoset(l) = ncoset(l-1)+nco(l)
      END DO

!   *** Build the Cartesian orbital pointer and the shell orbital pointer ***

      ALLOCATE (co(0:maxl, 0:maxl, 0:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 128, &
                          "co", (maxl+1)**3*int_size)
      END IF

      co(:, :, :) = 0

      ALLOCATE (coset(-1:maxl, -1:maxl, -1:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 136, &
                          "coset", (maxl+2)**3*int_size)
      END IF

      coset(:, :, :) = 0

      coset(-1, :, :) = 1
      coset(:, -1, :) = 1
      coset(:, :, -1) = 1

      DO lx = 0, maxl
         DO ly = 0, maxl
            DO lz = 0, maxl
               l = lx+ly+lz
               IF (l > maxl) CYCLE
               co(lx, ly, lz) = 1+(l-lx)*(l-lx+1)/2+lz
               coset(lx, ly, lz) = ncoset(l-1)+co(lx, ly, lz)
            END DO
         END DO
      END DO

      ALLOCATE (indco(3, ncoset(maxl)), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 159, &
                          "indco", 3*ncoset(maxl)*int_size)
      END IF

      indco(:, :) = 0

      DO l = 0, maxl
         DO lx = 0, l
            DO ly = 0, l-lx
               lz = l-lx-ly
               indco(1:3, coset(lx, ly, lz)) = (/lx, ly, lz/)
            END DO
         END DO
      END DO

!   *** Number of spherical orbitals for each l ***

      ALLOCATE (nso(-1:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 178, &
                          "nso", (maxl+2)*int_size)
      END IF

      nso(-1) = 0

      DO l = 0, maxl
         nso(l) = 2*l+1
      END DO

!   *** Number of spherical orbitals up to l ***

      ALLOCATE (nsoset(-1:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 192, &
                          "nsoset", (maxl+2)*int_size)
      END IF
      ALLOCATE (nsoset_pm(-1:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 197, &
                          "nsoset_pm", (maxl+2)*int_size)
      END IF

      nsoset(-1) = 0
      nsoset_pm(-1) = 0

      DO l = 0, maxl
         nsoset(l) = nsoset(l-1)+nso(l)
         nsoset_pm(l) = nsoset_pm(l-1)+l+1
      END DO

      ALLOCATE (indso(2, nsoset(maxl)), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 211, &
                          "indso", 2*nsoset(maxl)*int_size)
      END IF
      ! indso_inv: inverse to indso
      ALLOCATE (indso_inv(0:maxl, -maxl:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 217, &
                          "indso_inv", 2*(1+maxl)*(2*maxl+1)*int_size)
      END IF

      indso(:, :) = 0
      indso_inv(:, :) = 0

      iso = 0
      DO l = 0, maxl
         DO m = -l, l
            iso = iso+1
            indso(1:2, iso) = (/l, m/)
            indso_inv(l, m) = iso
         END DO
      END DO

      ALLOCATE (so(0:maxl, -maxl:maxl), soset(0:maxl, -maxl:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 235, &
                          "so,soset", 2*(1+maxl)*(2*maxl+1)*int_size)
      END IF

      soset(:, :) = 0
      DO l = 0, maxl
         DO m = -l, l
            so(l, m) = nso(l)-(l-m)
            soset(l, m) = nsoset(l-1)+nso(l)-(l-m)
         END DO
      END DO

!   *** Save initialization status ***

      current_maxl = maxl

   END SUBROUTINE create_orbital_pointers

! **************************************************************************************************
!> \brief   Deallocate the orbital pointers.
!> \date    20.05.2005
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_orbital_pointers()

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_orbital_pointers', &
         routineP = moduleN//':'//routineN

      IF (current_maxl > -1) THEN

         DEALLOCATE (co)

         DEALLOCATE (coset)

         DEALLOCATE (indco)

         DEALLOCATE (indso)

         DEALLOCATE (indso_inv)

         DEALLOCATE (nco)

         DEALLOCATE (ncoset)

         DEALLOCATE (nso)

         DEALLOCATE (nsoset)

         DEALLOCATE (nsoset_pm)

         DEALLOCATE (so)

         DEALLOCATE (soset)

         current_maxl = -1

      END IF

   END SUBROUTINE deallocate_orbital_pointers

! **************************************************************************************************
!> \brief   Initialize or update the orbital pointers.
!> \param maxl ...
!> \date    07.06.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_orbital_pointers(maxl)
      INTEGER, INTENT(IN)                                :: maxl

      CHARACTER(len=*), PARAMETER :: routineN = 'init_orbital_pointers', &
         routineP = moduleN//':'//routineN

      IF (maxl < 0) THEN
         CALL cp_abort(cp__l("aobasis/orbital_pointers.F",310), &
                       "A negative maximum angular momentum quantum "// &
                       "number is invalid")
      END IF

!   *** Check, if the current initialization is sufficient ***

      IF (maxl > current_maxl) THEN
         CALL deallocate_orbital_pointers()
         CALL create_orbital_pointers(maxl)
      END IF

   END SUBROUTINE init_orbital_pointers

END MODULE orbital_pointers
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief orbital_symbols
!> \par History
!>      none
!> \author Matthias Krack (08.06.2000)
! **************************************************************************************************
MODULE orbital_symbols

! Index:
! FUNCTION cgf_symbol(n,lxyz) RESULT(symbol)
! FUNCTION sgf_symbol(n,l,m) RESULT(symbol)
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'orbital_symbols'
   PRIVATE

   CHARACTER(LEN=1), PARAMETER, DIMENSION(0:11) :: l_sym = (/"s", "p", "d", &
                                                             "f", "g", "h", &
                                                             "i", "j", "k", &
                                                             "l", "m", "n"/)

   PUBLIC :: cgf_symbol, sgf_symbol
   PUBLIC :: l_sym

CONTAINS

! **************************************************************************************************
!> \brief   Build a Cartesian orbital symbol (orbital labels for printing).
!> \param n ...
!> \param lxyz ...
!> \retval symbol ...
!> \date    07.07.99
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   FUNCTION cgf_symbol(n, lxyz) RESULT(symbol)
      INTEGER, INTENT(IN)                                :: n
      INTEGER, DIMENSION(3), INTENT(IN)                  :: lxyz
      CHARACTER(LEN=12)                                  :: symbol

      CHARACTER(len=*), PARAMETER :: routineN = 'cgf_symbol', routineP = moduleN//':'//routineN
      CHARACTER(LEN=1), DIMENSION(3), PARAMETER          :: xyz = (/"x", "y", "z"/)

      INTEGER                                            :: i, ipos, l

      symbol = ""

      IF ((n > 0) .AND. (n < 100)) THEN
         WRITE (symbol(1:2), "(I2)") n
      ELSE
         CALL cp__b("aobasis/orbital_symbols.F",56,"Invalid principal quantum number specified")
      END IF

      l = SUM(lxyz(1:3))

      IF ((l >= 0) .AND. (l <= 11)) THEN
         symbol(3:3) = l_sym(l)
      ELSE
         CALL cp__b("aobasis/orbital_symbols.F",64,"Invalid angular momentum quantum number specified")
      END IF

      ipos = 4

      DO i = 1, 3
         IF (lxyz(i) > 0) THEN
            symbol(ipos:ipos) = xyz(i)
            ipos = ipos+1
            IF (lxyz(i) > 1) THEN
               IF (lxyz(i) < 10) THEN
                  WRITE (symbol(ipos:ipos), "(I1)") lxyz(i)
                  ipos = ipos+1
               ELSE IF (lxyz(i) < 100) THEN
                  WRITE (symbol(ipos:ipos+1), "(I2)") lxyz(i)
                  ipos = ipos+2
               ELSE
                  CALL cp__b("aobasis/orbital_symbols.F",81,"Invalid magnetic quantum number specified")
               END IF
            END IF
         END IF
      END DO

   END FUNCTION cgf_symbol

! **************************************************************************************************
!> \brief   Build a spherical orbital symbol (orbital labels for printing).
!> \param n ...
!> \param l ...
!> \param m ...
!> \retval symbol ...
!> \date    11.03.99
!> \par Variables
!>       - l: Angular momentum quantum number l of the orbital.
!>       - m: Magnetic quantum number m of the orbital.
!>       - n: Principle quantum number n of the orbital.
!> \par History
!>  - Ignore n value for n = 0 (16.02.2009,MK)
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   FUNCTION sgf_symbol(n, l, m) RESULT(symbol)
      INTEGER, INTENT(IN)                                :: n, l, m
      CHARACTER(LEN=6)                                   :: symbol

      CHARACTER(len=*), PARAMETER :: routineN = 'sgf_symbol', routineP = moduleN//':'//routineN
      CHARACTER(LEN=1), DIMENSION(-1:1), PARAMETER       :: yzx = (/"y", "z", "x"/)

      INTEGER                                            :: i

      symbol = ""

      IF (n == 0) THEN
         i = 1
      ELSE IF ((n > 0) .AND. (n < 100)) THEN
         WRITE (symbol(1:2), "(I2)") n
         i = 3
      ELSE
         CALL cp__b("aobasis/orbital_symbols.F",122,"Invalid principal quantum number specified")
      END IF

      IF ((l >= 0) .AND. (l <= 11)) THEN
         symbol(i:i) = l_sym(l)
         i = i+1
      ELSE
         CALL cp__b("aobasis/orbital_symbols.F",129,"Invalid angular momentum quantum number specified")
      END IF

      IF (ABS(m) <= l) THEN
         IF (l == 1) THEN
            symbol(i:i) = yzx(m)
         ELSE IF (l > 1) THEN
            IF (m == 0) THEN
               WRITE (symbol(i:i), "(I1)") m
            ELSE IF (ABS(m) < 10) THEN
               WRITE (symbol(i:i+1), "(SP,I2)") m
            ELSE IF (ABS(m) < 100) THEN
               WRITE (symbol(i:i+2), "(SP,I3)") m
            END IF
         END IF
      ELSE
         CALL cp__b("aobasis/orbital_symbols.F",145,"Invalid magnetic quantum number specified")
      END IF

   END FUNCTION sgf_symbol

END MODULE orbital_symbols
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the spherical harmonics and the corresponding orbital
!>      transformation matrices.
!> \par Literature
!>      H. B. Schlegel, M. J. Frisch, Int. J. Quantum Chem. 54, 83 (1995)
!> \par History
!>      - restructured and cleaned (20.05.2004,MK)
!> \author Matthias Krack (08.06.2000)
! **************************************************************************************************
MODULE orbital_transformation_matrices

   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              pi
   USE mathlib,                         ONLY: binomial
   USE orbital_pointers,                ONLY: co,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_symbols,                 ONLY: cgf_symbol,&
                                              sgf_symbol
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'orbital_transformation_matrices'

! **************************************************************************************************
   TYPE orbtramat_type
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: c2s, slm, s2c
   END TYPE orbtramat_type

   REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE  :: c2s_tramat, s2c_tramat
   TYPE(orbtramat_type), DIMENSION(:), POINTER :: orbtramat

   INTEGER, SAVE :: current_maxl = -1

! *** Public variables ***

   PUBLIC :: orbtramat, c2s_tramat

! *** Public data types ***

! *** Public subroutines ***

   PUBLIC :: deallocate_spherical_harmonics, &
             init_spherical_harmonics

CONTAINS

! **************************************************************************************************
!> \brief  Allocate and initialize the orbital transformation matrices for
!>         the transformation and for the backtransformation of orbitals
!>         from the Cartesian representation to the spherical representation.
!> \param maxl ...
!> \date    20.05.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE create_spherical_harmonics(maxl)

      INTEGER, INTENT(IN)                                :: maxl

      CHARACTER(len=*), PARAMETER :: routineN = 'create_spherical_harmonics', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: expo, i, ic, ic1, ic2, is, istat, j, k, &
                                                            l, lx, lx1, lx2, ly, ly1, ly2, lz, &
                                                            lz1, lz2, m, ma, nc, ns
      REAL(KIND=dp)                                      :: s, s1, s2

      IF (current_maxl > -1) THEN
         CALL cp_abort(cp__l("aobasis/orbital_transformation_matrices.F",85), &
                       "Spherical harmonics are already allocated. "// &
                       "Use the init routine for an update")
      END IF

      IF (maxl < 0) THEN
         CALL cp_abort(cp__l("aobasis/orbital_transformation_matrices.F",91), &
                       "A negative maximum angular momentum quantum "// &
                       "number is invalid")
      END IF

      ALLOCATE (orbtramat(0:maxl), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 98, &
                          "orbtramat", (maxl+1)*int_size)
      END IF
      nc = ncoset(maxl)
      ns = nsoset(maxl)
      ALLOCATE (c2s_tramat(ns, nc), s2c_tramat(ns, nc), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 105, &
                          "c2s_tramat, s2c_tramat", (2*ns*nc)*int_size)
      END IF
      c2s_tramat = 0.0_dp
      s2c_tramat = 0.0_dp

      DO l = 0, maxl

         nc = nco(l)
         ns = nso(l)

         ALLOCATE (orbtramat(l)%c2s(ns, nc), STAT=istat)
         IF (istat /= 0) THEN
            CALL stop_memory(routineN, moduleN, 118, &
                             "orbtramat(l)%c2s", ns*nc*dp_size)
         END IF
         orbtramat(l)%c2s = 0.0_dp

         ALLOCATE (orbtramat(l)%slm(ns, nc), STAT=istat)
         IF (istat /= 0) THEN
            CALL stop_memory(routineN, moduleN, 125, &
                             "orbtramat(l)%slm", ns*nc*dp_size)
         END IF
         orbtramat(l)%slm = 0.0_dp

         ALLOCATE (orbtramat(l)%s2c(ns, nc), STAT=istat)
         IF (istat /= 0) THEN
            CALL stop_memory(routineN, moduleN, 132, &
                             "orbtramat(l)%s2c", ns*nc*dp_size)
         END IF
         orbtramat(l)%s2c = 0.0_dp

      END DO

!   *** Loop over all angular momentum quantum numbers l ***

      DO l = 0, maxl

!     *** Build the orbital transformation matrix for the     ***
!     *** transformation from Cartesian to spherical orbitals ***
!     *** (c2s, formula 15)                                   ***

         DO lx = 0, l
            DO ly = 0, l-lx
               lz = l-lx-ly
               ic = co(lx, ly, lz)
               DO m = -l, l
                  is = l+m+1
                  ma = ABS(m)
                  j = lx+ly-ma
                  IF ((j >= 0) .AND. (MODULO(j, 2) == 0)) THEN
                     j = j/2
                     s1 = 0.0_dp
                     DO i = 0, (l-ma)/2
                        s2 = 0.0_dp
                        DO k = 0, j
                           IF (((m < 0) .AND. (MODULO(ABS(ma-lx), 2) == 1)) .OR. &
                               ((m > 0) .AND. (MODULO(ABS(ma-lx), 2) == 0))) THEN
                              expo = (ma-lx+2*k)/2
                              s = (-1.0_dp)**expo*SQRT(2.0_dp)
                           ELSE IF ((m == 0) .AND. (MODULO(lx, 2) == 0)) THEN
                              expo = k-lx/2
                              s = (-1.0_dp)**expo
                           ELSE
                              s = 0.0_dp
                           END IF
                           s2 = s2+binomial(j, k)*binomial(ma, lx-2*k)*s
                        END DO
                        s1 = s1+binomial(l, i)*binomial(i, j)* &
                             (-1.0_dp)**i*fac(2*l-2*i)/fac(l-ma-2*i)*s2
                     END DO
                     orbtramat(l)%c2s(is, ic) = &
                        SQRT((fac(2*lx)*fac(2*ly)*fac(2*lz)*fac(l)*fac(l-ma))/ &
                             (fac(lx)*fac(ly)*fac(lz)*fac(2*l)*fac(l+ma)))*s1/ &
                        (2.0_dp**l*fac(l))
                  ELSE
                     orbtramat(l)%c2s(is, ic) = 0.0_dp
                  END IF
               END DO
            END DO
         END DO

!     *** Build the corresponding transformation matrix for the ***
!     *** transformation from spherical to Cartesian orbitals   ***
!     *** (s2c = s*TRANSPOSE(c2s), formulas 18 and 19)          ***

         DO lx1 = 0, l
            DO ly1 = 0, l-lx1
               lz1 = l-lx1-ly1
               ic1 = co(lx1, ly1, lz1)
               s1 = SQRT((fac(lx1)*fac(ly1)*fac(lz1))/ &
                         (fac(2*lx1)*fac(2*ly1)*fac(2*lz1)))
               DO lx2 = 0, l
                  DO ly2 = 0, l-lx2
                     lz2 = l-lx2-ly2
                     ic2 = co(lx2, ly2, lz2)
                     lx = lx1+lx2
                     ly = ly1+ly2
                     lz = lz1+lz2
                     IF ((MODULO(lx, 2) == 0) .AND. &
                         (MODULO(ly, 2) == 0) .AND. &
                         (MODULO(lz, 2) == 0)) THEN
                        s2 = SQRT((fac(lx2)*fac(ly2)*fac(lz2))/ &
                                  (fac(2*lx2)*fac(2*ly2)*fac(2*lz2)))
                        s = fac(lx)*fac(ly)*fac(lz)*s1*s2/ &
                            (fac(lx/2)*fac(ly/2)*fac(lz/2))
                        DO is = 1, nso(l)
                           orbtramat(l)%s2c(is, ic1) = orbtramat(l)%s2c(is, ic1)+ &
                                                       s*orbtramat(l)%c2s(is, ic2)
                        END DO
                     END IF
                  END DO
               END DO
            END DO
         END DO

!     *** Build up the real spherical harmonics ***

         s = SQRT(0.25_dp*dfac(2*l+1)/pi)

         DO lx = 0, l
            DO ly = 0, l-lx
               lz = l-lx-ly
               ic = co(lx, ly, lz)
               DO m = -l, l
                  is = l+m+1
!MK         s2 = (-1.0_dp)**m*s !-> alternative S(lm) definition
                  orbtramat(l)%slm(is, ic) = &
                     s*orbtramat(l)%c2s(is, ic)/ &
                     !MK           s2*orbtramat(l)%c2s(is,ic)/&
                     SQRT(dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))
               END DO
            END DO
         END DO

!     *** Build up one shot transformation ***
!     *** for unnormalized functions       ***
         nc = nco(l)
         ns = nso(l)
         lx1 = ncoset(l-1)+1
         ly1 = nsoset(l-1)+1
         lx2 = ncoset(l)
         ly2 = nsoset(l)
         s2c_tramat(ly1:ly2, lx1:lx2) = orbtramat(l)%s2c(1:ns, 1:nc)
         c2s_tramat(ly1:ly2, lx1:lx2) = orbtramat(l)%c2s(1:ns, 1:nc)

         s1 = 8.0_dp*pi*fac(l+1)/fac(2*l+2)
         DO lx = 0, l
            DO ly = 0, l-lx
               lz = l-lx-ly
               s2 = fac(2*lx)*fac(2*ly)*fac(2*lz)/(fac(lx)*fac(ly)*fac(lz))
               s = SQRT(s1*s2)
               ic = ncoset(l-1)+co(lx, ly, lz)
               DO m = -l, l
                  is = nsoset(l-1)+l+m+1
                  s2c_tramat(is, ic) = s2c_tramat(is, ic)*s
                  c2s_tramat(is, ic) = c2s_tramat(is, ic)/s
               END DO
            END DO
         END DO

      END DO

!   *** Save initialization status ***

      current_maxl = maxl

   END SUBROUTINE create_spherical_harmonics

! **************************************************************************************************
!> \brief   Deallocate the orbital transformation matrices.
!> \date    20.05.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_spherical_harmonics()
      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_spherical_harmonics', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: l

      IF (current_maxl > -1) THEN

         DO l = 0, SIZE(orbtramat, 1)-1

            DEALLOCATE (orbtramat(l)%c2s)

            DEALLOCATE (orbtramat(l)%slm)

            DEALLOCATE (orbtramat(l)%s2c)

         END DO

         DEALLOCATE (orbtramat)

         DEALLOCATE (c2s_tramat, s2c_tramat)

         current_maxl = -1

      END IF

   END SUBROUTINE deallocate_spherical_harmonics

! **************************************************************************************************
!> \brief  Initialize or update the orbital transformation matrices.
!> \param maxl ...
!> \param output_unit ...
!> \date     09.07.1999
!> \par Variables
!>      - maxl   : Maximum angular momentum quantum number
!> \author MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_spherical_harmonics(maxl, output_unit)

      INTEGER, INTENT(IN)                                :: maxl
      INTEGER                                            :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'init_spherical_harmonics', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=78)                                  :: headline
      INTEGER                                            :: l, nc, ns

      IF (maxl < 0) THEN
         CALL cp_abort(cp__l("aobasis/orbital_transformation_matrices.F",330), &
                       "A negative maximum angular momentum quantum "// &
                       "number is invalid")
      END IF

      IF (maxl > current_maxl) THEN

         CALL deallocate_spherical_harmonics()
         CALL create_spherical_harmonics(maxl)

!     *** Print the spherical harmonics and   ***
!     *** the orbital transformation matrices ***
         IF (output_unit > 0) THEN

            DO l = 0, maxl

               nc = nco(l)
               ns = nso(l)

               headline = "SPHERICAL HARMONICS"
               CALL write_matrix(orbtramat(l)%slm, l, output_unit, headline)

               headline = "CARTESIAN ORBITAL TO SPHERICAL ORBITAL "// &
                          "TRANSFORMATION MATRIX"
               CALL write_matrix(orbtramat(l)%c2s, l, output_unit, headline)

               headline = "SPHERICAL ORBITAL TO CARTESIAN ORBITAL "// &
                          "TRANSFORMATION MATRIX"
               CALL write_matrix(orbtramat(l)%s2c, l, output_unit, headline)

            END DO

            WRITE (UNIT=output_unit, FMT="(A)") ""

         END IF

      END IF

   END SUBROUTINE init_spherical_harmonics

! **************************************************************************************************
!> \brief   Print a matrix to the logical unit "lunit".
!> \param matrix ...
!> \param l ...
!> \param lunit ...
!> \param headline ...
!> \date    07.06.2000
!> \par Variables
!>      - matrix  : Matrix to be printed.
!>      - l       : Angular momentum quantum number
!>      - lunit   : Logical unit number.
!>      - headline: Headline of the matrix.
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_matrix(matrix, l, lunit, headline)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: matrix
      INTEGER, INTENT(IN)                                :: l, lunit
      CHARACTER(LEN=*), INTENT(IN)                       :: headline

      CHARACTER(12)                                      :: symbol
      CHARACTER(LEN=78)                                  :: string
      INTEGER                                            :: from, i, ic, is, jc, lx, ly, lz, m, nc, &
                                                            to

!   *** Write headline ***

      WRITE (UNIT=lunit, FMT="(/,/,T2,A)") TRIM(headline)

!   *** Write the matrix in the defined format ***

      nc = nco(l)

      DO ic = 1, nc, 6
         from = ic
         to = MIN(nc, from+5)
         i = 1
         DO lx = l, 0, -1
            DO ly = l-lx, 0, -1
               lz = l-lx-ly
               jc = co(lx, ly, lz)
               IF ((jc >= from) .AND. (jc <= to)) THEN
                  symbol = cgf_symbol(1, (/lx, ly, lz/))
                  WRITE (UNIT=string(i:), FMT="(A)") TRIM(symbol(3:12))
                  i = i+12
               END IF
            END DO
         END DO
         WRITE (UNIT=lunit, FMT="(/,T13,A)") TRIM(string)
         symbol = ""
         DO m = -l, l
            is = l+m+1
            symbol = sgf_symbol(1, l, m)
            WRITE (UNIT=lunit, FMT="(T4,A4,6(1X,F11.6))") &
               symbol(3:6), (matrix(is, jc), jc=from, to)
         END DO
      END DO

   END SUBROUTINE write_matrix

END MODULE orbital_transformation_matrices
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief a module to allow simple internal preprocessing in input files.
!> \par History
!>      - standalone proof-of-concept implementation (20.02.2008,AK)
!>      - integration into cp2k (22.02.2008,tlaino)
!>      - variables added (25.02.2008,AK)
!> \author Axel Kohlmeyer [AK] - CMM/UPenn Philadelphia
!> \date 25.02.2008
! **************************************************************************************************
MODULE cp_parser_ilist_methods
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_parser_ilist_types,           ONLY: ilist_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: ilist_setup, ilist_update, ilist_reset
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_ilist_methods'

CONTAINS

! ****************************************************************************
!> \brief setup the integer listing type
!> \param ilist ...
!> \param token ...
!> \date  08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE ilist_setup(ilist, token)
      TYPE(ilist_type), POINTER                          :: ilist
      CHARACTER(LEN=*)                                   :: token

      CHARACTER(len=*), PARAMETER :: routineN = 'ilist_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: ind

      IF(.NOT.(ASSOCIATED(ilist)))CALL cp__a("input/cp_parser_ilist_methods.F",43)
      ind = INDEX(token, "..")
      READ (UNIT=token(:ind-1), FMT=*) ilist%istart
      READ (UNIT=token(ind+2:), FMT=*) ilist%iend
      IF (ilist%istart > ilist%iend) &
         CALL cp_abort(cp__l("input/cp_parser_ilist_methods.F",48), &
                       "Invalid list range specified: "// &
                       TRIM(ADJUSTL(cp_to_string(ilist%istart)))//".."// &
                       TRIM(ADJUSTL(cp_to_string(ilist%iend))))
      ilist%nel_list = ilist%iend-ilist%istart+1
      ilist%ipresent = ilist%istart
      ilist%in_use = .TRUE.

   END SUBROUTINE ilist_setup

! ****************************************************************************
!> \brief updates the integer listing type
!> \param ilist ...
!> \date  08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE ilist_update(ilist)
      TYPE(ilist_type), POINTER                          :: ilist

      CHARACTER(len=*), PARAMETER :: routineN = 'ilist_update', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ilist)))CALL cp__a("input/cp_parser_ilist_methods.F",69)
      ilist%ipresent = ilist%ipresent+1
      IF (ilist%ipresent > ilist%iend) THEN
         CALL ilist_reset(ilist)
      END IF
   END SUBROUTINE ilist_update

! ****************************************************************************
!> \brief updates the integer listing type
!> \param ilist ...
!> \date  08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE ilist_reset(ilist)
      TYPE(ilist_type), POINTER                          :: ilist

      CHARACTER(len=*), PARAMETER :: routineN = 'ilist_reset', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ilist)))CALL cp__a("input/cp_parser_ilist_methods.F",87)
      IF (ilist%ipresent == ilist%iend) THEN
         ilist%istart = HUGE(0)
         ilist%iend = HUGE(0)
         ilist%nel_list = HUGE(0)
         ilist%ipresent = HUGE(0)
         ilist%in_use = .FALSE.
      END IF
   END SUBROUTINE ilist_reset

END MODULE cp_parser_ilist_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief a module to allow simple internal preprocessing in input files.
!> \par History
!>      - standalone proof-of-concept implementation (20.02.2008,AK)
!>      - integration into cp2k (22.02.2008,tlaino)
!>      - variables added (23.02.2008,AK)
!>      - @IF/@ENDIF added (25.02.2008,AK)
!>      - @PRINT and debug ifdefs added (26.02.2008,AK)
!> \author Axel Kohlmeyer [AK] - CMM/UPenn Philadelphia
!> \date 20.02.2008
! **************************************************************************************************
MODULE cp_parser_inpp_methods
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_parser_inpp_types,            ONLY: inpp_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
   USE memory_utilities,                ONLY: reallocate
   USE string_utilities,                ONLY: is_whitespace,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_inpp_methods'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   INTEGER, PARAMETER, PRIVATE          :: max_message_length = 400

   PUBLIC  :: inpp_process_directive, inpp_end_include, inpp_expand_variables
   PRIVATE :: inpp_find_variable, inpp_list_variables

CONTAINS

! **************************************************************************************************
!> \brief process internal preprocessor directives like @INCLUDE, @SET, @IF/@ENDIF
!> \param inpp ...
!> \param input_line ...
!> \param input_file_name ...
!> \param input_line_number ...
!> \param input_unit ...
!> \par History
!>      - standalone proof-of-concept implementation (20.02.2008,AK)
!>      - integration into cp2k (22.02.2008,tlaino)
!>      - variables added (23.02.2008,AK)
!>      - @IF/@ENDIF added (25.02.2008,AK)
!> \author AK
! **************************************************************************************************
   SUBROUTINE inpp_process_directive(inpp, input_line, input_file_name, input_line_number, &
                                     input_unit)
      TYPE(inpp_type), POINTER                           :: inpp
      CHARACTER(LEN=*), INTENT(INOUT)                    :: input_line, input_file_name
      INTEGER, INTENT(INOUT)                             :: input_line_number, input_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'inpp_process_directive', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: cond1, cond2, filename, mytag, value, &
                                                            varname
      CHARACTER(LEN=max_message_length)                  :: message
      INTEGER                                            :: i, indf, indi, istat, output_unit, pos1, &
                                                            pos2, unit
      LOGICAL                                            :: check

      output_unit = cp_logger_get_default_io_unit()

      IF(.NOT.(ASSOCIATED(inpp)))CALL cp__a("input/cp_parser_inpp_methods.F",73)

      ! find location of directive in line and check whether it is commented out
      indi = INDEX(input_line, "@")
      pos1 = INDEX(input_line, "!")
      pos2 = INDEX(input_line, "#")
      IF (((pos1 > 0) .AND. (pos1 < indi)) .OR. ((pos2 > 0) .AND. (pos2 < indi))) THEN
         ! nothing to do here.
         RETURN
      ENDIF

      ! Get the start of the instruction and find "@KEYWORD" (or "@")
      indf = indi
      DO WHILE (.NOT. is_whitespace(input_line(indf:indf)))
         indf = indf+1
      END DO
      mytag = input_line(indi:indf-1)
      CALL uppercase(mytag)

      SELECT CASE (mytag)

      CASE ("@INCLUDE")
         ! Get the filename.. allow for " or ' or nothing..
         filename = TRIM(input_line(indf:))
         IF (LEN_TRIM(filename) == 0) THEN
            WRITE (UNIT=message, FMT="(3A,I6)") &
               "INPP_@INCLUDE: Incorrect @INCLUDE directive in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number
            CALL cp__b("input/cp_parser_inpp_methods.F",101,TRIM(message))
         ENDIF
         indi = 1
         DO WHILE (is_whitespace(filename(indi:indi)))
            indi = indi+1
         END DO
         filename = TRIM(filename(indi:))

         ! handle quoting of the filename
         pos1 = INDEX(filename, '"')
         pos2 = INDEX(filename(pos1+1:), '"')
         IF ((pos1 /= 0) .AND. (pos2 /= 0)) THEN
            filename = filename(pos1+1:pos1+pos2-1)
         ELSE
            pos1 = INDEX(filename, "'")
            pos2 = INDEX(filename(pos1+1:), "'")
            IF ((pos1 /= 0) .AND. (pos2 /= 0)) THEN
               filename = filename(pos1+1:pos1+pos2-1)
            ELSE
               ! incorrect quotes (only one of ' or ").
               pos2 = INDEX(filename, '"')
               IF ((pos1 /= 0) .OR. (pos2 /= 0)) THEN
                  WRITE (UNIT=message, FMT="(3A,I6)") &
                     "INPP_@INCLUDE: Incorrect quoting of include file in file: ", &
                     TRIM(input_file_name), "  Line:", input_line_number
                  CALL cp__b("input/cp_parser_inpp_methods.F",126,TRIM(message))
               ENDIF
               ! nothing to do. unquoted filename.
            ENDIF
         ENDIF

         ! Let's check that files already opened won't be again opened
         DO i = 1, inpp%io_stack_level
            check = TRIM(filename) /= TRIM(inpp%io_stack_filename(i))
            IF(.NOT.(check))CALL cp__a("input/cp_parser_inpp_methods.F",135)
         END DO

         ! this stops on so we can always assume success
         CALL open_file(file_name=TRIM(filename), &
                        file_status="OLD", &
                        file_form="FORMATTED", &
                        file_action="READ", &
                        unit_number=unit)

         IF (debug_this_module .AND. output_unit > 0) THEN
            WRITE (UNIT=message, FMT="(3A,I6,2A)") "INPP_@INCLUDE: in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number, &
               " Opened include file: ", TRIM(filename)
            WRITE (output_unit, *) TRIM(message)
         END IF

         ! make room, save status and position the parser at the beginning of new file.
         inpp%io_stack_level = inpp%io_stack_level+1
         CALL reallocate(inpp%io_stack_channel, 1, inpp%io_stack_level)
         CALL reallocate(inpp%io_stack_lineno, 1, inpp%io_stack_level)
         CALL reallocate(p_long=inpp%io_stack_filename, lb_new=1, ub_new=inpp%io_stack_level)

         inpp%io_stack_channel(inpp%io_stack_level) = input_unit
         inpp%io_stack_lineno(inpp%io_stack_level) = input_line_number
         inpp%io_stack_filename(inpp%io_stack_level) = input_file_name

         input_file_name = TRIM(filename)
         input_line_number = 0
         input_unit = unit

      CASE ("@SET")
         ! split directive into variable name and value data.
         varname = TRIM(input_line(indf:))
         IF (LEN_TRIM(varname) == 0) THEN
            WRITE (UNIT=message, FMT="(3A,I6)") &
               "INPP_@SET: Incorrect @SET directive in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number
            CALL cp__b("input/cp_parser_inpp_methods.F",173,TRIM(message))
         ENDIF

         indi = 1
         DO WHILE (is_whitespace(varname(indi:indi)))
            indi = indi+1
         END DO
         indf = indi
         DO WHILE (.NOT. is_whitespace(varname(indf:indf)))
            indf = indf+1
         END DO
         value = TRIM(varname(indf:))
         varname = TRIM(varname(indi:indf-1))
         indi = 1
         DO WHILE (is_whitespace(value(indi:indi)))
            indi = indi+1
         END DO
         value = TRIM(value(indi:))

         IF (LEN_TRIM(value) == 0) THEN
            WRITE (UNIT=message, FMT="(3A,I6)") &
               "INPP_@SET: Incorrect @SET directive in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number
            CALL cp__b("input/cp_parser_inpp_methods.F",196,TRIM(message))
         ENDIF

         ! sort into table of variables.
         indi = inpp_find_variable(inpp, varname)
         IF (indi == 0) THEN
            ! create new variable
            inpp%num_variables = inpp%num_variables+1
            CALL reallocate(p_long=inpp%variable_name, lb_new=1, ub_new=inpp%num_variables)
            CALL reallocate(p_long=inpp%variable_value, lb_new=1, ub_new=inpp%num_variables)
            inpp%variable_name(inpp%num_variables) = varname
            inpp%variable_value(inpp%num_variables) = value
            IF (debug_this_module .AND. output_unit > 0) THEN
               WRITE (UNIT=message, FMT="(3A,I6,4A)") "INPP_@SET: in file: ", &
                  TRIM(input_file_name), "  Line:", input_line_number, &
                  " Set new variable ", TRIM(varname), " to value: ", TRIM(value)
               WRITE (output_unit, *) TRIM(message)
            END IF
         ELSE
            ! reassign variable
            IF (debug_this_module .AND. output_unit > 0) THEN
               WRITE (UNIT=message, FMT="(3A,I6,6A)") "INPP_@SET: in file: ", &
                  TRIM(input_file_name), "  Line:", input_line_number, &
                  " Change variable ", TRIM(varname), " from value: ", &
                  TRIM(inpp%variable_value(indi)), " to value: ", TRIM(value)
               WRITE (output_unit, *) TRIM(message)
            END IF
            inpp%variable_value(indi) = value
         ENDIF

         IF (debug_this_module) CALL inpp_list_variables(inpp, 6)

      CASE ("@IF")
         ! detect IF expression.
         ! we recognize lexical equality or inequality, and presence of
         ! a string (true) vs. blank (false). in case the expression resolves
         ! to "false" we read lines here until we reach an @ENDIF or EOF.
         indi = indf
         pos1 = INDEX(input_line, "==")
         pos2 = INDEX(input_line, "/=")
         ! shave off leading whitespace
         DO WHILE (is_whitespace(input_line(indi:indi)))
            indi = indi+1
            IF (indi > LEN_TRIM(input_line)) EXIT
         END DO
         check = .FALSE.
         IF (pos1 > 0) THEN
            cond1 = input_line(indi:pos1-1)
            cond2 = input_line(pos1+2:)
            check = .TRUE.
            IF ((pos2 > 0) .OR. (INDEX(cond2, "==") > 0)) THEN
               WRITE (UNIT=message, FMT="(3A,I6)") &
                  "INPP_@IF: Incorrect @IF directive in file: ", &
                  TRIM(input_file_name), "  Line:", input_line_number
               CALL cp__b("input/cp_parser_inpp_methods.F",250,TRIM(message))
            ENDIF
         ELSEIF (pos2 > 0) THEN
            cond1 = input_line(indi:pos2-1)
            cond2 = input_line(pos2+2:)
            check = .FALSE.
            IF ((pos1 > 0) .OR. (INDEX(cond2, "/=") > 0)) THEN
               WRITE (UNIT=message, FMT="(3A,I6)") &
                  "INPP_@IF: Incorrect @IF directive in file: ", &
                  TRIM(input_file_name), "  Line:", input_line_number
               CALL cp__b("input/cp_parser_inpp_methods.F",260,TRIM(message))
            ENDIF
         ELSE
            IF (LEN_TRIM(input_line(indi:)) > 0) THEN
               IF (TRIM(input_line(indi:)) == '0') THEN
                  cond1 = 'XXX'
                  cond2 = 'XXX'
                  check = .FALSE.
               ELSE
                  cond1 = 'XXX'
                  cond2 = 'XXX'
                  check = .TRUE.
               ENDIF
            ELSE
               cond1 = 'XXX'
               cond2 = 'XXX'
               check = .FALSE.
            ENDIF
         ENDIF

         ! Get rid of possible parentheses
         IF (INDEX(cond1, "(") /= 0) cond1 = cond1(INDEX(cond1, "(")+1:)
         IF (INDEX(cond2, ")") /= 0) cond2 = cond2(1:INDEX(cond2, ")")-1)

         ! Shave off leading whitespace from cond1
         indi = 1
         DO WHILE (is_whitespace(cond1(indi:indi)))
            indi = indi+1
         END DO
         cond1 = cond1(indi:)

         ! Shave off leading whitespace from cond2
         indi = 1
         DO WHILE (is_whitespace(cond2(indi:indi)))
            indi = indi+1
         END DO
         cond2 = cond2(indi:)

         IF (LEN_TRIM(cond2) == 0) THEN
            WRITE (UNIT=message, FMT="(3A,I6)") &
               "INPP_@IF: Incorrect @IF directive in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number
            CALL cp__b("input/cp_parser_inpp_methods.F",302,TRIM(message))
         ENDIF

         IF ((TRIM(cond1) == TRIM(cond2)) .EQV. check) THEN
            IF (debug_this_module .AND. output_unit > 0) THEN
               WRITE (UNIT=message, FMT="(3A,I6,A)") "INPP_@IF: in file: ", &
                  TRIM(input_file_name), "  Line:", input_line_number, &
                  " Conditional ("//TRIM(cond1)//","//TRIM(cond2)// &
                  ") resolves to true. Continuing parsing."
               WRITE (output_unit, *) TRIM(message)
            END IF
            ! resolves to true. keep on reading normally...
            RETURN
         ELSE
            IF (debug_this_module .AND. output_unit > 0) THEN
               WRITE (UNIT=message, FMT="(3A,I6,A)") "INPP_@IF: in file: ", &
                  TRIM(input_file_name), "  Line:", input_line_number, &
                  " Conditional ("//TRIM(cond1)//","//TRIM(cond2)// &
                  ") resolves to false. Skipping Lines."
               WRITE (output_unit, *) TRIM(message)
            END IF
            istat = 0
            DO WHILE (istat == 0)
               input_line_number = input_line_number+1
               READ (UNIT=input_unit, FMT="(A)", IOSTAT=istat) input_line
               IF (debug_this_module .AND. output_unit > 0) THEN
                  WRITE (UNIT=message, FMT="(1A,I6,2A)") "INPP_@IF: skipping line ", &
                     input_line_number, ": ", TRIM(input_line)
                  WRITE (output_unit, *) TRIM(message)
               END IF

               indi = INDEX(input_line, "@")
               pos1 = INDEX(input_line, "!")
               pos2 = INDEX(input_line, "#")
               IF (((pos1 > 0) .AND. (pos1 < indi)) .OR. ((pos2 > 0) .AND. (pos2 < indi))) THEN
                  ! comment. nothing to do here.
                  CYCLE
               ENDIF

               ! Get the start of the instruction and find "@KEYWORD"
               indi = MAX(1, indi)
               indf = indi
               DO WHILE (input_line(indf:indf) /= " ")
                  indf = indf+1
               END DO
               IF(.NOT.((indf-indi) <= default_string_length))CALL cp__a("input/cp_parser_inpp_methods.F",347)
               mytag = input_line(indi:indf-1)
               CALL uppercase(mytag)
               IF (INDEX(mytag, "@ENDIF") > 0) THEN
                  ! ok found it. go back to normal
                  IF (debug_this_module .AND. output_unit > 0) THEN
                     WRITE (output_unit, *) "INPP_@IF: found @ENDIF. End of skipping."
                  END IF
                  RETURN
               ENDIF
            END DO
            IF (istat /= 0) THEN
               WRITE (UNIT=message, FMT="(3A,I6)") &
                  "INPP_@IF: Error while looking for @ENDIF directive in file: ", &
                  TRIM(input_file_name), "  Line:", input_line_number
               CALL cp__b("input/cp_parser_inpp_methods.F",362,TRIM(message))
            ENDIF
         ENDIF

      CASE ("@ENDIF")
         IF (debug_this_module .AND. output_unit > 0) THEN
            WRITE (output_unit, *) "INPP_@IF: found @ENDIF in normal parsing. Ignoring it."
         END IF
         ! nothing to do. just return to skip the line.
         RETURN

      CASE ("@PRINT")
         ! for debugging of variables etc.
         IF (output_unit > 0) THEN
            WRITE (UNIT=message, FMT="(3A,I6,2A)") "INPP_@PRINT: in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number, &
               " Text: ", TRIM(input_line(indf:))
            WRITE (output_unit, *) TRIM(message)
         END IF
         RETURN
         ! Do Nothing..
      END SELECT

   END SUBROUTINE inpp_process_directive

! **************************************************************************************************
!> \brief Restore older file status from stack after EOF on include file.
!> \param inpp ...
!> \param input_file_name ...
!> \param input_line_number ...
!> \param input_unit ...
!> \par History
!>      - standalone proof-of-concept implemenation (20.02.2008,AK)
!>      - integrated into cp2k (21.02.2008)
!> \author AK
! **************************************************************************************************
   SUBROUTINE inpp_end_include(inpp, input_file_name, input_line_number, input_unit)
      TYPE(inpp_type), POINTER                           :: inpp
      CHARACTER(LEN=*), INTENT(INOUT)                    :: input_file_name
      INTEGER, INTENT(INOUT)                             :: input_line_number, input_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'inpp_end_include', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(inpp)))CALL cp__a("input/cp_parser_inpp_methods.F",406)
      IF (inpp%io_stack_level > 0) THEN
         CALL close_file(input_unit)
         input_unit = inpp%io_stack_channel(inpp%io_stack_level)
         input_line_number = inpp%io_stack_lineno(inpp%io_stack_level)
         input_file_name = TRIM(inpp%io_stack_filename(inpp%io_stack_level))
         inpp%io_stack_level = inpp%io_stack_level-1
         CALL reallocate(inpp%io_stack_channel, 1, inpp%io_stack_level)
         CALL reallocate(inpp%io_stack_lineno, 1, inpp%io_stack_level)
         CALL reallocate(p_long=inpp%io_stack_filename, lb_new=1, ub_new=inpp%io_stack_level)
      ENDIF

   END SUBROUTINE inpp_end_include

! **************************************************************************************************
!> \brief expand all ${VAR} or $VAR variable entries on the input string
!> \param inpp ...
!> \param input_line ...
!> \param input_file_name ...
!> \param input_line_number ...
!> \par History
!>      - standalone proof-of-concept implemenation (22.02.2008,AK)
!>      - integrated into cp2k (23.02.2008)
!> \author AK
! **************************************************************************************************
   SUBROUTINE inpp_expand_variables(inpp, input_line, input_file_name, input_line_number)
      TYPE(inpp_type), POINTER                           :: inpp
      CHARACTER(LEN=*), INTENT(INOUT)                    :: input_line, input_file_name
      INTEGER, INTENT(IN)                                :: input_line_number

      CHARACTER(len=*), PARAMETER :: routineN = 'inpp_expand_variables', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: newline, varname
      CHARACTER(LEN=max_message_length)                  :: message
      INTEGER                                            :: idx, pos1, pos2

      IF(.NOT.(ASSOCIATED(inpp)))CALL cp__a("input/cp_parser_inpp_methods.F",443)

      ! process line until all variables named with the convention ${VAR} are expanded
      DO WHILE (INDEX(input_line, '${') > 0)
         pos1 = INDEX(input_line, '${')
         pos1 = pos1+2
         pos2 = INDEX(input_line(pos1:), '}')
         IF(.NOT.(pos2 > 0))CALL cp__a("input/cp_parser_inpp_methods.F",450)
         IF (pos2 == 0) THEN
            WRITE (UNIT=message, FMT="(3A,I6)") &
               "Missing '}' in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number
            CALL cp__b("input/cp_parser_inpp_methods.F",455,TRIM(message))
         ENDIF

         pos2 = pos1+pos2-2
         varname = input_line(pos1:pos2)
         idx = inpp_find_variable(inpp, varname)
         IF(.NOT.(idx >= 0))CALL cp__a("input/cp_parser_inpp_methods.F",461)
         IF (idx == 0) THEN
            WRITE (UNIT=message, FMT="(5A,I6)") &
               "Variable ${", TRIM(varname), "} not defined in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number
            CALL cp__b("input/cp_parser_inpp_methods.F",466,TRIM(message))
         ENDIF

         IF (pos1 > 3) THEN
            newline = input_line(1:pos1-3)//TRIM(inpp%variable_value(idx)) &
                      //input_line(pos2+2:)
         ELSE
            newline = TRIM(inpp%variable_value(idx))//input_line(pos2+2:)
         ENDIF

         input_line = newline
      END DO

      ! process line until all variables named with the convention $VAR are expanded
      DO WHILE (INDEX(input_line, '$') > 0)
         pos1 = INDEX(input_line, '$')
         pos1 = pos1+1
         pos2 = INDEX(input_line(pos1:), ' ')
         IF(.NOT.(pos2 >= 0))CALL cp__a("input/cp_parser_inpp_methods.F",484)
         IF (pos2 == 0) pos2 = LEN_TRIM(input_line(pos1:))+1

         pos2 = pos1+pos2-1
         varname = input_line(pos1:pos2)
         idx = inpp_find_variable(inpp, varname)
         IF(.NOT.(idx >= 0))CALL cp__a("input/cp_parser_inpp_methods.F",490)
         IF (idx == 0) THEN
            WRITE (UNIT=message, FMT="(5A,I6)") &
               "INPP_@SET: Variable $", TRIM(varname), " not defined in file: ", &
               TRIM(input_file_name), "  Line:", input_line_number
            CALL cp__b("input/cp_parser_inpp_methods.F",495,TRIM(message))
         ENDIF

         IF (pos1 > 2) THEN
            newline = input_line(1:pos1-2)//TRIM(inpp%variable_value(idx)) &
                      //input_line(pos2+1:)
         ELSE
            newline = TRIM(inpp%variable_value(idx))//input_line(pos2+1:)
         ENDIF

         input_line = newline
      END DO
   END SUBROUTINE inpp_expand_variables

! **************************************************************************************************
!> \brief return index position of a variable in dictionary. 0 if not found.
!> \param inpp ...
!> \param varname ...
!> \retval idx ...
!> \par History
!>      - standalone proof-of-concept implemenation (22.02.2008,AK)
!>      - integrated into cp2k (23.02.2008)
!> \author AK
! **************************************************************************************************
   FUNCTION inpp_find_variable(inpp, varname) RESULT(idx)
      TYPE(inpp_type), POINTER                           :: inpp
      CHARACTER(len=default_path_length)                 :: varname
      INTEGER                                            :: idx

      INTEGER                                            :: i

      idx = 0
      DO i = 1, inpp%num_variables
         IF (TRIM(varname) == TRIM(inpp%variable_name(i))) THEN
            idx = i
            RETURN
         ENDIF
      END DO
      RETURN
   END FUNCTION inpp_find_variable

! **************************************************************************************************
!> \brief print a list of the variable/value table
!> \param inpp ...
!> \param iochan ...
!> \par History
!>      - standalone proof-of-concept implemenation (22.02.2008,AK)
!>      - integrated into cp2k (23.02.2008)
!> \author AK
! **************************************************************************************************
   SUBROUTINE inpp_list_variables(inpp, iochan)
      TYPE(inpp_type), POINTER                           :: inpp
      INTEGER, INTENT(IN)                                :: iochan

      INTEGER                                            :: i

      WRITE (iochan, '(A)') '   #   NAME                   VALUE'
      DO i = 1, inpp%num_variables
         WRITE (iochan, '(I4," | ",A,T30," | ",A," |")') &
            i, TRIM(inpp%variable_name(i)), TRIM(inpp%variable_value(i))
      END DO
   END SUBROUTINE inpp_list_variables

END MODULE cp_parser_inpp_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility routines to read data from files.
!>      Kept as close as possible to the old parser because
!>        1. string handling is a weak point of fortran compilers, and it is
!>           easy to write correct things that do not work
!>        2. conversion of old code
!> \par History
!>      22.11.1999 first version of the old parser (called qs_parser)
!>                 Matthias Krack
!>      06.2004 removed module variables, cp_parser_type, new module [fawzi]
!> \author Fawzi Mohamed, Matthias Krack
! **************************************************************************************************
MODULE cp_parser_methods

   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_buffer_types,          ONLY: copy_buffer_type,&
                                              finalize_sub_buffer,&
                                              initialize_sub_buffer
   USE cp_parser_ilist_methods,         ONLY: ilist_reset,&
                                              ilist_setup,&
                                              ilist_update
   USE cp_parser_inpp_methods,          ONLY: inpp_end_include,&
                                              inpp_expand_variables,&
                                              inpp_process_directive
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_reset
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              max_line_length
   USE message_passing,                 ONLY: mp_bcast
   USE string_utilities,                ONLY: is_whitespace,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: parser_test_next_token, parser_get_object, parser_location, &
             parser_search_string, parser_get_next_line, parser_skip_space, &
             parser_read_line, read_float_object

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_methods'

   INTERFACE parser_get_object
      MODULE PROCEDURE parser_get_integer, &
         parser_get_logical, &
         parser_get_real, &
         parser_get_string
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief return a description of the part of the file acually parsed
!> \param parser the parser
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION parser_location(parser) RESULT(res)

      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER&
         (len=default_path_length+default_string_length) :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'parser_location', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",75)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",76)
      res = ", File: '"//TRIM(parser%input_file_name)//"', Line: "// &
            TRIM(ADJUSTL(cp_to_string(parser%input_line_number)))// &
            ", Column: "//TRIM(ADJUSTL(cp_to_string(parser%icol)))
      IF (parser%icol == -1) THEN
         res(LEN_TRIM(res):) = " (EOF)"
      ELSE IF (MAX(1, parser%icol1) <= parser%icol2) THEN
         res(LEN_TRIM(res):) = ", Chunk: <"// &
                               parser%input_line(MAX(1, parser%icol1):parser%icol2)//">"
      END IF

   END FUNCTION parser_location

! **************************************************************************************************
!> \brief   store the present status of the parser
!> \param parser ...
!> \date    08.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE parser_store_status(parser)

      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_store_status', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",102)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",103)
      IF(.NOT.(ASSOCIATED(parser%status)))CALL cp__a("input/cp_parser_methods.F",104)
      parser%status%in_use = .TRUE.
      parser%status%old_input_line = parser%input_line
      parser%status%old_input_line_number = parser%input_line_number
      parser%status%old_icol = parser%icol
      parser%status%old_icol1 = parser%icol1
      parser%status%old_icol2 = parser%icol2
      ! Store buffer info
      CALL copy_buffer_type(parser%buffer, parser%status%buffer)

   END SUBROUTINE parser_store_status

! **************************************************************************************************
!> \brief   retrieve the original status of the parser
!> \param parser ...
!> \date    08.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE parser_retrieve_status(parser)

      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_retrieve_status', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",129)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",130)

      ! Always store the new buffer (if it is really newly read)
      IF (parser%buffer%buffer_id /= parser%status%buffer%buffer_id) THEN
         CALL initialize_sub_buffer(parser%buffer%sub_buffer, parser%buffer)
      END IF
      parser%status%in_use = .FALSE.
      parser%input_line = parser%status%old_input_line
      parser%input_line_number = parser%status%old_input_line_number
      parser%icol = parser%status%old_icol
      parser%icol1 = parser%status%old_icol1
      parser%icol2 = parser%status%old_icol2

      ! Retrieve buffer info
      CALL copy_buffer_type(parser%status%buffer, parser%buffer)

   END SUBROUTINE parser_retrieve_status

! **************************************************************************************************
!> \brief   Read the next line from a logical unit "unit" (I/O node only).
!>          Skip (nline-1) lines and skip also all comment lines.
!> \param parser ...
!> \param nline ...
!> \param at_end ...
!> \date    22.11.1999
!> \author  Matthias Krack (MK)
!> \version 1.0
!> \note 08.2008 [tlaino] - Teodoro Laino UZH : updated for buffer
! **************************************************************************************************
   SUBROUTINE parser_read_line(parser, nline, at_end)

      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(IN)                                :: nline
      LOGICAL, INTENT(out), OPTIONAL                     :: at_end

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_read_line', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iline, istat

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",172)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",173)
      IF (PRESENT(at_end)) at_end = .FALSE.

      DO iline = 1, nline
         ! Try to read the next line from the buffer
         CALL parser_get_line_from_buffer(parser, istat)

         ! Handle (persisting) read errors
         IF (istat /= 0) THEN
            IF (istat < 0) THEN ! EOF/EOR is negative other errors positive
               IF (PRESENT(at_end)) THEN
                  at_end = .TRUE.
               ELSE
                  CALL cp__b("input/cp_parser_methods.F",186,"Unexpected EOF"//TRIM(parser_location(parser)))
               END IF
               parser%icol = -1
               parser%icol1 = 0
               parser%icol2 = -1
            ELSE
               CALL cp_abort(cp__l("input/cp_parser_methods.F",192), &
                             "An I/O error occurred (IOSTAT = "// &
                             TRIM(ADJUSTL(cp_to_string(istat)))//")"// &
                             TRIM(parser_location(parser)))
            END IF
            CALL timestop(handle)
            RETURN
         END IF
      END DO

      ! Reset column pointer, if a new line was read
      IF (nline > 0) parser%icol = 0

      CALL timestop(handle)
   END SUBROUTINE parser_read_line

! **************************************************************************************************
!> \brief   Retrieving lines from buffer
!> \param parser ...
!> \param istat ...
!> \date    08.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE parser_get_line_from_buffer(parser, istat)

      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(OUT)                               :: istat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_get_line_from_buffer', &
         routineP = moduleN//':'//routineN

      istat = 0
      ! Check buffer
      IF (parser%buffer%present_line_number == parser%buffer%size) THEN
         IF (ASSOCIATED(parser%buffer%sub_buffer)) THEN
            ! If the sub_buffer is initialized let's restore its buffer
            CALL finalize_sub_buffer(parser%buffer%sub_buffer, parser%buffer)
         ELSE
            ! Rebuffer input file if required
            CALL parser_read_line_low(parser)
         END IF
      END IF
      parser%buffer%present_line_number = parser%buffer%present_line_number+1
      parser%input_line_number = parser%buffer%input_line_numbers(parser%buffer%present_line_number)
      parser%input_line = parser%buffer%input_lines(parser%buffer%present_line_number)
      IF ((parser%buffer%istat /= 0) .AND. &
          (parser%buffer%last_line_number == parser%buffer%present_line_number)) THEN
         istat = parser%buffer%istat
      END IF

   END SUBROUTINE parser_get_line_from_buffer

! **************************************************************************************************
!> \brief   Low level reading subroutine with buffering
!> \param parser ...
!> \date    08.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE parser_read_line_low(parser)

      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_read_line_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iline, imark, islen, istat, &
                                                            last_buffered_line_number
      LOGICAL                                            :: non_white_found, &
                                                            this_line_is_white_or_comment

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",264)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",265)
      parser%buffer%input_lines = ""
      IF (parser%para_env%mepos == parser%para_env%source) THEN
         iline = 0
         istat = 0
         parser%buffer%buffer_id = parser%buffer%buffer_id+1
         parser%buffer%present_line_number = 0
         parser%buffer%last_line_number = parser%buffer%size
         last_buffered_line_number = parser%buffer%input_line_numbers(parser%buffer%size)
         DO WHILE (iline /= parser%buffer%size)
            ! Increment counters by 1
            iline = iline+1
            last_buffered_line_number = last_buffered_line_number+1

            ! Try to read the next line from file
            parser%buffer%input_line_numbers(iline) = last_buffered_line_number
            READ (UNIT=parser%input_unit, FMT="(A)", IOSTAT=istat) parser%buffer%input_lines(iline)
            islen = LEN_TRIM(parser%buffer%input_lines(iline))

            ! Pre-processing steps:
            ! 1. Expand variables 2. Process directives and read next line.
            ! On read failure try to go back from included file to previous i/o-stream.
            IF (istat == 0) THEN
               this_line_is_white_or_comment = is_comment_line(parser, parser%buffer%input_lines(iline))
               IF (.NOT. this_line_is_white_or_comment .AND. parser%apply_preprocessing) THEN
                  imark = INDEX(parser%buffer%input_lines(iline) (1:islen), "$")
                  IF (imark /= 0) THEN
                     CALL inpp_expand_variables(parser%inpp, parser%buffer%input_lines(iline), &
                                                parser%input_file_name, parser%buffer%input_line_numbers(iline))
                     islen = LEN_TRIM(parser%buffer%input_lines(iline))
                  END IF
                  imark = INDEX(parser%buffer%input_lines(iline) (1:islen), "@")
                  IF (imark /= 0) THEN
                     CALL inpp_process_directive(parser%inpp, parser%buffer%input_lines(iline), &
                                                 parser%input_file_name, parser%buffer%input_line_numbers(iline), &
                                                 parser%input_unit)
                     islen = LEN_TRIM(parser%buffer%input_lines(iline))
                     ! Handle index and cycle
                     last_buffered_line_number = 0
                     iline = iline-1
                     CYCLE
                  END IF
               END IF
            ELSE IF (istat < 0) THEN ! handle EOF
               IF (parser%inpp%io_stack_level > 0) THEN
                  ! We were reading from an included file. Go back one level.
                  CALL inpp_end_include(parser%inpp, parser%input_file_name, &
                                        parser%buffer%input_line_numbers(iline), parser%input_unit)
                  ! Handle index and cycle
                  last_buffered_line_number = parser%buffer%input_line_numbers(iline)
                  iline = iline-1
                  CYCLE
               END IF
            END IF

            ! Saving persisting read errors
            IF (istat /= 0) THEN
               parser%buffer%istat = istat
               parser%buffer%last_line_number = iline
               parser%buffer%input_line_numbers(iline:) = 0
               parser%buffer%input_lines(iline:) = ""
               EXIT
            END IF

            ! Pre-processing and error checking done. Ready for parsing.
            IF (.NOT. parser%parse_white_lines) THEN
               non_white_found = .NOT. this_line_is_white_or_comment
            ELSE
               non_white_found = .TRUE.
            END IF
            IF (.NOT. non_white_found) THEN
               iline = iline-1
               last_buffered_line_number = last_buffered_line_number-1
            END IF
         END DO
      END IF
      ! Broadcast buffer informations
      CALL broadcast_input_information(parser)

      CALL timestop(handle)

   END SUBROUTINE parser_read_line_low

! **************************************************************************************************
!> \brief   Broadcast the input information.
!> \param parser ...
!> \date    02.03.2001
!> \author  Matthias Krack (MK)
!> \note 08.2008 [tlaino] - Teodoro Laino UZH : updated for buffer
! **************************************************************************************************
   SUBROUTINE broadcast_input_information(parser)

      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(len=*), PARAMETER :: routineN = 'broadcast_input_information', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",367)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",368)
      para_env => parser%para_env
      IF (para_env%num_pe > 1) THEN
         CALL mp_bcast(parser%buffer%buffer_id, para_env%source, para_env%group)
         CALL mp_bcast(parser%buffer%present_line_number, para_env%source, para_env%group)
         CALL mp_bcast(parser%buffer%last_line_number, para_env%source, para_env%group)
         CALL mp_bcast(parser%buffer%istat, para_env%source, para_env%group)
         CALL mp_bcast(parser%buffer%input_line_numbers, para_env%source, para_env%group)
         CALL mp_bcast(parser%buffer%input_lines, para_env%source, para_env%group)
      END IF

      CALL timestop(handle)

   END SUBROUTINE broadcast_input_information

! **************************************************************************************************
!> \brief returns .true. if the line is a comment line or an empty line
!> \param parser ...
!> \param line ...
!> \retval resval ...
!> \par History
!>      03.2009 [tlaino] - Teodoro Laino
! **************************************************************************************************
   FUNCTION is_comment_line(parser, line) RESULT(resval)

      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=*), INTENT(IN)                       :: line
      LOGICAL                                            :: resval

      CHARACTER(len=*), PARAMETER :: routineN = 'is_comment_line', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: thischar
      INTEGER                                            :: icol

      resval = .TRUE.
      DO icol = 1, LEN(line)
         thischar = line(icol:icol)
         IF (.NOT. is_whitespace(thischar)) THEN
            IF (.NOT. is_comment(parser, thischar)) resval = .FALSE.
            EXIT
         END IF
      END DO

   END FUNCTION is_comment_line

! **************************************************************************************************
!> \brief returns .true. if the character passed is a comment character
!> \param parser ...
!> \param testchar ...
!> \retval resval ...
!> \par History
!>      02.2008 created, AK
!> \author AK
! **************************************************************************************************
   FUNCTION is_comment(parser, testchar) RESULT(resval)

      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=1), INTENT(IN)                       :: testchar
      LOGICAL                                            :: resval

      CHARACTER(len=*), PARAMETER :: routineN = 'is_comment', routineP = moduleN//':'//routineN

      resval = .FALSE.
      ! We are in a private function, and parser has been tested before...
      IF (ANY(parser%comment_character .EQ. testchar)) THEN
         resval = .TRUE.
      END IF

   END FUNCTION is_comment

! **************************************************************************************************
!> \brief   Read the next input line and broadcast the input information.
!>          Skip (nline-1) lines and skip also all comment lines.
!> \param parser ...
!> \param nline ...
!> \param at_end ...
!> \date    22.11.1999
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_get_next_line(parser, nline, at_end)

      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(IN)                                :: nline
      LOGICAL, INTENT(out), OPTIONAL                     :: at_end

      CHARACTER(len=*), PARAMETER :: routineN = 'parser_get_next_line', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_at_end

      IF (nline > 0) THEN
         CALL parser_read_line(parser, nline, at_end=my_at_end)
         IF (PRESENT(at_end)) THEN
            at_end = my_at_end
         ELSE
            IF (my_at_end) THEN
               CALL cp__b("input/cp_parser_methods.F",466,"Unexpected EOF"//TRIM(parser_location(parser)))
            END IF
         END IF
      ELSE IF (PRESENT(at_end)) THEN
         at_end = .FALSE.
      END IF

   END SUBROUTINE parser_get_next_line

! **************************************************************************************************
!> \brief   Skips the whitespaces
!> \param parser ...
!> \date    02.03.2001
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_skip_space(parser)
      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_skip_space', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",490)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",491)
      ! Variable input string length (automatic search)

      ! Check for EOF
      IF (parser%icol == -1) THEN
         parser%icol1 = 1
         parser%icol2 = -1
         RETURN
      END IF

      ! Search for the beginning of the next input string
      DO
         ! Increment the column counter
         parser%icol = parser%icol+1

         ! Quick return, if the end of line is found
         IF ((parser%icol > LEN_TRIM(parser%input_line)) .OR. &
             is_comment(parser, parser%input_line(parser%icol:parser%icol))) THEN
            parser%icol1 = 1
            parser%icol2 = -1
            RETURN
         END IF

         ! Check for input line continuation
         ! Ignore all white space and accept only one
         ! separator token or a string in quotation marks
         IF (.NOT. is_whitespace(parser%input_line(parser%icol:parser%icol))) THEN

            IF (parser%input_line(parser%icol:parser%icol) == parser%continuation_character) THEN
               DO i = LEN(parser%input_line), 1, -1
                  IF (.NOT. is_whitespace(parser%input_line(i:i))) EXIT
               END DO
               IF (parser%icol == i .OR. &
                   is_comment(parser, parser%input_line(i:i))) THEN
                  CALL parser_get_next_line(parser, 1)
                  ! does not pass at_end, it is an error to continue to EOF, change this behavoiur?
                  CYCLE
               ELSE
                  parser%icol1 = i
                  parser%icol2 = i
                  CALL cp_abort(cp__l("input/cp_parser_methods.F",531), &
                                "Found non-blank tokens after the line continuation character '"// &
                                parser%continuation_character//"' "// &
                                TRIM(parser_location(parser)))
                  EXIT
               END IF
            ELSE
               parser%icol = parser%icol-1
               parser%icol1 = parser%icol
               parser%icol2 = parser%icol
               EXIT
            END IF
         END IF
      END DO

   END SUBROUTINE parser_skip_space

! **************************************************************************************************
!> \brief   Get the next input string from the input line.
!> \param parser ...
!> \param string_length ...
!> \date    19.02.2001
!> \author  Matthias Krack (MK)
!> \version 1.0
!> \notes   -) this function MUST be private in this module!
! **************************************************************************************************
   SUBROUTINE parser_next_token(parser, string_length)

      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(IN), OPTIONAL                      :: string_length

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_next_token', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, len_inputline, len_trim_inputline, &
                                                            length

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",568)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",569)
      IF (PRESENT(string_length)) THEN
         length = MIN(string_length, max_line_length)
      ELSE
         length = 0
      END IF

      IF (length > 0) THEN
         ! Fixed input string length
         IF (parser%icol == -1) &
            CALL cp__b("input/cp_parser_methods.F",579,"Unexpectetly reached EOF"//TRIM(parser_location(parser)))

         length = MIN(LEN_TRIM(parser%input_line)-parser%icol1+1, length)
         parser%icol1 = parser%icol+1
         parser%icol2 = parser%icol+length
         parser%icol = parser%icol2
      ELSE

         ! Variable input string length (automatic search)
         ! Check for EOF
         IF (parser%icol == -1) THEN
            parser%icol1 = 1
            parser%icol2 = -1
            RETURN
         END IF

         ! Search for the beginning of the next input string
         ! we precompute those, they are expensive to get
         len_inputline = LEN(parser%input_line)
         len_trim_inputline = LEN_TRIM(parser%input_line)

         DO

            ! Increment the column counter
            parser%icol = parser%icol+1

            ! Quick return, if the end of line is found
            IF ((parser%icol > len_trim_inputline) .OR. &
                is_comment(parser, parser%input_line(parser%icol:parser%icol))) THEN
               parser%icol1 = 1
               parser%icol2 = -1
               RETURN
            END IF

            ! Check for input line continuation
            IF (parser%input_line(parser%icol:parser%icol) == &
                parser%continuation_character) THEN
               DO i = LEN(parser%input_line), 1, -1
                  IF (.NOT. is_whitespace(parser%input_line(i:i))) EXIT
               END DO
               IF (parser%icol == i .OR. &
                   is_comment(parser, parser%input_line(i:i))) THEN
                  CALL parser_get_next_line(parser, 1)
                  ! does not pass at_end, it is an error to continue to EOF, change this behavoiur?
                  CYCLE
               ELSE
                  parser%icol1 = i
                  parser%icol2 = i
                  CALL cp_abort(cp__l("input/cp_parser_methods.F",627), &
                                "Found non-blank tokens after the line continuation character '"// &
                                parser%continuation_character//"' "//TRIM(parser_location(parser)))
                  EXIT
               END IF
            END IF
            ! Ignore all white space and accept only one
            ! separator token or a string in quotation marks
            IF (is_whitespace(parser%input_line(parser%icol:parser%icol))) THEN
               CYCLE
            ELSE IF (INDEX(parser%separators, parser%input_line(parser%icol:parser%icol)) > 0) THEN
               IF (parser%first_separator) THEN
                  parser%first_separator = .FALSE.
                  CYCLE
               ELSE
                  parser%icol1 = parser%icol
                  parser%icol2 = parser%icol
                  CALL cp_abort(cp__l("input/cp_parser_methods.F",644), &
                                "Unexpected separator token <"//parser%input_line(parser%icol:parser%icol)// &
                                "> found"//TRIM(parser_location(parser)))
                  EXIT
               END IF
            ELSE IF (parser%input_line(parser%icol:parser%icol) == '"') THEN
               ! identify quotes and preserve them through-out the execution..
               parser%first_separator = .TRUE.
               parser%icol1 = parser%icol+1
               parser%icol2 = parser%icol+INDEX(parser%input_line(parser%icol1:), '"')
               IF (parser%icol2 == parser%icol) THEN
                  parser%icol1 = parser%icol
                  parser%icol2 = parser%icol
                  CALL cp__b("input/cp_parser_methods.F",657,"Unmatched quotation mark found"//TRIM(parser_location(parser)))
               ELSE
                  parser%icol1 = parser%icol1-1
                  parser%icol = parser%icol2
                  RETURN
               END IF
            ELSE
               parser%first_separator = .TRUE.
               parser%icol1 = parser%icol
               EXIT
            END IF

         END DO

         ! Search for the end of the next input string
         DO
            IF (parser%icol > len_trim_inputline .OR. parser%icol == len_inputline) EXIT
            parser%icol = parser%icol+1
            IF ((parser%icol > len_trim_inputline) .OR. &
                is_whitespace(parser%input_line(parser%icol:parser%icol)) .OR. &
                is_comment(parser, parser%input_line(parser%icol:parser%icol)) .OR. &
                (parser%input_line(parser%icol:parser%icol) == parser%continuation_character)) THEN
               EXIT
            ELSE IF (INDEX(parser%separators, parser%input_line(parser%icol:parser%icol)) > 0) THEN
               parser%first_separator = .FALSE.
               EXIT
            END IF
         END DO

         parser%icol2 = parser%icol-1

         IF (parser%input_line(parser%icol:parser%icol) == &
             parser%continuation_character) parser%icol = parser%icol2

      END IF

   END SUBROUTINE parser_next_token

! **************************************************************************************************
!> \brief   Test next input object.
!>           -  test_result : "EOL": End of line
!>           -  test_result : "EOS": End of section
!>           -  test_result : "FLT": Floating point number
!>           -  test_result : "INT": Integer number
!>           -  test_result : "STR": String
!> \param parser ...
!> \param string_length ...
!> \retval test_result ...
!> \date    23.11.1999
!> \author  Matthias Krack (MK)
!> \note - 08.2008 [tlaino] - Teodoro Laino UZH : updated for buffer
!>          - Major rewrite to parse also (multiple) products of integer or
!>            floating point numbers (23.11.2012,MK)
! **************************************************************************************************
   FUNCTION parser_test_next_token(parser, string_length) RESULT(test_result)

      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(IN), OPTIONAL                      :: string_length
      CHARACTER(LEN=3)                                   :: test_result

      CHARACTER(len=*), PARAMETER :: routineN = 'parser_test_next_token', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=max_line_length)                     :: error_message, string
      INTEGER                                            :: iz, n
      LOGICAL                                            :: ilist_in_use
      REAL(KIND=dp)                                      :: fz

      test_result = ""
      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",726)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",727)

      ! Store current status
      CALL parser_store_status(parser)

      ! Handle possible list of integers
      ilist_in_use = parser%ilist%in_use .AND. (parser%ilist%ipresent < parser%ilist%iend)
      IF (ilist_in_use) THEN
         test_result = "INT"
         CALL parser_retrieve_status(parser)
         RETURN
      END IF

      ! Otherwise continue normally
      CALL parser_next_token(parser, string_length=string_length)

      ! End of line
      IF (parser%icol1 > parser%icol2) THEN
         test_result = "EOL"
         CALL parser_retrieve_status(parser)
         RETURN
      END IF

      string = parser%input_line(parser%icol1:parser%icol2)
      n = LEN_TRIM(string)

      IF (n == 0) THEN
         test_result = "STR"
         CALL parser_retrieve_status(parser)
         RETURN
      END IF

      ! Check for end section string
      IF (string(1:n) == parser%end_section) THEN
         test_result = "EOS"
         CALL parser_retrieve_status(parser)
         RETURN
      END IF

      ! Check for integer object
      error_message = ""
      CALL read_integer_object(string(1:n), iz, error_message)
      IF (LEN_TRIM(error_message) == 0) THEN
         test_result = "INT"
         CALL parser_retrieve_status(parser)
         RETURN
      END IF

      ! Check for floating point object
      error_message = ""
      CALL read_float_object(string(1:n), fz, error_message)
      IF (LEN_TRIM(error_message) == 0) THEN
         test_result = "FLT"
         CALL parser_retrieve_status(parser)
         RETURN
      END IF

      test_result = "STR"
      CALL parser_retrieve_status(parser)

   END FUNCTION parser_test_next_token

! **************************************************************************************************
!> \brief   Search a string pattern in a file defined by its logical unit
!>          number "unit". A case sensitive search is performed, if
!>          ignore_case is .FALSE..
!>          begin_line: give back the parser at the beginning of the line
!>          matching the search
!> \param parser ...
!> \param string ...
!> \param ignore_case ...
!> \param found ...
!> \param line ...
!> \param begin_line ...
!> \param search_from_begin_of_file ...
!> \date    05.10.1999
!> \author  MK
!> \note 08.2008 [tlaino] - Teodoro Laino UZH : updated for buffer
! **************************************************************************************************
   SUBROUTINE parser_search_string(parser, string, ignore_case, found, line, begin_line, &
                                   search_from_begin_of_file)

      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=*), INTENT(IN)                       :: string
      LOGICAL, INTENT(IN)                                :: ignore_case
      LOGICAL, INTENT(OUT)                               :: found
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: line
      LOGICAL, INTENT(IN), OPTIONAL                      :: begin_line, search_from_begin_of_file

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_search_string', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=LEN(string))                         :: pattern
      CHARACTER(LEN=max_line_length+1)                   :: current_line
      INTEGER                                            :: ipattern
      LOGICAL                                            :: at_end, begin, do_reset

      found = .FALSE.
      begin = .FALSE.
      do_reset = .FALSE.
      IF (PRESENT(begin_line)) begin = begin_line
      IF (PRESENT(search_from_begin_of_file)) do_reset = search_from_begin_of_file
      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",829)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",830)
      IF (PRESENT(line)) line = ""

      ! Search for string pattern
      pattern = string
      IF (ignore_case) CALL uppercase(pattern)
      IF (do_reset) CALL parser_reset(parser)
      DO
         ! This call is buffered.. so should not represent any bottleneck
         CALL parser_get_next_line(parser, 1, at_end=at_end)

         ! Exit loop, if the end of file is reached
         IF (at_end) EXIT

         ! Check the current line for string pattern
         current_line = parser%input_line
         IF (ignore_case) CALL uppercase(current_line)
         ipattern = INDEX(current_line, TRIM(pattern))

         IF (ipattern > 0) THEN
            found = .TRUE.
            parser%icol = ipattern-1
            IF (PRESENT(line)) THEN
               IF (LEN(line) < LEN_TRIM(parser%input_line)) THEN
                  CALL cp_warn(cp__l("input/cp_parser_methods.F",854), &
                               "The returned input line has more than "// &
                               TRIM(ADJUSTL(cp_to_string(LEN(line))))// &
                               " characters and is therefore too long to fit in the "// &
                               "specified variable"// &
                               TRIM(parser_location(parser)))
               END IF
            END IF
            EXIT
         END IF

      END DO

      IF (found) THEN
         IF (begin) parser%icol = 0
      END IF

      IF (found) THEN
         IF (PRESENT(line)) line = parser%input_line
         IF (.NOT. begin) CALL parser_next_token(parser)
      END IF

   END SUBROUTINE parser_search_string

! **************************************************************************************************
!> \brief   Check, if the string object contains an object of type integer.
!> \param string ...
!> \retval contains_integer_object ...
!> \date    22.11.1999
!> \author  Matthias Krack (MK)
!> \version 1.0
!> \note - Introducing the possibility to parse a range of integers INT1..INT2
!>            Teodoro Laino [tlaino] - University of Zurich - 08.2008
!>          - Parse also a product of integer numbers (23.11.2012,MK)
! **************************************************************************************************
   FUNCTION integer_object(string) RESULT(contains_integer_object)

      CHARACTER(LEN=*), INTENT(IN)                       :: string
      LOGICAL                                            :: contains_integer_object

      INTEGER                                            :: i, idots, istar, n

      contains_integer_object = .TRUE.
      n = LEN_TRIM(string)

      IF (n == 0) THEN
         contains_integer_object = .FALSE.
         RETURN
      END IF

      idots = INDEX(string(1:n), "..")
      istar = INDEX(string(1:n), "*")

      IF (idots /= 0) THEN
         contains_integer_object = is_integer(string(1:idots-1)) .AND. &
                                   is_integer(string(idots+2:n))
      ELSE IF (istar /= 0) THEN
         i = 1
         DO WHILE (istar /= 0)
            IF (.NOT. is_integer(string(i:i+istar-2))) THEN
               contains_integer_object = .FALSE.
               RETURN
            END IF
            i = i+istar
            istar = INDEX(string(i:n), "*")
         END DO
         contains_integer_object = is_integer(string(i:n))
      ELSE
         contains_integer_object = is_integer(string(1:n))
      END IF

   END FUNCTION integer_object

! **************************************************************************************************
!> \brief ...
!> \param string ...
!> \retval check ...
! **************************************************************************************************
   FUNCTION is_integer(string) RESULT(check)

      CHARACTER(LEN=*), INTENT(IN)                       :: string
      LOGICAL                                            :: check

      INTEGER                                            :: i, n

      check = .TRUE.
      n = LEN_TRIM(string)

      IF (n == 0) THEN
         check = .FALSE.
         RETURN
      END IF

      IF ((INDEX("+-", string(1:1)) > 0) .AND. (n == 1)) THEN
         check = .FALSE.
         RETURN
      END IF

      IF (INDEX("+-0123456789", string(1:1)) == 0) THEN
         check = .FALSE.
         RETURN
      END IF

      DO i = 2, n
         IF (INDEX("0123456789", string(i:i)) == 0) THEN
            check = .FALSE.
            RETURN
         END IF
      END DO

   END FUNCTION is_integer

! **************************************************************************************************
!> \brief   Read an integer number.
!> \param parser ...
!> \param object ...
!> \param newline ...
!> \param skip_lines ...
!> \param string_length ...
!> \param at_end ...
!> \date    22.11.1999
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_get_integer(parser, object, newline, skip_lines, &
                                 string_length, at_end)

      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(OUT)                               :: object
      LOGICAL, INTENT(IN), OPTIONAL                      :: newline
      INTEGER, INTENT(IN), OPTIONAL                      :: skip_lines, string_length
      LOGICAL, INTENT(out), OPTIONAL                     :: at_end

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_get_integer', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=max_line_length)                     :: error_message
      INTEGER                                            :: nline
      LOGICAL                                            :: my_at_end

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",994)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",995)
      IF (PRESENT(skip_lines)) THEN
         nline = skip_lines
      ELSE
         nline = 0
      END IF

      IF (PRESENT(newline)) THEN
         IF (newline) nline = nline+1
      END IF

      CALL parser_get_next_line(parser, nline, at_end=my_at_end)
      IF (PRESENT(at_end)) THEN
         at_end = my_at_end
         IF (my_at_end) RETURN
      ELSE IF (my_at_end) THEN
         CALL cp__b("input/cp_parser_methods.F",1011,"Unexpected EOF"//TRIM(parser_location(parser)))
      END IF

      IF (parser%ilist%in_use) THEN
         CALL ilist_update(parser%ilist)
      ELSE
         CALL parser_next_token(parser, string_length=string_length)
         IF (parser%icol1 > parser%icol2) THEN
            parser%icol1 = parser%icol
            parser%icol2 = parser%icol
            CALL cp_abort(cp__l("input/cp_parser_methods.F",1021), &
                          "An integer type object was expected, found end of line"// &
                          TRIM(parser_location(parser)))
         END IF
         ! Checks for possible lists of integers
         IF (INDEX(parser%input_line(parser%icol1:parser%icol2), "..") /= 0) THEN
            CALL ilist_setup(parser%ilist, parser%input_line(parser%icol1:parser%icol2))
         END IF
      END IF

      IF (integer_object(parser%input_line(parser%icol1:parser%icol2))) THEN
         IF (parser%ilist%in_use) THEN
            object = parser%ilist%ipresent
            CALL ilist_reset(parser%ilist)
         ELSE
            CALL read_integer_object(parser%input_line(parser%icol1:parser%icol2), object, error_message)
            IF (LEN_TRIM(error_message) > 0) THEN
               CALL cp__b("input/cp_parser_methods.F",1038,TRIM(error_message)//TRIM(parser_location(parser)))
            END IF
         END IF
      ELSE
         CALL cp_abort(cp__l("input/cp_parser_methods.F",1042), &
                       "An integer type object was expected, found <"// &
                       parser%input_line(parser%icol1:parser%icol2)//">"// &
                       TRIM(parser_location(parser)))
      END IF

   END SUBROUTINE parser_get_integer

! **************************************************************************************************
!> \brief   Read a string representing logical object.
!> \param parser ...
!> \param object ...
!> \param newline ...
!> \param skip_lines ...
!> \param string_length ...
!> \param at_end ...
!> \date    01.04.2003
!> \par History
!>      - New version (08.07.2003,MK)
!> \author  FM
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_get_logical(parser, object, newline, skip_lines, &
                                 string_length, at_end)

      TYPE(cp_parser_type), POINTER                      :: parser
      LOGICAL, INTENT(OUT)                               :: object
      LOGICAL, INTENT(IN), OPTIONAL                      :: newline
      INTEGER, INTENT(IN), OPTIONAL                      :: skip_lines, string_length
      LOGICAL, INTENT(out), OPTIONAL                     :: at_end

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_get_logical', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=max_line_length)                     :: input_string
      INTEGER                                            :: input_string_length, nline
      LOGICAL                                            :: my_at_end

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",1080)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",1081)
      IF(.NOT.(.NOT. parser%ilist%in_use))CALL cp__a("input/cp_parser_methods.F",1082)
      IF (PRESENT(skip_lines)) THEN
         nline = skip_lines
      ELSE
         nline = 0
      END IF

      IF (PRESENT(newline)) THEN
         IF (newline) nline = nline+1
      END IF

      CALL parser_get_next_line(parser, nline, at_end=my_at_end)
      IF (PRESENT(at_end)) THEN
         at_end = my_at_end
         IF (my_at_end) RETURN
      ELSE IF (my_at_end) THEN
         CALL cp__b("input/cp_parser_methods.F",1098,"Unexpected EOF"//TRIM(parser_location(parser)))
      END IF

      CALL parser_next_token(parser, string_length=string_length)

      input_string_length = parser%icol2-parser%icol1+1

      IF (input_string_length == 0) THEN
         parser%icol1 = parser%icol
         parser%icol2 = parser%icol
         CALL cp_abort(cp__l("input/cp_parser_methods.F",1108), &
                       "A string representing a logical object was expected, found end of line"// &
                       TRIM(parser_location(parser)))
      ELSE
         input_string = ""
         input_string(:input_string_length) = parser%input_line(parser%icol1:parser%icol2)
      END IF
      CALL uppercase(input_string)

      SELECT CASE (TRIM (input_string))
      CASE ("0", "F", ".F.", "FALSE", ".FALSE.", "N", "NO", "OFF")
         object = .FALSE.
      CASE ("1", "T", ".T.", "TRUE", ".TRUE.", "Y", "YES", "ON")
         object = .TRUE.
      CASE DEFAULT
         CALL cp_abort(cp__l("input/cp_parser_methods.F",1123), &
                       "A string representing a logical object was expected, found <"// &
                       TRIM(input_string)//">"//TRIM(parser_location(parser)))
      END SELECT

   END SUBROUTINE parser_get_logical

! **************************************************************************************************
!> \brief   Read a floating point number.
!> \param parser ...
!> \param object ...
!> \param newline ...
!> \param skip_lines ...
!> \param string_length ...
!> \param at_end ...
!> \date    22.11.1999
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_get_real(parser, object, newline, skip_lines, string_length, &
                              at_end)

      TYPE(cp_parser_type), POINTER                      :: parser
      REAL(KIND=dp), INTENT(OUT)                         :: object
      LOGICAL, INTENT(IN), OPTIONAL                      :: newline
      INTEGER, INTENT(IN), OPTIONAL                      :: skip_lines, string_length
      LOGICAL, INTENT(out), OPTIONAL                     :: at_end

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_get_real', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=max_line_length)                     :: error_message
      INTEGER                                            :: nline
      LOGICAL                                            :: my_at_end

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",1158)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",1159)
      IF(.NOT.(.NOT. parser%ilist%in_use))CALL cp__a("input/cp_parser_methods.F",1160)

      IF (PRESENT(skip_lines)) THEN
         nline = skip_lines
      ELSE
         nline = 0
      END IF

      IF (PRESENT(newline)) THEN
         IF (newline) nline = nline+1
      END IF

      CALL parser_get_next_line(parser, nline, at_end=my_at_end)
      IF (PRESENT(at_end)) THEN
         at_end = my_at_end
         IF (my_at_end) RETURN
      ELSE IF (my_at_end) THEN
         CALL cp__b("input/cp_parser_methods.F",1177,"Unexpected EOF"//TRIM(parser_location(parser)))
      END IF

      CALL parser_next_token(parser, string_length=string_length)

      IF (parser%icol1 > parser%icol2) THEN
         parser%icol1 = parser%icol
         parser%icol2 = parser%icol
         CALL cp_abort(cp__l("input/cp_parser_methods.F",1185), &
                       "A floating point type object was expected, found end of the line"// &
                       TRIM(parser_location(parser)))
      END IF

      ! Possibility to have real numbers described in the input as division between two numbers
      CALL read_float_object(parser%input_line(parser%icol1:parser%icol2), object, error_message)
      IF (LEN_TRIM(error_message) > 0) THEN
         CALL cp__b("input/cp_parser_methods.F",1193,TRIM(error_message)//TRIM(parser_location(parser)))
      END IF

   END SUBROUTINE parser_get_real

! **************************************************************************************************
!> \brief   Read a string.
!> \param parser ...
!> \param object ...
!> \param lower_to_upper ...
!> \param newline ...
!> \param skip_lines ...
!> \param string_length ...
!> \param at_end ...
!> \date    22.11.1999
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_get_string(parser, object, lower_to_upper, newline, skip_lines, &
                                string_length, at_end)

      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=*), INTENT(OUT)                      :: object
      LOGICAL, INTENT(IN), OPTIONAL                      :: lower_to_upper, newline
      INTEGER, INTENT(IN), OPTIONAL                      :: skip_lines, string_length
      LOGICAL, INTENT(out), OPTIONAL                     :: at_end

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_get_string', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: input_string_length, nline
      LOGICAL                                            :: my_at_end

      object = ""
      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_methods.F",1227)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_methods.F",1228)
      IF(.NOT.(.NOT. parser%ilist%in_use))CALL cp__a("input/cp_parser_methods.F",1229)
      IF (PRESENT(skip_lines)) THEN
         nline = skip_lines
      ELSE
         nline = 0
      END IF

      IF (PRESENT(newline)) THEN
         IF (newline) nline = nline+1
      END IF

      CALL parser_get_next_line(parser, nline, at_end=my_at_end)
      IF (PRESENT(at_end)) THEN
         at_end = my_at_end
         IF (my_at_end) RETURN
      ELSE IF (my_at_end) THEN
         CALL cp__b("input/cp_parser_methods.F",1245,"Unexpected EOF"//TRIM(parser_location(parser)))
      END IF

      CALL parser_next_token(parser, string_length)

      input_string_length = parser%icol2-parser%icol1+1

      IF (input_string_length <= 0) THEN
         CALL cp_abort(cp__l("input/cp_parser_methods.F",1253), &
                       "A string type object was expected, found end of line"// &
                       TRIM(parser_location(parser)))
      ELSE IF (input_string_length > LEN(object)) THEN
         CALL cp_warn(cp__l("input/cp_parser_methods.F",1257), &
                      "The input string <"//parser%input_line(parser%icol1:parser%icol2)// &
                      "> has more than "//cp_to_string(LEN(object))// &
                      " characters and is therefore too long to fit in the "// &
                      "specified variable"//TRIM(parser_location(parser)))
         object = parser%input_line(parser%icol1:parser%icol1+LEN(object)-1)
      ELSE
         object(:input_string_length) = parser%input_line(parser%icol1:parser%icol2)
      END IF

      ! Convert lowercase to uppercase, if requested
      IF (PRESENT(lower_to_upper)) THEN
         IF (lower_to_upper) CALL uppercase(object)
      END IF

   END SUBROUTINE parser_get_string

! **************************************************************************************************
!> \brief   Returns a floating point number read from a string including
!>          fraction like z1/z2.
!> \param string ...
!> \param object ...
!> \param error_message ...
!> \date    11.01.2011 (MK)
!> \author  Matthias Krack
!> \version 1.0
!> \note - Parse also multiple products and fractions of floating point
!>            numbers (23.11.2012,MK)
! **************************************************************************************************
   SUBROUTINE read_float_object(string, object, error_message)

      CHARACTER(LEN=*), INTENT(IN)                       :: string
      REAL(KIND=dp), INTENT(OUT)                         :: object
      CHARACTER(LEN=*), INTENT(OUT)                      :: error_message

      INTEGER                                            :: i, iop, islash, istar, istat, n
      LOGICAL                                            :: parsing_done
      REAL(KIND=dp)                                      :: z

      error_message = ""

      i = 1
      iop = 0
      n = LEN_TRIM(string)

      parsing_done = .FALSE.

      DO WHILE (.NOT. parsing_done)
         i = i+iop
         islash = INDEX(string(i:n), "/")
         istar = INDEX(string(i:n), "*")
         IF ((islash == 0) .AND. (istar == 0)) THEN
            ! Last factor found: read it and then exit the loop
            iop = n-i+2
            parsing_done = .TRUE.
         ELSE IF ((islash > 0) .AND. (istar > 0)) THEN
            iop = MIN(islash, istar)
         ELSE IF (islash > 0) THEN
            iop = islash
         ELSE IF (istar > 0) THEN
            iop = istar
         END IF
         READ (UNIT=string(i:i+iop-2), FMT=*, IOSTAT=istat) z
         IF (istat /= 0) THEN
            error_message = "A floating point type object was expected, found <"// &
                            string(i:i+iop-2)//">"
            RETURN
         END IF
         IF (i == 1) THEN
            object = z
         ELSE IF (string(i-1:i-1) == "*") THEN
            object = object*z
         ELSE
            IF (z == 0.0_dp) THEN
               error_message = "Division by zero found <"// &
                               string(i:i+iop-2)//">"
               RETURN
            ELSE
               object = object/z
            END IF
         END IF
      END DO

   END SUBROUTINE read_float_object

! **************************************************************************************************
!> \brief   Returns an integer number read from a string including products of
!>          integer numbers like iz1*iz2*iz3
!> \param string ...
!> \param object ...
!> \param error_message ...
!> \date    23.11.2012 (MK)
!> \author  Matthias Krack
!> \version 1.0
!> \note - Parse also (multiple) products of integer numbers (23.11.2012,MK)
! **************************************************************************************************
   SUBROUTINE read_integer_object(string, object, error_message)

      CHARACTER(LEN=*), INTENT(IN)                       :: string
      INTEGER, INTENT(OUT)                               :: object
      CHARACTER(LEN=*), INTENT(OUT)                      :: error_message

      CHARACTER(LEN=20)                                  :: fmtstr
      INTEGER                                            :: i, iop, istat, iz, n
      LOGICAL                                            :: parsing_done

      error_message = ""

      i = 1
      iop = 0
      n = LEN_TRIM(string)

      parsing_done = .FALSE.

      DO WHILE (.NOT. parsing_done)
         i = i+iop
         iop = INDEX(string(i:n), "*")
         IF (iop == 0) THEN
            ! Last factor found: read it and then exit the loop
            iop = n-i+2
            parsing_done = .TRUE.
         END IF
         IF (iop-1 > 0) THEN
            ! need an explicit fmtstr here. With 'FMT=*' compilers from intel and pgi will also
            ! read float numbers as integers, without setting istat non-zero, i.e. string="0.3", istat=0, iz=0
            ! this leads to wrong CP2K results (e.g. parsing force fields).
            WRITE (fmtstr, FMT='(A,I0,A)') '(I', iop-1, ')'
            READ (UNIT=string(i:i+iop-2), FMT=fmtstr, IOSTAT=istat) iz
         ELSE
            istat = 1
         ENDIF
         IF (istat /= 0) THEN
            error_message = "An integer type object was expected, found <"// &
                            string(i:i+iop-2)//">"
            RETURN
         END IF
         IF (i == 1) THEN
            object = iz
         ELSE
            object = object*iz
         END IF
      END DO

   END SUBROUTINE read_integer_object

END MODULE cp_parser_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the incomplete Gamma function F_n(t) for multi-center
!>      integrals over Cartesian Gaussian functions.
!> \par History
!>      - restructured and cleaned (24.05.2004,MK)
!> \author Matthias Krack (07.01.1999)
! **************************************************************************************************
MODULE gamma

   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE mathconstants,                   ONLY: ifac,&
                                              pi
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gamma'
   REAL(KIND=dp), PARAMETER  :: teps = 1.0E-13_dp

! *** Maximum n value of the tabulated F_n(t) function values ***

   INTEGER, SAVE :: current_nmax = -1

! *** F_n(t) table ***

   REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE, SAVE :: ftable

! *** Public subroutines ***

   PUBLIC :: deallocate_md_ftable, &
             fgamma_0, &
             fgamma_ref, &
             init_md_ftable

CONTAINS

! **************************************************************************************************
!> \brief   Build a table of F_n(t) values in the range tmin <= t <= tmax
!>          with a stepsize of tdelta up to n equal to nmax.
!> \param nmax ...
!> \param tmin ...
!> \param tmax ...
!> \param tdelta ...
!> \date    11.01.1999
!> \par Parameters
!>       - nmax  : Maximum n value of F_n(t).
!>       - tdelta: Difference between two consecutive t abcissas (step size).
!>       - tmax  : Maximum t value.
!>       - tmin  : Minimum t value.
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE create_md_ftable(nmax, tmin, tmax, tdelta)

      INTEGER, INTENT(IN)                                :: nmax
      REAL(KIND=dp), INTENT(IN)                          :: tmin, tmax, tdelta

      CHARACTER(len=*), PARAMETER :: routineN = 'create_md_ftable', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: istat, itab, itabmax, itabmin, n
      REAL(KIND=dp)                                      :: t

      IF (current_nmax > -1) THEN
         CALL cp_abort(cp__l("common/gamma.F",76), &
                       "An incomplete Gamma function table is already "// &
                       " allocated. Use the init routine for an update")
      END IF

      IF (nmax < 0) THEN
         CALL cp_abort(cp__l("common/gamma.F",82), &
                       "A negative n value for the initialization of the "// &
                       "incomplete Gamma function is invalid")
      END IF

!   *** Check arguments ***

      IF ((tmax < 0.0_dp) .OR. &
          (tmin < 0.0_dp) .OR. &
          (tdelta <= 0.0_dp) .OR. &
          (tmin > tmax)) THEN
         CALL cp__b("common/gamma.F",93,"Invalid arguments")
      END IF

      n = nmax+6

      itabmin = FLOOR(tmin/tdelta)
      itabmax = CEILING((tmax-tmin)/tdelta)

      ALLOCATE (ftable(0:n, itabmin:itabmax), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 103, "ftable", &
                          (n+1)*(itabmax-itabmin+1)*dp_size)
      END IF
      ftable = 0.0_dp

!   *** Fill table ***

      DO itab = itabmin, itabmax
         t = REAL(itab, dp)*tdelta
         ftable(0:n, itab) = fgamma_ref(n, t)
      END DO

!   *** Save initialization status ***

      current_nmax = nmax

   END SUBROUTINE create_md_ftable

! **************************************************************************************************
!> \brief   Deallocate the table of F_n(t) values.
!> \date    24.05.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_md_ftable()

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_md_ftable', &
         routineP = moduleN//':'//routineN

      IF (current_nmax > -1) THEN

         DEALLOCATE (ftable)

         current_nmax = -1

      END IF

   END SUBROUTINE deallocate_md_ftable

! **************************************************************************************************
!> \brief   Calculation of the incomplete Gamma function F(t) for multicenter
!>          integrals over Gaussian functions. f returns a vector with all
!>          F_n(t) values for 0 <= n <= nmax.
!> \param nmax ...
!> \param t ...
!> \param f ...
!> \date    08.01.1999,
!> \par History
!>          09.06.1999, MK : Changed from a FUNCTION to a SUBROUTINE
!> \par Literature
!>       L. E. McMurchie, E. R. Davidson, J. Comp. Phys. 26, 218 (1978)
!> \par Parameters
!>       - f   : The incomplete Gamma function F_n(t).
!>       - nmax: Maximum n value of F_n(t).
!>       - t   : Argument of the incomplete Gamma function.
!>       - kmax: Maximum number of iterations.
!>       - expt: Exponential term in the upward recursion of F_n(t).
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE fgamma_0(nmax, t, f)

      INTEGER, INTENT(IN)                                :: nmax
      REAL(KIND=dp), INTENT(IN)                          :: t
      REAL(KIND=dp), DIMENSION(0:nmax), INTENT(OUT)      :: f

      INTEGER                                            :: itab, k, n
      REAL(KIND=dp)                                      :: expt, g, tdelta, tmp, ttab

!   *** Calculate F(t) ***

      IF (t < teps) THEN

!     *** Special cases: t = 0 ***

         DO n = 0, nmax
            f(n) = 1.0_dp/REAL(2*n+1, dp)
         END DO

      ELSE IF (t <= 12.0_dp) THEN

!     *** 0 < t < 12 -> Taylor expansion ***

         tdelta = 0.1_dp

!     *** Pretabulation of the F_n(t) function ***
!     *** for the Taylor series expansion      ***

         IF (nmax > current_nmax) THEN
            CALL init_md_ftable(nmax)
         END IF

         itab = NINT(t/tdelta)
         ttab = REAL(itab, dp)*tdelta

         f(nmax) = ftable(nmax, itab)

         tmp = 1.0_dp
         DO k = 1, 6
            tmp = tmp*(ttab-t)
            f(nmax) = f(nmax)+ftable(nmax+k, itab)*tmp*ifac(k)
         END DO

         expt = EXP(-t)

!     *** Use the downward recursion relation to ***
!     *** generate the remaining F_n(t) values   ***

         DO n = nmax-1, 0, -1
            f(n) = (2.0_dp*t*f(n+1)+expt)/REAL(2*n+1, dp)
         END DO

      ELSE

!     *** t > 12 ***

         IF (t <= 15.0_dp) THEN

!       *** 12 < t <= 15 -> Four term polynom expansion ***

            g = 0.4999489092_dp-0.2473631686_dp/t+ &
                0.321180909_dp/t**2-0.3811559346_dp/t**3
            f(0) = 0.5_dp*SQRT(pi/t)-g*EXP(-t)/t

         ELSE IF (t <= 18.0_dp) THEN

!       *** 15 < t <= 18 -> Three term polynom expansion ***

            g = 0.4998436875_dp-0.24249438_dp/t+0.24642845_dp/t**2
            f(0) = 0.5_dp*SQRT(pi/t)-g*EXP(-t)/t

         ELSE IF (t <= 24.0_dp) THEN

!       *** 18 < t <= 24 -> Two term polynom expansion ***

            g = 0.499093162_dp-0.2152832_dp/t
            f(0) = 0.5_dp*SQRT(pi/t)-g*EXP(-t)/t

         ELSE IF (t <= 30.0_dp) THEN

!       *** 24 < t <= 30 -> One term polynom expansion ***

            g = 0.49_dp
            f(0) = 0.5_dp*SQRT(pi/t)-g*EXP(-t)/t

         ELSE

!       *** t > 30 -> Asymptotic formula ***

            f(0) = 0.5_dp*SQRT(pi/t)

         END IF

         IF (t > REAL(2*nmax+36, dp)) THEN
            expt = 0.0_dp
         ELSE
            expt = EXP(-t)
         END IF

!     *** Use the upward recursion relation to ***
!     *** generate the remaining F_n(t) values ***

         DO n = 1, nmax
            f(n) = 0.5_dp*(REAL(2*n-1, dp)*f(n-1)-expt)/t
         END DO

      END IF

   END SUBROUTINE fgamma_0

! **************************************************************************************************
!> \brief   Calculation of the incomplete Gamma function F(t) for multicenter
!>          integrals over Gaussian functions. f returns a vector with all
!>          F_n(t) values for 0 <= n <= nmax.
!> \param nmax ...
!> \param t ...
!> \param f ...
!> \date    08.01.1999
!> \par Literature
!>       L. E. McMurchie, E. R. Davidson, J. Comp. Phys. 26, 218 (1978)
!> \par Parameters
!>       - f   : The incomplete Gamma function F_n(t).
!>       - nmax: Maximum n value of F_n(t).
!>       - t   : Argument of the incomplete Gamma function.
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE fgamma_1(nmax, t, f)

      INTEGER, INTENT(IN)                                :: nmax
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: t
      REAL(KIND=dp), DIMENSION(SIZE(t, 1), 0:nmax), &
         INTENT(OUT)                                     :: f

      INTEGER                                            :: i, itab, k, n
      REAL(KIND=dp)                                      :: expt, g, tdelta, tmp, ttab

      DO i = 1, SIZE(t, 1)

!     *** Calculate F(t) ***

         IF (t(i) < teps) THEN

!       *** Special cases: t = 0 ***

            DO n = 0, nmax
               f(i, n) = 1.0_dp/REAL(2*n+1, dp)
            END DO

         ELSE IF (t(i) <= 12.0_dp) THEN

!       *** 0 < t < 12 -> Taylor expansion ***

            tdelta = 0.1_dp

!       *** Pretabulation of the F_n(t) function ***
!       *** for the Taylor series expansion      ***

            IF (nmax > current_nmax) THEN
               CALL init_md_ftable(nmax)
            END IF

            itab = NINT(t(i)/tdelta)
            ttab = REAL(itab, dp)*tdelta

            f(i, nmax) = ftable(nmax, itab)

            tmp = 1.0_dp
            DO k = 1, 6
               tmp = tmp*(ttab-t(i))
               f(i, nmax) = f(i, nmax)+ftable(nmax+k, itab)*tmp*ifac(k)
            END DO

            expt = EXP(-t(i))

!       *** Use the downward recursion relation to ***
!       *** generate the remaining F_n(t) values   ***

            DO n = nmax-1, 0, -1
               f(i, n) = (2.0_dp*t(i)*f(i, n+1)+expt)/REAL(2*n+1, dp)
            END DO

         ELSE

!       *** t > 12 ***

            IF (t(i) <= 15.0_dp) THEN

!         *** 12 < t <= 15 -> Four term polynom expansion ***

               g = 0.4999489092_dp-0.2473631686_dp/t(i)+ &
                   0.321180909_dp/t(i)**2-0.3811559346_dp/t(i)**3
               f(i, 0) = 0.5_dp*SQRT(pi/t(i))-g*EXP(-t(i))/t(i)

            ELSE IF (t(i) <= 18.0_dp) THEN

!         *** 15 < t <= 18 -> Three term polynom expansion ***

               g = 0.4998436875_dp-0.24249438_dp/t(i)+0.24642845_dp/t(i)**2
               f(i, 0) = 0.5_dp*SQRT(pi/t(i))-g*EXP(-t(i))/t(i)

            ELSE IF (t(i) <= 24.0_dp) THEN

!         *** 18 < t <= 24 -> Two term polynom expansion ***

               g = 0.499093162_dp-0.2152832_dp/t(i)
               f(i, 0) = 0.5_dp*SQRT(pi/t(i))-g*EXP(-t(i))/t(i)

            ELSE IF (t(i) <= 30.0_dp) THEN

!         *** 24 < t <= 30 -> One term polynom expansion ***

               g = 0.49_dp
               f(i, 0) = 0.5_dp*SQRT(pi/t(i))-g*EXP(-t(i))/t(i)

            ELSE

!         *** t > 30 -> Asymptotic formula ***

               f(i, 0) = 0.5_dp*SQRT(pi/t(i))

            END IF

            IF (t(i) > REAL(2*nmax+36, dp)) THEN
               expt = 0.0_dp
            ELSE
               expt = EXP(-t(i))
            END IF

!       *** Use the upward recursion relation to ***
!       *** generate the remaining F_n(t) values ***

            DO n = 1, nmax
               f(i, n) = 0.5_dp*(REAL(2*n-1, dp)*f(i, n-1)-expt)/t(i)
            END DO

         END IF

      END DO

   END SUBROUTINE fgamma_1

! **************************************************************************************************
!> \brief   Calculation of the incomplete Gamma function F_n(t) using a
!>          spherical Bessel function expansion. fgamma_ref returns a
!>          vector with all F_n(t) values for 0 <= n <= nmax.
!>          For t values greater than 50 an asymptotic formula is used.
!>          This function is expected to return accurate F_n(t) values
!>          for any combination of n and t, but the calculation is slow
!>          and therefore the function may only be used for a pretabulation
!>          of F_n(t) values or for reference calculations.
!> \param nmax ...
!> \param t ...
!> \retval f ...
!> \date    07.01.1999
!> \par Literature
!>        F. E. Harris, Int. J. Quant. Chem. 23, 1469 (1983)
!> \par Parameters
!>       - expt   : Exponential term in the downward recursion of F_n(t).
!>       - factor : Prefactor of the Bessel function expansion.
!>       - nmax   : Maximum n value of F_n(t).
!>       - p      : Product of the Bessel function quotients.
!>       - r      : Quotients of the Bessel functions.
!>       - sumterm: One term in the sum over products of Bessel functions.
!>       - t      : Argument of the incomplete Gamma function.
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION fgamma_ref(nmax, t) RESULT(f)

      INTEGER, INTENT(IN)                                :: nmax
      REAL(KIND=dp), INTENT(IN)                          :: t
      REAL(KIND=dp), DIMENSION(0:nmax)                   :: f

      CHARACTER(len=*), PARAMETER :: routineN = 'fgamma_ref', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: kmax = 50
      REAL(KIND=dp), PARAMETER                           :: eps = EPSILON(0.0_dp)

      INTEGER                                            :: j, k, n
      REAL(KIND=dp)                                      :: expt, factor, p, sumterm, sumtot, term
      REAL(KIND=dp), DIMENSION(kmax+10)                  :: r

!   ------------------------------------------------------------------
!   *** Initialization ***

      f(:) = 0.0_dp

      IF (t < teps) THEN

!     *** Special case: t = 0 => analytic expression ***

         DO n = 0, nmax
            f(n) = 1.0_dp/REAL(2*n+1, dp)
         END DO

      ELSE IF (t <= 50.0_dp) THEN

!     *** Initialize ratios of Bessel functions ***

         r(kmax+10) = 0.0_dp

         DO j = kmax+9, 1, -1
            r(j) = -t/(REAL(4*j+2, dp)-t*r(j+1))
         END DO

         factor = 2.0_dp*SINH(0.5_dp*t)*EXP(-0.5_dp*t)/t

         DO n = 0, nmax

!       *** Initialize iteration ***

            sumtot = factor/REAL(2*n+1, dp)
            term = 1.0_dp

!       *** Begin the summation and recursion ***

            DO k = 1, kmax

               term = term*REAL(2*n-2*k+1, dp)/REAL(2*n+2*k+1, dp)

!         *** Product of Bessel function quotients ***

               p = 1.0_dp

               DO j = 1, k
                  p = p*r(j)
               END DO

               sumterm = factor*term*p*REAL(2*k+1, dp)/REAL(2*n+1, dp)

               IF (ABS(sumterm) < eps) THEN

!           *** Iteration converged ***

                  EXIT

               ELSE IF (k == kmax) THEN

!           *** No convergence with kmax iterations ***

                  CALL cp__b("common/gamma.F",503,"Maximum number of iterations reached")

               ELSE

!           *** Add the current term to the sum and continue the iteration ***

                  sumtot = sumtot+sumterm

               END IF

            END DO

            f(n) = sumtot

         END DO

      ELSE

!     *** Use asymptotic formula for t > 50 ***

         f(0) = 0.5_dp*SQRT(pi/t)

!     *** Use the upward recursion relation to ***
!     *** generate the remaining F_n(t) values ***

         expt = EXP(-t)

         DO n = 1, nmax
            f(n) = 0.5_dp*(REAL(2*n-1, dp)*f(n-1)-expt)/t
         END DO

      END IF

   END FUNCTION fgamma_ref

! **************************************************************************************************
!> \brief   Initalize a table of F_n(t) values in the range 0 <= t <= 12 with
!>            a stepsize of 0.1 up to n equal to nmax for the Taylor series
!>            expansion used by McMurchie-Davidson (MD).
!> \param nmax ...
!> \date    10.06.1999
!> \par Parameters
!>       - nmax   : Maximum n value of F_n(t).
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_md_ftable(nmax)
      INTEGER, INTENT(IN)                                :: nmax

      CHARACTER(len=*), PARAMETER :: routineN = 'init_md_ftable', routineP = moduleN//':'//routineN

      IF (nmax < 0) THEN
         CALL cp_abort(cp__l("common/gamma.F",555), &
                       "A negative n value for the initialization of the "// &
                       "incomplete Gamma function is invalid")
      END IF

!   *** Check, if the current initialization is sufficient ***

      IF (nmax > current_nmax) THEN

         CALL deallocate_md_ftable()

!     *** Pretabulation of the F_n(t) function ***
!     *** for the Taylor series expansion      ***

         CALL create_md_ftable(nmax, 0.0_dp, 12.0_dp, 0.1_dp)

      END IF

   END SUBROUTINE init_md_ftable

END MODULE gamma
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of Coulomb integrals over Cartesian Gaussian-type functions
!>      (electron repulsion integrals, ERIs).
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      none
!> \par Parameters
!>       - ax,ay,az    : Angular momentum index numbers of orbital a.
!>       - bx,by,bz    : Angular momentum index numbers of orbital b.
!>       - cx,cy,cz    : Angular momentum index numbers of orbital c.
!>       - coset       : Cartesian orbital set pointer.
!>       - dab         : Distance between the atomic centers a and b.
!>       - dac         : Distance between the atomic centers a and c.
!>       - dbc         : Distance between the atomic centers b and c.
!>       - gccc        : Prefactor of the primitive Gaussian function c.
!>       - l{a,b,c}    : Angular momentum quantum number of shell a, b or c.
!>       - l{a,b,c}_max: Maximum angular momentum quantum number of shell a, b or c.
!>       - l{a,b,c}_min: Minimum angular momentum quantum number of shell a, b or c.
!>       - ncoset      : Number of orbitals in a Cartesian orbital set.
!>       - npgf{a,b}   : Degree of contraction of shell a or b.
!>       - rab         : Distance vector between the atomic centers a and b.
!>       - rab2        : Square of the distance between the atomic centers a and b.
!>       - rac         : Distance vector between the atomic centers a and c.
!>       - rac2        : Square of the distance between the atomic centers a and c.
!>       - rbc         : Distance vector between the atomic centers b and c.
!>       - rbc2        : Square of the distance between the atomic centers b and c.
!>       - rpgf{a,b,c} : Radius of the primitive Gaussian-type function a, b or c.
!>       - zet{a,b,c}  : Exponents of the Gaussian-type functions a, b or c.
!>       - zetp        : Reciprocal of the sum of the exponents of orbital a and b.
!>       - zetw        : Reciprocal of the sum of the exponents of orbital a, b and c.
!> \author Matthias Krack (22.08.2000)
! **************************************************************************************************
MODULE ai_coulomb

   USE gamma,                           ONLY: fgamma=>fgamma_0
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_coulomb'
   PRIVATE

   ! *** Public subroutines ***

   PUBLIC :: coulomb2, coulomb2_new

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the primitive two-center Coulomb integrals over
!>          Cartesian Gaussian-type functions.
!> \param la_max ...
!> \param npgfa ...
!> \param zeta ...
!> \param rpgfa ...
!> \param la_min ...
!> \param lc_max ...
!> \param npgfc ...
!> \param zetc ...
!> \param rpgfc ...
!> \param lc_min ...
!> \param rac ...
!> \param rac2 ...
!> \param vac ...
!> \param v ...
!> \param f ...
!> \param maxder ...
!> \param vac_plus ...
!> \date    05.12.2000
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE coulomb2(la_max, npgfa, zeta, rpgfa, la_min, lc_max, npgfc, zetc, rpgfc, lc_min, &
                       rac, rac2, vac, v, f, maxder, vac_plus)
      INTEGER, INTENT(IN)                                :: la_max, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta, rpgfa
      INTEGER, INTENT(IN)                                :: la_min, lc_max, npgfc
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetc, rpgfc
      INTEGER, INTENT(IN)                                :: lc_min
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac
      REAL(KIND=dp), INTENT(IN)                          :: rac2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: vac
      REAL(KIND=dp), DIMENSION(:, :, :)                  :: v
      REAL(KIND=dp), DIMENSION(0:)                       :: f
      INTEGER, INTENT(IN), OPTIONAL                      :: maxder
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: vac_plus

      CHARACTER(len=*), PARAMETER :: routineN = 'coulomb2', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, coc, cocx, cocy, cocz, cx, &
                                                            cy, cz, i, ipgf, j, jpgf, la, lc, &
                                                            maxder_local, n, na, nap, nc, nmax
      REAL(KIND=dp)                                      :: dac, f0, f1, f2, f3, f4, f5, f6, fcx, &
                                                            fcy, fcz, rho, t, zetp, zetq, zetw
      REAL(KIND=dp), DIMENSION(3)                        :: raw, rcw

      v = 0.0_dp

      maxder_local = 0
      IF (PRESENT(maxder)) THEN
         maxder_local = maxder
         vac_plus = 0.0_dp
      END IF

      nmax = la_max+lc_max+1

      ! *** Calculate the distance of the centers a and c ***

      dac = SQRT(rac2)

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0
      nap = 0

      DO ipgf = 1, npgfa

         nc = 0

         DO jpgf = 1, npgfc

            ! *** Screening ***

            IF (rpgfa(ipgf)+rpgfc(jpgf) < dac) THEN
               DO j = nc+ncoset(lc_min-1)+1, nc+ncoset(lc_max)
                  DO i = na+ncoset(la_min-1)+1, na+ncoset(la_max-maxder_local)
                     vac(i, j) = 0.0_dp
                  END DO
               END DO
               nc = nc+ncoset(lc_max)
               CYCLE
            END IF

            ! *** Calculate some prefactors ***

            zetp = 1.0_dp/zeta(ipgf)
            zetq = 1.0_dp/zetc(jpgf)
            zetw = 1.0_dp/(zeta(ipgf)+zetc(jpgf))

            rho = zeta(ipgf)*zetc(jpgf)*zetw

            f0 = 2.0_dp*SQRT(pi**5*zetw)*zetp*zetq

            ! *** Calculate the incomplete Gamma function ***

            t = rho*rac2

            CALL fgamma(nmax-1, t, f)

            ! *** Calculate the basic two-center Coulomb integrals [s||s]{n} ***

            DO n = 1, nmax
               v(1, 1, n) = f0*f(n-1)
            END DO

            ! *** Vertical recurrence steps: [s||s] -> [s||c] ***

            IF (lc_max > 0) THEN

               f1 = 0.5_dp*zetq
               f2 = -rho*zetq

               rcw(:) = -zeta(ipgf)*zetw*rac(:)

               ! *** [s||p]{n} = (Wi - Ci)*[s||s]{n+1}  (i = x,y,z) ***

               DO n = 1, nmax-1
                  v(1, 2, n) = rcw(1)*v(1, 1, n+1)
                  v(1, 3, n) = rcw(2)*v(1, 1, n+1)
                  v(1, 4, n) = rcw(3)*v(1, 1, n+1)
               END DO

               ! ** [s||c]{n} = (Wi - Ci)*[s||c-1i]{n+1} +     ***
               ! **             f1*Ni(c-1i)*(   [s||c-2i]{n} + ***
               ! **                          f2*[s||c-2i]{n+1} ***

               DO lc = 2, lc_max

                  DO n = 1, nmax-lc

                     ! **** Increase the angular momentum component z of c ***

                     v(1, coset(0, 0, lc), n) = &
                        rcw(3)*v(1, coset(0, 0, lc-1), n+1)+ &
                        f1*REAL(lc-1, dp)*(v(1, coset(0, 0, lc-2), n)+ &
                                           f2*v(1, coset(0, 0, lc-2), n+1))

                     ! *** Increase the angular momentum component y of c ***

                     cz = lc-1
                     v(1, coset(0, 1, cz), n) = rcw(2)*v(1, coset(0, 0, cz), n+1)

                     DO cy = 2, lc
                        cz = lc-cy
                        v(1, coset(0, cy, cz), n) = &
                           rcw(2)*v(1, coset(0, cy-1, cz), n+1)+ &
                           f1*REAL(cy-1, dp)*(v(1, coset(0, cy-2, cz), n)+ &
                                              f2*v(1, coset(0, cy-2, cz), n+1))
                     END DO

                     ! *** Increase the angular momentum component x of c ***

                     DO cy = 0, lc-1
                        cz = lc-1-cy
                        v(1, coset(1, cy, cz), n) = rcw(1)*v(1, coset(0, cy, cz), n+1)
                     END DO

                     DO cx = 2, lc
                        f6 = f1*REAL(cx-1, dp)
                        DO cy = 0, lc-cx
                           cz = lc-cx-cy
                           v(1, coset(cx, cy, cz), n) = &
                              rcw(1)*v(1, coset(cx-1, cy, cz), n+1)+ &
                              f6*(v(1, coset(cx-2, cy, cz), n)+ &
                                  f2*v(1, coset(cx-2, cy, cz), n+1))
                        END DO
                     END DO

                  END DO

               END DO

            END IF

            ! *** Vertical recurrence steps: [s||c] -> [a||c] ***

            IF (la_max > 0) THEN

               f3 = 0.5_dp*zetp
               f4 = -rho*zetp
               f5 = 0.5_dp*zetw

               raw(:) = zetc(jpgf)*zetw*rac(:)

               ! *** [p||s]{n} = (Wi - Ai)*[s||s]{n+1}  (i = x,y,z) ***

               DO n = 1, nmax-1
                  v(2, 1, n) = raw(1)*v(1, 1, n+1)
                  v(3, 1, n) = raw(2)*v(1, 1, n+1)
                  v(4, 1, n) = raw(3)*v(1, 1, n+1)
               END DO

               ! *** [a||s]{n} = (Wi - Ai)*[a-1i||s]{n+1} +      ***
               ! ***             f3*Ni(a-1i)*(   [a-2i||s]{n} +  ***
               ! ***                          f4*[a-2i||s]{n+1}) ***

               DO la = 2, la_max

                  DO n = 1, nmax-la

                     ! *** Increase the angular momentum component z of a ***

                     v(coset(0, 0, la), 1, n) = &
                        raw(3)*v(coset(0, 0, la-1), 1, n+1)+ &
                        f3*REAL(la-1, dp)*(v(coset(0, 0, la-2), 1, n)+ &
                                           f4*v(coset(0, 0, la-2), 1, n+1))

                     ! *** Increase the angular momentum component y of a ***

                     az = la-1
                     v(coset(0, 1, az), 1, n) = raw(2)*v(coset(0, 0, az), 1, n+1)

                     DO ay = 2, la
                        az = la-ay
                        v(coset(0, ay, az), 1, n) = &
                           raw(2)*v(coset(0, ay-1, az), 1, n+1)+ &
                           f3*REAL(ay-1, dp)*(v(coset(0, ay-2, az), 1, n)+ &
                                              f4*v(coset(0, ay-2, az), 1, n+1))
                     END DO

                     ! *** Increase the angular momentum component x of a ***

                     DO ay = 0, la-1
                        az = la-1-ay
                        v(coset(1, ay, az), 1, n) = raw(1)*v(coset(0, ay, az), 1, n+1)
                     END DO

                     DO ax = 2, la
                        f6 = f3*REAL(ax-1, dp)
                        DO ay = 0, la-ax
                           az = la-ax-ay
                           v(coset(ax, ay, az), 1, n) = &
                              raw(1)*v(coset(ax-1, ay, az), 1, n+1)+ &
                              f6*(v(coset(ax-2, ay, az), 1, n)+ &
                                  f4*v(coset(ax-2, ay, az), 1, n+1))
                        END DO
                     END DO

                  END DO

               END DO

               DO lc = 1, lc_max

                  DO cx = 0, lc
                     DO cy = 0, lc-cx
                        cz = lc-cx-cy

                        coc = coset(cx, cy, cz)
                        cocx = coset(MAX(0, cx-1), cy, cz)
                        cocy = coset(cx, MAX(0, cy-1), cz)
                        cocz = coset(cx, cy, MAX(0, cz-1))

                        fcx = f5*REAL(cx, dp)
                        fcy = f5*REAL(cy, dp)
                        fcz = f5*REAL(cz, dp)

                        ! *** [p||c]{n} = (Wi - Ai)*[s||c]{n+1} + ***
                        ! ***             f5*Ni(c)*[s||c-1i]{n+1} ***

                        DO n = 1, nmax-1-lc
                           v(2, coc, n) = raw(1)*v(1, coc, n+1)+fcx*v(1, cocx, n+1)
                           v(3, coc, n) = raw(2)*v(1, coc, n+1)+fcy*v(1, cocy, n+1)
                           v(4, coc, n) = raw(3)*v(1, coc, n+1)+fcz*v(1, cocz, n+1)
                        END DO

                        ! *** [a||c]{n} = (Wi - Ai)*[a-1i||c]{n+1} +        ***
                        ! ***             f3*Ni(a-1i)*(   [a-2i||c]{n} +    ***
                        ! ***                          f4*[a-2i||c]{n+1}) + ***
                        ! ***             f5*Ni(c)*[a-1i||c-1i]{n+1}        ***

                        DO la = 2, la_max

                           DO n = 1, nmax-la-lc

                              ! *** Increase the angular momentum component z of a ***

                              v(coset(0, 0, la), coc, n) = &
                                 raw(3)*v(coset(0, 0, la-1), coc, n+1)+ &
                                 f3*REAL(la-1, dp)*(v(coset(0, 0, la-2), coc, n)+ &
                                                    f4*v(coset(0, 0, la-2), coc, n+1))+ &
                                 fcz*v(coset(0, 0, la-1), cocz, n+1)

                              ! *** Increase the angular momentum component y of a ***

                              az = la-1
                              v(coset(0, 1, az), coc, n) = &
                                 raw(2)*v(coset(0, 0, az), coc, n+1)+ &
                                 fcy*v(coset(0, 0, az), cocy, n+1)

                              DO ay = 2, la
                                 az = la-ay
                                 v(coset(0, ay, az), coc, n) = &
                                    raw(2)*v(coset(0, ay-1, az), coc, n+1)+ &
                                    f3*REAL(ay-1, dp)*(v(coset(0, ay-2, az), coc, n)+ &
                                                       f4*v(coset(0, ay-2, az), coc, n+1))+ &
                                    fcy*v(coset(0, ay-1, az), cocy, n+1)
                              END DO

                              ! *** Increase the angular momentum component x of a ***

                              DO ay = 0, la-1
                                 az = la-1-ay
                                 v(coset(1, ay, az), coc, n) = &
                                    raw(1)*v(coset(0, ay, az), coc, n+1)+ &
                                    fcx*v(coset(0, ay, az), cocx, n+1)
                              END DO

                              DO ax = 2, la
                                 f6 = f3*REAL(ax-1, dp)
                                 DO ay = 0, la-ax
                                    az = la-ax-ay
                                    v(coset(ax, ay, az), coc, n) = &
                                       raw(1)*v(coset(ax-1, ay, az), coc, n+1)+ &
                                       f6*(v(coset(ax-2, ay, az), coc, n)+ &
                                           f4*v(coset(ax-2, ay, az), coc, n+1))+ &
                                       fcx*v(coset(ax-1, ay, az), cocx, n+1)
                                 END DO
                              END DO

                           END DO

                        END DO

                     END DO
                  END DO

               END DO

            END IF

            DO j = ncoset(lc_min-1)+1, ncoset(lc_max)
               DO i = ncoset(la_min-1)+1, ncoset(la_max-maxder_local)
                  vac(na+i, nc+j) = v(i, j, 1)
               END DO
            END DO

            IF (PRESENT(maxder)) THEN
               DO j = ncoset(lc_min-1)+1, ncoset(lc_max)
                  DO i = 1, ncoset(la_max)
                     vac_plus(nap+i, nc+j) = v(i, j, 1)
                  END DO
               END DO
            END IF

            nc = nc+ncoset(lc_max)

         END DO

         na = na+ncoset(la_max-maxder_local)
         nap = nap+ncoset(la_max)

      END DO

   END SUBROUTINE coulomb2
! **************************************************************************************************
!> \brief   Calculation of the primitive two-center Coulomb integrals over
!>          Cartesian Gaussian-type functions. Same as coulomb2 treats derivative
!>          different (now vac_plus is symmetric)
!> \param la_max ...
!> \param npgfa ...
!> \param zeta ...
!> \param la_min ...
!> \param lc_max ...
!> \param npgfc ...
!> \param zetc ...
!> \param lc_min ...
!> \param rac ...
!> \param rac2 ...
!> \param vac ...
!> \param v ...
!> \param f ...
!> \param maxder ...
!> \param vac_plus ...
!> \date    6.02.2008
!> \author  CJM
!> \version 1.0
! **************************************************************************************************

   SUBROUTINE coulomb2_new(la_max, npgfa, zeta, la_min, lc_max, npgfc, zetc, lc_min, &
                           rac, rac2, vac, v, f, maxder, vac_plus)
      INTEGER, INTENT(IN)                                :: la_max, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta
      INTEGER, INTENT(IN)                                :: la_min, lc_max, npgfc
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetc
      INTEGER, INTENT(IN)                                :: lc_min
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac
      REAL(KIND=dp), INTENT(IN)                          :: rac2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: vac
      REAL(KIND=dp), DIMENSION(:, :, :)                  :: v
      REAL(KIND=dp), DIMENSION(0:)                       :: f
      INTEGER, INTENT(IN), OPTIONAL                      :: maxder
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: vac_plus

      CHARACTER(len=*), PARAMETER :: routineN = 'coulomb2_new', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, coc, cocx, cocy, cocz, cx, &
                                                            cy, cz, i, ipgf, j, jpgf, la, lc, &
                                                            maxder_local, n, na, nap, nc, ncp, nmax
      REAL(KIND=dp)                                      :: dac, f0, f1, f2, f3, f4, f5, f6, fcx, &
                                                            fcy, fcz, rho, t, zetp, zetq, zetw
      REAL(KIND=dp), DIMENSION(3)                        :: raw, rcw

      v = 0.0_dp

      maxder_local = 0
      IF (PRESENT(maxder)) THEN
         maxder_local = maxder
         vac_plus = 0.0_dp
      END IF

      nmax = la_max+lc_max+1

      ! *** Calculate the distance of the centers a and c ***

      dac = SQRT(rac2)

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0
      nap = 0

      DO ipgf = 1, npgfa

         nc = 0
         ncp = 0

         DO jpgf = 1, npgfc

            ! *** Calculate some prefactors ***

            zetp = 1.0_dp/zeta(ipgf)
            zetq = 1.0_dp/zetc(jpgf)
            zetw = 1.0_dp/(zeta(ipgf)+zetc(jpgf))

            rho = zeta(ipgf)*zetc(jpgf)*zetw

            f0 = 2.0_dp*SQRT(pi**5*zetw)*zetp*zetq

            ! *** Calculate the incomplete Gamma function ***

            t = rho*rac2

            CALL fgamma(nmax-1, t, f)

            ! *** Calculate the basic two-center Coulomb integrals [s||s]{n} ***

            DO n = 1, nmax
               v(1, 1, n) = f0*f(n-1)
            END DO

            ! *** Vertical recurrence steps: [s||s] -> [s||c] ***

            IF (lc_max > 0) THEN

               f1 = 0.5_dp*zetq
               f2 = -rho*zetq

               rcw(:) = -zeta(ipgf)*zetw*rac(:)

               ! *** [s||p]{n} = (Wi - Ci)*[s||s]{n+1}  (i = x,y,z) ***

               DO n = 1, nmax-1
                  v(1, 2, n) = rcw(1)*v(1, 1, n+1)
                  v(1, 3, n) = rcw(2)*v(1, 1, n+1)
                  v(1, 4, n) = rcw(3)*v(1, 1, n+1)
               END DO

               ! ** [s||c]{n} = (Wi - Ci)*[s||c-1i]{n+1} +     ***
               ! **             f1*Ni(c-1i)*(   [s||c-2i]{n} + ***
               ! **                          f2*[s||c-2i]{n+1} ***

               DO lc = 2, lc_max

                  DO n = 1, nmax-lc

                     ! **** Increase the angular momentum component z of c ***

                     v(1, coset(0, 0, lc), n) = &
                        rcw(3)*v(1, coset(0, 0, lc-1), n+1)+ &
                        f1*REAL(lc-1, dp)*(v(1, coset(0, 0, lc-2), n)+ &
                                           f2*v(1, coset(0, 0, lc-2), n+1))

                     ! *** Increase the angular momentum component y of c ***

                     cz = lc-1
                     v(1, coset(0, 1, cz), n) = rcw(2)*v(1, coset(0, 0, cz), n+1)

                     DO cy = 2, lc
                        cz = lc-cy
                        v(1, coset(0, cy, cz), n) = &
                           rcw(2)*v(1, coset(0, cy-1, cz), n+1)+ &
                           f1*REAL(cy-1, dp)*(v(1, coset(0, cy-2, cz), n)+ &
                                              f2*v(1, coset(0, cy-2, cz), n+1))
                     END DO

                     ! *** Increase the angular momentum component x of c ***

                     DO cy = 0, lc-1
                        cz = lc-1-cy
                        v(1, coset(1, cy, cz), n) = rcw(1)*v(1, coset(0, cy, cz), n+1)
                     END DO

                     DO cx = 2, lc
                        f6 = f1*REAL(cx-1, dp)
                        DO cy = 0, lc-cx
                           cz = lc-cx-cy
                           v(1, coset(cx, cy, cz), n) = &
                              rcw(1)*v(1, coset(cx-1, cy, cz), n+1)+ &
                              f6*(v(1, coset(cx-2, cy, cz), n)+ &
                                  f2*v(1, coset(cx-2, cy, cz), n+1))
                        END DO
                     END DO

                  END DO

               END DO

            END IF

            ! *** Vertical recurrence steps: [s||c] -> [a||c] ***

            IF (la_max > 0) THEN

               f3 = 0.5_dp*zetp
               f4 = -rho*zetp
               f5 = 0.5_dp*zetw

               raw(:) = zetc(jpgf)*zetw*rac(:)

               ! *** [p||s]{n} = (Wi - Ai)*[s||s]{n+1}  (i = x,y,z) ***

               DO n = 1, nmax-1
                  v(2, 1, n) = raw(1)*v(1, 1, n+1)
                  v(3, 1, n) = raw(2)*v(1, 1, n+1)
                  v(4, 1, n) = raw(3)*v(1, 1, n+1)
               END DO

               ! *** [a||s]{n} = (Wi - Ai)*[a-1i||s]{n+1} +      ***
               ! ***             f3*Ni(a-1i)*(   [a-2i||s]{n} +  ***
               ! ***                          f4*[a-2i||s]{n+1}) ***

               DO la = 2, la_max

                  DO n = 1, nmax-la

                     ! *** Increase the angular momentum component z of a ***

                     v(coset(0, 0, la), 1, n) = &
                        raw(3)*v(coset(0, 0, la-1), 1, n+1)+ &
                        f3*REAL(la-1, dp)*(v(coset(0, 0, la-2), 1, n)+ &
                                           f4*v(coset(0, 0, la-2), 1, n+1))

                     ! *** Increase the angular momentum component y of a ***

                     az = la-1
                     v(coset(0, 1, az), 1, n) = raw(2)*v(coset(0, 0, az), 1, n+1)

                     DO ay = 2, la
                        az = la-ay
                        v(coset(0, ay, az), 1, n) = &
                           raw(2)*v(coset(0, ay-1, az), 1, n+1)+ &
                           f3*REAL(ay-1, dp)*(v(coset(0, ay-2, az), 1, n)+ &
                                              f4*v(coset(0, ay-2, az), 1, n+1))
                     END DO

                     ! *** Increase the angular momentum component x of a ***

                     DO ay = 0, la-1
                        az = la-1-ay
                        v(coset(1, ay, az), 1, n) = raw(1)*v(coset(0, ay, az), 1, n+1)
                     END DO

                     DO ax = 2, la
                        f6 = f3*REAL(ax-1, dp)
                        DO ay = 0, la-ax
                           az = la-ax-ay
                           v(coset(ax, ay, az), 1, n) = &
                              raw(1)*v(coset(ax-1, ay, az), 1, n+1)+ &
                              f6*(v(coset(ax-2, ay, az), 1, n)+ &
                                  f4*v(coset(ax-2, ay, az), 1, n+1))
                        END DO
                     END DO

                  END DO

               END DO

               DO lc = 1, lc_max

                  DO cx = 0, lc
                     DO cy = 0, lc-cx
                        cz = lc-cx-cy

                        coc = coset(cx, cy, cz)
                        cocx = coset(MAX(0, cx-1), cy, cz)
                        cocy = coset(cx, MAX(0, cy-1), cz)
                        cocz = coset(cx, cy, MAX(0, cz-1))

                        fcx = f5*REAL(cx, dp)
                        fcy = f5*REAL(cy, dp)
                        fcz = f5*REAL(cz, dp)

                        ! *** [p||c]{n} = (Wi - Ai)*[s||c]{n+1} + ***
                        ! ***             f5*Ni(c)*[s||c-1i]{n+1} ***

                        DO n = 1, nmax-1-lc
                           v(2, coc, n) = raw(1)*v(1, coc, n+1)+fcx*v(1, cocx, n+1)
                           v(3, coc, n) = raw(2)*v(1, coc, n+1)+fcy*v(1, cocy, n+1)
                           v(4, coc, n) = raw(3)*v(1, coc, n+1)+fcz*v(1, cocz, n+1)
                        END DO

                        ! *** [a||c]{n} = (Wi - Ai)*[a-1i||c]{n+1} +        ***
                        ! ***             f3*Ni(a-1i)*(   [a-2i||c]{n} +    ***
                        ! ***                          f4*[a-2i||c]{n+1}) + ***
                        ! ***             f5*Ni(c)*[a-1i||c-1i]{n+1}        ***

                        DO la = 2, la_max

                           DO n = 1, nmax-la-lc

                              ! *** Increase the angular momentum component z of a ***

                              v(coset(0, 0, la), coc, n) = &
                                 raw(3)*v(coset(0, 0, la-1), coc, n+1)+ &
                                 f3*REAL(la-1, dp)*(v(coset(0, 0, la-2), coc, n)+ &
                                                    f4*v(coset(0, 0, la-2), coc, n+1))+ &
                                 fcz*v(coset(0, 0, la-1), cocz, n+1)

                              ! *** Increase the angular momentum component y of a ***

                              az = la-1
                              v(coset(0, 1, az), coc, n) = &
                                 raw(2)*v(coset(0, 0, az), coc, n+1)+ &
                                 fcy*v(coset(0, 0, az), cocy, n+1)

                              DO ay = 2, la
                                 az = la-ay
                                 v(coset(0, ay, az), coc, n) = &
                                    raw(2)*v(coset(0, ay-1, az), coc, n+1)+ &
                                    f3*REAL(ay-1, dp)*(v(coset(0, ay-2, az), coc, n)+ &
                                                       f4*v(coset(0, ay-2, az), coc, n+1))+ &
                                    fcy*v(coset(0, ay-1, az), cocy, n+1)
                              END DO

                              ! *** Increase the angular momentum component x of a ***

                              DO ay = 0, la-1
                                 az = la-1-ay
                                 v(coset(1, ay, az), coc, n) = &
                                    raw(1)*v(coset(0, ay, az), coc, n+1)+ &
                                    fcx*v(coset(0, ay, az), cocx, n+1)
                              END DO

                              DO ax = 2, la
                                 f6 = f3*REAL(ax-1, dp)
                                 DO ay = 0, la-ax
                                    az = la-ax-ay
                                    v(coset(ax, ay, az), coc, n) = &
                                       raw(1)*v(coset(ax-1, ay, az), coc, n+1)+ &
                                       f6*(v(coset(ax-2, ay, az), coc, n)+ &
                                           f4*v(coset(ax-2, ay, az), coc, n+1))+ &
                                       fcx*v(coset(ax-1, ay, az), cocx, n+1)
                                 END DO
                              END DO

                           END DO

                        END DO

                     END DO
                  END DO

               END DO

            END IF

            DO j = ncoset(lc_min-1)+1, ncoset(lc_max-maxder_local)
               DO i = ncoset(la_min-1)+1, ncoset(la_max-maxder_local)
                  vac(na+i, nc+j) = v(i, j, 1)
               END DO
            END DO

            IF (PRESENT(maxder)) THEN
               DO j = 1, ncoset(lc_max)
                  DO i = 1, ncoset(la_max)
                     vac_plus(nap+i, ncp+j) = v(i, j, 1)
                  END DO
               END DO
            END IF

            nc = nc+ncoset(lc_max-maxder_local)
            ncp = ncp+ncoset(lc_max)

         END DO

         na = na+ncoset(la_max-maxder_local)
         nap = nap+ncoset(la_max)

      END DO

   END SUBROUTINE coulomb2_new

! **************************************************************************************************
!> \brief   Calculation of the primitive three-center Coulomb integrals over
!>          Cartesian Gaussian-type functions (electron repulsion integrals,
!>          ERIs).
!> \param la_max ...
!> \param npgfa ...
!> \param zeta ...
!> \param rpgfa ...
!> \param la_min ...
!> \param lb_max ...
!> \param npgfb ...
!> \param zetb ...
!> \param rpgfb ...
!> \param lb_min ...
!> \param lc_max ...
!> \param zetc ...
!> \param rpgfc ...
!> \param lc_min ...
!> \param gccc ...
!> \param rab ...
!> \param rab2 ...
!> \param rac ...
!> \param rac2 ...
!> \param rbc2 ...
!> \param vabc ...
!> \param int_abc ...
!> \param v ...
!> \param f ...
!> \param maxder ...
!> \param vabc_plus ...
!> \date    06.11.2000
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE coulomb3(la_max, npgfa, zeta, rpgfa, la_min, lb_max, npgfb, zetb, rpgfb, lb_min, &
                       lc_max, zetc, rpgfc, lc_min, gccc, rab, rab2, rac, rac2, rbc2, vabc, int_abc, &
                       v, f, maxder, vabc_plus)
      INTEGER, INTENT(IN)                                :: la_max, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta, rpgfa
      INTEGER, INTENT(IN)                                :: la_min, lb_max, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb, rpgfb
      INTEGER, INTENT(IN)                                :: lb_min, lc_max
      REAL(KIND=dp), INTENT(IN)                          :: zetc, rpgfc
      INTEGER, INTENT(IN)                                :: lc_min
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: gccc
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: rab2
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac
      REAL(KIND=dp), INTENT(IN)                          :: rac2, rbc2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: vabc
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(OUT)     :: int_abc
      REAL(KIND=dp), DIMENSION(:, :, :, :)               :: v
      REAL(KIND=dp), DIMENSION(0:)                       :: f
      INTEGER, INTENT(IN), OPTIONAL                      :: maxder
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: vabc_plus

      CHARACTER(len=*), PARAMETER :: routineN = 'coulomb3', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, bx, by, bz, coc, cocx, cocy, &
                                                            cocz, cx, cy, cz, i, ipgf, j, jpgf, k, &
                                                            kk, la, la_start, lb, lc, &
                                                            maxder_local, n, na, nap, nb, nmax
      REAL(KIND=dp)                                      :: dab, dac, dbc, f0, f1, f2, f3, f4, f5, &
                                                            f6, f7, fcx, fcy, fcz, fx, fy, fz, t, &
                                                            zetp, zetq, zetw
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp, rcp, rcw, rpw

      v = 0.0_dp

      maxder_local = 0
      IF (PRESENT(maxder)) THEN
         maxder_local = maxder
      END IF

      nmax = la_max+lb_max+lc_max+1

      ! *** Calculate the distances of the centers a, b and c ***

      dab = SQRT(rab2)
      dac = SQRT(rac2)
      dbc = SQRT(rbc2)

      ! *** Initialize integrals array
      int_abc = 0.0_dp

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0
      nap = 0

      DO ipgf = 1, npgfa

         ! *** Screening ***
         IF (rpgfa(ipgf)+rpgfc < dac) THEN
            na = na+ncoset(la_max-maxder_local)
            nap = nap+ncoset(la_max)
            CYCLE
         END IF

         nb = 0

         DO jpgf = 1, npgfb

            ! *** Screening ***
            IF ( &
               (rpgfb(jpgf)+rpgfc < dbc) .OR. &
               (rpgfa(ipgf)+rpgfb(jpgf) < dab)) THEN
               nb = nb+ncoset(lb_max)
               CYCLE
            END IF

            ! *** Calculate some prefactors ***

            zetp = 1.0_dp/(zeta(ipgf)+zetb(jpgf))
            zetq = 1.0_dp/zetc
            zetw = 1.0_dp/(zeta(ipgf)+zetb(jpgf)+zetc)

            f0 = 2.0_dp*SQRT(pi**5*zetw)*zetp*zetq
            f1 = zetb(jpgf)*zetp
            f2 = 0.5_dp*zetp
            f4 = -zetc*zetw

            f0 = f0*EXP(-zeta(ipgf)*f1*rab2)

            rap(:) = f1*rab(:)
            rcp(:) = rap(:)-rac(:)
            rpw(:) = f4*rcp(:)

            ! *** Calculate the incomplete Gamma function ***

            t = -f4*(rcp(1)*rcp(1)+rcp(2)*rcp(2)+rcp(3)*rcp(3))/zetp

            CALL fgamma(nmax-1, t, f)

            ! *** Calculate the basic three-center Coulomb integrals [ss||s]{n} ***

            DO n = 1, nmax
               v(1, 1, 1, n) = f0*f(n-1)
            END DO

            ! *** Recurrence steps: [ss||s] -> [as||s] ***

            IF (la_max > 0) THEN

               ! *** Vertical recurrence steps: [ss||s] -> [as||s] ***

               ! *** [ps||s]{n} = (Pi - Ai)*[ss||s]{n} +              ***
               ! ***              (Wi - Pi)*[ss||s]{n+1}  (i = x,y,z) ***

               DO n = 1, nmax-1
                  v(2, 1, 1, n) = rap(1)*v(1, 1, 1, n)+rpw(1)*v(1, 1, 1, n+1)
                  v(3, 1, 1, n) = rap(2)*v(1, 1, 1, n)+rpw(2)*v(1, 1, 1, n+1)
                  v(4, 1, 1, n) = rap(3)*v(1, 1, 1, n)+rpw(3)*v(1, 1, 1, n+1)
               END DO

               ! *** [as||s]{n} = (Pi - Ai)*[(a-1i)s||s]{n} +        ***
               ! ***              (Wi - Pi)*[(a-1i)s||s]{n+1} +      ***
               ! ***              f2*Ni(a-1i)*(   [(a-2i)s||s]{n} +  ***
               ! ***                           f4*[(a-2i)s||s]{n+1}) ***

               DO la = 2, la_max

                  DO n = 1, nmax-la

                     ! *** Increase the angular momentum component z of a ***

                     v(coset(0, 0, la), 1, 1, n) = &
                        rap(3)*v(coset(0, 0, la-1), 1, 1, n)+ &
                        rpw(3)*v(coset(0, 0, la-1), 1, 1, n+1)+ &
                        f2*REAL(la-1, dp)*(v(coset(0, 0, la-2), 1, 1, n)+ &
                                           f4*v(coset(0, 0, la-2), 1, 1, n+1))

                     ! *** Increase the angular momentum component y of a ***

                     az = la-1
                     v(coset(0, 1, az), 1, 1, n) = &
                        rap(2)*v(coset(0, 0, az), 1, 1, n)+ &
                        rpw(2)*v(coset(0, 0, az), 1, 1, n+1)

                     DO ay = 2, la
                        az = la-ay
                        v(coset(0, ay, az), 1, 1, n) = &
                           rap(2)*v(coset(0, ay-1, az), 1, 1, n)+ &
                           rpw(2)*v(coset(0, ay-1, az), 1, 1, n+1)+ &
                           f2*REAL(ay-1, dp)*(v(coset(0, ay-2, az), 1, 1, n)+ &
                                              f4*v(coset(0, ay-2, az), 1, 1, n+1))
                     END DO

                     ! *** Increase the angular momentum component x of a ***

                     DO ay = 0, la-1
                        az = la-1-ay
                        v(coset(1, ay, az), 1, 1, n) = &
                           rap(1)*v(coset(0, ay, az), 1, 1, n)+ &
                           rpw(1)*v(coset(0, ay, az), 1, 1, n+1)
                     END DO

                     DO ax = 2, la
                        f3 = f2*REAL(ax-1, dp)
                        DO ay = 0, la-ax
                           az = la-ax-ay
                           v(coset(ax, ay, az), 1, 1, n) = &
                              rap(1)*v(coset(ax-1, ay, az), 1, 1, n)+ &
                              rpw(1)*v(coset(ax-1, ay, az), 1, 1, n+1)+ &
                              f3*(v(coset(ax-2, ay, az), 1, 1, n)+ &
                                  f4*v(coset(ax-2, ay, az), 1, 1, n+1))
                        END DO
                     END DO

                  END DO

               END DO

               ! *** Recurrence steps: [as||s] -> [ab||s] ***

               IF (lb_max > 0) THEN

                  ! *** Horizontal recurrence steps ***

                  rbp(:) = rap(:)-rab(:)

                  ! *** [ap||s]{n} = [(a+1i)s||s]{n} - (Bi - Ai)*[as||s]{n} ***

                  la_start = MAX(0, la_min-1)

                  DO la = la_start, la_max-1
                     DO n = 1, nmax-la-1
                        DO ax = 0, la
                           DO ay = 0, la-ax
                              az = la-ax-ay
                              v(coset(ax, ay, az), 2, 1, n) = &
                                 v(coset(ax+1, ay, az), 1, 1, n)- &
                                 rab(1)*v(coset(ax, ay, az), 1, 1, n)
                              v(coset(ax, ay, az), 3, 1, n) = &
                                 v(coset(ax, ay+1, az), 1, 1, n)- &
                                 rab(2)*v(coset(ax, ay, az), 1, 1, n)
                              v(coset(ax, ay, az), 4, 1, n) = &
                                 v(coset(ax, ay, az+1), 1, 1, n)- &
                                 rab(3)*v(coset(ax, ay, az), 1, 1, n)
                           END DO
                        END DO
                     END DO
                  END DO

                  ! *** Vertical recurrence step ***

                  ! *** [ap||s]{n} = (Pi - Bi)*[as||s]{n} +          ***
                  ! ***              (Wi - Pi)*[as||s]{n+1} +        ***
                  ! ***              f2*Ni(a)*(   [(a-1i)s||s]{n} +  ***
                  ! ***                        f4*[(a-1i)s||s]{n+1}) ***

                  DO n = 1, nmax-la_max-1
                     DO ax = 0, la_max
                        fx = f2*REAL(ax, dp)
                        DO ay = 0, la_max-ax
                           fy = f2*REAL(ay, dp)
                           az = la_max-ax-ay
                           fz = f2*REAL(az, dp)

                           IF (ax == 0) THEN
                              v(coset(ax, ay, az), 2, 1, n) = &
                                 rbp(1)*v(coset(ax, ay, az), 1, 1, n)+ &
                                 rpw(1)*v(coset(ax, ay, az), 1, 1, n+1)
                           ELSE
                              v(coset(ax, ay, az), 2, 1, n) = &
                                 rbp(1)*v(coset(ax, ay, az), 1, 1, n)+ &
                                 rpw(1)*v(coset(ax, ay, az), 1, 1, n+1)+ &
                                 fx*(v(coset(ax-1, ay, az), 1, 1, n)+ &
                                     f4*v(coset(ax-1, ay, az), 1, 1, n+1))
                           END IF

                           IF (ay == 0) THEN
                              v(coset(ax, ay, az), 3, 1, n) = &
                                 rbp(2)*v(coset(ax, ay, az), 1, 1, n)+ &
                                 rpw(2)*v(coset(ax, ay, az), 1, 1, n+1)
                           ELSE
                              v(coset(ax, ay, az), 3, 1, n) = &
                                 rbp(2)*v(coset(ax, ay, az), 1, 1, n)+ &
                                 rpw(2)*v(coset(ax, ay, az), 1, 1, n+1)+ &
                                 fy*(v(coset(ax, ay-1, az), 1, 1, n)+ &
                                     f4*v(coset(ax, ay-1, az), 1, 1, n+1))
                           END IF

                           IF (az == 0) THEN
                              v(coset(ax, ay, az), 4, 1, n) = &
                                 rbp(3)*v(coset(ax, ay, az), 1, 1, n)+ &
                                 rpw(3)*v(coset(ax, ay, az), 1, 1, n+1)
                           ELSE
                              v(coset(ax, ay, az), 4, 1, n) = &
                                 rbp(3)*v(coset(ax, ay, az), 1, 1, n)+ &
                                 rpw(3)*v(coset(ax, ay, az), 1, 1, n+1)+ &
                                 fz*(v(coset(ax, ay, az-1), 1, 1, n)+ &
                                     f4*v(coset(ax, ay, az-1), 1, 1, n+1))
                           END IF

                        END DO
                     END DO
                  END DO

                  ! *** Recurrence steps: [ap||s] -> [ab||s] ***

                  DO lb = 2, lb_max

                     ! *** Horizontal recurrence steps ***

                     ! *** [ab||s]{n} = [(a+1i)(b-1i)||s]{n} -    ***
                     ! ***              (Bi - Ai)*[a(b-1i)||s]{n} ***

                     la_start = MAX(0, la_min-1)

                     DO la = la_start, la_max-1
                        DO n = 1, nmax-la-lb
                           DO ax = 0, la
                              DO ay = 0, la-ax
                                 az = la-ax-ay

                                 ! *** Shift of angular momentum component z from a to b ***

                                 v(coset(ax, ay, az), coset(0, 0, lb), 1, n) = &
                                    v(coset(ax, ay, az+1), coset(0, 0, lb-1), 1, n)- &
                                    rab(3)*v(coset(ax, ay, az), coset(0, 0, lb-1), 1, n)

                                 ! *** Shift of angular momentum component y from a to b ***

                                 DO by = 1, lb
                                    bz = lb-by
                                    v(coset(ax, ay, az), coset(0, by, bz), 1, n) = &
                                       v(coset(ax, ay+1, az), coset(0, by-1, bz), 1, n)- &
                                       rab(2)*v(coset(ax, ay, az), coset(0, by-1, bz), 1, n)
                                 END DO

                                 ! *** Shift of angular momentum component x from a to b ***

                                 DO bx = 1, lb
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       v(coset(ax, ay, az), coset(bx, by, bz), 1, n) = &
                                          v(coset(ax+1, ay, az), coset(bx-1, by, bz), 1, n)- &
                                          rab(1)*v(coset(ax, ay, az), coset(bx-1, by, bz), 1, n)
                                    END DO
                                 END DO

                              END DO
                           END DO
                        END DO
                     END DO

                     ! *** Vertical recurrence step ***

                     ! *** [ab||s]{n} = (Pi - Bi)*[a(b-1i)||s]{n} +          ***
                     ! ***              (Wi - Pi)*[a(b-1i)||s]{n+1} +        ***
                     ! ***              f2*Ni(a)*(   [(a-1i)(b-1i)||s]{n} +  ***
                     ! ***                        f4*[(a-1i)(b-1i)||s]{n+1}) ***
                     ! ***              f2*Ni(b-1i)*(   [a(b-2i)||s]{n} +    ***
                     ! ***                           f4*[a(b-2i)||s]{n+1})   ***

                     DO n = 1, nmax-la_max-lb
                        DO ax = 0, la_max
                           fx = f2*REAL(ax, dp)
                           DO ay = 0, la_max-ax
                              fy = f2*REAL(ay, dp)
                              az = la_max-ax-ay
                              fz = f2*REAL(az, dp)

                              ! *** Shift of angular momentum component z from a to b ***

                              f3 = f2*REAL(lb-1, dp)

                              IF (az == 0) THEN
                                 v(coset(ax, ay, az), coset(0, 0, lb), 1, n) = &
                                    rbp(3)*v(coset(ax, ay, az), coset(0, 0, lb-1), 1, n)+ &
                                    rpw(3)*v(coset(ax, ay, az), coset(0, 0, lb-1), 1, n+1)+ &
                                    f3*(v(coset(ax, ay, az), coset(0, 0, lb-2), 1, n)+ &
                                        f4*v(coset(ax, ay, az), coset(0, 0, lb-2), 1, n+1))
                              ELSE
                                 v(coset(ax, ay, az), coset(0, 0, lb), 1, n) = &
                                    rbp(3)*v(coset(ax, ay, az), coset(0, 0, lb-1), 1, n)+ &
                                    rpw(3)*v(coset(ax, ay, az), coset(0, 0, lb-1), 1, n+1)+ &
                                    fz*(v(coset(ax, ay, az-1), coset(0, 0, lb-1), 1, n)+ &
                                        f4*v(coset(ax, ay, az-1), coset(0, 0, lb-1), 1, n+1))+ &
                                    f3*(v(coset(ax, ay, az), coset(0, 0, lb-2), 1, n)+ &
                                        f4*v(coset(ax, ay, az), coset(0, 0, lb-2), 1, n+1))
                              END IF

                              ! *** Shift of angular momentum component y from a to b ***

                              IF (ay == 0) THEN
                                 bz = lb-1
                                 v(coset(ax, ay, az), coset(0, 1, bz), 1, n) = &
                                    rbp(2)*v(coset(ax, ay, az), coset(0, 0, bz), 1, n)+ &
                                    rpw(2)*v(coset(ax, ay, az), coset(0, 0, bz), 1, n+1)
                                 DO by = 2, lb
                                    bz = lb-by
                                    f3 = f2*REAL(by-1, dp)
                                    v(coset(ax, ay, az), coset(0, by, bz), 1, n) = &
                                       rbp(2)*v(coset(ax, ay, az), coset(0, by-1, bz), 1, n)+ &
                                       rpw(2)*v(coset(ax, ay, az), coset(0, by-1, bz), 1, n+1)+ &
                                       f3*(v(coset(ax, ay, az), coset(0, by-2, bz), 1, n)+ &
                                           f4*v(coset(ax, ay, az), coset(0, by-2, bz), 1, n+1))
                                 END DO
                              ELSE
                                 bz = lb-1
                                 v(coset(ax, ay, az), coset(0, 1, bz), 1, n) = &
                                    rbp(2)*v(coset(ax, ay, az), coset(0, 0, bz), 1, n)+ &
                                    rpw(2)*v(coset(ax, ay, az), coset(0, 0, bz), 1, n+1)+ &
                                    fy*(v(coset(ax, ay-1, az), coset(0, 0, bz), 1, n)+ &
                                        f4*v(coset(ax, ay-1, az), coset(0, 0, bz), 1, n+1))
                                 DO by = 2, lb
                                    bz = lb-by
                                    f3 = f2*REAL(by-1, dp)
                                    v(coset(ax, ay, az), coset(0, by, bz), 1, n) = &
                                       rbp(2)*v(coset(ax, ay, az), coset(0, by-1, bz), 1, n)+ &
                                       rpw(2)*v(coset(ax, ay, az), coset(0, by-1, bz), 1, n+1)+ &
                                       fy*(v(coset(ax, ay-1, az), coset(0, by-1, bz), 1, n)+ &
                                           f4*v(coset(ax, ay-1, az), &
                                                coset(0, by-1, bz), 1, n+1))+ &
                                       f3*(v(coset(ax, ay, az), coset(0, by-2, bz), 1, n)+ &
                                           f4*v(coset(ax, ay, az), coset(0, by-2, bz), 1, n+1))
                                 END DO
                              END IF

                              ! *** Shift of angular momentum component x from a to b ***

                              IF (ax == 0) THEN
                                 DO by = 0, lb-1
                                    bz = lb-1-by
                                    v(coset(ax, ay, az), coset(1, by, bz), 1, n) = &
                                       rbp(1)*v(coset(ax, ay, az), coset(0, by, bz), 1, n)+ &
                                       rpw(1)*v(coset(ax, ay, az), coset(0, by, bz), 1, n+1)
                                 END DO
                                 DO bx = 2, lb
                                    f3 = f2*REAL(bx-1, dp)
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       v(coset(ax, ay, az), coset(bx, by, bz), 1, n) = &
                                          rbp(1)*v(coset(ax, ay, az), coset(bx-1, by, bz), 1, n)+ &
                                          rpw(1)*v(coset(ax, ay, az), &
                                                   coset(bx-1, by, bz), 1, n+1)+ &
                                          f3*(v(coset(ax, ay, az), coset(bx-2, by, bz), 1, n)+ &
                                              f4*v(coset(ax, ay, az), coset(bx-2, by, bz), 1, n+1))
                                    END DO
                                 END DO
                              ELSE
                                 DO by = 0, lb-1
                                    bz = lb-1-by
                                    v(coset(ax, ay, az), coset(1, by, bz), 1, n) = &
                                       rbp(1)*v(coset(ax, ay, az), coset(0, by, bz), 1, n)+ &
                                       rpw(1)*v(coset(ax, ay, az), coset(0, by, bz), 1, n+1)+ &
                                       fx*(v(coset(ax-1, ay, az), coset(0, by, bz), 1, n)+ &
                                           f4*v(coset(ax-1, ay, az), coset(0, by, bz), 1, n+1))
                                 END DO
                                 DO bx = 2, lb
                                    f3 = f2*REAL(bx-1, dp)
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       v(coset(ax, ay, az), coset(bx, by, bz), 1, n) = &
                                          rbp(1)*v(coset(ax, ay, az), coset(bx-1, by, bz), 1, n)+ &
                                          rpw(1)*v(coset(ax, ay, az), &
                                                   coset(bx-1, by, bz), 1, n+1)+ &
                                          fx*(v(coset(ax-1, ay, az), &
                                                coset(bx-1, by, bz), 1, n)+ &
                                              f4*v(coset(ax-1, ay, az), &
                                                   coset(bx-1, by, bz), 1, n+1))+ &
                                          f3*(v(coset(ax, ay, az), coset(bx-2, by, bz), 1, n)+ &
                                              f4*v(coset(ax, ay, az), coset(bx-2, by, bz), 1, n+1))
                                    END DO
                                 END DO
                              END IF

                           END DO
                        END DO
                     END DO

                  END DO

               END IF

            ELSE

               IF (lb_max > 0) THEN

                  ! *** Vertical recurrence steps: [ss||s] -> [sb||s] ***

                  rbp(:) = rap(:)-rab(:)

                  ! *** [sp||s]{n} = (Pi - Bi)*[ss||s]{n} + ***
                  ! ***              (Wi - Pi)*[ss||s]{n+1} ***

                  DO n = 1, nmax-1
                     v(1, 2, 1, n) = rbp(1)*v(1, 1, 1, n)+rpw(1)*v(1, 1, 1, n+1)
                     v(1, 3, 1, n) = rbp(2)*v(1, 1, 1, n)+rpw(2)*v(1, 1, 1, n+1)
                     v(1, 4, 1, n) = rbp(3)*v(1, 1, 1, n)+rpw(3)*v(1, 1, 1, n+1)
                  END DO

                  ! *** [sb||s]{n} = (Pi - Bi)*[s(b-1i)||s]{n} +        ***
                  ! ***              (Wi - Pi)*[s(b-1i)||s]{n+1} +      ***
                  ! ***              f2*Ni(b-1i)*(   [s(b-2i)||s]{n} +  ***
                  ! ***                           f4*[s(b-2i)||s]{n+1}) ***

                  DO lb = 2, lb_max

                     DO n = 1, nmax-lb

                        ! *** Increase the angular momentum component z of b ***

                        v(1, coset(0, 0, lb), 1, n) = &
                           rbp(3)*v(1, coset(0, 0, lb-1), 1, n)+ &
                           rpw(3)*v(1, coset(0, 0, lb-1), 1, n+1)+ &
                           f2*REAL(lb-1, dp)*(v(1, coset(0, 0, lb-2), 1, n)+ &
                                              f4*v(1, coset(0, 0, lb-2), 1, n+1))

                        ! *** Increase the angular momentum component y of b ***

                        bz = lb-1
                        v(1, coset(0, 1, bz), 1, n) = &
                           rbp(2)*v(1, coset(0, 0, bz), 1, n)+ &
                           rpw(2)*v(1, coset(0, 0, bz), 1, n+1)

                        DO by = 2, lb
                           bz = lb-by
                           v(1, coset(0, by, bz), 1, n) = &
                              rbp(2)*v(1, coset(0, by-1, bz), 1, n)+ &
                              rpw(2)*v(1, coset(0, by-1, bz), 1, n+1)+ &
                              f2*REAL(by-1, dp)*(v(1, coset(0, by-2, bz), 1, n)+ &
                                                 f4*v(1, coset(0, by-2, bz), 1, n+1))
                        END DO

                        ! *** Increase the angular momentum component x of b ***

                        DO by = 0, lb-1
                           bz = lb-1-by
                           v(1, coset(1, by, bz), 1, n) = &
                              rbp(1)*v(1, coset(0, by, bz), 1, n)+ &
                              rpw(1)*v(1, coset(0, by, bz), 1, n+1)
                        END DO

                        DO bx = 2, lb
                           f3 = f2*REAL(bx-1, dp)
                           DO by = 0, lb-bx
                              bz = lb-bx-by
                              v(1, coset(bx, by, bz), 1, n) = &
                                 rbp(1)*v(1, coset(bx-1, by, bz), 1, n)+ &
                                 rpw(1)*v(1, coset(bx-1, by, bz), 1, n+1)+ &
                                 f3*(v(1, coset(bx-2, by, bz), 1, n)+ &
                                     f4*v(1, coset(bx-2, by, bz), 1, n+1))
                           END DO
                        END DO

                     END DO

                  END DO

               END IF

            END IF

            ! *** Recurrence steps: [ab||s] -> [ab||c] ***

            IF (lc_max > 0) THEN

               ! *** Vertical recurrence steps: [ss||s] -> [ss||c] ***

               f5 = -zetw/zetp
               f6 = 0.5_dp*zetw
               f7 = 0.5_dp*zetq

               rcw(:) = rcp(:)+rpw(:)

               ! *** [ss||p]{n} = (Wi - Ci)*[ss||s]{n+1}  (i = x,y,z) ***

               DO n = 1, nmax-1
                  v(1, 1, 2, n) = rcw(1)*v(1, 1, 1, n+1)
                  v(1, 1, 3, n) = rcw(2)*v(1, 1, 1, n+1)
                  v(1, 1, 4, n) = rcw(3)*v(1, 1, 1, n+1)
               END DO

               ! *** [ss||c]{n} = (Wi - Ci)*[ss||c-1i]{n+1} + ***
               ! ***              f7*Ni(c-1i)*[ss||c-2i]{n} + ***
               ! ***              f5*[ss||c-2i]{n+1}          ***

               DO lc = 2, lc_max

                  DO n = 1, nmax-lc

                     ! *** Increase the angular momentum component z of c ***

                     v(1, 1, coset(0, 0, lc), n) = &
                        rcw(3)*v(1, 1, coset(0, 0, lc-1), n+1)+ &
                        f7*REAL(lc-1, dp)*(v(1, 1, coset(0, 0, lc-2), n)+ &
                                           f5*v(1, 1, coset(0, 0, lc-2), n+1))

                     ! *** Increase the angular momentum component y of c ***

                     cz = lc-1
                     v(1, 1, coset(0, 1, cz), n) = rcw(2)*v(1, 1, coset(0, 0, cz), n+1)

                     DO cy = 2, lc
                        cz = lc-cy
                        v(1, 1, coset(0, cy, cz), n) = &
                           rcw(2)*v(1, 1, coset(0, cy-1, cz), n+1)+ &
                           f7*REAL(cy-1, dp)*(v(1, 1, coset(0, cy-2, cz), n)+ &
                                              f5*v(1, 1, coset(0, cy-2, cz), n+1))
                     END DO

                     ! *** Increase the angular momentum component x of c ***

                     DO cy = 0, lc-1
                        cz = lc-1-cy
                        v(1, 1, coset(1, cy, cz), n) = rcw(1)*v(1, 1, coset(0, cy, cz), n+1)
                     END DO

                     DO cx = 2, lc
                        DO cy = 0, lc-cx
                           cz = lc-cx-cy
                           v(1, 1, coset(cx, cy, cz), n) = &
                              rcw(1)*v(1, 1, coset(cx-1, cy, cz), n+1)+ &
                              f7*REAL(cx-1, dp)*(v(1, 1, coset(cx-2, cy, cz), n)+ &
                                                 f5*v(1, 1, coset(cx-2, cy, cz), n+1))
                        END DO
                     END DO

                  END DO

               END DO

               ! *** Recurrence steps: [ss||c] -> [ab||c] ***

               DO lc = 1, lc_max

                  DO cx = 0, lc
                     DO cy = 0, lc-cx
                        cz = lc-cx-cy

                        coc = coset(cx, cy, cz)
                        cocx = coset(MAX(0, cx-1), cy, cz)
                        cocy = coset(cx, MAX(0, cy-1), cz)
                        cocz = coset(cx, cy, MAX(0, cz-1))

                        fcx = f6*REAL(cx, dp)
                        fcy = f6*REAL(cy, dp)
                        fcz = f6*REAL(cz, dp)

                        ! *** Recurrence steps: [ss||c] -> [as||c] ***

                        IF (la_max > 0) THEN

                           ! *** Vertical recurrence steps: [ss||c] -> [as||c] ***

                           ! *** [ps||c]{n} = (Pi - Ai)*[ss||c]{n} +                ***
                           ! ***              (Wi - Pi)*[ss||c]{n+1} +              ***
                           ! ***              f6*Ni(c)*[ss||c-1i]{n+1}  (i = x,y,z) ***

                           DO n = 1, nmax-1-lc
                              v(2, 1, coc, n) = rap(1)*v(1, 1, coc, n)+ &
                                                rpw(1)*v(1, 1, coc, n+1)+ &
                                                fcx*v(1, 1, cocx, n+1)
                              v(3, 1, coc, n) = rap(2)*v(1, 1, coc, n)+ &
                                                rpw(2)*v(1, 1, coc, n+1)+ &
                                                fcy*v(1, 1, cocy, n+1)
                              v(4, 1, coc, n) = rap(3)*v(1, 1, coc, n)+ &
                                                rpw(3)*v(1, 1, coc, n+1)+ &
                                                fcz*v(1, 1, cocz, n+1)
                           END DO

                           ! *** [as||c]{n} = (Pi - Ai)*[(a-1i)s||c]{n} +          ***
                           ! ***              (Wi - Pi)*[(a-1i)s||c]{n+1} +        ***
                           ! ***              f2*Ni(a-1i)*(   [(a-2i)s||c]{n} +    ***
                           ! ***                           f4*[(a-2i)s||c]{n+1}) + ***
                           ! ***              f6*Ni(c)*[(a-1i)s||c-1i]{n+1}        ***

                           DO la = 2, la_max

                              DO n = 1, nmax-la-lc

                                 ! *** Increase the angular momentum component z of a ***

                                 v(coset(0, 0, la), 1, coc, n) = &
                                    rap(3)*v(coset(0, 0, la-1), 1, coc, n)+ &
                                    rpw(3)*v(coset(0, 0, la-1), 1, coc, n+1)+ &
                                    f2*REAL(la-1, dp)*(v(coset(0, 0, la-2), 1, coc, n)+ &
                                                       f4*v(coset(0, 0, la-2), 1, coc, n+1))+ &
                                    fcz*v(coset(0, 0, la-1), 1, cocz, n+1)

                                 ! *** Increase the angular momentum component y of a ***

                                 az = la-1
                                 v(coset(0, 1, az), 1, coc, n) = &
                                    rap(2)*v(coset(0, 0, az), 1, coc, n)+ &
                                    rpw(2)*v(coset(0, 0, az), 1, coc, n+1)+ &
                                    fcy*v(coset(0, 0, az), 1, cocy, n+1)

                                 DO ay = 2, la
                                    f3 = f2*REAL(ay-1, dp)
                                    az = la-ay
                                    v(coset(0, ay, az), 1, coc, n) = &
                                       rap(2)*v(coset(0, ay-1, az), 1, coc, n)+ &
                                       rpw(2)*v(coset(0, ay-1, az), 1, coc, n+1)+ &
                                       f3*(v(coset(0, ay-2, az), 1, coc, n)+ &
                                           f4*v(coset(0, ay-2, az), 1, coc, n+1))+ &
                                       fcy*v(coset(0, ay-1, az), 1, cocy, n+1)
                                 END DO

                                 ! *** Increase the angular momentum component x of a ***

                                 DO ay = 0, la-1
                                    az = la-1-ay
                                    v(coset(1, ay, az), 1, coc, n) = &
                                       rap(1)*v(coset(0, ay, az), 1, coc, n)+ &
                                       rpw(1)*v(coset(0, ay, az), 1, coc, n+1)+ &
                                       fcx*v(coset(0, ay, az), 1, cocx, n+1)
                                 END DO

                                 DO ax = 2, la
                                    f3 = f2*REAL(ax-1, dp)
                                    DO ay = 0, la-ax
                                       az = la-ax-ay
                                       v(coset(ax, ay, az), 1, coc, n) = &
                                          rap(1)*v(coset(ax-1, ay, az), 1, coc, n)+ &
                                          rpw(1)*v(coset(ax-1, ay, az), 1, coc, n+1)+ &
                                          f3*(v(coset(ax-2, ay, az), 1, coc, n)+ &
                                              f4*v(coset(ax-2, ay, az), 1, coc, n+1))+ &
                                          fcx*v(coset(ax-1, ay, az), 1, cocx, n+1)
                                    END DO
                                 END DO

                              END DO

                           END DO

                           ! *** Recurrence steps: [as||c] -> [ab||c] ***

                           IF (lb_max > 0) THEN

                              ! *** Horizontal recurrence steps ***

                              ! *** [ap||c]{n} = [(a+1i)s||c]{n} - (Bi - Ai)*[as||c]{n} ***

                              la_start = MAX(0, la_min-1)

                              DO la = la_start, la_max-1
                                 DO n = 1, nmax-la-1-lc
                                    DO ax = 0, la
                                       DO ay = 0, la-ax
                                          az = la-ax-ay
                                          v(coset(ax, ay, az), 2, coc, n) = &
                                             v(coset(ax+1, ay, az), 1, coc, n)- &
                                             rab(1)*v(coset(ax, ay, az), 1, coc, n)
                                          v(coset(ax, ay, az), 3, coc, n) = &
                                             v(coset(ax, ay+1, az), 1, coc, n)- &
                                             rab(2)*v(coset(ax, ay, az), 1, coc, n)
                                          v(coset(ax, ay, az), 4, coc, n) = &
                                             v(coset(ax, ay, az+1), 1, coc, n)- &
                                             rab(3)*v(coset(ax, ay, az), 1, coc, n)
                                       END DO
                                    END DO
                                 END DO
                              END DO

                              ! *** Vertical recurrence step ***

                              ! *** [ap||c]{n} = (Pi - Bi)*[as||c]{n} +            ***
                              ! ***              (Wi - Pi)*[as||c]{n+1} +          ***
                              ! ***              f2*Ni(a)*(   [(a-1i)s||c]{n} +    ***
                              ! ***                        f4*[(a-1i)s||c]{n+1}) + ***
                              ! ***              f6*Ni(c)*[(as||c-1i]{n+1})        ***

                              DO n = 1, nmax-la_max-1-lc
                                 DO ax = 0, la_max
                                    fx = f2*REAL(ax, dp)
                                    DO ay = 0, la_max-ax
                                       fy = f2*REAL(ay, dp)
                                       az = la_max-ax-ay
                                       fz = f2*REAL(az, dp)

                                       IF (ax == 0) THEN
                                          v(coset(ax, ay, az), 2, coc, n) = &
                                             rbp(1)*v(coset(ax, ay, az), 1, coc, n)+ &
                                             rpw(1)*v(coset(ax, ay, az), 1, coc, n+1)+ &
                                             fcx*v(coset(ax, ay, az), 1, cocx, n+1)
                                       ELSE
                                          v(coset(ax, ay, az), 2, coc, n) = &
                                             rbp(1)*v(coset(ax, ay, az), 1, coc, n)+ &
                                             rpw(1)*v(coset(ax, ay, az), 1, coc, n+1)+ &
                                             fx*(v(coset(ax-1, ay, az), 1, coc, n)+ &
                                                 f4*v(coset(ax-1, ay, az), 1, coc, n+1))+ &
                                             fcx*v(coset(ax, ay, az), 1, cocx, n+1)
                                       END IF

                                       IF (ay == 0) THEN
                                          v(coset(ax, ay, az), 3, coc, n) = &
                                             rbp(2)*v(coset(ax, ay, az), 1, coc, n)+ &
                                             rpw(2)*v(coset(ax, ay, az), 1, coc, n+1)+ &
                                             fcy*v(coset(ax, ay, az), 1, cocy, n+1)
                                       ELSE
                                          v(coset(ax, ay, az), 3, coc, n) = &
                                             rbp(2)*v(coset(ax, ay, az), 1, coc, n)+ &
                                             rpw(2)*v(coset(ax, ay, az), 1, coc, n+1)+ &
                                             fy*(v(coset(ax, ay-1, az), 1, coc, n)+ &
                                                 f4*v(coset(ax, ay-1, az), 1, coc, n+1))+ &
                                             fcy*v(coset(ax, ay, az), 1, cocy, n+1)
                                       END IF

                                       IF (az == 0) THEN
                                          v(coset(ax, ay, az), 4, coc, n) = &
                                             rbp(3)*v(coset(ax, ay, az), 1, coc, n)+ &
                                             rpw(3)*v(coset(ax, ay, az), 1, coc, n+1)+ &
                                             fcz*v(coset(ax, ay, az), 1, cocz, n+1)
                                       ELSE
                                          v(coset(ax, ay, az), 4, coc, n) = &
                                             rbp(3)*v(coset(ax, ay, az), 1, coc, n)+ &
                                             rpw(3)*v(coset(ax, ay, az), 1, coc, n+1)+ &
                                             fz*(v(coset(ax, ay, az-1), 1, coc, n)+ &
                                                 f4*v(coset(ax, ay, az-1), 1, coc, n+1))+ &
                                             fcz*v(coset(ax, ay, az), 1, cocz, n+1)
                                       END IF

                                    END DO
                                 END DO
                              END DO

                              ! *** Recurrence steps: [ap||c] -> [ab||c] ***

                              DO lb = 2, lb_max

                                 ! *** Horizontal recurrence steps ***

                                 ! *** [ab||c]{n} = [(a+1i)(b-1i)||c]{n} -    ***
                                 ! ***              (Bi - Ai)*[a(b-1i)||c]{n} ***

                                 la_start = MAX(0, la_min-1)

                                 DO la = la_start, la_max-1
                                    DO n = 1, nmax-la-lb-lc
                                       DO ax = 0, la
                                          DO ay = 0, la-ax
                                             az = la-ax-ay

                                             ! *** Shift of angular momentum component z ***

                                             v(coset(ax, ay, az), coset(0, 0, lb), coc, n) = &
                                                v(coset(ax, ay, az+1), &
                                                  coset(0, 0, lb-1), coc, n)- &
                                                rab(3)*v(coset(ax, ay, az), &
                                                         coset(0, 0, lb-1), coc, n)

                                             ! *** Shift of angular momentum component y ***

                                             DO by = 1, lb
                                                bz = lb-by
                                                v(coset(ax, ay, az), coset(0, by, bz), coc, n) = &
                                                   v(coset(ax, ay+1, az), &
                                                     coset(0, by-1, bz), coc, n)- &
                                                   rab(2)*v(coset(ax, ay, az), &
                                                            coset(0, by-1, bz), coc, n)
                                             END DO

                                             ! *** Shift of angular momentum component x ***

                                             DO bx = 1, lb
                                                DO by = 0, lb-bx
                                                   bz = lb-bx-by
                                                   v(coset(ax, ay, az), coset(bx, by, bz), coc, n) = &
                                                      v(coset(ax+1, ay, az), &
                                                        coset(bx-1, by, bz), coc, n)- &
                                                      rab(1)*v(coset(ax, ay, az), &
                                                               coset(bx-1, by, bz), coc, n)
                                                END DO
                                             END DO

                                          END DO
                                       END DO
                                    END DO
                                 END DO

                                 ! *** Vertical recurrence step ***

                                 ! *** [ab||c]{n} = (Pi - Bi)*[a(b-1i)||c]{n} +          ***
                                 ! ***              (Wi - Pi)*[a(b-1i)||c]{n+1} +        ***
                                 ! ***              f2*Ni(a)*(   [(a-1i)(b-1i)||c]{n} +  ***
                                 ! ***                        f4*[(a-1i)(b-1i)||c]{n+1}) ***
                                 ! ***              f2*Ni(b-1i)*(   [a(b-2i)||c]{n} +    ***
                                 ! ***                           f4*[a(b-2i)||c]{n+1}) + ***
                                 ! ***              f6*Ni(c)*[a(b-1i)||c-1i]{n+1})       ***

                                 DO n = 1, nmax-la_max-lb-lc
                                    DO ax = 0, la_max
                                       fx = f2*REAL(ax, dp)
                                       DO ay = 0, la_max-ax
                                          fy = f2*REAL(ay, dp)
                                          az = la_max-ax-ay
                                          fz = f2*REAL(az, dp)

                                          ! *** Shift of angular momentum component z from a to b ***

                                          f3 = f2*REAL(lb-1, dp)

                                          IF (az == 0) THEN
                                             v(coset(ax, ay, az), coset(0, 0, lb), coc, n) = &
                                                rbp(3)*v(coset(ax, ay, az), &
                                                         coset(0, 0, lb-1), coc, n)+ &
                                                rpw(3)*v(coset(ax, ay, az), &
                                                         coset(0, 0, lb-1), coc, n+1)+ &
                                                f3*(v(coset(ax, ay, az), &
                                                      coset(0, 0, lb-2), coc, n)+ &
                                                    f4*v(coset(ax, ay, az), &
                                                         coset(0, 0, lb-2), coc, n+1))+ &
                                                fcz*v(coset(ax, ay, az), &
                                                      coset(0, 0, lb-1), cocz, n+1)
                                          ELSE
                                             v(coset(ax, ay, az), coset(0, 0, lb), coc, n) = &
                                                rbp(3)*v(coset(ax, ay, az), &
                                                         coset(0, 0, lb-1), coc, n)+ &
                                                rpw(3)*v(coset(ax, ay, az), &
                                                         coset(0, 0, lb-1), coc, n+1)+ &
                                                fz*(v(coset(ax, ay, az-1), &
                                                      coset(0, 0, lb-1), coc, n)+ &
                                                    f4*v(coset(ax, ay, az-1), &
                                                         coset(0, 0, lb-1), coc, n+1))+ &
                                                f3*(v(coset(ax, ay, az), &
                                                      coset(0, 0, lb-2), coc, n)+ &
                                                    f4*v(coset(ax, ay, az), &
                                                         coset(0, 0, lb-2), coc, n+1))+ &
                                                fcz*v(coset(ax, ay, az), &
                                                      coset(0, 0, lb-1), cocz, n+1)
                                          END IF

                                          ! *** Shift of angular momentum component y from a to b ***

                                          IF (ay == 0) THEN
                                             bz = lb-1
                                             v(coset(ax, ay, az), coset(0, 1, bz), coc, n) = &
                                                rbp(2)*v(coset(ax, ay, az), &
                                                         coset(0, 0, bz), coc, n)+ &
                                                rpw(2)*v(coset(ax, ay, az), &
                                                         coset(0, 0, bz), coc, n+1)+ &
                                                fcy*v(coset(ax, ay, az), &
                                                      coset(0, 0, bz), cocy, n+1)
                                             DO by = 2, lb
                                                bz = lb-by
                                                f3 = f2*REAL(by-1, dp)
                                                v(coset(ax, ay, az), coset(0, by, bz), coc, n) = &
                                                   rbp(2)*v(coset(ax, ay, az), &
                                                            coset(0, by-1, bz), coc, n)+ &
                                                   rpw(2)*v(coset(ax, ay, az), &
                                                            coset(0, by-1, bz), coc, n+1)+ &
                                                   f3*(v(coset(ax, ay, az), &
                                                         coset(0, by-2, bz), coc, n)+ &
                                                       f4*v(coset(ax, ay, az), &
                                                            coset(0, by-2, bz), coc, n+1))+ &
                                                   fcy*v(coset(ax, ay, az), &
                                                         coset(0, by-1, bz), cocy, n+1)
                                             END DO
                                          ELSE
                                             bz = lb-1
                                             v(coset(ax, ay, az), coset(0, 1, bz), coc, n) = &
                                                rbp(2)*v(coset(ax, ay, az), &
                                                         coset(0, 0, bz), coc, n)+ &
                                                rpw(2)*v(coset(ax, ay, az), &
                                                         coset(0, 0, bz), coc, n+1)+ &
                                                fy*(v(coset(ax, ay-1, az), &
                                                      coset(0, 0, bz), coc, n)+ &
                                                    f4*v(coset(ax, ay-1, az), &
                                                         coset(0, 0, bz), coc, n+1))+ &
                                                fcy*v(coset(ax, ay, az), &
                                                      coset(0, 0, bz), cocy, n+1)
                                             DO by = 2, lb
                                                bz = lb-by
                                                f3 = f2*REAL(by-1, dp)
                                                v(coset(ax, ay, az), coset(0, by, bz), coc, n) = &
                                                   rbp(2)*v(coset(ax, ay, az), &
                                                            coset(0, by-1, bz), coc, n)+ &
                                                   rpw(2)*v(coset(ax, ay, az), &
                                                            coset(0, by-1, bz), coc, n+1)+ &
                                                   fy*(v(coset(ax, ay-1, az), &
                                                         coset(0, by-1, bz), coc, n)+ &
                                                       f4*v(coset(ax, ay-1, az), &
                                                            coset(0, by-1, bz), coc, n+1))+ &
                                                   f3*(v(coset(ax, ay, az), &
                                                         coset(0, by-2, bz), coc, n)+ &
                                                       f4*v(coset(ax, ay, az), &
                                                            coset(0, by-2, bz), coc, n+1))+ &
                                                   fcy*v(coset(ax, ay, az), &
                                                         coset(0, by-1, bz), cocy, n+1)
                                             END DO
                                          END IF

                                          ! *** Shift of angular momentum component x from a to b ***

                                          IF (ax == 0) THEN
                                             DO by = 0, lb-1
                                                bz = lb-1-by
                                                v(coset(ax, ay, az), coset(1, by, bz), coc, n) = &
                                                   rbp(1)*v(coset(ax, ay, az), &
                                                            coset(0, by, bz), coc, n)+ &
                                                   rpw(1)*v(coset(ax, ay, az), &
                                                            coset(0, by, bz), coc, n+1)+ &
                                                   fcx*v(coset(ax, ay, az), &
                                                         coset(0, by, bz), cocx, n+1)
                                             END DO
                                             DO bx = 2, lb
                                                f3 = f2*REAL(bx-1, dp)
                                                DO by = 0, lb-bx
                                                   bz = lb-bx-by
                                                   v(coset(ax, ay, az), coset(bx, by, bz), coc, n) = &
                                                      rbp(1)*v(coset(ax, ay, az), &
                                                               coset(bx-1, by, bz), coc, n)+ &
                                                      rpw(1)*v(coset(ax, ay, az), &
                                                               coset(bx-1, by, bz), coc, n+1)+ &
                                                      f3*(v(coset(ax, ay, az), &
                                                            coset(bx-2, by, bz), coc, n)+ &
                                                          f4*v(coset(ax, ay, az), &
                                                               coset(bx-2, by, bz), coc, n+1))+ &
                                                      fcx*v(coset(ax, ay, az), &
                                                            coset(bx-1, by, bz), cocx, n+1)
                                                END DO
                                             END DO
                                          ELSE
                                             DO by = 0, lb-1
                                                bz = lb-1-by
                                                v(coset(ax, ay, az), coset(1, by, bz), coc, n) = &
                                                   rbp(1)*v(coset(ax, ay, az), &
                                                            coset(0, by, bz), coc, n)+ &
                                                   rpw(1)*v(coset(ax, ay, az), &
                                                            coset(0, by, bz), coc, n+1)+ &
                                                   fx*(v(coset(ax-1, ay, az), &
                                                         coset(0, by, bz), coc, n)+ &
                                                       f4*v(coset(ax-1, ay, az), &
                                                            coset(0, by, bz), coc, n+1))+ &
                                                   fcx*v(coset(ax, ay, az), &
                                                         coset(0, by, bz), cocx, n+1)
                                             END DO
                                             DO bx = 2, lb
                                                f3 = f2*REAL(bx-1, dp)
                                                DO by = 0, lb-bx
                                                   bz = lb-bx-by
                                                   v(coset(ax, ay, az), coset(bx, by, bz), coc, n) = &
                                                      rbp(1)*v(coset(ax, ay, az), &
                                                               coset(bx-1, by, bz), coc, n)+ &
                                                      rpw(1)*v(coset(ax, ay, az), &
                                                               coset(bx-1, by, bz), coc, n+1)+ &
                                                      fx*(v(coset(ax-1, ay, az), &
                                                            coset(bx-1, by, bz), coc, n)+ &
                                                          f4*v(coset(ax-1, ay, az), &
                                                               coset(bx-1, by, bz), coc, n+1))+ &
                                                      f3*(v(coset(ax, ay, az), &
                                                            coset(bx-2, by, bz), coc, n)+ &
                                                          f4*v(coset(ax, ay, az), &
                                                               coset(bx-2, by, bz), coc, n+1))+ &
                                                      fcx*v(coset(ax, ay, az), &
                                                            coset(bx-1, by, bz), cocx, n+1)
                                                END DO
                                             END DO
                                          END IF

                                       END DO
                                    END DO
                                 END DO

                              END DO
                           END IF

                        ELSE

                           IF (lb_max > 0) THEN

                              ! *** Vertical recurrence steps: [ss||c] -> [sb||c] ***

                              ! *** [sp||c]{n} = (Pi - Bi)*[ss||c]{n} +    ***
                              ! ***              (Wi - Pi)*[ss||c]{n+1} +  ***
                              ! ***              f6*Ni(c)**[ss||c-1i]{n+1} ***

                              DO n = 1, nmax-1-lc
                                 v(1, 2, coc, n) = rbp(1)*v(1, 1, coc, n)+ &
                                                   rpw(1)*v(1, 1, coc, n+1)+ &
                                                   fcx*v(1, 1, cocx, n+1)
                                 v(1, 3, coc, n) = rbp(2)*v(1, 1, coc, n)+ &
                                                   rpw(2)*v(1, 1, coc, n+1)+ &
                                                   fcy*v(1, 1, cocy, n+1)
                                 v(1, 4, coc, n) = rbp(3)*v(1, 1, coc, n)+ &
                                                   rpw(3)*v(1, 1, coc, n+1)+ &
                                                   fcz*v(1, 1, cocz, n+1)
                              END DO

                              ! *** [sb||c]{n} = (Pi - Bi)*[s(b-1i)||c]{n} +          ***
                              ! ***              (Wi - Pi)*[s(b-1i)||c]{n+1} +        ***
                              ! ***              f2*Ni(b-1i)*(   [s(b-2i)||c]{n} +    ***
                              ! ***                           f4*[s(b-2i)||c]{n+1}) + ***
                              ! ***              f6*Ni(c)**[s(b-1i)||c-1i]{n+1}       ***

                              DO lb = 2, lb_max

                                 DO n = 1, nmax-lb-lc

                                    ! *** Increase the angular momentum component z of b ***

                                    v(1, coset(0, 0, lb), coc, n) = &
                                       rbp(3)*v(1, coset(0, 0, lb-1), coc, n)+ &
                                       rpw(3)*v(1, coset(0, 0, lb-1), coc, n+1)+ &
                                       f2*REAL(lb-1, dp)*(v(1, coset(0, 0, lb-2), coc, n)+ &
                                                          f4*v(1, coset(0, 0, lb-2), coc, n+1))+ &
                                       fcz*v(1, coset(0, 0, lb-1), cocz, n+1)

                                    ! *** Increase the angular momentum component y of b ***

                                    bz = lb-1
                                    v(1, coset(0, 1, bz), coc, n) = &
                                       rbp(2)*v(1, coset(0, 0, bz), coc, n)+ &
                                       rpw(2)*v(1, coset(0, 0, bz), coc, n+1)+ &
                                       fcy*v(1, coset(0, 0, bz), cocy, n+1)

                                    DO by = 2, lb
                                       f3 = f2*REAL(by-1, dp)
                                       bz = lb-by
                                       v(1, coset(0, by, bz), coc, n) = &
                                          rbp(2)*v(1, coset(0, by-1, bz), coc, n)+ &
                                          rpw(2)*v(1, coset(0, by-1, bz), coc, n+1)+ &
                                          f3*(v(1, coset(0, by-2, bz), coc, n)+ &
                                              f4*v(1, coset(0, by-2, bz), coc, n+1))+ &
                                          fcy*v(1, coset(0, by-1, bz), cocy, n+1)
                                    END DO

                                    ! *** Increase the angular momentum component x of b ***

                                    DO by = 0, lb-1
                                       bz = lb-1-by
                                       v(1, coset(1, by, bz), coc, n) = &
                                          rbp(1)*v(1, coset(0, by, bz), coc, n)+ &
                                          rpw(1)*v(1, coset(0, by, bz), coc, n+1)+ &
                                          fcx*v(1, coset(0, by, bz), cocx, n+1)
                                    END DO

                                    DO bx = 2, lb
                                       f3 = f2*REAL(bx-1, dp)
                                       DO by = 0, lb-bx
                                          bz = lb-bx-by
                                          v(1, coset(bx, by, bz), coc, n) = &
                                             rbp(1)*v(1, coset(bx-1, by, bz), coc, n)+ &
                                             rpw(1)*v(1, coset(bx-1, by, bz), coc, n+1)+ &
                                             f3*(v(1, coset(bx-2, by, bz), coc, n)+ &
                                                 f4*v(1, coset(bx-2, by, bz), coc, n+1))+ &
                                             fcx*v(1, coset(bx-1, by, bz), cocx, n+1)
                                       END DO
                                    END DO

                                 END DO

                              END DO

                           END IF

                        END IF

                     END DO
                  END DO

               END DO

            END IF

            ! *** Add the contribution of the current pair ***
            ! *** of primitive Gaussian-type functions     ***

            DO k = ncoset(lc_min-1)+1, ncoset(lc_max)
               kk = k-ncoset(lc_min-1)
               DO j = ncoset(lb_min-1)+1, ncoset(lb_max)
                  DO i = ncoset(la_min-1)+1, ncoset(la_max-maxder_local)
                     vabc(na+i, nb+j) = vabc(na+i, nb+j)+gccc(kk)*v(i, j, k, 1)
                     int_abc(na+i, nb+j, kk) = v(i, j, k, 1)
                  END DO
               END DO
            END DO

            IF (PRESENT(maxder)) THEN
               DO k = ncoset(lc_min-1)+1, ncoset(lc_max)
                  kk = k-ncoset(lc_min-1)
                  DO j = 1, ncoset(lb_max)
                     DO i = 1, ncoset(la_max)
                        vabc_plus(nap+i, nb+j) = vabc_plus(nap+i, nb+j)+gccc(kk)*v(i, j, k, 1)
                     END DO
                  END DO
               END DO
            END IF

            nb = nb+ncoset(lb_max)

         END DO

         na = na+ncoset(la_max-maxder_local)
         nap = nap+ncoset(la_max)

      END DO

   END SUBROUTINE coulomb3

END MODULE ai_coulomb
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - 02.2004 flexible normalization of basis sets [jgh]
!>      - 07.2014 Add a set of contraction coefficient that only work on active
!>                functions
!> \author Matthias Krack (04.07.2000)
! **************************************************************************************************
MODULE basis_set_types

   USE ai_coulomb,                      ONLY: coulomb2
   USE bibliography,                    ONLY: VandeVondele2007,&
                                              cite_reference
   USE cp_linked_list_val,              ONLY: cp_sll_val_next,&
                                              cp_sll_val_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_object,&
                                              parser_search_string
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE input_section_types,             ONLY: section_vals_list_get,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE input_val_types,                 ONLY: val_get,&
                                              val_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: dfac,&
                                              pi
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: coset,&
                                              indco,&
                                              init_orbital_pointers,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_symbols,                 ONLY: cgf_symbol,&
                                              sgf_symbol
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE sto_ng,                          ONLY: get_sto_ng
   USE string_utilities,                ONLY: integer_to_string,&
                                              remove_word,&
                                              uppercase
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'basis_set_types'

! **************************************************************************************************
   ! Define the Gaussian-type orbital basis set type

   TYPE gto_basis_set_type
      !MK PRIVATE
      CHARACTER(LEN=default_string_length)       :: name
      REAL(KIND=dp)                            :: kind_radius
      REAL(KIND=dp)                            :: short_kind_radius
      INTEGER                                    :: norm_type = -1
      INTEGER                                    :: ncgf, nset, nsgf
      CHARACTER(LEN=12), DIMENSION(:), POINTER   :: cgf_symbol
      CHARACTER(LEN=6), DIMENSION(:), POINTER    :: sgf_symbol
      REAL(KIND=dp), DIMENSION(:), POINTER     :: norm_cgf, set_radius
      INTEGER, DIMENSION(:), POINTER             :: lmax, lmin, lx, ly, lz, m, ncgf_set, &
                                                    npgf, nsgf_set, nshell
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: cphi, pgf_radius, sphi, scon, zet
      INTEGER, DIMENSION(:, :), POINTER           :: first_cgf, first_sgf, l, &
                                                     last_cgf, last_sgf, n
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: gcc
   END TYPE gto_basis_set_type

   TYPE gto_basis_set_p_type
      TYPE(gto_basis_set_type), POINTER :: gto_basis_set
   END TYPE gto_basis_set_p_type

! **************************************************************************************************
   ! Define the Slater-type orbital basis set type

   TYPE sto_basis_set_type
      PRIVATE
      CHARACTER(LEN=default_string_length)       :: name
      INTEGER                                    :: nshell
      CHARACTER(LEN=6), DIMENSION(:), POINTER    :: symbol
      INTEGER, DIMENSION(:), POINTER             :: nq, lq
      REAL(KIND=dp), DIMENSION(:), POINTER     :: zet
   END TYPE sto_basis_set_type

   TYPE sto_basis_set_p_type
      TYPE(sto_basis_set_type), POINTER :: sto_basis_set
   END TYPE sto_basis_set_p_type

! **************************************************************************************************
   INTERFACE read_gto_basis_set
      MODULE PROCEDURE read_gto_basis_set1, read_gto_basis_set2
   END INTERFACE
! **************************************************************************************************

   ! Public subroutines
   PUBLIC :: allocate_gto_basis_set, &
             deallocate_gto_basis_set, &
             get_gto_basis_set, &
             init_aux_basis_set, &
             init_cphi_and_sphi, &
             init_orb_basis_set, &
             read_gto_basis_set, &
             copy_gto_basis_set, &
             create_primitive_basis_set, &
             combine_basis_sets, &
             set_gto_basis_set, &
             write_gto_basis_set, &
             write_orb_basis_set

   PUBLIC :: allocate_sto_basis_set, &
             create_gto_from_sto_basis, &
             deallocate_sto_basis_set, &
             set_sto_basis_set, &
             srules

   ! Public data types
   PUBLIC :: gto_basis_set_p_type, &
             gto_basis_set_type, &
             sto_basis_set_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE allocate_gto_basis_set(gto_basis_set)

      ! Allocate a Gaussian-type orbital (GTO) basis set data set.

      ! - Creation (26.10.2000,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_gto_basis_set', &
         routineP = moduleN//':'//routineN

      CALL deallocate_gto_basis_set(gto_basis_set)

      ALLOCATE (gto_basis_set)

      NULLIFY (gto_basis_set%cgf_symbol)
      NULLIFY (gto_basis_set%first_cgf)
      NULLIFY (gto_basis_set%first_sgf)
      NULLIFY (gto_basis_set%gcc)
      NULLIFY (gto_basis_set%l)
      NULLIFY (gto_basis_set%last_cgf)
      NULLIFY (gto_basis_set%last_sgf)
      NULLIFY (gto_basis_set%lmax)
      NULLIFY (gto_basis_set%lmin)
      NULLIFY (gto_basis_set%lx)
      NULLIFY (gto_basis_set%ly)
      NULLIFY (gto_basis_set%lz)
      NULLIFY (gto_basis_set%m)
      NULLIFY (gto_basis_set%n)
      NULLIFY (gto_basis_set%ncgf_set)
      NULLIFY (gto_basis_set%norm_cgf)
      NULLIFY (gto_basis_set%npgf)
      NULLIFY (gto_basis_set%nsgf_set)
      NULLIFY (gto_basis_set%nshell)
      NULLIFY (gto_basis_set%pgf_radius)
      NULLIFY (gto_basis_set%cphi)
      NULLIFY (gto_basis_set%sphi)
      NULLIFY (gto_basis_set%scon)
      NULLIFY (gto_basis_set%set_radius)
      NULLIFY (gto_basis_set%sgf_symbol)
      NULLIFY (gto_basis_set%zet)
      gto_basis_set%short_kind_radius = 0.0_dp

   END SUBROUTINE allocate_gto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_gto_basis_set(gto_basis_set)

      ! Deallocate a Gaussian-type orbital (GTO) basis set data set.

      ! - Creation (03.11.2000,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_gto_basis_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(gto_basis_set)) THEN
         IF (ASSOCIATED(gto_basis_set%cgf_symbol)) THEN
            DEALLOCATE (gto_basis_set%cgf_symbol)
         ENDIF
         IF (ASSOCIATED(gto_basis_set%sgf_symbol)) THEN
            DEALLOCATE (gto_basis_set%sgf_symbol)
         ENDIF
         DEALLOCATE (gto_basis_set%norm_cgf)
         DEALLOCATE (gto_basis_set%set_radius)
         DEALLOCATE (gto_basis_set%lmax)
         DEALLOCATE (gto_basis_set%lmin)
         DEALLOCATE (gto_basis_set%lx)
         DEALLOCATE (gto_basis_set%ly)
         DEALLOCATE (gto_basis_set%lz)
         DEALLOCATE (gto_basis_set%m)
         DEALLOCATE (gto_basis_set%ncgf_set)
         DEALLOCATE (gto_basis_set%npgf)
         DEALLOCATE (gto_basis_set%nsgf_set)
         DEALLOCATE (gto_basis_set%nshell)
         DEALLOCATE (gto_basis_set%cphi)
         DEALLOCATE (gto_basis_set%pgf_radius)
         DEALLOCATE (gto_basis_set%sphi)
         DEALLOCATE (gto_basis_set%scon)
         DEALLOCATE (gto_basis_set%zet)
         DEALLOCATE (gto_basis_set%first_cgf)
         DEALLOCATE (gto_basis_set%first_sgf)
         DEALLOCATE (gto_basis_set%l)
         DEALLOCATE (gto_basis_set%last_cgf)
         DEALLOCATE (gto_basis_set%last_sgf)
         DEALLOCATE (gto_basis_set%n)
         DEALLOCATE (gto_basis_set%gcc)
         DEALLOCATE (gto_basis_set)
      END IF
   END SUBROUTINE deallocate_gto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param basis_set_in ...
!> \param basis_set_out ...
! **************************************************************************************************
   SUBROUTINE copy_gto_basis_set(basis_set_in, basis_set_out)

      ! Copy a Gaussian-type orbital (GTO) basis set data set.

      TYPE(gto_basis_set_type), POINTER                  :: basis_set_in, basis_set_out

      CHARACTER(len=*), PARAMETER :: routineN = 'copy_gto_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: maxco, maxpgf, maxshell, ncgf, nset, nsgf
      TYPE(gto_basis_set_type), POINTER                  :: bin, bout

      IF(.NOT.(ASSOCIATED(basis_set_in)))CALL cp__a("aobasis/basis_set_types.F",253)

      CALL allocate_gto_basis_set(basis_set_out)

      bin => basis_set_in
      bout => basis_set_out

      bout%name = bin%name
      bout%kind_radius = bin%kind_radius
      bout%norm_type = bin%norm_type
      bout%nset = bin%nset
      bout%ncgf = bin%ncgf
      bout%nsgf = bin%nsgf
      nset = bin%nset
      ncgf = bin%ncgf
      nsgf = bin%nsgf
      ALLOCATE (bout%cgf_symbol(ncgf))
      ALLOCATE (bout%sgf_symbol(nsgf))
      bout%cgf_symbol = bin%cgf_symbol
      bout%sgf_symbol = bin%sgf_symbol
      ALLOCATE (bout%norm_cgf(ncgf))
      bout%norm_cgf = bin%norm_cgf
      ALLOCATE (bout%set_radius(nset))
      bout%set_radius = bin%set_radius
      ALLOCATE (bout%lmax(nset), bout%lmin(nset), bout%npgf(nset), bout%nshell(nset))
      bout%lmax = bin%lmax
      bout%lmin = bin%lmin
      bout%npgf = bin%npgf
      bout%nshell = bin%nshell
      ALLOCATE (bout%lx(ncgf), bout%ly(ncgf), bout%lz(ncgf), bout%m(nsgf))
      bout%lx = bin%lx
      bout%ly = bin%ly
      bout%lz = bin%lz
      bout%m = bin%m
      ALLOCATE (bout%ncgf_set(nset), bout%nsgf_set(nset))
      bout%ncgf_set = bin%ncgf_set
      bout%nsgf_set = bin%nsgf_set
      maxco = SIZE(bin%cphi, 1)
      ALLOCATE (bout%cphi(maxco, ncgf), bout%sphi(maxco, nsgf), bout%scon(maxco, nsgf))
      bout%cphi = bin%cphi
      bout%sphi = bin%sphi
      bout%scon = bin%scon
      maxpgf = MAXVAL(bin%npgf)
      ALLOCATE (bout%pgf_radius(maxpgf, nset), bout%zet(maxpgf, nset))
      bout%pgf_radius = bin%pgf_radius
      bout%zet = bin%zet
      maxshell = MAXVAL(bin%nshell)
      ALLOCATE (bout%first_cgf(maxshell, nset), bout%first_sgf(maxshell, nset))
      ALLOCATE (bout%last_cgf(maxshell, nset), bout%last_sgf(maxshell, nset))
      bout%first_cgf = bin%first_cgf
      bout%first_sgf = bin%first_sgf
      bout%last_cgf = bin%last_cgf
      bout%last_sgf = bin%last_sgf
      ALLOCATE (bout%n(maxshell, nset), bout%l(maxshell, nset))
      bout%n = bin%n
      bout%l = bin%l
      ALLOCATE (bout%gcc(maxpgf, maxshell, nset))
      bout%gcc = bin%gcc

   END SUBROUTINE copy_gto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param basis_set ...
!> \param pbasis ...
! **************************************************************************************************
   SUBROUTINE create_primitive_basis_set(basis_set, pbasis)

      ! Create a primitives only basis set

      TYPE(gto_basis_set_type), POINTER                  :: basis_set, pbasis

      CHARACTER(len=*), PARAMETER :: routineN = 'create_primitive_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ico, ip, ipgf, iset, ishell, l, lm, &
                                                            lshell, m, maxco, mpgf, nc, ncgf, ns, &
                                                            nset, nsgf
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nindex, nprim
      REAL(KIND=dp)                                      :: zet0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: zet, zeta

      IF(.NOT.(ASSOCIATED(basis_set)))CALL cp__a("aobasis/basis_set_types.F",335)

      mpgf = SUM(basis_set%npgf)
      lm = MAXVAL(basis_set%lmax)
      ALLOCATE (zet(mpgf, 0:lm), zeta(mpgf, lm+1), nindex(mpgf), nprim(0:lm))
      zet = 0.0_dp
      zeta = 0.0_dp
      DO l = 0, lm
         ip = 0
         DO iset = 1, basis_set%nset
            IF (basis_set%lmin(iset) <= l .AND. basis_set%lmax(iset) >= l) THEN
               DO ipgf = 1, basis_set%npgf(iset)
                  ip = ip+1
                  zet(ip, l) = basis_set%zet(ipgf, iset)
               END DO
            END IF
         END DO
         nprim(l) = ip
      END DO

      ! sort exponents
      DO l = 0, lm
         zet(1:nprim(l), l) = -zet(1:nprim(l), l)
         CALL sort(zet(1:nprim(l), l), nprim(l), nindex)
         ! remove duplicates
         ip = 0
         zet0 = 0.0_dp
         DO i = 1, nprim(l)
            IF (ABS(zet0-zet(i, l)) > 1.e-6_dp) THEN
               ip = ip+1
               zeta(ip, l+1) = zet(i, l)
            END IF
         END DO
         nprim(l) = ip
         !
         zeta(1:ip, l+1) = -zeta(1:ip, l+1)
      END DO

      CALL allocate_gto_basis_set(pbasis)

      pbasis%name = basis_set%name//"_primitive"
      pbasis%kind_radius = basis_set%kind_radius
      pbasis%short_kind_radius = basis_set%short_kind_radius
      pbasis%norm_type = basis_set%norm_type
      nset = lm+1
      pbasis%nset = nset
      ALLOCATE (pbasis%lmax(nset), pbasis%lmin(nset), pbasis%npgf(nset), pbasis%nshell(nset))
      DO iset = 1, nset
         pbasis%lmax(iset) = iset-1
         pbasis%lmin(iset) = iset-1
         pbasis%npgf(iset) = nprim(iset-1)
         pbasis%nshell(iset) = nprim(iset-1)
      END DO
      pbasis%ncgf = 0
      pbasis%nsgf = 0
      DO l = 0, lm
         pbasis%ncgf = pbasis%ncgf+nprim(l)*((l+1)*(l+2))/2
         pbasis%nsgf = pbasis%nsgf+nprim(l)*(2*l+1)
      END DO
      mpgf = MAXVAL(nprim)
      ALLOCATE (pbasis%zet(mpgf, nset))
      pbasis%zet(1:mpgf, 1:nset) = zeta(1:mpgf, 1:nset)

      ALLOCATE (pbasis%l(mpgf, nset), pbasis%n(mpgf, nset))
      DO iset = 1, nset
         DO ip = 1, nprim(iset-1)
            pbasis%l(ip, iset) = iset-1
            pbasis%n(ip, iset) = iset+ip-1
         END DO
      END DO

      ALLOCATE (pbasis%cgf_symbol(pbasis%ncgf))
      ALLOCATE (pbasis%lx(pbasis%ncgf))
      ALLOCATE (pbasis%ly(pbasis%ncgf))
      ALLOCATE (pbasis%lz(pbasis%ncgf))
      ALLOCATE (pbasis%m(pbasis%nsgf))
      ALLOCATE (pbasis%sgf_symbol(pbasis%nsgf))
      ALLOCATE (pbasis%ncgf_set(nset), pbasis%nsgf_set(nset))

      ncgf = 0
      nsgf = 0
      DO iset = 1, nset
         l = iset-1
         pbasis%ncgf_set(iset) = nprim(l)*((l+1)*(l+2))/2
         pbasis%nsgf_set(iset) = nprim(l)*(2*l+1)
         DO ishell = 1, pbasis%nshell(iset)
            lshell = pbasis%l(ishell, iset)
            DO ico = ncoset(lshell-1)+1, ncoset(lshell)
               ncgf = ncgf+1
               pbasis%lx(ncgf) = indco(1, ico)
               pbasis%ly(ncgf) = indco(2, ico)
               pbasis%lz(ncgf) = indco(3, ico)
               pbasis%cgf_symbol(ncgf) = &
                  cgf_symbol(pbasis%n(ishell, iset), (/pbasis%lx(ncgf), pbasis%ly(ncgf), pbasis%lz(ncgf)/))
            END DO
            DO m = -lshell, lshell
               nsgf = nsgf+1
               pbasis%m(nsgf) = m
               pbasis%sgf_symbol(nsgf) = sgf_symbol(pbasis%n(ishell, iset), lshell, m)
            END DO
         END DO
      END DO
      IF(.NOT.(ncgf == pbasis%ncgf))CALL cp__a("aobasis/basis_set_types.F",437)
      IF(.NOT.(nsgf == pbasis%nsgf))CALL cp__a("aobasis/basis_set_types.F",438)

      ALLOCATE (pbasis%gcc(mpgf, mpgf, nset))
      pbasis%gcc = 0.0_dp
      DO iset = 1, nset
         DO i = 1, mpgf
            pbasis%gcc(i, i, iset) = 1.0_dp
         END DO
      END DO

      ALLOCATE (pbasis%first_cgf(mpgf, nset))
      ALLOCATE (pbasis%first_sgf(mpgf, nset))
      ALLOCATE (pbasis%last_cgf(mpgf, nset))
      ALLOCATE (pbasis%last_sgf(mpgf, nset))
      nc = 0
      ns = 0
      maxco = 0
      DO iset = 1, nset
         DO ishell = 1, pbasis%nshell(iset)
            lshell = pbasis%l(ishell, iset)
            pbasis%first_cgf(ishell, iset) = nc+1
            nc = nc+nco(lshell)
            pbasis%last_cgf(ishell, iset) = nc
            pbasis%first_sgf(ishell, iset) = ns+1
            ns = ns+nso(lshell)
            pbasis%last_sgf(ishell, iset) = ns
         END DO
         maxco = MAX(maxco, pbasis%npgf(iset)*ncoset(pbasis%lmax(iset)))
      END DO

      ALLOCATE (pbasis%norm_cgf(ncgf))
      ALLOCATE (pbasis%cphi(maxco, ncgf))
      pbasis%cphi = 0.0_dp
      ALLOCATE (pbasis%sphi(maxco, ncgf))
      pbasis%sphi = 0.0_dp
      ALLOCATE (pbasis%scon(maxco, ncgf))
      pbasis%scon = 0.0_dp
      ALLOCATE (pbasis%set_radius(nset))
      ALLOCATE (pbasis%pgf_radius(mpgf, nset))
      pbasis%pgf_radius = 0.0_dp

      CALL init_orb_basis_set(pbasis)

      DEALLOCATE (zet, zeta, nindex, nprim)

   END SUBROUTINE create_primitive_basis_set

! **************************************************************************************************
!> \brief ...
!> \param basis_set ...
!> \param basis_set_add ...
! **************************************************************************************************
   SUBROUTINE combine_basis_sets(basis_set, basis_set_add)

      ! Combine two Gaussian-type orbital (GTO) basis sets.

      TYPE(gto_basis_set_type), POINTER                  :: basis_set, basis_set_add

      CHARACTER(len=*), PARAMETER :: routineN = 'combine_basis_sets', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=12), DIMENSION(:), POINTER           :: cgf_symbol
      CHARACTER(LEN=6), DIMENSION(:), POINTER            :: sgf_symbol
      INTEGER                                            :: iset, ishell, lshell, maxco, maxpgf, &
                                                            maxshell, nc, ncgf, ncgfn, ncgfo, ns, &
                                                            nset, nsetn, nseto, nsgf, nsgfn, nsgfo
      TYPE(gto_basis_set_type), POINTER                  :: bad, bout

      IF(.NOT.(ASSOCIATED(basis_set)))CALL cp__a("aobasis/basis_set_types.F",506)
      IF(.NOT.(ASSOCIATED(basis_set_add)))CALL cp__a("aobasis/basis_set_types.F",507)

      bad => basis_set_add
      bout => basis_set

      bout%name = bout%name//bad%name
      bout%nset = bout%nset+bad%nset
      bout%ncgf = bout%ncgf+bad%ncgf
      bout%nsgf = bout%nsgf+bad%nsgf
      nset = bout%nset
      ncgf = bout%ncgf
      nsgf = bout%nsgf

      nsetn = bad%nset
      nseto = nset-nsetn
      CALL reallocate(bout%set_radius, 1, nset) ! to be defined later
      CALL reallocate(bout%lmax, 1, nset)
      CALL reallocate(bout%lmin, 1, nset)
      CALL reallocate(bout%npgf, 1, nset)
      CALL reallocate(bout%nshell, 1, nset)
      bout%lmax(nseto+1:nset) = bad%lmax(1:nsetn)
      bout%lmin(nseto+1:nset) = bad%lmin(1:nsetn)
      bout%npgf(nseto+1:nset) = bad%npgf(1:nsetn)
      bout%nshell(nseto+1:nset) = bad%nshell(1:nsetn)
      CALL reallocate(bout%ncgf_set, 1, nset)
      CALL reallocate(bout%nsgf_set, 1, nset)
      bout%ncgf_set(nseto+1:nset) = bad%ncgf_set(1:nsetn)
      bout%nsgf_set(nseto+1:nset) = bad%nsgf_set(1:nsetn)

      nsgfn = bad%nsgf
      nsgfo = nsgf-nsgfn
      ncgfn = bad%ncgf
      ncgfo = ncgf-ncgfn

      ALLOCATE (cgf_symbol(ncgf), sgf_symbol(nsgf))
      cgf_symbol(1:ncgfo) = bout%cgf_symbol(1:ncgfo)
      cgf_symbol(ncgfo+1:ncgf) = bad%cgf_symbol(1:ncgfn)
      sgf_symbol(1:nsgfo) = bout%sgf_symbol(1:nsgfo)
      sgf_symbol(nsgfo+1:nsgf) = bad%sgf_symbol(1:nsgfn)
      DEALLOCATE (bout%cgf_symbol, bout%sgf_symbol)
      ALLOCATE (bout%cgf_symbol(ncgf), bout%sgf_symbol(nsgf))
      bout%cgf_symbol = cgf_symbol
      bout%sgf_symbol = sgf_symbol
      DEALLOCATE (cgf_symbol, sgf_symbol)

      CALL reallocate(bout%lx, 1, ncgf)
      CALL reallocate(bout%ly, 1, ncgf)
      CALL reallocate(bout%lz, 1, ncgf)
      CALL reallocate(bout%m, 1, nsgf)
      bout%lx(ncgfo+1:ncgf) = bad%lx(1:ncgfn)
      bout%ly(ncgfo+1:ncgf) = bad%ly(1:ncgfn)
      bout%lz(ncgfo+1:ncgf) = bad%lz(1:ncgfn)
      bout%m(nsgfo+1:nsgf) = bad%m(1:nsgfn)

      maxpgf = MAXVAL(bout%npgf)
      CALL reallocate(bout%zet, 1, maxpgf, 1, nset)
      bout%zet(:, nseto+1:nset) = bad%zet(:, 1:nsetn)

      maxshell = MAXVAL(bout%nshell)
      CALL reallocate(bout%l, 1, maxshell, 1, nset)
      CALL reallocate(bout%n, 1, maxshell, 1, nset)
      bout%l(:, nseto+1:nset) = bad%l(:, 1:nsetn)
      bout%n(:, nseto+1:nset) = bad%n(:, 1:nsetn)

      CALL reallocate(bout%first_cgf, 1, maxshell, 1, nset)
      CALL reallocate(bout%first_sgf, 1, maxshell, 1, nset)
      CALL reallocate(bout%last_cgf, 1, maxshell, 1, nset)
      CALL reallocate(bout%last_sgf, 1, maxshell, 1, nset)
      nc = 0
      ns = 0
      DO iset = 1, nset
         DO ishell = 1, bout%nshell(iset)
            lshell = bout%l(ishell, iset)
            bout%first_cgf(ishell, iset) = nc+1
            nc = nc+nco(lshell)
            bout%last_cgf(ishell, iset) = nc
            bout%first_sgf(ishell, iset) = ns+1
            ns = ns+nso(lshell)
            bout%last_sgf(ishell, iset) = ns
         END DO
      END DO

      CALL reallocate(bout%gcc, 1, maxpgf, 1, maxshell, 1, nset)
      bout%gcc(:, :, nseto+1:nset) = bad%gcc(:, :, 1:nsetn)

      ! these arrays are determined later using initialization calls
      CALL reallocate(bout%norm_cgf, 1, ncgf)
      maxco = MAX(SIZE(bout%cphi, 1), SIZE(bad%cphi, 1))
      CALL reallocate(bout%cphi, 1, maxco, 1, ncgf)
      CALL reallocate(bout%sphi, 1, maxco, 1, nsgf)
      CALL reallocate(bout%scon, 1, maxco, 1, nsgf)
      CALL reallocate(bout%pgf_radius, 1, maxpgf, 1, nset)

   END SUBROUTINE combine_basis_sets

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
!> \param name ...
!> \param norm_type ...
!> \param kind_radius ...
!> \param ncgf ...
!> \param nset ...
!> \param nsgf ...
!> \param cgf_symbol ...
!> \param sgf_symbol ...
!> \param norm_cgf ...
!> \param set_radius ...
!> \param lmax ...
!> \param lmin ...
!> \param lx ...
!> \param ly ...
!> \param lz ...
!> \param m ...
!> \param ncgf_set ...
!> \param npgf ...
!> \param nsgf_set ...
!> \param nshell ...
!> \param cphi ...
!> \param pgf_radius ...
!> \param sphi ...
!> \param scon ...
!> \param zet ...
!> \param first_cgf ...
!> \param first_sgf ...
!> \param l ...
!> \param last_cgf ...
!> \param last_sgf ...
!> \param n ...
!> \param gcc ...
!> \param maxco ...
!> \param maxl ...
!> \param maxpgf ...
!> \param maxsgf_set ...
!> \param maxshell ...
!> \param maxso ...
!> \param nco_sum ...
!> \param npgf_sum ...
!> \param nshell_sum ...
!> \param maxder ...
!> \param short_kind_radius ...
! **************************************************************************************************
   SUBROUTINE get_gto_basis_set(gto_basis_set, name, norm_type, kind_radius, ncgf, &
                                nset, nsgf, cgf_symbol, sgf_symbol, norm_cgf, set_radius, lmax, lmin, lx, ly, lz, &
                                m, ncgf_set, npgf, nsgf_set, nshell, cphi, pgf_radius, sphi, scon, zet, first_cgf, first_sgf, l, &
                                last_cgf, last_sgf, n, gcc, maxco, maxl, maxpgf, maxsgf_set, maxshell, maxso, nco_sum, &
                                npgf_sum, nshell_sum, maxder, short_kind_radius)

      ! Get informations about a Gaussian-type orbital (GTO) basis set.

      ! - Creation (10.01.2002,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      INTEGER, INTENT(OUT), OPTIONAL                     :: norm_type
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: kind_radius
      INTEGER, INTENT(OUT), OPTIONAL                     :: ncgf, nset, nsgf
      CHARACTER(LEN=12), DIMENSION(:), OPTIONAL, POINTER :: cgf_symbol
      CHARACTER(LEN=6), DIMENSION(:), OPTIONAL, POINTER  :: sgf_symbol
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: norm_cgf, set_radius
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: lmax, lmin, lx, ly, lz, m, ncgf_set, &
                                                            npgf, nsgf_set, nshell
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cphi, pgf_radius, sphi, scon, zet
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: first_cgf, first_sgf, l, last_cgf, &
                                                            last_sgf, n
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: gcc
      INTEGER, INTENT(OUT), OPTIONAL                     :: maxco, maxl, maxpgf, maxsgf_set, &
                                                            maxshell, maxso, nco_sum, npgf_sum, &
                                                            nshell_sum
      INTEGER, INTENT(IN), OPTIONAL                      :: maxder
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: short_kind_radius

      CHARACTER(len=*), PARAMETER :: routineN = 'get_gto_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iset, nder

! -------------------------------------------------------------------------

      IF (ASSOCIATED(gto_basis_set)) THEN
         IF (PRESENT(name)) name = gto_basis_set%name
         IF (PRESENT(norm_type)) norm_type = gto_basis_set%norm_type
         IF (PRESENT(kind_radius)) kind_radius = gto_basis_set%kind_radius
         IF (PRESENT(short_kind_radius)) short_kind_radius = gto_basis_set%short_kind_radius
         IF (PRESENT(ncgf)) ncgf = gto_basis_set%ncgf
         IF (PRESENT(nset)) nset = gto_basis_set%nset
         IF (PRESENT(nsgf)) nsgf = gto_basis_set%nsgf
         IF (PRESENT(cgf_symbol)) cgf_symbol => gto_basis_set%cgf_symbol
         IF (PRESENT(sgf_symbol)) sgf_symbol => gto_basis_set%sgf_symbol
         IF (PRESENT(norm_cgf)) norm_cgf => gto_basis_set%norm_cgf
         IF (PRESENT(set_radius)) set_radius => gto_basis_set%set_radius
         IF (PRESENT(lmax)) lmax => gto_basis_set%lmax
         IF (PRESENT(lmin)) lmin => gto_basis_set%lmin
         IF (PRESENT(lx)) lx => gto_basis_set%lx
         IF (PRESENT(ly)) ly => gto_basis_set%ly
         IF (PRESENT(lz)) lz => gto_basis_set%lz
         IF (PRESENT(m)) m => gto_basis_set%m
         IF (PRESENT(ncgf_set)) ncgf_set => gto_basis_set%ncgf_set
         IF (PRESENT(npgf)) npgf => gto_basis_set%npgf
         IF (PRESENT(nsgf_set)) nsgf_set => gto_basis_set%nsgf_set
         IF (PRESENT(nshell)) nshell => gto_basis_set%nshell
         IF (PRESENT(cphi)) cphi => gto_basis_set%cphi
         IF (PRESENT(pgf_radius)) pgf_radius => gto_basis_set%pgf_radius
         IF (PRESENT(sphi)) sphi => gto_basis_set%sphi
         IF (PRESENT(scon)) scon => gto_basis_set%scon
         IF (PRESENT(zet)) zet => gto_basis_set%zet
         IF (PRESENT(first_cgf)) first_cgf => gto_basis_set%first_cgf
         IF (PRESENT(first_sgf)) first_sgf => gto_basis_set%first_sgf
         IF (PRESENT(l)) l => gto_basis_set%l
         IF (PRESENT(last_cgf)) last_cgf => gto_basis_set%last_cgf
         IF (PRESENT(last_sgf)) last_sgf => gto_basis_set%last_sgf
         IF (PRESENT(n)) n => gto_basis_set%n
         IF (PRESENT(gcc)) gcc => gto_basis_set%gcc
         IF (PRESENT(maxco)) THEN
            maxco = 0
            IF (PRESENT(maxder)) THEN
               nder = maxder
            ELSE
               nder = 0
            END IF
            DO iset = 1, gto_basis_set%nset
               maxco = MAX(maxco, gto_basis_set%npgf(iset)* &
                           ncoset(gto_basis_set%lmax(iset)+nder))
            END DO
         END IF
         IF (PRESENT(maxl)) THEN
            maxl = -1
            DO iset = 1, gto_basis_set%nset
               maxl = MAX(maxl, gto_basis_set%lmax(iset))
            END DO
         END IF
         IF (PRESENT(maxpgf)) THEN
            maxpgf = 0
            DO iset = 1, gto_basis_set%nset
               maxpgf = MAX(maxpgf, gto_basis_set%npgf(iset))
            END DO
         END IF
         IF (PRESENT(maxsgf_set)) THEN
            maxsgf_set = 0
            DO iset = 1, gto_basis_set%nset
               maxsgf_set = MAX(maxsgf_set, gto_basis_set%nsgf_set(iset))
            END DO
         END IF
         IF (PRESENT(maxshell)) THEN ! MAXVAL on structure component avoided
            maxshell = 0
            DO iset = 1, gto_basis_set%nset
               maxshell = MAX(maxshell, gto_basis_set%nshell(iset))
            END DO
         END IF
         IF (PRESENT(maxso)) THEN
            maxso = 0
            DO iset = 1, gto_basis_set%nset
               maxso = MAX(maxso, gto_basis_set%npgf(iset)* &
                           nsoset(gto_basis_set%lmax(iset)))
            END DO
         END IF

         IF (PRESENT(nco_sum)) THEN
            nco_sum = 0
            DO iset = 1, gto_basis_set%nset
               nco_sum = nco_sum+gto_basis_set%npgf(iset)* &
                         ncoset(gto_basis_set%lmax(iset))
            END DO
         END IF
         IF (PRESENT(npgf_sum)) npgf_sum = SUM(gto_basis_set%npgf)
         IF (PRESENT(nshell_sum)) nshell_sum = SUM(gto_basis_set%nshell)
      ELSE
         CALL cp__b("aobasis/basis_set_types.F",776,"The pointer gto_basis_set is not associated")
      END IF

   END SUBROUTINE get_gto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE init_aux_basis_set(gto_basis_set)

      ! Initialise a Gaussian-type orbital (GTO) basis set data set.

      ! - Creation (06.12.2000,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'init_aux_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! -------------------------------------------------------------------------

      IF (.NOT. ASSOCIATED(gto_basis_set)) RETURN

      CALL timeset(routineN, handle)

      SELECT CASE (gto_basis_set%norm_type)
      CASE (0)
         ! No normalisation requested
      CASE (1)
         CALL init_norm_cgf_aux_2(gto_basis_set)
      CASE (2)
         ! WARNING this was never tested
         CALL init_norm_cgf_aux(gto_basis_set)
      CASE DEFAULT
         CALL cp__b("aobasis/basis_set_types.F",813,"Normalization method not specified")
      END SELECT

      ! Initialise the transformation matrices "pgf" -> "cgf"
      CALL init_cphi_and_sphi(gto_basis_set)

      CALL timestop(handle)

   END SUBROUTINE init_aux_basis_set

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE init_cphi_and_sphi(gto_basis_set)

      ! Initialise the matrices for the transformation of primitive Cartesian
      ! Gaussian-type functions to contracted Cartesian (cphi) and spherical
      ! (sphi) Gaussian-type functions.

      ! - Creation (20.09.2000,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'init_cphi_and_sphi', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_cgf, first_sgf, handle, icgf, ico, &
                                                            ipgf, iset, ishell, l, last_sgf, lmax, &
                                                            lmin, n, n1, n2, ncgf, nn, nn1, nn2, &
                                                            npgf, nsgf

! -------------------------------------------------------------------------
! Build the Cartesian transformation matrix "cphi"

      CALL timeset(routineN, handle)

      DO iset = 1, gto_basis_set%nset
         n = ncoset(gto_basis_set%lmax(iset))
         DO ishell = 1, gto_basis_set%nshell(iset)
            DO icgf = gto_basis_set%first_cgf(ishell, iset), &
                      gto_basis_set%last_cgf(ishell, iset)
               ico = coset(gto_basis_set%lx(icgf), &
                           gto_basis_set%ly(icgf), &
                           gto_basis_set%lz(icgf))
               DO ipgf = 1, gto_basis_set%npgf(iset)
                  gto_basis_set%cphi(ico, icgf) = gto_basis_set%norm_cgf(icgf)* &
                                                  gto_basis_set%gcc(ipgf, ishell, iset)
                  ico = ico+n
               END DO
            END DO
         END DO
      END DO

      ! Build the spherical transformation matrix "sphi"

      n = SIZE(gto_basis_set%cphi, 1)

      IF (n .GT. 0) THEN
         DO iset = 1, gto_basis_set%nset
            DO ishell = 1, gto_basis_set%nshell(iset)
               l = gto_basis_set%l(ishell, iset)
               first_cgf = gto_basis_set%first_cgf(ishell, iset)
               first_sgf = gto_basis_set%first_sgf(ishell, iset)
               ncgf = nco(l)
               nsgf = nso(l)
               CALL dgemm("N", "T", n, nsgf, ncgf, &
                          1.0_dp, gto_basis_set%cphi(1, first_cgf), n, &
                          orbtramat(l)%c2s(1, 1), nsgf, &
                          0.0_dp, gto_basis_set%sphi(1, first_sgf), n)
            END DO
         END DO
      ENDIF

      ! Build the reduced transformation matrix "scon"
      ! This matrix transforms from cartesian primitifs to spherical contracted functions
      ! "scon" only includes primitifs (lmin -> lmax), whereas sphi is (0 -> lmax)

      gto_basis_set%scon = 0.0_dp
      IF (n .GT. 0) THEN
         DO iset = 1, gto_basis_set%nset
            lmin = gto_basis_set%lmin(iset)
            lmax = gto_basis_set%lmax(iset)
            npgf = gto_basis_set%npgf(iset)
            nn = ncoset(lmax)-ncoset(lmin-1)
            DO ishell = 1, gto_basis_set%nshell(iset)
               first_sgf = gto_basis_set%first_sgf(ishell, iset)
               last_sgf = gto_basis_set%last_sgf(ishell, iset)
               DO ipgf = 1, npgf
                  nn1 = (ipgf-1)*ncoset(lmax)+ncoset(lmin-1)+1
                  nn2 = ipgf*ncoset(lmax)
                  n1 = (ipgf-1)*nn+1
                  n2 = ipgf*nn
                  gto_basis_set%scon(n1:n2, first_sgf:last_sgf) = gto_basis_set%sphi(nn1:nn2, first_sgf:last_sgf)
               END DO
            END DO
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE init_cphi_and_sphi

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE init_norm_cgf_aux(gto_basis_set)

      ! Initialise the normalization factors of the contracted Cartesian Gaussian
      ! functions, if the Gaussian functions represent charge distributions.

      ! - Creation (07.12.2000,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'init_norm_cgf_aux', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icgf, ico, ipgf, iset, ishell, jco, &
                                                            jpgf, ll, lmax, lmin, lx, ly, lz, n, &
                                                            npgfa
      REAL(KIND=dp)                                      :: fnorm, gcca, gccb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: gaa
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: vv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rpgfa, zeta

! -------------------------------------------------------------------------

      n = 0
      ll = 0
      DO iset = 1, gto_basis_set%nset
         n = MAX(n, gto_basis_set%npgf(iset)*ncoset(gto_basis_set%lmax(iset)))
         ll = MAX(ll, gto_basis_set%lmax(iset))
      END DO

      ALLOCATE (gaa(n, n))
      ALLOCATE (vv(ncoset(ll), ncoset(ll), ll+ll+1))
      ALLOCATE (ff(0:ll+ll))

      DO iset = 1, gto_basis_set%nset
         lmax = gto_basis_set%lmax(iset)
         lmin = gto_basis_set%lmin(iset)
         n = ncoset(lmax)
         npgfa = gto_basis_set%npgf(iset)
         rpgfa => gto_basis_set%pgf_radius(1:npgfa, iset)
         zeta => gto_basis_set%zet(1:npgfa, iset)
         CALL coulomb2(lmax, npgfa, zeta, rpgfa, lmin, &
                       lmax, npgfa, zeta, rpgfa, lmin, &
                       (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, gaa, vv, ff(0:))
         DO ishell = 1, gto_basis_set%nshell(iset)
            DO icgf = gto_basis_set%first_cgf(ishell, iset), &
                      gto_basis_set%last_cgf(ishell, iset)
               lx = gto_basis_set%lx(icgf)
               ly = gto_basis_set%ly(icgf)
               lz = gto_basis_set%lz(icgf)
               ico = coset(lx, ly, lz)
               fnorm = 0.0_dp
               DO ipgf = 1, npgfa
                  gcca = gto_basis_set%gcc(ipgf, ishell, iset)
                  jco = coset(lx, ly, lz)
                  DO jpgf = 1, npgfa
                     gccb = gto_basis_set%gcc(jpgf, ishell, iset)
                     fnorm = fnorm+gcca*gccb*gaa(ico, jco)
                     jco = jco+n
                  END DO
                  ico = ico+n
               END DO
               gto_basis_set%norm_cgf(icgf) = 1.0_dp/SQRT(fnorm)
            END DO
         END DO
      END DO

      DEALLOCATE (vv, ff)
      DEALLOCATE (gaa)

   END SUBROUTINE init_norm_cgf_aux

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE init_norm_cgf_aux_2(gto_basis_set)

      ! Initialise the normalization factors of the auxiliary Cartesian Gaussian
      ! functions (Kim-Gordon polarization basis) Norm = 1.

      ! - Creation (07.12.2000,GT)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      INTEGER                                            :: icgf, iset, ishell

! -------------------------------------------------------------------------

      DO iset = 1, gto_basis_set%nset
         DO ishell = 1, gto_basis_set%nshell(iset)
            DO icgf = gto_basis_set%first_cgf(ishell, iset), &
                      gto_basis_set%last_cgf(ishell, iset)
               gto_basis_set%norm_cgf(icgf) = 1.0_dp
            END DO
         END DO
      END DO

   END SUBROUTINE init_norm_cgf_aux_2

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE init_norm_cgf_orb(gto_basis_set)

      ! Initialise the normalization factors of the contracted Cartesian Gaussian
      ! functions.

      ! - Creation (14.04.2000,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      INTEGER                                            :: icgf, ipgf, iset, ishell, jpgf, l, lx, &
                                                            ly, lz
      REAL(KIND=dp)                                      :: expzet, fnorm, gcca, gccb, prefac, zeta, &
                                                            zetb

! -------------------------------------------------------------------------

      DO iset = 1, gto_basis_set%nset
         DO ishell = 1, gto_basis_set%nshell(iset)

            l = gto_basis_set%l(ishell, iset)

            expzet = 0.5_dp*REAL(2*l+3, dp)

            fnorm = 0.0_dp

            DO ipgf = 1, gto_basis_set%npgf(iset)
               gcca = gto_basis_set%gcc(ipgf, ishell, iset)
               zeta = gto_basis_set%zet(ipgf, iset)
               DO jpgf = 1, gto_basis_set%npgf(iset)
                  gccb = gto_basis_set%gcc(jpgf, ishell, iset)
                  zetb = gto_basis_set%zet(jpgf, iset)
                  fnorm = fnorm+gcca*gccb/(zeta+zetb)**expzet
               END DO
            END DO

            fnorm = 0.5_dp**l*pi**1.5_dp*fnorm

            DO icgf = gto_basis_set%first_cgf(ishell, iset), &
                      gto_basis_set%last_cgf(ishell, iset)
               lx = gto_basis_set%lx(icgf)
               ly = gto_basis_set%ly(icgf)
               lz = gto_basis_set%lz(icgf)
               prefac = dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1)
               gto_basis_set%norm_cgf(icgf) = 1.0_dp/SQRT(prefac*fnorm)
            END DO

         END DO
      END DO

   END SUBROUTINE init_norm_cgf_orb

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE init_norm_cgf_orb_den(gto_basis_set)

      ! Initialise the normalization factors of the contracted Cartesian Gaussian
      ! functions used for frozen density representation.

      ! - Creation (21.09.2002,GT)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      INTEGER                                            :: icgf, ipgf, iset, ishell, l
      REAL(KIND=dp)                                      :: expzet, gcca, prefac, zeta

! -------------------------------------------------------------------------

      DO iset = 1, gto_basis_set%nset
         DO ishell = 1, gto_basis_set%nshell(iset)
            l = gto_basis_set%l(ishell, iset)
            expzet = 0.5_dp*REAL(2*l+3, dp)
            prefac = (1.0_dp/pi)**1.5_dp
            DO ipgf = 1, gto_basis_set%npgf(iset)
               gcca = gto_basis_set%gcc(ipgf, ishell, iset)
               zeta = gto_basis_set%zet(ipgf, iset)
               gto_basis_set%gcc(ipgf, ishell, iset) = prefac*zeta**expzet*gcca
            END DO
            DO icgf = gto_basis_set%first_cgf(ishell, iset), &
                      gto_basis_set%last_cgf(ishell, iset)
               gto_basis_set%norm_cgf(icgf) = 1.0_dp
            END DO
         END DO
      END DO

   END SUBROUTINE init_norm_cgf_orb_den

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE init_orb_basis_set(gto_basis_set)

      ! Initialise a Gaussian-type orbital (GTO) basis set data set.

      ! - Creation (26.10.2000,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'init_orb_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! -------------------------------------------------------------------------

      IF (.NOT. ASSOCIATED(gto_basis_set)) RETURN

      CALL timeset(routineN, handle)

      SELECT CASE (gto_basis_set%norm_type)
      CASE (0)
         ! No normalisation requested
      CASE (1)
         CALL init_norm_cgf_orb_den(gto_basis_set)
      CASE (2)
         ! Normalise the primitive Gaussian functions
         CALL normalise_gcc_orb(gto_basis_set)
         ! Compute the normalization factors of the contracted Gaussian-type
         ! functions
         CALL init_norm_cgf_orb(gto_basis_set)
      CASE DEFAULT
         CALL cp__b("aobasis/basis_set_types.F",1147,"Normalization method not specified")
      END SELECT

      ! Initialise the transformation matrices "pgf" -> "cgf"

      CALL init_cphi_and_sphi(gto_basis_set)

      CALL timestop(handle)

   END SUBROUTINE init_orb_basis_set

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
! **************************************************************************************************
   SUBROUTINE normalise_gcc_orb(gto_basis_set)

      ! Normalise the primitive Cartesian Gaussian functions. The normalization
      ! factor is included in the Gaussian contraction coefficients.

      ! - Creation (20.08.1999,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set

      INTEGER                                            :: ipgf, iset, ishell, l
      REAL(KIND=dp)                                      :: expzet, gcca, prefac, zeta

! -------------------------------------------------------------------------

      DO iset = 1, gto_basis_set%nset
         DO ishell = 1, gto_basis_set%nshell(iset)
            l = gto_basis_set%l(ishell, iset)
            expzet = 0.25_dp*REAL(2*l+3, dp)
            prefac = 2.0_dp**l*(2.0_dp/pi)**0.75_dp
            DO ipgf = 1, gto_basis_set%npgf(iset)
               gcca = gto_basis_set%gcc(ipgf, ishell, iset)
               zeta = gto_basis_set%zet(ipgf, iset)
               gto_basis_set%gcc(ipgf, ishell, iset) = prefac*zeta**expzet*gcca
            END DO
         END DO
      END DO

   END SUBROUTINE normalise_gcc_orb

! **************************************************************************************************
!> \brief ...
!> \param element_symbol ...
!> \param basis_set_name ...
!> \param gto_basis_set ...
!> \param para_env ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE read_gto_basis_set1(element_symbol, basis_set_name, gto_basis_set, &
                                  para_env, dft_section)

      ! Read a Gaussian-type orbital (GTO) basis set from the database file.

      ! - Creation (13.04.2000,MK)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol, basis_set_name
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_gto_basis_set1', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(len=default_path_length)                 :: basis_set_file_name, tmp
      CHARACTER(LEN=default_path_length), DIMENSION(:), &
         POINTER                                         :: cbasis
      CHARACTER(LEN=LEN(basis_set_name))                 :: bsname
      CHARACTER(LEN=LEN(basis_set_name)+2)               :: bsname2
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(element_symbol)+2)               :: symbol2
      INTEGER                                            :: i, ibasis, ico, ipgf, irep, iset, &
                                                            ishell, lshell, m, maxco, maxl, &
                                                            maxpgf, maxshell, nbasis, ncgf, nmin, &
                                                            nset, nsgf, strlen1, strlen2
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l, n
      LOGICAL                                            :: basis_found, found, match
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc
      TYPE(cp_parser_type), POINTER                      :: parser

      line = ""
      line2 = ""
      symbol = ""
      symbol2 = ""
      bsname = ""
      bsname2 = ""

      nbasis = 1

      gto_basis_set%name = basis_set_name
      CALL section_vals_val_get(dft_section, "BASIS_SET_FILE_NAME", &
                                n_rep_val=nbasis)
      ALLOCATE (cbasis(nbasis))
      DO ibasis = 1, nbasis
         CALL section_vals_val_get(dft_section, "BASIS_SET_FILE_NAME", &
                                   i_rep_val=ibasis, c_val=cbasis(ibasis))
         NULLIFY (parser)
         basis_set_file_name = cbasis(ibasis)
         tmp = basis_set_file_name
         CALL uppercase(tmp)
         IF (INDEX(tmp, "MOLOPT") .NE. 0) CALL cite_reference(VandeVondele2007)
      END DO

      ! Search for the requested basis set in the basis set file
      ! until the basis set is found or the end of file is reached

      basis_found = .FALSE.
      basis_loop: DO ibasis = 1, nbasis
         IF (basis_found) EXIT basis_loop
         NULLIFY (parser)
         basis_set_file_name = cbasis(ibasis)
         CALL parser_create(parser, basis_set_file_name, para_env=para_env)

         bsname = basis_set_name
         symbol = element_symbol
         irep = 0

         tmp = basis_set_name
         CALL uppercase(tmp)
         IF (INDEX(tmp, "MOLOPT") .NE. 0) CALL cite_reference(VandeVondele2007)

         nset = 0
         maxshell = 0
         maxpgf = 0
         maxco = 0
         ncgf = 0
         nsgf = 0
         gto_basis_set%nset = nset
         gto_basis_set%ncgf = ncgf
         gto_basis_set%nsgf = nsgf
         CALL reallocate(gto_basis_set%lmax, 1, nset)
         CALL reallocate(gto_basis_set%lmin, 1, nset)
         CALL reallocate(gto_basis_set%npgf, 1, nset)
         CALL reallocate(gto_basis_set%nshell, 1, nset)
         CALL reallocate(gto_basis_set%n, 1, maxshell, 1, nset)
         CALL reallocate(gto_basis_set%l, 1, maxshell, 1, nset)
         CALL reallocate(gto_basis_set%zet, 1, maxpgf, 1, nset)
         CALL reallocate(gto_basis_set%gcc, 1, maxpgf, 1, maxshell, 1, nset)
         CALL reallocate(gto_basis_set%set_radius, 1, nset)
         CALL reallocate(gto_basis_set%pgf_radius, 1, maxpgf, 1, nset)
         CALL reallocate(gto_basis_set%first_cgf, 1, maxshell, 1, nset)
         CALL reallocate(gto_basis_set%first_sgf, 1, maxshell, 1, nset)
         CALL reallocate(gto_basis_set%last_cgf, 1, maxshell, 1, nset)
         CALL reallocate(gto_basis_set%last_sgf, 1, maxshell, 1, nset)
         CALL reallocate(gto_basis_set%ncgf_set, 1, nset)
         CALL reallocate(gto_basis_set%nsgf_set, 1, nset)
         CALL reallocate(gto_basis_set%cphi, 1, maxco, 1, ncgf)
         CALL reallocate(gto_basis_set%sphi, 1, maxco, 1, nsgf)
         CALL reallocate(gto_basis_set%scon, 1, maxco, 1, nsgf)
         CALL reallocate(gto_basis_set%lx, 1, ncgf)
         CALL reallocate(gto_basis_set%ly, 1, ncgf)
         CALL reallocate(gto_basis_set%lz, 1, ncgf)
         CALL reallocate(gto_basis_set%m, 1, nsgf)
         CALL reallocate(gto_basis_set%norm_cgf, 1, ncgf)

         IF (tmp .NE. "NONE") THEN
            search_loop: DO

               CALL parser_search_string(parser, TRIM(bsname), .TRUE., found, line)
               IF (found) THEN
                  CALL uppercase(symbol)
                  CALL uppercase(bsname)

                  match = .FALSE.
                  CALL uppercase(line)
                  ! Check both the element symbol and the basis set name
                  line2 = " "//line//" "
                  symbol2 = " "//TRIM(symbol)//" "
                  bsname2 = " "//TRIM(bsname)//" "
                  strlen1 = LEN_TRIM(symbol2)+1
                  strlen2 = LEN_TRIM(bsname2)+1

                  IF ((INDEX(line2, symbol2(:strlen1)) > 0) .AND. &
                      (INDEX(line2, bsname2(:strlen2)) > 0)) match = .TRUE.
                  IF (match) THEN
                     NULLIFY (gcc, l, lmax, lmin, n, npgf, nshell, zet)
                     ! Read the basis set information
                     CALL parser_get_object(parser, nset, newline=.TRUE.)

                     CALL reallocate(npgf, 1, nset)
                     CALL reallocate(nshell, 1, nset)
                     CALL reallocate(lmax, 1, nset)
                     CALL reallocate(lmin, 1, nset)
                     CALL reallocate(n, 1, 1, 1, nset)

                     maxl = 0
                     maxpgf = 0
                     maxshell = 0

                     DO iset = 1, nset
                        CALL parser_get_object(parser, n(1, iset), newline=.TRUE.)
                        CALL parser_get_object(parser, lmin(iset))
                        CALL parser_get_object(parser, lmax(iset))
                        CALL parser_get_object(parser, npgf(iset))
                        maxl = MAX(maxl, lmax(iset))
                        IF (npgf(iset) > maxpgf) THEN
                           maxpgf = npgf(iset)
                           CALL reallocate(zet, 1, maxpgf, 1, nset)
                           CALL reallocate(gcc, 1, maxpgf, 1, maxshell, 1, nset)
                        END IF
                        nshell(iset) = 0
                        DO lshell = lmin(iset), lmax(iset)
                           nmin = n(1, iset)+lshell-lmin(iset)
                           CALL parser_get_object(parser, ishell)
                           nshell(iset) = nshell(iset)+ishell
                           IF (nshell(iset) > maxshell) THEN
                              maxshell = nshell(iset)
                              CALL reallocate(n, 1, maxshell, 1, nset)
                              CALL reallocate(l, 1, maxshell, 1, nset)
                              CALL reallocate(gcc, 1, maxpgf, 1, maxshell, 1, nset)
                           END IF
                           DO i = 1, ishell
                              n(nshell(iset)-ishell+i, iset) = nmin+i-1
                              l(nshell(iset)-ishell+i, iset) = lshell
                           END DO
                        END DO
                        DO ipgf = 1, npgf(iset)
                           CALL parser_get_object(parser, zet(ipgf, iset), newline=.TRUE.)
                           DO ishell = 1, nshell(iset)
                              CALL parser_get_object(parser, gcc(ipgf, ishell, iset))
                           END DO
                        END DO
                     END DO

                     ! Maximum angular momentum quantum number of the atomic kind

                     CALL init_orbital_pointers(maxl)

                     ! Allocate the global variables

                     gto_basis_set%nset = nset
                     CALL reallocate(gto_basis_set%lmax, 1, nset)
                     CALL reallocate(gto_basis_set%lmin, 1, nset)
                     CALL reallocate(gto_basis_set%npgf, 1, nset)
                     CALL reallocate(gto_basis_set%nshell, 1, nset)
                     CALL reallocate(gto_basis_set%n, 1, maxshell, 1, nset)
                     CALL reallocate(gto_basis_set%l, 1, maxshell, 1, nset)
                     CALL reallocate(gto_basis_set%zet, 1, maxpgf, 1, nset)
                     CALL reallocate(gto_basis_set%gcc, 1, maxpgf, 1, maxshell, 1, nset)

                     ! Copy the basis set information into the data structure

                     DO iset = 1, nset
                        gto_basis_set%lmax(iset) = lmax(iset)
                        gto_basis_set%lmin(iset) = lmin(iset)
                        gto_basis_set%npgf(iset) = npgf(iset)
                        gto_basis_set%nshell(iset) = nshell(iset)
                        DO ishell = 1, nshell(iset)
                           gto_basis_set%n(ishell, iset) = n(ishell, iset)
                           gto_basis_set%l(ishell, iset) = l(ishell, iset)
                           DO ipgf = 1, npgf(iset)
                              gto_basis_set%gcc(ipgf, ishell, iset) = gcc(ipgf, ishell, iset)
                           END DO
                        END DO
                        DO ipgf = 1, npgf(iset)
                           gto_basis_set%zet(ipgf, iset) = zet(ipgf, iset)
                        END DO
                     END DO

                     ! Initialise the depending atomic kind information

                     CALL reallocate(gto_basis_set%set_radius, 1, nset)
                     CALL reallocate(gto_basis_set%pgf_radius, 1, maxpgf, 1, nset)
                     CALL reallocate(gto_basis_set%first_cgf, 1, maxshell, 1, nset)
                     CALL reallocate(gto_basis_set%first_sgf, 1, maxshell, 1, nset)
                     CALL reallocate(gto_basis_set%last_cgf, 1, maxshell, 1, nset)
                     CALL reallocate(gto_basis_set%last_sgf, 1, maxshell, 1, nset)
                     CALL reallocate(gto_basis_set%ncgf_set, 1, nset)
                     CALL reallocate(gto_basis_set%nsgf_set, 1, nset)

                     maxco = 0
                     ncgf = 0
                     nsgf = 0

                     DO iset = 1, nset
                        gto_basis_set%ncgf_set(iset) = 0
                        gto_basis_set%nsgf_set(iset) = 0
                        DO ishell = 1, nshell(iset)
                           lshell = gto_basis_set%l(ishell, iset)
                           gto_basis_set%first_cgf(ishell, iset) = ncgf+1
                           ncgf = ncgf+nco(lshell)
                           gto_basis_set%last_cgf(ishell, iset) = ncgf
                           gto_basis_set%ncgf_set(iset) = &
                              gto_basis_set%ncgf_set(iset)+nco(lshell)
                           gto_basis_set%first_sgf(ishell, iset) = nsgf+1
                           nsgf = nsgf+nso(lshell)
                           gto_basis_set%last_sgf(ishell, iset) = nsgf
                           gto_basis_set%nsgf_set(iset) = &
                              gto_basis_set%nsgf_set(iset)+nso(lshell)
                        END DO
                        maxco = MAX(maxco, npgf(iset)*ncoset(lmax(iset)))
                     END DO

                     gto_basis_set%ncgf = ncgf
                     gto_basis_set%nsgf = nsgf

                     CALL reallocate(gto_basis_set%cphi, 1, maxco, 1, ncgf)
                     CALL reallocate(gto_basis_set%sphi, 1, maxco, 1, nsgf)
                     CALL reallocate(gto_basis_set%scon, 1, maxco, 1, nsgf)
                     CALL reallocate(gto_basis_set%lx, 1, ncgf)
                     CALL reallocate(gto_basis_set%ly, 1, ncgf)
                     CALL reallocate(gto_basis_set%lz, 1, ncgf)
                     CALL reallocate(gto_basis_set%m, 1, nsgf)
                     CALL reallocate(gto_basis_set%norm_cgf, 1, ncgf)
                     ALLOCATE (gto_basis_set%cgf_symbol(ncgf))

                     ALLOCATE (gto_basis_set%sgf_symbol(nsgf))

                     ncgf = 0
                     nsgf = 0

                     DO iset = 1, nset
                        DO ishell = 1, nshell(iset)
                           lshell = gto_basis_set%l(ishell, iset)
                           DO ico = ncoset(lshell-1)+1, ncoset(lshell)
                              ncgf = ncgf+1
                              gto_basis_set%lx(ncgf) = indco(1, ico)
                              gto_basis_set%ly(ncgf) = indco(2, ico)
                              gto_basis_set%lz(ncgf) = indco(3, ico)
                              gto_basis_set%cgf_symbol(ncgf) = &
                                 cgf_symbol(n(ishell, iset), (/gto_basis_set%lx(ncgf), &
                                                               gto_basis_set%ly(ncgf), &
                                                               gto_basis_set%lz(ncgf)/))
                           END DO
                           DO m = -lshell, lshell
                              nsgf = nsgf+1
                              gto_basis_set%m(nsgf) = m
                              gto_basis_set%sgf_symbol(nsgf) = &
                                 sgf_symbol(n(ishell, iset), lshell, m)
                           END DO
                        END DO
                     END DO

                     DEALLOCATE (gcc, l, lmax, lmin, n, npgf, nshell, zet)

                     basis_found = .TRUE.
                     EXIT search_loop
                  END IF
               ELSE
                  EXIT search_loop
               END IF
            END DO search_loop
         ELSE
            match = .FALSE.
            ALLOCATE (gto_basis_set%cgf_symbol(ncgf))
            ALLOCATE (gto_basis_set%sgf_symbol(nsgf))
         ENDIF

         CALL parser_release(parser)

      END DO basis_loop

      IF (tmp .NE. "NONE") THEN
         IF (.NOT. basis_found) THEN
            basis_set_file_name = ""
            DO ibasis = 1, nbasis
               basis_set_file_name = TRIM(basis_set_file_name)//"<"//TRIM(cbasis(ibasis))//"> "
            END DO
            CALL cp_abort(cp__l("aobasis/basis_set_types.F",1512), &
                          "The requested basis set <"//TRIM(bsname)// &
                          "> for element <"//TRIM(symbol)//"> was not "// &
                          "found in the basis set files "// &
                          TRIM(basis_set_file_name))
         END IF
      END IF
      DEALLOCATE (cbasis)

   END SUBROUTINE read_gto_basis_set1

! **************************************************************************************************
!> \brief ...
!> \param element_symbol ...
!> \param basis_type ...
!> \param gto_basis_set ...
!> \param basis_section ...
!> \param irep ...
! **************************************************************************************************
   SUBROUTINE read_gto_basis_set2(element_symbol, basis_type, gto_basis_set, &
                                  basis_section, irep)

      ! Read a Gaussian-type orbital (GTO) basis set from the database file.

      ! - Creation (13.04.2000,MK)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol
      CHARACTER(LEN=*), INTENT(INOUT)                    :: basis_type
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      TYPE(section_vals_type), OPTIONAL, POINTER         :: basis_section
      INTEGER                                            :: irep

      CHARACTER(len=*), PARAMETER :: routineN = 'read_gto_basis_set2', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=20*default_string_length)            :: line_att
      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(LEN=default_path_length)                 :: bsname, bsname2
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(element_symbol)+2)               :: symbol2
      INTEGER                                            :: i, ico, ipgf, iset, ishell, lshell, m, &
                                                            maxco, maxl, maxpgf, maxshell, ncgf, &
                                                            nmin, nset, nsgf
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l, n
      LOGICAL                                            :: is_ok
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(val_type), POINTER                            :: val

      line = ""
      line2 = ""
      symbol = ""
      symbol2 = ""
      bsname = ""
      bsname2 = ""

      gto_basis_set%name = " "

      bsname = " "
      symbol = element_symbol

      nset = 0
      maxshell = 0
      maxpgf = 0
      maxco = 0
      ncgf = 0
      nsgf = 0
      gto_basis_set%nset = nset
      gto_basis_set%ncgf = ncgf
      gto_basis_set%nsgf = nsgf
      CALL reallocate(gto_basis_set%lmax, 1, nset)
      CALL reallocate(gto_basis_set%lmin, 1, nset)
      CALL reallocate(gto_basis_set%npgf, 1, nset)
      CALL reallocate(gto_basis_set%nshell, 1, nset)
      CALL reallocate(gto_basis_set%n, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%l, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%zet, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%gcc, 1, maxpgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%set_radius, 1, nset)
      CALL reallocate(gto_basis_set%pgf_radius, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%first_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%first_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%ncgf_set, 1, nset)
      CALL reallocate(gto_basis_set%nsgf_set, 1, nset)
      CALL reallocate(gto_basis_set%cphi, 1, maxco, 1, ncgf)
      CALL reallocate(gto_basis_set%sphi, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%scon, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%lx, 1, ncgf)
      CALL reallocate(gto_basis_set%ly, 1, ncgf)
      CALL reallocate(gto_basis_set%lz, 1, ncgf)
      CALL reallocate(gto_basis_set%m, 1, nsgf)
      CALL reallocate(gto_basis_set%norm_cgf, 1, ncgf)

      basis_type = ""
      CALL section_vals_val_get(basis_section, "_SECTION_PARAMETERS_", i_rep_val=irep, c_val=basis_type)
      IF (basis_type == "Orbital") basis_type = "ORB"

      NULLIFY (list, val)
      CALL section_vals_list_get(basis_section, "_DEFAULT_KEYWORD_", i_rep_section=irep, list=list)
      CALL uppercase(symbol)
      CALL uppercase(bsname)

      NULLIFY (gcc, l, lmax, lmin, n, npgf, nshell, zet)
      ! Read the basis set information
      is_ok = cp_sll_val_next(list, val)
      IF (.NOT. is_ok) CALL cp__b("aobasis/basis_set_types.F",1622,"Error reading the Basis set from input file!!")
      CALL val_get(val, c_val=line_att)
      READ (line_att, *) nset

      CALL reallocate(npgf, 1, nset)
      CALL reallocate(nshell, 1, nset)
      CALL reallocate(lmax, 1, nset)
      CALL reallocate(lmin, 1, nset)
      CALL reallocate(n, 1, 1, 1, nset)

      maxl = 0
      maxpgf = 0
      maxshell = 0

      DO iset = 1, nset
         is_ok = cp_sll_val_next(list, val)
         IF (.NOT. is_ok) CALL cp__b("aobasis/basis_set_types.F",1638,"Error reading the Basis set from input file!!")
         CALL val_get(val, c_val=line_att)
         READ (line_att, *) n(1, iset)
         CALL remove_word(line_att)
         READ (line_att, *) lmin(iset)
         CALL remove_word(line_att)
         READ (line_att, *) lmax(iset)
         CALL remove_word(line_att)
         READ (line_att, *) npgf(iset)
         CALL remove_word(line_att)
         maxl = MAX(maxl, lmax(iset))
         IF (npgf(iset) > maxpgf) THEN
            maxpgf = npgf(iset)
            CALL reallocate(zet, 1, maxpgf, 1, nset)
            CALL reallocate(gcc, 1, maxpgf, 1, maxshell, 1, nset)
         END IF
         nshell(iset) = 0
         DO lshell = lmin(iset), lmax(iset)
            nmin = n(1, iset)+lshell-lmin(iset)
            READ (line_att, *) ishell
            CALL remove_word(line_att)
            nshell(iset) = nshell(iset)+ishell
            IF (nshell(iset) > maxshell) THEN
               maxshell = nshell(iset)
               CALL reallocate(n, 1, maxshell, 1, nset)
               CALL reallocate(l, 1, maxshell, 1, nset)
               CALL reallocate(gcc, 1, maxpgf, 1, maxshell, 1, nset)
            END IF
            DO i = 1, ishell
               n(nshell(iset)-ishell+i, iset) = nmin+i-1
               l(nshell(iset)-ishell+i, iset) = lshell
            END DO
         END DO
         IF (LEN_TRIM(line_att) /= 0) &
            CALL cp__b("aobasis/basis_set_types.F",1672,"Error reading the Basis from input file!!")
         DO ipgf = 1, npgf(iset)
            is_ok = cp_sll_val_next(list, val)
            IF (.NOT. is_ok) CALL cp__b("aobasis/basis_set_types.F",1675,"Error reading the Basis set from input file!!")
            CALL val_get(val, c_val=line_att)
            READ (line_att, *) zet(ipgf, iset), (gcc(ipgf, ishell, iset), ishell=1, nshell(iset))
         END DO
      END DO

      ! Maximum angular momentum quantum number of the atomic kind

      CALL init_orbital_pointers(maxl)

      ! Allocate the global variables

      gto_basis_set%nset = nset
      CALL reallocate(gto_basis_set%lmax, 1, nset)
      CALL reallocate(gto_basis_set%lmin, 1, nset)
      CALL reallocate(gto_basis_set%npgf, 1, nset)
      CALL reallocate(gto_basis_set%nshell, 1, nset)
      CALL reallocate(gto_basis_set%n, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%l, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%zet, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%gcc, 1, maxpgf, 1, maxshell, 1, nset)

      ! Copy the basis set information into the data structure

      DO iset = 1, nset
         gto_basis_set%lmax(iset) = lmax(iset)
         gto_basis_set%lmin(iset) = lmin(iset)
         gto_basis_set%npgf(iset) = npgf(iset)
         gto_basis_set%nshell(iset) = nshell(iset)
         DO ishell = 1, nshell(iset)
            gto_basis_set%n(ishell, iset) = n(ishell, iset)
            gto_basis_set%l(ishell, iset) = l(ishell, iset)
            DO ipgf = 1, npgf(iset)
               gto_basis_set%gcc(ipgf, ishell, iset) = gcc(ipgf, ishell, iset)
            END DO
         END DO
         DO ipgf = 1, npgf(iset)
            gto_basis_set%zet(ipgf, iset) = zet(ipgf, iset)
         END DO
      END DO

      ! Initialise the depending atomic kind information

      CALL reallocate(gto_basis_set%set_radius, 1, nset)
      CALL reallocate(gto_basis_set%pgf_radius, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%first_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%first_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%ncgf_set, 1, nset)
      CALL reallocate(gto_basis_set%nsgf_set, 1, nset)

      maxco = 0
      ncgf = 0
      nsgf = 0

      DO iset = 1, nset
         gto_basis_set%ncgf_set(iset) = 0
         gto_basis_set%nsgf_set(iset) = 0
         DO ishell = 1, nshell(iset)
            lshell = gto_basis_set%l(ishell, iset)
            gto_basis_set%first_cgf(ishell, iset) = ncgf+1
            ncgf = ncgf+nco(lshell)
            gto_basis_set%last_cgf(ishell, iset) = ncgf
            gto_basis_set%ncgf_set(iset) = &
               gto_basis_set%ncgf_set(iset)+nco(lshell)
            gto_basis_set%first_sgf(ishell, iset) = nsgf+1
            nsgf = nsgf+nso(lshell)
            gto_basis_set%last_sgf(ishell, iset) = nsgf
            gto_basis_set%nsgf_set(iset) = &
               gto_basis_set%nsgf_set(iset)+nso(lshell)
         END DO
         maxco = MAX(maxco, npgf(iset)*ncoset(lmax(iset)))
      END DO

      gto_basis_set%ncgf = ncgf
      gto_basis_set%nsgf = nsgf

      CALL reallocate(gto_basis_set%cphi, 1, maxco, 1, ncgf)
      CALL reallocate(gto_basis_set%sphi, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%scon, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%lx, 1, ncgf)
      CALL reallocate(gto_basis_set%ly, 1, ncgf)
      CALL reallocate(gto_basis_set%lz, 1, ncgf)
      CALL reallocate(gto_basis_set%m, 1, nsgf)
      CALL reallocate(gto_basis_set%norm_cgf, 1, ncgf)
      ALLOCATE (gto_basis_set%cgf_symbol(ncgf))

      ALLOCATE (gto_basis_set%sgf_symbol(nsgf))

      ncgf = 0
      nsgf = 0

      DO iset = 1, nset
         DO ishell = 1, nshell(iset)
            lshell = gto_basis_set%l(ishell, iset)
            DO ico = ncoset(lshell-1)+1, ncoset(lshell)
               ncgf = ncgf+1
               gto_basis_set%lx(ncgf) = indco(1, ico)
               gto_basis_set%ly(ncgf) = indco(2, ico)
               gto_basis_set%lz(ncgf) = indco(3, ico)
               gto_basis_set%cgf_symbol(ncgf) = &
                  cgf_symbol(n(ishell, iset), (/gto_basis_set%lx(ncgf), &
                                                gto_basis_set%ly(ncgf), &
                                                gto_basis_set%lz(ncgf)/))
            END DO
            DO m = -lshell, lshell
               nsgf = nsgf+1
               gto_basis_set%m(nsgf) = m
               gto_basis_set%sgf_symbol(nsgf) = &
                  sgf_symbol(n(ishell, iset), lshell, m)
            END DO
         END DO
      END DO

      DEALLOCATE (gcc, l, lmax, lmin, n, npgf, nshell, zet)

   END SUBROUTINE read_gto_basis_set2

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
!> \param name ...
!> \param norm_type ...
!> \param kind_radius ...
!> \param ncgf ...
!> \param nset ...
!> \param nsgf ...
!> \param cgf_symbol ...
!> \param sgf_symbol ...
!> \param norm_cgf ...
!> \param set_radius ...
!> \param lmax ...
!> \param lmin ...
!> \param lx ...
!> \param ly ...
!> \param lz ...
!> \param m ...
!> \param ncgf_set ...
!> \param npgf ...
!> \param nsgf_set ...
!> \param nshell ...
!> \param cphi ...
!> \param pgf_radius ...
!> \param sphi ...
!> \param scon ...
!> \param zet ...
!> \param first_cgf ...
!> \param first_sgf ...
!> \param l ...
!> \param last_cgf ...
!> \param last_sgf ...
!> \param n ...
!> \param gcc ...
!> \param short_kind_radius ...
! **************************************************************************************************
   SUBROUTINE set_gto_basis_set(gto_basis_set, name, norm_type, kind_radius, ncgf, &
                                nset, nsgf, cgf_symbol, sgf_symbol, norm_cgf, set_radius, lmax, &
                                lmin, lx, ly, lz, m, ncgf_set, npgf, nsgf_set, nshell, &
                                cphi, pgf_radius, sphi, scon, zet, first_cgf, first_sgf, l, &
                                last_cgf, last_sgf, n, gcc, short_kind_radius)

      ! Set the components of Gaussian-type orbital (GTO) basis set data set.

      ! - Creation (10.01.2002,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: name
      INTEGER, INTENT(IN), OPTIONAL                      :: norm_type
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: kind_radius
      INTEGER, INTENT(IN), OPTIONAL                      :: ncgf, nset, nsgf
      CHARACTER(LEN=12), DIMENSION(:), OPTIONAL, POINTER :: cgf_symbol
      CHARACTER(LEN=6), DIMENSION(:), OPTIONAL, POINTER  :: sgf_symbol
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: norm_cgf, set_radius
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: lmax, lmin, lx, ly, lz, m, ncgf_set, &
                                                            npgf, nsgf_set, nshell
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cphi, pgf_radius, sphi, scon, zet
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: first_cgf, first_sgf, l, last_cgf, &
                                                            last_sgf, n
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: gcc
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: short_kind_radius

      CHARACTER(len=*), PARAMETER :: routineN = 'set_gto_basis_set', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF (ASSOCIATED(gto_basis_set)) THEN
         IF (PRESENT(name)) gto_basis_set%name = name
         IF (PRESENT(norm_type)) gto_basis_set%norm_type = norm_type
         IF (PRESENT(kind_radius)) gto_basis_set%kind_radius = kind_radius
         IF (PRESENT(short_kind_radius)) gto_basis_set%short_kind_radius = short_kind_radius
         IF (PRESENT(ncgf)) gto_basis_set%ncgf = ncgf
         IF (PRESENT(nset)) gto_basis_set%nset = nset
         IF (PRESENT(nsgf)) gto_basis_set%nsgf = nsgf
         IF (PRESENT(cgf_symbol)) gto_basis_set%cgf_symbol(:) = cgf_symbol(:)
         IF (PRESENT(sgf_symbol)) gto_basis_set%sgf_symbol(:) = sgf_symbol(:)
         IF (PRESENT(norm_cgf)) gto_basis_set%norm_cgf(:) = norm_cgf(:)
         IF (PRESENT(set_radius)) gto_basis_set%set_radius(:) = set_radius(:)
         IF (PRESENT(lmax)) gto_basis_set%lmax(:) = lmax(:)
         IF (PRESENT(lmin)) gto_basis_set%lmin(:) = lmin(:)
         IF (PRESENT(lx)) gto_basis_set%lx(:) = lx(:)
         IF (PRESENT(ly)) gto_basis_set%ly(:) = ly(:)
         IF (PRESENT(lz)) gto_basis_set%lz(:) = lz(:)
         IF (PRESENT(m)) gto_basis_set%m(:) = m(:)
         IF (PRESENT(ncgf_set)) gto_basis_set%ncgf_set(:) = ncgf_set(:)
         IF (PRESENT(npgf)) gto_basis_set%npgf(:) = npgf(:)
         IF (PRESENT(nsgf_set)) gto_basis_set%nsgf_set(:) = nsgf_set(:)
         IF (PRESENT(nshell)) gto_basis_set%nshell(:) = nshell(:)
         IF (PRESENT(cphi)) gto_basis_set%cphi(:, :) = cphi(:, :)
         IF (PRESENT(pgf_radius)) gto_basis_set%pgf_radius(:, :) = pgf_radius(:, :)
         IF (PRESENT(sphi)) gto_basis_set%sphi(:, :) = sphi(:, :)
         IF (PRESENT(scon)) gto_basis_set%scon(:, :) = scon(:, :)
         IF (PRESENT(zet)) gto_basis_set%zet(:, :) = zet(:, :)
         IF (PRESENT(first_cgf)) gto_basis_set%first_cgf(:, :) = first_cgf(:, :)
         IF (PRESENT(first_sgf)) gto_basis_set%first_sgf(:, :) = first_sgf(:, :)
         IF (PRESENT(l)) l(:, :) = gto_basis_set%l(:, :)
         IF (PRESENT(last_cgf)) gto_basis_set%last_cgf(:, :) = last_cgf(:, :)
         IF (PRESENT(last_sgf)) gto_basis_set%last_sgf(:, :) = last_sgf(:, :)
         IF (PRESENT(n)) gto_basis_set%n(:, :) = n(:, :)
         IF (PRESENT(gcc)) gto_basis_set%gcc(:, :, :) = gcc(:, :, :)
      ELSE
         CALL cp__b("aobasis/basis_set_types.F",1899,"The pointer gto_basis_set is not associated")
      END IF

   END SUBROUTINE set_gto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
!> \param output_unit ...
!> \param header ...
! **************************************************************************************************
   SUBROUTINE write_gto_basis_set(gto_basis_set, output_unit, header)

      ! Write a Gaussian-type orbital (GTO) basis set data set to the output
      ! unit.

      ! - Creation (09.01.2002,MK)

      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      INTEGER, INTENT(in)                                :: output_unit
      CHARACTER(len=*), OPTIONAL                         :: header

      INTEGER                                            :: ipgf, iset, ishell

! -------------------------------------------------------------------------

      IF (ASSOCIATED(gto_basis_set) .AND. (output_unit > 0)) THEN

         IF (PRESENT(header)) THEN
            WRITE (UNIT=output_unit, FMT="(/,T6,A,T41,A40)") &
               TRIM(header), TRIM(gto_basis_set%name)
         END IF

         WRITE (UNIT=output_unit, FMT="(/,(T8,A,T71,I10))") &
            "Number of orbital shell sets:            ", &
            gto_basis_set%nset, &
            "Number of orbital shells:                ", &
            SUM(gto_basis_set%nshell(:)), &
            "Number of primitive Cartesian functions: ", &
            SUM(gto_basis_set%npgf(:)), &
            "Number of Cartesian basis functions:     ", &
            gto_basis_set%ncgf, &
            "Number of spherical basis functions:     ", &
            gto_basis_set%nsgf, &
            "Norm type:                               ", &
            gto_basis_set%norm_type

         WRITE (UNIT=output_unit, FMT="(/,T6,A,T41,A40,/,/,T25,A)") &
            "GTO basis set information for", TRIM(gto_basis_set%name), &
            "Set   Shell     n   l            Exponent    Coefficient"

         DO iset = 1, gto_basis_set%nset
            WRITE (UNIT=output_unit, FMT="(A)") ""
            DO ishell = 1, gto_basis_set%nshell(iset)
               WRITE (UNIT=output_unit, &
                      FMT="(T25,I3,4X,I4,4X,I2,2X,I2,(T51,2F15.6))") &
                  iset, ishell, &
                  gto_basis_set%n(ishell, iset), &
                  gto_basis_set%l(ishell, iset), &
                  (gto_basis_set%zet(ipgf, iset), &
                   gto_basis_set%gcc(ipgf, ishell, iset), &
                   ipgf=1, gto_basis_set%npgf(iset))
            END DO
         END DO

      END IF

   END SUBROUTINE write_gto_basis_set

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param orb_basis_set ...
!> \param output_unit ...
!> \param header ...
! **************************************************************************************************
   SUBROUTINE write_orb_basis_set(orb_basis_set, output_unit, header)

      ! Write a Gaussian-type orbital (GTO) basis set data set to the output
      ! unit.

      ! - Creation (09.01.2002,MK)

      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      INTEGER, INTENT(in)                                :: output_unit
      CHARACTER(len=*), OPTIONAL                         :: header

      INTEGER                                            :: icgf, ico, ipgf, iset, ishell

      IF (ASSOCIATED(orb_basis_set) .AND. (output_unit > 0)) THEN
         IF (PRESENT(header)) THEN
            WRITE (UNIT=output_unit, FMT="(/,T6,A,T41,A40)") &
               TRIM(header), TRIM(orb_basis_set%name)
         END IF

         WRITE (UNIT=output_unit, FMT="(/,(T8,A,T71,I10))") &
            "Number of orbital shell sets:            ", &
            orb_basis_set%nset, &
            "Number of orbital shells:                ", &
            SUM(orb_basis_set%nshell(:)), &
            "Number of primitive Cartesian functions: ", &
            SUM(orb_basis_set%npgf(:)), &
            "Number of Cartesian basis functions:     ", &
            orb_basis_set%ncgf, &
            "Number of spherical basis functions:     ", &
            orb_basis_set%nsgf, &
            "Norm type:                               ", &
            orb_basis_set%norm_type

         WRITE (UNIT=output_unit, FMT="(/,T8,A,/,/,T25,A)") &
            "Normalised Cartesian orbitals:", &
            "Set   Shell   Orbital            Exponent    Coefficient"

         icgf = 0

         DO iset = 1, orb_basis_set%nset
            DO ishell = 1, orb_basis_set%nshell(iset)
               WRITE (UNIT=output_unit, FMT="(A)") ""
               DO ico = 1, nco(orb_basis_set%l(ishell, iset))
                  icgf = icgf+1
                  WRITE (UNIT=output_unit, &
                         FMT="(T25,I3,4X,I4,3X,A12,(T51,2F15.6))") &
                     iset, ishell, orb_basis_set%cgf_symbol(icgf), &
                     (orb_basis_set%zet(ipgf, iset), &
                      orb_basis_set%norm_cgf(icgf)* &
                      orb_basis_set%gcc(ipgf, ishell, iset), &
                      ipgf=1, orb_basis_set%npgf(iset))
               END DO
            END DO
         END DO
      END IF

   END SUBROUTINE write_orb_basis_set

! **************************************************************************************************
!> \brief ...
!> \param sto_basis_set ...
! **************************************************************************************************
   SUBROUTINE allocate_sto_basis_set(sto_basis_set)

      TYPE(sto_basis_set_type), POINTER                  :: sto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_sto_basis_set', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      CALL deallocate_sto_basis_set(sto_basis_set)

      ALLOCATE (sto_basis_set)

      sto_basis_set%name = "NONAME"
      NULLIFY (sto_basis_set%symbol)
      NULLIFY (sto_basis_set%nq)
      NULLIFY (sto_basis_set%lq)
      NULLIFY (sto_basis_set%zet)

   END SUBROUTINE allocate_sto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param sto_basis_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_sto_basis_set(sto_basis_set)

      TYPE(sto_basis_set_type), POINTER                  :: sto_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_sto_basis_set', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF (ASSOCIATED(sto_basis_set)) THEN
         IF (ASSOCIATED(sto_basis_set%symbol)) THEN
            DEALLOCATE (sto_basis_set%symbol)
         END IF
         IF (ASSOCIATED(sto_basis_set%nq)) THEN
            DEALLOCATE (sto_basis_set%nq)
         END IF
         IF (ASSOCIATED(sto_basis_set%lq)) THEN
            DEALLOCATE (sto_basis_set%lq)
         END IF
         IF (ASSOCIATED(sto_basis_set%zet)) THEN
            DEALLOCATE (sto_basis_set%zet)
         END IF

         DEALLOCATE (sto_basis_set)
      END IF
   END SUBROUTINE deallocate_sto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param sto_basis_set ...
!> \param name ...
!> \param nshell ...
!> \param symbol ...
!> \param nq ...
!> \param lq ...
!> \param zet ...
!> \param maxlq ...
!> \param numsto ...
! **************************************************************************************************
   SUBROUTINE get_sto_basis_set(sto_basis_set, name, nshell, symbol, nq, lq, zet, maxlq, numsto)

      TYPE(sto_basis_set_type), POINTER                  :: sto_basis_set
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      INTEGER, INTENT(OUT), OPTIONAL                     :: nshell
      CHARACTER(LEN=6), DIMENSION(:), OPTIONAL, POINTER  :: symbol
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nq, lq
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: zet
      INTEGER, INTENT(OUT), OPTIONAL                     :: maxlq, numsto

      CHARACTER(len=*), PARAMETER :: routineN = 'get_sto_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iset

! -------------------------------------------------------------------------

      IF (ASSOCIATED(sto_basis_set)) THEN
         IF (PRESENT(name)) name = sto_basis_set%name
         IF (PRESENT(nshell)) nshell = sto_basis_set%nshell
         IF (PRESENT(symbol)) symbol => sto_basis_set%symbol
         IF (PRESENT(nq)) nq => sto_basis_set%nq
         IF (PRESENT(lq)) lq => sto_basis_set%lq
         IF (PRESENT(zet)) zet => sto_basis_set%zet
         IF (PRESENT(maxlq)) THEN
            maxlq = MAXVAL(sto_basis_set%lq(1:sto_basis_set%nshell))
         END IF
         IF (PRESENT(numsto)) THEN
            numsto = 0
            DO iset = 1, sto_basis_set%nshell
               numsto = numsto+2*sto_basis_set%lq(iset)+1
            END DO
         END IF
      ELSE
         CALL cp__b("aobasis/basis_set_types.F",2137,"The pointer sto_basis_set is not associated")
      END IF

   END SUBROUTINE get_sto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param sto_basis_set ...
!> \param name ...
!> \param nshell ...
!> \param symbol ...
!> \param nq ...
!> \param lq ...
!> \param zet ...
! **************************************************************************************************
   SUBROUTINE set_sto_basis_set(sto_basis_set, name, nshell, symbol, nq, lq, zet)

      TYPE(sto_basis_set_type), POINTER                  :: sto_basis_set
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: name
      INTEGER, INTENT(IN), OPTIONAL                      :: nshell
      CHARACTER(LEN=6), DIMENSION(:), OPTIONAL, POINTER  :: symbol
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nq, lq
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: zet

      CHARACTER(len=*), PARAMETER :: routineN = 'set_sto_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ns

! -------------------------------------------------------------------------

      IF (ASSOCIATED(sto_basis_set)) THEN
         IF (PRESENT(name)) sto_basis_set%name = name
         IF (PRESENT(nshell)) sto_basis_set%nshell = nshell
         IF (PRESENT(symbol)) THEN
            ns = SIZE(symbol)
            IF (ASSOCIATED(sto_basis_set%symbol)) DEALLOCATE (sto_basis_set%symbol)
            ALLOCATE (sto_basis_set%symbol(1:ns))
            sto_basis_set%symbol(:) = symbol(:)
         END IF
         IF (PRESENT(nq)) THEN
            ns = SIZE(nq)
            CALL reallocate(sto_basis_set%nq, 1, ns)
            sto_basis_set%nq = nq(:)
         END IF
         IF (PRESENT(lq)) THEN
            ns = SIZE(lq)
            CALL reallocate(sto_basis_set%lq, 1, ns)
            sto_basis_set%lq = lq(:)
         END IF
         IF (PRESENT(zet)) THEN
            ns = SIZE(zet)
            CALL reallocate(sto_basis_set%zet, 1, ns)
            sto_basis_set%zet = zet(:)
         END IF
      ELSE
         CALL cp__b("aobasis/basis_set_types.F",2194,"The pointer sto_basis_set is not associated")
      END IF

   END SUBROUTINE set_sto_basis_set

! **************************************************************************************************
!> \brief ...
!> \param sto_basis_set ...
!> \param gto_basis_set ...
!> \param ngauss ...
! **************************************************************************************************
   SUBROUTINE create_gto_from_sto_basis(sto_basis_set, gto_basis_set, ngauss)

      TYPE(sto_basis_set_type), POINTER                  :: sto_basis_set
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      INTEGER, OPTIONAL                                  :: ngauss

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gto_from_sto_basis', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: maxng = 6

      CHARACTER(LEN=default_string_length)               :: name, sng
      INTEGER                                            :: ico, ipgf, iset, lshell, m, maxco, maxl, &
                                                            ncgf, ng, np, nset, nsgf, nshell
      INTEGER, DIMENSION(:), POINTER                     :: lq, nq
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet
      REAL(KIND=dp), DIMENSION(maxng)                    :: gcc, zetg

      ng = 6
      IF (PRESENT(ngauss)) ng = ngauss
      IF (ng > maxng) CALL cp__b("aobasis/basis_set_types.F",2224,"Too many Gaussian primitives requested")

      CALL allocate_gto_basis_set(gto_basis_set)

      CALL get_sto_basis_set(sto_basis_set, name=name, nshell=nshell, nq=nq, &
                             lq=lq, zet=zet)

      maxl = MAXVAL(lq)
      CALL init_orbital_pointers(maxl)

      CALL integer_to_string(ng, sng)
      gto_basis_set%name = TRIM(name)//"_STO-"//TRIM(sng)//"G"

      nset = nshell
      gto_basis_set%nset = nset
      CALL reallocate(gto_basis_set%lmax, 1, nset)
      CALL reallocate(gto_basis_set%lmin, 1, nset)
      CALL reallocate(gto_basis_set%npgf, 1, nset)
      CALL reallocate(gto_basis_set%nshell, 1, nset)
      CALL reallocate(gto_basis_set%n, 1, 1, 1, nset)
      CALL reallocate(gto_basis_set%l, 1, 1, 1, nset)
      CALL reallocate(gto_basis_set%zet, 1, ng, 1, nset)
      CALL reallocate(gto_basis_set%gcc, 1, ng, 1, 1, 1, nset)

      DO iset = 1, nset
         CALL get_sto_ng(zet(iset), ng, nq(iset), lq(iset), zetg, gcc)
         gto_basis_set%lmax(iset) = lq(iset)
         gto_basis_set%lmin(iset) = lq(iset)
         gto_basis_set%npgf(iset) = ng
         gto_basis_set%nshell(iset) = 1
         gto_basis_set%n(1, iset) = lq(iset)+1
         gto_basis_set%l(1, iset) = lq(iset)
         DO ipgf = 1, ng
            gto_basis_set%gcc(ipgf, 1, iset) = gcc(ipgf)
            gto_basis_set%zet(ipgf, iset) = zetg(ipgf)
         END DO
      END DO

      CALL reallocate(gto_basis_set%set_radius, 1, nset)
      CALL reallocate(gto_basis_set%pgf_radius, 1, ng, 1, nset)
      CALL reallocate(gto_basis_set%first_cgf, 1, 1, 1, nset)
      CALL reallocate(gto_basis_set%first_sgf, 1, 1, 1, nset)
      CALL reallocate(gto_basis_set%last_cgf, 1, 1, 1, nset)
      CALL reallocate(gto_basis_set%last_sgf, 1, 1, 1, nset)
      CALL reallocate(gto_basis_set%ncgf_set, 1, nset)
      CALL reallocate(gto_basis_set%nsgf_set, 1, nset)

      maxco = 0
      ncgf = 0
      nsgf = 0

      DO iset = 1, nset
         gto_basis_set%ncgf_set(iset) = 0
         gto_basis_set%nsgf_set(iset) = 0
         lshell = gto_basis_set%l(1, iset)
         gto_basis_set%first_cgf(1, iset) = ncgf+1
         ncgf = ncgf+nco(lshell)
         gto_basis_set%last_cgf(1, iset) = ncgf
         gto_basis_set%ncgf_set(iset) = &
            gto_basis_set%ncgf_set(iset)+nco(lshell)
         gto_basis_set%first_sgf(1, iset) = nsgf+1
         nsgf = nsgf+nso(lshell)
         gto_basis_set%last_sgf(1, iset) = nsgf
         gto_basis_set%nsgf_set(iset) = &
            gto_basis_set%nsgf_set(iset)+nso(lshell)
         maxco = MAX(maxco, ng*ncoset(lshell))
      END DO

      gto_basis_set%ncgf = ncgf
      gto_basis_set%nsgf = nsgf

      CALL reallocate(gto_basis_set%cphi, 1, maxco, 1, ncgf)
      CALL reallocate(gto_basis_set%sphi, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%scon, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%lx, 1, ncgf)
      CALL reallocate(gto_basis_set%ly, 1, ncgf)
      CALL reallocate(gto_basis_set%lz, 1, ncgf)
      CALL reallocate(gto_basis_set%m, 1, nsgf)
      CALL reallocate(gto_basis_set%norm_cgf, 1, ncgf)
      ALLOCATE (gto_basis_set%cgf_symbol(ncgf))
      ALLOCATE (gto_basis_set%sgf_symbol(nsgf))

      ncgf = 0
      nsgf = 0

      DO iset = 1, nset
         lshell = gto_basis_set%l(1, iset)
         np = lshell+1
         DO ico = ncoset(lshell-1)+1, ncoset(lshell)
            ncgf = ncgf+1
            gto_basis_set%lx(ncgf) = indco(1, ico)
            gto_basis_set%ly(ncgf) = indco(2, ico)
            gto_basis_set%lz(ncgf) = indco(3, ico)
            gto_basis_set%cgf_symbol(ncgf) = &
               cgf_symbol(np, (/gto_basis_set%lx(ncgf), &
                                gto_basis_set%ly(ncgf), &
                                gto_basis_set%lz(ncgf)/))
         END DO
         DO m = -lshell, lshell
            nsgf = nsgf+1
            gto_basis_set%m(nsgf) = m
            gto_basis_set%sgf_symbol(nsgf) = sgf_symbol(np, lshell, m)
         END DO
      END DO

      gto_basis_set%norm_type = -1

   END SUBROUTINE create_gto_from_sto_basis

! **************************************************************************************************
!> \brief ...
!> \param z ...
!> \param ne ...
!> \param n ...
!> \param l ...
!> \retval srules ...
! **************************************************************************************************
   FUNCTION srules(z, ne, n, l)
      ! Slater rules
      INTEGER                                            :: z
      INTEGER, DIMENSION(4, 7)                           :: ne
      INTEGER                                            :: n, l
      REAL(dp)                                           :: srules

      REAL(dp), DIMENSION(7), PARAMETER :: &
         xns = (/1.0_dp, 2.0_dp, 3.0_dp, 3.7_dp, 4.0_dp, 4.2_dp, 4.4_dp/)

      INTEGER                                            :: i, l1, l2, m, m1, m2
      REAL(dp)                                           :: s

      s = 0.0_dp
      ! The complete shell
      l1 = l+1
      IF (l1 == 1) l2 = 2
      IF (l1 == 2) l2 = 1
      IF (l1 == 3) l2 = 4
      IF (l1 == 4) l2 = 3
      ! Rule a) no contribution from shells further out
      ! Rule b) 0.35 (1s 0.3) from each other electron in the same shell
      IF (n == 1) THEN
         m = ne(1, 1)
         s = s+0.3_dp*REAL(m-1, dp)
      ELSE
         m = ne(l1, n)+ne(l2, n)
         s = s+0.35_dp*REAL(m-1, dp)
      END IF
      ! Rule c) if (s,p) shell 0.85 from each electron with n-1, and 1.0
      !      from all electrons further in
      IF (l1+l2 == 3) THEN
         IF (n > 1) THEN
            m1 = ne(1, n-1)+ne(2, n-1)+ne(3, n-1)+ne(4, n-1)
            m2 = 0
            DO i = 1, n-2
               m2 = m2+ne(1, i)+ne(2, i)+ne(3, i)+ne(4, I)
            END DO
            s = s+0.85_dp*REAL(m1, dp)+1._dp*REAL(m2, dp)
         END IF
      ELSE
         ! Rule d) if (d,f) shell 1.0 from each electron inside
         m = 0
         DO i = 1, n-1
            m = m+ne(1, i)+ne(2, i)+ne(3, i)+ne(4, i)
         END DO
         s = s+1._dp*REAL(m, dp)
      END IF
      ! Slater exponent is (Z-S)/NS
      srules = (REAL(z, dp)-s)/xns(n)
   END FUNCTION srules

! **************************************************************************************************

END MODULE basis_set_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define the neighbor list data types and the corresponding functionality
!> \par History
!>      - cleaned (23.07.2003,MK)
!>      - full refactoring, list iterators (20.10.2010, JGH)
!>      - add get_neighbor_list_set_p, return info for a set of neighborlists
!>                                                             (07.2014,JGH)
!> \author Matthias Krack (21.06.2000)
! **************************************************************************************************
MODULE qs_neighbor_list_types

   USE kinds,                           ONLY: dp,&
                                              int_size
   USE termination,                     ONLY: stop_memory
   USE util,                            ONLY: locate,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (in this module) ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_neighbor_list_types'

! *** Definition of the data types for a linked list of neighbors ***

! **************************************************************************************************
   TYPE neighbor_node_type
      PRIVATE
      TYPE(neighbor_node_type), POINTER :: next_neighbor_node
      REAL(dp), DIMENSION(3)            :: r
      INTEGER, DIMENSION(3)             :: cell
      INTEGER                           :: neighbor
   END TYPE neighbor_node_type

! **************************************************************************************************
   TYPE neighbor_list_type
      PRIVATE
      TYPE(neighbor_list_type), POINTER :: next_neighbor_list
      TYPE(neighbor_node_type), POINTER :: first_neighbor_node, &
                                           last_neighbor_node
      INTEGER                           :: atom, nnode
   END TYPE neighbor_list_type

! **************************************************************************************************
   TYPE neighbor_list_set_type
      PRIVATE
      TYPE(neighbor_list_type), POINTER :: first_neighbor_list, &
                                           last_neighbor_list
      REAL(dp)                          :: r_max
      INTEGER                           :: nlist
      LOGICAL                           :: symmetric
      LOGICAL                           :: mic
      LOGICAL                           :: molecular
   END TYPE neighbor_list_set_type

! **************************************************************************************************
   TYPE neighbor_list_p_type
      TYPE(neighbor_list_type), POINTER :: neighbor_list
   END TYPE neighbor_list_p_type

! **************************************************************************************************
   TYPE neighbor_list_set_p_type
      TYPE(neighbor_list_set_type), POINTER :: neighbor_list_set
   END TYPE neighbor_list_set_p_type

! **************************************************************************************************
   TYPE list_search_type
      PRIVATE
      INTEGER                               :: nlist
      INTEGER, DIMENSION(:), POINTER        :: atom_list
      INTEGER, DIMENSION(:), POINTER        :: atom_index
      TYPE(neighbor_list_p_type), &
         DIMENSION(:), POINTER               :: neighbor_list
   END TYPE list_search_type

! **************************************************************************************************
! Neighbor List Iterator
! **************************************************************************************************
   TYPE neighbor_list_iterator_type
      PRIVATE
      INTEGER                               :: ikind, jkind, ilist, inode
      INTEGER                               :: nkind, nlist, nnode
      INTEGER                               :: iatom, jatom
      TYPE(neighbor_list_set_p_type), &
         DIMENSION(:), POINTER               :: nl
      TYPE(neighbor_list_type), POINTER     :: neighbor_list
      TYPE(neighbor_node_type), POINTER     :: neighbor_node
      TYPE(list_search_type), &
         DIMENSION(:), POINTER               :: list_search
   END TYPE neighbor_list_iterator_type

   TYPE neighbor_list_iterator_p_type
      PRIVATE
      TYPE(neighbor_list_iterator_type), POINTER :: neighbor_list_iterator
      INTEGER                                    :: last
   END TYPE neighbor_list_iterator_p_type
! **************************************************************************************************

! *** Public data types ***

   PUBLIC :: neighbor_list_p_type, &
             neighbor_list_set_type, &
             neighbor_list_set_p_type

! *** Public subroutines ***

   PUBLIC :: add_neighbor_list, &
             add_neighbor_node, &
             allocate_neighbor_list_set, &
             deallocate_neighbor_list_set, &
             get_neighbor_list_set_p

! *** Iterator functions and types ***

   PUBLIC :: neighbor_list_iterator_p_type, &
             neighbor_list_iterator_create, &
             neighbor_list_iterator_release, &
             neighbor_list_iterate, &
             nl_set_sub_iterator, &
             nl_sub_iterate, &
             get_iterator_info

CONTAINS

! **************************************************************************************************
!> \brief   Neighbor list iterator functions
!> \param iterator_set ...
!> \param nl ...
!> \param search ...
!> \param nthread ...
!> \date    28.07.2010
!> \author  jhu
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE neighbor_list_iterator_create(iterator_set, nl, search, nthread)
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: iterator_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: nl
      LOGICAL, INTENT(IN), OPTIONAL                      :: search
      INTEGER, INTENT(IN), OPTIONAL                      :: nthread

      CHARACTER(len=*), PARAMETER :: routineN = 'neighbor_list_iterator_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, il, ilist, istat, mthread, nlist
      TYPE(list_search_type), DIMENSION(:), POINTER      :: list_search
      TYPE(neighbor_list_iterator_type), POINTER         :: iterator
      TYPE(neighbor_list_type), POINTER                  :: neighbor_list

      mthread = 1
      IF (PRESENT(nthread)) mthread = nthread

      ALLOCATE (iterator_set(0:mthread-1), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 163, &
                                       "iterator_set", int_size*mthread)

      DO il = 0, mthread-1
         ALLOCATE (iterator_set(il)%neighbor_list_iterator, STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 168, &
                                          "neighbor_list_iterator", 0)

         iterator => iterator_set(il)%neighbor_list_iterator

         iterator%nl => nl

         iterator%ikind = 0
         iterator%jkind = 0
         iterator%nkind = NINT(SQRT(REAL(SIZE(nl), dp)))

         iterator%ilist = 0
         iterator%nlist = 0
         iterator%inode = 0
         iterator%nnode = 0

         iterator%iatom = 0
         iterator%jatom = 0

         NULLIFY (iterator%neighbor_list)
         NULLIFY (iterator%neighbor_node)
         NULLIFY (iterator%list_search)
      END DO

      iterator_set(:)%last = 0

      IF (PRESENT(search)) THEN
         IF (search) THEN
            ALLOCATE (list_search(SIZE(nl)), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 197, &
                                             "list_search", 0)
            DO il = 1, SIZE(nl)
               IF (ASSOCIATED(nl(il)%neighbor_list_set)) THEN
                  CALL get_neighbor_list_set(neighbor_list_set=nl(il)%neighbor_list_set, nlist=nlist)
                  list_search(il)%nlist = nlist
                  ALLOCATE (list_search(il)%atom_list(nlist), STAT=istat)
                  IF (istat /= 0) CALL stop_memory(routineN, moduleN, 204, &
                                                   "list_search(il)%atom_list", int_size*nlist)
                  ALLOCATE (list_search(il)%atom_index(nlist), STAT=istat)
                  IF (istat /= 0) CALL stop_memory(routineN, moduleN, 207, &
                                                   "list_search(il)%atom_index", int_size*nlist)
                  ALLOCATE (list_search(il)%neighbor_list(nlist), STAT=istat)
                  IF (istat /= 0) CALL stop_memory(routineN, moduleN, 210, &
                                                   "list_search(il)%neighbor_list", int_size*nlist)

                  NULLIFY (neighbor_list)
                  DO ilist = 1, nlist
                     IF (.NOT. ASSOCIATED(neighbor_list)) THEN
                        neighbor_list => first_list(nl(il)%neighbor_list_set)
                     ELSE
                        neighbor_list => neighbor_list%next_neighbor_list
                     END IF
                     CALL get_neighbor_list(neighbor_list=neighbor_list, atom=iatom)
                     list_search(il)%atom_list(ilist) = iatom
                     list_search(il)%neighbor_list(ilist)%neighbor_list => neighbor_list
                  END DO
                  CALL sort(list_search(il)%atom_list, nlist, list_search(il)%atom_index)

               ELSE
                  list_search(il)%nlist = -1
                  NULLIFY (list_search(il)%atom_list, list_search(il)%atom_index, list_search(il)%neighbor_list)
               END IF
            END DO
            DO il = 0, mthread-1
               iterator => iterator_set(il)%neighbor_list_iterator
               iterator%list_search => list_search
            END DO
         END IF
      END IF

   END SUBROUTINE neighbor_list_iterator_create

! **************************************************************************************************
!> \brief ...
!> \param iterator_set ...
! **************************************************************************************************
   SUBROUTINE neighbor_list_iterator_release(iterator_set)
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: iterator_set

      CHARACTER(len=*), PARAMETER :: routineN = 'neighbor_list_iterator_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: il, mthread
      TYPE(neighbor_list_iterator_type), POINTER         :: iterator

!all threads have the same search list

      iterator => iterator_set(0)%neighbor_list_iterator
      IF (ASSOCIATED(iterator%list_search)) THEN
         DO il = 1, SIZE(iterator%list_search)
            IF (iterator%list_search(il)%nlist >= 0) THEN
               DEALLOCATE (iterator%list_search(il)%atom_list)
               DEALLOCATE (iterator%list_search(il)%atom_index)
               DEALLOCATE (iterator%list_search(il)%neighbor_list)
            END IF
         END DO
         DEALLOCATE (iterator%list_search)
      END IF

      mthread = SIZE(iterator_set)
      DO il = 0, mthread-1
         DEALLOCATE (iterator_set(il)%neighbor_list_iterator)
      END DO
      DEALLOCATE (iterator_set)

   END SUBROUTINE neighbor_list_iterator_release

! **************************************************************************************************
!> \brief ...
!> \param iterator_set ...
!> \param ikind ...
!> \param jkind ...
!> \param iatom ...
! **************************************************************************************************
   SUBROUTINE nl_set_sub_iterator(iterator_set, ikind, jkind, iatom)
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: iterator_set
      INTEGER, INTENT(IN)                                :: ikind, jkind, iatom

      CHARACTER(len=*), PARAMETER :: routineN = 'nl_set_sub_iterator', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, il, ilist, mthread, nlist, nnode
      TYPE(list_search_type), POINTER                    :: list_search
      TYPE(neighbor_list_iterator_type), POINTER         :: iterator
      TYPE(neighbor_list_type), POINTER                  :: neighbor_list

      iterator => iterator_set(0)%neighbor_list_iterator
      ij = ikind+iterator%nkind*(jkind-1)
      IF (ASSOCIATED(iterator%list_search)) THEN
         list_search => iterator%list_search(ij)
         nlist = list_search%nlist
         ilist = 0
         NULLIFY (neighbor_list)
         IF (nlist > 0) THEN
            i = locate(list_search%atom_list, iatom)
            i = list_search%atom_index(i)
            IF (i > 0) neighbor_list => list_search%neighbor_list(i)%neighbor_list
            ilist = i
         END IF
         IF (ASSOCIATED(neighbor_list)) THEN
            CALL get_neighbor_list(neighbor_list=neighbor_list, nnode=nnode)
         ELSE
            nnode = 0
         END IF
      ELSE
         CALL cp__b("qs_neighbor_list_types.F",315,"")
      END IF

      mthread = SIZE(iterator_set)
      DO il = 0, mthread-1
         iterator => iterator_set(il)%neighbor_list_iterator

         iterator%ikind = ikind
         iterator%jkind = jkind

         iterator%ilist = ilist
         iterator%nlist = nlist
         iterator%inode = 0
         iterator%nnode = nnode

         iterator%iatom = iatom
         iterator%jatom = 0

         iterator%neighbor_list => neighbor_list
         NULLIFY (iterator%neighbor_node)
      END DO

      iterator_set(:)%last = 0

   END SUBROUTINE nl_set_sub_iterator

! **************************************************************************************************
!> \brief ...
!> \param iterator_set ...
!> \param mepos ...
!> \retval istat ...
! **************************************************************************************************
   FUNCTION neighbor_list_iterate(iterator_set, mepos) RESULT(istat)
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: iterator_set
      INTEGER, OPTIONAL                                  :: mepos
      INTEGER                                            :: istat

      CHARACTER(len=*), PARAMETER :: routineN = 'neighbor_list_iterate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iab, last, me
      TYPE(neighbor_list_iterator_type), POINTER         :: iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: nl

      IF (SIZE(iterator_set) .NE. 1 .AND. .NOT. PRESENT(mepos)) &
         CALL cp__b("qs_neighbor_list_types.F",362,"Parallel iterator calls must include 'mepos'")

      IF (PRESENT(mepos)) THEN
         me = mepos
      ELSE
         me = 0
      END IF

      istat = 0

!$OMP CRITICAL(neighbour_list_iterate_critical)
      last = iterator_set(0)%last
      IF (last /= me) THEN
         iterator_set(me)%neighbor_list_iterator = iterator_set(last)%neighbor_list_iterator
      END IF
      iterator => iterator_set(me)%neighbor_list_iterator
      nl => iterator%nl

      IF (iterator%inode < iterator%nnode) THEN
         ! we can be sure that there is another node in this list
         iterator%inode = iterator%inode+1
         iterator%neighbor_node => iterator%neighbor_node%next_neighbor_node
      ELSE
         iab = MAX(iterator%ikind+iterator%nkind*(iterator%jkind-1), 0)
         kindloop: DO ! look for the next list with nnode /= 0
            listloop: DO
               IF (iterator%ilist >= iterator%nlist) EXIT listloop
               iterator%ilist = iterator%ilist+1
               IF (ASSOCIATED(iterator%neighbor_list)) THEN
                  iterator%neighbor_list => iterator%neighbor_list%next_neighbor_list
               ELSE
                  iterator%neighbor_list => first_list(nl(iab)%neighbor_list_set)
               END IF
               CALL get_neighbor_list(neighbor_list=iterator%neighbor_list, atom=iterator%iatom, &
                                      nnode=iterator%nnode)
               IF (iterator%nnode > 0) EXIT kindloop
            END DO listloop
            IF (iab >= iterator%nkind**2) THEN
               istat = 1
               EXIT kindloop
            ELSE
               iab = iab+1
               iterator%jkind = (iab-1)/iterator%nkind+1
               iterator%ikind = iab-iterator%nkind*(iterator%jkind-1)
               iterator%ilist = 0
               IF (.NOT. ASSOCIATED(nl(iab)%neighbor_list_set)) THEN
                  iterator%ilist = 0
                  iterator%nlist = 0
               ELSE
                  CALL get_neighbor_list_set(neighbor_list_set= &
                                             nl(iab)%neighbor_list_set, nlist=iterator%nlist)
                  iterator%ilist = 0
               END IF
               NULLIFY (iterator%neighbor_list)
            END IF
         END DO kindloop
         IF (istat == 0) THEN
            iterator%inode = 1
            iterator%neighbor_node => first_node(iterator%neighbor_list)
         END IF
      END IF
      IF (istat == 0) THEN
         CALL get_neighbor_node(neighbor_node=iterator%neighbor_node, neighbor=iterator%jatom)
      END IF

      ! mark the last iterator updated
      iterator_set(:)%last = me
!$OMP END CRITICAL(neighbour_list_iterate_critical)

   END FUNCTION neighbor_list_iterate

! **************************************************************************************************
!> \brief ...
!> \param iterator_set ...
!> \retval istat ...
! **************************************************************************************************
   FUNCTION nl_sub_iterate(iterator_set) RESULT(istat)
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: iterator_set
      INTEGER                                            :: istat

      CHARACTER(len=*), PARAMETER :: routineN = 'nl_sub_iterate', routineP = moduleN//':'//routineN

      INTEGER                                            :: last, me
      TYPE(neighbor_list_iterator_type), POINTER         :: iterator

! Sub-iterator single-threaded, or with private sub-iterators only

      me = 0
      istat = 0

      last = iterator_set(0)%last
      IF (last /= me) THEN
         iterator_set(me)%neighbor_list_iterator = iterator_set(last)%neighbor_list_iterator
      END IF
      iterator => iterator_set(me)%neighbor_list_iterator

      IF (ASSOCIATED(iterator%neighbor_list)) THEN
         IF (iterator%inode >= iterator%nnode) THEN
            ! end of loop
            istat = 1
         ELSEIF (iterator%inode == 0) THEN
            iterator%inode = 1
            iterator%neighbor_node => first_node(iterator%neighbor_list)
         ELSEIF (iterator%inode > 0) THEN
            ! we can be sure that there is another node in this list
            iterator%inode = iterator%inode+1
            iterator%neighbor_node => iterator%neighbor_node%next_neighbor_node
         ELSE
            CALL cp__b("qs_neighbor_list_types.F",471,"wrong")
         END IF
      ELSE
         ! no list available
         istat = 1
      END IF
      IF (istat == 0) THEN
         CALL get_neighbor_node(neighbor_node=iterator%neighbor_node, neighbor=iterator%jatom)
      END IF

      ! mark the last iterator updated
      iterator_set(:)%last = me

   END FUNCTION nl_sub_iterate

! **************************************************************************************************
!> \brief ...
!> \param iterator_set ...
!> \param mepos ...
!> \param ikind ...
!> \param jkind ...
!> \param nkind ...
!> \param ilist ...
!> \param nlist ...
!> \param inode ...
!> \param nnode ...
!> \param iatom ...
!> \param jatom ...
!> \param r ...
!> \param cell ...
! **************************************************************************************************
   SUBROUTINE get_iterator_info(iterator_set, mepos, &
                                ikind, jkind, nkind, ilist, nlist, inode, nnode, iatom, jatom, r, cell)
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: iterator_set
      INTEGER, OPTIONAL                                  :: mepos, ikind, jkind, nkind, ilist, &
                                                            nlist, inode, nnode, iatom, jatom
      REAL(dp), DIMENSION(3), OPTIONAL                   :: r
      INTEGER, DIMENSION(3), OPTIONAL                    :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'get_iterator_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: me
      TYPE(neighbor_list_iterator_type), POINTER         :: iterator

      IF (SIZE(iterator_set) .NE. 1 .AND. .NOT. PRESENT(mepos)) &
         CALL cp__b("qs_neighbor_list_types.F",518,"Parallel iterator calls must include 'mepos'")

      IF (PRESENT(mepos)) THEN
         me = mepos
      ELSE
         me = 0
      END IF
      iterator => iterator_set(me)%neighbor_list_iterator

      IF (PRESENT(ikind)) ikind = iterator%ikind
      IF (PRESENT(jkind)) jkind = iterator%jkind
      IF (PRESENT(nkind)) nkind = iterator%nkind
      IF (PRESENT(ilist)) ilist = iterator%ilist
      IF (PRESENT(nlist)) nlist = iterator%nlist
      IF (PRESENT(inode)) inode = iterator%inode
      IF (PRESENT(nnode)) nnode = iterator%nnode
      IF (PRESENT(iatom)) iatom = iterator%iatom
      IF (PRESENT(jatom)) jatom = iterator%jatom
      IF (PRESENT(r)) THEN
         CALL get_neighbor_node(neighbor_node=iterator%neighbor_node, r=r)
      END IF
      IF (PRESENT(cell)) THEN
         CALL get_neighbor_node(neighbor_node=iterator%neighbor_node, cell=cell)
      END IF

   END SUBROUTINE get_iterator_info

! **************************************************************************************************
!> \brief   Add a new neighbor list to a neighbor list set.
!> \param neighbor_list_set ...
!> \param atom ...
!> \param neighbor_list ...
!> \date    13.09.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE add_neighbor_list(neighbor_list_set, atom, neighbor_list)

      TYPE(neighbor_list_set_type), POINTER              :: neighbor_list_set
      INTEGER, INTENT(IN)                                :: atom
      TYPE(neighbor_list_type), POINTER                  :: neighbor_list

      CHARACTER(len=*), PARAMETER :: routineN = 'add_neighbor_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: istat
      TYPE(neighbor_list_type), POINTER                  :: new_neighbor_list

      IF (ASSOCIATED(neighbor_list_set)) THEN

         IF (ASSOCIATED(neighbor_list_set%last_neighbor_list)) THEN

            new_neighbor_list => &
               neighbor_list_set%last_neighbor_list%next_neighbor_list

            IF (.NOT. ASSOCIATED(new_neighbor_list)) THEN

!         *** Allocate a new neighbor list ***

               ALLOCATE (new_neighbor_list, STAT=istat)
               IF (istat /= 0) THEN
                  CALL stop_memory(routineN, moduleN, 579, &
                                   "new_neighbor_list", 0)
               END IF

               NULLIFY (new_neighbor_list%next_neighbor_list)
               NULLIFY (new_neighbor_list%first_neighbor_node)

!         *** Link the new neighbor list to the neighbor list set ***

               neighbor_list_set%last_neighbor_list%next_neighbor_list => new_neighbor_list

            END IF

         ELSE

            new_neighbor_list => neighbor_list_set%first_neighbor_list

            IF (.NOT. ASSOCIATED(new_neighbor_list)) THEN

!         *** Allocate a new first neighbor list ***

               ALLOCATE (new_neighbor_list, STAT=istat)
               IF (istat /= 0) THEN
                  CALL stop_memory(routineN, moduleN, 602, &
                                   "new_neighbor_list", 0)
               END IF

               NULLIFY (new_neighbor_list%next_neighbor_list)
               NULLIFY (new_neighbor_list%first_neighbor_node)

!         *** Link the new first neighbor list to the neighbor list set ***

               neighbor_list_set%first_neighbor_list => new_neighbor_list

            END IF

         END IF

!     *** Store the data set of the new neighbor list ***

         NULLIFY (new_neighbor_list%last_neighbor_node)
         new_neighbor_list%atom = atom
         new_neighbor_list%nnode = 0

!     *** Update the pointer to the last neighbor ***
!     *** list of the neighbor list set           ***

         neighbor_list_set%last_neighbor_list => new_neighbor_list

!     *** Increment the neighbor list counter ***

         neighbor_list_set%nlist = neighbor_list_set%nlist+1

!     *** Return a pointer to the new neighbor list ***

         neighbor_list => new_neighbor_list

      ELSE

         CALL cp__b("qs_neighbor_list_types.F",638,"The requested neighbor list set is not associated")

      END IF

   END SUBROUTINE add_neighbor_list

! **************************************************************************************************
!> \brief   Add a new neighbor list node to a neighbor list.
!> \param neighbor_list ...
!> \param neighbor ...
!> \param cell ...
!> \param r ...
!> \param exclusion_list ...
!> \param nkind ...
!> \date    23.06.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE add_neighbor_node(neighbor_list, neighbor, cell, r, exclusion_list, nkind)

      TYPE(neighbor_list_type), POINTER                  :: neighbor_list
      INTEGER, INTENT(IN)                                :: neighbor
      INTEGER, DIMENSION(3), INTENT(IN)                  :: cell
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: r
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: exclusion_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nkind

      CHARACTER(len=*), PARAMETER :: routineN = 'add_neighbor_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, istat, my_nkind
      TYPE(neighbor_node_type), POINTER                  :: new_neighbor_node

      IF (ASSOCIATED(neighbor_list)) THEN

!     *** Check for exclusions ***

         IF (PRESENT(exclusion_list)) THEN
            IF (ASSOCIATED(exclusion_list)) THEN
               DO iatom = 1, SIZE(exclusion_list)
                  IF (exclusion_list(iatom) == 0) EXIT
                  IF (exclusion_list(iatom) == neighbor) RETURN
               END DO
            END IF
         END IF

         my_nkind = 0
         IF (PRESENT(nkind)) my_nkind = nkind

         IF (ASSOCIATED(neighbor_list%last_neighbor_node)) THEN

            new_neighbor_node => neighbor_list%last_neighbor_node%next_neighbor_node

            IF (.NOT. ASSOCIATED(new_neighbor_node)) THEN

!         *** Allocate a new neighbor node ***

               ALLOCATE (new_neighbor_node, STAT=istat)
               IF (istat /= 0) THEN
                  CALL stop_memory(routineN, moduleN, 697, &
                                   "new_neighbor_node", 0)
               END IF

               NULLIFY (new_neighbor_node%next_neighbor_node)

!         *** Link the new neighbor node to the neighbor list ***

               neighbor_list%last_neighbor_node%next_neighbor_node => new_neighbor_node

            END IF

         ELSE

            new_neighbor_node => neighbor_list%first_neighbor_node

            IF (.NOT. ASSOCIATED(new_neighbor_node)) THEN

!         *** Allocate a new first neighbor node ***

               ALLOCATE (new_neighbor_node, STAT=istat)
               IF (istat /= 0) THEN
                  CALL stop_memory(routineN, moduleN, 719, &
                                   "new_neighbor_node", 0)
               END IF

               NULLIFY (new_neighbor_node%next_neighbor_node)

!         *** Link the new first neighbor node to the neighbor list ***

               neighbor_list%first_neighbor_node => new_neighbor_node

            END IF

         END IF

!     *** Store the data set of the new neighbor ***

         new_neighbor_node%neighbor = neighbor
         new_neighbor_node%cell(:) = cell(:)
         new_neighbor_node%r(:) = r(:)

!     *** Update the pointer to the last neighbor node of the neighbor list ***

         neighbor_list%last_neighbor_node => new_neighbor_node

!     *** Increment the neighbor node counter ***

         neighbor_list%nnode = neighbor_list%nnode+1

      ELSE

         CALL cp__b("qs_neighbor_list_types.F",749,"The requested neighbor list is not associated")

      END IF

   END SUBROUTINE add_neighbor_node

! **************************************************************************************************
!> \brief   Allocate and initialize a set of neighbor lists.
!> \param neighbor_list_set ...
!> \param r_max ...
!> \param symmetric ...
!> \param mic ...
!> \param molecular ...
!> \date    23.06.2000
!> \author MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_neighbor_list_set(neighbor_list_set, r_max, symmetric, mic, molecular)

      TYPE(neighbor_list_set_type), POINTER              :: neighbor_list_set
      REAL(dp), INTENT(IN)                               :: r_max
      LOGICAL, INTENT(IN)                                :: symmetric, mic, molecular

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_neighbor_list_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: istat

!   *** Deallocate the old neighbor list set ***

      IF (ASSOCIATED(neighbor_list_set)) THEN
         CALL deallocate_neighbor_list_set(neighbor_list_set)
      END IF

!   *** Allocate a set of neighbor lists ***

      ALLOCATE (neighbor_list_set, STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 787, &
                          "neighbor_list_set", 0)
      END IF

      NULLIFY (neighbor_list_set%first_neighbor_list)

!   *** Initialize the pointers to the first neighbor list ***

      CALL init_neighbor_list_set(neighbor_list_set, r_max, symmetric, mic, molecular)

   END SUBROUTINE allocate_neighbor_list_set

! **************************************************************************************************
!> \brief   Deallocate a neighbor list.
!> \param neighbor_list ...
!> \date    20.09.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_neighbor_list(neighbor_list)

      TYPE(neighbor_list_type), POINTER                  :: neighbor_list

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_neighbor_list', &
         routineP = moduleN//':'//routineN

      TYPE(neighbor_node_type), POINTER                  :: neighbor_node, next_neighbor_node

      IF (ASSOCIATED(neighbor_list)) THEN

         neighbor_node => neighbor_list%first_neighbor_node

         DO WHILE (ASSOCIATED(neighbor_node))
            next_neighbor_node => neighbor_node%next_neighbor_node
            DEALLOCATE (neighbor_node)
            neighbor_node => next_neighbor_node
         END DO

         DEALLOCATE (neighbor_list)

      END IF

   END SUBROUTINE deallocate_neighbor_list

! **************************************************************************************************
!> \brief   Deallocate a neighbor list set.
!> \param neighbor_list_set ...
!> \date    03.11.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_neighbor_list_set(neighbor_list_set)
      TYPE(neighbor_list_set_type), POINTER              :: neighbor_list_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_neighbor_list_set', &
         routineP = moduleN//':'//routineN

      TYPE(neighbor_list_type), POINTER                  :: neighbor_list, next_neighbor_list

      IF (ASSOCIATED(neighbor_list_set)) THEN

         neighbor_list => neighbor_list_set%first_neighbor_list

         DO WHILE (ASSOCIATED(neighbor_list))
            next_neighbor_list => neighbor_list%next_neighbor_list
            CALL deallocate_neighbor_list(neighbor_list)
            neighbor_list => next_neighbor_list
         END DO

         DEALLOCATE (neighbor_list_set)

      END IF

   END SUBROUTINE deallocate_neighbor_list_set

! **************************************************************************************************
!> \brief   Return a pointer to the first neighbor list of a neighbor list set.
!> \param neighbor_list_set ...
!> \retval first_neighbor_list ...
!> \date    13.09.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION first_list(neighbor_list_set) RESULT(first_neighbor_list)

      TYPE(neighbor_list_set_type), POINTER              :: neighbor_list_set
      TYPE(neighbor_list_type), POINTER                  :: first_neighbor_list

      first_neighbor_list => neighbor_list_set%first_neighbor_list

   END FUNCTION first_list

! **************************************************************************************************
!> \brief   Return a pointer to the first neighbor node of a neighbor list.
!> \param neighbor_list ...
!> \retval first_neighbor_node ...
!> \date    23.06.2000,
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION first_node(neighbor_list) RESULT(first_neighbor_node)

      TYPE(neighbor_list_type), POINTER                  :: neighbor_list
      TYPE(neighbor_node_type), POINTER                  :: first_neighbor_node

      first_neighbor_node => neighbor_list%first_neighbor_node

   END FUNCTION first_node

! **************************************************************************************************
!> \brief   Return the reqested data of a neighbor list.
!> \param neighbor_list ...
!> \param atom ...
!> \param nnode ...
!> \date    13.09.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_neighbor_list(neighbor_list, atom, nnode)

      TYPE(neighbor_list_type), POINTER                  :: neighbor_list
      INTEGER, INTENT(OUT), OPTIONAL                     :: atom, nnode

      CHARACTER(len=*), PARAMETER :: routineN = 'get_neighbor_list', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(neighbor_list)) THEN

         IF (PRESENT(atom)) atom = neighbor_list%atom
         IF (PRESENT(nnode)) nnode = neighbor_list%nnode

      ELSE

         CALL cp__b("qs_neighbor_list_types.F",920,"The requested neighbor list is not associated")

      END IF

   END SUBROUTINE get_neighbor_list

! **************************************************************************************************
!> \brief   Return the components of a neighbor list set.
!> \param neighbor_list_set ...
!> \param r_max ...
!> \param nlist ...
!> \param symmetric ...
!> \param mic ...
!> \param molecular ...
!> \date    10.11.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_neighbor_list_set(neighbor_list_set, r_max, nlist, symmetric, mic, molecular)

      TYPE(neighbor_list_set_type), POINTER              :: neighbor_list_set
      REAL(dp), INTENT(OUT), OPTIONAL                    :: r_max
      INTEGER, INTENT(OUT), OPTIONAL                     :: nlist
      LOGICAL, INTENT(OUT), OPTIONAL                     :: symmetric, mic, molecular

      CHARACTER(len=*), PARAMETER :: routineN = 'get_neighbor_list_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(neighbor_list_set)) THEN

         IF (PRESENT(r_max)) r_max = neighbor_list_set%r_max
         IF (PRESENT(nlist)) nlist = neighbor_list_set%nlist
         IF (PRESENT(symmetric)) symmetric = neighbor_list_set%symmetric
         IF (PRESENT(mic)) mic = neighbor_list_set%mic
         IF (PRESENT(molecular)) molecular = neighbor_list_set%molecular

      ELSE

         CALL cp__b("qs_neighbor_list_types.F",958,"The requested neighbor list set is not associated")

      END IF

   END SUBROUTINE get_neighbor_list_set

! **************************************************************************************************
!> \brief   Return the components of the first neighbor list set.
!> \param neighbor_list_sets ...
!> \param r_max ...
!> \param nlist ...
!> \param symmetric ...
!> \param mic ...
!> \param molecular ...
!> \date    07.2014
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_neighbor_list_set_p(neighbor_list_sets, r_max, nlist, symmetric, mic, molecular)

      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: neighbor_list_sets
      REAL(dp), INTENT(OUT), OPTIONAL                    :: r_max
      INTEGER, INTENT(OUT), OPTIONAL                     :: nlist
      LOGICAL, INTENT(OUT), OPTIONAL                     :: symmetric, mic, molecular

      CHARACTER(len=*), PARAMETER :: routineN = 'get_neighbor_list_set_p', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(neighbor_list_set_type), POINTER              :: neighbor_list_set

      IF (ASSOCIATED(neighbor_list_sets)) THEN

         NULLIFY (neighbor_list_set)
         DO i = 1, SIZE(neighbor_list_sets)
            neighbor_list_set => neighbor_list_sets(i)%neighbor_list_set
            IF (ASSOCIATED(neighbor_list_set)) EXIT
         END DO

         IF (ASSOCIATED(neighbor_list_set)) THEN
            IF (PRESENT(r_max)) r_max = neighbor_list_set%r_max
            IF (PRESENT(nlist)) nlist = neighbor_list_set%nlist
            IF (PRESENT(symmetric)) symmetric = neighbor_list_set%symmetric
            IF (PRESENT(mic)) mic = neighbor_list_set%mic
            IF (PRESENT(molecular)) molecular = neighbor_list_set%molecular
         ELSE
            CALL cp_abort(cp__l("qs_neighbor_list_types.F",1005), "No neighbor list set is associated. "// &
                          "Did you specify *all* required basis-sets, eg. for ADMM?")
         END IF

      ELSE

         CALL cp__b("qs_neighbor_list_types.F",1011,"The requested neighbor list sets are not associated")

      END IF

   END SUBROUTINE get_neighbor_list_set_p

! **************************************************************************************************
!> \brief   Return the reqested data of a neighbor node.
!> \param neighbor_node ...
!> \param neighbor ...
!> \param cell ...
!> \param r ...
!> \date    23.06.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_neighbor_node(neighbor_node, neighbor, cell, r)

      TYPE(neighbor_node_type), POINTER                  :: neighbor_node
      INTEGER, INTENT(OUT), OPTIONAL                     :: neighbor
      INTEGER, DIMENSION(3), INTENT(OUT), OPTIONAL       :: cell
      REAL(dp), DIMENSION(3), INTENT(OUT), OPTIONAL      :: r

      CHARACTER(len=*), PARAMETER :: routineN = 'get_neighbor_node', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(neighbor_node)) THEN

         IF (PRESENT(neighbor)) neighbor = neighbor_node%neighbor
         IF (PRESENT(r)) r(:) = neighbor_node%r(:)
         IF (PRESENT(cell)) cell(:) = neighbor_node%cell(:)

      ELSE

         CALL cp__b("qs_neighbor_list_types.F",1045,"The requested neighbor node is not associated")

      END IF

   END SUBROUTINE get_neighbor_node

! **************************************************************************************************
!> \brief Initialize a neighbor list set. Nothing is (de)allocated here.
!>         This routine is also used to prepare a neighbor list set for
!>         overwriting.
!> \param neighbor_list_set ...
!> \param r_max ...
!> \param symmetric ...
!> \param mic ...
!> \param molecular ...
!> \date  20.09.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_neighbor_list_set(neighbor_list_set, r_max, symmetric, mic, molecular)

      TYPE(neighbor_list_set_type), POINTER              :: neighbor_list_set
      REAL(dp), INTENT(IN)                               :: r_max
      LOGICAL, INTENT(IN)                                :: symmetric, mic, molecular

      CHARACTER(len=*), PARAMETER :: routineN = 'init_neighbor_list_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(neighbor_list_set)) THEN

         ! *** Initialize the pointers to the last neighbor list ***
         NULLIFY (neighbor_list_set%last_neighbor_list)

         ! *** Initialize the neighbor list counter ***
         neighbor_list_set%nlist = 0

         ! *** Initialize the maximum interaction radius (optional) ***
         neighbor_list_set%r_max = r_max

         ! *** Initialize the neighbor list build properties
         neighbor_list_set%symmetric = symmetric
         neighbor_list_set%mic = mic
         neighbor_list_set%molecular = molecular

      ELSE

         CALL cp__b("qs_neighbor_list_types.F",1091,"The requested neighbor list set is not associated")

      END IF

   END SUBROUTINE init_neighbor_list_set

END MODULE qs_neighbor_list_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of disperson types for DFT calculations
!> \author JGH (20.10.2008)
! **************************************************************************************************
MODULE qs_dispersion_types

   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE qs_neighbor_list_types,          ONLY: deallocate_neighbor_list_set,&
                                              neighbor_list_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dispersion_types'

! **************************************************************************************************
   INTEGER, PARAMETER                       :: dftd2_pp = 1
   INTEGER, PARAMETER                       :: dftd3_pp = 2

   TYPE qs_dispersion_type
      INTEGER                                :: TYPE
      INTEGER                                :: pp_type
      INTEGER                                :: nl_type
      CHARACTER(LEN=default_string_length)  :: ref_functional
      REAL(KIND=dp)                          :: scaling
      REAL(KIND=dp)                          :: rc_disp
      REAL(KIND=dp)                          :: exp_pre
      TYPE(section_vals_type), POINTER       :: dftd_section
      LOGICAL                                :: verbose !extended output
      CHARACTER(LEN=default_string_length)  :: parameter_file_name
      CHARACTER(LEN=default_string_length)  :: kernel_file_name
      !DFT-D3 global parameters
      INTEGER                                :: max_elem !elements parametrized
      INTEGER                                :: maxc !max coordination number references per element
      REAL(KIND=dp)                          :: k1, k2, k3 !ad hoc parameters
      REAL(KIND=dp)                          :: alp !ad hoc parameters
      REAL(KIND=dp)                          :: s6, s8, sr6 !scaling parameters
      REAL(KIND=dp)                          :: a1, a2 !BJ scaling parameters
      REAL(KIND=dp)                          :: eps_cn
      LOGICAL                                :: doabc !neglect C9 terms
      LOGICAL                                :: c9cnst !use constant c9 terms
      LOGICAL                                :: lrc !calculate a long range correction
      TYPE(neighbor_list_set_p_type), &
         DIMENSION(:), POINTER                :: sab_vdw, sab_cn ! neighborlists for pair interactions
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER &
         :: c6ab
      INTEGER, DIMENSION(:), POINTER           :: maxci
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: r0ab
      REAL(KIND=dp), DIMENSION(:), POINTER    :: rcov !covalent radii
      REAL(KIND=dp), DIMENSION(:), POINTER    :: r2r4 !atomic <r^2>/<r^4> values
      REAL(KIND=dp), DIMENSION(:), POINTER    :: cn !coordination numbers (defaults)
      TYPE(cn_kind_list), DIMENSION(:), POINTER &
         :: cnkind
      TYPE(cn_atom_list), DIMENSION(:), POINTER &
         :: cnlist
      !vdW-DF variables
      REAL(KIND=dp)                          :: pw_cutoff
      REAL(KIND=dp)                          :: b_value, c_value !parameters for the rVV10 functional
      INTEGER                                :: nqs, nr_points
      !! The number of q points and radial points
      !! used in generating the kernel phi(q1*r, q2*r)
      !! (see DION 14-16 and SOLER 3)
      REAL(KIND=dp)                          :: r_max, q_cut, q_min, dk
      !! The maximum value of r, the maximum and minimum
      !! values of q and the k-space spacing of grid points.
      !! Note that, during a vdW run, values of q0 found
      !! larger than q_cut will be saturated (SOLER 6-7) to
      !! q_cut
      REAL(KIND=dp), DIMENSION(:), POINTER    :: q_mesh !! The values of all the q points used
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER &
         :: kernel !! A matrix holding the Fourier transformed kernel function
      !! for each pair of q values.  The ordering is
      !! kernel(k_point, q1_value, q2_value)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER &
         :: d2phi_dk2 !! A matrix holding the second derivatives of the above
      !! kernel matrix at each of the q points.  Stored as
      !! d2phi_dk2(k_point, q1_value, q2_value)
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: d2y_dx2 !! 2nd derivatives of q_mesh for interpolation
   END TYPE qs_dispersion_type

   TYPE qs_atom_dispersion_type
      INTEGER                                :: TYPE
      LOGICAL                                :: defined
      REAL(KIND=dp)                          :: vdw_radii !van der Waals radii
      REAL(KIND=dp)                          :: c6 !c6 coefficients
   END TYPE qs_atom_dispersion_type

   TYPE cn_kind_list
      REAL(KIND=dp)                          :: cnum
      INTEGER                                :: kind
   END TYPE cn_kind_list
   TYPE cn_atom_list
      REAL(KIND=dp)                          :: cnum
      INTEGER                                :: natom
      INTEGER, DIMENSION(:), POINTER          :: atom
   END TYPE cn_atom_list

! **************************************************************************************************

   PUBLIC :: qs_atom_dispersion_type, qs_dispersion_type
   PUBLIC :: dftd2_pp, dftd3_pp
   PUBLIC :: qs_dispersion_release

! **************************************************************************************************
CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param dispersion_env ...
! **************************************************************************************************
   SUBROUTINE qs_dispersion_release(dispersion_env)

      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dispersion_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iab

      IF (ASSOCIATED(dispersion_env)) THEN
         IF (ASSOCIATED(dispersion_env%maxci)) THEN
            ! DFT-D3 arrays
            DEALLOCATE (dispersion_env%maxci)
            DEALLOCATE (dispersion_env%c6ab)
            DEALLOCATE (dispersion_env%r0ab)
            DEALLOCATE (dispersion_env%rcov)
            DEALLOCATE (dispersion_env%r2r4)
            DEALLOCATE (dispersion_env%cn)
            IF (ASSOCIATED(dispersion_env%cnkind)) THEN
               DEALLOCATE (dispersion_env%cnkind)
            END IF
            IF (ASSOCIATED(dispersion_env%cnlist)) THEN
               DO i = 1, SIZE(dispersion_env%cnlist)
                  DEALLOCATE (dispersion_env%cnlist(i)%atom)
               END DO
               DEALLOCATE (dispersion_env%cnlist)
            END IF
         END IF
         ! vdD-DF
         IF (ASSOCIATED(dispersion_env%q_mesh)) THEN
            DEALLOCATE (dispersion_env%q_mesh)
         END IF
         IF (ASSOCIATED(dispersion_env%kernel)) THEN
            DEALLOCATE (dispersion_env%kernel)
         END IF
         IF (ASSOCIATED(dispersion_env%d2phi_dk2)) THEN
            DEALLOCATE (dispersion_env%d2phi_dk2)
         END IF
         IF (ASSOCIATED(dispersion_env%d2y_dx2)) THEN
            DEALLOCATE (dispersion_env%d2y_dx2)
         END IF
         ! neighborlists
         IF (ASSOCIATED(dispersion_env%sab_vdw)) THEN
            DO iab = 1, SIZE(dispersion_env%sab_vdw)
               CALL deallocate_neighbor_list_set(dispersion_env%sab_vdw(iab)%neighbor_list_set)
            END DO
            DEALLOCATE (dispersion_env%sab_vdw)
         END IF
         IF (ASSOCIATED(dispersion_env%sab_cn)) THEN
            DO iab = 1, SIZE(dispersion_env%sab_cn)
               CALL deallocate_neighbor_list_set(dispersion_env%sab_cn(iab)%neighbor_list_set)
            END DO
            DEALLOCATE (dispersion_env%sab_cn)
         END IF

         DEALLOCATE (dispersion_env)

      END IF

   END SUBROUTINE qs_dispersion_release

END MODULE qs_dispersion_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of the DFTB parameter types.
!> \author JGH (24.02.2007)
! **************************************************************************************************
MODULE qs_dftb_types

   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dftb_types'

! **************************************************************************************************
   TYPE qs_dftb_atom_type
      ! PRIVATE
      CHARACTER(LEN=default_string_length)   :: typ
      CHARACTER(LEN=default_string_length)   :: name
      LOGICAL                                :: defined
      INTEGER                                :: z !atomic number
      REAL(KIND=dp)                          :: zeff !effective core charge
      INTEGER                                :: natorb !number of orbitals
      INTEGER                                :: lmax !max angular momentum
      REAL(KIND=dp), DIMENSION(0:3)          :: skself !orbital energy
      REAL(KIND=dp), DIMENSION(0:3)          :: occupation !free atom occupation
      REAL(KIND=dp), DIMENSION(0:3)          :: eta !orbital hardness
      REAL(KIND=dp)                          :: energy !free atom energy
      REAL(KIND=dp)                          :: cutoff !cutoff radius for f matrix
      REAL(KIND=dp)                          :: xi, di !London parameter
      REAL(KIND=dp)                          :: rcdisp !cutoff radius for vdW
      REAL(KIND=dp)                          :: dudq !DFTB3 hardness derivative
   END TYPE qs_dftb_atom_type

! **************************************************************************************************
   TYPE qs_dftb_pairpot_type
      REAL(KIND=dp)                          :: dgrd ! grid spacing
      INTEGER                                :: ngrd ! number of grid points
      INTEGER                                :: ngrdcut ! grid cutoff
      INTEGER                                :: llm ! number of interactions (l,l,m)
      INTEGER                                :: n_urpoly ! order of polynomial
      REAL(KIND=dp)                          :: urep_cut ! cutoff for repulsive pot.
      REAL(KIND=dp), DIMENSION(10)           :: urep ! coefficients for repulsive
      ! potential in polynomial form
      INTEGER                                :: spdim ! number of points for
      ! spline representation
      REAL(KIND=dp)                          :: s_cut ! left-hand cutoff
      REAL(KIND=dp), DIMENSION(3)            :: srep ! coefficients for extrapolation
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: spxr ! spline points
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: scoeff ! spline coefficients
      REAL(KIND=dp), DIMENSION(2)            :: surr ! coefficients for last point
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: fmat ! Slater-Koster table (Hamiltonian)
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: smat ! Slater-Koster table (overlap)
      ! van der Waals parameter
      REAL(KIND=dp)                          :: xij, dij ! standard LJ parameters
      REAL(KIND=dp)                          :: x0ij ! Evdw(x0) = 0
      REAL(KIND=dp)                          :: a, b, c ! Short range polynomial coeffs
   END TYPE qs_dftb_pairpot_type

! *** Public data types ***

   PUBLIC :: qs_dftb_atom_type, qs_dftb_pairpot_type, &
             qs_dftb_pairpot_init, qs_dftb_pairpot_create, qs_dftb_pairpot_release

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param pairpot ...
! **************************************************************************************************
   SUBROUTINE qs_dftb_pairpot_init(pairpot)
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), &
         POINTER                                         :: pairpot

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dftb_pairpot_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (ASSOCIATED(pairpot)) THEN
         DO i = 1, SIZE(pairpot, 1)
            DO j = 1, SIZE(pairpot, 2)
               NULLIFY (pairpot(i, j)%spxr, pairpot(i, j)%scoeff, &
                        pairpot(i, j)%smat, pairpot(i, j)%fmat)
            END DO
         END DO
      END IF

   END SUBROUTINE qs_dftb_pairpot_init

! **************************************************************************************************
!> \brief ...
!> \param pairpot ...
!> \param ngrd ...
!> \param llm ...
!> \param spdim ...
! **************************************************************************************************
   SUBROUTINE qs_dftb_pairpot_create(pairpot, ngrd, llm, spdim)
      TYPE(qs_dftb_pairpot_type)                         :: pairpot
      INTEGER, INTENT(IN)                                :: ngrd, llm, spdim

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dftb_pairpot_create', &
         routineP = moduleN//':'//routineN

      pairpot%ngrd = ngrd
      pairpot%spdim = spdim
      pairpot%llm = llm

      IF (spdim > 0) THEN
         ALLOCATE (pairpot%spxr(spdim, 2))

         ALLOCATE (pairpot%scoeff(spdim, 4))
      END IF

      ALLOCATE (pairpot%fmat(ngrd, llm))

      ALLOCATE (pairpot%smat(ngrd, llm))

   END SUBROUTINE qs_dftb_pairpot_create

! **************************************************************************************************
!> \brief ...
!> \param pairpot ...
! **************************************************************************************************
   SUBROUTINE qs_dftb_pairpot_release(pairpot)
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), &
         POINTER                                         :: pairpot

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dftb_pairpot_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, n1, n2

      IF (ASSOCIATED(pairpot)) THEN
         n1 = SIZE(pairpot, 1)
         n2 = SIZE(pairpot, 2)
         DO i = 1, n1
            DO j = 1, n2
               IF (ASSOCIATED(pairpot(i, j)%spxr)) THEN
                  DEALLOCATE (pairpot(i, j)%spxr)
               END IF
               IF (ASSOCIATED(pairpot(i, j)%scoeff)) THEN
                  DEALLOCATE (pairpot(i, j)%scoeff)
               END IF
               IF (ASSOCIATED(pairpot(i, j)%smat)) THEN
                  DEALLOCATE (pairpot(i, j)%smat)
               END IF
               IF (ASSOCIATED(pairpot(i, j)%fmat)) THEN
                  DEALLOCATE (pairpot(i, j)%fmat)
               END IF
            END DO
         END DO
         DEALLOCATE (pairpot)
      END IF

   END SUBROUTINE qs_dftb_pairpot_release

END MODULE qs_dftb_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author Rodolphe Vuilleumier (29.12.2009)
! **************************************************************************************************
MODULE damping_dipole_types

   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'damping_dipole_types'

! *** Global public parameters

   INTEGER, PUBLIC, PARAMETER :: no_damping = -1, &
                                 tang_toennies = 1

! *** Define the damping types ***
! **************************************************************************************************
   TYPE damping_info_type
      CHARACTER(LEN=default_string_length)   :: atm_name1, atm_name2
      CHARACTER(LEN=default_string_length)   :: dtype
      INTEGER                                 :: order
      REAL(KIND=dp)                           :: bij, cij
   END TYPE damping_info_type
! **************************************************************************************************
   TYPE damping_type
      INTEGER :: itype
      INTEGER :: order
      REAL(KIND=dp) :: bij, cij
   END TYPE damping_type

   TYPE damping_p_type
      TYPE(damping_type), DIMENSION(:), POINTER :: damp
   END TYPE

! **************************************************************************************************

! *** Public data types ***

   PUBLIC :: damping_info_type, damping_type

! *** Public subroutines ***

   PUBLIC :: damping_p_type, damping_p_create, damping_p_release

CONTAINS

! **************************************************************************************************
!> \brief Creates Data-structure that contains damping information
!> \param damping ...
!> \param nkinds ...
!> \author Rodolphe Vuilleumier
! **************************************************************************************************
   SUBROUTINE damping_p_create(damping, nkinds)
      TYPE(damping_p_type), POINTER                      :: damping
      INTEGER, INTENT(IN)                                :: nkinds

      CHARACTER(LEN=*), PARAMETER :: routineN = 'damping_p_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(.NOT. ASSOCIATED(damping)))CALL cp__a("subsys/damping_dipole_types.F",74)
      ALLOCATE (damping)
      ALLOCATE (damping%damp(nkinds))
      DO i = 1, nkinds
         CALL init_damping(damping%damp(i))
      END DO

   END SUBROUTINE damping_p_create
! **************************************************************************************************
!> \brief Release Data-structure that contains damping information
!> \param damping ...
!> \author Rodolphe Vuilleumier [RV]
! **************************************************************************************************
   SUBROUTINE damping_p_release(damping)
      TYPE(damping_p_type), POINTER                      :: damping

      CHARACTER(len=*), PARAMETER :: routineN = 'damping_p_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(damping)) THEN
         IF (ASSOCIATED(damping%damp)) THEN
            DEALLOCATE (damping%damp)
         END IF
         DEALLOCATE (damping)
      END IF
      NULLIFY (damping)

   END SUBROUTINE damping_p_release

! **************************************************************************************************
!> \brief ...
!> \param damping ...
! **************************************************************************************************
   SUBROUTINE init_damping(damping)
      TYPE(damping_type)                                 :: damping

      damping%itype = no_damping
      damping%order = 1
      damping%bij = HUGE(0.0_dp)
      damping%cij = 0.0_dp

   END SUBROUTINE init_damping

! **************************************************************************************************
END MODULE damping_dipole_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief All kind of helpfull little routines
!> \par History
!>      none
!> \author CJM & JGH
! **************************************************************************************************
MODULE ao_util

   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE mathconstants,                   ONLY: &
        dfac, fac, fourpi, oorootpi, rootpi, sqrt105, sqrt15, sqrt2, sqrt21, sqrt3, sqrt35, sqrt5, &
        sqrt7, sqrthalf
   USE orbital_pointers,                ONLY: coset,&
                                              indco,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ao_util'

   !MK sqrt* constants moved to mathconstants

   REAL(KIND=dp), PARAMETER ::  s_root1o4pi = 0.5_dp*oorootpi
   REAL(KIND=dp), PARAMETER :: root4pi = 2.0_dp*rootpi
   REAL(KIND=dp), PARAMETER ::  s_root3o4pi = sqrt3*s_root1o4pi
   REAL(KIND=dp), PARAMETER :: root4pio3 = root4pi/sqrt3
   REAL(KIND=dp), PARAMETER :: root4pio5 = root4pi/sqrt5
   REAL(KIND=dp), PARAMETER ::  s_root15o4pi = sqrt15*s_root1o4pi
   REAL(KIND=dp), PARAMETER :: root4pio15 = root4pi/sqrt15
   REAL(KIND=dp), PARAMETER ::  s_root105o4pi = sqrt105*s_root1o4pi
   REAL(KIND=dp), PARAMETER :: root4pio105 = root4pi/sqrt105
   REAL(KIND=dp), PARAMETER ::  s_root1o16pi = 0.25_dp*oorootpi
   REAL(KIND=dp), PARAMETER :: root16pi = 4.0_dp*rootpi
   REAL(KIND=dp), PARAMETER ::  s_root5o16pi = sqrt5*s_root1o16pi
   REAL(KIND=dp), PARAMETER :: root16pio5 = root16pi/sqrt5
   REAL(KIND=dp), PARAMETER ::  s_2root5o16pi = 2.0_dp*s_root5o16pi
   REAL(KIND=dp), PARAMETER :: root16pio5o2 = root16pio5*0.5_dp
   REAL(KIND=dp), PARAMETER ::  s_3root5o16pi = 3.0_dp*s_root5o16pi
   REAL(KIND=dp), PARAMETER :: root16pio5o3 = root16pio5/3.0_dp
   REAL(KIND=dp), PARAMETER ::  s_18root5o16pi = 18.0_dp*s_root5o16pi
   REAL(KIND=dp), PARAMETER :: root16pio5o18 = root16pio5/18.0_dp
   REAL(KIND=dp), PARAMETER ::  s_2root7o16pi = 2.0_dp*sqrt7*s_root1o16pi
   REAL(KIND=dp), PARAMETER :: root16pio7o2 = root16pi/sqrt7*0.5_dp
   REAL(KIND=dp), PARAMETER ::  s_3root7o16pi = 3.0_dp*sqrt7*s_root1o16pi
   REAL(KIND=dp), PARAMETER :: root16pio7o3 = root16pi/sqrt7/3.0_dp
   REAL(KIND=dp), PARAMETER ::  s_root15o16pi = sqrt15*s_root1o16pi
   REAL(KIND=dp), PARAMETER :: root16pio15 = root16pi/sqrt15
   REAL(KIND=dp), PARAMETER ::  s_3root35o16pi = sqrt5*s_3root7o16pi
   REAL(KIND=dp), PARAMETER :: root16pio35o3 = root16pio7o3/sqrt5
   REAL(KIND=dp), PARAMETER ::  s_root105o16pi = 0.5_dp*s_root105o4pi
   REAL(KIND=dp), PARAMETER ::  root16pio105 = root4pio105*2.0_dp
   REAL(KIND=dp), PARAMETER ::  s_root1o32pi = 0.25_dp*sqrthalf*oorootpi
   REAL(KIND=dp), PARAMETER ::  root32pi = root16pi*sqrt2
   REAL(KIND=dp), PARAMETER ::  s_3root5o32pi = 3.0_dp*sqrt5*s_root1o32pi
   REAL(KIND=dp), PARAMETER ::  root32pio5o3 = root32pi/sqrt5/3.0_dp
   REAL(KIND=dp), PARAMETER ::  s_9root5o32pi = 9.0_dp*sqrt5*s_root1o32pi
   REAL(KIND=dp), PARAMETER ::  root32pio5o9 = root32pi/sqrt5/9.0_dp
   REAL(KIND=dp), PARAMETER ::  s_12root5o32pi = 12.0_dp*sqrt5*s_root1o32pi
   REAL(KIND=dp), PARAMETER ::  root32pio5o12 = root32pi/sqrt5/12.0_dp
   REAL(KIND=dp), PARAMETER ::  s_root21o32pi = sqrt21*s_root1o32pi
   REAL(KIND=dp), PARAMETER ::  root32pio21 = root32pi/sqrt21
   REAL(KIND=dp), PARAMETER ::  s_4root21o32pi = 4.0_dp*s_root21o32pi
   REAL(KIND=dp), PARAMETER ::  root32pio21o4 = root32pio21/4.0_dp
   REAL(KIND=dp), PARAMETER ::  s_root35o32pi = sqrt35*s_root1o32pi
   REAL(KIND=dp), PARAMETER ::  root32pio35 = root32pi/sqrt35
   REAL(KIND=dp), PARAMETER ::  s_3root35o32pi = 3.0_dp*s_root35o32pi
   REAL(KIND=dp), PARAMETER ::  s_9root35o32pi = 9.0_dp*s_root35o32pi
   REAL(KIND=dp), PARAMETER ::  s_18root35o32pi = 18.0_dp*s_root35o32pi
   REAL(KIND=dp), PARAMETER ::  s_root1o64pi = 0.125_dp*oorootpi
   REAL(KIND=dp), PARAMETER ::  s_3root5o64pi = 3.0_dp*sqrt5*s_root1o64pi
   REAL(KIND=dp), PARAMETER ::  s_18root5o64pi = 18.0_dp*sqrt5*s_root1o64pi
   REAL(KIND=dp), PARAMETER ::  s_root1o256pi = 0.0625_dp*oorootpi
   REAL(KIND=dp), PARAMETER ::  s_3root1o256pi = 3.0_dp*s_root1o256pi
   REAL(KIND=dp), PARAMETER ::  s_9root1o256pi = 9.0_dp*s_root1o256pi
   REAL(KIND=dp), PARAMETER ::  s_18root1o256pi = 18.0_dp*s_root1o256pi
   REAL(KIND=dp), PARAMETER ::  s_24root1o256pi = 24.0_dp*s_root1o256pi
   REAL(KIND=dp), PARAMETER ::  s_72root1o256pi = 72.0_dp*s_root1o256pi
   REAL(KIND=dp), PARAMETER ::  s_3root35o256pi = 3.0_dp*sqrt35*s_root1o256pi
   REAL(KIND=dp), PARAMETER ::  s_18root35o256pi = 18.0_dp*sqrt35*s_root1o256pi

   ! *** Public subroutines ***

   PUBLIC :: exp_radius, &
             exp_radius_very_extended, &
             gauss_exponent, &
             gaussint_sph, &
             trace_r_AxB

CONTAINS

! **************************************************************************************************
!> \brief  The exponent of a primitive Gaussian function for a given radius
!>         and threshold is calculated.
!> \param l ...
!> \param radius ...
!> \param threshold ...
!> \param prefactor ...
!> \retval exponent ...
!> \date   07.03.1999
!> \par Variables
!>      - exponent : Exponent of the primitive Gaussian function.
!>      - l        : Angular momentum quantum number l.
!>      - prefactor: Prefactor of the Gaussian function (e.g. a contraction
!>                   coefficient).
!>      - radius   : Calculated radius of the Gaussian function.
!>      - threshold: Threshold for radius.
!> \author MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION gauss_exponent(l, radius, threshold, prefactor) RESULT(exponent)
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), INTENT(IN)                          :: radius, threshold, prefactor
      REAL(KIND=dp)                                      :: exponent

      exponent = 0.0_dp

      IF (radius < 1.0E-6_dp) RETURN
      IF (threshold < 1.0E-12_dp) RETURN

      exponent = LOG(ABS(prefactor)*radius**l/threshold)/radius**2

   END FUNCTION gauss_exponent

! **************************************************************************************************
!> \brief  The radius of a primitive Gaussian function for a given threshold
!>         is calculated.
!>               g(r) = prefactor*r**l*exp(-alpha*r**2) - threshold = 0
!> \param l Angular momentum quantum number l.
!> \param alpha Exponent of the primitive Gaussian function.
!> \param threshold Threshold for radius.
!> \param prefactor Prefactor of the Gaussian function (e.g. a contraction
!>                     coefficient).
!> \param epsin Convergence criterion
!> \retval radius Calculated radius of the Gaussian functio
!> \date    26.02.1999
!> \par History
!>            - Modified to bisection to avoid division by zero (01.2002, Joost)
!>            - that means the old version contains bugs .......(04.2002, Joost)
!>              so keep this one ...
!> \par Variables
!>        - g      : The function g(r).
!>        - maxiter: Maximum number of iterations.
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION exp_radius(l, alpha, threshold, prefactor, epsin) RESULT(radius)
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), INTENT(IN)                          :: alpha, threshold, prefactor
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: epsin
      REAL(KIND=dp)                                      :: radius

      CHARACTER(len=*), PARAMETER :: routineN = 'exp_radius', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: maxiter = 5000

      INTEGER                                            :: iter
      REAL(KIND=dp)                                      :: a, ar2, d, epsiter, g, r, rhigh, rlow, &
                                                            rmid, t

      IF (PRESENT(epsin)) THEN
         epsiter = epsin
      ELSE
         epsiter = EPSILON(epsiter)*(1.0E-12_dp/2.22044604925031E-16_dp)
      ENDIF

      ! Initialize function value

      radius = 0.0_dp

      ! Load and check parameter values

      IF (l < 0) THEN
         CALL cp__b("aobasis/ao_util.F",186,"The angular momentum quantum number is negative")
      END IF

      IF (alpha == 0.0_dp) THEN
         CALL cp__b("aobasis/ao_util.F",190,"The Gaussian function exponent is zero")
      ELSE
         a = ABS(alpha)
      END IF

      IF (threshold == 0.0_dp) THEN
         CALL cp__b("aobasis/ao_util.F",196,"The requested threshold is zero")
      ELSE
         t = ABS(threshold)
      END IF

      IF (prefactor == 0.0_dp) THEN
         RETURN
      ELSE
         d = ABS(prefactor)
      END IF

      ! Calculate the Maximum g(r)
      r = SQRT(0.5_dp*REAL(l, dp)/a)
      ar2 = a*r*r

      IF (l == 0) THEN
         g = d
      ELSE
         g = d*r**l*EXP(-ar2)
      END IF

      IF (t > g) THEN
         RETURN
      END IF

      rlow = r
      rhigh = 2.0_dp*rlow+1.0_dp
      iter = 0
      DO
         iter = iter+1
         IF (iter .GT. maxiter) THEN
            CALL cp__b("aobasis/ao_util.F",227,"Maximum number of iterations exceeded")
         END IF
         g = d*rhigh**l*EXP(-a*rhigh**2)
         IF (g < t) EXIT
         rlow = rhigh
         rhigh = 2.0_dp*rlow+1.0_dp
      ENDDO

      DO iter = 1, maxiter
         rmid = (rlow+rhigh)*0.5_dp
         ar2 = a*rmid*rmid
         g = d*rmid**l*EXP(-ar2)
         IF (g .LT. t) THEN
            rhigh = rmid
         ELSE
            rlow = rmid
         ENDIF
         IF (ABS(rhigh-rlow) .LT. epsiter) THEN
            radius = rhigh
            RETURN
         ENDIF
      ENDDO
      CALL cp__b("aobasis/ao_util.F",249,"Maximum number of iterations exceeded")

   END FUNCTION exp_radius

! **************************************************************************************************
!> \brief computes the radius of the Gaussian outside of which it is smaller
!>      than eps
!> \param la_min ...
!> \param la_max ...
!> \param lb_min ...
!> \param lb_max ...
!> \param pab ...
!> \param o1 ...
!> \param o2 ...
!> \param ra ...
!> \param rb ...
!> \param rp ...
!> \param zetp ...
!> \param eps ...
!> \param prefactor ...
!> \param cutoff ...
!> \param epsin ...
!> \retval radius ...
!> \par History
!>      03.2007 new version that assumes that the Gaussians origante from spherical
!>              Gaussians
!> \note
!>      can optionally screen by the maximum element of the pab block
! **************************************************************************************************
   FUNCTION exp_radius_very_extended(la_min, la_max, lb_min, lb_max, pab, o1, o2, ra, rb, rp, &
                                     zetp, eps, prefactor, cutoff, epsin) RESULT(radius)

      INTEGER, INTENT(IN)                                :: la_min, la_max, lb_min, lb_max
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: pab
      INTEGER, OPTIONAL                                  :: o1, o2
      REAL(KIND=dp), INTENT(IN)                          :: ra(3), rb(3), rp(3), zetp, eps, &
                                                            prefactor, cutoff
      REAL(KIND=dp), OPTIONAL                            :: epsin
      REAL(KIND=dp)                                      :: radius

      INTEGER                                            :: i, ico, j, jco, la(3), lb(3), lxa, lxb, &
                                                            lya, lyb, lza, lzb
      REAL(KIND=dp)                                      :: bini, binj, coef(0:20), epsin_local, &
                                                            polycoef(0:60), prefactor_local, &
                                                            rad_a, rad_b, s1, s2

! get the local prefactor, we'll now use the largest density matrix element of the block to screen

      epsin_local = 1.0E-2_dp
      IF (PRESENT(epsin)) epsin_local = epsin

      IF (PRESENT(pab)) THEN
         prefactor_local = cutoff
         DO lxa = 0, la_max
         DO lxb = 0, lb_max
            DO lya = 0, la_max-lxa
            DO lyb = 0, lb_max-lxb
               DO lza = MAX(la_min-lxa-lya, 0), la_max-lxa-lya
               DO lzb = MAX(lb_min-lxb-lyb, 0), lb_max-lxb-lyb
                  la = (/lxa, lya, lza/)
                  lb = (/lxb, lyb, lzb/)
                  ico = coset(lxa, lya, lza)
                  jco = coset(lxb, lyb, lzb)
                  prefactor_local = MAX(ABS(pab(o1+ico, o2+jco)), prefactor_local)
               ENDDO
               ENDDO
            ENDDO
            ENDDO
         ENDDO
         ENDDO
         prefactor_local = prefactor*prefactor_local
      ELSE
         prefactor_local = prefactor*MAX(1.0_dp, cutoff)
      ENDIF

      !
      ! assumes that we can compute the radius for the case where
      ! the Gaussians a and b are both on the z - axis, but at the same
      ! distance as the original a and b
      !
      rad_a = SQRT(SUM((ra-rp)**2))
      rad_b = SQRT(SUM((rb-rp)**2))

      polycoef(0:la_max+lb_max) = 0.0_dp
      DO lxa = 0, la_max
      DO lxb = 0, lb_max
         coef(0:la_max+lb_max) = 0.0_dp
         bini = 1.0_dp
         s1 = 1.0_dp
         DO i = 0, lxa
            binj = 1.0_dp
            s2 = 1.0_dp
            DO j = 0, lxb
               coef(lxa+lxb-i-j) = coef(lxa+lxb-i-j)+bini*binj*s1*s2
               binj = (binj*(lxb-j))/(j+1)
               s2 = s2*(rad_b)
            ENDDO
            bini = (bini*(lxa-i))/(i+1)
            s1 = s1*(rad_a)
         ENDDO
         DO i = 0, lxa+lxb
            polycoef(i) = MAX(polycoef(i), coef(i))
         ENDDO
      ENDDO
      ENDDO

      polycoef(0:la_max+lb_max) = polycoef(0:la_max+lb_max)*prefactor_local
      radius = 0.0_dp
      DO i = 0, la_max+lb_max
         radius = MAX(radius, exp_radius(i, zetp, eps, polycoef(i), epsin_local))
      ENDDO

   END FUNCTION exp_radius_very_extended

! **************************************************************************************************
!> \brief ...
!> \param alpha ...
!> \param l ...
!> \retval gaussint_sph ...
! **************************************************************************************************
   FUNCTION gaussint_sph(alpha, l)

      !  calculates the radial integral over a spherical Gaussian
      !  of the form
      !     r**(2+l) * exp(-alpha * r**2)
      !

      REAL(dp), INTENT(IN)                               :: alpha
      INTEGER, INTENT(IN)                                :: l
      REAL(dp)                                           :: gaussint_sph

      IF ((l/2)*2 == l) THEN
         !even l:
         gaussint_sph = ROOTPI*0.5_dp**(l/2+2)*dfac(l+1) &
                        /SQRT(alpha)**(l+3)
      ELSE
         !odd l:
         gaussint_sph = 0.5_dp*fac((l+1)/2)/SQRT(alpha)**(l+3)
      ENDIF

   END FUNCTION gaussint_sph

! **************************************************************************************************
!> \brief ...
!> \param A ...
!> \param lda ...
!> \param B ...
!> \param ldb ...
!> \param m ...
!> \param n ...
!> \retval trace_r_AxB ...
! **************************************************************************************************
   PURE FUNCTION trace_r_AxB(A, lda, B, ldb, m, n)

      INTEGER, INTENT(in)                                :: lda
      REAL(dp), INTENT(in)                               :: A(lda, *)
      INTEGER, INTENT(in)                                :: ldb
      REAL(dp), INTENT(in)                               :: B(ldb, *)
      INTEGER, INTENT(in)                                :: m, n
      REAL(dp)                                           :: trace_r_AxB

      INTEGER                                            :: i1, i2, imod, mminus3
      REAL(dp)                                           :: t1, t2, t3, t4

      t1 = 0._dp
      t2 = 0._dp
      t3 = 0._dp
      t4 = 0._dp
      imod = MODULO(m, 4)
      SELECT CASE (imod)
      CASE (0)
         DO i2 = 1, n
            DO i1 = 1, m, 4
               t1 = t1+A(i1, i2)*B(i1, i2)
               t2 = t2+A(i1+1, i2)*B(i1+1, i2)
               t3 = t3+A(i1+2, i2)*B(i1+2, i2)
               t4 = t4+A(i1+3, i2)*B(i1+3, i2)
            ENDDO
         ENDDO
      CASE (1)
         mminus3 = m-3
         DO i2 = 1, n
            DO i1 = 1, mminus3, 4
               t1 = t1+A(i1, i2)*B(i1, i2)
               t2 = t2+A(i1+1, i2)*B(i1+1, i2)
               t3 = t3+A(i1+2, i2)*B(i1+2, i2)
               t4 = t4+A(i1+3, i2)*B(i1+3, i2)
            ENDDO
            t1 = t1+A(m, i2)*B(m, i2)
         ENDDO
      CASE (2)
         mminus3 = m-3
         DO i2 = 1, n
            DO i1 = 1, mminus3, 4
               t1 = t1+A(i1, i2)*B(i1, i2)
               t2 = t2+A(i1+1, i2)*B(i1+1, i2)
               t3 = t3+A(i1+2, i2)*B(i1+2, i2)
               t4 = t4+A(i1+3, i2)*B(i1+3, i2)
            ENDDO
            t1 = t1+A(m-1, i2)*B(m-1, i2)
            t2 = t2+A(m, i2)*B(m, i2)
         ENDDO
      CASE (3)
         mminus3 = m-3
         DO i2 = 1, n
            DO i1 = 1, mminus3, 4
               t1 = t1+A(i1, i2)*B(i1, i2)
               t2 = t2+A(i1+1, i2)*B(i1+1, i2)
               t3 = t3+A(i1+2, i2)*B(i1+2, i2)
               t4 = t4+A(i1+3, i2)*B(i1+3, i2)
            ENDDO
            t1 = t1+A(m-2, i2)*B(m-2, i2)
            t2 = t2+A(m-1, i2)*B(m-1, i2)
            t3 = t3+A(m, i2)*B(m, i2)
         ENDDO
      END SELECT
      trace_r_AxB = t1+t2+t3+t4

   END FUNCTION trace_r_AxB

! **************************************************************************************************
!> \brief ...
!> \param CPC_co ...
!> \param CPC_so ...
!> \param maxl ...
!> \param lm1 ...
!> \param lm2 ...
! **************************************************************************************************
   SUBROUTINE transform_c2s(CPC_co, CPC_so, maxl, lm1, lm2)

      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: CPC_co
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: CPC_so
      INTEGER                                            :: maxl, lm1, lm2

      CHARACTER(len=*), PARAMETER :: routineN = 'transform_c2s', routineP = moduleN//':'//routineN

      INTEGER                                            :: ic1, ic2, is1, is2, istat, l, lx, ly, lz
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: work

      ALLOCATE (work(ncoset(maxl), nsoset(maxl)), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 490, &
                          "work", dp_size*ncoset(maxl)*nsoset(maxl))
      END IF
      CPC_so = 0.0_dp
      work = 0.0_dp

      !    DO l = 0,maxl
      !      ncgf = nco(l)
      !      nsgf = nso(l)
      !      ico = ncoset(l-1) + 1
      !      iso = nsoset(l-1) + 1
      !
      !      CALL dgemm("N","T",ncoset(maxl),nsgf,ncgf,&
      !                 1.0_dp,CPC_co(1,ico),ncoset(maxl),&
      !                 orbtramat(l)%c2s(1,1),nsgf,&
      !                 0.0_dp,work(1,iso),ncoset(maxl))
      !    ENDDO
      !
      !    DO l = 0,maxl
      !      ncgf = nco(l)
      !      nsgf = nso(l)
      !      ico = ncoset(l-1) + 1
      !      iso = nsoset(l-1) + 1
      !
      !      CALL dgemm("N","N",nsgf,nsoset(maxl),ncgf,&
      !                 1.0_dp,orbtramat(l)%c2s(1,1),nsgf,&
      !                 work(ico,1),ncoset(maxl),&
      !                 0.0_dp,CPC_so(iso,1),nsoset(maxl))
      !    ENDDO

      !  do l = 0,maxl
      !     do is1 = 1,nso(l)
      !      write(*,'(A,2I3,10f10.5)') 'or ', l, is1, orbtramat(l)%c2s(is1,1:nco(l))
      !     enddo
      !  enddo
      !
      !  stop

      CPC_so = 0.0_dp
      work = 0.0_dp
      DO ic1 = 1, ncoset(lm1)
         DO l = 0, lm2
            DO is2 = 1, nso(l)
               DO ic2 = 1, nco(l)
                  lx = indco(1, ic2+ncoset(l-1))
                  ly = indco(2, ic2+ncoset(l-1))
                  lz = indco(3, ic2+ncoset(l-1))
                  work(ic1, is2+nsoset(l-1)) = &
                     work(ic1, is2+nsoset(l-1))+ &
                     CPC_co(ic1, ic2+ncoset(l-1))* &
                     orbtramat(l)%c2s(is2, ic2)* &
                     SQRT(fourpi/dfac(2*l+1)* &
                          dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))
                  !         write(*,*) 'dfac 1', dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1)
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      DO is2 = 1, nsoset(lm2)
         DO l = 0, lm1
            DO is1 = 1, nso(l)
               DO ic1 = 1, nco(l)
                  lx = indco(1, ic1+ncoset(l-1))
                  ly = indco(2, ic1+ncoset(l-1))
                  lz = indco(3, ic1+ncoset(l-1))
                  CPC_so(is1+nsoset(l-1), is2) = &
                     CPC_so(is1+nsoset(l-1), is2)+ &
                     work(ic1+ncoset(l-1), is2)* &
                     orbtramat(l)%c2s(is1, ic1)* &
                     SQRT(fourpi/dfac(2*l+1)* &
                          dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))
                  !          write(*,*) 'dfac 2', dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1)
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      DEALLOCATE (work)

   END SUBROUTINE transform_c2s

! **************************************************************************************************
!> \brief ...
!> \param CPC_co ...
!> \param CPC_so ...
!> \param maxl ...
! **************************************************************************************************
   SUBROUTINE transform_c2s_new(CPC_co, CPC_so, maxl)

      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: CPC_co
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: CPC_so
      INTEGER                                            :: maxl

      CHARACTER(len=*), PARAMETER :: routineN = 'transform_c2s_new', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iso, istat, ldc, lds
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: work

      ALLOCATE (work(ncoset(maxl), ncoset(maxl), 1), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 592, &
                          "work", dp_size*ncoset(maxl)*nsoset(maxl))
      END IF
      ldc = ncoset(maxl)
      lds = nsoset(maxl)
      CPC_so = 0.0_dp
      work = 0.0_dp

      work(1:ldc, 1:ldc, 1) = CPC_co(1:ldc, 1:ldc)

      CALL cart2sph_mat(work, ldc, ldc, 1, maxl, maxl)

      DO iso = 1, nsoset(maxl)
         CPC_so(1:nsoset(maxl), iso) = work(1:nsoset(maxl), iso, 1)
      END DO

      DEALLOCATE (work)

   END SUBROUTINE transform_c2s_new

! **************************************************************************************************
!> \brief ...
!> \param matso ...
!> \param matco ...
!> \param maxl ...
!> \param lm1 ...
!> \param lm2 ...
! **************************************************************************************************
   SUBROUTINE transform_s2c(matso, matco, maxl, lm1, lm2)

      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: matso
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: matco
      INTEGER, INTENT(IN)                                :: maxl, lm1, lm2

      CHARACTER(len=*), PARAMETER :: routineN = 'transform_s2c', routineP = moduleN//':'//routineN

      INTEGER                                            :: ic1, ic2, ico, is1, is2, iso, istat, l, &
                                                            lx, ly, lz, nc1, nc2, ns1, ns2
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: matsc

      ALLOCATE (matsc(nsoset(maxl), ncoset(maxl)), STAT=istat)
      IF (istat /= 0) THEN
         CALL stop_memory(routineN, moduleN, 634, &
                          "work", nsoset(maxl)*ncoset(maxl))
      END IF
      matco = 0.0_dp
      matsc = 0.0_dp

      ns1 = nsoset(lm1)
      ns2 = nsoset(lm2)
      nc1 = ncoset(lm1)
      nc2 = ncoset(lm2)

      !   A = matso (ns1 x ns2)
      !   C = Tt x A x T = matco (nc1 x nc2)
      !   B = A x T      = matsc (ns1 x nc2)

      !   Calculate B
      DO is1 = 1, ns1
         DO l = 0, lm2
            DO ico = 1, nco(l)
               ic2 = ncoset(l-1)+ico
               lx = indco(1, ico+ncoset(l-1))
               ly = indco(2, ico+ncoset(l-1))
               lz = indco(3, ico+ncoset(l-1))
               DO iso = 1, nso(l)
                  is2 = nsoset(l-1)+iso
                  matsc(is1, ic2) = matsc(is1, ic2)+ &
                                    matso(is1, is2)*orbtramat(l)%s2c(iso, ico)* &
                                    SQRT((fourpi)/dfac(2*l+1)* &
                                         dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))
               ENDDO ! iso
            ENDDO ! ico
         ENDDO ! l
      ENDDO ! is1

      !   Calculate C
      DO ic2 = 1, nc2
         DO l = 0, lm1
            DO ico = 1, nco(l)
               ic1 = ncoset(l-1)+ico
               lx = indco(1, ico+ncoset(l-1))
               ly = indco(2, ico+ncoset(l-1))
               lz = indco(3, ico+ncoset(l-1))
               DO iso = 1, nso(l)
                  is1 = nsoset(l-1)+iso
                  matco(ic1, ic2) = matco(ic1, ic2)+ &
                                    matsc(is1, ic2)*orbtramat(l)%s2c(iso, ico)* &
                                    SQRT((fourpi)/dfac(2*l+1)* &
                                         dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))
               ENDDO ! iso
            ENDDO ! ico
         ENDDO ! l
      ENDDO ! ic2

      DEALLOCATE (matsc)

   END SUBROUTINE transform_s2c
   !------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief ...
!> \param mat ...
!> \param ld_mat ...
!> \param sd_mat ...
!> \param n ...
!> \param lmax1 ...
!> \param lmax2 ...
! **************************************************************************************************
   SUBROUTINE sph2cart_mat(mat, ld_mat, sd_mat, n, lmax1, lmax2)
      INTEGER                                            :: ld_mat, sd_mat, n
      REAL(dp)                                           :: mat(ld_mat, sd_mat, n)
      INTEGER                                            :: lmax1, lmax2

      CHARACTER(len=*), PARAMETER :: routineN = 'sph2cart_mat', routineP = moduleN//':'//routineN

      INTEGER                                            :: idx_l1, idx_l2
      REAL(dp) :: mat_aux(ncoset(lmax1), nsoset(lmax2), n)

      IF (lmax1 > 3 .OR. lmax2 > 3) THEN
         CALL cp__b("aobasis/ao_util.F",711,"l>3 not implemented")
      END IF
      DO idx_l2 = 1, nsoset(lmax2)

         mat_aux(1, idx_l2, :) &
            = s_root1o4pi*mat(1, idx_l2, :)

         IF (lmax1 == 0) CYCLE

         mat_aux(2, idx_l2, :) &
            = -s_root3o4pi*mat(4, idx_l2, :)

         mat_aux(3, idx_l2, :) &
            = -s_root3o4pi*mat(2, idx_l2, :)

         mat_aux(4, idx_l2, :) &
            = s_root3o4pi*mat(3, idx_l2, :)

         IF (lmax1 == 1) CYCLE

         mat_aux(5, idx_l2, :) &
            = -s_root5o16pi*mat(7, idx_l2, :) &
              +s_root15o16pi*mat(9, idx_l2, :)

         mat_aux(6, idx_l2, :) &
            = s_root15o4pi*mat(5, idx_l2, :)

         mat_aux(7, idx_l2, :) &
            = -s_root5o16pi*mat(7, idx_l2, :) &
              -s_root15o16pi*mat(9, idx_l2, :)

         mat_aux(8, idx_l2, :) &
            = -s_root15o4pi*mat(8, idx_l2, :)

         mat_aux(9, idx_l2, :) &
            = -s_root15o4pi*mat(6, idx_l2, :)

         mat_aux(10, idx_l2, :) &
            = s_2root5o16pi*mat(7, idx_l2, :)

         IF (lmax1 == 2) CYCLE

         mat_aux(11, idx_l2, :) &
            = -s_root21o32pi*mat(14, idx_l2, :) &
              +s_root35o32pi*mat(16, idx_l2, :)

         mat_aux(12, idx_l2, :) &
            = -s_3root35o32pi*mat(10, idx_l2, :) &
              +s_root21o32pi*mat(12, idx_l2, :)

         mat_aux(13, idx_l2, :) &
            = s_root21o32pi*mat(14, idx_l2, :) &
              +s_3root35o32pi*mat(16, idx_l2, :)

         mat_aux(14, idx_l2, :) &
            = s_root35o32pi*mat(10, idx_l2, :) &
              +s_root21o32pi*mat(12, idx_l2, :)

         mat_aux(15, idx_l2, :) &
            = -s_3root7o16pi*mat(13, idx_l2, :) &
              +s_root105o16pi*mat(15, idx_l2, :)

         mat_aux(16, idx_l2, :) &
            = s_root105o4pi*mat(11, idx_l2, :)

         mat_aux(17, idx_l2, :) &
            = -s_3root7o16pi*mat(13, idx_l2, :) &
              -s_root105o16pi*mat(15, idx_l2, :)

         mat_aux(18, idx_l2, :) &
            = -s_4root21o32pi*mat(14, idx_l2, :)

         mat_aux(19, idx_l2, :) &
            = -s_4root21o32pi*mat(12, idx_l2, :)

         mat_aux(20, idx_l2, :) &
            = s_2root7o16pi*mat(13, idx_l2, :)

      ENDDO

      DO idx_l1 = 1, ncoset(lmax1)

         mat(idx_l1, 1, :) &
            = s_root1o4pi*mat_aux(idx_l1, 1, :)

         IF (lmax2 == 0) CYCLE

         mat(idx_l1, 2, :) &
            = -s_root3o4pi*mat_aux(idx_l1, 4, :)

         mat(idx_l1, 3, :) &
            = -s_root3o4pi*mat_aux(idx_l1, 2, :)

         mat(idx_l1, 4, :) &
            = s_root3o4pi*mat_aux(idx_l1, 3, :)

         IF (lmax2 == 1) CYCLE

         mat(idx_l1, 5, :) &
            = -s_root5o16pi*mat_aux(idx_l1, 7, :) &
              +s_root15o16pi*mat_aux(idx_l1, 9, :)

         mat(idx_l1, 6, :) &
            = s_root15o4pi*mat_aux(idx_l1, 5, :)

         mat(idx_l1, 7, :) &
            = -s_root5o16pi*mat_aux(idx_l1, 7, :) &
              -s_root15o16pi*mat_aux(idx_l1, 9, :)

         mat(idx_l1, 8, :) &
            = -s_root15o4pi*mat_aux(idx_l1, 8, :)

         mat(idx_l1, 9, :) &
            = -s_root15o4pi*mat_aux(idx_l1, 6, :)

         mat(idx_l1, 10, :) &
            = s_2root5o16pi*mat_aux(idx_l1, 7, :)

         IF (lmax2 == 2) CYCLE

         mat(idx_l1, 11, :) &
            = -s_root21o32pi*mat_aux(idx_l1, 14, :) &
              +s_root35o32pi*mat_aux(idx_l1, 16, :)

         mat(idx_l1, 12, :) &
            = -s_3root35o32pi*mat_aux(idx_l1, 10, :) &
              +s_root21o32pi*mat_aux(idx_l1, 12, :)

         mat(idx_l1, 13, :) &
            = s_root21o32pi*mat_aux(idx_l1, 14, :) &
              +s_3root35o32pi*mat_aux(idx_l1, 16, :)

         mat(idx_l1, 14, :) &
            = s_root35o32pi*mat_aux(idx_l1, 10, :) &
              +s_root21o32pi*mat_aux(idx_l1, 12, :)

         mat(idx_l1, 15, :) &
            = -s_3root7o16pi*mat_aux(idx_l1, 13, :) &
              +s_root105o16pi*mat_aux(idx_l1, 15, :)

         mat(idx_l1, 16, :) &
            = s_root105o4pi*mat_aux(idx_l1, 11, :)

         mat(idx_l1, 17, :) &
            = -s_3root7o16pi*mat_aux(idx_l1, 13, :) &
              -s_root105o16pi*mat_aux(idx_l1, 15, :)

         mat(idx_l1, 18, :) &
            = -s_4root21o32pi*mat_aux(idx_l1, 14, :)

         mat(idx_l1, 19, :) &
            = -s_4root21o32pi*mat_aux(idx_l1, 12, :)

         mat(idx_l1, 20, :) &
            = s_2root7o16pi*mat_aux(idx_l1, 13, :)

      ENDDO

   END SUBROUTINE sph2cart_mat

! **************************************************************************************************
!> \brief ...
!> \param mat ...
!> \param ld_mat ...
!> \param sd_mat ...
!> \param n ...
!> \param lmax1 ...
!> \param lmax2 ...
! **************************************************************************************************
   SUBROUTINE cart2sph_mat(mat, ld_mat, sd_mat, n, lmax1, lmax2)

      !in:
      INTEGER                                            :: ld_mat, sd_mat, n
      REAL(dp)                                           :: mat(ld_mat, sd_mat, n)
      INTEGER                                            :: lmax1, lmax2

      CHARACTER(len=*), PARAMETER :: routineN = 'cart2sph_mat', routineP = moduleN//':'//routineN

      INTEGER                                            :: idx_l2, idx_lm1
      REAL(dp) :: mat_aux(nsoset(lmax1), ncoset(lmax2), n)

      IF ((lmax1 > 3) .OR. (lmax2 > 3)) THEN
         CALL cp__b("aobasis/ao_util.F",893,"l > 3 not implemented")
      END IF

      DO idx_l2 = 1, ncoset(lmax2)

         mat_aux(1, idx_l2, :) &
            = root4pi*mat(1, idx_l2, :)

         IF (lmax1 == 0) CYCLE

         mat_aux(2, idx_l2, :) &
            = -root4pio3*mat(3, idx_l2, :)

         mat_aux(3, idx_l2, :) &
            = root4pio3*mat(4, idx_l2, :)

         mat_aux(4, idx_l2, :) &
            = -root4pio3*mat(2, idx_l2, :)

         IF (lmax1 == 1) CYCLE

         mat_aux(5, idx_l2, :) &
            = root4pio15*mat(6, idx_l2, :)

         mat_aux(6, idx_l2, :) &
            = -root4pio15*mat(9, idx_l2, :)

         mat_aux(7, idx_l2, :) &
            = -0.5_dp*root4pio5*mat(5, idx_l2, :) &
              -0.5_dp*root4pio5*mat(7, idx_l2, :) &
              +root4pio5*mat(10, idx_l2, :)

         mat_aux(8, idx_l2, :) &
            = -root4pio15*mat(8, idx_l2, :)

         mat_aux(9, idx_l2, :) &
            = 0.5_dp*root4pio15*mat(5, idx_l2, :) &
              -0.5_dp*root4pio15*mat(7, idx_l2, :)

         IF (lmax1 == 2) CYCLE

         mat_aux(10, idx_l2, :) &
            = -s_3root35o32pi*mat(12, idx_l2, :) &
              +s_root35o32pi*mat(14, idx_l2, :)

         mat_aux(11, idx_l2, :) &
            = s_root105o4pi*mat(16, idx_l2, :)

         mat_aux(12, idx_l2, :) &
            = s_root21o32pi*mat(12, idx_l2, :) &
              +s_root21o32pi*mat(14, idx_l2, :) &
              -s_4root21o32pi*mat(19, idx_l2, :)

         mat_aux(13, idx_l2, :) &
            = -s_3root7o16pi*mat(15, idx_l2, :) &
              -s_3root7o16pi*mat(17, idx_l2, :) &
              +s_2root7o16pi*mat(20, idx_l2, :)

         mat_aux(14, idx_l2, :) &
            = s_root21o32pi*mat(11, idx_l2, :) &
              +s_root21o32pi*mat(13, idx_l2, :) &
              -s_4root21o32pi*mat(18, idx_l2, :)

         mat_aux(15, idx_l2, :) &
            = s_root105o16pi*mat(15, idx_l2, :) &
              -s_root105o16pi*mat(17, idx_l2, :)

         mat_aux(16, idx_l2, :) &
            = -s_root35o32pi*mat(11, idx_l2, :) &
              +s_3root35o32pi*mat(13, idx_l2, :)

      ENDDO

      DO idx_lm1 = 1, nsoset(lmax1)

         mat(idx_lm1, 1, :) &
            = root4pi*mat_aux(idx_lm1, 1, :)

         IF (lmax2 == 0) CYCLE

         mat(idx_lm1, 2, :) &
            = -root4pio3*mat_aux(idx_lm1, 3, :)

         mat(idx_lm1, 3, :) &
            = root4pio3*mat_aux(idx_lm1, 4, :)

         mat(idx_lm1, 4, :) &
            = -root4pio3*mat_aux(idx_lm1, 2, :)

         IF (lmax2 == 1) CYCLE

         mat(idx_lm1, 5, :) &
            = root4pio15*mat_aux(idx_lm1, 6, :)

         mat(idx_lm1, 6, :) &
            = -root4pio15*mat_aux(idx_lm1, 9, :)

         mat(idx_lm1, 7, :) &
            = -0.5_dp*root4pio5*mat_aux(idx_lm1, 5, :) &
              -0.5_dp*root4pio5*mat_aux(idx_lm1, 7, :) &
              +root4pio5*mat_aux(idx_lm1, 10, :)

         mat(idx_lm1, 8, :) &
            = -root4pio15*mat_aux(idx_lm1, 8, :)

         mat(idx_lm1, 9, :) &
            = 0.5_dp*root4pio15*mat_aux(idx_lm1, 5, :) &
              -0.5_dp*root4pio15*mat_aux(idx_lm1, 7, :)

         IF (lmax2 == 2) CYCLE

         mat(idx_lm1, 10, :) &
            = -s_3root35o32pi*mat_aux(idx_lm1, 12, :) &
              +s_root35o32pi*mat_aux(idx_lm1, 14, :)

         mat(idx_lm1, 11, :) &
            = s_root105o4pi*mat_aux(idx_lm1, 16, :)

         mat(idx_lm1, 12, :) &
            = s_root21o32pi*mat_aux(idx_lm1, 12, :) &
              +s_root21o32pi*mat_aux(idx_lm1, 14, :) &
              -s_4root21o32pi*mat_aux(idx_lm1, 19, :)

         mat(idx_lm1, 13, :) &
            = -s_3root7o16pi*mat_aux(idx_lm1, 15, :) &
              -s_3root7o16pi*mat_aux(idx_lm1, 17, :) &
              +s_2root7o16pi*mat_aux(idx_lm1, 20, :)

         mat(idx_lm1, 14, :) &
            = s_root21o32pi*mat_aux(idx_lm1, 11, :) &
              +s_root21o32pi*mat_aux(idx_lm1, 13, :) &
              -s_4root21o32pi*mat_aux(idx_lm1, 18, :)

         mat(idx_lm1, 15, :) &
            = s_root105o16pi*mat_aux(idx_lm1, 15, :) &
              -s_root105o16pi*mat_aux(idx_lm1, 17, :)

         mat(idx_lm1, 16, :) &
            = -s_root35o32pi*mat_aux(idx_lm1, 11, :) &
              +s_3root35o32pi*mat_aux(idx_lm1, 13, :)

      ENDDO

   END SUBROUTINE cart2sph_mat

END MODULE ao_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Periodic Table related data definitions
!> \par Literature
!>      M. E. Wieser and M. Berglund,
!>      Atomic weights of the elements 2007 (IUPAC Technical Report),
!>      Pure Appl. Chem. 81, 2131-2156 (2009)
!> \par History
!>      27.05.2010 Update of the atomic weights
!>      20.10.2009 Use vdW Radii from the Cambridge Structural Database
!> \author JGH, MK
! **************************************************************************************************
MODULE periodic_table

   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: init_periodic_table, ptable, nelem, &
             get_ptable_info

   TYPE atom
      CHARACTER(LEN=2)  :: symbol
      CHARACTER(LEN=14) :: name
      INTEGER           :: number
      REAL(KIND=dp)     :: amass ! Standard atomic weight, i.e. the relative atomic mass
      ! of an element derived by averaging the relative
      ! atomic masses of the (natural) isotopes of that
      ! element or the relative (possibly estimated)
      ! longest half-life period in formula units.
      REAL(KIND=dp)     :: covalent_radius ! in Angstroms
      REAL(KIND=dp)     :: vdw_radius ! in Angstroms
      INTEGER           :: e_conv(0:3)
      REAL(KIND=dp)     :: heat_of_formation ! in kcal/mol
      REAL(KIND=dp)     :: eht_param(0:3) ! in eV
      REAL(KIND=dp)     :: gyrom_ratio ! in MHz/Tesla
      INTEGER           :: gyrom_ratio_isotope ! isotope number corresponding to gyrom_ratio
   END TYPE atom

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'periodic_table'
   REAL(KIND=dp), PARAMETER             :: z = 0.0_dp
   INTEGER, PARAMETER                   :: nelem = 118

   TYPE(atom) :: ptable(0:nelem)

CONTAINS

! **************************************************************************************************
!> \brief Pass information about the kind given the element symbol
!> \param symbol ...
!> \param number ...
!> \param amass ...
!> \param ielement ...
!> \param covalent_radius ...
!> \param vdw_radius ...
!> \param found ...
!> \par History
!>      none
!> \author MI
! **************************************************************************************************
   SUBROUTINE get_ptable_info(symbol, number, amass, ielement, covalent_radius, &
                              vdw_radius, found)

      CHARACTER(LEN=2), INTENT(IN)                       :: symbol
      INTEGER, INTENT(OUT), OPTIONAL                     :: number
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: amass
      INTEGER, INTENT(OUT), OPTIONAL                     :: ielement
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: covalent_radius, vdw_radius
      LOGICAL, OPTIONAL                                  :: found

      CHARACTER(len=*), PARAMETER :: routineN = 'get_ptable_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: symb_ielem, symbol_in
      INTEGER                                            :: ielem, ielem_found

      ielem_found = -1
      symbol_in = symbol
      CALL uppercase(symbol_in)
      DO ielem = 1, nelem
         symb_ielem = ptable(ielem)%symbol
         CALL uppercase(symb_ielem)
         IF (symbol_in == symb_ielem) THEN
            ielem_found = ielem
            EXIT
         END IF
      END DO

      ! If requested give back the value and avoid failing
      IF (PRESENT(found)) THEN
         found = (ielem_found > 0)
      ELSE
         IF (ielem_found < 0) THEN
            CALL cp__b("common/periodic_table.F",102,"Unknown element symbol <"//TRIM(symbol)//"> found.")
         END IF
      END IF

      ! If the element has been identified provide the requested information
      IF (ielem_found > 0) THEN
         IF (PRESENT(ielement)) ielement = ielem_found
         IF (PRESENT(number)) number = ptable(ielem_found)%number
         IF (PRESENT(amass)) amass = ptable(ielem_found)%amass
         IF (PRESENT(covalent_radius)) covalent_radius = ptable(ielem_found)%covalent_radius
         IF (PRESENT(vdw_radius)) vdw_radius = ptable(ielem_found)%vdw_radius
      END IF

   END SUBROUTINE get_ptable_info

! **************************************************************************************************
!> \brief Initialization of Periodic Table related data
!> \par History
!>      27.05.2010 Update of the atomic weights
!> \author JGH, MK
! **************************************************************************************************
   SUBROUTINE init_periodic_table()

      ! Dummy
      ptable(0)%symbol = 'X '
      ptable(0)%name = 'Dummy'
      ptable(0)%number = 0
      ptable(0)%amass = z
      ptable(0)%covalent_radius = z
      ptable(0)%vdw_radius = z
      ptable(0)%e_conv(0:3) = (/0, 0, 0, 0/)
      ptable(0)%eht_param(0:3) = (/z, z, z, z/)

      ! Hydrogen
      ptable(1)%symbol = 'H '
      ptable(1)%name = 'Hydrogen'
      ptable(1)%number = 1
      ptable(1)%amass = 1.00794_dp
      ptable(1)%covalent_radius = 0.32_dp
      ptable(1)%vdw_radius = 1.09_dp
      ptable(1)%e_conv(0:3) = (/1, 0, 0, 0/)
      ptable(1)%eht_param(0:3) = (/-13.60_dp, z, z, z/)

      ! Helium
      ptable(2)%symbol = 'He'
      ptable(2)%name = 'Helium'
      ptable(2)%number = 2
      ptable(2)%amass = 4.002602_dp
      ptable(2)%covalent_radius = 0.9300_dp
      ptable(2)%vdw_radius = 1.40_dp
      ptable(2)%e_conv(0:3) = (/2, 0, 0, 0/)
      ptable(2)%eht_param(0:3) = (/-23.40_dp, z, z, z/)

      ! Lithium
      ptable(3)%symbol = 'Li'
      ptable(3)%name = 'Lithium'
      ptable(3)%number = 3
      ptable(3)%amass = 6.941_dp
      ptable(3)%covalent_radius = 1.2300_dp
      ptable(3)%vdw_radius = 1.82_dp
      ptable(3)%e_conv(0:3) = (/3, 0, 0, 0/)
      ptable(3)%eht_param(0:3) = (/-5.40_dp, -3.50_dp, z, z/)

      ! Beryllium
      ptable(4)%symbol = 'Be'
      ptable(4)%name = 'Beryllium'
      ptable(4)%number = 4
      ptable(4)%amass = 9.012182_dp
      ptable(4)%covalent_radius = 0.9000_dp
      ptable(4)%vdw_radius = 2.00_dp
      ptable(4)%e_conv(0:3) = (/4, 0, 0, 0/)
      ptable(4)%eht_param(0:3) = (/-10.00_dp, -6.00_dp, z, z/)

      ! Boron
      ptable(5)%symbol = 'B '
      ptable(5)%name = 'Boron'
      ptable(5)%number = 5
      ptable(5)%amass = 10.811_dp
      ptable(5)%covalent_radius = 0.8200_dp
      ptable(5)%vdw_radius = 2.00_dp
      ptable(5)%e_conv(0:3) = (/4, 1, 0, 0/)
      ptable(5)%eht_param(0:3) = (/-15.20_dp, -8.50_dp, z, z/)

      ! Carbon
      ptable(6)%symbol = 'C '
      ptable(6)%name = 'Carbon'
      ptable(6)%number = 6
      ptable(6)%amass = 12.0107_dp
      ptable(6)%covalent_radius = 0.7700_dp
      ptable(6)%vdw_radius = 1.7_dp
      ptable(6)%e_conv(0:3) = (/4, 2, 0, 0/)
      ptable(6)%eht_param(0:3) = (/-21.40_dp, -11.40_dp, z, z/)

      ! Nitrogen
      ptable(7)%symbol = 'N '
      ptable(7)%name = 'Nitrogen'
      ptable(7)%number = 7
      ptable(7)%amass = 14.0067_dp
      ptable(7)%covalent_radius = 0.7500_dp
      ptable(7)%vdw_radius = 1.55_dp
      ptable(7)%e_conv(0:3) = (/4, 3, 0, 0/)
      ptable(7)%eht_param(0:3) = (/-26.00_dp, -13.40_dp, z, z/)

      ! Oxygen
      ptable(8)%symbol = 'O '
      ptable(8)%name = 'Oxygen'
      ptable(8)%number = 8
      ptable(8)%amass = 15.9994_dp
      ptable(8)%covalent_radius = 0.7300_dp
      ptable(8)%vdw_radius = 1.52_dp
      ptable(8)%e_conv(0:3) = (/4, 4, 0, 0/)
      ptable(8)%eht_param(0:3) = (/-32.30_dp, -14.80_dp, z, z/)

      ! Fluorine
      ptable(9)%symbol = 'F '
      ptable(9)%name = 'Fluorine'
      ptable(9)%number = 9
      ptable(9)%amass = 18.9984032_dp
      ptable(9)%covalent_radius = 0.7200_dp
      ptable(9)%vdw_radius = 1.47_dp
      ptable(9)%e_conv(0:3) = (/4, 5, 0, 0/)
      ptable(9)%eht_param(0:3) = (/-40.00_dp, -18.10_dp, z, z/)

      ! Neon
      ptable(10)%symbol = 'Ne'
      ptable(10)%name = 'Neon'
      ptable(10)%number = 10
      ptable(10)%amass = 20.1797_dp
      ptable(10)%covalent_radius = 0.7100_dp
      ptable(10)%vdw_radius = 1.54_dp
      ptable(10)%e_conv(0:3) = (/4, 6, 0, 0/)
      ptable(10)%eht_param(0:3) = (/-43.20_dp, -20.00_dp, z, z/)

      ! Sodium
      ptable(11)%symbol = 'Na'
      ptable(11)%name = 'Sodium'
      ptable(11)%number = 11
      ptable(11)%amass = 22.98976928_dp
      ptable(11)%covalent_radius = 1.5400_dp
      ptable(11)%vdw_radius = 2.27_dp
      ptable(11)%e_conv(0:3) = (/5, 6, 0, 0/)
      ptable(11)%eht_param(0:3) = (/-5.10_dp, -3.00_dp, z, z/)

      ! Magnesium
      ptable(12)%symbol = 'Mg'
      ptable(12)%name = 'Magnesium'
      ptable(12)%number = 12
      ptable(12)%amass = 24.305_dp
      ptable(12)%covalent_radius = 1.3600_dp
      ptable(12)%vdw_radius = 1.73_dp
      ptable(12)%e_conv(0:3) = (/6, 6, 0, 0/)
      ptable(12)%eht_param(0:3) = (/-9.00_dp, -4.50_dp, z, z/)

      ! Aluminium
      ptable(13)%symbol = 'Al'
      ptable(13)%name = 'Aluminium'
      ptable(13)%number = 13
      ptable(13)%amass = 26.9815386_dp
      ptable(13)%covalent_radius = 1.1800_dp
      ptable(13)%vdw_radius = 2.00_dp
      ptable(13)%e_conv(0:3) = (/6, 7, 0, 0/)
      ptable(13)%eht_param(0:3) = (/-12.30_dp, -6.50_dp, z, z/)

      ! Silicon
      ptable(14)%symbol = 'Si'
      ptable(14)%name = 'Silicon'
      ptable(14)%number = 14
      ptable(14)%amass = 28.0855_dp
      ptable(14)%covalent_radius = 1.1100_dp
      ptable(14)%vdw_radius = 2.10_dp
      ptable(14)%e_conv(0:3) = (/6, 8, 0, 0/)
      ptable(14)%eht_param(0:3) = (/-17.30_dp, -9.20_dp, z, z/)

      ! Phosphorus
      ptable(15)%symbol = 'P '
      ptable(15)%name = 'Phosphorus'
      ptable(15)%number = 15
      ptable(15)%amass = 30.973762_dp
      ptable(15)%covalent_radius = 1.0600_dp
      ptable(15)%vdw_radius = 1.80_dp
      ptable(15)%e_conv(0:3) = (/6, 9, 0, 0/)
      ptable(15)%eht_param(0:3) = (/-18.60_dp, -14.00_dp, z, z/)

      ! Sulfur
      ptable(16)%symbol = 'S '
      ptable(16)%name = 'Sulfur'
      ptable(16)%number = 16
      ptable(16)%amass = 32.065_dp
      ptable(16)%covalent_radius = 1.0200_dp
      ptable(16)%vdw_radius = 1.80_dp
      ptable(16)%e_conv(0:3) = (/6, 10, 0, 0/)
      ptable(16)%eht_param(0:3) = (/-20.00_dp, -11.00_dp, z, z/)

      ! Chlorine
      ptable(17)%symbol = 'Cl'
      ptable(17)%name = 'Chlorine'
      ptable(17)%number = 17
      ptable(17)%amass = 35.453_dp
      ptable(17)%covalent_radius = 0.9900_dp
      ptable(17)%vdw_radius = 1.75_dp
      ptable(17)%e_conv(0:3) = (/6, 11, 0, 0/)
      ptable(17)%eht_param(0:3) = (/-26.30_dp, -14.20_dp, z, z/)

      ! Argon
      ptable(18)%symbol = 'Ar'
      ptable(18)%name = 'Argon'
      ptable(18)%number = 18
      ptable(18)%amass = 39.948_dp
      ptable(18)%covalent_radius = 0.9800_dp
      ptable(18)%vdw_radius = 1.88_dp
      ptable(18)%e_conv(0:3) = (/6, 12, 0, 0/)
      ptable(18)%eht_param(0:3) = (/z, z, z, z/)

      ! Potassium
      ptable(19)%symbol = 'K '
      ptable(19)%name = 'Potassium'
      ptable(19)%number = 19
      ptable(19)%amass = 39.0983_dp
      ptable(19)%covalent_radius = 2.0300_dp
      ptable(19)%vdw_radius = 2.75_dp
      ptable(19)%e_conv(0:3) = (/7, 12, 0, 0/)
      ptable(19)%eht_param(0:3) = (/-4.34_dp, -2.73_dp, z, z/)

      ! Calcium
      ptable(20)%symbol = 'Ca'
      ptable(20)%name = 'Calcium'
      ptable(20)%number = 20
      ptable(20)%amass = 40.078_dp
      ptable(20)%covalent_radius = 1.7400_dp
      ptable(20)%vdw_radius = 2.00_dp
      ptable(20)%e_conv(0:3) = (/8, 12, 0, 0/)
      ptable(20)%eht_param(0:3) = (/-7.00_dp, -4.00_dp, z, z/)

      ! Scandium
      ptable(21)%symbol = 'Sc'
      ptable(21)%name = 'Scandium'
      ptable(21)%number = 21
      ptable(21)%amass = 44.955912_dp
      ptable(21)%covalent_radius = 1.4400_dp
      ptable(21)%vdw_radius = 2.00_dp
      ptable(21)%e_conv(0:3) = (/8, 12, 1, 0/)
      ptable(21)%eht_param(0:3) = (/-8.87_dp, -2.75_dp, -8.51_dp, z/)

      ! Titanium
      ptable(22)%symbol = 'Ti'
      ptable(22)%name = 'Titanium'
      ptable(22)%number = 22
      ptable(22)%amass = 47.867_dp
      ptable(22)%covalent_radius = 1.3200_dp
      ptable(22)%vdw_radius = 2.00_dp
      ptable(22)%e_conv(0:3) = (/8, 12, 2, 0/)
      ptable(22)%eht_param(0:3) = (/-8.97_dp, -5.44_dp, -10.81_dp, z/)

      ! Vanadium
      ptable(23)%symbol = 'V '
      ptable(23)%name = 'Vanadium'
      ptable(23)%number = 23
      ptable(23)%amass = 50.9415_dp
      ptable(23)%covalent_radius = 1.2200_dp
      ptable(23)%vdw_radius = 2.00_dp
      ptable(23)%e_conv(0:3) = (/8, 12, 3, 0/)
      ptable(23)%eht_param(0:3) = (/-8.81_dp, -5.52_dp, -11.00_dp, z/)

      ! Chromium
      ptable(24)%symbol = 'Cr'
      ptable(24)%name = 'Chromium'
      ptable(24)%number = 24
      ptable(24)%amass = 51.9961_dp
      ptable(24)%covalent_radius = 1.1800_dp
      ptable(24)%vdw_radius = 2.00_dp
      ptable(24)%e_conv(0:3) = (/7, 12, 5, 0/)
      ptable(24)%eht_param(0:3) = (/-8.66_dp, -5.24_dp, -11.22_dp, z/)

      ! Manganese
      ptable(25)%symbol = 'Mn'
      ptable(25)%name = 'Manganese'
      ptable(25)%number = 25
      ptable(25)%amass = 54.938045_dp
      ptable(25)%covalent_radius = 1.1700_dp
      ptable(25)%vdw_radius = 2.00_dp
      ptable(25)%e_conv(0:3) = (/8, 12, 5, 0/)
      ptable(25)%eht_param(0:3) = (/-9.75_dp, -5.89_dp, -11.67_dp, z/)

      ! Iron
      ptable(26)%symbol = 'Fe'
      ptable(26)%name = 'Iron'
      ptable(26)%number = 26
      ptable(26)%amass = 55.845_dp
      ptable(26)%covalent_radius = 1.1700_dp
      ptable(26)%vdw_radius = 2.00_dp
      ptable(26)%e_conv(0:3) = (/8, 12, 6, 0/)
      ptable(26)%eht_param(0:3) = (/-9.10_dp, -5.32_dp, -12.60_dp, z/)

      ! Cobalt
      ptable(27)%symbol = 'Co'
      ptable(27)%name = 'Cobalt'
      ptable(27)%number = 27
      ptable(27)%amass = 58.933195_dp
      ptable(27)%covalent_radius = 1.1600_dp
      ptable(27)%vdw_radius = 2.00_dp
      ptable(27)%e_conv(0:3) = (/8, 12, 7, 0/)
      ptable(27)%eht_param(0:3) = (/-9.21_dp, -5.29_dp, -13.18_dp, z/)

      ! Nickel
      ptable(28)%symbol = 'Ni'
      ptable(28)%name = 'Nickel'
      ptable(28)%number = 28
      ptable(28)%amass = 58.6934_dp
      ptable(28)%covalent_radius = 1.1500_dp
      ptable(28)%vdw_radius = 1.63_dp
      ptable(28)%e_conv(0:3) = (/8, 12, 8, 0/)
      ptable(28)%eht_param(0:3) = (/-9.17_dp, -5.15_dp, -13.49_dp, z/)

      ! Copper
      ptable(29)%symbol = 'Cu'
      ptable(29)%name = 'Copper'
      ptable(29)%number = 29
      ptable(29)%amass = 63.546_dp
      ptable(29)%covalent_radius = 1.1700_dp
      ptable(29)%vdw_radius = 1.40_dp
      ptable(29)%e_conv(0:3) = (/7, 12, 10, 0/)
      ptable(29)%eht_param(0:3) = (/-11.40_dp, -6.06_dp, -14.00_dp, z/)

      ! Zinc
      ptable(30)%symbol = 'Zn'
      ptable(30)%name = 'Zinc'
      ptable(30)%number = 30
      ptable(30)%amass = 65.38_dp
      ptable(30)%covalent_radius = 1.2500_dp
      ptable(30)%vdw_radius = 1.39_dp
      ptable(30)%e_conv(0:3) = (/8, 12, 10, 0/)
      ptable(30)%eht_param(0:3) = (/-12.41_dp, -6.53_dp, z, z/)

      ! Gallium
      ptable(31)%symbol = 'Ga'
      ptable(31)%name = 'Gallium'
      ptable(31)%number = 31
      ptable(31)%amass = 69.723_dp
      ptable(31)%covalent_radius = 1.2600_dp
      ptable(31)%vdw_radius = 1.87_dp
      ptable(31)%e_conv(0:3) = (/8, 13, 10, 0/)
      ptable(31)%eht_param(0:3) = (/-14.58_dp, -6.75_dp, z, z/)

      ! Germanium
      ptable(32)%symbol = 'Ge'
      ptable(32)%name = 'Germanium'
      ptable(32)%number = 32
      ptable(32)%amass = 72.64_dp
      ptable(32)%covalent_radius = 1.2200_dp
      ptable(32)%vdw_radius = 2.00_dp
      ptable(32)%e_conv(0:3) = (/8, 14, 10, 0/)
      ptable(32)%eht_param(0:3) = (/-16.00_dp, -9.00_dp, z, z/)

      ! Arsenic
      ptable(33)%symbol = 'As'
      ptable(33)%name = 'Arsenic'
      ptable(33)%number = 33
      ptable(33)%amass = 74.9216_dp
      ptable(33)%covalent_radius = 1.2000_dp
      ptable(33)%vdw_radius = 1.85_dp
      ptable(33)%e_conv(0:3) = (/8, 15, 10, 0/)
      ptable(33)%eht_param(0:3) = (/-16.22_dp, -12.16_dp, z, z/)

      ! Selenium
      ptable(34)%symbol = 'Se'
      ptable(34)%name = 'Selenium'
      ptable(34)%number = 34
      ptable(34)%amass = 78.96_dp
      ptable(34)%covalent_radius = 1.1600_dp
      ptable(34)%vdw_radius = 1.90_dp
      ptable(34)%e_conv(0:3) = (/8, 16, 10, 0/)
      ptable(34)%eht_param(0:3) = (/-20.50_dp, -14.40_dp, z, z/)

      ! Bromine
      ptable(35)%symbol = 'Br'
      ptable(35)%name = 'Bromine'
      ptable(35)%number = 35
      ptable(35)%amass = 79.904_dp
      ptable(35)%covalent_radius = 1.1400_dp
      ptable(35)%vdw_radius = 1.85_dp
      ptable(35)%e_conv(0:3) = (/8, 17, 10, 0/)
      ptable(35)%eht_param(0:3) = (/-22.07_dp, -13.10_dp, z, z/)

      ! Krypton
      ptable(36)%symbol = 'Kr'
      ptable(36)%name = 'Krypton'
      ptable(36)%number = 36
      ptable(36)%amass = 83.798_dp
      ptable(36)%covalent_radius = 1.1200_dp
      ptable(36)%vdw_radius = 2.02_dp
      ptable(36)%e_conv(0:3) = (/8, 18, 10, 0/)
      ptable(36)%eht_param(0:3) = (/z, z, z, z/)

      ! Rubidium
      ptable(37)%symbol = 'Rb'
      ptable(37)%name = 'Rubidium'
      ptable(37)%number = 37
      ptable(37)%amass = 85.4678_dp
      ptable(37)%covalent_radius = 2.1600_dp
      ptable(37)%vdw_radius = 2.00_dp
      ptable(37)%e_conv(0:3) = (/9, 18, 10, 0/)
      ptable(37)%eht_param(0:3) = (/-4.18_dp, -2.60_dp, z, z/)

      ! Strontium
      ptable(38)%symbol = 'Sr'
      ptable(38)%name = 'Strontium'
      ptable(38)%number = 38
      ptable(38)%amass = 87.62_dp
      ptable(38)%covalent_radius = 1.9100_dp
      ptable(38)%vdw_radius = 2.00_dp
      ptable(38)%e_conv(0:3) = (/10, 18, 10, 0/)
      ptable(38)%eht_param(0:3) = (/-6.62_dp, -3.92_dp, z, z/)

      ! Yttrium
      ptable(39)%symbol = 'Y '
      ptable(39)%name = 'Yttrium'
      ptable(39)%number = 39
      ptable(39)%amass = 88.90585_dp
      ptable(39)%covalent_radius = 1.6200_dp
      ptable(39)%vdw_radius = 2.00_dp
      ptable(39)%e_conv(0:3) = (/10, 18, 11, 0/)
      ptable(39)%eht_param(0:3) = (/z, z, z, z/)

      ! Zirconium
      ptable(40)%symbol = 'Zr'
      ptable(40)%name = 'Zirconium'
      ptable(40)%number = 40
      ptable(40)%amass = 91.224_dp
      ptable(40)%covalent_radius = 1.4500_dp
      ptable(40)%vdw_radius = 2.00_dp
      ptable(40)%e_conv(0:3) = (/10, 18, 12, 0/)
      ptable(40)%eht_param(0:3) = (/-8.00_dp, -5.40_dp, -10.20_dp, z/)

      ! Niobium
      ptable(41)%symbol = 'Nb'
      ptable(41)%name = 'Niobium'
      ptable(41)%number = 41
      ptable(41)%amass = 92.90638_dp
      ptable(41)%covalent_radius = 1.3400_dp
      ptable(41)%vdw_radius = 2.00_dp
      ptable(41)%e_conv(0:3) = (/9, 18, 14, 0/)
      ptable(41)%eht_param(0:3) = (/-10.10_dp, -6.86_dp, -12.10_dp, z/)

      ! Molybdenum
      ptable(42)%symbol = 'Mo'
      ptable(42)%name = 'Molybdenum'
      ptable(42)%number = 42
      ptable(42)%amass = 95.96_dp
      ptable(42)%covalent_radius = 1.3000_dp
      ptable(42)%vdw_radius = 2.00_dp
      ptable(42)%e_conv(0:3) = (/9, 18, 15, 0/)
      ptable(42)%eht_param(0:3) = (/-8.34_dp, -5.25_dp, -10.50_dp, z/)

      ! Technetium
      ptable(43)%symbol = 'Tc'
      ptable(43)%name = 'Technetium'
      ptable(43)%number = 43
      ptable(43)%amass = 97.9072_dp ! (98)
      ptable(43)%covalent_radius = 1.2700_dp
      ptable(43)%vdw_radius = 2.00_dp
      ptable(43)%e_conv(0:3) = (/10, 18, 15, 0/)
      ptable(43)%eht_param(0:3) = (/-10.07_dp, -5.40_dp, -12.82_dp, z/)

      ! Ruthenium
      ptable(44)%symbol = 'Ru'
      ptable(44)%name = 'Ruthenium'
      ptable(44)%number = 44
      ptable(44)%amass = 101.07_dp
      ptable(44)%covalent_radius = 1.2500_dp
      ptable(44)%vdw_radius = 2.00_dp
      ptable(44)%e_conv(0:3) = (/9, 18, 17, 0/)
      ptable(44)%eht_param(0:3) = (/-10.40_dp, -6.87_dp, -14.90_dp, z/)

      ! Rhodium
      ptable(45)%symbol = 'Rh'
      ptable(45)%name = 'Rhodium'
      ptable(45)%number = 45
      ptable(45)%amass = 102.9055_dp
      ptable(45)%covalent_radius = 1.2500_dp
      ptable(45)%vdw_radius = 2.00_dp
      ptable(45)%e_conv(0:3) = (/9, 18, 18, 0/)
      ptable(45)%eht_param(0:3) = (/-8.09_dp, -4.57_dp, -12.50_dp, z/)

      ! Palladium
      ptable(46)%symbol = 'Pd'
      ptable(46)%name = 'Palladium'
      ptable(46)%number = 46
      ptable(46)%amass = 106.42_dp
      ptable(46)%covalent_radius = 1.2800_dp
      ptable(46)%vdw_radius = 1.63_dp
      ptable(46)%e_conv(0:3) = (/8, 18, 20, 0/)
      ptable(46)%eht_param(0:3) = (/-7.32_dp, -3.75_dp, -12.02_dp, z/)

      ! Silver
      ptable(47)%symbol = 'Ag'
      ptable(47)%name = 'Silver'
      ptable(47)%number = 47
      ptable(47)%amass = 107.8682_dp
      ptable(47)%covalent_radius = 1.3400_dp
      ptable(47)%vdw_radius = 1.72_dp
      ptable(47)%e_conv(0:3) = (/9, 18, 20, 0/)
      ptable(47)%eht_param(0:3) = (/z, z, z, z/)

      ! Cadmium
      ptable(48)%symbol = 'Cd'
      ptable(48)%name = 'Cadmium'
      ptable(48)%number = 48
      ptable(48)%amass = 112.411_dp
      ptable(48)%covalent_radius = 1.4800_dp
      ptable(48)%vdw_radius = 1.58_dp
      ptable(48)%e_conv(0:3) = (/10, 18, 20, 0/)
      ptable(48)%eht_param(0:3) = (/z, z, z, z/)

      ! Indium
      ptable(49)%symbol = 'In'
      ptable(49)%name = 'Indium'
      ptable(49)%number = 49
      ptable(49)%amass = 114.818_dp
      ptable(49)%covalent_radius = 1.4400_dp
      ptable(49)%vdw_radius = 1.93_dp
      ptable(49)%e_conv(0:3) = (/10, 19, 20, 0/)
      ptable(49)%eht_param(0:3) = (/-12.60_dp, -6.19_dp, z, z/)

      ! Tin
      ptable(50)%symbol = 'Sn'
      ptable(50)%name = 'Tin'
      ptable(50)%number = 50
      ptable(50)%amass = 118.71_dp
      ptable(50)%covalent_radius = 1.4100_dp
      ptable(50)%vdw_radius = 2.17_dp
      ptable(50)%e_conv(0:3) = (/10, 20, 20, 0/)
      ptable(50)%eht_param(0:3) = (/-16.16_dp, -8.32_dp, z, z/)

      ! Antimony
      ptable(51)%symbol = 'Sb'
      ptable(51)%name = 'Antimony'
      ptable(51)%number = 51
      ptable(51)%amass = 121.76_dp
      ptable(51)%covalent_radius = 1.4000_dp
      ptable(51)%vdw_radius = 2.00_dp
      ptable(51)%e_conv(0:3) = (/10, 21, 20, 0/)
      ptable(51)%eht_param(0:3) = (/-18.80_dp, -11.70_dp, z, z/)

      ! Tellurium
      ptable(52)%symbol = 'Te'
      ptable(52)%name = 'Tellurium'
      ptable(52)%number = 52
      ptable(52)%amass = 127.6_dp
      ptable(52)%covalent_radius = 1.3600_dp
      ptable(52)%vdw_radius = 2.06_dp
      ptable(52)%e_conv(0:3) = (/10, 22, 20, 0/)
      ptable(52)%eht_param(0:3) = (/-20.80_dp, -13.20_dp, z, z/)

      ! Iodine
      ptable(53)%symbol = 'I '
      ptable(53)%name = 'Iodine'
      ptable(53)%number = 53
      ptable(53)%amass = 126.90447_dp
      ptable(53)%covalent_radius = 1.3300_dp
      ptable(53)%vdw_radius = 1.98_dp
      ptable(53)%e_conv(0:3) = (/10, 23, 20, 0/)
      ptable(53)%eht_param(0:3) = (/-18.00_dp, -12.70_dp, z, z/)

      ! Xenon
      ptable(54)%symbol = 'Xe'
      ptable(54)%name = 'Xenon'
      ptable(54)%number = 54
      ptable(54)%amass = 131.293_dp
      ptable(54)%covalent_radius = 1.3100_dp
      ptable(54)%vdw_radius = 2.16_dp
      ptable(54)%e_conv(0:3) = (/10, 24, 20, 0/)
      ptable(54)%eht_param(0:3) = (/z, z, z, z/)

      ! Cesium
      ptable(55)%symbol = 'Cs'
      ptable(55)%name = 'Cesium'
      ptable(55)%number = 55
      ptable(55)%amass = 132.9054519_dp
      ptable(55)%covalent_radius = 2.3500_dp
      ptable(55)%vdw_radius = 2.00_dp
      ptable(55)%e_conv(0:3) = (/11, 24, 20, 0/)
      ptable(55)%eht_param(0:3) = (/-3.88_dp, -2.49_dp, z, z/)

      ! Barium
      ptable(56)%symbol = 'Ba'
      ptable(56)%name = 'Barium'
      ptable(56)%number = 56
      ptable(56)%amass = 137.327_dp
      ptable(56)%covalent_radius = 1.9800_dp
      ptable(56)%vdw_radius = 2.00_dp
      ptable(56)%e_conv(0:3) = (/12, 24, 20, 0/)
      ptable(56)%eht_param(0:3) = (/z, z, z, z/)

      ! Lanthanum
      ptable(57)%symbol = 'La'
      ptable(57)%name = 'Lanthanum'
      ptable(57)%number = 57
      ptable(57)%amass = 138.90547_dp
      ptable(57)%covalent_radius = 1.6900_dp
      ptable(57)%vdw_radius = 2.00_dp
      ptable(57)%e_conv(0:3) = (/12, 24, 21, 0/)
      ptable(57)%eht_param(0:3) = (/-7.67_dp, -5.01_dp, -8.21_dp, z/)

      ! Cerium
      ptable(58)%symbol = 'Ce'
      ptable(58)%name = 'Cerium'
      ptable(58)%number = 58
      ptable(58)%amass = 140.116_dp
      ptable(58)%covalent_radius = 1.6500_dp
      ptable(58)%vdw_radius = 2.00_dp
      ptable(58)%e_conv(0:3) = (/12, 24, 21, 1/)
      ptable(58)%eht_param(0:3) = (/z, z, z, z/)

      ! Praseodymium
      ptable(59)%symbol = 'Pr'
      ptable(59)%name = 'Praseodymium'
      ptable(59)%number = 59
      ptable(59)%amass = 140.90765_dp
      ptable(59)%covalent_radius = 1.6500_dp
      ptable(59)%vdw_radius = 2.00_dp
      ptable(59)%e_conv(0:3) = (/12, 24, 20, 3/)
      ptable(59)%eht_param(0:3) = (/z, z, z, z/)

      ! Neodymium
      ptable(60)%symbol = 'Nd'
      ptable(60)%name = 'Neodymium'
      ptable(60)%number = 60
      ptable(60)%amass = 144.242_dp
      ptable(60)%covalent_radius = 1.6400_dp
      ptable(60)%vdw_radius = 2.00_dp
      ptable(60)%e_conv(0:3) = (/12, 24, 20, 4/)
      ptable(60)%eht_param(0:3) = (/z, z, z, z/)

      ! Promethium
      ptable(61)%symbol = 'Pm'
      ptable(61)%name = 'Promethium'
      ptable(61)%number = 61
      ptable(61)%amass = 144.9127_dp ! (145)
      ptable(61)%covalent_radius = 1.6300_dp
      ptable(61)%vdw_radius = 2.00_dp
      ptable(61)%e_conv(0:3) = (/12, 24, 20, 5/)
      ptable(61)%eht_param(0:3) = (/z, z, z, z/)

      ! Samarium
      ptable(62)%symbol = 'Sm'
      ptable(62)%name = 'Samarium'
      ptable(62)%number = 62
      ptable(62)%amass = 150.36_dp
      ptable(62)%covalent_radius = 1.6200_dp
      ptable(62)%vdw_radius = 2.00_dp
      ptable(62)%e_conv(0:3) = (/12, 24, 20, 6/)
      ptable(62)%eht_param(0:3) = (/-4.86_dp, -4.86_dp, -6.06_dp, -11.28_dp/)

      ! Europium
      ptable(63)%symbol = 'Eu'
      ptable(63)%name = 'Europium'
      ptable(63)%number = 63
      ptable(63)%amass = 151.964_dp
      ptable(63)%covalent_radius = 1.8500_dp
      ptable(63)%vdw_radius = 2.00_dp
      ptable(63)%e_conv(0:3) = (/12, 24, 20, 7/)
      ptable(63)%eht_param(0:3) = (/z, z, z, z/)

      ! Gadolinium
      ptable(64)%symbol = 'Gd'
      ptable(64)%name = 'Gadolinium'
      ptable(64)%number = 64
      ptable(64)%amass = 157.25_dp
      ptable(64)%covalent_radius = 1.6100_dp
      ptable(64)%vdw_radius = 2.00_dp
      ptable(64)%e_conv(0:3) = (/12, 24, 21, 7/)
      ptable(64)%eht_param(0:3) = (/z, z, z, z/)

      ! Terbium
      ptable(65)%symbol = 'Tb'
      ptable(65)%name = 'Terbium'
      ptable(65)%number = 65
      ptable(65)%amass = 158.92535_dp
      ptable(65)%covalent_radius = 1.5900_dp
      ptable(65)%vdw_radius = 2.00_dp
      ptable(65)%e_conv(0:3) = (/12, 24, 20, 9/)
      ptable(65)%eht_param(0:3) = (/z, z, z, z/)

      ! Dysprosium
      ptable(66)%symbol = 'Dy'
      ptable(66)%name = 'Dysprosium'
      ptable(66)%number = 66
      ptable(66)%amass = 162.5_dp
      ptable(66)%covalent_radius = 1.5900_dp
      ptable(66)%vdw_radius = 2.00_dp
      ptable(66)%e_conv(0:3) = (/12, 24, 20, 10/)
      ptable(66)%eht_param(0:3) = (/z, z, z, z/)

      ! Holmium
      ptable(67)%symbol = 'Ho'
      ptable(67)%name = 'Holmium'
      ptable(67)%number = 67
      ptable(67)%amass = 164.93032_dp
      ptable(67)%covalent_radius = 1.5800_dp
      ptable(67)%vdw_radius = 2.00_dp
      ptable(67)%e_conv(0:3) = (/12, 24, 20, 11/)
      ptable(67)%eht_param(0:3) = (/z, z, z, z/)

      ! Erbium
      ptable(68)%symbol = 'Er'
      ptable(68)%name = 'Erbium'
      ptable(68)%number = 68
      ptable(68)%amass = 167.259_dp
      ptable(68)%covalent_radius = 1.5700_dp
      ptable(68)%vdw_radius = 2.00_dp
      ptable(68)%e_conv(0:3) = (/12, 24, 20, 12/)
      ptable(68)%eht_param(0:3) = (/z, z, z, z/)

      ! Thulium
      ptable(69)%symbol = 'Tm'
      ptable(69)%name = 'Thulium'
      ptable(69)%number = 69
      ptable(69)%amass = 168.93421_dp
      ptable(69)%covalent_radius = 1.5600_dp
      ptable(69)%vdw_radius = 2.00_dp
      ptable(69)%e_conv(0:3) = (/12, 24, 20, 13/)
      ptable(69)%eht_param(0:3) = (/z, z, z, z/)

      ! Ytterbium
      ptable(70)%symbol = 'Yb'
      ptable(70)%name = 'Ytterbium'
      ptable(70)%number = 70
      ptable(70)%amass = 173.054_dp
      ptable(70)%covalent_radius = 1.5600_dp
      ptable(70)%vdw_radius = 2.00_dp
      ptable(70)%e_conv(0:3) = (/12, 24, 20, 14/)
      ptable(70)%eht_param(0:3) = (/-5.35_dp, -5.35_dp, -5.21_dp, &
                                    -13.86_dp/)

      ! Lutetium
      ptable(71)%symbol = 'Lu'
      ptable(71)%name = 'Lutetium'
      ptable(71)%number = 71
      ptable(71)%amass = 174.9668_dp
      ptable(71)%covalent_radius = 1.5600_dp
      ptable(71)%vdw_radius = 2.00_dp
      ptable(71)%e_conv(0:3) = (/12, 24, 21, 14/)
      ptable(71)%eht_param(0:3) = (/-6.05_dp, -6.05_dp, -5.12_dp, -22.40_dp/)

      ! Hafnium
      ptable(72)%symbol = 'Hf'
      ptable(72)%name = 'Hafnium'
      ptable(72)%number = 72
      ptable(72)%amass = 178.49_dp
      ptable(72)%covalent_radius = 1.4400_dp
      ptable(72)%vdw_radius = 2.00_dp
      ptable(72)%e_conv(0:3) = (/12, 24, 22, 14/)
      ptable(72)%eht_param(0:3) = (/z, z, z, z/)

      ! Tantalum
      ptable(73)%symbol = 'Ta'
      ptable(73)%name = 'Tantalum'
      ptable(73)%number = 73
      ptable(73)%amass = 180.94788_dp
      ptable(73)%covalent_radius = 1.3400_dp
      ptable(73)%vdw_radius = 2.00_dp
      ptable(73)%e_conv(0:3) = (/12, 24, 23, 14/)
      ptable(73)%eht_param(0:3) = (/-10.10_dp, -6.86_dp, -12.10_dp, z/)

      ! Tungsten
      ptable(74)%symbol = 'W '
      ptable(74)%name = 'Tungsten'
      ptable(74)%number = 74
      ptable(74)%amass = 183.84_dp
      ptable(74)%covalent_radius = 1.3000_dp
      ptable(74)%vdw_radius = 2.00_dp
      ptable(74)%e_conv(0:3) = (/12, 24, 24, 14/)
      ptable(74)%eht_param(0:3) = (/-8.26_dp, -5.17_dp, -10.37_dp, z/)

      ! Rhenium
      ptable(75)%symbol = 'Re'
      ptable(75)%name = 'Rhenium'
      ptable(75)%number = 75
      ptable(75)%amass = 186.207_dp
      ptable(75)%covalent_radius = 1.2800_dp
      ptable(75)%vdw_radius = 2.00_dp
      ptable(75)%e_conv(0:3) = (/12, 24, 25, 14/)
      ptable(75)%eht_param(0:3) = (/-9.36_dp, -5.96_dp, -12.66_dp, z/)

      ! Osmium
      ptable(76)%symbol = 'Os'
      ptable(76)%name = 'Osmium'
      ptable(76)%number = 76
      ptable(76)%amass = 190.23_dp
      ptable(76)%covalent_radius = 1.2600_dp
      ptable(76)%vdw_radius = 2.00_dp
      ptable(76)%e_conv(0:3) = (/12, 24, 26, 14/)
      ptable(76)%eht_param(0:3) = (/-8.17_dp, -4.81_dp, -11.84_dp, z/)

      ! Iridium
      ptable(77)%symbol = 'Ir'
      ptable(77)%name = 'Iridium'
      ptable(77)%number = 77
      ptable(77)%amass = 192.217_dp
      ptable(77)%covalent_radius = 1.2700_dp
      ptable(77)%vdw_radius = 2.00_dp
      ptable(77)%e_conv(0:3) = (/12, 24, 27, 14/)
      ptable(77)%eht_param(0:3) = (/-11.36_dp, -4.50_dp, -12.17_dp, z/)

      ! Platinum
      ptable(78)%symbol = 'Pt'
      ptable(78)%name = 'Platinum'
      ptable(78)%number = 78
      ptable(78)%amass = 195.084_dp
      ptable(78)%covalent_radius = 1.3000_dp
      ptable(78)%vdw_radius = 1.72_dp
      ptable(78)%e_conv(0:3) = (/11, 24, 29, 14/)
      ptable(78)%eht_param(0:3) = (/-9.077_dp, -5.475_dp, -12.59_dp, z/)

      ! Gold
      ptable(79)%symbol = 'Au'
      ptable(79)%name = 'Gold'
      ptable(79)%number = 79
      ptable(79)%amass = 196.966569_dp
      ptable(79)%covalent_radius = 1.3400_dp
      ptable(79)%vdw_radius = 1.66_dp
      ptable(79)%e_conv(0:3) = (/11, 24, 30, 14/)
      ptable(79)%eht_param(0:3) = (/-10.92_dp, -5.55_dp, -15.076_dp, z/)

      ! Mercury
      ptable(80)%symbol = 'Hg'
      ptable(80)%name = 'Mercury'
      ptable(80)%number = 80
      ptable(80)%amass = 200.59_dp
      ptable(80)%covalent_radius = 1.4900_dp
      ptable(80)%vdw_radius = 1.55_dp
      ptable(80)%e_conv(0:3) = (/12, 24, 30, 14/)
      ptable(80)%eht_param(0:3) = (/-13.68_dp, -8.47_dp, -17.50_dp, z/)

      ! Thallium
      ptable(81)%symbol = 'Tl'
      ptable(81)%name = 'Thallium'
      ptable(81)%number = 81
      ptable(81)%amass = 204.3833_dp
      ptable(81)%covalent_radius = 1.4800_dp
      ptable(81)%vdw_radius = 1.96_dp
      ptable(81)%e_conv(0:3) = (/12, 25, 30, 14/)
      ptable(81)%eht_param(0:3) = (/-11.60_dp, -5.80_dp, z, z/)

      ! Lead
      ptable(82)%symbol = 'Pb'
      ptable(82)%name = 'Lead'
      ptable(82)%number = 82
      ptable(82)%amass = 207.2_dp
      ptable(82)%covalent_radius = 1.4700_dp
      ptable(82)%vdw_radius = 2.02_dp
      ptable(82)%e_conv(0:3) = (/12, 26, 30, 14/)
      ptable(82)%eht_param(0:3) = (/-15.70_dp, -8.00_dp, z, z/)

      ! Bismuth
      ptable(83)%symbol = 'Bi'
      ptable(83)%name = 'Bismuth'
      ptable(83)%number = 83
      ptable(83)%amass = 208.9804_dp
      ptable(83)%covalent_radius = 1.4600_dp
      ptable(83)%vdw_radius = 2.00_dp
      ptable(83)%e_conv(0:3) = (/12, 27, 30, 14/)
      ptable(83)%eht_param(0:3) = (/-15.19_dp, -7.79_dp, z, z/)

      ! Polonium
      ptable(84)%symbol = 'Po'
      ptable(84)%name = 'Polonium'
      ptable(84)%number = 84
      ptable(84)%amass = 208.9824_dp ! (209)
      ptable(84)%covalent_radius = 1.4600_dp
      ptable(84)%vdw_radius = 2.00_dp
      ptable(84)%e_conv(0:3) = (/12, 28, 30, 14/)
      ptable(84)%eht_param(0:3) = (/z, z, z, z/)

      ! Astatine
      ptable(85)%symbol = 'At'
      ptable(85)%name = 'Astatine'
      ptable(85)%number = 85
      ptable(85)%amass = 209.9871_dp ! (210)
      ptable(85)%covalent_radius = 1.4500_dp
      ptable(85)%vdw_radius = 2.00_dp
      ptable(85)%e_conv(0:3) = (/12, 29, 30, 14/)
      ptable(85)%eht_param(0:3) = (/z, z, z, z/)

      ! Radon
      ptable(86)%symbol = 'Rn'
      ptable(86)%name = 'Radon'
      ptable(86)%number = 86
      ptable(86)%amass = 222.0176_dp ! (222)
      ptable(86)%covalent_radius = 1.50_dp
      ptable(86)%vdw_radius = 2.00_dp
      ptable(86)%e_conv(0:3) = (/12, 30, 30, 14/)
      ptable(86)%eht_param(0:3) = (/z, z, z, z/)

      ! Francium
      ptable(87)%symbol = 'Fr'
      ptable(87)%name = 'Francium'
      ptable(87)%number = 87
      ptable(87)%amass = 223.0197_dp ! (223)
      ptable(87)%covalent_radius = 1.50_dp
      ptable(87)%vdw_radius = 2.00_dp
      ptable(87)%e_conv(0:3) = (/13, 30, 30, 14/)
      ptable(87)%eht_param(0:3) = (/z, z, z, z/)

      ! Radium
      ptable(88)%symbol = 'Ra'
      ptable(88)%name = 'Radium'
      ptable(88)%number = 88
      ptable(88)%amass = 226.0254_dp ! (226)
      ptable(88)%covalent_radius = 1.90_dp
      ptable(88)%vdw_radius = 2.00_dp
      ptable(88)%e_conv(0:3) = (/14, 30, 30, 14/)
      ptable(88)%eht_param(0:3) = (/z, z, z, z/)

      ! Actinium
      ptable(89)%symbol = 'Ac'
      ptable(89)%name = 'Actinium'
      ptable(89)%number = 89
      ptable(89)%amass = 227.0278_dp ! (227)
      ptable(89)%covalent_radius = 1.88_dp
      ptable(89)%vdw_radius = 2.00_dp
      ptable(89)%e_conv(0:3) = (/14, 30, 31, 14/)
      ptable(89)%eht_param(0:3) = (/z, z, z, z/)

      ! Thorium
      ptable(90)%symbol = 'Th'
      ptable(90)%name = 'Thorium'
      ptable(90)%number = 90
      ptable(90)%amass = 232.03806_dp
      ptable(90)%covalent_radius = 1.6500_dp
      ptable(90)%vdw_radius = 2.00_dp
      ptable(90)%e_conv(0:3) = (/14, 30, 32, 14/)
      ptable(90)%eht_param(0:3) = (/-5.39_dp, -5.39_dp, -10.11_dp, -9.64_dp/)

      ! Proctactinium
      ptable(91)%symbol = 'Pa'
      ptable(91)%name = 'Proctactinium'
      ptable(91)%number = 91
      ptable(91)%amass = 231.03588_dp
      ptable(91)%covalent_radius = 1.61_dp
      ptable(91)%vdw_radius = 2.00_dp
      ptable(91)%e_conv(0:3) = (/14, 30, 31, 16/)
      ptable(91)%eht_param(0:3) = (/z, z, z, z/)

      ! Uranium
      ptable(92)%symbol = 'U '
      ptable(92)%name = 'Uranium'
      ptable(92)%number = 92
      ptable(92)%amass = 238.02891_dp
      ptable(92)%covalent_radius = 1.4200_dp
      ptable(92)%vdw_radius = 1.86_dp
      ptable(92)%e_conv(0:3) = (/14, 30, 31, 17/)
      ptable(92)%eht_param(0:3) = (/-5.50_dp, -5.50_dp, -9.19_dp, -10.62_dp/)

      ! Neptunium
      ptable(93)%symbol = 'Np'
      ptable(93)%name = 'Neptunium'
      ptable(93)%number = 93
      ptable(93)%amass = 237.0482_dp ! (237)
      ptable(93)%covalent_radius = 1.55_dp
      ptable(93)%vdw_radius = 2.00_dp
      ptable(93)%e_conv(0:3) = (/14, 30, 31, 18/)
      ptable(93)%eht_param(0:3) = (/z, z, z, z/)

      ! Plutonium
      ptable(94)%symbol = 'Pu'
      ptable(94)%name = 'Plutonium'
      ptable(94)%number = 94
      ptable(94)%amass = 244.0642_dp ! (244)
      ptable(94)%covalent_radius = 1.53_dp
      ptable(94)%vdw_radius = 2.00_dp
      ptable(94)%e_conv(0:3) = (/14, 30, 30, 20/)
      ptable(94)%eht_param(0:3) = (/z, z, z, z/)

      ! Americum
      ptable(95)%symbol = 'Am'
      ptable(95)%name = 'Americum'
      ptable(95)%number = 95
      ptable(95)%amass = 243.0614_dp ! (243)
      ptable(95)%covalent_radius = 1.51_dp
      ptable(95)%vdw_radius = 2.00_dp
      ptable(95)%e_conv(0:3) = (/14, 30, 30, 21/)
      ptable(95)%eht_param(0:3) = (/z, z, z, z/)

      ! Curium
      ptable(96)%symbol = 'Cm'
      ptable(96)%name = 'Curium'
      ptable(96)%number = 96
      ptable(96)%amass = 247.0704_dp ! (247)
      ptable(96)%covalent_radius = 0.99_dp
      ptable(96)%vdw_radius = 2.00_dp
      ptable(96)%e_conv(0:3) = (/14, 30, 31, 21/)
      ptable(96)%eht_param(0:3) = (/z, z, z, z/)

      ! Berkelium
      ptable(97)%symbol = 'Bk'
      ptable(97)%name = 'Berkelium'
      ptable(97)%number = 97
      ptable(97)%amass = 247.0703_dp ! (247)
      ptable(97)%covalent_radius = 1.54_dp
      ptable(97)%vdw_radius = 2.00_dp
      ptable(97)%e_conv(0:3) = (/14, 30, 30, 23/)
      ptable(97)%eht_param(0:3) = (/z, z, z, z/)

      ! Californium
      ptable(98)%symbol = 'Cf'
      ptable(98)%name = 'Californium'
      ptable(98)%number = 98
      ptable(98)%amass = 251.0796_dp ! (251)
      ptable(98)%covalent_radius = 1.83_dp
      ptable(98)%vdw_radius = 2.00_dp
      ptable(98)%e_conv(0:3) = (/14, 30, 30, 24/)
      ptable(98)%eht_param(0:3) = (/z, z, z, z/)

      ! Einsteinium
      ptable(99)%symbol = 'Es'
      ptable(99)%name = 'Einsteinium'
      ptable(99)%number = 99
      ptable(99)%amass = 252.083_dp ! (252)
      ptable(99)%covalent_radius = 1.50_dp
      ptable(99)%vdw_radius = 2.00_dp
      ptable(99)%e_conv(0:3) = (/14, 30, 30, 25/)
      ptable(99)%eht_param(0:3) = (/z, z, z, z/)

      ! Fermium
      ptable(100)%symbol = 'Fm'
      ptable(100)%name = 'Fermium'
      ptable(100)%number = 100
      ptable(100)%amass = 257.0951_dp ! (257)
      ptable(100)%covalent_radius = 1.50_dp
      ptable(100)%vdw_radius = 2.00_dp
      ptable(100)%e_conv(0:3) = (/14, 30, 30, 26/)
      ptable(100)%eht_param(0:3) = (/z, z, z, z/)

      ! Mendelevium
      ptable(101)%symbol = 'Md'
      ptable(101)%name = 'Mendelevium'
      ptable(101)%number = 101
      ptable(101)%amass = 258.0984_dp ! (258)
      ptable(101)%covalent_radius = 1.50_dp
      ptable(101)%vdw_radius = 2.00_dp
      ptable(101)%e_conv(0:3) = (/14, 30, 30, 27/)
      ptable(101)%eht_param(0:3) = (/z, z, z, z/)

      ! Nobelium
      ptable(102)%symbol = 'No'
      ptable(102)%name = 'Nobelium'
      ptable(102)%number = 102
      ptable(102)%amass = 259.101_dp ! (259)
      ptable(102)%covalent_radius = 1.50_dp
      ptable(102)%vdw_radius = 2.00_dp
      ptable(102)%e_conv(0:3) = (/14, 30, 30, 28/)
      ptable(102)%eht_param(0:3) = (/z, z, z, z/)

      ! Lawrencium
      ptable(103)%symbol = 'Lr'
      ptable(103)%name = 'Lawrencium'
      ptable(103)%number = 103
      ptable(103)%amass = 262.1096_dp ! (262)
      ptable(103)%covalent_radius = 1.50_dp
      ptable(103)%vdw_radius = 2.00_dp
      ptable(103)%e_conv(0:3) = (/14, 31, 30, 28/)
      ptable(103)%eht_param(0:3) = (/z, z, z, z/)

      ! Rutherfordium
      ptable(104)%symbol = 'Rf'
      ptable(104)%name = 'Rutherfordium'
      ptable(104)%number = 104
      ptable(104)%amass = 267.122_dp ! (267)
      ptable(104)%covalent_radius = 1.50_dp
      ptable(104)%vdw_radius = 2.00_dp
      ptable(104)%e_conv(0:3) = (/14, 30, 32, 28/)
      ptable(104)%eht_param(0:3) = (/z, z, z, z/)

      ! Dubnium
      ptable(105)%symbol = 'Db'
      ptable(105)%name = 'Dubnium'
      ptable(105)%number = 105
      ptable(105)%amass = 268.125_dp ! (268)
      ptable(105)%covalent_radius = 1.50_dp
      ptable(105)%vdw_radius = 2.00_dp
      ptable(105)%e_conv(0:3) = (/14, 30, 33, 28/)
      ptable(105)%eht_param(0:3) = (/z, z, z, z/)

      ! Seaborgium
      ptable(106)%symbol = 'Sg'
      ptable(106)%name = 'Seaborgium'
      ptable(106)%number = 106
      ptable(106)%amass = 271.133_dp ! (271)
      ptable(106)%covalent_radius = 1.50_dp
      ptable(106)%vdw_radius = 2.00_dp
      ptable(106)%e_conv(0:3) = (/14, 30, 34, 28/)
      ptable(106)%eht_param(0:3) = (/z, z, z, z/)

      ! Bohrium
      ptable(107)%symbol = 'Bh'
      ptable(107)%name = 'Bohrium'
      ptable(107)%number = 107
      ptable(107)%amass = 272.0_dp ! (272)
      ptable(107)%covalent_radius = 1.50_dp
      ptable(107)%vdw_radius = 2.00_dp
      ptable(107)%e_conv(0:3) = (/14, 30, 35, 28/)
      ptable(107)%eht_param(0:3) = (/z, z, z, z/)

      ! Hassium
      ptable(108)%symbol = 'Hs'
      ptable(108)%name = 'Hassium'
      ptable(108)%number = 108
      ptable(108)%amass = 277.15_dp ! (277)
      ptable(108)%covalent_radius = 1.50_dp
      ptable(108)%vdw_radius = 2.00_dp
      ptable(108)%e_conv(0:3) = (/14, 30, 36, 28/)
      ptable(108)%eht_param(0:3) = (/z, z, z, z/)

      ! Meitnerium
      ptable(109)%symbol = 'Mt'
      ptable(109)%name = 'Meitnerium'
      ptable(109)%number = 109
      ptable(109)%amass = 276.151_dp ! (276)
      ptable(109)%covalent_radius = 1.50_dp
      ptable(109)%vdw_radius = 2.00_dp
      ptable(109)%e_conv(0:3) = (/14, 30, 37, 28/)
      ptable(109)%eht_param(0:3) = (/z, z, z, z/)

      ! Darmstadtium
      ptable(110)%symbol = 'Ds'
      ptable(110)%name = 'Darmstadtium'
      ptable(110)%number = 110
      ptable(110)%amass = 281.162_dp ! (281)
      ptable(110)%covalent_radius = 1.50_dp
      ptable(110)%vdw_radius = 2.00_dp
      ptable(110)%e_conv(0:3) = (/13, 30, 39, 28/)
      ptable(110)%eht_param(0:3) = (/z, z, z, z/)

      ! Roentgenium
      ptable(111)%symbol = 'Rg'
      ptable(111)%name = 'Roentgenium'
      ptable(111)%number = 111
      ptable(111)%amass = 280.164_dp ! (280)
      ptable(111)%covalent_radius = 1.50_dp
      ptable(111)%vdw_radius = 2.00_dp
      ptable(111)%e_conv(0:3) = (/13, 30, 40, 28/)
      ptable(111)%eht_param(0:3) = (/z, z, z, z/)

      ! Copernicium
      ptable(112)%symbol = 'Cn'
      ptable(112)%name = 'Copernicium'
      ptable(112)%number = 112
      ptable(112)%amass = 285.174_dp ! (285)
      ptable(112)%covalent_radius = 1.50_dp
      ptable(112)%vdw_radius = 2.00_dp
      ptable(112)%e_conv(0:3) = (/14, 30, 40, 28/)
      ptable(112)%eht_param(0:3) = (/z, z, z, z/)

      ! Ununtrium
      ptable(113)%symbol = 'Uut'
      ptable(113)%name = 'Ununtrium'
      ptable(113)%number = 113
      ptable(113)%amass = 284.178_dp ! (284)
      ptable(113)%covalent_radius = 1.50_dp
      ptable(113)%vdw_radius = 2.00_dp
      ptable(113)%e_conv(0:3) = (/14, 31, 40, 28/)
      ptable(113)%eht_param(0:3) = (/z, z, z, z/)

      ! Ununquadium
      ptable(114)%symbol = 'Uuq'
      ptable(114)%name = 'Ununquadium'
      ptable(114)%number = 114
      ptable(114)%amass = 289.189_dp ! (289)
      ptable(114)%covalent_radius = 1.50_dp
      ptable(114)%vdw_radius = 2.00_dp
      ptable(114)%e_conv(0:3) = (/14, 32, 40, 28/)
      ptable(114)%eht_param(0:3) = (/z, z, z, z/)

      ! Ununpentium
      ptable(115)%symbol = 'Uup'
      ptable(115)%name = 'Ununpentium'
      ptable(115)%number = 115
      ptable(115)%amass = 288.192_dp ! (288)
      ptable(115)%covalent_radius = 1.50_dp
      ptable(115)%vdw_radius = 2.00_dp
      ptable(115)%e_conv(0:3) = (/14, 33, 40, 28/)
      ptable(115)%eht_param(0:3) = (/z, z, z, z/)

      ! Ununhexium
      ptable(116)%symbol = 'Uuh'
      ptable(116)%name = 'Ununhexium'
      ptable(116)%number = 116
      ptable(116)%amass = 293.0_dp ! (293)
      ptable(116)%covalent_radius = 1.50_dp
      ptable(116)%vdw_radius = 2.00_dp
      ptable(116)%e_conv(0:3) = (/14, 34, 40, 28/)
      ptable(116)%eht_param(0:3) = (/z, z, z, z/)

      ! Ununseptium
      ptable(117)%symbol = 'Uus'
      ptable(117)%name = 'Ununseptium'
      ptable(117)%number = 117
      ptable(117)%amass = 294.0_dp ! (294)
      ptable(117)%covalent_radius = 1.50_dp
      ptable(117)%vdw_radius = 2.00_dp
      ptable(117)%e_conv(0:3) = (/14, 35, 40, 28/)
      ptable(117)%eht_param(0:3) = (/z, z, z, z/)

      ! Ununoctium
      ptable(118)%symbol = 'Uuo'
      ptable(118)%name = 'Ununoctium'
      ptable(118)%number = 118
      ptable(118)%amass = 294.0_dp ! (294)
      ptable(118)%covalent_radius = 1.50_dp
      ptable(118)%vdw_radius = 2.00_dp
      ptable(118)%e_conv(0:3) = (/14, 36, 40, 28/)
      ptable(118)%eht_param(0:3) = (/z, z, z, z/)

      ! Initialize heat of formation
      CALL init_eheat

      ! Initialize gyromagnetic ratio
      CALL init_gratio

   END SUBROUTINE init_periodic_table

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE init_eheat()
      ! All values in kcal/mol
      ! Dummy
      ptable(0)%heat_of_formation = z
      ! Hydrogen
      ptable(1)%heat_of_formation = 52.102_dp
      ! Helium
      ptable(2)%heat_of_formation = z
      ! Lithium
      ptable(3)%heat_of_formation = 38.410_dp
      ! Beryllium
      ptable(4)%heat_of_formation = 76.960_dp
      ! Boron
      ptable(5)%heat_of_formation = 135.700_dp
      ! Carbon
      ptable(6)%heat_of_formation = 170.890_dp
      ! Nitrogen
      ptable(7)%heat_of_formation = 113.000_dp
      ! Oxygen
      ptable(8)%heat_of_formation = 59.559_dp
      ! Fluorine
      ptable(9)%heat_of_formation = 18.890_dp
      ! Neon
      ptable(10)%heat_of_formation = z
      ! Sodium
      ptable(11)%heat_of_formation = 25.850_dp
      ! Magnesium
      ptable(12)%heat_of_formation = 35.000_dp
      ! Aluminium
      ptable(13)%heat_of_formation = 79.490_dp
      ! Silicon
      ptable(14)%heat_of_formation = 108.390_dp
      ! Phosphorus
      ptable(15)%heat_of_formation = 75.570_dp
      ! Sulfur
      ptable(16)%heat_of_formation = 66.400_dp
      ! Chlorine
      ptable(17)%heat_of_formation = 28.990_dp
      ! Argon
      ptable(18)%heat_of_formation = z
      ! Potassium
      ptable(19)%heat_of_formation = 21.420_dp
      ! Calcium
      ptable(20)%heat_of_formation = 42.600_dp
      ! Scandium
      ptable(21)%heat_of_formation = 90.300_dp
      ! Titanium
      ptable(22)%heat_of_formation = 112.300_dp
      ! Vanadium
      ptable(23)%heat_of_formation = 122.900_dp
      ! Chromium
      ptable(24)%heat_of_formation = 95.000_dp
      ! Manganese
      ptable(25)%heat_of_formation = 67.700_dp
      ! Iron
      ptable(26)%heat_of_formation = 99.300_dp
      ! Cobalt
      ptable(27)%heat_of_formation = 102.400_dp
      ! Nickel
      ptable(28)%heat_of_formation = 102.800_dp
      ! Copper
      ptable(29)%heat_of_formation = 80.700_dp
      ! Zinc
      ptable(30)%heat_of_formation = 31.170_dp
      ! Gallium
      ptable(31)%heat_of_formation = 65.400_dp
      ! Germanium
      ptable(32)%heat_of_formation = 89.500_dp
      ! Arsenic
      ptable(33)%heat_of_formation = 72.300_dp
      ! Selenium
      ptable(34)%heat_of_formation = 54.300_dp
      ! Bromine
      ptable(35)%heat_of_formation = 26.740_dp
      ! Krypton
      ptable(36)%heat_of_formation = z
      ! Rubidium
      ptable(37)%heat_of_formation = 19.600_dp
      ! Strontium
      ptable(38)%heat_of_formation = 39.100_dp
      ! Yttrium
      ptable(39)%heat_of_formation = 101.500_dp
      ! Zirconium
      ptable(40)%heat_of_formation = 145.500_dp
      ! Niobium
      ptable(41)%heat_of_formation = 172.400_dp
      ! Molybdenum
      ptable(42)%heat_of_formation = 157.300_dp
      ! Technetium
      ptable(43)%heat_of_formation = z
      ! Ruthenium
      ptable(44)%heat_of_formation = 155.500_dp
      ! Rhodium
      ptable(45)%heat_of_formation = 133.000_dp
      ! Palladium
      ptable(46)%heat_of_formation = 90.000_dp
      ! Silver
      ptable(47)%heat_of_formation = 68.100_dp
      ! Cadmium
      ptable(48)%heat_of_formation = 26.720_dp
      ! Indium
      ptable(49)%heat_of_formation = 58.000_dp
      ! Tin
      ptable(50)%heat_of_formation = 72.200_dp
      ! Antimony
      ptable(51)%heat_of_formation = 63.200_dp
      ! Tellurium
      ptable(52)%heat_of_formation = 47.000_dp
      ! Iodine
      ptable(53)%heat_of_formation = 25.517_dp
      ! Xenon
      ptable(54)%heat_of_formation = z
      ! Cesium
      ptable(55)%heat_of_formation = 18.700_dp
      ! Barium
      ptable(56)%heat_of_formation = 42.500_dp
      ! Lanthanum
      ptable(57)%heat_of_formation = z
      ! Cerium
      ptable(58)%heat_of_formation = 101.300_dp
      ! Praseodymium
      ptable(59)%heat_of_formation = z
      ! Neodymium
      ptable(60)%heat_of_formation = z
      ! Promethium
      ptable(61)%heat_of_formation = z
      ! Samarium
      ptable(62)%heat_of_formation = 49.400_dp
      ! Europium
      ptable(63)%heat_of_formation = z
      ! Gadolinium
      ptable(64)%heat_of_formation = z
      ! Terbium
      ptable(65)%heat_of_formation = z
      ! Dysprosium
      ptable(66)%heat_of_formation = z
      ! Holmium
      ptable(67)%heat_of_formation = z
      ! Erbium
      ptable(68)%heat_of_formation = 75.800_dp
      ! Thulium
      ptable(69)%heat_of_formation = z
      ! Ytterbium
      ptable(70)%heat_of_formation = 36.350_dp
      ! Lutetium
      ptable(71)%heat_of_formation = z
      ! Hafnium
      ptable(72)%heat_of_formation = 148.000_dp
      ! Tantalum
      ptable(73)%heat_of_formation = 186.900_dp
      ! Tungsten
      ptable(74)%heat_of_formation = 203.100_dp
      ! Rhenium
      ptable(75)%heat_of_formation = 185.000_dp
      ! Osmium
      ptable(76)%heat_of_formation = 188.000_dp
      ! Iridium
      ptable(77)%heat_of_formation = 160.000_dp
      ! Platinum
      ptable(78)%heat_of_formation = 135.200_dp
      ! Gold
      ptable(79)%heat_of_formation = 88.000_dp
      ! Mercury
      ptable(80)%heat_of_formation = 14.690_dp
      ! Thallium
      ptable(81)%heat_of_formation = 43.550_dp
      ! Lead
      ptable(82)%heat_of_formation = 46.620_dp
      ! Bismuth
      ptable(83)%heat_of_formation = 50.100_dp
      ! Polonium
      ptable(84)%heat_of_formation = z
      ! Astatine
      ptable(85)%heat_of_formation = z
      ! Radon
      ! from Radon no parametrisation in dynamo
      ptable(86:nelem)%heat_of_formation = z

   END SUBROUTINE init_eheat

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE init_gratio()
      ! D. M. Granty and R. K. Harris, Encyclopedia of Nuclear
      ! Magnetic Resonance vol. 5 (Wiley, Chichester UK, 1996)
      !
      ! gyrom_ratio values in MHz/Tesla
      !
      ! Dummy
      ptable(0)%gyrom_ratio = 0.0_dp
      ptable(0)%gyrom_ratio_isotope = 0
      ! Hydrogen
      ptable(1)%gyrom_ratio = 42.5774690577_dp
      ptable(1)%gyrom_ratio_isotope = 1
      ! Helium
      ptable(2)%gyrom_ratio = -32.4360299810_dp
      ptable(2)%gyrom_ratio_isotope = 3
      ! Lithium
      ptable(3)%gyrom_ratio = 16.5484555869_dp
      ptable(3)%gyrom_ratio_isotope = 7
      ! Beryllium
      ptable(4)%gyrom_ratio = -5.9836942827_dp
      ptable(4)%gyrom_ratio_isotope = 9
      ! Boron
      ptable(5)%gyrom_ratio = 13.6629814024_dp
      ptable(5)%gyrom_ratio_isotope = 11
      ! Carbon
      ptable(6)%gyrom_ratio = 10.7083965713_dp
      ptable(6)%gyrom_ratio_isotope = 13
      ! Nitrogen
      ptable(7)%gyrom_ratio = 3.0777051853_dp
      ptable(7)%gyrom_ratio_isotope = 14
      ! Oxygen
      ptable(8)%gyrom_ratio = -5.7742686593_dp
      ptable(8)%gyrom_ratio_isotope = 17
      ! Fluorine
      ptable(9)%gyrom_ratio = 40.0775701637_dp
      ptable(9)%gyrom_ratio_isotope = 19
      ! Neon
      ptable(10)%gyrom_ratio = -3.3630712715_dp
      ptable(10)%gyrom_ratio_isotope = 21
      ! Sodium
      ptable(11)%gyrom_ratio = 11.2695216738_dp
      ptable(11)%gyrom_ratio_isotope = 23
      ! Magnesium
      ptable(12)%gyrom_ratio = -2.6083426159_dp
      ptable(12)%gyrom_ratio_isotope = 25
      ! Aluminium
      ptable(13)%gyrom_ratio = 11.1030809358_dp
      ptable(13)%gyrom_ratio_isotope = 27
      ! Silicon
      ptable(14)%gyrom_ratio = -8.4654514231_dp
      ptable(14)%gyrom_ratio_isotope = 29
      ! Phosphorus
      ptable(15)%gyrom_ratio = 17.2514409015_dp
      ptable(15)%gyrom_ratio_isotope = 31
      ! Sulfur
      ptable(16)%gyrom_ratio = 3.2717242919_dp
      ptable(16)%gyrom_ratio_isotope = 33
      ! Chlorine
      ptable(17)%gyrom_ratio = 4.1765408335_dp
      ptable(17)%gyrom_ratio_isotope = 35
      ! Argon
      ptable(18)%gyrom_ratio = 0.0_dp
      ptable(18)%gyrom_ratio_isotope = 0
      ! Potassium
      ptable(19)%gyrom_ratio = 1.9895335549_dp
      ptable(19)%gyrom_ratio_isotope = 39
      ! Calcium
      ptable(20)%gyrom_ratio = -2.8696734409_dp
      ptable(20)%gyrom_ratio_isotope = 43
      ! Scandium
      ptable(21)%gyrom_ratio = 10.3590726388_dp
      ptable(21)%gyrom_ratio_isotope = 45
      ! Titanium
      ptable(22)%gyrom_ratio = -2.4040354154_dp
      ptable(22)%gyrom_ratio_isotope = 47
      ! Vanadium
      ptable(23)%gyrom_ratio = 11.2132801367_dp
      ptable(23)%gyrom_ratio_isotope = 51
      ! Chromium
      ptable(24)%gyrom_ratio = -2.4115156977_dp
      ptable(24)%gyrom_ratio_isotope = 53
      ! Manganese
      ptable(25)%gyrom_ratio = 10.5762511769_dp
      ptable(25)%gyrom_ratio_isotope = 55
      ! Iron
      ptable(26)%gyrom_ratio = 1.3815642187_dp
      ptable(26)%gyrom_ratio_isotope = 57
      ! Cobalt
      ptable(27)%gyrom_ratio = 10.0776909966_dp
      ptable(27)%gyrom_ratio_isotope = 59
      ! Nickel
      ptable(28)%gyrom_ratio = -3.8114425772_dp
      ptable(28)%gyrom_ratio_isotope = 61
      ! Copper
      ptable(29)%gyrom_ratio = 11.3187637358_dp
      ptable(29)%gyrom_ratio_isotope = 63
      ! Zinc
      ptable(30)%gyrom_ratio = 2.6685318322_dp
      ptable(30)%gyrom_ratio_isotope = 67
      ! Gallium
      ptable(31)%gyrom_ratio = 10.2477560110_dp
      ptable(31)%gyrom_ratio_isotope = 69
      ! Germanium
      ptable(32)%gyrom_ratio = -1.4897384913_dp
      ptable(32)%gyrom_ratio_isotope = 73
      ! Arsenic
      ptable(33)%gyrom_ratio = 7.3150206071_dp
      ptable(33)%gyrom_ratio_isotope = 75
      ! Selenium
      ptable(34)%gyrom_ratio = 8.1573046941_dp
      ptable(34)%gyrom_ratio_isotope = 77
      ! Bromine
      ptable(35)%gyrom_ratio = 10.7041503174_dp
      ptable(35)%gyrom_ratio_isotope = 79
      ! Krypton
      ptable(36)%gyrom_ratio = -1.6442297171_dp
      ptable(36)%gyrom_ratio_isotope = 83
      ! Rubidium
      ptable(37)%gyrom_ratio = 4.1264181673_dp
      ptable(37)%gyrom_ratio_isotope = 85
      ! Strontium
      ptable(38)%gyrom_ratio = -1.8524642249_dp
      ptable(38)%gyrom_ratio_isotope = 87
      ! Yttrium
      ptable(39)%gyrom_ratio = -2.0949232525_dp
      ptable(39)%gyrom_ratio_isotope = 89
      ! Zirconium
      ptable(40)%gyrom_ratio = -3.9747832953_dp
      ptable(40)%gyrom_ratio_isotope = 91
      ! Niobium
      ptable(41)%gyrom_ratio = 10.4523417326_dp
      ptable(41)%gyrom_ratio_isotope = 93
      ! Molybdenum
      ptable(42)%gyrom_ratio = -2.7868030535_dp
      ptable(42)%gyrom_ratio_isotope = 95
      ! Technetium
      ptable(43)%gyrom_ratio = 9.6225078593_dp
      ptable(43)%gyrom_ratio_isotope = 99
      ! Ruthenium
      ptable(44)%gyrom_ratio = -2.1915635664_dp
      ptable(44)%gyrom_ratio_isotope = 101
      ! Rhodium
      ptable(45)%gyrom_ratio = -1.3477240581_dp
      ptable(45)%gyrom_ratio_isotope = 103
      ! Palladium
      ptable(46)%gyrom_ratio = -1.9576058000_dp
      ptable(46)%gyrom_ratio_isotope = 105
      ! Silver
      ptable(47)%gyrom_ratio = -1.7330669824_dp
      ptable(47)%gyrom_ratio_isotope = 107
      ! Cadmium
      ptable(48)%gyrom_ratio = -9.0691469715_dp
      ptable(48)%gyrom_ratio_isotope = 111
      ! Indium
      ptable(49)%gyrom_ratio = 9.3856853040_dp
      ptable(49)%gyrom_ratio_isotope = 115
      ! Tin
      ptable(50)%gyrom_ratio = -15.9659464261_dp
      ptable(50)%gyrom_ratio_isotope = 119
      ! Antimony
      ptable(51)%gyrom_ratio = 10.2551487581_dp
      ptable(51)%gyrom_ratio_isotope = 121
      ! Tellurium
      ptable(52)%gyrom_ratio = -13.5454231953_dp
      ptable(52)%gyrom_ratio_isotope = 125
      ! Iodine
      ptable(53)%gyrom_ratio = 8.5777718410_dp
      ptable(53)%gyrom_ratio_isotope = 127
      ! Xenon
      ptable(54)%gyrom_ratio = -11.8603902888_dp
      ptable(54)%gyrom_ratio_isotope = 129
      ! Cesium
      ptable(55)%gyrom_ratio = 5.6233482338_dp
      ptable(55)%gyrom_ratio_isotope = 133
      ! Barium
      ptable(56)%gyrom_ratio = 4.7634278693_dp
      ptable(56)%gyrom_ratio_isotope = 137
      ! Lantanum
      ptable(57)%gyrom_ratio = 6.0611483090_dp
      ptable(57)%gyrom_ratio_isotope = 139
      ! Cerium
      ptable(58)%gyrom_ratio = 0.0_dp
      ptable(58)%gyrom_ratio_isotope = 0
      ! Praseodymium
      ptable(59)%gyrom_ratio = 13.0359039238_dp
      ptable(59)%gyrom_ratio_isotope = 141
      ! Neodymium
      ptable(60)%gyrom_ratio = -2.3188875208_dp
      ptable(60)%gyrom_ratio_isotope = 143
      ! Promethium
      ptable(61)%gyrom_ratio = 5.7502680939_dp
      ptable(61)%gyrom_ratio_isotope = 147
      ! Samarium
      ptable(62)%gyrom_ratio = -1.7745776155_dp
      ptable(62)%gyrom_ratio_isotope = 147
      ! Europium
      ptable(63)%gyrom_ratio = 4.6742215237_dp
      ptable(63)%gyrom_ratio_isotope = 153
      ! Gadolinium
      ptable(64)%gyrom_ratio = -1.7139395822_dp
      ptable(64)%gyrom_ratio_isotope = 157
      ! Terbium
      ptable(65)%gyrom_ratio = 10.2352543902_dp
      ptable(65)%gyrom_ratio_isotope = 159
      ! Dysprosium
      ptable(66)%gyrom_ratio = 2.0515072165_dp
      ptable(66)%gyrom_ratio_isotope = 163
      ! Holmium
      ptable(67)%gyrom_ratio = 9.0877472505_dp
      ptable(67)%gyrom_ratio_isotope = 165
      ! Erbium
      ptable(68)%gyrom_ratio = -1.2279917944_dp
      ptable(68)%gyrom_ratio_isotope = 167
      ! Thulium
      ptable(69)%gyrom_ratio = -3.5300566378_dp
      ptable(69)%gyrom_ratio_isotope = 169
      ! Ytterbium
      ptable(70)%gyrom_ratio = -2.0729931338_dp
      ptable(70)%gyrom_ratio_isotope = 173
      ! Lutetium
      ptable(71)%gyrom_ratio = 4.8625018213_dp
      ptable(71)%gyrom_ratio_isotope = 175
      ! Hafnium
      ptable(72)%gyrom_ratio = 1.7284226820_dp
      ptable(72)%gyrom_ratio_isotope = 177
      ! Tantalum
      ptable(73)%gyrom_ratio = 5.1626680440_dp
      ptable(73)%gyrom_ratio_isotope = 181
      ! Tungsten
      ptable(74)%gyrom_ratio = 1.7956502074_dp
      ptable(74)%gyrom_ratio_isotope = 183
      ! Rhenium
      ptable(75)%gyrom_ratio = 9.8169951998_dp
      ptable(75)%gyrom_ratio_isotope = 187
      ! Osmium
      ptable(76)%gyrom_ratio = 3.3536015524_dp
      ptable(76)%gyrom_ratio_isotope = 189
      ! Iridium
      ptable(77)%gyrom_ratio = 0.8319028875_dp
      ptable(77)%gyrom_ratio_isotope = 193
      ! Platinum
      ptable(78)%gyrom_ratio = 9.2922613524_dp
      ptable(78)%gyrom_ratio_isotope = 195
      ! Gold
      ptable(79)%gyrom_ratio = 0.7528983738_dp
      ptable(79)%gyrom_ratio_isotope = 197
      ! Mercury
      ptable(80)%gyrom_ratio = 7.7123168633_dp
      ptable(80)%gyrom_ratio_isotope = 199
      ! Thallium
      ptable(81)%gyrom_ratio = 24.9748814221_dp
      ptable(81)%gyrom_ratio_isotope = 205
      ! Lead
      ptable(82)%gyrom_ratio = 8.8815779373_dp
      ptable(82)%gyrom_ratio_isotope = 207
      ! Bismuth
      ptable(83)%gyrom_ratio = 6.9630287603_dp
      ptable(83)%gyrom_ratio_isotope = 209
      ! Polonium
      ptable(84)%gyrom_ratio = 11.7774657888_dp
      ptable(84)%gyrom_ratio_isotope = 209
      ! Astatine
      ptable(85)%gyrom_ratio = 0.0_dp
      ptable(85)%gyrom_ratio_isotope = 0
      ! Radon
      ptable(86)%gyrom_ratio = 0.0_dp
      ptable(86)%gyrom_ratio_isotope = 0
      ! Francium
      ptable(87)%gyrom_ratio = 0.0_dp
      ptable(87)%gyrom_ratio_isotope = 0
      ! Radium
      ptable(88)%gyrom_ratio = 0.0_dp
      ptable(88)%gyrom_ratio_isotope = 0
      ! Actinium
      ptable(89)%gyrom_ratio = 5.5704230082_dp
      ptable(89)%gyrom_ratio_isotope = 227
      ! Thorium
      ptable(90)%gyrom_ratio = 0.6366197724_dp
      ptable(90)%gyrom_ratio_isotope = 229
      ! Proctactinium
      ptable(91)%gyrom_ratio = 5.1088736732_dp
      ptable(91)%gyrom_ratio_isotope = 231
      ! Uranium
      ptable(92)%gyrom_ratio = -0.8276057041_dp
      ptable(92)%gyrom_ratio_isotope = 235
      ! Neptunium
      ptable(93)%gyrom_ratio = 4.9338032358_dp
      ptable(93)%gyrom_ratio_isotope = 237
      ! Plutonium
      ptable(94)%gyrom_ratio = 1.5469860469_dp
      ptable(94)%gyrom_ratio_isotope = 239
      ! Americium
      ptable(95)%gyrom_ratio = 2.4509861236_dp
      ptable(95)%gyrom_ratio_isotope = 243
      ! Curium
      ptable(96)%gyrom_ratio = 0.3183098862_dp
      ptable(96)%gyrom_ratio_isotope = 247
      ! No data available
      ptable(97:nelem)%gyrom_ratio = 0.0_dp
      ptable(97:nelem)%gyrom_ratio_isotope = 0

   END SUBROUTINE init_gratio

END MODULE periodic_table
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of the atomic potential types.
!> \par History
!>      GT, 22.09.2002: added elp_potential_types
!> \author Matthias Krack (04.07.2000)
! **************************************************************************************************
MODULE external_potential_types
   USE ao_util,                         ONLY: exp_radius
   USE bibliography,                    ONLY: Goedecker1996,&
                                              Hartwigsen1998,&
                                              Krack2000,&
                                              Krack2005,&
                                              cite_reference
   USE cp_linked_list_val,              ONLY: cp_sll_val_next,&
                                              cp_sll_val_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_search_string,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_list_get,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE input_val_types,                 ONLY: val_get,&
                                              val_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              pi,&
                                              rootpi
   USE mathlib,                         ONLY: symmetrize_matrix
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: co,&
                                              coset,&
                                              init_orbital_pointers,&
                                              nco,&
                                              ncoset,&
                                              nso
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE periodic_table,                  ONLY: ptable
   USE string_utilities,                ONLY: remove_word,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'external_potential_types'

   ! Define the all-electron potential type

   ! Literature: M. Krack and M. Parrinello,
   !             Phys. Chem. Chem. Phys. 2, 2105 (2000)

! **************************************************************************************************
   TYPE all_potential_type
      !MK PRIVATE
      CHARACTER(LEN=default_string_length)   :: name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(2)                      :: description
      REAL(KIND=dp)                        :: alpha_core_charge, &
                                              ccore_charge, &
                                              core_charge_radius, &
                                              zeff, zeff_correction
      INTEGER                                :: z
      INTEGER, DIMENSION(:), POINTER         :: elec_conf
   END TYPE all_potential_type

   ! Define the effective charge & inducible dipole potential type (for Fist)
! **************************************************************************************************
   TYPE fist_potential_type
      PRIVATE
      CHARACTER(LEN=default_string_length)     :: name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(1)                        :: description
      REAL(KIND=dp)                          :: apol, cpol, mm_radius, qeff, &
                                                qmmm_corr_radius, qmmm_radius

   END TYPE fist_potential_type

! **************************************************************************************************
! local potential type
! V(r) = SUM_i exp(0.5*(r/rci)**2) * ( C1i + C2i (r/rci)**2 + C3i (r/rci)**4 ...)
! alpha = 0.5/rci**2
   TYPE local_potential_type
      !PRIVATE
      CHARACTER(LEN=default_string_length)       :: name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(4)                          :: description
      INTEGER                                    :: ngau, npol
      REAL(KIND=dp)                            :: radius
      REAL(KIND=dp), DIMENSION(:), POINTER     :: alpha
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: cval
   END TYPE local_potential_type

   ! Define the GTH potential type

   ! Literature: - S. Goedecker, M. Teter and J. Hutter,
   !               Phys. Rev. B 54, 1703 (1996)
   !             - C. Hartwigsen, S. Goedecker and J. Hutter,
   !               Phys. Rev. B 58, 3641 (1998)
   !             - M. Krack,
   !               Theor. Chem. Acc. 114, 145 (2005)

! **************************************************************************************************
   TYPE gth_potential_type
      !PRIVATE
      CHARACTER(LEN=default_string_length)       :: name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(4)                          :: description
      REAL(KIND=dp)                            :: alpha_core_charge, &
                                                  alpha_ppl, ccore_charge, &
                                                  cerf_ppl, zeff, &
                                                  core_charge_radius, &
                                                  ppl_radius, ppnl_radius, &
                                                  zeff_correction
      INTEGER                                    :: lppnl, lprj_ppnl_max, &
                                                    nexp_ppl, nppnl, &
                                                    nprj_ppnl_max, z
      REAL(KIND=dp), DIMENSION(:), POINTER     :: alpha_ppnl, cexp_ppl
      INTEGER, DIMENSION(:), POINTER             :: elec_conf
      ! nonlocal projectors
      INTEGER, DIMENSION(:), POINTER             :: nprj_ppnl
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: cprj, cprj_ppnl, vprj_ppnl
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: hprj_ppnl
      ! type extensions
      ! NLCC
      LOGICAL                                    :: nlcc
      INTEGER                                    :: nexp_nlcc
      REAL(KIND=dp), DIMENSION(:), POINTER     :: alpha_nlcc
      INTEGER, DIMENSION(:), POINTER             :: nct_nlcc
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: cval_nlcc
      ! LSD potential
      LOGICAL                                    :: lsdpot
      INTEGER                                    :: nexp_lsd
      REAL(KIND=dp), DIMENSION(:), POINTER     :: alpha_lsd
      INTEGER, DIMENSION(:), POINTER             :: nct_lsd
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: cval_lsd
      ! extended local potential
      LOGICAL                                    :: lpotextended
      INTEGER                                    :: nexp_lpot
      REAL(KIND=dp), DIMENSION(:), POINTER     :: alpha_lpot
      INTEGER, DIMENSION(:), POINTER             :: nct_lpot
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: cval_lpot

   END TYPE gth_potential_type

   TYPE all_potential_p_type
      TYPE(all_potential_type), POINTER          :: all_potential
   END TYPE all_potential_p_type

   TYPE gth_potential_p_type
      TYPE(gth_potential_type), POINTER          :: gth_potential
   END TYPE gth_potential_p_type

   TYPE local_potential_p_type
      TYPE(local_potential_type), POINTER        :: local_potential
   END TYPE local_potential_p_type

! **************************************************************************************************

   ! Public subroutines
   PUBLIC :: allocate_potential, &
             deallocate_potential, &
             get_potential, &
             init_potential, &
             read_potential, &
             set_potential, &
             set_default_all_potential, &
             write_potential

   ! Public data types

   PUBLIC :: all_potential_type, &
             fist_potential_type, &
             local_potential_type, &
             gth_potential_type
   PUBLIC :: gth_potential_p_type

   INTERFACE allocate_potential
      MODULE PROCEDURE allocate_all_potential, &
         allocate_fist_potential, &
         allocate_local_potential, &
         allocate_gth_potential
   END INTERFACE

   INTERFACE deallocate_potential
      MODULE PROCEDURE deallocate_all_potential, &
         deallocate_fist_potential, &
         deallocate_local_potential, &
         deallocate_gth_potential
   END INTERFACE

   INTERFACE get_potential
      MODULE PROCEDURE get_all_potential, &
         get_fist_potential, &
         get_local_potential, &
         get_gth_potential
   END INTERFACE

   INTERFACE init_potential
      MODULE PROCEDURE init_all_potential, &
         init_gth_potential
   END INTERFACE

   INTERFACE read_potential
      MODULE PROCEDURE read_all_potential, &
         read_local_potential, &
         read_gth_potential
   END INTERFACE

   INTERFACE set_potential
      MODULE PROCEDURE set_all_potential, &
         set_fist_potential, &
         set_local_potential, &
         set_gth_potential
   END INTERFACE

   INTERFACE write_potential
      MODULE PROCEDURE write_all_potential, &
         write_local_potential, &
         write_gth_potential
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief   Allocate an atomic all-electron potential data set.
!> \param potential ...
!> \date    25.07.2000,
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_all_potential(potential)
      TYPE(all_potential_type), POINTER                  :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_all_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) CALL deallocate_potential(potential)

      ALLOCATE (potential)

      NULLIFY (potential%elec_conf)

      potential%description(1) = "All-electron potential"
      potential%description(2) = "Krack, Parrinello, PCCP 2, 2105 (2000)"

   END SUBROUTINE allocate_all_potential

! **************************************************************************************************
!> \brief   Allocate an effective charge and inducible dipole potential data set.
!> \param potential ...
!> \date    05.03.2010
!> \author  Toon.Verstraelen@gmail.com
! **************************************************************************************************
   SUBROUTINE allocate_fist_potential(potential)
      TYPE(fist_potential_type), POINTER                 :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_fist_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) CALL deallocate_potential(potential)

      ALLOCATE (potential)

      potential%apol = 0.0_dp
      potential%cpol = 0.0_dp
      potential%mm_radius = 0.0_dp
      potential%qeff = 0.0_dp
      potential%qmmm_radius = 0.0_dp
      potential%qmmm_corr_radius = 0.0_dp

      potential%description(1) = "Effective charge and inducible dipole potential"

   END SUBROUTINE allocate_fist_potential

! **************************************************************************************************
!> \brief   Allocate an atomic local potential data set.
!> \param potential ...
!> \date    24.01.2014
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_local_potential(potential)
      TYPE(local_potential_type), POINTER                :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_local_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) CALL deallocate_potential(potential)

      ALLOCATE (potential)

      NULLIFY (potential%alpha)
      NULLIFY (potential%cval)

      potential%description(1) = "Local short-range pseudopotential"
      potential%ngau = 0
      potential%npol = 0
      potential%radius = 0.0_dp

   END SUBROUTINE allocate_local_potential

! **************************************************************************************************
!> \brief   Allocate an atomic GTH potential data set.
!> \param potential ...
!> \date    25.07.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_gth_potential(potential)
      TYPE(gth_potential_type), POINTER                  :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_gth_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) CALL deallocate_potential(potential)

      ALLOCATE (potential)

      NULLIFY (potential%alpha_ppnl)
      NULLIFY (potential%cexp_ppl)
      NULLIFY (potential%elec_conf)
      NULLIFY (potential%nprj_ppnl)
      NULLIFY (potential%cprj)
      NULLIFY (potential%cprj_ppnl)
      NULLIFY (potential%vprj_ppnl)
      NULLIFY (potential%hprj_ppnl)

      NULLIFY (potential%alpha_lpot)
      NULLIFY (potential%nct_lpot)
      NULLIFY (potential%cval_lpot)
      NULLIFY (potential%alpha_lsd)
      NULLIFY (potential%nct_lsd)
      NULLIFY (potential%cval_lsd)
      NULLIFY (potential%alpha_nlcc)
      NULLIFY (potential%nct_nlcc)
      NULLIFY (potential%cval_nlcc)

      potential%description(1) = "Goedecker-Teter-Hutter pseudopotential"
      potential%description(2) = "Goedecker et al., PRB 54, 1703 (1996)"
      potential%description(3) = "Hartwigsen et al., PRB 58, 3641 (1998)"
      potential%description(4) = "Krack, TCA 114, 145 (2005)"

   END SUBROUTINE allocate_gth_potential

! **************************************************************************************************
!> \brief   Deallocate an atomic all-electron potential data set.
!> \param potential ...
!> \date    03.11.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_all_potential(potential)
      TYPE(all_potential_type), POINTER                  :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_all_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN
         DEALLOCATE (potential%elec_conf)
         DEALLOCATE (potential)
      ELSE
         CALL cp__b("subsys/external_potential_types.F",380,"The pointer potential is not associated.")
      END IF

   END SUBROUTINE deallocate_all_potential

! **************************************************************************************************
!> \brief   Deallocate an effective charge and inducible dipole potential data set.
!> \param potential ...
!> \date    05.03.2010
!> \author  Toon.Verstraelen@gmail.com
! **************************************************************************************************
   SUBROUTINE deallocate_fist_potential(potential)
      TYPE(fist_potential_type), POINTER                 :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_fist_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN
         ! Nothing exciting here yet.
         DEALLOCATE (potential)
      ELSE
         CALL cp__b("subsys/external_potential_types.F",401,"The pointer potential is not associated.")
      END IF

   END SUBROUTINE deallocate_fist_potential

! **************************************************************************************************
!> \brief   Deallocate an atomic local potential data set.
!> \param potential ...
!> \date    24.01.2014
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_local_potential(potential)
      TYPE(local_potential_type), POINTER                :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_local_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (ASSOCIATED(potential%alpha)) THEN
            DEALLOCATE (potential%alpha)
         END IF
         IF (ASSOCIATED(potential%cval)) THEN
            DEALLOCATE (potential%cval)
         END IF

         DEALLOCATE (potential)

      ELSE

         CALL cp__b("subsys/external_potential_types.F",432,"The pointer potential is not associated.")

      END IF

   END SUBROUTINE deallocate_local_potential

! **************************************************************************************************
!> \brief   Deallocate an atomic GTH potential data set.
!> \param potential ...
!> \date    03.11.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_gth_potential(potential)
      TYPE(gth_potential_type), POINTER                  :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_gth_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         DEALLOCATE (potential%elec_conf)
         !     *** Deallocate the parameters of the local part ***

         IF (ASSOCIATED(potential%cexp_ppl)) THEN
            DEALLOCATE (potential%cexp_ppl)
         END IF

         !     *** Deallocate the parameters of the non-local part ***
         IF (ASSOCIATED(potential%alpha_ppnl)) THEN
            DEALLOCATE (potential%alpha_ppnl)
            DEALLOCATE (potential%cprj)
            DEALLOCATE (potential%cprj_ppnl)
            DEALLOCATE (potential%hprj_ppnl)
            DEALLOCATE (potential%nprj_ppnl)
            DEALLOCATE (potential%vprj_ppnl)
         END IF

         IF (ASSOCIATED(potential%alpha_lpot)) THEN
            DEALLOCATE (potential%alpha_lpot)
            DEALLOCATE (potential%nct_lpot)
            DEALLOCATE (potential%cval_lpot)
         END IF

         IF (ASSOCIATED(potential%alpha_lsd)) THEN
            DEALLOCATE (potential%alpha_lsd)
            DEALLOCATE (potential%nct_lsd)
            DEALLOCATE (potential%cval_lsd)
         END IF

         IF (ASSOCIATED(potential%alpha_nlcc)) THEN
            DEALLOCATE (potential%alpha_nlcc)
            DEALLOCATE (potential%nct_nlcc)
            DEALLOCATE (potential%cval_nlcc)
         END IF

         DEALLOCATE (potential)
      ELSE

         CALL cp__b("subsys/external_potential_types.F",491,"The pointer potential is not associated.")
      END IF

   END SUBROUTINE deallocate_gth_potential

! **************************************************************************************************
!> \brief   Get attributes of an all-electron potential data set.
!> \param potential ...
!> \param name ...
!> \param alpha_core_charge ...
!> \param ccore_charge ...
!> \param core_charge_radius ...
!> \param z ...
!> \param zeff ...
!> \param zeff_correction ...
!> \param elec_conf ...
!> \date    11.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_all_potential(potential, name, alpha_core_charge, &
                                ccore_charge, core_charge_radius, z, zeff, &
                                zeff_correction, elec_conf)
      TYPE(all_potential_type), POINTER                  :: potential
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: alpha_core_charge, ccore_charge, &
                                                            core_charge_radius
      INTEGER, INTENT(OUT), OPTIONAL                     :: z
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: zeff, zeff_correction
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: elec_conf

      CHARACTER(len=*), PARAMETER :: routineN = 'get_all_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(name)) name = potential%name
         IF (PRESENT(alpha_core_charge)) &
            alpha_core_charge = potential%alpha_core_charge
         IF (PRESENT(ccore_charge)) ccore_charge = potential%ccore_charge
         IF (PRESENT(core_charge_radius)) &
            core_charge_radius = potential%core_charge_radius
         IF (PRESENT(z)) z = potential%z
         IF (PRESENT(zeff)) zeff = potential%zeff
         IF (PRESENT(zeff_correction)) zeff_correction = potential%zeff_correction
         IF (PRESENT(elec_conf)) elec_conf => potential%elec_conf

      ELSE

         CALL cp__b("subsys/external_potential_types.F",541,"The pointer potential is not associated.")

      END IF

   END SUBROUTINE get_all_potential

! **************************************************************************************************
!> \brief   Get attributes of an effective point charge and inducible dipole
!>          potential.
!> \param potential ...
!> \param name ...
!> \param apol ...
!> \param cpol ...
!> \param mm_radius ...
!> \param qeff ...
!> \param qmmm_corr_radius ...
!> \param qmmm_radius ...
!> \date    05.03-2010
!> \author  Toon.Verstraelen@UGent.be
! **************************************************************************************************
   SUBROUTINE get_fist_potential(potential, name, apol, cpol, mm_radius, qeff, &
                                 qmmm_corr_radius, qmmm_radius)
      TYPE(fist_potential_type), POINTER                 :: potential
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: apol, cpol, mm_radius, qeff, &
                                                            qmmm_corr_radius, qmmm_radius

      CHARACTER(len=*), PARAMETER :: routineN = 'get_fist_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(name)) name = potential%name
         IF (PRESENT(apol)) apol = potential%apol
         IF (PRESENT(cpol)) cpol = potential%cpol
         IF (PRESENT(mm_radius)) mm_radius = potential%mm_radius
         IF (PRESENT(qeff)) qeff = potential%qeff
         IF (PRESENT(qmmm_corr_radius)) qmmm_corr_radius = potential%qmmm_corr_radius
         IF (PRESENT(qmmm_radius)) qmmm_radius = potential%qmmm_radius

      ELSE

         CALL cp__b("subsys/external_potential_types.F",584,"The pointer potential is not associated.")

      END IF

   END SUBROUTINE get_fist_potential

! **************************************************************************************************
!> \brief   Get attributes of an atomic local potential data set.
!> \param potential ...
!> \param name ...
!> \param ngau ...
!> \param npol ...
!> \param alpha ...
!> \param cval ...
!> \param radius ...
!> \date    24.01.2014
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_local_potential(potential, name, ngau, npol, alpha, cval, radius)
      TYPE(local_potential_type), POINTER                :: potential
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      INTEGER, INTENT(OUT), OPTIONAL                     :: ngau, npol
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: alpha
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cval
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: radius

      CHARACTER(len=*), PARAMETER :: routineN = 'get_local_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(name)) name = potential%name
         IF (PRESENT(ngau)) ngau = potential%ngau
         IF (PRESENT(npol)) npol = potential%npol
         IF (PRESENT(alpha)) alpha => potential%alpha
         IF (PRESENT(cval)) cval => potential%cval
         IF (PRESENT(radius)) radius = potential%radius

      ELSE

         CALL cp__b("subsys/external_potential_types.F",626,"The pointer potential is not associated.")

      END IF

   END SUBROUTINE get_local_potential

! **************************************************************************************************
!> \brief   Get attributes of a GTH potential data set.
!> \param potential ...
!> \param name ...
!> \param alpha_core_charge ...
!> \param alpha_ppl ...
!> \param ccore_charge ...
!> \param cerf_ppl ...
!> \param core_charge_radius ...
!> \param ppl_radius ...
!> \param ppnl_radius ...
!> \param lppnl ...
!> \param lprj_ppnl_max ...
!> \param nexp_ppl ...
!> \param nppnl ...
!> \param nprj_ppnl_max ...
!> \param z ...
!> \param zeff ...
!> \param zeff_correction ...
!> \param ppl_present ...
!> \param ppnl_present ...
!> \param alpha_ppnl ...
!> \param cexp_ppl ...
!> \param elec_conf ...
!> \param nprj_ppnl ...
!> \param cprj ...
!> \param cprj_ppnl ...
!> \param vprj_ppnl ...
!> \param hprj_ppnl ...
!> \param lpot_present ...
!> \param nexp_lpot ...
!> \param alpha_lpot ...
!> \param nct_lpot ...
!> \param cval_lpot ...
!> \param lsd_present ...
!> \param nexp_lsd ...
!> \param alpha_lsd ...
!> \param nct_lsd ...
!> \param cval_lsd ...
!> \param nlcc_present ...
!> \param nexp_nlcc ...
!> \param alpha_nlcc ...
!> \param nct_nlcc ...
!> \param cval_nlcc ...
!> \date    11.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_gth_potential(potential, name, alpha_core_charge, &
                                alpha_ppl, ccore_charge, cerf_ppl, &
                                core_charge_radius, ppl_radius, ppnl_radius, &
                                lppnl, lprj_ppnl_max, nexp_ppl, nppnl, &
                                nprj_ppnl_max, z, zeff, zeff_correction, &
                                ppl_present, ppnl_present, &
                                alpha_ppnl, cexp_ppl, elec_conf, nprj_ppnl, cprj, &
                                cprj_ppnl, vprj_ppnl, hprj_ppnl, &
                                lpot_present, nexp_lpot, alpha_lpot, nct_lpot, cval_lpot, &
                                lsd_present, nexp_lsd, alpha_lsd, nct_lsd, cval_lsd, &
                                nlcc_present, nexp_nlcc, alpha_nlcc, nct_nlcc, cval_nlcc)

      TYPE(gth_potential_type), POINTER                  :: potential
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: alpha_core_charge, alpha_ppl, &
                                                            ccore_charge, cerf_ppl, &
                                                            core_charge_radius, ppl_radius, &
                                                            ppnl_radius
      INTEGER, INTENT(OUT), OPTIONAL                     :: lppnl, lprj_ppnl_max, nexp_ppl, nppnl, &
                                                            nprj_ppnl_max, z
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: zeff, zeff_correction
      LOGICAL, INTENT(OUT), OPTIONAL                     :: ppl_present, ppnl_present
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: alpha_ppnl, cexp_ppl
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: elec_conf, nprj_ppnl
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cprj, cprj_ppnl, vprj_ppnl
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: hprj_ppnl
      LOGICAL, INTENT(OUT), OPTIONAL                     :: lpot_present
      INTEGER, INTENT(OUT), OPTIONAL                     :: nexp_lpot
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: alpha_lpot
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nct_lpot
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cval_lpot
      LOGICAL, INTENT(OUT), OPTIONAL                     :: lsd_present
      INTEGER, INTENT(OUT), OPTIONAL                     :: nexp_lsd
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: alpha_lsd
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nct_lsd
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cval_lsd
      LOGICAL, INTENT(OUT), OPTIONAL                     :: nlcc_present
      INTEGER, INTENT(OUT), OPTIONAL                     :: nexp_nlcc
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: alpha_nlcc
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nct_nlcc
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cval_nlcc

      CHARACTER(len=*), PARAMETER :: routineN = 'get_gth_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(name)) name = potential%name
         IF (PRESENT(alpha_core_charge)) &
            alpha_core_charge = potential%alpha_core_charge
         IF (PRESENT(alpha_ppl)) alpha_ppl = potential%alpha_ppl
         IF (PRESENT(ccore_charge)) ccore_charge = potential%ccore_charge
         IF (PRESENT(cerf_ppl)) cerf_ppl = potential%cerf_ppl
         IF (PRESENT(core_charge_radius)) &
            core_charge_radius = potential%core_charge_radius
         IF (PRESENT(ppl_radius)) ppl_radius = potential%ppl_radius
         IF (PRESENT(ppnl_radius)) ppnl_radius = potential%ppnl_radius
         IF (PRESENT(lppnl)) lppnl = potential%lppnl
         IF (PRESENT(lprj_ppnl_max)) lprj_ppnl_max = potential%lprj_ppnl_max
         IF (PRESENT(nexp_ppl)) nexp_ppl = potential%nexp_ppl
         IF (PRESENT(nppnl)) nppnl = potential%nppnl
         IF (PRESENT(nprj_ppnl_max)) nprj_ppnl_max = potential%nprj_ppnl_max
         IF (PRESENT(z)) z = potential%z
         IF (PRESENT(zeff)) zeff = potential%zeff
         IF (PRESENT(zeff_correction)) zeff_correction = potential%zeff_correction
         IF (PRESENT(ppl_present)) ppl_present = (potential%nexp_ppl > 0)
         IF (PRESENT(ppnl_present)) ppnl_present = (potential%nppnl > 0)
         IF (PRESENT(alpha_ppnl)) alpha_ppnl => potential%alpha_ppnl
         IF (PRESENT(cexp_ppl)) cexp_ppl => potential%cexp_ppl
         IF (PRESENT(elec_conf)) elec_conf => potential%elec_conf
         IF (PRESENT(nprj_ppnl)) nprj_ppnl => potential%nprj_ppnl
         IF (PRESENT(cprj)) cprj => potential%cprj
         IF (PRESENT(cprj_ppnl)) cprj_ppnl => potential%cprj_ppnl
         IF (PRESENT(vprj_ppnl)) vprj_ppnl => potential%vprj_ppnl
         IF (PRESENT(hprj_ppnl)) hprj_ppnl => potential%hprj_ppnl

         IF (PRESENT(lpot_present)) lpot_present = potential%lpotextended
         IF (PRESENT(nexp_lpot)) nexp_lpot = potential%nexp_lpot
         IF (PRESENT(alpha_lpot)) alpha_lpot => potential%alpha_lpot
         IF (PRESENT(nct_lpot)) nct_lpot => potential%nct_lpot
         IF (PRESENT(cval_lpot)) cval_lpot => potential%cval_lpot

         IF (PRESENT(lsd_present)) lsd_present = potential%lsdpot
         IF (PRESENT(nexp_lsd)) nexp_lsd = potential%nexp_lsd
         IF (PRESENT(alpha_lsd)) alpha_lsd => potential%alpha_lsd
         IF (PRESENT(nct_lsd)) nct_lsd => potential%nct_lsd
         IF (PRESENT(cval_lsd)) cval_lsd => potential%cval_lsd

         IF (PRESENT(nlcc_present)) nlcc_present = potential%nlcc
         IF (PRESENT(nexp_nlcc)) nexp_nlcc = potential%nexp_nlcc
         IF (PRESENT(alpha_nlcc)) alpha_nlcc => potential%alpha_nlcc
         IF (PRESENT(nct_nlcc)) nct_nlcc => potential%nct_nlcc
         IF (PRESENT(cval_nlcc)) cval_nlcc => potential%cval_nlcc

      ELSE

         CALL cp__b("subsys/external_potential_types.F",778,"The pointer potential is not associated.")

      END IF

   END SUBROUTINE get_gth_potential

! **************************************************************************************************
!> \brief   Initialise the coefficients of the projectors of the non-local
!>          part of the GTH pseudopotential and the transformation matrices
!>          for Cartesian overlap integrals between the orbital basis
!>          functions and the projector functions.
!> \param potential ...
!> \date    16.10.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_cprj_ppnl(potential)
      TYPE(gth_potential_type), POINTER                  :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'init_cprj_ppnl', routineP = moduleN//':'//routineN

      INTEGER                                            :: cpx, cpy, cpz, cx, cy, cz, ico, iprj, &
                                                            iprj_ppnl, l, lp, lprj_ppnl, nprj, px, &
                                                            py, pz
      REAL(KIND=dp)                                      :: alpha_ppnl, cp

      nprj = 0

      DO l = 0, potential%lppnl
         alpha_ppnl = potential%alpha_ppnl(l)
         DO iprj_ppnl = 1, potential%nprj_ppnl(l)
            lp = iprj_ppnl-1
            lprj_ppnl = l+2*lp
            cp = SQRT(2.0_dp**(2.0_dp*REAL(lprj_ppnl, dp)+3.5_dp)* &
                      alpha_ppnl**(REAL(lprj_ppnl, dp)+1.5_dp)/ &
                      (rootpi*dfac(2*lprj_ppnl+1)))
            potential%cprj_ppnl(iprj_ppnl, l) = cp
            DO cx = 0, l
               DO cy = 0, l-cx
                  cz = l-cx-cy
                  iprj = nprj+co(cx, cy, cz)
                  DO px = 0, lp
                     DO py = 0, lp-px
                        pz = lp-px-py
                        cpx = cx+2*px
                        cpy = cy+2*py
                        cpz = cz+2*pz
                        ico = coset(cpx, cpy, cpz)
                        potential%cprj(ico, iprj) = cp*fac(lp)/(fac(px)*fac(py)*fac(pz))
                     END DO
                  END DO
               END DO
            END DO
            nprj = nprj+nco(l)
         END DO
      END DO

   END SUBROUTINE init_cprj_ppnl

! **************************************************************************************************
!> \brief   Initialise a GTH potential data set structure.
!> \param potential ...
!> \date    27.10.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_gth_potential(potential)
      TYPE(gth_potential_type), POINTER                  :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'init_gth_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF (.NOT. ASSOCIATED(potential)) RETURN

      CALL timeset(routineN, handle)

      IF (potential%nppnl > 0) THEN

         !     *** Initialise the projector coefficients of the    ***
         !     *** non-local part of the GTH pseudopotential and   ***
         !     *** the transformation matrices "pgf" -> "prj_ppnl" ***

         CALL init_cprj_ppnl(potential)

         !     *** Initialise the h(i,j) projector coefficients of ***
         !     *** the non-local part of the GTH pseudopotential   ***

         CALL init_vprj_ppnl(potential)

      END IF

      CALL timestop(handle)

   END SUBROUTINE init_gth_potential

! **************************************************************************************************
!> \brief   Initialise the h(i,j) projector coefficients of the non-local part
!>          of the GTH pseudopotential.
!> \param potential ...
!> \date    24.10.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_vprj_ppnl(potential)
      TYPE(gth_potential_type), POINTER                  :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'init_vprj_ppnl', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ico, iprj, iprj_ppnl, iso, j, jco, &
                                                            jprj, jprj_ppnl, l, nprj

      nprj = 0

      DO l = 0, potential%lppnl
         DO iprj_ppnl = 1, potential%nprj_ppnl(l)
            iprj = nprj+(iprj_ppnl-1)*nco(l)
            DO jprj_ppnl = 1, potential%nprj_ppnl(l)
               jprj = nprj+(jprj_ppnl-1)*nco(l)
               DO ico = 1, nco(l)
                  i = iprj+ico
                  DO jco = 1, nco(l)
                     j = jprj+jco
                     DO iso = 1, nso(l)
                        potential%vprj_ppnl(i, j) = potential%vprj_ppnl(i, j)+ &
                                                    orbtramat(l)%slm(iso, ico)* &
                                                    potential%hprj_ppnl(iprj_ppnl, &
                                                                        jprj_ppnl, l)* &
                                                    orbtramat(l)%slm(iso, jco)
                     END DO
                  END DO
               END DO
            END DO
         END DO
         nprj = nprj+potential%nprj_ppnl(l)*nco(l)
      END DO

   END SUBROUTINE init_vprj_ppnl

! **************************************************************************************************
!> \brief ...
!> \param potential ...
!> \param itype ...
!> \param zeff ...
!> \param zeff_correction ...
! **************************************************************************************************
   SUBROUTINE init_all_potential(potential, itype, zeff, zeff_correction)

      TYPE(all_potential_type), POINTER                  :: potential
      CHARACTER(LEN=*), OPTIONAL                         :: itype
      REAL(KIND=dp), OPTIONAL                            :: zeff, zeff_correction

      CHARACTER(len=*), PARAMETER :: routineN = 'init_all_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dz, handle

      IF (.NOT. ASSOCIATED(potential)) RETURN

      CALL timeset(routineN, handle)

      IF (PRESENT(zeff)) potential%zeff = zeff
      IF (PRESENT(zeff_correction)) potential%zeff_correction = zeff_correction
      dz = potential%z-INT(potential%zeff-potential%zeff_correction)
      SELECT CASE (dz)
      CASE DEFAULT
      CASE (2)
         potential%elec_conf(0) = potential%elec_conf(0)-2
      CASE (10)
         potential%elec_conf(0) = potential%elec_conf(0)-4
         potential%elec_conf(1) = potential%elec_conf(1)-6
      CASE (18)
         potential%elec_conf(0) = potential%elec_conf(0)-6
         potential%elec_conf(1) = potential%elec_conf(1)-12
      CASE (28)
         potential%elec_conf(0) = potential%elec_conf(0)-6
         potential%elec_conf(1) = potential%elec_conf(1)-12
         potential%elec_conf(2) = potential%elec_conf(2)-10
      CASE (30)
         potential%elec_conf(0) = potential%elec_conf(0)-8
         potential%elec_conf(1) = potential%elec_conf(1)-12
         potential%elec_conf(2) = potential%elec_conf(2)-10
      CASE (36)
         potential%elec_conf(0) = potential%elec_conf(0)-8
         potential%elec_conf(1) = potential%elec_conf(1)-18
         potential%elec_conf(2) = potential%elec_conf(2)-10
      CASE (46)
         potential%elec_conf(0) = potential%elec_conf(0)-8
         potential%elec_conf(1) = potential%elec_conf(1)-18
         potential%elec_conf(2) = potential%elec_conf(2)-20
      CASE (48)
         potential%elec_conf(0) = potential%elec_conf(0)-10
         potential%elec_conf(1) = potential%elec_conf(1)-18
         potential%elec_conf(2) = potential%elec_conf(2)-20
      CASE (54)
         potential%elec_conf(0) = potential%elec_conf(0)-10
         potential%elec_conf(1) = potential%elec_conf(1)-24
         potential%elec_conf(2) = potential%elec_conf(2)-20
      CASE (68)
         potential%elec_conf(0) = potential%elec_conf(0)-10
         potential%elec_conf(1) = potential%elec_conf(1)-24
         potential%elec_conf(2) = potential%elec_conf(2)-20
         potential%elec_conf(3) = potential%elec_conf(3)-14
      CASE (78)
         potential%elec_conf(0) = potential%elec_conf(0)-10
         potential%elec_conf(1) = potential%elec_conf(1)-24
         potential%elec_conf(2) = potential%elec_conf(2)-30
         potential%elec_conf(3) = potential%elec_conf(3)-14
      CASE (80)
         potential%elec_conf(0) = potential%elec_conf(0)-12
         potential%elec_conf(1) = potential%elec_conf(1)-24
         potential%elec_conf(2) = potential%elec_conf(2)-30
         potential%elec_conf(3) = potential%elec_conf(3)-14
      CASE (86)
         potential%elec_conf(0) = potential%elec_conf(0)-12
         potential%elec_conf(1) = potential%elec_conf(1)-30
         potential%elec_conf(2) = potential%elec_conf(2)-30
         potential%elec_conf(3) = potential%elec_conf(3)-14
      CASE (100)
         potential%elec_conf(0) = potential%elec_conf(0)-12
         potential%elec_conf(1) = potential%elec_conf(1)-30
         potential%elec_conf(2) = potential%elec_conf(2)-30
         potential%elec_conf(3) = potential%elec_conf(3)-28
      END SELECT

      IF (PRESENT(itype)) THEN
         IF (itype == "BARE") THEN
            potential%description(1) = "Bare Coulomb Potential"
            IF (dz > 0) THEN
               potential%description(2) = "Valence charge only"
            ELSE
               potential%description(2) = "Full atomic charge"
            END IF
         ENDIF
      END IF

      CALL timestop(handle)

   END SUBROUTINE init_all_potential

! **************************************************************************************************
!> \brief   Read an atomic all-electron potential data set.
!> \param element_symbol ...
!> \param potential_name ...
!> \param potential ...
!> \param zeff_correction ...
!> \param para_env ...
!> \param dft_section ...
!> \param potential_section ...
!> \date    14.05.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_all_potential(element_symbol, potential_name, potential, zeff_correction, &
                                 para_env, dft_section, potential_section)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol, potential_name
      TYPE(all_potential_type), POINTER                  :: potential
      REAL(KIND=dp)                                      :: zeff_correction
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: dft_section, potential_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_all_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(len=5*default_string_length)             :: line_att
      CHARACTER(len=default_path_length)                 :: potential_file_name
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(element_symbol)+2)               :: symbol2
      CHARACTER(LEN=LEN(potential_name))                 :: apname
      CHARACTER(LEN=LEN(potential_name)+2)               :: apname2
      INTEGER                                            :: irep, l, strlen1, strlen2
      INTEGER, DIMENSION(:), POINTER                     :: elec_conf
      LOGICAL                                            :: found, is_ok, match, read_from_input
      REAL(KIND=dp)                                      :: alpha, r
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(val_type), POINTER                            :: val

      line2 = ""
      symbol2 = ""
      apname2 = ""
      NULLIFY (parser)
      CALL cite_reference(Krack2000)

      potential%name = potential_name
      read_from_input = .FALSE.
      CALL section_vals_get(potential_section, explicit=read_from_input)
      IF (.NOT. read_from_input) THEN
         CALL section_vals_val_get(dft_section, "POTENTIAL_FILE_NAME", c_val=potential_file_name)
         CALL parser_create(parser, potential_file_name, para_env=para_env)
      END IF

      !   *** Search for the requested potential in the potential file   ***
      !   *** until the potential is found or the end of file is reached ***

      apname = potential_name
      symbol = element_symbol
      irep = 0
      search_loop: DO
         IF (read_from_input) THEN
            NULLIFY (list, val)
            found = .TRUE.
            CALL section_vals_list_get(potential_section, "_DEFAULT_KEYWORD_", list=list)
         ELSE
            CALL parser_search_string(parser, TRIM(apname), .TRUE., found, line)
         END IF
         IF (found) THEN
            CALL uppercase(symbol)
            CALL uppercase(apname)

            IF (read_from_input) THEN
               match = .TRUE.
            ELSE
               ! Check both the element symbol and the atomic potential name
               match = .FALSE.
               CALL uppercase(line)
               line2 = " "//line//" "
               symbol2 = " "//TRIM(symbol)//" "
               apname2 = " "//TRIM(apname)//" "
               strlen1 = LEN_TRIM(symbol2)+1
               strlen2 = LEN_TRIM(apname2)+1

               IF ((INDEX(line2, symbol2(:strlen1)) > 0) .AND. &
                   (INDEX(line2, apname2(:strlen2)) > 0)) match = .TRUE.
            END IF
            IF (match) THEN
               ! Read the electronic configuration
               NULLIFY (elec_conf)
               l = 0
               CALL reallocate(elec_conf, 0, l)
               IF (read_from_input) THEN
                  is_ok = cp_sll_val_next(list, val)
                  IF (.NOT. is_ok) &
                     CALL cp_abort(cp__l("subsys/external_potential_types.F",1115), &
                                   "Error reading the Potential from input file!!")
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) elec_conf(l)
                  CALL remove_word(line_att)
                  DO WHILE (LEN_TRIM(line_att) /= 0)
                     l = l+1
                     CALL reallocate(elec_conf, 0, l)
                     READ (line_att, *) elec_conf(l)
                     CALL remove_word(line_att)
                  END DO
               ELSE
                  CALL parser_get_object(parser, elec_conf(l), newline=.TRUE.)
                  DO WHILE (parser_test_next_token(parser) == "INT")
                     l = l+1
                     CALL reallocate(elec_conf, 0, l)
                     CALL parser_get_object(parser, elec_conf(l))
                  END DO
                  irep = irep+1
                  WRITE (line_att, '(100(1X,I0))') elec_conf
                  CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                            c_val=TRIM(line_att))
               END IF

               CALL reallocate(potential%elec_conf, 0, l)
               potential%elec_conf(:) = elec_conf(:)

               potential%zeff_correction = zeff_correction
               potential%zeff = REAL(SUM(elec_conf), dp)+zeff_correction

               DEALLOCATE (elec_conf)

               ! Read r(loc) to define the exponent of the core charge
               ! distribution and calculate the corresponding coefficient

               IF (read_from_input) THEN
                  is_ok = cp_sll_val_next(list, val)
                  IF (.NOT. is_ok) &
                     CALL cp_abort(cp__l("subsys/external_potential_types.F",1153), &
                                   "Error reading the Potential from input file!!")
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) r
               ELSE
                  CALL parser_get_object(parser, r, newline=.TRUE.)
                  irep = irep+1
                  WRITE (line_att, '(E24.16)') r
                  CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                            c_val=TRIM(line_att))
               END IF
               alpha = 1.0_dp/(2.0_dp*r**2)

               potential%alpha_core_charge = alpha
               potential%ccore_charge = potential%zeff*SQRT((alpha/pi)**3)

               EXIT search_loop
            END IF
         ELSE
            ! Stop program, if the end of file is reached
            CALL cp_abort(cp__l("subsys/external_potential_types.F",1173), &
                          "The requested atomic potential <"// &
                          TRIM(potential_name)// &
                          "> for element <"// &
                          TRIM(symbol)// &
                          "> was not found in the potential file <"// &
                          TRIM(potential_file_name)//">")
         END IF
      END DO search_loop

      IF (.NOT. read_from_input) THEN
         ! Dump the potential info the in potential section
         IF (match) THEN
            irep = irep+1
            WRITE (line_att, '(A)') "         # Potential name: "//apname2(:strlen2)//" for symbol: "//symbol2(:strlen1)
            CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                      c_val=TRIM(line_att))
            irep = irep+1
            WRITE (line_att, '(A)') "         # Potential read from the potential filename: "//TRIM(potential_file_name)
            CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                      c_val=TRIM(line_att))
         END IF
         CALL parser_release(parser)
      END IF
   END SUBROUTINE read_all_potential

! **************************************************************************************************
!> \brief   Read an atomic local potential data set.
!> \param element_symbol ...
!> \param potential_name ...
!> \param potential ...
!> \param para_env ...
!> \param dft_section ...
!> \param potential_section ...
!> \date    24.12.2014
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_local_potential(element_symbol, potential_name, potential, &
                                   para_env, dft_section, potential_section)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol, potential_name
      TYPE(local_potential_type), POINTER                :: potential
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: dft_section, potential_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_local_potential', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: eps_tpot = 1.e-10_dp

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(len=5*default_string_length)             :: line_att
      CHARACTER(len=default_path_length)                 :: potential_file_name
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(element_symbol)+2)               :: symbol2
      CHARACTER(LEN=LEN(potential_name))                 :: apname
      CHARACTER(LEN=LEN(potential_name)+2)               :: apname2
      INTEGER                                            :: igau, ipol, irep, l, ngau, npol, &
                                                            strlen1, strlen2
      LOGICAL                                            :: found, is_ok, match, read_from_input
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cval
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(val_type), POINTER                            :: val

      line2 = ""
      symbol2 = ""
      apname2 = ""
      NULLIFY (parser, alpha, cval)

      potential%name = potential_name
      read_from_input = .FALSE.
      CALL section_vals_get(potential_section, explicit=read_from_input)
      IF (.NOT. read_from_input) THEN
         CALL section_vals_val_get(dft_section, "POTENTIAL_FILE_NAME", c_val=potential_file_name)
         CALL parser_create(parser, potential_file_name, para_env=para_env)
      END IF

      !   *** Search for the requested potential in the potential file   ***
      !   *** until the potential is found or the end of file is reached ***

      apname = potential_name
      symbol = element_symbol
      irep = 0
      search_loop: DO
         IF (read_from_input) THEN
            NULLIFY (list, val)
            found = .TRUE.
            CALL section_vals_list_get(potential_section, "_DEFAULT_KEYWORD_", list=list)
         ELSE
            CALL parser_search_string(parser, TRIM(apname), .TRUE., found, line)
         END IF
         IF (found) THEN
            CALL uppercase(symbol)
            CALL uppercase(apname)

            IF (read_from_input) THEN
               match = .TRUE.
            ELSE
               ! Check both the element symbol and the atomic potential name
               match = .FALSE.
               CALL uppercase(line)
               line2 = " "//line//" "
               symbol2 = " "//TRIM(symbol)//" "
               apname2 = " "//TRIM(apname)//" "
               strlen1 = LEN_TRIM(symbol2)+1
               strlen2 = LEN_TRIM(apname2)+1

               IF ((INDEX(line2, symbol2(:strlen1)) > 0) .AND. &
                   (INDEX(line2, apname2(:strlen2)) > 0)) match = .TRUE.
            END IF
            IF (match) THEN

               ! Read ngau and npol
               IF (read_from_input) THEN
                  is_ok = cp_sll_val_next(list, val)
                  IF (.NOT. is_ok) &
                     CALL cp_abort(cp__l("subsys/external_potential_types.F",1292), &
                                   "Error reading the Potential from input file!!")
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) ngau, npol
                  CALL remove_word(line_att)
               ELSE
                  CALL parser_get_object(parser, ngau, newline=.TRUE.)
                  CALL parser_get_object(parser, npol)
                  irep = irep+1
                  WRITE (line_att, '(2(1X,I0))') ngau, npol
                  CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                            c_val=TRIM(line_att))
               END IF

               CALL reallocate(alpha, 1, ngau)
               CALL reallocate(cval, 1, ngau, 1, npol)
               DO igau = 1, ngau
                  IF (read_from_input) THEN
                     is_ok = cp_sll_val_next(list, val)
                     IF (.NOT. is_ok) &
                        CALL cp_abort(cp__l("subsys/external_potential_types.F",1312), &
                                      "Error reading the Potential from input file!!")
                     CALL val_get(val, c_val=line_att)
                     READ (line_att, *) alpha(igau), (cval(igau, ipol), ipol=1, npol)
                  ELSE
                     CALL parser_get_object(parser, alpha(igau), newline=.TRUE.)
                     DO ipol = 1, npol
                        CALL parser_get_object(parser, cval(igau, ipol), newline=.FALSE.)
                     END DO
                     irep = irep+1
                     WRITE (line_att, '(8(E24.16))') alpha(igau), (cval(igau, ipol), ipol=1, npol)
                     CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                               c_val=TRIM(line_att))
                  END IF
               END DO
               alpha = 1.0_dp/(2.0_dp*alpha**2)

               potential%ngau = ngau
               potential%npol = npol

               potential%alpha => alpha
               potential%cval => cval

               potential%radius = 0.0_dp
               DO igau = 1, ngau
                  DO ipol = 1, npol
                     l = 2*(ipol-1)
                     potential%radius = MAX(potential%radius, &
                                            exp_radius(l, alpha(igau), eps_tpot, cval(igau, ipol)))
                  END DO
               END DO

               EXIT search_loop
            END IF
         ELSE
            ! Stop program, if the end of file is reached
            CALL cp_abort(cp__l("subsys/external_potential_types.F",1348), &
                          "The requested local atomic potential <"// &
                          TRIM(potential_name)// &
                          "> for element <"// &
                          TRIM(symbol)// &
                          "> was not found in the potential file <"// &
                          TRIM(potential_file_name)//">")
         END IF
      END DO search_loop

      IF (.NOT. read_from_input) THEN
         ! Dump the potential info in the potential section
         IF (match) THEN
            irep = irep+1
            WRITE (line_att, '(A)') "         # Potential name: "//apname2(:strlen2)//" for symbol: "//symbol2(:strlen1)
            CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                      c_val=TRIM(line_att))
            irep = irep+1
            WRITE (line_att, '(A)') "         # Potential read from the potential filename: "//TRIM(potential_file_name)
            CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                      c_val=TRIM(line_att))
         END IF
         CALL parser_release(parser)
      END IF

   END SUBROUTINE read_local_potential

! **************************************************************************************************
!> \brief   Read an atomic GTH potential data set.
!> \param element_symbol ...
!> \param potential_name ...
!> \param potential ...
!> \param zeff_correction ...
!> \param para_env ...
!> \param dft_section ...
!> \param potential_section ...
!> \date    14.05.2000
!> \par  Literature
!>         - S. Goedecker, M. Teter and J. Hutter,
!>                Phys. Rev. B 54, 1703 (1996)
!>         - C. Hartwigsen, S. Goedecker and J. Hutter,
!>                Phys. Rev. B 58, 3641 (1998)
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_gth_potential(element_symbol, potential_name, potential, zeff_correction, &
                                 para_env, dft_section, potential_section)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol, potential_name
      TYPE(gth_potential_type), POINTER                  :: potential
      REAL(KIND=dp), INTENT(IN)                          :: zeff_correction
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: dft_section, potential_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_gth_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(len=5*default_string_length)             :: line_att
      CHARACTER(len=default_path_length)                 :: potential_file_name
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(element_symbol)+2)               :: symbol2
      CHARACTER(LEN=LEN(potential_name))                 :: apname
      CHARACTER(LEN=LEN(potential_name)+2)               :: apname2
      INTEGER                                            :: i, ic, ipot, irep, istr, j, l, lppnl, &
                                                            lprj_ppnl_max, maxlppl, n, nppnl, &
                                                            nprj_ppnl, nprj_ppnl_max, strlen1, &
                                                            strlen2
      INTEGER, DIMENSION(:), POINTER                     :: elec_conf
      LOGICAL                                            :: found, is_ok, match, read_from_input
      REAL(KIND=dp)                                      :: alpha, ci, r, rc2
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tmp_vals
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: hprj_ppnl
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(val_type), POINTER                            :: val

      line2 = ""
      symbol2 = ""
      apname2 = ""
      NULLIFY (parser, tmp_vals)
      CALL cite_reference(Goedecker1996)
      CALL cite_reference(Hartwigsen1998)
      CALL cite_reference(Krack2005)

      potential%name = potential_name
      read_from_input = .FALSE.
      CALL section_vals_get(potential_section, explicit=read_from_input)
      IF (.NOT. read_from_input) THEN
         CALL section_vals_val_get(dft_section, "POTENTIAL_FILE_NAME", c_val=potential_file_name)
         CALL parser_create(parser, potential_file_name, para_env=para_env)
      END IF

      !initialize extended form
      potential%lpotextended = .FALSE.
      potential%nexp_lpot = 0
      potential%lsdpot = .FALSE.
      potential%nexp_lsd = 0
      potential%nlcc = .FALSE.
      potential%nexp_nlcc = 0

      !   *** Search for the requested potential in the potential file   ***
      !   *** until the potential is found or the end of file is reached ***

      apname = potential_name
      symbol = element_symbol
      irep = 0
      search_loop: DO
         IF (read_from_input) THEN
            NULLIFY (list, val)
            found = .TRUE.
            CALL section_vals_list_get(potential_section, "_DEFAULT_KEYWORD_", list=list)
         ELSE
            CALL parser_search_string(parser, TRIM(apname), .TRUE., found, line)
         END IF
         IF (found) THEN
            CALL uppercase(symbol)
            CALL uppercase(apname)
            IF (read_from_input) THEN
               match = .TRUE.
            ELSE
               !       *** Check both the element symbol and the atomic potential name ***
               match = .FALSE.
               CALL uppercase(line)
               line2 = " "//line//" "
               symbol2 = " "//TRIM(symbol)//" "
               apname2 = " "//TRIM(apname)//" "
               strlen1 = LEN_TRIM(symbol2)+1
               strlen2 = LEN_TRIM(apname2)+1

               IF ((INDEX(line2, symbol2(:strlen1)) > 0) .AND. &
                   (INDEX(line2, apname2(:strlen2)) > 0)) match = .TRUE.
            END IF
            IF (match) THEN
               !         *** Read the electronic configuration ***
               NULLIFY (elec_conf)
               l = 0
               CALL reallocate(elec_conf, 0, l)
               IF (read_from_input) THEN
                  is_ok = cp_sll_val_next(list, val)
                  IF (.NOT. is_ok) &
                     CALL cp_abort(cp__l("subsys/external_potential_types.F",1490), &
                                   "Error reading the Potential from input file!!")
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) elec_conf(l)
                  CALL remove_word(line_att)
                  DO WHILE (LEN_TRIM(line_att) /= 0)
                     l = l+1
                     CALL reallocate(elec_conf, 0, l)
                     READ (line_att, *) elec_conf(l)
                     CALL remove_word(line_att)
                  END DO
               ELSE
                  CALL parser_get_object(parser, elec_conf(l), newline=.TRUE.)
                  DO WHILE (parser_test_next_token(parser) == "INT")
                     l = l+1
                     CALL reallocate(elec_conf, 0, l)
                     CALL parser_get_object(parser, elec_conf(l))
                  END DO
                  irep = irep+1
                  WRITE (line_att, '(100(1X,I0))') elec_conf
                  CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                            c_val=TRIM(line_att))
               END IF

               CALL reallocate(potential%elec_conf, 0, l)
               potential%elec_conf(:) = elec_conf(:)

               potential%zeff_correction = zeff_correction
               potential%zeff = REAL(SUM(elec_conf), dp)+zeff_correction

               DEALLOCATE (elec_conf)

               !         *** Read r(loc) to define the exponent of the core charge    ***
               !         *** distribution and calculate the corresponding coefficient ***

               IF (read_from_input) THEN
                  is_ok = cp_sll_val_next(list, val)
                  IF (.NOT. is_ok) &
                     CALL cp_abort(cp__l("subsys/external_potential_types.F",1528), &
                                   "Error reading the Potential from input file!!")
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) r
                  CALL remove_word(line_att)
               ELSE
                  line_att = ""
                  CALL parser_get_object(parser, r, newline=.TRUE.)
                  istr = LEN_TRIM(line_att)+1
                  WRITE (line_att(istr:), '(E24.16)') r
               END IF
               alpha = 1.0_dp/(2.0_dp*r**2)

               potential%alpha_core_charge = alpha
               potential%ccore_charge = potential%zeff*SQRT((alpha/pi)**3)

               potential%alpha_ppl = alpha
               potential%cerf_ppl = potential%zeff*SQRT((alpha/pi)**3)

               !         *** Read the parameters for the local part ***
               !         *** of the GTH pseudopotential (ppl)       ***

               IF (read_from_input) THEN
                  READ (line_att, *) n
                  CALL remove_word(line_att)
               ELSE
                  CALL parser_get_object(parser, n)
                  istr = LEN_TRIM(line_att)+1
                  WRITE (line_att(istr:), '(1X,I0)') n
               END IF
               potential%nexp_ppl = n
               CALL reallocate(potential%cexp_ppl, 1, n)

               DO i = 1, n
                  IF (read_from_input) THEN
                     READ (line_att, *) ci
                     CALL remove_word(line_att)
                  ELSE
                     CALL parser_get_object(parser, ci)
                     istr = LEN_TRIM(line_att)+1
                     WRITE (line_att(istr:), '(E24.16)') ci
                  END IF
                  rc2 = (2.0_dp*potential%alpha_ppl)
                  potential%cexp_ppl(i) = rc2**(i-1)*ci
               END DO

               IF (.NOT. read_from_input) THEN
                  irep = irep+1
                  CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                            c_val=TRIM(line_att))
                  line_att = ""
               ELSE
                  IF (LEN_TRIM(line_att) /= 0) THEN
                     CALL cp_abort(cp__l("subsys/external_potential_types.F",1581), &
                                   "Error reading the Potential from input file!!")
                  END IF
               END IF
               maxlppl = 2*(n-1)

               IF (maxlppl > -1) CALL init_orbital_pointers(maxlppl)

               ! Read extended form of GTH pseudopotential
               ! local potential, NLCC, LSD potential
               IF (read_from_input) THEN
                  DO
                     is_ok = cp_sll_val_next(list, val)
                     IF(.NOT.(is_ok))CALL cp__a("subsys/external_potential_types.F",1594)
                     CALL val_get(val, c_val=line_att)
                     IF (INDEX(line_att, "LPOT") /= 0) THEN
                        potential%lpotextended = .TRUE.
                        CALL remove_word(line_att)
                        READ (line_att, *) potential%nexp_lpot
                        n = potential%nexp_lpot
                        maxlppl = 2*(n-1)
                        IF (maxlppl > -1) CALL init_orbital_pointers(maxlppl)
                        NULLIFY (potential%alpha_lpot, potential%nct_lpot, potential%cval_lpot)
                        CALL reallocate(potential%alpha_lpot, 1, n)
                        CALL reallocate(potential%nct_lpot, 1, n)
                        CALL reallocate(potential%cval_lpot, 1, 4, 1, n)
                        DO ipot = 1, potential%nexp_lpot
                           is_ok = cp_sll_val_next(list, val)
                           IF(.NOT.(is_ok))CALL cp__a("subsys/external_potential_types.F",1609)
                           CALL val_get(val, c_val=line_att)
                           READ (line_att, *) r
                           potential%alpha_lpot(ipot) = 0.5_dp/(r*r)
                           CALL remove_word(line_att)
                           READ (line_att, *) potential%nct_lpot(ipot)
                           CALL remove_word(line_att)
                           DO ic = 1, potential%nct_lpot(ipot)
                              READ (line_att, *) ci
                              rc2 = (2._dp*potential%alpha_lpot(ipot))**(ic-1)
                              potential%cval_lpot(ic, ipot) = ci*rc2
                              CALL remove_word(line_att)
                           END DO
                        END DO
                     ELSEIF (INDEX(line_att, "NLCC") /= 0) THEN
                        potential%nlcc = .TRUE.
                        CALL remove_word(line_att)
                        READ (line_att, *) potential%nexp_nlcc
                        n = potential%nexp_nlcc
                        NULLIFY (potential%alpha_nlcc, potential%nct_nlcc, potential%cval_nlcc)
                        CALL reallocate(potential%alpha_nlcc, 1, n)
                        CALL reallocate(potential%nct_nlcc, 1, n)
                        CALL reallocate(potential%cval_nlcc, 1, 4, 1, n)
                        DO ipot = 1, potential%nexp_nlcc
                           is_ok = cp_sll_val_next(list, val)
                           IF(.NOT.(is_ok))CALL cp__a("subsys/external_potential_types.F",1634)
                           CALL val_get(val, c_val=line_att)
                           READ (line_att, *) potential%alpha_nlcc(ipot)
                           CALL remove_word(line_att)
                           READ (line_att, *) potential%nct_nlcc(ipot)
                           CALL remove_word(line_att)
                           DO ic = 1, potential%nct_nlcc(ipot)
                              READ (line_att, *) potential%cval_nlcc(ic, ipot)
                              !make it compatible with bigdft style
                              potential%cval_nlcc(ic, ipot) = potential%cval_nlcc(ic, ipot)/(4.0_dp*pi)
                              CALL remove_word(line_att)
                           END DO
                        END DO
                     ELSEIF (INDEX(line_att, "LSD") /= 0) THEN
                        potential%lsdpot = .TRUE.
                        CALL remove_word(line_att)
                        READ (line_att, *) potential%nexp_lsd
                        n = potential%nexp_lsd
                        NULLIFY (potential%alpha_lsd, potential%nct_lsd, potential%cval_lsd)
                        CALL reallocate(potential%alpha_lsd, 1, n)
                        CALL reallocate(potential%nct_lsd, 1, n)
                        CALL reallocate(potential%cval_lsd, 1, 4, 1, n)
                        DO ipot = 1, potential%nexp_lsd
                           is_ok = cp_sll_val_next(list, val)
                           IF(.NOT.(is_ok))CALL cp__a("subsys/external_potential_types.F",1658)
                           CALL val_get(val, c_val=line_att)
                           READ (line_att, *) r
                           potential%alpha_lsd(ipot) = 0.5_dp/(r*r)
                           CALL remove_word(line_att)
                           READ (line_att, *) potential%nct_lsd(ipot)
                           CALL remove_word(line_att)
                           DO ic = 1, potential%nct_lsd(ipot)
                              READ (line_att, *) ci
                              rc2 = (2._dp*potential%alpha_lsd(ipot))**(ic-1)
                              potential%cval_lsd(ic, ipot) = ci*rc2
                              CALL remove_word(line_att)
                           END DO
                        END DO
                     ELSE
                        EXIT
                     END IF
                  END DO
               ELSE
                  DO
                     CALL parser_get_next_line(parser, 1)
                     IF (parser_test_next_token(parser) == "INT") THEN
                        EXIT
                     ELSEIF (parser_test_next_token(parser) == "STR") THEN
                        CALL parser_get_object(parser, line)
                        IF (INDEX(LINE, "LPOT") /= 0) THEN
                           ! local potential
                           potential%lpotextended = .TRUE.
                           CALL parser_get_object(parser, potential%nexp_lpot)
                           n = potential%nexp_lpot
                           NULLIFY (potential%alpha_lpot, potential%nct_lpot, potential%cval_lpot)
                           CALL reallocate(potential%alpha_lpot, 1, n)
                           CALL reallocate(potential%nct_lpot, 1, n)
                           CALL reallocate(potential%cval_lpot, 1, 4, 1, n)
                           ! add to input section
                           WRITE (line_att, '(A,1X,I0)') "LPOT", n
                           irep = irep+1
                           CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                     c_val=TRIM(line_att))
                           DO ipot = 1, potential%nexp_lpot
                              CALL parser_get_object(parser, r, newline=.TRUE.)
                              potential%alpha_lpot(ipot) = 0.5_dp/(r*r)
                              CALL parser_get_object(parser, potential%nct_lpot(ipot))
                              CALL reallocate(tmp_vals, 1, potential%nct_lpot(ipot))
                              DO ic = 1, potential%nct_lpot(ipot)
                                 CALL parser_get_object(parser, ci)
                                 tmp_vals(ic) = ci
                                 rc2 = (2._dp*potential%alpha_lpot(ipot))**(ic-1)
                                 potential%cval_lpot(ic, ipot) = ci*rc2
                              END DO
                              ! add to input section
                              irep = irep+1
                              WRITE (line_att, '(E24.16,1X,I0,100(1X,E24.16))') r, potential%nct_lpot(ipot), &
                                 tmp_vals(1:potential%nct_lpot(ipot))
                              CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                        c_val=TRIM(line_att))
                           END DO
                        ELSEIF (INDEX(LINE, "NLCC") /= 0) THEN
                           ! NLCC
                           potential%nlcc = .TRUE.
                           CALL parser_get_object(parser, potential%nexp_nlcc)
                           n = potential%nexp_nlcc
                           NULLIFY (potential%alpha_nlcc, potential%nct_nlcc, potential%cval_nlcc)
                           CALL reallocate(potential%alpha_nlcc, 1, n)
                           CALL reallocate(potential%nct_nlcc, 1, n)
                           CALL reallocate(potential%cval_nlcc, 1, 4, 1, n)
                           ! add to input section
                           WRITE (line_att, '(A,1X,I0)') "NLCC", n
                           irep = irep+1
                           CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                     c_val=TRIM(line_att))
                           DO ipot = 1, potential%nexp_nlcc
                              CALL parser_get_object(parser, potential%alpha_nlcc(ipot), newline=.TRUE.)
                              CALL parser_get_object(parser, potential%nct_nlcc(ipot))
                              CALL reallocate(tmp_vals, 1, potential%nct_nlcc(ipot))
                              DO ic = 1, potential%nct_nlcc(ipot)
                                 CALL parser_get_object(parser, potential%cval_nlcc(ic, ipot))
                                 tmp_vals(ic) = potential%cval_nlcc(ic, ipot)
                                 !make it compatible with bigdft style
                                 potential%cval_nlcc(ic, ipot) = potential%cval_nlcc(ic, ipot)/(4.0_dp*pi)
                              END DO
                              ! add to input section
                              irep = irep+1
                              WRITE (line_att, '(E24.16,1X,I0,100(1X,E24.16))') &
                                 potential%alpha_nlcc(ipot), potential%nct_nlcc(ipot), &
                                 tmp_vals(1:potential%nct_nlcc(ipot))
                              CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                        c_val=TRIM(line_att))
                           END DO
                        ELSEIF (INDEX(LINE, "LSD") /= 0) THEN
                           ! LSD potential
                           potential%lsdpot = .TRUE.
                           CALL parser_get_object(parser, potential%nexp_lsd)
                           n = potential%nexp_lsd
                           NULLIFY (potential%alpha_lsd, potential%nct_lsd, potential%cval_lsd)
                           CALL reallocate(potential%alpha_lsd, 1, n)
                           CALL reallocate(potential%nct_lsd, 1, n)
                           CALL reallocate(potential%cval_lsd, 1, 4, 1, n)
                           ! add to input section
                           WRITE (line_att, '(A,1X,I0)') "LSD", n
                           irep = irep+1
                           CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                     c_val=TRIM(line_att))
                           DO ipot = 1, potential%nexp_lsd
                              CALL parser_get_object(parser, r, newline=.TRUE.)
                              potential%alpha_lsd(ipot) = 0.5_dp/(r*r)
                              CALL parser_get_object(parser, potential%nct_lsd(ipot))
                              CALL reallocate(tmp_vals, 1, potential%nct_lsd(ipot))
                              DO ic = 1, potential%nct_lsd(ipot)
                                 CALL parser_get_object(parser, ci)
                                 tmp_vals(ic) = ci
                                 rc2 = (2._dp*potential%alpha_lsd(ipot))**(ic-1)
                                 potential%cval_lsd(ic, ipot) = ci*rc2
                              END DO
                              ! add to input section
                              irep = irep+1
                              WRITE (line_att, '(E24.16,1X,I0,100(1X,E24.16))') r, potential%nct_lsd(ipot), &
                                 tmp_vals(1:potential%nct_lsd(ipot))
                              CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                        c_val=TRIM(line_att))
                           END DO
                        ELSE
                           CALL cp__b("subsys/external_potential_types.F",1780,"")
                        END IF
                     ELSE
                        CALL cp__b("subsys/external_potential_types.F",1783,"")
                     END IF
                  END DO
               END IF
               !         *** Read the parameters for the non-local  ***
               !         *** part of the GTH pseudopotential (ppnl) ***
               IF (read_from_input) THEN
                  READ (line_att, *) n
                  CALL remove_word(line_att)
               ELSE
                  CALL parser_get_object(parser, n)
                  irep = irep+1
                  WRITE (line_att, '(1X,I0)') n
                  CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                            c_val=TRIM(line_att))
               END IF
               potential%lppnl = n-1
               potential%nppnl = 0

               potential%lprj_ppnl_max = n-1
               potential%nprj_ppnl_max = 0

               IF (n > 0) THEN

                  lppnl = potential%lppnl
                  nppnl = potential%nppnl

                  CALL init_orbital_pointers(lppnl)

                  NULLIFY (hprj_ppnl)

                  ! Load the parameter for n non-local projectors

                  CALL reallocate(potential%alpha_ppnl, 0, lppnl)
                  CALL reallocate(potential%nprj_ppnl, 0, lppnl)

                  lprj_ppnl_max = -1
                  nprj_ppnl_max = 0

                  DO l = 0, lppnl
                     IF (read_from_input) THEN
                        is_ok = cp_sll_val_next(list, val)
                        IF (.NOT. is_ok) &
                           CALL cp_abort(cp__l("subsys/external_potential_types.F",1826), &
                                         "Error reading the Potential from input file!!")
                        CALL val_get(val, c_val=line_att)
                        READ (line_att, *) r
                        CALL remove_word(line_att)
                        READ (line_att, *) nprj_ppnl
                        CALL remove_word(line_att)
                     ELSE
                        line_att = ""
                        CALL parser_get_object(parser, r, newline=.TRUE.)
                        CALL parser_get_object(parser, nprj_ppnl)
                        istr = LEN_TRIM(line_att)+1
                        WRITE (line_att(istr:), '(E24.16,1X,I0)') r, nprj_ppnl
                     END IF
                     IF (r == 0.0_dp .AND. nprj_ppnl /= 0) THEN
                        CALL cp_abort(cp__l("subsys/external_potential_types.F",1841), &
                                      "An error was detected in the atomic potential <"// &
                                      TRIM(potential_name)// &
                                      "> potential file <"// &
                                      TRIM(potential_file_name)//">")
                     END IF
                     potential%alpha_ppnl(l) = 0.0_dp
                     IF (r /= 0.0_dp .AND. n /= 0) potential%alpha_ppnl(l) = 1.0_dp/(2.0_dp*r**2)
                     potential%nprj_ppnl(l) = nprj_ppnl
                     nppnl = nppnl+nprj_ppnl*nco(l)
                     IF (nprj_ppnl > nprj_ppnl_max) THEN
                        nprj_ppnl_max = nprj_ppnl
                        CALL reallocate(hprj_ppnl, 1, nprj_ppnl_max, &
                                        1, nprj_ppnl_max, &
                                        0, lppnl)
                     END IF
                     DO i = 1, nprj_ppnl
                        IF (i == 1) THEN
                           IF (read_from_input) THEN
                              READ (line_att, *) hprj_ppnl(i, i, l)
                              CALL remove_word(line_att)
                           ELSE
                              CALL parser_get_object(parser, hprj_ppnl(i, i, l))
                              istr = LEN_TRIM(line_att)+1
                              WRITE (line_att(istr:), '(E24.16)') hprj_ppnl(i, i, l)
                           END IF
                        ELSE
                           IF (read_from_input) THEN
                              IF (LEN_TRIM(line_att) /= 0) &
                                 CALL cp_abort(cp__l("subsys/external_potential_types.F",1870), &
                                               "Error reading the Potential from input file!!")
                              is_ok = cp_sll_val_next(list, val)
                              IF (.NOT. is_ok) &
                                 CALL cp_abort(cp__l("subsys/external_potential_types.F",1874), &
                                               "Error reading the Potential from input file!!")
                              CALL val_get(val, c_val=line_att)
                              READ (line_att, *) hprj_ppnl(i, i, l)
                              CALL remove_word(line_att)
                           ELSE
                              irep = irep+1
                              CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                        c_val=TRIM(line_att))
                              line_att = ""
                              CALL parser_get_object(parser, hprj_ppnl(i, i, l), newline=.TRUE.)
                              istr = LEN_TRIM(line_att)+1
                              WRITE (line_att(istr:), '(E24.16)') hprj_ppnl(i, i, l)
                           END IF
                        END IF
                        DO j = i+1, nprj_ppnl
                           IF (read_from_input) THEN
                              READ (line_att, *) hprj_ppnl(i, j, l)
                              CALL remove_word(line_att)
                           ELSE
                              CALL parser_get_object(parser, hprj_ppnl(i, j, l))
                              istr = LEN_TRIM(line_att)+1
                              WRITE (line_att(istr:), '(E24.16)') hprj_ppnl(i, j, l)
                           END IF
                        END DO
                     END DO
                     IF (.NOT. read_from_input) THEN
                        irep = irep+1
                        CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                                  c_val=TRIM(line_att))
                        line_att = ""
                     ELSE
                        IF (LEN_TRIM(line_att) /= 0) THEN
                           CALL cp_abort(cp__l("subsys/external_potential_types.F",1907), &
                                         "Error reading the Potential from input file!!")
                        END IF
                     END IF
                     IF (nprj_ppnl > 1) THEN
                        CALL symmetrize_matrix(hprj_ppnl(:, :, l), "upper_to_lower")
                     END IF
                     lprj_ppnl_max = MAX(lprj_ppnl_max, l+2*(nprj_ppnl-1))
                  END DO

                  potential%nppnl = nppnl
                  CALL init_orbital_pointers(lprj_ppnl_max)

                  potential%lprj_ppnl_max = lprj_ppnl_max
                  potential%nprj_ppnl_max = nprj_ppnl_max
                  CALL reallocate(potential%hprj_ppnl, 1, nprj_ppnl_max, &
                                  1, nprj_ppnl_max, &
                                  0, lppnl)
                  potential%hprj_ppnl(:, :, :) = hprj_ppnl(:, :, :)

                  CALL reallocate(potential%cprj, 1, ncoset(lprj_ppnl_max), 1, nppnl)
                  CALL reallocate(potential%cprj_ppnl, 1, nprj_ppnl_max, 0, lppnl)
                  CALL reallocate(potential%vprj_ppnl, 1, nppnl, 1, nppnl)

                  DEALLOCATE (hprj_ppnl)
               END IF
               EXIT search_loop
            END IF
         ELSE
            ! Stop program, if the end of file is reached
            CALL cp_abort(cp__l("subsys/external_potential_types.F",1937), &
                          "The requested atomic potential <"// &
                          TRIM(potential_name)// &
                          "> for element <"// &
                          TRIM(symbol)// &
                          "> was not found in the potential file <"// &
                          TRIM(potential_file_name)//">")
         END IF
      END DO search_loop

      IF (.NOT. read_from_input) THEN
         ! Dump the potential info the in potential section
         IF (match) THEN
            irep = irep+1
            WRITE (line_att, '(A)') "         # Potential name: "//apname2(:strlen2)//" for symbol: "//symbol2(:strlen1)
            CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                      c_val=TRIM(line_att))
            irep = irep+1
            WRITE (line_att, '(A)') "         # Potential read from the potential filename: "//TRIM(potential_file_name)
            CALL section_vals_val_set(potential_section, "_DEFAULT_KEYWORD_", i_rep_val=irep, &
                                      c_val=TRIM(line_att))
         END IF
         CALL parser_release(parser)
      END IF

      IF (ASSOCIATED(tmp_vals)) DEALLOCATE (tmp_vals)

   END SUBROUTINE read_gth_potential

! **************************************************************************************************
!> \brief ...
!> \param potential ...
!> \param z ...
!> \param zeff_correction ...
! **************************************************************************************************
   SUBROUTINE set_default_all_potential(potential, z, zeff_correction)

      TYPE(all_potential_type), POINTER                  :: potential
      INTEGER, INTENT(IN)                                :: z
      REAL(KIND=dp)                                      :: zeff_correction

      CHARACTER(len=*), PARAMETER :: routineN = 'set_default_all_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name
      INTEGER, DIMENSION(:), POINTER                     :: elec_conf
      REAL(KIND=dp)                                      :: alpha, alpha_core_charge, ccore_charge, &
                                                            core_charge_radius, r, zeff

      ALLOCATE (elec_conf(0:3))
      elec_conf(0:3) = ptable(z)%e_conv(0:3)
      zeff = REAL(SUM(elec_conf), dp)+zeff_correction
      name = ptable(z)%name

      r = ptable(z)%covalent_radius*0.5_dp
      r = MAX(r, 0.2_dp)
      r = MIN(r, 1.0_dp)
      alpha = 1.0_dp/(2.0_dp*r**2)

      core_charge_radius = r
      alpha_core_charge = alpha
      ccore_charge = zeff*SQRT((alpha/pi)**3)

      CALL set_all_potential(potential, &
                             name=name, &
                             alpha_core_charge=alpha_core_charge, &
                             ccore_charge=ccore_charge, &
                             core_charge_radius=core_charge_radius, &
                             z=z, &
                             zeff=zeff, &
                             zeff_correction=zeff_correction, &
                             elec_conf=elec_conf)

      DEALLOCATE (elec_conf)

   END SUBROUTINE set_default_all_potential

! **************************************************************************************************
!> \brief   Set the attributes of an all-electron potential data set.
!> \param potential ...
!> \param name ...
!> \param alpha_core_charge ...
!> \param ccore_charge ...
!> \param core_charge_radius ...
!> \param z ...
!> \param zeff ...
!> \param zeff_correction ...
!> \param elec_conf ...
!> \date    11.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_all_potential(potential, name, alpha_core_charge, &
                                ccore_charge, core_charge_radius, z, zeff, &
                                zeff_correction, elec_conf)

      TYPE(all_potential_type), POINTER                  :: potential
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: name
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: alpha_core_charge, ccore_charge, &
                                                            core_charge_radius
      INTEGER, INTENT(IN), OPTIONAL                      :: z
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: zeff, zeff_correction
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: elec_conf

      CHARACTER(len=*), PARAMETER :: routineN = 'set_all_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(name)) potential%name = name
         IF (PRESENT(alpha_core_charge)) &
            potential%alpha_core_charge = alpha_core_charge
         IF (PRESENT(ccore_charge)) potential%ccore_charge = ccore_charge
         IF (PRESENT(core_charge_radius)) &
            potential%core_charge_radius = core_charge_radius
         IF (PRESENT(z)) potential%z = z
         IF (PRESENT(zeff)) potential%zeff = zeff
         IF (PRESENT(zeff_correction)) potential%zeff_correction = zeff_correction
         IF (PRESENT(elec_conf)) THEN
            IF (.NOT. ASSOCIATED(potential%elec_conf)) THEN
               CALL reallocate(potential%elec_conf, 0, SIZE(elec_conf)-1)
            END IF
            potential%elec_conf(:) = elec_conf(:)
         END IF

      ELSE

         CALL cp__b("subsys/external_potential_types.F",2065,"The pointer potential is not associated")

      END IF

   END SUBROUTINE set_all_potential

! **************************************************************************************************
!> \brief   Set the attributes of an atomic local potential data set.
!> \param potential ...
!> \param name ...
!> \param alpha ...
!> \param cval ...
!> \param radius ...
!> \date    24.01.2014
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_local_potential(potential, name, alpha, cval, radius)

      TYPE(local_potential_type), POINTER                :: potential
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: name
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: alpha
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cval
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: radius

      CHARACTER(len=*), PARAMETER :: routineN = 'set_local_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(name)) potential%name = name
         IF (PRESENT(alpha)) potential%alpha => alpha
         IF (PRESENT(cval)) potential%cval => cval
         IF (PRESENT(radius)) potential%radius = radius

      ELSE

         CALL cp__b("subsys/external_potential_types.F",2103,"The pointer potential is not associated")

      END IF

   END SUBROUTINE set_local_potential

! **************************************************************************************************
!> \brief   Set the attributes of an effective charge and inducible point
!>          dipole potential data set.
!> \param potential ...
!> \param apol ...
!> \param cpol ...
!> \param qeff ...
!> \param mm_radius ...
!> \param qmmm_corr_radius ...
!> \param qmmm_radius ...
!> \date    05.03.2010
!> \author  Toon.Verstraelen@gmail.com
! **************************************************************************************************
   SUBROUTINE set_fist_potential(potential, apol, cpol, qeff, mm_radius, &
                                 qmmm_corr_radius, qmmm_radius)
      TYPE(fist_potential_type), POINTER                 :: potential
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: apol, cpol, qeff, mm_radius, &
                                                            qmmm_corr_radius, qmmm_radius

      CHARACTER(len=*), PARAMETER :: routineN = 'set_fist_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(apol)) potential%apol = apol
         IF (PRESENT(cpol)) potential%cpol = cpol
         IF (PRESENT(mm_radius)) potential%mm_radius = mm_radius
         IF (PRESENT(qeff)) potential%qeff = qeff
         IF (PRESENT(qmmm_corr_radius)) potential%qmmm_corr_radius = qmmm_corr_radius
         IF (PRESENT(qmmm_radius)) potential%qmmm_radius = qmmm_radius

      ELSE

         CALL cp__b("subsys/external_potential_types.F",2142,"The pointer potential is not associated")

      END IF

   END SUBROUTINE set_fist_potential

! **************************************************************************************************
!> \brief   Set the attributes of a GTH potential data set.
!> \param potential ...
!> \param name ...
!> \param alpha_core_charge ...
!> \param alpha_ppl ...
!> \param ccore_charge ...
!> \param cerf_ppl ...
!> \param core_charge_radius ...
!> \param ppl_radius ...
!> \param ppnl_radius ...
!> \param lppnl ...
!> \param lprj_ppnl_max ...
!> \param nexp_ppl ...
!> \param nppnl ...
!> \param nprj_ppnl_max ...
!> \param z ...
!> \param zeff ...
!> \param zeff_correction ...
!> \param alpha_ppnl ...
!> \param cexp_ppl ...
!> \param elec_conf ...
!> \param nprj_ppnl ...
!> \param cprj ...
!> \param cprj_ppnl ...
!> \param vprj_ppnl ...
!> \param hprj_ppnl ...
!> \date    11.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_gth_potential(potential, name, alpha_core_charge, alpha_ppl, &
                                ccore_charge, cerf_ppl, core_charge_radius, &
                                ppl_radius, ppnl_radius, lppnl, lprj_ppnl_max, &
                                nexp_ppl, nppnl, nprj_ppnl_max, z, zeff, zeff_correction, &
                                alpha_ppnl, cexp_ppl, elec_conf, nprj_ppnl, cprj, cprj_ppnl, &
                                vprj_ppnl, hprj_ppnl)
      TYPE(gth_potential_type), POINTER                  :: potential
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: name
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: alpha_core_charge, alpha_ppl, &
                                                            ccore_charge, cerf_ppl, &
                                                            core_charge_radius, ppl_radius, &
                                                            ppnl_radius
      INTEGER, INTENT(IN), OPTIONAL                      :: lppnl, lprj_ppnl_max, nexp_ppl, nppnl, &
                                                            nprj_ppnl_max, z
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: zeff, zeff_correction
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: alpha_ppnl, cexp_ppl
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: elec_conf, nprj_ppnl
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: cprj, cprj_ppnl, vprj_ppnl
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: hprj_ppnl

      CHARACTER(len=*), PARAMETER :: routineN = 'set_gth_potential', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(potential)) THEN

         IF (PRESENT(name)) potential%name = name
         IF (PRESENT(alpha_core_charge)) &
            potential%alpha_core_charge = alpha_core_charge
         IF (PRESENT(alpha_ppl)) potential%alpha_ppl = alpha_ppl
         IF (PRESENT(ccore_charge)) potential%ccore_charge = ccore_charge
         IF (PRESENT(cerf_ppl)) potential%cerf_ppl = cerf_ppl
         IF (PRESENT(core_charge_radius)) &
            potential%core_charge_radius = core_charge_radius
         IF (PRESENT(ppl_radius)) potential%ppl_radius = ppl_radius
         IF (PRESENT(ppnl_radius)) potential%ppnl_radius = ppnl_radius
         IF (PRESENT(lppnl)) potential%lppnl = lppnl
         IF (PRESENT(lprj_ppnl_max)) potential%lprj_ppnl_max = lprj_ppnl_max
         IF (PRESENT(nexp_ppl)) potential%nexp_ppl = nexp_ppl
         IF (PRESENT(nppnl)) potential%nppnl = nppnl
         IF (PRESENT(nprj_ppnl_max)) potential%nprj_ppnl_max = nprj_ppnl_max
         IF (PRESENT(z)) potential%z = z
         IF (PRESENT(zeff)) potential%zeff = zeff
         IF (PRESENT(zeff_correction)) potential%zeff_correction = zeff_correction
         IF (PRESENT(alpha_ppnl)) potential%alpha_ppnl => alpha_ppnl
         IF (PRESENT(cexp_ppl)) potential%cexp_ppl => cexp_ppl
         IF (PRESENT(elec_conf)) potential%elec_conf => elec_conf
         IF (PRESENT(nprj_ppnl)) potential%nprj_ppnl => nprj_ppnl
         IF (PRESENT(cprj)) potential%cprj => cprj
         IF (PRESENT(cprj_ppnl)) potential%cprj_ppnl => cprj_ppnl
         IF (PRESENT(vprj_ppnl)) potential%vprj_ppnl => vprj_ppnl
         IF (PRESENT(hprj_ppnl)) potential%hprj_ppnl => hprj_ppnl

      ELSE

         CALL cp__b("subsys/external_potential_types.F",2235,"The pointer potential is not associated")

      END IF

   END SUBROUTINE set_gth_potential

! **************************************************************************************************
!> \brief ...
!> \param potential ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE write_all_potential(potential, output_unit)

      ! Write an atomic all-electron potential data set to the output unit.

      ! - Creation (09.02.2002,MK)

      TYPE(all_potential_type), POINTER                  :: potential
      INTEGER, INTENT(in)                                :: output_unit

      CHARACTER(LEN=20)                                  :: string

      IF (output_unit > 0 .AND. ASSOCIATED(potential)) THEN
         WRITE (UNIT=output_unit, FMT="(/,T6,A,T41,A40,/)") &
            "Potential information for", ADJUSTR(TRIM(potential%name))
         WRITE (UNIT=output_unit, FMT="(T8,A,T41,A40)") &
            "Description: ", TRIM(potential%description(1)), &
            "             ", TRIM(potential%description(2))
         WRITE (UNIT=output_unit, FMT="(/,T8,A,T69,F12.6)") &
            "Gaussian exponent of the core charge distribution: ", &
            potential%alpha_core_charge
         WRITE (UNIT=string, FMT="(5I4)") potential%elec_conf
         WRITE (UNIT=output_unit, FMT="(T8,A,T61,A20)") &
            "Electronic configuration (s p d ...):", &
            ADJUSTR(TRIM(string))
      END IF

   END SUBROUTINE write_all_potential

! **************************************************************************************************
!> \brief ...
!> \param potential ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE write_local_potential(potential, output_unit)

      ! Write an atomic local potential data set to the output unit.

      ! - Creation (24.01.2014,JGH)

      TYPE(local_potential_type), POINTER                :: potential
      INTEGER, INTENT(in)                                :: output_unit

      INTEGER                                            :: igau, ipol

      IF (output_unit > 0 .AND. ASSOCIATED(potential)) THEN
         WRITE (UNIT=output_unit, FMT="(/,T6,A,T41,A40)") &
            "Potential information for", ADJUSTR(TRIM(potential%name))
         WRITE (UNIT=output_unit, FMT="(T8,A,T41,A40)") &
            "Description: ", TRIM(potential%description(1))
         DO igau = 1, potential%ngau
            WRITE (UNIT=output_unit, FMT="(T8,A,F12.6,T50,A,4(T68,I2,F10.4))") &
               "Exponent: ", potential%alpha(igau), &
               "Coefficients: ", (2*ipol-2, potential%cval(igau, ipol), ipol=1, potential%npol)
         END DO
      END IF

   END SUBROUTINE write_local_potential

! **************************************************************************************************
!> \brief ...
!> \param potential ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE write_gth_potential(potential, output_unit)

      ! Write an atomic GTH potential data set to the output unit.
      ! - Creation (09.02.2002,MK)

      TYPE(gth_potential_type), POINTER                  :: potential
      INTEGER, INTENT(in)                                :: output_unit

      CHARACTER(LEN=20)                                  :: string
      INTEGER                                            :: i, j, l
      REAL(KIND=dp)                                      :: r

      IF (output_unit > 0 .AND. ASSOCIATED(potential)) THEN
         WRITE (UNIT=output_unit, FMT="(/,T6,A,T41,A40,/)") &
            "Potential information for", ADJUSTR(TRIM(potential%name))
         WRITE (UNIT=output_unit, FMT="(T8,A,T41,A40)") &
            "Description: ", ADJUSTR(TRIM(potential%description(1))), &
            "             ", ADJUSTR(TRIM(potential%description(2))), &
            "             ", ADJUSTR(TRIM(potential%description(3))), &
            "             ", ADJUSTR(TRIM(potential%description(4)))
         WRITE (UNIT=output_unit, FMT="(/,T8,A,T69,F12.6)") &
            "Gaussian exponent of the core charge distribution: ", &
            potential%alpha_core_charge
         WRITE (UNIT=string, FMT="(5I4)") potential%elec_conf
         WRITE (UNIT=output_unit, FMT="(T8,A,T61,A20)") &
            "Electronic configuration (s p d ...):", &
            ADJUSTR(TRIM(string))

         r = 1.0_dp/SQRT(2.0_dp*potential%alpha_ppl)

         WRITE (UNIT=output_unit, FMT="(/,T8,A,/,/,T27,A,/,T21,5F12.6)") &
            "Parameters of the local part of the GTH pseudopotential:", &
            "rloc        C1          C2          C3          C4", &
            r, (potential%cexp_ppl(i)*r**(2*(i-1)), i=1, potential%nexp_ppl)

         IF (potential%lppnl > -1) THEN
            WRITE (UNIT=output_unit, FMT="(/,T8,A,/,/,T20,A,/)") &
               "Parameters of the non-local part of the GTH pseudopotential:", &
               "l      r(l)      h(i,j,l)"
            DO l = 0, potential%lppnl
               r = SQRT(0.5_dp/potential%alpha_ppnl(l))
               WRITE (UNIT=output_unit, FMT="(T19,I2,5F12.6)") &
                  l, r, (potential%hprj_ppnl(1, j, l), j=1, potential%nprj_ppnl(l))
               DO i = 2, potential%nprj_ppnl(l)
                  WRITE (UNIT=output_unit, FMT="(T33,4F12.6)") &
                     (potential%hprj_ppnl(i, j, l), j=1, potential%nprj_ppnl(l))
               END DO
            END DO
         END IF
      END IF

   END SUBROUTINE write_gth_potential

! **************************************************************************************************

END MODULE external_potential_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author MI (12.01.2007)
! **************************************************************************************************
MODULE shell_potential_types

   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'shell_potential_types'

! **************************************************************************************************
!> \brief Define the shell type
! **************************************************************************************************
   TYPE shell_kind_type
      INTEGER                                :: ref_count
      REAL(dp)                               :: charge_core, &
                                                charge_shell
      REAL(dp)                               :: mass_core, &
                                                massfrac, &
                                                mass_shell
      REAL(dp)                               :: k2_spring, k4_spring
      REAL(dp)                               :: max_dist
      REAL(dp)                               :: shell_cutoff
   END TYPE shell_kind_type

! **************************************************************************************************
   TYPE shell_p_type
      CHARACTER(LEN=default_string_length)   :: atm_name
      TYPE(shell_kind_type), POINTER          :: shell
   END TYPE shell_p_type

! Public subroutines

   PUBLIC :: get_shell, shell_create, shell_p_create, &
             shell_p_release, shell_release, shell_retain

! Public data types

   PUBLIC :: shell_p_type, shell_kind_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param shell ...
!> \param charge ...
!> \param charge_core ...
!> \param charge_shell ...
!> \param mass_core ...
!> \param mass_shell ...
!> \param k2_spring ...
!> \param k4_spring ...
!> \param max_dist ...
!> \param shell_cutoff ...
! **************************************************************************************************
   SUBROUTINE get_shell(shell, charge, charge_core, charge_shell, mass_core, &
                        mass_shell, k2_spring, k4_spring, max_dist, shell_cutoff)

      TYPE(shell_kind_type), POINTER                     :: shell
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: charge, charge_core, charge_shell, &
                                                            mass_core, mass_shell, k2_spring, &
                                                            k4_spring, max_dist, shell_cutoff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_shell', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(shell)) THEN
         IF (PRESENT(charge)) charge = shell%charge_core+shell%charge_shell
         IF (PRESENT(charge_core)) charge_core = shell%charge_core
         IF (PRESENT(charge_shell)) charge_shell = shell%charge_shell
         IF (PRESENT(mass_core)) mass_core = shell%mass_core
         IF (PRESENT(mass_shell)) mass_shell = shell%mass_shell
         IF (PRESENT(k2_spring)) k2_spring = shell%k2_spring
         IF (PRESENT(k4_spring)) k4_spring = shell%k4_spring
         IF (PRESENT(max_dist)) max_dist = shell%max_dist
         IF (PRESENT(shell_cutoff)) shell_cutoff = shell%shell_cutoff
      END IF

   END SUBROUTINE
! **************************************************************************************************
!> \brief ...
!> \param shell ...
! **************************************************************************************************
   SUBROUTINE shell_create(shell)

      TYPE(shell_kind_type), POINTER                     :: shell

      CHARACTER(len=*), PARAMETER :: routineN = 'shell_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(shell)))CALL cp__a("subsys/shell_potential_types.F",101)
      ALLOCATE (shell)
      shell%ref_count = 1

   END SUBROUTINE shell_create

! **************************************************************************************************
!> \brief ...
!> \param shell_list ...
!> \param ndim ...
! **************************************************************************************************
   SUBROUTINE shell_p_create(shell_list, ndim)

      TYPE(shell_p_type), DIMENSION(:), POINTER          :: shell_list
      INTEGER, INTENT(IN)                                :: ndim

      CHARACTER(len=*), PARAMETER :: routineN = 'shell_p_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(.NOT. ASSOCIATED(shell_list)))CALL cp__a("subsys/shell_potential_types.F",121)
      ALLOCATE (shell_list(ndim))

      DO i = 1, ndim
         NULLIFY (shell_list(i)%shell)
         CALL shell_create(shell_list(i)%shell)
         shell_list(i)%atm_name = ''
      END DO

   END SUBROUTINE shell_p_create

! **************************************************************************************************
!> \brief ...
!> \param shell ...
! **************************************************************************************************
   SUBROUTINE shell_retain(shell)

      TYPE(shell_kind_type), POINTER                     :: shell

      CHARACTER(len=*), PARAMETER :: routineN = 'shell_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(shell)))CALL cp__a("subsys/shell_potential_types.F",142)
      IF(.NOT.(shell%ref_count > 0))CALL cp__a("subsys/shell_potential_types.F",143)
      shell%ref_count = shell%ref_count+1

   END SUBROUTINE shell_retain

! **************************************************************************************************
!> \brief ...
!> \param shell ...
! **************************************************************************************************
   SUBROUTINE shell_release(shell)

      TYPE(shell_kind_type), POINTER                     :: shell

      CHARACTER(len=*), PARAMETER :: routineN = 'shell_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(shell)) THEN
         IF(.NOT.(shell%ref_count > 0))CALL cp__a("subsys/shell_potential_types.F",159)
         shell%ref_count = shell%ref_count-1
         IF (shell%ref_count == 0) THEN
            DEALLOCATE (shell)
         END IF
      END IF
      NULLIFY (shell)

   END SUBROUTINE shell_release

! **************************************************************************************************
!> \brief ...
!> \param shell_list ...
! **************************************************************************************************
   SUBROUTINE shell_p_release(shell_list)
      TYPE(shell_p_type), DIMENSION(:), POINTER          :: shell_list

      CHARACTER(len=*), PARAMETER :: routineN = 'shell_p_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(shell_list)) THEN
         DO i = 1, SIZE(shell_list)
            CALL shell_release(shell_list(i)%shell)
         END DO
         DEALLOCATE (shell_list)
      END IF

      NULLIFY (shell_list)

   END SUBROUTINE shell_p_release

END MODULE shell_potential_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Define the atomic kind types and their sub types
!> \author  Matthias Krack (MK)
!> \date    02.01.2002
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - 01.2002 creation [MK]
!> - 04.2002 added pao [fawzi]
!> - 09.2002 adapted for POL/KG use [GT]
!> - 02.2004 flexible normalization of basis sets [jgh]
!> - 03.2004 attach/detach routines [jgh]
!> - 10.2004 removed pao [fawzi]
!> - 08.2014 moevd qs-related stuff into new qs_kind_types.F [Ole Schuett]
! **************************************************************************************************
MODULE atomic_kind_types
   USE damping_dipole_types,            ONLY: damping_p_release,&
                                              damping_p_type
   USE external_potential_types,        ONLY: deallocate_potential,&
                                              fist_potential_type,&
                                              get_potential
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE periodic_table,                  ONLY: get_ptable_info
   USE shell_potential_types,           ONLY: shell_kind_type,&
                                              shell_release,&
                                              shell_retain
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atomic_kind_types'

!> Provides all information about an atomic kind
! **************************************************************************************************
   TYPE atomic_kind_type
      TYPE(fist_potential_type), POINTER     :: fist_potential => Null()
      CHARACTER(LEN=default_string_length)   :: name = ""
      CHARACTER(LEN=2)                       :: element_symbol = ""
      REAL(KIND=dp)                        :: mass = 0.0_dp
      INTEGER                                :: kind_number = -1
      INTEGER                                :: natom = -1
      INTEGER, DIMENSION(:), POINTER         :: atom_list => Null()
      LOGICAL                                :: shell_active = .FALSE.
      TYPE(shell_kind_type), POINTER         :: shell => Null()
      TYPE(damping_p_type), POINTER          :: damping => Null()
   END TYPE atomic_kind_type

!> Provides a vector of pointers of type atomic_kind_type
! **************************************************************************************************
   TYPE atomic_kind_p_type
      TYPE(atomic_kind_type), DIMENSION(:), &
         POINTER                             :: atomic_kind_set
   END TYPE atomic_kind_p_type

   ! Public subroutines

   PUBLIC :: deallocate_atomic_kind_set, &
             get_atomic_kind, &
             get_atomic_kind_set, &
             set_atomic_kind, &
             is_hydrogen

   ! Public data types
   PUBLIC :: atomic_kind_type

CONTAINS

! **************************************************************************************************
!> \brief   Destructor routine for a set of atomic kinds
!> \param atomic_kind_set ...
!> \date    02.01.2002
!> \author  Matthias Krack (MK)
!> \version 2.0
! **************************************************************************************************
   SUBROUTINE deallocate_atomic_kind_set(atomic_kind_set)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_atomic_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nkind

      IF (ASSOCIATED(atomic_kind_set)) THEN

         nkind = SIZE(atomic_kind_set)

         DO ikind = 1, nkind
            IF (ASSOCIATED(atomic_kind_set(ikind)%fist_potential)) THEN
               CALL deallocate_potential(atomic_kind_set(ikind)%fist_potential)
            END IF
            IF (ASSOCIATED(atomic_kind_set(ikind)%atom_list)) THEN
               DEALLOCATE (atomic_kind_set(ikind)%atom_list)
            END IF
            CALL shell_release(atomic_kind_set(ikind)%shell)

            CALL damping_p_release(atomic_kind_set(ikind)%damping)
         END DO
         DEALLOCATE (atomic_kind_set)
      ELSE
         CALL cp_abort(cp__l("subsys/atomic_kind_types.F",111), &
                       "The pointer atomic_kind_set is not associated and "// &
                       "cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_atomic_kind_set

! **************************************************************************************************
!> \brief Get attributes of an atomic kind.
!> \param atomic_kind ...
!> \param fist_potential ...
!> \param element_symbol ...
!> \param name ...
!> \param mass ...
!> \param kind_number ...
!> \param natom ...
!> \param atom_list ...
!> \param rcov ...
!> \param rvdw ...
!> \param z ...
!> \param qeff ...
!> \param apol ...
!> \param cpol ...
!> \param mm_radius ...
!> \param shell ...
!> \param shell_active ...
!> \param damping ...
! **************************************************************************************************
   SUBROUTINE get_atomic_kind(atomic_kind, fist_potential, &
                              element_symbol, name, mass, kind_number, natom, atom_list, &
                              rcov, rvdw, z, qeff, apol, cpol, mm_radius, &
                              shell, shell_active, damping)

      TYPE(atomic_kind_type)                             :: atomic_kind
      TYPE(fist_potential_type), OPTIONAL, POINTER       :: fist_potential
      CHARACTER(LEN=2), INTENT(OUT), OPTIONAL            :: element_symbol
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: mass
      INTEGER, INTENT(OUT), OPTIONAL                     :: kind_number, natom
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: atom_list
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: rcov, rvdw
      INTEGER, INTENT(OUT), OPTIONAL                     :: z
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: qeff, apol, cpol, mm_radius
      TYPE(shell_kind_type), OPTIONAL, POINTER           :: shell
      LOGICAL, INTENT(OUT), OPTIONAL                     :: shell_active
      TYPE(damping_p_type), OPTIONAL, POINTER            :: damping

      CHARACTER(len=*), PARAMETER :: routineN = 'get_atomic_kind', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(fist_potential)) fist_potential => atomic_kind%fist_potential
      IF (PRESENT(element_symbol)) element_symbol = atomic_kind%element_symbol
      IF (PRESENT(name)) name = atomic_kind%name
      IF (PRESENT(mass)) mass = atomic_kind%mass
      IF (PRESENT(kind_number)) kind_number = atomic_kind%kind_number
      IF (PRESENT(natom)) natom = atomic_kind%natom
      IF (PRESENT(atom_list)) atom_list => atomic_kind%atom_list

      IF (PRESENT(z)) THEN
         CALL get_ptable_info(atomic_kind%element_symbol, number=z)
      END IF
      IF (PRESENT(rcov)) THEN
         CALL get_ptable_info(atomic_kind%element_symbol, covalent_radius=rcov)
      END IF
      IF (PRESENT(rvdw)) THEN
         CALL get_ptable_info(atomic_kind%element_symbol, vdw_radius=rvdw)
      END IF
      IF (PRESENT(qeff)) THEN
         IF (ASSOCIATED(atomic_kind%fist_potential)) THEN
            CALL get_potential(potential=atomic_kind%fist_potential, qeff=qeff)
         ELSE
            qeff = -HUGE(0.0_dp)
         END IF
      END IF
      IF (PRESENT(apol)) THEN
         IF (ASSOCIATED(atomic_kind%fist_potential)) THEN
            CALL get_potential(potential=atomic_kind%fist_potential, apol=apol)
         ELSE
            apol = -HUGE(0.0_dp)
         END IF
      END IF
      IF (PRESENT(cpol)) THEN
         IF (ASSOCIATED(atomic_kind%fist_potential)) THEN
            CALL get_potential(potential=atomic_kind%fist_potential, cpol=cpol)
         ELSE
            cpol = -HUGE(0.0_dp)
         END IF
      END IF
      IF (PRESENT(mm_radius)) THEN
         IF (ASSOCIATED(atomic_kind%fist_potential)) THEN
            CALL get_potential(potential=atomic_kind%fist_potential, mm_radius=mm_radius)
         ELSE
            mm_radius = -HUGE(0.0_dp)
         END IF
      END IF
      IF (PRESENT(shell)) shell => atomic_kind%shell
      IF (PRESENT(shell_active)) shell_active = atomic_kind%shell_active
      IF (PRESENT(damping)) damping => atomic_kind%damping

   END SUBROUTINE get_atomic_kind

! **************************************************************************************************
!> \brief Get attributes of an atomic kind set.
!> \param atomic_kind_set ...
!> \param atom_of_kind ...
!> \param kind_of ...
!> \param natom_of_kind ...
!> \param maxatom ...
!> \param natom ...
!> \param nshell ...
!> \param fist_potential_present ...
!> \param shell_present ...
!> \param shell_adiabatic ...
!> \param shell_check_distance ...
!> \param damping_present ...
! **************************************************************************************************
   SUBROUTINE get_atomic_kind_set(atomic_kind_set, &
                                  atom_of_kind, kind_of, natom_of_kind, &
                                  maxatom, &
                                  natom, &
                                  nshell, &
                                  fist_potential_present, &
                                  shell_present, shell_adiabatic, &
                                  shell_check_distance, &
                                  damping_present)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL       :: atom_of_kind, kind_of, natom_of_kind
      INTEGER, INTENT(OUT), OPTIONAL                     :: maxatom, natom, nshell
      LOGICAL, INTENT(OUT), OPTIONAL                     :: fist_potential_present, shell_present, &
                                                            shell_adiabatic, shell_check_distance, &
                                                            damping_present

      CHARACTER(len=*), PARAMETER :: routineN = 'get_atomic_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, iatom, ikind, nkind
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(damping_p_type), POINTER                      :: damping
      TYPE(fist_potential_type), POINTER                 :: fist_potential
      TYPE(shell_kind_type), POINTER                     :: shell

      IF (ASSOCIATED(atomic_kind_set)) THEN

         IF (PRESENT(maxatom)) maxatom = 0
         IF (PRESENT(natom)) natom = 0
         IF (PRESENT(nshell)) nshell = 0
         IF (PRESENT(shell_present)) shell_present = .FALSE.
         IF (PRESENT(shell_adiabatic)) shell_adiabatic = .FALSE.
         IF (PRESENT(shell_check_distance)) shell_check_distance = .FALSE.
         IF (PRESENT(damping_present)) damping_present = .FALSE.
         IF (PRESENT(atom_of_kind)) atom_of_kind(:) = 0
         IF (PRESENT(kind_of)) kind_of(:) = 0
         IF (PRESENT(natom_of_kind)) natom_of_kind(:) = 0

         nkind = SIZE(atomic_kind_set)
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 fist_potential=fist_potential, &
                                 shell=shell, &
                                 damping=damping)
            IF (PRESENT(maxatom)) THEN
               maxatom = MAX(maxatom, atomic_kind%natom)
            END IF
            IF (PRESENT(natom)) THEN
               natom = natom+atomic_kind_set(ikind)%natom
            END IF
            IF (PRESENT(fist_potential_present)) THEN
               IF (ASSOCIATED(fist_potential)) THEN
                  fist_potential_present = .TRUE.
               END IF
            END IF
            IF (PRESENT(shell_present)) THEN
               IF (ASSOCIATED(shell)) THEN
                  shell_present = .TRUE.
               END IF
            END IF
            IF (PRESENT(shell_adiabatic) .AND. ASSOCIATED(shell)) THEN
               IF (.NOT. shell_adiabatic) shell_adiabatic = (shell%massfrac /= 0.0_dp)
            END IF
            IF (PRESENT(shell_check_distance) .AND. ASSOCIATED(shell)) THEN
               IF (.NOT. shell_check_distance) shell_check_distance = (shell%max_dist > 0.0_dp)
            END IF
            IF (PRESENT(damping_present)) THEN
               IF (ASSOCIATED(damping)) THEN
                  damping_present = .TRUE.
               END IF
            END IF
            IF (PRESENT(atom_of_kind)) THEN
               DO iatom = 1, atomic_kind%natom
                  atom_a = atomic_kind%atom_list(iatom)
                  atom_of_kind(atom_a) = iatom
               END DO
            END IF
            IF (PRESENT(kind_of)) THEN
               DO iatom = 1, atomic_kind%natom
                  atom_a = atomic_kind%atom_list(iatom)
                  kind_of(atom_a) = ikind
               END DO
            END IF
            IF (PRESENT(natom_of_kind)) THEN
               natom_of_kind(ikind) = atomic_kind_set(ikind)%natom
            END IF
         END DO
      ELSE
         CALL cp__b("subsys/atomic_kind_types.F",318,"The pointer atomic_kind_set is not associated")
      END IF

   END SUBROUTINE get_atomic_kind_set

! **************************************************************************************************
!> \brief Set the components of an atomic kind data set.
!> \param atomic_kind ...
!> \param element_symbol ...
!> \param name ...
!> \param mass ...
!> \param kind_number ...
!> \param natom ...
!> \param atom_list ...
!> \param fist_potential ...
!> \param shell ...
!> \param shell_active ...
!> \param damping ...
! **************************************************************************************************
   SUBROUTINE set_atomic_kind(atomic_kind, element_symbol, name, mass, kind_number, &
                              natom, atom_list, &
                              fist_potential, shell, &
                              shell_active, damping)

      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: element_symbol, name
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: mass
      INTEGER, INTENT(IN), OPTIONAL                      :: kind_number, natom
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: atom_list
      TYPE(fist_potential_type), OPTIONAL, POINTER       :: fist_potential
      TYPE(shell_kind_type), OPTIONAL, POINTER           :: shell
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell_active
      TYPE(damping_p_type), OPTIONAL, POINTER            :: damping

      CHARACTER(len=*), PARAMETER :: routineN = 'set_atomic_kind', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n

      IF (ASSOCIATED(atomic_kind)) THEN

         IF (PRESENT(element_symbol)) atomic_kind%element_symbol = element_symbol
         IF (PRESENT(name)) atomic_kind%name = name
         IF (PRESENT(mass)) atomic_kind%mass = mass
         IF (PRESENT(kind_number)) atomic_kind%kind_number = kind_number
         IF (PRESENT(natom)) atomic_kind%natom = natom
         IF (PRESENT(atom_list)) THEN
            n = SIZE(atom_list)
            IF (n > 0) THEN
               IF (ASSOCIATED(atomic_kind%atom_list)) THEN
                  DEALLOCATE (atomic_kind%atom_list)
               END IF
               ALLOCATE (atomic_kind%atom_list(n))
               atomic_kind%atom_list(:) = atom_list(:)
               atomic_kind%natom = n
            ELSE
               CALL cp__b("subsys/atomic_kind_types.F",374,"An invalid atom_list was supplied")
            END IF
         END IF
         IF (PRESENT(fist_potential)) atomic_kind%fist_potential => fist_potential
         IF (PRESENT(shell)) THEN
            atomic_kind%shell => shell
            CALL shell_retain(shell)
         END IF
         IF (PRESENT(shell_active)) atomic_kind%shell_active = shell_active

         IF (PRESENT(damping)) atomic_kind%damping => damping

      ELSE
         CALL cp__b("subsys/atomic_kind_types.F",387,"The pointer atomic_kind is not associated")
      END IF

   END SUBROUTINE set_atomic_kind

! **************************************************************************************************
!> \brief Determines if the atomic_kind is HYDROGEN
!> \param atomic_kind ...
!> \retval res ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   PURE FUNCTION is_hydrogen(atomic_kind) RESULT(res)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      LOGICAL                                            :: res

      res = TRIM(atomic_kind%element_symbol) == "H"
   END FUNCTION is_hydrogen

END MODULE atomic_kind_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
!> \author MI (08.01.2004)
! **************************************************************************************************
MODULE paw_proj_set_types

   USE ao_util,                         ONLY: exp_radius,&
                                              gauss_exponent
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: qs_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: dfac,&
                                              pi,&
                                              rootpi
   USE mathlib,                         ONLY: invert_matrix
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: indco,&
                                              indso,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: orbtramat
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'paw_proj_set_types'

   INTEGER, PARAMETER :: max_name_length = 60

   ! *** Define the projectors types ***

! **************************************************************************************************
   TYPE paw_proj_set_type
      INTEGER                                  :: maxl, ncgauprj, nsgauprj
      INTEGER, DIMENSION(:), POINTER           :: nprj ! 0:maxl
      INTEGER, DIMENSION(:), POINTER           :: lx, ly, lz ! ncgauprj
      INTEGER, DIMENSION(:), POINTER           :: ll, m ! nsgauprj
      INTEGER, DIMENSION(:), POINTER           :: first_prj, last_prj ! 0:maxl
      INTEGER, DIMENSION(:), POINTER           :: first_prjs ! 0:maxl
      REAL(dp)                                 :: rcprj
      REAL(dp), DIMENSION(:), POINTER          :: zisomin, zprjisomin
      REAL(dp), DIMENSION(:, :), POINTER        :: zetprj ! maxnprj,0:maxl
      REAL(dp), DIMENSION(:, :), POINTER        :: rzetprj ! maxnprj,0:maxl
      REAL(dp), DIMENSION(:, :), POINTER        :: cprj ! ncgauprj, maxco*nset
      REAL(dp), DIMENSION(:, :), POINTER        :: cprj_s ! nsgauprj, maxso*nset
      REAL(dp), DIMENSION(:, :), POINTER        :: csprj ! ncgauprj, np_so
      REAL(dp), DIMENSION(:, :), POINTER        :: local_oce_cphi_h, local_oce_cphi_s ! maxco,ncgf
      REAL(dp), DIMENSION(:, :), POINTER        :: local_oce_sphi_h, local_oce_sphi_s ! maxco,nsgf
      REAL(dp), DIMENSION(:, :), POINTER        :: sphi_h, sphi_s
      REAL(dp), DIMENSION(:, :, :, :), POINTER    :: gccprj ! maxnprj,maxnpgf,0:maxl,nset
      LOGICAL, DIMENSION(:, :), POINTER         :: isoprj ! maxnprj,0:maxl
      INTEGER                                  :: nsatbas
      INTEGER                                  :: nsotot
      INTEGER, DIMENSION(:), POINTER           :: o2nindex ! maxso*nset
      INTEGER, DIMENSION(:), POINTER           :: n2oindex ! maxso*nset

   END TYPE paw_proj_set_type

   ! *** Public subroutines ***

   PUBLIC ::  allocate_paw_proj_set, &
             deallocate_paw_proj_set, &
             get_paw_proj_set, &
             projectors, &
             set_paw_proj_set

   ! *** Public data types ***

   PUBLIC ::  paw_proj_set_type

CONTAINS

! **************************************************************************************************
!> \brief   Allocate projector type for GAPW
!> \param paw_proj_set ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_paw_proj_set(paw_proj_set)

      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_set

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_paw_proj_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(paw_proj_set)) CALL deallocate_paw_proj_set(paw_proj_set)

      ALLOCATE (paw_proj_set)

      NULLIFY (paw_proj_set%nprj)
      NULLIFY (paw_proj_set%lx)
      NULLIFY (paw_proj_set%ly)
      NULLIFY (paw_proj_set%lz)
      NULLIFY (paw_proj_set%ll)
      NULLIFY (paw_proj_set%m)
      NULLIFY (paw_proj_set%first_prj)
      NULLIFY (paw_proj_set%last_prj)
      NULLIFY (paw_proj_set%first_prjs)

      NULLIFY (paw_proj_set%zisomin)
      NULLIFY (paw_proj_set%zprjisomin)
      NULLIFY (paw_proj_set%zetprj)
      NULLIFY (paw_proj_set%cprj)
      NULLIFY (paw_proj_set%cprj_s)
      NULLIFY (paw_proj_set%csprj)
      NULLIFY (paw_proj_set%local_oce_cphi_h)
      NULLIFY (paw_proj_set%local_oce_cphi_s)
      NULLIFY (paw_proj_set%local_oce_sphi_h)
      NULLIFY (paw_proj_set%local_oce_sphi_s)
      NULLIFY (paw_proj_set%sphi_h)
      NULLIFY (paw_proj_set%sphi_s)
      NULLIFY (paw_proj_set%gccprj)
      NULLIFY (paw_proj_set%rzetprj)

      NULLIFY (paw_proj_set%isoprj)

      NULLIFY (paw_proj_set%o2nindex)
      NULLIFY (paw_proj_set%n2oindex)

   END SUBROUTINE allocate_paw_proj_set

! **************************************************************************************************
!> \brief   Deallocate a projector-type set data set.
!> \param paw_proj_set ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_paw_proj_set(paw_proj_set)
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_paw_proj_set', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (paw_proj_set%zisomin)
      DEALLOCATE (paw_proj_set%zprjisomin)
      DEALLOCATE (paw_proj_set%nprj)
      DEALLOCATE (paw_proj_set%lx)
      DEALLOCATE (paw_proj_set%ly)
      DEALLOCATE (paw_proj_set%lz)
      DEALLOCATE (paw_proj_set%ll)
      DEALLOCATE (paw_proj_set%m)
      DEALLOCATE (paw_proj_set%first_prj)
      DEALLOCATE (paw_proj_set%last_prj)
      DEALLOCATE (paw_proj_set%first_prjs)
      DEALLOCATE (paw_proj_set%zetprj)
      DEALLOCATE (paw_proj_set%cprj)
      DEALLOCATE (paw_proj_set%cprj_s)
      DEALLOCATE (paw_proj_set%csprj)
      DEALLOCATE (paw_proj_set%local_oce_cphi_h)
      DEALLOCATE (paw_proj_set%local_oce_cphi_s)
      DEALLOCATE (paw_proj_set%local_oce_sphi_h)
      DEALLOCATE (paw_proj_set%local_oce_sphi_s)
      DEALLOCATE (paw_proj_set%sphi_h)
      DEALLOCATE (paw_proj_set%sphi_s)
      DEALLOCATE (paw_proj_set%gccprj)
      DEALLOCATE (paw_proj_set%isoprj)
      DEALLOCATE (paw_proj_set%rzetprj)
      DEALLOCATE (paw_proj_set%o2nindex)
      DEALLOCATE (paw_proj_set%n2oindex)

      DEALLOCATE (paw_proj_set)
   END SUBROUTINE deallocate_paw_proj_set

! **************************************************************************************************
!> \brief   Initialize the projector-type set data set.
!> \param paw_proj ...
!> \param orb_basis ...
!> \param rc ...
!> \param qs_control ...
!> \param max_rad_local_type ...
!> \param force_env_section ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE projectors(paw_proj, orb_basis, rc, qs_control, max_rad_local_type, &
                         force_env_section)

      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      REAL(dp)                                           :: rc
      TYPE(qs_control_type), INTENT(IN)                  :: qs_control
      REAL(dp), INTENT(IN)                               :: max_rad_local_type
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'projectors', routineP = moduleN//':'//routineN

      REAL(dp)                                           :: eps_fit, eps_iso, eps_orb, eps_svd, &
                                                            max_rad_local

      eps_fit = qs_control%gapw_control%eps_fit
      eps_iso = qs_control%gapw_control%eps_iso
      eps_svd = qs_control%gapw_control%eps_svd
      max_rad_local = qs_control%gapw_control%max_rad_local
      IF (max_rad_local_type .LT. max_rad_local) THEN
         max_rad_local = max_rad_local_type
      END IF
      eps_orb = qs_control%eps_pgf_orb

      CALL build_projector(paw_proj, orb_basis, eps_fit, eps_iso, eps_svd, &
                           rc, eps_orb, max_rad_local, force_env_section)

   END SUBROUTINE projectors

! **************************************************************************************************
!> \brief   initialize the projector-type set data set.
!> \param paw_proj ...
!> \param orb_basis ...
!> \param eps_fit ...
!> \param eps_iso ...
!> \param eps_svd ...
!> \param rc ...
!> \param eps_orb ...
!> \param max_rad_local ...
!> \param force_env_section ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE build_projector(paw_proj, orb_basis, eps_fit, eps_iso, eps_svd, &
                              rc, eps_orb, max_rad_local, force_env_section)

      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      REAL(dp), INTENT(IN)                               :: eps_fit, eps_iso, eps_svd, rc, eps_orb, &
                                                            max_rad_local
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'build_projector', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: bsname
      INTEGER :: ic, icgf, icgfmax, icgfmin, ico, icomax, icomin, il, info, ip, ipgf, ipp, iprj, &
         iprjfirst, iprjs, is, iset, isgf, isgfmax, isgfmin, ishell, iso, iso_pgf, iso_set, &
         isomin, jp, k, kp, lshell, lwork, lx, ly, lz, maxco, maxl, maxnprj, maxpgf, maxso, mp, &
         ms, n, ncgauprj, ncgf, nisop, notprj, np, npgfg, ns, nset, nsgauprj, nsgf, nsox, &
         output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: IWORK
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: first_cgf, first_sgf, l, last_cgf, &
                                                            last_sgf
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: isoprj
      REAL(dp)                                           :: expzet, my_error, prefac, radius, x, &
                                                            zetmin, zetval
      REAL(dp), DIMENSION(:), POINTER                    :: zet, zetp
      REAL(dp), DIMENSION(:, :), POINTER                 :: cphi, gcc, smat, sphi, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: S, Work_dgesdd
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: U, VT
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcca, set_radius2
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      NULLIFY (first_cgf, first_sgf, last_cgf, last_sgf, gcc, l, set_radius, set_radius2)
      NULLIFY (cphi, sphi, lmax, lmin, npgf, nshell, zet, zetp, smat, work, gcca)

      IF(.NOT.(ASSOCIATED(paw_proj)))CALL cp__a("paw_proj_set_types.F",272)
      IF(.NOT.(ASSOCIATED(orb_basis)))CALL cp__a("paw_proj_set_types.F",273)

      CALL get_gto_basis_set(gto_basis_set=orb_basis, name=bsname, &
                             ncgf=ncgf, nset=nset, nsgf=nsgf, &
                             lmax=lmax, lmin=lmin, npgf=npgf, &
                             nshell=nshell, cphi=cphi, sphi=sphi, &
                             first_cgf=first_cgf, first_sgf=first_sgf, &
                             l=l, last_cgf=last_cgf, last_sgf=last_sgf, &
                             maxco=maxco, maxso=maxso, maxl=maxl, maxpgf=maxpgf, &
                             gcc=gcca)

      paw_proj%maxl = maxl
      IF(.NOT.(.NOT. ASSOCIATED(paw_proj%zisomin)))CALL cp__a("paw_proj_set_types.F",285)
      IF(.NOT.(.NOT. ASSOCIATED(paw_proj%zprjisomin)))CALL cp__a("paw_proj_set_types.F",286)
      IF(.NOT.(.NOT. ASSOCIATED(paw_proj%nprj)))CALL cp__a("paw_proj_set_types.F",287)

      ALLOCATE (paw_proj%zisomin(0:maxl))
      paw_proj%zisomin(0:maxl) = 0.0_dp
      ALLOCATE (paw_proj%zprjisomin(0:maxl))
      paw_proj%zprjisomin(0:maxl) = 0.0_dp
      ALLOCATE (paw_proj%nprj(0:maxl))
      paw_proj%nprj(0:maxl) = 0

      output_unit = cp_print_key_unit_nr(logger, force_env_section, &
                                         "DFT%PRINT%GAPW%PROJECTORS", extension=".Log")

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "Projectors for the basis functions of "//TRIM(bsname)
      END IF

      ALLOCATE (set_radius(nset))
      set_radius = 0.0_dp
      DO iset = 1, nset
         DO is = 1, nshell(iset)
            set_radius(iset) = MAX(set_radius(iset), &
                                   exp_radius(l(is, iset), orb_basis%zet(npgf(iset), iset), eps_orb, gcca(npgf(iset), is, iset)))
         END DO ! is
      END DO ! iset

      ALLOCATE (set_radius2(maxpgf, 0:maxl, nset))
      set_radius2 = 0.0_dp
      DO iset = 1, nset
         DO lshell = lmin(iset), lmax(iset)
            DO ip = 1, npgf(iset)
               set_radius2(ip, lshell, iset) = &
                  exp_radius(lshell, orb_basis%zet(ip, iset), eps_orb, 1.0_dp)
            END DO
         END DO ! is
      END DO ! iset

      maxnprj = 0
      DO lshell = 0, maxl ! lshell
         np = 0
         DO iset = 1, nset
            IF (lshell >= lmin(iset) .AND. lshell <= lmax(iset)) THEN
               DO ip = 1, npgf(iset)
                  IF (set_radius2(ip, lshell, iset) < max_rad_local) THEN
                     np = np+1
                  END IF
               END DO
            ENDIF
         ENDDO
         maxnprj = MAX(maxnprj, np)
         paw_proj%nprj(lshell) = np
      ENDDO ! lshell

      !   *** Allocate exponents and coefficients ***
      ALLOCATE (paw_proj%zetprj(maxnprj, 0:maxl))
      paw_proj%zetprj(1:maxnprj, 0:maxl) = 0.0_dp
      ALLOCATE (paw_proj%gccprj(maxnprj, maxpgf, 0:maxl, nset))
      paw_proj%gccprj = 0.0_dp
      ALLOCATE (paw_proj%rzetprj(maxnprj, 0:maxl))
      paw_proj%rzetprj(1:maxnprj, 0:maxl) = 0.0_dp
      ALLOCATE (paw_proj%isoprj(maxnprj, 0:maxl))
      paw_proj%isoprj = .FALSE.

      mp = 0
      DO lshell = 0, maxl ! lshell
         mp = MAX(mp, paw_proj%nprj(lshell)*nco(lshell))
      END DO

      NULLIFY (zet, zetp, gcc, smat, work)
      !   *** Generate the projetor basis for each ang. mom. q.n. ***
      DO lshell = 0, maxl ! lshell

         np = paw_proj%nprj(lshell)

         ALLOCATE (isoprj(np))
         isoprj = .FALSE.

         ALLOCATE (zet(np), zetp(np), gcc(np, np), smat(np, np), work(np, np))

         zet(:) = 0.0_dp
         zetp(:) = 0.0_dp
         gcc(:, :) = 0.0_dp
         smat(:, :) = 0.0_dp
         work(:, :) = 0.0_dp

         npgfg = 0
         notprj = 0
         !     *** Collect all the exponent which contribute to lshell ***
         DO iset = 1, nset ! iset
            IF (lshell >= lmin(iset) .AND. lshell <= lmax(iset)) THEN
               DO ip = 1, npgf(iset)
                  IF (set_radius2(ip, lshell, iset) < max_rad_local) THEN
                     npgfg = npgfg+1
                     zet(npgfg) = orb_basis%zet(ip, iset)
                  ELSE
                     notprj = notprj+1
                  END IF
               END DO
            ENDIF
         ENDDO ! iset

         !     *** Smallest exp. due to eps_iso: concerned as an isolated projector ***
         paw_proj%zisomin(lshell) = gauss_exponent(lshell, rc, eps_iso, 1.0_dp)

         nisop = 0
         DO ip = 1, np
            ! Check for equal exponents
            DO ipp = 1, ip-1
               IF (zet(ip) == zet(ipp)) THEN
                  CALL cp_abort(cp__l("paw_proj_set_types.F",396), &
                                "Linear dependency in the construction of the GAPW projectors:"// &
                                " different sets of the BASIS SET contain identical exponents")
               END IF
            END DO

            IF (zet(ip) >= paw_proj%zisomin(lshell)) THEN
               isoprj(ip) = .TRUE.
               nisop = nisop+1
            ELSE
               isoprj(ip) = .FALSE.
            END IF
         END DO

         !     *** Smallest exp. due to eps_fit: where to start geometric progression ***
         zetmin = gauss_exponent(lshell, rc, eps_fit, 1.0_dp)

         !     *** Generate the projectors by the geometric progression ***

         IF (np-nisop-1 > 2) THEN
            x = (80.0_dp/zetmin)**(1.0_dp/REAL(np-nisop-1, dp))
         ELSE
            x = 2.0_dp
         END IF
         IF (x > 2.0_dp) x = 2.0_dp

         zetval = zetmin
         DO ip = np, 1, -1
            IF (.NOT. isoprj(ip)) THEN
               zetp(ip) = zetval
               zetval = x*zetval
            ENDIF
         ENDDO

         paw_proj%zprjisomin(lshell) = zetval

         nisop = 0
         DO ip = np, 1, -1
            IF (isoprj(ip)) THEN
               zetp(ip) = zetval
               zetval = x*zetval
               nisop = nisop+1
            ENDIF
         ENDDO

         !     *** Build the overlap matrix: <projector|primitive> ***
         prefac = 0.5_dp**(lshell+2)*rootpi*dfac(2*lshell+1)
         expzet = REAL(lshell, dp)+1.5_dp

         DO ip = 1, np
            IF (isoprj(ip)) THEN
               smat(ip, ip) = prefac/(zetp(ip)+zet(ip))**expzet
            ELSE
               DO jp = 1, np
                  IF (.NOT. isoprj(jp)) THEN
                     smat(ip, jp) = prefac/(zetp(ip)+zet(jp))**expzet
                  ENDIF
               ENDDO
            ENDIF
         ENDDO

         !     FIXME ... catch the case where np == 0
         IF(.NOT.(np > 0))CALL cp__a("paw_proj_set_types.F",458)

         !     *** Compute inverse of the transpose ***
         IF (eps_svd .EQ. 0.0_dp) THEN
            CALL invert_matrix(smat, gcc, my_error, "T")
         ELSE
            work = TRANSPOSE(smat)
            ! workspace query
            ALLOCATE (iwork(8*np), S(np), U(np, np), VT(np, np), work_dgesdd(1))
            lwork = -1
            CALL DGESDD('S', np, np, work, np, S, U, np, vt, np, work_dgesdd, lwork, iwork, info)
            lwork = INT(work_dgesdd(1))
            DEALLOCATE (work_dgesdd); ALLOCATE (work_dgesdd(lwork))
            CALL DGESDD('S', np, np, work, np, S, U, np, vt, np, work_dgesdd, lwork, iwork, info)
            ! construct the inverse
            DO k = 1, np
               ! invert SV
               IF (S(k) < eps_svd) THEN
                  S(k) = 0.0_dp
               ELSE
                  S(k) = 1.0_dp/S(k)
               ENDIF
               VT(k, :) = VT(k, :)*S(k)
            ENDDO
            CALL DGEMM('T', 'T', np, np, np, 1.0_dp, VT, np, U, np, 0.0_dp, gcc, np)
            DEALLOCATE (iwork, S, U, VT, work_dgesdd)
         ENDIF

         !     *** Set the coefficient of the isolated projectors to 0 ***
         DO ip = 1, np
            IF (isoprj(ip)) THEN
               gcc(ip, ip) = 0.0_dp
            ENDIF
         ENDDO

         !     *** Transfer data from local to global variables ***

         paw_proj%zetprj(1:np, lshell) = zetp(1:np)
         paw_proj%isoprj(1:np, lshell) = isoprj(1:np)

         npgfg = 0
         notprj = 0
         DO iset = 1, nset ! iset
            IF (lshell >= lmin(iset) .AND. lshell <= lmax(iset)) THEN
               DO ip = 1, npgf(iset)
                  IF (set_radius2(ip, lshell, iset) < max_rad_local) THEN
                     npgfg = npgfg+1
                     paw_proj%gccprj(1:np, ip, lshell, iset) = gcc(1:np, npgfg)
                  ELSE
                     notprj = notprj+1
                     paw_proj%gccprj(1:np, ip, lshell, iset) = 0.0_dp
                  END IF
               END DO
            ENDIF
         ENDDO ! iset

         !     *** Print exponents and coefficients of the projectors ***
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A,I2)") &
               "Built projector for angular momentum quantum number l= ", lshell
            WRITE (UNIT=output_unit, FMT="(T2,A,I2)") &
               "Number of isolated projectors = ", nisop
            DO iset = 1, nset ! iset
               IF (lshell >= lmin(iset) .AND. lshell <= lmax(iset)) THEN
                  WRITE (UNIT=output_unit, FMT="(/,T2,A,I5,/,/,T4,A9,(T13,4f15.6))") &
                     "Set ", iset, "exp prj: ", &
                     (paw_proj%zetprj(ip, lshell), ip=1, np)
                  DO jp = 1, npgf(iset)
                     WRITE (UNIT=output_unit, FMT="(/,T4,A9,F15.6,/,T4,A9,(t13,4E15.6))") &
                        "exp gto: ", orb_basis%zet(jp, iset), &
                        "coeff.:  ", (paw_proj%gccprj(ip, jp, &
                                                      lshell, iset), ip=1, np)
                  ENDDO
               ENDIF
            ENDDO ! iset
         END IF

         !   *** Release the working storage for the current value lshell ***
         DEALLOCATE (isoprj)
         DEALLOCATE (gcc, zet, zetp, smat, work)

      ENDDO ! lshell
      CALL cp_print_key_finished_output(output_unit, logger, force_env_section, &
                                        "DFT%PRINT%GAPW%PROJECTORS")

      !   *** Release the working storage for the current value lshell ***
      DEALLOCATE (set_radius)
      DEALLOCATE (set_radius2)

      !   *** Count primitives basis functions for the projectors
      paw_proj%ncgauprj = 0
      paw_proj%nsgauprj = 0
      DO lshell = 0, maxl
         paw_proj%ncgauprj = paw_proj%ncgauprj+nco(lshell)*paw_proj%nprj(lshell)
         paw_proj%nsgauprj = paw_proj%nsgauprj+nso(lshell)*paw_proj%nprj(lshell)
      ENDDO

      ncgauprj = paw_proj%ncgauprj
      nsgauprj = paw_proj%nsgauprj
      CALL reallocate(paw_proj%cprj, 1, ncgauprj, 1, maxso*nset)
      CALL reallocate(paw_proj%lx, 1, ncgauprj)
      CALL reallocate(paw_proj%ly, 1, ncgauprj)
      CALL reallocate(paw_proj%lz, 1, ncgauprj)
      CALL reallocate(paw_proj%first_prj, 0, maxl)
      CALL reallocate(paw_proj%last_prj, 0, maxl)
      CALL reallocate(paw_proj%cprj_s, 1, nsgauprj, 1, maxso*nset)
      CALL reallocate(paw_proj%ll, 1, nsgauprj)
      CALL reallocate(paw_proj%m, 1, nsgauprj)
      CALL reallocate(paw_proj%first_prjs, 0, maxl)

      ncgauprj = 0
      nsgauprj = 0
      DO lshell = 0, maxl
         np = paw_proj%nprj(lshell)
         paw_proj%first_prj(lshell) = ncgauprj+1
         paw_proj%first_prjs(lshell) = nsgauprj+1
         paw_proj%last_prj(lshell) = ncgauprj+nco(lshell)*np
         DO ip = 1, np
            DO ico = ncoset(lshell-1)+1, ncoset(lshell)
               ncgauprj = ncgauprj+1
               paw_proj%lx(ncgauprj) = indco(1, ico)
               paw_proj%ly(ncgauprj) = indco(2, ico)
               paw_proj%lz(ncgauprj) = indco(3, ico)
            ENDDO ! ico
            DO iso = nsoset(lshell-1)+1, nsoset(lshell)
               nsgauprj = nsgauprj+1
               paw_proj%ll(nsgauprj) = indso(1, iso)
               paw_proj%m(nsgauprj) = indso(2, iso)
            ENDDO
         ENDDO ! ip
      ENDDO ! lshell

      ms = 0
      DO iset = 1, orb_basis%nset
         ns = nsoset(lmax(iset))
         DO lshell = lmin(iset), lmax(iset)
            iprjfirst = paw_proj%first_prjs(lshell)
            np = paw_proj%nprj(lshell)
            DO ipgf = 1, npgf(iset)
               DO ip = 1, np
                  DO il = 1, nso(lshell)
                     iprjs = iprjfirst-1+il+(ip-1)*nso(lshell)
                     iso = nsoset(lshell-1)+1+(lshell+paw_proj%m(iprjs))

                     iso = iso+(ipgf-1)*ns+ms
                     paw_proj%cprj_s(iprjs, iso) = &
                        paw_proj%gccprj(ip, ipgf, lshell, iset)
                  ENDDO ! iprjs
               ENDDO ! ip
            ENDDO ! ipgf
         ENDDO ! lshell
         ms = ms+maxso
      ENDDO ! iset

      ms = 0
      DO iset = 1, orb_basis%nset
         ns = nsoset(lmax(iset))
         DO ipgf = 1, npgf(iset)
            DO iso = nsoset(lmin(iset)-1)+1+ns*(ipgf-1)+ms, ns*ipgf+ms
               DO lshell = 0, maxl
                  np = paw_proj%nprj(lshell)
                  DO ip = 1, np
                     jp = paw_proj%first_prj(lshell)+(ip-1)*nco(lshell)-1
                     kp = paw_proj%first_prjs(lshell)+(ip-1)*nso(lshell)-1
                     DO il = 1, nco(lshell)
                        lx = indco(1, il+ncoset(lshell-1))
                        ly = indco(2, il+ncoset(lshell-1))
                        lz = indco(3, il+ncoset(lshell-1))
                        iprj = jp+il
                        DO ic = 1, nso(lshell)
                           iprjs = kp+ic
                           paw_proj%cprj(iprj, iso) = paw_proj%cprj(iprj, iso)+ &
                                                      orbtramat(lshell)%s2c(ic, il)*paw_proj%cprj_s(iprjs, iso)* &
                                                      SQRT(dfac(2*lshell+1)/(4.0_dp*pi)/ &
                                                           dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))
                        ENDDO ! ic
                     ENDDO ! il
                  ENDDO ! ip
               ENDDO ! lshell
            ENDDO ! iso
         ENDDO ! ipgf
         ms = ms+maxso
      ENDDO ! iset

      !   local coefficients for the one center expansions : oce
      !   the coefficients are calculated for the full and soft expansions
      ALLOCATE (paw_proj%local_oce_cphi_h(maxco, ncgf))
      paw_proj%local_oce_cphi_h = 0.0_dp
      ALLOCATE (paw_proj%local_oce_sphi_h(maxco, nsgf))
      paw_proj%local_oce_sphi_h = 0.0_dp
      ALLOCATE (paw_proj%sphi_h(maxco, nsgf))
      paw_proj%sphi_h = 0.0_dp

      ALLOCATE (paw_proj%local_oce_cphi_s(maxco, ncgf))
      paw_proj%local_oce_cphi_s = 0.0_dp
      ALLOCATE (paw_proj%local_oce_sphi_s(maxco, nsgf))
      paw_proj%local_oce_sphi_s = 0.0_dp
      ALLOCATE (paw_proj%sphi_s(maxco, nsgf))
      paw_proj%sphi_s = 0.0_dp

      !   *** Cartesian ***
      DO iset = 1, nset
         n = ncoset(lmax(iset))
         DO ipgf = 1, npgf(iset)
            DO ishell = 1, nshell(iset)
               lshell = l(ishell, iset)
               icomin = ncoset(lshell-1)+1+n*(ipgf-1)
               icomax = ncoset(lshell)+n*(ipgf-1)
               icgfmin = first_cgf(ishell, iset)
               icgfmax = last_cgf(ishell, iset)
               radius = exp_radius(lshell, orb_basis%zet(ipgf, iset), &
                                   eps_fit, 1.0_dp)
               DO icgf = icgfmin, icgfmax
                  paw_proj%local_oce_cphi_h(icomin:icomax, icgf) = &
                     cphi(icomin:icomax, icgf)
                  IF (radius < rc) THEN
                     paw_proj%local_oce_cphi_s(icomin:icomax, icgf) = 0.0_dp
                  ELSE
                     paw_proj%local_oce_cphi_s(icomin:icomax, icgf) = &
                        cphi(icomin:icomax, icgf)
                  ENDIF
               END DO
            ENDDO ! ishell
         ENDDO ! ipgf
      ENDDO ! iset

      DO iset = 1, nset
         n = ncoset(lmax(iset))
         DO ipgf = 1, npgf(iset)
            DO ishell = 1, nshell(iset)
               lshell = l(ishell, iset)
               icomin = ncoset(lshell-1)+1+n*(ipgf-1)
               icomax = ncoset(lshell)+n*(ipgf-1)
               isgfmin = first_sgf(ishell, iset)
               isgfmax = last_sgf(ishell, iset)
               radius = exp_radius(lshell, orb_basis%zet(ipgf, iset), &
                                   eps_fit, 1.0_dp)
               DO isgf = isgfmin, isgfmax
                  paw_proj%sphi_h(icomin:icomax, isgf) = &
                     sphi(icomin:icomax, isgf)
                  IF (radius < rc) THEN
                     paw_proj%sphi_s(icomin:icomax, isgf) = 0.0_dp
                  ELSE
                     paw_proj%sphi_s(icomin:icomax, isgf) = &
                        sphi(icomin:icomax, isgf)
                  ENDIF
               END DO
            ENDDO ! ishell
         ENDDO ! ipgf
      ENDDO ! iset

      DO iset = 1, nset
         n = ncoset(lmax(iset))
         ns = nsoset(lmax(iset))
         DO ipgf = 1, npgf(iset)
            DO ishell = 1, nshell(iset)
               lshell = l(ishell, iset)
               icomin = ncoset(lshell-1)+1+n*(ipgf-1)
               icomax = ncoset(lshell)+n*(ipgf-1)
               isgfmin = first_sgf(ishell, iset)
               isgfmax = last_sgf(ishell, iset)
               isomin = nsoset(lshell-1)+1+ns*(ipgf-1)
               DO is = 1, nso(lshell)
                  iso = isomin+is-1
                  DO ic = 1, nco(lshell)
                     ico = icomin+ic-1
                     lx = indco(1, ic+ncoset(lshell-1))
                     ly = indco(2, ic+ncoset(lshell-1))
                     lz = indco(3, ic+ncoset(lshell-1))
                     DO isgf = isgfmin, isgfmax
                        paw_proj%local_oce_sphi_h(iso, isgf) = &
                           paw_proj%local_oce_sphi_h(iso, isgf)+ &
                           orbtramat(lshell)%s2c(is, ic)* &
                           paw_proj%sphi_h(ico, isgf)*SQRT((4.0_dp*pi* &
                                                            dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))/dfac(2*lshell+1))
                        paw_proj%local_oce_sphi_s(iso, isgf) = &
                           paw_proj%local_oce_sphi_s(iso, isgf)+ &
                           orbtramat(lshell)%s2c(is, ic)* &
                           paw_proj%sphi_s(ico, isgf)*SQRT((4.0_dp*pi* &
                                                            dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1))/dfac(2*lshell+1))
                     END DO ! isgf
                  END DO ! ic
               END DO ! is
            END DO ! ishell
         END DO ! ipgf
      END DO ! iset

      !Index transformation OLD-NEW
      ALLOCATE (paw_proj%o2nindex(maxso*nset))
      ALLOCATE (paw_proj%n2oindex(maxso*nset))
      paw_proj%o2nindex = 0
      paw_proj%n2oindex = 0
      ico = 1
      DO iset = 1, nset
         iso_set = (iset-1)*maxso+1
         nsox = nsoset(lmax(iset))
         DO ipgf = 1, npgf(iset)
            iso_pgf = iso_set+(ipgf-1)*nsox
            iso = iso_pgf+nsoset(lmin(iset)-1)
            DO lx = lmin(iset), lmax(iset)
               DO k = 1, nso(lx)
                  paw_proj%n2oindex(ico) = iso
                  paw_proj%o2nindex(iso) = ico
                  iso = iso+1
                  ico = ico+1
               END DO
            END DO
         END DO
      END DO
      mp = ico-1
      paw_proj%nsatbas = mp
      paw_proj%nsotot = nset*maxso
      ALLOCATE (paw_proj%csprj(nsgauprj, mp))
      paw_proj%csprj = 0._dp
      DO k = 1, mp
         ico = paw_proj%n2oindex(k)
         paw_proj%csprj(:, k) = paw_proj%cprj_s(:, ico)
      END DO
   END SUBROUTINE build_projector

! **************************************************************************************************
!> \brief   Get informations about a paw projectors set.
!> \param paw_proj_set ...
!> \param cprj ...
!> \param cprj_s ...
!> \param csprj ...
!> \param first_prj ...
!> \param first_prjs ...
!> \param last_prj ...
!> \param local_oce_sphi_h ...
!> \param local_oce_sphi_s ...
!> \param maxl ...
!> \param ncgauprj ...
!> \param nsgauprj ...
!> \param nsatbas ...
!> \param nsotot ...
!> \param nprj ...
!> \param o2nindex ...
!> \param n2oindex ...
!> \param rcprj ...
!> \param rzetprj ...
!> \param zisomin ...
!> \param zetprj ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_paw_proj_set(paw_proj_set, cprj, cprj_s, csprj, &
                               first_prj, first_prjs, last_prj, &
                               local_oce_sphi_h, local_oce_sphi_s, &
                               maxl, ncgauprj, nsgauprj, nsatbas, nsotot, nprj, &
                               o2nindex, n2oindex, &
                               rcprj, rzetprj, zisomin, zetprj)

      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_set
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: cprj, cprj_s, csprj
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: first_prj, first_prjs, last_prj
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: local_oce_sphi_h, local_oce_sphi_s
      INTEGER, INTENT(OUT), OPTIONAL                     :: maxl, ncgauprj, nsgauprj, nsatbas, nsotot
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nprj, o2nindex, n2oindex
      REAL(dp), INTENT(OUT), OPTIONAL                    :: rcprj
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: rzetprj
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: zisomin
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: zetprj

      CHARACTER(len=*), PARAMETER :: routineN = 'get_paw_proj_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(paw_proj_set)) THEN
         IF (PRESENT(cprj)) cprj => paw_proj_set%cprj
         IF (PRESENT(cprj_s)) cprj_s => paw_proj_set%cprj_s
         IF (PRESENT(csprj)) csprj => paw_proj_set%csprj
         IF (PRESENT(local_oce_sphi_h)) local_oce_sphi_h => paw_proj_set%local_oce_sphi_h
         IF (PRESENT(local_oce_sphi_s)) local_oce_sphi_s => paw_proj_set%local_oce_sphi_s
         IF (PRESENT(first_prj)) first_prj => paw_proj_set%first_prj
         IF (PRESENT(last_prj)) last_prj => paw_proj_set%last_prj
         IF (PRESENT(first_prjs)) first_prjs => paw_proj_set%first_prjs
         IF (PRESENT(maxl)) maxl = paw_proj_set%maxl
         IF (PRESENT(ncgauprj)) ncgauprj = paw_proj_set%ncgauprj
         IF (PRESENT(nsgauprj)) nsgauprj = paw_proj_set%nsgauprj
         IF (PRESENT(nsatbas)) nsatbas = paw_proj_set%nsatbas
         IF (PRESENT(nsotot)) nsotot = paw_proj_set%nsotot
         IF (PRESENT(nprj)) nprj => paw_proj_set%nprj
         IF (PRESENT(rcprj)) rcprj = paw_proj_set%rcprj
         IF (PRESENT(rzetprj)) rzetprj => paw_proj_set%rzetprj
         IF (PRESENT(zisomin)) zisomin => paw_proj_set%zisomin
         IF (PRESENT(zetprj)) zetprj => paw_proj_set%zetprj
         IF (PRESENT(o2nindex)) o2nindex => paw_proj_set%o2nindex
         IF (PRESENT(n2oindex)) n2oindex => paw_proj_set%n2oindex

      ELSE
         CALL cp__b("paw_proj_set_types.F",847,"The pointer  paw_proj_set is not associated")
      ENDIF

   END SUBROUTINE get_paw_proj_set

! **************************************************************************************************
!> \brief   Set informations about a paw projectors set.
!> \param paw_proj_set ...
!> \param rzetprj ...
!> \param rcprj ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_paw_proj_set(paw_proj_set, rzetprj, rcprj)

      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_set
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: rzetprj
      REAL(dp), INTENT(IN), OPTIONAL                     :: rcprj

      CHARACTER(len=*), PARAMETER :: routineN = 'set_paw_proj_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(paw_proj_set)) THEN
         IF (PRESENT(rzetprj)) paw_proj_set%rzetprj(:, 0:) = rzetprj(:, 0:)
         IF (PRESENT(rcprj)) paw_proj_set%rcprj = rcprj
      ELSE
         CALL cp__b("paw_proj_set_types.F",872,"The pointer paw_proj_set is not associated")
      END IF

   END SUBROUTINE set_paw_proj_set

END MODULE paw_proj_set_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of the SCPTB parameter types.
!> \author JGH (12.2011)
! **************************************************************************************************
MODULE scptb_types

   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'scptb_types'

! **************************************************************************************************
   TYPE scptb_parameter_type
      ! PRIVATE
      CHARACTER(LEN=default_string_length)   :: key
      CHARACTER(LEN=default_string_length)   :: parameterization
      CHARACTER(LEN=default_string_length)   :: atomname
      LOGICAL                                :: defined
      INTEGER                                :: z !atomic number
      REAL(KIND=dp)                          :: zeff !effective core charge
      INTEGER                                :: natorb !total number of orbitals
      INTEGER                                :: lmaxorb !max angular momentum orbitals
      INTEGER                                :: lmaxscp !max angular momentum scp density
      INTEGER, DIMENSION(0:3)                :: norb !number of orbitals per l QN
      INTEGER, DIMENSION(10, 0:3)             :: nqm !principal quantum numbers
      REAL(KIND=dp), DIMENSION(10, 0:3)       :: zeta !orbital exponents
      REAL(KIND=dp), DIMENSION(10, 0:3)       :: hcore !core potential energy
      REAL(KIND=dp), DIMENSION(10, 0:3)       :: occupation !free atom occupation
      REAL(KIND=dp)                          :: energy !free atom energy
      REAL(KIND=dp), DIMENSION(3)            :: crep !core repulsion parameters
      REAL(KIND=dp), DIMENSION(1:3)          :: pol !polarization for SCP
      REAL(KIND=dp)                          :: ag !width of SCP charge
      REAL(KIND=dp)                          :: rcpair !cutoff radius for core pair potentials
   END TYPE scptb_parameter_type

   TYPE scptb_parameter_p_type
      TYPE(scptb_parameter_type), POINTER    :: scptb_param
   END TYPE scptb_parameter_p_type

! **************************************************************************************************

   TYPE scp_kind_vector_type
      INTEGER                                           :: natom
      INTEGER                                           :: nbasis
      REAL(dp), DIMENSION(:, :), POINTER                 :: vmat
   END TYPE scp_kind_vector_type

   TYPE scp_vector_type
      INTEGER                                           :: total_length
      TYPE(scp_kind_vector_type), DIMENSION(:), POINTER :: vector
   END TYPE scp_vector_type

   TYPE scp_vector_p_type
      TYPE(scp_vector_type), POINTER    :: vec
   END TYPE scp_vector_p_type

! **************************************************************************************************

   PUBLIC :: scptb_parameter_type, &
             scptb_parameter_p_type, &
             allocate_scptb_parameter, &
             deallocate_scptb_parameter, &
             get_scptb_parameter, &
             set_scptb_parameter, &
             write_scptb_parameter

   PUBLIC :: scp_vector_type, &
             scp_vector_p_type, &
             scp_vector_create, &
             scp_vector_release, &
             scp_vector_set, &
             scp_vector_add, &
             scp_vector_mult, &
             scp_vector_copy, &
             scp_vector_dot

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param scptb_parameter ...
! **************************************************************************************************
   SUBROUTINE allocate_scptb_parameter(scptb_parameter)

      TYPE(scptb_parameter_type), POINTER                :: scptb_parameter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_scptb_parameter', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(scptb_parameter)) &
         CALL deallocate_scptb_parameter(scptb_parameter)

      ALLOCATE (scptb_parameter)

      scptb_parameter%key = ""
      scptb_parameter%parameterization = ""
      scptb_parameter%atomname = ""
      scptb_parameter%defined = .FALSE.
      scptb_parameter%z = -1
      scptb_parameter%zeff = -1.0_dp
      scptb_parameter%natorb = 0
      scptb_parameter%lmaxorb = -1
      scptb_parameter%lmaxscp = -1
      scptb_parameter%norb = 0
      scptb_parameter%nqm = 0
      scptb_parameter%zeta = 0.0_dp
      scptb_parameter%hcore = 0.0_dp
      scptb_parameter%occupation = 0.0_dp
      scptb_parameter%energy = 0.0_dp
      scptb_parameter%crep = 0.0_dp
      scptb_parameter%pol = 0.0_dp
      scptb_parameter%ag = 0.0_dp
      scptb_parameter%rcpair = 0.0_dp

   END SUBROUTINE allocate_scptb_parameter

! **************************************************************************************************
!> \brief ...
!> \param scptb_parameter ...
! **************************************************************************************************
   SUBROUTINE deallocate_scptb_parameter(scptb_parameter)

      TYPE(scptb_parameter_type), POINTER                :: scptb_parameter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_scptb_parameter', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(scptb_parameter)))CALL cp__a("scptb_types.F",151)
      DEALLOCATE (scptb_parameter)

   END SUBROUTINE deallocate_scptb_parameter

! **************************************************************************************************
!> \brief ...
!> \param scptb_parameter ...
!> \param key ...
!> \param atomname ...
!> \param parameterization ...
!> \param defined ...
!> \param z ...
!> \param zeff ...
!> \param natorb ...
!> \param lmaxorb ...
!> \param lmaxscp ...
!> \param norb ...
!> \param nqm ...
!> \param zeta ...
!> \param hcore ...
!> \param occupation ...
!> \param energy ...
!> \param crep ...
!> \param pol ...
!> \param ag ...
!> \param rcpair ...
! **************************************************************************************************
   SUBROUTINE get_scptb_parameter(scptb_parameter, key, atomname, parameterization, defined, z, zeff, natorb, &
                                  lmaxorb, lmaxscp, norb, nqm, zeta, hcore, occupation, energy, crep, pol, ag, rcpair)

      TYPE(scptb_parameter_type), POINTER                :: scptb_parameter
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: key, atomname, parameterization
      LOGICAL, INTENT(OUT), OPTIONAL                     :: defined
      INTEGER, INTENT(OUT), OPTIONAL                     :: z
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: zeff
      INTEGER, INTENT(OUT), OPTIONAL                     :: natorb, lmaxorb, lmaxscp
      INTEGER, DIMENSION(0:3), OPTIONAL                  :: norb
      INTEGER, DIMENSION(10, 0:3), OPTIONAL              :: nqm
      REAL(KIND=dp), DIMENSION(10, 0:3), OPTIONAL        :: zeta, hcore, occupation
      REAL(KIND=dp), OPTIONAL                            :: energy
      REAL(KIND=dp), DIMENSION(3), OPTIONAL              :: crep
      REAL(KIND=dp), DIMENSION(1:3), OPTIONAL            :: pol
      REAL(KIND=dp), OPTIONAL                            :: ag, rcpair

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_scptb_parameter', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(scptb_parameter)))CALL cp__a("scptb_types.F",199)

      IF (PRESENT(key)) key = scptb_parameter%key
      IF (PRESENT(atomname)) atomname = scptb_parameter%atomname
      IF (PRESENT(parameterization)) parameterization = scptb_parameter%parameterization
      IF (PRESENT(defined)) defined = scptb_parameter%defined
      IF (PRESENT(z)) z = scptb_parameter%z
      IF (PRESENT(zeff)) zeff = scptb_parameter%zeff
      IF (PRESENT(natorb)) natorb = scptb_parameter%natorb
      IF (PRESENT(lmaxorb)) lmaxorb = scptb_parameter%lmaxorb
      IF (PRESENT(lmaxscp)) lmaxscp = scptb_parameter%lmaxscp
      IF (PRESENT(norb)) norb = scptb_parameter%norb
      IF (PRESENT(nqm)) nqm = scptb_parameter%nqm
      IF (PRESENT(zeta)) zeta = scptb_parameter%zeta
      IF (PRESENT(hcore)) hcore = scptb_parameter%hcore
      IF (PRESENT(occupation)) occupation = scptb_parameter%occupation
      IF (PRESENT(energy)) energy = scptb_parameter%energy
      IF (PRESENT(crep)) crep = scptb_parameter%crep
      IF (PRESENT(pol)) pol = scptb_parameter%pol
      IF (PRESENT(ag)) ag = scptb_parameter%ag
      IF (PRESENT(rcpair)) rcpair = scptb_parameter%rcpair

   END SUBROUTINE get_scptb_parameter

! **************************************************************************************************
!> \brief ...
!> \param scptb_parameter ...
!> \param key ...
!> \param atomname ...
!> \param parameterization ...
!> \param defined ...
!> \param z ...
!> \param zeff ...
!> \param natorb ...
!> \param lmaxorb ...
!> \param lmaxscp ...
!> \param norb ...
!> \param nqm ...
!> \param zeta ...
!> \param hcore ...
!> \param occupation ...
!> \param energy ...
!> \param crep ...
!> \param pol ...
!> \param ag ...
!> \param rcpair ...
! **************************************************************************************************
   SUBROUTINE set_scptb_parameter(scptb_parameter, key, atomname, parameterization, defined, z, zeff, natorb, &
                                  lmaxorb, lmaxscp, norb, nqm, zeta, hcore, occupation, energy, crep, pol, ag, rcpair)

      TYPE(scptb_parameter_type), POINTER                :: scptb_parameter
      CHARACTER(LEN=*), OPTIONAL                         :: key, atomname, parameterization
      LOGICAL, OPTIONAL                                  :: defined
      INTEGER, OPTIONAL                                  :: z
      REAL(KIND=dp), OPTIONAL                            :: zeff
      INTEGER, OPTIONAL                                  :: natorb, lmaxorb, lmaxscp
      INTEGER, DIMENSION(0:3), OPTIONAL                  :: norb
      INTEGER, DIMENSION(10, 0:3), OPTIONAL              :: nqm
      REAL(KIND=dp), DIMENSION(10, 0:3), OPTIONAL        :: zeta, hcore, occupation
      REAL(KIND=dp), OPTIONAL                            :: energy
      REAL(KIND=dp), DIMENSION(3), OPTIONAL              :: crep
      REAL(KIND=dp), DIMENSION(1:3), OPTIONAL            :: pol
      REAL(KIND=dp), OPTIONAL                            :: ag, rcpair

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_scptb_parameter', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(scptb_parameter)))CALL cp__a("scptb_types.F",266)

      IF (PRESENT(key)) scptb_parameter%key = key
      IF (PRESENT(atomname)) scptb_parameter%atomname = atomname
      IF (PRESENT(parameterization)) scptb_parameter%parameterization = parameterization
      IF (PRESENT(defined)) scptb_parameter%defined = defined
      IF (PRESENT(z)) scptb_parameter%z = z
      IF (PRESENT(zeff)) scptb_parameter%zeff = zeff
      IF (PRESENT(natorb)) scptb_parameter%natorb = natorb
      IF (PRESENT(lmaxorb)) scptb_parameter%lmaxorb = lmaxorb
      IF (PRESENT(lmaxscp)) scptb_parameter%lmaxscp = lmaxscp
      IF (PRESENT(norb)) scptb_parameter%norb = norb
      IF (PRESENT(nqm)) scptb_parameter%nqm = nqm
      IF (PRESENT(zeta)) scptb_parameter%zeta = zeta
      IF (PRESENT(hcore)) scptb_parameter%hcore = hcore
      IF (PRESENT(occupation)) scptb_parameter%occupation = occupation
      IF (PRESENT(energy)) scptb_parameter%energy = energy
      IF (PRESENT(crep)) scptb_parameter%crep = crep
      IF (PRESENT(pol)) scptb_parameter%pol = pol
      IF (PRESENT(ag)) scptb_parameter%ag = ag
      IF (PRESENT(rcpair)) scptb_parameter%rcpair = rcpair

   END SUBROUTINE set_scptb_parameter

! **************************************************************************************************
!> \brief ...
!> \param scptb_parameter ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE write_scptb_parameter(scptb_parameter, section)

      TYPE(scptb_parameter_type), POINTER                :: scptb_parameter
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_scptb_parameter', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atomname, key, parameterization
      INTEGER                                            :: i, l, lmaxorb, lmaxscp, natorb, &
                                                            output_unit
      INTEGER, DIMENSION(0:3)                            :: norb
      INTEGER, DIMENSION(10, 0:3)                        :: nqm
      LOGICAL                                            :: defined
      REAL(KIND=dp)                                      :: ag, energy, rcpair, zeff
      REAL(KIND=dp), DIMENSION(10, 0:3)                  :: hcore, zeta
      REAL(KIND=dp), DIMENSION(1:3)                      :: pol
      REAL(KIND=dp), DIMENSION(3)                        :: crep
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      IF (ASSOCIATED(scptb_parameter) .AND. &
          BTEST(cp_print_key_should_output(logger%iter_info, section, &
                                           "PRINT%KINDS/POTENTIAL"), cp_p_file)) THEN

         output_unit = cp_print_key_unit_nr(logger, section, "PRINT%KINDS", &
                                            extension=".Log")

         IF (output_unit > 0) THEN
            CALL get_scptb_parameter(scptb_parameter, key=key, atomname=atomname, &
                                     parameterization=parameterization, defined=defined)

            WRITE (UNIT=output_unit, FMT="(/,T10,A,T67,A14)") " SCPTB  parameters: ", TRIM(atomname)
            WRITE (UNIT=output_unit, FMT="(T67,A14)") TRIM(key)
            WRITE (UNIT=output_unit, FMT="(T67,A14)") TRIM(parameterization)
            IF (defined) THEN
               CALL get_scptb_parameter(scptb_parameter, &
                                        zeff=zeff, natorb=natorb, lmaxorb=lmaxorb, lmaxscp=lmaxscp, &
                                        norb=norb, nqm=nqm, zeta=zeta, hcore=hcore, &
                                        energy=energy, crep=crep, pol=pol, ag=ag, rcpair=rcpair)
               WRITE (UNIT=output_unit, FMT="(T16,A,T71,F10.2)") "Effective core charge:", zeff
               WRITE (UNIT=output_unit, FMT="(T16,A,T71,I10)") "Total number of orbitals:", natorb
               WRITE (UNIT=output_unit, FMT="(T16,A,T30,A,T71,A)") "l-QM n-QM  i", "Exponent", "H core"
               DO l = 0, lmaxorb
                  DO i = 1, norb(l)
                     WRITE (UNIT=output_unit, FMT="(T16,3I4,T30,F12.6,T71,F10.3)") &
                        l, i, nqm(i, l), zeta(i, l), hcore(i, l)
                  END DO
               END DO
               WRITE (UNIT=output_unit, FMT="(T16,A,T61,F20.10)") "Energy of free atom [au]:", energy
               WRITE (UNIT=output_unit, FMT="(T16,A,T41,4F10.4)") "Core repulsion potential ", (crep(i), i=1, 4)
               WRITE (UNIT=output_unit, FMT="(T16,A,T41,4F10.4)") "SCP Polarization ", (pol(l), l=1, lmaxscp)
               WRITE (UNIT=output_unit, FMT="(T16,A,T61,F20.5)") "Gaussian exponents for SCP ", ag
               WRITE (UNIT=output_unit, FMT="(T16,A,T61,F20.5)") "Pair potential cutoff (core) [bohr]:", rcpair
            ELSE
               WRITE (UNIT=output_unit, FMT="(T55,A)") &
                  "Parameters are not defined"
            END IF
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, section, &
                                           "PRINT%KINDS")
      END IF

   END SUBROUTINE write_scptb_parameter

! **************************************************************************************************
!> \brief ...
!> \param scp_vec ...
!> \param nkind ...
!> \param natoms ...
!> \param nbasis ...
! **************************************************************************************************
   SUBROUTINE scp_vector_create(scp_vec, nkind, natoms, nbasis)
      TYPE(scp_vector_type), POINTER                     :: scp_vec
      INTEGER, INTENT(IN)                                :: nkind
      INTEGER, DIMENSION(:), INTENT(IN)                  :: natoms, nbasis

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, tl

      CALL scp_vector_release(scp_vec)

      ALLOCATE (scp_vec)
      ALLOCATE (scp_vec%vector(nkind))

      tl = 0
      DO ikind = 1, nkind
         scp_vec%vector(ikind)%natom = natoms(ikind)
         scp_vec%vector(ikind)%nbasis = nbasis(ikind)
         tl = tl+natoms(ikind)*nbasis(ikind)
         ALLOCATE (scp_vec%vector(ikind)%vmat(nbasis(ikind), natoms(ikind)))
         scp_vec%vector(ikind)%vmat = 0._dp
      END DO
      scp_vec%total_length = tl

   END SUBROUTINE scp_vector_create

! **************************************************************************************************
!> \brief ...
!> \param scp_vec ...
! **************************************************************************************************
   SUBROUTINE scp_vector_release(scp_vec)
      TYPE(scp_vector_type), POINTER                     :: scp_vec

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      IF (ASSOCIATED(scp_vec)) THEN
         IF (ASSOCIATED(scp_vec%vector)) THEN
            DO ikind = 1, SIZE(scp_vec%vector)
               DEALLOCATE (scp_vec%vector(ikind)%vmat)
            END DO
            DEALLOCATE (scp_vec%vector)
         END IF
         DEALLOCATE (scp_vec)
      END IF

   END SUBROUTINE scp_vector_release

! **************************************************************************************************
!> \brief ...
!> \param scp_vec ...
!> \param value ...
! **************************************************************************************************
   SUBROUTINE scp_vector_set(scp_vec, value)
      TYPE(scp_vector_type)                              :: scp_vec
      REAL(KIND=dp), INTENT(IN)                          :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_set', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, SIZE(scp_vec%vector)
         scp_vec%vector(i)%vmat(:, :) = value
      END DO

   END SUBROUTINE scp_vector_set

! **************************************************************************************************
!> \brief ...
!> \param alpha ...
!> \param vecx ...
!> \param vecy ...
! **************************************************************************************************
   SUBROUTINE scp_vector_add(alpha, vecx, vecy)
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(scp_vector_type)                              :: vecx, vecy

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_add', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, SIZE(vecx%vector)
         vecy%vector(i)%vmat(:, :) = vecy%vector(i)%vmat(:, :)+alpha*vecx%vector(i)%vmat(:, :)
      END DO

   END SUBROUTINE scp_vector_add

! **************************************************************************************************
!> \brief ...
!> \param vec_in ...
!> \param vec_out ...
! **************************************************************************************************
   SUBROUTINE scp_vector_copy(vec_in, vec_out)
      TYPE(scp_vector_type)                              :: vec_in, vec_out

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, SIZE(vec_in%vector)
         vec_out%vector(i)%vmat(:, :) = vec_in%vector(i)%vmat(:, :)
      END DO

   END SUBROUTINE scp_vector_copy

! **************************************************************************************************
!> \brief ...
!> \param scale ...
!> \param vec ...
! **************************************************************************************************
   SUBROUTINE scp_vector_scale(scale, vec)
      REAL(KIND=dp), INTENT(IN)                          :: scale
      TYPE(scp_vector_type)                              :: vec

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_scale', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, SIZE(vec%vector)
         vec%vector(i)%vmat(:, :) = scale*vec%vector(i)%vmat(:, :)
      END DO

   END SUBROUTINE scp_vector_scale

! **************************************************************************************************
!> \brief ...
!> \param vecx ...
!> \param vecy ...
! **************************************************************************************************
   SUBROUTINE scp_vector_mult(vecx, vecy)
      TYPE(scp_vector_type)                              :: vecx, vecy

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_mult', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, SIZE(vecx%vector)
         vecy%vector(i)%vmat(:, :) = vecy%vector(i)%vmat(:, :)*vecx%vector(i)%vmat(:, :)
      END DO

   END SUBROUTINE scp_vector_mult

! **************************************************************************************************
!> \brief ...
!> \param RESULT ...
!> \param vec1 ...
!> \param vec2 ...
! **************************************************************************************************
   SUBROUTINE scp_vector_dot(RESULT, vec1, vec2)
      REAL(KIND=dp), INTENT(OUT)                         :: RESULT
      TYPE(scp_vector_type)                              :: vec1, vec2

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_dot', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      RESULT = 0._dp
      DO i = 1, SIZE(vec1%vector)
         RESULT = RESULT+SUM(vec1%vector(i)%vmat(:, :)*vec2%vector(i)%vmat(:, :))
      END DO

   END SUBROUTINE scp_vector_dot

! **************************************************************************************************
!> \brief ...
!> \param RESULT ...
!> \param vec ...
! **************************************************************************************************
   SUBROUTINE scp_vector_norm(RESULT, vec)
      REAL(KIND=dp), INTENT(OUT)                         :: RESULT
      TYPE(scp_vector_type)                              :: vec

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_norm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      RESULT = 0._dp
      DO i = 1, SIZE(vec%vector)
         RESULT = RESULT+SUM(vec%vector(i)%vmat(:, :)*vec%vector(i)%vmat(:, :))
      END DO
      RESULT = SQRT(RESULT/REAL(vec%total_length, KIND=dp))

   END SUBROUTINE scp_vector_norm

! **************************************************************************************************
!> \brief ...
!> \param vec ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE scp_vector_sync(vec, para_env)
      TYPE(scp_vector_type)                              :: vec
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_sync', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, SIZE(vec%vector)
         CALL mp_sum(vec%vector(i)%vmat(:, :), para_env%group)
      END DO

   END SUBROUTINE scp_vector_sync

! **************************************************************************************************
!> \brief ...
!> \param vec ...
! **************************************************************************************************
   SUBROUTINE scp_vector_print(vec)
      TYPE(scp_vector_type)                              :: vec

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_vector_print', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ia, ib, j, na, nb

      WRITE (*, *) "Index   Type   Atom    Basis       Value"
      j = 0
      DO i = 1, SIZE(vec%vector)
         na = vec%vector(i)%natom
         nb = vec%vector(i)%nbasis
         DO ia = 1, na
            DO ib = 1, nb
               j = j+1
               WRITE (*, "(i5,i7,i7,i7,F20.10)") j, i, ia, ib, vec%vector(i)%vmat(ib, ia)
            END DO
         END DO
      END DO

   END SUBROUTINE scp_vector_print

! **************************************************************************************************

END MODULE scptb_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - Container to hold basis sets
!> \author JGH (09.07.2015)
! **************************************************************************************************
MODULE basis_set_container_types

   USE basis_set_types,                 ONLY: deallocate_gto_basis_set,&
                                              gto_basis_set_type
   USE kinds,                           ONLY: default_string_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'basis_set_container_types'

! **************************************************************************************************
   INTEGER, PARAMETER                       :: unknown_basis = 100, &
                                               orbital_basis = 101, &
                                               auxiliary_basis = 102, &
                                               ri_aux_basis = 103, &
                                               lri_aux_basis = 104, &
                                               aux_fit_basis = 105, &
                                               soft_basis = 106, &
                                               hard_basis = 107
! **************************************************************************************************
   TYPE basis_set_container_type
      PRIVATE
      CHARACTER(LEN=default_string_length)       :: basis_type = ""
      INTEGER                                    :: basis_type_nr = 0
      TYPE(gto_basis_set_type), POINTER          :: basis_set => NULL()
   END TYPE basis_set_container_type
! **************************************************************************************************

   PUBLIC :: basis_set_container_type

   PUBLIC :: remove_basis_set_container, &
             add_basis_set_to_container, get_basis_from_container, &
             remove_basis_from_container, translate_basis_type

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE remove_basis_set_container(basis)
      TYPE(basis_set_container_type), DIMENSION(:), &
         INTENT(inout)                                   :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'remove_basis_set_container', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      DO i = 1, SIZE(basis)
         basis(i)%basis_type = ""
         basis(i)%basis_type_nr = 0
         IF (ASSOCIATED(basis(i)%basis_set)) THEN
            CALL deallocate_gto_basis_set(basis(i)%basis_set)
         END IF
      END DO

   END SUBROUTINE remove_basis_set_container

! **************************************************************************************************
!> \brief ...
!> \param basis_set_type ...
!> \retval basis_type_nr ...
! **************************************************************************************************
   FUNCTION get_basis_type(basis_set_type) RESULT(basis_type_nr)
      CHARACTER(len=*)                                   :: basis_set_type
      INTEGER                                            :: basis_type_nr

      SELECT CASE (basis_set_type)
      CASE ("ORB")
         basis_type_nr = orbital_basis
      CASE ("AUX")
         basis_type_nr = auxiliary_basis
      CASE ("RI_AUX")
         basis_type_nr = ri_aux_basis
      CASE ("LRI")
         basis_type_nr = lri_aux_basis
      CASE ("AUX_FIT")
         basis_type_nr = aux_fit_basis
      CASE ("SOFT")
         basis_type_nr = soft_basis
      CASE ("HARD")
         basis_type_nr = hard_basis
      CASE DEFAULT
         basis_type_nr = unknown_basis
      END SELECT

   END FUNCTION get_basis_type
! **************************************************************************************************
!> \brief ...
!> \param basis_set_id ...
!> \retval basis_set_type ...
! **************************************************************************************************
   FUNCTION translate_basis_type(basis_set_id) RESULT(basis_set_type)
      INTEGER                                            :: basis_set_id
      CHARACTER(len=default_string_length)               :: basis_set_type

      INTEGER, PARAMETER :: use_aux_basis_set = 3, use_aux_fit_basis_set = 2, &
         use_lri_basis_set = 5, use_orb_basis_set = 1, use_ri_aux_basis_set = 4

! This parameter identifies basis sets

      SELECT CASE (basis_set_id)
      CASE (use_orb_basis_set)
         basis_set_type = "ORB"
      CASE (use_ri_aux_basis_set)
         basis_set_type = "RI_AUX"
      CASE (use_aux_basis_set)
         basis_set_type = "AUX"
      CASE (use_aux_fit_basis_set)
         basis_set_type = "AUX_FIT"
      CASE (use_lri_basis_set)
         basis_set_type = "LRI"
      CASE DEFAULT
         basis_set_type = ""
      END SELECT

   END FUNCTION translate_basis_type

! **************************************************************************************************
!> \brief ...
!> \param container ...
!> \param basis_set ...
!> \param basis_set_type ...
! **************************************************************************************************
   SUBROUTINE add_basis_set_to_container(container, basis_set, basis_set_type)
      TYPE(basis_set_container_type), DIMENSION(:), &
         INTENT(inout)                                   :: container
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      CHARACTER(len=*)                                   :: basis_set_type

      CHARACTER(len=*), PARAMETER :: routineN = 'add_basis_set_to_container', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: success

      success = .FALSE.
      DO i = 1, SIZE(container)
         IF (container(i)%basis_type_nr == 0) THEN
            container(i)%basis_type = basis_set_type
            container(i)%basis_set => basis_set
            container(i)%basis_type_nr = get_basis_type(basis_set_type)
            success = .TRUE.
            EXIT
         END IF
      END DO
      IF(.NOT.(success))CALL cp__a("aobasis/basis_set_container_types.F",165)

   END SUBROUTINE add_basis_set_to_container

! **************************************************************************************************
!> \brief ...
!> \param container ...
!> \param inum ...
!> \param basis_type ...
! **************************************************************************************************
   SUBROUTINE remove_basis_from_container(container, inum, basis_type)
      TYPE(basis_set_container_type), DIMENSION(:), &
         INTENT(inout)                                   :: container
      INTEGER, INTENT(IN), OPTIONAL                      :: inum
      CHARACTER(len=*), OPTIONAL                         :: basis_type

      CHARACTER(len=*), PARAMETER :: routineN = 'remove_basis_from_container', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: basis_nr, i, ibas

      IF (PRESENT(inum)) THEN
         IF(.NOT.(inum <= SIZE(container)))CALL cp__a("aobasis/basis_set_container_types.F",187)
         IF(.NOT.(inum >= 1))CALL cp__a("aobasis/basis_set_container_types.F",188)
         ibas = inum
      ELSE IF (PRESENT(basis_type)) THEN
         basis_nr = get_basis_type(basis_type)
         ibas = 0
         DO i = 1, SIZE(container)
            IF (container(i)%basis_type_nr == basis_nr) THEN
               ibas = i
               EXIT
            END IF
         END DO
      ELSE
         CALL cp__b("aobasis/basis_set_container_types.F",200,"")
      END IF
      !
      IF (ibas /= 0) THEN
         container(ibas)%basis_type = ""
         container(ibas)%basis_type_nr = 0
         IF (ASSOCIATED(container(ibas)%basis_set)) THEN
            CALL deallocate_gto_basis_set(container(ibas)%basis_set)
         END IF
      END IF
      ! shift other basis sets
      DO i = ibas+1, SIZE(container)
         IF (container(i)%basis_type_nr == 0) CYCLE
         container(i-1)%basis_type = container(i)%basis_type
         container(i-1)%basis_set => container(i)%basis_set
         container(i-1)%basis_type_nr = container(i)%basis_type_nr
      END DO

   END SUBROUTINE remove_basis_from_container

! **************************************************************************************************
!> \brief Retrieve a basis set from the container
!> \param container ...
!> \param basis_set ...
!> \param inumbas ...
!> \param basis_type ...
! **************************************************************************************************
   SUBROUTINE get_basis_from_container(container, basis_set, inumbas, basis_type)
      TYPE(basis_set_container_type), DIMENSION(:), &
         INTENT(inout)                                   :: container
      TYPE(gto_basis_set_type), POINTER                  :: basis_set
      INTEGER, OPTIONAL                                  :: inumbas
      CHARACTER(len=*), OPTIONAL                         :: basis_type

      CHARACTER(len=*), PARAMETER :: routineN = 'get_basis_from_container', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: basis_nr, i

      IF (PRESENT(inumbas)) THEN
         IF(.NOT.(inumbas <= SIZE(container)))CALL cp__a("aobasis/basis_set_container_types.F",240)
         IF(.NOT.(inumbas >= 1))CALL cp__a("aobasis/basis_set_container_types.F",241)
         basis_set => container(inumbas)%basis_set
         IF (PRESENT(basis_type)) THEN
            basis_type = container(inumbas)%basis_type
         END IF
      ELSE IF (PRESENT(basis_type)) THEN
         NULLIFY (basis_set)
         basis_nr = get_basis_type(basis_type)
         DO i = 1, SIZE(container)
            IF (container(i)%basis_type_nr == basis_nr) THEN
               basis_set => container(i)%basis_set
               EXIT
            END IF
         END DO
      ELSE
         CALL cp__b("aobasis/basis_set_container_types.F",256,"")
      END IF

   END SUBROUTINE get_basis_from_container
! **************************************************************************************************

END MODULE basis_set_container_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of the type to handle the 1/R^3 residual integral part
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
MODULE semi_empirical_expns3_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

! **************************************************************************************************
!> \brief 1/R^3 expansion type
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   TYPE semi_empirical_expns3_type
      REAL(KIND=dp)                                  :: core_core
      REAL(KIND=dp), DIMENSION(9)                    :: e1b, e2a
      REAL(KIND=dp), DIMENSION(81)                   :: w
   END TYPE semi_empirical_expns3_type

! **************************************************************************************************
!> \brief 1/R^3 expansion type: array of pointers
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   TYPE semi_empirical_expns3_p_type
      TYPE(semi_empirical_expns3_type), POINTER      :: expns3
   END TYPE semi_empirical_expns3_p_type

   ! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_expns3_types'

   PUBLIC :: semi_empirical_expns3_p_type, &
             semi_empirical_expns3_create, &
             semi_empirical_expns3_release

CONTAINS

! **************************************************************************************************
!> \brief Allocate semi-empirical 1/R^3 expansion type
!> \param expns3 ...
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   SUBROUTINE semi_empirical_expns3_create(expns3)
      TYPE(semi_empirical_expns3_type), POINTER          :: expns3

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_expns3_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(expns3)))CALL cp__a("semi_empirical_expns3_types.F",56)
      ALLOCATE (expns3)
      expns3%core_core = 0.0_dp
      expns3%e1b = 0.0_dp
      expns3%e2a = 0.0_dp
      expns3%w = 0.0_dp
   END SUBROUTINE semi_empirical_expns3_create

! **************************************************************************************************
!> \brief Deallocate the semi-empirical type
!> \param expns3 ...
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   SUBROUTINE semi_empirical_expns3_release(expns3)
      TYPE(semi_empirical_expns3_type), POINTER          :: expns3

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_expns3_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(expns3)) THEN
         DEALLOCATE (expns3)
      END IF
   END SUBROUTINE semi_empirical_expns3_release

END MODULE semi_empirical_expns3_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of the semi empirical multipole integral expansions types
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
MODULE semi_empirical_mpole_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_mpole_types'

! **************************************************************************************************
!> \brief Semi-empirical integral multipole expansion type
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   TYPE semi_empirical_mpole_type
      LOGICAL, DIMENSION(3)                    :: task
      INTEGER                                  :: indi, indj
      REAL(KIND=dp)                            :: c
      REAL(KIND=dp), DIMENSION(3)              :: d
      REAL(KIND=dp), DIMENSION(3, 3)            :: qc ! quadrupole cartesian
      REAL(KIND=dp), DIMENSION(5)              :: qs ! quadrupole spherical
      ! alternative definition used in GKS integral routines
      REAL(KIND=dp)                            :: cs
      REAL(KIND=dp), DIMENSION(3)              :: ds
      REAL(KIND=dp), DIMENSION(3, 3)            :: qq ! quadrupole cartesian
   END TYPE semi_empirical_mpole_type

! **************************************************************************************************
!> \brief Semi-empirical integral multipole expansion type - pointer type
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   TYPE semi_empirical_mpole_p_type
      TYPE(semi_empirical_mpole_type), POINTER :: mpole
   END TYPE semi_empirical_mpole_p_type

! **************************************************************************************************
!> \brief Global Multipolar NDDO information type
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   TYPE nddo_mpole_type
      REAL(KIND=dp), DIMENSION(:), POINTER     :: charge, efield0
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: dipole, efield1, efield2
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: quadrupole
   END TYPE nddo_mpole_type

   PUBLIC :: semi_empirical_mpole_type, &
             semi_empirical_mpole_p_type, &
             semi_empirical_mpole_p_create, &
             semi_empirical_mpole_p_release, &
             nddo_mpole_type, &
             nddo_mpole_create, &
             nddo_mpole_release

CONTAINS

! **************************************************************************************************
!> \brief Allocate semi-empirical mpole type
!> \param mpole ...
!> \param ndim ...
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   SUBROUTINE semi_empirical_mpole_p_create(mpole, ndim)
      TYPE(semi_empirical_mpole_p_type), DIMENSION(:), &
         POINTER                                         :: mpole
      INTEGER, INTENT(IN)                                :: ndim

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_mpole_p_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(.NOT. ASSOCIATED(mpole)))CALL cp__a("semi_empirical_mpole_types.F",84)
      ALLOCATE (mpole(ndim))
      DO i = 1, ndim
         NULLIFY (mpole(i)%mpole)
         CALL semi_empirical_mpole_create(mpole(i)%mpole)
      END DO

   END SUBROUTINE semi_empirical_mpole_p_create

! **************************************************************************************************
!> \brief Deallocate the semi-empirical mpole type
!> \param mpole ...
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   SUBROUTINE semi_empirical_mpole_p_release(mpole)
      TYPE(semi_empirical_mpole_p_type), DIMENSION(:), &
         POINTER                                         :: mpole

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_mpole_p_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(mpole)) THEN
         DO i = 1, SIZE(mpole)
            CALL semi_empirical_mpole_release(mpole(i)%mpole)
         END DO
         DEALLOCATE (mpole)
      END IF

   END SUBROUTINE semi_empirical_mpole_p_release

! **************************************************************************************************
!> \brief Allocate semi-empirical mpole type
!> \param mpole ...
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   SUBROUTINE semi_empirical_mpole_create(mpole)
      TYPE(semi_empirical_mpole_type), POINTER           :: mpole

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_mpole_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(mpole)))CALL cp__a("semi_empirical_mpole_types.F",127)
      ALLOCATE (mpole)
      mpole%task = .FALSE.
      mpole%indi = 0
      mpole%indj = 0
      mpole%c = HUGE(0.0_dp)
      mpole%d = HUGE(0.0_dp)
      mpole%qc = HUGE(0.0_dp)
      mpole%qs = HUGE(0.0_dp)
      mpole%cs = HUGE(0.0_dp)
      mpole%ds = HUGE(0.0_dp)
      mpole%qq = HUGE(0.0_dp)
   END SUBROUTINE semi_empirical_mpole_create

! **************************************************************************************************
!> \brief Deallocate the semi-empirical mpole type
!> \param mpole ...
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   SUBROUTINE semi_empirical_mpole_release(mpole)
      TYPE(semi_empirical_mpole_type), POINTER           :: mpole

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_mpole_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mpole)) THEN
         DEALLOCATE (mpole)
      END IF

   END SUBROUTINE semi_empirical_mpole_release

! **************************************************************************************************
!> \brief Allocate NDDO multipole type
!> \param nddo_mpole ...
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   SUBROUTINE nddo_mpole_create(nddo_mpole)
      TYPE(nddo_mpole_type), POINTER                     :: nddo_mpole

      CHARACTER(len=*), PARAMETER :: routineN = 'nddo_mpole_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(nddo_mpole)))CALL cp__a("semi_empirical_mpole_types.F",169)
      ALLOCATE (nddo_mpole)
      NULLIFY (nddo_mpole%charge)
      NULLIFY (nddo_mpole%dipole)
      NULLIFY (nddo_mpole%quadrupole)
      NULLIFY (nddo_mpole%efield0)
      NULLIFY (nddo_mpole%efield1)
      NULLIFY (nddo_mpole%efield2)
   END SUBROUTINE nddo_mpole_create

! **************************************************************************************************
!> \brief Deallocate NDDO multipole type
!> \param nddo_mpole ...
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
   SUBROUTINE nddo_mpole_release(nddo_mpole)
      TYPE(nddo_mpole_type), POINTER                     :: nddo_mpole

      CHARACTER(len=*), PARAMETER :: routineN = 'nddo_mpole_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(nddo_mpole)) THEN
         IF (ASSOCIATED(nddo_mpole%charge)) THEN
            DEALLOCATE (nddo_mpole%charge)
         END IF
         IF (ASSOCIATED(nddo_mpole%dipole)) THEN
            DEALLOCATE (nddo_mpole%dipole)
         END IF
         IF (ASSOCIATED(nddo_mpole%quadrupole)) THEN
            DEALLOCATE (nddo_mpole%quadrupole)
         END IF
         IF (ASSOCIATED(nddo_mpole%efield0)) THEN
            DEALLOCATE (nddo_mpole%efield0)
         END IF
         IF (ASSOCIATED(nddo_mpole%efield1)) THEN
            DEALLOCATE (nddo_mpole%efield1)
         END IF
         IF (ASSOCIATED(nddo_mpole%efield2)) THEN
            DEALLOCATE (nddo_mpole%efield2)
         END IF
         DEALLOCATE (nddo_mpole)
      END IF

   END SUBROUTINE nddo_mpole_release

END MODULE semi_empirical_mpole_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of the semi empirical parameter types.
!> \author Teodoro Laino [tlaino] - 10.2008 University of Zurich
! **************************************************************************************************
MODULE taper_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'taper_types'

! **************************************************************************************************
!> \brief Taper type
! **************************************************************************************************
   TYPE taper_type
      LOGICAL                               :: apply_taper
      REAL(KIND=dp)                         :: r0, rscale
   END TYPE taper_type

   PUBLIC :: taper_type, taper_create, taper_release, taper_eval, dtaper_eval

CONTAINS

! **************************************************************************************************
!> \brief Creates taper type
!> \param taper ...
!> \param rc ...
!> \param range ...
! **************************************************************************************************
   SUBROUTINE taper_create(taper, rc, range)
      TYPE(taper_type), POINTER                          :: taper
      REAL(KIND=dp), INTENT(IN)                          :: rc, range

      CHARACTER(len=*), PARAMETER :: routineN = 'taper_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(taper)))CALL cp__a("taper_types.F",47)
      ALLOCATE (taper)
      IF (range > EPSILON(0.0_dp)) THEN
         taper%apply_taper = .TRUE.
         IF(.NOT.(range > 0.0_dp))CALL cp__a("taper_types.F",51)
         taper%r0 = 2.0_dp*rc-20.0_dp*range
         taper%rscale = 1.0_dp/range
      ELSE
         taper%apply_taper = .FALSE.
      END IF

   END SUBROUTINE taper_create

! **************************************************************************************************
!> \brief Releases taper type
!> \param taper ...
! **************************************************************************************************
   SUBROUTINE taper_release(taper)
      TYPE(taper_type), POINTER                          :: taper

      CHARACTER(len=*), PARAMETER :: routineN = 'taper_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(taper)) THEN
         DEALLOCATE (taper)
      END IF
   END SUBROUTINE taper_release

! **************************************************************************************************
!> \brief Taper functions
!> \param taper ...
!> \param rij ...
!> \retval ft ...
! **************************************************************************************************
   FUNCTION taper_eval(taper, rij) RESULT(ft)
      TYPE(taper_type), POINTER                          :: taper
      REAL(KIND=dp), INTENT(IN)                          :: rij
      REAL(KIND=dp)                                      :: ft

      CHARACTER(len=*), PARAMETER :: routineN = 'taper_eval', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: dr

      ft = 1._dp
      IF (taper%apply_taper) THEN
         dr = taper%rscale*(rij-taper%r0)
         ft = 0.5_dp*(1.0_dp-TANH(dr))
      END IF
   END FUNCTION taper_eval

! **************************************************************************************************
!> \brief Analytical derivatives for taper function
!> \param taper ...
!> \param rij ...
!> \retval dft ...
! **************************************************************************************************
   FUNCTION dtaper_eval(taper, rij) RESULT(dft)
      TYPE(taper_type), POINTER                          :: taper
      REAL(KIND=dp), INTENT(IN)                          :: rij
      REAL(KIND=dp)                                      :: dft

      CHARACTER(len=*), PARAMETER :: routineN = 'dtaper_eval', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: dr

      dft = 0.0_dp
      IF (taper%apply_taper) THEN
         dr = taper%rscale*(rij-taper%r0)
         dft = -0.5_dp*(1.0_dp-TANH(dr)**2)*taper%rscale
      END IF
   END FUNCTION dtaper_eval

END MODULE taper_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! less not much meaningful...

! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_pw
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_pw'

!API type
   PUBLIC :: cp_sll_pw_type, cp_sll_pw_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_pw_create, cp_sll_pw_dealloc, &
             cp_sll_pw_next, &
             cp_sll_pw_get_length, &
             cp_sll_pw_get_el_at, cp_sll_pw_set_el_at, &
             cp_sll_pw_get_first_el, cp_sll_pw_get_rest, &
             cp_sll_pw_insert_el, cp_sll_pw_insert_el_at, &
             cp_sll_pw_rm_first_el, cp_sll_pw_rm_el_at, &
             cp_sll_pw_rm_all_el, &
             cp_sll_pw_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_pw_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_pw_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_pw_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_pw_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_pw_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_pw_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_pw_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_pw_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_pw_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_pw_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_pw_rm_first_el, &
         cp_sll_pw_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_pw_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_pw_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_pw_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_pw_type
      TYPE(pw_type), POINTER :: first_el
      TYPE(cp_sll_pw_type), POINTER :: rest
   END TYPE cp_sll_pw_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_pw_p_type
      TYPE(cp_sll_pw_type), POINTER :: list
   END TYPE cp_sll_pw_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_create(sll, first_el, rest)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      TYPE(pw_type), OPTIONAL, POINTER                   :: first_el
      TYPE(cp_sll_pw_type), OPTIONAL, POINTER            :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el => first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_pw_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_dealloc(sll)
      TYPE(cp_sll_pw_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_pw_rm_all_el(sll)
   END SUBROUTINE cp_sll_pw_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_dealloc_node(sll)
      TYPE(cp_sll_pw_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_pw_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_pw_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      TYPE(pw_type), POINTER                             :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_get_first_el', &
         routineP = moduleN//':'//routineN

      res => sll%first_el
   END FUNCTION cp_sll_pw_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_pw_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_pw_type), POINTER                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("pw/cp_linked_list_pw.F",295,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_pw_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_pw_get_length(sll) RESULT(res)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_pw_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_pw_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(pw_type), POINTER                             :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: pos

      IF (index == -1) THEN
         pos => cp_sll_pw_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_pw_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_pw.F",367)

      res => pos%first_el
   END FUNCTION cp_sll_pw_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_set_el_at(sll, index, value)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(pw_type), POINTER                             :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: pos

      IF (index == -1) THEN
         pos => cp_sll_pw_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_pw_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_pw.F",399)

      pos%first_el => value
   END SUBROUTINE cp_sll_pw_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_pw_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_pw_type), POINTER                      :: iterator
      TYPE(pw_type), OPTIONAL, POINTER                   :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_next', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att => iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_pw_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_insert_el(sll, el)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      TYPE(pw_type), POINTER                             :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_pw_create(newSlot, first_el=el, &
                            rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_pw_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_rm_first_el(sll)
      TYPE(cp_sll_pw_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_pw_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("pw/cp_linked_list_pw.F",484,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_pw_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_insert_el_at(sll, el, index)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      TYPE(pw_type), POINTER                             :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: pos

      IF (index == 1) THEN
         CALL cp_sll_pw_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_pw_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_pw_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_pw.F",518)
         CALL cp_sll_pw_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_pw_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_rm_el_at(sll, index)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: pos

      IF (index == 1) THEN
         CALL cp_sll_pw_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_pw_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_pw_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_pw.F",550)
         CALL cp_sll_pw_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_pw_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_pw_rm_all_el(sll)
      TYPE(cp_sll_pw_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_pw_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_pw_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_pw_to_array(sll) RESULT(res)
      TYPE(cp_sll_pw_type), POINTER                      :: sll
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_pw_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_pw_type), POINTER                      :: iter

      len = cp_sll_pw_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i)%pw => iter%first_el
         IF(.NOT.(cp_sll_pw_next(iter) .OR. i == len))CALL cp__a("pw/cp_linked_list_pw.F",606)
      END DO
   END FUNCTION cp_sll_pw_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "=>"
!  USE = "USE pw_types, ONLY: pw_type, pw_p_type"
!  array= = "=>"
!  arrayEl = "%pw"
!  common_dir = "../common"
!  defines =
!    "! less not much meaningful...
!     #define ( el1 %id_nr < el2 %id_nr ) ( el1 %id_nr < el2 %id_nr )
!     #define ( el1 %id_nr == el2 %id_nr ) ( el1 %id_nr == el2 %id_nr )
!     "
!  equalQ = "CP_SLL_PW_EQUAL_Q"
!  lessQ = "CP_SLL_PW_LESS_Q"
!  nametype1 = "pw"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "type(pw_type),pointer"
!  type1arrayEl = "type(pw_p_type)"
!  type1in = "type(pw_type), pointer"
!  type1out = "type(pw_type), pointer"
!  write_el = ""

END MODULE cp_linked_list_pw

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief various utilities that regard array of different kinds:
!>      output, allocation,...
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!>
!>      maybe it is not a good idea mixing output and memeory utils...
!> \par History
!>      12.2001 first version [fawzi]
!>      3.2002 templatized [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_array_utils_r
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_array_utils_r'

   ! types
   PUBLIC :: cp_1d_r_p_type, cp_2d_r_p_type, cp_3d_r_p_type

   ! generic interfaces
   PUBLIC :: cp_guarantee_size

   ! the underlying procedures
   PUBLIC :: cp_1d_r_guarantee_size, &
             cp_1d_r_write, cp_2d_r_write, &
             cp_2d_r_guarantee_size, &
             cp_1d_r_bsearch

   INTERFACE cp_guarantee_size
      MODULE PROCEDURE cp_1d_r_guarantee_size, &
         cp_2d_r_guarantee_size
   END INTERFACE

!***

! **************************************************************************************************
!> \brief represent a pointer to a 1d array
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_1d_r_p_type
      REAL(kind=dp), DIMENSION(:), POINTER :: array
   END TYPE cp_1d_r_p_type

! **************************************************************************************************
!> \brief represent a pointer to a 2d array
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_2d_r_p_type
      REAL(kind=dp), DIMENSION(:, :), POINTER :: array
   END TYPE cp_2d_r_p_type

! **************************************************************************************************
!> \brief represent a pointer to a 3d array
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_3d_r_p_type
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: array
   END TYPE cp_3d_r_p_type

CONTAINS

! **************************************************************************************************
!> \brief writes an array to the given unit
!> \param array the array to write
!> \param unit_nr the unit to write to (defaults to the standard out)
!> \param el_format the format of a single element
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      maybe I will move to a comma separated paretized list
! **************************************************************************************************
   SUBROUTINE cp_1d_r_write(array, unit_nr, el_format)
      REAL(kind=dp), INTENT(in)                          :: array(:)
      INTEGER, INTENT(in)                                :: unit_nr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: el_format

      CHARACTER(len=*), PARAMETER :: defaultFormat = "(es14.6)", routineN = 'cp_1d_r_write', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iostat

      WRITE (unit=unit_nr, fmt="('( ')", advance="no", iostat=iostat)
      IF(.NOT.(iostat == 0))CALL cp__a("common/cp_array_utils_r.F",126)
      IF (PRESENT(el_format)) THEN
         DO i = 1, SIZE(array)-1
            WRITE (unit=unit_nr, fmt=el_format, advance="no") array(i)
            IF (MOD(i, 5) .EQ. 0) THEN ! only a few elements per line
               WRITE (unit=unit_nr, fmt="(',')")
            ELSE
               WRITE (unit=unit_nr, fmt="(',')", advance="no")
            ENDIF
         END DO
         IF (SIZE(array) > 0) &
            WRITE (unit=unit_nr, fmt=el_format, advance="no") array(SIZE(array))
      ELSE
         DO i = 1, SIZE(array)-1
            WRITE (unit=unit_nr, fmt=defaultFormat, advance="no") array(i)
            IF (MOD(i, 5) .EQ. 0) THEN ! only a few elements per line
               WRITE (unit=unit_nr, fmt="(',')")
            ELSE
               WRITE (unit=unit_nr, fmt="(',')", advance="no")
            ENDIF
         END DO
         IF (SIZE(array) > 0) &
            WRITE (unit=unit_nr, fmt=defaultFormat, advance="no") array(SIZE(array))
      END IF
      WRITE (unit=unit_nr, fmt="(' )')")
      CALL m_flush(unit_nr)

   END SUBROUTINE cp_1d_r_write

! **************************************************************************************************
!> \brief writes an array to the given unit
!> \param array the array to write
!> \param unit_nr the unit to write to (defaults to the standard out)
!> \param el_format the format of a single element
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      maybe I will move to a comma separated parentized list
! **************************************************************************************************
   SUBROUTINE cp_2d_r_write(array, unit_nr, el_format)
      REAL(kind=dp), INTENT(in)                          :: array(:, :)
      INTEGER, INTENT(in)                                :: unit_nr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: el_format

      CHARACTER(len=*), PARAMETER :: defaultFormat = "(es14.6)", routineN = 'cp_2d_r_write', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=10)                                  :: nRiga
      CHARACTER(len=200)                                 :: fmtstr
      INTEGER                                            :: i, iostat

      nRiga = cp_to_string(SIZE(array, 2))
      DO i = 1, SIZE(array, 1)
         IF (PRESENT(el_format)) THEN
            fmtstr = '(" ",'//nRiga//el_format//')'
            WRITE (unit=unit_nr, fmt=fmtstr, iostat=iostat) array(i, :)
         ELSE
            fmtstr = '(" ",'//nRiga//defaultFormat//')'
            WRITE (unit=unit_nr, fmt=fmtstr, iostat=iostat) array(i, :)
         END IF
         IF(.NOT.(iostat == 0))CALL cp__a("common/cp_array_utils_r.F",187)
      END DO
      CALL m_flush(unit_nr)
   END SUBROUTINE cp_2d_r_write

! **************************************************************************************************
!> \brief If the size of the array is changes reallocate it.
!>      Issues a warning when the size changes (but not on allocation
!>      and deallocation).
!>
!>      The data is NOT preserved (if you want to preserve the data see
!>      the realloc in the module memory_utilities)
!> \param array the array to reallocate if necessary
!> \param n the wanted size
!> \par History
!>      12.2001 first version [fawzi]
!>      3.2002 templatized [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      this is a different behaviour than the realloc in the module
!>      memory_utilities. It is quite low level
! **************************************************************************************************
   SUBROUTINE cp_1d_r_guarantee_size(array, n)
      REAL(kind=dp), POINTER                             :: array(:)
      INTEGER, INTENT(in)                                :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_1d_r_guarantee_size', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(n >= 0))CALL cp__a("common/cp_array_utils_r.F",216)
      IF (ASSOCIATED(array)) THEN
         IF (SIZE(array) /= n) THEN
            CALL cp__w("common/cp_array_utils_r.F",219,'size has changed')
            DEALLOCATE (array)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(array)) THEN
         ALLOCATE (array(n))
      END IF
   END SUBROUTINE cp_1d_r_guarantee_size

! **************************************************************************************************
!> \brief If the size of the array is changes reallocate it.
!>      Issues a warning when the size changes (but not on allocation
!>      and deallocation).
!>
!>      The data is NOT preserved (if you want to preserve the data see
!>      the realloc in the module memory_utilities)
!> \param array the array to reallocate if necessary
!> \param n_rows the wanted number of rows
!> \param n_cols the wanted number of cols
!> \par History
!>      5.2001 first version [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      this is a different behaviour than the realloc in the module
!>      memory_utilities. It is quite low level
! **************************************************************************************************
   SUBROUTINE cp_2d_r_guarantee_size(array, n_rows, n_cols)
      REAL(kind=dp), POINTER                             :: array(:, :)
      INTEGER, INTENT(in)                                :: n_rows, n_cols

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_2d_r_guarantee_size', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(n_cols >= 0))CALL cp__a("common/cp_array_utils_r.F",252)
      IF(.NOT.(n_rows >= 0))CALL cp__a("common/cp_array_utils_r.F",253)
      IF (ASSOCIATED(array)) THEN
         IF (SIZE(array, 1) /= n_rows .OR. SIZE(array, 2) /= n_cols) THEN
            CALL cp__w("common/cp_array_utils_r.F",256,'size has changed')
            DEALLOCATE (array)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(array)) THEN
         ALLOCATE (array(n_rows, n_cols))
      END IF
   END SUBROUTINE cp_2d_r_guarantee_size

! **************************************************************************************************
!> \brief returns the index at which the element el should be inserted in the
!>      array to keep it ordered (array(i)>=el).
!>      If the element is bigger than all the elements in the array returns
!>      the last index+1.
!> \param array the array to search
!> \param el the element to look for
!> \param l_index the lower index for binary search (defaults to 1)
!> \param u_index the upper index for binary search (defaults to size(array))
!> \retval res ...
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      the array should be ordered in growing order
! **************************************************************************************************
   FUNCTION cp_1d_r_bsearch(array, el, l_index, u_index) &
      RESULT(res)
      REAL(kind=dp), INTENT(in)                          :: array(:), el
      INTEGER, INTENT(in), OPTIONAL                      :: l_index, u_index
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_1d_r_bsearch', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: aindex, lindex, uindex

      lindex = 1
      uindex = SIZE(array)
      IF (PRESENT(l_index)) lindex = l_index
      IF (PRESENT(u_index)) uindex = u_index
      DO WHILE (lindex <= uindex)
         aindex = (lindex+uindex)/2
         IF (( array(aindex) <  el )) THEN
            lindex = aindex+1
         ELSE
            uindex = aindex-1
         END IF
      END DO
      res = lindex
   END FUNCTION cp_1d_r_bsearch

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,defaultFormatType1,lessQ,defines)]
! ARGS:
!  USE = "USE kinds, only: dp"
!  defaultFormatType1 = ""(es14.6)""
!  defines =
!    "#define CP_R_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_R_LESS_Q"
!  nametype1 = "r"
!  type1 = "REAL(kind=dp)"

END MODULE cp_array_utils_r
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! less not much meningful... and defines oly a partial ordering.

! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_3d_r
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_3d_r'

!API type
   PUBLIC :: cp_sll_3d_r_type, cp_sll_3d_r_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_3d_r_create, cp_sll_3d_r_dealloc, &
             cp_sll_3d_r_next, &
             cp_sll_3d_r_get_length, &
             cp_sll_3d_r_get_el_at, cp_sll_3d_r_set_el_at, &
             cp_sll_3d_r_get_first_el, cp_sll_3d_r_get_rest, &
             cp_sll_3d_r_insert_el, cp_sll_3d_r_insert_el_at, &
             cp_sll_3d_r_rm_first_el, cp_sll_3d_r_rm_el_at, &
             cp_sll_3d_r_rm_all_el, &
             cp_sll_3d_r_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_3d_r_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_3d_r_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_3d_r_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_3d_r_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_3d_r_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_3d_r_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_3d_r_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_3d_r_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_3d_r_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_3d_r_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_3d_r_rm_first_el, &
         cp_sll_3d_r_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_3d_r_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_3d_r_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_3d_r_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_3d_r_type
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: first_el
      TYPE(cp_sll_3d_r_type), POINTER :: rest
   END TYPE cp_sll_3d_r_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_3d_r_p_type
      TYPE(cp_sll_3d_r_type), POINTER :: list
   END TYPE cp_sll_3d_r_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_create(sll, first_el, rest)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      REAL(kind=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: first_el
      TYPE(cp_sll_3d_r_type), OPTIONAL, POINTER          :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el => first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_3d_r_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_dealloc(sll)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_3d_r_rm_all_el(sll)
   END SUBROUTINE cp_sll_3d_r_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_dealloc_node(sll)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_3d_r_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_3d_r_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_get_first_el', &
         routineP = moduleN//':'//routineN

      res => sll%first_el
   END FUNCTION cp_sll_3d_r_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_3d_r_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_3d_r_type), POINTER                    :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("pw/cp_linked_list_3d_r.F",296,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_3d_r_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_3d_r_get_length(sll) RESULT(res)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_3d_r_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_3d_r_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: pos

      IF (index == -1) THEN
         pos => cp_sll_3d_r_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_3d_r_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_3d_r.F",368)

      res => pos%first_el
   END FUNCTION cp_sll_3d_r_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_set_el_at(sll, index, value)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: pos

      IF (index == -1) THEN
         pos => cp_sll_3d_r_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_3d_r_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_3d_r.F",400)

      pos%first_el => value
   END SUBROUTINE cp_sll_3d_r_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_3d_r_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_3d_r_type), POINTER                    :: iterator
      REAL(kind=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_next', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att => iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_3d_r_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_insert_el(sll, el)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_3d_r_create(newSlot, first_el=el, &
                              rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_3d_r_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_rm_first_el(sll)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_3d_r_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("pw/cp_linked_list_3d_r.F",487,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_3d_r_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_insert_el_at(sll, el, index)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: pos

      IF (index == 1) THEN
         CALL cp_sll_3d_r_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_3d_r_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_3d_r_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_3d_r.F",521)
         CALL cp_sll_3d_r_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_3d_r_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_rm_el_at(sll, index)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: pos

      IF (index == 1) THEN
         CALL cp_sll_3d_r_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_3d_r_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_3d_r_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_3d_r.F",553)
         CALL cp_sll_3d_r_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_3d_r_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_3d_r_rm_all_el(sll)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_3d_r_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_3d_r_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_3d_r_to_array(sll) RESULT(res)
      TYPE(cp_sll_3d_r_type), POINTER                    :: sll
      TYPE(cp_3d_r_p_type), DIMENSION(:), POINTER        :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_3d_r_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_3d_r_type), POINTER                    :: iter

      len = cp_sll_3d_r_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i)%array => iter%first_el
         IF(.NOT.(cp_sll_3d_r_next(iter) .OR. i == len))CALL cp__a("pw/cp_linked_list_3d_r.F",609)
      END DO
   END FUNCTION cp_sll_3d_r_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "=>"
!  USE =
!    "USE kinds, only: dp
!     USE cp_array_utils_r, only: cp_3d_r_p_type"
!  array= = "=>"
!  arrayEl = "%array"
!  common_dir = "../common"
!  defines =
!    "! less not much meningful... and defines oly a partial ordering.
!     #define ( ALL(el1 < el2) ) ( all(el1 < el2) )
!     #define ( ALL(el1 == el2) ) ( all(el1 == el2) )
!     "
!  equalQ = "CP_SLL_R_EQUAL_Q"
!  lessQ = "CP_SLL_R_LESS_Q"
!  nametype1 = "3d_r"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "REAL(kind=dp),dimension(:,:,:),pointer"
!  type1arrayEl = "type(cp_3d_r_p_type)"
!  type1in = "REAL(kind=dp),dimension(:,:,:), pointer"
!  type1out = "REAL(kind=dp),dimension(:,:,:), pointer"
!  write_el = ""

END MODULE cp_linked_list_3d_r

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \note
!> This module contains routines necessary to operate on plane waves on NVIDIA
!> GPUs using CUDA. It depends at execution time on NVIDIA's CUFFT library.
!> \par History
!>      BGL (06-Mar-2008)  : Created
!>      AG  (18-May-2012)  : Refacturing:
!>                           - added explicit interfaces to C routines
!>                           - enable double precision complex transformations
!>      AG  (11-Sept-2012) : Modifications:
!>                          - use pointers if precision mapping is not required
!>                          - use OMP for mapping
!> \author Benjamin G. Levine
! **************************************************************************************************
MODULE pw_cuda
   USE ISO_C_BINDING,                   ONLY: C_DOUBLE,&
                                              C_INT,&
                                              C_LOC,&
                                              C_PTR
   USE fast,                            ONLY: zero_c
   USE fft_tools,                       ONLY: &
        cube_transpose_1, cube_transpose_2, fft_scratch_sizes, fft_scratch_type, get_fft_scratch, &
        release_fft_scratch, x_to_yz, xz_to_yz, yz_to_x, yz_to_xz
   USE kinds,                           ONLY: dp,&
                                              int_size
   USE message_passing,                 ONLY: mp_comm_compare,&
                                              mp_environ,&
                                              mp_rank_compare
   USE pw_grid_types,                   ONLY: FULLSPACE
   USE pw_types,                        ONLY: REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pw_cuda_r3dc1d_3d
   PUBLIC :: pw_cuda_c1dr3d_3d
   PUBLIC :: pw_cuda_r3dc1d_3d_ps
   PUBLIC :: pw_cuda_c1dr3d_3d_ps
   PUBLIC :: pw_cuda_init, pw_cuda_finalize

   ! Explicit interfaces to double precision complex transformation
   ! routines (C+CUDA). For more details, see: ../cuda_tools!
   INTERFACE pw_cuda_cfffg_cu
! **************************************************************************************************
!> \brief ...
!> \param din ...
!> \param zout ...
!> \param ghatmap ...
!> \param npts ...
!> \param ngpts ...
!> \param scale ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_cfffg_z(din, zout, ghatmap, npts, ngpts, scale) &
         BIND(C, name="pw_cuda_cfffg_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: din
         TYPE(C_PTR), VALUE                       :: zout
         TYPE(C_PTR), INTENT(IN), VALUE           :: ghatmap
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts
         INTEGER(KIND=C_INT), INTENT(IN), VALUE   :: ngpts
         REAL(KIND=C_DOUBLE), INTENT(IN), VALUE   :: scale

      END SUBROUTINE pw_cuda_cfffg_z
   END INTERFACE

   INTERFACE pw_cuda_sfffc_cu
! **************************************************************************************************
!> \brief ...
!> \param zin ...
!> \param dout ...
!> \param ghatmap ...
!> \param npts ...
!> \param ngpts ...
!> \param nmaps ...
!> \param scale ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_sfffc_z(zin, dout, ghatmap, npts, ngpts, nmaps, scale) &
         BIND(C, name="pw_cuda_sfffc_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: zin
         TYPE(C_PTR), VALUE                       :: dout
         TYPE(C_PTR), INTENT(IN), VALUE           :: ghatmap
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts
         INTEGER(KIND=C_INT), INTENT(IN), VALUE   :: ngpts, nmaps
         REAL(KIND=C_DOUBLE), INTENT(IN), VALUE   :: scale

      END SUBROUTINE pw_cuda_sfffc_z
   END INTERFACE

   INTERFACE pw_cuda_cff_cu
! **************************************************************************************************
!> \brief ...
!> \param din ...
!> \param zout ...
!> \param npts ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_cff_z(din, zout, npts) &
         BIND(C, name="pw_cuda_cff_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: din
         TYPE(C_PTR), VALUE                       :: zout
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts

      END SUBROUTINE pw_cuda_cff_z
   END INTERFACE

   INTERFACE pw_cuda_ffc_cu
! **************************************************************************************************
!> \brief ...
!> \param zin ...
!> \param dout ...
!> \param npts ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_ffc_z(zin, dout, npts) &
         BIND(C, name="pw_cuda_ffc_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: zin
         TYPE(C_PTR), VALUE                       :: dout
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts

      END SUBROUTINE pw_cuda_ffc_z
   END INTERFACE

   INTERFACE pw_cuda_cf_cu
! **************************************************************************************************
!> \brief ...
!> \param din ...
!> \param zout ...
!> \param npts ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_cf_z(din, zout, npts) &
         BIND(C, name="pw_cuda_cf_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: din
         TYPE(C_PTR), VALUE                       :: zout
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts

      END SUBROUTINE pw_cuda_cf_z
   END INTERFACE

   INTERFACE pw_cuda_fc_cu
! **************************************************************************************************
!> \brief ...
!> \param zin ...
!> \param dout ...
!> \param npts ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_fc_z(zin, dout, npts) &
         BIND(C, name="pw_cuda_fc_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: zin
         TYPE(C_PTR), VALUE                       :: dout
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts

      END SUBROUTINE pw_cuda_fc_z
   END INTERFACE

   INTERFACE pw_cuda_f_cu
! **************************************************************************************************
!> \brief ...
!> \param zin ...
!> \param zout ...
!> \param dir ...
!> \param n ...
!> \param m ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_f_z(zin, zout, dir, n, m) &
         BIND(C, name="pw_cuda_f_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: zin
         TYPE(C_PTR), VALUE                       :: zout
         INTEGER(KIND=C_INT), INTENT(IN), VALUE   :: dir, n, m

      END SUBROUTINE pw_cuda_f_z
   END INTERFACE

   INTERFACE pw_cuda_fg_cu
! **************************************************************************************************
!> \brief ...
!> \param zin ...
!> \param zout ...
!> \param ghatmap ...
!> \param npts ...
!> \param mmax ...
!> \param ngpts ...
!> \param scale ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_fg_z(zin, zout, ghatmap, npts, mmax, ngpts, scale) &
         BIND(C, name="pw_cuda_fg_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: zin
         TYPE(C_PTR), VALUE                       :: zout
         TYPE(C_PTR), INTENT(IN), VALUE           :: ghatmap
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts
         INTEGER(KIND=C_INT), INTENT(IN), VALUE   :: mmax, ngpts
         REAL(KIND=C_DOUBLE), INTENT(IN), VALUE   :: scale

      END SUBROUTINE pw_cuda_fg_z
   END INTERFACE

   INTERFACE pw_cuda_sf_cu
! **************************************************************************************************
!> \brief ...
!> \param zin ...
!> \param zout ...
!> \param ghatmap ...
!> \param npts ...
!> \param mmax ...
!> \param ngpts ...
!> \param nmaps ...
!> \param scale ...
! **************************************************************************************************
      SUBROUTINE pw_cuda_sf_z(zin, zout, ghatmap, npts, mmax, ngpts, nmaps, scale) &
         BIND(C, name="pw_cuda_sf_z_")
         IMPORT
         TYPE(C_PTR), INTENT(IN), VALUE           :: zin
         TYPE(C_PTR), VALUE                       :: zout
         TYPE(C_PTR), INTENT(IN), VALUE           :: ghatmap
         INTEGER(KIND=C_INT), DIMENSION(*), &
            INTENT(IN)                             :: npts
         INTEGER(KIND=C_INT), INTENT(IN), VALUE   :: mmax, ngpts, nmaps
         REAL(KIND=C_DOUBLE), INTENT(IN), VALUE   :: scale

      END SUBROUTINE pw_cuda_sf_z
   END INTERFACE

   INTERFACE
      FUNCTION pw_cuda_init_cu() RESULT(istat) BIND(C, name="pw_cuda_init")
         IMPORT
         INTEGER(KIND=C_INT)                    :: istat
      END FUNCTION pw_cuda_init_cu
      SUBROUTINE pw_cuda_finalize_cu() BIND(C, name="pw_cuda_finalize")
      END SUBROUTINE pw_cuda_finalize_cu
   END INTERFACE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_methods_cuda'
   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief Allocates resources on the cuda device for cuda fft acceleration
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE pw_cuda_init()
   END SUBROUTINE pw_cuda_init

! **************************************************************************************************
!> \brief Releases resources on the cuda device for cuda fft acceleration
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE pw_cuda_finalize()
   END SUBROUTINE pw_cuda_finalize

! **************************************************************************************************
!> \brief perform an fft followed by a gather on the gpu
!> \param pw1 ...
!> \param pw2 ...
!> \param scale ...
!> \author Benjamin G Levine
! **************************************************************************************************
   SUBROUTINE pw_cuda_r3dc1d_3d(pw1, pw2, scale)
      TYPE(pw_type), TARGET, INTENT(IN)        :: pw1
      TYPE(pw_type), TARGET, INTENT(INOUT)     :: pw2
      REAL(KIND=dp)                            :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_r3dc1d_3d', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw2))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_r3dc1d_3d

! **************************************************************************************************
!> \brief perform an scatter followed by a fft on the gpu
!> \param pw1 ...
!> \param pw2 ...
!> \param scale ...
!> \author Benjamin G Levine
! **************************************************************************************************
   SUBROUTINE pw_cuda_c1dr3d_3d(pw1, pw2, scale)
      TYPE(pw_type), TARGET, INTENT(IN)        :: pw1
      TYPE(pw_type), TARGET, INTENT(INOUT)     :: pw2
      REAL(KIND=dp)                            :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_c1dr3d_3d', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw2))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_c1dr3d_3d

! **************************************************************************************************
!> \brief perform an parallel fft followed by a gather on the gpu
!> \param pw1 ...
!> \param pw2 ...
!> \param scale ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_r3dc1d_3d_ps(pw1, pw2, scale)
      TYPE(pw_type), TARGET, INTENT(IN)        :: pw1
      TYPE(pw_type), TARGET, INTENT(INOUT)     :: pw2
      REAL(KIND=dp)                            :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_r3dc1d_3d_ps', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw2))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_r3dc1d_3d_ps

! **************************************************************************************************
!> \brief perform an parallel scatter followed by a fft on the gpu
!> \param pw1 ...
!> \param pw2 ...
!> \param scale ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_c1dr3d_3d_ps(pw1, pw2, scale)
      TYPE(pw_type), TARGET, INTENT(IN)        :: pw1
      TYPE(pw_type), TARGET, INTENT(INOUT)     :: pw2
      REAL(KIND=dp)                            :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_c1dr3d_3d_ps', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw2))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_c1dr3d_3d_ps

! **************************************************************************************************
!> \brief perform a parallel real_to_complex copy followed by a 2D-FFT on the gpu
!> \param pw1 ...
!> \param pwbuf ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_cff(pw1, pwbuf)
      TYPE(pw_type), TARGET, INTENT(IN)        :: pw1
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         POINTER, INTENT(INOUT)                 :: pwbuf

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_cff', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_cff

! **************************************************************************************************
!> \brief perform a parallel 2D-FFT followed by a complex_to_real copy on the gpu
!> \param pwbuf ...
!> \param pw2 ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_ffc(pwbuf, pw2)
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         POINTER, INTENT(IN)                    :: pwbuf
      TYPE(pw_type), TARGET, INTENT(INOUT)     :: pw2

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_ffc', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw2))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_ffc

! **************************************************************************************************
!> \brief perform a parallel real_to_complex copy followed by a 1D-FFT on the gpu
!> \param pw1 ...
!> \param pwbuf ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_cf(pw1, pwbuf)
      TYPE(pw_type), TARGET, INTENT(IN)        :: pw1
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER, INTENT(INOUT)                 :: pwbuf

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_cf', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_cf

! **************************************************************************************************
!> \brief perform a parallel 1D-FFT followed by a complex_to_real copy on the gpu
!> \param pwbuf ...
!> \param pw2 ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_fc(pwbuf, pw2)
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER, INTENT(IN)                    :: pwbuf
      TYPE(pw_type), TARGET, INTENT(INOUT)     :: pw2

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_fc', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw2))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_fc

! **************************************************************************************************
!> \brief perform a parallel 1D-FFT on the gpu
!> \param pwbuf1 ...
!> \param pwbuf2 ...
!> \param dir ...
!> \param n ...
!> \param m ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_f(pwbuf1, pwbuf2, dir, n, m)
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER, INTENT(IN)                    :: pwbuf1
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER, INTENT(INOUT)                 :: pwbuf2
      INTEGER, INTENT(IN)                      :: dir
      INTEGER, INTENT(IN)                      :: n
      INTEGER, INTENT(IN)                      :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_f', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf2))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dir))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(m))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_f
! **************************************************************************************************
!> \brief perform a parallel 1D-FFT followed by a gather on the gpu
!> \param pwbuf ...
!> \param pw2 ...
!> \param scale ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_fg(pwbuf, pw2, scale)
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER, INTENT(IN)                    :: pwbuf
      TYPE(pw_type), TARGET, INTENT(INOUT)     :: pw2
      REAL(KIND=dp), INTENT(IN)                :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_fg', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw2))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_fg

! **************************************************************************************************
!> \brief perform a parallel scatter followed by a 1D-FFT on the gpu
!> \param pw1 ...
!> \param pwbuf ...
!> \param scale ...
!> \author Andreas Gloess
! **************************************************************************************************
   SUBROUTINE pw_cuda_sf(pw1, pwbuf, scale)
      TYPE(pw_type), TARGET, INTENT(IN)        :: pw1
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER, INTENT(INOUT)                 :: pwbuf
      REAL(KIND=dp), INTENT(IN)                :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_cuda_sf', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pw1))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pwbuf))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE pw_cuda_sf
END MODULE pw_cuda

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief sums arrays of real/complex numbers with *much* reduced round-off as compared to
!>      a naive implementation (or the one found in most compiler's SUM intrinsic)
!>      using an implementation of Kahan's algorithm for summing real numbers
!>      that can be used instead of the standard Fortran SUM(array[,mask]).
!>
!>      see also http://en.wikipedia.org/wiki/Kahan_summation_algorithm
!> \note
!>      if the compiler optimises away the 'tricky' bit, no accuracy is gained,
!>      if the compiler uses extended precision inconsistently even worse results might be obtained.
!>      This has not been observed.
!>      This algorithm is not fast, and thus not recommended for cases where round-off is not a
!>      concern but performance is.
!>
!>      the standard intrinsic sum can be 'replaced' using the following use statement
!>
!>      USE kahan_sum, ONLY: sum => kahan_sum
!> \par History
!>      03.2006 [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE kahan_sum

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: accurate_sum
   INTEGER, PARAMETER :: sp = KIND(0.0), dp = KIND(0.0D0)
   REAL(KIND=sp), PARAMETER :: szero = 0.0_sp
   REAL(KIND=dp), PARAMETER :: dzero = 0.0_dp
   COMPLEX(KIND=sp), PARAMETER :: czero = (0.0_sp, 0.0_sp)
   COMPLEX(KIND=dp), PARAMETER :: zzero = (0.0_dp, 0.0_dp)
   INTERFACE accurate_sum
      MODULE PROCEDURE &
         kahan_sum_s1, kahan_sum_d1, kahan_sum_c1, kahan_sum_z1, &
         kahan_sum_s2, kahan_sum_d2, kahan_sum_c2, kahan_sum_z2, &
         kahan_sum_s3, kahan_sum_d3, kahan_sum_c3, kahan_sum_z3, &
         kahan_sum_s4, kahan_sum_d4, kahan_sum_c4, kahan_sum_z4, &
         kahan_sum_s5, kahan_sum_d5, kahan_sum_c5, kahan_sum_z5, &
         kahan_sum_s6, kahan_sum_d6, kahan_sum_c6, kahan_sum_z6, &
         kahan_sum_s7, kahan_sum_d7, kahan_sum_c7, kahan_sum_z7
   END INTERFACE accurate_sum
CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_s1(array, mask) RESULT(ks)
      REAL(KIND=sp), DIMENSION(:), INTENT(IN)            :: array
      LOGICAL, DIMENSION(:), INTENT(IN), OPTIONAL        :: mask
      REAL(KIND=sp)                                      :: ks

      INTEGER                                            :: i1
      REAL(KIND=sp)                                      :: c, t, y

      ks = szero; t = szero; y = szero; c = szero

      IF (PRESENT(mask)) THEN
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1)) THEN
               y = array(i1)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
      ELSE
         DO i1 = 1, SIZE(array, 1)
            y = array(i1)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_s1

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_d1(array, mask) RESULT(ks)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: array
      LOGICAL, DIMENSION(:), INTENT(IN), OPTIONAL        :: mask
      REAL(KIND=dp)                                      :: ks

      INTEGER                                            :: i1
      REAL(KIND=dp)                                      :: c, t, y

      ks = dzero; t = dzero; y = dzero; c = dzero

      IF (PRESENT(mask)) THEN
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1)) THEN
               y = array(i1)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
      ELSE
         DO i1 = 1, SIZE(array, 1)
            y = array(i1)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_d1

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_c1(array, mask) RESULT(ks)
      COMPLEX(KIND=sp), DIMENSION(:), INTENT(IN)         :: array
      LOGICAL, DIMENSION(:), INTENT(IN), OPTIONAL        :: mask
      COMPLEX(KIND=sp)                                   :: ks

      COMPLEX(KIND=sp)                                   :: c, t, y
      INTEGER                                            :: i1

      ks = czero; t = czero; y = czero; c = czero

      IF (PRESENT(mask)) THEN
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1)) THEN
               y = array(i1)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
      ELSE
         DO i1 = 1, SIZE(array, 1)
            y = array(i1)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_c1

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_z1(array, mask) RESULT(ks)
      COMPLEX(KIND=dp), DIMENSION(:), INTENT(IN)         :: array
      LOGICAL, DIMENSION(:), INTENT(IN), OPTIONAL        :: mask
      COMPLEX(KIND=dp)                                   :: ks

      COMPLEX(KIND=dp)                                   :: c, t, y
      INTEGER                                            :: i1

      ks = zzero; t = zzero; y = zzero; c = zzero

      IF (PRESENT(mask)) THEN
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1)) THEN
               y = array(i1)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
      ELSE
         DO i1 = 1, SIZE(array, 1)
            y = array(i1)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_z1

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_s2(array, mask) RESULT(ks)
      REAL(KIND=sp), DIMENSION(:, :), INTENT(IN)         :: array
      LOGICAL, DIMENSION(:, :), INTENT(IN), OPTIONAL     :: mask
      REAL(KIND=sp)                                      :: ks

      INTEGER                                            :: i1, i2
      REAL(KIND=sp)                                      :: c, t, y

      ks = szero; t = szero; y = szero; c = szero

      IF (PRESENT(mask)) THEN
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2)) THEN
               y = array(i1, i2)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
      ELSE
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_s2

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_d2(array, mask) RESULT(ks)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: array
      LOGICAL, DIMENSION(:, :), INTENT(IN), OPTIONAL     :: mask
      REAL(KIND=dp)                                      :: ks

      INTEGER                                            :: i1, i2
      REAL(KIND=dp)                                      :: c, t, y

      ks = dzero; t = dzero; y = dzero; c = dzero

      IF (PRESENT(mask)) THEN
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2)) THEN
               y = array(i1, i2)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
      ELSE
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_d2

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_c2(array, mask) RESULT(ks)
      COMPLEX(KIND=sp), DIMENSION(:, :), INTENT(IN)      :: array
      LOGICAL, DIMENSION(:, :), INTENT(IN), OPTIONAL     :: mask
      COMPLEX(KIND=sp)                                   :: ks

      COMPLEX(KIND=sp)                                   :: c, t, y
      INTEGER                                            :: i1, i2

      ks = czero; t = czero; y = czero; c = czero

      IF (PRESENT(mask)) THEN
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2)) THEN
               y = array(i1, i2)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
      ELSE
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_c2

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_z2(array, mask) RESULT(ks)
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: array
      LOGICAL, DIMENSION(:, :), INTENT(IN), OPTIONAL     :: mask
      COMPLEX(KIND=dp)                                   :: ks

      COMPLEX(KIND=dp)                                   :: c, t, y
      INTEGER                                            :: i1, i2

      ks = zzero; t = zzero; y = zzero; c = zzero

      IF (PRESENT(mask)) THEN
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2)) THEN
               y = array(i1, i2)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
      ELSE
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_z2

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_s3(array, mask) RESULT(ks)
      REAL(KIND=sp), DIMENSION(:, :, :), INTENT(IN)      :: array
      LOGICAL, DIMENSION(:, :, :), INTENT(IN), OPTIONAL  :: mask
      REAL(KIND=sp)                                      :: ks

      INTEGER                                            :: i1, i2, i3
      REAL(KIND=sp)                                      :: c, t, y

      ks = szero; t = szero; y = szero; c = szero

      IF (PRESENT(mask)) THEN
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3)) THEN
               y = array(i1, i2, i3)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_s3

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_d3(array, mask) RESULT(ks)
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: array
      LOGICAL, DIMENSION(:, :, :), INTENT(IN), OPTIONAL  :: mask
      REAL(KIND=dp)                                      :: ks

      INTEGER                                            :: i1, i2, i3
      REAL(KIND=dp)                                      :: c, t, y

      ks = dzero; t = dzero; y = dzero; c = dzero

      IF (PRESENT(mask)) THEN
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3)) THEN
               y = array(i1, i2, i3)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_d3

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_c3(array, mask) RESULT(ks)
      COMPLEX(KIND=sp), DIMENSION(:, :, :), INTENT(IN)   :: array
      LOGICAL, DIMENSION(:, :, :), INTENT(IN), OPTIONAL  :: mask
      COMPLEX(KIND=sp)                                   :: ks

      COMPLEX(KIND=sp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3

      ks = czero; t = czero; y = czero; c = czero

      IF (PRESENT(mask)) THEN
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3)) THEN
               y = array(i1, i2, i3)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_c3

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_z3(array, mask) RESULT(ks)
      COMPLEX(KIND=dp), DIMENSION(:, :, :), INTENT(IN)   :: array
      LOGICAL, DIMENSION(:, :, :), INTENT(IN), OPTIONAL  :: mask
      COMPLEX(KIND=dp)                                   :: ks

      COMPLEX(KIND=dp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3

      ks = zzero; t = zzero; y = zzero; c = zzero

      IF (PRESENT(mask)) THEN
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3)) THEN
               y = array(i1, i2, i3)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_z3

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_s4(array, mask) RESULT(ks)
      REAL(KIND=sp), DIMENSION(:, :, :, :), INTENT(IN)   :: array
      LOGICAL, DIMENSION(:, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      REAL(KIND=sp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4
      REAL(KIND=sp)                                      :: c, t, y

      ks = szero; t = szero; y = szero; c = szero

      IF (PRESENT(mask)) THEN
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4)) THEN
               y = array(i1, i2, i3, i4)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_s4

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_d4(array, mask) RESULT(ks)
      REAL(KIND=dp), DIMENSION(:, :, :, :), INTENT(IN)   :: array
      LOGICAL, DIMENSION(:, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      REAL(KIND=dp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4
      REAL(KIND=dp)                                      :: c, t, y

      ks = dzero; t = dzero; y = dzero; c = dzero

      IF (PRESENT(mask)) THEN
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4)) THEN
               y = array(i1, i2, i3, i4)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_d4

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_c4(array, mask) RESULT(ks)
      COMPLEX(KIND=sp), DIMENSION(:, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      COMPLEX(KIND=sp)                                   :: ks

      COMPLEX(KIND=sp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4

      ks = czero; t = czero; y = czero; c = czero

      IF (PRESENT(mask)) THEN
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4)) THEN
               y = array(i1, i2, i3, i4)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_c4

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_z4(array, mask) RESULT(ks)
      COMPLEX(KIND=dp), DIMENSION(:, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      COMPLEX(KIND=dp)                                   :: ks

      COMPLEX(KIND=dp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4

      ks = zzero; t = zzero; y = zzero; c = zzero

      IF (PRESENT(mask)) THEN
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4)) THEN
               y = array(i1, i2, i3, i4)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_z4

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_s5(array, mask) RESULT(ks)
      REAL(KIND=sp), DIMENSION(:, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      REAL(KIND=sp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4, i5
      REAL(KIND=sp)                                      :: c, t, y

      ks = szero; t = szero; y = szero; c = szero

      IF (PRESENT(mask)) THEN
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5)) THEN
               y = array(i1, i2, i3, i4, i5)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_s5

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_d5(array, mask) RESULT(ks)
      REAL(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      REAL(KIND=dp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4, i5
      REAL(KIND=dp)                                      :: c, t, y

      ks = dzero; t = dzero; y = dzero; c = dzero

      IF (PRESENT(mask)) THEN
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5)) THEN
               y = array(i1, i2, i3, i4, i5)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_d5

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_c5(array, mask) RESULT(ks)
      COMPLEX(KIND=sp), DIMENSION(:, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      COMPLEX(KIND=sp)                                   :: ks

      COMPLEX(KIND=sp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4, i5

      ks = czero; t = czero; y = czero; c = czero

      IF (PRESENT(mask)) THEN
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5)) THEN
               y = array(i1, i2, i3, i4, i5)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_c5

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_z5(array, mask) RESULT(ks)
      COMPLEX(KIND=dp), DIMENSION(:, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      COMPLEX(KIND=dp)                                   :: ks

      COMPLEX(KIND=dp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4, i5

      ks = zzero; t = zzero; y = zzero; c = zzero

      IF (PRESENT(mask)) THEN
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5)) THEN
               y = array(i1, i2, i3, i4, i5)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_z5

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_s6(array, mask) RESULT(ks)
      REAL(KIND=sp), DIMENSION(:, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      REAL(KIND=sp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4, i5, i6
      REAL(KIND=sp)                                      :: c, t, y

      ks = szero; t = szero; y = szero; c = szero

      IF (PRESENT(mask)) THEN
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6)) THEN
               y = array(i1, i2, i3, i4, i5, i6)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_s6

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_d6(array, mask) RESULT(ks)
      REAL(KIND=dp), DIMENSION(:, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      REAL(KIND=dp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4, i5, i6
      REAL(KIND=dp)                                      :: c, t, y

      ks = dzero; t = dzero; y = dzero; c = dzero

      IF (PRESENT(mask)) THEN
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6)) THEN
               y = array(i1, i2, i3, i4, i5, i6)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_d6

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_c6(array, mask) RESULT(ks)
      COMPLEX(KIND=sp), DIMENSION(:, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      COMPLEX(KIND=sp)                                   :: ks

      COMPLEX(KIND=sp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4, i5, i6

      ks = czero; t = czero; y = czero; c = czero

      IF (PRESENT(mask)) THEN
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6)) THEN
               y = array(i1, i2, i3, i4, i5, i6)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_c6

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_z6(array, mask) RESULT(ks)
      COMPLEX(KIND=dp), DIMENSION(:, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :), INTENT(IN), &
         OPTIONAL                                        :: mask
      COMPLEX(KIND=dp)                                   :: ks

      COMPLEX(KIND=dp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4, i5, i6

      ks = zzero; t = zzero; y = zzero; c = zzero

      IF (PRESENT(mask)) THEN
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6)) THEN
               y = array(i1, i2, i3, i4, i5, i6)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_z6

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_s7(array, mask) RESULT(ks)
      REAL(KIND=sp), DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN), OPTIONAL                            :: mask
      REAL(KIND=sp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4, i5, i6, i7
      REAL(KIND=sp)                                      :: c, t, y

      ks = szero; t = szero; y = szero; c = szero

      IF (PRESENT(mask)) THEN
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6, i7)) THEN
               y = array(i1, i2, i3, i4, i5, i6, i7)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6, i7)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_s7

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_d7(array, mask) RESULT(ks)
      REAL(KIND=dp), DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN), OPTIONAL                            :: mask
      REAL(KIND=dp)                                      :: ks

      INTEGER                                            :: i1, i2, i3, i4, i5, i6, i7
      REAL(KIND=dp)                                      :: c, t, y

      ks = dzero; t = dzero; y = dzero; c = dzero

      IF (PRESENT(mask)) THEN
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6, i7)) THEN
               y = array(i1, i2, i3, i4, i5, i6, i7)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6, i7)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_d7

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_c7(array, mask) RESULT(ks)
      COMPLEX(KIND=sp), DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN), OPTIONAL                            :: mask
      COMPLEX(KIND=sp)                                   :: ks

      COMPLEX(KIND=sp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4, i5, i6, i7

      ks = czero; t = czero; y = czero; c = czero

      IF (PRESENT(mask)) THEN
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6, i7)) THEN
               y = array(i1, i2, i3, i4, i5, i6, i7)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6, i7)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_c7

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param mask ...
!> \retval ks ...
! **************************************************************************************************
   FUNCTION kahan_sum_z7(array, mask) RESULT(ks)
      COMPLEX(KIND=dp), DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN)                                      :: array
      LOGICAL, DIMENSION(:, :, :, :, :, :, :), &
         INTENT(IN), OPTIONAL                            :: mask
      COMPLEX(KIND=dp)                                   :: ks

      COMPLEX(KIND=dp)                                   :: c, t, y
      INTEGER                                            :: i1, i2, i3, i4, i5, i6, i7

      ks = zzero; t = zzero; y = zzero; c = zzero

      IF (PRESENT(mask)) THEN
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            IF (mask(i1, i2, i3, i4, i5, i6, i7)) THEN
               y = array(i1, i2, i3, i4, i5, i6, i7)-c
               t = ks+y
               c = (t-ks)-y
               ks = t
            END IF
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ELSE
         DO i7 = 1, SIZE(array, 7)
         DO i6 = 1, SIZE(array, 6)
         DO i5 = 1, SIZE(array, 5)
         DO i4 = 1, SIZE(array, 4)
         DO i3 = 1, SIZE(array, 3)
         DO i2 = 1, SIZE(array, 2)
         DO i1 = 1, SIZE(array, 1)
            y = array(i1, i2, i3, i4, i5, i6, i7)-c
            t = ks+y
            c = (t-ks)-y
            ks = t
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
         ENDDO
      ENDIF
   END FUNCTION kahan_sum_z7

END MODULE kahan_sum
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \note
!>      If parallel mode is distributed certain combination of
!>      "in_use" and "in_space" can not be used.
!>      For performance reasons it would be better to have the loops
!>      over g-vectros in the gather/scatter routines in new subprograms
!>      with the actual arrays (also the adressing) in the parameter list
!> \par History
!>      JGH (29-Dec-2000) : Changes for parallel use
!>      JGH (13-Mar-2001) : added timing calls
!>      JGH (26-Feb-2003) : OpenMP enabled
!>      JGH (17-Nov-2007) : Removed mass arrays
!>      JGH (01-Dec-2007) : Removed and renamed routines
!>      03.2008 [tlaino] : Splitting pw_types into pw_types and pw_methods
!> \author apsi
! **************************************************************************************************
MODULE pw_methods
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_to_string
   USE fast,                            ONLY: copy_cr,&
                                              copy_rc,&
                                              zero_c
   USE fft_tools,                       ONLY: BWFFT,&
                                              FWFFT,&
                                              fft3d
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE machine,                         ONLY: m_memory
   USE message_passing,                 ONLY: mp_sum
   USE pw_cuda,                         ONLY: pw_cuda_c1dr3d_3d,&
                                              pw_cuda_c1dr3d_3d_ps,&
                                              pw_cuda_r3dc1d_3d,&
                                              pw_cuda_r3dc1d_3d_ps
   USE pw_grid_types,                   ONLY: HALFSPACE,&
                                              PW_MODE_DISTRIBUTED,&
                                              PW_MODE_LOCAL,&
                                              pw_grid_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              COMPLEXDATA3D,&
                                              NOSPACE,&
                                              REALDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pw_zero, pw_structure_factor, pw_smoothing
   PUBLIC :: pw_copy, pw_axpy, pw_transfer, pw_scale
   PUBLIC :: pw_derive, pw_dr2, pw_write
   PUBLIC :: pw_integral_ab, pw_integral_a2b
   PUBLIC :: pw_dr2_gg, pw_integrate_function

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_methods'
   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .FALSE.

   INTERFACE pw_gather
      MODULE PROCEDURE pw_gather_s, pw_gather_p
   END INTERFACE

   INTERFACE pw_scatter
      MODULE PROCEDURE pw_scatter_s, pw_scatter_p
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Set values of a pw type to zero
!> \param pw ...
!> \par History
!>      none
!> \author apsi
! **************************************************************************************************
   SUBROUTINE pw_zero(pw)

      TYPE(pw_type), INTENT(INOUT)                       :: pw

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_zero', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ns
      REAL(KIND=dp)                                      :: zr

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",95)
      IF (pw%in_use == REALDATA1D) THEN
         ns = SIZE(pw%cr)
         pw%cr(:) = 0._dp
      ELSE IF (pw%in_use == COMPLEXDATA1D) THEN
         ns = SIZE(pw%cc)
         pw%cc(:) = CMPLX(0._dp, 0._dp, KIND=dp)
      ELSE IF (pw%in_use == REALDATA3D) THEN
         ns = SIZE(pw%cr3d)
         pw%cr3d(:, :, :) = 0._dp
      ELSE IF (pw%in_use == COMPLEXDATA3D) THEN
         ns = SIZE(pw%cc3d)
         pw%cc3d(:, :, :) = CMPLX(0._dp, 0._dp, KIND=dp)
      ELSE
         CALL cp__b("pw/pw_methods.F",109,"No possible data field!")
      END IF

      zr = REAL(ns, KIND=dp)*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_zero

! **************************************************************************************************
!> \brief copy a pw type variable
!> \param pw1 ...
!> \param pw2 ...
!> \par History
!>      JGH (7-Mar-2001) : check for pw_grid %id_nr, allow copy if
!>        in_use == COMPLEXDATA1D and in_space == RECIPROCALSPACE
!>      JGH (21-Feb-2003) : Code for generalized reference grids
!> \author apsi
!> \note
!>      Currently only copying of respective types allowed,
!>      in order to avoid errors
! **************************************************************************************************
   SUBROUTINE pw_copy(pw1, pw2)
      TYPE(pw_type), INTENT(IN)                          :: pw1
      TYPE(pw_type), INTENT(INOUT)                       :: pw2

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_copy', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ng, ng1, ng2, ns, &
                                                            output_unit
      REAL(KIND=dp)                                      :: zc

      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit()
      IF(.NOT.(pw1%ref_count > 0))CALL cp__a("pw/pw_methods.F",142)
      IF(.NOT.(pw2%ref_count > 0))CALL cp__a("pw/pw_methods.F",143)
      IF (pw1%pw_grid%id_nr /= pw2%pw_grid%id_nr) THEN

         IF (pw1%pw_grid%spherical .AND. pw2%pw_grid%spherical) THEN

            IF (pw_compatible(pw1%pw_grid, pw2%pw_grid)) THEN

               IF (pw1%in_use == COMPLEXDATA1D .AND. &
                   pw2%in_use == COMPLEXDATA1D .AND. &
                   pw1%in_space == RECIPROCALSPACE) THEN
                  ng1 = SIZE(pw1%cc)
                  ng2 = SIZE(pw2%cc)
                  ng = MIN(ng1, ng2)
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(ng,pw1,pw2)
                  DO i = 1, ng
                     pw2%cc(i) = pw1%cc(i)
                  END DO
                  IF (ng2 > ng) THEN
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(ng,ng2,pw2)
                     DO i = ng+1, ng2
                        pw2%cc(i) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
                     END DO
                  END IF
                  ns = 2*MAX(ng1, ng2)
               ELSE
                  CALL cp__b("pw/pw_methods.F",168,"No suitable data field")
               END IF

            ELSE
               IF (output_unit > 0) THEN
                  WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                     " grid 1 :", pw1%pw_grid%id_nr, &
                     " spherical :", pw1%pw_grid%spherical, &
                     " reference :", pw1%pw_grid%reference
                  WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                     " grid 2 :", pw2%pw_grid%id_nr, &
                     " spherical :", pw2%pw_grid%spherical, &
                     " reference :", pw2%pw_grid%reference
               END IF
               CALL cp__b("pw/pw_methods.F",182,"Incompatible grids")
            END IF

         ELSE IF (.NOT. (pw1%pw_grid%spherical .OR. &
                         pw2%pw_grid%spherical)) THEN

            ng1 = SIZE(pw1%cc)
            ng2 = SIZE(pw2%cc)
            ns = 2*MAX(ng1, ng2)

            IF (pw1%in_use == COMPLEXDATA1D .AND. &
                pw2%in_use == COMPLEXDATA1D .AND. &
                pw1%in_space == RECIPROCALSPACE) THEN

               IF ((pw1%pw_grid%id_nr == pw2%pw_grid%reference)) THEN
                  IF (ng1 >= ng2) THEN
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(ng2,pw1,pw2)
                     DO i = 1, ng2
                        j = pw2%pw_grid%gidx(i)
                        pw2%cc(i) = pw1%cc(j)
                     END DO
                  ELSE
                     CALL pw_zero(pw2)
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(ng1,pw1,pw2)
                     DO i = 1, ng1
                        j = pw2%pw_grid%gidx(i)
                        pw2%cc(j) = pw1%cc(i)
                     END DO
                  END IF
               ELSE IF ((pw2%pw_grid%id_nr == pw1%pw_grid%reference)) THEN
                  IF (ng1 >= ng2) THEN
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw1,pw2,ng2)
                     DO i = 1, ng2
                        j = pw1%pw_grid%gidx(i)
                        pw2%cc(i) = pw1%cc(j)
                     END DO
                  ELSE
                     CALL pw_zero(pw2)
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw1,pw2,ng1)
                     DO i = 1, ng1
                        j = pw1%pw_grid%gidx(i)
                        pw2%cc(j) = pw1%cc(i)
                     END DO
                  END IF
               ELSE
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                        " grid 1 :", pw1%pw_grid%id_nr, &
                        " spherical :", pw1%pw_grid%spherical, &
                        " reference :", pw1%pw_grid%reference
                     WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                        " grid 2 :", pw2%pw_grid%id_nr, &
                        " spherical :", pw2%pw_grid%spherical, &
                        " reference :", pw2%pw_grid%reference
                  ENDIF
                  CALL cp__b("pw/pw_methods.F",237,"Incompatible grids")
               END IF

            ELSE
               CALL cp__b("pw/pw_methods.F",241,"No suitable data field")
            END IF

            pw2%in_space = RECIPROCALSPACE

         ELSE
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                  " grid 1 :", pw1%pw_grid%id_nr, &
                  " spherical :", pw1%pw_grid%spherical, &
                  " reference :", pw1%pw_grid%reference
               WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                  " grid 2 :", pw2%pw_grid%id_nr, &
                  " spherical :", pw2%pw_grid%spherical, &
                  " reference :", pw2%pw_grid%reference
            ENDIF
            CALL cp__b("pw/pw_methods.F",257,"Incompatible grids")
         END IF

      ELSE

         IF (pw1%in_use == REALDATA1D .AND. pw2%in_use == REALDATA1D) THEN
            ns = SIZE(pw1%cr)
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(ns,pw1,pw2)
            DO i = 1, ns
               pw2%cr(i) = pw1%cr(i)
            END DO
         ELSE IF (pw1%in_use == COMPLEXDATA1D .AND. &
                  pw2%in_use == COMPLEXDATA1D) THEN
            ns = SIZE(pw1%cc)
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(ns,pw2,pw1)
            DO i = 1, ns
               pw2%cc(i) = pw1%cc(i)
            END DO
         ELSE IF (pw1%in_use == REALDATA3D .AND. pw2%in_use == REALDATA3D) THEN
            ns = SIZE(pw1%cr3d)
            pw2%cr3d(:, :, :) = pw1%cr3d(:, :, :)
         ELSE IF (pw1%in_use == COMPLEXDATA3D .AND. &
                  pw2%in_use == COMPLEXDATA3D) THEN
            ns = SIZE(pw1%cc3d)
            pw2%cc3d(:, :, :) = pw1%cc3d(:, :, :)
         ELSE
            CALL cp__b("pw/pw_methods.F",283,"No suitable data field")
         END IF

      END IF

      pw2%in_space = pw1%in_space
      zc = REAL(ns, KIND=dp)*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_copy

! **************************************************************************************************
!> \brief multiplies pw coeffs with a number
!> \param pw ...
!> \param a ...
!> \par History
!>      11.2004 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE pw_scale(pw, a)

      TYPE(pw_type), INTENT(INOUT)                       :: pw
      REAL(KIND=dp), INTENT(IN)                          :: a

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_scale', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ns
      REAL(KIND=dp)                                      :: flop

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",312)

      SELECT CASE (pw%in_use)
      CASE (REALDATA1D)
         ns = SIZE(pw%cr)
         pw%cr(:) = a*pw%cr(:)
      CASE (COMPLEXDATA1D)
         ns = 2*SIZE(pw%cc)
         pw%cc(:) = a*pw%cc(:)
      CASE (REALDATA3D)
         ns = SIZE(pw%cr3d)
         pw%cr3d(:, :, :) = a*pw%cr3d(:, :, :)
      CASE (COMPLEXDATA3D)
         ns = 2*SIZE(pw%cc3d)
         pw%cc3d(:, :, :) = a*pw%cc3d(:, :, :)
      CASE DEFAULT
         CALL cp__b("pw/pw_methods.F",328,"No suitable data field")
      END SELECT

      flop = REAL(ns, KIND=dp)*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_scale

! **************************************************************************************************
!> \brief Calculate the derivative of a plane wave vector
!> \param pw ...
!> \param n ...
!> \par History
!>      JGH (06-10-2002) allow only for inplace derivatives
!> \author JGH (25-Feb-2001)
!> \note
!>      Calculate the derivative dx^n(1) dy^n(2) dz^n(3) PW
!>      PW has to be in RECIPROCALSPACE and data in use is COMPLEXDATA1D
! **************************************************************************************************
   SUBROUTINE pw_derive(pw, n)

      TYPE(pw_type), INTENT(INOUT)                       :: pw
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_derive', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: im
      INTEGER                                            :: cnt, handle, i, m
      REAL(KIND=dp)                                      :: flop

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",359)
      IF(.NOT.(ALL(n >= 0)))CALL cp__a("pw/pw_methods.F",360)

      m = SUM(n)
      im = CMPLX(0.0_dp, 1.0_dp, KIND=dp)**m

      flop = 0.0_dp

      IF (pw%in_space == RECIPROCALSPACE .AND. &
          pw%in_use == COMPLEXDATA1D) THEN

         cnt = SIZE(pw%cc)

         IF (n(1) == 1) THEN
!$OMP PARALLEL DO PRIVATE (i) DEFAULT(NONE) SHARED(cnt, pw)
            DO i = 1, cnt
               pw%cc(i) = pw%cc(i)*pw%pw_grid%g(1, i)
            END DO
            flop = flop+6*cnt
         ELSE IF (n(1) > 1) THEN
!$OMP PARALLEL DO PRIVATE (i) DEFAULT(NONE) SHARED(cnt, pw,n)
            DO i = 1, cnt
               pw%cc(i) = pw%cc(i)*(pw%pw_grid%g(1, i)**n(1))
            END DO
            flop = flop+7*cnt
         END IF
         IF (n(2) == 1) THEN
!$OMP PARALLEL DO PRIVATE (i) DEFAULT(NONE) SHARED(pw, cnt)
            DO i = 1, cnt
               pw%cc(i) = pw%cc(i)*pw%pw_grid%g(2, i)
            END DO
            flop = flop+6*cnt
         ELSE IF (n(2) > 1) THEN
!$OMP PARALLEL DO PRIVATE (i) DEFAULT(NONE) SHARED(pw, cnt,n)
            DO i = 1, cnt
               pw%cc(i) = pw%cc(i)*(pw%pw_grid%g(2, i)**n(2))
            END DO
            flop = flop+7*cnt
         END IF
         IF (n(3) == 1) THEN
!$OMP PARALLEL DO PRIVATE (i) DEFAULT(NONE) SHARED(pw,cnt)
            DO i = 1, cnt
               pw%cc(i) = pw%cc(i)*pw%pw_grid%g(3, i)
            END DO
            flop = flop+6*cnt
         ELSE IF (n(3) > 1) THEN
!$OMP PARALLEL DO PRIVATE (i) DEFAULT(NONE) SHARED(pw,cnt,n)
            DO i = 1, cnt
               pw%cc(i) = pw%cc(i)*(pw%pw_grid%g(3, i)**n(3))
            END DO
            flop = flop+7*cnt
         END IF

         ! im can take the values 1, -1, i, -i
         ! skip this if im == 1
         IF (ABS(REAL(im, KIND=dp)-1.0_dp) > 1.e-10) THEN
!$OMP PARALLEL DO PRIVATE (i) DEFAULT(NONE) SHARED(pw,cnt,im)
            DO i = 1, cnt
               pw%cc(i) = im*pw%cc(i)
            END DO
            flop = flop+6*cnt
         END IF

      ELSE

         CALL cp__b("pw/pw_methods.F",424,"No suitable data field")

      END IF

      flop = flop*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_derive

! **************************************************************************************************
!> \brief Calculate the tensorial 2nd derivative of a plane wave vector
!> \param pw ...
!> \param pwdr2 ...
!> \param i ...
!> \param j ...
!> \par History
!>      none
!> \author JGH (05-May-2006)
!> \note
!>      PW has to be in RECIPROCALSPACE and data in use is COMPLEXDATA1D
! **************************************************************************************************
   SUBROUTINE pw_dr2(pw, pwdr2, i, j)

      TYPE(pw_type), INTENT(IN)                          :: pw
      TYPE(pw_type), INTENT(INOUT)                       :: pwdr2
      INTEGER, INTENT(IN)                                :: i, j

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_dr2', routineP = moduleN//':'//routineN

      INTEGER                                            :: cnt, handle, ig
      REAL(KIND=dp)                                      :: flop, gg, o3

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",457)

      flop = 0.0_dp
      o3 = 1._dp/3._dp

      IF (pw%in_space == RECIPROCALSPACE .AND. &
          pw%in_use == COMPLEXDATA1D) THEN

         cnt = SIZE(pw%cc)

         IF (i == j) THEN
!$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(ig,gg) SHARED(i,o3,pw,pwdr2,cnt)
            DO ig = 1, cnt
               gg = pw%pw_grid%g(i, ig)**2-o3*pw%pw_grid%gsq(ig)
               pwdr2%cc(ig) = gg*pw%cc(ig)
            END DO
            flop = flop+5*cnt
         ELSE
!$OMP PARALLEL DO PRIVATE (ig) DEFAULT(NONE) SHARED(pwdr2,pw,i,j,cnt)
            DO ig = 1, cnt
               pwdr2%cc(ig) = pw%cc(ig)*(pw%pw_grid%g(i, ig)*pw%pw_grid%g(j, ig))
            END DO
            flop = flop+4*cnt
         END IF

      ELSE

         CALL cp__b("pw/pw_methods.F",484,"No suitable data field")

      END IF

      flop = flop*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_dr2

! **************************************************************************************************
!> \brief Calculate the tensorial 2nd derivative of a plane wave vector
!>      and divides by |G|^2. pwdr2_gg(G=0) is put to zero.
!> \param pw ...
!> \param pwdr2_gg ...
!> \param i ...
!> \param j ...
!> \par History
!>      none
!> \author RD (20-Nov-2006)
!> \note
!>      PW has to be in RECIPROCALSPACE and data in use is COMPLEXDATA1D
!>      Adapted from pw_dr2
! **************************************************************************************************
   SUBROUTINE pw_dr2_gg(pw, pwdr2_gg, i, j)

      TYPE(pw_type), INTENT(IN)                          :: pw
      TYPE(pw_type), INTENT(INOUT)                       :: pwdr2_gg
      INTEGER, INTENT(IN)                                :: i, j

      INTEGER                                            :: cnt, handle, ig
      REAL(KIND=dp)                                      :: flop, gg, o3
      CHARACTER(len=*), PARAMETER :: routineN = 'pw_dr2_gg', routineP = moduleN//':'//routineN

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",518)

      flop = 0.0_dp
      o3 = 1._dp/3._dp

      IF (pw%in_space == RECIPROCALSPACE .AND. &
          pw%in_use == COMPLEXDATA1D) THEN

         cnt = SIZE(pw%cc)

         IF (i == j) THEN
!$OMP PARALLEL DO PRIVATE (ig) DEFAULT(NONE) PRIVATE(gg) SHARED(cnt,pw,o3,pwdr2_gg,i)
            DO ig = pw%pw_grid%first_gne0, cnt
               gg = pw%pw_grid%g(i, ig)**2-o3*pw%pw_grid%gsq(ig)
               pwdr2_gg%cc(ig) = gg*pw%cc(ig)/pw%pw_grid%gsq(ig)
            END DO
            flop = flop+6*cnt
         ELSE
!$OMP PARALLEL DO PRIVATE (ig) DEFAULT(NONE) SHARED(pwdr2_gg,pw,i,j,cnt)
            DO ig = pw%pw_grid%first_gne0, cnt
               pwdr2_gg%cc(ig) = pw%cc(ig)*(pw%pw_grid%g(i, ig)*pw%pw_grid%g(j, ig)) &
                                 /pw%pw_grid%gsq(ig)
            END DO
            flop = flop+5*cnt
         END IF

         IF (pw%pw_grid%have_g0) pwdr2_gg%cc(1) = 0.0_dp

      ELSE

         CALL cp__b("pw/pw_methods.F",548,"No suitable data field")

      END IF

      flop = flop*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_dr2_gg

! **************************************************************************************************
!> \brief Multiplies a G-space function with a smoothing factor of the form
!>      f(|G|) = exp((ecut - G^2)/sigma)/(1+exp((ecut - G^2)/sigma))
!> \param pw ...
!> \param ecut ...
!> \param sigma ...
!> \par History
!>      none
!> \author JGH (09-June-2006)
!> \note
!>      PW has to be in RECIPROCALSPACE and data in use is COMPLEXDATA1D
! **************************************************************************************************
   SUBROUTINE pw_smoothing(pw, ecut, sigma)

      TYPE(pw_type), INTENT(INOUT)                       :: pw
      REAL(KIND=dp), INTENT(IN)                          :: ecut, sigma

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_smoothing', routineP = moduleN//':'//routineN

      INTEGER                                            :: cnt, handle, ig
      REAL(KIND=dp)                                      :: arg, f, flop

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",580)

      flop = 0.0_dp

      IF (pw%in_space == RECIPROCALSPACE .AND. &
          pw%in_use == COMPLEXDATA1D) THEN

         cnt = SIZE(pw%cc)

!$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(ig,arg,f) SHARED(sigma,ecut,pw,cnt)
         DO ig = 1, cnt
            arg = (ecut-pw%pw_grid%gsq(ig))/sigma
            f = EXP(arg)/(1+EXP(arg))
            pw%cc(ig) = f*pw%cc(ig)
         END DO
         flop = flop+6*cnt

      ELSE

         CALL cp__b("pw/pw_methods.F",599,"No suitable data field")

      END IF

      flop = flop*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_smoothing

! **************************************************************************************************
!> \brief Generalize copy of pw types
!> \param pw1 ...
!> \param pw2 ...
!> \param debug ...
!> \par History
!>      JGH (13-Mar-2001) : added gather/scatter cases
!> \author JGH (25-Feb-2001)
!> \note
!>      Copy routine that allows for in_space changes
! **************************************************************************************************
   SUBROUTINE pw_transfer(pw1, pw2, debug)

      TYPE(pw_type), INTENT(IN), TARGET                  :: pw1
      TYPE(pw_type), INTENT(INOUT), TARGET               :: pw2
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_transfer', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      !sample peak memory
      CALL m_memory()

      IF(.NOT.(pw1%ref_count > 0))CALL cp__a("pw/pw_methods.F",633)
      IF(.NOT.(pw2%ref_count > 0))CALL cp__a("pw/pw_methods.F",634)

      IF (pw1%in_space == REALSPACE .AND. pw2%in_space == REALSPACE) THEN

         ! simple copy should do
         CALL pw_copy(pw1, pw2)

      ELSEIF (pw1%in_space == RECIPROCALSPACE .AND. &
              pw2%in_space == RECIPROCALSPACE) THEN

         IF (pw1%in_use == pw2%in_use) THEN

            ! simple copy should do
            CALL pw_copy(pw1, pw2)

         ELSE

            ! we have to gather/scatter the data
            IF (pw1%in_use == COMPLEXDATA1D) THEN
               CALL pw_scatter(pw1, pw2%cc3d)
            ELSEIF (pw2%in_use == COMPLEXDATA1D) THEN
               CALL pw_gather(pw2, pw1%cc3d)
            ELSE
               CALL cp__b("pw/pw_methods.F",657,"Do not know what to do")
            END IF

         END IF

      ELSE

         ! FFT needed, all further tests done in fft_wrap_pw1pw2
         CALL fft_wrap_pw1pw2(pw1, pw2, debug)

      END IF

      CALL timestop(handle)

   END SUBROUTINE pw_transfer

! **************************************************************************************************
!> \brief pw2 = alpha*pw1 + pw2
!>      alpha defaults to 1
!> \param pw1 ...
!> \param pw2 ...
!> \param alpha ...
!> \par History
!>      JGH (21-Feb-2003) : added reference grid functionality
!>      JGH (01-Dec-2007) : rename and remove complex alpha
!> \author apsi
!> \note
!>      Currently only summing up of respective types allowed,
!>      in order to avoid errors
! **************************************************************************************************
   SUBROUTINE pw_axpy(pw1, pw2, alpha)

      TYPE(pw_type), INTENT(IN)                          :: pw1
      TYPE(pw_type), INTENT(INOUT)                       :: pw2
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_axpy', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ng, ng1, ng2, output_unit
      REAL(KIND=dp)                                      :: flop, my_alpha

      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit()

      IF(.NOT.(pw1%ref_count > 0))CALL cp__a("pw/pw_methods.F",701)
      IF(.NOT.(pw2%ref_count > 0))CALL cp__a("pw/pw_methods.F",702)

      my_alpha = 1.0_dp
      IF (PRESENT(alpha)) my_alpha = alpha

      IF (pw1%pw_grid%id_nr == pw2%pw_grid%id_nr) THEN

         IF (pw1%in_use == REALDATA1D .AND. pw2%in_use == REALDATA1D) THEN
            IF (my_alpha == 1.0_dp) THEN
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(pw2,pw1)
               DO i = 1, SIZE(pw2%cr)
                  pw2%cr(i) = pw2%cr(i)+pw1%cr(i)
               END DO
               flop = REAL(SIZE(pw2%cr), KIND=dp)
            ELSE
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(pw2,pw1,my_alpha)
               DO i = 1, SIZE(pw2%cr)
                  pw2%cr(i) = pw2%cr(i)+my_alpha*pw1%cr(i)
               END DO
               flop = REAL(2*SIZE(pw2%cr), KIND=dp)
            END IF
         ELSE IF (pw1%in_use == COMPLEXDATA1D .AND. &
                  pw2%in_use == COMPLEXDATA1D) THEN
            IF (my_alpha == 1.0_dp) THEN
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(pw2,pw1)
               DO i = 1, SIZE(pw2%cc)
                  pw2%cc(i) = pw2%cc(i)+pw1%cc(i)
               END DO
               flop = REAL(2*SIZE(pw2%cc), KIND=dp)
            ELSE
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(pw2,my_alpha,pw1)
               DO i = 1, SIZE(pw2%cc)
                  pw2%cc(i) = pw2%cc(i)+my_alpha*pw1%cc(i)
               END DO
               flop = REAL(4*SIZE(pw2%cc), KIND=dp)
            END IF
         ELSE IF (pw1%in_use == REALDATA3D .AND. pw2%in_use == REALDATA3D) THEN
            IF (my_alpha == 1.0_dp) THEN
               pw2%cr3d = pw2%cr3d+pw1%cr3d
               flop = REAL(SIZE(pw2%cr3d), KIND=dp)
            ELSE
               pw2%cr3d = pw2%cr3d+my_alpha*pw1%cr3d
               flop = REAL(2*SIZE(pw2%cr3d), KIND=dp)
            END IF
         ELSE IF (pw1%in_use == COMPLEXDATA3D .AND. &
                  pw2%in_use == COMPLEXDATA3D) THEN
            IF (my_alpha == 1.0_dp) THEN
               pw2%cc3d = pw2%cc3d+pw1%cc3d
               flop = REAL(2*SIZE(pw2%cc3d), KIND=dp)
            ELSE
               pw2%cc3d = pw2%cc3d+my_alpha*pw1%cc3d
               flop = REAL(4*SIZE(pw2%cc3d), KIND=dp)
            END IF
         ELSE
            CALL cp__b("pw/pw_methods.F",756,"No suitable data field")
         END IF

      ELSE IF (pw_compatible(pw1%pw_grid, pw2%pw_grid)) THEN

         IF (pw1%in_use == COMPLEXDATA1D .AND. &
             pw2%in_use == COMPLEXDATA1D .AND. &
             pw1%in_space == RECIPROCALSPACE .AND. &
             pw1%in_space == RECIPROCALSPACE) THEN

            ng1 = SIZE(pw1%cc)
            ng2 = SIZE(pw2%cc)
            ng = MIN(ng1, ng2)
            flop = REAL(2*ng, KIND=dp)

            IF (pw1%pw_grid%spherical) THEN

               IF (my_alpha == 1.0_dp) THEN
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(pw2,pw1,ng)
                  DO i = 1, ng
                     pw2%cc(i) = pw2%cc(i)+pw1%cc(i)
                  END DO
               ELSE
!$OMP PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(pw2,pw1,my_alpha,ng)
                  DO i = 1, ng
                     pw2%cc(i) = pw2%cc(i)+my_alpha*pw1%cc(i)
                  END DO
               END IF

            ELSEIF ((pw1%pw_grid%id_nr == pw2%pw_grid%reference)) THEN

               IF (my_alpha == 1.0_dp) THEN
                  IF (ng1 >= ng2) THEN
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,ng2)
                     DO i = 1, ng2
                        j = pw2%pw_grid%gidx(i)
                        pw2%cc(i) = pw2%cc(i)+pw1%cc(j)
                     END DO
                  ELSE
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,ng1)
                     DO i = 1, ng1
                        j = pw2%pw_grid%gidx(i)
                        pw2%cc(j) = pw2%cc(j)+pw1%cc(i)
                     END DO
                  END IF
               ELSE
                  IF (ng1 >= ng2) THEN
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,my_alpha,ng2)
                     DO i = 1, ng2
                        j = pw2%pw_grid%gidx(i)
                        pw2%cc(i) = pw2%cc(i)+my_alpha*pw1%cc(j)
                     END DO
                  ELSE
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,my_alpha,ng1)
                     DO i = 1, ng1
                        j = pw2%pw_grid%gidx(i)
                        pw2%cc(j) = pw2%cc(j)+my_alpha*pw1%cc(i)
                     END DO
                  END IF
               END IF
            ELSEIF ((pw2%pw_grid%id_nr == pw1%pw_grid%reference)) THEN

               IF (my_alpha == 1.0_dp) THEN
                  IF (ng1 >= ng2) THEN
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,ng2)
                     DO i = 1, ng2
                        j = pw1%pw_grid%gidx(i)
                        pw2%cc(i) = pw2%cc(i)+pw1%cc(j)
                     END DO
                  ELSE
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,ng1)
                     DO i = 1, ng1
                        j = pw1%pw_grid%gidx(i)
                        pw2%cc(j) = pw2%cc(j)+pw1%cc(i)
                     END DO
                  END IF
               ELSE
                  IF (ng1 >= ng2) THEN
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,my_alpha,ng2)
                     DO i = 1, ng2
                        j = pw1%pw_grid%gidx(i)
                        pw2%cc(i) = pw2%cc(i)+my_alpha*pw1%cc(j)
                     END DO
                  ELSE
!$OMP PARALLEL DO PRIVATE(i,j) DEFAULT(NONE) SHARED(pw2,pw1,my_alpha,ng1)
                     DO i = 1, ng1
                        j = pw1%pw_grid%gidx(i)
                        pw2%cc(j) = pw2%cc(j)+my_alpha*pw1%cc(i)
                     END DO
                  END IF
               END IF
            ELSE
               IF (output_unit > 0) THEN
                  WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                     " grid 1 :", pw1%pw_grid%id_nr, &
                     " sperical :", pw1%pw_grid%spherical, &
                     " reference :", pw1%pw_grid%reference
                  WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
                     " grid 2 :", pw2%pw_grid%id_nr, &
                     " sperical :", pw2%pw_grid%spherical, &
                     " reference :", pw2%pw_grid%reference
               ENDIF
               CALL cp__b("pw/pw_methods.F",858,"Grids not compatible")

            END IF

         ELSE
            CALL cp__b("pw/pw_methods.F",863,"No suitable data field")
         END IF

      ELSE
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
               " grid 1 :", pw1%pw_grid%id_nr, &
               " sperical :", pw1%pw_grid%spherical, &
               " reference :", pw1%pw_grid%reference
            WRITE (output_unit, "(A,I5,T30,A,L1,T60,A,I5)") &
               " grid 2 :", pw2%pw_grid%id_nr, &
               " sperical :", pw2%pw_grid%spherical, &
               " reference :", pw2%pw_grid%reference
         ENDIF
         CALL cp__b("pw/pw_methods.F",877,"Grids not compatible")

      END IF

      flop = flop*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_axpy

! **************************************************************************************************
!> \brief Gathers the pw vector from a 3d data field
!> \param pw ...
!> \param c ...
!> \param scale ...
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   SUBROUTINE pw_gather_s(pw, c, scale)

      TYPE(pw_type), INTENT(INOUT)                       :: pw
      COMPLEX(KIND=dp), DIMENSION(:, :, :), INTENT(IN)   :: c
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_gather_s', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, handle, l, m, n, ngpts
      INTEGER, DIMENSION(:), POINTER                     :: mapl, mapm, mapn
      INTEGER, DIMENSION(:, :), POINTER                  :: ghat
      REAL(KIND=dp)                                      :: cpy

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",909)

      IF (pw%in_use /= COMPLEXDATA1D) THEN
         CALL cp__b("pw/pw_methods.F",912,"Data field has to be COMPLEXDATA1D")
      ENDIF

      ! after the gather we are in g-space
      pw%in_space = RECIPROCALSPACE

      mapl => pw%pw_grid%mapl%pos
      mapm => pw%pw_grid%mapm%pos
      mapn => pw%pw_grid%mapn%pos

      ngpts = SIZE(pw%pw_grid%gsq)

      ghat => pw%pw_grid%g_hat

      IF (PRESENT(scale)) THEN

!$OMP PARALLEL DO PRIVATE(gpt,l,m,n) DEFAULT(NONE) SHARED(ngpts, mapl,mapm,mapn,ghat,scale,pw,c)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            pw%cc(gpt) = scale*c(l, m, n)

         END DO

      ELSE

!$OMP PARALLEL DO PRIVATE(gpt,l,m,n) DEFAULT(NONE) SHARED(ngpts, mapl,mapm,mapn,ghat,pw,c)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            pw%cc(gpt) = c(l, m, n)

         END DO

      END IF

      cpy = REAL(ngpts, KIND=dp)*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_gather_s

! **************************************************************************************************
!> \brief ...
!> \param pw ...
!> \param c ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE pw_gather_p(pw, c, scale)

      TYPE(pw_type), INTENT(INOUT), TARGET               :: pw
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: c
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_gather_p', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, handle, l, m, mn, n, ngpts
      INTEGER, DIMENSION(:), POINTER                     :: mapl, mapm, mapn
      INTEGER, DIMENSION(:, :), POINTER                  :: ghat, yzq
      REAL(KIND=dp)                                      :: cpy

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",977)

      IF (pw%in_use /= COMPLEXDATA1D) THEN
         CALL cp__b("pw/pw_methods.F",980,"Data field has to be COMPLEXDATA1D")
      ENDIF

      IF (pw%pw_grid%para%mode /= PW_MODE_DISTRIBUTED) THEN
         CALL cp__b("pw/pw_methods.F",984,"This grid type is not distributed")
      ENDIF

      ! after the gather we are in g-space
      pw%in_space = RECIPROCALSPACE

      mapl => pw%pw_grid%mapl%pos
      mapm => pw%pw_grid%mapm%pos
      mapn => pw%pw_grid%mapn%pos

      ngpts = SIZE(pw%pw_grid%gsq)

      ghat => pw%pw_grid%g_hat
      yzq => pw%pw_grid%para%yzq

      IF (PRESENT(scale)) THEN
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(l,m,n,mn), &
!$OMP             SHARED(ngpts,mapl,mapm,mapn,ghat,yzq,pw,c,scale)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            mn = yzq(m, n)
            pw%cc(gpt) = scale*c(l, mn)

         END DO
!$OMP END PARALLEL DO
      ELSE
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(l,m,n,mn), &
!$OMP             SHARED(ngpts,mapl,mapm,mapn,ghat,yzq,pw,c)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            mn = yzq(m, n)
            pw%cc(gpt) = c(l, mn)

         END DO
!$OMP END PARALLEL DO
      END IF

      cpy = REAL(ngpts, KIND=dp)*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_gather_p

! **************************************************************************************************
!> \brief Scatters a pw vector to a 3d data field
!> \param pw ...
!> \param c ...
!> \param scale ...
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   SUBROUTINE pw_scatter_s(pw, c, scale)

      TYPE(pw_type), INTENT(IN)                          :: pw
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         INTENT(INOUT)                                   :: c
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_scatter_s', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, handle, l, m, n, ngpts
      INTEGER, DIMENSION(:), POINTER                     :: mapl, mapm, mapn
      INTEGER, DIMENSION(:, :), POINTER                  :: ghat
      REAL(KIND=dp)                                      :: cpy

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",1058)

      IF (pw%in_use /= COMPLEXDATA1D) THEN
         CALL cp__b("pw/pw_methods.F",1061,"Data field has to be COMPLEXDATA1D")
      ENDIF

      IF (pw%in_space /= RECIPROCALSPACE) THEN
         CALL cp__b("pw/pw_methods.F",1065,"Data has to be in RECIPROCALSPACE")
      ENDIF

      mapl => pw%pw_grid%mapl%pos
      mapm => pw%pw_grid%mapm%pos
      mapn => pw%pw_grid%mapn%pos

      ghat => pw%pw_grid%g_hat

      ngpts = SIZE(pw%pw_grid%gsq)

      ! should only zero the unused bits (but the zero is needed)
      IF (.NOT. PRESENT(scale)) c = 0.0_dp

      IF (PRESENT(scale)) THEN

!$OMP PARALLEL DO PRIVATE(gpt,l,m,n) DEFAULT(NONE) SHARED(ngpts, mapl,mapm,mapn,ghat,scale,pw,c)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            c(l, m, n) = scale*pw%cc(gpt)

         END DO

      ELSE

!$OMP PARALLEL DO PRIVATE(gpt,l,m,n) DEFAULT(NONE) SHARED(ngpts, mapl,mapm,mapn,ghat,pw,c)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            c(l, m, n) = pw%cc(gpt)

         END DO

      END IF

      IF (pw%pw_grid%grid_span == HALFSPACE) THEN

         mapl => pw%pw_grid%mapl%neg
         mapm => pw%pw_grid%mapm%neg
         mapn => pw%pw_grid%mapn%neg

         IF (PRESENT(scale)) THEN

!$OMP PARALLEL DO PRIVATE(gpt,l,m,n) DEFAULT(NONE) SHARED(ngpts, mapl,mapm,mapn,ghat,scale,pw,c)
            DO gpt = 1, ngpts

               l = mapl(ghat(1, gpt))+1
               m = mapm(ghat(2, gpt))+1
               n = mapn(ghat(3, gpt))+1
               c(l, m, n) = scale*CONJG(pw%cc(gpt))

            END DO

         ELSE

!$OMP PARALLEL DO PRIVATE(gpt,l,m,n) DEFAULT(NONE) SHARED(ngpts, mapl,mapm,mapn,ghat,pw,c)
            DO gpt = 1, ngpts

               l = mapl(ghat(1, gpt))+1
               m = mapm(ghat(2, gpt))+1
               n = mapn(ghat(3, gpt))+1
               c(l, m, n) = CONJG(pw%cc(gpt))

            END DO

         END IF

      END IF

      cpy = REAL(ngpts, KIND=dp)*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_scatter_s

! **************************************************************************************************
!> \brief ...
!> \param pw ...
!> \param c ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE pw_scatter_p(pw, c, scale)

      TYPE(pw_type), INTENT(IN), TARGET                  :: pw
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: c
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_scatter_p', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, handle, l, m, mn, n, ngpts
      INTEGER, DIMENSION(:), POINTER                     :: mapl, mapm, mapn
      INTEGER, DIMENSION(:, :), POINTER                  :: ghat, yzq
      REAL(KIND=dp)                                      :: cpy

      CALL timeset(routineN, handle)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_methods.F",1164)

      IF (pw%in_use /= COMPLEXDATA1D) THEN
         CALL cp__b("pw/pw_methods.F",1167,"Data field has to be COMPLEXDATA1D")
      ENDIF

      IF (pw%in_space /= RECIPROCALSPACE) THEN
         CALL cp__b("pw/pw_methods.F",1171,"Data has to be in RECIPROCALSPACE")
      ENDIF

      IF (pw%pw_grid%para%mode /= PW_MODE_DISTRIBUTED) THEN
         CALL cp__b("pw/pw_methods.F",1175,"This grid type is not distributed")
      ENDIF

      mapl => pw%pw_grid%mapl%pos
      mapm => pw%pw_grid%mapm%pos
      mapn => pw%pw_grid%mapn%pos

      ghat => pw%pw_grid%g_hat
      yzq => pw%pw_grid%para%yzq

      ngpts = SIZE(pw%pw_grid%gsq)

      IF (.NOT. PRESENT(scale)) CALL zero_c(c)

      IF (PRESENT(scale)) THEN
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(l,m,n,mn), &
!$OMP             SHARED(ngpts,mapl,mapm,mapn,ghat,yzq,pw,c,scale)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            mn = yzq(m, n)
            c(l, mn) = scale*pw%cc(gpt)

         END DO
!$OMP END PARALLEL DO
      ELSE
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(l,m,n,mn), &
!$OMP             SHARED(ngpts,mapl,mapm,mapn,ghat,yzq,pw,c)
         DO gpt = 1, ngpts

            l = mapl(ghat(1, gpt))+1
            m = mapm(ghat(2, gpt))+1
            n = mapn(ghat(3, gpt))+1
            mn = yzq(m, n)
            c(l, mn) = pw%cc(gpt)

         END DO
!$OMP END PARALLEL DO
      END IF

      IF (pw%pw_grid%grid_span == HALFSPACE) THEN

         mapm => pw%pw_grid%mapm%neg
         mapn => pw%pw_grid%mapn%neg
         mapl => pw%pw_grid%mapl%neg

         IF (PRESENT(scale)) THEN
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(l,m,n,mn), &
!$OMP             SHARED(ngpts,mapl,mapm,mapn,ghat,yzq,pw,c,scale)
            DO gpt = 1, ngpts

               l = mapl(ghat(1, gpt))+1
               m = mapm(ghat(2, gpt))+1
               n = mapn(ghat(3, gpt))+1
               mn = yzq(m, n)
               c(l, mn) = scale*CONJG(pw%cc(gpt))

            END DO
!$OMP END PARALLEL DO
         ELSE
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(l,m,n,mn), &
!$OMP             SHARED(ngpts,mapl,mapm,mapn,ghat,yzq,pw,c)
            DO gpt = 1, ngpts

               l = mapl(ghat(1, gpt))+1
               m = mapm(ghat(2, gpt))+1
               n = mapn(ghat(3, gpt))+1
               mn = yzq(m, n)
               c(l, mn) = CONJG(pw%cc(gpt))

            END DO
!$OMP END PARALLEL DO
         END IF

      END IF

      cpy = REAL(ngpts, KIND=dp)*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_scatter_p

! **************************************************************************************************
!> \brief Generic function for 3d FFT of a coefficient_type or pw_type
!> \param pw1 ...
!> \param pw2 ...
!> \param debug ...
!> \par History
!>      JGH (30-12-2000): New setup of functions and adaptation to parallelism
!>      JGH (04-01-2001): Moved routine from pws to this module, only covers
!>                        pw_types, no more coefficient types
!> \author apsi
!> \note
!>       fft_wrap_pw1pw2
! **************************************************************************************************
   SUBROUTINE fft_wrap_pw1pw2(pw1, pw2, debug)

      TYPE(pw_type), INTENT(IN), TARGET                  :: pw1
      TYPE(pw_type), INTENT(INOUT), TARGET               :: pw2
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      CHARACTER(len=*), PARAMETER :: routineN = 'fft_wrap_pw1pw2', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=9)                                   :: mode
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: grays
      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER      :: c_in, c_out
      INTEGER                                            :: dir, handle, handle2, my_pos, nrays, &
                                                            out_space, output_unit, stat
      INTEGER, DIMENSION(3)                              :: nloc
      INTEGER, DIMENSION(:), POINTER                     :: n
      LOGICAL                                            :: test, use_pw_cuda
      REAL(KIND=dp)                                      :: norm

      CALL timeset(routineN, handle2)
      output_unit = cp_logger_get_default_io_unit()
      CALL timeset(routineN//"_"//TRIM(ADJUSTL(cp_to_string( &
                                               CEILING(pw1%pw_grid%cutoff/10)*10))), handle)

      IF(.NOT.(pw1%ref_count > 0))CALL cp__a("pw/pw_methods.F",1299)
      IF(.NOT.(pw2%ref_count > 0))CALL cp__a("pw/pw_methods.F",1300)
      NULLIFY (c_in)
      NULLIFY (c_out)

      IF (PRESENT(debug)) THEN
         test = debug
      ELSE
         test = .FALSE.
      END IF

      !..check if grids are compatible
      IF (pw1%pw_grid%id_nr /= pw2%pw_grid%id_nr) THEN
         IF (pw1%pw_grid%dvol /= pw2%pw_grid%dvol) THEN
            CALL cp__b("pw/pw_methods.F",1313,"PW grids not compatible")
         END IF
         IF (pw1%pw_grid%para%group /= pw2%pw_grid%para%group) THEN
            CALL cp__b("pw/pw_methods.F",1316,"PW grids have not compatible MPI groups")
         END IF
      END IF

      !..prepare input
      IF (pw1%in_space == REALSPACE) THEN
         dir = FWFFT
         norm = 1.0_dp/pw1%pw_grid%ngpts
         out_space = RECIPROCALSPACE
      ELSE IF (pw1%in_space == RECIPROCALSPACE) THEN
         dir = BWFFT
         norm = 1.0_dp
         out_space = REALSPACE
      ELSE
         CALL cp__b("pw/pw_methods.F",1330,"Error in space tag")
      END IF

      n => pw1%pw_grid%npts

      mode = fftselect(pw1%in_use, pw2%in_use, pw1%in_space)

      IF (pw1%pw_grid%para%mode == PW_MODE_LOCAL) THEN

         !
         !..replicated data, use local FFT
         !

         IF (test .AND. output_unit > 0) THEN
            WRITE (output_unit, '(A)') " FFT Protocol "
            IF (dir == FWFFT) WRITE (output_unit, '(A,T76,A)') "  Transform direction ", "FWFFT"
            IF (dir == BWFFT) WRITE (output_unit, '(A,T76,A)') "  Transform direction ", "BWFFT"
            IF (pw1%in_space == REALSPACE) &
               WRITE (output_unit, '(A,T72,A)') "  in space ", "REALSPACE"
            IF (pw1%in_space == RECIPROCALSPACE) &
               WRITE (output_unit, '(A,T66,A)') "  in space ", "RECIPROCALSPACE"
            IF (out_space == REALSPACE) &
               WRITE (output_unit, '(A,T72,A)') "  out space ", "REALSPACE"
            IF (out_space == RECIPROCALSPACE) &
               WRITE (output_unit, '(A,T66,A)') "  out space ", "RECIPROCALSPACE"
            WRITE (output_unit, '(A,T66,E15.6)') "  scale factor", norm
         END IF

         SELECT CASE (mode)
         CASE DEFAULT
            CALL cp__b("pw/pw_methods.F",1360,"Illegal combination of in_use and in_space")
         CASE ("FW_C3DC3D")
            c_in => pw1%cc3d
            c_out => pw2%cc3d
            CALL fft3d(dir, n, c_in, c_out, scale=norm, debug=test)
         CASE ("FW_R3DC3D")
            pw2%cc3d = CMPLX(pw1%cr3d, 0.0_dp, KIND=dp)
            c_out => pw2%cc3d
            CALL fft3d(dir, n, c_out, scale=norm, debug=test)
         CASE ("FW_C3DC1D")
            c_in => pw1%cc3d
            ALLOCATE (c_out(n(1), n(2), n(3)), STAT=stat)
            IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1372, &
                                            "c_out", 2*dp_size*PRODUCT(n(1:3)))
            ! transform
            CALL fft3d(dir, n, c_in, c_out, scale=norm, debug=test)
            ! gather results
            IF (test .AND. output_unit > 0) WRITE (output_unit, '(A)') "  PW_GATHER : 3d -> 1d "
            CALL pw_gather(pw2, c_out)
            DEALLOCATE (c_out)
         CASE ("FW_R3DC1D")
            ALLOCATE (c_out(n(1), n(2), n(3)), STAT=stat)
            IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1385, &
                                            "c_out", 2*dp_size*PRODUCT(n(1:3)))
            CALL copy_rc(pw1%cr3d, c_out)
            CALL fft3d(dir, n, c_out, scale=norm, debug=test)
            CALL pw_gather(pw2, c_out)
            DEALLOCATE (c_out)
         CASE ("BW_C3DC3D")
            c_in => pw1%cc3d
            c_out => pw2%cc3d
            CALL fft3d(dir, n, c_in, c_out, scale=norm, debug=test)
         CASE ("BW_C3DR3D")
            c_in => pw1%cc3d
            ALLOCATE (c_out(n(1), n(2), n(3)), STAT=stat)
            IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1399, &
                                            "c_out", 2*dp_size*PRODUCT(n(1:3)))
            CALL fft3d(dir, n, c_in, c_out, scale=norm, debug=test)
            ! use real part only
            IF (test .AND. output_unit > 0) WRITE (output_unit, '(A)') "  REAL part "
            pw2%cr3d = REAL(c_out, KIND=dp)
            DEALLOCATE (c_out)
         CASE ("BW_C1DC3D")
            c_out => pw2%cc3d
            IF (test .AND. output_unit > 0) WRITE (output_unit, '(A)') "  PW_SCATTER : 3d -> 1d "
            CALL pw_scatter(pw1, c_out)
            CALL fft3d(dir, n, c_out, scale=norm, debug=test)
         CASE ("BW_C1DR3D")
            ALLOCATE (c_out(n(1), n(2), n(3)), STAT=stat)
            IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1416, &
                                            "c_out", 2*dp_size*PRODUCT(n(1:3)))
            IF (test .AND. output_unit > 0) WRITE (output_unit, '(A)') "  PW_SCATTER : 3d -> 1d "
            CALL pw_scatter(pw1, c_out)
            ! transform
            CALL fft3d(dir, n, c_out, scale=norm, debug=test)
            ! use real part only
            IF (test .AND. output_unit > 0) WRITE (output_unit, '(A)') "  REAL part "
            CALL copy_cr(c_out, pw2%cr3d)
            DEALLOCATE (c_out)
         END SELECT

         IF (test .AND. output_unit > 0) WRITE (output_unit, '(A)') " End of FFT Protocol "

      ELSE

         !
         !..parallel FFT
         !

         IF (test .AND. pw1%pw_grid%para%group_head .AND. output_unit > 0) THEN
            WRITE (output_unit, '(A)') " FFT Protocol "
            IF (dir == FWFFT) WRITE (output_unit, '(A,T76,A)') "  Transform direction ", "FWFFT"
            IF (dir == BWFFT) WRITE (output_unit, '(A,T76,A)') "  Transform direction ", "BWFFT"
            IF (pw1%in_space == REALSPACE) &
               WRITE (output_unit, '(A,T72,A)') "  in space ", "REALSPACE"
            IF (pw1%in_space == RECIPROCALSPACE) &
               WRITE (output_unit, '(A,T66,A)') "  in space ", "RECIPROCALSPACE"
            IF (out_space == REALSPACE) &
               WRITE (output_unit, '(A,T72,A)') "  out space ", "REALSPACE"
            IF (out_space == RECIPROCALSPACE) &
               WRITE (output_unit, '(A,T66,A)') "  out space ", "RECIPROCALSPACE"
            WRITE (output_unit, '(A,T66,E15.6)') "  scale factor", norm
         END IF

         my_pos = pw1%pw_grid%para%my_pos
         nrays = pw1%pw_grid%para%nyzray(my_pos)
         grays => pw1%pw_grid%grays
         IF(.NOT.(SIZE(grays, 1) == n(1)))CALL cp__a("pw/pw_methods.F",1455)
         IF(.NOT.(SIZE(grays, 2) == nrays))CALL cp__a("pw/pw_methods.F",1456)

         SELECT CASE (mode)
         CASE DEFAULT
            CALL cp_abort(cp__l("pw/pw_methods.F",1460), &
                          "Illegal combination of in_use and in_space "// &
                          "in parallel 3d FFT")
         CASE ("FW_C3DC1D")
            !..prepare input
            c_in => pw1%cc3d
            CALL zero_c(grays)
            !..transform
            IF (pw1%pw_grid%para%ray_distribution) THEN
               CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%group, &
                          pw1%pw_grid%para%rs_group, &
                          pw1%pw_grid%para%yzp, pw1%pw_grid%para%nyzray, &
                          pw1%pw_grid%para%bo, scale=norm, debug=test)
            ELSE
               CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%rs_group, &
                          pw1%pw_grid%para%bo, scale=norm, debug=test)
            END IF
            !..prepare output
            IF (test .AND. pw1%pw_grid%para%group_head .AND. output_unit > 0) &
               WRITE (output_unit, '(A)') "  PW_GATHER : 2d -> 1d "
            CALL pw_gather(pw2, grays)
         CASE ("FW_R3DC1D")
            use_pw_cuda = .FALSE.
            IF (use_pw_cuda) THEN
               CALL pw_cuda_r3dc1d_3d_ps(pw1, pw2, scale=norm)
            ELSE
               !..prepare input
               nloc = pw1%pw_grid%npts_local
               ALLOCATE (c_in(nloc(1), nloc(2), nloc(3)), STAT=stat)
               IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1494, &
                                               "c_in", 2*dp_size*PRODUCT(nloc(1:3)))
               CALL copy_rc(pw1%cr3d, c_in)
               CALL zero_c(grays)
               !..transform
               IF (pw1%pw_grid%para%ray_distribution) THEN
                  CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%group, &
                             pw1%pw_grid%para%rs_group, &
                             pw1%pw_grid%para%yzp, pw1%pw_grid%para%nyzray, &
                             pw1%pw_grid%para%bo, scale=norm, debug=test)
               ELSE
                  CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%rs_group, &
                             pw1%pw_grid%para%bo, scale=norm, debug=test)
               END IF
               !..prepare output
               IF (test .AND. pw1%pw_grid%para%group_head .AND. output_unit > 0) &
                  WRITE (output_unit, '(A)') "  PW_GATHER : 2d -> 1d "
               CALL pw_gather(pw2, grays)
               DEALLOCATE (c_in)
            END IF
         CASE ("BW_C1DC3D")
            !..prepare input
            IF (test .AND. pw1%pw_grid%para%group_head .AND. output_unit > 0) &
               WRITE (output_unit, '(A)') "  PW_SCATTER : 2d -> 1d "
            CALL zero_c(grays)
            CALL pw_scatter(pw1, grays)
            c_in => pw2%cc3d
            !..transform
            IF (pw1%pw_grid%para%ray_distribution) THEN
               CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%group, &
                          pw1%pw_grid%para%rs_group, &
                          pw1%pw_grid%para%yzp, pw1%pw_grid%para%nyzray, &
                          pw1%pw_grid%para%bo, scale=norm, debug=test)
            ELSE
               CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%rs_group, &
                          pw1%pw_grid%para%bo, scale=norm, debug=test)
            END IF
            !..prepare output (nothing to do)
         CASE ("BW_C1DR3D")
            use_pw_cuda = .FALSE.
            IF (use_pw_cuda) THEN
               CALL pw_cuda_c1dr3d_3d_ps(pw1, pw2, scale=norm)
            ELSE
               !..prepare input
               IF (test .AND. pw1%pw_grid%para%group_head .AND. output_unit > 0) &
                  WRITE (output_unit, '(A)') "  PW_SCATTER : 2d -> 1d "
               CALL zero_c(grays)
               CALL pw_scatter(pw1, grays)
               nloc = pw2%pw_grid%npts_local
               ALLOCATE (c_in(nloc(1), nloc(2), nloc(3)), STAT=stat)
               IF (stat /= 0) CALL stop_memory(routineN, moduleN, 1549, &
                                               "c_in", 2*dp_size*PRODUCT(nloc(1:3)))
               !..transform
               IF (pw1%pw_grid%para%ray_distribution) THEN
                  CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%group, &
                             pw1%pw_grid%para%rs_group, &
                             pw1%pw_grid%para%yzp, pw1%pw_grid%para%nyzray, &
                             pw1%pw_grid%para%bo, scale=norm, debug=test)
               ELSE
                  CALL fft3d(dir, n, c_in, grays, pw1%pw_grid%para%rs_group, &
                             pw1%pw_grid%para%bo, scale=norm, debug=test)
               END IF
               !..prepare output
               IF (test .AND. pw1%pw_grid%para%group_head .AND. output_unit > 0) &
                  WRITE (output_unit, '(A)') "  Real part "
               CALL copy_cr(c_in, pw2%cr3d)
               DEALLOCATE (c_in)
            END IF
         END SELECT

      END IF

      !..update the space tag for pw2
      pw2%in_space = out_space

      IF (test .AND. pw1%pw_grid%para%group_head .AND. output_unit > 0) THEN
         WRITE (output_unit, '(A)') " End of FFT Protocol "
      END IF
      CALL timestop(handle)
      CALL timestop(handle2)

   END SUBROUTINE fft_wrap_pw1pw2

! **************************************************************************************************
!> \brief ...
!> \param use1 ...
!> \param use2 ...
!> \param space1 ...
!> \retval mode ...
! **************************************************************************************************
   FUNCTION fftselect(use1, use2, space1) RESULT(mode)
      INTEGER, INTENT(IN)                                :: use1, use2, space1
      CHARACTER(LEN=9)                                   :: mode

      CHARACTER(len=*), PARAMETER :: routineN = 'fftselect', routineP = moduleN//':'//routineN

      IF (space1 == REALSPACE) THEN
         mode(1:3) = "FW_"
      ELSE IF (space1 == RECIPROCALSPACE) THEN
         mode(1:3) = "BW_"
      ELSE
         CALL cp__b("pw/pw_methods.F",1600,"Error in space tag")
      END IF

      SELECT CASE (use1)
      CASE (COMPLEXDATA3D)
         mode(4:6) = "C3D"
      CASE (REALDATA3D)
         mode(4:6) = "R3D"
      CASE (COMPLEXDATA1D)
         mode(4:6) = "C1D"
      CASE (REALDATA1D)
         mode(4:6) = "R1D"
      CASE DEFAULT
         CALL cp__b("pw/pw_methods.F",1613,"Error in use1 tag")
      END SELECT

      SELECT CASE (use2)
      CASE (COMPLEXDATA3D)
         mode(7:9) = "C3D"
      CASE (REALDATA3D)
         mode(7:9) = "R3D"
      CASE (COMPLEXDATA1D)
         mode(7:9) = "C1D"
      CASE (REALDATA1D)
         mode(7:9) = "R1D"
      CASE DEFAULT
         CALL cp__b("pw/pw_methods.F",1626,"Error in use1 tag")
      END SELECT

   END FUNCTION fftselect

! **************************************************************************************************
!> \brief writes a small description of the actual grid
!>      (change to output the data as cube file, maybe with an
!>      optional long_description arg?)
!> \param pw the pw data to output
!> \param unit_nr the unit to output to
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_write(pw, unit_nr)
      TYPE(pw_type), INTENT(in)                          :: pw
      INTEGER, INTENT(in)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_write', routineP = moduleN//':'//routineN

      INTEGER                                            :: iostat

      WRITE (unit=unit_nr, fmt="('<pw>:{ id_nr=',i8,',')", iostat=iostat) &
         pw%id_nr

      SELECT CASE (pw%in_use)
      CASE (REALDATA1D)
         WRITE (unit=unit_nr, fmt="(a)", iostat=iostat) " in_use=REALDATA1D"
         IF (ASSOCIATED(pw%cr)) THEN
            WRITE (unit=unit_nr, fmt="(' cr=<real(',i8,':',i8,')>')") &
               LBOUND(pw%cr, 1), UBOUND(pw%cr, 1)
         ELSE
            WRITE (unit=unit_nr, fmt="(' cr=*null*')")
         END IF
      CASE (REALDATA3D)
         WRITE (unit=unit_nr, fmt="(a)", iostat=iostat) " in_use=REALDATA3D"
         IF (ASSOCIATED(pw%cr3d)) THEN
            WRITE (unit=unit_nr, fmt="(' cr3d=<real(',i8,':',i8,',',i8,':',i8,',',i8,':',i8,')>')") &
               LBOUND(pw%cr3d, 1), UBOUND(pw%cr3d, 1), LBOUND(pw%cr3d, 2), UBOUND(pw%cr3d, 2), &
               LBOUND(pw%cr3d, 3), UBOUND(pw%cr3d, 3)
         ELSE
            WRITE (unit=unit_nr, fmt="(' cr3d=*null*')")
         END IF
      CASE (COMPLEXDATA1D)
         WRITE (unit=unit_nr, fmt="(a)", iostat=iostat) " in_use=COMPLEXDATA1D"
         IF (ASSOCIATED(pw%cc)) THEN
            WRITE (unit=unit_nr, fmt="(' cc=<real(',i8,':',i8,')>')") &
               LBOUND(pw%cc, 1), UBOUND(pw%cc, 1)
         ELSE
            WRITE (unit=unit_nr, fmt="(' cc=*null*')")
         END IF
      CASE (COMPLEXDATA3D)
         WRITE (unit=unit_nr, fmt="(a)", iostat=iostat) " in_use=COMPLEXDATA3D"
         IF (ASSOCIATED(pw%cc3d)) THEN
            WRITE (unit=unit_nr, fmt="(' cc3d=<real(',i8,':',i8,',',i8,':',i8,',',i8,':',i8,')>')") &
               LBOUND(pw%cc3d, 1), UBOUND(pw%cc3d, 1), LBOUND(pw%cc3d, 2), UBOUND(pw%cc3d, 2), &
               LBOUND(pw%cc3d, 3), UBOUND(pw%cc3d, 3)
         ELSE
            WRITE (unit=unit_nr, fmt="(' cr3d=*null*')")
         END IF
      CASE default
         WRITE (unit=unit_nr, fmt="(' in_use=',i8,',')", iostat=iostat) &
            pw%in_use
      END SELECT

      SELECT CASE (pw%in_space)
      CASE (NOSPACE)
         WRITE (unit=unit_nr, fmt="(a)", iostat=iostat) " in_space=NOSPACE"
      CASE (REALSPACE)
         WRITE (unit=unit_nr, fmt="(a)", iostat=iostat) " in_space=REALSPACE"
      CASE (RECIPROCALSPACE)
         WRITE (unit=unit_nr, fmt="(a)", iostat=iostat) " in_space=RECIPROCALSPACE"
      CASE default
         WRITE (unit=unit_nr, fmt="(' in_space=',i8,',')", iostat=iostat) &
            pw%in_space
      END SELECT

      WRITE (unit=unit_nr, fmt="(' pw_grid%id_nr=',i8,/,' }')", iostat=iostat) &
         pw%pw_grid%id_nr

   END SUBROUTINE pw_write

! **************************************************************************************************
!> \brief ...
!> \param grida ...
!> \param gridb ...
!> \retval compat ...
! **************************************************************************************************
   FUNCTION pw_compatible(grida, gridb) RESULT(compat)
      TYPE(pw_grid_type), INTENT(IN)                     :: grida, gridb
      LOGICAL                                            :: compat

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_compatible', routineP = moduleN//':'//routineN

      compat = .FALSE.
      IF (grida%id_nr == gridb%id_nr) THEN
         compat = .TRUE.
      ELSE IF (grida%reference == gridb%id_nr) THEN
         compat = .TRUE.
      ELSE IF (gridb%reference == grida%id_nr) THEN
         compat = .TRUE.
      END IF

   END FUNCTION pw_compatible

! **************************************************************************************************
!> \brief Calculate integral over unit cell for functions in plane wave basis
!>      only returns the real part of it ......
!> \param pw1 ...
!> \param pw2 ...
!> \retval integral_value ...
!> \par History
!>      JGH (14-Mar-2001) : Parallel sum and some tests, HALFSPACE case
!> \author apsi
! **************************************************************************************************
   FUNCTION pw_integral_ab(pw1, pw2) RESULT(integral_value)

      TYPE(pw_type), INTENT(IN)                          :: pw1, pw2
      REAL(KIND=dp)                                      :: integral_value

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_integral_ab', routineP = moduleN//':'//routineN

      IF (pw1%pw_grid%id_nr /= pw2%pw_grid%id_nr) THEN
         CALL cp__b("pw/pw_methods.F",1750,"Grids incompatible")
      END IF

      ! since the return value is real, only do accurate sum on the real bit ?
      IF (pw1%in_use == REALDATA3D .AND. pw2%in_use == REALDATA3D) THEN
         integral_value = accurate_sum(pw1%cr3d(:, :, :) &
                                       *pw2%cr3d(:, :, :))
      ELSE IF (pw1%in_use == REALDATA3D &
               .AND. pw2%in_use == COMPLEXDATA3D) THEN
         integral_value = REAL(accurate_sum(pw1%cr3d(:, :, :) &
                                            *pw2%cc3d(:, :, :)), KIND=dp) !? complex bit
      ELSE IF (pw1%in_use == COMPLEXDATA3D &
               .AND. pw2%in_use == REALDATA3D) THEN
         integral_value = REAL(accurate_sum(pw1%cc3d(:, :, :) &
                                            *pw2%cr3d(:, :, :)), KIND=dp) !? complex bit
      ELSE IF (pw1%in_use == COMPLEXDATA3D &
               .AND. pw2%in_use == COMPLEXDATA3D) THEN
         integral_value = REAL(accurate_sum(CONJG(pw1%cc3d(:, :, :)) &
                                            *pw2%cc3d(:, :, :)), KIND=dp) !? complex bit

      ELSE IF (pw1%in_use == REALDATA1D &
               .AND. pw2%in_use == REALDATA1D) THEN
         integral_value = accurate_sum(pw1%cr(:)*pw2%cr(:))
      ELSE IF (pw1%in_use == REALDATA1D &
               .AND. pw2%in_use == COMPLEXDATA1D) THEN
         integral_value = REAL(accurate_sum(pw1%cr(:)*pw2%cc(:)), KIND=dp) !? complex bit
      ELSE IF (pw1%in_use == COMPLEXDATA1D &
               .AND. pw2%in_use == REALDATA1D) THEN
         integral_value = REAL(accurate_sum(pw1%cc(:)*pw2%cr(:)), KIND=dp) !? complex bit
      ELSE IF (pw1%in_use == COMPLEXDATA1D &
               .AND. pw2%in_use == COMPLEXDATA1D) THEN
         integral_value = REAL(accurate_sum(CONJG(pw1%cc(:))*pw2%cc(:)), KIND=dp) !? complex bit
      ELSE
         CALL cp__b("pw/pw_methods.F",1783,"No possible DATA")
      END IF

      IF (pw1%in_use == REALDATA3D .OR. pw1%in_use == COMPLEXDATA3D) THEN
         integral_value = integral_value*pw1%pw_grid%dvol
      ELSE
         integral_value = integral_value*pw1%pw_grid%vol
      ENDIF
      IF (pw1%in_use == COMPLEXDATA1D) THEN
         IF (pw1%pw_grid%grid_span == HALFSPACE) THEN
            integral_value = 2.0_dp*integral_value
            IF (pw1%pw_grid%have_g0) integral_value = integral_value- &
                                                      REAL(CONJG(pw1%cc(1))*pw2%cc(1), KIND=dp)
         END IF
      END IF

      IF (pw1%pw_grid%para%mode == PW_MODE_DISTRIBUTED) &
         CALL mp_sum(integral_value, pw1%pw_grid%para%group)

   END FUNCTION pw_integral_ab

! **************************************************************************************************
!> \brief ...
!> \param pw1 ...
!> \param pw2 ...
!> \retval integral_value ...
! **************************************************************************************************
   FUNCTION pw_integral_a2b(pw1, pw2) RESULT(integral_value)

      TYPE(pw_type), INTENT(IN)                          :: pw1, pw2
      REAL(KIND=dp)                                      :: integral_value

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_integral_a2b', &
         routineP = moduleN//':'//routineN

      IF (pw1%pw_grid%id_nr /= pw2%pw_grid%id_nr) THEN
         CALL cp__b("pw/pw_methods.F",1819,"Grids incompatible")
      END IF
      IF (pw1%in_use == REALDATA1D .AND. &
          pw2%in_use == REALDATA1D) THEN
         integral_value = accurate_sum(pw1%cr(:)*pw2%cr(:) &
                                       *pw1%pw_grid%gsq(:))
      ELSE IF (pw1%in_use == COMPLEXDATA1D .AND. &
               pw2%in_use == COMPLEXDATA1D) THEN
         integral_value = accurate_sum(REAL(CONJG(pw1%cc(:)) &
                                            *pw2%cc(:), KIND=dp)*pw1%pw_grid%gsq(:))
         IF (pw1%pw_grid%grid_span == HALFSPACE) THEN
            integral_value = 2.0_dp*integral_value
         END IF
      ELSE
         CALL cp__b("pw/pw_methods.F",1833,"No possible DATA")
      END IF

      IF (pw1%in_use == REALDATA3D .OR. pw1%in_use == COMPLEXDATA3D) THEN
         integral_value = integral_value*pw1%pw_grid%dvol
      ELSE
         integral_value = integral_value*pw1%pw_grid%vol
      END IF

      IF (pw1%pw_grid%para%mode == PW_MODE_DISTRIBUTED) &
         CALL mp_sum(integral_value, pw1%pw_grid%para%group)

   END FUNCTION pw_integral_a2b

! **************************************************************************************************
!> \brief Calculate the structure factor for point r
!> \param sf ...
!> \param r ...
!> \par History
!>      none
!> \author JGH (05-May-2006)
!> \note
!>      PW has to be in RECIPROCALSPACE and data in use is COMPLEXDATA1D
! **************************************************************************************************
   SUBROUTINE pw_structure_factor(sf, r)

      TYPE(pw_type), INTENT(INOUT)                       :: sf
      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: r

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_structure_factor', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: cnt, handle, ig
      REAL(KIND=dp)                                      :: arg, flop

      CALL timeset(routineN, handle)
      IF(.NOT.(sf%ref_count > 0))CALL cp__a("pw/pw_methods.F",1869)
      flop = 0.0_dp

      IF (sf%in_space == RECIPROCALSPACE .AND. &
          sf%in_use == COMPLEXDATA1D) THEN

         cnt = SIZE(sf%cc)

!$OMP PARALLEL DO PRIVATE (ig,arg) DEFAULT(NONE) SHARED(sf,r,cnt)
         DO ig = 1, cnt
            arg = DOT_PRODUCT(sf%pw_grid%g(:, ig), r)
            sf%cc(ig) = CMPLX(COS(arg), -SIN(arg), KIND=dp)
         END DO
         flop = flop+7*cnt
      ELSE

         CALL cp__b("pw/pw_methods.F",1885,"No suitable data field")

      END IF

      flop = flop*1.e-6_dp
      CALL timestop(handle)

   END SUBROUTINE pw_structure_factor

! **************************************************************************************************
!> \brief ...
!> \param fun ...
!> \param isign ...
!> \param oprt ...
!> \retval total_fun ...
! **************************************************************************************************
   FUNCTION pw_integrate_function(fun, isign, oprt) RESULT(total_fun)

      TYPE(pw_type), INTENT(IN)                          :: fun
      INTEGER, INTENT(IN), OPTIONAL                      :: isign
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: oprt
      REAL(KIND=dp)                                      :: total_fun

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_integrate_function', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iop

      iop = 0
      IF (PRESENT(oprt)) THEN
         SELECT CASE (oprt)
         CASE ("ABS", "abs")
            iop = 1
         CASE DEFAULT
            CALL cp__b("pw/pw_methods.F",1919,"Unknown operator")
         END SELECT
      END IF

      total_fun = 0._dp

      IF (fun%in_space == REALSPACE) THEN
         IF (fun%in_use == REALDATA3D) THEN
            ! do reduction using maximum accuracy
            IF (iop == 1) THEN
               total_fun = fun%pw_grid%dvol*accurate_sum(ABS(fun%cr3d))
            ELSE
               total_fun = fun%pw_grid%dvol*accurate_sum(fun%cr3d)
            END IF
         ELSE
            CALL cp__b("pw/pw_methods.F",1934,"In_space/in_use combination not implemented")
         END IF
      ELSEIF (fun%in_space == RECIPROCALSPACE) THEN
         IF (iop == 1) &
            CALL cp__b("pw/pw_methods.F",1938,"Operator ABS not implemented")
         IF (fun%in_use == COMPLEXDATA1D) THEN
            IF (fun%pw_grid%have_g0) total_fun = REAL(fun%pw_grid%vol*fun%cc(1), KIND=dp)
         ELSE
            CALL cp__b("pw/pw_methods.F",1942,"In_space/in_use combination not implemented")
         END IF
      ELSE
         CALL cp__b("pw/pw_methods.F",1945,"No space defined")
      END IF
      IF (fun%pw_grid%para%mode /= PW_MODE_LOCAL) THEN
         CALL mp_sum(total_fun, fun%pw_grid%para%group)
      END IF
      IF (PRESENT(isign)) THEN
         total_fun = total_fun*SIGN(1._dp, REAL(isign, dp))
      END IF

   END FUNCTION pw_integrate_function

END MODULE pw_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Manages a pool of grids (to be used for example as tmp objects),
!>      but can also be used to instantiate grids that are never given back.
!>
!>      Multigrid pools are just an array of pw_pools
!> \note
!>      The pool could also work without pointers (doing = each time),
!>      but I find it *very* ugly.
!>
!>      The pool could be integrated into pw_grid_type, I don't know if
!>      it would be a good or bad idea (but would add a circular dependence
!>      between pw and pw_grid types).
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE pw_pool_types
   USE cp_linked_list_3d_r,             ONLY: cp_sll_3d_r_dealloc,&
                                              cp_sll_3d_r_get_first_el,&
                                              cp_sll_3d_r_get_length,&
                                              cp_sll_3d_r_insert_el,&
                                              cp_sll_3d_r_next,&
                                              cp_sll_3d_r_rm_first_el,&
                                              cp_sll_3d_r_type
   USE cp_linked_list_pw,               ONLY: cp_sll_pw_dealloc,&
                                              cp_sll_pw_get_first_el,&
                                              cp_sll_pw_get_length,&
                                              cp_sll_pw_insert_el,&
                                              cp_sll_pw_next,&
                                              cp_sll_pw_rm_first_el,&
                                              cp_sll_pw_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE kinds,                           ONLY: dp
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_compare,&
                                              pw_grid_release,&
                                              pw_grid_retain
   USE pw_methods,                      ONLY: pw_write
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              COMPLEXDATA3D,&
                                              REALDATA1D,&
                                              REALDATA3D,&
                                              pw_create,&
                                              pw_p_type,&
                                              pw_release,&
                                              pw_type
   USE timings,                         ONLY: print_stack
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_pool_types'
   INTEGER, SAVE, PRIVATE :: last_pw_pool_id_nr = 0
   INTEGER, PARAMETER :: default_max_cache = 75, max_max_cache = 150

   PUBLIC :: pw_pool_type, pw_pool_p_type
   PUBLIC :: pw_pool_create, pw_pool_retain, pw_pool_release, &
             pw_pool_create_pw, pw_pool_give_back_pw, &
             pw_pool_create_cr3d, pw_pool_give_back_cr3d
   PUBLIC :: pw_pools_copy, pw_pools_dealloc, &
             pw_pools_create_pws, pw_pools_give_back_pws

! **************************************************************************************************
!> \brief Manages a pool of grids (to be used for example as tmp objects),
!>      but can also be used to intantiate grids that are never given back.
!> \param ref_count reference count (see /cp2k/doc/ReferenceCounting.html)
!> \param id_nr number that identifies each pool
!> \param real 1d_pw, real3d_pw, complex1d_pw, complex3d_pw: liked list with
!>        the cached grids of the corresponding type
!> \note
!>      As of now I would like replace the linked lists by arrays
!>      (no annoying list elements that are allocated would show up when
!>      tracking leaks) [fawzi]
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE pw_pool_type
      INTEGER :: ref_count, id_nr, max_cache
      TYPE(pw_grid_type), POINTER :: pw_grid
      TYPE(cp_sll_pw_type), POINTER :: real1d_pw, real3d_pw, &
                                       complex1d_pw, complex3d_pw
      TYPE(cp_sll_3d_r_type), POINTER :: real3d_array
   END TYPE pw_pool_type

! **************************************************************************************************
!> \brief to create arrays of pools
!> \param pool the pool
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE pw_pool_p_type
      TYPE(pw_pool_type), POINTER :: pool
   END TYPE pw_pool_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a pool for pw
!> \param pool the pool to create
!> \param pw_grid the grid that is used to create the pw
!> \param max_cache ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pool_create(pool, pw_grid, max_cache)
      TYPE(pw_pool_type), POINTER                        :: pool
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, OPTIONAL                                  :: max_cache

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_create', routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()

      ALLOCATE (pool)
      pool%pw_grid => pw_grid
      CALL pw_grid_retain(pw_grid)
      last_pw_pool_id_nr = last_pw_pool_id_nr+1
      pool%id_nr = last_pw_pool_id_nr
      pool%ref_count = 1
      pool%max_cache = default_max_cache
      IF (PRESENT(max_cache)) pool%max_cache = max_cache
      pool%max_cache = MIN(max_max_cache, pool%max_cache)
      IF (debug_this_module) THEN
         WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                fmt="(' *** pw_pool ',i4,' has been created')") pool%id_nr
         CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
      END IF
      NULLIFY (pool%real1d_pw, pool%real3d_pw, &
               pool%complex1d_pw, pool%complex3d_pw, pool%real3d_array)
   END SUBROUTINE pw_pool_create

! **************************************************************************************************
!> \brief retains the pool (see cp2k/doc/ReferenceCounting.html)
!> \param pool the pool to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pool_retain(pool)
      TYPE(pw_pool_type), POINTER                        :: pool

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_retain', routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("pw/pw_pool_types.F",162)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("pw/pw_pool_types.F",163)

      pool%ref_count = pool%ref_count+1
      IF (debug_this_module) THEN
         WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                fmt="(' *** pw_pool ',i4,' has been retained, ref_count=',i4)") &
            pool%id_nr, pool%ref_count
         CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
      END IF
   END SUBROUTINE pw_pool_retain

! **************************************************************************************************
!> \brief deallocates all the cached grids
!> \param pool the pool to flush
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pool_flush_cache(pool)
      TYPE(pw_pool_type), POINTER                        :: pool

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_flush_cache', &
         routineP = moduleN//':'//routineN

      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: array_att
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_sll_3d_r_type), POINTER                    :: array_iterator
      TYPE(cp_sll_pw_type), POINTER                      :: iterator
      TYPE(pw_type), POINTER                             :: pw_el

      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("pw/pw_pool_types.F",193)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("pw/pw_pool_types.F",194)
      NULLIFY (iterator, array_iterator, pw_el, array_att)
      logger => cp_get_default_logger()
      IF (debug_this_module) THEN
         WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                fmt="(' *** pw_pool ',i4,' is flushing the cache')") pool%id_nr
         CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
      END IF

      iterator => pool%real1d_pw
      DO
         IF (.NOT. cp_sll_pw_next(iterator, el_att=pw_el)) EXIT
         IF(.NOT.(pw_el%ref_count == 0))CALL cp__a("pw/pw_pool_types.F",206)
         pw_el%ref_count = 1
         CALL pw_release(pw_el)
      END DO
      CALL cp_sll_pw_dealloc(pool%real1d_pw)

      iterator => pool%real3d_pw
      DO
         IF (.NOT. cp_sll_pw_next(iterator, el_att=pw_el)) EXIT
         IF(.NOT.(pw_el%ref_count == 0))CALL cp__a("pw/pw_pool_types.F",215)
         pw_el%ref_count = 1
         CALL pw_release(pw_el)
      END DO
      CALL cp_sll_pw_dealloc(pool%real3d_pw)

      iterator => pool%complex1d_pw
      DO
         IF (.NOT. cp_sll_pw_next(iterator, el_att=pw_el)) EXIT
         IF(.NOT.(pw_el%ref_count == 0))CALL cp__a("pw/pw_pool_types.F",224)
         pw_el%ref_count = 1
         CALL pw_release(pw_el)
      END DO
      CALL cp_sll_pw_dealloc(pool%complex1d_pw)

      iterator => pool%complex3d_pw
      DO
         IF (.NOT. cp_sll_pw_next(iterator, el_att=pw_el)) EXIT
         IF(.NOT.(pw_el%ref_count == 0))CALL cp__a("pw/pw_pool_types.F",233)
         pw_el%ref_count = 1
         CALL pw_release(pw_el)
      END DO
      CALL cp_sll_pw_dealloc(pool%complex3d_pw)

      array_iterator => pool%real3d_array
      DO
         IF (.NOT. cp_sll_3d_r_next(array_iterator, el_att=array_att) &
             ) EXIT
         DEALLOCATE (array_att)
      END DO
      CALL cp_sll_3d_r_dealloc(pool%real3d_array)

   END SUBROUTINE pw_pool_flush_cache

! **************************************************************************************************
!> \brief releases the given pool (see cp2k/doc/ReferenceCounting.html)
!> \param pool the pool to release
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pool_release(pool)
      TYPE(pw_pool_type), POINTER                        :: pool

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_release', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()

      IF (ASSOCIATED(pool)) THEN
         IF(.NOT.(pool%ref_count > 0))CALL cp__a("pw/pw_pool_types.F",267)
         pool%ref_count = pool%ref_count-1
         IF (debug_this_module) THEN
            WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                   fmt="(' *** pw_pool ',i4,' released ref_count=',i4)") &
               pool%id_nr, pool%ref_count
            CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
         END IF
         IF (pool%ref_count == 0) THEN
            pool%ref_count = 1
            CALL pw_pool_flush_cache(pool)
            pool%ref_count = 0
            IF(.NOT.(ASSOCIATED(pool%pw_grid)))CALL cp__a("pw/pw_pool_types.F",279)
            CALL pw_grid_release(pool%pw_grid)

            DEALLOCATE (pool)
         END IF
      END IF
      NULLIFY (pool)
   END SUBROUTINE pw_pool_release

! **************************************************************************************************
!> \brief tries to pop an element from the given list (no error on failure)
!> \param list the list to pop
!> \retval res ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      private function
! **************************************************************************************************
   FUNCTION try_pop(list) RESULT(res)
      TYPE(cp_sll_pw_type), POINTER                      :: list
      TYPE(pw_type), POINTER                             :: res

      IF (ASSOCIATED(list)) THEN
         res => cp_sll_pw_get_first_el(list)
         CALL cp_sll_pw_rm_first_el(list)
      ELSE
         NULLIFY (res)
      END IF
   END FUNCTION try_pop

! **************************************************************************************************
!> \brief returns a pw, allocating it if none is in the pool
!> \param pool the pool from where you get the pw
!> \param pw will contain the new pw
!> \param use_data which data it uses: REALDATA1D, COMPLEXDATA1D,
!>        REALDATA3D, COMPLEXDATA3D
!> \param in_space in which space it is: REALSPACE, RECIPROCALSPACE
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pool_create_pw(pool, pw, use_data, in_space)
      TYPE(pw_pool_type), POINTER                        :: pool
      TYPE(pw_type), POINTER                             :: pw
      INTEGER, INTENT(in)                                :: use_data
      INTEGER, INTENT(in), OPTIONAL                      :: in_space

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_create_pw', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: cr3d_ptr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      NULLIFY (pw)
      NULLIFY (cr3d_ptr)
      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("pw/pw_pool_types.F",338)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("pw/pw_pool_types.F",339)

      SELECT CASE (use_data)
      CASE (REALDATA1D)
         pw => try_pop(pool%real1d_pw)
      CASE (REALDATA3D)
         pw => try_pop(pool%real3d_pw)
         IF (.NOT. ASSOCIATED(pw)) THEN
            IF (ASSOCIATED(pool%real3d_array)) THEN
               cr3d_ptr => cp_sll_3d_r_get_first_el(pool%real3d_array)
               CALL cp_sll_3d_r_rm_first_el(pool%real3d_array)
            END IF
         END IF
      CASE (COMPLEXDATA1D)
         pw => try_pop(pool%complex1d_pw)
      CASE (COMPLEXDATA3D)
         pw => try_pop(pool%complex3d_pw)
      CASE default
! unknown use_data
         CALL cp__b("pw/pw_pool_types.F",358,"")
      END SELECT

      IF (.NOT. ASSOCIATED(pw)) THEN
         CALL pw_create(pw, pool%pw_grid, use_data=use_data, &
                        cr3d_ptr=cr3d_ptr)
         IF (debug_this_module) THEN
            IF (ASSOCIATED(cr3d_ptr)) THEN
               WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                      fmt="(' *** pw_pool ',i4,' created pw ',i4,'reusing 3dr array')") &
                  pool%id_nr, pw%id_nr
            ELSE
               WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                      fmt="(' *** pw_pool ',i4,' created pw ',i4)") &
                  pool%id_nr, pw%id_nr
            END IF
            CALL pw_write( &
               unit_nr=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
               pw=pw)
            CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
         END IF
      ELSE
         IF(.NOT.(pw%ref_count == 0))CALL cp__a("pw/pw_pool_types.F",380)
         pw%ref_count = 1
         IF (debug_this_module) THEN
            WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                   fmt="(' *** pw_pool ',i4,' created pw reusing old ',i4)") &
               pool%id_nr, pw%id_nr
            CALL pw_write( &
               unit_nr=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
               pw=pw)
            CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
         END IF
      END IF

      pw%in_space = 0
      IF (PRESENT(in_space)) pw%in_space = in_space

      CALL timestop(handle)

   END SUBROUTINE pw_pool_create_pw

! **************************************************************************************************
!> \brief returns the pw to the pool
!> \param pool the pool where to reintegrate the pw
!> \param pw the pw to give back
!> \param accept_non_compatible if non compatible pw should be accepted
!>        (they will be destroied). Defaults to false (and thus stops with
!>        an error)
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pool_give_back_pw(pool, pw, accept_non_compatible)
      TYPE(pw_pool_type), POINTER                        :: pool
      TYPE(pw_type), POINTER                             :: pw
      LOGICAL, INTENT(in), OPTIONAL                      :: accept_non_compatible

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_give_back_pw', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: failure, my_accept_non_compatible
      TYPE(cp_logger_type), POINTER                      :: logger

      failure = .FALSE.

      my_accept_non_compatible = .FALSE.
      logger => cp_get_default_logger()
      IF (PRESENT(accept_non_compatible)) my_accept_non_compatible = accept_non_compatible

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("pw/pw_pool_types.F",430)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("pw/pw_pool_types.F",431)
      IF (.NOT. ASSOCIATED(pw)) THEN
         IF(.NOT.(my_accept_non_compatible))CALL cp__a("pw/pw_pool_types.F",433)
         failure = .TRUE.
      END IF
      IF (.NOT. failure) THEN
         IF(.NOT.(pw%ref_count == 1))CALL cp__a("pw/pw_pool_types.F",437)
         IF (.NOT. pw_grid_compare(pw%pw_grid, pool%pw_grid)) THEN
            IF (debug_this_module) THEN
               WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                      fmt="(' *** pw_pool ',i4,' giving back incompatible pw ',i4)") &
                  pool%id_nr, pw%id_nr
               CALL pw_write(unit_nr=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                             pw=pw)
               CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
            END IF

            IF (.NOT. my_accept_non_compatible) &
               CALL cp__b("pw/pw_pool_types.F",449,"pool cannot reuse pw of another grid")
            CALL pw_release(pw)
            failure = .TRUE.
         END IF
      END IF
      IF (.NOT. failure) THEN
         IF (debug_this_module) THEN
            WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                   fmt="(' *** pw_pool ',i4,' giving back pw ',i4)") &
               pool%id_nr, pw%id_nr
            CALL pw_write(unit_nr=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                          pw=pw)
            CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
         END IF

         SELECT CASE (pw%in_use)
         CASE (REALDATA1D)
            IF (cp_sll_pw_get_length(pool%real1d_pw) < pool%max_cache) THEN
               CALL cp_sll_pw_insert_el(pool%real1d_pw, el=pw)
            ELSE
               IF (max_max_cache >= 0) &
                  CALL cp__w("pw/pw_pool_types.F",470,"hit max_cache")
               CALL pw_release(pw)
            END IF
         CASE (REALDATA3D)
            IF (ASSOCIATED(pw%cr3d)) THEN
               IF (cp_sll_pw_get_length(pool%real3d_pw) < pool%max_cache) THEN
                  CALL cp_sll_pw_insert_el(pool%real3d_pw, el=pw)
               ELSE
                  IF (max_max_cache >= 0) &
                     CALL cp__w("pw/pw_pool_types.F",479,"hit max_cache")
                  CALL pw_release(pw)
               END IF
            ELSE
               IF (debug_this_module) THEN
                  WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                         fmt="(' *** pw_pool ',i4,' pw ',i4,' cr3d is not associated, discarding')") &
                     pool%id_nr, pw%id_nr
                  CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
               END IF
               IF(.NOT.(my_accept_non_compatible))CALL cp__a("pw/pw_pool_types.F",489)
               CALL pw_release(pw)
            END IF
         CASE (COMPLEXDATA1D)
            IF (cp_sll_pw_get_length(pool%complex1d_pw) < pool%max_cache) THEN
               CALL cp_sll_pw_insert_el(pool%complex1d_pw, el=pw)
            ELSE
               IF (max_max_cache >= 0) &
                  CALL cp__w("pw/pw_pool_types.F",497,"hit max_cache")
               CALL pw_release(pw)
            END IF
         CASE (COMPLEXDATA3D)
            IF (cp_sll_pw_get_length(pool%complex3d_pw) < pool%max_cache) THEN
               CALL cp_sll_pw_insert_el(pool%complex3d_pw, el=pw)
            ELSE
               IF (max_max_cache >= 0) &
                  CALL cp__w("pw/pw_pool_types.F",505,"hit max_cache")
               CALL pw_release(pw)
            END IF
         CASE default
            ! unknown in_use
            CALL cp__b("pw/pw_pool_types.F",510,"")
         END SELECT
         IF (ASSOCIATED(pw)) pw%ref_count = 0
         !FM so that if someone tries to use a pw that is in the pool
         !FM (s)he gets problems
      END IF
      NULLIFY (pw)
      CALL timestop(handle)
   END SUBROUTINE pw_pool_give_back_pw

! **************************************************************************************************
!> \brief returns a 3d real array of coefficents as the one used by pw with
!>      REALDATA3D, allocating it if none is present in the pool
!> \param pw_pool the pool that caches the cr3d
!> \param cr3d the pointer that will contain the array
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_pool_create_cr3d(pw_pool, cr3d)
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: cr3d

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_create_cr3d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: stat
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(pw_type), POINTER                             :: pw

      NULLIFY (pw)
      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(pw_pool)))CALL cp__a("pw/pw_pool_types.F",543)
      IF(.NOT.(pw_pool%ref_count > 0))CALL cp__a("pw/pw_pool_types.F",544)
      IF(.NOT.(.NOT. ASSOCIATED(cr3d)))CALL cp__a("pw/pw_pool_types.F",545)
      IF (ASSOCIATED(pw_pool%real3d_array)) THEN
         cr3d => cp_sll_3d_r_get_first_el(pw_pool%real3d_array)
         CALL cp_sll_3d_r_rm_first_el(pw_pool%real3d_array)
      ELSE
         pw => try_pop(pw_pool%real3d_pw)
         IF (ASSOCIATED(pw)) THEN
            IF(.NOT.(pw%ref_count == 0))CALL cp__a("pw/pw_pool_types.F",552)
            pw%ref_count = 1
            cr3d => pw%cr3d
            NULLIFY (pw%cr3d)
            CALL pw_release(pw)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(cr3d)) THEN
         ALLOCATE (cr3d(pw_pool%pw_grid%bounds_local(1, 1):pw_pool%pw_grid%bounds_local(2, 1), &
                        pw_pool%pw_grid%bounds_local(1, 2):pw_pool%pw_grid%bounds_local(2, 2), &
                        pw_pool%pw_grid%bounds_local(1, 3):pw_pool%pw_grid%bounds_local(2, 3)), &
                   stat=stat)
         IF(.NOT.(stat == 0))CALL cp__a("pw/pw_pool_types.F",564)
         IF (debug_this_module) THEN
            WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                   fmt="(' *** pw_pool ',i4,' created cr3d')") pw_pool%id_nr
            CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
         END IF
      ELSEIF (debug_this_module) THEN
         WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                fmt="(' *** pw_pool ',i4,' created cr3d')") pw_pool%id_nr
         CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
      END IF
   END SUBROUTINE pw_pool_create_cr3d

! **************************************************************************************************
!> \brief returns a 3d real array of coefficents as the one used by pw with
!>      REALDATA3D, allocating it if none is present in the pool
!> \param pw_pool the pool that caches the cr3d
!> \param cr3d the pointer that will contain the array
!> \param accept_non_compatible if true deallocates the non compatible
!>        arrays passed in, if false (the default) stops with an error
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_pool_give_back_cr3d(pw_pool, cr3d, accept_non_compatible)
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: cr3d
      LOGICAL, INTENT(in), OPTIONAL                      :: accept_non_compatible

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pool_give_back_cr3d', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: compatible, my_accept_non_compatible
      TYPE(cp_logger_type), POINTER                      :: logger

      my_accept_non_compatible = .FALSE.
      logger => cp_get_default_logger()
      IF (PRESENT(accept_non_compatible)) my_accept_non_compatible = accept_non_compatible

      IF(.NOT.(ASSOCIATED(pw_pool)))CALL cp__a("pw/pw_pool_types.F",603)
      IF(.NOT.(pw_pool%ref_count > 0))CALL cp__a("pw/pw_pool_types.F",604)
      IF (ASSOCIATED(cr3d)) THEN
         IF (debug_this_module) THEN
            WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                   fmt="(' *** pw_pool ',i4,' received back a cr3d')") pw_pool%id_nr
            CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
         END IF
         compatible = ALL(MERGE(pw_pool%pw_grid%bounds_local(1, :) == LBOUND(cr3d) .AND. &
                                pw_pool%pw_grid%bounds_local(2, :) == UBOUND(cr3d), &
                                pw_pool%pw_grid%bounds_local(2, :) < pw_pool%pw_grid%bounds_local(1, :), &
                                UBOUND(cr3d) >= LBOUND(cr3d)))
         IF(.NOT.(compatible .OR. my_accept_non_compatible))CALL cp__a("pw/pw_pool_types.F",615)
         IF (compatible) THEN
            IF (cp_sll_3d_r_get_length(pw_pool%real3d_array) < pw_pool%max_cache) THEN
               CALL cp_sll_3d_r_insert_el(pw_pool%real3d_array, el=cr3d)
            ELSE
               IF (max_max_cache >= 0) &
                  CALL cp__w("pw/pw_pool_types.F",621,"hit max_cache")
               DEALLOCATE (cr3d)
            END IF
         ELSE
            IF (debug_this_module) THEN
               WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                      fmt="(' *** pw_pool ',i4,' cr3d not accepted, deallocating')") &
                  pw_pool%id_nr
               CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
            END IF
            DEALLOCATE (cr3d)
         END IF
      ELSE
         IF (debug_this_module) THEN
            WRITE (unit=cp_logger_get_default_unit_nr(logger, local=.TRUE.), &
                   fmt="(' *** pw_pool ',i4,' received back a null cr3d')") &
               pw_pool%id_nr
            CALL print_stack(cp_logger_get_default_unit_nr(logger, local=.TRUE.))
         END IF
         IF(.NOT.(my_accept_non_compatible))CALL cp__a("pw/pw_pool_types.F",640)
      END IF
      NULLIFY (cr3d)
   END SUBROUTINE pw_pool_give_back_cr3d

! **************************************************************************************************
!> \brief creates a multigrid structure
!> \param pools the multigrid pool (i.e. an array of pw_pool)
!> \param pws the multigrid of coefficent you want to initialize
!> \param use_data which data it uses: REALDATA1D, COMPLEXDATA1D,
!>        REALDATA3D, COMPLEXDATA3D
!> \param in_space ...
!> \par History
!>      07.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pools_create_pws(pools, pws, use_data, in_space)
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pools
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: pws
      INTEGER, INTENT(in)                                :: use_data
      INTEGER, INTENT(in), OPTIONAL                      :: in_space

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pools_create_pws', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(pools)))CALL cp__a("pw/pw_pool_types.F",667)
      ALLOCATE (pws(SIZE(pools)))
      DO i = 1, SIZE(pools)
         NULLIFY (pws(i)%pw)
         CALL pw_pool_create_pw(pools(i)%pool, pws(i)%pw, use_data, &
                                in_space=in_space)
      END DO
   END SUBROUTINE pw_pools_create_pws

! **************************************************************************************************
!> \brief returns the pw part of the coefficents into the pools
!> \param pools the pools that will cache the pws %pw
!> \param pws the coefficents to give back
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pools_give_back_pws(pools, pws)
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pools
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: pws

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pools_give_back_pws', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(pools)))CALL cp__a("pw/pw_pool_types.F",693)
      IF(.NOT.(ASSOCIATED(pws)))CALL cp__a("pw/pw_pool_types.F",694)
      IF(.NOT.(SIZE(pws) == SIZE(pools)))CALL cp__a("pw/pw_pool_types.F",695)
      DO i = 1, SIZE(pools)
         CALL pw_pool_give_back_pw(pools(i)%pool, pws(i)%pw)
      END DO
      DEALLOCATE (pws)
   END SUBROUTINE pw_pools_give_back_pws

! **************************************************************************************************
!> \brief copies a multigrid pool, the underlying pools are shared
!> \param source_pools the pools to copy
!> \param target_pools will hold the copy of the pools
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pools_copy(source_pools, target_pools)
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: source_pools, target_pools

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pools_copy', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(source_pools)))CALL cp__a("pw/pw_pool_types.F",717)
      ALLOCATE (target_pools(SIZE(source_pools)))
      DO i = 1, SIZE(source_pools)
         target_pools(i)%pool => source_pools(i)%pool
         CALL pw_pool_retain(source_pools(i)%pool)
      END DO
   END SUBROUTINE pw_pools_copy

! **************************************************************************************************
!> \brief deallocates the given pools (releasing each of the underlying
!>      pools)
!> \param pools the pols to deallocate
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_pools_dealloc(pools)
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pools

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_pools_dealloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(pools)) THEN
         DO i = 1, SIZE(pools)
            CALL pw_pool_release(pools(i)%pool)
         END DO
         DEALLOCATE (pools)
      END IF
      NULLIFY (pools)
   END SUBROUTINE pw_pools_dealloc

END MODULE pw_pool_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Dirichlet boundary condition data types
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!>       10.2015 completely revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE dirichlet_bc_types

   USE kinds,                           ONLY: dp
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: pw_release,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dirichlet_bc_types'

! main types
   TYPE :: dirichlet_bc_parameters
      LOGICAL                                  :: verbose_output
      LOGICAL                                  :: do_dbc_cube = .FALSE.

      INTEGER                                  :: n_aa_planar
      INTEGER, DIMENSION(:, :), POINTER         :: aa_planar_nprtn => NULL()
      INTEGER, DIMENSION(:), POINTER           :: aa_planar_pplane => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: aa_planar_xxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: aa_planar_yxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: aa_planar_zxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: aa_planar_vD => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: aa_planar_sigma => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: aa_planar_thickness => NULL()

      INTEGER                                  :: n_planar
      INTEGER, DIMENSION(:, :), POINTER         :: planar_nprtn => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: planar_Avtx => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: planar_Bvtx => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: planar_Cvtx => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: planar_Dvtx => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: planar_vD => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: planar_sigma => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: planar_thickness => NULL()

      INTEGER                                  :: n_xaa_cylindrical
      INTEGER, DIMENSION(:, :), POINTER         :: xaa_cylindrical_nprtn => NULL()
      INTEGER, DIMENSION(:), POINTER           :: xaa_cylindrical_nsides => NULL()
      INTEGER, DIMENSION(:), POINTER           :: xaa_cylindrical_apxtyp => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: xaa_cylindrical_xxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: xaa_cylindrical_bctr => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: xaa_cylindrical_brad => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: xaa_cylindrical_vD => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: xaa_cylindrical_sigma => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: xaa_cylindrical_thickness => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: xaa_cylindrical_sgap => NULL()

      INTEGER                                  :: n_aa_cuboidal
      INTEGER, DIMENSION(:, :), POINTER         :: aa_cuboidal_nprtn => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: aa_cuboidal_xxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: aa_cuboidal_yxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: aa_cuboidal_zxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: aa_cuboidal_vD => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER     :: aa_cuboidal_sigma => NULL()

   END TYPE dirichlet_bc_parameters

   TYPE :: dirichlet_bc_type
      REAL(dp)                          :: v_D
      INTEGER                           :: dbc_id
      INTEGER                           :: dbc_geom
      REAL(dp), DIMENSION(3, 8)          :: vertices
      TYPE(tile_p_type), &
         DIMENSION(:), POINTER           :: tiles => NULL()
      INTEGER                           :: n_tiles
      REAL(dp)                          :: smoothing_width
   END TYPE dirichlet_bc_type

   TYPE :: dirichlet_bc_p_type
      TYPE(dirichlet_bc_type), POINTER  :: dirichlet_bc
   END TYPE dirichlet_bc_p_type

   TYPE :: tile_type
      INTEGER                           :: tile_id
      REAL(dp), DIMENSION(3, 8)          :: vertices
      TYPE(pw_type), POINTER            :: tile_pw => NULL()
      REAL(dp)                          :: volume
   END TYPE tile_type

   TYPE :: tile_p_type
      TYPE(tile_type), POINTER          :: tile
   END TYPE tile_p_type

! publicly accessible entities
   PUBLIC dirichlet_bc_type, dirichlet_bc_p_type, tile_type, tile_p_type
   PUBLIC dirichlet_bc_parameters

   PUBLIC dbc_release, &
      dbc_tile_release, &
      dbc_parameters_dealloc

! parameters
   INTEGER, PARAMETER, PUBLIC           :: AA_PLANAR = 10, &
                                           PLANAR = 11, &
                                           CYLINDRICAL = 20, &
                                           AA_CUBOIDAL = 30

   INTEGER, PARAMETER, PUBLIC           :: CIRCUMSCRIBED = 1, &
                                           INSCRIBED = 2

   INTEGER, PARAMETER, PUBLIC           :: xy_plane = 110, &
                                           xz_plane = 101, &
                                           yz_plane = 011

CONTAINS

! **************************************************************************************************
!> \brief   releases the defined Dirichlet boundary region
!> \param contacts the DBC region to be released
!> \param do_dbc_cube ...
!> \param pw_pool pool of the plane wave grid
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dbc_release(contacts, do_dbc_cube, pw_pool)

      TYPE(dirichlet_bc_p_type), DIMENSION(:), &
         INTENT(INOUT), POINTER                          :: contacts
      LOGICAL, INTENT(IN)                                :: do_dbc_cube
      TYPE(pw_pool_type), INTENT(IN), OPTIONAL, POINTER  :: pw_pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbc_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, n_contacts

      CALL timeset(routineN, handle)

      n_contacts = SIZE(contacts)

      IF (ASSOCIATED(contacts)) THEN
         IF (do_dbc_cube) THEN
            DO i = 1, n_contacts
               CALL dbc_tile_release(contacts(i)%dirichlet_bc, pw_pool)
               DEALLOCATE (contacts(i)%dirichlet_bc)
            END DO
         ELSE
            DO i = 1, n_contacts
               DEALLOCATE (contacts(i)%dirichlet_bc)
            END DO
         END IF
         DEALLOCATE (contacts)
      END IF

      CALL timestop(handle)

   END SUBROUTINE dbc_release

! **************************************************************************************************
!> \brief   releases tiles
!> \param dbc ...
!> \param pw_pool pool of the plane wave grid
!> \par History
!>       11.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dbc_tile_release(dbc, pw_pool)

      TYPE(dirichlet_bc_type), INTENT(INOUT), POINTER    :: dbc
      TYPE(pw_pool_type), INTENT(IN), OPTIONAL, POINTER  :: pw_pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbc_tile_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k, n_tiles

      CALL timeset(routineN, handle)

      n_tiles = dbc%n_tiles
      IF (PRESENT(pw_pool)) THEN
         DO k = 1, n_tiles
            CALL pw_pool_give_back_pw(pw_pool, dbc%tiles(k)%tile%tile_pw)
            DEALLOCATE (dbc%tiles(k)%tile)
         END DO
         DEALLOCATE (dbc%tiles)
      ELSE
         DO k = 1, n_tiles
            CALL pw_release(dbc%tiles(k)%tile%tile_pw)
            DEALLOCATE (dbc%tiles(k)%tile)
         END DO
         DEALLOCATE (dbc%tiles)
      END IF

      CALL timestop(handle)

   END SUBROUTINE dbc_tile_release

! **************************************************************************************************
!> \brief   deallocates dirichlet_bc_parameters type
!> \param dbc_params dbc parameters
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dbc_parameters_dealloc(dbc_params)

      TYPE(dirichlet_bc_parameters), INTENT(INOUT)       :: dbc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbc_parameters_dealloc', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dbc_params%aa_planar_nprtn)) DEALLOCATE (dbc_params%aa_planar_nprtn)
      IF (ASSOCIATED(dbc_params%aa_planar_pplane)) DEALLOCATE (dbc_params%aa_planar_pplane)
      IF (ASSOCIATED(dbc_params%aa_planar_xxtnt)) DEALLOCATE (dbc_params%aa_planar_xxtnt)
      IF (ASSOCIATED(dbc_params%aa_planar_yxtnt)) DEALLOCATE (dbc_params%aa_planar_yxtnt)
      IF (ASSOCIATED(dbc_params%aa_planar_zxtnt)) DEALLOCATE (dbc_params%aa_planar_zxtnt)
      IF (ASSOCIATED(dbc_params%aa_planar_vD)) DEALLOCATE (dbc_params%aa_planar_vD)
      IF (ASSOCIATED(dbc_params%aa_planar_sigma)) DEALLOCATE (dbc_params%aa_planar_sigma)
      IF (ASSOCIATED(dbc_params%aa_planar_thickness)) DEALLOCATE (dbc_params%aa_planar_thickness)

      IF (ASSOCIATED(dbc_params%planar_nprtn)) DEALLOCATE (dbc_params%planar_nprtn)
      IF (ASSOCIATED(dbc_params%planar_Avtx)) DEALLOCATE (dbc_params%planar_Avtx)
      IF (ASSOCIATED(dbc_params%planar_Bvtx)) DEALLOCATE (dbc_params%planar_Bvtx)
      IF (ASSOCIATED(dbc_params%planar_Cvtx)) DEALLOCATE (dbc_params%planar_Cvtx)
      IF (ASSOCIATED(dbc_params%planar_Dvtx)) DEALLOCATE (dbc_params%planar_Dvtx)
      IF (ASSOCIATED(dbc_params%planar_vD)) DEALLOCATE (dbc_params%planar_vD)
      IF (ASSOCIATED(dbc_params%planar_sigma)) DEALLOCATE (dbc_params%planar_sigma)
      IF (ASSOCIATED(dbc_params%planar_thickness)) DEALLOCATE (dbc_params%planar_thickness)

      IF (ASSOCIATED(dbc_params%xaa_cylindrical_nprtn)) DEALLOCATE (dbc_params%xaa_cylindrical_nprtn)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_nsides)) DEALLOCATE (dbc_params%xaa_cylindrical_nsides)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_apxtyp)) DEALLOCATE (dbc_params%xaa_cylindrical_apxtyp)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_xxtnt)) DEALLOCATE (dbc_params%xaa_cylindrical_xxtnt)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_bctr)) DEALLOCATE (dbc_params%xaa_cylindrical_bctr)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_brad)) DEALLOCATE (dbc_params%xaa_cylindrical_brad)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_vD)) DEALLOCATE (dbc_params%xaa_cylindrical_vD)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_sigma)) DEALLOCATE (dbc_params%xaa_cylindrical_sigma)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_thickness)) DEALLOCATE (dbc_params%xaa_cylindrical_thickness)
      IF (ASSOCIATED(dbc_params%xaa_cylindrical_sgap)) DEALLOCATE (dbc_params%xaa_cylindrical_sgap)

      IF (ASSOCIATED(dbc_params%aa_cuboidal_nprtn)) DEALLOCATE (dbc_params%aa_cuboidal_nprtn)
      IF (ASSOCIATED(dbc_params%aa_cuboidal_xxtnt)) DEALLOCATE (dbc_params%aa_cuboidal_xxtnt)
      IF (ASSOCIATED(dbc_params%aa_cuboidal_yxtnt)) DEALLOCATE (dbc_params%aa_cuboidal_yxtnt)
      IF (ASSOCIATED(dbc_params%aa_cuboidal_zxtnt)) DEALLOCATE (dbc_params%aa_cuboidal_zxtnt)
      IF (ASSOCIATED(dbc_params%aa_cuboidal_vD)) DEALLOCATE (dbc_params%aa_cuboidal_vD)
      IF (ASSOCIATED(dbc_params%aa_cuboidal_sigma)) DEALLOCATE (dbc_params%aa_cuboidal_sigma)

   END SUBROUTINE dbc_parameters_dealloc

END MODULE dirichlet_bc_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief dielectric constant data type
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE dielectric_types

   USE kinds,                           ONLY: dp
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_release,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dielectric_types'

   TYPE dielectric_parameters
      LOGICAL                                :: dielec_core_correction
      INTEGER                                :: dielec_functiontype
      REAL(KIND=dp)                          :: eps0
      REAL(KIND=dp)                          :: rho_min
      REAL(KIND=dp)                          :: rho_max
      INTEGER                                :: derivative_method
      INTEGER                                :: times_called

      INTEGER                                :: n_aa_cuboidal
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: aa_cuboidal_xxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: aa_cuboidal_yxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: aa_cuboidal_zxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER   :: aa_cuboidal_eps => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER   :: aa_cuboidal_zeta => NULL()

      INTEGER                                :: n_xaa_annular
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: xaa_annular_xxtnt => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: xaa_annular_bctr => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: xaa_annular_brad => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER   :: xaa_annular_eps => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER   :: xaa_annular_zeta => NULL()
   END TYPE dielectric_parameters

   TYPE dielectric_type
      TYPE(dielectric_parameters)       :: params
      TYPE(pw_type), POINTER            :: eps
      TYPE(pw_type), POINTER            :: deps_drho
      TYPE(pw_p_type), DIMENSION(3)     :: dln_eps
   END TYPE dielectric_type

   PUBLIC :: dielectric_type, dielectric_parameters
   PUBLIC :: dielectric_release

   INTEGER, PARAMETER, PUBLIC           :: derivative_cd3 = 1, &
                                           derivative_cd5 = 2, &
                                           derivative_cd7 = 3, &
                                           derivative_fft = 100, &
                                           derivative_fft_use_deps = 101, &
                                           derivative_fft_use_drho = 102

   INTEGER, PARAMETER, PUBLIC           :: rho_dependent = 10, &
                                           spatially_dependent = 20, &
                                           spatially_rho_dependent = 30
CONTAINS

! **************************************************************************************************
!> \brief   deallocates dielectric data type
!> \param dielectric the dielectric data type to be released
!> \param pw_pool pool of the plane wave grid
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_release(dielectric, pw_pool)
      TYPE(dielectric_type), POINTER                     :: dielectric
      TYPE(pw_pool_type), INTENT(IN), OPTIONAL, POINTER  :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'dielectric_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      LOGICAL                                            :: can_give_back

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(dielectric)) THEN
         can_give_back = PRESENT(pw_pool)
         IF (can_give_back) can_give_back = ASSOCIATED(pw_pool)
         IF (can_give_back) THEN
            CALL pw_pool_give_back_pw(pw_pool, dielectric%eps, &
                                      accept_non_compatible=.TRUE.)
            CALL pw_pool_give_back_pw(pw_pool, dielectric%deps_drho, &
                                      accept_non_compatible=.TRUE.)
            DO i = 1, 3
               CALL pw_pool_give_back_pw(pw_pool, dielectric%dln_eps(i)%pw, &
                                         accept_non_compatible=.TRUE.)
            END DO
         ELSE
            CALL pw_release(dielectric%eps)
            CALL pw_release(dielectric%deps_drho)
            DO i = 1, 3
               CALL pw_release(dielectric%dln_eps(i)%pw)
            END DO
         END IF
         CALL dielectric_parameters_dealloc(dielectric%params)
         DEALLOCATE (dielectric)
      END IF

      CALL timestop(handle)

   END SUBROUTINE dielectric_release

! **************************************************************************************************
!> \brief   deallocates dielectric_parameters type
!> \param dielec_params dielectric parameters
!> \par History
!>       07.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_parameters_dealloc(dielec_params)

      TYPE(dielectric_parameters), INTENT(INOUT)         :: dielec_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_parameters_dealloc', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dielec_params%aa_cuboidal_xxtnt)) DEALLOCATE (dielec_params%aa_cuboidal_xxtnt)
      IF (ASSOCIATED(dielec_params%aa_cuboidal_yxtnt)) DEALLOCATE (dielec_params%aa_cuboidal_yxtnt)
      IF (ASSOCIATED(dielec_params%aa_cuboidal_zxtnt)) DEALLOCATE (dielec_params%aa_cuboidal_zxtnt)
      IF (ASSOCIATED(dielec_params%aa_cuboidal_eps)) DEALLOCATE (dielec_params%aa_cuboidal_eps)
      IF (ASSOCIATED(dielec_params%aa_cuboidal_zeta)) DEALLOCATE (dielec_params%aa_cuboidal_zeta)

      IF (ASSOCIATED(dielec_params%xaa_annular_xxtnt)) DEALLOCATE (dielec_params%xaa_annular_xxtnt)
      IF (ASSOCIATED(dielec_params%xaa_annular_bctr)) DEALLOCATE (dielec_params%xaa_annular_bctr)
      IF (ASSOCIATED(dielec_params%xaa_annular_brad)) DEALLOCATE (dielec_params%xaa_annular_brad)
      IF (ASSOCIATED(dielec_params%xaa_annular_eps)) DEALLOCATE (dielec_params%xaa_annular_eps)
      IF (ASSOCIATED(dielec_params%xaa_annular_zeta)) DEALLOCATE (dielec_params%xaa_annular_zeta)

   END SUBROUTINE dielectric_parameters_dealloc

END MODULE dielectric_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief the type I Discrete Cosine Transfrom (DCT-I)
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!>       11.2015 dealt with periodic grids [Hossein Bani-Hashemian]
!>       03.2016 dct in one or two directions [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE dct

   USE fast,                            ONLY: copy_cr
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_cart_rank,&
                                              mp_irecv,&
                                              mp_isend,&
                                              mp_request_null,&
                                              mp_wait,&
                                              mp_waitall
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_create,&
                                              pw_grid_setup
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA1D,&
                                              RECIPROCALSPACE,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dct'

   TYPE :: dct_type
      INTEGER, DIMENSION(:), POINTER     :: dests_expand => NULL()
      INTEGER, DIMENSION(:), POINTER     :: srcs_expand => NULL()
      INTEGER, DIMENSION(:), POINTER     :: flipg_stat => NULL()
      INTEGER, DIMENSION(:), POINTER     :: dests_shrink => NULL()
      INTEGER                            :: srcs_shrink
      INTEGER, DIMENSION(:, :, :), POINTER :: recv_msgs_bnds => NULL()
      INTEGER, DIMENSION(2, 3)            :: dct_bounds
      INTEGER, DIMENSION(2, 3)            :: dct_bounds_local
      INTEGER, DIMENSION(2, 3)            :: bounds_shftd
      INTEGER, DIMENSION(2, 3)            :: bounds_local_shftd
   END TYPE dct_type

   TYPE dct_msg_type
      PRIVATE
      REAL(dp), DIMENSION(:, :, :), POINTER :: msg => NULL()
   END TYPE dct_msg_type

   PUBLIC dct_type, &
      dct_type_init, &
      dct_type_release, &
      setup_dct_pw_grids, &
      pw_shrink, &
      pw_expand

   INTEGER, PARAMETER, PRIVATE    :: NOT_FLIPPED = 0, &
                                     UD_FLIPPED = 1, &
                                     LR_FLIPPED = 2, &
                                     BF_FLIPPED = 3, &
                                     ROTATED = 4

   INTEGER, PARAMETER, PUBLIC     :: neumannXYZ = 111, &
                                     neumannXY = 110, &
                                     neumannXZ = 101, &
                                     neumannYZ = 011, &
                                     neumannX = 100, &
                                     neumannY = 010, &
                                     neumannZ = 001

CONTAINS

! **************************************************************************************************
!> \brief  Initializes a dct_type
!> \param pw_grid the original plane wave grid
!> \param neumann_directions directions in which dct should be performed
!> \param dct_env dct_type to be initialized
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dct_type_init(pw_grid, neumann_directions, dct_env)

      TYPE(pw_grid_type), INTENT(IN), POINTER            :: pw_grid
      INTEGER, INTENT(IN)                                :: neumann_directions
      TYPE(dct_type), INTENT(INOUT)                      :: dct_env

      CHARACTER(len=*), PARAMETER :: routineN = 'dct_type_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, maxn_sendrecv

      CALL timeset(routineN, handle)

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ, neumannXY)
         maxn_sendrecv = 4
      CASE (neumannX, neumannY, neumannXZ, neumannYZ)
         maxn_sendrecv = 2
      CASE (neumannZ)
         maxn_sendrecv = 1
      CASE DEFAULT
         CALL cp__b("pw/dct.F",108,"Invalid combination of Neumann and periodic conditions.")
      END SELECT

      ALLOCATE (dct_env%flipg_stat(maxn_sendrecv))
      ALLOCATE (dct_env%dests_expand(maxn_sendrecv), dct_env%srcs_expand(maxn_sendrecv))
      ALLOCATE (dct_env%dests_shrink(maxn_sendrecv))
      ALLOCATE (dct_env%recv_msgs_bnds(2, 3, maxn_sendrecv))

      CALL set_dests_srcs_pid(pw_grid, neumann_directions, &
                              dct_env%dests_expand, dct_env%srcs_expand, dct_env%flipg_stat, &
                              dct_env%dests_shrink, dct_env%srcs_shrink)
      CALL expansion_bounds(pw_grid, neumann_directions, &
                            dct_env%srcs_expand, dct_env%flipg_stat, &
                            dct_env%bounds_shftd, dct_env%bounds_local_shftd, &
                            dct_env%recv_msgs_bnds, dct_env%dct_bounds, &
                            dct_env%dct_bounds_local)

      CALL timestop(handle)

   END SUBROUTINE dct_type_init

! **************************************************************************************************
!> \brief  Releases a dct_type
!> \param dct_env dct_type to be released
!> \par History
!>       03.2016 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dct_type_release(dct_env)

      TYPE(dct_type), INTENT(INOUT)                      :: dct_env

      CHARACTER(len=*), PARAMETER :: routineN = 'dct_type_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(dct_env%dests_shrink)) DEALLOCATE (dct_env%dests_shrink)
      IF (ASSOCIATED(dct_env%dests_expand)) DEALLOCATE (dct_env%dests_expand)
      IF (ASSOCIATED(dct_env%srcs_expand)) DEALLOCATE (dct_env%srcs_expand)
      IF (ASSOCIATED(dct_env%flipg_stat)) DEALLOCATE (dct_env%flipg_stat)
      IF (ASSOCIATED(dct_env%recv_msgs_bnds)) DEALLOCATE (dct_env%recv_msgs_bnds)

      CALL timestop(handle)

   END SUBROUTINE dct_type_release

! **************************************************************************************************
!> \brief   sets up an extended pw_grid for Discrete Cosine Transform (DCT)
!>          calculations
!> \param pw_grid the original plane wave grid
!> \param cell_hmat cell hmat
!> \param neumann_directions directions in which dct should be performed
!> \param dct_pw_grid DCT plane-wave grid
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE setup_dct_pw_grids(pw_grid, cell_hmat, neumann_directions, dct_pw_grid)

      TYPE(pw_grid_type), INTENT(IN), POINTER            :: pw_grid
      REAL(dp), DIMENSION(3, 3), INTENT(IN)              :: cell_hmat
      INTEGER, INTENT(IN)                                :: neumann_directions
      TYPE(pw_grid_type), INTENT(INOUT), POINTER         :: dct_pw_grid

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_dct_pw_grids', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blocked, handle, maxn_sendrecv, &
                                                            srcs_shrink
      INTEGER, DIMENSION(2, 3)                           :: bounds_local_new, bounds_local_shftd, &
                                                            bounds_new, bounds_shftd
      INTEGER, DIMENSION(:), POINTER                     :: dests_expand, dests_shrink, flipg_stat, &
                                                            srcs_expand
      INTEGER, DIMENSION(:, :, :), POINTER               :: recv_msgs_bnds
      REAL(KIND=dp), DIMENSION(3)                        :: scfac
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat2

      CALL timeset(routineN, handle)

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ)
         maxn_sendrecv = 4
         scfac = (/2.0_dp, 2.0_dp, 2.0_dp/)
      CASE (neumannXY)
         maxn_sendrecv = 4
         scfac = (/2.0_dp, 2.0_dp, 1.0_dp/)
      CASE (neumannXZ)
         maxn_sendrecv = 2
         scfac = (/2.0_dp, 1.0_dp, 2.0_dp/)
      CASE (neumannYZ)
         maxn_sendrecv = 2
         scfac = (/1.0_dp, 2.0_dp, 2.0_dp/)
      CASE (neumannX)
         maxn_sendrecv = 2
         scfac = (/2.0_dp, 1.0_dp, 1.0_dp/)
      CASE (neumannY)
         maxn_sendrecv = 2
         scfac = (/1.0_dp, 2.0_dp, 1.0_dp/)
      CASE (neumannZ)
         maxn_sendrecv = 1
         scfac = (/1.0_dp, 1.0_dp, 2.0_dp/)
      CASE DEFAULT
         CALL cp__b("pw/dct.F",213,"Invalid combination of Neumann and periodic conditions.")
      END SELECT

      ALLOCATE (flipg_stat(maxn_sendrecv))
      ALLOCATE (dests_expand(maxn_sendrecv), srcs_expand(maxn_sendrecv), dests_shrink(maxn_sendrecv))
      ALLOCATE (recv_msgs_bnds(2, 3, maxn_sendrecv))

      CALL set_dests_srcs_pid(pw_grid, neumann_directions, dests_expand, srcs_expand, flipg_stat, &
                              dests_shrink, srcs_shrink)
      CALL expansion_bounds(pw_grid, neumann_directions, srcs_expand, flipg_stat, &
                            bounds_shftd, bounds_local_shftd, recv_msgs_bnds, bounds_new, bounds_local_new)
      CALL pw_grid_create(dct_pw_grid, pw_grid%para%rs_group, local=.FALSE.)

      hmat2 = 0.0_dp
      hmat2(1, 1) = scfac(1)*cell_hmat(1, 1)
      hmat2(2, 2) = scfac(2)*cell_hmat(2, 2)
      hmat2(3, 3) = scfac(3)*cell_hmat(3, 3)

      ! uses bounds_local_new that is 2*n-2 in size....this is only rarely fft-able by fftsg, and needs fftw3,
      ! where it might use sizes that are not declared available in fft_get_radix.

      IF (pw_grid%para%blocked) THEN
         blocked = 1
      ELSE IF (pw_grid%para%ray_distribution) THEN
         blocked = 0
      END IF

      CALL pw_grid_setup(hmat2, dct_pw_grid, &
                         bounds=bounds_new, &
                         rs_dims=pw_grid%para%rs_dims, &
                         blocked=blocked, &
                         bounds_local=bounds_local_new)

      DEALLOCATE (flipg_stat, dests_expand, srcs_expand, dests_shrink, recv_msgs_bnds)

      CALL timestop(handle)

   END SUBROUTINE setup_dct_pw_grids

! **************************************************************************************************
!> \brief Finds the process ids for mpi_isend destiations and mpi_irecv sources
!>   for expanding and shrinking a pw_type data
!> \param pw_grid the original plane wave grid
!> \param neumann_directions directions in which dct should be performed
!> \param dests_expand list of the destination processes (pw_expand)
!> \param srcs_expand list of the source processes (pw_expand)
!> \param flipg_stat flipping status for the received data chunks (pw_expand)
!> \param dests_shrink list of the destination processes (pw_shrink)
!> \param srcs_shrink list of the source proceses (pw_shrink)
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE set_dests_srcs_pid(pw_grid, neumann_directions, dests_expand, srcs_expand, &
                                 flipg_stat, dests_shrink, srcs_shrink)

      TYPE(pw_grid_type), INTENT(IN), POINTER            :: pw_grid
      INTEGER, INTENT(IN)                                :: neumann_directions
      INTEGER, DIMENSION(:), INTENT(INOUT), POINTER      :: dests_expand, srcs_expand, flipg_stat, &
                                                            dests_shrink
      INTEGER, INTENT(OUT)                               :: srcs_shrink

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_dests_srcs_pid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group_size, handle, i, j, k, &
                                                            maxn_sendrecv, rs_dim1, rs_dim2, &
                                                            rs_group, rs_mpo, tmp_size1, tmp_size2
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: src_pos1_onesdd, src_pos2_onesdd, &
                                                            tmp1_arr, tmp2_arr
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: dests_shrink_all, src_pos1, src_pos2, &
                                                            srcs_coord, srcs_expand_all
      INTEGER, DIMENSION(2)                              :: rs_dims, rs_pos

      CALL timeset(routineN, handle)

! example: 3x4 process grid
! XYZ or XY
! rs_dim1 = 3 -->  src_pos1 = [1  3  -2]
!                             [2 -3  -1]
! rs_dim2 = 4 -->  src_pos2 = [1  3  -4  -2]
!                             [2  4  -3  -1]
! => (1,1) receives from (1,1) ; (1,2) ; (2,1) ; (2,2) | flipping status 0 0 0 0
!    (2,4) receives from (3,2) ; (3,1) ; (3,2) ; (3,1) | flipping status 2 2 4 4
! and so on ...
! or equivalently
! =>   0   receives from 0 ; 1 ; 4 ; 5 | flipping status 0 0 0 0
!      7   receives from 9 ; 8 ; 9 ; 8 | flipping status 2 2 4 4
! from srcs_coord :
! => rs_mpo = 0 -> rs_pos = 0,0 -> srcs_coord = [ 1  1  2  2] -> 0(0) 1(0) 4(0) 5(0)
!                                               [ 1  2  1  2]
!    rs_mpo = 7 -> rs_pos = 1,3 -> srcs_coord = [ 3  3 -3 -3] -> 9(2) 8(2) 9(4) 8(4)
!                                               [-2 -1 -2 -1]
! schematically :
! ij : coordinates in a 2D process grid (starting from 1 just for demonstration)
! () : to be flipped from left to right
! [] : to be flipped from up to down
! {} : to be rotated 180 degrees
!    11 | 12 | 13 | 14          11   12  |  13   14  | (14) (13) | (12) (11)
!    -----------------    ==>   21   22  |  23   24  | (24) (23) | (22) (21)
!    21 | 22 | 23 | 24         ---------------------------------------------
!    -----------------          31   32  |  33   34  | (34) (33) | (32) (31)
!    31 | 32 | 33 | 34         [31] [32] | [33] [34] | {34} {33} | {32} {31}
!                              ---------------------------------------------
!                              [21] [22] | [23] [24] | {24} {23} | {22} {21}
!                              [11] [12] | [13] [14] | {14} {13} | {12} {11}
! one(two)-sided :
! YZ or Y
! rs_dim1 = 3 -->  src_pos1 = [1  2  3]
! rs_dim2 = 4 -->  src_pos2 = [1  3  -4  -2]
!                             [2  4  -3  -1]
! XZ or X
! rs_dim1 = 3 -->  src_pos1 = [1  3  -2]
!                             [2 -3  -1]
! rs_dim2 = 4 -->  src_pos2 = [1  2   3  4]
! Z
! rs_dim1 = 3 -->  src_pos1 = [1  2  3]
! rs_dim2 = 4 -->  src_pos2 = [1  2   3  4]

      rs_group = pw_grid%para%rs_group
      rs_mpo = pw_grid%para%rs_mpo
      group_size = pw_grid%para%group_size
      rs_dims = pw_grid%para%rs_dims

      rs_pos = pw_grid%para%rs_pos
      rs_dim1 = rs_dims(1); rs_dim2 = rs_dims(2)

! prepare srcs_coord
      SELECT CASE (neumann_directions)
      CASE (neumannXYZ, neumannXY)
         maxn_sendrecv = 4
         ALLOCATE (srcs_coord(2, maxn_sendrecv))

         IF (MOD(rs_dim1, 2) .EQ. 0) THEN
            tmp_size1 = rs_dim1
         ELSE
            tmp_size1 = rs_dim1+1
         END IF
         ALLOCATE (tmp1_arr(tmp_size1), src_pos1(2, 0:rs_dim1-1))

         IF (MOD(rs_dim1, 2) .EQ. 0) THEN
            tmp1_arr(:) = (/(i, i=1, rs_dim1)/)
            src_pos1(:, :) = RESHAPE((/tmp1_arr, -tmp1_arr(tmp_size1:1:-1)/), (/2, rs_dim1/))
         ELSE
            tmp1_arr(:) = (/(i, i=1, rs_dim1), -rs_dim1/)
            src_pos1(:, :) = RESHAPE((/tmp1_arr, -tmp1_arr(tmp_size1-2:1:-1)/), (/2, rs_dim1/))
         END IF
!---
         IF (MOD(rs_dim2, 2) .EQ. 0) THEN
            tmp_size2 = rs_dim2
         ELSE
            tmp_size2 = rs_dim2+1
         END IF
         ALLOCATE (tmp2_arr(tmp_size2), src_pos2(2, 0:rs_dim2-1))

         IF (MOD(rs_dim2, 2) .EQ. 0) THEN
            tmp2_arr(:) = (/(i, i=1, rs_dim2)/)
            src_pos2(:, :) = RESHAPE((/tmp2_arr, -tmp2_arr(tmp_size2:1:-1)/), (/2, rs_dim2/))
         ELSE
            tmp2_arr(:) = (/(i, i=1, rs_dim2), -rs_dim2/)
            src_pos2(:, :) = RESHAPE((/tmp2_arr, -tmp2_arr(tmp_size2-2:1:-1)/), (/2, rs_dim2/))
         END IF
!---
         srcs_coord(:, 1) = (/src_pos1(1, rs_pos(1)), src_pos2(1, rs_pos(2))/)
         srcs_coord(:, 2) = (/src_pos1(1, rs_pos(1)), src_pos2(2, rs_pos(2))/)
         srcs_coord(:, 3) = (/src_pos1(2, rs_pos(1)), src_pos2(1, rs_pos(2))/)
         srcs_coord(:, 4) = (/src_pos1(2, rs_pos(1)), src_pos2(2, rs_pos(2))/)
      CASE (neumannXZ, neumannX)
         maxn_sendrecv = 2
         ALLOCATE (srcs_coord(2, maxn_sendrecv))

         IF (MOD(rs_dim1, 2) .EQ. 0) THEN
            tmp_size1 = rs_dim1
         ELSE
            tmp_size1 = rs_dim1+1
         END IF
         ALLOCATE (tmp1_arr(tmp_size1), src_pos1(2, 0:rs_dim1-1))

         IF (MOD(rs_dim1, 2) .EQ. 0) THEN
            tmp1_arr(:) = (/(i, i=1, rs_dim1)/)
            src_pos1(:, :) = RESHAPE((/tmp1_arr, -tmp1_arr(tmp_size1:1:-1)/), (/2, rs_dim1/))
         ELSE
            tmp1_arr(:) = (/(i, i=1, rs_dim1), -rs_dim1/)
            src_pos1(:, :) = RESHAPE((/tmp1_arr, -tmp1_arr(tmp_size1-2:1:-1)/), (/2, rs_dim1/))
         END IF
!---
         ALLOCATE (src_pos2_onesdd(0:rs_dim2-1))
         src_pos2_onesdd(:) = (/(i, i=1, rs_dim2)/)
!---
         srcs_coord(:, 1) = (/src_pos1(1, rs_pos(1)), src_pos2_onesdd(rs_pos(2))/)
         srcs_coord(:, 2) = (/src_pos1(2, rs_pos(1)), src_pos2_onesdd(rs_pos(2))/)
      CASE (neumannYZ, neumannY)
         maxn_sendrecv = 2
         ALLOCATE (srcs_coord(2, maxn_sendrecv))

         ALLOCATE (src_pos1_onesdd(0:rs_dim1-1))
         src_pos1_onesdd(:) = (/(i, i=1, rs_dim1)/)
!---
         IF (MOD(rs_dim2, 2) .EQ. 0) THEN
            tmp_size2 = rs_dim2
         ELSE
            tmp_size2 = rs_dim2+1
         END IF
         ALLOCATE (tmp2_arr(tmp_size2), src_pos2(2, 0:rs_dim2-1))

         IF (MOD(rs_dim2, 2) .EQ. 0) THEN
            tmp2_arr(:) = (/(i, i=1, rs_dim2)/)
            src_pos2(:, :) = RESHAPE((/tmp2_arr, -tmp2_arr(tmp_size2:1:-1)/), (/2, rs_dim2/))
         ELSE
            tmp2_arr(:) = (/(i, i=1, rs_dim2), -rs_dim2/)
            src_pos2(:, :) = RESHAPE((/tmp2_arr, -tmp2_arr(tmp_size2-2:1:-1)/), (/2, rs_dim2/))
         END IF
!---
         srcs_coord(:, 1) = (/src_pos1_onesdd(rs_pos(1)), src_pos2(1, rs_pos(2))/)
         srcs_coord(:, 2) = (/src_pos1_onesdd(rs_pos(1)), src_pos2(2, rs_pos(2))/)
      CASE (neumannZ)
         maxn_sendrecv = 1
         ALLOCATE (srcs_coord(2, maxn_sendrecv))
         ALLOCATE (src_pos1_onesdd(0:rs_dim1-1))
         ALLOCATE (src_pos2_onesdd(0:rs_dim2-1))

         src_pos1_onesdd(:) = (/(i, i=1, rs_dim1)/)
!---
         src_pos2_onesdd(:) = (/(i, i=1, rs_dim2)/)
!---
         srcs_coord(:, 1) = (/src_pos1_onesdd(rs_pos(1)), src_pos2_onesdd(rs_pos(2))/)
      END SELECT

! default flipping status
      flipg_stat = NOT_FLIPPED

      DO k = 1, maxn_sendrecv
! convert srcs_coord to pid
         CALL mp_cart_rank(pw_grid%para%rs_group, ABS(srcs_coord(:, k))-1, srcs_expand(k))
! find out the flipping status
         IF ((srcs_coord(1, k) .GT. 0) .AND. (srcs_coord(2, k) .GT. 0)) THEN
            flipg_stat(k) = NOT_FLIPPED
         ELSE IF ((srcs_coord(1, k) .LT. 0) .AND. (srcs_coord(2, k) .GT. 0)) THEN
            flipg_stat(k) = UD_FLIPPED
         ELSE IF ((srcs_coord(1, k) .GT. 0) .AND. (srcs_coord(2, k) .LT. 0)) THEN
            flipg_stat(k) = LR_FLIPPED
         ELSE
            flipg_stat(k) = ROTATED
         END IF
      END DO

! let all the nodes know about each others srcs_expand list
      ALLOCATE (srcs_expand_all(maxn_sendrecv, group_size))
      CALL mp_allgather(srcs_expand, srcs_expand_all, rs_group)
! now scan the srcs_expand_all list and check if I am on the srcs_expand list of the other nodes
! if that is the case then I am obliged to send data to those nodes (the nodes are on my dests_expand list)
      k = 1
      DO i = 1, group_size
         DO j = 1, maxn_sendrecv
            IF (srcs_expand_all(j, i) .EQ. rs_mpo) THEN
               dests_expand(k) = i-1
               k = k+1
            END IF
         END DO
      END DO

! find srcs and dests for the reverse procedure :
! initialize dests_shrink and srcs_shrink with invalid process id
      dests_shrink = -1
      srcs_shrink = -1
! scan the flipping status of the data that I am supposed to receive
! if the flipping status for a process is NOT_FLIPPED that means I will have to resend
! data to that process in the reverse procedure (the process is on my dests_shrink list)
      DO i = 1, maxn_sendrecv
         IF (flipg_stat(i) .EQ. NOT_FLIPPED) dests_shrink(i) = srcs_expand(i)
      END DO

! let all the nodes know about each others dests_shrink list
      ALLOCATE (dests_shrink_all(maxn_sendrecv, group_size))
      CALL mp_allgather(dests_shrink, dests_shrink_all, rs_group)
! now scan the dests_shrink_all list and check if I am on the dests_shrink list of any other node
! if that is the case then I'll receive data from that node (the node is on my srcs_shrink list)
! note that in the shrinking procedure I will receive from only one node
      DO i = 1, group_size
         DO j = 1, maxn_sendrecv
            IF (dests_shrink_all(j, i) .EQ. rs_mpo) THEN
               srcs_shrink = i-1
               EXIT
            END IF
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE set_dests_srcs_pid

! **************************************************************************************************
!> \brief expands a pw_type data to an evenly symmetric pw_type data that is 8 times
!>   larger than the original one:
!>   the even symmetry for a 1D sequence of length n is defined as:
!>      1 2 3 ... n-2 n-1 n --> 1 2 3 ... n-2 n-1 n n-1 n-2 ... 3 2
!>   and is generalized to 3D by applying the same rule in all three directions
!>
!> \param neumann_directions directions in which dct should be performed
!> \param recv_msgs_bnds bounds of the messages to be received
!> \param dests_expand list of the destination processes
!> \param srcs_expand list of the source processes
!> \param flipg_stat flipping status for the received data chunks
!> \param bounds_shftd bounds of the original grid shifted to have g0 in the middle of the cell
!> \param pw_in the original plane wave data
!> \param pw_expanded the pw data after expansion
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE pw_expand(neumann_directions, recv_msgs_bnds, dests_expand, srcs_expand, &
                        flipg_stat, bounds_shftd, pw_in, pw_expanded)

      INTEGER, INTENT(IN)                                :: neumann_directions
      INTEGER, DIMENSION(:, :, :), INTENT(IN), POINTER   :: recv_msgs_bnds
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: dests_expand, srcs_expand, flipg_stat
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bounds_shftd
      TYPE(pw_type), INTENT(IN), POINTER                 :: pw_in
      TYPE(pw_type), INTENT(INOUT), POINTER              :: pw_expanded

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pw_expand', routineP = moduleN//':'//routineN

      INTEGER :: group_size, handle, i, ind, lb1, lb1_loc, lb1_new, lb2, lb2_loc, lb2_new, lb3, &
         lb3_loc, lb3_new, loc, maxn_sendrecv, rs_group, rs_mpo, ub1, ub1_loc, ub1_new, ub2, &
         ub2_loc, ub2_new, ub3, ub3_loc, ub3_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: dest_hist, recv_reqs, send_reqs, src_hist
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: pcs_bnds
      INTEGER, DIMENSION(2, 3)                           :: bounds_local_new
      REAL(dp), DIMENSION(:, :, :), POINTER              :: catd, catd_flipdbf, cr3d_xpndd, send_msg
      TYPE(dct_msg_type), DIMENSION(:), POINTER          :: pcs, recv_msgs
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CALL timeset(routineN, handle)

      pw_grid => pw_in%pw_grid
      rs_group = pw_grid%para%rs_group
      rs_mpo = pw_grid%para%my_pos
      group_size = pw_grid%para%group_size

      bounds_local_new = pw_expanded%pw_grid%bounds_local

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ, neumannXY)
         maxn_sendrecv = 4
      CASE (neumannX, neumannY, neumannXZ, neumannYZ)
         maxn_sendrecv = 2
      CASE (neumannZ)
         maxn_sendrecv = 1
      END SELECT

      ALLOCATE (recv_reqs(maxn_sendrecv), send_reqs(maxn_sendrecv))
      ALLOCATE (dest_hist(maxn_sendrecv), src_hist(maxn_sendrecv))
      ALLOCATE (pcs_bnds(2, 3, maxn_sendrecv))

      NULLIFY (pcs, recv_msgs)
      ALLOCATE (pcs(maxn_sendrecv), recv_msgs(maxn_sendrecv))

      send_reqs = mp_request_null
      recv_reqs = mp_request_null

      send_msg => pw_in%cr3d

      src_hist = -1 ! keeps the history of sources
      dest_hist = -1 ! keeps the history of destinations

      DO i = 1, maxn_sendrecv
! no need to send to myself or to the destination that I have already sent to
         IF ((dests_expand(i) .NE. rs_mpo) .AND. .NOT. ANY(dest_hist .EQ. dests_expand(i))) THEN
            CALL mp_isend(send_msg, dests_expand(i), rs_group, send_reqs(i))
         END IF
         dest_hist(i) = dests_expand(i)
      END DO

      DO i = 1, maxn_sendrecv
         lb1 = recv_msgs_bnds(1, 1, i)
         ub1 = recv_msgs_bnds(2, 1, i)
         lb2 = recv_msgs_bnds(1, 2, i)
         ub2 = recv_msgs_bnds(2, 2, i)
         lb3 = recv_msgs_bnds(1, 3, i)
         ub3 = recv_msgs_bnds(2, 3, i)
! no need to receive from myself
         IF (srcs_expand(i) .EQ. rs_mpo) THEN
            ALLOCATE (recv_msgs(i)%msg(lb1:ub1, lb2:ub2, lb3:ub3))
            recv_msgs(i)%msg = send_msg
! if I have already received data from the source, just use the one from the last time
         ELSE IF (ANY(src_hist .EQ. srcs_expand(i))) THEN
            loc = MINLOC(ABS(src_hist-srcs_expand(i)), 1)
            lb1_loc = recv_msgs_bnds(1, 1, loc)
            ub1_loc = recv_msgs_bnds(2, 1, loc)
            lb2_loc = recv_msgs_bnds(1, 2, loc)
            ub2_loc = recv_msgs_bnds(2, 2, loc)
            lb3_loc = recv_msgs_bnds(1, 3, loc)
            ub3_loc = recv_msgs_bnds(2, 3, loc)
            ALLOCATE (recv_msgs(i)%msg(lb1_loc:ub1_loc, lb2_loc:ub2_loc, lb3_loc:ub3_loc))
            recv_msgs(i)%msg = recv_msgs(loc)%msg
         ELSE
            ALLOCATE (recv_msgs(i)%msg(lb1:ub1, lb2:ub2, lb3:ub3))
            CALL mp_irecv(recv_msgs(i)%msg, srcs_expand(i), rs_group, recv_reqs(i))
            CALL mp_wait(recv_reqs(i))
         END IF
         src_hist(i) = srcs_expand(i)
      END DO
! cleanup mpi_request to prevent memory leak
      CALL mp_waitall(send_reqs(:))

! flip the received data according on the flipping status
      DO i = 1, maxn_sendrecv
         SELECT CASE (flipg_stat (i))
         CASE (NOT_FLIPPED)
            lb1 = recv_msgs_bnds(1, 1, i)
            ub1 = recv_msgs_bnds(2, 1, i)
            lb2 = recv_msgs_bnds(1, 2, i)
            ub2 = recv_msgs_bnds(2, 2, i)
            lb3 = recv_msgs_bnds(1, 3, i)
            ub3 = recv_msgs_bnds(2, 3, i)
            ALLOCATE (pcs(i)%msg(lb1:ub1, lb2:ub2, lb3:ub3))
            pcs(i)%msg = recv_msgs(i)%msg
         CASE (UD_FLIPPED)
            CALL flipud(recv_msgs(i)%msg, pcs(i)%msg, bounds_shftd)
         CASE (LR_FLIPPED)
            CALL fliplr(recv_msgs(i)%msg, pcs(i)%msg, bounds_shftd)
         CASE (BF_FLIPPED)
            CALL flipbf(recv_msgs(i)%msg, pcs(i)%msg, bounds_shftd)
         CASE (ROTATED)
            CALL rot180(recv_msgs(i)%msg, pcs(i)%msg, bounds_shftd)
         END SELECT
      END DO
! concatenate the received (flipped) data store the result as catd
! need the bounds of the four pieces for concatenation
      DO i = 1, maxn_sendrecv
         pcs_bnds(1, 1, i) = LBOUND(pcs(i)%msg, 1)
         pcs_bnds(2, 1, i) = UBOUND(pcs(i)%msg, 1)
         pcs_bnds(1, 2, i) = LBOUND(pcs(i)%msg, 2)
         pcs_bnds(2, 2, i) = UBOUND(pcs(i)%msg, 2)
         pcs_bnds(1, 3, i) = LBOUND(pcs(i)%msg, 3)
         pcs_bnds(2, 3, i) = UBOUND(pcs(i)%msg, 3)
      END DO

      lb1_new = bounds_local_new(1, 1); ub1_new = bounds_local_new(2, 1)
      lb2_new = bounds_local_new(1, 2); ub2_new = bounds_local_new(2, 2)
      lb3_new = bounds_local_new(1, 3); ub3_new = bounds_local_new(2, 3)

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ, neumannXZ, neumannYZ, neumannZ)
         ind = INT(0.5*(ub3_new+lb3_new+1))
         ALLOCATE (catd(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ind-1))
      CASE (neumannXY, neumannX, neumannY)
         ALLOCATE (catd(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new))
      END SELECT

      DO i = 1, maxn_sendrecv
         catd(pcs_bnds(1, 1, i):pcs_bnds(2, 1, i), &
              pcs_bnds(1, 2, i):pcs_bnds(2, 2, i), &
              pcs_bnds(1, 3, i):pcs_bnds(2, 3, i)) = pcs(i)%msg
      END DO

! flip catd from back to front
      CALL flipbf(catd, catd_flipdbf, bounds_shftd)
! concatenate catd and catd_flipdbf to get cr3d_xpndd
      ALLOCATE (cr3d_xpndd(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new))
      SELECT CASE (neumann_directions)
      CASE (neumannXYZ, neumannXZ, neumannYZ, neumannZ)
         cr3d_xpndd(:, :, lb3_new:ind-1) = catd
         cr3d_xpndd(:, :, ind:ub3_new) = catd_flipdbf
      CASE (neumannXY, neumannX, neumannY)
         cr3d_xpndd(:, :, :) = catd
      END SELECT

      pw_expanded%cr3d = cr3d_xpndd

      DO i = 1, maxn_sendrecv
         DEALLOCATE (pcs(i)%msg)
         DEALLOCATE (recv_msgs(i)%msg)
      END DO
      DEALLOCATE (pcs, recv_msgs)
      DEALLOCATE (catd, catd_flipdbf, cr3d_xpndd)

      CALL timestop(handle)

   END SUBROUTINE pw_expand

! **************************************************************************************************
!> \brief shrinks an evenly symmetric pw_type data to a pw_type data that is 8
!>        times smaller (the reverse procedure of pw_expand).
!>
!> \param neumann_directions directions in which dct should be performed
!> \param dests_shrink list of the destination processes
!> \param srcs_shrink list of the source proceses
!> \param bounds_local_shftd local bounds of the original grid after shifting
!> \param pw_in the original plane wave data
!> \param pw_shrinked the shrinked plane wave data
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE pw_shrink(neumann_directions, dests_shrink, srcs_shrink, bounds_local_shftd, &
                        pw_in, pw_shrinked)

      INTEGER, INTENT(IN)                                :: neumann_directions
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: dests_shrink
      INTEGER, INTENT(IN)                                :: srcs_shrink
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bounds_local_shftd
      TYPE(pw_type), INTENT(IN), POINTER                 :: pw_in
      TYPE(pw_type), INTENT(INOUT), POINTER              :: pw_shrinked

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pw_shrink', routineP = moduleN//':'//routineN

      COMPLEX(dp), DIMENSION(:, :, :), POINTER           :: cc3d
      INTEGER :: group_size, handle, i, in_space, in_use, lb1_orig, lb1_xpnd, lb2_orig, lb2_xpnd, &
         lb3_orig, lb3_xpnd, maxn_sendrecv, recv_req, rs_group, rs_mpo, send_lb1, send_lb2, &
         send_lb3, send_req, send_ub1, send_ub2, send_ub3, ub1_orig, ub1_xpnd, ub2_orig, ub2_xpnd, &
         ub3_orig, ub3_xpnd
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: bounds_local_all
      INTEGER, DIMENSION(2, 3)                           :: bounds_local_xpnd
      REAL(dp), DIMENSION(:, :, :), POINTER              :: cr3d, send_crmsg
      TYPE(pw_grid_type), POINTER                        :: pw_grid_orig

      CALL timeset(routineN, handle)

      pw_grid_orig => pw_shrinked%pw_grid
      rs_group = pw_grid_orig%para%rs_group
      rs_mpo = pw_grid_orig%para%my_pos
      group_size = pw_grid_orig%para%group_size
      bounds_local_xpnd = pw_in%pw_grid%bounds_local
      in_space = pw_in%in_space
      in_use = pw_in%in_use

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ, neumannXY)
         maxn_sendrecv = 4
      CASE (neumannX, neumannY, neumannXZ, neumannYZ)
         maxn_sendrecv = 2
      CASE (neumannZ)
         maxn_sendrecv = 1
      END SELECT

! cosine transform is a real transform. The cosine transfrom of a 3D data must be real and 3D.
      NULLIFY (cr3d, cc3d)
      lb1_xpnd = bounds_local_xpnd(1, 1)
      ub1_xpnd = bounds_local_xpnd(2, 1)
      lb2_xpnd = bounds_local_xpnd(1, 2)
      ub2_xpnd = bounds_local_xpnd(2, 2)
      lb3_xpnd = bounds_local_xpnd(1, 3)
      ub3_xpnd = bounds_local_xpnd(2, 3)
      IF ((in_use .EQ. REALDATA1D) .OR. (in_use .EQ. COMPLEXDATA1D)) THEN
         IF (in_space .EQ. RECIPROCALSPACE) THEN
            ALLOCATE (cr3d(lb1_xpnd:ub1_xpnd, lb2_xpnd:ub2_xpnd, lb3_xpnd:ub3_xpnd))
            ALLOCATE (cc3d(lb1_xpnd:ub1_xpnd, lb2_xpnd:ub2_xpnd, lb3_xpnd:ub3_xpnd))
            cc3d = RESHAPE(pw_in%cc, (/ub1_xpnd-lb1_xpnd+1, ub2_xpnd-lb2_xpnd+1, ub3_xpnd-lb3_xpnd+1/))
            CALL copy_cr(cc3d, cr3d)
            DEALLOCATE (cc3d)
         ELSE
            ALLOCATE (cr3d(lb1_xpnd:ub1_xpnd, lb2_xpnd:ub2_xpnd, lb3_xpnd:ub3_xpnd))
            cr3d = RESHAPE(pw_in%cr, (/ub1_xpnd-lb1_xpnd+1, ub2_xpnd-lb2_xpnd+1, ub3_xpnd-lb3_xpnd+1/))
         END IF
      ELSE
         IF (in_space .EQ. RECIPROCALSPACE) THEN
            CALL copy_cr(pw_in%cc3d, cr3d)
         ELSE
            ALLOCATE (cr3d(lb1_xpnd:ub1_xpnd, lb2_xpnd:ub2_xpnd, lb3_xpnd:ub3_xpnd))
            cr3d = pw_in%cr3d
         END IF
      END IF

! let all the nodes know about each others shifted local bounds
      ALLOCATE (bounds_local_all(2, 3, group_size))
      CALL mp_allgather(bounds_local_shftd, bounds_local_all, rs_group)

      DO i = 1, maxn_sendrecv
! no need to send to myself or to an invalid destination (pid = -1)
         IF ((dests_shrink(i) .NE. rs_mpo) .AND. (dests_shrink(i) .NE. -1)) THEN
            send_lb1 = bounds_local_all(1, 1, dests_shrink(i)+1)
            send_ub1 = bounds_local_all(2, 1, dests_shrink(i)+1)
            send_lb2 = bounds_local_all(1, 2, dests_shrink(i)+1)
            send_ub2 = bounds_local_all(2, 2, dests_shrink(i)+1)
            send_lb3 = bounds_local_all(1, 3, dests_shrink(i)+1)
            send_ub3 = bounds_local_all(2, 3, dests_shrink(i)+1)

            ALLOCATE (send_crmsg(send_lb1:send_ub1, send_lb2:send_ub2, send_lb3:send_ub3))
            send_crmsg = cr3d(send_lb1:send_ub1, send_lb2:send_ub2, send_lb3:send_ub3)
            CALL mp_isend(send_crmsg, dests_shrink(i), rs_group, send_req)
            CALL mp_wait(send_req)
            DEALLOCATE (send_crmsg)
         END IF
      END DO

      lb1_orig = bounds_local_shftd(1, 1)
      ub1_orig = bounds_local_shftd(2, 1)
      lb2_orig = bounds_local_shftd(1, 2)
      ub2_orig = bounds_local_shftd(2, 2)
      lb3_orig = bounds_local_shftd(1, 3)
      ub3_orig = bounds_local_shftd(2, 3)

! no need to receive from myself
      IF (srcs_shrink .EQ. rs_mpo) THEN
         pw_shrinked%cr3d = cr3d(lb1_orig:ub1_orig, lb2_orig:ub2_orig, lb3_orig:ub3_orig)
      ELSE IF (srcs_shrink .EQ. -1) THEN
! the source is invalid ... do nothing
      ELSE
         CALL mp_irecv(pw_shrinked%cr3d, srcs_shrink, rs_group, recv_req)
         CALL mp_wait(recv_req)
      END IF

      DEALLOCATE (bounds_local_all)
      DEALLOCATE (cr3d)
      CALL timestop(handle)

   END SUBROUTINE pw_shrink

! **************************************************************************************************
!> \brief   flips a 3d (real dp) array up to down (the way needed to expand data
!>          as explained in the description of the afore-defined subroutines)
!> \param cr3d_in input array
!> \param cr3d_out output array
!> \param bounds global lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE flipud(cr3d_in, cr3d_out, bounds)

      REAL(dp), DIMENSION(:, :, :), INTENT(IN), POINTER  :: cr3d_in
      REAL(dp), DIMENSION(:, :, :), INTENT(OUT), POINTER :: cr3d_out
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bounds

      CHARACTER(LEN=*), PARAMETER :: routineN = 'flipud', routineP = moduleN//':'//routineN

      INTEGER :: handle, i, lb1, lb1_glbl, lb1_new, lb2, lb2_glbl, lb2_new, lb3, lb3_glbl, &
         lb3_new, ub1, ub1_glbl, ub1_new, ub2, ub2_glbl, ub2_new, ub3, ub3_glbl, ub3_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indx
      INTEGER, DIMENSION(2, 3)                           :: bndsl, bndsl_new

      CALL timeset(routineN, handle)

      lb1 = LBOUND(cr3d_in, 1); ub1 = UBOUND(cr3d_in, 1)
      lb2 = LBOUND(cr3d_in, 2); ub2 = UBOUND(cr3d_in, 2)
      lb3 = LBOUND(cr3d_in, 3); ub3 = UBOUND(cr3d_in, 3)

      lb1_glbl = bounds(1, 1); ub1_glbl = bounds(2, 1)
      lb2_glbl = bounds(1, 2); ub2_glbl = bounds(2, 2)
      lb3_glbl = bounds(1, 3); ub3_glbl = bounds(2, 3)

      bndsl = RESHAPE((/lb1, ub1, lb2, ub2, lb3, ub3/), (/2, 3/))
      bndsl_new = flipud_bounds_local(bndsl, bounds)

      lb1_new = bndsl_new(1, 1); ub1_new = bndsl_new(2, 1)
      lb2_new = bndsl_new(1, 2); ub2_new = bndsl_new(2, 2)
      lb3_new = bndsl_new(1, 3); ub3_new = bndsl_new(2, 3)

      ALLOCATE (cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new))
      cr3d_out = 0.0_dp

! set the data at the missing grid points (in a periodic grid) equal to the data at
! the last existing grid points
      ALLOCATE (indx(ub1_new-lb1_new+1))
      indx(:) = (/(i, i=2*(ub1_glbl+1)-lb1_new, 2*(ub1_glbl+1)-ub1_new, -1)/)
      IF (lb1_new .EQ. ub1_glbl+1) indx(1) = indx(2)
      cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new) = cr3d_in(indx, :, :)

      CALL timestop(handle)

   END SUBROUTINE flipud

! **************************************************************************************************
!> \brief   flips a 3d (real dp) array left to right (the way needed to expand data
!>          as explained in the description of the afore-defined subroutines)
!> \param cr3d_in input array
!> \param cr3d_out output array
!> \param bounds global lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE fliplr(cr3d_in, cr3d_out, bounds)

      REAL(dp), DIMENSION(:, :, :), INTENT(IN), POINTER  :: cr3d_in
      REAL(dp), DIMENSION(:, :, :), INTENT(OUT), POINTER :: cr3d_out
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bounds

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fliplr', routineP = moduleN//':'//routineN

      INTEGER :: handle, i, lb1, lb1_glbl, lb1_new, lb2, lb2_glbl, lb2_new, lb3, lb3_glbl, &
         lb3_new, ub1, ub1_glbl, ub1_new, ub2, ub2_glbl, ub2_new, ub3, ub3_glbl, ub3_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indy
      INTEGER, DIMENSION(2, 3)                           :: bndsl, bndsl_new

      CALL timeset(routineN, handle)

      lb1 = LBOUND(cr3d_in, 1); ub1 = UBOUND(cr3d_in, 1)
      lb2 = LBOUND(cr3d_in, 2); ub2 = UBOUND(cr3d_in, 2)
      lb3 = LBOUND(cr3d_in, 3); ub3 = UBOUND(cr3d_in, 3)

      lb1_glbl = bounds(1, 1); ub1_glbl = bounds(2, 1)
      lb2_glbl = bounds(1, 2); ub2_glbl = bounds(2, 2)
      lb3_glbl = bounds(1, 3); ub3_glbl = bounds(2, 3)

      bndsl = RESHAPE((/lb1, ub1, lb2, ub2, lb3, ub3/), (/2, 3/))
      bndsl_new = fliplr_bounds_local(bndsl, bounds)

      lb1_new = bndsl_new(1, 1); ub1_new = bndsl_new(2, 1)
      lb2_new = bndsl_new(1, 2); ub2_new = bndsl_new(2, 2)
      lb3_new = bndsl_new(1, 3); ub3_new = bndsl_new(2, 3)

      ALLOCATE (cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new))
      cr3d_out = 0.0_dp

! set the data at the missing grid points (in a periodic grid) equal to the data at
! the last existing grid points
      ALLOCATE (indy(ub2_new-lb2_new+1))
      indy(:) = (/(i, i=2*(ub2_glbl+1)-lb2_new, 2*(ub2_glbl+1)-ub2_new, -1)/)
      IF (lb2_new .EQ. ub2_glbl+1) indy(1) = indy(2)
      cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new) = cr3d_in(:, indy, :)

      CALL timestop(handle)

   END SUBROUTINE fliplr

! **************************************************************************************************
!> \brief   flips a 3d (real dp) array back to front (the way needed to expand data
!>          as explained in the description of the afore-defined subroutines)
!> \param cr3d_in input array
!> \param cr3d_out output array
!> \param bounds global lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE flipbf(cr3d_in, cr3d_out, bounds)

      REAL(dp), DIMENSION(:, :, :), INTENT(IN), POINTER  :: cr3d_in
      REAL(dp), DIMENSION(:, :, :), INTENT(OUT), POINTER :: cr3d_out
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bounds

      CHARACTER(LEN=*), PARAMETER :: routineN = 'flipbf', routineP = moduleN//':'//routineN

      INTEGER :: handle, i, lb1, lb1_glbl, lb1_new, lb2, lb2_glbl, lb2_new, lb3, lb3_glbl, &
         lb3_new, ub1, ub1_glbl, ub1_new, ub2, ub2_glbl, ub2_new, ub3, ub3_glbl, ub3_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indz
      INTEGER, DIMENSION(2, 3)                           :: bndsl, bndsl_new

      CALL timeset(routineN, handle)

      lb1 = LBOUND(cr3d_in, 1); ub1 = UBOUND(cr3d_in, 1)
      lb2 = LBOUND(cr3d_in, 2); ub2 = UBOUND(cr3d_in, 2)
      lb3 = LBOUND(cr3d_in, 3); ub3 = UBOUND(cr3d_in, 3)

      lb1_glbl = bounds(1, 1); ub1_glbl = bounds(2, 1)
      lb2_glbl = bounds(1, 2); ub2_glbl = bounds(2, 2)
      lb3_glbl = bounds(1, 3); ub3_glbl = bounds(2, 3)

      bndsl = RESHAPE((/lb1, ub1, lb2, ub2, lb3, ub3/), (/2, 3/))
      bndsl_new = flipbf_bounds_local(bndsl, bounds)

      lb1_new = bndsl_new(1, 1); ub1_new = bndsl_new(2, 1)
      lb2_new = bndsl_new(1, 2); ub2_new = bndsl_new(2, 2)
      lb3_new = bndsl_new(1, 3); ub3_new = bndsl_new(2, 3)

      ALLOCATE (cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new))
      cr3d_out = 0.0_dp

! set the data at the missing grid points (in a periodic grid) equal to the data at
! the last existing grid points
      ALLOCATE (indz(ub3_new-lb3_new+1))
      indz(:) = (/(i, i=2*(ub3_glbl+1)-lb3_new, 2*(ub3_glbl+1)-ub3_new, -1)/)
      IF (lb3_new .EQ. ub3_glbl+1) indz(1) = indz(2)
      cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new) = cr3d_in(:, :, indz)

      CALL timestop(handle)

   END SUBROUTINE flipbf

! **************************************************************************************************
!> \brief   rotates a 3d (real dp) array by 180 degrees (the way needed to expand data
!>          as explained in the description of the afore-defined subroutines)
!> \param cr3d_in input array
!> \param cr3d_out output array
!> \param bounds global lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE rot180(cr3d_in, cr3d_out, bounds)

      REAL(dp), DIMENSION(:, :, :), INTENT(IN), POINTER  :: cr3d_in
      REAL(dp), DIMENSION(:, :, :), INTENT(OUT), POINTER :: cr3d_out
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bounds

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rot180', routineP = moduleN//':'//routineN

      INTEGER :: handle, i, lb1, lb1_glbl, lb1_new, lb2, lb2_glbl, lb2_new, lb3, lb3_glbl, &
         lb3_new, ub1, ub1_glbl, ub1_new, ub2, ub2_glbl, ub2_new, ub3, ub3_glbl, ub3_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indx, indy
      INTEGER, DIMENSION(2, 3)                           :: bndsl, bndsl_new

      CALL timeset(routineN, handle)

      lb1 = LBOUND(cr3d_in, 1); ub1 = UBOUND(cr3d_in, 1)
      lb2 = LBOUND(cr3d_in, 2); ub2 = UBOUND(cr3d_in, 2)
      lb3 = LBOUND(cr3d_in, 3); ub3 = UBOUND(cr3d_in, 3)

      lb1_glbl = bounds(1, 1); ub1_glbl = bounds(2, 1)
      lb2_glbl = bounds(1, 2); ub2_glbl = bounds(2, 2)
      lb3_glbl = bounds(1, 3); ub3_glbl = bounds(2, 3)

      bndsl = RESHAPE((/lb1, ub1, lb2, ub2, lb3, ub3/), (/2, 3/))
      bndsl_new = rot180_bounds_local(bndsl, bounds)

      lb1_new = bndsl_new(1, 1); ub1_new = bndsl_new(2, 1)
      lb2_new = bndsl_new(1, 2); ub2_new = bndsl_new(2, 2)
      lb3_new = bndsl_new(1, 3); ub3_new = bndsl_new(2, 3)

      ALLOCATE (cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new))
      cr3d_out = 0.0_dp

! set the data at the missing grid points (in a periodic grid) equal to the data at
! the last existing grid points
      ALLOCATE (indx(ub1_new-lb1_new+1), indy(ub2_new-lb2_new+1))
      indx(:) = (/(i, i=2*(ub1_glbl+1)-lb1_new, 2*(ub1_glbl+1)-ub1_new, -1)/)
      indy(:) = (/(i, i=2*(ub2_glbl+1)-lb2_new, 2*(ub2_glbl+1)-ub2_new, -1)/)
      IF (lb1_new .EQ. ub1_glbl+1) indx(1) = indx(2)
      IF (lb2_new .EQ. ub2_glbl+1) indy(1) = indy(2)
      cr3d_out(lb1_new:ub1_new, lb2_new:ub2_new, lb3_new:ub3_new) = cr3d_in(indx, indy, :)

      CALL timestop(handle)

   END SUBROUTINE rot180

! **************************************************************************************************
!> \brief   calculates the global and local bounds of the expanded data
!> \param pw_grid original plane-wave grid
!> \param neumann_directions directions in which dct should be performed
!> \param srcs_expand list of the source processes (pw_expand)
!> \param flipg_stat flipping status for the received data chunks (pw_expand)
!> \param bounds_shftd bounds of the original grid shifted to have g0 in the middle of the cell
!> \param bounds_local_shftd local bounds of the original grid after shifting
!> \param recv_msgs_bnds bounds of the messages to be received (pw_expand)
!> \param bounds_new new global lower and upper bounds
!> \param bounds_local_new new local lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE expansion_bounds(pw_grid, neumann_directions, srcs_expand, flipg_stat, &
                               bounds_shftd, bounds_local_shftd, &
                               recv_msgs_bnds, bounds_new, bounds_local_new)

      TYPE(pw_grid_type), INTENT(IN), POINTER            :: pw_grid
      INTEGER, INTENT(IN)                                :: neumann_directions
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: srcs_expand, flipg_stat
      INTEGER, DIMENSION(2, 3), INTENT(OUT)              :: bounds_shftd, bounds_local_shftd
      INTEGER, DIMENSION(:, :, :), INTENT(INOUT), &
         POINTER                                         :: recv_msgs_bnds
      INTEGER, DIMENSION(2, 3), INTENT(OUT)              :: bounds_new, bounds_local_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'expansion_bounds', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group_size, handle, i, lb1_new, lb2_new, &
                                                            lb3_new, loc, maxn_sendrecv, rs_group, &
                                                            rs_mpo, ub1_new, ub2_new, ub3_new
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: src_hist
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: bounds_local_all, bounds_local_new_all, &
                                                            pcs_bnds
      INTEGER, DIMENSION(2, 3)                           :: bounds, bounds_local
      INTEGER, DIMENSION(3)                              :: npts_new, shf_yesno, shift

      CALL timeset(routineN, handle)

      rs_group = pw_grid%para%rs_group
      rs_mpo = pw_grid%para%my_pos
      group_size = pw_grid%para%group_size
      bounds = pw_grid%bounds
      bounds_local = pw_grid%bounds_local

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ)
         maxn_sendrecv = 4
         shf_yesno = (/1, 1, 1/)
      CASE (neumannXY)
         maxn_sendrecv = 4
         shf_yesno = (/1, 1, 0/)
      CASE (neumannXZ)
         maxn_sendrecv = 2
         shf_yesno = (/1, 0, 1/)
      CASE (neumannYZ)
         maxn_sendrecv = 2
         shf_yesno = (/0, 1, 1/)
      CASE (neumannX)
         maxn_sendrecv = 2
         shf_yesno = (/1, 0, 0/)
      CASE (neumannY)
         maxn_sendrecv = 2
         shf_yesno = (/0, 1, 0/)
      CASE (neumannZ)
         maxn_sendrecv = 1
         shf_yesno = (/0, 0, 1/)
      END SELECT

      ALLOCATE (pcs_bnds(2, 3, maxn_sendrecv))
      ALLOCATE (src_hist(maxn_sendrecv))

      ! Note that this is not easily FFT-able ... needed anyway, so link in FFTW.
      npts_new = 2*pw_grid%npts
      shift = -npts_new/2
      shift = shift-bounds(1, :)
      bounds_shftd(:, 1) = bounds(:, 1)+shf_yesno(1)*shift(1)
      bounds_shftd(:, 2) = bounds(:, 2)+shf_yesno(2)*shift(2)
      bounds_shftd(:, 3) = bounds(:, 3)+shf_yesno(3)*shift(3)
      bounds_local_shftd(:, 1) = bounds_local(:, 1)+shf_yesno(1)*shift(1)
      bounds_local_shftd(:, 2) = bounds_local(:, 2)+shf_yesno(2)*shift(2)
      bounds_local_shftd(:, 3) = bounds_local(:, 3)+shf_yesno(3)*shift(3)

! let all the nodes know about each others local shifted bounds
      ALLOCATE (bounds_local_all(2, 3, group_size))
      CALL mp_allgather(bounds_local_shftd, bounds_local_all, rs_group)

      src_hist = -1 ! keeps the history of sources

      DO i = 1, maxn_sendrecv
! no need to receive from myself
         IF (srcs_expand(i) .EQ. rs_mpo) THEN
            recv_msgs_bnds(1, 1, i) = bounds_local_shftd(1, 1)
            recv_msgs_bnds(2, 1, i) = bounds_local_shftd(2, 1)
            recv_msgs_bnds(1, 2, i) = bounds_local_shftd(1, 2)
            recv_msgs_bnds(2, 2, i) = bounds_local_shftd(2, 2)
            recv_msgs_bnds(1, 3, i) = bounds_local_shftd(1, 3)
            recv_msgs_bnds(2, 3, i) = bounds_local_shftd(2, 3)
! if I have already received data from the source, just use the one from the last time
         ELSE IF (ANY(src_hist .EQ. srcs_expand(i))) THEN
            loc = MINLOC(ABS(src_hist-srcs_expand(i)), 1)
            recv_msgs_bnds(1, 1, i) = bounds_local_all(1, 1, srcs_expand(loc)+1)
            recv_msgs_bnds(2, 1, i) = bounds_local_all(2, 1, srcs_expand(loc)+1)
            recv_msgs_bnds(1, 2, i) = bounds_local_all(1, 2, srcs_expand(loc)+1)
            recv_msgs_bnds(2, 2, i) = bounds_local_all(2, 2, srcs_expand(loc)+1)
            recv_msgs_bnds(1, 3, i) = bounds_local_all(1, 3, srcs_expand(loc)+1)
            recv_msgs_bnds(2, 3, i) = bounds_local_all(2, 3, srcs_expand(loc)+1)
         ELSE
            recv_msgs_bnds(1, 1, i) = bounds_local_all(1, 1, srcs_expand(i)+1)
            recv_msgs_bnds(2, 1, i) = bounds_local_all(2, 1, srcs_expand(i)+1)
            recv_msgs_bnds(1, 2, i) = bounds_local_all(1, 2, srcs_expand(i)+1)
            recv_msgs_bnds(2, 2, i) = bounds_local_all(2, 2, srcs_expand(i)+1)
            recv_msgs_bnds(1, 3, i) = bounds_local_all(1, 3, srcs_expand(i)+1)
            recv_msgs_bnds(2, 3, i) = bounds_local_all(2, 3, srcs_expand(i)+1)
         END IF
         src_hist(i) = srcs_expand(i)
      END DO

! flip the received data based on the flipping status
      DO i = 1, maxn_sendrecv
         SELECT CASE (flipg_stat (i))
         CASE (NOT_FLIPPED)
            pcs_bnds(:, :, i) = recv_msgs_bnds(:, :, i)
         CASE (UD_FLIPPED)
            pcs_bnds(:, :, i) = flipud_bounds_local(recv_msgs_bnds(:, :, i), bounds_shftd)
         CASE (LR_FLIPPED)
            pcs_bnds(:, :, i) = fliplr_bounds_local(recv_msgs_bnds(:, :, i), bounds_shftd)
         CASE (BF_FLIPPED)
            pcs_bnds(:, :, i) = flipbf_bounds_local(recv_msgs_bnds(:, :, i), bounds_shftd)
         CASE (ROTATED)
            pcs_bnds(:, :, i) = rot180_bounds_local(recv_msgs_bnds(:, :, i), bounds_shftd)
         END SELECT
      END DO

      lb1_new = MINVAL(pcs_bnds(1, 1, :)); ub1_new = MAXVAL(pcs_bnds(2, 1, :))
      lb2_new = MINVAL(pcs_bnds(1, 2, :)); ub2_new = MAXVAL(pcs_bnds(2, 2, :))
      lb3_new = MINVAL(pcs_bnds(1, 3, :)); ub3_new = MAXVAL(pcs_bnds(2, 3, :))

! calculate the new local and global bounds
      bounds_local_new(1, 1) = MINVAL(pcs_bnds(1, 1, :))
      bounds_local_new(2, 1) = MAXVAL(pcs_bnds(2, 1, :))
      bounds_local_new(1, 2) = MINVAL(pcs_bnds(1, 2, :))
      bounds_local_new(2, 2) = MAXVAL(pcs_bnds(2, 2, :))
      bounds_local_new(1, 3) = MINVAL(pcs_bnds(1, 3, :))
      SELECT CASE (neumann_directions)
      CASE (neumannXYZ, neumannXZ, neumannYZ, neumannZ)
         bounds_local_new(2, 3) = 2*(MAXVAL(pcs_bnds(2, 3, :))+1)-bounds_local_new(1, 3)-1
      CASE (neumannXY, neumannX, neumannY)
         bounds_local_new(2, 3) = MAXVAL(pcs_bnds(2, 3, :))
      END SELECT

      ALLOCATE (bounds_local_new_all(2, 3, group_size))
      CALL mp_allgather(bounds_local_new, bounds_local_new_all, rs_group)
      bounds_new(1, 1) = MINVAL(bounds_local_new_all(1, 1, :))
      bounds_new(2, 1) = MAXVAL(bounds_local_new_all(2, 1, :))
      bounds_new(1, 2) = MINVAL(bounds_local_new_all(1, 2, :))
      bounds_new(2, 2) = MAXVAL(bounds_local_new_all(2, 2, :))
      bounds_new(1, 3) = MINVAL(bounds_local_new_all(1, 3, :))
      bounds_new(2, 3) = MAXVAL(bounds_local_new_all(2, 3, :))

      DEALLOCATE (bounds_local_all, bounds_local_new_all)

      CALL timestop(handle)

   END SUBROUTINE expansion_bounds

! **************************************************************************************************
!> \brief   precalculates the local bounds of a 3d array after applying flipud
!> \param bndsl_in current local lower and upper bounds
!> \param bounds global lower and upper bounds
!> \retval bndsl_out new local lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   FUNCTION flipud_bounds_local(bndsl_in, bounds) RESULT(bndsl_out)

      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bndsl_in, bounds
      INTEGER, DIMENSION(2, 3)                           :: bndsl_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'flipud_bounds_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      bndsl_out(1, 1) = 2*(bounds(2, 1)+1)-bndsl_in(2, 1)
      bndsl_out(2, 1) = 2*(bounds(2, 1)+1)-bndsl_in(1, 1)
      IF (bndsl_out(1, 1) .EQ. bounds(2, 1)+2) bndsl_out(1, 1) = bndsl_out(1, 1)-1
      IF (bndsl_out(2, 1) .EQ. 2*(bounds(2, 1)+1)-bounds(1, 1)) bndsl_out(2, 1) = bndsl_out(2, 1)-1

      bndsl_out(1, 2) = bndsl_in(1, 2)
      bndsl_out(2, 2) = bndsl_in(2, 2)

      bndsl_out(1, 3) = bndsl_in(1, 3)
      bndsl_out(2, 3) = bndsl_in(2, 3)

      CALL timestop(handle)

   END FUNCTION flipud_bounds_local

! **************************************************************************************************
!> \brief   precalculates the local bounds of a 3d array after applying fliplr
!> \param bndsl_in current local lower and upper bounds
!> \param bounds global lower and upper bounds
!> \retval bndsl_out new local lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   FUNCTION fliplr_bounds_local(bndsl_in, bounds) RESULT(bndsl_out)

      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bndsl_in, bounds
      INTEGER, DIMENSION(2, 3)                           :: bndsl_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fliplr_bounds_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      bndsl_out(1, 1) = bndsl_in(1, 1)
      bndsl_out(2, 1) = bndsl_in(2, 1)

      bndsl_out(1, 2) = 2*(bounds(2, 2)+1)-bndsl_in(2, 2)
      bndsl_out(2, 2) = 2*(bounds(2, 2)+1)-bndsl_in(1, 2)
      IF (bndsl_out(1, 2) .EQ. bounds(2, 2)+2) bndsl_out(1, 2) = bndsl_out(1, 2)-1
      IF (bndsl_out(2, 2) .EQ. 2*(bounds(2, 2)+1)-bounds(1, 2)) bndsl_out(2, 2) = bndsl_out(2, 2)-1

      bndsl_out(1, 3) = bndsl_in(1, 3)
      bndsl_out(2, 3) = bndsl_in(2, 3)

      CALL timestop(handle)

   END FUNCTION fliplr_bounds_local

! **************************************************************************************************
!> \brief   precalculates the local bounds of a 3d array after applying flipbf
!> \param bndsl_in current local lower and upper bounds
!> \param bounds global lower and upper bounds
!> \retval bndsl_out new local lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   FUNCTION flipbf_bounds_local(bndsl_in, bounds) RESULT(bndsl_out)

      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bndsl_in, bounds
      INTEGER, DIMENSION(2, 3)                           :: bndsl_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'flipbf_bounds_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      bndsl_out(1, 1) = bndsl_in(1, 1)
      bndsl_out(2, 1) = bndsl_in(2, 1)

      bndsl_out(1, 2) = bndsl_in(1, 2)
      bndsl_out(2, 2) = bndsl_in(2, 2)

      bndsl_out(1, 3) = 2*(bounds(2, 3)+1)-bndsl_in(2, 3)
      bndsl_out(2, 3) = 2*(bounds(2, 3)+1)-bndsl_in(1, 3)
      IF (bndsl_out(1, 3) .EQ. bounds(2, 3)+2) bndsl_out(1, 3) = bndsl_out(1, 3)-1
      IF (bndsl_out(2, 3) .EQ. 2*(bounds(2, 3)+1)-bounds(1, 3)) bndsl_out(2, 3) = bndsl_out(2, 3)-1

      CALL timestop(handle)

   END FUNCTION flipbf_bounds_local

! **************************************************************************************************
!> \brief   precalculates the local bounds of a 3d array after applying rot180
!> \param bndsl_in current local lower and upper bounds
!> \param bounds global lower and upper bounds
!> \retval bndsl_out new local lower and upper bounds
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   FUNCTION rot180_bounds_local(bndsl_in, bounds) RESULT(bndsl_out)

      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bndsl_in, bounds
      INTEGER, DIMENSION(2, 3)                           :: bndsl_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rot180_bounds_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      bndsl_out(1, 1) = 2*(bounds(2, 1)+1)-bndsl_in(2, 1)
      bndsl_out(2, 1) = 2*(bounds(2, 1)+1)-bndsl_in(1, 1)
      IF (bndsl_out(1, 1) .EQ. bounds(2, 1)+2) bndsl_out(1, 1) = bndsl_out(1, 1)-1
      IF (bndsl_out(2, 1) .EQ. 2*(bounds(2, 1)+1)-bounds(1, 1)) bndsl_out(2, 1) = bndsl_out(2, 1)-1

      bndsl_out(1, 2) = 2*(bounds(2, 2)+1)-bndsl_in(2, 2)
      bndsl_out(2, 2) = 2*(bounds(2, 2)+1)-bndsl_in(1, 2)
      IF (bndsl_out(1, 2) .EQ. bounds(2, 2)+2) bndsl_out(1, 2) = bndsl_out(1, 2)-1
      IF (bndsl_out(2, 2) .EQ. 2*(bounds(2, 2)+1)-bounds(1, 2)) bndsl_out(2, 2) = bndsl_out(2, 2)-1

      bndsl_out(1, 3) = bndsl_in(1, 3)
      bndsl_out(2, 3) = bndsl_in(2, 3)

      CALL timestop(handle)

   END FUNCTION rot180_bounds_local

END MODULE dct

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types containing essential information for running implicit (iterative)
!>        Poisson solver
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Hossein Bani-Hashemian
! **************************************************************************************************
MODULE ps_implicit_types

   USE dct,                             ONLY: dct_type,&
                                              dct_type_release
   USE dielectric_types,                ONLY: dielectric_release,&
                                              dielectric_type
   USE dirichlet_bc_types,              ONLY: dbc_release,&
                                              dirichlet_bc_p_type
   USE kinds,                           ONLY: dp
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: pw_release,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_implicit_types'

   TYPE :: ps_implicit_parameters
      INTEGER                           :: boundary_condition
      LOGICAL                           :: zero_initial_guess
      INTEGER                           :: max_iter
      REAL(KIND=dp)                     :: tol
      REAL(KIND=dp)                     :: omega
      INTEGER                           :: neumann_directions
   END TYPE ps_implicit_parameters

   TYPE :: ps_implicit_type
      LOGICAL                           :: do_dbc_cube
      TYPE(dielectric_type), POINTER    :: dielectric
      TYPE(dct_type)                    :: dct_env
      TYPE(pw_type), POINTER            :: initial_guess
      TYPE(pw_type), POINTER            :: v_eps
      TYPE(pw_type), POINTER            :: cstr_charge
      REAL(dp), DIMENSION(:), POINTER   :: initial_lambda => NULL()
      REAL(dp)                          :: ehartree
      REAL(dp)                          :: electric_enthalpy
      INTEGER                           :: times_called
      TYPE(dirichlet_bc_p_type), &
         DIMENSION(:), POINTER           :: contacts => NULL()
      REAL(dp), DIMENSION(:, :), POINTER :: QS => NULL()
      REAL(dp), DIMENSION(:, :), POINTER :: Rinv => NULL()
      REAL(dp), DIMENSION(:, :), POINTER :: B => NULL()
      REAL(dp), DIMENSION(:, :), POINTER :: Bt => NULL()
      REAL(dp), DIMENSION(:), POINTER   :: v_D => NULL()
      INTEGER, DIMENSION(:), POINTER    :: idx_1dto3d => NULL()
   END TYPE ps_implicit_type

   PUBLIC ps_implicit_type, ps_implicit_parameters
   PUBLIC ps_implicit_release

   INTEGER, PARAMETER, PUBLIC           :: PERIODIC_BC = 0, &
                                           MIXED_BC = 1, &
                                           MIXED_PERIODIC_BC = 2, &
                                           NEUMANN_BC = 3
CONTAINS

! **************************************************************************************************
!> \brief   Deallocates ps_implicit
!> \param ps_implicit_env the implicit_env to be deallocated
!> \param ps_implicit_params implicit env parameters
!> \param pw_pool pool of plane-wave grid
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_release(ps_implicit_env, ps_implicit_params, pw_pool)
      TYPE(ps_implicit_type), POINTER                    :: ps_implicit_env
      TYPE(ps_implicit_parameters), INTENT(INOUT)        :: ps_implicit_params
      TYPE(pw_pool_type), INTENT(IN), OPTIONAL, POINTER  :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'ps_implicit_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: boundary_condition, handle
      LOGICAL                                            :: can_give_back, do_dbc_cube

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(ps_implicit_env)) THEN
         can_give_back = PRESENT(pw_pool)
         IF (can_give_back) can_give_back = ASSOCIATED(pw_pool)
         do_dbc_cube = ps_implicit_env%do_dbc_cube

         IF (can_give_back) THEN
            CALL pw_pool_give_back_pw(pw_pool, ps_implicit_env%initial_guess, &
                                      accept_non_compatible=.TRUE.)
            CALL pw_pool_give_back_pw(pw_pool, ps_implicit_env%v_eps, &
                                      accept_non_compatible=.TRUE.)
            CALL pw_pool_give_back_pw(pw_pool, ps_implicit_env%cstr_charge, &
                                      accept_non_compatible=.TRUE.)
            CALL dbc_release(ps_implicit_env%contacts, do_dbc_cube, pw_pool=pw_pool)
         ELSE
            CALL pw_release(ps_implicit_env%initial_guess)
            CALL pw_release(ps_implicit_env%v_eps)
            CALL pw_release(ps_implicit_env%cstr_charge)
            CALL dbc_release(ps_implicit_env%contacts, do_dbc_cube)
         END IF

         DEALLOCATE (ps_implicit_env%initial_lambda)
         DEALLOCATE (ps_implicit_env%B)
         DEALLOCATE (ps_implicit_env%Bt)
         DEALLOCATE (ps_implicit_env%QS)
         DEALLOCATE (ps_implicit_env%Rinv)
         DEALLOCATE (ps_implicit_env%v_D)
         DEALLOCATE (ps_implicit_env%idx_1dto3d)

         CALL dielectric_release(ps_implicit_env%dielectric, pw_pool)
         boundary_condition = ps_implicit_params%boundary_condition
         IF (boundary_condition .EQ. MIXED_BC .OR. boundary_condition .EQ. NEUMANN_BC) THEN
            CALL dct_type_release(ps_implicit_env%dct_env)
         END IF

         DEALLOCATE (ps_implicit_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_release

END MODULE ps_implicit_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition and initialisation of the ps_wavelet data type.
!> \author Florian Schiffmann (09.2007,fschiff)
! **************************************************************************************************
MODULE ps_wavelet_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_wavelet_types'

   INTEGER, PARAMETER, PUBLIC               :: WAVELET3D = 1400, &
                                               WAVELET2D = 1401, &
                                               WAVELET1D = 1402, &
                                               WAVELET0D = 1403

   PUBLIC :: ps_wavelet_type, &
             ps_wavelet_release

! **************************************************************************************************
!> \par History
!>      09.2007 created [Florian Schiffmann]
!> \author fschiff
! **************************************************************************************************
   TYPE ps_wavelet_type
      CHARACTER(LEN=1)                                  :: geocode
      CHARACTER(LEN=1)                                  :: datacode
      INTEGER                                           :: itype_scf
      INTEGER                                           :: method, special_dimension
      REAL(kind=dp), POINTER, DIMENSION(:)             :: karray
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER    :: rho_z_sliced
      INTEGER, DIMENSION(3)                              :: PS_grid
   END TYPE ps_wavelet_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param wavelet ...
! **************************************************************************************************
   SUBROUTINE ps_wavelet_release(wavelet)

      TYPE(ps_wavelet_type), POINTER                     :: wavelet

      CHARACTER(len=*), PARAMETER :: routineN = 'ps_wavelet_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(wavelet)) THEN
         IF (ASSOCIATED(wavelet%karray)) &
            DEALLOCATE (wavelet%karray)
         IF (ASSOCIATED(wavelet%rho_z_sliced)) &
            DEALLOCATE (wavelet%rho_z_sliced)
         DEALLOCATE (wavelet)
      END IF
   END SUBROUTINE ps_wavelet_release

END MODULE ps_wavelet_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \note
!>      Basic type for real space grid methods
!> \par History
!>      JGH (22-May-2002) : New routine rs_grid_zero
!>      JGH (12-Jun-2002) : Bug fix for mpi groups
!>      JGH (19-Jun-2003) : Added routine for task distribution
!>      JGH (23-Nov-2003) : Added routine for task loop separation
!> \author JGH (18-Mar-2001)
! **************************************************************************************************
MODULE realspace_grid_types
   USE cp_array_utils_r,                ONLY: cp_1d_r_p_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE fast,                            ONLY: copy_cr,&
                                              copy_rc
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: m_memory
   USE mathlib,                         ONLY: det_3x3
   USE message_passing,                 ONLY: &
        mp_comm_dup, mp_comm_free, mp_environ, mp_irecv, mp_isend, mp_isendrecv, mp_max, mp_min, &
        mp_request_null, mp_sum, mp_sync, mp_waitall, mp_waitany
   USE pw_grid_types,                   ONLY: PW_MODE_LOCAL,&
                                              pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_release,&
                                              pw_grid_retain
   USE pw_methods,                      ONLY: pw_integrate_function
   USE pw_types,                        ONLY: COMPLEXDATA3D,&
                                              REALDATA3D,&
                                              pw_type
   USE util,                            ONLY: get_limit

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: realspace_grid_type, &
             realspace_grid_desc_type, &
             realspace_grid_p_type, &
             realspace_grid_desc_p_type, &
             realspace_grid_input_type

   PUBLIC :: rs_pw_transfer, &
             rs_grid_zero, &
             rs_grid_set_box, &
             rs_grid_create, &
             rs_grid_create_descriptor, &
             rs_grid_retain, &
             rs_grid_retain_descriptor, &
             rs_grid_release, &
             rs_grid_release_descriptor, &
             rs_grid_reorder_ranks, &
             rs_grid_print, &
             rs_grid_locate_rank, &
             rs_grid_max_ngpts, &
             rs_grid_mult_and_add

   INTEGER, PARAMETER, PUBLIC               :: rsgrid_distributed = 0, &
                                               rsgrid_replicated = 1, &
                                               rsgrid_automatic = 2

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'realspace_grid_types'
   INTEGER, SAVE, PRIVATE :: last_rs_id = 0
   INTEGER, SAVE, PRIVATE :: allocated_rs_grid_count = 0
   INTEGER, PARAMETER, PUBLIC :: rs2pw = 11, pw2rs = 12

! **************************************************************************************************
   TYPE realspace_grid_input_type
      INTEGER       :: distribution_type
      INTEGER       :: distribution_layout(3)
      REAL(KIND=dp) :: memory_factor
      LOGICAL       :: lock_distribution
      INTEGER       :: nsmax
      REAL(KIND=dp) :: halo_reduction_factor
   END TYPE realspace_grid_input_type

! **************************************************************************************************
   TYPE realspace_grid_desc_type
      INTEGER :: grid_id ! tag of the pw_grid

      TYPE(pw_grid_type), POINTER   :: pw ! the pw grid

      INTEGER :: ref_count ! reference count

      INTEGER(int_8) :: ngpts ! # grid points
      INTEGER, DIMENSION(3) :: npts ! # grid points per dimension
      INTEGER, DIMENSION(3) :: lb ! lower bounds
      INTEGER, DIMENSION(3) :: ub ! upper bounds

      INTEGER :: border ! border points

      INTEGER, DIMENSION(3) :: perd ! periodicity enforced
      REAL(KIND=dp), DIMENSION(3, 3) :: dh ! incremental grid matrix
      REAL(KIND=dp), DIMENSION(3, 3) :: dh_inv ! inverse incremental grid matrix
      LOGICAL :: orthorhombic ! grid symmetry

      LOGICAL :: parallel ! whether the corresponding pw grid is distributed
      LOGICAL :: distributed ! whether the rs grid is distributed
      ! these MPI related quantities are only meaningful depending on how the grid has been layed out
      ! they are most useful for fully distributed grids, where they reflect the topology of the grid
      INTEGER :: group
      INTEGER :: my_pos
      LOGICAL :: group_head
      INTEGER :: group_size
      INTEGER, DIMENSION(3) :: group_dim
      INTEGER, DIMENSION(3) :: group_coor
      INTEGER, DIMENSION(3) :: neighbours
      ! only meaningful on distributed grids
      ! a list of bounds for each CPU
      INTEGER, DIMENSION(:, :), POINTER :: lb_global
      INTEGER, DIMENSION(:, :), POINTER :: ub_global
      ! a mapping from linear rank to 3d coord
      INTEGER, DIMENSION(:, :), POINTER :: rank2coord
      INTEGER, DIMENSION(:, :, :), POINTER :: coord2rank
      ! a mapping from index to rank (which allows to figure out easily on which rank a given point of the grid is)
      INTEGER, DIMENSION(:), POINTER :: x2coord
      INTEGER, DIMENSION(:), POINTER :: y2coord
      INTEGER, DIMENSION(:), POINTER :: z2coord

      INTEGER                :: my_virtual_pos
      INTEGER, DIMENSION(3) :: virtual_group_coor

      INTEGER, DIMENSION(:), ALLOCATABLE :: virtual2real, real2virtual

   END TYPE realspace_grid_desc_type

   TYPE realspace_grid_type

      TYPE(realspace_grid_desc_type), POINTER :: desc

      INTEGER :: id_nr ! unique identifier of rs
      INTEGER :: ref_count ! reference count

      INTEGER :: ngpts_local ! local dimensions
      INTEGER, DIMENSION(3) :: npts_local
      INTEGER, DIMENSION(3) :: lb_local
      INTEGER, DIMENSION(3) :: ub_local
      INTEGER, DIMENSION(3) :: lb_real ! lower bounds of the real local data
      INTEGER, DIMENSION(3) :: ub_real ! upper bounds of the real local data

      INTEGER, DIMENSION(:), POINTER :: px, py, pz ! index translators
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: r ! the grid

   END TYPE realspace_grid_type

! **************************************************************************************************
   TYPE realspace_grid_p_type
      TYPE(realspace_grid_type), POINTER :: rs_grid
   END TYPE realspace_grid_p_type

   TYPE realspace_grid_desc_p_type
      TYPE(realspace_grid_desc_type), POINTER :: rs_desc
   END TYPE realspace_grid_desc_p_type

CONTAINS

! **************************************************************************************************
!> \brief returns the 1D rank of the task which is a cartesian shift away from 1D rank rank_in
!>        only possible if rs_grid is a distributed grid
!> \param rs_desc ...
!> \param rank_in ...
!> \param shift ...
!> \retval rank_out ...
! **************************************************************************************************
   FUNCTION rs_grid_locate_rank(rs_desc, rank_in, shift) RESULT(rank_out)
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      INTEGER, INTENT(IN)                                :: rank_in
      INTEGER, DIMENSION(3), INTENT(IN)                  :: shift
      INTEGER                                            :: rank_out

      INTEGER                                            :: coord(3)

      coord = MODULO(rs_desc%rank2coord(:, rank_in)+shift, rs_desc%group_dim)
      rank_out = rs_desc%coord2rank(coord(1), coord(2), coord(3))
   END FUNCTION rs_grid_locate_rank

! **************************************************************************************************
!> \brief Determine the setup of real space grids - this is divided up into the
!>        creation of a descriptor and the actual grid itself (see rs_grid_create)
!> \param desc ...
!> \param pw_grid ...
!> \param input_settings ...
!> \param border_points ...
!> \par History
!>      JGH (08-Jun-2003) : nsmax <= 0 indicates fully replicated grid
!>      Iain Bethune (05-Sep-2008) : modified cut heuristic
!>      (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
!>      - Create a descriptor for realspace grids with a number of border
!>        points as exactly given by the optional argument border_points.
!>        These grids are always distributed.
!>        (27.11.2013, Matthias Krack)
!> \author JGH (18-Mar-2001)
! **************************************************************************************************
   SUBROUTINE rs_grid_create_descriptor(desc, pw_grid, input_settings, border_points)
      TYPE(realspace_grid_desc_type), POINTER            :: desc
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(realspace_grid_input_type), INTENT(IN)        :: input_settings
      INTEGER, INTENT(IN), OPTIONAL                      :: border_points

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rs_grid_create_descriptor', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: border_size, dir, handle, i, j, k, l, &
                                                            lb(2), n_slices(3), n_slices_tmp(3), &
                                                            neighbours(3), nmin
      LOGICAL                                            :: overlap
      REAL(KIND=dp)                                      :: ratio, ratio_best, volume, volume_dist

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(pw_grid)))CALL cp__a("pw/realspace_grid_types.F",221)

      IF (PRESENT(border_points)) THEN
         border_size = border_points
      ELSE
         border_size = 0
      END IF

      ALLOCATE (desc)

      CALL mp_sync(pw_grid%para%group)

      NULLIFY (desc%rank2coord, desc%coord2rank, desc%lb_global, desc%ub_global, desc%x2coord, desc%y2coord, desc%z2coord)

      desc%pw => pw_grid
      CALL pw_grid_retain(desc%pw)

      desc%dh = pw_grid%dh
      desc%dh_inv = pw_grid%dh_inv
      desc%orthorhombic = pw_grid%orthorhombic
      desc%grid_id = pw_grid%id_nr
      desc%ref_count = 1

      IF (pw_grid%para%mode == PW_MODE_LOCAL) THEN
         ! The corresponding group has dimension 1
         ! All operations will be done localy
         desc%npts = pw_grid%npts
         desc%ngpts = PRODUCT(INT(desc%npts, KIND=int_8))
         desc%lb = pw_grid%bounds(1, :)
         desc%ub = pw_grid%bounds(2, :)
         desc%border = border_size
         IF (border_size == 0) THEN
            desc%perd = 1
         ELSE
            desc%perd = 0
         END IF
         desc%parallel = .FALSE.
         desc%distributed = .FALSE.
         desc%group = -1
         desc%group_size = 1
         desc%group_head = .TRUE.
         desc%group_dim = 1
         desc%group_coor = 0
         desc%my_pos = 0
      ELSE
         ! group size of desc grid
         ! global grid dimensions are still the same
         desc%group_size = pw_grid%para%group_size
         desc%npts = pw_grid%npts
         desc%ngpts = PRODUCT(INT(desc%npts, KIND=int_8))
         desc%lb = pw_grid%bounds(1, :)
         desc%ub = pw_grid%bounds(2, :)

         ! this is the eventual border size
         IF (border_size == 0) THEN
            nmin = (input_settings%nsmax+1)/2
            nmin = MAX(0, NINT(nmin*input_settings%halo_reduction_factor))
         ELSE
            ! Set explicitly the requested border size
            nmin = border_size
         END IF

         IF (input_settings%distribution_type == rsgrid_replicated) THEN

            n_slices = 1
            IF (border_size > 0) THEN
               CALL cp_abort(cp__l("pw/realspace_grid_types.F",287), &
                             "An explicit border size > 0 is not yet working for "// &
                             "replicated realspace grids. Request DISTRIBUTION_TYPE "// &
                             "distributed for RS_GRID explicitly.")
            END IF

         ELSE

            n_slices = 1
            ratio_best = -HUGE(ratio_best)

            ! don't allow distributions with more processors than real grid points
            DO k = 1, MIN(desc%npts(3), desc%group_size)
            DO j = 1, MIN(desc%npts(2), desc%group_size)
               i = MIN(desc%npts(1), desc%group_size/(j*k))
               n_slices_tmp = (/i, j, k/)

               ! we don't match the actual number of CPUs
               IF (PRODUCT(n_slices_tmp) .NE. desc%group_size) CYCLE

               ! we see if there has been a input constraint
               ! i.e. if the layout is not -1 we need to fullfil it
               IF (.NOT. ALL(PACK(n_slices_tmp == input_settings%distribution_layout, &
                                  (/-1, -1, -1/) /= input_settings%distribution_layout) &
                             )) CYCLE

               ! we currently can not work with a grid that has borders that overlaps with the local data of the grid itself
               overlap = .FALSE.
               DO dir = 1, 3
                  IF (n_slices_tmp(dir) > 1) THEN
                     neighbours(dir) = HUGE(0)
                     DO l = 0, n_slices_tmp(dir)-1
                        lb = get_limit(desc%npts(dir), n_slices_tmp(dir), l)
                        neighbours(dir) = MIN(lb(2)-lb(1)+1, neighbours(dir))
                     ENDDO
                     desc%neighbours(dir) = (nmin+neighbours(dir)-1)/neighbours(dir)
                     IF (desc%neighbours(dir) .GE. n_slices_tmp(dir)) overlap = .TRUE.
                  ELSE
                     neighbours(dir) = 0
                  ENDIF
               ENDDO
               ! XXXXXXX I think the overlap criterium / neighbour calculation is too conservative
               ! write(*,*) n_slices_tmp,desc % neighbours, overlap
               IF (overlap) CYCLE

               ! a heuristic optimisation to reduce the memory usage
               ! we go for the smallest local to real volume
               ! volume of the box without the wings / volume of the box with the wings
               ! with prefactodesc to promote less cuts in Z dimension
               ratio = PRODUCT(REAL(desc%npts, KIND=dp)/n_slices_tmp)/ &
                       PRODUCT(REAL(desc%npts, KIND=dp)/n_slices_tmp+ &
                               MERGE((/0.0, 0.0, 0.0/), 2*(/1.06*nmin, 1.05*nmin, 1.03*nmin/), n_slices_tmp == (/1, 1, 1/)))
               IF (ratio > ratio_best) THEN
                  ratio_best = ratio
                  n_slices = n_slices_tmp
               ENDIF

            ENDDO
            ENDDO

            ! if automatic we can still decide this is a replicated grid
            ! if the memory gain (or the gain is messages) is too small.
            IF (input_settings%distribution_type == rsgrid_automatic) THEN
               volume = PRODUCT(REAL(desc%npts, KIND=dp))
               volume_dist = PRODUCT(REAL(desc%npts, KIND=dp)/n_slices+ &
                                     MERGE((/0, 0, 0/), 2*(/nmin, nmin, nmin/), n_slices == (/1, 1, 1/)))
               IF (volume < volume_dist*input_settings%memory_factor) THEN
                  n_slices = 1
               END IF
            END IF

         END IF

         desc%group_dim(:) = n_slices(:)
         CALL mp_comm_dup(pw_grid%para%group, desc%group)
         CALL mp_environ(desc%group_size, desc%my_pos, desc%group)

         IF (ALL(n_slices == 1)) THEN
            ! CASE 1 : only one slice: we do not need overlapping regions and special
            !          recombination of the total density
            desc%border = border_size
            IF (border_size == 0) THEN
               desc%perd = 1
            ELSE
               desc%perd = 0
            END IF
            desc%distributed = .FALSE.
            desc%parallel = .TRUE.
            desc%group_head = pw_grid%para%group_head
            desc%group_coor(:) = 0
            desc%my_virtual_pos = 0

            ALLOCATE (desc%virtual2real(0:desc%group_size-1))
            ALLOCATE (desc%real2virtual(0:desc%group_size-1))
            ! Start with no reordering
            DO i = 0, desc%group_size-1
               desc%virtual2real(i) = i
               desc%real2virtual(i) = i
            END DO
         ELSE
            ! CASE 2 : general case
            ! periodicity is no longer enforced arbritary directions
            IF (border_size == 0) THEN
               desc%perd = 1
               DO dir = 1, 3
                  IF (n_slices(dir) > 1) desc%perd(dir) = 0
               END DO
            ELSE
               desc%perd(:) = 0
            END IF
            ! we keep a border of nmin points
            desc%border = nmin
            ! we are going parallel on the real space grid
            desc%parallel = .TRUE.
            desc%distributed = .TRUE.
            desc%group_head = (desc%my_pos == 0)

            ! set up global info about the distribution
            ALLOCATE (desc%rank2coord(3, 0:desc%group_size-1))
            ALLOCATE (desc%coord2rank(0:desc%group_dim(1)-1, 0:desc%group_dim(2)-1, 0:desc%group_dim(3)-1))
            ALLOCATE (desc%lb_global(3, 0:desc%group_size-1))
            ALLOCATE (desc%ub_global(3, 0:desc%group_size-1))
            ALLOCATE (desc%x2coord(desc%lb(1):desc%ub(1)))
            ALLOCATE (desc%y2coord(desc%lb(2):desc%ub(2)))
            ALLOCATE (desc%z2coord(desc%lb(3):desc%ub(3)))

            DO i = 0, desc%group_size-1
               ! Calculate coordinates in a row-major order (to be SMP-friendly)
               desc%rank2coord(1, i) = i/(desc%group_dim(2)*desc%group_dim(3))
               desc%rank2coord(2, i) = MODULO(i, desc%group_dim(2)*desc%group_dim(3)) &
                                       /desc%group_dim(3)
               desc%rank2coord(3, i) = MODULO(i, desc%group_dim(3))

               IF (i == desc%my_pos) THEN
                  desc%group_coor = desc%rank2coord(:, i)
               END IF

               desc%coord2rank(desc%rank2coord(1, i), desc%rank2coord(2, i), desc%rank2coord(3, i)) = i
               ! the lb_global and ub_global correspond to lb_real and ub_real of each task
               desc%lb_global(:, i) = desc%lb
               desc%ub_global(:, i) = desc%ub
               DO dir = 1, 3
                  IF (desc%group_dim(dir) .GT. 1) THEN
                     lb = get_limit(desc%npts(dir), desc%group_dim(dir), desc%rank2coord(dir, i))
                     desc%lb_global(dir, i) = lb(1)+desc%lb(dir)-1
                     desc%ub_global(dir, i) = lb(2)+desc%lb(dir)-1
                  ENDIF
               ENDDO
            ENDDO

            ! map a grid point to a CPU coord
            DO dir = 1, 3
               DO l = 0, desc%group_dim(dir)-1
                  IF (desc%group_dim(dir) .GT. 1) THEN
                     lb = get_limit(desc%npts(dir), desc%group_dim(dir), l)
                     lb = lb+desc%lb(dir)-1
                  ELSE
                     lb(1) = desc%lb(dir)
                     lb(2) = desc%ub(dir)
                  ENDIF
                  SELECT CASE (dir)
                  CASE (1)
                     desc%x2coord(lb(1):lb(2)) = l
                  CASE (2)
                     desc%y2coord(lb(1):lb(2)) = l
                  CASE (3)
                     desc%z2coord(lb(1):lb(2)) = l
                  END SELECT
               ENDDO
            ENDDO

            ! an upper bound for the number of neighbours the border is overlapping with
            DO dir = 1, 3
               desc%neighbours(dir) = 0
               IF ((n_slices(dir) > 1) .OR. (border_size > 0)) THEN
                  neighbours(dir) = HUGE(0)
                  DO l = 0, n_slices(dir)-1
                     lb = get_limit(desc%npts(dir), n_slices(dir), l)
                     neighbours(dir) = MIN(lb(2)-lb(1)+1, neighbours(dir))
                  END DO
                  desc%neighbours(dir) = (desc%border+neighbours(dir)-1)/neighbours(dir)
               END IF
            END DO

            ALLOCATE (desc%virtual2real(0:desc%group_size-1))
            ALLOCATE (desc%real2virtual(0:desc%group_size-1))
            ! Start with no reordering
            DO i = 0, desc%group_size-1
               desc%virtual2real(i) = i
               desc%real2virtual(i) = i
            END DO

            desc%my_virtual_pos = desc%real2virtual(desc%my_pos)
            desc%virtual_group_coor(:) = desc%rank2coord(:, desc%my_virtual_pos)

         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE rs_grid_create_descriptor

! **************************************************************************************************
!> \brief ...
!> \param rs ...
!> \param desc ...
! **************************************************************************************************
   SUBROUTINE rs_grid_create(rs, desc)
      TYPE(realspace_grid_type), POINTER                 :: rs
      TYPE(realspace_grid_desc_type), POINTER            :: desc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rs_grid_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      ALLOCATE (rs)

      last_rs_id = last_rs_id+1
      rs%id_nr = last_rs_id
      rs%ref_count = 1
      rs%desc => desc
      CALL rs_grid_retain_descriptor(rs%desc)

      IF (desc%pw%para%mode == PW_MODE_LOCAL) THEN
         ! The corresponding group has dimension 1
         ! All operations will be done locally
         rs%lb_real = desc%lb
         rs%ub_real = desc%ub
         rs%lb_local = rs%lb_real-desc%border*(1-desc%perd)
         rs%ub_local = rs%ub_real+desc%border*(1-desc%perd)
         rs%npts_local = rs%ub_local-rs%lb_local+1
         rs%ngpts_local = PRODUCT(rs%npts_local)
      END IF

      IF (ALL(rs%desc%group_dim == 1)) THEN
         ! CASE 1 : only one slice: we do not need overlapping regions and special
         !          recombination of the total density
         rs%lb_real = desc%lb
         rs%ub_real = desc%ub
         rs%lb_local = rs%lb_real-desc%border*(1-desc%perd)
         rs%ub_local = rs%ub_real+desc%border*(1-desc%perd)
         rs%npts_local = rs%ub_local-rs%lb_local+1
         rs%ngpts_local = PRODUCT(rs%npts_local)
      ELSE
         ! CASE 2 : general case
         ! extract some more derived quantities about the local grid
         rs%lb_real = desc%lb_global(:, desc%my_virtual_pos)
         rs%ub_real = desc%ub_global(:, desc%my_virtual_pos)
         rs%lb_local = rs%lb_real-desc%border*(1-desc%perd)
         rs%ub_local = rs%ub_real+desc%border*(1-desc%perd)
         rs%npts_local = rs%ub_local-rs%lb_local+1
         rs%ngpts_local = PRODUCT(rs%npts_local)
      END IF

      allocated_rs_grid_count = allocated_rs_grid_count+1

      ALLOCATE (rs%r(rs%lb_local(1):rs%ub_local(1), &
                     rs%lb_local(2):rs%ub_local(2), &
                     rs%lb_local(3):rs%ub_local(3)))
      ALLOCATE (rs%px(desc%npts(1)))
      ALLOCATE (rs%py(desc%npts(2)))
      ALLOCATE (rs%pz(desc%npts(3)))

      CALL timestop(handle)

   END SUBROUTINE rs_grid_create

! **************************************************************************************************
!> \brief Defines a new ordering of ranks on this realspace grid, recalculating
!>        the data bounds and reallocating the grid.  As a result, each MPI process
!>        now has a real rank (i.e. it's rank in the MPI communicator from the pw grid)
!>        and a virtual rank (the rank of the process where the data now owned by this
!>        process would reside in an ordinary cartesian distribution).
!>        NB. Since the grid size required may change, the caller should be sure to release
!>        and recreate the corresponding rs_grids
!>        The desc%real2virtual and desc%virtual2real arrays can be used to map
!>        a physical rank to the 'rank' of data owned by that process and vice versa
!> \param desc ...
!> \param real2virtual ...
!> \par History
!>        04-2009 created [Iain Bethune]
!>          (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
! **************************************************************************************************
   SUBROUTINE rs_grid_reorder_ranks(desc, real2virtual)

      TYPE(realspace_grid_desc_type), POINTER            :: desc
      INTEGER, DIMENSION(:), INTENT(IN)                  :: real2virtual

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rs_grid_reorder_ranks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      desc%real2virtual(:) = real2virtual

      DO i = 0, desc%group_size-1
         desc%virtual2real(desc%real2virtual(i)) = i
      END DO

      desc%my_virtual_pos = desc%real2virtual(desc%my_pos)

      IF (.NOT. ALL(desc%group_dim == 1)) THEN
         desc%virtual_group_coor(:) = desc%rank2coord(:, desc%my_virtual_pos)
      END IF

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Print information on grids to output
!> \param rs ...
!> \param iounit ...
!> \author JGH (17-May-2007)
! **************************************************************************************************
   SUBROUTINE rs_grid_print(rs, iounit)
      TYPE(realspace_grid_type), POINTER                 :: rs
      INTEGER, INTENT(in)                                :: iounit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rs_grid_print', routineP = moduleN//':'//routineN

      INTEGER                                            :: dir, i, nn
      REAL(KIND=dp)                                      :: pp(3)

      IF (rs%desc%parallel) THEN
         IF (iounit > 0) THEN
            WRITE (iounit, '(/,A,T71,I10)') &
               " RS_GRID| Information for grid number ", rs%desc%grid_id
            DO i = 1, 3
               WRITE (iounit, '(A,I3,T30,2I8,T62,A,T71,I10)') " RS_GRID|   Bounds ", &
                  i, rs%desc%lb(i), rs%desc%ub(i), "Points:", rs%desc%npts(i)
            END DO
            IF (.NOT. rs%desc%distributed) THEN
               WRITE (iounit, '(A)') " RS_GRID| Real space fully replicated"
               WRITE (iounit, '(A,T71,I10)') &
                  " RS_GRID| Group size ", rs%desc%group_dim(2)
            ELSE
               DO dir = 1, 3
                  IF (rs%desc%perd(dir) /= 1) THEN
                     WRITE (iounit, '(A,T71,I3,A)') &
                        " RS_GRID| Real space distribution over ", rs%desc%group_dim(dir), " groups"
                     WRITE (iounit, '(A,T71,I10)') &
                        " RS_GRID| Real space distribution along direction ", dir
                     WRITE (iounit, '(A,T71,I10)') &
                        " RS_GRID| Border size ", rs%desc%border
                  END IF
               END DO
            END IF
         END IF
         IF (rs%desc%distributed) THEN
            DO dir = 1, 3
               IF (rs%desc%perd(dir) /= 1) THEN
                  nn = rs%npts_local(dir)
                  CALL mp_sum(nn, rs%desc%group)
                  pp(1) = REAL(nn, KIND=dp)/REAL(PRODUCT(rs%desc%group_dim), KIND=dp)
                  nn = rs%npts_local(dir)
                  CALL mp_max(nn, rs%desc%group)
                  pp(2) = REAL(nn, KIND=dp)
                  nn = rs%npts_local(dir)
                  CALL mp_min(nn, rs%desc%group)
                  pp(3) = REAL(nn, KIND=dp)
                  IF (iounit > 0) THEN
                     WRITE (iounit, '(A,T48,A)') " RS_GRID|   Distribution", &
                        "  Average         Max         Min"
                     WRITE (iounit, '(A,T45,F12.1,2I12)') " RS_GRID|   Planes   ", &
                        pp(1), NINT(pp(2)), NINT(pp(3))
                  END IF
               END IF
            END DO
!          WRITE ( iounit, '(/)' )
         END IF
      ELSE
         IF (iounit > 0) THEN
            WRITE (iounit, '(/,A,T71,I10)') &
               " RS_GRID| Information for grid number ", rs%desc%grid_id
            DO i = 1, 3
               WRITE (iounit, '(A,I3,T30,2I8,T62,A,T71,I10)') " RS_GRID|   Bounds ", &
                  i, rs%desc%lb(i), rs%desc%ub(i), "Points:", rs%desc%npts(i)
            END DO
!         WRITE ( iounit, '(/)' )
         ENDIF
      END IF

   END SUBROUTINE rs_grid_print

! **************************************************************************************************
!> \brief Copy a function from/to a PW grid type to/from a real
!>      space type
!>      dir is the named constant rs2pw or pw2rs
!> \param rs ...
!> \param pw ...
!> \param dir ...
!> \par History
!>      JGH (15-Feb-2003) reduced additional memory usage
!>      Joost VandeVondele (Sep-2003) moved from sum/bcast to shift
!> \author JGH (18-Mar-2001)
! **************************************************************************************************
   SUBROUTINE rs_pw_transfer(rs, pw, dir)

      TYPE(realspace_grid_type), POINTER                 :: rs
      TYPE(pw_type), POINTER                             :: pw
      INTEGER, INTENT(IN)                                :: dir

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_pw_transfer', routineP = moduleN//':'//routineN

      CHARACTER(LEN=5)                                   :: dirname
      INTEGER                                            :: handle, handle2, i, im, j, jm, k, km

      CALL timeset(routineN, handle2)
      IF (dir .EQ. rs2pw) dirname = "RS2PW"
      IF (dir .EQ. pw2rs) dirname = "PW2RS"
      CALL timeset(routineN//"_"//TRIM(dirname)//"_"//TRIM(ADJUSTL( &
                                                           cp_to_string(CEILING(pw%pw_grid%cutoff/10)*10))), handle)

      IF (rs%desc%grid_id /= pw%pw_grid%id_nr) &
         CALL cp__b("pw/realspace_grid_types.F",706,"Different rs and pw indentitfiers")
      IF ((pw%in_use .NE. REALDATA3D) .AND. (pw%in_use .NE. COMPLEXDATA3D)) &
         CALL cp__b("pw/realspace_grid_types.F",708,"Need REALDATA3D or COMPLEXDATA3D")
      IF ((dir .NE. rs2pw) .AND. (dir .NE. pw2rs)) &
         CALL cp__b("pw/realspace_grid_types.F",710,"Direction must be rs2pw or pw2rs")

      IF (rs%desc%distributed) THEN
         CALL rs_pw_transfer_distributed(rs, pw, dir)
      ELSE IF (rs%desc%parallel) THEN
         CALL rs_pw_transfer_replicated(rs, pw, dir)
      ELSE ! treat simple serial case locally
         IF (dir == rs2pw) THEN
            IF (pw%in_use == REALDATA3D) THEN
               IF (rs%desc%border == 0) THEN
                  CALL dcopy(SIZE(rs%r), rs%r, 1, pw%cr3d, 1)
               ELSE
                  IF(.NOT.(LBOUND(pw%cr3d, 3) .EQ. rs%lb_real(3)))CALL cp__a("pw/realspace_grid_types.F",722)
!$OMP                 PARALLEL DO DEFAULT(NONE) SHARED(pw,rs)
                  DO i = rs%lb_real(3), rs%ub_real(3)
                     pw%cr3d(:, :, i) = rs%r(rs%lb_real(1):rs%ub_real(1), &
                                             rs%lb_real(2):rs%ub_real(2), i)
                  END DO
!$OMP                 END PARALLEL DO
               END IF
            ELSEIF (pw%in_use == COMPLEXDATA3D) THEN
               IF (rs%desc%border == 0) THEN
                  CALL copy_rc(rs%r, pw%cc3d)
               ELSE
                  IF(.NOT.(LBOUND(pw%cr3d, 3) .EQ. rs%lb_real(3)))CALL cp__a("pw/realspace_grid_types.F",734)
!$OMP                 PARALLEL DO DEFAULT(NONE) SHARED(pw,rs)
                  DO i = rs%lb_real(3), rs%ub_real(3)
                     pw%cc3d(:, :, i) = CMPLX(rs%r(rs%lb_real(1):rs%ub_real(1), &
                                                   rs%lb_real(2):rs%ub_real(2), i), &
                                              0.0_dp, KIND=dp)
                  END DO
!$OMP                 END PARALLEL DO
               END IF
            ELSE
               CALL cp__b("pw/realspace_grid_types.F",744,"PW type not compatible")
            END IF
         ELSE
            IF (pw%in_use == REALDATA3D) THEN
               IF (rs%desc%border == 0) THEN
                  CALL dcopy(SIZE(rs%r), pw%cr3d, 1, rs%r, 1)
               ELSE
!$OMP                 PARALLEL DO DEFAULT(NONE) &
!$OMP                             PRIVATE(i,im,j,jm,k,km) &
!$OMP                             SHARED(pw,rs)
                  DO k = rs%lb_local(3), rs%ub_local(3)
                     IF (k < rs%lb_real(3)) THEN
                        km = k+rs%desc%npts(3)
                     ELSE IF (k > rs%ub_real(3)) THEN
                        km = k-rs%desc%npts(3)
                     ELSE
                        km = k
                     END IF
                     DO j = rs%lb_local(2), rs%ub_local(2)
                        IF (j < rs%lb_real(2)) THEN
                           jm = j+rs%desc%npts(2)
                        ELSE IF (j > rs%ub_real(2)) THEN
                           jm = j-rs%desc%npts(2)
                        ELSE
                           jm = j
                        END IF
                        DO i = rs%lb_local(1), rs%ub_local(1)
                           IF (i < rs%lb_real(1)) THEN
                              im = i+rs%desc%npts(1)
                           ELSE IF (i > rs%ub_real(1)) THEN
                              im = i-rs%desc%npts(1)
                           ELSE
                              im = i
                           END IF
                           rs%r(i, j, k) = pw%cr3d(im, jm, km)
                        END DO
                     END DO
                  END DO
!$OMP                 END PARALLEL DO
               END IF
            ELSEIF (pw%in_use == COMPLEXDATA3D) THEN
               IF (rs%desc%border == 0) THEN
                  CALL copy_cr(pw%cc3d, rs%r)
               ELSE
!$OMP                 PARALLEL DO DEFAULT(NONE) &
!$OMP                             PRIVATE(i,im,j,jm,k,km) &
!$OMP                             SHARED(pw,rs)
                  DO k = rs%lb_local(3), rs%ub_local(3)
                     IF (k < rs%lb_real(3)) THEN
                        km = k+rs%desc%npts(3)
                     ELSE IF (k > rs%ub_real(3)) THEN
                        km = k-rs%desc%npts(3)
                     ELSE
                        km = k
                     END IF
                     DO j = rs%lb_local(2), rs%ub_local(2)
                        IF (j < rs%lb_real(2)) THEN
                           jm = j+rs%desc%npts(2)
                        ELSE IF (j > rs%ub_real(2)) THEN
                           jm = j-rs%desc%npts(2)
                        ELSE
                           jm = j
                        END IF
                        DO i = rs%lb_local(1), rs%ub_local(1)
                           IF (i < rs%lb_real(1)) THEN
                              im = i+rs%desc%npts(1)
                           ELSE IF (i > rs%ub_real(1)) THEN
                              im = i-rs%desc%npts(1)
                           ELSE
                              im = i
                           END IF
                           rs%r(i, j, k) = REAL(pw%cc3d(im, jm, km), KIND=dp)
                        END DO
                     END DO
                  END DO
!$OMP                 END PARALLEL DO
               END IF
            ELSE
               CALL cp__b("pw/realspace_grid_types.F",822,"PW type not compatible")
            END IF
         END IF
      END IF

      CALL timestop(handle)
      CALL timestop(handle2)

   END SUBROUTINE rs_pw_transfer

! **************************************************************************************************
!> \brief transfer from a realspace grid to a planewave grid or the other way around
!> \param rs ...
!> \param pw ...
!> \param dir == rs2pw or dir == pw2rs
!> \note
!>      rs2pw sums all data on the rs grid into the respective pw grid
!>      pw2rs will scatter all pw data to the rs grids
! **************************************************************************************************
   SUBROUTINE rs_pw_transfer_replicated(rs, pw, dir)
      TYPE(realspace_grid_type), POINTER                 :: rs
      TYPE(pw_type), POINTER                             :: pw
      INTEGER, INTENT(IN)                                :: dir

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_pw_transfer_replicated', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dest, group, i, ii, im, ip, ix, iy, iz, &
                                                            j, jm, k, km, mepos, nma, nn, np, &
                                                            req(2), s(3), source
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rcount
      INTEGER, DIMENSION(3)                              :: lb, ub
      INTEGER, DIMENSION(:, :), POINTER                  :: pbo
      INTEGER, DIMENSION(:, :, :), POINTER               :: bo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: recvbuf, sendbuf, swapptr
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid

      np = pw%pw_grid%para%group_size
      bo => pw%pw_grid%para%bo(1:2, 1:3, 0:np-1, 1)
      pbo => pw%pw_grid%bounds
      group = pw%pw_grid%para%rs_group
      mepos = pw%pw_grid%para%rs_mpo
      ALLOCATE (rcount(0:np-1))
      DO ip = 1, np
         rcount(ip-1) = PRODUCT(bo(2, :, ip)-bo(1, :, ip)+1)
      END DO
      nma = MAXVAL(rcount(0:np-1))
      ALLOCATE (sendbuf(nma), recvbuf(nma))
      sendbuf = 1.0E99_dp; recvbuf = 1.0E99_dp ! init mpi'ed buffers to silence warnings under valgrind
      grid => rs%r

      !sample peak memory
      CALL m_memory()

      IF (dir == rs2pw) THEN
         dest = MODULO(mepos+1, np)
         source = MODULO(mepos-1, np)
         sendbuf = 0.0_dp

         DO ip = 1, np

            lb = pbo(1, :)+bo(1, :, MODULO(mepos-ip, np)+1)-1
            ub = pbo(1, :)+bo(2, :, MODULO(mepos-ip, np)+1)-1
            ! this loop takes about the same time as the message passing call
            ! notice that the range of ix is only a small fraction of the first index of grid
            ! therefore it seems faster to have the second index as the innermost loop
            ! if this runs on many cpus
            ! tested on itanium, pentium4, opteron, ultrasparc...
            s = ub-lb+1
            DO iz = lb(3), ub(3)
               DO ix = lb(1), ub(1)
                  ii = (iz-lb(3))*s(1)*s(2)+(ix-lb(1))+1
                  DO iy = lb(2), ub(2)
                     sendbuf(ii) = sendbuf(ii)+grid(ix, iy, iz)
                     ii = ii+s(1)
                  END DO
               END DO
            END DO
            IF (ip .EQ. np) EXIT
            CALL mp_isendrecv(sendbuf, dest, recvbuf, source, &
                              group, req(1), req(2), 13)
            CALL mp_waitall(req)
            swapptr => sendbuf
            sendbuf => recvbuf
            recvbuf => swapptr
         ENDDO
         nn = rcount(mepos)
         IF (pw%in_use == REALDATA3D) THEN
            CALL dcopy(nn, sendbuf, 1, pw%cr3d, 1)
         ELSEIF (pw%in_use == COMPLEXDATA3D) THEN
            CALL copy_rc(RESHAPE(sendbuf, SHAPE(pw%cc3d)), pw%cc3d)
         ELSE
            CALL cp__b("pw/realspace_grid_types.F",914,"PW type not compatible")
         END IF
      ELSE
         nn = rcount(mepos)
         IF (pw%in_use == REALDATA3D) THEN
            CALL dcopy(nn, pw%cr3d, 1, sendbuf, 1)
         ELSEIF (pw%in_use == COMPLEXDATA3D) THEN
            CALL dcopy(nn, pw%cc3d, 2, sendbuf, 1)
         ELSE
            CALL cp__b("pw/realspace_grid_types.F",923,"PW type not compatible")
         END IF

         dest = MODULO(mepos+1, np)
         source = MODULO(mepos-1, np)

         DO ip = 0, np-1
            ! we must shift the buffer only np-1 times around
            IF (ip .NE. np-1) THEN
               CALL mp_isendrecv(sendbuf, dest, recvbuf, source, &
                                 group, req(1), req(2), 13)
            ENDIF
            lb = pbo(1, :)+bo(1, :, MODULO(mepos-ip, np)+1)-1
            ub = pbo(1, :)+bo(2, :, MODULO(mepos-ip, np)+1)-1
            ii = 0
            ! this loop takes about the same time as the message passing call
            ! If I read the code correctly then:
            ! BUG/FIXME: Unfortunately MPI standard says we are not allowed
            ! to access the sendbuf/recvbuf while the sendrecv is still not
            ! finished. This EXPLICITLY includes READING sendbuf!
            ! I don't know any MPI implementation that really requires this
            ! but the standard explicitly mandates it.
            DO iz = lb(3), ub(3)
               DO iy = lb(2), ub(2)
                  DO ix = lb(1), ub(1)
                     ii = ii+1
                     grid(ix, iy, iz) = sendbuf(ii)
                  END DO
               END DO
            END DO
            IF (ip .NE. np-1) THEN
               CALL mp_waitall(req)
            ENDIF
            swapptr => sendbuf
            sendbuf => recvbuf
            recvbuf => swapptr
         END DO
         IF (rs%desc%border > 0) THEN
!$OMP           PARALLEL DO DEFAULT(NONE) &
!$OMP                       PRIVATE(i,im,j,jm,k,km) &
!$OMP                       SHARED(rs)
            DO k = rs%lb_local(3), rs%ub_local(3)
               IF (k < rs%lb_real(3)) THEN
                  km = k+rs%desc%npts(3)
               ELSE IF (k > rs%ub_real(3)) THEN
                  km = k-rs%desc%npts(3)
               ELSE
                  km = k
               END IF
               DO j = rs%lb_local(2), rs%ub_local(2)
                  IF (j < rs%lb_real(2)) THEN
                     jm = j+rs%desc%npts(2)
                  ELSE IF (j > rs%ub_real(2)) THEN
                     jm = j-rs%desc%npts(2)
                  ELSE
                     jm = j
                  END IF
                  DO i = rs%lb_local(1), rs%ub_local(1)
                     IF (i < rs%lb_real(1)) THEN
                        im = i+rs%desc%npts(1)
                     ELSE IF (i > rs%ub_real(1)) THEN
                        im = i-rs%desc%npts(1)
                     ELSE
                        im = i
                     END IF
                     rs%r(i, j, k) = rs%r(im, jm, km)
                  END DO
               END DO
            END DO
!$OMP           END PARALLEL DO
         END IF
      END IF

      DEALLOCATE (rcount)
      DEALLOCATE (sendbuf)
      DEALLOCATE (recvbuf)

   END SUBROUTINE rs_pw_transfer_replicated

! **************************************************************************************************
!> \brief does the rs2pw and pw2rs transfer in the case where the rs grid is
!>       distributed (3D domain decomposition)
!> \param rs ...
!> \param pw ...
!> \param dir ...
!> \par History
!>      12.2007 created [Matt Watkins]
!>      9.2008 reduced amount of halo data sent [Iain Bethune]
!>      10.2008 added non-blocking communication [Iain Bethune]
!>      4.2009 added support for rank-reordering on the grid [Iain Bethune]
!>      12.2009 added OMP and sparse alltoall [Iain Bethune]
!>              (c) The Numerical Algorithms Group (NAG) Ltd, 2008-2009 on behalf of the HECToR project
!> \note
!>       the transfer is a two step procedure. For example, for the rs2pw transfer:
!>
!>       1) Halo-exchange in 3D so that the local part of the rs_grid contains the full data
!>       2) an alltoall communication to redistribute the local rs_grid to the local pw_grid
!>
!>       the halo exchange is most expensive on a large number of CPUs. Particular in this halo
!>       exchange is that the border region is rather large (e.g. 20 points) and that it might overlap
!>       with the central domain of several CPUs (i.e. next nearest neighbors)
! **************************************************************************************************
   SUBROUTINE rs_pw_transfer_distributed(rs, pw, dir)
      TYPE(realspace_grid_type), POINTER                 :: rs
      TYPE(pw_type), POINTER                             :: pw
      INTEGER, INTENT(IN)                                :: dir

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_pw_transfer_distributed', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=200)                                 :: error_string
      INTEGER :: completed, dest_down, dest_up, i, idir, j, k, lb, my_id, my_pw_rank, my_rs_rank, &
         n_shifts, nn, num_threads, position, source_down, source_up, ub, x, y, z
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: dshifts, recv_disps, recv_reqs, &
                                                            recv_sizes, send_disps, send_reqs, &
                                                            send_sizes, ushifts
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: bounds, recv_tasks, send_tasks
      INTEGER, DIMENSION(2)                              :: neighbours, pos
      INTEGER, DIMENSION(3) :: coords, lb_recv, lb_recv_down, lb_recv_up, lb_send, lb_send_down, &
         lb_send_up, ub_recv, ub_recv_down, ub_recv_up, ub_send, ub_send_down, ub_send_up
      INTEGER, DIMENSION(4)                              :: req
      LOGICAL, DIMENSION(3)                              :: halo_swapped
      REAL(KIND=dp)                                      :: pw_sum, rs_sum
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: recv_buf_3d_down, recv_buf_3d_up, &
                                                            send_buf_3d_down, send_buf_3d_up
      TYPE(cp_1d_r_p_type), ALLOCATABLE, DIMENSION(:)    :: recv_bufs, send_bufs

      num_threads = 1
      my_id = 0

      IF (dir == rs2pw) THEN

         ! safety check, to be removed once we're absolute sure the routine is correct
         IF (debug_this_module) THEN
            rs_sum = accurate_sum(rs%r)*ABS(det_3x3(rs%desc%dh))
            CALL mp_sum(rs_sum, rs%desc%group)
         ENDIF

         halo_swapped = .FALSE.
         ! We don't need to send the 'edges' of the halos that have already been sent
         ! Halos are contiguous in memory in z-direction only, so swap these first,
         ! and send less data in the y and x directions which are more expensive

         DO idir = 3, 1, -1

            IF (rs%desc%perd(idir) .NE. 1) THEN

               ALLOCATE (dshifts(0:rs%desc%neighbours(idir)))
               ALLOCATE (ushifts(0:rs%desc%neighbours(idir)))

               ushifts = 0
               dshifts = 0

               ! check that we don't try to send data to ourself
               DO n_shifts = 1, MIN(rs%desc%neighbours(idir), rs%desc%group_dim(idir)-1)

                  ! need to take into account the possible varying widths of neighbouring cells
                  ! offset_up and offset_down hold the real size of the neighbouring cells
                  position = MODULO(rs%desc%virtual_group_coor(idir)-n_shifts, rs%desc%group_dim(idir))
                  neighbours = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), position)
                  dshifts(n_shifts) = dshifts(n_shifts-1)+(neighbours(2)-neighbours(1)+1)

                  position = MODULO(rs%desc%virtual_group_coor(idir)+n_shifts, rs%desc%group_dim(idir))
                  neighbours = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), position)
                  ushifts(n_shifts) = ushifts(n_shifts-1)+(neighbours(2)-neighbours(1)+1)

                  ! The border data has to be send/received from the neighbours
                  ! First we calculate the source and destination processes for the shift
                  ! We do both shifts at once to allow for more overlap of communication and buffer packing/unpacking

                  CALL cart_shift(rs, idir, -1*n_shifts, source_down, dest_down)

                  lb_send_down(:) = rs%lb_local(:)
                  lb_recv_down(:) = rs%lb_local(:)
                  ub_recv_down(:) = rs%ub_local(:)
                  ub_send_down(:) = rs%ub_local(:)

                  IF (dshifts(n_shifts-1) .LE. rs%desc%border) THEN
                     ub_send_down(idir) = lb_send_down(idir)+rs%desc%border-1-dshifts(n_shifts-1)
                     lb_send_down(idir) = MAX(lb_send_down(idir), &
                                              lb_send_down(idir)+rs%desc%border-dshifts(n_shifts))

                     ub_recv_down(idir) = ub_recv_down(idir)-rs%desc%border
                     lb_recv_down(idir) = MAX(lb_recv_down(idir)+rs%desc%border, &
                                              ub_recv_down(idir)-rs%desc%border+1+ushifts(n_shifts-1))
                  ELSE
                     lb_send_down(idir) = 0
                     ub_send_down(idir) = -1
                     lb_recv_down(idir) = 0
                     ub_recv_down(idir) = -1
                  ENDIF

                  DO i = 1, 3
                     IF (halo_swapped(i)) THEN
                        lb_send_down(i) = rs%lb_real(i)
                        ub_send_down(i) = rs%ub_real(i)
                        lb_recv_down(i) = rs%lb_real(i)
                        ub_recv_down(i) = rs%ub_real(i)
                     ENDIF
                  ENDDO

                  ! post the recieve
                  ALLOCATE (recv_buf_3d_down(lb_recv_down(1):ub_recv_down(1), &
                                             lb_recv_down(2):ub_recv_down(2), lb_recv_down(3):ub_recv_down(3)))
                  CALL mp_irecv(recv_buf_3d_down, source_down, rs%desc%group, req(1))

                  ! now allocate, pack and send the send buffer
                  nn = PRODUCT(ub_send_down-lb_send_down+1)
                  ALLOCATE (send_buf_3d_down(lb_send_down(1):ub_send_down(1), &
                                             lb_send_down(2):ub_send_down(2), lb_send_down(3):ub_send_down(3)))

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(send_buf_3d_down,rs,lb_send_down,ub_send_down)
!$                num_threads = MIN(omp_get_max_threads(), ub_send_down(3)-lb_send_down(3)+1)
!$                my_id = omp_get_thread_num()
                  IF (my_id < num_threads) THEN
                     lb = lb_send_down(3)+((ub_send_down(3)-lb_send_down(3)+1)*my_id)/num_threads
                     ub = lb_send_down(3)+((ub_send_down(3)-lb_send_down(3)+1)*(my_id+1))/num_threads-1

                     send_buf_3d_down(lb_send_down(1):ub_send_down(1), lb_send_down(2):ub_send_down(2), &
                                      lb:ub) = rs%r(lb_send_down(1):ub_send_down(1), &
                                                    lb_send_down(2):ub_send_down(2), lb:ub)
                  END IF
!$OMP END PARALLEL

                  CALL mp_isend(send_buf_3d_down, dest_down, rs%desc%group, req(3))

                  ! Now for the other direction
                  CALL cart_shift(rs, idir, n_shifts, source_up, dest_up)

                  lb_send_up(:) = rs%lb_local(:)
                  lb_recv_up(:) = rs%lb_local(:)
                  ub_recv_up(:) = rs%ub_local(:)
                  ub_send_up(:) = rs%ub_local(:)

                  IF (ushifts(n_shifts-1) .LE. rs%desc%border) THEN

                     lb_send_up(idir) = ub_send_up(idir)-rs%desc%border+1+ushifts(n_shifts-1)
                     ub_send_up(idir) = MIN(ub_send_up(idir), &
                                            ub_send_up(idir)-rs%desc%border+ushifts(n_shifts))

                     lb_recv_up(idir) = lb_recv_up(idir)+rs%desc%border
                     ub_recv_up(idir) = MIN(ub_recv_up(idir)-rs%desc%border, &
                                            lb_recv_up(idir)+rs%desc%border-1-dshifts(n_shifts-1))
                  ELSE
                     lb_send_up(idir) = 0
                     ub_send_up(idir) = -1
                     lb_recv_up(idir) = 0
                     ub_recv_up(idir) = -1
                  ENDIF

                  DO i = 1, 3
                     IF (halo_swapped(i)) THEN
                        lb_send_up(i) = rs%lb_real(i)
                        ub_send_up(i) = rs%ub_real(i)
                        lb_recv_up(i) = rs%lb_real(i)
                        ub_recv_up(i) = rs%ub_real(i)
                     ENDIF
                  ENDDO

                  ! post the recieve
                  ALLOCATE (recv_buf_3d_up(lb_recv_up(1):ub_recv_up(1), &
                                           lb_recv_up(2):ub_recv_up(2), lb_recv_up(3):ub_recv_up(3)))
                  CALL mp_irecv(recv_buf_3d_up, source_up, rs%desc%group, req(2))

                  ! now allocate,pack and send the send buffer
                  nn = PRODUCT(ub_send_up-lb_send_up+1)
                  ALLOCATE (send_buf_3d_up(lb_send_up(1):ub_send_up(1), &
                                           lb_send_up(2):ub_send_up(2), lb_send_up(3):ub_send_up(3)))

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(send_buf_3d_up,rs,lb_send_up,ub_send_up)
!$                num_threads = MIN(omp_get_max_threads(), ub_send_up(3)-lb_send_up(3)+1)
!$                my_id = omp_get_thread_num()
                  IF (my_id < num_threads) THEN
                     lb = lb_send_up(3)+((ub_send_up(3)-lb_send_up(3)+1)*my_id)/num_threads
                     ub = lb_send_up(3)+((ub_send_up(3)-lb_send_up(3)+1)*(my_id+1))/num_threads-1

                     send_buf_3d_up(lb_send_up(1):ub_send_up(1), lb_send_up(2):ub_send_up(2), &
                                    lb:ub) = rs%r(lb_send_up(1):ub_send_up(1), &
                                                  lb_send_up(2):ub_send_up(2), lb:ub)
                  END IF
!$OMP END PARALLEL

                  CALL mp_isend(send_buf_3d_up, dest_up, rs%desc%group, req(4))

                  ! wait for a recv to complete, then we can unpack

                  DO i = 1, 2

                     CALL mp_waitany(req(1:2), completed)

                     IF (completed .EQ. 1) THEN

                        ! only some procs may need later shifts
                        IF (ub_recv_down(idir) .GE. lb_recv_down(idir)) THEN
                           ! Sum the data in the RS Grid
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(recv_buf_3d_down,rs,lb_recv_down,ub_recv_down)
!$                         num_threads = MIN(omp_get_max_threads(), ub_recv_down(3)-lb_recv_down(3)+1)
!$                         my_id = omp_get_thread_num()
                           IF (my_id < num_threads) THEN
                              lb = lb_recv_down(3)+((ub_recv_down(3)-lb_recv_down(3)+1)*my_id)/num_threads
                              ub = lb_recv_down(3)+((ub_recv_down(3)-lb_recv_down(3)+1)*(my_id+1))/num_threads-1

                              rs%r(lb_recv_down(1):ub_recv_down(1), &
                                   lb_recv_down(2):ub_recv_down(2), lb:ub) = &
                                 rs%r(lb_recv_down(1):ub_recv_down(1), &
                                      lb_recv_down(2):ub_recv_down(2), lb:ub)+ &
                                 recv_buf_3d_down(:, :, lb:ub)
                           END IF
!$OMP END PARALLEL
                        END IF
                        DEALLOCATE (recv_buf_3d_down)
                     ELSE

                        ! only some procs may need later shifts
                        IF (ub_recv_up(idir) .GE. lb_recv_up(idir)) THEN
                           ! Sum the data in the RS Grid
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(recv_buf_3d_up,rs,lb_recv_up,ub_recv_up)
!$                         num_threads = MIN(omp_get_max_threads(), ub_recv_up(3)-lb_recv_up(3)+1)
!$                         my_id = omp_get_thread_num()
                           IF (my_id < num_threads) THEN
                              lb = lb_recv_up(3)+((ub_recv_up(3)-lb_recv_up(3)+1)*my_id)/num_threads
                              ub = lb_recv_up(3)+((ub_recv_up(3)-lb_recv_up(3)+1)*(my_id+1))/num_threads-1

                              rs%r(lb_recv_up(1):ub_recv_up(1), &
                                   lb_recv_up(2):ub_recv_up(2), lb:ub) = &
                                 rs%r(lb_recv_up(1):ub_recv_up(1), &
                                      lb_recv_up(2):ub_recv_up(2), lb:ub)+ &
                                 recv_buf_3d_up(:, :, lb:ub)
                           END IF
!$OMP END PARALLEL
                        END IF
                        DEALLOCATE (recv_buf_3d_up)
                     END IF

                  END DO

                  ! make sure the sends have completed before we deallocate

                  CALL mp_waitall(req(3:4))

                  DEALLOCATE (send_buf_3d_down)
                  DEALLOCATE (send_buf_3d_up)
               END DO

               DEALLOCATE (dshifts)
               DEALLOCATE (ushifts)

            END IF

            halo_swapped(idir) = .TRUE.

         END DO

         ! This is the real redistribution
         ALLOCATE (bounds(0:pw%pw_grid%para%group_size-1, 1:4))

         ! work out the pw grid points each proc holds
         DO i = 0, pw%pw_grid%para%group_size-1
            bounds(i, 1:2) = pw%pw_grid%para%bo(1:2, 1, i, 1)
            bounds(i, 3:4) = pw%pw_grid%para%bo(1:2, 2, i, 1)
            bounds(i, 1:2) = bounds(i, 1:2)-pw%pw_grid%npts(1)/2-1
            bounds(i, 3:4) = bounds(i, 3:4)-pw%pw_grid%npts(2)/2-1
         ENDDO

         ALLOCATE (send_tasks(0:pw%pw_grid%para%group_size-1, 1:6))
         ALLOCATE (send_sizes(0:pw%pw_grid%para%group_size-1))
         ALLOCATE (send_disps(0:pw%pw_grid%para%group_size-1))
         ALLOCATE (recv_tasks(0:pw%pw_grid%para%group_size-1, 1:6))
         ALLOCATE (recv_sizes(0:pw%pw_grid%para%group_size-1))
         ALLOCATE (recv_disps(0:pw%pw_grid%para%group_size-1))
         send_tasks(:, 1) = 1
         send_tasks(:, 2) = 0
         send_tasks(:, 3) = 1
         send_tasks(:, 4) = 0
         send_tasks(:, 5) = 1
         send_tasks(:, 6) = 0
         send_sizes = 0
         recv_sizes = 0

         my_rs_rank = rs%desc%my_pos
         my_pw_rank = pw%pw_grid%para%rs_mpo

         ! find the processors that should hold our data
         ! should be part of the rs grid type
         ! this is a loop over real ranks (i.e. the in-order cartesian ranks)
         ! do the recv and send tasks in two seperate loops which will
         ! load balance better for OpenMP with large numbers of MPI tasks

!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(coords,idir,pos,lb_send,ub_send), &
!$OMP             SHARED(rs,bounds,my_rs_rank,recv_tasks,recv_sizes)
         DO i = 0, rs%desc%group_size-1

            coords(:) = rs%desc%rank2coord(:, rs%desc%real2virtual(i))
            !calculate the rs grid points on each processor
            !coords is the part of the grid that rank i actually holds
            DO idir = 1, 3
               pos(:) = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), coords(idir))
               pos(:) = pos(:)-rs%desc%npts(idir)/2-1
               lb_send(idir) = pos(1)
               ub_send(idir) = pos(2)
            ENDDO

            IF (lb_send(1) .GT. bounds(my_rs_rank, 2)) CYCLE
            IF (ub_send(1) .LT. bounds(my_rs_rank, 1)) CYCLE
            IF (lb_send(2) .GT. bounds(my_rs_rank, 4)) CYCLE
            IF (ub_send(2) .LT. bounds(my_rs_rank, 3)) CYCLE

            recv_tasks(i, 1) = MAX(lb_send(1), bounds(my_rs_rank, 1))
            recv_tasks(i, 2) = MIN(ub_send(1), bounds(my_rs_rank, 2))
            recv_tasks(i, 3) = MAX(lb_send(2), bounds(my_rs_rank, 3))
            recv_tasks(i, 4) = MIN(ub_send(2), bounds(my_rs_rank, 4))
            recv_tasks(i, 5) = lb_send(3)
            recv_tasks(i, 6) = ub_send(3)
            recv_sizes(i) = (recv_tasks(i, 2)-recv_tasks(i, 1)+1)* &
                            (recv_tasks(i, 4)-recv_tasks(i, 3)+1)*(recv_tasks(i, 6)-recv_tasks(i, 5)+1)

         ENDDO
!$OMP END PARALLEL DO

         coords(:) = rs%desc%rank2coord(:, rs%desc%real2virtual(my_rs_rank))
         DO idir = 1, 3
            pos(:) = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), coords(idir))
            pos(:) = pos(:)-rs%desc%npts(idir)/2-1
            lb_send(idir) = pos(1)
            ub_send(idir) = pos(2)
         ENDDO

         lb_recv(:) = lb_send(:)
         ub_recv(:) = ub_send(:)
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             SHARED(pw,lb_send,ub_send,bounds,send_tasks,send_sizes)
         DO j = 0, pw%pw_grid%para%group_size-1

            IF (lb_send(1) .GT. bounds(j, 2)) CYCLE
            IF (ub_send(1) .LT. bounds(j, 1)) CYCLE
            IF (lb_send(2) .GT. bounds(j, 4)) CYCLE
            IF (ub_send(2) .LT. bounds(j, 3)) CYCLE

            send_tasks(j, 1) = MAX(lb_send(1), bounds(j, 1))
            send_tasks(j, 2) = MIN(ub_send(1), bounds(j, 2))
            send_tasks(j, 3) = MAX(lb_send(2), bounds(j, 3))
            send_tasks(j, 4) = MIN(ub_send(2), bounds(j, 4))
            send_tasks(j, 5) = lb_send(3)
            send_tasks(j, 6) = ub_send(3)
            send_sizes(j) = (send_tasks(j, 2)-send_tasks(j, 1)+1)* &
                            (send_tasks(j, 4)-send_tasks(j, 3)+1)*(send_tasks(j, 6)-send_tasks(j, 5)+1)

         END DO
!$OMP END PARALLEL DO

         send_disps(0) = 0
         recv_disps(0) = 0
         DO i = 1, pw%pw_grid%para%group_size-1
            send_disps(i) = send_disps(i-1)+send_sizes(i-1)
            recv_disps(i) = recv_disps(i-1)+recv_sizes(i-1)
         ENDDO

         IF(.NOT.(SUM(send_sizes) == PRODUCT(ub_recv-lb_recv+1)))CALL cp__a("pw/realspace_grid_types.F",1389)

         ALLOCATE (send_bufs(0:rs%desc%group_size-1))
         ALLOCATE (recv_bufs(0:rs%desc%group_size-1))

         DO i = 0, rs%desc%group_size-1
            IF (send_sizes(i) .NE. 0) THEN
               ALLOCATE (send_bufs(i)%array(send_sizes(i)))
            ELSE
               NULLIFY (send_bufs(i)%array)
            END IF
            IF (recv_sizes(i) .NE. 0) THEN
               ALLOCATE (recv_bufs(i)%array(recv_sizes(i)))
            ELSE
               NULLIFY (recv_bufs(i)%array)
            END IF
         END DO

         ALLOCATE (recv_reqs(0:rs%desc%group_size-1))
         recv_reqs = mp_request_null

         DO i = 0, rs%desc%group_size-1
            IF (recv_sizes(i) .NE. 0) THEN
               CALL mp_irecv(recv_bufs(i)%array, i, rs%desc%group, recv_reqs(i))
            END IF
         END DO

         ! do packing
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(k,z,y,x), &
!$OMP             SHARED(rs,send_tasks,send_bufs,send_disps)
         DO i = 0, rs%desc%group_size-1
            k = 0
            DO z = send_tasks(i, 5), send_tasks(i, 6)
               DO y = send_tasks(i, 3), send_tasks(i, 4)
                  DO x = send_tasks(i, 1), send_tasks(i, 2)
                     k = k+1
                     send_bufs(i)%array(k) = rs%r(x, y, z)
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
!$OMP END PARALLEL DO

         ALLOCATE (send_reqs(0:rs%desc%group_size-1))
         send_reqs = mp_request_null

         DO i = 0, rs%desc%group_size-1
            IF (send_sizes(i) .NE. 0) THEN
               CALL mp_isend(send_bufs(i)%array, i, rs%desc%group, send_reqs(i))
            END IF
         END DO

         ! do unpacking
         ! no OMP here so we can unpack each message as it arrives
         DO i = 0, rs%desc%group_size-1
            IF (recv_sizes(i) .EQ. 0) CYCLE

            CALL mp_waitany(recv_reqs, completed)
            k = 0
            DO z = recv_tasks(completed-1, 5), recv_tasks(completed-1, 6)
               DO y = recv_tasks(completed-1, 3), recv_tasks(completed-1, 4)
                  DO x = recv_tasks(completed-1, 1), recv_tasks(completed-1, 2)
                     k = k+1
                     pw%cr3d(x, y, z) = recv_bufs(completed-1)%array(k)
                  ENDDO
               ENDDO
            ENDDO
         ENDDO

         CALL mp_waitall(send_reqs)

         DEALLOCATE (recv_reqs)
         DEALLOCATE (send_reqs)

         DO i = 0, rs%desc%group_size-1
            IF (ASSOCIATED(send_bufs(i)%array)) THEN
               DEALLOCATE (send_bufs(i)%array)
            END IF
            IF (ASSOCIATED(recv_bufs(i)%array)) THEN
               DEALLOCATE (recv_bufs(i)%array)
            END IF
         END DO

         DEALLOCATE (send_bufs)
         DEALLOCATE (recv_bufs)
         DEALLOCATE (send_tasks)
         DEALLOCATE (send_sizes)
         DEALLOCATE (send_disps)
         DEALLOCATE (recv_tasks)
         DEALLOCATE (recv_sizes)
         DEALLOCATE (recv_disps)

         IF (debug_this_module) THEN
            ! safety check, to be removed once we're absolute sure the routine is correct
            pw_sum = pw_integrate_function(pw)
            IF (ABS(pw_sum-rs_sum)/MAX(1.0_dp, ABS(pw_sum), ABS(rs_sum)) > EPSILON(rs_sum)*1000) THEN
               WRITE (error_string, '(A,6(1X,I4.4),3F25.16)') "rs_pw_transfer_distributed", &
                  rs%desc%npts, rs%desc%group_dim, pw_sum, rs_sum, ABS(pw_sum-rs_sum)
               CALL cp_abort(cp__l("pw/realspace_grid_types.F",1488), &
                             error_string//" Please report this bug ... quick workaround: use "// &
                             "DISTRIBUTION_TYPE REPLICATED")
            ENDIF
         ENDIF

      ELSE

         ! pw to rs transfer

         CALL rs_grid_zero(rs)

         ! This is the real redistribution

         ALLOCATE (bounds(0:pw%pw_grid%para%group_size-1, 1:4))

         DO i = 0, pw%pw_grid%para%group_size-1
            bounds(i, 1:2) = pw%pw_grid%para%bo(1:2, 1, i, 1)
            bounds(i, 3:4) = pw%pw_grid%para%bo(1:2, 2, i, 1)
            bounds(i, 1:2) = bounds(i, 1:2)-pw%pw_grid%npts(1)/2-1
            bounds(i, 3:4) = bounds(i, 3:4)-pw%pw_grid%npts(2)/2-1
         ENDDO

         ALLOCATE (send_tasks(0:pw%pw_grid%para%group_size-1, 1:6))
         ALLOCATE (send_sizes(0:pw%pw_grid%para%group_size-1))
         ALLOCATE (send_disps(0:pw%pw_grid%para%group_size-1))
         ALLOCATE (recv_tasks(0:pw%pw_grid%para%group_size-1, 1:6))
         ALLOCATE (recv_sizes(0:pw%pw_grid%para%group_size-1))
         ALLOCATE (recv_disps(0:pw%pw_grid%para%group_size-1))

         send_tasks = 0
         send_tasks(:, 1) = 1
         send_tasks(:, 2) = 0
         send_tasks(:, 3) = 1
         send_tasks(:, 4) = 0
         send_tasks(:, 5) = 1
         send_tasks(:, 6) = 0
         send_sizes = 0

         recv_tasks = 0
         recv_tasks(:, 1) = 1
         recv_tasks(:, 2) = 0
         send_tasks(:, 3) = 1
         send_tasks(:, 4) = 0
         send_tasks(:, 5) = 1
         send_tasks(:, 6) = 0
         recv_sizes = 0

         my_rs_rank = rs%desc%my_pos
         my_pw_rank = pw%pw_grid%para%rs_mpo

         ! find the processors that should hold our data
         ! should be part of the rs grid type
         ! this is a loop over real ranks (i.e. the in-order cartesian ranks)
         ! do the recv and send tasks in two seperate loops which will
         ! load balance better for OpenMP with large numbers of MPI tasks

         ! this is the reverse of rs2pw: what were the sends are now the recvs

!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(coords,idir,pos,lb_send,ub_send), &
!$OMP             SHARED(rs,bounds,my_rs_rank,send_tasks,send_sizes,pw)
         DO i = 0, pw%pw_grid%para%group_size-1

            coords(:) = rs%desc%rank2coord(:, rs%desc%real2virtual(i))
            !calculate the real rs grid points on each processor
            !coords is the part of the grid that rank i actually holds
            DO idir = 1, 3
               pos(:) = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), coords(idir))
               pos(:) = pos(:)-rs%desc%npts(idir)/2-1
               lb_send(idir) = pos(1)
               ub_send(idir) = pos(2)
            ENDDO

            IF (ub_send(1) .LT. bounds(my_rs_rank, 1)) CYCLE
            IF (lb_send(1) .GT. bounds(my_rs_rank, 2)) CYCLE
            IF (ub_send(2) .LT. bounds(my_rs_rank, 3)) CYCLE
            IF (lb_send(2) .GT. bounds(my_rs_rank, 4)) CYCLE

            send_tasks(i, 1) = MAX(lb_send(1), bounds(my_rs_rank, 1))
            send_tasks(i, 2) = MIN(ub_send(1), bounds(my_rs_rank, 2))
            send_tasks(i, 3) = MAX(lb_send(2), bounds(my_rs_rank, 3))
            send_tasks(i, 4) = MIN(ub_send(2), bounds(my_rs_rank, 4))
            send_tasks(i, 5) = lb_send(3)
            send_tasks(i, 6) = ub_send(3)
            send_sizes(i) = (send_tasks(i, 2)-send_tasks(i, 1)+1)* &
                            (send_tasks(i, 4)-send_tasks(i, 3)+1)*(send_tasks(i, 6)-send_tasks(i, 5)+1)

         ENDDO
!$OMP END PARALLEL DO

         coords(:) = rs%desc%rank2coord(:, rs%desc%real2virtual(my_rs_rank))
         DO idir = 1, 3
            pos(:) = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), coords(idir))
            pos(:) = pos(:)-rs%desc%npts(idir)/2-1
            lb_send(idir) = pos(1)
            ub_send(idir) = pos(2)
         ENDDO

         lb_recv(:) = lb_send(:)
         ub_recv(:) = ub_send(:)

!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             SHARED(pw,lb_send,ub_send,bounds,recv_tasks,recv_sizes)
         DO j = 0, pw%pw_grid%para%group_size-1

            IF (ub_send(1) .LT. bounds(j, 1)) CYCLE
            IF (lb_send(1) .GT. bounds(j, 2)) CYCLE
            IF (ub_send(2) .LT. bounds(j, 3)) CYCLE
            IF (lb_send(2) .GT. bounds(j, 4)) CYCLE

            recv_tasks(j, 1) = MAX(lb_send(1), bounds(j, 1))
            recv_tasks(j, 2) = MIN(ub_send(1), bounds(j, 2))
            recv_tasks(j, 3) = MAX(lb_send(2), bounds(j, 3))
            recv_tasks(j, 4) = MIN(ub_send(2), bounds(j, 4))
            recv_tasks(j, 5) = lb_send(3)
            recv_tasks(j, 6) = ub_send(3)
            recv_sizes(j) = (recv_tasks(j, 2)-recv_tasks(j, 1)+1)* &
                            (recv_tasks(j, 4)-recv_tasks(j, 3)+1)*(recv_tasks(j, 6)-recv_tasks(j, 5)+1)

         ENDDO
!$OMP END PARALLEL DO

         send_disps(0) = 0
         recv_disps(0) = 0
         DO i = 1, pw%pw_grid%para%group_size-1
            send_disps(i) = send_disps(i-1)+send_sizes(i-1)
            recv_disps(i) = recv_disps(i-1)+recv_sizes(i-1)
         ENDDO

         IF(.NOT.(SUM(recv_sizes) == PRODUCT(ub_recv-lb_recv+1)))CALL cp__a("pw/realspace_grid_types.F",1618)

         ALLOCATE (send_bufs(0:rs%desc%group_size-1))
         ALLOCATE (recv_bufs(0:rs%desc%group_size-1))

         DO i = 0, rs%desc%group_size-1
            IF (send_sizes(i) .NE. 0) THEN
               ALLOCATE (send_bufs(i)%array(send_sizes(i)))
            ELSE
               NULLIFY (send_bufs(i)%array)
            END IF
            IF (recv_sizes(i) .NE. 0) THEN
               ALLOCATE (recv_bufs(i)%array(recv_sizes(i)))
            ELSE
               NULLIFY (recv_bufs(i)%array)
            END IF
         END DO

         ALLOCATE (recv_reqs(0:rs%desc%group_size-1))
         recv_reqs = mp_request_null

         DO i = 0, rs%desc%group_size-1
            IF (recv_sizes(i) .NE. 0) THEN
               CALL mp_irecv(recv_bufs(i)%array, i, rs%desc%group, recv_reqs(i))
            END IF
         END DO

         ! do packing
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(k,z,y,x), &
!$OMP             SHARED(pw,rs,send_tasks,send_bufs,send_disps)
         DO i = 0, rs%desc%group_size-1
            k = 0
            DO z = send_tasks(i, 5), send_tasks(i, 6)
               DO y = send_tasks(i, 3), send_tasks(i, 4)
                  DO x = send_tasks(i, 1), send_tasks(i, 2)
                     k = k+1
                     send_bufs(i)%array(k) = pw%cr3d(x, y, z)
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
!$OMP END PARALLEL DO

         ALLOCATE (send_reqs(0:rs%desc%group_size-1))
         send_reqs = mp_request_null

         DO i = 0, rs%desc%group_size-1
            IF (send_sizes(i) .NE. 0) THEN
               CALL mp_isend(send_bufs(i)%array, i, rs%desc%group, send_reqs(i))
            END IF
         END DO

         ! do unpacking
         ! no OMP here so we can unpack each message as it arrives

         DO i = 0, rs%desc%group_size-1
            IF (recv_sizes(i) .EQ. 0) CYCLE

            CALL mp_waitany(recv_reqs, completed)
            k = 0
            DO z = recv_tasks(completed-1, 5), recv_tasks(completed-1, 6)
               DO y = recv_tasks(completed-1, 3), recv_tasks(completed-1, 4)
                  DO x = recv_tasks(completed-1, 1), recv_tasks(completed-1, 2)
                     k = k+1
                     rs%r(x, y, z) = recv_bufs(completed-1)%array(k)
                  ENDDO
               ENDDO
            ENDDO
         ENDDO

         CALL mp_waitall(send_reqs)

         DEALLOCATE (recv_reqs)
         DEALLOCATE (send_reqs)

         DO i = 0, rs%desc%group_size-1
            IF (ASSOCIATED(send_bufs(i)%array)) THEN
               DEALLOCATE (send_bufs(i)%array)
            END IF
            IF (ASSOCIATED(recv_bufs(i)%array)) THEN
               DEALLOCATE (recv_bufs(i)%array)
            END IF
         END DO

         DEALLOCATE (send_bufs)
         DEALLOCATE (recv_bufs)
         DEALLOCATE (send_tasks)
         DEALLOCATE (send_sizes)
         DEALLOCATE (send_disps)
         DEALLOCATE (recv_tasks)
         DEALLOCATE (recv_sizes)
         DEALLOCATE (recv_disps)

         ! now pass wings around
         halo_swapped = .FALSE.

         DO idir = 1, 3

            IF (rs%desc%perd(idir) /= 1) THEN

               ALLOCATE (dshifts(0:rs%desc%neighbours(idir)))
               ALLOCATE (ushifts(0:rs%desc%neighbours(idir)))
               ushifts = 0
               dshifts = 0

               DO n_shifts = 1, rs%desc%neighbours(idir)

                  ! need to take into account the possible varying widths of neighbouring cells
                  ! ushifts and dshifts hold the real size of the neighbouring cells

                  position = MODULO(rs%desc%virtual_group_coor(idir)-n_shifts, rs%desc%group_dim(idir))
                  neighbours = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), position)
                  dshifts(n_shifts) = dshifts(n_shifts-1)+(neighbours(2)-neighbours(1)+1)

                  position = MODULO(rs%desc%virtual_group_coor(idir)+n_shifts, rs%desc%group_dim(idir))
                  neighbours = get_limit(rs%desc%npts(idir), rs%desc%group_dim(idir), position)
                  ushifts(n_shifts) = ushifts(n_shifts-1)+(neighbours(2)-neighbours(1)+1)

                  ! The border data has to be send/received from the neighbors
                  ! First we calculate the source and destination processes for the shift
                  ! The first shift is "downwards"

                  CALL cart_shift(rs, idir, -1*n_shifts, source_down, dest_down)

                  lb_send_down(:) = rs%lb_local(:)
                  ub_send_down(:) = rs%ub_local(:)
                  lb_recv_down(:) = rs%lb_local(:)
                  ub_recv_down(:) = rs%ub_local(:)

                  IF (dshifts(n_shifts-1) .LE. rs%desc%border) THEN
                     lb_send_down(idir) = lb_send_down(idir)+rs%desc%border
                     ub_send_down(idir) = MIN(ub_send_down(idir)-rs%desc%border, &
                                              lb_send_down(idir)+rs%desc%border-1-dshifts(n_shifts-1))

                     lb_recv_down(idir) = ub_recv_down(idir)-rs%desc%border+1+ushifts(n_shifts-1)
                     ub_recv_down(idir) = MIN(ub_recv_down(idir), &
                                              ub_recv_down(idir)-rs%desc%border+ushifts(n_shifts))
                  ELSE
                     lb_send_down(idir) = 0
                     ub_send_down(idir) = -1
                     lb_recv_down(idir) = 0
                     ub_recv_down(idir) = -1
                  ENDIF

                  DO i = 1, 3
                     IF (.NOT. (halo_swapped(i) .OR. i .EQ. idir)) THEN
                        lb_send_down(i) = rs%lb_real(i)
                        ub_send_down(i) = rs%ub_real(i)
                        lb_recv_down(i) = rs%lb_real(i)
                        ub_recv_down(i) = rs%ub_real(i)
                     ENDIF
                  ENDDO

                  ! allocate the recv buffer
                  nn = PRODUCT(ub_recv_down-lb_recv_down+1)
                  ALLOCATE (recv_buf_3d_down(lb_recv_down(1):ub_recv_down(1), &
                                             lb_recv_down(2):ub_recv_down(2), lb_recv_down(3):ub_recv_down(3)))

                  ! recv buffer is now ready, so post the recieve
                  CALL mp_irecv(recv_buf_3d_down, source_down, rs%desc%group, req(1))

                  ! now allocate,pack and send the send buffer
                  nn = PRODUCT(ub_send_down-lb_send_down+1)
                  ALLOCATE (send_buf_3d_down(lb_send_down(1):ub_send_down(1), &
                                             lb_send_down(2):ub_send_down(2), lb_send_down(3):ub_send_down(3)))

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(send_buf_3d_down,rs,lb_send_down,ub_send_down)
!$                num_threads = MIN(omp_get_max_threads(), ub_send_down(3)-lb_send_down(3)+1)
!$                my_id = omp_get_thread_num()
                  IF (my_id < num_threads) THEN
                     lb = lb_send_down(3)+((ub_send_down(3)-lb_send_down(3)+1)*my_id)/num_threads
                     ub = lb_send_down(3)+((ub_send_down(3)-lb_send_down(3)+1)*(my_id+1))/num_threads-1

                     send_buf_3d_down(lb_send_down(1):ub_send_down(1), lb_send_down(2):ub_send_down(2), &
                                      lb:ub) = rs%r(lb_send_down(1):ub_send_down(1), &
                                                    lb_send_down(2):ub_send_down(2), lb:ub)
                  END IF
!$OMP END PARALLEL

                  CALL mp_isend(send_buf_3d_down, dest_down, rs%desc%group, req(3))

                  ! Now for the other direction

                  CALL cart_shift(rs, idir, n_shifts, source_up, dest_up)

                  lb_send_up(:) = rs%lb_local(:)
                  ub_send_up(:) = rs%ub_local(:)
                  lb_recv_up(:) = rs%lb_local(:)
                  ub_recv_up(:) = rs%ub_local(:)

                  IF (ushifts(n_shifts-1) .LE. rs%desc%border) THEN
                     ub_send_up(idir) = ub_send_up(idir)-rs%desc%border
                     lb_send_up(idir) = MAX(lb_send_up(idir)+rs%desc%border, &
                                            ub_send_up(idir)-rs%desc%border+1+ushifts(n_shifts-1))

                     ub_recv_up(idir) = lb_recv_up(idir)+rs%desc%border-1-dshifts(n_shifts-1)
                     lb_recv_up(idir) = MAX(lb_recv_up(idir), &
                                            lb_recv_up(idir)+rs%desc%border-dshifts(n_shifts))
                  ELSE
                     lb_send_up(idir) = 0
                     ub_send_up(idir) = -1
                     lb_recv_up(idir) = 0
                     ub_recv_up(idir) = -1
                  ENDIF

                  DO i = 1, 3
                     IF (.NOT. (halo_swapped(i) .OR. i .EQ. idir)) THEN
                        lb_send_up(i) = rs%lb_real(i)
                        ub_send_up(i) = rs%ub_real(i)
                        lb_recv_up(i) = rs%lb_real(i)
                        ub_recv_up(i) = rs%ub_real(i)
                     ENDIF
                  ENDDO

                  ! allocate the recv buffer
                  nn = PRODUCT(ub_recv_up-lb_recv_up+1)
                  ALLOCATE (recv_buf_3d_up(lb_recv_up(1):ub_recv_up(1), &
                                           lb_recv_up(2):ub_recv_up(2), lb_recv_up(3):ub_recv_up(3)))

                  ! recv buffer is now ready, so post the recieve

                  CALL mp_irecv(recv_buf_3d_up, source_up, rs%desc%group, req(2))

                  ! now allocate,pack and send the send buffer
                  nn = PRODUCT(ub_send_up-lb_send_up+1)
                  ALLOCATE (send_buf_3d_up(lb_send_up(1):ub_send_up(1), &
                                           lb_send_up(2):ub_send_up(2), lb_send_up(3):ub_send_up(3)))

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(send_buf_3d_up,rs,lb_send_up,ub_send_up)
!$                num_threads = MIN(omp_get_max_threads(), ub_send_up(3)-lb_send_up(3)+1)
!$                my_id = omp_get_thread_num()
                  IF (my_id < num_threads) THEN
                     lb = lb_send_up(3)+((ub_send_up(3)-lb_send_up(3)+1)*my_id)/num_threads
                     ub = lb_send_up(3)+((ub_send_up(3)-lb_send_up(3)+1)*(my_id+1))/num_threads-1

                     send_buf_3d_up(lb_send_up(1):ub_send_up(1), lb_send_up(2):ub_send_up(2), &
                                    lb:ub) = rs%r(lb_send_up(1):ub_send_up(1), &
                                                  lb_send_up(2):ub_send_up(2), lb:ub)
                  END IF
!$OMP END PARALLEL

                  CALL mp_isend(send_buf_3d_up, dest_up, rs%desc%group, req(4))

                  ! wait for a recv to complete, then we can unpack

                  DO i = 1, 2

                     CALL mp_waitany(req(1:2), completed)

                     IF (completed .EQ. 1) THEN

                        ! only some procs may need later shifts
                        IF (ub_recv_down(idir) .GE. lb_recv_down(idir)) THEN

                           ! Add the data to the RS Grid
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(recv_buf_3d_down,rs,lb_recv_down,ub_recv_down)
!$                         num_threads = MIN(omp_get_max_threads(), ub_recv_down(3)-lb_recv_down(3)+1)
!$                         my_id = omp_get_thread_num()
                           IF (my_id < num_threads) THEN
                              lb = lb_recv_down(3)+((ub_recv_down(3)-lb_recv_down(3)+1)*my_id)/num_threads
                              ub = lb_recv_down(3)+((ub_recv_down(3)-lb_recv_down(3)+1)*(my_id+1))/num_threads-1

                              rs%r(lb_recv_down(1):ub_recv_down(1), lb_recv_down(2):ub_recv_down(2), &
                                   lb:ub) = recv_buf_3d_down(:, :, lb:ub)
                           END IF
!$OMP END PARALLEL
                        END IF

                        DEALLOCATE (recv_buf_3d_down)
                     ELSE

                        ! only some procs may need later shifts
                        IF (ub_recv_up(idir) .GE. lb_recv_up(idir)) THEN

                           ! Add the data to the RS Grid
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(lb,ub,my_id,NUM_THREADS), &
!$OMP          SHARED(recv_buf_3d_up,rs,lb_recv_up,ub_recv_up)
!$                         num_threads = MIN(omp_get_max_threads(), ub_recv_up(3)-lb_recv_up(3)+1)
!$                         my_id = omp_get_thread_num()
                           IF (my_id < num_threads) THEN
                              lb = lb_recv_up(3)+((ub_recv_up(3)-lb_recv_up(3)+1)*my_id)/num_threads
                              ub = lb_recv_up(3)+((ub_recv_up(3)-lb_recv_up(3)+1)*(my_id+1))/num_threads-1

                              rs%r(lb_recv_up(1):ub_recv_up(1), lb_recv_up(2):ub_recv_up(2), &
                                   lb:ub) = recv_buf_3d_up(:, :, lb:ub)
                           END IF
!$OMP END PARALLEL
                        END IF

                        DEALLOCATE (recv_buf_3d_up)
                     END IF
                  END DO

                  CALL mp_waitall(req(3:4))

                  DEALLOCATE (send_buf_3d_down)
                  DEALLOCATE (send_buf_3d_up)
               END DO

               DEALLOCATE (ushifts)
               DEALLOCATE (dshifts)
            END IF

            halo_swapped(idir) = .TRUE.

         END DO
      END IF

   END SUBROUTINE rs_pw_transfer_distributed

! **************************************************************************************************
!> \brief Initialize grid to zero
!> \param rs ...
!> \par History
!>      none
!> \author JGH (23-Mar-2002)
! **************************************************************************************************
   SUBROUTINE rs_grid_zero(rs)

      TYPE(realspace_grid_type), POINTER                 :: rs

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_grid_zero', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, l(3), u(3)

      CALL timeset(routineN, handle)
      l(1) = LBOUND(rs%r, 1); l(2) = LBOUND(rs%r, 2); l(3) = LBOUND(rs%r, 3)
      u(1) = UBOUND(rs%r, 1); u(2) = UBOUND(rs%r, 2); u(3) = UBOUND(rs%r, 3)
!$OMP PARALLEL DO DEFAULT(NONE) COLLAPSE(3) &
!$OMP             PRIVATE(i,j,k) &
!$OMP             SHARED(rs,l,u)
      DO k = l(3), u(3)
      DO j = l(2), u(2)
      DO i = l(1), u(1)
         rs%r(i, j, k) = 0.0_dp
      ENDDO
      ENDDO
      ENDDO
!$OMP END PARALLEL DO
      CALL timestop(handle)

   END SUBROUTINE rs_grid_zero

! **************************************************************************************************
!> \brief rs1(i) = rs1(i) + rs2(i)*rs3(i)
!> \param rs1 ...
!> \param rs2 ...
!> \param rs3 ...
!> \param scalar ...
!> \par History
!>      none
!> \author
! **************************************************************************************************
   SUBROUTINE rs_grid_mult_and_add(rs1, rs2, rs3, scalar)

      TYPE(realspace_grid_type), POINTER                 :: rs1, rs2, rs3
      REAL(dp), INTENT(IN)                               :: scalar

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_grid_mult_and_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, l(3), u(3)

!-----------------------------------------------------------------------------!

      CALL timeset(routineN, handle)
      IF (scalar .NE. 0.0_dp) THEN
         l(1) = LBOUND(rs1%r, 1); l(2) = LBOUND(rs1%r, 2); l(3) = LBOUND(rs1%r, 3)
         u(1) = UBOUND(rs1%r, 1); u(2) = UBOUND(rs1%r, 2); u(3) = UBOUND(rs1%r, 3)
!$OMP PARALLEL DO DEFAULT(NONE) COLLAPSE(3) &
!$OMP             PRIVATE(i,j,k) &
!$OMP             SHARED(rs1,rs2,rs3,scalar,l,u)
         DO k = l(3), u(3)
         DO j = l(2), u(2)
         DO i = l(1), u(1)
            rs1%r(i, j, k) = rs1%r(i, j, k)+scalar*rs2%r(i, j, k)*rs3%r(i, j, k)
         ENDDO
         ENDDO
         ENDDO
!$OMP END PARALLEL DO
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE rs_grid_mult_and_add

! **************************************************************************************************
!> \brief Set box matrix info for real space grid
!>      This is needed for variable cell simulations
!> \param pw_grid ...
!> \param rs ...
!> \par History
!>      none
!> \author JGH (15-May-2007)
! **************************************************************************************************
   SUBROUTINE rs_grid_set_box(pw_grid, rs)

      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(realspace_grid_type), POINTER                 :: rs

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_grid_set_box', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pw_grid)))CALL cp__a("pw/realspace_grid_types.F",2027)
      IF(.NOT.(ASSOCIATED(rs)))CALL cp__a("pw/realspace_grid_types.F",2028)
      IF(.NOT.(rs%desc%grid_id == pw_grid%id_nr))CALL cp__a("pw/realspace_grid_types.F",2029)
      rs%desc%dh = pw_grid%dh
      rs%desc%dh_inv = pw_grid%dh_inv

   END SUBROUTINE rs_grid_set_box

! **************************************************************************************************
!> \brief retains the given rs grid (see doc/ReferenceCounting.html)
!> \param rs_grid the grid to retain
!> \par History
!>      03.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE rs_grid_retain(rs_grid)
      TYPE(realspace_grid_type), POINTER                 :: rs_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_grid_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rs_grid)))CALL cp__a("pw/realspace_grid_types.F",2047)
      IF(.NOT.(rs_grid%ref_count > 0))CALL cp__a("pw/realspace_grid_types.F",2048)
      rs_grid%ref_count = rs_grid%ref_count+1
   END SUBROUTINE rs_grid_retain

! **************************************************************************************************
!> \brief retains the given rs grid descriptor (see doc/ReferenceCounting.html)
!> \param rs_desc the grid descriptor to retain
!> \par History
!>      04.2009 created [Iain Bethune]
!>        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
! **************************************************************************************************
   SUBROUTINE rs_grid_retain_descriptor(rs_desc)
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_grid_retain_descriptor', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rs_desc)))CALL cp__a("pw/realspace_grid_types.F",2065)
      IF(.NOT.(rs_desc%ref_count > 0))CALL cp__a("pw/realspace_grid_types.F",2066)
      rs_desc%ref_count = rs_desc%ref_count+1
   END SUBROUTINE rs_grid_retain_descriptor

! **************************************************************************************************
!> \brief releases the given rs grid (see doc/ReferenceCounting.html)
!> \param rs_grid the rs grid to release
!> \par History
!>      03.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE rs_grid_release(rs_grid)
      TYPE(realspace_grid_type), POINTER                 :: rs_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_grid_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rs_grid)) THEN
         IF(.NOT.(rs_grid%ref_count > 0))CALL cp__a("pw/realspace_grid_types.F",2084)
         rs_grid%ref_count = rs_grid%ref_count-1
         IF (rs_grid%ref_count == 0) THEN

            CALL rs_grid_release_descriptor(rs_grid%desc)

            allocated_rs_grid_count = allocated_rs_grid_count-1
            DEALLOCATE (rs_grid%r)
            DEALLOCATE (rs_grid%px)
            DEALLOCATE (rs_grid%py)
            DEALLOCATE (rs_grid%pz)
            DEALLOCATE (rs_grid)
         END IF
      END IF
   END SUBROUTINE rs_grid_release

! **************************************************************************************************
!> \brief releases the given rs grid descriptor (see doc/ReferenceCounting.html)
!> \param rs_desc the rs grid descriptor to release
!> \par History
!>      04.2009 created [Iain Bethune]
!>        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
! **************************************************************************************************
   SUBROUTINE rs_grid_release_descriptor(rs_desc)
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc

      CHARACTER(len=*), PARAMETER :: routineN = 'rs_grid_release_descriptor', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rs_desc)) THEN
         IF(.NOT.(rs_desc%ref_count > 0))CALL cp__a("pw/realspace_grid_types.F",2114)
         rs_desc%ref_count = rs_desc%ref_count-1
         IF (rs_desc%ref_count == 0) THEN

            CALL pw_grid_release(rs_desc%pw)

            IF (rs_desc%parallel) THEN
               ! release the group communicator
               CALL mp_comm_free(rs_desc%group)

               DEALLOCATE (rs_desc%virtual2real)
               DEALLOCATE (rs_desc%real2virtual)
            END IF

            IF (rs_desc%distributed) THEN
               DEALLOCATE (rs_desc%rank2coord)
               DEALLOCATE (rs_desc%coord2rank)
               DEALLOCATE (rs_desc%lb_global)
               DEALLOCATE (rs_desc%ub_global)
               DEALLOCATE (rs_desc%x2coord)
               DEALLOCATE (rs_desc%y2coord)
               DEALLOCATE (rs_desc%z2coord)
            ENDIF

            DEALLOCATE (rs_desc)
         END IF
      END IF
   END SUBROUTINE rs_grid_release_descriptor

! **************************************************************************************************
!> \brief emulates the function of an MPI_cart_shift operation, but the shift is
!>        done in virtual coordinates, and the corresponding real ranks are returned
!> \param rs_grid ...
!> \param dir ...
!> \param disp ...
!> \param source ...
!> \param dest ...
!> \par History
!>      04.2009 created [Iain Bethune]
!>        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
! **************************************************************************************************
   SUBROUTINE cart_shift(rs_grid, dir, disp, source, dest)

      TYPE(realspace_grid_type), POINTER                 :: rs_grid
      INTEGER, INTENT(IN)                                :: dir, disp
      INTEGER, INTENT(OUT)                               :: source, dest

      CHARACTER(len=*), PARAMETER :: routineN = 'cart_shift', routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(3)                              :: shift_coords

      shift_coords = rs_grid%desc%virtual_group_coor
      shift_coords(dir) = MODULO(shift_coords(dir)+disp, rs_grid%desc%group_dim(dir))
      dest = rs_grid%desc%virtual2real(rs_grid%desc%coord2rank(shift_coords(1), shift_coords(2), shift_coords(3)))
      shift_coords = rs_grid%desc%virtual_group_coor
      shift_coords(dir) = MODULO(shift_coords(dir)-disp, rs_grid%desc%group_dim(dir))
      source = rs_grid%desc%virtual2real(rs_grid%desc%coord2rank(shift_coords(1), shift_coords(2), shift_coords(3)))

   END SUBROUTINE

! **************************************************************************************************
!> \brief returns the maximum number of points in the local grid of any process
!>        to account for the case where the grid may later be reordered
!> \param desc ...
!> \retval max_ngpts ...
!> \par History
!>      10.2011 created [Iain Bethune]
! **************************************************************************************************
   FUNCTION rs_grid_max_ngpts(desc) RESULT(max_ngpts)
      TYPE(realspace_grid_desc_type), POINTER            :: desc
      INTEGER                                            :: max_ngpts

      INTEGER                                            :: i
      INTEGER, DIMENSION(3)                              :: lb, ub

      max_ngpts = 0
      IF ((desc%pw%para%mode == PW_MODE_LOCAL) .OR. &
          (ALL(desc%group_dim == 1))) THEN
         max_ngpts = PRODUCT(desc%npts)
      ELSE
         DO i = 0, desc%group_size-1
            lb = desc%lb_global(:, i)
            ub = desc%ub_global(:, i)
            lb = lb-desc%border*(1-desc%perd)
            ub = ub+desc%border*(1-desc%perd)
            max_ngpts = MAX(max_ngpts, PRODUCT(ub-lb+1))
         END DO
      END IF

   END FUNCTION rs_grid_max_ngpts

END MODULE realspace_grid_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates error function and complementary error function
!> \par History
!>      adapted to F95
! **************************************************************************************************
MODULE erf_fn
  
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

      IMPLICIT NONE
      PRIVATE
  PUBLIC :: erf, erfc

    CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ARG ...
!> \param RESULT ...
!> \param JINT ...
! **************************************************************************************************
SUBROUTINE CALERF(ARG,RESULT,JINT)

!------------------------------------------------------------------
!
! This packet evaluates  erf(x),  erfc(x),  and  exp(x*x)*erfc(x)
!   for a real argument  x.  It contains three FUNCTION type
!   subprograms: ERF, ERFC, and ERFCX (or DERF, DERFC, and DERFCX),
!   and one SUBROUTINE type subprogram, CALERF.  The calling
!   statements for the primary entries are:
!
!                   Y=ERF(X)     (or   Y=DERF(X)),
!
!                   Y=ERFC(X)    (or   Y=DERFC(X)),
!   and
!                   Y=ERFCX(X)   (or   Y=DERFCX(X)).
!
!   The routine  CALERF  is intended for internal packet use only,
!   all computations within the packet being concentrated in this
!   routine.  The function subprograms invoke  CALERF  with the
!   statement
!
!          CALL CALERF(ARG,RESULT,JINT)
!
!   where the parameter usage is as follows
!
!      Function                     Parameters for CALERF
!       call              ARG                  Result          JINT
!
!     ERF(ARG)      ANY REAL ARGUMENT         ERF(ARG)          0
!     ERFC(ARG)     ABS(ARG) < XBIG           ERFC(ARG)         1
!     ERFCX(ARG)    XNEG < ARG < XMAX         ERFCX(ARG)        2
!
!   The main computation evaluates near-minimax approximations
!   from "Rational Chebyshev approximations for the error function"
!   by W. J. Cody, Math. Comp., 1969, PP. 631-638.  This
!   transportable program uses rational functions that theoretically
!   approximate  erf(x)  and  erfc(x)  to at least 18 significant
!   decimal digits.  The accuracy achieved depends on the arithmetic
!   system, the compiler, the intrinsic functions, and proper
!   selection of the machine-dependent constants.
!
!
! Explanation of machine-dependent constants
!
!   XMIN   = the smallest positive floating-point number.
!   XINF   = the largest positive finite floating-point number.
!   XNEG   = the largest negative argument acceptable to ERFCX;
!            the negative of the solution to the equation
!            2*exp(x*x) = XINF.
!   XSMALL = argument below which erf(x) may be represented by
!            2*x/sqrt(pi)  and above which  x*x  will not underflow.
!            A conservative value is the largest machine number X
!            such that   1.0 + X = 1.0   to machine precision.
!   XBIG   = largest argument acceptable to ERFC;  solution to
!            the equation:  W(x) * (1-0.5/x**2) = XMIN,  where
!            W(x) = exp(-x*x)/[x*sqrt(pi)].
!   XHUGE  = argument above which  1.0 - 1/(2*x*x) = 1.0  to
!            machine precision.  A conservative value is
!            1/[2*sqrt(XSMALL)]
!   XMAX   = largest acceptable argument to ERFCX; the minimum
!            of XINF and 1/[sqrt(pi)*XMIN].
!
!   Approximate values for some important machines are:
!
!                          XMIN            XINF        XNEG     XSMALL
!
!  CDC 7600      (S.P.)  3.13E-294      1.26E+322     -27.220  7.11E-15
!  CRAY-1        (S.P.)  4.58E-2467     5.45E+2465    -75.345  7.11E-15
!  IEEE (IBM/XT,
!    SUN, etc.)  (S.P.)  1.18E-38       3.40E+38       -9.382  5.96E-8
!  IEEE (IBM/XT,
!    SUN, etc.)  (D.P.)  2.23e-308_dp   1.79e+308_dp  -26.628  1.11e-16_dp
!  IBM 195       (D.P.)  5.40e-79_dp    7.23E+75      -13.190  1.39e-17_dp
!  UNIVAC 1108   (D.P.)  2.78e-309_dp   8.98e+307_dp  -26.615  1.73e-18_dp
!  VAX D-Format  (D.P.)  2.94e-39_dp    1.70e+38_dp    -9.345  1.39e-17_dp
!  VAX G-Format  (D.P.)  5.56e-309_dp   8.98e+307_dp  -26.615  1.11e-16_dp
!
!
!                          XBIG       XHUGE       XMAX
!
!  CDC 7600      (S.P.)  25.922      8.39E+6     1.80X+293
!  CRAY-1        (S.P.)  75.326      8.39E+6     5.45E+2465
!  IEEE (IBM/XT,
!    SUN, etc.)  (S.P.)   9.194      2.90E+3     4.79E+37
!  IEEE (IBM/XT,
!    SUN, etc.)  (D.P.)  26.543      6.71e+7_dp  2.53e+307_dp
!  IBM 195       (D.P.)  13.306      1.90e+8_dp  7.23E+75
!  UNIVAC 1108   (D.P.)  26.582      5.37e+8_dp  8.98e+307_dp
!  VAX D-Format  (D.P.)   9.269      1.90e+8_dp  1.70e+38_dp
!  VAX G-Format  (D.P.)  26.569      6.71e+7_dp  8.98e+307_dp
!
!
! Error returns
!
!  The program returns  ERFC = 0      for  ARG >= XBIG;
!
!                       ERFCX = XINF  for  ARG < XNEG;
!      and
!                       ERFCX = 0     for  ARG >= XMAX.
!
!
! Intrinsic functions required are:
!
!     ABS, AINT, EXP
!
!
!  Author: W. J. Cody
!          Mathematics and Computer Science Division
!          Argonne National Laboratory
!          Argonne, IL 60439
!
!  Latest modification: March 19, 1990
!
!------------------------------------------------------------------

      REAL(KIND=dp)                                      :: ARG, RESULT
      INTEGER                                            :: JINT

      INTEGER                                            :: I
      REAL(KIND=dp)                                      :: A, B, C, D, DEL, FOUR, HALF, ONE, P, Q, &
                                                            SIXTEN, SQRPI, THRESH, TWO, X, XBIG, &
                                                            XDEN, XHUGE, XINF, XMAX, XNEG, XNUM, &
                                                            XSMALL, Y, YSQ, ZERO

  DIMENSION A(5),B(4),C(9),D(8),P(6),Q(5)
!------------------------------------------------------------------
!  Mathematical constants
!------------------------------------------------------------------
  DATA FOUR,ONE,HALF,TWO,ZERO/4.0_dp,1.0_dp,0.5_dp,2.0_dp,0.0_dp/, &
       SQRPI/5.6418958354775628695E-1_dp/,THRESH/0.46875_dp/, &
       SIXTEN/16.0_dp/
!------------------------------------------------------------------
!  Machine-dependent constants
!------------------------------------------------------------------
  DATA XINF,XNEG,XSMALL/1.79E+308_dp,-26.628_dp,1.11E-16_dp/, &
       XBIG,XHUGE,XMAX/26.543_dp,6.71E+7_dp,2.53E+307_dp/
!------------------------------------------------------------------
!  Coefficients for approximation to  erf  in first interval
!------------------------------------------------------------------
  DATA A/3.16112374387056560E+00_dp,1.13864154151050156E+02_dp, &
       3.77485237685302021E+02_dp,3.20937758913846947E+03_dp, &
       1.85777706184603153E-1_dp/
  DATA B/2.36012909523441209E+01_dp,2.44024637934444173E+02_dp, &
       1.28261652607737228E+03_dp,2.84423683343917062E+03_dp/
!------------------------------------------------------------------
!  Coefficients for approximation to  erfc  in second interval
!------------------------------------------------------------------
  DATA C/5.64188496988670089E-1_dp,8.88314979438837594E+0_dp, &
       6.61191906371416295E+01_dp,2.98635138197400131E+02_dp, &
       8.81952221241769090E+02_dp,1.71204761263407058E+03_dp, &
       2.05107837782607147E+03_dp,1.23033935479799725E+03_dp, &
       2.15311535474403846E-8_dp/
  DATA D/1.57449261107098347E+01_dp,1.17693950891312499E+02_dp, &
       5.37181101862009858E+02_dp,1.62138957456669019E+03_dp, &
       3.29079923573345963E+03_dp,4.36261909014324716E+03_dp, &
       3.43936767414372164E+03_dp,1.23033935480374942E+03_dp/
!------------------------------------------------------------------
!  Coefficients for approximation to  erfc  in third interval
!------------------------------------------------------------------
  DATA P /3.05326634961232344E-1_dp,3.60344899949804439E-1_dp, &
       1.25781726111229246E-1_dp,1.60837851487422766E-2_dp, &
       6.58749161529837803E-4_dp,1.63153871373020978E-2_dp/
  DATA Q /2.56852019228982242E+00_dp,1.87295284992346047E+00_dp, &
       5.27905102951428412E-1_dp,6.05183413124413191E-2_dp, &
       2.33520497626869185E-3_dp/

!------------------------------------------------------------------

  X = ARG
  Y = ABS(X)
  IF (Y <= THRESH) THEN
!------------------------------------------------------------------
!  Evaluate  erf  for  |X| <= 0.46875
!------------------------------------------------------------------
     YSQ = ZERO
     IF (Y > XSMALL) YSQ = Y * Y
     XNUM = A(5)*YSQ
     XDEN = YSQ
     DO I = 1, 3
        XNUM = (XNUM + A(I)) * YSQ
        XDEN = (XDEN + B(I)) * YSQ
     END DO
     RESULT = X * (XNUM + A(4)) / (XDEN + B(4))
     IF (JINT /= 0) RESULT = ONE - RESULT
     IF (JINT == 2) RESULT = EXP(YSQ) * RESULT
     RETURN
!------------------------------------------------------------------
!  Evaluate  erfc  for 0.46875 <= |X| <= 4.0
!------------------------------------------------------------------
  ELSE IF (Y <= FOUR) THEN
     XNUM = C(9)*Y
     XDEN = Y
     DO I = 1, 7
        XNUM = (XNUM + C(I)) * Y
        XDEN = (XDEN + D(I)) * Y
     END DO
     RESULT = (XNUM + C(8)) / (XDEN + D(8))
     IF (JINT /= 2) THEN
        YSQ = AINT(Y*SIXTEN)/SIXTEN
        DEL = (Y-YSQ)*(Y+YSQ)
        RESULT = EXP(-YSQ*YSQ) * EXP(-DEL) * RESULT
     END IF
!------------------------------------------------------------------
!  Evaluate  erfc  for |X| > 4.0
!------------------------------------------------------------------
  ELSE
     RESULT = ZERO
     IF (Y >= XBIG) THEN
        IF ((JINT /= 2) .OR. (Y >= XMAX)) GO TO 300
        IF (Y >= XHUGE) THEN
           RESULT = SQRPI / Y
           GO TO 300
        END IF
     END IF
     YSQ = ONE / (Y * Y)
     XNUM = P(6)*YSQ
     XDEN = YSQ
     DO I = 1, 4
        XNUM = (XNUM + P(I)) * YSQ
        XDEN = (XDEN + Q(I)) * YSQ
     END DO
     RESULT = YSQ *(XNUM + P(5)) / (XDEN + Q(5))
     RESULT = (SQRPI -  RESULT) / Y
     IF (JINT /= 2) THEN
        YSQ = AINT(Y*SIXTEN)/SIXTEN
        DEL = (Y-YSQ)*(Y+YSQ)
        RESULT = EXP(-YSQ*YSQ) * EXP(-DEL) * RESULT
     END IF
  END IF
!------------------------------------------------------------------
!  Fix up for negative argument, erf, etc.
!------------------------------------------------------------------
300 IF (JINT == 0) THEN
     RESULT = (HALF - RESULT) + HALF
     IF (X < ZERO) RESULT = -RESULT
  ELSE IF (JINT == 1) THEN
     IF (X < ZERO) RESULT = TWO - RESULT
  ELSE
     IF (X < ZERO) THEN
        IF (X < XNEG) THEN
           RESULT = XINF
        ELSE
           YSQ = AINT(X*SIXTEN)/SIXTEN
           DEL = (X-YSQ)*(X+YSQ)
           Y = EXP(YSQ*YSQ) * EXP(DEL)
           RESULT = (Y+Y) - RESULT
        END IF
     END IF
  END IF

!---------- Last card of CALERF ----------
END SUBROUTINE CALERF

! **************************************************************************************************
!> \brief ...
!> \param X ...
!> \retval ERF ...
! **************************************************************************************************
FUNCTION ERF(X)
!--------------------------------------------------------------------
!
! This subprogram computes approximate values for erf(x).
!   (see comments heading CALERF).
!
!   Author/date: W. J. Cody, January 8, 1985
!
!--------------------------------------------------------------------
      REAL(KIND=dp)                                      :: X, ERF

      INTEGER                                            :: JINT
      REAL(KIND=dp)                                      :: RESULT

  JINT = 0
  CALL CALERF(X,RESULT,JINT)
  ERF = RESULT
  RETURN
!---------- Last card of DERF ----------
END FUNCTION erf

! **************************************************************************************************
!> \brief ...
!> \param X ...
!> \retval ERFC ...
! **************************************************************************************************
FUNCTION ERFC(X)
!--------------------------------------------------------------------
!
! This subprogram computes approximate values for erfc(x).
!   (see comments heading CALERF).
!
!   Author/date: W. J. Cody, January 8, 1985
!
!--------------------------------------------------------------------
      REAL(KIND=dp)                                      :: X, ERFC

      INTEGER                                            :: JINT
      REAL(KIND=dp)                                      :: RESULT

  JINT = 1
  CALL CALERF(X,RESULT,JINT)
  ERFC = RESULT
  RETURN
!---------- Last card of DERFC ----------
END FUNCTION erfc

! **************************************************************************************************
!> \brief ...
!> \param X ...
!> \retval ERFCX ...
! **************************************************************************************************
FUNCTION ERFCX(X)
!------------------------------------------------------------------
!
! This subprogram computes approximate values for exp(x*x) * erfc(x).
!   (see comments heading CALERF).
!
!   Author/date: W. J. Cody, March 30, 1987
!
!------------------------------------------------------------------
      REAL(KIND=dp)                                      :: X, ERFCX

      INTEGER                                            :: JINT
      REAL(KIND=dp)                                      :: RESULT

  JINT = 2
  CALL CALERF(X,RESULT,JINT)
  ERFCX = RESULT
  RETURN
!---------- Last card of DERFCX ----------
END FUNCTION erfcx
END  MODULE erf_fn
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE mt_util
   USE bibliography,                    ONLY: Martyna1999,&
                                              cite_reference
   USE erf_fn,                          ONLY: erf
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              oorootpi,&
                                              pi
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_release,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mt_util'

   INTEGER, PARAMETER, PUBLIC               :: MT2D = 1101, &
                                               MT1D = 1102, &
                                               MT0D = 1103

   PUBLIC :: MTin_create_screen_fn
CONTAINS

! **************************************************************************************************
!> \brief Initialize the Martyna && Tuckerman Poisson Solver
!> \param screen_function ...
!> \param pw_pool ...
!> \param method ...
!> \param alpha ...
!> \param special_dimension ...
!> \param slab_size ...
!> \param super_ref_pw_grid ...
!> \author Teodoro Laino (16.06.2004)
! **************************************************************************************************
   SUBROUTINE MTin_create_screen_fn(screen_function, pw_pool, method, alpha, &
                                    special_dimension, slab_size, super_ref_pw_grid)
      TYPE(pw_type), POINTER                             :: screen_function
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      INTEGER, INTENT(IN)                                :: method
      REAL(KIND=dp), INTENT(in)                          :: alpha
      INTEGER, INTENT(IN)                                :: special_dimension
      REAL(KIND=dp), INTENT(in)                          :: slab_size
      TYPE(pw_grid_type), POINTER                        :: super_ref_pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'MTin_create_screen_fn', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ig, iz
      REAL(KIND=dp)                                      :: alpha2, g2, g3d, gxy, gz, zlength
      TYPE(pw_pool_type), POINTER                        :: pw_pool_aux
      TYPE(pw_type), POINTER                             :: Vloc, Vlocg

      CALL timeset(routineN, handle)
      NULLIFY (Vloc, Vlocg, pw_pool_aux)
      !
      ! For Martyna-Tuckerman we set up an auxiliary pw_pool at an higher cutoff
      !
      CALL cite_reference(Martyna1999)
      IF (ASSOCIATED(super_ref_pw_grid)) THEN
         CALL pw_pool_create(pw_pool_aux, pw_grid=super_ref_pw_grid)
      END IF
      CALL pw_pool_create_pw(pw_pool, screen_function, use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_zero(screen_function)
      SELECT CASE (method)
      CASE (MT0D)
         IF (ASSOCIATED(pw_pool_aux)) THEN
            CALL pw_pool_create_pw(pw_pool_aux, Vloc, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_pool_create_pw(pw_pool_aux, Vlocg, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         ELSE
            CALL pw_pool_create_pw(pw_pool, Vloc, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_pool_create_pw(pw_pool, Vlocg, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         END IF
         CALL mt0din(Vloc, alpha)
         CALL pw_transfer(Vloc, Vlocg)
         CALL pw_axpy(Vlocg, screen_function)
         IF (ASSOCIATED(pw_pool_aux)) THEN
            CALL pw_pool_give_back_pw(pw_pool_aux, Vloc)
            CALL pw_pool_give_back_pw(pw_pool_aux, Vlocg)
         ELSE
            CALL pw_pool_give_back_pw(pw_pool, Vloc)
            CALL pw_pool_give_back_pw(pw_pool, Vlocg)
         END IF
         !
         ! Get rid of the analytical FT of the erf(a*r)/r
         !
         alpha2 = alpha*alpha
         DO ig = screen_function%pw_grid%first_gne0, screen_function%pw_grid%ngpts_cut_local
            g2 = screen_function%pw_grid%gsq(ig)
            g3d = fourpi/g2
            screen_function%cc(ig) = screen_function%cc(ig)-g3d*EXP(-g2/(4.0E0_dp*alpha2))
         END DO
         IF (screen_function%pw_grid%have_g0) &
            screen_function%cc(1) = screen_function%cc(1)+fourpi/(4.0E0_dp*alpha2)
      CASE (MT2D)
         iz = special_dimension ! iz is the direction with NO PBC
         zlength = slab_size ! zlength is the thickness of the cell
         DO ig = screen_function%pw_grid%first_gne0, screen_function%pw_grid%ngpts_cut_local
            gz = screen_function%pw_grid%g(iz, ig)
            g2 = screen_function%pw_grid%gsq(ig)
            gxy = SQRT(ABS(g2-gz*gz))
            g3d = fourpi/g2
            screen_function%cc(ig) = -g3d*COS(gz*zlength/2.0_dp)*EXP(-gxy*zlength/2.0_dp)
         END DO
         IF (screen_function%pw_grid%have_g0) screen_function%cc(1) = pi*zlength*zlength/2.0_dp
      CASE (MT1D)
         iz = special_dimension ! iz is the direction with PBC
         CALL mt1din(screen_function)
         CALL cp__b("pw/mt_util.F",129,"MT1D unimplemented")
      END SELECT
      CALL pw_pool_release(pw_pool_aux)
      CALL timestop(handle)
   END SUBROUTINE MTin_create_screen_fn

! **************************************************************************************************
!> \brief Calculates the Tuckerman Green's function in reciprocal space
!>      according the scheme published on:
!>      Martyna and Tuckerman, J. Chem. Phys. Vol. 110, No. 6, 2810-2821
!> \param Vloc ...
!> \param alpha ...
!> \author Teodoro Laino (09.03.2005)
! **************************************************************************************************
   SUBROUTINE mt0din(Vloc, alpha)
      TYPE(pw_type), POINTER                             :: Vloc
      REAL(KIND=dp), INTENT(in)                          :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'mt0din', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, j, jj, k, kk
      INTEGER, DIMENSION(:), POINTER                     :: glb
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: dx, dy, dz, fact, omega, r, r2, x, y, &
                                                            y2, z, z2
      REAL(KIND=dp), DIMENSION(3)                        :: box, box2
      TYPE(pw_grid_type), POINTER                        :: grid

      CALL timeset(routineN, handle)

      grid => Vloc%pw_grid
      bo => grid%bounds_local
      glb => grid%bounds(1, :)
      Vloc%cr3d = 0.0_dp
      box = REAL(grid%npts, kind=dp)*grid%dr
      box2 = box/2.0_dp
      omega = PRODUCT(box)
      fact = omega
      dx = grid%dr(1)
      dy = grid%dr(2)
      dz = grid%dr(3)
      kk = bo(1, 3)
      DO k = bo(1, 3), bo(2, 3)
         z = REAL(k-glb(3), dp)*dz; IF (z .GT. box2(3)) z = box(3)-z
         z2 = z*z
         jj = bo(1, 2)
         DO j = bo(1, 2), bo(2, 2)
            y = REAL(j-glb(2), dp)*dy; IF (y .GT. box2(2)) y = box(2)-y
            y2 = y*y
            ii = bo(1, 1)
            DO i = bo(1, 1), bo(2, 1)
               x = REAL(i-glb(1), dp)*dx; IF (x .GT. box2(1)) x = box(1)-x
               r2 = x*x+y2+z2
               r = SQRT(r2)
               IF (r .GT. 1.0E-10_dp) THEN
                  Vloc%cr3d(ii, jj, kk) = erf(alpha*r)/r*fact
               ELSE
                  Vloc%cr3d(ii, jj, kk) = 2.0_dp*alpha*oorootpi*fact
               END IF
               ii = ii+1
            END DO
            jj = jj+1
         END DO
         kk = kk+1
      END DO
      CALL timestop(handle)
   END SUBROUTINE Mt0din

! **************************************************************************************************
!> \brief Calculates the Tuckerman Green's function in reciprocal space
!>      according the scheme published on:
!>      Martyna and Tuckerman, J. Chem. Phys. Vol. 121, No. 23, 11949
!> \param screen_function ...
!> \author Teodoro Laino (11.2005)
! **************************************************************************************************
   SUBROUTINE mt1din(screen_function)
      TYPE(pw_type), POINTER                             :: screen_function

      CHARACTER(len=*), PARAMETER :: routineN = 'mt1din', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: dx, dy, dz, omega
      REAL(KIND=dp), DIMENSION(3)                        :: box, box2
      TYPE(pw_grid_type), POINTER                        :: grid

      CALL timeset(routineN, handle)
      grid => screen_function%pw_grid
      box = REAL(grid%npts, kind=dp)*grid%dr
      box2 = box/2.0_dp
      omega = PRODUCT(box)
      dx = grid%dr(1)
      dy = grid%dr(2)
      dz = grid%dr(3)

      CALL timestop(handle)
   END SUBROUTINE mt1din

END MODULE mt_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates Bessel functions
!> \note
!>      Functions adapted from netlib
!> \par History
!>      March-2006: Bessel Transform (JGH)
!> \author JGH (10-02-2001)
! **************************************************************************************************
MODULE bessel_lib

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fac,&
                                              pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: bessj0, bessj1, bessk0, bessk1

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval bessj0 ...
! **************************************************************************************************
   FUNCTION bessj0(x)

      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: bessj0

      REAL(KIND=dp), PARAMETER :: p1 = 1.0_dp, p2 = -0.1098628627e-2_dp, p3 = 0.2734510407e-4_dp, &
         p4 = -0.2073370639e-5_dp, p5 = 0.2093887211e-6_dp, q1 = -0.1562499995e-1_dp, &
         q2 = 0.1430488765e-3_dp, q3 = -0.6911147651e-5_dp, q4 = 0.7621095161e-6_dp, &
         q5 = -0.934945152e-7_dp, r1 = 57568490574.0_dp, r2 = -13362590354.0_dp, &
         r3 = 651619640.7_dp, r4 = -11214424.18_dp, r5 = 77392.33017_dp, r6 = -184.9052456_dp, &
         s1 = 57568490411.0_dp, s2 = 1029532985.0_dp, s3 = 9494680.718_dp, s4 = 59272.64853_dp, &
         s5 = 267.8532712_dp, s6 = 1.0_dp

      REAL(KIND=dp)                                      :: ax, xx, y, z

      IF (ABS(x) < 8.0_dp) THEN
         y = x*x
         bessj0 = (r1+y*(r2+y*(r3+y*(r4+y* &
                                     (r5+y*r6)))))/(s1+y*(s2+y*(s3+y* &
                                                                (s4+y*(s5+y*s6)))))
      ELSE
         ax = ABS(x)
         z = 8.0_dp/ax
         y = z*z
         xx = ax-0.785398164_dp
         bessj0 = SQRT(0.636619772_dp/ax)*(COS(xx)* &
                                           (p1+y*(p2+y*(p3+y*(p4+y*p5)))) &
                                           -z*SIN(xx)*(q1+y*(q2+y*(q3+y* &
                                                                   (q4+y*q5)))))
      END IF

   END FUNCTION bessj0

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval bessj1 ...
! **************************************************************************************************
   FUNCTION bessj1(x)

      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: bessj1

      REAL(KIND=dp), PARAMETER :: p1 = 1.0_dp, p2 = -0.183105e-2_dp, p3 = -0.3516396496e-4_dp, &
         p4 = 0.2457520174e-5_dp, p5 = -0.240337019e-6_dp, q1 = 0.04687499995_dp, &
         q2 = -0.2002690873e-3_dp, q3 = 0.8449199096e-5_dp, q4 = -0.88228987e-6_dp, &
         q5 = 0.105787412e-6_dp, r1 = 72362614232.0_dp, r2 = -7895059235.0_dp, &
         r3 = 242396853.1_dp, r4 = -2972611.439_dp, r5 = 15704.48260_dp, r6 = -30.16036606_dp, &
         s1 = 144725228442.0_dp, s2 = 2300535178.0_dp, s3 = 18583304.74_dp, s4 = 99447.43394_dp, &
         s5 = 376.9991397_dp, s6 = 1.0_dp

      REAL(KIND=dp)                                      :: ax, xx, y, z

      IF (ABS(x) < 8.0_dp) THEN
         y = x*x
         bessj1 = x*(r1+y*(r2+y*(r3+y*(r4+y* &
                                       (r5+y*r6)))))/(s1+y*(s2+y*(s3+ &
                                                                  y*(s4+y*(s5+y*s6)))))
      ELSE
         ax = ABS(x)
         z = 8.0_dp/ax
         y = z*z
         xx = ax-2.356194491_dp
         bessj1 = SQRT(0.636619772_dp/ax)*(COS(xx)* &
                                           (p1+y*(p2+y*(p3+y*(p4+y*p5)))) &
                                           -z*SIN(xx)*(q1+y*(q2+y*(q3+y* &
                                                                   (q4+y*q5)))))*SIGN(1.0_dp, x)
      END IF
   END FUNCTION bessj1

! **************************************************************************************************
!> \brief ...
!> \param x must be positive
!> \retval bessk0 ...
! **************************************************************************************************
   FUNCTION bessk0(x)

      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: bessk0

      REAL(KIND=dp), PARAMETER :: p1 = -0.57721566_dp, p2 = 0.42278420_dp, p3 = 0.23069756_dp, &
         p4 = 0.3488590e-1_dp, p5 = 0.262698e-2_dp, p6 = 0.10750e-3_dp, p7 = 0.74e-5_dp, &
         q1 = 1.25331414_dp, q2 = -0.7832358e-1_dp, q3 = 0.2189568e-1_dp, q4 = -0.1062446e-1_dp, &
         q5 = 0.587872e-2_dp, q6 = -0.251540e-2_dp, q7 = 0.53208e-3_dp

      REAL(KIND=dp)                                      :: y

      IF (x < 2.0_dp) THEN
         y = x*x/4.0_dp
         bessk0 = (-LOG(x/2.0_dp)*bessi0(x))+(p1+y* &
                                              (p2+y*(p3+y*(p4+y*(p5+y*(p6+y*p7))))))
      ELSE
         y = (2.0_dp/x)
         bessk0 = (EXP(-x)/SQRT(x))*(q1+y*(q2+y* &
                                           (q3+y*(q4+y*(q5+y*(q6+y*q7))))))
      END IF

   END FUNCTION bessk0

! **************************************************************************************************
!> \brief ...
!> \param x must be positive
!> \retval bessk1 ...
! **************************************************************************************************
   FUNCTION bessk1(x)

      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: bessk1

      REAL(KIND=dp), PARAMETER :: p1 = 1.0_dp, p2 = 0.15443144_dp, p3 = -0.67278579_dp, &
         p4 = -0.18156897_dp, p5 = -0.1919402e-1_dp, p6 = -0.110404e-2_dp, p7 = -0.4686e-4_dp, &
         q1 = 1.25331414_dp, q2 = 0.23498619_dp, q3 = -0.3655620e-1_dp, q4 = 0.1504268e-1_dp, &
         q5 = -0.780353e-2_dp, q6 = 0.325614e-2_dp, q7 = -0.68245e-3_dp

      REAL(KIND=dp)                                      :: y

      IF (x < 2.0_dp) THEN
         y = x*x/4.0_dp
         bessk1 = (LOG(x/2.0_dp)*bessi1(x))+(1.0_dp/x)* &
                  (p1+y*(p2+y*(p3+y*(p4+y*(p5+y* &
                                           (p6+y*p7))))))
      ELSE
         y = 2.0_dp/x
         bessk1 = (EXP(-x)/SQRT(x))*(q1+y*(q2+y* &
                                           (q3+y*(q4+y*(q5+y*(q6+y*q7))))))
      END IF

   END FUNCTION bessk1

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval bessi0 ...
! **************************************************************************************************
   FUNCTION bessi0(x)

      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: bessi0

      REAL(KIND=dp), PARAMETER :: p1 = 1.0_dp, p2 = 3.5156229_dp, p3 = 3.0899424_dp, &
         p4 = 1.2067492_dp, p5 = 0.2659732_dp, p6 = 0.360768e-1_dp, p7 = 0.45813e-2_dp, &
         q1 = 0.39894228_dp, q2 = 0.1328592e-1_dp, q3 = 0.225319e-2_dp, q4 = -0.157565e-2_dp, &
         q5 = 0.916281e-2_dp, q6 = -0.2057706e-1_dp, q7 = 0.2635537e-1_dp, q8 = -0.1647633e-1_dp, &
         q9 = 0.392377e-2_dp

      REAL(KIND=dp)                                      :: ax, y

      IF (ABS(x) < 3.75_dp) THEN
         y = (x/3.75_dp)**2
         bessi0 = p1+y*(p2+y*(p3+y*(p4+y* &
                                    (p5+y*(p6+y*p7)))))
      ELSE
         ax = ABS(x)
         y = 3.75_dp/ax
         bessi0 = (EXP(ax)/SQRT(ax))*(q1+y*(q2+y* &
                                            (q3+y*(q4+y*(q5+y*(q6+y*(q7+y* &
                                                                     (q8+y*q9))))))))
      END IF

   END FUNCTION bessi0

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval bessi1 ...
! **************************************************************************************************
   FUNCTION bessi1(x)

      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: bessi1

      REAL(KIND=dp), PARAMETER :: p1 = 0.5_dp, p2 = 0.87890594_dp, p3 = 0.51498869_dp, &
         p4 = 0.15084934e0_dp, p5 = 0.2658733e-1_dp, p6 = 0.301532e-2_dp, p7 = 0.32411e-3_dp, &
         q1 = 0.39894228_dp, q2 = -0.3988024e-1_dp, q3 = -0.362018e-2_dp, q4 = 0.163801e-2_dp, &
         q5 = -0.1031555e-1_dp, q6 = 0.2282967e-1_dp, q7 = -0.2895312e-1_dp, q8 = 0.1787654e-1_dp, &
         q9 = -0.420059e-2_dp

      REAL(KIND=dp)                                      :: ax, y

      IF (ABS(x) < 3.75_dp) THEN
         y = (x/3.75_dp)**2
         bessi1 = p1+y*(p2+y*(p3+y*(p4+y* &
                                    (p5+y*(p6+y*p7)))))
      ELSE
         ax = ABS(x)
         y = 3.75_dp/ax
         bessi1 = (EXP(ax)/SQRT(ax))*(q1+y*(q2+y* &
                                            (q3+y*(q4+y*(q5+y*(q6+y*(q7+y* &
                                                                     (q8+y*q9))))))))
         IF (x < 0.0_dp) bessi1 = -bessi1
      END IF

   END FUNCTION bessi1

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param l ...
!> \retval bessel0 ...
! **************************************************************************************************
   FUNCTION bessel0(x, l)
      !
      ! Calculates spherical Bessel functions
      ! Abramowitz & Stegun using Formulas 10.1.2, 10.1.8, 10.1.9
      ! Adapted from P. Bloechl
      !
      REAL(KIND=dp), INTENT(IN)                          :: x
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp)                                      :: bessel0

      REAL(KIND=dp), PARAMETER                           :: tol = 1.e-12_dp

      INTEGER                                            :: i, ii, il, isvar, k
      REAL(KIND=dp)                                      :: arg, fact, xsq
      REAL(KIND=dp), DIMENSION(4)                        :: trig

      IF (x > REAL(l, KIND=dp)) THEN
         arg = x-0.5_dp*REAL(l, KIND=dp)*pi
         trig(1) = SIN(arg)/x
         trig(2) = COS(arg)/x
         trig(3) = -trig(1)
         trig(4) = -trig(2)
         bessel0 = trig(1)
         IF (l /= 0) THEN
            xsq = 0.5_dp/x
            fact = 1._dp
            DO k = 1, l
               ii = MOD(k, 4)+1
               fact = fac(k+l)/fac(k)/fac(l-k)*xsq**k
               bessel0 = bessel0+fact*trig(ii)
            END DO
         END IF
      ELSE
         ! Taylor expansion for small arguments
         isvar = 1
         DO il = 1, l
            isvar = isvar*(2*il+1)
         END DO
         IF (l /= 0._dp) THEN
            fact = x**l/REAL(isvar, KIND=dp)
         ELSE
            fact = 1._dp/REAL(isvar, KIND=dp)
         END IF
         bessel0 = fact
         xsq = -0.5_dp*x*x
         isvar = 2*l+1
         DO i = 1, 1000
            isvar = isvar+2
            fact = fact*xsq/REAL(i*isvar, KIND=dp)
            bessel0 = bessel0+fact
            IF (ABS(fact) < tol) EXIT
         ENDDO
         IF (ABS(fact) > tol) CALL cp__b("common/bessel_lib.F",285,"BESSEL0 NOT CONVERGED")
      END IF

   END FUNCTION bessel0

END MODULE bessel_lib

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief functions related to the poisson solver on regular grids
!> \par History
!>      greens_fn: JGH (9-Mar-2001) : include influence_fn into
!>                         greens_fn_type add cell volume
!>                         as indicator for updates
!>      greens_fn: JGH (30-Mar-2001) : Added B-spline routines
!>      pws      : JGH (13-Mar-2001) : new pw_poisson_solver, delete
!>                         pw_greens_fn
!>      12.2004 condensed from pws, greens_fn and green_fns, by apsi and JGH,
!>              made thread safe, new input [fawzi]
!>      14-Mar-2006 : added short range screening function for SE codes
!> \author fawzi
! **************************************************************************************************
MODULE pw_poisson_types
   USE bessel_lib,                      ONLY: bessj0,&
                                              bessj1,&
                                              bessk0,&
                                              bessk1
   USE dielectric_types,                ONLY: dielectric_parameters
   USE dirichlet_bc_types,              ONLY: dirichlet_bc_parameters
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              twopi
   USE mt_util,                         ONLY: MT0D,&
                                              MT1D,&
                                              MT2D,&
                                              MTin_create_screen_fn
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              NEUMANN_BC,&
                                              ps_implicit_parameters,&
                                              ps_implicit_release,&
                                              ps_implicit_type
   USE ps_wavelet_types,                ONLY: ps_wavelet_release,&
                                              ps_wavelet_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_release
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_release,&
                                              pw_pool_type,&
                                              pw_pools_dealloc
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA1D,&
                                              RECIPROCALSPACE,&
                                              pw_release,&
                                              pw_type
   USE realspace_grid_types,            ONLY: realspace_grid_type,&
                                              rs_grid_release
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_poisson_types'

   PUBLIC :: pw_poisson_type
   PUBLIC :: pw_poisson_create, pw_poisson_retain, &
             pw_poisson_release
   PUBLIC :: greens_fn_type, pw_green_create, &
             pw_green_release
   PUBLIC :: pw_poisson_parameter_type

   INTEGER, SAVE, PRIVATE :: last_greens_fn_id_nr = 0
   INTEGER, SAVE, PRIVATE :: last_poisson_id = 0

   INTEGER, PARAMETER, PUBLIC               :: pw_poisson_none = 0, &
                                               pw_poisson_periodic = 1, &
                                               pw_poisson_analytic = 2, &
                                               pw_poisson_mt = 3, &
                                               pw_poisson_hockney = 5, &
                                               pw_poisson_multipole = 4, &
                                               pw_poisson_wavelet = 6, &
                                               pw_poisson_implicit = 7
   ! EWALD methods
   INTEGER, PARAMETER, PUBLIC               :: do_ewald_none = 1, &
                                               do_ewald_ewald = 2, &
                                               do_ewald_pme = 3, &
                                               do_ewald_spme = 4

   INTEGER, PARAMETER, PUBLIC               :: PERIODIC3D = 1000, &
                                               ANALYTIC2D = 1001, &
                                               ANALYTIC1D = 1002, &
                                               ANALYTIC0D = 1003, &
                                               HOCKNEY2D = 1201, &
                                               HOCKNEY1D = 1202, &
                                               HOCKNEY0D = 1203, &
                                               MULTIPOLE2D = 1301, &
                                               MULTIPOLE1D = 1302, &
                                               MULTIPOLE0D = 1303, &
                                               PS_IMPLICIT = 1400

! **************************************************************************************************
!> \brief parameters for the poisson solver independet of input_section
!> \author Ole Schuett
! **************************************************************************************************
   TYPE pw_poisson_parameter_type
      INTEGER                        :: solver

      INTEGER, DIMENSION(3)          :: periodic
      INTEGER                        :: ewald_type = do_ewald_none
      INTEGER                        :: ewald_o_spline
      REAL(KIND=dp)                 :: ewald_alpha

      REAL(KIND=dp)                 :: mt_rel_cutoff
      REAL(KIND=dp)                 :: mt_alpha

      INTEGER                        :: wavelet_scf_type
      INTEGER                        :: wavelet_method
      INTEGER                        :: wavelet_special_dimension
      CHARACTER(LEN=1)               :: wavelet_geocode

      LOGICAL                        :: has_dielectric
      TYPE(dielectric_parameters)    :: dielectric_params
      TYPE(ps_implicit_parameters)   :: ps_implicit_params
      TYPE(dirichlet_bc_parameters)  :: dbc_params
   END TYPE pw_poisson_parameter_type

! **************************************************************************************************
!> \brief environment for the poisson solver
!> \author fawzi
! **************************************************************************************************
   TYPE pw_poisson_type
      INTEGER :: ref_count, id_nr
      INTEGER :: pw_level
      INTEGER :: method
      INTEGER :: used_grid
      LOGICAL :: rebuild
      TYPE(greens_fn_type), POINTER               :: green_fft
      TYPE(ps_wavelet_type), POINTER               :: wavelet
      TYPE(pw_poisson_parameter_type)             :: parameters
      REAL(KIND=dp), DIMENSION(3, 3)             :: cell_hmat = 0.0_dp
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER :: pw_pools
      TYPE(pw_grid_type), POINTER                 :: mt_super_ref_pw_grid
      TYPE(ps_implicit_type), POINTER             :: implicit_env
      TYPE(pw_grid_type), POINTER                 :: dct_pw_grid
      TYPE(realspace_grid_type), POINTER          :: diel_rs_grid
   END TYPE pw_poisson_type

! **************************************************************************************************
!> \brief contains all the informations needed by the fft based poisson solvers
!> \author JGH,Teo,fawzi
! **************************************************************************************************
   TYPE greens_fn_type
      INTEGER :: method
      INTEGER :: special_dimension
      INTEGER :: id_nr
      INTEGER :: ref_count
      REAL(KIND=dp) :: radius
      REAL(KIND=dp) :: MT_alpha
      REAL(KIND=dp) :: MT_rel_cutoff
      REAL(KIND=dp) :: slab_size
      REAL(KIND=dp) :: alpha
      LOGICAL :: p3m
      INTEGER :: p3m_order
      REAL(KIND=dp) :: p3m_alpha
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: p3m_coeff
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: p3m_bm2
      LOGICAL :: sr_screening
      REAL(KIND=dp) :: sr_alpha
      REAL(KIND=dp) :: sr_rc
      TYPE(pw_type), POINTER :: influence_fn
      TYPE(pw_type), POINTER :: dct_influence_fn
      TYPE(pw_type), POINTER :: screen_fn
      TYPE(pw_type), POINTER :: p3m_charge
   END TYPE greens_fn_type

CONTAINS

! **************************************************************************************************
!> \brief Allocates and sets up the green functions for the fft based poisson
!>      solvers
!> \param green ...
!> \param poisson_params ...
!> \param cell_hmat ...
!> \param pw_pool ...
!> \param mt_super_ref_pw_grid ...
!> \param dct_pw_grid ...
!> \author Fawzi, based on previous functions by JGH and Teo
! **************************************************************************************************
   SUBROUTINE pw_green_create(green, poisson_params, cell_hmat, pw_pool, &
                              mt_super_ref_pw_grid, dct_pw_grid)
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(pw_poisson_parameter_type), INTENT(IN)        :: poisson_params
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_grid_type), POINTER                        :: mt_super_ref_pw_grid, dct_pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_green_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dim, i, ig, iz, n, nz
      REAL(KIND=dp)                                      :: g2, g3d, gg, gxy, gz, j0g, j1g, k0g, &
                                                            k1g, rlength, zlength
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(pw_grid_type), POINTER                        :: dct_grid, grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool_xpndd
      TYPE(pw_type), POINTER                             :: dct_gf, gf

      IF(.NOT.(.NOT. (ASSOCIATED(green))))CALL cp__a("pw/pw_poisson_types.F",208)
      ALLOCATE (green)
      green%p3m = .FALSE.
      green%special_dimension = 0
      green%radius = 0.0_dp
      green%slab_size = 0.0_dp
      green%alpha = 0.0_dp
      green%method = PERIODIC3D
      last_greens_fn_id_nr = last_greens_fn_id_nr+1
      green%id_nr = last_greens_fn_id_nr
      green%ref_count = 1
      green%MT_alpha = 1.0_dp
      green%MT_rel_cutoff = 1.0_dp
      green%p3m = .FALSE.
      green%p3m_order = 0
      green%p3m_alpha = 0.0_dp
      green%sr_screening = .FALSE.
      green%sr_alpha = 1.0_dp
      green%sr_rc = 0.0_dp

      NULLIFY (green%influence_fn, green%p3m_charge)
      NULLIFY (green%p3m_coeff, green%p3m_bm2)
      NULLIFY (green%dct_influence_fn)
      NULLIFY (green%screen_fn)

      !IF(.NOT.(cell%orthorhombic))CALL cp__a("pw/pw_poisson_types.F",233)
      DO i = 1, 3
         abc(i) = cell_hmat(i, i)
      END DO
      dim = COUNT(poisson_params%periodic == 1)

      SELECT CASE (poisson_params%solver)
      CASE (pw_poisson_periodic)
         green%method = PERIODIC3D
         IF (dim /= 3) THEN
            CALL cp__b("pw/pw_poisson_types.F",243,"Illegal combination of periodicity and Poisson solver periodic3d")
         END IF
      CASE (pw_poisson_multipole)
         green%method = MULTIPOLE0D
         IF (dim /= 0) THEN
            CALL cp__b("pw/pw_poisson_types.F",248,"Illegal combination of periodicity and Poisson solver mulipole0d")
         END IF
      CASE (pw_poisson_analytic)
         SELECT CASE (dim)
         CASE (0)
            green%method = ANALYTIC0D
            green%radius = 0.5_dp*MINVAL(abc)
         CASE (1)
            green%method = ANALYTIC1D
            green%special_dimension = MAXLOC(poisson_params%periodic(1:3), 1)
            green%radius = MAXVAL(abc(1:3))
            DO i = 1, 3
               IF (i == green%special_dimension) CYCLE
               green%radius = MIN(green%radius, 0.5_dp*abc(i))
            END DO
         CASE (2)
            green%method = ANALYTIC2D
            i = MINLOC(poisson_params%periodic, 1)
            green%special_dimension = i
            green%slab_size = abc(i)
         CASE (3)
            green%method = PERIODIC3D
         CASE DEFAULT
            CALL cp__b("pw/pw_poisson_types.F",271,"")
         END SELECT
      CASE (pw_poisson_mt)
         green%MT_rel_cutoff = poisson_params%mt_rel_cutoff
         green%MT_alpha = poisson_params%mt_alpha
         green%MT_alpha = green%MT_alpha/MINVAL(abc)
         SELECT CASE (dim)
         CASE (0)
            green%method = MT0D
            green%radius = 0.5_dp*MINVAL(abc)
         CASE (1)
            green%method = MT1D
            green%special_dimension = MAXLOC(poisson_params%periodic(1:3), 1)
            green%radius = MAXVAL(abc(1:3))
            DO i = 1, 3
               IF (i == green%special_dimension) CYCLE
               green%radius = MIN(green%radius, 0.5_dp*abc(i))
            END DO
         CASE (2)
            green%method = MT2D
            i = MINLOC(poisson_params%periodic, 1)
            green%special_dimension = i
            green%slab_size = abc(i)
         CASE (3)
            CALL cp__b("pw/pw_poisson_types.F",295,"Illegal combination of periodicity and Poisson solver (MT)")
         CASE DEFAULT
            CALL cp__b("pw/pw_poisson_types.F",297,"")
         END SELECT
      CASE (pw_poisson_implicit)
         green%method = PS_IMPLICIT
      CASE DEFAULT
         CALL cp__b("pw/pw_poisson_types.F",302,"An unknown Poisson solver was specified")
      END SELECT

      ! allocate influence function,...
      SELECT CASE (green%method)
      CASE (PERIODIC3D, ANALYTIC2D, ANALYTIC1D, ANALYTIC0D, MT2D, MT1D, MT0D, MULTIPOLE0D, PS_IMPLICIT)
         CALL pw_pool_create_pw(pw_pool, green%influence_fn, &
                                use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

         IF (poisson_params%ewald_type == do_ewald_spme) THEN
            green%p3m = .TRUE.
            green%p3m_order = poisson_params%ewald_o_spline
            green%p3m_alpha = poisson_params%ewald_alpha
            n = green%p3m_order
            ALLOCATE (green%p3m_coeff(-(n-1):n-1, 0:n-1))
            CALL spme_coeff_calculate(n, green%p3m_coeff)
            CALL pw_pool_create_pw(pw_pool, green%p3m_charge, use_data=REALDATA1D, &
                                   in_space=RECIPROCALSPACE)
            CALL influence_factor(green)
            CALL calc_p3m_charge(green)
         ELSE
            green%p3m = .FALSE.
         END IF
         !
         SELECT CASE (green%method)
         CASE (MT0D, MT1D, MT2D)
            CALL MTin_create_screen_fn(green%screen_fn, pw_pool=pw_pool, method=green%method, &
                                       alpha=green%MT_alpha, &
                                       special_dimension=green%special_dimension, slab_size=green%slab_size, &
                                       super_ref_pw_grid=mt_super_ref_pw_grid)
         CASE (PS_IMPLICIT)
            IF ((poisson_params%ps_implicit_params%boundary_condition .EQ. MIXED_BC) .OR. &
                (poisson_params%ps_implicit_params%boundary_condition .EQ. NEUMANN_BC)) THEN
               CALL pw_pool_create(pw_pool_xpndd, pw_grid=dct_pw_grid)
               CALL pw_pool_create_pw(pw_pool_xpndd, green%dct_influence_fn, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
               CALL pw_pool_release(pw_pool_xpndd)
            END IF
         END SELECT

      CASE DEFAULT
         CALL cp__b("pw/pw_poisson_types.F",343,"")
      END SELECT

      ! initialize influence function
      gf => green%influence_fn
      grid => green%influence_fn%pw_grid
      SELECT CASE (green%method)
      CASE (PERIODIC3D, MULTIPOLE0D)

         DO ig = grid%first_gne0, grid%ngpts_cut_local
            g2 = grid%gsq(ig)
            gf%cc(ig) = fourpi/g2
         END DO
         IF (grid%have_g0) gf%cc(1) = 0.0_dp

      CASE (ANALYTIC2D)

         iz = green%special_dimension ! iz is the direction with NO PBC
         zlength = green%slab_size ! zlength is the thickness of the cell
         DO ig = grid%first_gne0, grid%ngpts_cut_local
            nz = grid%g_hat(iz, ig)
            g2 = grid%gsq(ig)
            g3d = fourpi/g2
            gg = 0.5_dp*SQRT(g2)
            gf%cc(ig) = g3d*(1.0_dp-(-1.0_dp)**nz*EXP(-gg*zlength))
         END DO
         IF (grid%have_g0) gf%cc(1) = 0.0_dp

      CASE (ANALYTIC1D)
         ! see 'ab initio molecular dynamics' table 3.1
         ! iz is the direction of the PBC ( can be 1,2,3 -> x,y,z )
         iz = green%special_dimension
         ! rlength is the radius of the tube
         rlength = green%radius
         DO ig = grid%first_gne0, grid%ngpts_cut_local
            g2 = grid%gsq(ig)
            g3d = fourpi/g2
            gxy = SQRT(MAX(0.0_dp, g2-grid%g(iz, ig)*grid%g(iz, ig)))
            gz = ABS(grid%g(iz, ig))
            j0g = bessj0(rlength*gxy)
            j1g = bessj1(rlength*gxy)
            IF (gz > 0) THEN
               k0g = bessk0(rlength*gz)
               k1g = bessk1(rlength*gz)
            ELSE
               k0g = 0
               k1g = 0
            ENDIF
            gf%cc(ig) = g3d*(1.0_dp+rlength* &
                             (gxy*j1g*k0g-gz*j0g*k1g))
         END DO
         IF (grid%have_g0) gf%cc(1) = 0.0_dp

      CASE (ANALYTIC0D)

         rlength = green%radius ! rlength is the radius of the sphere
         DO ig = grid%first_gne0, grid%ngpts_cut_local
            g2 = grid%gsq(ig)
            gg = SQRT(g2)
            g3d = fourpi/g2
            gf%cc(ig) = g3d*(1.0_dp-COS(rlength*gg))
         END DO
         IF (grid%have_g0) &
            gf%cc(1) = 0.5_dp*fourpi*rlength*rlength

      CASE (MT2D, MT1D, MT0D)

         DO ig = grid%first_gne0, grid%ngpts_cut_local
            g2 = grid%gsq(ig)
            g3d = fourpi/g2
            gf%cc(ig) = g3d+green%screen_fn%cc(ig)
         END DO
         IF (grid%have_g0) &
            gf%cc(1) = green%screen_fn%cc(1)

      CASE (PS_IMPLICIT)

         DO ig = grid%first_gne0, grid%ngpts_cut_local
            g2 = grid%gsq(ig)
            gf%cc(ig) = fourpi/g2
         END DO
         IF (grid%have_g0) gf%cc(1) = 0.0_dp

         IF (ASSOCIATED(green%dct_influence_fn)) THEN
            dct_gf => green%dct_influence_fn
            dct_grid => green%dct_influence_fn%pw_grid

            DO ig = dct_grid%first_gne0, dct_grid%ngpts_cut_local
               g2 = dct_grid%gsq(ig)
               dct_gf%cc(ig) = fourpi/g2
            END DO
            IF (dct_grid%have_g0) dct_gf%cc(1) = 0.0_dp
         END IF

      CASE DEFAULT
         CALL cp__b("pw/pw_poisson_types.F",438,"")
      END SELECT

   END SUBROUTINE pw_green_create

! **************************************************************************************************
!> \brief retains the type
!> \param gftype ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE pw_green_retain(gftype)
      TYPE(greens_fn_type), POINTER                      :: gftype

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_green_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(gftype)))CALL cp__a("pw/pw_poisson_types.F",454)
      IF(.NOT.(gftype%ref_count > 0))CALL cp__a("pw/pw_poisson_types.F",455)
      gftype%ref_count = gftype%ref_count+1
   END SUBROUTINE pw_green_retain

! **************************************************************************************************
!> \brief destroys the type (deallocates data)
!> \param gftype ...
!> \param pw_pool ...
!> \par History
!>      none
!> \author Joost VandeVondele
!>      Teodoro Laino
! **************************************************************************************************
   SUBROUTINE pw_green_release(gftype, pw_pool)
      TYPE(greens_fn_type), POINTER                      :: gftype
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_green_release', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: can_give_back

      IF (ASSOCIATED(gftype)) THEN
         IF(.NOT.(gftype%ref_count > 0))CALL cp__a("pw/pw_poisson_types.F",478)
         gftype%ref_count = gftype%ref_count-1
         IF (gftype%ref_count == 0) THEN
            can_give_back = PRESENT(pw_pool)
            IF (can_give_back) can_give_back = ASSOCIATED(pw_pool)
            IF (can_give_back) THEN
               CALL pw_pool_give_back_pw(pw_pool, gftype%influence_fn, &
                                         accept_non_compatible=.TRUE.)
               CALL pw_pool_give_back_pw(pw_pool, gftype%dct_influence_fn, &
                                         accept_non_compatible=.TRUE.)
               CALL pw_pool_give_back_pw(pw_pool, gftype%screen_fn, &
                                         accept_non_compatible=.TRUE.)
               CALL pw_pool_give_back_pw(pw_pool, gftype%p3m_charge, &
                                         accept_non_compatible=.TRUE.)
            ELSE
               CALL pw_release(gftype%influence_fn)
               CALL pw_release(gftype%dct_influence_fn)
               CALL pw_release(gftype%screen_fn)
               CALL pw_release(gftype%p3m_charge)
            END IF
            IF (ASSOCIATED(gftype%p3m_bm2)) &
               DEALLOCATE (gftype%p3m_bm2)
            IF (ASSOCIATED(gftype%p3m_coeff)) &
               DEALLOCATE (gftype%p3m_coeff)
            DEALLOCATE (gftype)
         END IF
      END IF
      NULLIFY (gftype)
   END SUBROUTINE pw_green_release

! **************************************************************************************************
!> \brief Calculates the influence_factor for the
!>      SPME Green's function in reciprocal space'''
!> \param gftype ...
!> \par History
!>      none
!> \author DH (29-Mar-2001)
! **************************************************************************************************
   SUBROUTINE influence_factor(gftype)
      TYPE(greens_fn_type), POINTER                      :: gftype

      CHARACTER(len=*), PARAMETER :: routineN = 'influence_factor', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: b_m, exp_m, sum_m
      INTEGER                                            :: dim, j, k, l, n, pt
      INTEGER, DIMENSION(3)                              :: npts
      INTEGER, DIMENSION(:), POINTER                     :: lb, ub
      REAL(KIND=dp)                                      :: l_arg, prod_arg, val
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: m_assign

      IF(.NOT.(ASSOCIATED(gftype)))CALL cp__a("pw/pw_poisson_types.F",529)
      IF(.NOT.(gftype%ref_count > 0))CALL cp__a("pw/pw_poisson_types.F",530)
      n = gftype%p3m_order

      ! calculate the assignment function values

      lb => gftype%influence_fn%pw_grid%bounds(1, :)
      ub => gftype%influence_fn%pw_grid%bounds(2, :)
      IF (ASSOCIATED(gftype%p3m_bm2)) THEN
         IF (LBOUND(gftype%p3m_bm2, 2) /= MINVAL(lb(:)) .OR. &
             UBOUND(gftype%p3m_bm2, 2) /= MAXVAL(ub(:))) THEN
            DEALLOCATE (gftype%p3m_bm2)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(gftype%p3m_bm2)) THEN
         ALLOCATE (gftype%p3m_bm2(3, MINVAL(lb(:)):MAXVAL(ub(:))))
      END IF

      ALLOCATE (m_assign(0:n-2))
      m_assign = 0.0_dp
      DO k = 0, n-2
         j = -(n-1)+2*k
         DO l = 0, n-1
            l_arg = 0.5_dp**l
            prod_arg = gftype%p3m_coeff(j, l)*l_arg
            m_assign(k) = m_assign(k)+prod_arg
         END DO
      END DO

      ! calculate the absolute b values

      npts(:) = ub(:)-lb(:)+1
      DO dim = 1, 3
         DO pt = lb(dim), ub(dim)
            val = twopi*(REAL(pt, KIND=dp)/REAL(npts(dim), KIND=dp))
            exp_m = CMPLX(COS(val), -SIN(val), KIND=dp)
            sum_m = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
            DO k = 0, n-2
               sum_m = sum_m+m_assign(k)*exp_m**k
            END DO
            b_m = exp_m**(n-1)/sum_m
            gftype%p3m_bm2(dim, pt) = SQRT(REAL(b_m*CONJG(b_m), KIND=dp))
         END DO
      END DO

      DEALLOCATE (m_assign)
   END SUBROUTINE influence_factor

! **************************************************************************************************
!> \brief ...
!> \param gf ...
! **************************************************************************************************
   SUBROUTINE calc_p3m_charge(gf)

      TYPE(greens_fn_type), POINTER                      :: gf

      INTEGER                                            :: ig, l, m, n
      REAL(KIND=dp)                                      :: arg, novol
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: bm2
      TYPE(pw_grid_type), POINTER                        :: grid
      TYPE(pw_type), POINTER                             :: pc

      grid => gf%influence_fn%pw_grid

      ! check if charge function is consistent with current box volume

      pc => gf%p3m_charge
      bm2 => gf%p3m_bm2
      arg = 1.0_dp/(8.0_dp*gf%p3m_alpha**2)
      novol = REAL(grid%ngpts, KIND=dp)/grid%vol
      DO ig = 1, grid%ngpts_cut_local
         l = grid%g_hat(1, ig)
         m = grid%g_hat(2, ig)
         n = grid%g_hat(3, ig)
         pc%cr(ig) = novol*EXP(-arg*grid%gsq(ig))* &
                     bm2(1, l)*bm2(2, m)*bm2(3, n)
      END DO

   END SUBROUTINE calc_p3m_charge

! **************************************************************************************************
!> \brief Initialize the poisson solver
!>      You should call this just before calling the work routine
!>      pw_poisson_solver
!>      Call pw_poisson_release when you have finished
!> \param poisson_env ...
!> \par History
!>      none
!> \author JGH (12-Mar-2001)
! **************************************************************************************************
   SUBROUTINE pw_poisson_create(poisson_env)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(poisson_env)))CALL cp__a("pw/pw_poisson_types.F",626)
      ALLOCATE (poisson_env)
      last_poisson_id = last_poisson_id+1
      poisson_env%id_nr = last_poisson_id
      poisson_env%ref_count = 1
      poisson_env%method = pw_poisson_none
      poisson_env%rebuild = .TRUE.
      NULLIFY (poisson_env%green_fft, &
               poisson_env%pw_pools, &
               poisson_env%mt_super_ref_pw_grid, &
               poisson_env%wavelet, &
               poisson_env%implicit_env, &
               poisson_env%dct_pw_grid, &
               poisson_env%diel_rs_grid)
      poisson_env%pw_level = -1
      poisson_env%ref_count = 1

   END SUBROUTINE pw_poisson_create

! **************************************************************************************************
!> \brief retains the pw_poisson_env
!> \param poisson_env ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_poisson_retain(poisson_env)
      TYPE(pw_poisson_type), POINTER                     :: poisson_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(poisson_env)))CALL cp__a("pw/pw_poisson_types.F",656)
      IF(.NOT.(poisson_env%ref_count > 0))CALL cp__a("pw/pw_poisson_types.F",657)
      poisson_env%ref_count = poisson_env%ref_count+1
   END SUBROUTINE pw_poisson_retain

! **************************************************************************************************
!> \brief releases the poisson solver
!> \param poisson_env ...
!> \par History
!>      none
!> \author fawzi (11.2002)
! **************************************************************************************************
   SUBROUTINE pw_poisson_release(poisson_env)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(poisson_env)) THEN
         IF(.NOT.(poisson_env%ref_count > 0))CALL cp__a("pw/pw_poisson_types.F",676)
         poisson_env%ref_count = poisson_env%ref_count-1
         IF (poisson_env%ref_count == 0) THEN
            IF (ASSOCIATED(poisson_env%pw_pools)) THEN
               CALL pw_pools_dealloc(poisson_env%pw_pools)
            END IF

            CALL pw_green_release(poisson_env%green_fft)
            CALL pw_grid_release(poisson_env%mt_super_ref_pw_grid)
            CALL ps_wavelet_release(poisson_env%wavelet)
            CALL ps_implicit_release(poisson_env%implicit_env, &
                                     poisson_env%parameters%ps_implicit_params)
            CALL pw_grid_release(poisson_env%dct_pw_grid)
            CALL rs_grid_release(poisson_env%diel_rs_grid)
            DEALLOCATE (poisson_env)

         END IF
      END IF
      NULLIFY (poisson_env)

   END SUBROUTINE pw_poisson_release

! **************************************************************************************************
!> \brief Calculates the coefficients for the charge assignment function
!> \param n ...
!> \param coeff ...
!> \par History
!>      none
!> \author DG (29-Mar-2001)
! **************************************************************************************************
   SUBROUTINE spme_coeff_calculate(n, coeff)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(-(n-1):n-1, 0:n-1), &
         INTENT(OUT)                                     :: coeff

      INTEGER                                            :: i, j, l, m
      REAL(KIND=dp)                                      :: b
      REAL(KIND=dp), DIMENSION(n, -n:n, 0:n-1)           :: a

      a = 0.0_dp
      a(1, 0, 0) = 1.0_dp

      DO i = 2, n
         m = i-1
         DO j = -m, m, 2
            DO l = 0, m-1
               b = (a(m, j-1, l)+ &
                    REAL((-1)**l, KIND=dp)*a(m, j+1, l))/ &
                   REAL((l+1)*2**(l+1), KIND=dp)
               a(i, j, 0) = a(i, j, 0)+b
            END DO
            DO l = 0, m-1
               a(i, j, l+1) = (a(m, j+1, l)- &
                               a(m, j-1, l))/REAL(l+1, KIND=dp)
            END DO
         END DO
      END DO

      coeff = 0.0_dp
      DO i = 0, n-1
         DO j = -(n-1), n-1, 2
            coeff(j, i) = a(n, j, i)
         END DO
      END DO

   END SUBROUTINE spme_coeff_calculate

END MODULE pw_poisson_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
! **************************************************************************************************
MODULE dg_rho0_types

   USE kinds,                           ONLY: dp
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_zero
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_none,&
                                              do_ewald_pme,&
                                              do_ewald_spme
   USE pw_types,                        ONLY: REALDATA3D,&
                                              pw_create,&
                                              pw_p_type,&
                                              pw_release
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC:: dg_rho0_type, dg_rho0_init, dg_rho0_set, dg_rho0_get, &
            dg_rho0_create, dg_rho0_retain, dg_rho0_release

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dg_rho0_types'
   INTEGER, PRIVATE, SAVE :: last_dg_rho0_id_nr = 0

! **************************************************************************************************
!> \brief   Type for Gaussian Densities
!>              type = type of gaussian (PME)
!>              grid = grid number
!>              gcc = Gaussian contraction coefficient
!>              zet = Gaussian exponent
! **************************************************************************************************
   TYPE dg_rho0_type
      INTEGER :: ref_count, id_nr
      INTEGER :: TYPE
      INTEGER :: grid
      INTEGER :: kind
      REAL(KIND=dp) :: cutoff_radius
      REAL(KIND=dp), DIMENSION(:), POINTER :: gcc
      REAL(KIND=dp), DIMENSION(:), POINTER :: zet
      TYPE(pw_p_type) :: density
   END TYPE dg_rho0_type

CONTAINS

! **************************************************************************************************
!> \brief   Set the dg_rho0_type
!> \param dg_rho0 ...
!> \param TYPE ...
!> \param grid ...
!> \param kind ...
!> \param cutoff_radius ...
!> \param gcc ...
!> \param zet ...
!> \param density ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE dg_rho0_set(dg_rho0, TYPE, grid, kind, cutoff_radius, &
                          gcc, zet, density)
      INTEGER, OPTIONAL                                  :: TYPE
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      INTEGER, OPTIONAL                                  :: grid, kind
      REAL(KIND=dp), OPTIONAL                            :: cutoff_radius
      REAL(KIND=dp), OPTIONAL, POINTER                   :: gcc(:), zet(:)
      TYPE(pw_p_type), OPTIONAL                          :: density

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_rho0_set', routineP = moduleN//':'//routineN

      IF (PRESENT(grid)) dg_rho0%grid = grid
      IF (PRESENT(kind)) dg_rho0%kind = kind
      IF (PRESENT(density)) dg_rho0%density = density
      IF (PRESENT(gcc)) dg_rho0%gcc => gcc
      IF (PRESENT(zet)) dg_rho0%zet => zet
      IF (PRESENT(TYPE)) dg_rho0%type = TYPE
      IF (PRESENT(cutoff_radius)) dg_rho0%cutoff_radius = cutoff_radius

   END SUBROUTINE dg_rho0_set

! **************************************************************************************************
!> \brief  Get the dg_rho0_type
!> \param dg_rho0 ...
!> \param cutoff_radius ...
!> \param TYPE ...
!> \param grid ...
!> \param kind ...
!> \param gcc ...
!> \param zet ...
!> \param density ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE dg_rho0_get(dg_rho0, cutoff_radius, TYPE, grid, kind, gcc, zet, density)
      INTEGER, OPTIONAL                                  :: TYPE, kind, grid
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      REAL(KIND=dp), OPTIONAL                            :: cutoff_radius
      REAL(KIND=dp), OPTIONAL, POINTER                   :: gcc(:), zet(:)
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: density

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_rho0_get', routineP = moduleN//':'//routineN

      IF (PRESENT(grid)) grid = dg_rho0%grid
      IF (PRESENT(kind)) kind = dg_rho0%kind
      IF (PRESENT(density)) density = dg_rho0%density
      IF (PRESENT(gcc)) gcc => dg_rho0%gcc
      IF (PRESENT(zet)) zet => dg_rho0%zet
      IF (PRESENT(TYPE)) TYPE = dg_rho0%type
      IF (PRESENT(cutoff_radius)) cutoff_radius = dg_rho0%cutoff_radius

   END SUBROUTINE dg_rho0_get

! **************************************************************************************************
!> \brief   create the dg_rho0 structure
!> \param dg_rho0 ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE dg_rho0_create(dg_rho0)
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_rho0_create', routineP = moduleN//':'//routineN

      ALLOCATE (dg_rho0)
      NULLIFY (dg_rho0%gcc)
      NULLIFY (dg_rho0%zet)
      dg_rho0%cutoff_radius = 0.0_dp
      dg_rho0%grid = 0
      dg_rho0%kind = 0
      dg_rho0%type = do_ewald_none
      last_dg_rho0_id_nr = last_dg_rho0_id_nr+1
      dg_rho0%id_nr = last_dg_rho0_id_nr
      dg_rho0%ref_count = 1
      NULLIFY (dg_rho0%density%pw)

   END SUBROUTINE dg_rho0_create

! **************************************************************************************************
!> \brief retains the given dg_rho0_type
!> \param dg_rho0 the dg_rho0_type to retain
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE dg_rho0_retain(dg_rho0)
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0

      CHARACTER(len=*), PARAMETER :: routineN = 'dg_rho0_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(dg_rho0)))CALL cp__a("pw/dg_rho0_types.F",156)
      IF(.NOT.(dg_rho0%ref_count > 0))CALL cp__a("pw/dg_rho0_types.F",157)
      dg_rho0%ref_count = dg_rho0%ref_count+1
   END SUBROUTINE dg_rho0_retain

! **************************************************************************************************
!> \brief releases the given dg_rho0_type
!> \param dg_rho0 the dg_rho0_type to release
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE dg_rho0_release(dg_rho0)
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0

      CHARACTER(len=*), PARAMETER :: routineN = 'dg_rho0_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dg_rho0)) THEN
         IF(.NOT.(dg_rho0%ref_count > 0))CALL cp__a("pw/dg_rho0_types.F",177)
         dg_rho0%ref_count = dg_rho0%ref_count-1
         IF (dg_rho0%ref_count == 0) THEN
            IF (ASSOCIATED(dg_rho0%gcc)) THEN
               DEALLOCATE (dg_rho0%gcc)
            END IF
            IF (ASSOCIATED(dg_rho0%zet)) THEN
               DEALLOCATE (dg_rho0%zet)
            END IF
            CALL pw_release(dg_rho0%density%pw)
            NULLIFY (dg_rho0%gcc)
            NULLIFY (dg_rho0%zet)
            DEALLOCATE (dg_rho0)
         END IF
      END IF
      NULLIFY (dg_rho0)
   END SUBROUTINE dg_rho0_release

! **************************************************************************************************
!> \brief ...
!> \param dg_rho0 ...
!> \param pw_grid ...
! **************************************************************************************************
   SUBROUTINE dg_rho0_init(dg_rho0, pw_grid)
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_rho0_init', routineP = moduleN//':'//routineN

      CALL pw_release(dg_rho0%density%pw)
      SELECT CASE (dg_rho0%type)
      CASE (do_ewald_ewald)
         CALL pw_create(dg_rho0%density%pw, pw_grid, REALDATA3D)
         CALL dg_rho0_pme_gauss(dg_rho0%density, dg_rho0%zet(1))
      CASE (do_ewald_pme)
         CALL pw_create(dg_rho0%density%pw, pw_grid, REALDATA3D)
         CALL dg_rho0_pme_gauss(dg_rho0%density, dg_rho0%zet(1))
      CASE (do_ewald_spme)
         CALL cp__b("pw/dg_rho0_types.F",215,'SPME type not implemented')
      END SELECT

   END SUBROUTINE dg_rho0_init

! **************************************************************************************************
!> \brief ...
!> \param dg_rho0 ...
!> \param alpha ...
! **************************************************************************************************
   SUBROUTINE dg_rho0_pme_gauss(dg_rho0, alpha)

      TYPE(pw_p_type), INTENT(INOUT)                     :: dg_rho0
      REAL(KIND=dp), INTENT(IN)                          :: alpha

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_rho0_pme_gauss', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: IMPOSSIBLE = 10000

      INTEGER                                            :: gpt, l0, ln, lp, m0, mn, mp, n0, nn, np
      INTEGER, DIMENSION(:), POINTER                     :: ghat
      INTEGER, DIMENSION(:, :), POINTER                  :: bds
      REAL(KIND=dp)                                      :: const, e_gsq
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: rho0
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      const = 1.0_dp/(8.0_dp*alpha**2)

      pw_grid => dg_rho0%pw%pw_grid
      bds => pw_grid%bounds

      IF (-bds(1, 1) == bds(2, 1)) THEN
         l0 = IMPOSSIBLE
      ELSE
         l0 = bds(1, 1)
      END IF

      IF (-bds(1, 2) == bds(2, 2)) THEN
         m0 = IMPOSSIBLE
      ELSE
         m0 = bds(1, 2)
      END IF

      IF (-bds(1, 3) == bds(2, 3)) THEN
         n0 = IMPOSSIBLE
      ELSE
         n0 = bds(1, 3)
      END IF

      CALL pw_zero(dg_rho0%pw)

      rho0 => dg_rho0%pw%cr3d

      DO gpt = 1, pw_grid%ngpts_cut_local
         ghat => pw_grid%g_hat(:, gpt)

         lp = pw_grid%mapl%pos(ghat(1))
         ln = pw_grid%mapl%neg(ghat(1))
         mp = pw_grid%mapm%pos(ghat(2))
         mn = pw_grid%mapm%neg(ghat(2))
         np = pw_grid%mapn%pos(ghat(3))
         nn = pw_grid%mapn%neg(ghat(3))

         e_gsq = EXP(-const*pw_grid%gsq(gpt))/pw_grid%vol

         !*apsi
         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)
         ln = ln+bds(1, 1)
         mn = mn+bds(1, 2)
         nn = nn+bds(1, 3)

         rho0(lp, mp, np) = e_gsq
         rho0(ln, mn, nn) = e_gsq

         IF (ghat(1) == l0 .OR. ghat(2) == m0 .OR. ghat(3) == n0) THEN
            rho0(lp, mp, np) = 0.0_dp
            rho0(ln, mn, nn) = 0.0_dp
         END IF

      END DO

   END SUBROUTINE dg_rho0_pme_gauss

END MODULE dg_rho0_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
! **************************************************************************************************
MODULE dg_types

   USE dg_rho0_types,                   ONLY: dg_rho0_create,&
                                              dg_rho0_release,&
                                              dg_rho0_retain,&
                                              dg_rho0_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

! Information on the assignment function for Ewald
! **************************************************************************************************
   TYPE dg_type
      PRIVATE
      INTEGER :: ref_count, id_nr
      INTEGER :: grid_index
      TYPE(dg_rho0_type), POINTER :: dg_rho0
   END TYPE dg_type

! **************************************************************************************************
   TYPE dg_p_type
      TYPE(dg_type), POINTER :: dg
   END TYPE dg_p_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dg_types'
   INTEGER, SAVE, PRIVATE :: last_dg_id = 0

   PRIVATE
   PUBLIC :: dg_type, dg_get, &
             dg_set, dg_retain, dg_release, &
             dg_create

CONTAINS

! **************************************************************************************************
!> \brief   Get the dg_type
!> \param dg ...
!> \param id_nr ...
!> \param grid_index ...
!> \param dg_rho0 ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE dg_get(dg, id_nr, grid_index, dg_rho0)
      TYPE(dg_type), POINTER                             :: dg
      INTEGER, OPTIONAL                                  :: id_nr, grid_index
      TYPE(dg_rho0_type), OPTIONAL, POINTER              :: dg_rho0

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_get', routineP = moduleN//':'//routineN

      IF (PRESENT(id_nr)) id_nr = dg%id_nr
      IF (PRESENT(grid_index)) grid_index = dg%grid_index
      IF (PRESENT(dg_rho0)) dg_rho0 => dg%dg_rho0

   END SUBROUTINE dg_get

! **************************************************************************************************
!> \brief   create the dg structure
!> \param dg ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE dg_create(dg)
      TYPE(dg_type), POINTER                             :: dg

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_create', routineP = moduleN//':'//routineN

      TYPE(dg_rho0_type), POINTER                        :: dg_rho0

      ALLOCATE (dg)
      NULLIFY (dg_rho0)
      CALL dg_rho0_create(dg_rho0)
      dg%dg_rho0 => dg_rho0
      last_dg_id = last_dg_id+1
      dg%id_nr = last_dg_id
      dg%ref_count = 1

   END SUBROUTINE dg_create

! **************************************************************************************************
!> \brief retains the given dg_type
!> \param dg the dg_type to retain
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE dg_retain(dg)
      TYPE(dg_type), POINTER                             :: dg

      CHARACTER(len=*), PARAMETER :: routineN = 'dg_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(dg)))CALL cp__a("pw/dg_types.F",101)
      IF(.NOT.(dg%ref_count > 0))CALL cp__a("pw/dg_types.F",102)
      dg%ref_count = dg%ref_count+1
   END SUBROUTINE dg_retain

! **************************************************************************************************
!> \brief releases the given dg_type
!> \param dg the dg_type to release
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE dg_release(dg)
      TYPE(dg_type), POINTER                             :: dg

      CHARACTER(len=*), PARAMETER :: routineN = 'dg_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dg)) THEN
         IF(.NOT.(dg%ref_count > 0))CALL cp__a("pw/dg_types.F",121)
         dg%ref_count = dg%ref_count-1
         IF (dg%ref_count == 0) THEN
            CALL dg_rho0_release(dg%dg_rho0)
            DEALLOCATE (dg)
         END IF
      END IF
      NULLIFY (dg)
   END SUBROUTINE dg_release

! **************************************************************************************************
!> \brief   Set the double grid environment
!> \param dg ...
!> \param dg_rho0 ...
!> \param grid_index ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE dg_set(dg, dg_rho0, grid_index)
      TYPE(dg_type), POINTER                             :: dg
      TYPE(dg_rho0_type), OPTIONAL, POINTER              :: dg_rho0
      INTEGER, OPTIONAL                                  :: grid_index

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_set', routineP = moduleN//':'//routineN

      IF (PRESENT(dg_rho0)) THEN
         CALL dg_rho0_retain(dg_rho0)
         CALL dg_rho0_release(dg%dg_rho0)
         dg%dg_rho0 => dg_rho0
      END IF
      IF (PRESENT(grid_index)) dg%grid_index = grid_index
   END SUBROUTINE dg_set

END MODULE dg_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define the data structure for the particle information.
!> \par History
!>      - Atomic kind added in particle_type (MK,08.01.2002)
!>      - Functionality for particle_type added (MK,14.01.2002)
!>      - Allow for general coordinate input (MK,13.09.2003)
!>      - Molecule concept introduced (MK,26.09.2003)
!>      - Last atom information added (jgh,23.05.2004)
!>      - particle_type cleaned (MK,03.02.2005)
!> \author CJM, MK
! **************************************************************************************************
MODULE particle_types
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters (in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'particle_types'

   ! Data types
! **************************************************************************************************
   TYPE particle_type
      TYPE(atomic_kind_type), POINTER       :: atomic_kind => Null() ! atomic kind information
      REAL(KIND=dp), DIMENSION(3)           :: f = 0.0_dp, & ! force
                                               r = 0.0_dp, & ! position
                                               v = 0.0_dp ! velocity
      ! Particle dependent terms for shell-model
      INTEGER                               :: atom_index = -1, &
                                               t_region_index = -1, &
                                               shell_index = -1
   END TYPE particle_type

   ! Public data types

   PUBLIC :: particle_type

   ! Public subroutines

   PUBLIC :: allocate_particle_set, &
             deallocate_particle_set, &
             update_particle_set, &
             update_particle_pos_or_vel, &
             get_particle_pos_or_vel

CONTAINS

! **************************************************************************************************
!> \brief   Allocate a particle set.
!> \param particle_set ...
!> \param nparticle ...
!> \date    14.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_particle_set(particle_set, nparticle)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: nparticle

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_particle_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iparticle

      IF (ASSOCIATED(particle_set)) THEN
         CALL deallocate_particle_set(particle_set)
      END IF
      ALLOCATE (particle_set(nparticle))

      DO iparticle = 1, nparticle
         NULLIFY (particle_set(iparticle)%atomic_kind)
         particle_set(iparticle)%f(:) = 0.0_dp
         particle_set(iparticle)%r(:) = 0.0_dp
         particle_set(iparticle)%v(:) = 0.0_dp
         particle_set(iparticle)%shell_index = 0
         particle_set(iparticle)%atom_index = 0
         particle_set(iparticle)%t_region_index = 0
      END DO

   END SUBROUTINE allocate_particle_set

! **************************************************************************************************
!> \brief   Deallocate a particle set.
!> \param particle_set ...
!> \date    14.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_particle_set(particle_set)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_particle_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(particle_set)) THEN
         DEALLOCATE (particle_set)
      ELSE
         CALL cp_abort(cp__l("subsys/particle_types.F",108), &
                       "The pointer particle_set is not associated and "// &
                       "cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_particle_set

! **************************************************************************************************
!> \brief ...
!> \param particle_set ...
!> \param int_group ...
!> \param pos ...
!> \param vel ...
!> \param for ...
!> \param add ...
! **************************************************************************************************
   SUBROUTINE update_particle_set(particle_set, int_group, pos, vel, for, add)

      TYPE(particle_type), POINTER                       :: particle_set(:)
      INTEGER, INTENT(IN)                                :: int_group
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: pos(:, :), vel(:, :), for(:, :)
      LOGICAL, INTENT(IN), OPTIONAL                      :: add

      CHARACTER(len=*), PARAMETER :: routineN = 'update_particle_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iparticle, nparticle
      LOGICAL                                            :: my_add, update_for, update_pos, &
                                                            update_vel

      CALL timeset(routineN, handle)

      nparticle = SIZE(particle_set)
      update_pos = PRESENT(pos)
      update_vel = PRESENT(vel)
      update_for = PRESENT(for)
      my_add = .FALSE.
      IF (PRESENT(add)) my_add = add

      IF (update_pos) THEN
         CALL mp_sum(pos, int_group)
         IF (my_add) THEN
            DO iparticle = 1, nparticle
               particle_set(iparticle)%r(:) = particle_set(iparticle)%r(:)+pos(:, iparticle)
            END DO
         ELSE
            DO iparticle = 1, nparticle
               particle_set(iparticle)%r(:) = pos(:, iparticle)
            END DO
         END IF
      END IF
      IF (update_vel) THEN
         CALL mp_sum(vel, int_group)
         IF (my_add) THEN
            DO iparticle = 1, nparticle
               particle_set(iparticle)%v(:) = particle_set(iparticle)%v(:)+vel(:, iparticle)
            END DO
         ELSE
            DO iparticle = 1, nparticle
               particle_set(iparticle)%v(:) = vel(:, iparticle)
            END DO
         END IF
      END IF
      IF (update_for) THEN
         CALL mp_sum(for, int_group)
         IF (my_add) THEN
            DO iparticle = 1, nparticle
               particle_set(iparticle)%f(:) = particle_set(iparticle)%f(:)+for(:, iparticle)
            END DO
         ELSE
            DO iparticle = 1, nparticle
               particle_set(iparticle)%f(:) = for(:, iparticle)
            END DO
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE update_particle_set

! **************************************************************************************************
!> \brief   Return the atomic position or velocity of atom iatom in x from a
!>          packed vector even if core-shell particles are present
!> \param iatom ...
!> \param particle_set ...
!> \param vector ...
!> \retval x ...
!> \date    25.11.2010
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   FUNCTION get_particle_pos_or_vel(iatom, particle_set, vector) RESULT(x)

      INTEGER, INTENT(IN)                                :: iatom
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: vector
      REAL(KIND=dp), DIMENSION(3)                        :: x

      INTEGER                                            :: ic, is
      REAL(KIND=dp)                                      :: fc, fs, mass

      ic = 3*(iatom-1)
      IF (particle_set(iatom)%shell_index == 0) THEN
         x(1:3) = vector(ic+1:ic+3)
      ELSE
         is = 3*(SIZE(particle_set)+particle_set(iatom)%shell_index-1)
         mass = particle_set(iatom)%atomic_kind%mass
         fc = particle_set(iatom)%atomic_kind%shell%mass_core/mass
         fs = particle_set(iatom)%atomic_kind%shell%mass_shell/mass
         x(1:3) = fc*vector(ic+1:ic+3)+fs*vector(is+1:is+3)
      END IF

   END FUNCTION get_particle_pos_or_vel

! **************************************************************************************************
!> \brief   Update the atomic position or velocity by x and return the updated
!>          atomic position or velocity in x even if core-shell particles are
!>          present
!> \param iatom ...
!> \param particle_set ...
!> \param x ...
!> \param vector ...
!> \date    26.11.2010
!> \author  Matthias Krack
!> \version 1.0
!> \note    particle-set is not changed, only the positions or velocities in
!>          the packed vector are updated
! **************************************************************************************************
   SUBROUTINE update_particle_pos_or_vel(iatom, particle_set, x, vector)

      INTEGER, INTENT(IN)                                :: iatom
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: x
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: vector

      INTEGER                                            :: ic, is
      REAL(KIND=dp)                                      :: fc, fs, mass

      ic = 3*(iatom-1)
      IF (particle_set(iatom)%shell_index == 0) THEN
         vector(ic+1:ic+3) = vector(ic+1:ic+3)+x(1:3)
         x(1:3) = vector(ic+1:ic+3)
      ELSE
         is = 3*(SIZE(particle_set)+particle_set(iatom)%shell_index-1)
         mass = particle_set(iatom)%atomic_kind%mass
         fc = particle_set(iatom)%atomic_kind%shell%mass_core/mass
         fs = particle_set(iatom)%atomic_kind%shell%mass_shell/mass
         vector(ic+1:ic+3) = vector(ic+1:ic+3)+x(1:3)
         vector(is+1:is+3) = vector(is+1:is+3)+x(1:3)
         x(1:3) = fc*vector(ic+1:ic+3)+fs*vector(is+1:is+3)
      END IF

   END SUBROUTINE update_particle_pos_or_vel

END MODULE particle_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Multipole structure: for multipole (fixed and induced) in FF based MD
!> \author Teodoro Laino [tlaino] - University of Zurich - 12.2007
! **************************************************************************************************
MODULE multipole_types
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE external_potential_types,        ONLY: fist_potential_type,&
                                              get_potential
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: multipole_type, &
             create_multipole_type, &
             release_multipole_type, &
             retain_multipole_type

   INTEGER, PARAMETER, PUBLIC               :: do_multipole_none = -1, &
                                               do_multipole_charge = 0, &
                                               do_multipole_dipole = 1, &
                                               do_multipole_quadrupole = 2

! **************************************************************************************************
!> \brief Define multipole type
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      12.2007 created [tlaino] - Teodoro Laino - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   TYPE multipole_type
      INTEGER :: id_nr, ref_count
      LOGICAL, DIMENSION(3)                    :: task
      REAL(KIND=dp), DIMENSION(:), POINTER     :: charges
      REAL(KIND=dp), DIMENSION(:), POINTER     :: radii
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: quadrupoles
   END TYPE multipole_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'multipole_types'
   INTEGER, PRIVATE, SAVE :: last_multipole_id_nr = 0

CONTAINS

! **************************************************************************************************
!> \brief Create a multipole type
!> \param multipoles ...
!> \param particle_set ...
!> \param subsys_section ...
!> \param max_multipole ...
!> \par History
!>      12.2007 created [tlaino] - Teodoro Laino - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE create_multipole_type(multipoles, particle_set, subsys_section, max_multipole)
      TYPE(multipole_type), POINTER                      :: multipoles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: subsys_section
      INTEGER, INTENT(IN)                                :: max_multipole

      CHARACTER(len=*), PARAMETER :: routineN = 'create_multipole_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ind2, iparticle, j, n_rep, nparticles
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: work
      TYPE(fist_potential_type), POINTER                 :: fist_potential
      TYPE(section_vals_type), POINTER                   :: work_section

      ALLOCATE (multipoles)

      last_multipole_id_nr = last_multipole_id_nr+1
      multipoles%id_nr = last_multipole_id_nr
      multipoles%ref_count = 1
      multipoles%task = .FALSE.
      NULLIFY (multipoles%charges)
      NULLIFY (multipoles%radii)
      NULLIFY (multipoles%dipoles)
      NULLIFY (multipoles%quadrupoles)
      SELECT CASE (max_multipole)
      CASE (do_multipole_none)
         ! Do nothing..
      CASE (do_multipole_charge)
         multipoles%task(1:1) = .TRUE.
      CASE (do_multipole_dipole)
         multipoles%task(1:2) = .TRUE.
      CASE (do_multipole_quadrupole)
         multipoles%task(1:3) = .TRUE.
      CASE DEFAULT
         CALL cp__b("subsys/multipole_types.F",102,"")
      END SELECT
      nparticles = SIZE(particle_set)
      IF (multipoles%task(1)) THEN
         ALLOCATE (multipoles%charges(nparticles))
         ALLOCATE (multipoles%radii(nparticles))
         ! Fill in charge array
         DO iparticle = 1, nparticles
            !atomic_kind =>
            CALL get_atomic_kind(particle_set(iparticle)%atomic_kind, &
                                 fist_potential=fist_potential)
            CALL get_potential(fist_potential, qeff=multipoles%charges(iparticle), &
                               mm_radius=multipoles%radii(iparticle))
         END DO
      END IF
      IF (multipoles%task(2)) THEN
         ALLOCATE (multipoles%dipoles(3, nparticles))
         ! Fill in dipole array (if specified)
         work_section => section_vals_get_subs_vals(subsys_section, "MULTIPOLES%DIPOLES")
         CALL section_vals_get(work_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(work_section, "_DEFAULT_KEYWORD_", n_rep_val=n_rep)
            IF(.NOT.(n_rep == nparticles))CALL cp__a("subsys/multipole_types.F",124)
            DO iparticle = 1, n_rep
               CALL section_vals_val_get(work_section, "_DEFAULT_KEYWORD_", i_rep_val=iparticle, r_vals=work)
               multipoles%dipoles(1:3, iparticle) = work
            END DO
         ELSE
            multipoles%dipoles = 0.0_dp
         END IF
      END IF
      IF (multipoles%task(3)) THEN
         ALLOCATE (multipoles%quadrupoles(3, 3, nparticles))
         ! Fill in quadrupole array (if specified)
         work_section => section_vals_get_subs_vals(subsys_section, "MULTIPOLES%QUADRUPOLES")
         CALL section_vals_get(work_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(work_section, "_DEFAULT_KEYWORD_", n_rep_val=n_rep)
            IF(.NOT.(n_rep == nparticles))CALL cp__a("subsys/multipole_types.F",140)
            DO iparticle = 1, n_rep
               CALL section_vals_val_get(work_section, "_DEFAULT_KEYWORD_", i_rep_val=iparticle, r_vals=work)
               DO i = 1, 3
                  DO j = 1, 3
                     ind2 = 3*(MIN(i, j)-1)-(MIN(i, j)*(MIN(i, j)-1))/2+MAX(i, j)
                     multipoles%quadrupoles(i, j, iparticle) = work(ind2)
                  END DO
               END DO
            END DO
         ELSE
            multipoles%quadrupoles = 0.0_dp
         END IF
      END IF
   END SUBROUTINE create_multipole_type

! **************************************************************************************************
!> \brief ...
!> \param multipoles ...
!> \par History
!>      12.2007 created [tlaino] - Teodoro Laino - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE release_multipole_type(multipoles)
      TYPE(multipole_type), POINTER                      :: multipoles

      CHARACTER(len=*), PARAMETER :: routineN = 'release_multipole_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(multipoles)) THEN
         IF(.NOT.(multipoles%ref_count > 0))CALL cp__a("subsys/multipole_types.F",170)
         multipoles%ref_count = multipoles%ref_count-1
         IF (multipoles%ref_count == 0) THEN
            IF (ASSOCIATED(multipoles%charges)) THEN
               DEALLOCATE (multipoles%charges)
            END IF
            IF (ASSOCIATED(multipoles%radii)) THEN
               DEALLOCATE (multipoles%radii)
            END IF
            IF (ASSOCIATED(multipoles%dipoles)) THEN
               DEALLOCATE (multipoles%dipoles)
            END IF
            IF (ASSOCIATED(multipoles%quadrupoles)) THEN
               DEALLOCATE (multipoles%quadrupoles)
            END IF
            DEALLOCATE (multipoles)
         END IF
      END IF
   END SUBROUTINE release_multipole_type

! **************************************************************************************************
!> \brief ...
!> \param multipoles ...
!> \par History
!>      12.2007 created [tlaino] - Teodoro Laino - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE retain_multipole_type(multipoles)
      TYPE(multipole_type), POINTER                      :: multipoles

      CHARACTER(len=*), PARAMETER :: routineN = 'retain_multipole_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(multipoles)) THEN
         IF(.NOT.(multipoles%ref_count > 0))CALL cp__a("subsys/multipole_types.F",204)
         multipoles%ref_count = multipoles%ref_count+1
      END IF
   END SUBROUTINE retain_multipole_type

END MODULE multipole_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of the semi empirical parameter types.
!> \author JGH (14.08.2004)
! **************************************************************************************************
MODULE semi_empirical_types
   USE basis_set_types,                 ONLY: deallocate_sto_basis_set,&
                                              sto_basis_set_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE dg_types,                        ONLY: dg_type
   USE input_constants,                 ONLY: &
        do_method_am1, do_method_mndo, do_method_mndod, do_method_pdg, do_method_pm3, &
        do_method_pm6, do_method_pnnl, do_method_rm1, do_se_IS_kdso_d, do_se_IS_slater
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE multipole_types,                 ONLY: do_multipole_charge,&
                                              do_multipole_dipole,&
                                              do_multipole_none,&
                                              do_multipole_quadrupole
   USE physcon,                         ONLY: angstrom,&
                                              evolt,&
                                              kcalmol
   USE pw_pool_types,                   ONLY: pw_pool_type
   USE semi_empirical_expns3_types,     ONLY: semi_empirical_expns3_p_type,&
                                              semi_empirical_expns3_release
   USE semi_empirical_mpole_types,      ONLY: semi_empirical_mpole_p_release,&
                                              semi_empirical_mpole_p_type
   USE taper_types,                     ONLY: taper_create,&
                                              taper_release,&
                                              taper_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_types'

! **************************************************************************************************
!> \brief Semi-empirical type
! **************************************************************************************************
   TYPE semi_empirical_type
      INTEGER                                :: typ
      INTEGER                                :: nr
      INTEGER                                :: core_size, atm_int_size
      CHARACTER(LEN=default_string_length)   :: name
      LOGICAL                                :: defined, dorb, extended_basis_set
      LOGICAL                                :: p_orbitals_on_h
      INTEGER                                :: z
      REAL(KIND=dp)                          :: zeff
      INTEGER                                :: natorb
      REAL(KIND=dp), DIMENSION(:), POINTER :: beta
      REAL(KIND=dp), DIMENSION(:), POINTER :: sto_exponents
      REAL(KIND=dp), DIMENSION(:), POINTER :: zn
      TYPE(sto_basis_set_type), POINTER      :: basis
      INTEGER                                :: ngauss
      REAL(KIND=dp)                        :: eheat
      REAL(KIND=dp)                        :: uss, upp, udd, uff
      REAL(KIND=dp)                        :: alp
      REAL(KIND=dp)                        :: eisol
      REAL(KIND=dp)                        :: ass, asp, app, de, acoul
      REAL(KIND=dp)                        :: gss, gsp, gpp, gp2
      REAL(KIND=dp)                        :: gsd, gpd, gdd
      REAL(KIND=dp)                        :: hsp
      REAL(KIND=dp)                        :: dd, qq, am, ad, aq
      REAL(KIND=dp), DIMENSION(2)           :: pre, d
      REAL(KIND=dp), DIMENSION(4)           :: fn1, fn2, fn3
      REAL(KIND=dp), DIMENSION(4, 4)         :: bfn1, bfn2, bfn3
      REAL(KIND=dp)                        :: f0dd, f2dd, f4dd, f0sd, f0pd, f2pd, &
                                              g1pd, g2sd, g3pd
      REAL(KIND=dp), DIMENSION(9)          :: ko
      REAL(KIND=dp), DIMENSION(6)          :: cs
      REAL(KIND=dp), DIMENSION(52)         :: onec2el
      ! Specific for PM6
      REAL(KIND=dp), DIMENSION(0:115)      :: xab
      REAL(KIND=dp), DIMENSION(0:115)      :: aab
      REAL(KIND=dp)                        :: a, b, c, rho
      ! One center - two electron integrals
      REAL(KIND=dp), DIMENSION(:, :), &
         POINTER                           :: w
      TYPE(semi_empirical_mpole_p_type), &
         POINTER, DIMENSION(:)             :: w_mpole
      ! 1/R^3 residual integral part
      TYPE(semi_empirical_expns3_p_type), &
         POINTER, DIMENSION(:)             :: expns3_int
   END TYPE semi_empirical_type

   TYPE semi_empirical_p_type
      TYPE(semi_empirical_type), POINTER    :: se_param
   END TYPE semi_empirical_p_type

! **************************************************************************************************
!> \brief  Rotation Matrix Type
!> \author 05.2008 Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   TYPE rotmat_type
      ! Value of Rotation Matrices
      REAL(KIND=dp), DIMENSION(3, 3)      :: sp
      REAL(KIND=dp), DIMENSION(5, 5)      :: sd
      REAL(KIND=dp), DIMENSION(6, 3, 3)      :: pp
      REAL(KIND=dp), DIMENSION(15, 5, 3)      :: pd
      REAL(KIND=dp), DIMENSION(15, 5, 5)      :: dd
      ! Derivatives of Rotation Matrices
      REAL(KIND=dp), DIMENSION(3, 3, 3)   :: sp_d
      REAL(KIND=dp), DIMENSION(3, 5, 5)   :: sd_d
      REAL(KIND=dp), DIMENSION(3, 6, 3, 3)   :: pp_d
      REAL(KIND=dp), DIMENSION(3, 15, 5, 3)   :: pd_d
      REAL(KIND=dp), DIMENSION(3, 15, 5, 5)   :: dd_d
   END TYPE rotmat_type

! **************************************************************************************************
!> \brief  Ewald control type (for periodic SE)
!> \author Teodoro Laino [tlaino]  - 12.2008
! **************************************************************************************************
   TYPE ewald_gks_type
      REAL(KIND=dp)                            :: alpha
      TYPE(dg_type), POINTER                   :: dg
      TYPE(pw_pool_type), POINTER              :: pw_pool
   END TYPE ewald_gks_type

   TYPE se_int_control_type
      LOGICAL                                  :: shortrange
      LOGICAL                                  :: do_ewald_r3
      LOGICAL                                  :: do_ewald_gks
      LOGICAL                                  :: pc_coulomb_int
      INTEGER                                  :: integral_screening
      INTEGER                                  :: max_multipole
      TYPE(ewald_gks_type)                     :: ewald_gks
   END TYPE se_int_control_type

! **************************************************************************************************
!> \brief Store the value of the tapering function and possibly its derivative
!>        for screened integrals
! **************************************************************************************************
   TYPE se_int_screen_type
      REAL(KIND=dp)                            :: ft, dft
   END TYPE se_int_screen_type

! **************************************************************************************************
!> \brief Taper type use in semi-empirical calculations
! **************************************************************************************************
   TYPE se_taper_type
      TYPE(taper_type), POINTER             :: taper
      TYPE(taper_type), POINTER             :: taper_cou
      TYPE(taper_type), POINTER             :: taper_exc
      TYPE(taper_type), POINTER             :: taper_lrc
      ! This taper is for KDSO-D integrals
      TYPE(taper_type), POINTER             :: taper_add
   END TYPE se_taper_type

   PUBLIC :: semi_empirical_type, &
             semi_empirical_p_type, &
             semi_empirical_create, &
             semi_empirical_release, &
             rotmat_type, &
             rotmat_create, &
             rotmat_release, &
             get_se_param, &
             write_se_param, &
             se_int_control_type, &
             setup_se_int_control_type, &
             se_int_screen_type, &
             se_taper_type, &
             se_taper_release, &
             se_taper_create

CONTAINS

! **************************************************************************************************
!> \brief Allocate semi-empirical type
!> \param sep ...
! **************************************************************************************************
   SUBROUTINE semi_empirical_create(sep)
      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(sep)))CALL cp__a("semi_empirical_types.F",191)
      ALLOCATE (sep)
      ALLOCATE (sep%beta(0:3))
      ALLOCATE (sep%sto_exponents(0:3))
      ALLOCATE (sep%zn(0:3))
      NULLIFY (sep%basis)
      NULLIFY (sep%w)
      NULLIFY (sep%w_mpole)
      NULLIFY (sep%expns3_int)
      CALL zero_se_param(sep)

   END SUBROUTINE semi_empirical_create

! **************************************************************************************************
!> \brief Deallocate the semi-empirical type
!> \param sep ...
! **************************************************************************************************
   SUBROUTINE semi_empirical_release(sep)

      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(sep)) THEN
         CALL deallocate_sto_basis_set(sep%basis)
         CALL semi_empirical_mpole_p_release(sep%w_mpole)
         IF (ASSOCIATED(sep%beta)) THEN
            DEALLOCATE (sep%beta)
         END IF
         IF (ASSOCIATED(sep%sto_exponents)) THEN
            DEALLOCATE (sep%sto_exponents)
         END IF
         IF (ASSOCIATED(sep%zn)) THEN
            DEALLOCATE (sep%zn)
         END IF
         IF (ASSOCIATED(sep%w)) THEN
            DEALLOCATE (sep%w)
         END IF
         IF (ASSOCIATED(sep%expns3_int)) THEN
            DO i = 1, SIZE(sep%expns3_int)
               CALL semi_empirical_expns3_release(sep%expns3_int(i)%expns3)
            END DO
            DEALLOCATE (sep%expns3_int)
         END IF
         DEALLOCATE (sep)
      END IF

   END SUBROUTINE semi_empirical_release

! **************************************************************************************************
!> \brief Zero the whole semi-empirical type
!> \param sep ...
! **************************************************************************************************
   SUBROUTINE zero_se_param(sep)
      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'zero_se_param', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(sep)))CALL cp__a("semi_empirical_types.F",252)
      sep%defined = .FALSE.
      sep%dorb = .FALSE.
      sep%extended_basis_set = .FALSE.
      sep%p_orbitals_on_h = .FALSE.
      sep%name = ""
      sep%typ = HUGE(0)
      sep%core_size = HUGE(0)
      sep%atm_int_size = HUGE(0)
      sep%z = HUGE(0)
      sep%zeff = HUGE(0.0_dp)
      sep%natorb = 0
      sep%ngauss = 0
      sep%eheat = HUGE(0.0_dp)

      sep%zn = 0.0_dp
      sep%sto_exponents = 0.0_dp
      sep%beta = 0.0_dp

      sep%uss = 0.0_dp !eV
      sep%upp = 0.0_dp !eV
      sep%udd = 0.0_dp !eV
      sep%uff = 0.0_dp
      sep%alp = 0.0_dp
      sep%eisol = 0.0_dp
      sep%nr = 1
      sep%acoul = 0.0_dp
      sep%de = 0.0_dp
      sep%ass = 0.0_dp
      sep%asp = 0.0_dp
      sep%app = 0.0_dp
      sep%gss = 0.0_dp
      sep%gsp = 0.0_dp
      sep%gpp = 0.0_dp
      sep%gp2 = 0.0_dp
      sep%gsd = 0.0_dp
      sep%gpd = 0.0_dp
      sep%gdd = 0.0_dp
      sep%hsp = 0.0_dp
      sep%dd = 0.0_dp
      sep%qq = 0.0_dp
      sep%am = 0.0_dp
      sep%ad = 0.0_dp
      sep%aq = 0.0_dp

      sep%fn1 = 0.0_dp
      sep%fn2 = 0.0_dp
      sep%fn3 = 0.0_dp
      sep%bfn1 = 0.0_dp
      sep%bfn2 = 0.0_dp
      sep%bfn3 = 0.0_dp

      sep%pre = 0.0_dp
      sep%d = 0.0_dp

      sep%xab = 0.0_dp
      sep%aab = 0.0_dp
      sep%a = 0.0_dp
      sep%b = 0.0_dp
      sep%c = 0.0_dp
      sep%rho = 0.0_dp

      sep%f0dd = 0.0_dp
      sep%f2dd = 0.0_dp
      sep%f4dd = 0.0_dp
      sep%f0sd = 0.0_dp
      sep%f0pd = 0.0_dp
      sep%f2pd = 0.0_dp
      sep%g1pd = 0.0_dp
      sep%g2sd = 0.0_dp
      sep%g3pd = 0.0_dp
      sep%ko = 0.0_dp
      sep%cs = 0.0_dp
      sep%onec2el = 0.0_dp

   END SUBROUTINE zero_se_param

! **************************************************************************************************
!> \brief Get info from the semi-empirical type
!> \param sep ...
!> \param name ...
!> \param typ ...
!> \param defined ...
!> \param z ...
!> \param zeff ...
!> \param natorb ...
!> \param eheat ...
!> \param beta ...
!> \param sto_exponents ...
!> \param uss ...
!> \param upp ...
!> \param udd ...
!> \param uff ...
!> \param alp ...
!> \param eisol ...
!> \param gss ...
!> \param gsp ...
!> \param gpp ...
!> \param gp2 ...
!> \param acoul ...
!> \param nr ...
!> \param de ...
!> \param ass ...
!> \param asp ...
!> \param app ...
!> \param hsp ...
!> \param gsd ...
!> \param gpd ...
!> \param gdd ...
!> \param ppddg ...
!> \param dpddg ...
!> \param ngauss ...
! **************************************************************************************************
   SUBROUTINE get_se_param(sep, name, typ, defined, z, zeff, natorb, eheat, &
                           beta, sto_exponents, uss, upp, udd, uff, alp, eisol, gss, gsp, gpp, gp2, &
                           acoul, nr, de, ass, asp, app, hsp, gsd, gpd, gdd, ppddg, dpddg, ngauss)

      TYPE(semi_empirical_type), POINTER                 :: sep
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      INTEGER, INTENT(OUT), OPTIONAL                     :: typ
      LOGICAL, INTENT(OUT), OPTIONAL                     :: defined
      INTEGER, INTENT(OUT), OPTIONAL                     :: z
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: zeff
      INTEGER, INTENT(OUT), OPTIONAL                     :: natorb
      REAL(KIND=dp), OPTIONAL                            :: eheat
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: beta, sto_exponents
      REAL(KIND=dp), OPTIONAL                            :: uss, upp, udd, uff, alp, eisol, gss, &
                                                            gsp, gpp, gp2, acoul
      INTEGER, INTENT(OUT), OPTIONAL                     :: nr
      REAL(KIND=dp), OPTIONAL                            :: de, ass, asp, app, hsp, gsd, gpd, gdd
      REAL(KIND=dp), DIMENSION(2), OPTIONAL              :: ppddg, dpddg
      INTEGER, INTENT(OUT), OPTIONAL                     :: ngauss

      CHARACTER(len=*), PARAMETER :: routineN = 'get_se_param', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(sep)) THEN
         IF (PRESENT(name)) name = sep%name
         IF (PRESENT(typ)) typ = sep%typ
         IF (PRESENT(defined)) defined = sep%defined
         IF (PRESENT(z)) z = sep%z
         IF (PRESENT(zeff)) zeff = sep%zeff
         IF (PRESENT(natorb)) natorb = sep%natorb
         IF (PRESENT(eheat)) eheat = sep%eheat
         IF (PRESENT(beta)) beta => sep%beta
         IF (PRESENT(sto_exponents)) sto_exponents => sep%sto_exponents
         IF (PRESENT(ngauss)) ngauss = sep%ngauss
         IF (PRESENT(uss)) uss = sep%uss
         IF (PRESENT(upp)) upp = sep%upp
         IF (PRESENT(udd)) udd = sep%udd
         IF (PRESENT(uff)) uff = sep%uff
         IF (PRESENT(alp)) alp = sep%alp
         IF (PRESENT(eisol)) eisol = sep%eisol
         IF (PRESENT(nr)) nr = sep%nr
         IF (PRESENT(acoul)) acoul = sep%acoul
         IF (PRESENT(de)) de = sep%de
         IF (PRESENT(ass)) ass = sep%ass
         IF (PRESENT(asp)) asp = sep%asp
         IF (PRESENT(app)) app = sep%app
         IF (PRESENT(gss)) gss = sep%gss
         IF (PRESENT(gsp)) gsp = sep%gsp
         IF (PRESENT(gpp)) gpp = sep%gpp
         IF (PRESENT(gp2)) gp2 = sep%gp2
         IF (PRESENT(hsp)) hsp = sep%hsp
         IF (PRESENT(gsd)) gsd = sep%gsd
         IF (PRESENT(gpd)) gpd = sep%gpd
         IF (PRESENT(gdd)) gdd = sep%gdd
         IF (PRESENT(ppddg)) ppddg = sep%pre
         IF (PRESENT(dpddg)) dpddg = sep%d
      ELSE
         CALL cp__b("semi_empirical_types.F",422,"The pointer sep is not associated")
      END IF

   END SUBROUTINE get_se_param

! **************************************************************************************************
!> \brief Set info from the semi-empirical type
!> \param sep ...
!> \param name ...
!> \param typ ...
!> \param defined ...
!> \param z ...
!> \param zeff ...
!> \param natorb ...
!> \param eheat ...
!> \param beta ...
!> \param sto_exponents ...
!> \param uss ...
!> \param upp ...
!> \param udd ...
!> \param uff ...
!> \param alp ...
!> \param eisol ...
!> \param gss ...
!> \param gsp ...
!> \param gpp ...
!> \param gp2 ...
!> \param acoul ...
!> \param nr ...
!> \param de ...
!> \param ass ...
!> \param asp ...
!> \param app ...
!> \param hsp ...
!> \param gsd ...
!> \param gpd ...
!> \param gdd ...
!> \param ppddg ...
!> \param dpddg ...
!> \param ngauss ...
! **************************************************************************************************
   SUBROUTINE set_se_param(sep, name, typ, defined, z, zeff, natorb, eheat, &
                           beta, sto_exponents, uss, upp, udd, uff, alp, eisol, gss, gsp, gpp, gp2, &
                           acoul, nr, de, ass, asp, app, hsp, gsd, gpd, gdd, ppddg, dpddg, ngauss)

      TYPE(semi_empirical_type), POINTER                 :: sep
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: name
      INTEGER, INTENT(IN), OPTIONAL                      :: typ
      LOGICAL, INTENT(IN), OPTIONAL                      :: defined
      INTEGER, INTENT(IN), OPTIONAL                      :: z
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: zeff
      INTEGER, INTENT(IN), OPTIONAL                      :: natorb
      REAL(KIND=dp), OPTIONAL                            :: eheat
      REAL(dp), DIMENSION(0:), OPTIONAL                  :: beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: sto_exponents
      REAL(KIND=dp), OPTIONAL                            :: uss, upp, udd, uff, alp, eisol, gss, &
                                                            gsp, gpp, gp2, acoul
      INTEGER, INTENT(IN), OPTIONAL                      :: nr
      REAL(KIND=dp), OPTIONAL                            :: de, ass, asp, app, hsp, gsd, gpd, gdd
      REAL(dp), DIMENSION(2), OPTIONAL                   :: ppddg, dpddg
      INTEGER, INTENT(IN), OPTIONAL                      :: ngauss

      CHARACTER(len=*), PARAMETER :: routineN = 'set_se_param', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(sep)) THEN
         IF (PRESENT(name)) sep%name = name
         IF (PRESENT(typ)) sep%typ = typ
         IF (PRESENT(defined)) sep%defined = defined
         IF (PRESENT(z)) sep%z = z
         IF (PRESENT(zeff)) sep%zeff = zeff
         IF (PRESENT(natorb)) sep%natorb = natorb
         IF (PRESENT(eheat)) sep%eheat = eheat
         IF (PRESENT(beta)) sep%beta = beta
         IF (PRESENT(sto_exponents)) sep%sto_exponents = sto_exponents
         IF (PRESENT(ngauss)) sep%ngauss = ngauss
         IF (PRESENT(uss)) sep%uss = uss
         IF (PRESENT(upp)) sep%upp = upp
         IF (PRESENT(udd)) sep%udd = udd
         IF (PRESENT(uff)) sep%uff = uff
         IF (PRESENT(alp)) sep%alp = alp
         IF (PRESENT(eisol)) sep%eisol = eisol
         IF (PRESENT(acoul)) sep%acoul = acoul
         IF (PRESENT(nr)) sep%nr = nr
         IF (PRESENT(de)) sep%de = de
         IF (PRESENT(ass)) sep%ass = ass
         IF (PRESENT(asp)) sep%asp = asp
         IF (PRESENT(app)) sep%app = app
         IF (PRESENT(gss)) sep%gss = gss
         IF (PRESENT(gsp)) sep%gsp = gsp
         IF (PRESENT(gpp)) sep%gpp = gpp
         IF (PRESENT(gp2)) sep%gp2 = gp2
         IF (PRESENT(hsp)) sep%hsp = hsp
         IF (PRESENT(gsd)) sep%gsd = gsd
         IF (PRESENT(gpd)) sep%gpd = gpd
         IF (PRESENT(gdd)) sep%gdd = gdd
         IF (PRESENT(ppddg)) sep%pre = ppddg
         IF (PRESENT(dpddg)) sep%d = dpddg
      ELSE
         CALL cp__b("semi_empirical_types.F",521,"The pointer sep is not associated")
      END IF

   END SUBROUTINE set_se_param

! **************************************************************************************************
!> \brief Creates rotmat type
!> \param rotmat ...
! **************************************************************************************************
   SUBROUTINE rotmat_create(rotmat)
      TYPE(rotmat_type), POINTER                         :: rotmat

      CHARACTER(len=*), PARAMETER :: routineN = 'rotmat_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(rotmat)))CALL cp__a("semi_empirical_types.F",535)
      ALLOCATE (rotmat)

   END SUBROUTINE rotmat_create

! **************************************************************************************************
!> \brief Releases rotmat type
!> \param rotmat ...
! **************************************************************************************************
   SUBROUTINE rotmat_release(rotmat)
      TYPE(rotmat_type), POINTER                         :: rotmat

      CHARACTER(len=*), PARAMETER :: routineN = 'rotmat_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rotmat)) THEN
         DEALLOCATE (rotmat)
      END IF

   END SUBROUTINE rotmat_release

! **************************************************************************************************
!> \brief Setup the Semiempirical integral control type
!> \param se_int_control ...
!> \param shortrange ...
!> \param do_ewald_r3 ...
!> \param do_ewald_gks ...
!> \param integral_screening ...
!> \param max_multipole ...
!> \param pc_coulomb_int ...
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   SUBROUTINE setup_se_int_control_type(se_int_control, shortrange, do_ewald_r3, &
                                        do_ewald_gks, integral_screening, max_multipole, pc_coulomb_int)
      TYPE(se_int_control_type)                          :: se_int_control
      LOGICAL, INTENT(IN)                                :: shortrange, do_ewald_r3, do_ewald_gks
      INTEGER, INTENT(IN)                                :: integral_screening, max_multipole
      LOGICAL, INTENT(IN)                                :: pc_coulomb_int

      se_int_control%shortrange = shortrange
      se_int_control%do_ewald_r3 = do_ewald_r3
      se_int_control%integral_screening = integral_screening
      ! This makes the assignment independent of the value of the different constants
      SELECT CASE (max_multipole)
      CASE (do_multipole_none)
         se_int_control%max_multipole = -1
      CASE (do_multipole_charge)
         se_int_control%max_multipole = 0
      CASE (do_multipole_dipole)
         se_int_control%max_multipole = 1
      CASE (do_multipole_quadrupole)
         se_int_control%max_multipole = 2
      END SELECT

      se_int_control%do_ewald_gks = do_ewald_gks
      se_int_control%pc_coulomb_int = pc_coulomb_int
      NULLIFY (se_int_control%ewald_gks%dg, se_int_control%ewald_gks%pw_pool)

   END SUBROUTINE setup_se_int_control_type

! **************************************************************************************************
!> \brief Creates the taper type used in SE calculations
!> \param se_taper ...
!> \param integral_screening ...
!> \param do_ewald ...
!> \param taper_cou ...
!> \param range_cou ...
!> \param taper_exc ...
!> \param range_exc ...
!> \param taper_scr ...
!> \param range_scr ...
!> \param taper_lrc ...
!> \param range_lrc ...
!> \author Teodoro Laino [tlaino] - 03.2009
! **************************************************************************************************
   SUBROUTINE se_taper_create(se_taper, integral_screening, do_ewald, &
                              taper_cou, range_cou, taper_exc, range_exc, taper_scr, range_scr, &
                              taper_lrc, range_lrc)
      TYPE(se_taper_type), POINTER                       :: se_taper
      INTEGER, INTENT(IN)                                :: integral_screening
      LOGICAL, INTENT(IN)                                :: do_ewald
      REAL(KIND=dp), INTENT(IN)                          :: taper_cou, range_cou, taper_exc, &
                                                            range_exc, taper_scr, range_scr, &
                                                            taper_lrc, range_lrc

      CHARACTER(len=*), PARAMETER :: routineN = 'se_taper_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(se_taper)))CALL cp__a("semi_empirical_types.F",622)
      ALLOCATE (se_taper)
      NULLIFY (se_taper%taper)
      NULLIFY (se_taper%taper_cou)
      NULLIFY (se_taper%taper_exc)
      NULLIFY (se_taper%taper_lrc)
      NULLIFY (se_taper%taper_add)
      ! Create the sub-typo taper
      CALL taper_create(se_taper%taper_cou, taper_cou, range_cou)
      CALL taper_create(se_taper%taper_exc, taper_exc, range_exc)
      IF (integral_screening == do_se_IS_kdso_d) THEN
         CALL taper_create(se_taper%taper_add, taper_scr, range_scr)
      END IF
      IF ((integral_screening /= do_se_IS_slater) .AND. do_ewald) THEN
         CALL taper_create(se_taper%taper_lrc, taper_lrc, range_lrc)
      END IF
   END SUBROUTINE se_taper_create

! **************************************************************************************************
!> \brief Releases the taper type used in SE calculations
!> \param se_taper ...
!> \author Teodoro Laino [tlaino] - 03.2009
! **************************************************************************************************
   SUBROUTINE se_taper_release(se_taper)
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'se_taper_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(se_taper)) THEN
         CALL taper_release(se_taper%taper_cou)
         CALL taper_release(se_taper%taper_exc)
         CALL taper_release(se_taper%taper_lrc)
         CALL taper_release(se_taper%taper_add)

         DEALLOCATE (se_taper)
      END IF
   END SUBROUTINE se_taper_release

! **************************************************************************************************
!> \brief Writes the semi-empirical type
!> \param sep ...
!> \param subsys_section ...
!> \par History
!>        04.2008 Teodoro Laino [tlaino] - University of Zurich: rewriting with
!>                support for the whole set of parameters
! **************************************************************************************************
   SUBROUTINE write_se_param(sep, subsys_section)

      TYPE(semi_empirical_type), POINTER                 :: sep
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_se_param', routineP = moduleN//':'//routineN
      CHARACTER(LEN=1), DIMENSION(0:3), PARAMETER :: orb_lab = (/"S", "P", "D", "F"/)
      CHARACTER(LEN=2), DIMENSION(0:3), PARAMETER :: z_lab = (/"ZS", "ZP", "ZD", "ZF"/)
      CHARACTER(LEN=3), DIMENSION(0:3), PARAMETER :: zeta_lab = (/"ZSN", "ZPN", "ZDN", "ZFN"/)
      CHARACTER(LEN=5), DIMENSION(0:3), PARAMETER :: &
         beta_lab = (/"BETAS", "BETAP", "BETAD", "BETAF"/)
      CHARACTER(LEN=default_string_length)               :: i_string, name
      INTEGER                                            :: i, l, natorb, ngauss, nr, output_unit, &
                                                            typ, z
      LOGICAL                                            :: defined
      REAL(KIND=dp)                                      :: acoul, alp, app, asp, ass, de, eheat, &
                                                            eisol, gp2, gpp, gsp, gss, hsp, udd, &
                                                            uff, upp, uss, zeff
      CHARACTER(LEN=3), DIMENSION(0:3), PARAMETER :: u_lab = (/"USS", "UPP", "UDD", "UFF"/)

      REAL(KIND=dp), DIMENSION(0:3)                      :: u
      REAL(KIND=dp), DIMENSION(2)                        :: dpddg, ppddg
      REAL(KIND=dp), DIMENSION(:), POINTER               :: beta, sexp
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      IF (ASSOCIATED(sep) .AND. BTEST(cp_print_key_should_output(logger%iter_info, subsys_section, &
                                                                 "PRINT%KINDS/SE_PARAMETERS"), cp_p_file)) THEN

         output_unit = cp_print_key_unit_nr(logger, subsys_section, "PRINT%KINDS/SE_PARAMETERS", &
                                            extension=".Log")

         IF (output_unit > 0) THEN
            CALL get_se_param(sep, name=name, typ=typ, defined=defined, &
                              z=z, zeff=zeff, natorb=natorb, eheat=eheat, beta=beta, &
                              sto_exponents=sexp, uss=uss, upp=upp, udd=udd, uff=uff, &
                              alp=alp, eisol=eisol, gss=gss, gsp=gsp, gpp=gpp, gp2=gp2, &
                              de=de, ass=ass, asp=asp, app=app, hsp=hsp, ppddg=ppddg, &
                              acoul=acoul, nr=nr, dpddg=dpddg, ngauss=ngauss)

            u(0) = uss
            u(1) = upp
            u(2) = udd
            u(3) = uff

            SELECT CASE (typ)
            CASE DEFAULT
               CALL cp__b("semi_empirical_types.F",717,"Semiempirical method unknown")
            CASE (do_method_am1)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "Austin Model 1 (AM1)", TRIM(name)
            CASE (do_method_rm1)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "Recife Model 1 (RM1)", TRIM(name)
            CASE (do_method_pm3)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "Parametric Method 3 (PM3) ", TRIM(name)
            CASE (do_method_pnnl)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "PNNL method ", TRIM(name)
            CASE (do_method_pm6)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "Parametric Method 6 (PM6) ", TRIM(name)
            CASE (do_method_pdg)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "PDDG/PM3 ", TRIM(name)
            CASE (do_method_mndo)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "MNDO ", TRIM(name)
            CASE (do_method_mndod)
               WRITE (UNIT=output_unit, FMT="(/,A,T35,A,T67,A14)") &
                  " Semi empirical parameters: ", "MNDOD", TRIM(name)
            END SELECT

            ! If defined print all its semi-empirical parameters
            IF (defined) THEN
               WRITE (UNIT=output_unit, FMT="(T16,A,T71,F10.2)") &
                  "Effective core charge:", zeff
               WRITE (UNIT=output_unit, FMT="(T16,A,T71,I10)") &
                  "Number of orbitals:", natorb, &
                  "Basis set expansion (STO-NG)", ngauss
               WRITE (UNIT=output_unit, FMT="(T16,A,T66,F15.5)") &
                  "Atomic heat of formation [kcal/mol]:", eheat*kcalmol
               DO l = 0, 3
                  IF (ABS(beta(l)) > 0._dp) THEN
                     WRITE (UNIT=output_unit, FMT="(T16,A,I2)") "Parameters for Shell: ", l
                     WRITE (UNIT=output_unit, FMT="(T22,A5,T30,A,T64,F17.4)") &
                        ADJUSTR(z_lab(l)), "- "//"Slater  Exponent for "//orb_lab(l)//"     [A]: ", sexp(l)
                     WRITE (UNIT=output_unit, FMT="(T22,A5,T30,A,T64,F17.4)") &
                        ADJUSTR(u_lab(l)), "- "//"One Center Energy for "//orb_lab(l)//"   [eV]: ", u(l)*evolt
                     WRITE (UNIT=output_unit, FMT="(T22,A5,T30,A,T64,F17.4)") &
                        ADJUSTR(beta_lab(l)), "- "//"Beta Parameter for "//orb_lab(l)//"      [eV]: ", beta(l)*evolt
                     WRITE (UNIT=output_unit, FMT="(T22,A5,T30,A,T64,F17.4)") &
                        ADJUSTR(zeta_lab(l)), "- "//"Internal Exponent for "//orb_lab(l)//" [a.u.]: ", sep%zn(l)
                  END IF
               END DO
               WRITE (UNIT=output_unit, FMT="(/,T16,A)") "Additional Parameters (Derived or Fitted):"
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("ALP"), "- "//"Alpha Parameter for Core    [A^-1]: ", alp/angstrom
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("EISOL"), "- "//"Atomic Energy (Calculated)    [eV]: ", eisol*evolt
               ! One center Two electron Integrals
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("GSS"), "- "//"One Center Integral (SS ,SS ) [eV]: ", gss*evolt
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("GSP"), "- "//"One Center Integral (SS ,PP ) [eV]: ", gsp*evolt
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("GPP"), "- "//"One Center Integral (PP ,PP ) [eV]: ", gpp*evolt
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("GP2"), "- "//"One Center Integral (PP*,PP*) [eV]: ", gp2*evolt
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("HSP"), "- "//"One Center Integral (SP ,SP ) [eV]: ", hsp*evolt
               ! Slater Condon Parameters
               IF (sep%dorb) THEN
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("F0DD"), "- "//"Slater Condon Parameter F0DD  [eV]: ", sep%f0dd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("F2DD"), "- "//"Slater Condon Parameter F2DD  [eV]: ", sep%f2dd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("F4DD"), "- "//"Slater Condon Parameter F4DD  [eV]: ", sep%f4dd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("FOSD"), "- "//"Slater Condon Parameter FOSD  [eV]: ", sep%f0sd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("G2SD"), "- "//"Slater Condon Parameter G2SD  [eV]: ", sep%g2sd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("F0PD"), "- "//"Slater Condon Parameter F0PD  [eV]: ", sep%f0pd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("F2PD"), "- "//"Slater Condon Parameter F2PD  [eV]: ", sep%f2pd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("G1PD"), "- "//"Slater Condon Parameter G1PD  [eV]: ", sep%g1pd
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("G3PD"), "- "//"Slater Condon Parameter G3PD  [eV]: ", sep%g3pd
               END IF
               ! Charge Separation
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("DD2"), "- "//"Charge Separation  SP, L=1  [bohr]: ", sep%cs(2)
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("DD3"), "- "//"Charge Separation  PP, L=2  [bohr]: ", sep%cs(3)
               IF (sep%dorb) THEN
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("DD4"), "- "//"Charge Separation  SD, L=2  [bohr]: ", sep%cs(4)
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("DD5"), "- "//"Charge Separation  PD, L=1  [bohr]: ", sep%cs(5)
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("DD6"), "- "//"Charge Separation  DD, L=2  [bohr]: ", sep%cs(6)
               END IF
               ! Klopman-Ohno Terms
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("PO1"), "- "//"Klopman-Ohno term, SS, L=0  [bohr]: ", sep%ko(1)
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("PO2"), "- "//"Klopman-Ohno term, SP, L=1  [bohr]: ", sep%ko(2)
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("PO3"), "- "//"Klopman-Ohno term, PP, L=2  [bohr]: ", sep%ko(3)
               IF (sep%dorb) THEN
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("PO4"), "- "//"Klopman-Ohno term, SD, L=2  [bohr]: ", sep%ko(4)
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("PO5"), "- "//"Klopman-Ohno term, PD, L=1  [bohr]: ", sep%ko(5)
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("PO6"), "- "//"Klopman-Ohno term, DD, L=2  [bohr]: ", sep%ko(6)
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("PO7"), "- "//"Klopman-Ohno term, PP, L=0  [bohr]: ", sep%ko(7)
                  WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                     ADJUSTR("PO8"), "- "//"Klopman-Ohno term, DD, L=0  [bohr]: ", sep%ko(8)
               END IF
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                  ADJUSTR("PO9"), "- "//"Klopman-Ohno term, CORE     [bohr]: ", sep%ko(9)
               SELECT CASE (typ)
               CASE (do_method_am1, do_method_rm1, do_method_pm3, do_method_pdg, do_method_pnnl)
                  IF (typ == do_method_pnnl) THEN
                     WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                        ADJUSTR("ASS"), "- "//" SS polarization [au]: ", sep%ass
                     WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                        ADJUSTR("ASP"), "- "//" SP polarization [au]: ", sep%asp
                     WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                        ADJUSTR("APP"), "- "//" PP polarization[au]: ", sep%app
                     WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                        ADJUSTR("DE"), "- "//" Dispersion Parameter [eV]: ", sep%de*evolt
                     WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                        ADJUSTR("ACOUL"), "- "//" Slater parameter: ", sep%acoul
                     WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,I12)") &
                        ADJUSTR("NR"), "- "//" Slater parameter: ", sep%nr
                  ELSEIF ((typ == do_method_am1 .OR. typ == do_method_rm1) .AND. sep%z == 5) THEN
                     ! Standard case
                     DO i = 1, SIZE(sep%bfn1, 1)
                        i_string = cp_to_string(i)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN1"//TRIM(ADJUSTL(i_string))//"_ALL"), &
                           "- "//"Core-Core VDW, Multiplier   [a.u.]: ", sep%bfn1(i, 1)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN2"//TRIM(ADJUSTL(i_string))//"_ALL"), &
                           "- "//"Core-Core VDW, Exponent     [a.u.]: ", sep%bfn2(i, 1)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN3"//TRIM(ADJUSTL(i_string))//"_ALL"), &
                           "- "//"Core-Core VDW, Position     [a.u.]: ", sep%bfn3(i, 1)
                     END DO
                     ! Special Case : Hydrogen
                     DO i = 1, SIZE(sep%bfn1, 1)
                        i_string = cp_to_string(i)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN1"//TRIM(ADJUSTL(i_string))//"_H"), &
                           "- "//"Core-Core VDW, Multiplier   [a.u.]: ", sep%bfn1(i, 2)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN2"//TRIM(ADJUSTL(i_string))//"_H"), &
                           "- "//"Core-Core VDW, Exponent     [a.u.]: ", sep%bfn2(i, 2)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN3"//TRIM(ADJUSTL(i_string))//"_H"), &
                           "- "//"Core-Core VDW, Position     [a.u.]: ", sep%bfn3(i, 2)
                     END DO
                     ! Special Case : Carbon
                     DO i = 1, SIZE(sep%bfn1, 1)
                        i_string = cp_to_string(i)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN1"//TRIM(ADJUSTL(i_string))//"_C"), &
                           "- "//"Core-Core VDW, Multiplier   [a.u.]: ", sep%bfn1(i, 3)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN2"//TRIM(ADJUSTL(i_string))//"_C"), &
                           "- "//"Core-Core VDW, Exponent     [a.u.]: ", sep%bfn2(i, 3)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN3"//TRIM(ADJUSTL(i_string))//"_C"), &
                           "- "//"Core-Core VDW, Position     [a.u.]: ", sep%bfn3(i, 3)
                     END DO
                     ! Special Case : Halogens
                     DO i = 1, SIZE(sep%bfn1, 1)
                        i_string = cp_to_string(i)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN1"//TRIM(ADJUSTL(i_string))//"_HALO"), &
                           "- "//"Core-Core VDW, Multiplier   [a.u.]: ", sep%bfn1(i, 4)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN2"//TRIM(ADJUSTL(i_string))//"_HALO"), &
                           "- "//"Core-Core VDW, Exponent     [a.u.]: ", sep%bfn2(i, 4)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN3"//TRIM(ADJUSTL(i_string))//"_HALO"), &
                           "- "//"Core-Core VDW, Position     [a.u.]: ", sep%bfn3(i, 4)
                     END DO
                  ELSE
                     DO i = 1, SIZE(sep%fn1, 1)
                        i_string = cp_to_string(i)
                        ! Skip the printing of params that are zero..
                        IF (sep%fn1(i) == 0.0_dp .AND. sep%fn2(i) == 0.0_dp .AND. sep%fn3(i) == 0.0_dp) CYCLE
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN1"//TRIM(ADJUSTL(i_string))), &
                           "- "//"Core-Core VDW, Multiplier   [a.u.]: ", sep%fn1(i)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN2"//TRIM(ADJUSTL(i_string))), &
                           "- "//"Core-Core VDW, Exponent     [a.u.]: ", sep%fn2(i)
                        WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T69,F12.4)") &
                           ADJUSTR("FN3"//TRIM(ADJUSTL(i_string))), &
                           "- "//"Core-Core VDW, Position     [a.u.]: ", sep%fn3(i)
                     END DO
                  END IF
               END SELECT
            ELSE
               WRITE (UNIT=output_unit, FMT="(T55,A)") "Parameters are not defined"
            END IF

            ! Additional Parameters not common to all semi-empirical methods
            SELECT CASE (typ)
            CASE (do_method_pdg)
               WRITE (UNIT=output_unit, FMT="(T16,A11,T30,A,T52,F14.10,T67,F14.10)") &
                  ADJUSTR("d_PDDG"), "- "//"Exponent [A^-1]:", dpddg/angstrom, &
                  ADJUSTR("P_PDDG"), "- "//"Parameter  [eV]:", ppddg*evolt
            END SELECT
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                           "PRINT%KINDS/SE_PARAMETERS")
      END IF
   END SUBROUTINE write_se_param

END MODULE semi_empirical_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Generation of the spherical Lebedev grids.
!>          All Lebedev grids were generated with a precision of at least
!>          33 digits (Mathematica). Thus the Lebedev abscissas and weights
!>          are accurate to 32 digits (quadruple precision).
!> \version 1.0
!>
!> \par Literature
!>          - V. I. Lebedev, Zh. Vychisl. Mat. Mat. Fiz. 15, 48 (1975)
!>          - V. I. Lebedev, Zh. Vychisl. Mat. Mat. Fiz. 16, 293 (1976)
!>          - V. I. Lebedev, Sibirsk Mat. Zh. 18, 132 (1977)
!>          - V. I. Lebedev and A. L. Skorokhodov, Russ. Acad. Sci. Dokl.
!>            Math. 45, 587 (1992)
!>          - V. I. Lebedev, Russ. Acad. Sci. Dokl. Math. 50, 283 (1995)
!> \par Notes
!>          Implemented Lebedev grids:
!>            No.    l    n(nonred.)    n(total)
!>             1     3         1             6
!>             2     5         2            14
!>             3     7         3            26
!>             4     9         3            38
!>             5    11         4            50
!>             6    15         5            86
!>             7    17         6           110
!>             8    19         7           146
!>             9    23         9           194
!>            10    29        12           302
!>            11    35        16           434
!>            12    41        20           590
!>            13    47        25           770
!>            14    53        30           974
!> \par Variables:
!>        - lebedev_grid: Lebedev grids.
!>                        l: Angular momentum quantum number l.
!>                        n: Number of grid points.
!>                        w: Grid point weights.
!>                        r: Abscissas (grid point coordinates)
!>        - max_np      : Maximum number of nonredundant grid points.
!>        - na1         : Number of point for subsystem A1.
!>        - na2         : Number of point for subsystem A2.
!>        - na3         : Number of point for subsystem A3.
!>        - nb          : Number of point for subsystem B.
!>        - nc          : Number of point for subsystem C.
!>        - nd          : Number of point for subsystem D.
!>        - nlg         : Number of implemented Lebedev grids.
!>        - nlgp        : Grid point counter.
!>        - r           : Working storage (grid point coordinates).
!>        - w           : Working storage (grid point weights).
! **************************************************************************************************
MODULE lebedev

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   SAVE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lebedev'

   INTEGER, PARAMETER  :: na1 = 6, &
                          na2 = 12, &
                          na3 = 8, &
                          nb = 24, &
                          nc = 24, &
                          nd = 48, &
                          nlg = 14, &
                          max_np = 36

! **************************************************************************************************
   TYPE oh_grid
      INTEGER                           :: l, n
      REAL(KIND=dp), DIMENSION(:), POINTER   :: w => NULL()
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: r => NULL()
   END TYPE oh_grid

   TYPE(oh_grid), DIMENSION(nlg), TARGET :: lebedev_grid

   REAL(KIND=dp) :: one, rs2, rs3, zero
   INTEGER  :: nlgp
   LOGICAL  :: init_lebedev_grids_done = .FALSE.

   REAL(KIND=dp), DIMENSION(max_np/3) :: w
   REAL(KIND=dp), DIMENSION(max_np)   :: r

   PUBLIC :: lebedev_grid

   PUBLIC :: deallocate_lebedev_grids, get_number_of_lebedev_grid, init_lebedev_grids

CONTAINS

! **************************************************************************************************
!> \brief  Get the number of the Lebedev grid, which has the requested
!>          angular momentum quantnum number l or size n.
!> \param l ...
!> \param n ...
!> \retval number_of_lebedev_grid ...
!> \date   05.05.99
!> \par Variables
!> \author Matthias Krack
!> \version 1.0
!> \note         - l : Minimal angular momentum quantum number of the requested Lebedev grid.
!> \note         - n : Minimal size of the requested Lebedev grid.
! **************************************************************************************************
   FUNCTION get_number_of_lebedev_grid(l, n) RESULT(number_of_lebedev_grid)

      INTEGER, INTENT(IN), OPTIONAL                      :: l, n
      INTEGER                                            :: number_of_lebedev_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'get_number_of_lebedev_grid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ll, nn
      INTEGER, DIMENSION(1)                              :: lgnum

      IF (.NOT. init_lebedev_grids_done) CALL init_lebedev_grids

      lgnum(1) = 0
      ALLOCATE (nn(SIZE(lebedev_grid, 1)), ll(SIZE(lebedev_grid, 1)))
      DO i = 1, SIZE(nn, 1)
         nn(i) = lebedev_grid(i)%n
         ll(i) = lebedev_grid(i)%l
      END DO
      IF (PRESENT(l)) THEN
!      lgnum(:) = MINLOC(lebedev_grid(:)%n,MASK=(lebedev_grid(:)%l >= l))
         lgnum(:) = MINLOC(nn(:), MASK=(ll(:) >= l))
      ELSE IF (PRESENT(n)) THEN
!      lgnum(:) = MINLOC(lebedev_grid(:)%n,MASK=(lebedev_grid(:)%n >= n))
         lgnum(:) = MINLOC(nn(:), MASK=(nn(:) >= n))
      ELSE
         CALL cp__b("common/lebedev.F",138,"*** ERROR in FUNCTION get_number_of_lebedev_grid ***")
      END IF

      IF (lgnum(1) == 0) THEN
         CALL cp__b("common/lebedev.F",142,"*** ERROR in FUNCTION get_number_of_lebedev_grid ***")
      ELSE
         number_of_lebedev_grid = lgnum(1)
      END IF
      DEALLOCATE (nn, ll)
   END FUNCTION get_number_of_lebedev_grid

! **************************************************************************************************
!> \brief  Load the Lebedev grid points of a subsystem.
!> \param subsystem ...
!> \param lgnum ...
!> \param np ...
!> \date    06.05.99
!> \par Variables
!>       - subsystem: String, which specifies the selected subsystem.
!>       - lgnum    : Number of the current Lebedev grid.
!>       - np       : Number of the nonredundant grid points.
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE load_sub_grid(subsystem, lgnum, np)
      CHARACTER(*), INTENT(IN)                           :: subsystem
      INTEGER, INTENT(IN)                                :: lgnum, np

      CHARACTER(len=*), PARAMETER :: routineN = 'load_sub_grid', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      REAL(KIND=dp)                                      :: x, y, z

!   *** Check argument values ***

      IF ((lgnum < 1) .OR. (lgnum > nlg)) THEN
         CALL cp__b("common/lebedev.F",174,"Argument value #2 (lgnum) is out of range")
      END IF

      IF ((np < 0) .OR. (np > max_np)) THEN
         CALL cp__b("common/lebedev.F",178,"Argument value #3 (np) is out of range")
      END IF

      SELECT CASE (subsystem)
      CASE ("A1")
         lebedev_grid(lgnum)%w(1:na1) = w(1)
         lebedev_grid(lgnum)%r(1:3, 1:na1) = &
            RESHAPE((/zero, zero, one, &
                      zero, zero, -one, &
                      zero, one, zero, &
                      zero, -one, zero, &
                      one, zero, zero, &
                      -one, zero, zero/), (/3, na1/))
         nlgp = na1
      CASE ("A2")
         lebedev_grid(lgnum)%w(nlgp+1:nlgp+na2) = w(1)
         lebedev_grid(lgnum)%r(1:3, nlgp+1:nlgp+na2) = &
            RESHAPE((/zero, rs2, rs2, &
                      zero, rs2, -rs2, &
                      zero, -rs2, rs2, &
                      zero, -rs2, -rs2, &
                      rs2, zero, rs2, &
                      rs2, zero, -rs2, &
                      -rs2, zero, rs2, &
                      -rs2, zero, -rs2, &
                      rs2, rs2, zero, &
                      rs2, -rs2, zero, &
                      -rs2, rs2, zero, &
                      -rs2, -rs2, zero/), (/3, na2/))
         nlgp = nlgp+na2
      CASE ("A3")
         lebedev_grid(lgnum)%w(nlgp+1:nlgp+na3) = w(1)
         lebedev_grid(lgnum)%r(1:3, nlgp+1:nlgp+na3) = &
            RESHAPE((/rs3, rs3, rs3, &
                      rs3, rs3, -rs3, &
                      rs3, -rs3, rs3, &
                      -rs3, rs3, rs3, &
                      rs3, -rs3, -rs3, &
                      -rs3, rs3, -rs3, &
                      -rs3, -rs3, rs3, &
                      -rs3, -rs3, -rs3/), (/3, na3/))
         nlgp = nlgp+na3
      CASE ("B")
         DO i = 1, np
            x = r(i)
            y = rs2*SQRT(one-x**2)
            lebedev_grid(lgnum)%w(nlgp+nb*(i-1)+1:nlgp+nb*i) = w(i)
            lebedev_grid(lgnum)%r(1:3, nlgp+nb*(i-1)+1:nlgp+nb*i) = &
               RESHAPE((/x, y, y, &
                         x, y, -y, &
                         x, -y, y, &
                         x, -y, -y, &
                         -x, y, y, &
                         -x, y, -y, &
                         -x, -y, y, &
                         -x, -y, -y, &
                         y, x, y, &
                         y, x, -y, &
                         -y, x, y, &
                         -y, x, -y, &
                         y, -x, y, &
                         y, -x, -y, &
                         -y, -x, y, &
                         -y, -x, -y, &
                         y, y, x, &
                         y, -y, x, &
                         -y, y, x, &
                         -y, -y, x, &
                         y, y, -x, &
                         y, -y, -x, &
                         -y, y, -x, &
                         -y, -y, -x/), (/3, nb/))
         END DO
         nlgp = nlgp+nb*np
      CASE ("C")
         DO i = 1, np
            x = r(i)
            y = SQRT(one-x**2)
            lebedev_grid(lgnum)%w(nlgp+nc*(i-1)+1:nlgp+nc*i) = w(i)
            lebedev_grid(lgnum)%r(1:3, nlgp+nc*(i-1)+1:nlgp+nc*i) = &
               RESHAPE((/x, y, zero, &
                         x, -y, zero, &
                         -x, y, zero, &
                         -x, -y, zero, &
                         y, x, zero, &
                         y, -x, zero, &
                         -y, x, zero, &
                         -y, -x, zero, &
                         x, zero, y, &
                         x, zero, -y, &
                         -x, zero, y, &
                         -x, zero, -y, &
                         y, zero, x, &
                         y, zero, -x, &
                         -y, zero, x, &
                         -y, zero, -x, &
                         zero, x, y, &
                         zero, x, -y, &
                         zero, -x, y, &
                         zero, -x, -y, &
                         zero, y, x, &
                         zero, y, -x, &
                         zero, -y, x, &
                         zero, -y, -x/), (/3, nc/))
         END DO
         nlgp = nlgp+nc*np
      CASE ("D")
         IF (MODULO(np, 3) == 0) THEN
            DO i = 1, np, 3
               j = (i+2)/3
               x = r(i)
               y = r(i+1)
               z = r(i+2)
               lebedev_grid(lgnum)%w(nlgp+nd*(j-1)+1:nlgp+nd*j) = w(j)
               lebedev_grid(lgnum)%r(1:3, nlgp+nd*(j-1)+1:nlgp+nd*j) = &
                  RESHAPE((/x, y, z, x, y, -z, x, -y, z, -x, y, z, &
                            x, -y, -z, -x, y, -z, -x, -y, z, -x, -y, -z, &
                            x, z, y, x, z, -y, x, -z, y, -x, z, y, &
                            x, -z, -y, -x, z, -y, -x, -z, y, -x, -z, -y, &
                            y, x, z, y, x, -z, y, -x, z, -y, x, z, &
                            y, -x, -z, -y, x, -z, -y, -x, z, -y, -x, -z, &
                            y, z, x, y, z, -x, y, -z, x, -y, z, x, &
                            y, -z, -x, -y, z, -x, -y, -z, x, -y, -z, -x, &
                            z, x, y, z, x, -y, z, -x, y, -z, x, y, &
                            z, -x, -y, -z, x, -y, -z, -x, y, -z, -x, -y, &
                            z, y, x, z, y, -x, z, -y, x, -z, y, x, &
                            z, -y, -x, -z, y, -x, -z, -y, x, -z, -y, -x/), (/3, nd/))
            END DO
         ELSE
            CALL cp__b("common/lebedev.F",307,"Subsytem D: np is not modulo 3 (check argument #3)")
         END IF
         nlgp = nlgp+nd*np/3
      CASE DEFAULT
         CALL cp_abort(cp__l("common/lebedev.F",311), &
                       "The invalid subsystem <"//TRIM(subsystem)//"> was "// &
                       "specified (check argument #1)")
      END SELECT

!   *** Reinitialize working storage ***

      w(:) = zero
      r(:) = zero

   END SUBROUTINE load_sub_grid

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE deallocate_lebedev_grids()

      INTEGER                                            :: ilg

      DO ilg = 1, nlg
         DEALLOCATE (lebedev_grid(ilg)%r, lebedev_grid(ilg)%w)
      END DO

   END SUBROUTINE deallocate_lebedev_grids

! **************************************************************************************************
!> \brief Load the coordinates and weights of the nonredundant Lebedev grid
!>         points.
!> \date    04.05.99
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_lebedev_grids()

      CHARACTER(len=*), PARAMETER :: routineN = 'init_lebedev_grids', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ilg

      CALL timeset(routineN, handle)

      DO ilg = 1, nlg
         IF (ASSOCIATED(lebedev_grid(ilg)%r)) DEALLOCATE (lebedev_grid(ilg)%r)
         IF (ASSOCIATED(lebedev_grid(ilg)%w)) DEALLOCATE (lebedev_grid(ilg)%w)
      END DO

!   *** Initialize constants ***

      one = 1.0_dp
      rs2 = one/SQRT(2.0_dp)
      rs3 = one/SQRT(3.0_dp)
      zero = 0.0_dp

!   *** Initialize working storage ***

      w(:) = zero
      r(:) = zero

!   *** Load the angular momentum quantum numbers l of the Lebedev grids ***

      lebedev_grid(1:nlg)%l = (/3, 5, 7, 9, 11, 15, 17, 19, 23, 29, 35, 41, 47, 53/)

!   *** Load the total number of grid points for each Lebedev grid ***

      lebedev_grid(1:nlg)%n = (/6, 14, 26, 38, 50, 86, 110, 146, 194, 302, 434, 590, 770, &
                                974/)

!   *** Allocate storage for the Lebedev grids ***

      DO ilg = 1, nlg
         ALLOCATE (lebedev_grid(ilg)%r(3, lebedev_grid(ilg)%n), &
                   lebedev_grid(ilg)%w(lebedev_grid(ilg)%n))
      END DO

!   *** Load the coordinates and weights of the Lebedev grid points ***

!   *** 1. l = 3 (6 points, octahedron) ***

      w(1) = 1.0_dp/6.0_dp
      CALL load_sub_grid("A1", 1, 0)

!   *** 2. l = 5 (14 points, capped octahedron) ***

      w(1) = 1.0_dp/15.0_dp
      CALL load_sub_grid("A1", 2, 0)

      w(1) = 3.0_dp/40.0_dp
      CALL load_sub_grid("A3", 2, 0)

!   *** 3. l = 7 (26 points) ***

      w(1) = 1.0_dp/21.0_dp
      CALL load_sub_grid("A1", 3, 0)

      w(1) = 4.0_dp/105.0_dp
      CALL load_sub_grid("A2", 3, 0)

      w(1) = 9.0_dp/280.0_dp
      CALL load_sub_grid("A3", 3, 0)

!   *** 4. l = 9 (38 points) ***

      w(1) = 1.0_dp/105.0_dp
      CALL load_sub_grid("A1", 4, 0)

      w(1) = 9.0_dp/280.0_dp
      CALL load_sub_grid("A3", 4, 0)

      w(1) = 1.0_dp/35.0_dp
      r(1) = rs2*SQRT(1.0_dp+rs3)
      CALL load_sub_grid("C", 4, 1)

!   *** 5. l = 11 (50 points) ***

      w(1) = 4.0_dp/315.0_dp
      CALL load_sub_grid("A1", 5, 0)

      w(1) = 64.0_dp/2835.0_dp
      CALL load_sub_grid("A2", 5, 0)

      w(1) = 27.0_dp/1280.0_dp
      CALL load_sub_grid("A3", 5, 0)

      w(1) = 14641.0_dp/725760.0_dp
      r(1) = 3.0_dp/SQRT(11.0_dp)
      CALL load_sub_grid("B", 5, 1)

!   *** 6. l = 15 (86 points) ***

      w(1) = 1.15440115440115440115440115440115E-2_dp
      CALL load_sub_grid("A1", 6, 0)

      w(1) = 1.19439090858562823236989259736470E-2_dp
      CALL load_sub_grid("A3", 6, 0)

      w(1) = 1.11105557106034025109468482160140E-2_dp
      r(1) = 8.52518311701267605338736780155357E-1_dp
      w(2) = 1.18765012945371420137882805994025E-2_dp
      r(2) = 1.89063552885395482707075847005288E-1_dp
      CALL load_sub_grid("B", 6, 2)

      w(1) = 1.18123037469044753644792263073650E-2_dp
      r(1) = 9.27330657151172465678969739310097E-1_dp
      CALL load_sub_grid("C", 6, 1)

!   *** 7. l = 17 (110 points) ***

      w(1) = 3.82827049493716160382827049493716E-3_dp
      CALL load_sub_grid("A1", 7, 0)

      w(1) = 9.79373751248751248751248751248751E-3_dp
      CALL load_sub_grid("A3", 7, 0)

      w(1) = 8.21173728319111097598993405227308E-3_dp
      r(1) = 9.65124035086594105655529546158531E-1_dp
      w(2) = 9.59547133607096284945318117290259E-3_dp
      r(2) = 8.28769981252592210694031500711749E-1_dp
      w(3) = 9.94281489117810328140065828526450E-3_dp
      r(3) = 2.15957291845848832354930328945946E-1_dp
      CALL load_sub_grid("B", 7, 3)

      w(1) = 9.69499636166302832969499636166303E-3_dp
      r(1) = 8.78158910604066133449110592671618E-1_dp
      CALL load_sub_grid("C", 7, 1)

!   *** 8. l = 19 (146 points) ***

      w(1) = 5.99631368862138092907323676554446E-4_dp
      CALL load_sub_grid("A1", 8, 0)

      w(1) = 7.37299971862075642305743268410561E-3_dp
      CALL load_sub_grid("A2", 8, 0)

      w(1) = 7.21051536014448777763305996834282E-3_dp
      CALL load_sub_grid("A3", 8, 0)

      w(1) = 7.57439415905403372268748574713806E-3_dp
      r(1) = 9.74888643677173235480043489928723E-1_dp
      w(2) = 6.75382948631447744073541732486459E-3_dp
      r(2) = 8.07089818359582501629574989264033E-1_dp
      w(3) = 7.11635549311755538760089284953968E-3_dp
      r(3) = 2.91298882209526746288335878313128E-1_dp
      CALL load_sub_grid("B", 8, 3)

      w(1) = 6.99108735330326239417148508057599E-3_dp
      r(1) = 1.40355381171318328571556780746292E-1_dp
      r(2) = 4.49332832326955734884695382705101E-1_dp
      r(3) = 8.82270011260322631916665753046583E-1_dp
      CALL load_sub_grid("D", 8, 3)

!   *** 9. l = 23 (194 points) ***

      w(1) = 1.78234044724461115736727104869868E-3_dp
      CALL load_sub_grid("A1", 9, 0)

      w(1) = 5.71690594997710189299212838832099E-3_dp
      CALL load_sub_grid("A2", 9, 0)

      w(1) = 5.57338317884873796836784958446647E-3_dp
      CALL load_sub_grid("A3", 9, 0)

      w(1) = 5.51877146727361369172768460119380E-3_dp
      r(1) = 7.77493219314767127213777704028801E-1_dp
      w(2) = 5.15823771180538310324916154718792E-3_dp
      r(2) = 9.12509096867473692992168387555434E-1_dp
      w(3) = 5.60870408258799684374936673855184E-3_dp
      r(3) = 3.14196994182586079225390955831946E-1_dp
      w(4) = 4.10677702816939409072861128564582E-3_dp
      r(4) = 9.82972302707253296863729593965148E-1_dp
      CALL load_sub_grid("B", 9, 4)

      w(1) = 5.05184606461480847598931196006390E-3_dp
      r(1) = 9.38319218137591520905616389195671E-1_dp
      CALL load_sub_grid("C", 9, 1)

      w(1) = 5.53024891623309370129768269143303E-3_dp
      r(1) = 1.59041710538352952424263581362096E-1_dp
      r(2) = 5.25118572443642024905268207753317E-1_dp
      r(3) = 8.36036015482458885943746377793309E-1_dp
      CALL load_sub_grid("D", 9, 3)

!   *** 10. l = 29 (302 points) ***

      w(1) = 8.54591172512814813423121032618880E-4_dp
      CALL load_sub_grid("A1", 10, 0)

      w(1) = 3.59911928502557145886397858961119E-3_dp
      CALL load_sub_grid("A3", 10, 0)

      w(1) = 3.65004580767725542865433220112651E-3_dp
      r(1) = 1.29238672710514925339493976600550E-1_dp
      w(2) = 3.60482260141988171131480913104353E-3_dp
      r(2) = 3.71034178384821189405344300671593E-1_dp
      w(3) = 3.57672966174336707556208137560857E-3_dp
      r(3) = 7.43452042987555751133204566760209E-1_dp
      w(4) = 3.44978842430588331001302771048283E-3_dp
      r(4) = 8.67643624544083327473890295837421E-1_dp
      w(5) = 3.10895312241367525484587698082774E-3_dp
      r(5) = 9.49454317226443084214869821724903E-1_dp
      w(6) = 2.35210141368916437879217118337424E-3_dp
      r(6) = 9.90705621379408123821774729602922E-1_dp
      CALL load_sub_grid("B", 10, 6)

      w(1) = 3.60082093221646027279920634177099E-3_dp
      r(1) = 8.20326419827759303328870367500864E-1_dp
      w(2) = 2.98234496317180385195111046924520E-3_dp
      r(2) = 9.64408914879206014987053264857406E-1_dp
      CALL load_sub_grid("C", 10, 2)

      w(1) = 3.57154055427338708123297920312395E-3_dp
      r(1) = 2.51003475177046506904110494820777E-1_dp
      r(2) = 5.44867737258077380269062623720797E-1_dp
      r(3) = 8.00072749407395172686232883863933E-1_dp
      w(2) = 3.39231220500617018197882653945702E-3_dp
      r(4) = 1.23354853258332742165467555277499E-1_dp
      r(5) = 4.12772408316853095996383989497138E-1_dp
      r(6) = 9.02442529533000401095678759353334E-1_dp
      CALL load_sub_grid("D", 10, 6)

!   *** 11. l = 35 (434 points) ***

      w(1) = 5.26589796822443623921598397444095E-4_dp
      CALL load_sub_grid("A1", 11, 0)

      w(1) = 2.54821997200260718024899528063707E-3_dp
      CALL load_sub_grid("A2", 11, 0)

      w(1) = 2.51231741892730716751285677008112E-3_dp
      CALL load_sub_grid("A3", 11, 0)

      w(1) = 1.46249562159461384222611198788353E-3_dp
      r(1) = 9.94255912631277812357804047239411E-1_dp
      w(2) = 2.01427902091852819554717351102614E-3_dp
      r(2) = 9.67987158791472796844440561853288E-1_dp
      w(3) = 2.30269478222741577118243755624539E-3_dp
      r(3) = 9.14472801120872429366691177286487E-1_dp
      w(4) = 2.44537343731298000188170923067793E-3_dp
      r(4) = 8.31584400419232294499745594669204E-1_dp
      w(5) = 2.50172516840293614549327311336461E-3_dp
      r(5) = 7.19016501040843432905788800294849E-1_dp
      w(6) = 2.51326717459756436118956600426018E-3_dp
      r(6) = 4.07712664897769512483055658274718E-1_dp
      w(7) = 2.53040380118635500209853222454729E-3_dp
      r(7) = 2.12646824707552073524018894972799E-1_dp
      CALL load_sub_grid("B", 11, 7)

      w(1) = 1.91095128217953227363116198076372E-3_dp
      r(1) = 9.77642811118264871364737333213224E-1_dp
      w(2) = 2.41744237563898077608784837021936E-3_dp
      r(2) = 8.81813287779428812899447039158355E-1_dp
      CALL load_sub_grid("C", 11, 2)

      w(1) = 2.23660776043784866397235325901501E-3_dp
      r(1) = 9.92176963642923726861469072682696E-2_dp
      r(2) = 3.34436314534345491435541212141297E-1_dp
      r(3) = 9.37180985855372235789161206437430E-1_dp
      w(2) = 2.41693004432477530950382860647420E-3_dp
      r(4) = 2.05482369640304370826435780923978E-1_dp
      r(5) = 4.50233038258262537898138529993494E-1_dp
      r(6) = 8.68946032287241197946412945028781E-1_dp
      w(3) = 2.51223685456349510228853571762152E-3_dp
      r(7) = 1.06801826075804827808519981079378E-1_dp
      r(8) = 5.90515704892527108597639202779953E-1_dp
      r(9) = 7.99927854385728570494616568630393E-1_dp
      w(4) = 2.49664405455308596359511734587523E-3_dp
      r(10) = 3.10428403516654146818973681387686E-1_dp
      r(11) = 5.55015236107680716116395510647568E-1_dp
      r(12) = 7.71746262691590088133470564383333E-1_dp
      CALL load_sub_grid("D", 11, 12)

!   *** 12. l = 41 (590 points) ***

      w(1) = 3.09512129530618734224885916984167E-4_dp
      CALL load_sub_grid("A1", 12, 0)

      w(1) = 1.85237969859748902097779927120955E-3_dp
      CALL load_sub_grid("A3", 12, 0)

      w(1) = 9.76433116505105003063946798281256E-4_dp
      r(1) = 9.96278129754016372294779397439010E-1_dp
      w(2) = 1.38473723485169190049558481046633E-3_dp
      r(2) = 9.78480583762693839428807199465825E-1_dp
      w(3) = 1.61721064725441119207240564027320E-3_dp
      r(3) = 9.41414158220402542954812123169147E-1_dp
      w(4) = 1.74956465728115411858337657547711E-3_dp
      r(4) = 8.83078727934132546639472279693012E-1_dp
      w(5) = 1.81847177816276877817839645350131E-3_dp
      r(5) = 8.02836877335273789544708232847842E-1_dp
      w(6) = 1.84671595615124182496431643927999E-3_dp
      r(6) = 7.00768575373572920755474238992223E-1_dp
      w(7) = 1.85202882829621309713188574624172E-3_dp
      r(7) = 4.33373868777154278470427918169993E-1_dp
      w(8) = 1.85881258543831701688713685607976E-3_dp
      r(8) = 2.70356088359165054450644840302674E-1_dp
      w(9) = 1.87179063927774375088198703268059E-3_dp
      r(9) = 9.21904070768989460773325829762845E-2_dp
      CALL load_sub_grid("B", 12, 9)

      w(1) = 1.30032168588604773254689500223665E-3_dp
      r(1) = 9.85013335028001910431569760301413E-1_dp
      w(2) = 1.70515399639586401999426890257133E-3_dp
      r(2) = 9.18045287711453949033261544688478E-1_dp
      w(3) = 1.85716119677407798248150549753942E-3_dp
      r(3) = 7.91101929626901988791072435283520E-1_dp
      CALL load_sub_grid("C", 12, 3)

      w(1) = 1.55521360339680849695800543399046E-3_dp
      r(1) = 8.21302158193251139256011433978433E-2_dp
      r(2) = 2.77867319058624428710339115104831E-1_dp
      r(3) = 9.57102074310072578548507660672158E-1_dp
      w(2) = 1.80223912800852549116687840059028E-3_dp
      r(4) = 8.99920584207487492736600381248490E-2_dp
      r(5) = 5.03356427107511721798695016575784E-1_dp
      r(6) = 8.59379855890721226547417613471625E-1_dp
      w(3) = 1.84983056044366016523678105066982E-3_dp
      r(7) = 1.81664084036020946133369276322890E-1_dp
      r(8) = 5.98412649788537962548021778158698E-1_dp
      r(9) = 7.80320742479920327279647082439668E-1_dp
      w(4) = 1.71390450710670866804340083221112E-3_dp
      r(10) = 1.72079522565687812382959749101837E-1_dp
      r(11) = 3.79103540769556328153581995776882E-1_dp
      r(12) = 9.09213475092373607255560905376387E-1_dp
      w(5) = 1.80265893437745115894760427403985E-3_dp
      r(13) = 2.63471665593794963182134911777324E-1_dp
      r(14) = 4.74239284255198024276965888711084E-1_dp
      r(15) = 8.40047488359050429417090817376105E-1_dp
      w(6) = 1.84286647290528563239873317969286E-3_dp
      r(16) = 3.51828092773351899746468433493901E-1_dp
      r(17) = 5.61026380862206018993247924039616E-1_dp
      r(18) = 7.49310611904115932021924350883623E-1_dp
      CALL load_sub_grid("D", 12, 18)

!   *** 13. l = 47 (770 points) ***

      w(1) = 2.19294208818118413191132531546907E-4_dp
      CALL load_sub_grid("A1", 13, 0)

      w(1) = 1.43643361731907982089311290725143E-3_dp
      CALL load_sub_grid("A2", 13, 0)

      w(1) = 1.42194034433587736471699229343586E-3_dp
      CALL load_sub_grid("A3", 13, 0)

      w(1) = 6.79812351105050201620332392505941E-4_dp
      r(1) = 9.97408677652823043744886496130338E-1_dp
      w(2) = 9.91318423529491220896771270468706E-4_dp
      r(2) = 9.84799753572301179668683394920869E-1_dp
      w(3) = 1.18020783323894879580204544917243E-3_dp
      r(3) = 9.58036675983391439732458619060098E-1_dp
      w(4) = 1.29659960208092067423594734151509E-3_dp
      r(4) = 9.15317950483154831654282453968846E-1_dp
      w(5) = 1.36587142742831640181200652406724E-3_dp
      r(5) = 8.55901928697886435270454067917142E-1_dp
      w(6) = 1.40298860477532526267417750582642E-3_dp
      r(6) = 7.79621319527635171330859492071650E-1_dp
      w(7) = 1.41864556359560936142032713234525E-3_dp
      r(7) = 6.86644447264154195285405155618666E-1_dp
      w(8) = 1.42137674185166176103362636565150E-3_dp
      r(8) = 4.52311920313658441442123883818590E-1_dp
      w(9) = 1.42399647549096162940277119054908E-3_dp
      r(9) = 3.12521305001653125265256223110153E-1_dp
      w(10) = 1.43155404217856675285551659613462E-3_dp
      r(10) = 1.60155803498828974610387120985563E-1_dp
      CALL load_sub_grid("B", 13, 10)

      w(1) = 9.25440149986536789398468419487000E-4_dp
      r(1) = 9.89477537495598496934466234590453E-1_dp
      w(2) = 1.25023999505350931588903916766368E-3_dp
      r(2) = 9.40776878793758755393692765525837E-1_dp
      w(3) = 1.39436584332923012313782491233921E-3_dp
      r(3) = 8.45749305193653306776973233373898E-1_dp
      CALL load_sub_grid("C", 13, 3)

      w(1) = 1.12708909467174883469499826293069E-3_dp
      r(1) = 6.94402439334941301856132689108761E-2_dp
      r(2) = 2.35518789424232641745294975361896E-1_dp
      r(3) = 9.69385863498432080572651262482050E-1_dp
      w(2) = 1.34575376091067007284708159178371E-3_dp
      r(4) = 2.26900410952945985812144550510454E-1_dp
      r(5) = 4.10218247404573022502091360294078E-1_dp
      r(6) = 8.83310360522112719095559130461882E-1_dp
      w(3) = 1.42495728331678280442457616193820E-3_dp
      r(7) = 8.02557460777533890494192154117455E-2_dp
      r(8) = 6.21430241748160458082227723944198E-1_dp
      r(9) = 7.79348105702660973885784203335648E-1_dp
      w(4) = 1.26152334123774998512114871120604E-3_dp
      r(10) = 1.46799952789657197013820141783396E-1_dp
      r(11) = 3.24528434571739443574137218155171E-1_dp
      r(12) = 9.34414827052402165196138643907493E-1_dp
      w(5) = 1.39254710605269594012443620424449E-3_dp
      r(13) = 1.57150776982472710564392240262888E-1_dp
      r(14) = 5.22448218969662967424959711591943E-1_dp
      r(15) = 8.38064133458312497889043403162912E-1_dp
      w(6) = 1.41876167787765638610216672064623E-3_dp
      r(16) = 2.36570299315724563080851921006450E-1_dp
      r(17) = 6.01754663408955808458505175650313E-1_dp
      r(18) = 7.62840624604669826410820436887813E-1_dp
      w(7) = 1.33836668447955413932118842030907E-3_dp
      r(19) = 7.71481586676573204146393256189282E-2_dp
      r(20) = 4.34657551614116278920169057936696E-1_dp
      r(21) = 8.97285336132833396499244991766618E-1_dp
      w(8) = 1.39370086267613137196129133534721E-3_dp
      r(22) = 3.06293666621073020830889612421600E-1_dp
      r(23) = 4.90882658903761621546950183900134E-1_dp
      r(24) = 8.15609223203975422051589658409773E-1_dp
      w(9) = 1.41591475746693200265535028223575E-3_dp
      r(25) = 3.82247737952478700050265491776632E-1_dp
      r(26) = 5.64876814909950046651375249432877E-1_dp
      r(27) = 7.31300793659765701947278263710234E-1_dp
      CALL load_sub_grid("D", 13, 27)

!   *** 14. l = 53 (974 points) ***

      w(1) = 1.43829419052743111472634384792800E-4_dp
      CALL load_sub_grid("A1", 14, 0)

      w(1) = 1.12577228828700411922446003897444E-3_dp
      CALL load_sub_grid("A3", 14, 0)

      w(1) = 4.94802934194924095056628951683154E-4_dp
      r(1) = 9.98155345023846501191494527575073E-1_dp
      w(2) = 7.35799010912547049057334785150092E-4_dp
      r(2) = 9.88883224354685545466876075112182E-1_dp
      w(3) = 8.88913277130438430936840534873175E-4_dp
      r(3) = 9.68890220434707404463853311788372E-1_dp
      w(4) = 9.88834783892143487490379018339493E-4_dp
      r(4) = 9.36602730407163187902332491169310E-1_dp
      w(5) = 1.05329968170947064968554154238842E-3_dp
      r(5) = 8.91267942647606054508357503868536E-1_dp
      w(6) = 1.09277880701457848752501645357106E-3_dp
      r(6) = 8.32596723702351872809065001851482E-1_dp
      w(7) = 1.11438939406322716311544826926779E-3_dp
      r(7) = 7.60582905315251495676259855686005E-1_dp
      w(8) = 1.12372478805155528926212353749792E-3_dp
      r(8) = 6.75400969108414309734349145075511E-1_dp
      w(9) = 1.12523932524381359890224350348300E-3_dp
      r(9) = 4.66858905695743283152769352096062E-1_dp
      w(10) = 1.12615327181590500171763113615557E-3_dp
      r(10) = 3.44613654237438223690005816795002E-1_dp
      w(11) = 1.13028693112384079609639852411886E-3_dp
      r(11) = 2.11954151850184645694140401473565E-1_dp
      w(12) = 1.13498653436395488643714589963293E-3_dp
      r(12) = 7.16244014499556615845369803814026E-2_dp
      CALL load_sub_grid("B", 14, 12)

      w(1) = 6.82336792710993096637138363442284E-4_dp
      r(1) = 9.92323565431490196070394244122238E-1_dp
      w(2) = 9.45415816044709583567874030695969E-4_dp
      r(2) = 9.55781512496548488574596747552425E-1_dp
      w(3) = 1.07442997538567914061819358161104E-3_dp
      r(3) = 8.82785980701181710576015026411856E-1_dp
      w(4) = 1.12930008656913165822354345410170E-3_dp
      r(4) = 7.73778447257374736807517446423091E-1_dp
      CALL load_sub_grid("C", 14, 4)

      w(1) = 8.43688450090195438498026316671055E-4_dp
      r(1) = 5.97404861418134181366741615801962E-2_dp
      r(2) = 2.02912875277752280821562682821390E-1_dp
      r(3) = 9.77372722845309994986723236843860E-1_dp
      w(2) = 1.07525572044888463296011392358171E-3_dp
      r(4) = 1.37576040847363647998707229269443E-1_dp
      r(5) = 4.60262194248405393969672054729315E-1_dp
      r(6) = 8.77058461865802689033970747916718E-1_dp
      w(3) = 1.10857723686446203483870096411918E-3_dp
      r(7) = 3.39101652633628571879780724877562E-1_dp
      r(8) = 5.03067399966203571325278413235681E-1_dp
      r(9) = 7.94942299964208514029117644252761E-1_dp
      w(4) = 9.56647532378335729987022997768218E-4_dp
      r(10) = 1.27167519143981950887694563391455E-1_dp
      r(11) = 2.81760642244213432084049569237994E-1_dp
      r(12) = 9.51020169374389952612600596224493E-1_dp
      w(5) = 1.08066325071739071464463316273343E-3_dp
      r(13) = 2.69312074041351249441688496163892E-1_dp
      r(14) = 4.33156129172015739916409763707103E-1_dp
      r(15) = 8.60143461601761963548074094490711E-1_dp
      w(6) = 1.12679713119629459182839576027292E-3_dp
      r(16) = 1.41978645260191825726693154370171E-1_dp
      r(17) = 6.25616735858081415467234398457286E-1_dp
      r(18) = 7.67102186220558360145428572382623E-1_dp
      w(7) = 1.02256871535806121135514702022316E-3_dp
      r(19) = 6.70928460073825495111114896527700E-2_dp
      r(20) = 3.79839521685915669705644712365302E-1_dp
      r(21) = 9.22616110730809020138255560099254E-1_dp
      w(8) = 1.10896026771310753764243318312336E-3_dp
      r(22) = 7.05773818325617225225155364651637E-2_dp
      r(23) = 5.51750542142352049132113965320893E-1_dp
      r(24) = 8.31017552413474225750006298786621E-1_dp
      w(9) = 1.12279065343576582686387381292198E-3_dp
      r(25) = 2.78388847788215460217462722823059E-1_dp
      r(26) = 6.02961915615918685584436745852250E-1_dp
      r(27) = 7.47620610834085792324588316104014E-1_dp
      w(10) = 1.03240184711745983700951101978720E-3_dp
      r(28) = 1.97957893891740690574322079335028E-1_dp
      r(29) = 3.58960632958909578333772677316348E-1_dp
      r(30) = 9.12118378409121483660044157731511E-1_dp
      w(11) = 1.10724938228385393683388780275631E-3_dp
      r(31) = 2.08730706110327399856945819943227E-1_dp
      r(32) = 5.34866643813547649971323458627733E-1_dp
      r(33) = 8.18748536281021806431067610000242E-1_dp
      w(12) = 1.12178004851997205823535727177375E-3_dp
      r(34) = 4.05512213787283588596898056051174E-1_dp
      r(35) = 5.67499754607437348401393912269671E-1_dp
      r(36) = 7.16591845467023718833743633176626E-1_dp
      CALL load_sub_grid("D", 14, 36)

      init_lebedev_grids_done = .TRUE.

      CALL timestop(handle)

   END SUBROUTINE init_lebedev_grids

END MODULE lebedev
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE qs_grid_atom

   USE input_constants,                 ONLY: do_gapw_gcs,&
                                              do_gapw_gct,&
                                              do_gapw_log
   USE kinds,                           ONLY: dp
   USE lebedev,                         ONLY: lebedev_grid
   USE mathconstants,                   ONLY: pi
   USE memory_utilities,                ONLY: reallocate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_grid_atom'

   TYPE grid_atom_type

      INTEGER                         :: nr, ng_sphere
      REAL(dp), DIMENSION(:), POINTER :: rad, rad2, &
                                         wr, wa, &
                                         azi, cos_azi, sin_azi, cotan_pol, &
                                         pol, cos_pol, sin_pol, usin_azi
      REAL(dp), DIMENSION(:, :), &
         POINTER :: rad2l, oorad2l, weight

   END TYPE grid_atom_type

   PUBLIC :: allocate_grid_atom, &
             create_grid_atom, &
             deallocate_grid_atom

   PUBLIC :: grid_atom_type

CONTAINS

! **************************************************************************************************
!> \brief   Deallocate a Gaussian-type orbital (GTO) basis set data set.
!> \param grid_atom ...
!> \date    03.11.2000
!> \author  MK
!> \author Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_grid_atom(grid_atom)

      TYPE(grid_atom_type), POINTER                      :: grid_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_grid_atom', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(grid_atom)) CALL deallocate_grid_atom(grid_atom)

      ALLOCATE (grid_atom)

      NULLIFY (grid_atom%rad)
      NULLIFY (grid_atom%rad2)
      NULLIFY (grid_atom%wr)
      NULLIFY (grid_atom%wa)
      NULLIFY (grid_atom%weight)
      NULLIFY (grid_atom%azi)
      NULLIFY (grid_atom%cos_azi)
      NULLIFY (grid_atom%sin_azi)
      NULLIFY (grid_atom%cotan_pol)
      NULLIFY (grid_atom%pol)
      NULLIFY (grid_atom%cos_pol)
      NULLIFY (grid_atom%sin_pol)
      NULLIFY (grid_atom%usin_azi)
      NULLIFY (grid_atom%rad2l)
      NULLIFY (grid_atom%oorad2l)

   END SUBROUTINE allocate_grid_atom

! **************************************************************************************************
!> \brief   Deallocate a Gaussian-type orbital (GTO) basis set data set.
!> \param grid_atom ...
!> \date    03.11.2000
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_grid_atom(grid_atom)
      TYPE(grid_atom_type), POINTER                      :: grid_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_grid_atom', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(grid_atom)) THEN

         IF (ASSOCIATED(grid_atom%rad)) THEN
            DEALLOCATE (grid_atom%rad)
         END IF

         IF (ASSOCIATED(grid_atom%rad2)) THEN
            DEALLOCATE (grid_atom%rad2)
         END IF

         IF (ASSOCIATED(grid_atom%wr)) THEN
            DEALLOCATE (grid_atom%wr)
         END IF

         IF (ASSOCIATED(grid_atom%wa)) THEN
            DEALLOCATE (grid_atom%wa)
         END IF

         IF (ASSOCIATED(grid_atom%weight)) THEN
            DEALLOCATE (grid_atom%weight)
         END IF

         IF (ASSOCIATED(grid_atom%azi)) THEN
            DEALLOCATE (grid_atom%azi)
         END IF

         IF (ASSOCIATED(grid_atom%cos_azi)) THEN
            DEALLOCATE (grid_atom%cos_azi)
         END IF

         IF (ASSOCIATED(grid_atom%sin_azi)) THEN
            DEALLOCATE (grid_atom%sin_azi)
         END IF

         IF (ASSOCIATED(grid_atom%cotan_pol)) THEN
            DEALLOCATE (grid_atom%cotan_pol)
         END IF

         IF (ASSOCIATED(grid_atom%pol)) THEN
            DEALLOCATE (grid_atom%pol)
         END IF

         IF (ASSOCIATED(grid_atom%cos_pol)) THEN
            DEALLOCATE (grid_atom%cos_pol)
         END IF

         IF (ASSOCIATED(grid_atom%sin_pol)) THEN
            DEALLOCATE (grid_atom%sin_pol)
         END IF

         IF (ASSOCIATED(grid_atom%usin_azi)) THEN
            DEALLOCATE (grid_atom%usin_azi)
         END IF

         IF (ASSOCIATED(grid_atom%rad2l)) THEN
            DEALLOCATE (grid_atom%rad2l)
         END IF

         IF (ASSOCIATED(grid_atom%oorad2l)) THEN
            DEALLOCATE (grid_atom%oorad2l)
         END IF

         DEALLOCATE (grid_atom)
      ELSE
         CALL cp_abort(cp__l("qs_grid_atom.F",157), &
                       "The pointer grid_atom is not associated and "// &
                       "cannot be deallocated")
      END IF
   END SUBROUTINE deallocate_grid_atom

! **************************************************************************************************
!> \brief ...
!> \param grid_atom ...
!> \param nr ...
!> \param na ...
!> \param llmax ...
!> \param ll ...
!> \param quadrature ...
! **************************************************************************************************
   SUBROUTINE create_grid_atom(grid_atom, nr, na, llmax, ll, quadrature)

      TYPE(grid_atom_type), POINTER                      :: grid_atom
      INTEGER, INTENT(IN)                                :: nr, na, llmax, ll, quadrature

      CHARACTER(len=*), PARAMETER :: routineN = 'create_grid_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ia, ir, l
      REAL(dp)                                           :: cosia, pol
      REAL(dp), DIMENSION(:), POINTER                    :: rad, rad2, wr

      CALL timeset(routineN, handle)

      NULLIFY (rad, rad2, wr)

      IF (ASSOCIATED(grid_atom)) THEN

         ! Allocate the radial grid arrays
         CALL reallocate(grid_atom%rad, 1, nr)
         CALL reallocate(grid_atom%rad2, 1, nr)
         CALL reallocate(grid_atom%wr, 1, nr)
         CALL reallocate(grid_atom%wa, 1, na)
         CALL reallocate(grid_atom%weight, 1, na, 1, nr)
         CALL reallocate(grid_atom%azi, 1, na)
         CALL reallocate(grid_atom%cos_azi, 1, na)
         CALL reallocate(grid_atom%sin_azi, 1, na)
         CALL reallocate(grid_atom%cotan_pol, 1, na)
         CALL reallocate(grid_atom%pol, 1, na)
         CALL reallocate(grid_atom%cos_pol, 1, na)
         CALL reallocate(grid_atom%sin_pol, 1, na)
         CALL reallocate(grid_atom%usin_azi, 1, na)
         CALL reallocate(grid_atom%rad2l, 1, nr, 0, llmax+1)
         CALL reallocate(grid_atom%oorad2l, 1, nr, 0, llmax+1)

         ! Calculate the radial grid for this kind
         rad => grid_atom%rad
         rad2 => grid_atom%rad2
         wr => grid_atom%wr

         CALL radial_grid(nr, rad, rad2, wr, quadrature)

         grid_atom%rad2l(:, 0) = 1._dp
         grid_atom%oorad2l(:, 0) = 1._dp
         DO l = 1, llmax+1
            grid_atom%rad2l(:, l) = grid_atom%rad2l(:, l-1)*rad(:)
            grid_atom%oorad2l(:, l) = grid_atom%oorad2l(:, l-1)/rad(:)
         ENDDO

         IF (ll > 0) THEN
            grid_atom%wa(1:na) = 4._dp*pi*lebedev_grid(ll)%w(1:na)
            DO ir = 1, nr
               DO ia = 1, na
                  grid_atom%weight(ia, ir) = grid_atom%wr(ir)*grid_atom%wa(ia)
               END DO
            END DO

            DO ia = 1, na
               ! polar angle: pol = acos(r(3))
               cosia = lebedev_grid(ll)%r(3, ia)
               grid_atom%cos_pol(ia) = cosia
               ! azimuthal angle: pol = atan(r(2)/r(1))
               IF (ABS(lebedev_grid(ll)%r(2, ia)) < EPSILON(1.0_dp) .AND. &
                   ABS(lebedev_grid(ll)%r(1, ia)) < EPSILON(1.0_dp)) THEN
                  grid_atom%azi(ia) = 0.0_dp
               ELSE
                  grid_atom%azi(ia) = ATAN2(lebedev_grid(ll)%r(2, ia), lebedev_grid(ll)%r(1, ia))
               END IF
               grid_atom%cos_azi(ia) = COS(grid_atom%azi(ia))
               pol = ACOS(cosia)
               IF (grid_atom%sin_pol(ia) < 0.0_dp) pol = -pol
               grid_atom%pol(ia) = pol
               grid_atom%sin_pol(ia) = SIN(grid_atom%pol(ia))
               IF (ABS(grid_atom%sin_pol(ia)) > EPSILON(1.0_dp)) THEN
                  grid_atom%cotan_pol(ia) = &
                     grid_atom%cos_pol(ia)/grid_atom%sin_pol(ia)
               ELSE
                  grid_atom%cotan_pol(ia) = 0.0_dp
               END IF

               grid_atom%sin_azi(ia) = SIN(grid_atom%azi(ia))
               IF (ABS(grid_atom%sin_azi(ia)) > EPSILON(1.0_dp)) THEN
                  grid_atom%usin_azi(ia) = 1.0_dp/grid_atom%sin_azi(ia)
               ELSE
                  grid_atom%usin_azi(ia) = 1.0_dp
               END IF

            ENDDO

         END IF

      ELSE
         CALL cp__b("qs_grid_atom.F",264,"The pointer grid_atom is not associated")
      END IF

      CALL timestop(handle)

   END SUBROUTINE create_grid_atom

! **************************************************************************************************
!> \brief   Generate a radial grid with n points by a quadrature rule.
!> \param n ...
!> \param r ...
!> \param r2 ...
!> \param wr ...
!> \param radial_quadrature ...
!> \date    20.09.1999
!> \par Literature
!>           - A. D. Becke, J. Chem. Phys. 88, 2547 (1988)
!>           - J. M. Perez-Jorda, A. D. Becke and E. San-Fabian,
!>             J. Chem. Phys. 100, 6520 (1994)
!>           - M. Krack and A. M. Koester, J. Chem. Phys. 108, 3226 (1998)
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE radial_grid(n, r, r2, wr, radial_quadrature)

      INTEGER, INTENT(IN)                                :: n
      REAL(dp), DIMENSION(:), POINTER                    :: r, r2, wr
      INTEGER, INTENT(IN)                                :: radial_quadrature

      CHARACTER(len=*), PARAMETER :: routineN = 'radial_grid', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(dp)                                           :: cost, f, sint, sint2, t, w, x

      f = pi/REAL(n+1, dp)

      SELECT CASE (radial_quadrature)

      CASE (do_gapw_gcs)

         !     *** Gauss-Chebyshev quadrature formula of the second kind ***
         !     *** u [-1,+1] -> r [0,infinity] =>  r = (1 + u)/(1 - u)   ***

         DO i = 1, n
            t = REAL(i, dp)*f
            x = COS(t)
            w = f*SIN(t)**2
            r(i) = (1.0_dp+x)/(1.0_dp-x)
            r2(i) = r(i)**2
            wr(i) = w/SQRT(1.0_dp-x**2)
            wr(i) = 2.0_dp*wr(i)*r2(i)/(1.0_dp-x)**2
         END DO

      CASE (do_gapw_gct)

         !     *** Transformed Gauss-Chebyshev quadrature formula of the second kind ***
         !     *** u [-1,+1] -> r [0,infinity] => r = (1 + u)/(1 - u)                ***

         DO i = 1, n
            t = REAL(i, dp)*f
            cost = COS(t)
            sint = SIN(t)
            sint2 = sint**2
            x = REAL(2*i-n-1, dp)/REAL(n+1, dp)- &
                2.0_dp*(1.0_dp+2.0_dp*sint2/3.0_dp)*cost*sint/pi
            w = 16.0_dp*sint2**2/REAL(3*(n+1), dp)
            r(n+1-i) = (1.0_dp+x)/(1.0_dp-x)
            r2(n+1-i) = r(n+1-i)**2
            wr(n+1-i) = 2.0_dp*w*r2(n+1-i)/(1.0_dp-x)**2
         END DO

      CASE (do_gapw_log)

         !     *** Transformed Gauss-Chebyshev quadrature formula of the second kind ***
         !     *** u [-1,+1] -> r [0,infinity] => r = ln(2/(1 - u))/ln(2)            ***

         DO i = 1, n
            t = REAL(i, dp)*f
            cost = COS(t)
            sint = SIN(t)
            sint2 = sint**2
            x = REAL(2*i-n-1, dp)/REAL(n+1, dp)- &
                2.0_dp*(1.0_dp+2.0_dp*sint2/3.0_dp)*cost*sint/pi
            w = 16.0_dp*sint2**2/REAL(3*(n+1), dp)
            r(n+1-i) = LOG(2.0_dp/(1.0_dp-x))/LOG(2.0_dp)
            r2(n+1-i) = r(n+1-i)**2
            wr(n+1-i) = w*r2(n+1-i)/(LOG(2.0_dp)*(1.0_dp-x))
         END DO

      CASE DEFAULT

         CALL cp__b("qs_grid_atom.F",355,"Invalid radial quadrature type specified")

      END SELECT

   END SUBROUTINE radial_grid

END MODULE qs_grid_atom
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Handles all functions related to the CELL
!> \par History
!>      11.2008 Teodoro Laino [tlaino] - deeply cleaning cell_type from units
!>      10.2014 Moved many routines from cell_types.F here.
!> \author Matthias KracK (16.01.2002, based on a earlier version of CJM, JGH)
! **************************************************************************************************
MODULE cell_types
   USE cp_units,                        ONLY: cp_unit_to_cp2k,&
                                              cp_units_rad
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: degree,&
                                              sqrt3
   USE mathlib,                         ONLY: angle,&
                                              det_3x3,&
                                              inv_3x3
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cell_types'

   INTEGER, SAVE, PRIVATE :: last_cell_id = 0

   ! Impose cell symmetry
   INTEGER, PARAMETER, PUBLIC               :: cell_sym_none = 0, &
                                               cell_sym_triclinic = 1, &
                                               cell_sym_monoclinic = 2, &
                                               cell_sym_orthorhombic = 3, &
                                               cell_sym_tetragonal_ab = 4, &
                                               cell_sym_tetragonal_ac = 5, &
                                               cell_sym_tetragonal_bc = 6, &
                                               cell_sym_rhombohedral = 7, &
                                               cell_sym_hexagonal = 8, &
                                               cell_sym_cubic = 9

   INTEGER, PARAMETER, PUBLIC               :: use_perd_x = 0, &
                                               use_perd_y = 1, &
                                               use_perd_z = 2, &
                                               use_perd_xy = 3, &
                                               use_perd_xz = 4, &
                                               use_perd_yz = 5, &
                                               use_perd_xyz = 6, &
                                               use_perd_none = 7

! **************************************************************************************************
!> \brief   Type defining parameters related to the simulation cell
!> \version 1.0
! **************************************************************************************************
   TYPE cell_type
      INTEGER                           :: ref_count, id_nr
      INTEGER                           :: symmetry_id
      LOGICAL                           :: orthorhombic ! actually means a diagonal hmat
      REAL(KIND=dp)                   :: deth
      INTEGER, DIMENSION(3)             :: perd
      REAL(KIND=dp), DIMENSION(3, 3)   :: hmat, h_inv
   END TYPE cell_type

   TYPE cell_p_type
      TYPE(cell_type), POINTER :: cell
   END TYPE cell_p_type

   ! Public data types
   PUBLIC :: cell_type, cell_p_type

   ! Public subroutines
   PUBLIC :: get_cell, get_cell_param, init_cell, &
             cell_create, cell_retain, cell_release, &
             cell_clone, cell_copy, parse_cell_line, set_cell_param


   ! Public functions
   PUBLIC :: plane_distance, pbc, real_to_scaled, scaled_to_real

   INTERFACE pbc
      MODULE PROCEDURE pbc1, pbc2, pbc3, pbc4
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cell_in ...
!> \param cell_out ...
! **************************************************************************************************
   SUBROUTINE cell_clone(cell_in, cell_out)

      TYPE(cell_type), POINTER                           :: cell_in, cell_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cell_clone', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(cell_in)))CALL cp__a("subsys/cell_types.F",102)
      IF(.NOT.(ASSOCIATED(cell_out)))CALL cp__a("subsys/cell_types.F",103)

      cell_out%deth = cell_in%deth
      cell_out%perd = cell_in%perd
      cell_out%hmat = cell_in%hmat
      cell_out%h_inv = cell_in%h_inv
      cell_out%orthorhombic = cell_in%orthorhombic
      cell_out%symmetry_id = cell_in%symmetry_id
      cell_out%ref_count = 1
      last_cell_id = last_cell_id+1
      cell_out%id_nr = last_cell_id

   END SUBROUTINE cell_clone

! **************************************************************************************************
!> \brief ...
!> \param cell_in ...
!> \param cell_out ...
! **************************************************************************************************
   SUBROUTINE cell_copy(cell_in, cell_out)

      TYPE(cell_type), POINTER                           :: cell_in, cell_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cell_copy', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(cell_in)))CALL cp__a("subsys/cell_types.F",128)
      IF(.NOT.(ASSOCIATED(cell_out)))CALL cp__a("subsys/cell_types.F",129)

      cell_out%deth = cell_in%deth
      cell_out%perd = cell_in%perd
      cell_out%hmat = cell_in%hmat
      cell_out%h_inv = cell_in%h_inv
      cell_out%orthorhombic = cell_in%orthorhombic
      cell_out%symmetry_id = cell_in%symmetry_id

   END SUBROUTINE cell_copy

! **************************************************************************************************
!> \brief   Read cell info from a line (parsed from a file)
!> \param input_line ...
!> \param cell_itimes ...
!> \param cell_time ...
!> \param h ...
!> \param vol ...
!> \date    19.02.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parse_cell_line(input_line, cell_itimes, cell_time, h, vol)
      CHARACTER(LEN=*), INTENT(IN)                       :: input_line
      INTEGER, INTENT(OUT)                               :: cell_itimes
      REAL(KIND=dp), INTENT(OUT)                         :: cell_time
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: h
      REAL(KIND=dp), INTENT(OUT)                         :: vol

      CHARACTER(len=*), PARAMETER :: routineN = 'parse_cell_line', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      READ (input_line, *) cell_itimes, cell_time, &
         h(1, 1), h(2, 1), h(3, 1), h(1, 2), h(2, 2), h(3, 2), h(1, 3), h(2, 3), h(3, 3), vol
      DO i = 1, 3
         DO j = 1, 3
            h(j, i) = cp_unit_to_cp2k(h(j, i), "angstrom")
         END DO
      END DO

   END SUBROUTINE parse_cell_line

! **************************************************************************************************
!> \brief   Get informations about a simulation cell.
!> \param cell ...
!> \param alpha ...
!> \param beta ...
!> \param gamma ...
!> \param deth ...
!> \param orthorhombic ...
!> \param abc ...
!> \param periodic ...
!> \param h ...
!> \param h_inv ...
!> \param id_nr ...
!> \param symmetry_id ...
!> \date    16.01.2002
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_cell(cell, alpha, beta, gamma, deth, orthorhombic, abc, periodic, &
                       h, h_inv, id_nr, symmetry_id)

      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: alpha, beta, gamma, deth
      LOGICAL, INTENT(OUT), OPTIONAL                     :: orthorhombic
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT), OPTIONAL :: abc
      INTEGER, DIMENSION(3), INTENT(OUT), OPTIONAL       :: periodic
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT), &
         OPTIONAL                                        :: h, h_inv
      INTEGER, INTENT(out), OPTIONAL                     :: id_nr, symmetry_id

      IF (PRESENT(deth)) deth = cell%deth ! the volume
      IF (PRESENT(orthorhombic)) orthorhombic = cell%orthorhombic
      IF (PRESENT(periodic)) periodic(:) = cell%perd(:)
      IF (PRESENT(h)) h(:, :) = cell%hmat(:, :)
      IF (PRESENT(h_inv)) h_inv(:, :) = cell%h_inv(:, :)

      ! Calculate the lengths of the cell vectors a, b, and c
      IF (PRESENT(abc)) THEN
         abc(1) = SQRT(cell%hmat(1, 1)*cell%hmat(1, 1)+ &
                       cell%hmat(2, 1)*cell%hmat(2, 1)+ &
                       cell%hmat(3, 1)*cell%hmat(3, 1))
         abc(2) = SQRT(cell%hmat(1, 2)*cell%hmat(1, 2)+ &
                       cell%hmat(2, 2)*cell%hmat(2, 2)+ &
                       cell%hmat(3, 2)*cell%hmat(3, 2))
         abc(3) = SQRT(cell%hmat(1, 3)*cell%hmat(1, 3)+ &
                       cell%hmat(2, 3)*cell%hmat(2, 3)+ &
                       cell%hmat(3, 3)*cell%hmat(3, 3))
      END IF

      ! Angles between the cell vectors a, b, and c
      ! alpha = <(b,c)
      IF (PRESENT(alpha)) alpha = angle(cell%hmat(:, 2), cell%hmat(:, 3))*degree
      ! beta = <(a,c)
      IF (PRESENT(beta)) beta = angle(cell%hmat(:, 1), cell%hmat(:, 3))*degree
      ! gamma = <(a,b)
      IF (PRESENT(gamma)) gamma = angle(cell%hmat(:, 1), cell%hmat(:, 2))*degree
      IF (PRESENT(id_nr)) id_nr = cell%id_nr
      IF (PRESENT(symmetry_id)) symmetry_id = cell%symmetry_id
   END SUBROUTINE get_cell

! **************************************************************************************************
!> \brief   Access internal type variables
!> \param cell ...
!> \param cell_length ...
!> \param cell_angle ...
!> \param units_angle ...
!> \param periodic ...
!> \date    04.04.2002
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_cell_param(cell, cell_length, cell_angle, units_angle, periodic)

      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: cell_length
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT), OPTIONAL :: cell_angle
      INTEGER, INTENT(IN), OPTIONAL                      :: units_angle
      INTEGER, DIMENSION(3), INTENT(OUT), OPTIONAL       :: periodic

      CHARACTER(len=*), PARAMETER :: routineN = 'get_cell_param', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: alpha, beta, gamma

      CALL get_cell(cell=cell, abc=cell_length)

      IF (PRESENT(cell_angle)) THEN
         CALL get_cell(cell=cell, alpha=alpha, beta=beta, gamma=gamma)
         cell_angle(:) = (/alpha, beta, gamma/)
         IF (PRESENT(units_angle)) THEN
            IF (units_angle == cp_units_rad) cell_angle = cell_angle/degree
         END IF
      END IF

      IF (PRESENT(periodic)) CALL get_cell(cell=cell, periodic=periodic)

   END SUBROUTINE get_cell_param

! **************************************************************************************************
!> \brief   Sets the cell using the internal parameters (a,b,c) (alpha,beta,gamma)
!>          using the convention: a parallel to the x axis, b in the x-y plane and
!>          and c univoquely determined; gamma is the angle between a and b; beta
!>          is the angle between c and a and alpha is the angle between c and b
!> \param cell ...
!> \param cell_length ...
!> \param cell_angle ...
!> \param periodic ...
!> \param do_init_cell ...
!> \date    03.2008
!> \author  Teodoro Laino
! **************************************************************************************************
   SUBROUTINE set_cell_param(cell, cell_length, cell_angle, periodic, do_init_cell)

      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: cell_length, cell_angle
      INTEGER, DIMENSION(3), INTENT(IN), OPTIONAL        :: periodic
      LOGICAL, INTENT(IN)                                :: do_init_cell

      CHARACTER(len=*), PARAMETER :: routineN = 'set_cell_param', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: cos_alpha, cos_beta, cos_gamma, eps, &
                                                            sin_gamma

      IF(.NOT.(ALL(cell_angle /= 0.0_dp)))CALL cp__a("subsys/cell_types.F",295)
      eps = EPSILON(0.0_dp)
      cos_gamma = COS(cell_angle(3)); IF (ABS(cos_gamma) < eps) cos_gamma = 0.0_dp
      IF (ABS(ABS(cos_gamma)-1.0_dp) < eps) cos_gamma = SIGN(1.0_dp, cos_gamma)
      sin_gamma = SIN(cell_angle(3)); IF (ABS(sin_gamma) < eps) sin_gamma = 0.0_dp
      IF (ABS(ABS(sin_gamma)-1.0_dp) < eps) sin_gamma = SIGN(1.0_dp, sin_gamma)
      cos_beta = COS(cell_angle(2)); IF (ABS(cos_beta) < eps) cos_beta = 0.0_dp
      IF (ABS(ABS(cos_beta)-1.0_dp) < eps) cos_beta = SIGN(1.0_dp, cos_beta)
      cos_alpha = COS(cell_angle(1)); IF (ABS(cos_alpha) < eps) cos_alpha = 0.0_dp
      IF (ABS(ABS(cos_alpha)-1.0_dp) < eps) cos_alpha = SIGN(1.0_dp, cos_alpha)

      cell%hmat(:, 1) = (/1.0_dp, 0.0_dp, 0.0_dp/)
      cell%hmat(:, 2) = (/cos_gamma, sin_gamma, 0.0_dp/)
      cell%hmat(:, 3) = (/cos_beta, (cos_alpha-cos_gamma*cos_beta)/sin_gamma, 0.0_dp/)
      cell%hmat(3, 3) = SQRT(1.0_dp-cell%hmat(1, 3)**2-cell%hmat(2, 3)**2)

      cell%hmat(:, 1) = cell%hmat(:, 1)*cell_length(1)
      cell%hmat(:, 2) = cell%hmat(:, 2)*cell_length(2)
      cell%hmat(:, 3) = cell%hmat(:, 3)*cell_length(3)

      IF (do_init_cell) THEN
         IF (PRESENT(periodic)) THEN
            CALL init_cell(cell=cell, periodic=periodic)
         ELSE
            CALL init_cell(cell=cell)
         END IF
      END IF

   END SUBROUTINE set_cell_param

! **************************************************************************************************
!> \brief   Initialise/readjust a simulation cell after hmat has been changed
!> \param cell ...
!> \param hmat ...
!> \param periodic ...
!> \date    16.01.2002
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_cell(cell, hmat, periodic)

      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN), &
         OPTIONAL                                        :: hmat
      INTEGER, DIMENSION(3), INTENT(IN), OPTIONAL        :: periodic

      CHARACTER(len=*), PARAMETER :: routineN = 'init_cell', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: eps_hmat = 1.0E-14_dp

      INTEGER                                            :: dim
      REAL(KIND=dp)                                      :: a, acosa, acosah, alpha, asina, asinah, &
                                                            beta, norm, norm_c
      REAL(KIND=dp), DIMENSION(3)                        :: abc

      IF (PRESENT(hmat)) cell%hmat(:, :) = hmat(:, :)
      IF (PRESENT(periodic)) cell%perd(:) = periodic(:)

      cell%deth = ABS(det_3x3(cell%hmat))

      IF (cell%deth < 1.0E-10_dp) THEN
         CALL cp_abort(cp__l("subsys/cell_types.F",355), &
                       "An invalid set of cell vectors was specified. "// &
                       "The determinant det(h) is too small")
      END IF
      cell%h_inv = inv_3x3(cell%hmat)

      SELECT CASE (cell%symmetry_id)
      CASE (cell_sym_cubic, &
            cell_sym_tetragonal_ab, &
            cell_sym_tetragonal_ac, &
            cell_sym_tetragonal_bc, &
            cell_sym_orthorhombic)
         CALL get_cell(cell=cell, abc=abc)
         abc(2) = plane_distance(0, 1, 0, cell=cell)
         abc(3) = plane_distance(0, 0, 1, cell=cell)
         SELECT CASE (cell%symmetry_id)
         CASE (cell_sym_cubic)
            abc(1:3) = SUM(abc(1:3))/3.0_dp
         CASE (cell_sym_tetragonal_ab, &
               cell_sym_tetragonal_ac, &
               cell_sym_tetragonal_bc)
            SELECT CASE (cell%symmetry_id)
            CASE (cell_sym_tetragonal_ab)
               a = 0.5_dp*(abc(1)+abc(2))
               abc(1) = a
               abc(2) = a
            CASE (cell_sym_tetragonal_ac)
               a = 0.5_dp*(abc(1)+abc(3))
               abc(1) = a
               abc(3) = a
            CASE (cell_sym_tetragonal_bc)
               a = 0.5_dp*(abc(2)+abc(3))
               abc(2) = a
               abc(3) = a
            END SELECT
         END SELECT
         cell%hmat(1, 1) = abc(1); cell%hmat(1, 2) = 0.0_dp; cell%hmat(1, 3) = 0.0_dp
         cell%hmat(2, 1) = 0.0_dp; cell%hmat(2, 2) = abc(2); cell%hmat(2, 3) = 0.0_dp
         cell%hmat(3, 1) = 0.0_dp; cell%hmat(3, 2) = 0.0_dp; cell%hmat(3, 3) = abc(3)
      CASE (cell_sym_hexagonal)
         CALL get_cell(cell=cell, abc=abc)
         a = 0.5_dp*(abc(1)+abc(2))
         acosa = 0.5_dp*a
         asina = sqrt3*acosa
         cell%hmat(1, 1) = a; cell%hmat(1, 2) = acosa; cell%hmat(1, 3) = 0.0_dp
         cell%hmat(2, 1) = 0.0_dp; cell%hmat(2, 2) = asina; cell%hmat(2, 3) = 0.0_dp
         cell%hmat(3, 1) = 0.0_dp; cell%hmat(3, 2) = 0.0_dp; cell%hmat(3, 3) = abc(3)
      CASE (cell_sym_rhombohedral)
         CALL get_cell(cell=cell, abc=abc)
         a = SUM(abc(1:3))/3.0_dp
         alpha = (angle(cell%hmat(:, 3), cell%hmat(:, 2))+ &
                  angle(cell%hmat(:, 1), cell%hmat(:, 3))+ &
                  angle(cell%hmat(:, 1), cell%hmat(:, 2)))/3.0_dp
         acosa = a*COS(alpha)
         asina = a*SIN(alpha)
         acosah = a*COS(0.5_dp*alpha)
         asinah = a*SIN(0.5_dp*alpha)
         norm = acosa/acosah
         norm_c = SQRT(1.0_dp-norm*norm)
         cell%hmat(1, 1) = a; cell%hmat(1, 2) = acosa; cell%hmat(1, 3) = acosah*norm
         cell%hmat(2, 1) = 0.0_dp; cell%hmat(2, 2) = asina; cell%hmat(2, 3) = asinah*norm
         cell%hmat(3, 1) = 0.0_dp; cell%hmat(3, 2) = 0.0_dp; cell%hmat(3, 3) = a*norm_c
      CASE (cell_sym_monoclinic)
         CALL get_cell(cell=cell, abc=abc)
         beta = angle(cell%hmat(:, 1), cell%hmat(:, 3))
         cell%hmat(1, 1) = abc(1); cell%hmat(1, 2) = 0.0_dp; cell%hmat(1, 3) = abc(3)*COS(beta)
         cell%hmat(2, 1) = 0.0_dp; cell%hmat(2, 2) = abc(2); cell%hmat(2, 3) = 0.0_dp
         cell%hmat(3, 1) = 0.0_dp; cell%hmat(3, 2) = 0.0_dp; cell%hmat(3, 3) = abc(3)*SIN(beta)
      CASE (cell_sym_triclinic)
         ! Nothing to do
      END SELECT

      ! Do we have an (almost) orthorhombic cell?
      IF ((ABS(cell%hmat(1, 2)) < eps_hmat) .AND. (ABS(cell%hmat(1, 3)) < eps_hmat) .AND. &
          (ABS(cell%hmat(2, 1)) < eps_hmat) .AND. (ABS(cell%hmat(2, 3)) < eps_hmat) .AND. &
          (ABS(cell%hmat(3, 1)) < eps_hmat) .AND. (ABS(cell%hmat(3, 2)) < eps_hmat)) THEN
         cell%orthorhombic = .TRUE.
      ELSE
         cell%orthorhombic = .FALSE.
      END IF

      ! Retain an exact orthorhombic cell
      ! (off-diagonal elements must remain zero identically to keep QS fast)
      IF (cell%orthorhombic) THEN
         cell%hmat(1, 2) = 0.0_dp
         cell%hmat(1, 3) = 0.0_dp
         cell%hmat(2, 1) = 0.0_dp
         cell%hmat(2, 3) = 0.0_dp
         cell%hmat(3, 1) = 0.0_dp
         cell%hmat(3, 2) = 0.0_dp
      END IF

      dim = COUNT(cell%perd == 1)
      IF ((dim == 1) .AND. (.NOT. cell%orthorhombic)) THEN
         CALL cp__b("subsys/cell_types.F",449,"Non-orthorhombic and not periodic")
      END IF

   END SUBROUTINE init_cell

! **************************************************************************************************
!> \brief   Calculate the distance between two lattice planes as defined by
!>          a triple of Miller indices (hkl).
!> \param h ...
!> \param k ...
!> \param l ...
!> \param cell ...
!> \retval distance ...
!> \date    18.11.2004
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   FUNCTION plane_distance(h, k, l, cell) RESULT(distance)

      INTEGER, INTENT(IN)                                :: h, k, l
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp)                                      :: distance

      REAL(KIND=dp)                                      :: a, alpha, b, beta, c, cosa, cosb, cosg, &
                                                            d, gamma, x, y, z
      REAL(KIND=dp), DIMENSION(3)                        :: abc

      x = REAL(h, KIND=dp)
      y = REAL(k, KIND=dp)
      z = REAL(l, KIND=dp)

      CALL get_cell(cell=cell, abc=abc)

      a = abc(1)
      b = abc(2)
      c = abc(3)

      IF (cell%orthorhombic) THEN

         d = (x/a)**2+(y/b)**2+(z/c)**2

      ELSE

         CALL get_cell(cell=cell, &
                       alpha=alpha, &
                       beta=beta, &
                       gamma=gamma)

         alpha = alpha/degree
         beta = beta/degree
         gamma = gamma/degree

         cosa = COS(alpha)
         cosb = COS(beta)
         cosg = COS(gamma)

         d = ((x*b*c*SIN(alpha))**2+ &
              (y*c*a*SIN(beta))**2+ &
              (z*a*b*SIN(gamma))**2+ &
              2.0_dp*a*b*c*(x*y*c*(cosa*cosb-cosg)+ &
                            z*x*b*(cosg*cosa-cosb)+ &
                            y*z*a*(cosb*cosg-cosa)))/ &
             ((a*b*c)**2*(1.0_dp-cosa**2-cosb**2-cosg**2+ &
                          2.0_dp*cosa*cosb*cosg))

      END IF

      distance = 1.0_dp/SQRT(d)

   END FUNCTION plane_distance

! **************************************************************************************************
!> \brief   Apply the periodic boundary conditions defined by a simulation
!>          cell to a position vector r.
!> \param r ...
!> \param cell ...
!> \retval r_pbc ...
!> \date    16.01.2002
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   FUNCTION pbc1(r, cell) RESULT(r_pbc)

      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3)                        :: r_pbc

      REAL(KIND=dp), DIMENSION(3)                        :: s

      IF (cell%orthorhombic) THEN
         r_pbc(1) = r(1)-cell%hmat(1, 1)*cell%perd(1)*ANINT(cell%h_inv(1, 1)*r(1))
         r_pbc(2) = r(2)-cell%hmat(2, 2)*cell%perd(2)*ANINT(cell%h_inv(2, 2)*r(2))
         r_pbc(3) = r(3)-cell%hmat(3, 3)*cell%perd(3)*ANINT(cell%h_inv(3, 3)*r(3))
      ELSE
         s(1) = cell%h_inv(1, 1)*r(1)+cell%h_inv(1, 2)*r(2)+cell%h_inv(1, 3)*r(3)
         s(2) = cell%h_inv(2, 1)*r(1)+cell%h_inv(2, 2)*r(2)+cell%h_inv(2, 3)*r(3)
         s(3) = cell%h_inv(3, 1)*r(1)+cell%h_inv(3, 2)*r(2)+cell%h_inv(3, 3)*r(3)
         s(1) = s(1)-cell%perd(1)*ANINT(s(1))
         s(2) = s(2)-cell%perd(2)*ANINT(s(2))
         s(3) = s(3)-cell%perd(3)*ANINT(s(3))
         r_pbc(1) = cell%hmat(1, 1)*s(1)+cell%hmat(1, 2)*s(2)+cell%hmat(1, 3)*s(3)
         r_pbc(2) = cell%hmat(2, 1)*s(1)+cell%hmat(2, 2)*s(2)+cell%hmat(2, 3)*s(3)
         r_pbc(3) = cell%hmat(3, 1)*s(1)+cell%hmat(3, 2)*s(2)+cell%hmat(3, 3)*s(3)
      END IF

   END FUNCTION pbc1

! **************************************************************************************************
!> \brief   Apply the periodic boundary conditions defined by a simulation
!>          cell to a position vector r subtracting nl from the periodic images
!> \param r ...
!> \param cell ...
!> \param nl ...
!> \retval r_pbc ...
!> \date    16.01.2002
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   FUNCTION pbc2(r, cell, nl) RESULT(r_pbc)

      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, DIMENSION(3), INTENT(IN)                  :: nl
      REAL(KIND=dp), DIMENSION(3)                        :: r_pbc

      REAL(KIND=dp), DIMENSION(3)                        :: s

      IF (cell%orthorhombic) THEN
         r_pbc(1) = r(1)-cell%hmat(1, 1)*cell%perd(1)* &
                    REAL(NINT(cell%h_inv(1, 1)*r(1))-nl(1), dp)
         r_pbc(2) = r(2)-cell%hmat(2, 2)*cell%perd(2)* &
                    REAL(NINT(cell%h_inv(2, 2)*r(2))-nl(2), dp)
         r_pbc(3) = r(3)-cell%hmat(3, 3)*cell%perd(3)* &
                    REAL(NINT(cell%h_inv(3, 3)*r(3))-nl(3), dp)
      ELSE
         s(1) = cell%h_inv(1, 1)*r(1)+cell%h_inv(1, 2)*r(2)+cell%h_inv(1, 3)*r(3)
         s(2) = cell%h_inv(2, 1)*r(1)+cell%h_inv(2, 2)*r(2)+cell%h_inv(2, 3)*r(3)
         s(3) = cell%h_inv(3, 1)*r(1)+cell%h_inv(3, 2)*r(2)+cell%h_inv(3, 3)*r(3)
         s(1) = s(1)-cell%perd(1)*REAL(NINT(s(1))-nl(1), dp)
         s(2) = s(2)-cell%perd(2)*REAL(NINT(s(2))-nl(2), dp)
         s(3) = s(3)-cell%perd(3)*REAL(NINT(s(3))-nl(3), dp)
         r_pbc(1) = cell%hmat(1, 1)*s(1)+cell%hmat(1, 2)*s(2)+cell%hmat(1, 3)*s(3)
         r_pbc(2) = cell%hmat(2, 1)*s(1)+cell%hmat(2, 2)*s(2)+cell%hmat(2, 3)*s(3)
         r_pbc(3) = cell%hmat(3, 1)*s(1)+cell%hmat(3, 2)*s(2)+cell%hmat(3, 3)*s(3)
      END IF

   END FUNCTION pbc2

! **************************************************************************************************
!> \brief   Apply the periodic boundary conditions defined by the simulation
!>          cell cell to the vector pointing from atom a to atom b.
!> \param ra ...
!> \param rb ...
!> \param cell ...
!> \retval rab_pbc ...
!> \date    11.03.2004
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   FUNCTION pbc3(ra, rb, cell) RESULT(rab_pbc)

      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ra, rb
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3)                        :: rab_pbc

      INTEGER                                            :: icell, jcell, kcell
      INTEGER, DIMENSION(3)                              :: periodic
      REAL(KIND=dp)                                      :: rab2, rab2_pbc
      REAL(KIND=dp), DIMENSION(3)                        :: r, ra_pbc, rab, rb_image, rb_pbc, s2r

      CALL get_cell(cell=cell, periodic=periodic)

      ra_pbc(:) = pbc(ra(:), cell)
      rb_pbc(:) = pbc(rb(:), cell)

      rab2_pbc = HUGE(1.0_dp)

      DO icell = -periodic(1), periodic(1)
         DO jcell = -periodic(2), periodic(2)
            DO kcell = -periodic(3), periodic(3)
               r = REAL((/icell, jcell, kcell/), dp)
               CALL scaled_to_real(s2r, r, cell)
               rb_image(:) = rb_pbc(:)+s2r
               rab(:) = rb_image(:)-ra_pbc(:)
               rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
               IF (rab2 < rab2_pbc) THEN
                  rab2_pbc = rab2
                  rab_pbc(:) = rab(:)
               END IF
            END DO
         END DO
      END DO

   END FUNCTION pbc3

   !if positive_range == true, r(i) (or s(i)) in range [0, hmat(i,i)],
   !else, r(i) (s(i)) in range [-hmat(i,i)/2, hmat(i,i)/2]
! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param cell ...
!> \param positive_range ...
!> \retval r_pbc ...
! **************************************************************************************************
   FUNCTION pbc4(r, cell, positive_range) RESULT(r_pbc)

      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r
      TYPE(cell_type), POINTER                           :: cell
      LOGICAL                                            :: positive_range
      REAL(KIND=dp), DIMENSION(3)                        :: r_pbc

      REAL(KIND=dp), DIMENSION(3)                        :: s

      IF (positive_range) THEN
         IF (cell%orthorhombic) THEN
            r_pbc(1) = r(1)-cell%hmat(1, 1)*cell%perd(1)*FLOOR(cell%h_inv(1, 1)*r(1))
            r_pbc(2) = r(2)-cell%hmat(2, 2)*cell%perd(2)*FLOOR(cell%h_inv(2, 2)*r(2))
            r_pbc(3) = r(3)-cell%hmat(3, 3)*cell%perd(3)*FLOOR(cell%h_inv(3, 3)*r(3))
         ELSE
            s(1) = cell%h_inv(1, 1)*r(1)+cell%h_inv(1, 2)*r(2)+cell%h_inv(1, 3)*r(3)
            s(2) = cell%h_inv(2, 1)*r(1)+cell%h_inv(2, 2)*r(2)+cell%h_inv(2, 3)*r(3)
            s(3) = cell%h_inv(3, 1)*r(1)+cell%h_inv(3, 2)*r(2)+cell%h_inv(3, 3)*r(3)
            s(1) = s(1)-cell%perd(1)*FLOOR(s(1))
            s(2) = s(2)-cell%perd(2)*FLOOR(s(2))
            s(3) = s(3)-cell%perd(3)*FLOOR(s(3))
            r_pbc(1) = cell%hmat(1, 1)*s(1)+cell%hmat(1, 2)*s(2)+cell%hmat(1, 3)*s(3)
            r_pbc(2) = cell%hmat(2, 1)*s(1)+cell%hmat(2, 2)*s(2)+cell%hmat(2, 3)*s(3)
            r_pbc(3) = cell%hmat(3, 1)*s(1)+cell%hmat(3, 2)*s(2)+cell%hmat(3, 3)*s(3)
         END IF
      ELSE
         r_pbc = pbc1(r, cell)
      END IF

   END FUNCTION pbc4

! **************************************************************************************************
!> \brief   Transform real to scaled cell coordinates.
!>          s=h_inv*r
!> \param s ...
!> \param r ...
!> \param cell ...
!> \date    16.01.2002
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE real_to_scaled(s, r, cell)

      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: s
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r
      TYPE(cell_type), POINTER                           :: cell

      IF (cell%orthorhombic) THEN
         s(1) = cell%h_inv(1, 1)*r(1)
         s(2) = cell%h_inv(2, 2)*r(2)
         s(3) = cell%h_inv(3, 3)*r(3)
      ELSE
         s(1) = cell%h_inv(1, 1)*r(1)+cell%h_inv(1, 2)*r(2)+cell%h_inv(1, 3)*r(3)
         s(2) = cell%h_inv(2, 1)*r(1)+cell%h_inv(2, 2)*r(2)+cell%h_inv(2, 3)*r(3)
         s(3) = cell%h_inv(3, 1)*r(1)+cell%h_inv(3, 2)*r(2)+cell%h_inv(3, 3)*r(3)
      END IF

   END SUBROUTINE real_to_scaled

! **************************************************************************************************
!> \brief   Transform scaled cell coordinates real coordinates.
!>          r=h*s
!> \param r ...
!> \param s ...
!> \param cell ...
!> \date    16.01.2002
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE scaled_to_real(r, s, cell)

      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: r
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: s
      TYPE(cell_type), POINTER                           :: cell

      IF (cell%orthorhombic) THEN
         r(1) = cell%hmat(1, 1)*s(1)
         r(2) = cell%hmat(2, 2)*s(2)
         r(3) = cell%hmat(3, 3)*s(3)
      ELSE
         r(1) = cell%hmat(1, 1)*s(1)+cell%hmat(1, 2)*s(2)+cell%hmat(1, 3)*s(3)
         r(2) = cell%hmat(2, 1)*s(1)+cell%hmat(2, 2)*s(2)+cell%hmat(2, 3)*s(3)
         r(3) = cell%hmat(3, 1)*s(1)+cell%hmat(3, 2)*s(2)+cell%hmat(3, 3)*s(3)
      END IF

   END SUBROUTINE scaled_to_real

! **************************************************************************************************
!> \brief allocates and initializes a cell
!> \param cell the cell to initialize
!> \param hmat the h matrix that defines the cell
!> \param periodic periodicity of the cell
!> \par History
!>      09.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cell_create(cell, hmat, periodic)

      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN), &
         OPTIONAL                                        :: hmat
      INTEGER, DIMENSION(3), INTENT(IN), OPTIONAL        :: periodic

      CHARACTER(len=*), PARAMETER :: routineN = 'cell_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(cell)))CALL cp__a("subsys/cell_types.F",758)
      ALLOCATE (cell)
      last_cell_id = last_cell_id+1
      cell%id_nr = last_cell_id
      cell%ref_count = 1
      IF (PRESENT(periodic)) THEN
         cell%perd = periodic
      ELSE
         cell%perd = 1
      END IF
      cell%orthorhombic = .FALSE.
      cell%symmetry_id = cell_sym_none
      IF (.NOT. PRESENT(hmat)) RETURN
      CALL init_cell(cell, hmat, periodic)

   END SUBROUTINE cell_create

! **************************************************************************************************
!> \brief retains the given cell (see doc/ReferenceCounting.html)
!> \param cell the cell to retain
!> \par History
!>      09.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cell_retain(cell)

      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'cell_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("subsys/cell_types.F",788)
      IF(.NOT.(cell%ref_count > 0))CALL cp__a("subsys/cell_types.F",789)
      cell%ref_count = cell%ref_count+1

   END SUBROUTINE cell_retain

! **************************************************************************************************
!> \brief releases the given cell (see doc/ReferenceCounting.html)
!> \param cell the cell to release
!> \par History
!>      09.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cell_release(cell)

      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'cell_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(cell)) THEN
         IF(.NOT.(cell%ref_count > 0))CALL cp__a("subsys/cell_types.F",808)
         cell%ref_count = cell%ref_count-1
         IF (cell%ref_count == 0) THEN
            DEALLOCATE (cell)
         END IF
         NULLIFY (cell)
      END IF

   END SUBROUTINE cell_release


END MODULE cell_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Arrays of parameters used in the semi-empirical calculations
!> \References  Everywhere in this module TCA stands for:
!>   - TCA:   W. Thiel and  A. A. Voityuk - Teor. Chim. Acta (1992) 81:391-404
!>   - TCA77: M.J.S. Dewar and W. Thiel - Teor. Chim. Acta (1977) 46:89-104
!>
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date   03.2008 [tlaino]
! **************************************************************************************************
MODULE semi_empirical_int_arrays

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_int_arrays'
   REAL(KINd=dp), PARAMETER, PUBLIC     :: rij_threshold = 0.00002_dp

   ! Mapping index for orbital ordering
   INTEGER, DIMENSION(9), PUBLIC        :: se_orbital_pointer = (/1, 4, 2, 3, 9, 8, 7, 6, 5/)
   INTEGER, DIMENSION(9), PUBLIC        :: se_map_alm = (/1, 3, 4, 2, 8, 6, 5, 7, 9/)

   ! Arrays to treat the invertion of the reference frame for the integrals
   ! Using the same indexing convention of the integrals
   INTEGER, PARAMETER, DIMENSION(2, 9), PUBLIC        :: map_x_to_z = RESHAPE((/ &
                                                                              1, 0, & ! s    <-> s
                                                                              4, 0, & ! px   <-> pz
                                                                              3, 0, & ! py   <-> py
                                                                              2, 0, & ! pz   <-> px
                                                                              7, 5, & ! dx2  <-> SQR3/2*dz2 +    0.5*dx2
                                                                              6, 0, & ! dzx  <-> dzx
                                                                              7, 5, & ! dz2  <->   -0.5*dz2 + SQR3/2*dx2
                                                                              9, 0, & ! dzy  <-> dxy
                                                                              8, 0 & ! dxy  <-> dzy
                                                                              /), (/2, 9/))
   REAL(KIND=dp), PARAMETER, DIMENSION(2, 9), PUBLIC  :: fac_x_to_z = RESHAPE((/ &
                                                                              1.0_dp, 0.0_dp, &
                                                                              1.0_dp, 0.0_dp, &
                                                                              1.0_dp, 0.0_dp, &
                                                                              1.0_dp, 0.0_dp, &
                                                                              0.8660254037844386_dp, 0.5_dp, &
                                                                              1.0_dp, 0.0_dp, &
                                                                              -0.5_dp, 0.8660254037844386_dp, &
                                                                              1.0_dp, 0.0_dp, &
                                                                              1.0_dp, 0.0_dp &
                                                                              /), (/2, 9/))

   ! Clm coefficients for d-orbitals:  see  Table [1] and [2] of TCA
   REAL(KIND=dp), DIMENSION(45, 0:2, -2:2), PUBLIC :: clm_d
   ! Clm coefficients for sp-orbitals: see original paper TCA77
   INTEGER, DIMENSION(45, 0:2, -2:2), PUBLIC :: clm_sp
   ! alm coefficients: see Laino and Hutter (periodic SE)
   REAL(KIND=dp), DIMENSION(45, 0:2, -2:2), PUBLIC :: alm

   ! These values are absolutely arbitrary and are used only for a proper
   ! tag of the integrals
   INTEGER, PARAMETER, PUBLIC :: &
      CLMz = 10, CLMp = 11, CLMzz = 12, CLMzp = 13, CLMyy = 14, CLMxy = 15, CLMxx = 16

   ! Indexes for diagonal storage of ij and kl multipoles
   INTEGER, DIMENSION(9, 9), PUBLIC                 :: indexa, indexb

   ! Type of integral for 2electron 2centers integrals
   INTEGER, DIMENSION(45), PARAMETER, PUBLIC       :: int2c_type = (/ &
                                                      1, 2, 3, 2, 3, 3, 2, 3, 3, 3, 4, 5, 5, 5, 6, 4, 5, 5, 5, &
                                                      6, 6, 4, 5, 5, 5, 6, 6, 6, 4, 5, 5, 5, 6, 6, 6, 6, 4, 5, &
                                                      5, 5, 6, 6, 6, 6, 6/)

   ! Mappinf of shell index
   INTEGER, DIMENSION(9), PARAMETER, PUBLIC        :: l_index = (/ &
                                                      0, 1, 1, 1, 2, 2, 2, 2, 2/)

   ! Index for <ij|kl>
   INTEGER, DIMENSION(45, 45), PUBLIC             :: ijkl_ind

   ! Symmetry index for <ij|kl>
   INTEGER, DIMENSION(491), PUBLIC                :: ijkl_sym

   ! Index for integral rotations
   INTEGER, DIMENSION(3, 3), PUBLIC               :: indpp
   INTEGER, DIMENSION(5, 3), PUBLIC               :: inddp
   INTEGER, DIMENSION(5, 5), PUBLIC               :: inddd

   ! Indexes use for the construction of the one-center two-electron integrals
   INTEGER, DIMENSION(243), PUBLIC ::  int_ij = (/ &
                                      1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, &
                                      5, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, &
                                      10, 10, 10, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 15, &
                                      15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 17, 17, 17, 17, 17, 18, 18, 18, 19, &
                                      19, 19, 19, 19, 20, 20, 20, 20, 20, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 22, 22, &
                                      22, 22, 22, 22, 22, 22, 22, 23, 23, 23, 23, 23, 24, 24, 24, 24, 24, 25, 25, 25, 25, 26, 26, &
                                      26, 26, 26, 26, 27, 27, 27, 27, 27, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 29, 29, 29, 29, &
                                      29, 30, 30, 30, 31, 31, 31, 31, 31, 32, 32, 32, 32, 32, 33, 33, 33, 33, 33, 34, 34, 34, 34, &
                                      35, 35, 35, 35, 35, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 37, 37, 37, 37, 38, 38, &
                                      38, 38, 38, 39, 39, 39, 39, 39, 40, 40, 40, 41, 42, 42, 42, 42, 42, 43, 43, 43, 43, 44, 44, &
                                      44, 44, 44, 45, 45, 45, 45, 45, 45, 45, 45, 45, 45/)
   INTEGER, DIMENSION(243), PUBLIC ::  int_kl = (/ &
                                      15, 21, 28, 36, 45, 12, 19, 23, 39, 11, 15, 21, 22, 26, 28, 36, 45, 13, 24, 32, 38, 34, 37, &
                                      43, 11, 15, 21, 22, 26, 28, 36, 45, 17, 25, 31, 16, 20, 27, 44, 29, 33, 35, 42, 15, 21, 22, &
                                      28, 36, 45, 3, 6, 11, 21, 26, 36, 2, 12, 19, 23, 39, 4, 13, 24, 32, 38, 14, 17, 31, 1, &
                                      3, 6, 10, 15, 21, 22, 28, 36, 45, 8, 16, 20, 27, 44, 7, 14, 17, 25, 31, 18, 30, 40, 2, &
                                      12, 19, 23, 39, 8, 16, 20, 27, 44, 1, 3, 6, 10, 11, 15, 21, 22, 26, 28, 36, 45, 3, 6, &
                                      10, 15, 21, 22, 28, 36, 45, 2, 12, 19, 23, 39, 4, 13, 24, 32, 38, 7, 17, 25, 31, 3, 6, &
                                      11, 21, 26, 36, 8, 16, 20, 27, 44, 1, 3, 6, 10, 15, 21, 22, 28, 36, 45, 9, 29, 33, 35, &
                                      42, 18, 30, 40, 7, 14, 17, 25, 31, 4, 13, 24, 32, 38, 9, 29, 33, 35, 42, 5, 34, 37, 43, &
                                      9, 29, 33, 35, 42, 1, 3, 6, 10, 11, 15, 21, 22, 26, 28, 36, 45, 5, 34, 37, 43, 4, 13, &
                                      24, 32, 38, 2, 12, 19, 23, 39, 18, 30, 40, 41, 9, 29, 33, 35, 42, 5, 34, 37, 43, 8, 16, &
                                      20, 27, 44, 1, 3, 6, 10, 15, 21, 22, 28, 36, 45/)
   INTEGER, DIMENSION(243), PUBLIC :: int_onec2el = (/ &
                                      1, 1, 1, 1, 1, 3, 3, 8, 3, 9, 6, 6, 12, 14, 13, 7, 6, 15, 8, 3, 3, 11, 9, &
                                      14, 17, 6, 7, 12, 18, 13, 6, 6, 3, 2, 3, 9, 11, 10, 11, 9, 16, 10, 11, 7, 6, 4, &
                                      5, 6, 7, 9, 17, 19, 32, 22, 40, 3, 33, 34, 27, 46, 15, 33, 28, 41, 47, 35, 35, 42, 1, &
                                      6, 6, 7, 29, 38, 22, 31, 38, 51, 9, 19, 32, 21, 32, 3, 35, 33, 24, 34, 35, 35, 35, 3, &
                                      34, 33, 26, 34, 11, 32, 44, 37, 49, 1, 6, 7, 6, 32, 38, 29, 21, 39, 30, 38, 38, 12, 12, &
                                      4, 22, 21, 19, 20, 21, 22, 8, 27, 26, 25, 27, 8, 28, 25, 26, 27, 2, 24, 23, 24, 14, 18, &
                                      22, 39, 48, 45, 10, 21, 37, 36, 37, 1, 13, 13, 5, 31, 30, 20, 29, 30, 31, 9, 19, 40, 21, &
                                      32, 35, 35, 35, 3, 42, 34, 24, 33, 3, 41, 26, 33, 34, 16, 40, 44, 43, 50, 11, 44, 32, 39, &
                                      10, 21, 43, 36, 37, 1, 7, 6, 6, 40, 38, 38, 21, 45, 30, 29, 38, 9, 32, 19, 22, 3, 47, &
                                      27, 34, 33, 3, 46, 34, 27, 33, 35, 35, 35, 52, 11, 32, 50, 37, 44, 14, 39, 22, 48, 11, 32, &
                                      49, 37, 44, 1, 6, 6, 7, 51, 38, 22, 31, 38, 29/)

   PUBLIC :: init_se_intd_array

CONTAINS
! **************************************************************************************************
!> \brief  Initialize all arrays used for the evaluation of the integrals
!>
!> \date   04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE init_se_intd_array()

      CHARACTER(len=*), PARAMETER :: routineN = 'init_se_intd_array', &
         routineP = moduleN//':'//routineN

      CALL setup_index_array()
      CALL setup_indrot_array()
      CALL setup_clm_array()
      CALL setup_ijkl_array()

   END SUBROUTINE init_se_intd_array

! **************************************************************************************************
!> \brief Fills in array for the diagonal storage of the ij and kl multipoles term
!>
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_index_array()

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_index_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      DO i = 1, 9
         DO j = 1, i
            ! indexa:
            !           s   pz   px   py  dz2  dzx  dzy dx2-y2 dxy
            !      s    1    2    3    4    5    6    7    8    9
            !      pz   2   10   11   12   13   14   15   16   17
            !      px   3   11   18   19   20   21   22   23   24
            !      py   4   12   19   25   26   27   28   29   30
            !     dz2   5   13   20   26   31   32   33   34   35
            !     dzx   6   14   21   27   32   36   37   38   39
            !     dzy   7   15   22   28   33   37   40   41   42
            !  dx2-y2   8   16   23   29   34   38   41   43   44
            !     dxy   9   17   24   30   35   39   42   44   45
            indexa(i, j) = (9*(j-1))-(j*(j-1))/2+i
            indexa(j, i) = indexa(i, j)
            ! indexb:
            !           s   pz   px   py  dz2  dzx  dzy dx2-y2 dxy
            !      s    1    2    4    7   11   16   22   29   37
            !      pz   2    3    5    8   12   17   23   30   38
            !      px   4    5    6    9   13   18   24   31   39
            !      py   7    8    9   10   14   19   25   32   40
            !     dz2  11   12   13   14   15   20   26   33   41
            !     dzx  16   17   18   19   20   21   27   34   42
            !     dzy  22   23   24   25   26   27   28   35   43
            !  dx2-y2  29   30   31   32   33   34   35   36   44
            !     dxy  37   38   39   40   41   42   43   44   45
            indexb(i, j) = (i*(i-1))/2+j
            indexb(j, i) = indexb(i, j)
         END DO
      END DO
   END SUBROUTINE setup_index_array

! **************************************************************************************************
!> \brief Fills in array for the rotation of the integrals
!>
!> \date   04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_indrot_array()

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_indrot_array', &
         routineP = moduleN//':'//routineN

! Setup indexes for integral rotations
! INDPP

      indpp(1, 1) = 1
      indpp(2, 1) = 4
      indpp(3, 1) = 5
      indpp(1, 2) = 4
      indpp(2, 2) = 2
      indpp(3, 2) = 6
      indpp(1, 3) = 5
      indpp(2, 3) = 6
      indpp(3, 3) = 3
      ! INDDP
      inddp(1, 1) = 1
      inddp(2, 1) = 4
      inddp(3, 1) = 7
      inddp(4, 1) = 10
      inddp(5, 1) = 13
      inddp(1, 2) = 2
      inddp(2, 2) = 5
      inddp(3, 2) = 8
      inddp(4, 2) = 11
      inddp(5, 2) = 14
      inddp(1, 3) = 3
      inddp(2, 3) = 6
      inddp(3, 3) = 9
      inddp(4, 3) = 12
      inddp(5, 3) = 15
      ! INDDD
      inddd(1, 1) = 1
      inddd(2, 1) = 6
      inddd(3, 1) = 7
      inddd(4, 1) = 9
      inddd(5, 1) = 12
      inddd(1, 2) = 6
      inddd(2, 2) = 2
      inddd(3, 2) = 8
      inddd(4, 2) = 10
      inddd(5, 2) = 13
      inddd(1, 3) = 7
      inddd(2, 3) = 8
      inddd(3, 3) = 3
      inddd(4, 3) = 11
      inddd(5, 3) = 14
      inddd(1, 4) = 9
      inddd(2, 4) = 10
      inddd(3, 4) = 11
      inddd(4, 4) = 4
      inddd(5, 4) = 15
      inddd(1, 5) = 12
      inddd(2, 5) = 13
      inddd(3, 5) = 14
      inddd(4, 5) = 15
      inddd(5, 5) = 5
   END SUBROUTINE setup_indrot_array

! **************************************************************************************************
!> \brief Fills in Clm coefficients (see Table [2] of TCA)
!>
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_clm_array()

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_clm_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: CLM1, CLM1m
      REAL(KIND=dp) :: ALM1, ALMs15_49, ALMs15_49m, ALMs15m, ALMs20_49, ALMs20_49m, ALMs35, &
         ALMs35m, ALMs45, ALMs5_49, ALMs5_49m, CLM23, CLM23m, CLM43, CLM43m, CLMs13, CLMs13m, &
         CLMs43, CLMs43m

      CLM1 = 1
      CLM1m = -1
      CLMs13 = SQRT(1.0_dp/3.0_dp)
      CLMs13m = -SQRT(1.0_dp/3.0_dp)
      CLM23 = 2.0_dp/3.0_dp
      CLM23m = -2.0_dp/3.0_dp
      CLM43 = 4.0_dp/3.0_dp
      CLM43m = -4.0_dp/3.0_dp
      CLMs43 = SQRT(4.0_dp/3.0_dp)
      CLMs43m = -SQRT(4.0_dp/3.0_dp)
      ALM1 = 1.0_dp
      ALMs45 = SQRT(4.0_dp/5.0_dp)
      ALMs35 = SQRT(3.0_dp/5.0_dp)
      ALMs35m = -ALMs35
      ALMs15m = -SQRT(1.0_dp/5.0_dp)
      ALMs20_49 = SQRT(20.0_dp/49.0_dp)
      ALMs20_49m = -ALMs20_49
      ALMs5_49 = SQRT(5.0_dp/49.0_dp)
      ALMs5_49m = -ALMs5_49
      ALMs15_49 = SQRT(15.0_dp/49.0_dp)
      ALMs15_49m = -ALMs15_49

      ! Notation (1) s
      !          (2) pz = p_sigma             (5) d_sigma           = dz2  (8) d_delta               = dx2-y2
      !          (3) px = p_pi                (6) d_pi              = dzx  (9) d_{\overline{delta}}  = dxy
      !          (4) py = pi_{\overline{pi}}  (7) d_{\overline{pi}} = dzy

      clm_d = 0.0_dp
      clm_sp = 0
      alm = 0.0_dp
      ! Let's fill all element of table 1 with resulting multipole lesser than 2
      ! Important Note: the value of the clm_sp does not reflect any phyisical
      !                 rule for decomposing/summing multipoles. It's just a
      !                 computational trick to put some order where conceptual mess
      !                 has been created..
      ! s s
      clm_d(1, 0, 0) = CLM1
      clm_sp(1, 0, 0) = CLM1
      alm(1, 0, 0) = ALM1
      ! s pz
      clm_d(2, 1, 0) = CLM1
      clm_sp(2, 1, 0) = CLMz
      alm(2, 1, 0) = ALM1
      ! s px
      clm_d(3, 1, 1) = CLM1
      clm_sp(3, 1, 1) = CLMp
      alm(3, 1, 1) = ALM1
      ! s py
      clm_d(4, 1, -1) = CLM1
      clm_sp(4, 1, -1) = CLMp
      alm(4, 1, -1) = ALM1
      ! s dz2
      clm_d(5, 2, 0) = CLMs43
      alm(5, 2, 0) = ALM1
      ! s dzx
      clm_d(6, 2, 1) = CLM1
      alm(6, 2, 1) = ALM1
      ! s dzy
      clm_d(7, 2, -1) = CLM1
      alm(7, 2, -1) = ALM1
      ! s dx2-y2
      clm_d(8, 2, 2) = CLM1
      alm(8, 2, 2) = ALM1
      ! s dxy
      clm_d(9, 2, -2) = CLM1
      alm(9, 2, -2) = ALM1
      ! pz pz
      clm_d(10, 0, 0) = CLM1
      clm_d(10, 2, 0) = CLM43
      clm_sp(10, 0, 0) = CLM1
      clm_sp(10, 2, 0) = CLMzz
      alm(10, 0, 0) = ALM1
      alm(10, 2, 0) = ALMs45
      ! pz px
      clm_d(11, 2, 1) = CLM1
      clm_sp(11, 2, 1) = CLMzp
      alm(11, 2, 1) = ALMs35
      ! pz py
      clm_d(12, 2, -1) = CLM1
      clm_sp(12, 2, -1) = CLMzp
      alm(12, 2, -1) = ALMs35
      ! pz dz2
      clm_d(13, 1, 0) = CLMs43
      alm(13, 1, 0) = ALMs45
      ! pz dzx
      clm_d(14, 1, 1) = CLM1
      alm(14, 1, 1) = ALMs35
      ! pz dzy
      clm_d(15, 1, -1) = CLM1
      alm(15, 1, -1) = ALMs35
      ! px px
      clm_d(18, 0, 0) = CLM1
      clm_d(18, 2, 0) = CLM23m
      clm_d(18, 2, 2) = CLM1
      clm_sp(18, 0, 0) = CLM1
      clm_sp(18, 2, 0) = CLMyy
      alm(18, 0, 0) = ALM1
      alm(18, 2, 0) = ALMs15m
      alm(18, 2, 2) = ALMs35
      ! px py
      clm_d(19, 2, -2) = CLM1
      clm_sp(19, 2, -2) = CLMxy
      alm(19, 2, -2) = ALMs35
      ! px dz2
      clm_d(20, 1, 1) = CLMs13m
      alm(20, 1, 1) = ALMs15m
      ! px dzx
      clm_d(21, 1, 0) = CLM1
      alm(21, 1, 0) = ALMs35
      ! px dx2-y2
      clm_d(23, 1, 1) = CLM1
      alm(23, 1, 1) = ALMs35
      ! px dxy
      clm_d(24, 1, -1) = CLM1
      alm(24, 1, -1) = ALMs35
      ! py py
      clm_d(25, 0, 0) = CLM1
      clm_d(25, 2, 0) = CLM23m
      clm_d(25, 2, 2) = CLM1m
      clm_sp(25, 0, 0) = CLM1
      clm_sp(25, 2, 0) = CLMxx
      alm(25, 0, 0) = ALM1
      alm(25, 2, 0) = ALMs15m
      alm(25, 2, 2) = ALMs35m
      ! py dz2
      clm_d(26, 1, -1) = CLMs13m
      alm(26, 1, -1) = ALMs15m
      ! py dzy
      clm_d(28, 1, 0) = CLM1
      alm(28, 1, 0) = ALMs35
      ! py dx2-y2
      clm_d(29, 1, -1) = CLM1m
      alm(29, 1, -1) = ALMs35m
      ! py dxy
      clm_d(30, 1, 1) = CLM1
      alm(30, 1, 1) = ALMs35
      ! dz2 dz2
      clm_d(31, 0, 0) = CLM1
      clm_d(31, 2, 0) = CLM43
      alm(31, 0, 0) = ALM1
      alm(31, 2, 0) = ALMs20_49
      ! dz2 dzx
      clm_d(32, 2, 1) = CLMs13
      alm(32, 2, 1) = ALMs5_49
      ! dz2 dzy
      clm_d(33, 2, -1) = CLMs13
      alm(33, 2, -1) = ALMs5_49
      ! dz2 dx2-y2
      clm_d(34, 2, 2) = CLMs43m
      alm(34, 2, 2) = ALMs20_49m
      ! dz2 dxy
      clm_d(35, 2, -2) = CLMs43m
      alm(35, 2, -2) = ALMs20_49m
      ! dzx dzx
      clm_d(36, 0, 0) = CLM1
      clm_d(36, 2, 0) = CLM23
      clm_d(36, 2, 2) = CLM1
      alm(36, 0, 0) = ALM1
      alm(36, 2, 0) = ALMs5_49
      alm(36, 2, 2) = ALMs15_49
      ! dzx dzy
      clm_d(37, 2, -2) = CLM1
      alm(37, 2, -2) = ALMs15_49m
      ! dzx dxy-y2
      clm_d(38, 2, 1) = CLM1
      alm(38, 2, 1) = ALMs15_49
      ! dzx dxy
      clm_d(39, 2, -1) = CLM1
      alm(38, 2, -1) = ALMs15_49
      ! dzy dzy
      clm_d(40, 0, 0) = CLM1
      clm_d(40, 2, 0) = CLM23
      clm_d(40, 2, 2) = CLM1m
      alm(40, 0, 0) = ALM1
      alm(40, 2, 0) = ALMs5_49
      alm(40, 2, 2) = ALMs5_49m
      ! dzy dx2-y2
      clm_d(41, 2, -1) = CLM1m
      alm(41, 2, -1) = ALMs15_49m
      ! dzy dxy
      clm_d(42, 2, 1) = CLM1
      alm(42, 2, 1) = ALMs15_49
      ! dx2-y2 dx2-y2
      clm_d(43, 0, 0) = CLM1
      clm_d(43, 2, 0) = CLM43m
      alm(43, 0, 0) = ALM1
      alm(43, 2, 0) = ALMs20_49m
      ! dxy dxy
      clm_d(45, 0, 0) = CLM1
      clm_d(45, 2, 0) = CLM43m
      alm(45, 0, 0) = ALM1
      alm(45, 2, 0) = ALMs20_49m
   END SUBROUTINE setup_clm_array

! **************************************************************************************************
!> \brief Fills in the index number for the <ij|kl> integral as well as the
!>        symmetry index
!>
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_ijkl_array()

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_ijkl_array', &
         routineP = moduleN//':'//routineN

! Address unique indexes (excluding those related by rotations)
! Indexes according:
!           s   pz   px   py  dz2  dzx  dzy dx2-y2 dxy
!      s    1    2    3    4    5    6    7    8    9
!      pz   2   10   11   12   13   14   15   16   17
!      px   3   11   18   19   20   21   22   23   24
!      py   4   12   19   25   26   27   28   29   30
!     dz2   5   13   20   26   31   32   33   34   35
!     dzx   6   14   21   27   32   36   37   38   39
!     dzy   7   15   22   28   33   37   40   41   42
!  dx2-y2   8   16   23   29   34   38   41   43   44
!     dxy   9   17   24   30   35   39   42   44   45
! ################  Zero the Arrays #######################

      ijkl_ind = 0
      ijkl_sym = 0
      ! ################      s       s   #######################
      !                       s       s   -        s       s
      ijkl_ind(1, 1) = 1
      !                       s       s   -       pz       s
      ijkl_ind(1, 2) = 2
      !                       s       s   -       pz      pz
      ijkl_ind(1, 10) = 3
      !                       s       s   -       px      px
      ijkl_ind(1, 18) = 4
      !                       s       s   -       py      py
      ijkl_ind(1, 25) = 5
      ijkl_sym(5) = 4
      !                       s       s   -      dz2       s
      ijkl_ind(1, 5) = 35
      !                       s       s   -      dz2      pz
      ijkl_ind(1, 13) = 36
      !                       s       s   -      dz2     dz2
      ijkl_ind(1, 31) = 37
      !                       s       s   -      dzx      px
      ijkl_ind(1, 21) = 38
      !                       s       s   -      dzx     dzx
      ijkl_ind(1, 36) = 39
      !                       s       s   -      dzy      py
      ijkl_ind(1, 28) = 40
      ijkl_sym(40) = 38
      !                       s       s   -      dzy     dzy
      ijkl_ind(1, 40) = 41
      ijkl_sym(41) = 39
      !                       s       s   -   dx2-y2  dx2-y2
      ijkl_ind(1, 43) = 42
      !                       s       s   -      dxy     dxy
      ijkl_ind(1, 45) = 43
      ijkl_sym(43) = 42
      ! ################     pz       s   #######################
      !                      pz       s   -        s       s
      ijkl_ind(2, 1) = 6
      !                      pz       s   -       pz       s
      ijkl_ind(2, 2) = 7
      !                      pz       s   -       pz      pz
      ijkl_ind(2, 10) = 8
      !                      pz       s   -       px      px
      ijkl_ind(2, 18) = 9
      !                      pz       s   -       py      py
      ijkl_ind(2, 25) = 10
      ijkl_sym(10) = 9
      !                      pz       s   -      dz2       s
      ijkl_ind(2, 5) = 44
      !                      pz       s   -      dz2      pz
      ijkl_ind(2, 13) = 45
      !                      pz       s   -      dz2     dz2
      ijkl_ind(2, 31) = 46
      !                      pz       s   -      dzx      px
      ijkl_ind(2, 21) = 47
      !                      pz       s   -      dzx     dzx
      ijkl_ind(2, 36) = 48
      !                      pz       s   -      dzy      py
      ijkl_ind(2, 28) = 49
      ijkl_sym(49) = 47
      !                      pz       s   -      dzy     dzy
      ijkl_ind(2, 40) = 50
      ijkl_sym(50) = 48
      !                      pz       s   -   dx2-y2  dx2-y2
      ijkl_ind(2, 43) = 51
      !                      pz       s   -      dxy     dxy
      ijkl_ind(2, 45) = 52
      ijkl_sym(52) = 51
      ! ################     pz      pz   #######################
      !                      pz      pz   -        s       s
      ijkl_ind(10, 1) = 11
      !                      pz      pz   -       pz       s
      ijkl_ind(10, 2) = 12
      !                      pz      pz   -       pz      pz
      ijkl_ind(10, 10) = 13
      !                      pz      pz   -       px      px
      ijkl_ind(10, 18) = 14
      !                      pz      pz   -       py      py
      ijkl_ind(10, 25) = 15
      ijkl_sym(15) = 14
      !                      pz      pz   -      dz2       s
      ijkl_ind(10, 5) = 53
      !                      pz      pz   -      dz2      pz
      ijkl_ind(10, 13) = 54
      !                      pz      pz   -      dz2     dz2
      ijkl_ind(10, 31) = 55
      !                      pz      pz   -      dzx      px
      ijkl_ind(10, 21) = 56
      !                      pz      pz   -      dzx     dzx
      ijkl_ind(10, 36) = 57
      !                      pz      pz   -      dzy      py
      ijkl_ind(10, 28) = 58
      ijkl_sym(58) = 56
      !                      pz      pz   -      dzy     dzy
      ijkl_ind(10, 40) = 59
      ijkl_sym(59) = 57
      !                      pz      pz   -   dx2-y2  dx2-y2
      ijkl_ind(10, 43) = 60
      !                      pz      pz   -      dxy     dxy
      ijkl_ind(10, 45) = 61
      ijkl_sym(61) = 60
      ! ################     px       s   #######################
      !                      px       s   -       px       s
      ijkl_ind(3, 3) = 16
      !                      px       s   -       px      pz
      ijkl_ind(3, 11) = 17
      !                      px       s   -      dz2      px
      ijkl_ind(3, 20) = 62
      !                      px       s   -      dzx       s
      ijkl_ind(3, 6) = 63
      !                      px       s   -      dzx      pz
      ijkl_ind(3, 14) = 64
      !                      px       s   -      dzx     dz2
      ijkl_ind(3, 32) = 65
      !                      px       s   -   dx2-y2      px
      ijkl_ind(3, 23) = 66
      !                      px       s   -   dx2-y2     dzx
      ijkl_ind(3, 38) = 67
      !                      px       s   -      dxy      py
      ijkl_ind(3, 30) = 68
      ijkl_sym(68) = 66
      !                      px       s   -      dxy     dzy
      ijkl_ind(3, 42) = 69
      ijkl_sym(69) = 67
      ! ################     px      pz   #######################
      !                      px      pz   -       px       s
      ijkl_ind(11, 3) = 18
      !                      px      pz   -       px      pz
      ijkl_ind(11, 11) = 19
      !                      px      pz   -      dz2      px
      ijkl_ind(11, 20) = 70
      !                      px      pz   -      dzx       s
      ijkl_ind(11, 6) = 71
      !                      px      pz   -      dzx      pz
      ijkl_ind(11, 14) = 72
      !                      px      pz   -      dzx     dz2
      ijkl_ind(11, 32) = 73
      !                      px      pz   -   dx2-y2      px
      ijkl_ind(11, 23) = 74
      !                      px      pz   -   dx2-y2     dzx
      ijkl_ind(11, 38) = 75
      !                      px      pz   -      dxy      py
      ijkl_ind(11, 30) = 76
      ijkl_sym(76) = 74
      !                      px      pz   -      dxy     dzy
      ijkl_ind(11, 42) = 77
      ijkl_sym(77) = 75
      ! ################     px      px   #######################
      !                      px      px   -        s       s
      ijkl_ind(18, 1) = 20
      !                      px      px   -       pz       s
      ijkl_ind(18, 2) = 21
      !                      px      px   -       pz      pz
      ijkl_ind(18, 10) = 22
      !                      px      px   -       px      px
      ijkl_ind(18, 18) = 23
      !                      px      px   -       py      py
      ijkl_ind(18, 25) = 24
      !                      px      px   -      dz2       s
      ijkl_ind(18, 5) = 78
      !                      px      px   -      dz2      pz
      ijkl_ind(18, 13) = 79
      !                      px      px   -      dz2     dz2
      ijkl_ind(18, 31) = 80
      !                      px      px   -      dzx      px
      ijkl_ind(18, 21) = 81
      !                      px      px   -      dzx     dzx
      ijkl_ind(18, 36) = 82
      !                      px      px   -      dzy      py
      ijkl_ind(18, 28) = 83
      !                      px      px   -      dzy     dzy
      ijkl_ind(18, 40) = 84
      !                      px      px   -   dx2-y2       s
      ijkl_ind(18, 8) = 85
      !                      px      px   -   dx2-y2      pz
      ijkl_ind(18, 16) = 86
      !                      px      px   -   dx2-y2     dz2
      ijkl_ind(18, 34) = 87
      !                      px      px   -   dx2-y2  dx2-y2
      ijkl_ind(18, 43) = 88
      !                      px      px   -      dxy     dxy
      ijkl_ind(18, 45) = 89
      ijkl_sym(89) = 88
      ! ################     py       s   #######################
      !                      py       s   -       py       s
      ijkl_ind(4, 4) = 25
      ijkl_sym(25) = 16
      !                      py       s   -       py      pz
      ijkl_ind(4, 12) = 26
      ijkl_sym(26) = 17
      !                      py       s   -      dz2      py
      ijkl_ind(4, 26) = 90
      ijkl_sym(90) = 62
      !                      py       s   -      dzy       s
      ijkl_ind(4, 7) = 91
      ijkl_sym(91) = 63
      !                      py       s   -      dzy      pz
      ijkl_ind(4, 15) = 92
      ijkl_sym(92) = 64
      !                      py       s   -      dzy     dz2
      ijkl_ind(4, 33) = 93
      ijkl_sym(93) = 65
      !                      py       s   -   dx2-y2      py
      ijkl_ind(4, 29) = 94
      ijkl_sym(94) = 66*(-1)
      !                      py       s   -   dx2-y2     dzy
      ijkl_ind(4, 41) = 95
      ijkl_sym(95) = 67*(-1)
      !                      py       s   -      dxy      px
      ijkl_ind(4, 24) = 96
      ijkl_sym(96) = 66
      !                      py       s   -      dxy     dzx
      ijkl_ind(4, 39) = 97
      ijkl_sym(97) = 67
      ! ################     py      pz   #######################
      !                      py      pz   -       py       s
      ijkl_ind(12, 4) = 27
      ijkl_sym(27) = 18
      !                      py      pz   -       py      pz
      ijkl_ind(12, 12) = 28
      ijkl_sym(28) = 19
      !                      py      pz   -      dzy       s
      ijkl_ind(12, 26) = 98
      ijkl_sym(98) = 70
      !                      py      pz   -      dzy      pz
      ijkl_ind(12, 7) = 99
      ijkl_sym(99) = 71
      !                      py      pz   -      dzy     dz2
      ijkl_ind(12, 15) = 100
      ijkl_sym(100) = 72
      !                      py      pz   -      dzy     dz2
      ijkl_ind(12, 33) = 101
      ijkl_sym(101) = 73
      !                      py      pz   -   dx2-y2      py
      ijkl_ind(12, 29) = 102
      ijkl_sym(102) = 74*(-1)
      !                      py      pz   -   dx2-y2     dzy
      ijkl_ind(12, 41) = 103
      ijkl_sym(103) = 75*(-1)
      !                      py      pz   -      dxy      px
      ijkl_ind(12, 24) = 104
      ijkl_sym(104) = 74
      !                      py      pz   -      dxy     dzx
      ijkl_ind(12, 39) = 105
      ijkl_sym(105) = 75
      ! ################     py      px   #######################
      !                      py      px   -       py      px
      ijkl_ind(19, 19) = 29
      !                      py      px   -      dzx      py
      ijkl_ind(19, 27) = 106
      ijkl_sym(106) = 86
      !                      py      px   -      dzy      px
      ijkl_ind(19, 22) = 107
      ijkl_sym(107) = 86
      !                      py      px   -      dzy     dzx
      ijkl_ind(19, 37) = 108
      !                      py      px   -      dxy       s
      ijkl_ind(19, 9) = 109
      ijkl_sym(109) = 85
      !                      py      px   -      dxy      pz
      ijkl_ind(19, 17) = 110
      ijkl_sym(110) = 86
      !                      py      px   -      dxy     dz2
      ijkl_ind(19, 35) = 111
      ijkl_sym(111) = 87
      ! ################     py      py   #######################
      !                      py      py   -        s       s
      ijkl_ind(25, 1) = 30
      ijkl_sym(30) = 20
      !                      py      py   -       pz       s
      ijkl_ind(25, 2) = 31
      ijkl_sym(31) = 21
      !                      py      py   -       pz      pz
      ijkl_ind(25, 10) = 32
      ijkl_sym(32) = 22
      !                      py      py   -       px      px
      ijkl_ind(25, 18) = 33
      ijkl_sym(33) = 24
      !                      py      py   -       py      py
      ijkl_ind(25, 25) = 34
      ijkl_sym(34) = 23
      !                      py      py   -      dz2       s
      ijkl_ind(25, 5) = 112
      ijkl_sym(112) = 78
      !                      py      py   -      dz2      pz
      ijkl_ind(25, 13) = 113
      ijkl_sym(113) = 79
      !                      py      py   -      dz2     dz2
      ijkl_ind(25, 31) = 114
      ijkl_sym(114) = 80
      !                      py      py   -      dzx      px
      ijkl_ind(25, 21) = 115
      ijkl_sym(115) = 83
      !                      py      py   -      dzx     dzx
      ijkl_ind(25, 36) = 116
      ijkl_sym(116) = 84
      !                      py      py   -      dzy      py
      ijkl_ind(25, 28) = 117
      ijkl_sym(117) = 81
      !                      py      py   -      dzy     dzy
      ijkl_ind(25, 40) = 118
      ijkl_sym(118) = 82
      !                      py      py   -   dx2-y2       s
      ijkl_ind(25, 8) = 119
      ijkl_sym(119) = 85*(-1)
      !                      py      py   -   dx2-y2      pz
      ijkl_ind(25, 16) = 120
      ijkl_sym(120) = 86*(-1)
      !                      py      py   -   dx2-y2     dz2
      ijkl_ind(25, 34) = 121
      ijkl_sym(121) = 87*(-1)
      !                      py      py   -   dx2-y2  dx2-y2
      ijkl_ind(25, 43) = 122
      ijkl_sym(122) = 88
      !                      py      py   -      dxy     dxy
      ijkl_ind(25, 45) = 123
      ijkl_sym(123) = 88
      ! ################    dz2     dz2   #######################
      !                     dz2       s   -        s       s
      ijkl_ind(5, 1) = 124
      !                     dz2       s   -       pz       s
      ijkl_ind(5, 2) = 125
      !                     dz2       s   -       pz      pz
      ijkl_ind(5, 10) = 126
      !                     dz2       s   -       px      px
      ijkl_ind(5, 18) = 127
      !                     dz2       s   -       py      py
      ijkl_ind(5, 25) = 128
      ijkl_sym(128) = 127
      !                     dz2       s   -      dz2       s
      ijkl_ind(5, 5) = 129
      !                     dz2       s   -      dz2      pz
      ijkl_ind(5, 13) = 130
      !                     dz2       s   -      dz2     dz2
      ijkl_ind(5, 31) = 131
      !                     dz2       s   -      dzx      px
      ijkl_ind(5, 21) = 132
      !                     dz2       s   -      dzx     dzx
      ijkl_ind(5, 36) = 133
      !                     dz2       s   -      dzy      py
      ijkl_ind(5, 28) = 134
      ijkl_sym(134) = 132
      !                     dz2       s   -      dzy     dzy
      ijkl_ind(5, 40) = 135
      ijkl_sym(135) = 133
      !                     dz2       s   -   dx2-y2  dx2-y2
      ijkl_ind(5, 43) = 136
      !                     dz2       s   -      dxy     dxy
      ijkl_ind(5, 45) = 137
      ijkl_sym(137) = 136
      ! ################    dz2      pz   #######################
      !                     dz2      pz   -        s       s
      ijkl_ind(13, 1) = 138
      !                     dz2      pz   -       pz       s
      ijkl_ind(13, 2) = 139
      !                     dz2      pz   -       pz      pz
      ijkl_ind(13, 10) = 140
      !                     dz2      pz   -       px      px
      ijkl_ind(13, 18) = 141
      !                     dz2      pz   -       py      py
      ijkl_ind(13, 25) = 142
      ijkl_sym(142) = 141
      !                     dz2      pz   -      dz2       s
      ijkl_ind(13, 5) = 143
      !                     dz2      pz   -      dz2      pz
      ijkl_ind(13, 13) = 144
      !                     dz2      pz   -      dz2     dz2
      ijkl_ind(13, 31) = 145
      !                     dz2      pz   -      dzx      px
      ijkl_ind(13, 21) = 146
      !                     dz2      pz   -      dzx     dzx
      ijkl_ind(13, 36) = 147
      !                     dz2      pz   -      dzy      py
      ijkl_ind(13, 28) = 148
      ijkl_sym(148) = 146
      !                     dz2      pz   -      dzy     dzy
      ijkl_ind(13, 40) = 149
      ijkl_sym(149) = 147
      !                     dz2      pz   -   dx2-y2  dx2-y2
      ijkl_ind(13, 43) = 150
      !                     dz2      pz   -      dxy     dxy
      ijkl_ind(13, 45) = 151
      ijkl_sym(151) = 150
      ! ################    dz2      px   #######################
      !                     dz2      px   -       px       s
      ijkl_ind(20, 3) = 152
      !                     dz2      px   -       px      pz
      ijkl_ind(20, 11) = 153
      !                     dz2      px   -      dz2      px
      ijkl_ind(20, 20) = 154
      !                     dz2      px   -      dzx       s
      ijkl_ind(20, 6) = 155
      !                     dz2      px   -      dzx      pz
      ijkl_ind(20, 14) = 156
      !                     dz2      px   -      dzx     dz2
      ijkl_ind(20, 32) = 157
      !                     dz2      px   -   dx2-y2      px
      ijkl_ind(20, 23) = 158
      !                     dz2      px   -   dx2-y2     dzx
      ijkl_ind(20, 38) = 159
      !                     dz2      px   -      dxy      py
      ijkl_ind(20, 30) = 160
      ijkl_sym(160) = 158
      !                     dz2      px   -      dxy     dxy
      ijkl_ind(20, 42) = 161
      ijkl_sym(161) = 159
      ! ################    dz2      py   #######################
      !                     dz2      py   -       py       s
      ijkl_ind(26, 4) = 162
      ijkl_sym(162) = 152
      !                     dz2      py   -       py      pz
      ijkl_ind(26, 12) = 163
      ijkl_sym(163) = 153
      !                     dz2      py   -      dz2      py
      ijkl_ind(26, 26) = 164
      ijkl_sym(164) = 154
      !                     dz2      py   -      dzy       s
      ijkl_ind(26, 7) = 165
      ijkl_sym(165) = 155
      !                     dz2      py   -      dzy      pz
      ijkl_ind(26, 15) = 166
      ijkl_sym(166) = 156
      !                     dz2      py   -      dzy     dz2
      ijkl_ind(26, 33) = 167
      ijkl_sym(167) = 157
      !                     dz2      py   -   dx2-y2      py
      ijkl_ind(26, 29) = 168
      ijkl_sym(168) = 158*(-1)
      !                     dz2      py   -   dx2-y2     dzy
      ijkl_ind(26, 41) = 169
      ijkl_sym(169) = 159*(-1)
      !                     dz2      py   -      dxy      px
      ijkl_ind(26, 24) = 170
      ijkl_sym(170) = 158
      !                     dz2      py   -      dxy     dzx
      ijkl_ind(26, 39) = 171
      ijkl_sym(171) = 159
      ! ################    dz2     dz2   #######################
      !                     dz2     dz2   -        s       s
      ijkl_ind(31, 1) = 172
      !                     dz2     dz2   -       pz       s
      ijkl_ind(31, 2) = 173
      !                     dz2     dz2   -       pz      pz
      ijkl_ind(31, 10) = 174
      !                     dz2     dz2   -       px      px
      ijkl_ind(31, 18) = 175
      !                     dz2     dz2   -       py      py
      ijkl_ind(31, 25) = 176
      ijkl_sym(176) = 175
      !                     dz2     dz2   -      dz2       s
      ijkl_ind(31, 5) = 177
      !                     dz2     dz2   -      dz2      pz
      ijkl_ind(31, 13) = 178
      !                     dz2     dz2   -      dz2     dz2
      ijkl_ind(31, 31) = 179
      !                     dz2     dz2   -      dzx      px
      ijkl_ind(31, 21) = 180
      !                     dz2     dz2   -      dzx     dzx
      ijkl_ind(31, 36) = 181
      !                     dz2     dz2   -      dzy      py
      ijkl_ind(31, 28) = 182
      ijkl_sym(182) = 180
      !                     dz2     dz2   -      dzy     dzy
      ijkl_ind(31, 40) = 183
      ijkl_sym(183) = 181
      !                     dz2     dz2   -   dx2-y2  dx2-y2
      ijkl_ind(31, 43) = 184
      !                     dz2     dz2   -      dxy     dxy
      ijkl_ind(31, 45) = 185
      ijkl_sym(185) = 184
      ! ################    dzx       s   #######################
      !                     dzx       s   -       px       s
      ijkl_ind(6, 3) = 186
      !                     dzx       s   -       px      pz
      ijkl_ind(6, 11) = 187
      !                     dzx       s   -      dz2      px
      ijkl_ind(6, 20) = 188
      !                     dzx       s   -      dzx       s
      ijkl_ind(6, 6) = 189
      !                     dzx       s   -      dzx      pz
      ijkl_ind(6, 14) = 190
      !                     dzx       s   -      dzx     dz2
      ijkl_ind(6, 32) = 191
      !                     dzx       s   -   dx2-y2      px
      ijkl_ind(6, 23) = 192
      !                     dzx       s   -   dx2-y2  dx2-y2
      ijkl_ind(6, 38) = 193
      !                     dzx       s   -      dxy      py
      ijkl_ind(6, 30) = 194
      ijkl_sym(194) = 192
      !                     dzx       s   -      dxy     dzy
      ijkl_ind(6, 42) = 195
      ijkl_sym(195) = 193
      ! ################    dzx      pz   #######################
      !                     dzx      pz   -       px       s
      ijkl_ind(14, 3) = 196
      !                     dzx      pz   -       px      pz
      ijkl_ind(14, 11) = 197
      !                     dzx      pz   -      dz2      px
      ijkl_ind(14, 20) = 198
      !                     dzx      pz   -      dzx       s
      ijkl_ind(14, 6) = 199
      !                     dzx      pz   -      dzx      pz
      ijkl_ind(14, 14) = 200
      !                     dzx      pz   -      dzx     dz2
      ijkl_ind(14, 32) = 201
      !                     dzx      pz   -   dx2-y2      px
      ijkl_ind(14, 23) = 202
      !                     dzx      pz   -   dx2-y2  dx2-y2
      ijkl_ind(14, 38) = 203
      !                     dzx      pz   -      dxy      py
      ijkl_ind(14, 30) = 204
      ijkl_sym(204) = 202
      !                     dzx      pz   -      dxy     dzy
      ijkl_ind(14, 42) = 205
      ijkl_sym(205) = 203
      ! ################    dzx      px   #######################
      !                     dzx      px   -        s       s
      ijkl_ind(21, 1) = 206
      !                     dzx      px   -       pz       s
      ijkl_ind(21, 2) = 207
      !                     dzx      px   -       pz      pz
      ijkl_ind(21, 10) = 208
      !                     dzx      px   -       px      px
      ijkl_ind(21, 18) = 209
      !                     dzx      px   -       py      py
      ijkl_ind(21, 25) = 210
      !                     dzx      px   -      dz2       s
      ijkl_ind(21, 5) = 211
      !                     dzx      px   -      dz2      pz
      ijkl_ind(21, 13) = 212
      !                     dzx      px   -      dz2     dz2
      ijkl_ind(21, 31) = 213
      !                     dzx      px   -      dzx      px
      ijkl_ind(21, 21) = 214
      !                     dzx      px   -      dzx     dzx
      ijkl_ind(21, 36) = 215
      !                     dzx      px   -      dzy      py
      ijkl_ind(21, 28) = 216
      !                     dzx      px   -      dzy     dzy
      ijkl_ind(21, 40) = 217
      !                     dzx      px   -   dx2-y2       s
      ijkl_ind(21, 8) = 218
      !                     dzx      px   -   dx2-y2      pz
      ijkl_ind(21, 16) = 219
      !                     dzx      px   -   dx2-y2     dz2
      ijkl_ind(21, 34) = 220
      !                     dzx      px   -   dx2-y2  dx2-y2
      ijkl_ind(21, 43) = 221
      !                     dzx      px   -      dxy     dxy
      ijkl_ind(21, 45) = 222
      ijkl_sym(222) = 221
      ! ################    dzx      py   #######################
      !                     dzx      py   -       px      py
      ijkl_ind(27, 19) = 223
      !                     dzx      py   -      dzx      py
      ijkl_ind(27, 27) = 224
      ijkl_sym(224) = 219
      !                     dzx      py   -      dzy      px
      ijkl_ind(27, 22) = 225
      ijkl_sym(225) = 219
      !                     dzx      py   -      dzy     dzx
      ijkl_ind(27, 37) = 226
      !                     dzx      py   -      dxy       s
      ijkl_ind(27, 9) = 227
      ijkl_sym(227) = 218
      !                     dzx      py   -      dxy      pz
      ijkl_ind(27, 17) = 228
      ijkl_sym(228) = 219
      !                     dzx      py   -      dxy     dz2
      ijkl_ind(27, 35) = 229
      ijkl_sym(229) = 220
      ! ################    dzx     dz2   #######################
      !                     dzx     dz2   -       px       s
      ijkl_ind(32, 3) = 230
      !                     dzx     dz2   -       px      pz
      ijkl_ind(32, 11) = 231
      !                     dzx     dz2   -      dz2      px
      ijkl_ind(32, 20) = 232
      !                     dzx     dz2   -      dzx       s
      ijkl_ind(32, 6) = 233
      !                     dzx     dz2   -      dzx      pz
      ijkl_ind(32, 14) = 234
      !                     dzx     dz2   -      dzx     dz2
      ijkl_ind(32, 32) = 235
      !                     dzx     dz2   -   dx2-y2      px
      ijkl_ind(32, 23) = 236
      !                     dzx     dz2   -   dx2-y2  dx2-y2
      ijkl_ind(32, 38) = 237
      !                     dzx     dz2   -      dxy      py
      ijkl_ind(32, 30) = 238
      ijkl_sym(238) = 236
      !                     dzx     dz2   -      dxy     dzy
      ijkl_ind(32, 42) = 239
      ijkl_sym(239) = 237
      ! ################    dzx     dzx   #######################
      !                     dzx     dzx   -        s       s
      ijkl_ind(36, 1) = 240
      !                     dzx     dzx   -       pz       s
      ijkl_ind(36, 2) = 241
      !                     dzx     dzx   -       pz      pz
      ijkl_ind(36, 10) = 242
      !                     dzx     dzx   -       px      px
      ijkl_ind(36, 18) = 243
      !                     dzx     dzx   -       py      py
      ijkl_ind(36, 25) = 244
      !                     dzx     dzx   -      dz2       s
      ijkl_ind(36, 5) = 245
      !                     dzx     dzx   -      dz2      pz
      ijkl_ind(36, 13) = 246
      !                     dzx     dzx   -      dz2     dz2
      ijkl_ind(36, 31) = 247
      !                     dzx     dzx   -      dzx      px
      ijkl_ind(36, 21) = 248
      !                     dzx     dzx   -      dzx     dzx
      ijkl_ind(36, 36) = 249
      !                     dzx     dzx   -      dzy      py
      ijkl_ind(36, 28) = 250
      !                     dzx     dzx   -      dzy     dzy
      ijkl_ind(36, 40) = 251
      !                     dzx     dzx   -   dx2-y2       s
      ijkl_ind(36, 8) = 252
      !                     dzx     dzx   -   dx2-y2      pz
      ijkl_ind(36, 16) = 253
      !                     dzx     dzx   -   dx2-y2     dz2
      ijkl_ind(36, 34) = 254
      !                     dzx     dzx   -   dx2-y2  dx2-y2
      ijkl_ind(36, 43) = 255
      !                     dzx     dzx   -      dxy     dxy
      ijkl_ind(36, 45) = 256
      ijkl_sym(256) = 255
      ! ################    dzy       s   #######################
      !                     dzy       s   -       py       s
      ijkl_ind(7, 4) = 257
      ijkl_sym(257) = 186
      !                     dzy       s   -       py      pz
      ijkl_ind(7, 12) = 258
      ijkl_sym(258) = 187
      !                     dzy       s   -      dz2      py
      ijkl_ind(7, 26) = 259
      ijkl_sym(259) = 188
      !                     dzy       s   -      dzy       s
      ijkl_ind(7, 7) = 260
      ijkl_sym(260) = 189
      !                     dzy       s   -      dzy      pz
      ijkl_ind(7, 15) = 261
      ijkl_sym(261) = 190
      !                     dzy       s   -      dzy     dz2
      ijkl_ind(7, 33) = 262
      ijkl_sym(262) = 191
      !                     dzy       s   -   dx2-y2      py
      ijkl_ind(7, 29) = 263
      ijkl_sym(263) = 192*(-1)
      !                     dzy       s   -   dx2-y2     dzy
      ijkl_ind(7, 41) = 264
      ijkl_sym(264) = 193*(-1)
      !                     dzy       s   -      dxy      px
      ijkl_ind(7, 24) = 265
      ijkl_sym(265) = 192
      !                     dzy       s   -      dxy     dzx
      ijkl_ind(7, 39) = 266
      ijkl_sym(266) = 193
      ! ################    dzy      pz   #######################
      !                     dzy      pz   -       py       s
      ijkl_ind(15, 4) = 267
      ijkl_sym(267) = 196
      !                     dzy      pz   -       py      pz
      ijkl_ind(15, 12) = 268
      ijkl_sym(268) = 197
      !                     dzy      pz   -      dz2      py
      ijkl_ind(15, 26) = 269
      ijkl_sym(269) = 198
      !                     dzy      pz   -      dzy       s
      ijkl_ind(15, 7) = 270
      ijkl_sym(270) = 199
      !                     dzy      pz   -      dzy      pz
      ijkl_ind(15, 15) = 271
      ijkl_sym(271) = 200
      !                     dzy      pz   -      dzy     dz2
      ijkl_ind(15, 33) = 272
      ijkl_sym(272) = 201
      !                     dzy      pz   -   dx2-y2      py
      ijkl_ind(15, 29) = 273
      ijkl_sym(273) = 202*(-1)
      !                     dzy      pz   -   dx2-y2     dzy
      ijkl_ind(15, 41) = 274
      ijkl_sym(274) = 203*(-1)
      !                     dzy      pz   -      dxy      px
      ijkl_ind(15, 24) = 275
      ijkl_sym(275) = 202
      !                     dzy      pz   -      dxy     dzx
      ijkl_ind(15, 39) = 276
      ijkl_sym(276) = 203
      ! ################    dzy      px   #######################
      !                     dzy      px   -       px      py
      ijkl_ind(22, 19) = 277
      ijkl_sym(277) = 223
      !                     dzy      px   -      dzx      py
      ijkl_ind(22, 27) = 278
      ijkl_sym(278) = 219
      !                     dzy      px   -      dzy      px
      ijkl_ind(22, 22) = 279
      ijkl_sym(279) = 219
      !                     dzy      px   -      dzy     dzx
      ijkl_ind(22, 37) = 280
      ijkl_sym(280) = 226
      !                     dzy      px   -      dxy       s
      ijkl_ind(22, 9) = 281
      ijkl_sym(281) = 218
      !                     dzy      px   -      dxy      pz
      ijkl_ind(22, 17) = 282
      ijkl_sym(282) = 219
      !                     dzy      px   -      dxy     dz2
      ijkl_ind(22, 35) = 283
      ijkl_sym(283) = 220
      ! ################    dzy      py   #######################
      !                     dzy      py   -        s       s
      ijkl_ind(28, 1) = 284
      ijkl_sym(284) = 206
      !                     dzy      py   -       pz       s
      ijkl_ind(28, 2) = 285
      ijkl_sym(285) = 207
      !                     dzy      py   -       pz      pz
      ijkl_ind(28, 10) = 286
      ijkl_sym(286) = 208
      !                     dzy      py   -       px      px
      ijkl_ind(28, 18) = 287
      ijkl_sym(287) = 210
      !                     dzy      py   -       py      py
      ijkl_ind(28, 25) = 288
      ijkl_sym(288) = 209
      !                     dzy      py   -      dz2       s
      ijkl_ind(28, 5) = 289
      ijkl_sym(289) = 211
      !                     dzy      py   -      dz2      pz
      ijkl_ind(28, 13) = 290
      ijkl_sym(290) = 212
      !                     dzy      py   -      dz2     dz2
      ijkl_ind(28, 31) = 291
      ijkl_sym(291) = 213
      !                     dzy      py   -      dzx      px
      ijkl_ind(28, 21) = 292
      ijkl_sym(292) = 216
      !                     dzy      py   -      dzx     dzx
      ijkl_ind(28, 36) = 293
      ijkl_sym(293) = 217
      !                     dzy      py   -      dzy      py
      ijkl_ind(28, 28) = 294
      ijkl_sym(294) = 214
      !                     dzy      py   -      dzy     dzy
      ijkl_ind(28, 40) = 295
      ijkl_sym(295) = 215
      !                     dzy      py   -   dx2-y2       s
      ijkl_ind(28, 8) = 296
      ijkl_sym(296) = 218*(-1)
      !                     dzy      py   -   dx2-y2      pz
      ijkl_ind(28, 16) = 297
      ijkl_sym(297) = 219*(-1)
      !                     dzy      py   -   dx2-y2     dz2
      ijkl_ind(28, 34) = 298
      ijkl_sym(298) = 220*(-1)
      !                     dzy      py   -   dx2-y2  dx2-y2
      ijkl_ind(28, 43) = 299
      ijkl_sym(299) = 221
      !                     dzy      py   -      dxy     dxy
      ijkl_ind(28, 45) = 300
      ijkl_sym(300) = 221
      ! ################    dzy     dz2   #######################
      !                     dzy     dz2   -       py       s
      ijkl_ind(33, 4) = 301
      ijkl_sym(301) = 230
      !                     dzy     dz2   -       py      pz
      ijkl_ind(33, 12) = 302
      ijkl_sym(302) = 231
      !                     dzy     dz2   -      dz2      py
      ijkl_ind(33, 26) = 303
      ijkl_sym(303) = 232
      !                     dzy     dz2   -      dzy       s
      ijkl_ind(33, 7) = 304
      ijkl_sym(304) = 233
      !                     dzy     dz2   -      dzy      pz
      ijkl_ind(33, 15) = 305
      ijkl_sym(305) = 234
      !                     dzy     dz2   -      dzy     dz2
      ijkl_ind(33, 33) = 306
      ijkl_sym(306) = 235
      !                     dzy     dz2   -   dx2-y2      py
      ijkl_ind(33, 29) = 307
      ijkl_sym(307) = 236*(-1)
      !                     dzy     dz2   -   dx2-y2     dzy
      ijkl_ind(33, 41) = 308
      ijkl_sym(308) = 237*(-1)
      !                     dzy     dz2   -      dxy      px
      ijkl_ind(33, 24) = 309
      ijkl_sym(309) = 236
      !                     dzy     dz2   -      dxy     dzx
      ijkl_ind(33, 39) = 310
      ijkl_sym(310) = 237
      ! ################    dzy     dzx   #######################
      !                     dzy     dzx   -       px      py
      ijkl_ind(37, 19) = 311
      !                     dzy     dzx   -      dzx      py
      ijkl_ind(37, 27) = 312
      ijkl_sym(312) = 253
      !                     dzy     dzx   -      dzy      px
      ijkl_ind(37, 22) = 313
      ijkl_sym(313) = 253
      !                     dzy     dzx   -      dzy     dzx
      ijkl_ind(37, 37) = 314
      !                     dzy     dzx   -      dxy       s
      ijkl_ind(37, 9) = 315
      ijkl_sym(315) = 252
      !                     dzy     dzx   -      dxy      pz
      ijkl_ind(37, 17) = 316
      ijkl_sym(316) = 253
      !                     dzy     dzx   -      dxy     dz2
      ijkl_ind(37, 35) = 317
      ijkl_sym(317) = 254
      ! ################    dzy     dzy   #######################
      !                     dzy     dzy   -        s       s
      ijkl_ind(40, 1) = 318
      ijkl_sym(318) = 240
      !                     dzy     dzy   -       pz       s
      ijkl_ind(40, 2) = 319
      ijkl_sym(319) = 241
      !                     dzy     dzy   -       pz      pz
      ijkl_ind(40, 10) = 320
      ijkl_sym(320) = 242
      !                     dzy     dzy   -       px      px
      ijkl_ind(40, 18) = 321
      ijkl_sym(321) = 244
      !                     dzy     dzy   -       py      py
      ijkl_ind(40, 25) = 322
      ijkl_sym(322) = 243
      !                     dzy     dzy   -      dz2       s
      ijkl_ind(40, 5) = 323
      ijkl_sym(323) = 245
      !                     dzy     dzy   -      dz2      pz
      ijkl_ind(40, 13) = 324
      ijkl_sym(324) = 246
      !                     dzy     dzy   -      dz2     dz2
      ijkl_ind(40, 31) = 325
      ijkl_sym(325) = 247
      !                     dzy     dzy   -      dzx      px
      ijkl_ind(40, 21) = 326
      ijkl_sym(326) = 250
      !                     dzy     dzy   -      dzx     dzx
      ijkl_ind(40, 36) = 327
      ijkl_sym(327) = 251
      !                     dzy     dzy   -      dzy      py
      ijkl_ind(40, 28) = 328
      ijkl_sym(328) = 248
      !                     dzy     dzy   -      dzy     dzy
      ijkl_ind(40, 40) = 329
      ijkl_sym(329) = 249
      !                     dzy     dzy   -   dx2-y2       s
      ijkl_ind(40, 8) = 330
      ijkl_sym(330) = 252*(-1)
      !                     dzy     dzy   -   dx2-y2      pz
      ijkl_ind(40, 16) = 331
      ijkl_sym(331) = 253*(-1)
      !                     dzy     dzy   -   dx2-y2     dz2
      ijkl_ind(40, 34) = 332
      ijkl_sym(332) = 254*(-1)
      !                     dzy     dzy   -   dx2-y2  dx2-y2
      ijkl_ind(40, 43) = 333
      ijkl_sym(333) = 255
      !                     dzy     dzy   -      dxy     dxy
      ijkl_ind(40, 45) = 334
      ijkl_sym(334) = 255
      ! ################ dx2-y2       s   #######################
      !                  dx2-y2       s   -       px      px
      ijkl_ind(8, 18) = 335
      !                  dx2-y2       s   -       py      py
      ijkl_ind(8, 25) = 336
      ijkl_sym(336) = 335*(-1)
      !                  dx2-y2       s   -      dzx      px
      ijkl_ind(8, 21) = 337
      !                  dx2-y2       s   -      dzx     dzx
      ijkl_ind(8, 36) = 338
      !                  dx2-y2       s   -      dzy      py
      ijkl_ind(8, 28) = 339
      ijkl_sym(339) = 337*(-1)
      !                  dx2-y2       s   -      dzy     dzy
      ijkl_ind(8, 40) = 340
      ijkl_sym(340) = 338*(-1)
      !                  dx2-y2       s   -   dx2-y2       s
      ijkl_ind(8, 8) = 341
      !                  dx2-y2       s   -   dx2-y2      pz
      ijkl_ind(8, 16) = 342
      ijkl_sym(342) = 337
      !                  dx2-y2       s   -   dx2-y2     dz2
      ijkl_ind(8, 34) = 343
      ! ################ dx2-y2      pz   #######################
      !                  dx2-y2      pz   -       px      px
      ijkl_ind(16, 18) = 344
      ijkl_sym(344) = 223
      !                  dx2-y2      pz   -       py      py
      ijkl_ind(16, 25) = 345
      ijkl_sym(345) = 223*(-1)
      !                  dx2-y2      pz   -      dzx      px
      ijkl_ind(16, 21) = 346
      ijkl_sym(346) = 219
      !                  dx2-y2      pz   -      dzx     dzx
      ijkl_ind(16, 36) = 347
      ijkl_sym(347) = 226
      !                  dx2-y2      pz   -      dzy      py
      ijkl_ind(16, 28) = 348
      ijkl_sym(348) = 219*(-1)
      !                  dx2-y2      pz   -      dzy     dzy
      ijkl_ind(16, 40) = 349
      ijkl_sym(349) = 226*(-1)
      !                  dx2-y2      pz   -   dx2-y2       s
      ijkl_ind(16, 8) = 350
      ijkl_sym(350) = 218
      !                  dx2-y2      pz   -   dx2-y2      pz
      ijkl_ind(16, 16) = 351
      ijkl_sym(351) = 219
      !                  dx2-y2      pz   -   dx2-y2     dz2
      ijkl_ind(16, 34) = 352
      ijkl_sym(352) = 220
      ! ################ dx2-y2      px   #######################
      !                  dx2-y2      px   -       px       s
      ijkl_ind(23, 3) = 353
      !                  dx2-y2      px   -       px      pz
      ijkl_ind(23, 11) = 354
      !                  dx2-y2      px   -      dz2      px
      ijkl_ind(23, 20) = 355
      !                  dx2-y2      px   -      dzx       s
      ijkl_ind(23, 6) = 356
      !                  dx2-y2      px   -      dzx      pz
      ijkl_ind(23, 14) = 357
      !                  dx2-y2      px   -      dzx     dz2
      ijkl_ind(23, 32) = 358
      !                  dx2-y2      px   -   dx2-y2      px
      ijkl_ind(23, 23) = 359
      !                  dx2-y2      px   -   dx2-y2  dx2-y2
      ijkl_ind(23, 38) = 360
      !                  dx2-y2      px   -      dxy      py
      ijkl_ind(23, 30) = 361
      !                  dx2-y2      px   -      dxy     dzy
      ijkl_ind(23, 42) = 362
      ! ################ dx2-y2      py   #######################
      !                  dx2-y2      py   -       py       s
      ijkl_ind(29, 4) = 363
      ijkl_sym(363) = 353*(-1)
      !                  dx2-y2      py   -       py      pz
      ijkl_ind(29, 12) = 364
      ijkl_sym(364) = 354*(-1)
      !                  dx2-y2      py   -      dz2      py
      ijkl_ind(29, 26) = 365
      ijkl_sym(365) = 355*(-1)
      !                  dx2-y2      py   -      dzy       s
      ijkl_ind(29, 7) = 366
      ijkl_sym(366) = 356*(-1)
      !                  dx2-y2      py   -      dzy      pz
      ijkl_ind(29, 15) = 367
      ijkl_sym(367) = 357*(-1)
      !                  dx2-y2      py   -      dzy     dz2
      ijkl_ind(29, 33) = 368
      ijkl_sym(368) = 358*(-1)
      !                  dx2-y2      py   -   dx2-y2      py
      ijkl_ind(29, 29) = 369
      ijkl_sym(369) = 359
      !                  dx2-y2      py   -   dx2-y2     dzy
      ijkl_ind(29, 41) = 370
      ijkl_sym(370) = 360
      !                  dx2-y2      py   -      dxy      px
      ijkl_ind(29, 24) = 371
      ijkl_sym(371) = 361*(-1)
      !                  dx2-y2      py   -      dxy     dzx
      ijkl_ind(29, 39) = 372
      ijkl_sym(372) = 362*(-1)
      ! ################ dx2-y2     dz2   #######################
      !                  dx2-y2     dz2   -       px      px
      ijkl_ind(34, 18) = 373
      !                  dx2-y2     dz2   -       py      py
      ijkl_ind(34, 25) = 374
      ijkl_sym(374) = 373*(-1)
      !                  dx2-y2     dz2   -      dzx      px
      ijkl_ind(34, 21) = 375
      !                  dx2-y2     dz2   -      dzx     dzx
      ijkl_ind(34, 36) = 376
      !                  dx2-y2     dz2   -      dzy      py
      ijkl_ind(34, 28) = 377
      ijkl_sym(377) = 375*(-1)
      !                  dx2-y2     dz2   -      dzy     dzy
      ijkl_ind(34, 40) = 378
      ijkl_sym(378) = 376*(-1)
      !                  dx2-y2     dz2   -   dx2-y2       s
      ijkl_ind(34, 8) = 379
      !                  dx2-y2     dz2   -   dx2-y2      pz
      ijkl_ind(34, 16) = 380
      ijkl_sym(380) = 375
      !                  dx2-y2     dz2   -   dx2-y2     dz2
      ijkl_ind(34, 34) = 381
      ! ################ dx2-y2     dzx   #######################
      !                  dx2-y2     dzx   -       px       s
      ijkl_ind(38, 3) = 382
      !                  dx2-y2     dzx   -       px      pz
      ijkl_ind(38, 11) = 383
      !                  dx2-y2     dzx   -      dz2      px
      ijkl_ind(38, 20) = 384
      !                  dx2-y2     dzx   -      dzx       s
      ijkl_ind(38, 6) = 385
      !                  dx2-y2     dzx   -      dzx      pz
      ijkl_ind(38, 14) = 386
      !                  dx2-y2     dzx   -      dzx     dz2
      ijkl_ind(38, 32) = 387
      !                  dx2-y2     dzx   -   dx2-y2      px
      ijkl_ind(38, 23) = 388
      !                  dx2-y2     dzx   -   dx2-y2  dx2-y2
      ijkl_ind(38, 38) = 389
      !                  dx2-y2     dzx   -      dxy      py
      ijkl_ind(38, 30) = 390
      !                  dx2-y2     dzx   -      dxy     dzy
      ijkl_ind(38, 42) = 391
      ! ################ dx2-y2     dzy   #######################
      !                  dx2-y2     dzy   -       py       s
      ijkl_ind(41, 4) = 392
      ijkl_sym(392) = 382*(-1)
      !                  dx2-y2     dzy   -       py      pz
      ijkl_ind(41, 12) = 393
      ijkl_sym(393) = 383*(-1)
      !                  dx2-y2     dzy   -      dz2      py
      ijkl_ind(41, 26) = 394
      ijkl_sym(394) = 384*(-1)
      !                  dx2-y2     dzy   -      dzy       s
      ijkl_ind(41, 7) = 395
      ijkl_sym(395) = 385*(-1)
      !                  dx2-y2     dzy   -      dzy      pz
      ijkl_ind(41, 15) = 396
      ijkl_sym(396) = 386*(-1)
      !                  dx2-y2     dzy   -      dzy     dz2
      ijkl_ind(41, 33) = 397
      ijkl_sym(397) = 387*(-1)
      !                  dx2-y2     dzy   -   dx2-y2      py
      ijkl_ind(41, 29) = 398
      ijkl_sym(398) = 388
      !                  dx2-y2     dzy   -   dx2-y2     dzy
      ijkl_ind(41, 41) = 399
      ijkl_sym(399) = 389
      !                  dx2-y2     dzy   -      dxy      px
      ijkl_ind(41, 24) = 400
      ijkl_sym(400) = 390*(-1)
      !                  dx2-y2     dzy   -      dxy     dzx
      ijkl_ind(41, 39) = 401
      ijkl_sym(401) = 391*(-1)
      ! ################ dx2-y2  dx2-y2   #######################
      !                  dx2-y2  dx2-y2   -        s       s
      ijkl_ind(43, 1) = 402
      !                  dx2-y2  dx2-y2   -       pz       s
      ijkl_ind(43, 2) = 403
      !                  dx2-y2  dx2-y2   -       pz      pz
      ijkl_ind(43, 10) = 404
      !                  dx2-y2  dx2-y2   -       px      px
      ijkl_ind(43, 18) = 405
      !                  dx2-y2  dx2-y2   -       py      py
      ijkl_ind(43, 25) = 406
      ijkl_sym(406) = 405
      !                  dx2-y2  dx2-y2   -      dz2       s
      ijkl_ind(43, 5) = 407
      !                  dx2-y2  dx2-y2   -      dz2      pz
      ijkl_ind(43, 13) = 408
      !                  dx2-y2  dx2-y2   -      dz2     dz2
      ijkl_ind(43, 31) = 409
      !                  dx2-y2  dx2-y2   -      dzx      px
      ijkl_ind(43, 21) = 410
      !                  dx2-y2  dx2-y2   -      dzx     dzx
      ijkl_ind(43, 36) = 411
      !                  dx2-y2  dx2-y2   -      dzy      py
      ijkl_ind(43, 28) = 412
      ijkl_sym(412) = 410
      !                  dx2-y2  dx2-y2   -      dzy     dzy
      ijkl_ind(43, 40) = 413
      ijkl_sym(413) = 411
      !                  dx2-y2  dx2-y2   -   dx2-y2  dx2-y2
      ijkl_ind(43, 43) = 414
      !                  dx2-y2  dx2-y2   -      dxy     dxy
      ijkl_ind(43, 45) = 415
      ! ################    dxy       s   #######################
      !                     dxy       s   -       px      py
      ijkl_ind(9, 19) = 416
      ijkl_sym(416) = 335
      !                     dxy       s   -      dzx      py
      ijkl_ind(9, 27) = 417
      ijkl_sym(417) = 337
      !                     dxy       s   -      dzy      px
      ijkl_ind(9, 22) = 418
      ijkl_sym(418) = 337
      !                     dxy       s   -      dzy     dzx
      ijkl_ind(9, 37) = 419
      ijkl_sym(419) = 338
      !                     dxy       s   -      dxy       s
      ijkl_ind(9, 9) = 420
      ijkl_sym(420) = 341
      !                     dxy       s   -      dxy      pz
      ijkl_ind(9, 17) = 421
      ijkl_sym(421) = 337
      !                     dxy       s   -      dxy     dz2
      ijkl_ind(9, 35) = 422
      ijkl_sym(422) = 343
      ! ################    dxy      pz   #######################
      !                     dxy      pz   -       px      py
      ijkl_ind(17, 19) = 423
      ijkl_sym(423) = 223
      !                     dxy      pz   -      dzx      py
      ijkl_ind(17, 27) = 424
      ijkl_sym(424) = 219
      !                     dxy      pz   -      dzy      px
      ijkl_ind(17, 22) = 425
      ijkl_sym(425) = 219
      !                     dxy      pz   -      dzy     dzx
      ijkl_ind(17, 37) = 426
      ijkl_sym(426) = 226
      !                     dxy      pz   -      dxy       s
      ijkl_ind(17, 9) = 427
      ijkl_sym(427) = 218
      !                     dxy      pz   -      dxy      pz
      ijkl_ind(17, 17) = 428
      ijkl_sym(428) = 219
      !                     dxy      pz   -      dxy     dz2
      ijkl_ind(17, 35) = 429
      ijkl_sym(429) = 220
      ! ################    dxy      px   #######################
      !                     dxy      px   -       py       s
      ijkl_ind(24, 4) = 430
      ijkl_sym(430) = 353
      !                     dxy      px   -       py      pz
      ijkl_ind(24, 12) = 431
      ijkl_sym(431) = 354
      !                     dxy      px   -      dz2      py
      ijkl_ind(24, 26) = 432
      ijkl_sym(432) = 355
      !                     dxy      px   -      dzy       s
      ijkl_ind(24, 7) = 433
      ijkl_sym(433) = 356
      !                     dxy      px   -      dzy      pz
      ijkl_ind(24, 15) = 434
      ijkl_sym(434) = 357
      !                     dxy      px   -      dzy     dz2
      ijkl_ind(24, 33) = 435
      ijkl_sym(435) = 358
      !                     dxy      px   -   dx2-y2      py
      ijkl_ind(24, 29) = 436
      ijkl_sym(436) = 361*(-1)
      !                     dxy      px   -   dx2-y2     dzy
      ijkl_ind(24, 41) = 437
      ijkl_sym(437) = 362*(-1)
      !                     dxy      px   -      dxy      px
      ijkl_ind(24, 24) = 438
      ijkl_sym(438) = 359
      !                     dxy      px   -      dxy     dzx
      ijkl_ind(24, 39) = 439
      ijkl_sym(439) = 360
      ! ################    dxy      py   #######################
      !                     dxy      py   -       px       s
      ijkl_ind(30, 3) = 440
      ijkl_sym(440) = 353
      !                     dxy      py   -       px      pz
      ijkl_ind(30, 11) = 441
      ijkl_sym(441) = 354
      !                     dxy      py   -      dz2      px
      ijkl_ind(30, 20) = 442
      ijkl_sym(442) = 355
      !                     dxy      py   -      dzx       s
      ijkl_ind(30, 6) = 443
      ijkl_sym(443) = 356
      !                     dxy      py   -      dzx      pz
      ijkl_ind(30, 14) = 444
      ijkl_sym(444) = 357
      !                     dxy      py   -      dzx     dz2
      ijkl_ind(30, 32) = 445
      ijkl_sym(445) = 358
      !                     dxy      py   -   dx2-y2      px
      ijkl_ind(30, 23) = 446
      ijkl_sym(446) = 361
      !                     dxy      py   -   dx2-y2  dx2-y2
      ijkl_ind(30, 38) = 447
      ijkl_sym(447) = 362
      !                     dxy      py   -      dxy      py
      ijkl_ind(30, 30) = 448
      ijkl_sym(448) = 359
      !                     dxy      py   -      dxy     dzy
      ijkl_ind(30, 42) = 449
      ijkl_sym(449) = 360
      ! ################    dxy     dz2   #######################
      !                     dxy     dz2   -       px      py
      ijkl_ind(35, 19) = 450
      ijkl_sym(450) = 373
      !                     dxy     dz2   -      dzx      py
      ijkl_ind(35, 27) = 451
      ijkl_sym(451) = 375
      !                     dxy     dz2   -      dzy      px
      ijkl_ind(35, 22) = 452
      ijkl_sym(452) = 375
      !                     dxy     dz2   -      dzy     dzx
      ijkl_ind(35, 37) = 453
      ijkl_sym(453) = 376
      !                     dxy     dz2   -      dxy       s
      ijkl_ind(35, 9) = 454
      ijkl_sym(454) = 379
      !                     dxy     dz2   -      dxy      pz
      ijkl_ind(35, 17) = 455
      ijkl_sym(455) = 375
      !                     dxy     dz2   -      dxy     dz2
      ijkl_ind(35, 35) = 456
      ijkl_sym(456) = 381
      ! ################    dxy     dzx   #######################
      !                     dxy     dzx   -       py       s
      ijkl_ind(39, 4) = 457
      ijkl_sym(457) = 382
      !                     dxy     dzx   -       py      pz
      ijkl_ind(39, 12) = 458
      ijkl_sym(458) = 383
      !                     dxy     dzx   -      dz2      py
      ijkl_ind(39, 26) = 459
      ijkl_sym(459) = 384
      !                     dxy     dzx   -      dzy       s
      ijkl_ind(39, 7) = 460
      ijkl_sym(460) = 385
      !                     dxy     dzx   -      dzy      pz
      ijkl_ind(39, 15) = 461
      ijkl_sym(461) = 386
      !                     dxy     dzx   -      dzy     dz2
      ijkl_ind(39, 33) = 462
      ijkl_sym(462) = 387
      !                     dxy     dzx   -   dx2-y2      py
      ijkl_ind(39, 29) = 463
      ijkl_sym(463) = 390*(-1)
      !                     dxy     dzx   -   dx2-y2     dzy
      ijkl_ind(39, 41) = 464
      ijkl_sym(464) = 391*(-1)
      !                     dxy     dzx   -      dxy      px
      ijkl_ind(39, 24) = 465
      ijkl_sym(465) = 388
      !                     dxy     dzx   -      dxy     dzx
      ijkl_ind(39, 39) = 466
      ijkl_sym(466) = 389
      ! ################    dxy     dzy   #######################
      !                     dxy     dzy   -       px       s
      ijkl_ind(42, 3) = 467
      ijkl_sym(467) = 382
      !                     dxy     dzy   -       px      pz
      ijkl_ind(42, 11) = 468
      ijkl_sym(468) = 383
      !                     dxy     dzy   -      dz2      px
      ijkl_ind(42, 20) = 469
      ijkl_sym(469) = 384
      !                     dxy     dzy   -      dzx       s
      ijkl_ind(42, 6) = 470
      ijkl_sym(470) = 385
      !                     dxy     dzy   -      dzx      pz
      ijkl_ind(42, 14) = 471
      ijkl_sym(471) = 386
      !                     dxy     dzy   -      dzx     dz2
      ijkl_ind(42, 32) = 472
      ijkl_sym(472) = 387
      !                     dxy     dzy   -   dx2-y2      px
      ijkl_ind(42, 23) = 473
      ijkl_sym(473) = 390
      !                     dxy     dzy   -   dx2-y2  dx2-y2
      ijkl_ind(42, 38) = 474
      ijkl_sym(474) = 391
      !                     dxy     dzy   -      dxy      py
      ijkl_ind(42, 30) = 475
      ijkl_sym(475) = 388
      !                     dxy     dzy   -      dxy     dzy
      ijkl_ind(42, 42) = 476
      ijkl_sym(476) = 389
      ! ################    dxy  dx2-y2   #######################
      !                     dxy  dx2-y2   -      dxy  dx2-y2
      ijkl_ind(44, 44) = 477
      ! ################    dxy     dxy   #######################
      !                     dxy     dxy   -        s       s
      ijkl_ind(45, 1) = 478
      ijkl_sym(478) = 402
      !                     dxy     dxy   -       pz       s
      ijkl_ind(45, 2) = 479
      ijkl_sym(479) = 403
      !                     dxy     dxy   -       pz      pz
      ijkl_ind(45, 10) = 480
      ijkl_sym(480) = 404
      !                     dxy     dxy   -       px      px
      ijkl_ind(45, 18) = 481
      ijkl_sym(481) = 405
      !                     dxy     dxy   -       py      py
      ijkl_ind(45, 25) = 482
      ijkl_sym(482) = 405
      !                     dxy     dxy   -      dz2       s
      ijkl_ind(45, 5) = 483
      ijkl_sym(483) = 407
      !                     dxy     dxy   -      dz2      pz
      ijkl_ind(45, 13) = 484
      ijkl_sym(484) = 408
      !                     dxy     dxy   -      dz2     dz2
      ijkl_ind(45, 31) = 485
      ijkl_sym(485) = 409
      !                     dxy     dxy   -      dzx      px
      ijkl_ind(45, 21) = 486
      ijkl_sym(486) = 410
      !                     dxy     dxy   -      dzx     dzx
      ijkl_ind(45, 36) = 487
      ijkl_sym(487) = 411
      !                     dxy     dxy   -      dzy      py
      ijkl_ind(45, 28) = 488
      ijkl_sym(488) = 410
      !                     dxy     dxy   -      dzy     dzy
      ijkl_ind(45, 40) = 489
      ijkl_sym(489) = 411
      !                     dxy     dxy   -   dx2-y2  dx2-y2
      ijkl_ind(45, 43) = 490
      ijkl_sym(490) = 415
      !                     dxy     dxy   -      dxy     dxy
      ijkl_ind(45, 45) = 491
      ijkl_sym(491) = 414
   END SUBROUTINE setup_ijkl_array

END MODULE semi_empirical_int_arrays
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities to post-process semi-empirical parameters
!> \par History
!>         [tlaino] 03.2008 - Splitting from semi_empirical_parameters and
!>                            keeping there only the setting of the SE params
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date   03.2008 [tlaino]
! **************************************************************************************************
MODULE semi_empirical_par_utils

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fac
   USE mathlib,                         ONLY: binomial
   USE physcon,                         ONLY: bohr,&
                                              evolt
   USE semi_empirical_int_arrays,       ONLY: int_ij,&
                                              int_kl,&
                                              int_onec2el
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   INTEGER, PARAMETER, PRIVATE :: nelem = 106

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_par_utils'

!                STANDARD MOPAC PARAMETERS USED FOR AM1, RM1, MNDO, PM3, PM6
!
!      H                                                                      He
!      Li Be                                                 B  C  N  O  F    Ne
!      Na Mg                                                 Al Si P  S  Cl   Ar
!      K  Ca Sc                Ti V  Cr Mn Fe Co Ni Cu Zn    Ga Ge As Se Br   Kr
!      Rb Sr Y                 Zr Nb Mo Tc Ru Rh Pd Ag Cd    In Sn Sb Te I    Xe
!      Cs Ba La Ce-Lu          Hf Ta W  Re Os Ir Pt Au Hg    Tl Pb Bi Po At   Rn
!      Fr Ra Ac Th Pa U        Np Pu Am Cm Bk Cf Es Fm Md    No Lr Rf Ha 106

!                                      "s" shell
   INTEGER, DIMENSION(0:nelem), PRIVATE :: Nos = (/-1, & !    0
                                                   1, 2, & !    2
                                                   1, 2, 2, 2, 2, 2, 2, 0, & !   10
                                                   1, 2, 2, 2, 2, 2, 2, 0, & !   18
                                                   1, 2, 2, 2, 2, 1, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 0, & !   36
                                                   1, 2, 2, 2, 1, 1, 2, 1, 1, 0, 1, 2, 2, 2, 2, 2, 2, 0, & !   54
                                                   1, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, &
                                                   2, 2, 1, 2, 2, 2, 1, 1, 2, 2, 2, 2, 2, 2, 0, & !   86
                                                   1, 1, 2, 4, 2, 2, 2, 2, 2, 2, 2, 2, 0, 3, -3, 1, 2, 1, -2, -1/)

!                                      "p" shell
   INTEGER, DIMENSION(0:nelem), PRIVATE :: Nop = (/-1, & !    0
                                                   0, 0, & !    2
                                                   0, 0, 1, 2, 3, 4, 5, 6, & !   10
                                                   0, 0, 1, 2, 3, 4, 5, 6, & !   18
                                                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, & !   36
                                                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, & !   54
                                                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
                                                   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, & !   86
                                                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)

!                                      "d" shell
   INTEGER, DIMENSION(0:nelem), PRIVATE :: Nod = (/-1, & !    0
                                                   0, 0, & !    2
                                                   0, 0, 0, 0, 0, 0, 0, 0, & !   10
                                                   0, 0, 0, 0, 0, 0, 0, 0, & !   18
                                                   0, 0, 1, 2, 3, 5, 5, 6, 7, 8, 10, 0, 0, 0, 0, 0, 0, 0, & !   36
                                                   0, 0, 1, 2, 4, 5, 5, 7, 8, 10, 10, 0, 0, 0, 0, 0, 0, 0, & !   54
                                                   0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, &
                                                   2, 3, 5, 5, 6, 7, 9, 10, 0, 0, 0, 0, 0, 0, 0, & !   86
                                                   0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)

!      H          <Quantum Numbers for s, p, d and f orbitals>                He
!      Li Be                                                 B  C  N  O  F    Ne
!      Na Mg                                                 Al Si P  S  Cl   Ar
!      K  Ca Sc                Ti V  Cr Mn Fe Co Ni Cu Zn    Ga Ge As Se Br   Kr
!      Rb Sr Y                 Zr Nb Mo Tc Ru Rh Pd Ag Cd    In Sn Sb Te I    Xe
!      Cs Ba La Ce-Lu          Hf Ta W  Re Os Ir Pt Au Hg    Tl Pb Bi Po At   Rn
!      Fr Ra Ac Th Pa U        Np Pu Am Cm Bk Cf Es Fm Md    No Lr Rf Ha 106

   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE ::   nqs = (/-1, & !    0
                                                                1, 1, & !    2
                                                                2, 2, 2, 2, 2, 2, 2, 2, & !   10
                                                                3, 3, 3, 3, 3, 3, 3, 3, & !   18
                                                                4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, & !   36
                                                                5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, & !   54
                                                                6, 6, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 6, &
                                                                6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, -1, -1, -1, & !   86
                                                    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1/)

   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE ::   nqp = (/-1, & !    0
                                                                -1, -1, & !    2
                                                                2, 2, 2, 2, 2, 2, 2, 2, & !   10
                                                                3, 3, 3, 3, 3, 3, 3, 3, & !   18
                                                                4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, & !   36
                                                                5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, & !   54
                                                                6, 6, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 6, &
                                                                6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, -1, -1, -1, & !   86
                                                    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1/)

   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE ::   nqd = (/-1, & !    0
                                                                -1, -1, & !    2
                                                                -1, -1, -1, -1, -1, -1, -1, -1, & !   10
                                                                -1, -1, 3, 3, 3, 3, 3, -1, & !   18
                                                                -1, -1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, -1, & !   36
                                                                -1, -1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, -1, & !   54
                                                                -1, -1, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 5, &
                                                                5, 5, 5, 5, 5, 5, 5, 5, 5, -1, -1, -1, -1, -1, -1, & !   86
                                                    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1/)

   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE ::   nqf = (/-1, & !    0
                                                                -1, -1, & !    2
                                                                -1, -1, -1, -1, -1, -1, -1, -1, & !   10
                                                                -1, -1, -1, -1, -1, -1, -1, -1, & !   18
                                                    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, & !   36
                                                    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, & !   54
                                                               -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, &
                                                                -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, & !   86
                                                    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1/)

   ! Element Valence
   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE :: zval = (/-1, & !    0
                                                               1, 2, & !    2
                                                               1, 2, 3, 4, 5, 6, 7, 8, & !   10
                                                               1, 2, 3, 4, 5, 6, 7, 8, & !   18
                                                               1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 8, & !   36
                                                               1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 8, & !   54
                                                               1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 3, &
                                                               4, 5, 6, 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, -1, & !   86
                                                      -1, -1, -1, 4, -1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1/)

   ! Number of 1 center 2 electron integrals involving partially filled d shells
   ! r016:  <SS|DD>
   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE :: ir016 = (/0, & !    0
                                                                0, 0, & !    2
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   10
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   18
                                                                0, 0, 2, 4, 6, 5, 10, 12, 14, 16, 10, 0, 0, 0, 0, 0, 0, 0, & !   36
                                                                0, 0, 4, 4, 4, 5, 10, 7, 8, 0, 10, 0, 0, 0, 0, 0, 0, 0, & !   54
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
                                                                4, 6, 8, 10, 12, 14, 9, 10, 0, 0, 0, 0, 0, 0, 0, & !   86
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)

   ! r066:  <DD|DD> "0" term
   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE :: ir066 = (/0, & !    0
                                                                0, 0, & !    2
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   10
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   18
                                                                0, 0, 0, 1, 3, 10, 10, 15, 21, 28, 45, 0, 0, 0, 0, 0, 0, 0, & !   36
                                                                0, 0, 0, 1, 6, 10, 10, 21, 28, 45, 45, 0, 0, 0, 0, 0, 0, 0, & !   54
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
                                                                1, 3, 6, 10, 15, 21, 36, 45, 0, 0, 0, 0, 0, 0, 0, & !   86
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)

   ! r244:  <SD|SD>
   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE :: ir244 = (/0, & !    0
                                                                0, 0, & !    2
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   10
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   18
                                                                0, 0, 1, 2, 3, 5, 5, 6, 7, 8, 5, 0, 0, 0, 0, 0, 0, 0, & !   36
                                                                0, 0, 1, 2, 4, 5, 5, 5, 5, 0, 5, 0, 0, 0, 0, 0, 0, 0, & !   54
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
                                                                2, 3, 4, 5, 6, 7, 5, 5, 0, 0, 0, 0, 0, 0, 0, & !   86
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)

   ! r266:  <DD|DD> "2" term
   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE :: ir266 = (/0, & !    0
                                                                0, 0, & !    2
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   10
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   18
                                                                0, 0, 0, 8, 15, 35, 35, 35, 43, 50, 70, 0, 0, 0, 0, 0, 0, 0, & !   36
                                                                0, 0, 0, 8, 21, 35, 35, 43, 50, 70, 70, 0, 0, 0, 0, 0, 0, 0, & !   54
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
                                                                8, 15, 21, 35, 35, 43, 56, 70, 0, 0, 0, 0, 0, 0, 0, & !   86
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)

   ! r466:  <DD|DD> "4" term
   INTEGER, DIMENSION(0:nelem), PARAMETER, PRIVATE :: ir466 = (/0, & !    0
                                                                0, 0, & !    2
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   10
                                                                0, 0, 0, 0, 0, 0, 0, 0, & !   18
                                                                0, 0, 0, 1, 8, 35, 35, 35, 36, 43, 70, 0, 0, 0, 0, 0, 0, 0, & !   36
                                                                0, 0, 0, 1, 21, 35, 35, 36, 43, 70, 70, 0, 0, 0, 0, 0, 0, 0, & !   54
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
                                                                1, 8, 21, 35, 35, 36, 56, 70, 0, 0, 0, 0, 0, 0, 0, & !   86
                                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/)

   INTERFACE amn_l
      MODULE PROCEDURE amn_l1, amn_l2
   END INTERFACE

   PUBLIC :: convert_param_to_cp2k, calpar, valence_electrons, get_se_basis, &
             setup_1c_2el_int, amn_l

CONTAINS

! **************************************************************************************************
!> \brief  Gives back the number of valence electrons for element z and also the
!>         number of atomic orbitals for that specific element
!> \param sep ...
!> \param extended_basis_set ...
! **************************************************************************************************
   SUBROUTINE valence_electrons(sep, extended_basis_set)
      TYPE(semi_empirical_type), POINTER                 :: sep
      LOGICAL, INTENT(IN)                                :: extended_basis_set

      CHARACTER(len=*), PARAMETER :: routineN = 'valence_electrons', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: natorb, z
      LOGICAL                                            :: check, use_p_orbitals
      REAL(KIND=dp)                                      :: zeff

      use_p_orbitals = .TRUE.
      z = sep%z
      IF(.NOT.(z >= 0))CALL cp__a("semi_empirical_par_utils.F",222)
      ! Special case for Hydrogen.. If requested allow p-orbitals on it..
      SELECT CASE (z)
      CASE (0, 2)
         use_p_orbitals = .FALSE.
      CASE (1)
         use_p_orbitals = sep%p_orbitals_on_h
      CASE DEFAULT
         ! Nothing to do..
      END SELECT
      ! Determine the number of atomic orbitals
      natorb = 0
      IF (nqs(z) > 0) natorb = natorb+1
      IF ((nqp(z) > 0) .OR. use_p_orbitals) natorb = natorb+3
      IF (extended_basis_set .AND. element_has_d(sep)) natorb = natorb+5
      IF (extended_basis_set .AND. element_has_f(sep)) natorb = natorb+7
      ! Check and assignemnt
      check = (natorb <= 4) .OR. (extended_basis_set)
      IF(.NOT.(check))CALL cp__a("semi_empirical_par_utils.F",240)
      sep%natorb = natorb
      sep%extended_basis_set = extended_basis_set
      ! Determine the Z eff
      zeff = REAL(zval(z), KIND=dp)
      sep%zeff = zeff
   END SUBROUTINE valence_electrons

! **************************************************************************************************
!> \brief  Gives back the number of basis function for each l
!> \param sep ...
!> \param l ...
!> \retval n ...
! **************************************************************************************************
   FUNCTION get_se_basis(sep, l) RESULT(n)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER, INTENT(IN)                                :: l
      INTEGER                                            :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'get_se_basis', routineP = moduleN//':'//routineN

      IF (sep%z < 0 .OR. sep%z > nelem) THEN
         CALL cp__b("semi_empirical_par_utils.F",262,"Invalid atomic number !")
      ELSE
         IF (l == 0) THEN
            n = nqs(sep%z)
         ELSEIF (l == 1) THEN
            ! Special case for Hydrogen.. If requested allow p-orbitals on it..
            IF ((sep%z == 1) .AND. sep%p_orbitals_on_h) THEN
               n = 1
            ELSE
               n = nqp(sep%z)
            END IF
         ELSEIF (l == 2) THEN
            n = nqd(sep%z)
         ELSEIF (l == 3) THEN
            n = nqf(sep%z)
         ELSE
            CALL cp__b("semi_empirical_par_utils.F",278,"Invalid l quantum number !")
         END IF
         IF (n < 0) THEN
            CALL cp__b("semi_empirical_par_utils.F",281,"Invalid n quantum number !")
         END IF
      END IF
   END FUNCTION get_se_basis

! **************************************************************************************************
!> \brief  Converts parameter units to internal
!> \param sep ...
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE convert_param_to_cp2k(sep)
      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'convert_param_to_cp2k', &
         routineP = moduleN//':'//routineN

      sep%beta = sep%beta/evolt
      sep%uss = sep%uss/evolt
      sep%upp = sep%upp/evolt
      sep%udd = sep%udd/evolt
      sep%alp = sep%alp/bohr
      sep%eisol = sep%eisol/evolt
      sep%gss = sep%gss/evolt
      sep%gsp = sep%gsp/evolt
      sep%gpp = sep%gpp/evolt
      sep%gp2 = sep%gp2/evolt
      sep%gsd = sep%gsd/evolt
      sep%gpd = sep%gpd/evolt
      sep%gdd = sep%gdd/evolt
      sep%hsp = sep%hsp/evolt
      sep%fn1 = sep%fn1*bohr/evolt
      sep%fn2 = sep%fn2/bohr/bohr
      sep%fn3 = sep%fn3*bohr
      sep%bfn1 = sep%bfn1*bohr/evolt
      sep%bfn2 = sep%bfn2/bohr/bohr
      sep%bfn3 = sep%bfn3*bohr
      sep%f0sd = sep%f0sd
      sep%g2sd = sep%g2sd
      sep%a = sep%a*bohr/evolt
      sep%b = sep%b/bohr/bohr
      sep%c = sep%c*bohr
      sep%pre = sep%pre/evolt
      sep%d = sep%d/bohr

   END SUBROUTINE convert_param_to_cp2k

! **************************************************************************************************
!> \brief  Calculates missing parameters
!> \param z ...
!> \param sep ...
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE calpar(z, sep)
      INTEGER                                            :: z
      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'calpar', routineP = moduleN//':'//routineN

      INTEGER                                            :: iod, iop, ios, j, jmax, k, l
      REAL(KIND=dp) :: ad, am, aq, d1, d2, d3, dd, df, eisol, gdd1, gp2, gp2c, gpp, gppc, gqq, &
         gsp, gspc, gss, gssc, hpp, hpp1, hpp2, hsp, hsp1, hsp2, hspc, p, p4, q1, q2, q3, qf, qn, &
         qq, udd, upp, uss, zp, zs

      IF (.NOT. sep%defined) RETURN
      uss = sep%uss
      upp = sep%upp
      udd = sep%udd
      gss = sep%gss
      gpp = sep%gpp
      gsp = sep%gsp
      gp2 = sep%gp2
      hsp = sep%hsp
      zs = sep%sto_exponents(0)
      zp = sep%sto_exponents(1)
      ios = Nos(z)
      iop = Nop(z)
      iod = Nod(z)

      p = 2.0_dp
      p4 = p**4
      !  GSSC is the number of two-electron terms of type <SS|SS>
      gssc = REAL(MAX(ios-1, 0), KIND=dp)
      k = iop
      !  GSPC is the number of two-electron terms of type <SS|PP>
      gspc = REAL(ios*k, KIND=dp)
      l = MIN(k, 6-k)
      !  GP2C is the number of two-electron terms of type <PP|PP>
      !       plus 0.5 of the number of HPP integrals.
      !  (HPP is not used; instead it is replaced by 0.5(GPP-GP2))
      gp2c = REAL((k*(k-1))/2, KIND=dp)+0.5_dp*REAL((l*(l-1))/2, KIND=dp)
      !  GPPC is minus 0.5 times the number of HPP integrals.
      gppc = -0.5_dp*REAL((l*(l-1))/2, KIND=dp)
      !  HSPC is the number of two-electron terms of type <SP|SP>.
      !       (S and P must have the same spin.  In all cases, if
      !  P is non-zero, there are two S electrons)
      hspc = REAL(-k, KIND=dp)
      !  Constraint the value of the STO exponent
      zp = MAX(0.3_dp, zp)
      !  Take into account constraints on the values of the integrals
      hpp = 0.5_dp*(gpp-gp2)
      hpp = MAX(0.1_dp, hpp)
      hsp = MAX(1.E-7_dp, hsp)

      ! Evaluation of EISOL
      eisol = uss*ios+upp*iop+udd*iod+gss*gssc+gpp*gppc+gsp*gspc+gp2*gp2c+hsp*hspc

      ! Principal quantum number
      qn = REAL(nqs(z), KIND=dp)
      IF(.NOT.(qn > 0))CALL cp__a("semi_empirical_par_utils.F",391)

      ! Charge separation evaluation
      dd = (2.0_dp*qn+1)*(4.0_dp*zs*zp)**(qn+0.5_dp)/(zs+zp)**(2.0_dp*qn+2)/SQRT(3.0_dp)
      qq = SQRT((4.0_dp*qn*qn+6.0_dp*qn+2.0_dp)/20.0_dp)/zp

      ! Calculation of the additive terms in atomic units
      jmax = 5
      gdd1 = (hsp/(evolt*dd**2))**(1.0_dp/3.0_dp)
      d1 = gdd1
      d2 = gdd1+0.04_dp
      DO j = 1, jmax
         df = d2-d1
         hsp1 = 0.5_dp*d1-0.5_dp/SQRT(4.0_dp*dd**2+1.0_dp/d1**2)
         hsp2 = 0.5_dp*d2-0.5_dp/SQRT(4.0_dp*dd**2+1.0_dp/d2**2)
         IF (ABS(hsp2-hsp1) < EPSILON(0.0_dp)) EXIT
         d3 = d1+df*(hsp/evolt-hsp1)/(hsp2-hsp1)
         d1 = d2
         d2 = d3
      END DO
      gqq = (p4*hpp/(evolt*48.0_dp*qq**4))**0.2_dp
      q1 = gqq
      q2 = gqq+0.04_dp
      DO j = 1, jmax
         qf = q2-q1
         hpp1 = 0.25_dp*q1-0.5_dp/SQRT(4.0_dp*qq**2+1.0_dp/q1**2)+0.25_dp/SQRT(8.0_dp*qq**2+1.0_dp/q1**2)
         hpp2 = 0.25_dp*q2-0.5_dp/SQRT(4.0_dp*qq**2+1.0_dp/q2**2)+0.25_dp/SQRT(8.0_dp*qq**2+1.0_dp/q2**2)
         IF (ABS(hpp2-hpp1) < EPSILON(0.0_dp)) EXIT
         q3 = q1+qf*(hpp/evolt-hpp1)/(hpp2-hpp1)
         q1 = q2
         q2 = q3
      END DO
      am = gss/evolt
      ad = d2
      aq = q2
      IF (z == 1) THEN
         ad = am
         aq = am
         dd = 0.0_dp
         qq = 0.0_dp
      END IF
      ! Overwrite these parameters if they were undefined.. otherwise keep the defined
      ! value
      IF (ABS(sep%eisol) < EPSILON(0.0_dp)) sep%eisol = eisol
      IF (ABS(sep%dd) < EPSILON(0.0_dp)) sep%dd = dd
      IF (ABS(sep%qq) < EPSILON(0.0_dp)) sep%qq = qq
      IF (ABS(sep%am) < EPSILON(0.0_dp)) sep%am = am
      IF (ABS(sep%ad) < EPSILON(0.0_dp)) sep%ad = ad
      IF (ABS(sep%aq) < EPSILON(0.0_dp)) sep%aq = aq
      ! Proceed with d-orbitals and fill the Kolpman-Ohno and Charge Separation
      ! arrays
      CALL calpar_d(sep)
   END SUBROUTINE calpar

! **************************************************************************************************
!> \brief  Finalize the initialization of parameters, defining additional
!>         parameters for d-orbitals
!>
!> \param sep ...
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE calpar_d(sep)
      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'calpar_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(6)                        :: amn

! Determine if this element owns d-orbitals (only if the parametrization
! supports the d-orbitals)

      IF (sep%extended_basis_set) sep%dorb = element_has_d(sep)
      IF (sep%dorb) THEN
         CALL amn_l(sep, amn)
         CALL eval_1c_2el_spd(sep)
         CALL eval_cs_ko(sep, amn)
      END IF
      IF (.NOT. sep%dorb) THEN
         ! Use the old integral module
         IF (ABS(sep%am) > EPSILON(0.0_dp)) THEN
            sep%ko(1) = 0.5_dp/sep%am
         END IF
         IF (ABS(sep%ad) > EPSILON(0.0_dp) .AND. (sep%z /= 1)) THEN
            sep%ko(2) = 0.5_dp/sep%ad
         END IF
         IF (ABS(sep%aq) > EPSILON(0.0_dp) .AND. (sep%z /= 1)) THEN
            sep%ko(3) = 0.5_dp/sep%aq
         END IF
         sep%ko(7) = sep%ko(1)
         sep%ko(9) = sep%ko(1)
         sep%cs(2) = sep%dd
         sep%cs(3) = sep%qq*SQRT(2.0_dp)
      ELSE
         ! Use the new integral module
         sep%ko(9) = sep%ko(1)
         sep%aq = 0.5_dp/sep%ko(3)
      END IF
      ! In case the Klopman-Ohno CORE therm is provided let's overwrite the
      ! computed one
      IF (ABS(sep%rho) > EPSILON(0.0_dp)) THEN
         sep%ko(9) = sep%rho
      END IF
   END SUBROUTINE calpar_d

! **************************************************************************************************
!> \brief  Determines if the elements has d-orbitals
!>
!> \param sep ...
!> \retval res ...
!> \date   05.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION element_has_d(sep) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sep
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'element_has_d', routineP = moduleN//':'//routineN

      res = (nqd(sep%z) > 0 .AND. sep%sto_exponents(2) > EPSILON(0.0_dp))
   END FUNCTION element_has_d

! **************************************************************************************************
!> \brief  Determines if the elements has f-orbitals
!>
!> \param sep ...
!> \retval res ...
!> \date   05.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION element_has_f(sep) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sep
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'element_has_f', routineP = moduleN//':'//routineN

      res = (nqf(sep%z) > 0 .AND. sep%sto_exponents(3) > EPSILON(0.0_dp))
   END FUNCTION element_has_f

! **************************************************************************************************
!> \brief  Computes the A^{\mu \nu}_l values for the evaluation of the two-center
!>          two-electron integrals. The term is the one reported in Eq.(7) of TCA
!>
!> \param sep ...
!> \param amn ...
!> \date   03.2008 [tlaino]
!> \par    Notation Index: 1 (SS), 2 (SP), 3 (SD), 4 (PP), 5 (PD), 6 (DD)
!>
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE amn_l1(sep, amn)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), DIMENSION(6), INTENT(OUT)           :: amn

      CHARACTER(len=*), PARAMETER :: routineN = 'amn_l1', routineP = moduleN//':'//routineN

      INTEGER                                            :: nd, nsp
      REAL(KIND=dp)                                      :: z1, z2, z3

      z1 = sep%sto_exponents(0)
      z2 = sep%sto_exponents(1)
      z3 = sep%sto_exponents(2)
      IF (z1 <= 0.0_dp) &
         CALL cp_abort(cp__l("semi_empirical_par_utils.F",554), &
                       "Trying to use s-orbitals, but the STO exponents is set to 0. "// &
                       "Please check if your parameterization supports the usage of s orbitals! ")
      amn = 0.0_dp
      nsp = nqs(sep%z)
      IF (sep%natorb >= 4) THEN
         IF (z2 <= 0.0_dp) &
            CALL cp_abort(cp__l("semi_empirical_par_utils.F",561), &
                          "Trying to use p-orbitals, but the STO exponents is set to 0. "// &
                          "Please check if your parameterization supports the usage of p orbitals! ")
         amn(2) = amn_l_low(z1, z2, nsp, nsp, 1)
         amn(3) = amn_l_low(z2, z2, nsp, nsp, 2)
         IF (sep%dorb) THEN
            IF (z3 <= 0.0_dp) &
               CALL cp_abort(cp__l("semi_empirical_par_utils.F",568), &
                             "Trying to use d-orbitals, but the STO exponents is set to 0. "// &
                             "Please check if your parameterization supports the usage of d orbitals! ")
            nd = nqd(sep%z)
            amn(4) = amn_l_low(z1, z3, nsp, nd, 2)
            amn(5) = amn_l_low(z2, z3, nsp, nd, 1)
            amn(6) = amn_l_low(z3, z3, nd, nd, 2)
         END IF
      END IF
   END SUBROUTINE amn_l1

! **************************************************************************************************
!> \brief  Computes the A^{\mu \nu}_l values for the evaluation of the two-center
!>          two-electron integrals. The term is the one reported in Eq.(7) of TCA
!>
!> \param sep ...
!> \param amn ...
!> \date   09.2008 [tlaino]
!> \par
!>
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE amn_l2(sep, amn)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), DIMENSION(6, 0:2), INTENT(OUT)      :: amn

      CHARACTER(len=*), PARAMETER :: routineN = 'amn_l2', routineP = moduleN//':'//routineN

      INTEGER                                            :: nd, nsp
      REAL(KIND=dp)                                      :: z1, z2, z3

      z1 = sep%sto_exponents(0)
      z2 = sep%sto_exponents(1)
      z3 = sep%sto_exponents(2)
      IF(.NOT.(z1 > 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",602)
      amn = 0.0_dp
      nsp = nqs(sep%z)
      amn(1, 0) = amn_l_low(z1, z1, nsp, nsp, 0)
      IF (sep%natorb >= 4) THEN
         IF(.NOT.(z2 > 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",607)
         amn(2, 1) = amn_l_low(z1, z2, nsp, nsp, 1)
         amn(3, 0) = amn_l_low(z2, z2, nsp, nsp, 0)
         amn(3, 2) = amn_l_low(z2, z2, nsp, nsp, 2)
         IF (sep%dorb) THEN
            IF(.NOT.(z3 > 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",612)
            nd = nqd(sep%z)
            amn(4, 2) = amn_l_low(z1, z3, nsp, nd, 2)
            amn(5, 1) = amn_l_low(z2, z3, nsp, nd, 1)
            amn(6, 0) = amn_l_low(z3, z3, nd, nd, 0)
            amn(6, 2) = amn_l_low(z3, z3, nd, nd, 2)
         END IF
      END IF
   END SUBROUTINE amn_l2

! **************************************************************************************************
!> \brief  Low level for computing Eq.(7) of TCA
!> \param z1 ...
!> \param z2 ...
!> \param n1 ...
!> \param n2 ...
!> \param l ...
!> \retval amnl ...
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION amn_l_low(z1, z2, n1, n2, l) RESULT(amnl)
      REAL(KIND=dp), INTENT(IN)                          :: z1, z2
      INTEGER, INTENT(IN)                                :: n1, n2, l
      REAL(KIND=dp)                                      :: amnl

      amnl = fac(n1+n2+l)/SQRT(fac(2*n1)*fac(2*n2))*(2.0_dp*z1/(z1+z2))**n1* &
             (2.0_dp*z2/(z1+z2))**n2*2.0_dp*SQRT(z1*z2)/(z1+z2)**(l+1)

   END FUNCTION amn_l_low

! **************************************************************************************************
!> \brief  Calculation of chare separations and additive terms used for computing
!>         the two-center two-electron integrals with d-orbitals
!> \param sep ...
!> \param amn ...
!> \date   03.2008 [tlaino]
!> \par    Notation
!>         -) Charge separations [sep%cs(1:6)]  [see equations (12)-(16) of TCA]
!>         -) Additive terms of Klopman-Ohno terms [sep%ko(1:9)] [see equations
!>            (19)-(26) of TCA]
!>         -) Atomic core additive term stored in sep%ko(9): used in the calculation
!>            of the core-electron attractions and core-core repulsions
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE eval_cs_ko(sep, amn)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), DIMENSION(6), INTENT(IN)            :: amn

      CHARACTER(len=*), PARAMETER :: routineN = 'eval_cs_ko', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: d, fg

! SS term

      fg = sep%gss
      sep%ko(1) = ko_ij(0, 1.0_dp, fg)
      IF (sep%natorb >= 4) THEN
         ! Other terms for SP basis
         ! SP
         d = amn(2)/SQRT(3.0_dp)
         fg = sep%hsp
         sep%cs(2) = d
         sep%ko(2) = ko_ij(1, d, fg)
         ! PP
         sep%ko(7) = sep%ko(1)
         d = SQRT(amn(3)*2.0_dp/5.0_dp)
         fg = 0.5_dp*(sep%gpp-sep%gp2)
         sep%cs(3) = d
         sep%ko(3) = ko_ij(2, d, fg)
         ! Terms involving d-orbitals
         IF (sep%dorb) THEN
            ! SD
            d = SQRT(amn(4)*2.0_dp/SQRT(15.0_dp))
            fg = sep%onec2el(19)
            sep%cs(4) = d
            sep%ko(4) = ko_ij(2, d, fg)
            ! PD
            d = amn(5)/SQRT(5.0_dp)
            fg = sep%onec2el(23)-1.8_dp*sep%onec2el(35)
            sep%cs(5) = d
            sep%ko(5) = ko_ij(1, d, fg)
            ! DD
            fg = 0.2_dp*(sep%onec2el(29)+2.0_dp*sep%onec2el(30)+2.0_dp*sep%onec2el(31))
            sep%ko(8) = ko_ij(0, 1.0_dp, fg)
            d = SQRT(amn(6)*2.0_dp/7.0_dp)
            fg = sep%onec2el(44)-(20.0_dp/35.0_dp)*sep%onec2el(52)
            sep%cs(6) = d
            sep%ko(6) = ko_ij(2, d, fg)
         END IF
      END IF
   END SUBROUTINE eval_cs_ko

! **************************************************************************************************
!> \brief  Computes the 1 center two-electrons integrals for a SPD basis
!> \param sep ...
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE eval_1c_2el_spd(sep)
      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'eval_1c_2el_spd', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: r016, r036, r066, r125, r155, r234, &
                                                            r236, r244, r246, r266, r355, r466, &
                                                            s15, s3, s5

      IF (sep%dorb) THEN
         s3 = SQRT(3.0_dp)
         s5 = SQRT(5.0_dp)
         s15 = SQRT(15.0_dp)

         ! We evaluate now the Slater-Condon parameters (Rlij)
         CALL sc_param(sep, r066, r266, r466, r016, r244, r036, r236, r155, r355, r125, &
                       r234, r246)

         IF (ABS(sep%f0sd) > EPSILON(0.0_dp)) THEN
            r016 = sep%f0sd
         END IF
         IF (ABS(sep%g2sd) > EPSILON(0.0_dp)) THEN
            r244 = sep%g2sd
         END IF
         CALL eisol_corr(sep, r016, r066, r244, r266, r466)
         sep%onec2el(1) = r016
         sep%onec2el(2) = 2.0_dp/(3.0_dp*s5)*r125
         sep%onec2el(3) = 1.0_dp/s15*r125
         sep%onec2el(4) = 2.0_dp/(5.0_dp*s5)*r234
         sep%onec2el(5) = r036+4.0_dp/35.0_dp*r236
         sep%onec2el(6) = r036+2.0_dp/35.0_dp*r236
         sep%onec2el(7) = r036-4.0_dp/35.0_dp*r236
         sep%onec2el(8) = -1.0_dp/(3.0_dp*s5)*r125
         sep%onec2el(9) = SQRT(3.0_dp/125.0_dp)*r234
         sep%onec2el(10) = s3/35.0_dp*r236
         sep%onec2el(11) = 3.0_dp/35.0_dp*r236
         sep%onec2el(12) = -1.0_dp/(5.0_dp*s5)*r234
         sep%onec2el(13) = r036-2.0_dp/35.0_dp*r236
         sep%onec2el(14) = -2.0_dp*s3/35.0_dp*r236
         sep%onec2el(15) = -sep%onec2el(3)
         sep%onec2el(16) = -sep%onec2el(11)
         sep%onec2el(17) = -sep%onec2el(9)
         sep%onec2el(18) = -sep%onec2el(14)
         sep%onec2el(19) = 1.0_dp/5.0_dp*r244
         sep%onec2el(20) = 2.0_dp/(7.0_dp*s5)*r246
         sep%onec2el(21) = sep%onec2el(20)/2.0_dp
         sep%onec2el(22) = -sep%onec2el(20)
         sep%onec2el(23) = 4.0_dp/15.0_dp*r155+27.0_dp/245.0_dp*r355
         sep%onec2el(24) = 2.0_dp*s3/15.0_dp*r155-9.0_dp*s3/245.0_dp*r355
         sep%onec2el(25) = 1.0_dp/15.0_dp*r155+18.0_dp/245.0_dp*r355
         sep%onec2el(26) = -s3/15.0_dp*r155+12.0_dp*s3/245.0_dp*r355
         sep%onec2el(27) = -s3/15.0_dp*r155-3.0_dp*s3/245.0_dp*r355
         sep%onec2el(28) = -sep%onec2el(27)
         sep%onec2el(29) = r066+4.0_dp/49.0_dp*r266+4.0_dp/49.0_dp*r466
         sep%onec2el(30) = r066+2.0_dp/49.0_dp*r266-24.0_dp/441.0_dp*r466
         sep%onec2el(31) = r066-4.0_dp/49.0_dp*r266+6.0_dp/441.0_dp*r466
         sep%onec2el(32) = SQRT(3.0_dp/245.0_dp)*r246
         sep%onec2el(33) = 1.0_dp/5.0_dp*r155+24.0_dp/245.0_dp*r355
         sep%onec2el(34) = 1.0_dp/5.0_dp*r155-6.0_dp/245.0_dp*r355
         sep%onec2el(35) = 3.0_dp/49.0_dp*r355
         sep%onec2el(36) = 1.0_dp/49.0_dp*r266+30.0_dp/441.0_dp*r466
         sep%onec2el(37) = s3/49.0_dp*r266-5.0_dp*s3/441.0_dp*r466
         sep%onec2el(38) = r066-2.0_dp/49.0_dp*r266-4.0_dp/441.0_dp*r466
         sep%onec2el(39) = -2.0_dp*s3/49.0_dp*r266+10.0_dp*s3/441.0_dp*r466
         sep%onec2el(40) = -sep%onec2el(32)
         sep%onec2el(41) = -sep%onec2el(34)
         sep%onec2el(42) = -sep%onec2el(35)
         sep%onec2el(43) = -sep%onec2el(37)
         sep%onec2el(44) = 3.0_dp/49.0_dp*r266+20.0_dp/441.0_dp*r466
         sep%onec2el(45) = -sep%onec2el(39)
         sep%onec2el(46) = 1.0_dp/5.0_dp*r155-3.0_dp/35.0_dp*r355
         sep%onec2el(47) = -sep%onec2el(46)
         sep%onec2el(48) = 4.0_dp/49.0_dp*r266+15.0_dp/441.0_dp*r466
         sep%onec2el(49) = 3.0_dp/49.0_dp*r266-5.0_dp/147.0_dp*r466
         sep%onec2el(50) = -sep%onec2el(49)
         sep%onec2el(51) = r066+4.0_dp/49.0_dp*r266-34.0_dp/441.0_dp*r466
         sep%onec2el(52) = 35.0_dp/441.0_dp*r466
         sep%f0dd = r066
         sep%f2dd = r266
         sep%f4dd = r466
         sep%f0sd = r016
         sep%g2sd = r244
         sep%f0pd = r036
         sep%f2pd = r236
         sep%g1pd = r155
         sep%g3pd = r355
      END IF
   END SUBROUTINE eval_1c_2el_spd

! **************************************************************************************************
!> \brief  Slater-Condon parameters for 1 center 2 electrons integrals
!> \param sep ...
!> \param r066 ...
!> \param r266 ...
!> \param r466 ...
!> \param r016 ...
!> \param r244 ...
!> \param r036 ...
!> \param r236 ...
!> \param r155 ...
!> \param r355 ...
!> \param r125 ...
!> \param r234 ...
!> \param r246 ...
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE sc_param(sep, r066, r266, r466, r016, r244, r036, r236, r155, r355, &
                       r125, r234, r246)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), INTENT(out)                         :: r066, r266, r466, r016, r244, r036, &
                                                            r236, r155, r355, r125, r234, r246

      CHARACTER(len=*), PARAMETER :: routineN = 'sc_param', routineP = moduleN//':'//routineN

      INTEGER                                            :: nd, ns
      REAL(KIND=dp)                                      :: ed, ep, es

      ns = nqs(sep%z)
      nd = nqd(sep%z)
      es = sep%zn(0)
      ep = sep%zn(1)
      ed = sep%zn(2)
      r016 = sc_param_low(0, ns, es, ns, es, nd, ed, nd, ed)
      r036 = sc_param_low(0, ns, ep, ns, ep, nd, ed, nd, ed)
      r066 = sc_param_low(0, nd, ed, nd, ed, nd, ed, nd, ed)
      r155 = sc_param_low(1, ns, ep, nd, ed, ns, ep, nd, ed)
      r125 = sc_param_low(1, ns, es, ns, ep, ns, ep, nd, ed)
      r244 = sc_param_low(2, ns, es, nd, ed, ns, es, nd, ed)
      r236 = sc_param_low(2, ns, ep, ns, ep, nd, ed, nd, ed)
      r266 = sc_param_low(2, nd, ed, nd, ed, nd, ed, nd, ed)
      r234 = sc_param_low(2, ns, ep, ns, ep, ns, es, nd, ed)
      r246 = sc_param_low(2, ns, es, nd, ed, nd, ed, nd, ed)
      r355 = sc_param_low(3, ns, ep, nd, ed, ns, ep, nd, ed)
      r466 = sc_param_low(4, nd, ed, nd, ed, nd, ed, nd, ed)
   END SUBROUTINE sc_param

! **************************************************************************************************
!> \brief  Slater-Condon parameters for 1 center 2 electrons integrals - Low level
!> \param k ...
!> \param na ...
!> \param ea ...
!> \param nb ...
!> \param eb ...
!> \param nc ...
!> \param ec ...
!> \param nd ...
!> \param ed ...
!> \retval res ...
!> \date   03.2008 [tlaino]
!> \par    Notation
!>         -) k:      Type of integral
!>         -) na,na:  Principle Quantum Number of AO,corresponding to electron 1
!>         -) ea,eb:  Exponents of AO,corresponding to electron 1
!>         -) nb,nc:  Principle Quantum Number of AO,corresponding to electron 2
!>         -) ec,ed:  Exponents of AO,corresponding to electron 2
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION sc_param_low(k, na, ea, nb, eb, nc, ec, nd, ed) RESULT(res)
      INTEGER, INTENT(in)                                :: k, na
      REAL(KIND=dp), INTENT(in)                          :: ea
      INTEGER, INTENT(in)                                :: nb
      REAL(KIND=dp), INTENT(in)                          :: eb
      INTEGER, INTENT(in)                                :: nc
      REAL(KIND=dp), INTENT(in)                          :: ec
      INTEGER, INTENT(in)                                :: nd
      REAL(KIND=dp), INTENT(in)                          :: ed
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'sc_param_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, m, m1, m2, n, nab, ncd
      REAL(KIND=dp)                                      :: a2, aab, acd, ae, aea, aeb, aec, aed, c, &
                                                            e, eab, ecd, ff, s0, s1, s2, s3, tmp

      IF(.NOT.(ea > 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",887)
      IF(.NOT.(eb > 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",888)
      IF(.NOT.(ec > 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",889)
      IF(.NOT.(ed > 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",890)
      aea = LOG(ea)
      aeb = LOG(eb)
      aec = LOG(ec)
      aed = LOG(ed)
      nab = na+nb
      ncd = nc+nd
      ecd = ec+ed
      eab = ea+eb
      e = ecd+eab
      n = nab+ncd
      ae = LOG(e)
      a2 = LOG(2.0_dp)
      acd = LOG(ecd)
      aab = LOG(eab)
      ff = fac(n-1)/SQRT(fac(2*na)*fac(2*nb)*fac(2*nc)*fac(2*nd))
      tmp = na*aea+nb*aeb+nc*aec+nd*aed+0.5_dp*(aea+aeb+aec+aed)+a2*(n+2)-ae*n
      c = evolt*ff*EXP(tmp)
      s0 = 1.0_dp/e
      s1 = 0.0_dp
      s2 = 0.0_dp
      m = ncd-k
      DO i = 1, m
         s0 = s0*e/ecd
         s1 = s1+s0*(binomial(ncd-k-1, i-1)-binomial(ncd+k, i-1))/binomial(n-1, i-1)
      END DO
      m1 = m
      m2 = ncd+k
      DO i = m1, m2
         s0 = s0*e/ecd
         s2 = s2+s0*binomial(m2, i)/binomial(n-1, i)
      END DO
      s3 = EXP(ae*n-acd*(m2+1)-aab*(nab-k))/binomial(n-1, m2)
      res = c*(s1-s2+s3)
   END FUNCTION sc_param_low

! **************************************************************************************************
!> \brief  Corrects the EISOL fo the one-center terms coming from those atoms
!>         that have partially filled "d" shells
!> \param sep ...
!> \param r016 ...
!> \param r066 ...
!> \param r244 ...
!> \param r266 ...
!> \param r466 ...
!> \date   03.2008 [tlaino]
!> \par    Notation
!>         r016:  <SS|DD>
!>         r066:  <DD|DD> "0" term
!>         r244:  <SD|SD>
!>         r266:  <DD|DD> "2" term
!>         r466:  <DD|DD> "4" term
!>
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE eisol_corr(sep, r016, r066, r244, r266, r466)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), INTENT(in)                          :: r016, r066, r244, r266, r466

      CHARACTER(len=*), PARAMETER :: routineN = 'eisol_corr', routineP = moduleN//':'//routineN

      sep%eisol = sep%eisol+ir016(sep%z)*r016+ &
                  ir066(sep%z)*r066- &
                  ir244(sep%z)*r244/5.0_dp- &
                  ir266(sep%z)*r266/49.0_dp- &
                  ir466(sep%z)*r466/49.0_dp
   END SUBROUTINE eisol_corr

! **************************************************************************************************
!> \brief  Computes the Klopman-Ohno additive terms for 2-center 2-electron
!>         integrals requiring that the corresponding 1-center 2-electron integral
!>         is reproduced from the 2-center one for r->0
!> \param l ...
!> \param d ...
!> \param fg ...
!> \retval res ...
!> \date   03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION ko_ij(l, d, fg) RESULT(res)
      INTEGER, INTENT(in)                                :: l
      REAL(KIND=dp), INTENT(in)                          :: d, fg
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'ko_ij', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: niter = 100
      REAL(KIND=dp), PARAMETER                           :: epsil = 1.0E-08_dp, g1 = 0.382_dp, &
                                                            g2 = 0.618_dp

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: a1, a2, delta, dsq, ev4, ev8, f1, f2, &
                                                            y1, y2

      IF(.NOT.(fg /= 0.0_dp))CALL cp__a("semi_empirical_par_utils.F",983)
      ! Term for SS
      IF (l == 0) THEN
         res = 0.5_dp*evolt/fg
         RETURN
      END IF
      ! Term for Higher angular momentum
      dsq = d*d
      ev4 = evolt*0.25_dp
      ev8 = evolt/8.0_dp
      a1 = 0.1_dp
      a2 = 5.0_dp
      DO i = 1, niter
         delta = a2-a1
         IF (delta < epsil) EXIT
         y1 = a1+delta*g1
         y2 = a1+delta*g2
         IF (l == 1) THEN
            f1 = (ev4*(1/y1-1/SQRT(y1**2+dsq))-fg)**2
            f2 = (ev4*(1/y2-1/SQRT(y2**2+dsq))-fg)**2
         ELSE IF (l == 2) THEN
            f1 = (ev8*(1.0_dp/y1-2.0_dp/SQRT(y1**2+dsq*0.5_dp)+1.0_dp/SQRT(y1**2+dsq))-fg)**2
            f2 = (ev8*(1/y2-2.0_dp/SQRT(y2**2+dsq*0.5_dp)+1.0_dp/SQRT(y2**2+dsq))-fg)**2
         END IF
         IF (f1 < f2) THEN
            a2 = y2
         ELSE
            a1 = y1
         END IF
      END DO
      ! Convergence reached.. define additive terms
      IF (f1 >= f2) THEN
         res = a2
      ELSE
         res = a1
      END IF
   END FUNCTION ko_ij

! **************************************************************************************************
!> \brief  Fills the 1 center 2 electron integrals for the construction of the
!>         one-electron fock matrix
!> \param sep ...
!> \date   04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_1c_2el_int(sep)
      TYPE(semi_empirical_type), POINTER                 :: sep

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_1c_2el_int', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, ij0, ind, ip, ipx, ipy, ipz, &
                                                            isize, kl, natorb
      LOGICAL                                            :: defined
      REAL(KIND=dp)                                      :: gp2, gpp, gsp, gss, hsp

      CALL get_se_param(sep, defined=defined, natorb=natorb, &
                        gss=gss, gsp=gsp, gpp=gpp, gp2=gp2, hsp=hsp)
      IF(.NOT.(defined))CALL cp__a("semi_empirical_par_utils.F",1041)

      isize = natorb*(natorb+1)/2
      ALLOCATE (sep%w(isize, isize))
      ! Initialize array
      sep%w = 0.0_dp
      ! Fill the array
      IF (natorb > 0) THEN
         ip = 1
         sep%w(ip, ip) = gss
         IF (natorb > 2) THEN
            ipx = ip+2
            ipy = ip+5
            ipz = ip+9
            sep%w(ipx, ip) = gsp
            sep%w(ipy, ip) = gsp
            sep%w(ipz, ip) = gsp
            sep%w(ip, ipx) = gsp
            sep%w(ip, ipy) = gsp
            sep%w(ip, ipz) = gsp
            sep%w(ipx, ipx) = gpp
            sep%w(ipy, ipy) = gpp
            sep%w(ipz, ipz) = gpp
            sep%w(ipy, ipx) = gp2
            sep%w(ipz, ipx) = gp2
            sep%w(ipz, ipy) = gp2
            sep%w(ipx, ipy) = gp2
            sep%w(ipx, ipz) = gp2
            sep%w(ipy, ipz) = gp2
            sep%w(ip+1, ip+1) = hsp
            sep%w(ip+3, ip+3) = hsp
            sep%w(ip+6, ip+6) = hsp
            sep%w(ip+4, ip+4) = 0.5_dp*(gpp-gp2)
            sep%w(ip+7, ip+7) = 0.5_dp*(gpp-gp2)
            sep%w(ip+8, ip+8) = 0.5_dp*(gpp-gp2)
            IF (sep%dorb) THEN
               ij0 = ip-1
               DO i = 1, 243
                  ij = int_ij(i)
                  kl = int_kl(i)
                  ind = int_onec2el(i)
                  sep%w(ij+ij0, kl+ij0) = sep%onec2el(ind)/evolt
               END DO
            END IF
         END IF
      END IF
   END SUBROUTINE setup_1c_2el_int

END MODULE semi_empirical_par_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Setup and Methods for semi-empirical multipole types
!> \author Teodoro Laino [tlaino] - 08.2008 Zurich University
! **************************************************************************************************
MODULE semi_empirical_mpole_methods

   USE input_constants,                 ONLY: do_method_pnnl
   USE kinds,                           ONLY: dp
   USE semi_empirical_int_arrays,       ONLY: alm,&
                                              indexa,&
                                              indexb,&
                                              se_map_alm
   USE semi_empirical_mpole_types,      ONLY: nddo_mpole_create,&
                                              nddo_mpole_release,&
                                              nddo_mpole_type,&
                                              semi_empirical_mpole_p_create,&
                                              semi_empirical_mpole_p_type,&
                                              semi_empirical_mpole_type
   USE semi_empirical_par_utils,        ONLY: amn_l
   USE semi_empirical_types,            ONLY: semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_mpole_methods'

   PUBLIC :: semi_empirical_mpole_p_setup, &
             nddo_mpole_setup, &
             quadrupole_sph_to_cart

CONTAINS

! **************************************************************************************************
!> \brief Setup semi-empirical mpole type
!>        This function setup for each semi-empirical type a structure containing
!>        the multipolar expansion for all possible combination on-site of atomic
!>        orbitals ( \mu \nu |
!> \param mpoles ...
!> \param se_parameter ...
!> \param method ...
!> \date   09.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE semi_empirical_mpole_p_setup(mpoles, se_parameter, method)
      TYPE(semi_empirical_mpole_p_type), DIMENSION(:), &
         POINTER                                         :: mpoles
      TYPE(semi_empirical_type), POINTER                 :: se_parameter
      INTEGER, INTENT(IN)                                :: method

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_mpole_p_setup', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=3), DIMENSION(9), PARAMETER :: &
         label_print_orb = (/"  s", " px", " py", " pz", "dx2", "dzx", "dz2", "dzy", "dxy"/)
      INTEGER, DIMENSION(9), PARAMETER :: loc_index = (/1, 2, 2, 2, 3, 3, 3, 3, 3/)

      INTEGER                                            :: a, b, i, ind1, ind2, j, k, k1, k2, mu, &
                                                            natorb, ndim, nr
      REAL(KIND=dp)                                      :: dlm, tmp, wp, ws, zb, ZP, ZS, zt
      REAL(KIND=dp), DIMENSION(3, 3, 45)                 :: M2
      REAL(KIND=dp), DIMENSION(3, 45)                    :: M1
      REAL(KIND=dp), DIMENSION(45)                       :: M0
      REAL(KIND=dp), DIMENSION(6, 0:2)                   :: amn
      TYPE(semi_empirical_mpole_type), POINTER           :: mpole

      IF(.NOT.(.NOT. ASSOCIATED(mpoles)))CALL cp__a("semi_empirical_mpole_methods.F",74)
      ! If there are atomic orbitals proceed with the expansion in multipoles
      natorb = se_parameter%natorb
      IF (natorb /= 0) THEN
         ndim = natorb*(natorb+1)/2
         CALL semi_empirical_mpole_p_create(mpoles, ndim)

         ! Select method for multipolar expansion

         ! Fill in information on multipole expansion due to atomic orbitals charge
         ! distribution
         NULLIFY (mpole)
         CALL amn_l(se_parameter, amn)
         DO i = 1, natorb
            DO j = 1, i
               ind1 = indexa(se_map_alm(i), se_map_alm(j))
               ind2 = indexb(i, j)
               ! the order in the mpoles structure is like the standard one for the
               ! integrals: s px py pz dx2-y2 dzx dz2 dzy dxy (lower triangular)
               ! which differs from the order of the Hamiltonian in CP2K. But I
               ! preferred to keep this order for consistency with the integrals
               mpole => mpoles(ind2)%mpole
               mpole%indi = i
               mpole%indj = j
               a = loc_index(i)
               b = loc_index(j)
               mpole%c = HUGE(0.0_dp)
               mpole%d = HUGE(0.0_dp)
               mpole%qs = HUGE(0.0_dp)
               mpole%qc = HUGE(0.0_dp)

               ! Charge
               IF (alm(ind1, 0, 0) /= 0.0_dp) THEN
                  dlm = 1.0_dp/SQRT(REAL((2*0+1), KIND=dp))
                  tmp = -dlm*amn(indexb(a, b), 0)
                  mpole%c = tmp*alm(ind1, 0, 0)
                  mpole%task(1) = .TRUE.
               END IF

               ! Dipole
               IF (ANY(alm(ind1, 1, -1:1) /= 0.0_dp)) THEN
                  dlm = 1.0_dp/SQRT(REAL((2*1+1), KIND=dp))
                  tmp = -dlm*amn(indexb(a, b), 1)
                  mpole%d(1) = tmp*alm(ind1, 1, 1)
                  mpole%d(2) = tmp*alm(ind1, 1, -1)
                  mpole%d(3) = tmp*alm(ind1, 1, 0)
                  mpole%task(2) = .TRUE.
               END IF

               ! Quadrupole
               IF (ANY(alm(ind1, 2, -2:2) /= 0.0_dp)) THEN
                  dlm = 1.0_dp/SQRT(REAL((2*2+1), KIND=dp))
                  tmp = -dlm*amn(indexb(a, b), 2)

                  ! Spherical components
                  mpole%qs(1) = tmp*alm(ind1, 2, 0) ! d3z2-r2
                  mpole%qs(2) = tmp*alm(ind1, 2, 1) ! dzx
                  mpole%qs(3) = tmp*alm(ind1, 2, -1) ! dzy
                  mpole%qs(4) = tmp*alm(ind1, 2, 2) ! dx2-y2
                  mpole%qs(5) = tmp*alm(ind1, 2, -2) ! dxy

                  ! Convert into cartesian components
                  CALL quadrupole_sph_to_cart(mpole%qc, mpole%qs)
                  mpole%task(3) = .TRUE.
               END IF

               IF (debug_this_module) THEN
                  WRITE (*, '(A,2I6,A)') "Orbitals ", i, j, &
                     " ("//label_print_orb(i)//","//label_print_orb(j)//")"
                  IF (mpole%task(1)) WRITE (*, '(9F12.6)') mpole%c
                  IF (mpole%task(2)) WRITE (*, '(9F12.6)') mpole%d
                  IF (mpole%task(3)) WRITE (*, '(9F12.6)') mpole%qc
                  WRITE (*, *)
               END IF
            END DO
         END DO

         IF (method == do_method_pnnl) THEN
            ! No d-function for Schenter type integrals
            IF(.NOT.(natorb <= 4))CALL cp__a("semi_empirical_mpole_methods.F",153)

            M0 = 0.0_dp
            M1 = 0.0_dp
            M2 = 0.0_dp

            DO mu = 1, se_parameter%natorb
               M0(indexb(mu, mu)) = 1.0_dp
            ENDDO

            ZS = se_parameter%sto_exponents(0)
            ZP = se_parameter%sto_exponents(1)
            nr = se_parameter%nr

            ws = REAL((2*nr+2)*(2*nr+1), dp)/(24.0_dp*ZS**2)
            DO k = 1, 3
               M2(k, k, indexb(1, 1)) = ws
            ENDDO

            IF (ZP > 0._dp) THEN
               zt = SQRT(ZS*ZP)
               zb = 0.5_dp*(ZS+ZP)
               DO k = 1, 3
                  M1(k, indexb(1, 1+k)) = (zt/zb)**(2*nr+1)*REAL(2*nr+1, dp)/(2.0*zb*SQRT(3.0_dp))
               ENDDO

               wp = REAL((2*nr+2)*(2*nr+1), dp)/(40.0_dp*ZP**2)
               DO k1 = 1, 3
                  DO k2 = 1, 3
                     IF (k1 == k2) THEN
                        M2(k2, k2, indexb(1+k1, 1+k1)) = 3.0_dp*wp
                     ELSE
                        M2(k2, k2, indexb(1+k1, 1+k1)) = wp
                     ENDIF
                  ENDDO
               ENDDO
               M2(1, 2, indexb(1+1, 1+2)) = wp
               M2(2, 1, indexb(1+1, 1+2)) = wp
               M2(2, 3, indexb(1+2, 1+3)) = wp
               M2(3, 2, indexb(1+2, 1+3)) = wp
               M2(3, 1, indexb(1+3, 1+1)) = wp
               M2(1, 3, indexb(1+3, 1+1)) = wp
            ENDIF

            DO i = 1, natorb
               DO j = 1, i
                  ind1 = indexa(se_map_alm(i), se_map_alm(j))
                  ind2 = indexb(i, j)
                  mpole => mpoles(ind2)%mpole
                  mpole%indi = i
                  mpole%indj = j
                  ! Charge
                  mpole%cs = -M0(indexb(i, j))
                  ! Dipole
                  mpole%ds = -M1(1:3, indexb(i, j))
                  ! Quadrupole
                  mpole%qq = -3._dp*M2(1:3, 1:3, indexb(i, j))
                  IF (debug_this_module) THEN
                     WRITE (*, '(A,2I6,A)') "Orbitals ", i, j, &
                        " ("//label_print_orb(i)//","//label_print_orb(j)//")"
                     WRITE (*, '(9F12.6)') mpole%cs
                     WRITE (*, '(9F12.6)') mpole%ds
                     WRITE (*, '(9F12.6)') mpole%qq
                     WRITE (*, *)
                  END IF
               END DO
            END DO
         ELSE
            mpole%cs = mpole%c
            mpole%ds = mpole%d
            mpole%qq = mpole%qc
         END IF
      END IF

   END SUBROUTINE semi_empirical_mpole_p_setup

! **************************************************************************************************
!> \brief  Transforms the quadrupole components from sphericals to cartesians
!> \param qcart ...
!> \param qsph ...
!> \date   09.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE quadrupole_sph_to_cart(qcart, qsph)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: qcart
      REAL(KIND=dp), DIMENSION(5), INTENT(IN)            :: qsph

      CHARACTER(len=*), PARAMETER :: routineN = 'quadrupole_sph_to_cart', &
         routineP = moduleN//':'//routineN

! Notation
!          qs(1) - d3z2-r2
!          qs(2) - dzx
!          qs(3) - dzy
!          qs(4) - dx2-y2
!          qs(5) - dxy
! Cartesian components

      qcart(1, 1) = (qsph(4)-qsph(1)/SQRT(3.0_dp))*SQRT(3.0_dp)/2.0_dp
      qcart(2, 1) = qsph(5)*SQRT(3.0_dp)/2.0_dp
      qcart(3, 1) = qsph(2)*SQRT(3.0_dp)/2.0_dp
      qcart(2, 2) = -(qsph(4)+qsph(1)/SQRT(3.0_dp))*SQRT(3.0_dp)/2.0_dp
      qcart(3, 2) = qsph(3)*SQRT(3.0_dp)/2.0_dp
      qcart(3, 3) = qsph(1)
      ! Symmetrize tensor
      qcart(1, 2) = qcart(2, 1)
      qcart(1, 3) = qcart(3, 1)
      qcart(2, 3) = qcart(3, 2)

   END SUBROUTINE quadrupole_sph_to_cart

! **************************************************************************************************
!> \brief Setup NDDO multipole type
!> \param nddo_mpole ...
!> \param natom ...
!> \date   09.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE nddo_mpole_setup(nddo_mpole, natom)
      TYPE(nddo_mpole_type), POINTER                     :: nddo_mpole
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'nddo_mpole_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(nddo_mpole)) THEN
         CALL nddo_mpole_release(nddo_mpole)
      END IF
      CALL nddo_mpole_create(nddo_mpole)
      ! Allocate Global Arrays
      ALLOCATE (nddo_mpole%charge(natom))
      ALLOCATE (nddo_mpole%dipole(3, natom))
      ALLOCATE (nddo_mpole%quadrupole(3, 3, natom))

      ALLOCATE (nddo_mpole%efield0(natom))
      ALLOCATE (nddo_mpole%efield1(3, natom))
      ALLOCATE (nddo_mpole%efield2(9, natom))

      CALL timestop(handle)

   END SUBROUTINE nddo_mpole_setup

END MODULE semi_empirical_mpole_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: sep%...
!>      ass, asp, app, a.u.      parameters for the SCP-NDDO calculations
!>      ad, am, aq,    a.u.      parameters for the multipole approximation to
!>      dd, qq         a.u.      the two-centre, two-electron integrals.
!>      alp            A**(-1)   the exponents used to calculate the core-core
!>                               repulsion terms.
!>      beta           eV        parameters for the two-centre, one-electron core
!>                               resonance integrals.
!>      z                        the CORE charge on each element.
!>      eheat          kcal/mol  the experimental heat of formation for each element.
!>      eisol          eV        the calculated electronic energies for each element.
!>      fn1            A*eV      AM1/PM3 specific parameters
!>      fn2            A**(-2)   for the core-core repulsion
!>      fn3            A         interaction.
!>      pre            eV        PDDG function
!>      d              A         PDDG function
!>      bfn1-3         special core-core interactions
!>
!>      gdd, gpd, gsd, eV        parameters for the Coulomb and exchange one-centre,
!>      gp2, gpp, gsp, eV        two-electron integrals.
!>      gss, hsp       eV
!>      uss, upp, udd  eV        electron kinetic energy integrals.
!>      sto_exponents  a.u.      the Slater exponents of the basis functions.
!>
!>      all dimensions will be converted to a.u. at the end of the routines
!> \par History
!>      BK   (12.10.2004)
!>      JGH  (36.03.2006) added PDDG function
!>      Samuele Giani - Teo (28.01.2008) : Parameters for PM6
!>      Teodoro Laino [tlaino] - 03.2008 - University of Zurich: preparing for d-orbitals
!> \author JGH (14.08.2004)
! **************************************************************************************************
MODULE semi_empirical_parameters

   USE input_constants,                 ONLY: do_method_am1,&
                                              do_method_mndod,&
                                              do_method_pchg,&
                                              do_method_pdg,&
                                              do_method_pm3,&
                                              do_method_pm6,&
                                              do_method_pnnl,&
                                              do_method_rm1
   USE kinds,                           ONLY: dp
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: angstrom,&
                                              evolt,&
                                              kcalmol
   USE semi_empirical_par_utils,        ONLY: calpar,&
                                              convert_param_to_cp2k,&
                                              valence_electrons
   USE semi_empirical_types,            ONLY: semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_parameters'

   INTEGER, PARAMETER, PRIVATE :: nelem = 106

   REAL(KIND=dp), DIMENSION(0:nelem, 0:nelem) :: xab_pm6, aab_pm6

   PUBLIC :: mndo_default_parameter, &
             am1_default_parameter, &
             rm1_default_parameter, &
             pdg_default_parameter, &
             pm3_default_parameter, &
             pm6_default_parameter, &
             pcharge_default_parameter, &
             pnnl_default_parameter

CONTAINS

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: POINT_CHARGE
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE pcharge_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER                                            :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'pcharge_default_parameter', &
         routineP = moduleN//':'//routineN

      sep%typ = do_method_pchg
      sep%name = "CLASSICAL"
      sep%eheat = 0.0_dp
      sep%z = z
      sep%defined = .TRUE.
      CALL valence_electrons(sep, extended_basis_set=.FALSE.)

   END SUBROUTINE pcharge_default_parameter

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: MNDO
!> \param sep ...
!> \param z ...
!> \param itype ...
! **************************************************************************************************
   SUBROUTINE mndo_default_parameter(sep, z, itype)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER, INTENT(IN)                                :: z, itype

      CHARACTER(len=*), PARAMETER :: routineN = 'mndo_default_parameter', &
         routineP = moduleN//':'//routineN

      sep%typ = itype
      sep%name = ptable(z)%name
      sep%eheat = ptable(z)%heat_of_formation/kcalmol
      sep%z = z

      SELECT CASE (z)
      CASE DEFAULT
         CALL cp__b("semi_empirical_parameters.F",122,"invalid atomic number")
      CASE (0)
         sep%defined = .TRUE.
      CASE (1)
         !Hydrogen
         sep%defined = .TRUE.
         sep%alp = 2.5441341_dp
         sep%beta(0) = -6.9890640_dp
         sep%sto_exponents(0) = 1.3319670_dp
         sep%uss = -11.9062760_dp
         sep%gss = 12.8480000_dp
         sep%ass = 1.0e-16_dp
         sep%asp = 1.0e-16_dp
         sep%app = 1.0e-16_dp
      CASE (2)
      CASE (3)
         !Lithium
         sep%defined = .TRUE.
         sep%alp = 1.2083987_dp
         sep%beta(0) = -0.1904126_dp
         sep%beta(1) = -0.1904126_dp
         sep%sto_exponents(0) = 0.4296141_dp
         sep%sto_exponents(1) = 0.7554884_dp
         sep%uss = -4.8578570_dp
         sep%upp = -2.0266084_dp
         sep%gss = 7.5947069_dp
         sep%gsp = 6.7259856_dp
         sep%gpp = 8.6596829_dp
         sep%gp2 = 3.8714751_dp
         sep%hsp = 5.0003381_dp
      CASE (4)
         !Beryllium
         sep%defined = .TRUE.
         sep%uss = -16.6023780_dp
         sep%upp = -10.7037710_dp
         sep%beta(0) = -4.0170960_dp
         sep%beta(1) = -4.0170960_dp
         sep%sto_exponents(0) = 1.0042100_dp
         sep%sto_exponents(1) = 1.0042100_dp
         sep%alp = 1.6694340_dp
         sep%gss = 9.0000000_dp
         sep%gsp = 7.4300000_dp
         sep%gpp = 6.9700000_dp
         sep%gp2 = 6.2200000_dp
         sep%hsp = 1.2800000_dp
      CASE (5)
         !Boron
         sep%defined = .TRUE.
         sep%uss = -34.5471300_dp
         sep%upp = -23.1216900_dp
         sep%beta(0) = -8.2520540_dp
         sep%beta(1) = -8.2520540_dp
         sep%sto_exponents(0) = 1.5068010_dp
         sep%sto_exponents(1) = 1.5068010_dp
         sep%alp = 2.1349930_dp
         sep%gss = 10.5900000_dp
         sep%gsp = 9.5600000_dp
         sep%gpp = 8.8600000_dp
         sep%gp2 = 7.8600000_dp
         sep%hsp = 1.8100000_dp
      CASE (6)
         !Carbon
         sep%defined = .TRUE.
         sep%uss = -52.2797450_dp
         sep%upp = -39.2055580_dp
         sep%beta(0) = -18.9850440_dp
         sep%beta(1) = -7.9341220_dp
         sep%sto_exponents(0) = 1.7875370_dp
         sep%sto_exponents(1) = 1.7875370_dp
         sep%alp = 2.5463800_dp
         sep%gss = 12.2300000_dp
         sep%gsp = 11.4700000_dp
         sep%gpp = 11.0800000_dp
         sep%gp2 = 9.8400000_dp
         sep%hsp = 2.4300000_dp
      CASE (7)
         !Nitrogen
         sep%defined = .TRUE.
         sep%uss = -71.9321220_dp
         sep%upp = -57.1723190_dp
         sep%beta(0) = -20.4957580_dp
         sep%beta(1) = -20.4957580_dp
         sep%sto_exponents(0) = 2.2556140_dp
         sep%sto_exponents(1) = 2.2556140_dp
         sep%alp = 2.8613420_dp
         sep%gss = 13.5900000_dp
         sep%gsp = 12.6600000_dp
         sep%gpp = 12.9800000_dp
         sep%gp2 = 11.5900000_dp
         sep%hsp = 3.1400000_dp
      CASE (8)
         !Oxygen
         sep%defined = .TRUE.
         sep%uss = -99.6443090_dp
         sep%upp = -77.7974720_dp
         sep%beta(0) = -32.6880820_dp
         sep%beta(1) = -32.6880820_dp
         sep%sto_exponents(0) = 2.6999050_dp
         sep%sto_exponents(1) = 2.6999050_dp
         sep%alp = 3.1606040_dp
         sep%gss = 15.4200000_dp
         sep%gsp = 14.4800000_dp
         sep%gpp = 14.5200000_dp
         sep%gp2 = 12.9800000_dp
         sep%hsp = 3.9400000_dp
         sep%ass = 1.0e-16_dp
         sep%asp = 6.0_dp
         sep%app = 1.0e-16_dp
      CASE (9)
         !Fluorine
         sep%defined = .TRUE.
         sep%uss = -131.0715480_dp
         sep%upp = -105.7821370_dp
         sep%beta(0) = -48.2904660_dp
         sep%beta(1) = -36.5085400_dp
         sep%sto_exponents(0) = 2.8484870_dp
         sep%sto_exponents(1) = 2.8484870_dp
         sep%alp = 3.4196606_dp
         sep%gss = 16.9200000_dp
         sep%gsp = 17.2500000_dp
         sep%gpp = 16.7100000_dp
         sep%gp2 = 14.9100000_dp
         sep%hsp = 4.8300000_dp
      CASE (10)
      CASE (11)
         !Sodium
         sep%defined = .TRUE.
         sep%uss = -5.1235942_dp
         sep%upp = -3.0124713_dp
         sep%beta(0) = -1.4916657_dp
         sep%beta(1) = -0.2208233_dp
         sep%sto_exponents(0) = 0.8213124_dp
         sep%sto_exponents(1) = 1.0303270_dp
         sep%alp = 5.9940638_dp
         sep%gss = 6.9934520_dp
         sep%gsp = 5.4380530_dp
         sep%gpp = 6.9285408_dp
         sep%gp2 = 2.4299952_dp
         sep%hsp = 3.1374231_dp
      CASE (12)
         !Magnesium
         sep%defined = .TRUE.
         sep%uss = -14.8057921_dp
         sep%upp = -12.7545494_dp
         sep%beta(0) = -0.0999626_dp
         sep%beta(1) = -7.6952692_dp
         sep%sto_exponents(0) = 0.9394811_dp
         sep%sto_exponents(1) = 1.3103428_dp
         sep%alp = 0.5354201_dp
         sep%gss = 6.9632774_dp
         sep%gsp = 8.2410934_dp
         sep%gpp = 10.0000349_dp
         sep%gp2 = 9.0593304_dp
         sep%hsp = 0.7165063_dp
      CASE (13)
         !Aluminium
         sep%defined = .TRUE.
         sep%uss = -23.8070970_dp
         sep%upp = -17.5198780_dp
         sep%beta(0) = -2.6702840_dp
         sep%beta(1) = -2.6702840_dp
         sep%sto_exponents(0) = 1.4441610_dp
         sep%sto_exponents(1) = 1.4441610_dp
         sep%alp = 1.8688394_dp
         sep%gss = 8.0900000_dp
         sep%gsp = 6.6300000_dp
         sep%gpp = 5.9800000_dp
         sep%gp2 = 5.4000000_dp
         sep%hsp = 0.7000000_dp
      CASE (14)
         !Silicon
         sep%defined = .TRUE.
         sep%uss = -37.0375330_dp
         sep%upp = -27.7696780_dp
         sep%beta(0) = -9.0868040_dp
         sep%beta(1) = -1.0758270_dp
         sep%sto_exponents(0) = 1.3159860_dp
         sep%sto_exponents(1) = 1.7099430_dp
         sep%alp = 2.2053160_dp
         sep%gss = 9.8200000_dp
         sep%gsp = 8.3600000_dp
         sep%gpp = 7.3100000_dp
         sep%gp2 = 6.5400000_dp
         sep%hsp = 1.3200000_dp
      CASE (15)
         !Phosphorus
         sep%defined = .TRUE.
         sep%uss = -56.1433600_dp
         sep%upp = -42.8510800_dp
         sep%beta(0) = -6.7916000_dp
         sep%beta(1) = -6.7916000_dp
         sep%sto_exponents(0) = 2.1087200_dp
         sep%sto_exponents(1) = 1.7858100_dp
         sep%alp = 2.4152800_dp
         sep%gss = 11.5600000_dp
         sep%gsp = 10.0800000_dp
         sep%gpp = 8.6400000_dp
         sep%gp2 = 7.6800000_dp
         sep%hsp = 1.9200000_dp
      CASE (16)
         !Sulphur
         sep%defined = .TRUE.
         sep%uss = -72.2422810_dp
         sep%upp = -56.9732070_dp
         sep%beta(0) = -10.7616700_dp
         sep%beta(1) = -10.1084330_dp
         sep%sto_exponents(0) = 2.3129620_dp
         sep%sto_exponents(1) = 2.0091460_dp
         sep%alp = 2.4780260_dp
         sep%gss = 12.8800000_dp
         sep%gsp = 11.2600000_dp
         sep%gpp = 9.9000000_dp
         sep%gp2 = 8.8300000_dp
         sep%hsp = 2.2600000_dp
      CASE (17)
         !Chlorine
         sep%defined = .TRUE.
         sep%uss = -100.2271660_dp
         sep%upp = -77.3786670_dp
         sep%beta(0) = -14.2623200_dp
         sep%beta(1) = -14.2623200_dp
         sep%sto_exponents(0) = 3.7846450_dp
         sep%sto_exponents(1) = 2.0362630_dp
         sep%alp = 2.5422010_dp
         sep%gss = 15.0300000_dp
         sep%gsp = 13.1600000_dp
         sep%gpp = 11.3000000_dp
         sep%gp2 = 9.9700000_dp
         sep%hsp = 2.4200000_dp
      CASE (18)
      CASE (19)
         !Potassium
         sep%defined = .TRUE.
         sep%uss = -3.6401731_dp
         sep%upp = -2.0249253_dp
         sep%beta(0) = -0.1361851_dp
         sep%beta(1) = -2.8142350_dp
         sep%sto_exponents(0) = 0.7276039_dp
         sep%sto_exponents(1) = 0.9871174_dp
         sep%alp = 0.5616422_dp
         sep%gss = 3.7939792_dp
         sep%gsp = 6.4170233_dp
         sep%gpp = 5.0972823_dp
         sep%gp2 = 2.1945567_dp
         sep%hsp = 1.5788130_dp
      CASE (20)
         !Calcium
         sep%defined = .TRUE.
         sep%uss = -12.3919094_dp
         sep%upp = -9.9348289_dp
         sep%beta(0) = -8.6404687_dp
         sep%beta(1) = -9.9515712_dp
         sep%sto_exponents(0) = 1.0034161_dp
         sep%sto_exponents(1) = 1.3102564_dp
         sep%alp = 0.4999997_dp
         sep%gss = 6.5321649_dp
         sep%gsp = 6.5424442_dp
         sep%gpp = 6.4627059_dp
         sep%gp2 = 6.3842472_dp
         sep%hsp = 0.5789676_dp
      CASE (21:29)
      CASE (30)
         !Zinc
         sep%defined = .TRUE.
         sep%uss = -20.8397160_dp
         sep%upp = -19.6252240_dp
         sep%beta(0) = -1.0000000_dp
         sep%beta(1) = -2.0000000_dp
         sep%sto_exponents(0) = 2.0473590_dp
         sep%sto_exponents(1) = 1.4609460_dp
         sep%alp = 1.5064570_dp
         sep%gss = 11.8000000_dp
         sep%gsp = 11.1820180_dp
         sep%gpp = 13.3000000_dp
         sep%gp2 = 12.9305200_dp
         sep%hsp = 0.4846060_dp
      CASE (31)
         !Gallium
         sep%defined = .TRUE.
         sep%uss = -28.3044924_dp
         sep%upp = -27.2063910_dp
         sep%beta(0) = -3.9987435_dp
         sep%beta(1) = -4.3146711_dp
         sep%sto_exponents(0) = 0.6986316_dp
         sep%sto_exponents(1) = 1.8386933_dp
         sep%alp = 2.7577991_dp
         sep%gss = 7.5468114_dp
         sep%gsp = 10.4697612_dp
         sep%gpp = 8.4599454_dp
         sep%gp2 = 10.4251148_dp
         sep%hsp = 1.0628013_dp
      CASE (32)
         !Germanium
         sep%defined = .TRUE.
         sep%uss = -33.9493670_dp
         sep%upp = -27.4251050_dp
         sep%beta(0) = -4.5164790_dp
         sep%beta(1) = -1.7555170_dp
         sep%sto_exponents(0) = 1.2931800_dp
         sep%sto_exponents(1) = 2.0205640_dp
         sep%alp = 1.9784980_dp
         sep%gss = 9.8000000_dp
         sep%gsp = 8.3000000_dp
         sep%gpp = 7.3000000_dp
         sep%gp2 = 6.5000000_dp
         sep%hsp = 1.3000000_dp
      CASE (33)
         !Arsenic
         sep%defined = .TRUE.
         sep%uss = -38.6240790_dp
         sep%upp = -33.4995395_dp
         sep%beta(0) = -3.9998231_dp
         sep%beta(1) = -4.9056176_dp
         sep%sto_exponents(0) = 2.5614338_dp
         sep%sto_exponents(1) = 1.6117315_dp
         sep%alp = 1.9381219_dp
         sep%gss = 6.7464011_dp
         sep%gsp = 5.6174985_dp
         sep%gpp = 6.9333925_dp
         sep%gp2 = 6.3054798_dp
         sep%hsp = 0.5994346_dp
      CASE (34)
         !Selenium
         sep%defined = .TRUE.
         sep%uss = -49.8117347_dp
         sep%upp = -38.0475911_dp
         sep%beta(0) = -12.4685178_dp
         sep%beta(1) = -5.1744376_dp
         sep%sto_exponents(0) = 0.7242956_dp
         sep%sto_exponents(1) = 1.9267288_dp
         sep%alp = 2.3513155_dp
         sep%gss = 10.3549483_dp
         sep%gsp = 5.2801360_dp
         sep%gpp = 7.3611317_dp
         sep%gp2 = 6.1897284_dp
         sep%hsp = 0.5996560_dp
      CASE (35)
         !Bromine
         sep%defined = .TRUE.
         sep%uss = -99.98644050_dp
         sep%upp = -75.67130750_dp
         sep%beta(0) = -8.91710700_dp
         sep%beta(1) = -9.94374000_dp
         sep%sto_exponents(0) = 3.85430190_dp
         sep%sto_exponents(1) = 2.19920910_dp
         sep%alp = 2.44570510_dp
         sep%gss = 15.03643948_dp
         sep%gsp = 13.03468242_dp
         sep%gpp = 11.27632539_dp
         sep%gp2 = 9.85442552_dp
         sep%hsp = 2.45586832_dp
      CASE (36)
      CASE (37)
         !Rubidium
         sep%defined = .TRUE.
         sep%uss = -4.3098071_dp
         sep%upp = -2.7381921_dp
         sep%beta(0) = -2.2062173_dp
         sep%beta(1) = -6.2176392_dp
         sep%sto_exponents(0) = 4.0001632_dp
         sep%sto_exponents(1) = 0.9187408_dp
         sep%alp = 0.9976197_dp
         sep%gss = 10.7409462_dp
         sep%gsp = 11.4853623_dp
         sep%gpp = 8.9878480_dp
         sep%gp2 = 7.7258368_dp
         sep%hsp = 0.1999294_dp
      CASE (38)
         !Strontium
         sep%defined = .TRUE.
         sep%uss = -10.8451287_dp
         sep%upp = -8.3129821_dp
         sep%beta(0) = -9.9683427_dp
         sep%beta(1) = -9.9946390_dp
         sep%sto_exponents(0) = 1.3729266_dp
         sep%sto_exponents(1) = 1.1118128_dp
         sep%alp = 0.5082703_dp
         sep%gss = 4.9305520_dp
         sep%gsp = 4.4249843_dp
         sep%gpp = 4.0178045_dp
         sep%gp2 = 4.0335640_dp
         sep%hsp = 0.6272993_dp
      CASE (39:47)
      CASE (48)
         !Cadmium
         sep%defined = .TRUE.
         sep%uss = -26.1908325_dp
         sep%upp = -22.4523736_dp
         sep%beta(0) = -11.9610608_dp
         sep%beta(1) = -3.9999848_dp
         sep%sto_exponents(0) = 1.4192491_dp
         sep%sto_exponents(1) = 1.0480637_dp
         sep%alp = 1.1507745_dp
         sep%gss = 17.2196544_dp
         sep%gsp = 17.9900180_dp
         sep%gpp = 19.2007795_dp
         sep%gp2 = 19.0000212_dp
         sep%hsp = 0.7384511_dp
      CASE (49)
         !Indium
         sep%defined = .TRUE.
         sep%uss = -28.0980892_dp
         sep%upp = -19.2780588_dp
         sep%beta(0) = -6.3107479_dp
         sep%beta(1) = -2.7025837_dp
         sep%sto_exponents(0) = 1.7625740_dp
         sep%sto_exponents(1) = 1.8648962_dp
         sep%alp = 2.3438756_dp
         sep%gss = 9.4928794_dp
         sep%gsp = 7.0094241_dp
         sep%gpp = 9.6640986_dp
         sep%gp2 = 7.0100315_dp
         sep%hsp = 0.5995894_dp
      CASE (50)
         !Tin
         sep%defined = .TRUE.
         sep%uss = -40.8518020_dp
         sep%upp = -28.5602490_dp
         sep%beta(0) = -3.2351470_dp
         sep%beta(1) = -4.2904160_dp
         sep%sto_exponents(0) = 2.0803800_dp
         sep%sto_exponents(1) = 1.9371060_dp
         sep%alp = 1.8008140_dp
         sep%gss = 9.8000000_dp
         sep%gsp = 8.3000000_dp
         sep%gpp = 7.3000000_dp
         sep%gp2 = 6.5000000_dp
         sep%hsp = 1.3000000_dp
      CASE (51)
         !Antimony
         sep%defined = .TRUE.
         sep%uss = -42.0643435_dp
         sep%upp = -35.0626031_dp
         sep%beta(0) = -0.9999715_dp
         sep%beta(1) = -4.0920176_dp
         sep%sto_exponents(0) = 3.6458835_dp
         sep%sto_exponents(1) = 1.9733156_dp
         sep%alp = 1.9763403_dp
         sep%gss = 10.6739308_dp
         sep%gsp = 7.0477648_dp
         sep%gpp = 6.7446162_dp
         sep%gp2 = 6.3408531_dp
         sep%hsp = 0.5997512_dp
      CASE (52)
         !Tellurium
         sep%defined = .TRUE.
         sep%uss = -84.2274722_dp
         sep%upp = -46.5332871_dp
         sep%beta(0) = -8.5622652_dp
         sep%beta(1) = -2.6942963_dp
         sep%sto_exponents(0) = 2.7461609_dp
         sep%sto_exponents(1) = 1.6160376_dp
         sep%alp = 2.2924145_dp
         sep%gss = 5.1367706_dp
         sep%gsp = 11.0720752_dp
         sep%gpp = 5.8447934_dp
         sep%gp2 = 5.0720495_dp
         sep%hsp = 0.5997994_dp
      CASE (53)
         !Iodine
         sep%defined = .TRUE.
         sep%uss = -100.00305380_dp
         sep%upp = -74.61146920_dp
         sep%beta(0) = -7.41445100_dp
         sep%beta(1) = -6.19678100_dp
         sep%sto_exponents(0) = 2.27296100_dp
         sep%sto_exponents(1) = 2.16949800_dp
         sep%alp = 2.20732000_dp
         sep%gss = 15.04044855_dp
         sep%gsp = 13.05655798_dp
         sep%gpp = 11.14778369_dp
         sep%gp2 = 9.91409071_dp
         sep%hsp = 2.45638202_dp
      CASE (54)
      CASE (55)
         !Cesium
         sep%defined = .TRUE.
         sep%uss = -3.2184078_dp
         sep%upp = -1.7699113_dp
         sep%beta(0) = -1.6043600_dp
         sep%beta(1) = -4.2698040_dp
         sep%sto_exponents(0) = 6.0004170_dp
         sep%sto_exponents(1) = 0.8986916_dp
         sep%alp = 0.4981646_dp
         sep%gss = 7.6447851_dp
         sep%gsp = 3.0454989_dp
         sep%gpp = 10.0000745_dp
         sep%gp2 = 6.1761092_dp
         sep%hsp = 0.4647853_dp
      CASE (56)
         !Barium
         sep%defined = .TRUE.
         sep%uss = -10.1125345_dp
         sep%upp = -8.2347224_dp
         sep%beta(0) = -9.9994459_dp
         sep%beta(1) = -9.6197255_dp
         sep%sto_exponents(0) = 1.9765973_dp
         sep%sto_exponents(1) = 1.3157348_dp
         sep%alp = 0.8594840_dp
         sep%gss = 4.8486178_dp
         sep%gsp = 4.5659982_dp
         sep%gpp = 5.0937708_dp
         sep%gp2 = 5.2125824_dp
         sep%hsp = 0.5237082_dp
      CASE (57:79)
      CASE (80)
         !Mercury
         sep%defined = .TRUE.
         sep%uss = -19.8095740_dp
         sep%upp = -13.1025300_dp
         sep%beta(0) = -0.4045250_dp
         sep%beta(1) = -6.2066830_dp
         sep%sto_exponents(0) = 2.2181840_dp
         sep%sto_exponents(1) = 2.0650380_dp
         sep%alp = 1.3356410_dp
         sep%gss = 10.8000000_dp
         sep%gsp = 9.3000000_dp
         sep%gpp = 14.3000000_dp
         sep%gp2 = 13.5000000_dp
         sep%hsp = 1.3000000_dp
      CASE (81)
         !Thallium
         sep%defined = .TRUE.
         sep%uss = -29.7009655_dp
         sep%upp = -29.5937539_dp
         sep%beta(0) = -4.9667442_dp
         sep%beta(1) = -7.7616060_dp
         sep%sto_exponents(0) = 4.0000447_dp
         sep%sto_exponents(1) = 1.8076332_dp
         sep%alp = 1.3116968_dp
         sep%gss = 8.8675337_dp
         sep%gsp = 12.1148290_dp
         sep%gpp = 10.6532769_dp
         sep%gp2 = 13.5333191_dp
         sep%hsp = 0.5997565_dp
      CASE (82)
         !Lead
         sep%defined = .TRUE.
         sep%uss = -47.3196920_dp
         sep%upp = -28.8475600_dp
         sep%beta(0) = -8.0423870_dp
         sep%beta(1) = -3.0000000_dp
         sep%sto_exponents(0) = 2.4982860_dp
         sep%sto_exponents(1) = 2.0820710_dp
         sep%alp = 1.7283330_dp
         sep%gss = 9.8000000_dp
         sep%gsp = 8.3000000_dp
         sep%gpp = 7.3000000_dp
         sep%gp2 = 6.5000000_dp
         sep%hsp = 1.3000000_dp
      CASE (83)
         !Bismuth
         sep%defined = .TRUE.
         sep%uss = -53.5827147_dp
         sep%upp = -39.4572213_dp
         sep%beta(0) = -9.0000249_dp
         sep%beta(1) = -1.9830269_dp
         sep%sto_exponents(0) = 2.6772255_dp
         sep%sto_exponents(1) = 0.6936864_dp
         sep%alp = 5.7660628_dp
         sep%gss = 8.3702778_dp
         sep%gsp = 7.7974668_dp
         sep%gpp = 9.8303165_dp
         sep%gp2 = 8.9291355_dp
         sep%hsp = 0.5999908_dp
      CASE (84:nelem)
      END SELECT
      ! In case overwrite with MNDO-d parameters
      IF (itype == do_method_mndod) CALL mndod_default_parameter(sep, z)
      ! Finalize parameters for the element
      CALL valence_electrons(sep, extended_basis_set=(itype == do_method_mndod))
      CALL calpar(z, sep)
      CALL convert_param_to_cp2k(sep)

   END SUBROUTINE mndo_default_parameter

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models developed at PNNL
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE pnnl_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER, INTENT(IN)                                :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'pnnl_default_parameter', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: Ang = 0.529167_dp, eV = 27.21_dp

      sep%typ = do_method_pnnl
      sep%name = ptable(z)%name
      sep%eheat = ptable(z)%heat_of_formation/kcalmol
      sep%z = z

      SELECT CASE (z)
      CASE DEFAULT
         CALL cp__b("semi_empirical_parameters.F",718,"invalid atomic number")
      CASE (0)
         sep%defined = .TRUE.
      CASE (1)
         !Hydrogen
         sep%defined = .TRUE.
         sep%alp = 3.323150026265647_dp*Ang/angstrom
         sep%beta(0) = -6.070138572957370_dp*evolt/eV
         sep%beta(1) = 0.0_dp*evolt/eV
         sep%sto_exponents(0) = 1.220643961907751_dp
         sep%sto_exponents(1) = 1.351277996861119_dp
         sep%uss = -11.185667_dp*evolt/eV
         sep%upp = 1000000.0_dp*evolt/eV
         sep%gss = 14.794208_dp*evolt/eV
         sep%gsp = 0.0_dp*evolt/eV
         sep%gpp = 0.0_dp*evolt/eV
         sep%gp2 = 0.0_dp*evolt/eV
         sep%hsp = 0.0_dp*evolt/eV
         sep%ass = 1.0e-16_dp
         sep%asp = 3.355792629670590_dp
         sep%app = 0.03941680815458009_dp
         sep%de = 20.06411392885370_dp/eV
         sep%nr = 1
         sep%acoul = 16.27368453086666_dp/eV
         ! This parameterization explicitly supports
         ! p-orbitals for H. So by default we turn them
         ! on for this element.
         sep%p_orbitals_on_h = .TRUE.
      CASE (8)
         !Oxygen
         sep%defined = .TRUE.
         sep%alp = 2.980042911001686_dp*Ang/angstrom
         sep%beta(0) = -40.62654394990248_dp*evolt/eV
         sep%beta(1) = -28.39467965792422_dp*evolt/eV
         sep%sto_exponents(0) = 3.814650982832301_dp
         sep%sto_exponents(1) = 2.446497244623426_dp
         sep%uss = -79.081852_dp*evolt/eV
         sep%upp = -72.59032_dp*evolt/eV
         sep%gss = 15.75576_dp*evolt/eV
         sep%gsp = 10.62116_dp*evolt/eV
         sep%gpp = 13.654016_dp*evolt/eV
         sep%gp2 = 12.406095_dp*evolt/eV
         sep%hsp = 0.593883_dp*evolt/eV
         sep%ass = 1.0e-16_dp
         sep%asp = 3.583554483196465_dp
         sep%app = 0.0008605218877175237_dp
         sep%de = 7.159002319741392_dp/eV
         sep%nr = 2
         sep%acoul = 16.22361790703965_dp/eV
      END SELECT
      ! Finalize parameters for the element
      CALL valence_electrons(sep, extended_basis_set=.FALSE.)
      CALL calpar(z, sep)
      CALL convert_param_to_cp2k(sep)

   END SUBROUTINE pnnl_default_parameter

! **************************************************************************************************
!> \brief Additional default parameter sets for semi empirical models: MNDO-d
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE mndod_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER                                            :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'mndod_default_parameter', &
         routineP = moduleN//':'//routineN

      SELECT CASE (z)
      CASE (11)
         ! Sodium
         sep%defined = .TRUE.
         sep%uss = -5.20100000_dp
         sep%upp = -2.71257317_dp
         sep%beta(0) = -1.08738166_dp
         sep%beta(1) = -0.48623935_dp
         sep%sto_exponents(0) = 0.98750834_dp
         sep%sto_exponents(1) = 0.89334983_dp
         sep%zn(0) = 0.65411258_dp
         sep%zn(1) = 0.56440874_dp
         sep%alp = 1.17010202_dp
         sep%gss = 4.59444476_dp
         sep%gsp = 4.14757400_dp
         sep%gpp = 4.29919761_dp
         sep%gp2 = 3.79695732_dp
         sep%hsp = 0.53440874_dp
      CASE (12)
         ! Magnesium
         sep%defined = .TRUE.
         sep%uss = -15.09700000_dp
         sep%upp = -10.65000000_dp
         sep%beta(0) = -1.89588355_dp
         sep%beta(1) = -2.14108943_dp
         sep%sto_exponents(0) = 1.44890446_dp
         sep%sto_exponents(1) = 0.95293002_dp
         sep%zn(0) = 1.05000000_dp
         sep%zn(1) = 0.92527190_dp
         sep%alp = 1.62146984_dp
         sep%gss = 7.37513258_dp
         sep%gsp = 6.88890741_dp
         sep%gpp = 7.04795383_dp
         sep%gp2 = 6.22459871_dp
         sep%hsp = 0.72673390_dp
      CASE (13)
         !Aluminium
         sep%defined = .TRUE.
         sep%uss = -24.01792910_dp
         sep%upp = -20.79597967_dp
         sep%udd = -5.22082737_dp
         sep%beta(0) = -7.10185851_dp
         sep%beta(1) = -2.31809618_dp
         sep%beta(2) = -3.35638545_dp
         sep%sto_exponents(0) = 1.79402273_dp
         sep%sto_exponents(1) = 1.37130919_dp
         sep%sto_exponents(2) = 0.80591133_dp
         sep%zn(0) = 1.22249269_dp
         sep%zn(1) = 1.09291990_dp
         sep%zn(2) = 0.80038285_dp
         sep%alp = 1.44301676_dp
         sep%gss = 8.58671016_dp
         sep%gsp = 7.66469306_dp
         sep%gpp = 8.32495723_dp
         sep%gp2 = 7.35242020_dp
         sep%hsp = 0.54401293_dp
      CASE (14)
         !Silicon
         sep%defined = .TRUE.
         sep%uss = -36.05153000_dp
         sep%upp = -27.53569100_dp
         sep%udd = -14.67743900_dp
         sep%beta(0) = -8.21073420_dp
         sep%beta(1) = -4.88462030_dp
         sep%beta(2) = -2.60801150_dp
         sep%sto_exponents(0) = 1.91565460_dp
         sep%sto_exponents(1) = 1.68161130_dp
         sep%sto_exponents(2) = 0.96677166_dp
         sep%zn(0) = 1.52929180_dp
         sep%zn(1) = 0.97628075_dp
         sep%zn(2) = 0.93816441_dp
         sep%alp = 1.66006930_dp
         sep%gss = 10.74164700_dp
         sep%gsp = 7.56066640_dp
         sep%gpp = 7.43649690_dp
         sep%gp2 = 6.56775150_dp
         sep%hsp = 0.87753880_dp
      CASE (15)
         !Phosphorus
         sep%defined = .TRUE.
         sep%uss = -47.05552900_dp
         sep%upp = -38.06705900_dp
         sep%udd = -23.69159700_dp
         sep%beta(0) = -8.90210430_dp
         sep%beta(1) = -9.38611080_dp
         sep%beta(2) = -2.09170080_dp
         sep%sto_exponents(0) = 2.26646290_dp
         sep%sto_exponents(1) = 1.94001490_dp
         sep%sto_exponents(2) = 1.10010900_dp
         sep%zn(0) = 1.63437610_dp
         sep%zn(1) = 1.08291170_dp
         sep%zn(2) = 1.00651470_dp
         sep%alp = 1.85255120_dp
         sep%gss = 11.4797530_dp
         sep%gsp = 8.55756910_dp
         sep%gpp = 8.24872280_dp
         sep%gp2 = 7.28509170_dp
         sep%hsp = 2.10780440_dp
      CASE (16)
         !Sulphur
         sep%defined = .TRUE.
         sep%uss = -56.88912800_dp
         sep%upp = -47.27474500_dp
         sep%udd = -25.09511800_dp
         sep%beta(0) = -10.99954500_dp
         sep%beta(1) = -12.21543700_dp
         sep%beta(2) = -1.88066950_dp
         sep%sto_exponents(0) = 2.22585050_dp
         sep%sto_exponents(1) = 2.09970560_dp
         sep%sto_exponents(2) = 1.23147250_dp
         sep%zn(0) = 1.73639140_dp
         sep%zn(1) = 1.12118170_dp
         sep%zn(2) = 1.05084670_dp
         sep%alp = 2.02305950_dp
         sep%gss = 12.19630200_dp
         sep%gsp = 8.85390090_dp
         sep%gpp = 8.54023240_dp
         sep%gp2 = 7.54254650_dp
         sep%hsp = 2.64635230_dp
      CASE (17)
         !Chlorine
         sep%defined = .TRUE.
         sep%uss = -69.62297275_dp
         sep%upp = -59.10072899_dp
         sep%udd = -36.67457320_dp
         sep%beta(0) = -6.03729165_dp
         sep%beta(1) = -19.18338497_dp
         sep%beta(2) = -1.87778198_dp
         sep%sto_exponents(0) = 2.56161065_dp
         sep%sto_exponents(1) = 2.38933800_dp
         sep%sto_exponents(2) = 1.25139777_dp
         sep%zn(0) = 1.88087547_dp
         sep%zn(1) = 1.18104227_dp
         sep%zn(2) = 1.14061555_dp
         sep%alp = 2.18030019_dp
         sep%gss = 13.21114854_dp
         sep%gsp = 9.41949513_dp
         sep%gpp = 8.99620033_dp
         sep%gp2 = 7.94524745_dp
         sep%hsp = 3.08149862_dp
      CASE (30)
         !Zinc
         sep%defined = .TRUE.
         sep%uss = -18.02300000_dp
         sep%upp = -12.24216585_dp
         sep%beta(0) = -5.01726076_dp
         sep%beta(1) = -0.71205972_dp
         sep%sto_exponents(0) = 1.73150352_dp
         sep%sto_exponents(1) = 1.39358305_dp
         sep%zn(0) = 1.56600000_dp
         sep%zn(1) = 0.862839817_dp
         sep%alp = 1.51763697_dp
         sep%gss = 8.56072836_dp
         sep%gsp = 7.49003598_dp
         sep%gpp = 5.13964830_dp
         sep%gp2 = 4.50540309_dp
         sep%hsp = 0.53294610_dp
      CASE (35)
         !Bromine
         sep%defined = .TRUE.
         sep%uss = -65.40277790_dp
         sep%upp = -54.55375352_dp
         sep%udd = -13.72809929_dp
         sep%beta(0) = -8.31497607_dp
         sep%beta(1) = -10.50704145_dp
         sep%beta(2) = -0.96259930_dp
         sep%sto_exponents(0) = 2.59054101_dp
         sep%sto_exponents(1) = 2.33085649_dp
         sep%sto_exponents(2) = 1.35736115_dp
         sep%zn(0) = 2.23581544_dp
         sep%zn(1) = 1.43292654_dp
         sep%zn(2) = 1.24257826_dp
         sep%alp = 2.09105000_dp
         sep%gss = 12.22235546_dp
         sep%gsp = 8.26372010_dp
         sep%gpp = 8.53546437_dp
         sep%gp2 = 7.48216712_dp
         sep%hsp = 2.74952230_dp
      CASE (48)
         ! Cadmium
         sep%defined = .TRUE.
         sep%uss = -16.96970000_dp
         sep%upp = -12.40096476_dp
         sep%beta(0) = -2.77154379_dp
         sep%beta(1) = -1.80565019_dp
         sep%sto_exponents(0) = 1.74880559_dp
         sep%sto_exponents(1) = 1.56321473_dp
         sep%zn(0) = 1.76314840_dp
         sep%zn(1) = 1.52551900_dp
         sep%alp = 1.42461329_dp
         sep%gss = 7.90443438_dp
         sep%gsp = 7.51570687_dp
         sep%gpp = 7.47999993_dp
         sep%gp2 = 6.51866416_dp
         sep%hsp = 0.63674441_dp
      CASE (53)
         !Iodine
         sep%defined = .TRUE.
         sep%uss = -62.76535256_dp
         sep%upp = -50.29211568_dp
         sep%udd = -12.24830501_dp
         sep%beta(0) = -10.69948666_dp
         sep%beta(1) = -4.94117776_dp
         sep%beta(2) = -2.35046098_dp
         sep%sto_exponents(0) = 2.75654324_dp
         sep%sto_exponents(1) = 2.25307954_dp
         sep%sto_exponents(2) = 1.50233509_dp
         sep%zn(0) = 2.67241100_dp
         sep%zn(1) = 1.57229871_dp
         sep%zn(2) = 1.25884802_dp
         sep%alp = 1.90617441_dp
         sep%gss = 11.98078196_dp
         sep%gsp = 7.85590192_dp
         sep%gpp = 7.70937227_dp
         sep%gp2 = 6.71855729_dp
         sep%hsp = 2.07147462_dp
      CASE (80)
         !Mercury
         sep%defined = .TRUE.
         sep%uss = -18.81564903_dp
         sep%upp = -13.39711352_dp
         sep%beta(0) = -2.21872239_dp
         sep%beta(1) = -2.90978573_dp
         sep%sto_exponents(0) = 2.33310757_dp
         sep%sto_exponents(1) = 1.70831069_dp
         sep%zn(0) = 2.18600011_dp
         sep%zn(1) = 1.70500462_dp
         sep%alp = 1.38224172_dp
         sep%gss = 8.31564948_dp
         sep%gsp = 8.21217300_dp
         sep%gpp = 7.11525878_dp
         sep%gp2 = 6.17124983_dp
         sep%hsp = 0.83594100_dp
      END SELECT

   END SUBROUTINE mndod_default_parameter

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: AM1
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE am1_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER                                            :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'am1_default_parameter', &
         routineP = moduleN//':'//routineN

      sep%typ = do_method_am1
      sep%name = ptable(z)%name
      sep%eheat = ptable(z)%heat_of_formation/kcalmol
      sep%z = z

      SELECT CASE (z)
      CASE DEFAULT
         CALL cp__b("semi_empirical_parameters.F",1043,"invalid atomic number")
      CASE (0)
         sep%defined = .TRUE.
      CASE (1)
         !Hydrogen
         sep%defined = .TRUE.
         sep%alp = 2.8823240_dp
         sep%beta(0) = -6.1737870_dp
         sep%sto_exponents(0) = 1.1880780_dp
         sep%uss = -11.3964270_dp
         sep%gss = 12.8480000_dp
         sep%fn1(1) = 0.1227960_dp; sep%fn2(1) = 5.0000000_dp; sep%fn3(1) = 1.2000000_dp
         sep%fn1(2) = 0.0050900_dp; sep%fn2(2) = 5.0000000_dp; sep%fn3(2) = 1.8000000_dp
         sep%fn1(3) = -0.0183360_dp; sep%fn2(3) = 2.0000000_dp; sep%fn3(3) = 2.1000000_dp
      CASE (2)
      CASE (3)
         !Lithium
         sep%defined = .TRUE.
         sep%alp = 1.2501400_dp
         sep%beta(0) = -1.3500400_dp
         sep%beta(1) = -1.3500400_dp
         sep%sto_exponents(0) = 0.7023800_dp
         sep%sto_exponents(1) = 0.7023800_dp
         sep%uss = -5.1280000_dp
         sep%upp = -2.7212000_dp
         sep%gss = 7.3000000_dp
         sep%gsp = 5.4200000_dp
         sep%gpp = 5.0000000_dp
         sep%gp2 = 4.5200000_dp
         sep%hsp = 0.8300000_dp
      CASE (4)
         !Beryllium
         sep%defined = .TRUE.
         sep%alp = 1.6694340_dp
         sep%beta(0) = -4.0170960_dp
         sep%beta(1) = -4.0170960_dp
         sep%sto_exponents(0) = 1.0042100_dp
         sep%sto_exponents(1) = 1.0042100_dp
         sep%uss = -16.6023780_dp
         sep%upp = -10.7037710_dp
         sep%gss = 9.0000000_dp
         sep%gsp = 7.4300000_dp
         sep%gpp = 6.9700000_dp
         sep%gp2 = 6.2200000_dp
         sep%hsp = 1.2800000_dp
      CASE (5)
         !Boron
         sep%defined = .TRUE.
         sep%alp = 2.4469090_dp
         sep%beta(0) = -9.5991140_dp
         sep%beta(1) = -6.2737570_dp
         sep%sto_exponents(0) = 1.6117090_dp
         sep%sto_exponents(1) = 1.5553850_dp
         sep%uss = -34.4928700_dp
         sep%upp = -22.6315250_dp
         sep%gss = 10.5900000_dp
         sep%gsp = 9.5600000_dp
         sep%gpp = 8.8600000_dp
         sep%gp2 = 7.8600000_dp
         sep%hsp = 1.8100000_dp
         sep%bfn1(1, 1) = 0.1826130_dp; sep%bfn2(1, 1) = 6.0000000_dp; sep%bfn3(1, 1) = 0.7275920_dp
         sep%bfn1(2, 1) = 0.1185870_dp; sep%bfn2(2, 1) = 6.0000000_dp; sep%bfn3(2, 1) = 1.4666390_dp
         sep%bfn1(3, 1) = -0.0732800_dp; sep%bfn2(3, 1) = 5.0000000_dp; sep%bfn3(3, 1) = 1.5709750_dp
         sep%bfn1(1, 2) = 0.4122530_dp; sep%bfn2(1, 2) = 10.0000000_dp; sep%bfn3(1, 2) = 0.8325860_dp
         sep%bfn1(2, 2) = -0.1499170_dp; sep%bfn2(2, 2) = 6.0000000_dp; sep%bfn3(2, 2) = 1.1862200_dp
         sep%bfn1(3, 2) = 0.0000000_dp; sep%bfn2(3, 2) = 0.0000000_dp; sep%bfn3(3, 2) = 0.0000000_dp
         sep%bfn1(1, 3) = 0.2617510_dp; sep%bfn2(1, 3) = 8.0000000_dp; sep%bfn3(1, 3) = 1.0639950_dp
         sep%bfn1(2, 3) = 0.0502750_dp; sep%bfn2(2, 3) = 5.0000000_dp; sep%bfn3(2, 3) = 1.9364920_dp
         sep%bfn1(3, 3) = 0.0000000_dp; sep%bfn2(3, 3) = 0.0000000_dp; sep%bfn3(3, 3) = 0.0000000_dp
         sep%bfn1(1, 4) = 0.3592440_dp; sep%bfn2(1, 4) = 9.0000000_dp; sep%bfn3(1, 4) = 0.8193510_dp
         sep%bfn1(2, 4) = 0.0747290_dp; sep%bfn2(2, 4) = 9.0000000_dp; sep%bfn3(2, 4) = 1.5744140_dp
         sep%bfn1(3, 4) = 0.0000000_dp; sep%bfn2(3, 4) = 0.0000000_dp; sep%bfn3(3, 4) = 0.0000000_dp
      CASE (6)
         !Carbon
         sep%defined = .TRUE.
         sep%alp = 2.6482740_dp
         sep%beta(0) = -15.7157830_dp
         sep%beta(1) = -7.7192830_dp
         sep%sto_exponents(0) = 1.8086650_dp
         sep%sto_exponents(1) = 1.6851160_dp
         sep%uss = -52.0286580_dp
         sep%upp = -39.6142390_dp
         sep%gss = 12.2300000_dp
         sep%gsp = 11.4700000_dp
         sep%gpp = 11.0800000_dp
         sep%gp2 = 9.8400000_dp
         sep%hsp = 2.4300000_dp
         sep%fn1(1) = 0.0113550_dp; sep%fn2(1) = 5.0000000_dp; sep%fn3(1) = 1.6000000_dp
         sep%fn1(2) = 0.0459240_dp; sep%fn2(2) = 5.0000000_dp; sep%fn3(2) = 1.8500000_dp
         sep%fn1(3) = -0.0200610_dp; sep%fn2(3) = 5.0000000_dp; sep%fn3(3) = 2.0500000_dp
         sep%fn1(4) = -0.0012600_dp; sep%fn2(4) = 5.0000000_dp; sep%fn3(4) = 2.6500000_dp
      CASE (7)
         !Nitrogen
         sep%defined = .TRUE.
         sep%alp = 2.9472860_dp
         sep%beta(0) = -20.2991100_dp
         sep%beta(1) = -18.2386660_dp
         sep%sto_exponents(0) = 2.3154100_dp
         sep%sto_exponents(1) = 2.1579400_dp
         sep%uss = -71.8600000_dp
         sep%upp = -57.1675810_dp
         sep%gss = 13.5900000_dp
         sep%gsp = 12.6600000_dp
         sep%gpp = 12.9800000_dp
         sep%gp2 = 11.5900000_dp
         sep%hsp = 3.1400000_dp
         sep%fn1(1) = 0.0252510_dp; sep%fn2(1) = 5.0000000_dp; sep%fn3(1) = 1.5000000_dp
         sep%fn1(2) = 0.0289530_dp; sep%fn2(2) = 5.0000000_dp; sep%fn3(2) = 2.1000000_dp
         sep%fn1(3) = -0.0058060_dp; sep%fn2(3) = 2.0000000_dp; sep%fn3(3) = 2.4000000_dp
      CASE (8)
         !Oxygen
         sep%defined = .TRUE.
         sep%alp = 4.4553710_dp
         sep%beta(0) = -29.2727730_dp
         sep%beta(1) = -29.2727730_dp
         sep%sto_exponents(0) = 3.1080320_dp
         sep%sto_exponents(1) = 2.5240390_dp
         sep%uss = -97.8300000_dp
         sep%upp = -78.2623800_dp
         sep%gss = 15.4200000_dp
         sep%gsp = 14.4800000_dp
         sep%gpp = 14.5200000_dp
         sep%gp2 = 12.9800000_dp
         sep%hsp = 3.9400000_dp
         sep%fn1(1) = 0.2809620_dp; sep%fn2(1) = 5.0000000_dp; sep%fn3(1) = 0.8479180_dp
         sep%fn1(2) = 0.0814300_dp; sep%fn2(2) = 7.0000000_dp; sep%fn3(2) = 1.4450710_dp
      CASE (9)
         !Fluorine
         sep%defined = .TRUE.
         sep%alp = 5.5178000_dp
         sep%beta(0) = -69.5902770_dp
         sep%beta(1) = -27.9223600_dp
         sep%sto_exponents(0) = 3.7700820_dp
         sep%sto_exponents(1) = 2.4946700_dp
         sep%uss = -136.1055790_dp
         sep%upp = -104.8898850_dp
         sep%gss = 16.9200000_dp
         sep%gsp = 17.2500000_dp
         sep%gpp = 16.7100000_dp
         sep%gp2 = 14.9100000_dp
         sep%hsp = 4.8300000_dp
         sep%fn1(1) = 0.2420790_dp; sep%fn2(1) = 4.8000000_dp; sep%fn3(1) = 0.9300000_dp
         sep%fn1(2) = 0.0036070_dp; sep%fn2(2) = 4.6000000_dp; sep%fn3(2) = 1.6600000_dp
      CASE (10:11)
      CASE (12)
         !Magnesium
         sep%defined = .TRUE.
         sep%uss = -14.96959313_dp
         sep%upp = -11.56229248_dp
         sep%beta(0) = -1.25974355_dp
         sep%beta(1) = -0.77836604_dp
         sep%sto_exponents(0) = 1.22339270_dp
         sep%sto_exponents(1) = 1.02030798_dp
         sep%alp = 1.67049799_dp
         sep%gss = 7.50132277_dp
         sep%gsp = 6.34591536_dp
         sep%gpp = 4.77534467_dp
         sep%gp2 = 4.34017279_dp
         sep%hsp = 0.48930466_dp
         sep%fn1(1) = 2.55017735_dp; sep%fn2(1) = 4.29397225_dp; sep%fn3(1) = 0.79989601_dp
         sep%fn1(2) = -0.00565806_dp; sep%fn2(2) = 2.96053910_dp; sep%fn3(2) = 1.47499983_dp
         sep%fn1(3) = -0.00610286_dp; sep%fn2(3) = 2.61416919_dp; sep%fn3(3) = 2.42604040_dp
      CASE (13)
         !Aluminium
         sep%defined = .TRUE.
         sep%alp = 1.9765860_dp
         sep%beta(0) = -3.8668220_dp
         sep%beta(1) = -2.3171460_dp
         sep%sto_exponents(0) = 1.5165930_dp
         sep%sto_exponents(1) = 1.3063470_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -24.3535850_dp
         sep%upp = -18.3636450_dp
         sep%gss = 8.0900000_dp
         sep%gsp = 6.6300000_dp
         sep%gpp = 5.9800000_dp
         sep%gp2 = 5.4000000_dp
         sep%hsp = 0.7000000_dp
         sep%fn1(1) = 0.0900000_dp; sep%fn2(1) = 12.3924430_dp; sep%fn3(1) = 2.0503940_dp
      CASE (14)
         !Silicon
         sep%defined = .TRUE.
         sep%alp = 2.257816_dp
         sep%beta(0) = -3.784852_dp
         sep%beta(1) = -1.968123_dp
         sep%sto_exponents(0) = 1.830697_dp
         sep%sto_exponents(1) = 1.2849530_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -33.9536220_dp
         sep%upp = -28.9347490_dp
         sep%gss = 9.8200000_dp
         sep%gsp = 8.3600000_dp
         sep%gpp = 7.3100000_dp
         sep%gp2 = 6.5400000_dp
         sep%hsp = 1.3200000_dp
         sep%fn1(1) = 0.25000000_dp; sep%fn2(1) = 9.00000000_dp; sep%fn3(1) = 0.91145300_dp
         sep%fn1(2) = 0.06151300_dp; sep%fn2(2) = 5.00000000_dp; sep%fn3(2) = 1.99556900_dp
         sep%fn1(3) = 0.00207890_dp; sep%fn2(3) = 5.00000000_dp; sep%fn3(3) = 2.99061000_dp
      CASE (15)
         !Phosphorus
         sep%defined = .TRUE.
         sep%alp = 2.4553220_dp
         sep%beta(0) = -6.3537640_dp
         sep%beta(1) = -6.5907090_dp
         sep%sto_exponents(0) = 1.9812800_dp
         sep%sto_exponents(1) = 1.8751500_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -42.0298630_dp
         sep%upp = -34.0307090_dp
         sep%gss = 11.5600050_dp
         sep%gsp = 5.2374490_dp
         sep%gpp = 7.8775890_dp
         sep%gp2 = 7.3076480_dp
         sep%hsp = 0.7792380_dp
         sep%fn1(1) = -0.03182700_dp; sep%fn2(1) = 6.00000000_dp; sep%fn3(1) = 1.47432300_dp
         sep%fn1(2) = 0.01847000_dp; sep%fn2(2) = 7.00000000_dp; sep%fn3(2) = 1.77935400_dp
         sep%fn1(3) = 0.03329000_dp; sep%fn2(3) = 9.00000000_dp; sep%fn3(3) = 3.00657600_dp
      CASE (16)
         !Sulphur
         sep%defined = .TRUE.
         sep%alp = 2.4616480_dp
         sep%beta(0) = -3.9205660_dp
         sep%beta(1) = -7.9052780_dp
         sep%sto_exponents(0) = 2.3665150_dp
         sep%sto_exponents(1) = 1.6672630_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -56.6940560_dp
         sep%upp = -48.7170490_dp
         sep%gss = 11.7863290_dp
         sep%gsp = 8.6631270_dp
         sep%gpp = 10.0393080_dp
         sep%gp2 = 7.7816880_dp
         sep%hsp = 2.5321370_dp
         sep%fn1(1) = -0.50919500_dp; sep%fn2(1) = 4.59369100_dp; sep%fn3(1) = 0.77066500_dp
         sep%fn1(2) = -0.01186300_dp; sep%fn2(2) = 5.86573100_dp; sep%fn3(2) = 1.50331300_dp
         sep%fn1(3) = 0.01233400_dp; sep%fn2(3) = 13.55733600_dp; sep%fn3(3) = 2.00917300_dp
      CASE (17)
         !Chlorine
         sep%defined = .TRUE.
         sep%alp = 2.9193680_dp
         sep%beta(0) = -24.5946700_dp
         sep%beta(1) = -14.6372160_dp
         sep%sto_exponents(0) = 3.6313760_dp
         sep%sto_exponents(1) = 2.0767990_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -111.6139480_dp
         sep%upp = -76.6401070_dp
         sep%gss = 15.0300000_dp
         sep%gsp = 13.1600000_dp
         sep%gpp = 11.3000000_dp
         sep%gp2 = 9.9700000_dp
         sep%hsp = 2.4200000_dp
         sep%fn1(1) = 0.0942430_dp; sep%fn2(1) = 4.0000000_dp; sep%fn3(1) = 1.3000000_dp
         sep%fn1(2) = 0.0271680_dp; sep%fn2(2) = 4.0000000_dp; sep%fn3(2) = 2.1000000_dp
      CASE (18:29)
      CASE (30)
         !Zinc
         sep%defined = .TRUE.
         sep%uss = -21.0400080_dp
         sep%upp = -17.6555740_dp
         sep%beta(0) = -1.9974290_dp
         sep%beta(1) = -4.7581190_dp
         sep%sto_exponents(0) = 1.9542990_dp
         sep%sto_exponents(1) = 1.3723650_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 1.4845630_dp
         sep%gss = 11.8000000_dp
         sep%gsp = 11.1820180_dp
         sep%gpp = 13.3000000_dp
         sep%gp2 = 12.9305200_dp
         sep%hsp = 0.4846060_dp
      CASE (31)
      CASE (32)
         !Germanium
         sep%defined = .TRUE.
         sep%uss = -34.1838890_dp
         sep%upp = -28.6408110_dp
         sep%beta(0) = -4.3566070_dp
         sep%beta(1) = -0.9910910_dp
         sep%sto_exponents(0) = 1.2196310_dp
         sep%sto_exponents(1) = 1.9827940_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 2.1364050_dp
         sep%gss = 10.1686050_dp
         sep%gsp = 8.1444730_dp
         sep%gpp = 6.6719020_dp
         sep%gp2 = 6.2697060_dp
         sep%hsp = 0.9370930_dp
      CASE (33:34)
         !see calpar.F90 for additional constraint params
         !                    DATA FOR ELEMENT 33      ARSENIC
         !     data ussam1(33)/  - 41.6817510e0_dp/
         !     data uppam1(33)/  - 33.4506152e0_dp/
         !     data betasa(33)/  - 5.6481504e0_dp/
         !     data betapa(33)/  - 4.9979109e0_dp/
         !     data zsam1(33)/ 2.2576897e0_dp/
         !     data zpam1(33)/ 1.7249710e0_dp/
         !     data alpam1(33)/ 2.2405380e0_dp/
         !     data gssam1(33)/ 11.0962258e0_dp/
         !     data gspam1(33)/ 4.9259328e0_dp/
         !     data gppam1(33)/ 7.8781648e0_dp/
         !     data gp2am1(33)/ 7.5961088e0_dp/
         !     data hspam1(33)/ 0.6246173e0_dp/
         !     data guesa1(33,1)/  - 0.0073614e0_dp/
         !     data guesa2(33,1)/ 4.9433993e0_dp/
         !     data guesa3(33,1)/ 1.4544264e0_dp/
         !     data guesa1(33,2)/ 0.0437629e0_dp/
         !     data guesa2(33,2)/ 3.1944613e0_dp/
         !     data guesa3(33,2)/ 2.0144939e0_dp/
         !                    DATA FOR ELEMENT 34        SELENIUM
         !     data ussam1(34)/  - 41.9984056e0_dp/
         !     data uppam1(34)/  - 32.8575485e0_dp/
         !     data betasa(34)/  - 3.1470826e0_dp/
         !     data betapa(34)/  - 6.1468406e0_dp/
         !     data zsam1(34)/ 2.6841570e0_dp/
         !     data zpam1(34)/ 2.0506164e0_dp/
         !     data alpam1(34)/ 2.6375694e0_dp/
         !     data gssam1(34)/ 6.7908891e0_dp/
         !     data gspam1(34)/ 6.4812786e0_dp/
         !     data gppam1(34)/ 6.4769273e0_dp/
         !     data gp2am1(34)/ 5.2796993e0_dp/
         !     data hspam1(34)/ 4.4548356e0_dp/
         !     data guesa1(34,1)/ 0.1116681e0_dp/
         !     data guesa2(34,1)/ 6.5086644e0_dp/
         !     data guesa3(34,1)/ 1.4981077e0_dp/
         !     data guesa1(34,2)/ 0.0396143e0_dp/
         !     data guesa2(34,2)/ 6.5241228e0_dp/
         !     data guesa3(34,2)/ 2.0751916e0_dp/
         !
      CASE (35)
         !Bromine
         sep%defined = .TRUE.
         sep%alp = 2.5765460_dp
         sep%beta(0) = -19.3998800_dp
         sep%beta(1) = -8.9571950_dp
         sep%sto_exponents(0) = 3.0641330_dp
         sep%sto_exponents(1) = 2.0383330_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -104.6560630_dp
         sep%upp = -74.9300520_dp
         sep%gss = 15.0364395_dp
         sep%gsp = 13.0346824_dp
         sep%gpp = 11.2763254_dp
         sep%gp2 = 9.8544255_dp
         sep%hsp = 2.4558683_dp
         sep%fn1(1) = 0.0666850_dp; sep%fn2(1) = 4.0000000_dp; sep%fn3(1) = 1.5000000_dp
         sep%fn1(2) = 0.0255680_dp; sep%fn2(2) = 4.0000000_dp; sep%fn3(2) = 2.3000000_dp
      CASE (36:49)
      CASE (50)
         !Tin
         sep%defined = .TRUE.
         sep%alp = 1.8369360_dp
         sep%beta(0) = -3.2350000_dp
         sep%beta(1) = -2.5778900_dp
         sep%sto_exponents(0) = 2.5993760_dp
         sep%sto_exponents(1) = 1.6959620_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -35.4967410_dp
         sep%upp = -28.0976360_dp
         sep%gss = 9.8000000_dp
         sep%gsp = 8.3000000_dp
         sep%gpp = 7.3000000_dp
         sep%gp2 = 6.5000000_dp
         sep%hsp = 1.3000000_dp
      CASE (51:52)
         !                    DATA FOR ELEMENT 51        ANTIMONY
         !     data ussam1(51)/  - 44.4381620e0_dp/
         !     data uppam1(51)/  - 32.3895140e0_dp/
         !     data betasa(51)/  - 7.3823300e0_dp/
         !     data betapa(51)/  - 3.6331190e0_dp/
         !     data zsam1(51)/ 2.2548230e0_dp/
         !     data zpam1(51)/ 2.2185920e0_dp/
         !     data alpam1(51)/ 2.2763310e0_dp/
         !     data gssam1(51)/ 11.4302510e0_dp/
         !     data gspam1(51)/ 5.7879220e0_dp/
         !     data gppam1(51)/ 6.4240940e0_dp/
         !     data gp2am1(51)/ 6.8491810e0_dp/
         !     data hspam1(51)/ 0.5883400e0_dp/
         !     data guesa1(51,1)/  - 0.5964470e0_dp/
         !     data guesa2(51,1)/ 6.0279500e0_dp/
         !     data guesa3(51,1)/ 1.7103670e0_dp/
         !     data guesa1(51,2)/ 0.8955130e0_dp/
         !     data guesa2(51,2)/ 3.0281090e0_dp/
         !     data guesa3(51,2)/ 1.5383180e0_dp/
         !                    DATA FOR ELEMENT 52        TELLURIUM
         !     data ussam1(52)/  - 39.2454230e0_dp/
         !     data uppam1(52)/  - 30.8515845e0_dp/
         !     data betasa(52)/  - 8.3897294e0_dp/
         !     data betapa(52)/  - 5.1065429e0_dp/
         !     data zsam1(52)/ 2.1321165e0_dp/
         !     data zpam1(52)/ 1.9712680e0_dp/
         !     data alpam1(52)/ 6.0171167e0_dp/
         !     data gssam1(52)/ 4.9925231e0_dp/
         !     data gspam1(52)/ 4.9721484e0_dp/
         !     data gppam1(52)/ 7.2097852e0_dp/
         !     data gp2am1(52)/ 5.6211521e0_dp/
         !     data hspam1(52)/ 4.0071821e0_dp/
         !     data guesa1(52,1)/ 0.4873378e0_dp/
         !     data guesa2(52,1)/ 6.0519413e0_dp/
         !     data guesa3(52,1)/ 1.3079857e0_dp/
         !     data guesa1(52,2)/ 0.1520464e0_dp/
         !     data guesa2(52,2)/ 3.8304067e0_dp/
         !     data guesa3(52,2)/ 2.0899707e0_dp/
      CASE (53)
         !Iodine
         sep%defined = .TRUE.
         sep%uss = -103.5896630_dp
         sep%upp = -74.4299970_dp
         sep%beta(0) = -8.4433270_dp
         sep%beta(1) = -6.3234050_dp
         sep%sto_exponents(0) = 2.1028580_dp
         sep%sto_exponents(1) = 2.1611530_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 2.2994240_dp
         sep%gss = 15.0404486_dp
         sep%gsp = 13.0565580_dp
         sep%gpp = 11.1477837_dp
         sep%gp2 = 9.9140907_dp
         sep%hsp = 2.4563820_dp
         sep%fn1(1) = 0.0043610_dp; sep%fn2(1) = 2.3000000_dp; sep%fn3(1) = 1.8000000_dp
         sep%fn1(2) = 0.0157060_dp; sep%fn2(2) = 3.0000000_dp; sep%fn3(2) = 2.2400000_dp
      CASE (80)
         !Mercury
         sep%defined = .TRUE.
         sep%uss = -19.9415780_dp
         sep%upp = -11.1108700_dp
         sep%beta(0) = -0.9086570_dp
         sep%beta(1) = -4.9093840_dp
         sep%sto_exponents(0) = 2.0364130_dp
         sep%sto_exponents(1) = 1.9557660_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 1.4847340_dp
         sep%gss = 10.8000000_dp
         sep%gsp = 9.3000000_dp
         sep%gpp = 14.3000000_dp
         sep%gp2 = 13.5000000_dp
         sep%hsp = 1.3000000_dp
      CASE (81)
      CASE (82)
         !Lead
         sep%defined = .TRUE.
         sep%uss = -47.3196920_dp
         sep%upp = -28.8475600_dp
         sep%beta(0) = -8.0423870_dp
         sep%beta(1) = -3.0000000_dp
         sep%sto_exponents(0) = 2.4982860_dp
         sep%sto_exponents(1) = 1.7283330_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 1.7283330_dp
         sep%gss = 9.8000000_dp
         sep%gsp = 8.3000000_dp
         sep%gpp = 7.3000000_dp
         sep%gp2 = 6.5000000_dp
         sep%hsp = 1.3000000_dp
      CASE (83:nelem)
      END SELECT
      ! Finalize parameters for the element
      CALL valence_electrons(sep, extended_basis_set=.FALSE.)
      CALL calpar(z, sep)
      CALL convert_param_to_cp2k(sep)

   END SUBROUTINE am1_default_parameter

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: RM1
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE rm1_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER                                            :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'rm1_default_parameter', &
         routineP = moduleN//':'//routineN

      sep%typ = do_method_rm1
      sep%name = ptable(z)%name
      sep%eheat = ptable(z)%heat_of_formation/kcalmol
      sep%z = z

      SELECT CASE (z)
      CASE DEFAULT
         CALL cp__b("semi_empirical_parameters.F",1525,"invalid atomic number")
      CASE (0)
         sep%defined = .TRUE.
      CASE (1)
         !Hydrogen
         sep%defined = .TRUE.
         sep%alp = 3.06835947_dp
         sep%beta(0) = -5.76544469_dp
         sep%sto_exponents(0) = 1.08267366_dp
         sep%uss = -11.96067697_dp
         sep%gss = 13.98321296_dp
         sep%fn1(1) = 0.10288875_dp; sep%fn2(1) = 5.90172268_dp; sep%fn3(1) = 1.17501185_dp
         sep%fn1(2) = 0.06457449_dp; sep%fn2(2) = 6.41785671_dp; sep%fn3(2) = 1.93844484_dp
         sep%fn1(3) = -0.03567387_dp; sep%fn2(3) = 2.80473127_dp; sep%fn3(3) = 1.63655241_dp
      CASE (2:5)
      CASE (6)
         !Carbon
         sep%defined = .TRUE.
         sep%alp = 2.79282078_dp
         sep%beta(0) = -15.45932428_dp
         sep%beta(1) = -8.23608638_dp
         sep%sto_exponents(0) = 1.85018803_dp
         sep%sto_exponents(1) = 1.76830093_dp
         sep%uss = -51.72556032_dp
         sep%upp = -39.40728943_dp
         sep%gss = 13.05312440_dp
         sep%gsp = 11.33479389_dp
         sep%gpp = 10.95113739_dp
         sep%gp2 = 9.72395099_dp
         sep%hsp = 1.55215133_dp
         sep%fn1(1) = 0.07462271_dp; sep%fn2(1) = 5.73921605_dp; sep%fn3(1) = 1.04396983_dp
         sep%fn1(2) = 0.01177053_dp; sep%fn2(2) = 6.92401726_dp; sep%fn3(2) = 1.66159571_dp
         sep%fn1(3) = 0.03720662_dp; sep%fn2(3) = 6.26158944_dp; sep%fn3(3) = 1.63158721_dp
         sep%fn1(4) = -0.00270657_dp; sep%fn2(4) = 9.00003735_dp; sep%fn3(4) = 2.79557901_dp
      CASE (7)
         !Nitrogen
         sep%defined = .TRUE.
         sep%alp = 2.96422542_dp
         sep%beta(0) = -20.87124548_dp
         sep%beta(1) = -16.67171853_dp
         sep%sto_exponents(0) = 2.37447159_dp
         sep%sto_exponents(1) = 1.97812569_dp
         sep%uss = -70.85123715_dp
         sep%upp = -57.97730920_dp
         sep%gss = 13.08736234_dp
         sep%gsp = 13.21226834_dp
         sep%gpp = 13.69924324_dp
         sep%gp2 = 11.94103953_dp
         sep%hsp = 5.00000846_dp
         sep%fn1(1) = 0.06073380_dp; sep%fn2(1) = 4.58892946_dp; sep%fn3(1) = 1.37873881_dp
         sep%fn1(2) = 0.02438558_dp; sep%fn2(2) = 4.62730519_dp; sep%fn3(2) = 2.08370698_dp
         sep%fn1(3) = -0.02283430_dp; sep%fn2(3) = 2.05274659_dp; sep%fn3(3) = 1.86763816_dp
      CASE (8)
         !Oxygen
         sep%defined = .TRUE.
         sep%alp = 4.17196717_dp
         sep%beta(0) = -29.85101212_dp
         sep%beta(1) = -29.15101314_dp
         sep%sto_exponents(0) = 3.17936914_dp
         sep%sto_exponents(1) = 2.55361907_dp
         sep%uss = -96.94948069_dp
         sep%upp = -77.89092978_dp
         sep%gss = 14.00242788_dp
         sep%gsp = 14.95625043_dp
         sep%gpp = 14.14515138_dp
         sep%gp2 = 12.70325497_dp
         sep%hsp = 3.93217161_dp
         sep%fn1(1) = 0.23093552_dp; sep%fn2(1) = 5.21828736_dp; sep%fn3(1) = 0.90363555_dp
         sep%fn1(2) = 0.05859873_dp; sep%fn2(2) = 7.42932932_dp; sep%fn3(2) = 1.51754610_dp
      CASE (9)
         !Fluorine
         sep%defined = .TRUE.
         sep%alp = 6.00000062_dp
         sep%beta(0) = -70.00000512_dp
         sep%beta(1) = -32.67982711_dp
         sep%sto_exponents(0) = 4.40337913_dp
         sep%sto_exponents(1) = 2.64841556_dp
         sep%uss = -134.18369591_dp
         sep%upp = -107.84660920_dp
         sep%gss = 16.72091319_dp
         sep%gsp = 16.76142629_dp
         sep%gpp = 15.22581028_dp
         sep%gp2 = 14.86578679_dp
         sep%hsp = 1.99766171_dp
         sep%fn1(1) = 0.40302025_dp; sep%fn2(1) = 7.20441959_dp; sep%fn3(1) = 0.81653013_dp
         sep%fn1(2) = 0.07085831_dp; sep%fn2(2) = 9.00001562_dp; sep%fn3(2) = 1.43802381_dp
      CASE (10:14)
      CASE (15)
         !Phosphorus
         sep%defined = .TRUE.
         sep%alp = 1.90993294_dp
         sep%beta(0) = -6.13514969_dp
         sep%beta(1) = -5.94442127_dp
         sep%sto_exponents(0) = 2.12240118_dp
         sep%sto_exponents(1) = 1.74327954_dp
         sep%sto_exponents(2) = 1.00000000_dp
         sep%uss = -41.81533184_dp
         sep%upp = -34.38342529_dp
         sep%gss = 11.08059265_dp
         sep%gsp = 5.68339201_dp
         sep%gpp = 7.60417563_dp
         sep%gp2 = 7.40265182_dp
         sep%hsp = 1.16181792_dp
         sep%fn1(1) = -0.41063467_dp; sep%fn2(1) = 6.08752832_dp; sep%fn3(1) = 1.31650261_dp
         sep%fn1(2) = -0.16299288_dp; sep%fn2(2) = 7.09472602_dp; sep%fn3(2) = 1.90721319_dp
         sep%fn1(3) = -0.04887125_dp; sep%fn2(3) = 8.99979308_dp; sep%fn3(3) = 2.65857780_dp
      CASE (16)
         !Sulphur
         sep%defined = .TRUE.
         sep%alp = 2.44015636_dp
         sep%beta(0) = -1.95910719_dp
         sep%beta(1) = -8.77430652_dp
         sep%sto_exponents(0) = 2.13344308_dp
         sep%sto_exponents(1) = 1.87460650_dp
         sep%sto_exponents(2) = 1.00000000_dp
         sep%uss = -55.16775121_dp
         sep%upp = -46.52930422_dp
         sep%gss = 12.48828408_dp
         sep%gsp = 8.56910574_dp
         sep%gpp = 8.52301167_dp
         sep%gp2 = 7.66863296_dp
         sep%hsp = 3.88978932_dp
         sep%fn1(1) = -0.74601055_dp; sep%fn2(1) = 4.81038002_dp; sep%fn3(1) = 0.59380129_dp
         sep%fn1(2) = -0.06519286_dp; sep%fn2(2) = 7.20760864_dp; sep%fn3(2) = 1.29492008_dp
         sep%fn1(3) = -0.00655977_dp; sep%fn2(3) = 9.00000180_dp; sep%fn3(3) = 1.80060151_dp
      CASE (17)
         !Chlorine
         sep%defined = .TRUE.
         sep%alp = 3.69358828_dp
         sep%beta(0) = -19.92430432_dp
         sep%beta(1) = -11.52935197_dp
         sep%sto_exponents(0) = 3.86491071_dp
         sep%sto_exponents(1) = 1.89593144_dp
         sep%sto_exponents(2) = 1.00000000_dp
         sep%uss = -118.47306918_dp
         sep%upp = -76.35330340_dp
         sep%gss = 15.36023105_dp
         sep%gsp = 13.30671171_dp
         sep%gpp = 12.56502640_dp
         sep%gp2 = 9.66397083_dp
         sep%hsp = 1.76489897_dp
         sep%fn1(1) = 0.12947108_dp; sep%fn2(1) = 2.97724424_dp; sep%fn3(1) = 1.46749784_dp
         sep%fn1(2) = 0.00288899_dp; sep%fn2(2) = 7.09827589_dp; sep%fn3(2) = 2.50002723_dp
      CASE (18:34)
      CASE (35)
         !Bromine
         sep%defined = .TRUE.
         sep%alp = 2.86710532_dp
         sep%beta(0) = -1.34139841_dp
         sep%beta(1) = -8.20225991_dp
         sep%sto_exponents(0) = 5.73157215_dp
         sep%sto_exponents(1) = 2.03147582_dp
         sep%sto_exponents(2) = 1.00000000_dp
         sep%uss = -113.48398183_dp
         sep%upp = -76.18720023_dp
         sep%gss = 17.11563074_dp
         sep%gsp = 15.62419251_dp
         sep%gpp = 10.73546293_dp
         sep%gp2 = 8.86056199_dp
         sep%hsp = 2.23512762_dp
         sep%fn1(1) = 0.98689937_dp; sep%fn2(1) = 4.28484191_dp; sep%fn3(1) = 2.00019696_dp
         sep%fn1(2) = -0.92731247_dp; sep%fn2(2) = 4.54005910_dp; sep%fn3(2) = 2.01617695_dp
      CASE (36:49)
      CASE (50)
      CASE (51:52)
      CASE (53)
         !Iodine
         sep%defined = .TRUE.
         sep%alp = 2.14157092_dp
         sep%beta(0) = -4.19316149_dp
         sep%beta(1) = -4.40038412_dp
         sep%sto_exponents(0) = 2.53003753_dp
         sep%sto_exponents(1) = 2.31738678_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%uss = -74.89997837_dp
         sep%upp = -51.41023805_dp
         sep%gss = 19.99974131_dp
         sep%gsp = 7.68957672_dp
         sep%gpp = 7.30488343_dp
         sep%gp2 = 6.85424614_dp
         sep%hsp = 1.41602940_dp
         sep%fn1(1) = -0.08147724_dp; sep%fn2(1) = 1.56065072_dp; sep%fn3(1) = 2.00002063_dp
         sep%fn1(2) = 0.05914991_dp; sep%fn2(2) = 5.76111270_dp; sep%fn3(2) = 2.20488800_dp
      CASE (80:nelem)
      END SELECT
      ! Finalize parameters for the element
      CALL valence_electrons(sep, extended_basis_set=.FALSE.)
      CALL calpar(z, sep)
      CALL convert_param_to_cp2k(sep)

   END SUBROUTINE rm1_default_parameter

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: PM3
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE pm3_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER                                            :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'pm3_default_parameter', &
         routineP = moduleN//':'//routineN

      sep%typ = do_method_pm3
      sep%name = ptable(z)%name
      sep%eheat = ptable(z)%heat_of_formation/kcalmol
      sep%z = z

      SELECT CASE (z)
      CASE DEFAULT
         CALL cp__b("semi_empirical_parameters.F",1736,"invalid atomic number")
      CASE (0)
         sep%defined = .TRUE.
      CASE (1)
         !Hydrogen
         sep%defined = .TRUE.
         sep%uss = -13.0733210_dp
         sep%beta(0) = -5.6265120_dp
         sep%sto_exponents(0) = 0.9678070_dp
         sep%alp = 3.3563860_dp
         sep%gss = 14.7942080_dp
         sep%fn1(1) = 1.1287500_dp; sep%fn2(1) = 5.0962820_dp; sep%fn3(1) = 1.5374650_dp
         sep%fn1(2) = -1.0603290_dp; sep%fn2(2) = 6.0037880_dp; sep%fn3(2) = 1.5701890_dp
         sep%ass = 1.0e-10_dp
         sep%asp = 3.13469_dp
         sep%app = 0.604262_dp
      CASE (2)
      CASE (3)
         !Lithium
         sep%defined = .TRUE.
         sep%uss = -5.3000000_dp
         sep%upp = -3.4000000_dp
         sep%beta(0) = -0.5500000_dp
         sep%beta(1) = -1.5000000_dp
         sep%sto_exponents(0) = 0.6500000_dp
         sep%sto_exponents(1) = 0.7500000_dp
         sep%alp = 1.2550000_dp
         sep%gss = 4.5000000_dp
         sep%gsp = 3.0000000_dp
         sep%gpp = 5.2500000_dp
         sep%gp2 = 4.5000000_dp
         sep%hsp = 0.1500000_dp
         sep%fn1(1) = -0.4500000_dp; sep%fn2(1) = 5.0000000_dp; sep%fn3(1) = 1.0000000_dp
         sep%fn1(2) = 0.8000000_dp; sep%fn2(2) = 6.5000000_dp; sep%fn3(2) = 1.0000000_dp
      CASE (4)
         !Beryllium
         sep%defined = .TRUE.
         sep%uss = -17.2647520_dp
         sep%upp = -11.3042430_dp
         sep%beta(0) = -3.9620530_dp
         sep%beta(1) = -2.7806840_dp
         sep%sto_exponents(0) = 0.8774390_dp
         sep%sto_exponents(1) = 1.5087550_dp
         sep%alp = 1.5935360_dp
         sep%gss = 9.0128510_dp
         sep%gsp = 6.5761990_dp
         sep%gpp = 6.0571820_dp
         sep%gp2 = 9.0052190_dp
         sep%hsp = 0.5446790_dp
         sep%fn1(1) = 1.6315720_dp; sep%fn2(1) = 2.6729620_dp; sep%fn3(1) = 1.7916860_dp
         sep%fn1(2) = -2.1109590_dp; sep%fn2(2) = 1.9685940_dp; sep%fn3(2) = 1.7558710_dp
      CASE (5)
         !Boron
         sep%defined = .TRUE.
         sep%uss = -28.3539160_dp
         sep%upp = -23.9738780_dp
         sep%beta(0) = -5.5215820_dp
         sep%beta(1) = -5.9611770_dp
         sep%sto_exponents(0) = 0.9885410_dp
         sep%sto_exponents(1) = 1.9094410_dp
         sep%alp = 2.4981070_dp
         sep%gss = 8.8104530_dp
         sep%gsp = 10.2992280_dp
         sep%gpp = 4.9891680_dp
         sep%gp2 = 7.2899770_dp
         sep%hsp = 0.8370320_dp
         sep%fn1(1) = 3.0046640_dp; sep%fn2(1) = 6.0288690_dp; sep%fn3(1) = 0.3472230_dp
         sep%fn1(2) = 0.0201950_dp; sep%fn2(2) = 6.0106580_dp; sep%fn3(2) = 1.2140670_dp
      CASE (6)
         !Carbon
         sep%defined = .TRUE.
         sep%uss = -47.2703200_dp
         sep%upp = -36.2669180_dp
         sep%beta(0) = -11.9100150_dp
         sep%beta(1) = -9.8027550_dp
         sep%sto_exponents(0) = 1.5650850_dp
         sep%sto_exponents(1) = 1.8423450_dp
         sep%alp = 2.7078070_dp
         sep%gss = 11.2007080_dp
         sep%gsp = 10.2650270_dp
         sep%gpp = 10.7962920_dp
         sep%gp2 = 9.0425660_dp
         sep%hsp = 2.2909800_dp
         sep%fn1(1) = 0.0501070_dp; sep%fn2(1) = 6.0031650_dp; sep%fn3(1) = 1.6422140_dp
         sep%fn1(2) = 0.0507330_dp; sep%fn2(2) = 6.0029790_dp; sep%fn3(2) = 0.8924880_dp
      CASE (7)
         !Nitrogen
         sep%defined = .TRUE.
         sep%uss = -49.3356720_dp
         sep%upp = -47.5097360_dp
         sep%beta(0) = -14.0625210_dp
         sep%beta(1) = -20.0438480_dp
         sep%sto_exponents(0) = 2.0280940_dp
         sep%sto_exponents(1) = 2.3137280_dp
         sep%alp = 2.8305450_dp
         sep%gss = 11.9047870_dp
         sep%gsp = 7.3485650_dp
         sep%gpp = 11.7546720_dp
         sep%gp2 = 10.8072770_dp
         sep%hsp = 1.1367130_dp
         sep%fn1(1) = 1.5016740_dp; sep%fn2(1) = 5.9011480_dp; sep%fn3(1) = 1.7107400_dp
         sep%fn1(2) = -1.5057720_dp; sep%fn2(2) = 6.0046580_dp; sep%fn3(2) = 1.7161490_dp
      CASE (8)
         !Oxygen
         sep%defined = .TRUE.
         sep%uss = -86.9930020_dp
         sep%upp = -71.8795800_dp
         sep%beta(0) = -45.2026510_dp
         sep%beta(1) = -24.7525150_dp
         sep%sto_exponents(0) = 3.7965440_dp
         sep%sto_exponents(1) = 2.3894020_dp
         sep%alp = 3.2171020_dp
         sep%gss = 15.7557600_dp
         sep%gsp = 10.6211600_dp
         sep%gpp = 13.6540160_dp
         sep%gp2 = 12.4060950_dp
         sep%hsp = 0.5938830_dp
         sep%fn1(1) = -1.1311280_dp; sep%fn2(1) = 6.0024770_dp; sep%fn3(1) = 1.6073110_dp
         sep%fn1(2) = 1.1378910_dp; sep%fn2(2) = 5.9505120_dp; sep%fn3(2) = 1.5983950_dp
         sep%ass = 1.0e-10_dp
         sep%asp = 6.906169_dp
         sep%app = 0.911779_dp
      CASE (9)
         !Fluorine
         sep%defined = .TRUE.
         sep%uss = -110.4353030_dp
         sep%upp = -105.6850470_dp
         sep%beta(0) = -48.4059390_dp
         sep%beta(1) = -27.7446600_dp
         sep%sto_exponents(0) = 4.7085550_dp
         sep%sto_exponents(1) = 2.4911780_dp
         sep%alp = 3.3589210_dp
         sep%gss = 10.4966670_dp
         sep%gsp = 16.0736890_dp
         sep%gpp = 14.8172560_dp
         sep%gp2 = 14.4183930_dp
         sep%hsp = 0.7277630_dp
         sep%fn1(1) = -0.0121660_dp; sep%fn2(1) = 6.0235740_dp; sep%fn3(1) = 1.8568590_dp
         sep%fn1(2) = -0.0028520_dp; sep%fn2(2) = 6.0037170_dp; sep%fn3(2) = 2.6361580_dp
      CASE (10)
      CASE (11)
         !Sodium
         sep%defined = .TRUE.
         sep%uss = -4.7683450_dp
         sep%upp = -4.7703090_dp
         sep%beta(0) = -0.5354170_dp
         sep%beta(1) = 0.0548760_dp
         sep%sto_exponents(0) = 0.6767940_dp
         sep%sto_exponents(1) = 0.9876530_dp
         sep%alp = 1.6810000_dp
         sep%gss = 5.1531420_dp
         sep%gsp = 3.1778570_dp
         sep%gpp = 3.7335330_dp
         sep%gp2 = 11.5386740_dp
         sep%hsp = 0.5455510_dp
         sep%fn1(1) = 0.4175960_dp; sep%fn2(1) = 8.5969780_dp; sep%fn3(1) = 1.6731780_dp
         sep%fn1(2) = 0.4432600_dp; sep%fn2(2) = 3.3594960_dp; sep%fn3(2) = 2.2315110_dp
      CASE (12)
         !Magnesium
         sep%defined = .TRUE.
         sep%uss = -14.6236880_dp
         sep%upp = -14.1734600_dp
         sep%beta(0) = -2.0716910_dp
         sep%beta(1) = -0.5695810_dp
         sep%sto_exponents(0) = 0.6985520_dp
         sep%sto_exponents(1) = 1.4834530_dp
         sep%alp = 1.3291470_dp
         sep%gss = 6.6943000_dp
         sep%gsp = 6.7939950_dp
         sep%gpp = 6.9104460_dp
         sep%gp2 = 7.0908230_dp
         sep%hsp = 0.5433000_dp
         sep%fn1(1) = 2.1170500_dp; sep%fn2(1) = 6.0094770_dp; sep%fn3(1) = 2.0844060_dp
         sep%fn1(2) = -2.5477670_dp; sep%fn2(2) = 4.3953700_dp; sep%fn3(2) = 2.0636740_dp
      CASE (13)
         !Aluminium
         sep%defined = .TRUE.
         sep%uss = -24.8454040_dp
         sep%upp = -22.2641590_dp
         sep%beta(0) = -0.5943010_dp
         sep%beta(1) = -0.9565500_dp
         sep%sto_exponents(0) = 1.7028880_dp
         sep%sto_exponents(1) = 1.0736290_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 1.5217030_dp
         sep%gss = 5.7767370_dp
         sep%gsp = 11.6598560_dp
         sep%gpp = 6.3477900_dp
         sep%gp2 = 6.1210770_dp
         sep%hsp = 4.0062450_dp
         sep%fn1(1) = -0.4730900_dp; sep%fn2(1) = 1.9158250_dp; sep%fn3(1) = 1.4517280_dp
         sep%fn1(2) = -0.1540510_dp; sep%fn2(2) = 6.0050860_dp; sep%fn3(2) = 2.5199970_dp
      CASE (14)
         !Silicon
         sep%defined = .TRUE.
         sep%uss = -26.7634830_dp
         sep%upp = -22.8136350_dp
         sep%beta(0) = -2.8621450_dp
         sep%beta(1) = -3.9331480_dp
         sep%sto_exponents(0) = 1.6350750_dp
         sep%sto_exponents(1) = 1.3130880_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 2.1358090_dp
         sep%gss = 5.0471960_dp
         sep%gsp = 5.9490570_dp
         sep%gpp = 6.7593670_dp
         sep%gp2 = 5.1612970_dp
         sep%hsp = 0.9198320_dp
         sep%fn1(1) = -0.3906000_dp; sep%fn2(1) = 6.0000540_dp; sep%fn3(1) = 0.6322620_dp
         sep%fn1(2) = 0.0572590_dp; sep%fn2(2) = 6.0071830_dp; sep%fn3(2) = 2.0199870_dp
      CASE (15)
         !Phosphorus
         sep%defined = .TRUE.
         sep%uss = -40.4130960_dp
         sep%upp = -29.5930520_dp
         sep%beta(0) = -12.6158790_dp
         sep%beta(1) = -4.1600400_dp
         sep%sto_exponents(0) = 2.0175630_dp
         sep%sto_exponents(1) = 1.5047320_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 1.9405340_dp
         sep%gss = 7.8016150_dp
         sep%gsp = 5.1869490_dp
         sep%gpp = 6.6184780_dp
         sep%gp2 = 6.0620020_dp
         sep%hsp = 1.5428090_dp
         sep%fn1(1) = -0.6114210_dp; sep%fn2(1) = 1.9972720_dp; sep%fn3(1) = 0.7946240_dp
         sep%fn1(2) = -0.0939350_dp; sep%fn2(2) = 1.9983600_dp; sep%fn3(2) = 1.9106770_dp
      CASE (16)
         !Sulphur
         sep%defined = .TRUE.
         sep%uss = -49.8953710_dp
         sep%upp = -44.3925830_dp
         sep%beta(0) = -8.8274650_dp
         sep%beta(1) = -8.0914150_dp
         sep%sto_exponents(0) = 1.8911850_dp
         sep%sto_exponents(1) = 1.6589720_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 2.2697060_dp
         sep%gss = 8.9646670_dp
         sep%gsp = 6.7859360_dp
         sep%gpp = 9.9681640_dp
         sep%gp2 = 7.9702470_dp
         sep%hsp = 4.0418360_dp
         sep%fn1(1) = -0.3991910_dp; sep%fn2(1) = 6.0006690_dp; sep%fn3(1) = 0.9621230_dp
         sep%fn1(2) = -0.0548990_dp; sep%fn2(2) = 6.0018450_dp; sep%fn3(2) = 1.5799440_dp
      CASE (17)
         !Chlorine
         sep%defined = .TRUE.
         sep%uss = -100.6267470_dp
         sep%upp = -53.6143960_dp
         sep%beta(0) = -27.5285600_dp
         sep%beta(1) = -11.5939220_dp
         sep%sto_exponents(0) = 2.2462100_dp
         sep%sto_exponents(1) = 2.1510100_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 2.5172960_dp
         sep%gss = 16.0136010_dp
         sep%gsp = 8.0481150_dp
         sep%gpp = 7.5222150_dp
         sep%gp2 = 7.5041540_dp
         sep%hsp = 3.4811530_dp
         sep%fn1(1) = -0.1715910_dp; sep%fn2(1) = 6.0008020_dp; sep%fn3(1) = 1.0875020_dp
         sep%fn1(2) = -0.0134580_dp; sep%fn2(2) = 1.9666180_dp; sep%fn3(2) = 2.2928910_dp
      CASE (18)
      CASE (19)
         !Potassium
         sep%defined = .TRUE.
         sep%uss = -4.1079950_dp
         sep%upp = -1.9126360_dp
         sep%beta(0) = -0.7235020_dp
         sep%beta(1) = 0.0109380_dp
         sep%sto_exponents(0) = 0.9158730_dp
         sep%sto_exponents(1) = 1.3430990_dp
         sep%alp = 4.4452360_dp
         sep%gss = 4.0502660_dp
         sep%gsp = 2.9879670_dp
         sep%gpp = 2.9983050_dp
         sep%gp2 = 4.9936140_dp
         sep%hsp = 0.5653070_dp
         sep%fn1(1) = 6.0409170_dp; sep%fn2(1) = 10.0465540_dp; sep%fn3(1) = 1.6005120_dp
         sep%fn1(2) = 0.7242240_dp; sep%fn2(2) = 3.0082750_dp; sep%fn3(2) = 2.4643520_dp
      CASE (20)
         !Calcium
         sep%defined = .TRUE.
         sep%uss = -11.6466770_dp
         sep%upp = -11.7003040_dp
         sep%beta(0) = -1.3663020_dp
         sep%beta(1) = -0.6800840_dp
         sep%sto_exponents(0) = 0.8224530_dp
         sep%sto_exponents(1) = 1.2791170_dp
         sep%alp = 3.2251430_dp
         sep%gss = 5.5661490_dp
         sep%gsp = 15.8559960_dp
         sep%gpp = 6.7057840_dp
         sep%gp2 = 4.8642700_dp
         sep%hsp = 1.0053200_dp
         sep%fn1(1) = 0.7940250_dp; sep%fn2(1) = 3.8548560_dp; sep%fn3(1) = 1.5284410_dp
         sep%fn1(2) = 0.3400220_dp; sep%fn2(2) = 5.7455560_dp; sep%fn3(2) = 2.3571110_dp
      CASE (21:23)
         !                    DATA FOR ELEMENT 22 Titanium (Russell-Saunders only)
         !  data alppm3(22)/3.0683070e0_dp/
         !  data betapp(22)/-0.1000000e0_dp/
         !  data betasp(22)/-0.1000000e0_dp/
         !  data gp2pm3(22)/3.5000000e0_dp/
         !  data gpppm3(22)/5.0000000e0_dp/
         !  data gsppm3(22)/4.1500000e0_dp/
         !  data gsspm3(22)/6.0000000e0_dp/
         !  data hsppm3(22)/1.0000000e0_dp/
         !  data uddpm3(22)/-30.0000000e0_dp/
         !  data upppm3(22)/10.0000000e0_dp/
         !  data usspm3(22)/10.0000000e0_dp/
         !  data zdpm3(22)/2.8845490e0_dp/
         !  data zppm3(22)/1.5000000e0_dp/
         !  data zspm3(22)/1.5000000e0_dp/
      CASE (24)
         !Chromuim
         sep%defined = .TRUE.
         sep%uss = -17.5170270_dp
         sep%upp = -12.5337290_dp
         sep%udd = -44.1249280_dp
         sep%beta(0) = -0.1000000_dp
         sep%beta(1) = -0.1000000_dp
         sep%beta(2) = -8.7766360_dp
         sep%sto_exponents(0) = 1.5000000_dp
         sep%sto_exponents(1) = 1.5000000_dp
         sep%sto_exponents(1) = 2.8845490_dp
         sep%alp = 3.0683070_dp
         sep%gss = 6.0000000_dp
         sep%gsp = 4.1500000_dp
         sep%gpp = 5.0000000_dp
         sep%gp2 = 3.5000000_dp
         sep%hsp = 1.0000000_dp
      CASE (25:29)
      CASE (30)
         !Zinc
         sep%defined = .TRUE.
         sep%uss = -18.5321980_dp
         sep%upp = -11.0474090_dp
         sep%beta(0) = -0.7155780_dp
         sep%beta(1) = -6.3518640_dp
         sep%sto_exponents(0) = 1.8199890_dp
         sep%sto_exponents(1) = 1.5069220_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 1.3501260_dp
         sep%gss = 9.6771960_dp
         sep%gsp = 7.7362040_dp
         sep%gpp = 4.9801740_dp
         sep%gp2 = 4.6696560_dp
         sep%hsp = 0.6004130_dp
         sep%fn1(1) = -0.1112340_dp; sep%fn2(1) = 6.0014780_dp; sep%fn3(1) = 1.5160320_dp
         sep%fn1(2) = -0.1323700_dp; sep%fn2(2) = 1.9958390_dp; sep%fn3(2) = 2.5196420_dp
      CASE (31)
         !Gallium
         sep%defined = .TRUE.
         sep%uss = -29.8555930_dp
         sep%upp = -21.8753710_dp
         sep%beta(0) = -4.9456180_dp
         sep%beta(1) = -0.4070530_dp
         sep%sto_exponents(0) = 1.8470400_dp
         sep%sto_exponents(1) = 0.8394110_dp
         sep%alp = 1.6051150_dp
         sep%gss = 8.4585540_dp
         sep%gsp = 8.9256190_dp
         sep%gpp = 5.0868550_dp
         sep%gp2 = 4.9830450_dp
         sep%hsp = 2.0512600_dp
         sep%fn1(1) = -0.5601790_dp; sep%fn2(1) = 5.6232730_dp; sep%fn3(1) = 1.5317800_dp
         sep%fn1(2) = -0.2727310_dp; sep%fn2(2) = 1.9918430_dp; sep%fn3(2) = 2.1838640_dp
      CASE (32)
         !Germanium
         sep%defined = .TRUE.
         sep%uss = -35.4671955_dp
         sep%upp = -31.5863583_dp
         sep%beta(0) = -5.3250024_dp
         sep%beta(1) = -2.2501567_dp
         sep%sto_exponents(0) = 2.2373526_dp
         sep%sto_exponents(1) = 1.5924319_dp
         sep%alp = 1.9723370_dp
         sep%gss = 5.3769635_dp
         sep%gsp = 10.2095293_dp
         sep%gpp = 7.6718647_dp
         sep%gp2 = 6.9242663_dp
         sep%hsp = 1.3370204_dp
         sep%fn1(1) = 0.9631726_dp; sep%fn2(1) = 6.0120134_dp; sep%fn3(1) = 2.1633655_dp
         sep%fn1(2) = -0.9593891_dp; sep%fn2(2) = 5.7491802_dp; sep%fn3(2) = 2.1693724_dp
      CASE (33)
         !Arsenic
         sep%defined = .TRUE.
         sep%uss = -38.5074240_dp
         sep%upp = -35.1524150_dp
         sep%beta(0) = -8.2321650_dp
         sep%beta(1) = -5.0173860_dp
         sep%sto_exponents(0) = 2.6361770_dp
         sep%sto_exponents(1) = 1.7038890_dp
         sep%alp = 1.7944770_dp
         sep%gss = 8.7890010_dp
         sep%gsp = 5.3979830_dp
         sep%gpp = 8.2872500_dp
         sep%gp2 = 8.2103460_dp
         sep%hsp = 1.9510340_dp
         sep%fn1(1) = -0.4600950_dp; sep%fn2(1) = 1.9831150_dp; sep%fn3(1) = 1.0867930_dp
         sep%fn1(2) = -0.0889960_dp; sep%fn2(2) = 1.9929440_dp; sep%fn3(2) = 2.1400580_dp
      CASE (34)
         !Selenium
         sep%defined = .TRUE.
         sep%uss = -55.3781350_dp
         sep%upp = -49.8230760_dp
         sep%beta(0) = -6.1578220_dp
         sep%beta(1) = -5.4930390_dp
         sep%sto_exponents(0) = 2.8280510_dp
         sep%sto_exponents(1) = 1.7325360_dp
         sep%alp = 3.0439570_dp
         sep%gss = 7.4325910_dp
         sep%gsp = 10.0604610_dp
         sep%gpp = 9.5683260_dp
         sep%gp2 = 7.7242890_dp
         sep%hsp = 4.0165580_dp
         sep%fn1(1) = 0.0478730_dp; sep%fn2(1) = 6.0074000_dp; sep%fn3(1) = 2.0817170_dp
         sep%fn1(2) = 0.1147200_dp; sep%fn2(2) = 6.0086720_dp; sep%fn3(2) = 1.5164230_dp
      CASE (35)
         !Bromine
         sep%defined = .TRUE.
         sep%uss = -116.6193110_dp
         sep%upp = -74.2271290_dp
         sep%beta(0) = -31.1713420_dp
         sep%beta(1) = -6.8140130_dp
         sep%sto_exponents(0) = 5.3484570_dp
         sep%sto_exponents(1) = 2.1275900_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 2.5118420_dp
         sep%gss = 15.9434250_dp
         sep%gsp = 16.0616800_dp
         sep%gpp = 8.2827630_dp
         sep%gp2 = 7.8168490_dp
         sep%hsp = 0.5788690_dp
         sep%fn1(1) = 0.9604580_dp; sep%fn2(1) = 5.9765080_dp; sep%fn3(1) = 2.3216540_dp
         sep%fn1(2) = -0.9549160_dp; sep%fn2(2) = 5.9447030_dp; sep%fn3(2) = 2.3281420_dp
      CASE (36)
      CASE (37)
         !     DATA USSPM3( 37)/      -4.4841970e0_dp/
         !     DATA UPPPM3( 37)/      -3.0722520e0_dp/
         !     DATA BETASP( 37)/      -0.7442240e0_dp/
         !     DATA BETAPP( 37)/      -0.5677060e0_dp/
         !     DATA ZSPM3 ( 37)/       0.6801340e0_dp/
         !     DATA ZPPM3 ( 37)/       0.9658140e0_dp/
         !     DATA ALPPM3( 37)/       5.7668440e0_dp/
         !     DATA EISOLP( 37)/      -4.4841970e0_dp/
         !     DATA GSSPM3( 37)/       4.9848710e0_dp/
         !     DATA GSPPM3( 37)/       4.9983590e0_dp/
         !     DATA GPPPM3( 37)/       6.9857170e0_dp/
         !     DATA GP2PM3( 37)/       5.5079160e0_dp/
         !     DATA HSPPM3( 37)/       0.6211430e0_dp/
         !     DATA DDPM3 ( 37)/       3.2610134e0_dp/
         !     DATA QQPM3 ( 37)/       2.6599806e0_dp/
         !     DATA AMPM3 ( 37)/       0.1832000e0_dp/
         !     DATA ADPM3 ( 37)/       0.1544789e0_dp/
         !     DATA AQPM3 ( 37)/       0.3073402e0_dp/
         !     DATA GuesP1( 37,1)/       4.7724480e0_dp/
         !     DATA GuesP2( 37,1)/       5.8039230e0_dp/
         !     DATA GuesP3( 37,1)/       1.4442230e0_dp/
         !     DATA GuesP1( 37,2)/       0.5062010e0_dp/
         !     DATA GuesP2( 37,2)/       2.8303820e0_dp/
         !     DATA GuesP3( 37,2)/       2.6433260e0_dp/
         !     Data (GuesP1(37,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP2(37,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP3(37,J),J=3,10)/8*0.0e0_dp/
      CASE (38)
         !     DATA USSPM3( 38)/     -11.4870450e0_dp/
         !     DATA UPPPM3( 38)/      -7.7152160e0_dp/
         !     DATA BETASP( 38)/      -0.7533420e0_dp/
         !     DATA BETAPP( 38)/      -0.5760720e0_dp/
         !     DATA ZSPM3 ( 38)/       1.5592340e0_dp/
         !     DATA ZPPM3 ( 38)/       1.2129820e0_dp/
         !     DATA ALPPM3( 38)/       4.5905860e0_dp/
         !     DATA EISOLP( 38)/     -16.9990050e0_dp/
         !     DATA GSSPM3( 38)/       5.9750850e0_dp/
         !     DATA GSPPM3( 38)/       4.8869600e0_dp/
         !     DATA GPPPM3( 38)/       7.3407630e0_dp/
         !     DATA GP2PM3( 38)/      15.8140850e0_dp/
         !     DATA HSPPM3( 38)/       4.0512180e0_dp/
         !     DATA DDPM3 ( 38)/       2.1011077e0_dp/
         !     DATA QQPM3 ( 38)/       2.1179593e0_dp/
         !     DATA AMPM3 ( 38)/       0.2195915e0_dp/
         !     DATA ADPM3 ( 38)/       0.5137005e0_dp/
         !     DATA AQPM3 ( 38)/       0.1752812e0_dp/
         !     DATA GuesP1( 38,1)/       0.4398870e0_dp/
         !     DATA GuesP2( 38,1)/       5.7510950e0_dp/
         !     DATA GuesP3( 38,1)/       1.6731940e0_dp/
         !     DATA GuesP1( 38,2)/       0.3957700e0_dp/
         !     DATA GuesP2( 38,2)/       4.4993190e0_dp/
         !     DATA GuesP3( 38,2)/       2.3281550e0_dp/
         !     Data (GuesP1(38,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP2(38,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP3(38,J),J=3,10)/8*0.0e0_dp/
      CASE (39:47)
      CASE (48)
         !Cadmium
         sep%defined = .TRUE.
         sep%uss = -15.8285840_dp
         sep%upp = 8.7497950_dp
         sep%beta(0) = -8.5819440_dp
         sep%beta(1) = -0.6010340_dp
         sep%sto_exponents(0) = 1.6793510_dp
         sep%sto_exponents(1) = 2.0664120_dp
         sep%alp = 1.5253820_dp
         sep%gss = 9.2069600_dp
         sep%gsp = 8.2315390_dp
         sep%gpp = 4.9481040_dp
         sep%gp2 = 4.6696560_dp
         sep%hsp = 1.6562340_dp
      CASE (49)
         !Indium
         sep%defined = .TRUE.
         sep%uss = -26.1762050_dp
         sep%upp = -20.0058220_dp
         sep%beta(0) = -2.9933190_dp
         sep%beta(1) = -1.8289080_dp
         sep%sto_exponents(0) = 2.0161160_dp
         sep%sto_exponents(1) = 1.4453500_dp
         sep%alp = 1.4183850_dp
         sep%gss = 6.5549000_dp
         sep%gsp = 8.2298730_dp
         sep%gpp = 6.2992690_dp
         sep%gp2 = 4.9842110_dp
         sep%hsp = 2.6314610_dp
         sep%fn1(1) = -0.3431380_dp; sep%fn2(1) = 1.9940340_dp; sep%fn3(1) = 1.6255160_dp
         sep%fn1(2) = -0.1095320_dp; sep%fn2(2) = 5.6832170_dp; sep%fn3(2) = 2.8670090_dp
      CASE (50)
         !Tin
         sep%defined = .TRUE.
         sep%uss = -34.5501920_dp
         sep%upp = -25.8944190_dp
         sep%beta(0) = -2.7858020_dp
         sep%beta(1) = -2.0059990_dp
         sep%sto_exponents(0) = 2.3733280_dp
         sep%sto_exponents(1) = 1.6382330_dp
         sep%alp = 1.6996500_dp
         sep%gss = 10.1900330_dp
         sep%gsp = 7.2353270_dp
         sep%gpp = 5.6738100_dp
         sep%gp2 = 5.1822140_dp
         sep%hsp = 1.0331570_dp
         sep%fn1(1) = -0.1503530_dp; sep%fn2(1) = 6.0056940_dp; sep%fn3(1) = 1.7046420_dp
         sep%fn1(2) = -0.0444170_dp; sep%fn2(2) = 2.2573810_dp; sep%fn3(2) = 2.4698690_dp
      CASE (51)
         !Antimony
         sep%defined = .TRUE.
         sep%uss = -56.4321960_dp
         sep%upp = -29.4349540_dp
         sep%beta(0) = -14.7942170_dp
         sep%beta(1) = -2.8179480_dp
         sep%sto_exponents(0) = 2.3430390_dp
         sep%sto_exponents(1) = 1.8999920_dp
         sep%alp = 2.0343010_dp
         sep%gss = 9.2382770_dp
         sep%gsp = 5.2776800_dp
         sep%gpp = 6.3500000_dp
         sep%gp2 = 6.2500000_dp
         sep%hsp = 2.4244640_dp
         sep%fn1(1) = 3.0020280_dp; sep%fn2(1) = 6.0053420_dp; sep%fn3(1) = 0.8530600_dp
         sep%fn1(2) = -0.0188920_dp; sep%fn2(2) = 6.0114780_dp; sep%fn3(2) = 2.7933110_dp
      CASE (52)
         !Tellurium
         sep%defined = .TRUE.
         sep%uss = -44.9380360_dp
         sep%upp = -46.3140990_dp
         sep%beta(0) = -2.6651460_dp
         sep%beta(1) = -3.8954300_dp
         sep%sto_exponents(0) = 4.1654920_dp
         sep%sto_exponents(1) = 1.6475550_dp
         sep%alp = 2.4850190_dp
         sep%gss = 10.2550730_dp
         sep%gsp = 8.1691450_dp
         sep%gpp = 7.7775920_dp
         sep%gp2 = 7.7551210_dp
         sep%hsp = 3.7724620_dp
         sep%fn1(1) = 0.0333910_dp; sep%fn2(1) = 5.9563790_dp; sep%fn3(1) = 2.2775750_dp
         sep%fn1(2) = -1.9218670_dp; sep%fn2(2) = 4.9732190_dp; sep%fn3(2) = 0.5242430_dp
      CASE (53)
         !Iodine
         sep%defined = .TRUE.
         sep%uss = -96.4540370_dp
         sep%upp = -61.0915820_dp
         sep%beta(0) = -14.4942340_dp
         sep%beta(1) = -5.8947030_dp
         sep%sto_exponents(0) = 7.0010130_dp
         sep%sto_exponents(1) = 2.4543540_dp
         sep%sto_exponents(2) = 1.0000000_dp
         sep%alp = 1.9901850_dp
         sep%gss = 13.6319430_dp
         sep%gsp = 14.9904060_dp
         sep%gpp = 7.2883300_dp
         sep%gp2 = 5.9664070_dp
         sep%hsp = 2.6300350_dp
         sep%fn1(1) = -0.1314810_dp; sep%fn2(1) = 5.2064170_dp; sep%fn3(1) = 1.7488240_dp
         sep%fn1(2) = -0.0368970_dp; sep%fn2(2) = 6.0101170_dp; sep%fn3(2) = 2.7103730_dp
      CASE (54)
      CASE (55)
         !     DATA USSPM3( 55)/      -4.3207390e0_dp/
         !     DATA UPPPM3( 55)/      -3.3475800e0_dp/
         !     DATA BETASP( 55)/      -0.9775130e0_dp/
         !     DATA BETAPP( 55)/      -0.5484990e0_dp/
         !     DATA ZSPM3 ( 55)/       0.9188020e0_dp/
         !     DATA ZPPM3 ( 55)/       1.5522530e0_dp/
         !     DATA ALPPM3( 55)/       2.5439870e0_dp/
         !     DATA EISOLP( 55)/      -4.3207390e0_dp/
         !     DATA GSSPM3( 55)/       4.9615320e0_dp/
         !     DATA GSPPM3( 55)/       4.9943210e0_dp/
         !     DATA GPPPM3( 55)/       5.0115170e0_dp/
         !     DATA GP2PM3( 55)/      16.0009950e0_dp/
         !     DATA HSPPM3( 55)/       0.5830050e0_dp/
         !     DATA DDPM3 ( 55)/       1.9526201e0_dp/
         !     DATA QQPM3 ( 55)/       1.9433821e0_dp/
         !     DATA AMPM3 ( 55)/       0.1823422e0_dp/
         !     DATA ADPM3 ( 55)/       0.2009234e0_dp/
         !     DATA AQPM3 ( 55)/       0.1864057e0_dp/
         !     DATA GuesP1( 55,1)/       3.3960620e0_dp/
         !     DATA GuesP2( 55,1)/       6.0007540e0_dp/
         !     DATA GuesP3( 55,1)/       1.6762910e0_dp/
         !     DATA GuesP1( 55,2)/       0.6677680e0_dp/
         !     DATA GuesP2( 55,2)/       3.0084010e0_dp/
         !     DATA GuesP3( 55,2)/       2.7188720e0_dp/
         !     Data (GuesP1(55,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP2(55,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP3(55,J),J=3,10)/8*0.0e0_dp/
      CASE (56)
         !     DATA USSPM3( 56)/     -13.5286010e0_dp/
         !     DATA UPPPM3( 56)/     -10.1346170e0_dp/
         !     DATA BETASP( 56)/      -3.9502870e0_dp/
         !     DATA BETAPP( 56)/      -0.5749730e0_dp/
         !     DATA ZSPM3 ( 56)/       1.7594520e0_dp/
         !     DATA ZPPM3 ( 56)/       1.4916880e0_dp/
         !     DATA ALPPM3( 56)/       1.6584300e0_dp/
         !     DATA EISOLP( 56)/     -16.9427990e0_dp/
         !     DATA GSSPM3( 56)/      10.1144030e0_dp/
         !     DATA GSPPM3( 56)/       4.8886780e0_dp/
         !     DATA GPPPM3( 56)/       8.8650930e0_dp/
         !     DATA GP2PM3( 56)/       8.7503680e0_dp/
         !     DATA HSPPM3( 56)/       1.4151370e0_dp/
         !     DATA DDPM3 ( 56)/       2.2086835e0_dp/
         !     DATA QQPM3 ( 56)/       2.0222866e0_dp/
         !     DATA AMPM3 ( 56)/       0.3717164e0_dp/
         !     DATA ADPM3 ( 56)/       0.2800750e0_dp/
         !     DATA AQPM3 ( 56)/       0.1811672e0_dp/
         !     DATA GuesP1( 56,1)/       0.3501710e0_dp/
         !     DATA GuesP2( 56,1)/       5.8729050e0_dp/
         !     DATA GuesP3( 56,1)/       1.9750730e0_dp/
         !     DATA GuesP1( 56,2)/       0.3339000e0_dp/
         !     DATA GuesP2( 56,2)/       4.0328190e0_dp/
         !     DATA GuesP3( 56,2)/       2.6276060e0_dp/
         !     Data (GuesP1(56,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP2(56,J),J=3,10)/8*0.0e0_dp/
         !     Data (GuesP3(56,J),J=3,10)/8*0.0e0_dp/
      CASE (57:79)
      CASE (80)
         !Mercury
         sep%defined = .TRUE.
         sep%uss = -17.7622290_dp
         sep%upp = -18.3307510_dp
         sep%beta(0) = -3.1013650_dp
         sep%beta(1) = -3.4640310_dp
         sep%sto_exponents(0) = 1.4768850_dp
         sep%sto_exponents(1) = 2.4799510_dp
         sep%alp = 1.5293770_dp
         sep%gss = 6.6247200_dp
         sep%gsp = 10.6392970_dp
         sep%gpp = 14.7092830_dp
         sep%gp2 = 16.0007400_dp
         sep%hsp = 2.0363110_dp
         sep%fn1(1) = 1.0827200_dp; sep%fn2(1) = 6.4965980_dp; sep%fn3(1) = 1.1951460_dp
         sep%fn1(2) = -0.0965530_dp; sep%fn2(2) = 3.9262810_dp; sep%fn3(2) = 2.6271600_dp
      CASE (81)
         !Thallium
         sep%defined = .TRUE.
         sep%uss = -30.0531700_dp
         sep%upp = -26.9206370_dp
         sep%beta(0) = -1.0844950_dp
         sep%beta(1) = -7.9467990_dp
         sep%sto_exponents(0) = 6.8679210_dp
         sep%sto_exponents(1) = 1.9694450_dp
         sep%alp = 1.3409510_dp
         sep%gss = 10.4604120_dp
         sep%gsp = 11.2238830_dp
         sep%gpp = 4.9927850_dp
         sep%gp2 = 8.9627270_dp
         sep%hsp = 2.5304060_dp
         sep%fn1(1) = -1.3613990_dp; sep%fn2(1) = 3.5572260_dp; sep%fn3(1) = 1.0928020_dp
         sep%fn1(2) = -0.0454010_dp; sep%fn2(2) = 2.3069950_dp; sep%fn3(2) = 2.9650290_dp
      CASE (82)
         !Lead
         sep%defined = .TRUE.
         sep%uss = -30.3227560_dp
         sep%upp = -24.4258340_dp
         sep%beta(0) = -6.1260240_dp
         sep%beta(1) = -1.3954300_dp
         sep%sto_exponents(0) = 3.1412890_dp
         sep%sto_exponents(1) = 1.8924180_dp
         sep%alp = 1.6200450_dp
         sep%gss = 7.0119920_dp
         sep%gsp = 6.7937820_dp
         sep%gpp = 5.1837800_dp
         sep%gp2 = 5.0456510_dp
         sep%hsp = 1.5663020_dp
         sep%fn1(1) = -0.1225760_dp; sep%fn2(1) = 6.0030620_dp; sep%fn3(1) = 1.9015970_dp
         sep%fn1(2) = -0.0566480_dp; sep%fn2(2) = 4.7437050_dp; sep%fn3(2) = 2.8618790_dp
      CASE (83)
         !Bismuth
         sep%defined = .TRUE.
         sep%uss = -33.4959380_dp
         sep%upp = -35.5210260_dp
         sep%beta(0) = -5.6072830_dp
         sep%beta(1) = -5.8001520_dp
         sep%sto_exponents(0) = 4.9164510_dp
         sep%sto_exponents(1) = 1.9349350_dp
         sep%alp = 1.8574310_dp
         sep%gss = 4.9894800_dp
         sep%gsp = 6.1033080_dp
         sep%gpp = 8.6960070_dp
         sep%gp2 = 8.3354470_dp
         sep%hsp = 0.5991220_dp
         sep%fn1(1) = 2.5816930_dp; sep%fn2(1) = 5.0940220_dp; sep%fn3(1) = 0.4997870_dp
         sep%fn1(2) = 0.0603200_dp; sep%fn2(2) = 6.0015380_dp; sep%fn3(2) = 2.4279700_dp
      CASE (84:89)
      CASE (90)
         !     DATA USSPM3( 90)/     -40.5682920e0_dp/
         !     DATA UPPPM3( 90)/     -28.0891870e0_dp/
         !     DATA BETASP( 90)/      -4.2562180e0_dp/
         !     DATA BETAPP( 90)/      -4.2562180e0_dp/
         !     DATA ZSPM3 ( 90)/       1.4353060e0_dp/
         !     DATA ZPPM3 ( 90)/       1.4353060e0_dp/
         !     DATA ZDPM3 ( 90)/       1.0000000e0_dp/
         !     DATA ALPPM3( 90)/       2.1961078e0_dp/
         !     DATA EISOLP( 90)/       0.0000000e0_dp/
         !     DATA GSSPM3( 90)/       9.8200000e0_dp/
         !     DATA GSPPM3( 90)/       8.3600000e0_dp/
         !     DATA GPPPM3( 90)/       7.3100000e0_dp/
         !     DATA GP2PM3( 90)/       6.5400000e0_dp/
         !     DATA HSPPM3( 90)/       1.3200000e0_dp/
         !     DATA DDPM3 ( 90)/       0.2011245e0_dp/
         !     DATA QQPM3 ( 90)/       0.2203208e0_dp/
         !     DATA AMPM3 ( 90)/       0.3608967e0_dp/
         !     DATA ADPM3 ( 90)/       1.1124287e0_dp/
         !     DATA AQPM3 ( 90)/       1.3153602e0_dp/
         !     Data (GuesP1(90,J),J=1,10)/10*0.0e0_dp/
         !     Data (GuesP2(90,J),J=1,10)/10*0.0e0_dp/
         !     Data (GuesP3(90,J),J=1,10)/10*0.0e0_dp/
      CASE (91)
         !     DATA USSPM3( 91)/     -75.2391520e0_dp/
         !     DATA UPPPM3( 91)/     -57.8320130e0_dp/
         !     DATA BETASP( 91)/     -11.1422310e0_dp/
         !     DATA BETAPP( 91)/     -11.1422310e0_dp/
         !     DATA ZSPM3 ( 91)/       2.6135910e0_dp/
         !     DATA ZPPM3 ( 91)/       2.0343930e0_dp/
         !     DATA ZDPM3 ( 91)/       1.0000000e0_dp/
         !     DATA ALPPM3( 91)/       2.4916445e0_dp/
         !     DATA EISOLP( 91)/       0.0000000e0_dp/
         !     DATA HSPPM3( 91)/       0.0000001e0_dp/
         !     DATA DDPM3 ( 91)/       0.1232470e0_dp/
         !     DATA QQPM3 ( 91)/       0.1554408e0_dp/
         !     DATA AMPM3 ( 91)/       0.0000000e0_dp/
         !     DATA ADPM3 ( 91)/       0.0062312e0_dp/
         !     DATA AQPM3 ( 91)/       1.2399130e0_dp/
         !     Data (GuesP1(91,J),J=1,10)/10*0.0e0_dp/
         !     Data (GuesP2(91,J),J=1,10)/10*0.0e0_dp/
         !     Data (GuesP3(91,J),J=1,10)/10*0.0e0_dp/
      CASE (92:nelem)
      END SELECT
      ! Finalize parameters for the element
      CALL valence_electrons(sep, extended_basis_set=.FALSE.)
      CALL calpar(z, sep)
      CALL convert_param_to_cp2k(sep)

   END SUBROUTINE pm3_default_parameter

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: PM6
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE pm6_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER                                            :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'pm6_default_parameter', &
         routineP = moduleN//':'//routineN

      sep%typ = do_method_pm6
      sep%name = ptable(z)%name
      sep%eheat = ptable(z)%heat_of_formation/kcalmol
      sep%z = z

      SELECT CASE (z)
      CASE DEFAULT
         CALL cp__b("semi_empirical_parameters.F",2530,"invalid atomic number")
      CASE (0)
         sep%defined = .TRUE.
      CASE (1)
         ! Hydrogen
         sep%defined = .TRUE.
         sep%uss = -11.246958_dp
         sep%beta(0) = -8.352984_dp
         sep%sto_exponents(0) = 1.268641_dp
         sep%gss = 14.448686_dp
         sep%a = 0.024184_dp
         sep%b = 3.055953_dp
         sep%c = 1.786011_dp
      CASE (2)
         ! Helium
         sep%defined = .TRUE.
         sep%uss = -31.770969_dp
         sep%upp = -5.856352_dp
         sep%beta(0) = -58.903774_dp
         sep%beta(1) = -37.039974_dp
         sep%sto_exponents(0) = 3.313204_dp
         sep%sto_exponents(1) = 3.657133_dp
         sep%gss = 9.445299_dp
         sep%gsp = 11.201419_dp
         sep%gpp = 9.214548_dp
         sep%gp2 = 13.046115_dp
         sep%hsp = 0.299954_dp
      CASE (3)
         ! Lithium
         sep%defined = .TRUE.
         sep%uss = -4.709912_dp
         sep%upp = -2.722581_dp
         sep%beta(0) = -2.283946_dp
         sep%beta(1) = -7.535573_dp
         sep%sto_exponents(0) = 0.981041_dp
         sep%sto_exponents(1) = 2.953445_dp
         sep%gss = 11.035907_dp
         sep%gsp = 19.998647_dp
         sep%gpp = 11.543650_dp
         sep%gp2 = 9.059036_dp
         sep%hsp = 1.641886_dp
      CASE (4)
         ! Beryllium
         sep%defined = .TRUE.
         sep%uss = -16.360315_dp
         sep%upp = -16.339216_dp
         sep%beta(0) = -3.199549_dp
         sep%beta(1) = -4.451920_dp
         sep%sto_exponents(0) = 1.212539_dp
         sep%sto_exponents(1) = 1.276487_dp
         sep%gss = 7.552804_dp
         sep%gsp = 10.203146_dp
         sep%gpp = 12.862153_dp
         sep%gp2 = 13.602858_dp
         sep%hsp = 1.501452_dp
         sep%a = 0.164180_dp
         sep%b = 1.704828_dp
         sep%c = 1.785591_dp
      CASE (5)
         ! Boron
         sep%defined = .TRUE.
         sep%uss = -25.967679_dp
         sep%upp = -19.115864_dp
         sep%beta(0) = -4.959706_dp
         sep%beta(1) = -4.656753_dp
         sep%sto_exponents(0) = 1.634174_dp
         sep%sto_exponents(1) = 1.479195_dp
         sep%gss = 8.179341_dp
         sep%gsp = 7.294021_dp
         sep%gpp = 7.829395_dp
         sep%gp2 = 6.401072_dp
         sep%hsp = 1.252845_dp
      CASE (6)
         ! Carbon
         sep%defined = .TRUE.
         sep%uss = -51.089653_dp
         sep%upp = -39.937920_dp
         sep%beta(0) = -15.385236_dp
         sep%beta(1) = -7.471929_dp
         sep%sto_exponents(0) = 2.047558_dp
         sep%sto_exponents(1) = 1.702841_dp
         sep%gss = 13.335519_dp
         sep%gsp = 11.528134_dp
         sep%gpp = 10.778326_dp
         sep%gp2 = 9.486212_dp
         sep%hsp = 0.717322_dp
         sep%a = 0.046302_dp
         sep%b = 2.100206_dp
         sep%c = 1.333959_dp
      CASE (7)
         ! Nitrogen
         sep%defined = .TRUE.
         sep%uss = -57.784823_dp
         sep%upp = -49.893036_dp
         sep%beta(0) = -17.979377_dp
         sep%beta(1) = -15.055017_dp
         sep%sto_exponents(0) = 2.380406_dp
         sep%sto_exponents(1) = 1.999246_dp
         sep%gss = 12.357026_dp
         sep%gsp = 9.636190_dp
         sep%gpp = 12.570756_dp
         sep%gp2 = 10.576425_dp
         sep%hsp = 2.871545_dp
         sep%a = -0.001436_dp
         sep%b = 0.495196_dp
         sep%c = 1.704857_dp
      CASE (8)
         ! Oxygen
         sep%defined = .TRUE.
         sep%uss = -91.678761_dp
         sep%upp = -70.460949_dp
         sep%beta(0) = -65.635137_dp
         sep%beta(1) = -21.622604_dp
         sep%sto_exponents(0) = 5.421751_dp
         sep%sto_exponents(1) = 2.270960_dp
         sep%gss = 11.304042_dp
         sep%gsp = 15.807424_dp
         sep%gpp = 13.618205_dp
         sep%gp2 = 10.332765_dp
         sep%hsp = 5.010801_dp
         sep%a = -0.017771_dp
         sep%b = 3.058310_dp
         sep%c = 1.896435_dp
      CASE (9)
         ! Fluorine
         sep%defined = .TRUE.
         sep%uss = -140.225626_dp
         sep%upp = -98.778044_dp
         sep%beta(0) = -69.922593_dp
         sep%beta(1) = -30.448165_dp
         sep%sto_exponents(0) = 6.043849_dp
         sep%sto_exponents(1) = 2.906722_dp
         sep%gss = 12.446818_dp
         sep%gsp = 18.496082_dp
         sep%gpp = 8.417366_dp
         sep%gp2 = 12.179816_dp
         sep%hsp = 2.604382_dp
         sep%a = -0.010792_dp
         sep%b = 6.004648_dp
         sep%c = 1.847724_dp
      CASE (10)
         ! Neon
         sep%defined = .TRUE.
         sep%uss = -2.978729_dp
         sep%upp = -85.441118_dp
         sep%beta(0) = -69.793475_dp
         sep%beta(1) = -33.261962_dp
         sep%sto_exponents(0) = 6.000148_dp
         sep%sto_exponents(1) = 3.834528_dp
         sep%gss = 19.999574_dp
         sep%gsp = 16.896951_dp
         sep%gpp = 8.963560_dp
         sep%gp2 = 16.027799_dp
         sep%hsp = 1.779280_dp
      CASE (11)
         ! Sodium
         sep%defined = .TRUE.
         sep%uss = -4.537153_dp
         sep%upp = -2.433015_dp
         sep%beta(0) = 0.244853_dp
         sep%beta(1) = 0.491998_dp
         sep%sto_exponents(0) = 0.686327_dp
         sep%sto_exponents(1) = 0.950068_dp
         sep%gss = 4.059972_dp
         sep%gsp = 7.061183_dp
         sep%gpp = 9.283540_dp
         sep%gp2 = 17.034978_dp
         sep%hsp = 0.640715_dp
         sep%a = -1.026036_dp
         sep%b = 2.014506_dp
         sep%c = 1.271202_dp
      CASE (12)
         ! Magnesium
         sep%defined = .TRUE.
         sep%uss = -14.574226_dp
         sep%upp = -7.583850_dp
         sep%beta(0) = -9.604932_dp
         sep%beta(1) = 3.416908_dp
         sep%sto_exponents(0) = 1.310830_dp
         sep%sto_exponents(1) = 1.388897_dp
         sep%gss = 7.115328_dp
         sep%gsp = 3.253024_dp
         sep%gpp = 4.737311_dp
         sep%gp2 = 8.428485_dp
         sep%hsp = 0.877379_dp
      CASE (13)
         ! Aluminium
         sep%defined = .TRUE.
         sep%uss = -24.546778_dp
         sep%upp = -20.104434_dp
         sep%udd = 8.004394_dp
         sep%beta(0) = -18.375229_dp
         sep%beta(1) = -9.382700_dp
         sep%beta(2) = -20.840474_dp
         sep%sto_exponents(0) = 2.364264_dp
         sep%sto_exponents(1) = 1.749102_dp
         sep%sto_exponents(2) = 1.269384_dp
         sep%zn(0) = 4.742341_dp
         sep%zn(1) = 4.669626_dp
         sep%zn(2) = 7.131138_dp
         sep%gss = 6.652155_dp
         sep%gsp = 7.459435_dp
         sep%gpp = 7.668857_dp
         sep%gp2 = 6.673299_dp
         sep%hsp = 0.435060_dp
         sep%a = 1.002222_dp
         sep%b = 1.517400_dp
         sep%c = 0.659101_dp
      CASE (14)
         ! Silicon
         sep%defined = .TRUE.
         sep%uss = -27.358058_dp
         sep%upp = -20.490578_dp
         sep%udd = -22.751900_dp
         sep%beta(0) = -8.686909_dp
         sep%beta(1) = -1.856482_dp
         sep%beta(2) = -6.360627_dp
         sep%sto_exponents(0) = 1.752741_dp
         sep%sto_exponents(1) = 1.198413_dp
         sep%sto_exponents(2) = 2.128593_dp
         sep%zn(0) = 8.388111_dp
         sep%zn(1) = 1.843048_dp
         sep%zn(2) = 0.708600_dp
         sep%gss = 5.194805_dp
         sep%gsp = 5.090534_dp
         sep%gpp = 5.185150_dp
         sep%gp2 = 4.769775_dp
         sep%hsp = 1.425012_dp
         sep%a = 0.208571_dp
         sep%b = 6.000483_dp
         sep%c = 1.185245_dp
      CASE (15)
         ! Phosphorus
         sep%defined = .TRUE.
         sep%uss = -48.729905_dp
         sep%upp = -40.354689_dp
         sep%udd = -7.349246_dp
         sep%beta(0) = -14.583780_dp
         sep%beta(1) = -11.744725_dp
         sep%beta(2) = -20.099893_dp
         sep%sto_exponents(0) = 2.158033_dp
         sep%sto_exponents(1) = 1.805343_dp
         sep%sto_exponents(2) = 1.230358_dp
         sep%zn(0) = 6.042706_dp
         sep%zn(1) = 2.376473_dp
         sep%zn(2) = 7.147750_dp
         sep%gss = 8.758856_dp
         sep%gsp = 8.483679_dp
         sep%gpp = 8.662754_dp
         sep%gp2 = 7.734264_dp
         sep%hsp = 0.871681_dp
         sep%a = -0.034320_dp
         sep%b = 6.001394_dp
         sep%c = 2.296737_dp
      CASE (16)
         ! Sulfur
         sep%defined = .TRUE.
         sep%uss = -47.530706_dp
         sep%upp = -39.191045_dp
         sep%udd = -46.306944_dp
         sep%beta(0) = -13.827440_dp
         sep%beta(1) = -7.664613_dp
         sep%beta(2) = -9.986172_dp
         sep%sto_exponents(0) = 2.192844_dp
         sep%sto_exponents(1) = 1.841078_dp
         sep%sto_exponents(2) = 3.109401_dp
         sep%zn(0) = 0.479722_dp
         sep%zn(1) = 1.015507_dp
         sep%zn(2) = 4.317470_dp
         sep%gss = 9.170350_dp
         sep%gsp = 5.944296_dp
         sep%gpp = 8.165473_dp
         sep%gp2 = 7.301878_dp
         sep%hsp = 5.005404_dp
         sep%a = -0.036928_dp
         sep%b = 1.795067_dp
         sep%c = 2.082618_dp
      CASE (17)
         ! Chlorine
         sep%defined = .TRUE.
         sep%uss = -61.389930_dp
         sep%upp = -54.482801_dp
         sep%udd = -38.258155_dp
         sep%beta(0) = -2.367988_dp
         sep%beta(1) = -13.802139_dp
         sep%beta(2) = -4.037751_dp
         sep%sto_exponents(0) = 2.637050_dp
         sep%sto_exponents(1) = 2.118146_dp
         sep%sto_exponents(2) = 1.324033_dp
         sep%zn(0) = 0.956297_dp
         sep%zn(1) = 2.464067_dp
         sep%zn(2) = 6.410325_dp
         sep%gss = 11.142654_dp
         sep%gsp = 7.487881_dp
         sep%gpp = 9.551886_dp
         sep%gp2 = 8.128436_dp
         sep%hsp = 5.004267_dp
         sep%a = -0.013213_dp
         sep%b = 3.687022_dp
         sep%c = 2.544635_dp
      CASE (18)
         ! Argon
         sep%defined = .TRUE.
         sep%uss = -7.797931_dp
         sep%upp = -83.211487_dp
         sep%beta(0) = -8.839842_dp
         sep%beta(1) = -28.427303_dp
         sep%sto_exponents(0) = 6.000272_dp
         sep%sto_exponents(1) = 5.949170_dp
         sep%gss = 17.858776_dp
         sep%gsp = 4.168451_dp
         sep%gpp = 11.852500_dp
         sep%gp2 = 15.669543_dp
         sep%hsp = 4.574549_dp
      CASE (19)
         ! Potassium
         sep%defined = .TRUE.
         sep%uss = -3.801108_dp
         sep%upp = -3.339656_dp
         sep%beta(0) = -8.755195_dp
         sep%beta(1) = -1.788061_dp
         sep%sto_exponents(0) = 6.000478_dp
         sep%sto_exponents(1) = 1.127503_dp
         sep%gss = 3.369251_dp
         sep%gsp = 6.129351_dp
         sep%gpp = 0.999505_dp
         sep%gp2 = 18.999148_dp
         sep%hsp = 0.300325_dp
         sep%a = 0.157519_dp
         sep%b = 6.000566_dp
         sep%c = 2.047539_dp
      CASE (20)
         ! Calcium
         sep%defined = .TRUE.
         sep%uss = -10.770058_dp
         sep%upp = -9.754177_dp
         sep%beta(0) = -4.343881_dp
         sep%beta(1) = -1.296612_dp
         sep%sto_exponents(0) = 1.528258_dp
         sep%sto_exponents(1) = 2.060094_dp
         sep%gss = 5.725773_dp
         sep%gsp = 4.781065_dp
         sep%gpp = 7.172103_dp
         sep%gp2 = 7.431876_dp
         sep%hsp = 1.240572_dp
         sep%a = -0.025275_dp
         sep%b = 0.500017_dp
         sep%c = 2.329051_dp
      CASE (21)
         ! Scandium
         sep%defined = .TRUE.
         sep%uss = -15.544461_dp
         sep%upp = -18.646295_dp
         sep%udd = -16.069444_dp
         sep%beta(0) = -8.620944_dp
         sep%beta(1) = 3.075948_dp
         sep%beta(2) = -9.768661_dp
         sep%sto_exponents(0) = 1.402469_dp
         sep%sto_exponents(1) = 1.345196_dp
         sep%sto_exponents(2) = 1.859012_dp
         sep%zn(0) = 0.848418_dp
         sep%zn(1) = 2.451729_dp
         sep%zn(2) = 0.789372_dp
         sep%gss = 4.638215_dp
         sep%gsp = 5.739164_dp
         sep%gpp = 14.604872_dp
         sep%gp2 = 12.802595_dp
         sep%hsp = 0.193835_dp
         sep%rho = 3.173734_dp
         sep%f0sd = 4.798313_dp
         sep%g2sd = 5.380136_dp
      CASE (22)
         ! Titanium
         sep%defined = .TRUE.
         sep%uss = -25.507973_dp
         sep%upp = -17.260909_dp
         sep%udd = -23.809486_dp
         sep%beta(0) = 3.389142_dp
         sep%beta(1) = -3.355350_dp
         sep%beta(2) = -1.842829_dp
         sep%sto_exponents(0) = 5.324777_dp
         sep%sto_exponents(1) = 1.164068_dp
         sep%sto_exponents(2) = 1.418280_dp
         sep%zn(0) = 1.045904_dp
         sep%zn(1) = 1.076844_dp
         sep%zn(2) = 0.717945_dp
         sep%gss = 5.717851_dp
         sep%gsp = 5.800015_dp
         sep%gpp = 6.414726_dp
         sep%gp2 = 5.623133_dp
         sep%hsp = 1.403732_dp
         sep%f0sd = 6.560562_dp
         sep%g2sd = 3.396235_dp
      CASE (23)
         ! Vanadium
         sep%defined = .TRUE.
         sep%uss = -32.162276_dp
         sep%upp = -21.572501_dp
         sep%udd = -34.506245_dp
         sep%beta(0) = -1.211330_dp
         sep%beta(1) = 0.740746_dp
         sep%beta(2) = 3.153669_dp
         sep%sto_exponents(0) = 1.974330_dp
         sep%sto_exponents(1) = 1.063106_dp
         sep%sto_exponents(2) = 1.394806_dp
         sep%zn(0) = 1.094426_dp
         sep%zn(1) = 0.755378_dp
         sep%zn(2) = 1.099367_dp
         sep%gss = 5.983116_dp
         sep%gsp = 4.736769_dp
         sep%gpp = 4.499763_dp
         sep%gp2 = 3.944481_dp
         sep%hsp = 0.901105_dp
         sep%f0sd = 6.810021_dp
         sep%g2sd = 1.831407_dp
      CASE (24)
         ! Chromium
         sep%defined = .TRUE.
         sep%uss = -34.864339_dp
         sep%upp = -26.978615_dp
         sep%udd = -54.431036_dp
         sep%beta(0) = -5.122615_dp
         sep%beta(1) = 3.926711_dp
         sep%beta(2) = -4.230550_dp
         sep%sto_exponents(0) = 3.283460_dp
         sep%sto_exponents(1) = 1.029394_dp
         sep%sto_exponents(2) = 1.623119_dp
         sep%zn(0) = 1.619853_dp
         sep%zn(1) = 0.848266_dp
         sep%zn(2) = 1.405015_dp
         sep%gss = 8.855572_dp
         sep%gsp = 5.588631_dp
         sep%gpp = 5.053094_dp
         sep%gp2 = 4.429530_dp
         sep%hsp = 0.648039_dp
         sep%f0sd = 6.150136_dp
         sep%g2sd = 2.000300_dp
      CASE (25)
         ! Manganese
         sep%defined = .TRUE.
         sep%uss = -51.460000_dp
         sep%upp = -37.543990_dp
         sep%udd = -47.655370_dp
         sep%beta(0) = -4.185290_dp
         sep%beta(1) = -3.479630_dp
         sep%beta(2) = -13.473190_dp
         sep%sto_exponents(0) = 2.131680_dp
         sep%sto_exponents(1) = 1.525880_dp
         sep%sto_exponents(2) = 2.607800_dp
         sep%zn(0) = 1.132450_dp
         sep%zn(1) = 1.390740_dp
         sep%zn(2) = 0.962550_dp
         sep%gss = 6.190990_dp
         sep%gsp = 6.757427_dp
         sep%gpp = 8.284594_dp
         sep%gp2 = 7.262255_dp
         sep%hsp = 1.520518_dp
         sep%f0sd = 7.690920_dp
         sep%g2sd = 1.105330_dp
      CASE (26)
         ! Iron
         sep%defined = .TRUE.
         sep%uss = -70.515047_dp
         sep%upp = -62.963069_dp
         sep%udd = -103.631790_dp
         sep%beta(0) = 8.027621_dp
         sep%beta(1) = -1.125760_dp
         sep%beta(2) = -3.507531_dp
         sep%sto_exponents(0) = 1.479150_dp
         sep%sto_exponents(1) = 6.002246_dp
         sep%sto_exponents(2) = 1.080747_dp
         sep%zn(0) = 1.459152_dp
         sep%zn(1) = 1.392614_dp
         sep%zn(2) = 2.161909_dp
         sep%gss = 7.977036_dp
         sep%gsp = 7.786867_dp
         sep%gpp = 8.295758_dp
         sep%gp2 = 7.272041_dp
         sep%hsp = 1.880189_dp
         sep%rho = 1.272092_dp
         sep%f0sd = 9.300165_dp
         sep%g2sd = 1.601345_dp
      CASE (27)
         ! Cobalt
         sep%defined = .TRUE.
         sep%uss = -21.039413_dp
         sep%upp = 10.000000_dp
         sep%udd = -28.068971_dp
         sep%beta(0) = -8.992062_dp
         sep%beta(1) = -0.100000_dp
         sep%beta(2) = -2.481509_dp
         sep%sto_exponents(0) = 1.166613_dp
         sep%sto_exponents(1) = 3.000000_dp
         sep%sto_exponents(2) = 1.860218_dp
         sep%zn(0) = 0.519518_dp
         sep%zn(1) = 1.000000_dp
         sep%zn(2) = 0.352115_dp
         sep%gss = 2.840152_dp
         sep%gsp = 3.425933_dp
         sep%gpp = 5.956968_dp
         sep%gp2 = 5.221864_dp
         sep%hsp = 0.390087_dp
         sep%f0sd = 1.446283_dp
         sep%g2sd = 1.680225_dp
      CASE (28)
         ! Nickel
         sep%defined = .TRUE.
         sep%uss = -47.620247_dp
         sep%upp = -32.878408_dp
         sep%udd = -93.026395_dp
         sep%beta(0) = -9.151521_dp
         sep%beta(1) = -8.086696_dp
         sep%beta(2) = -8.655910_dp
         sep%sto_exponents(0) = 1.591828_dp
         sep%sto_exponents(1) = 2.304739_dp
         sep%sto_exponents(2) = 2.514761_dp
         sep%zn(0) = 0.746470_dp
         sep%zn(1) = 0.753327_dp
         sep%zn(2) = 1.461345_dp
         sep%gss = 4.080876_dp
         sep%gsp = 4.099452_dp
         sep%gpp = 4.487545_dp
         sep%gp2 = 3.933771_dp
         sep%hsp = 0.993498_dp
         sep%rho = 1.586979_dp
         sep%f0sd = 4.651664_dp
         sep%g2sd = 1.880502_dp
      CASE (29)
         ! Copper
         sep%defined = .TRUE.
         sep%uss = -97.002205_dp
         sep%upp = -1.000000_dp
         sep%udd = -110.442592_dp
         sep%beta(0) = -9.369508_dp
         sep%beta(1) = -0.100000_dp
         sep%beta(2) = -16.982092_dp
         sep%sto_exponents(0) = 1.669096_dp
         sep%sto_exponents(1) = 3.000000_dp
         sep%sto_exponents(2) = 2.734990_dp
         sep%zn(0) = 1.899598_dp
         sep%zn(1) = 3.000000_dp
         sep%zn(2) = 1.484317_dp
         sep%gss = 10.384910_dp
         sep%gsp = 12.145361_dp
         sep%gpp = 17.870905_dp
         sep%gp2 = 15.665592_dp
         sep%hsp = 2.037394_dp
         sep%f0sd = 9.848807_dp
         sep%g2sd = 9.847577_dp
      CASE (30)
         ! Zinc
         sep%defined = .TRUE.
         sep%uss = -18.040862_dp
         sep%upp = -7.834895_dp
         sep%beta(0) = -13.276583_dp
         sep%beta(1) = 1.479642_dp
         sep%sto_exponents(0) = 1.512875_dp
         sep%sto_exponents(1) = 1.789482_dp
         sep%gss = 8.707424_dp
         sep%gsp = 3.436116_dp
         sep%gpp = 20.000041_dp
         sep%gp2 = 6.782785_dp
         sep%hsp = 0.662036_dp
      CASE (31)
         ! Gallium
         sep%defined = .TRUE.
         sep%uss = -30.600226_dp
         sep%upp = -21.032425_dp
         sep%beta(0) = -10.808320_dp
         sep%beta(1) = -4.185500_dp
         sep%sto_exponents(0) = 2.339067_dp
         sep%sto_exponents(1) = 1.729592_dp
         sep%gss = 10.354885_dp
         sep%gsp = 7.993674_dp
         sep%gpp = 6.090184_dp
         sep%gp2 = 6.299226_dp
         sep%hsp = 1.295974_dp
      CASE (32)
         ! Germanium
         sep%defined = .TRUE.
         sep%uss = -32.747338_dp
         sep%upp = -24.709016_dp
         sep%beta(0) = -14.854297_dp
         sep%beta(1) = -2.591260_dp
         sep%sto_exponents(0) = 2.546073_dp
         sep%sto_exponents(1) = 1.709130_dp
         sep%gss = 7.518301_dp
         sep%gsp = 6.594443_dp
         sep%gpp = 6.066801_dp
         sep%gp2 = 5.305947_dp
         sep%hsp = 0.290742_dp
      CASE (33)
         ! Arsenic
         sep%defined = .TRUE.
         sep%uss = -37.956965_dp
         sep%upp = -38.453701_dp
         sep%udd = -30.282658_dp
         sep%beta(0) = -11.963725_dp
         sep%beta(1) = -7.340073_dp
         sep%beta(2) = 3.753005_dp
         sep%sto_exponents(0) = 2.926171_dp
         sep%sto_exponents(1) = 1.765191_dp
         sep%sto_exponents(2) = 1.392142_dp
         sep%zn(0) = 2.006543_dp
         sep%zn(1) = 3.316832_dp
         sep%zn(2) = 4.653440_dp
         sep%gss = 6.665030_dp
         sep%gsp = 6.213867_dp
         sep%gpp = 9.310836_dp
         sep%gp2 = 8.712542_dp
         sep%hsp = 0.280662_dp
      CASE (34)
         ! Selenium
         sep%defined = .TRUE.
         sep%uss = -32.671088_dp
         sep%upp = -32.522220_dp
         sep%beta(0) = 2.636001_dp
         sep%beta(1) = -9.557700_dp
         sep%sto_exponents(0) = 2.512366_dp
         sep%sto_exponents(1) = 2.007576_dp
         sep%gss = 5.522356_dp
         sep%gsp = 2.907562_dp
         sep%gpp = 8.042391_dp
         sep%gp2 = 6.735106_dp
         sep%hsp = 3.095789_dp
      CASE (35)
         ! Bromine
         sep%defined = .TRUE.
         sep%uss = -45.834364_dp
         sep%upp = -50.293675_dp
         sep%udd = 7.086738_dp
         sep%beta(0) = -32.131665_dp
         sep%beta(1) = -9.514484_dp
         sep%beta(2) = -9.839124_dp
         sep%sto_exponents(0) = 4.670684_dp
         sep%sto_exponents(1) = 2.035626_dp
         sep%sto_exponents(2) = 1.521031_dp
         sep%zn(0) = 3.094777_dp
         sep%zn(1) = 3.065764_dp
         sep%zn(2) = 2.820003_dp
         sep%gss = 7.616791_dp
         sep%gsp = 5.010425_dp
         sep%gpp = 9.649216_dp
         sep%gp2 = 8.343792_dp
         sep%hsp = 4.996553_dp
         sep%a = -0.004996_dp
         sep%b = 6.001292_dp
         sep%c = 2.895153_dp
      CASE (36)
         ! Krypton
         sep%defined = .TRUE.
         sep%uss = 8.535384_dp
         sep%upp = -80.484321_dp
         sep%beta(0) = -2.727088_dp
         sep%beta(1) = -16.142951_dp
         sep%sto_exponents(0) = 1.312248_dp
         sep%sto_exponents(1) = 4.491371_dp
         sep%gss = 19.999857_dp
         sep%gsp = 1.175304_dp
         sep%gpp = 9.174784_dp
         sep%gp2 = 14.926948_dp
         sep%hsp = 0.299867_dp
      CASE (37)
         ! Rubidium
         sep%defined = .TRUE.
         sep%uss = -3.636505_dp
         sep%upp = -2.500671_dp
         sep%beta(0) = 9.998744_dp
         sep%beta(1) = 1.343004_dp
         sep%sto_exponents(0) = 5.510145_dp
         sep%sto_exponents(1) = 1.335170_dp
         sep%gss = 6.680824_dp
         sep%gsp = 20.001098_dp
         sep%gpp = 5.068874_dp
         sep%gp2 = 2.747860_dp
         sep%hsp = 3.602834_dp
      CASE (38)
         ! Strontium
         sep%defined = .TRUE.
         sep%uss = -10.427671_dp
         sep%upp = -9.943751_dp
         sep%beta(0) = -6.253108_dp
         sep%beta(1) = -9.844498_dp
         sep%sto_exponents(0) = 2.197303_dp
         sep%sto_exponents(1) = 1.730137_dp
         sep%gss = 4.603664_dp
         sep%gsp = 5.716069_dp
         sep%gpp = 7.334620_dp
         sep%gp2 = 7.443088_dp
         sep%hsp = 0.831527_dp
         sep%a = -0.012948_dp
         sep%b = 6.000126_dp
         sep%c = 3.011964_dp
      CASE (39)
         ! Yttrium
         sep%defined = .TRUE.
         sep%uss = -14.247809_dp
         sep%upp = -14.817140_dp
         sep%udd = -16.394302_dp
         sep%beta(0) = 0.343336_dp
         sep%beta(1) = -3.180807_dp
         sep%beta(2) = -4.508957_dp
         sep%sto_exponents(0) = 0.593368_dp
         sep%sto_exponents(1) = 1.490422_dp
         sep%sto_exponents(2) = 1.650893_dp
         sep%zn(0) = 0.902611_dp
         sep%zn(1) = 1.484400_dp
         sep%zn(2) = 1.384238_dp
         sep%gss = 4.046733_dp
         sep%gsp = 4.726277_dp
         sep%gpp = 7.278752_dp
         sep%gp2 = 6.343281_dp
         sep%hsp = 0.679228_dp
         sep%rho = 2.773703_dp
         sep%f0sd = 4.972716_dp
         sep%g2sd = 5.016364_dp
      CASE (40)
         ! Zirconium
         sep%defined = .TRUE.
         sep%uss = -20.008884_dp
         sep%upp = -14.559692_dp
         sep%udd = -21.302657_dp
         sep%beta(0) = 9.551952_dp
         sep%beta(1) = -4.551915_dp
         sep%beta(2) = -3.213274_dp
         sep%sto_exponents(0) = 1.692590_dp
         sep%sto_exponents(1) = 1.694916_dp
         sep%sto_exponents(2) = 1.567392_dp
         sep%zn(0) = 1.189109_dp
         sep%zn(1) = 0.809092_dp
         sep%zn(2) = 1.190249_dp
         sep%gss = 5.331208_dp
         sep%gsp = 4.150579_dp
         sep%gpp = 3.967381_dp
         sep%gp2 = 3.457490_dp
         sep%hsp = 0.743676_dp
         sep%f0sd = 5.010704_dp
         sep%g2sd = 2.943652_dp
      CASE (41)
         ! Niobium
         sep%defined = .TRUE.
         sep%uss = -31.269298_dp
         sep%upp = -20.151277_dp
         sep%udd = -35.893116_dp
         sep%beta(0) = -12.045244_dp
         sep%beta(1) = 1.465762_dp
         sep%beta(2) = -5.920160_dp
         sep%sto_exponents(0) = 2.355562_dp
         sep%sto_exponents(1) = 1.386907_dp
         sep%sto_exponents(2) = 1.977324_dp
         sep%zn(0) = 1.490754_dp
         sep%zn(1) = 0.892760_dp
         sep%zn(2) = 1.443837_dp
         sep%gss = 6.683592_dp
         sep%gsp = 4.685339_dp
         sep%gpp = 4.377647_dp
         sep%gp2 = 3.815028_dp
         sep%hsp = 0.650679_dp
         sep%f0sd = 6.550674_dp
         sep%g2sd = 1.065577_dp
      CASE (42)
         ! Molybdenum
         sep%defined = .TRUE.
         sep%uss = -53.467728_dp
         sep%upp = -35.291951_dp
         sep%udd = -55.836977_dp
         sep%beta(0) = -0.189344_dp
         sep%beta(1) = 7.017762_dp
         sep%beta(2) = -10.941126_dp
         sep%sto_exponents(0) = 1.060429_dp
         sep%sto_exponents(1) = 1.350412_dp
         sep%sto_exponents(2) = 1.827152_dp
         sep%zn(0) = 1.912995_dp
         sep%zn(1) = 1.355055_dp
         sep%zn(2) = 1.876231_dp
         sep%gss = 8.576652_dp
         sep%gsp = 6.888293_dp
         sep%gpp = 6.644509_dp
         sep%gp2 = 5.790552_dp
         sep%hsp = 1.317368_dp
         sep%f0sd = 10.000608_dp
         sep%g2sd = 1.216752_dp
      CASE (43)
         ! Technetium
         sep%defined = .TRUE.
         sep%uss = -41.850292_dp
         sep%upp = -34.910293_dp
         sep%udd = -45.530412_dp
         sep%beta(0) = -2.791024_dp
         sep%beta(1) = -8.086697_dp
         sep%beta(2) = -5.724335_dp
         sep%sto_exponents(0) = 1.956245_dp
         sep%sto_exponents(1) = 6.006299_dp
         sep%sto_exponents(2) = 1.767360_dp
         sep%zn(0) = 1.411033_dp
         sep%zn(1) = 1.141313_dp
         sep%zn(2) = 1.159312_dp
         sep%gss = 6.326174_dp
         sep%gsp = 5.587138_dp
         sep%gpp = 5.596426_dp
         sep%gp2 = 4.877169_dp
         sep%hsp = 1.258989_dp
         sep%f0sd = 5.434886_dp
         sep%g2sd = 1.106875_dp
      CASE (44)
         ! Ruthenium
         sep%defined = .TRUE.
         sep%uss = -44.901521_dp
         sep%upp = -41.424409_dp
         sep%udd = -37.934514_dp
         sep%beta(0) = -12.859508_dp
         sep%beta(1) = -8.475518_dp
         sep%beta(2) = -3.830797_dp
         sep%sto_exponents(0) = 1.459195_dp
         sep%sto_exponents(1) = 5.537201_dp
         sep%sto_exponents(2) = 2.093164_dp
         sep%zn(0) = 0.984449_dp
         sep%zn(1) = 4.586613_dp
         sep%zn(2) = 0.765332_dp
         sep%gss = 4.413643_dp
         sep%gsp = 5.356996_dp
         sep%gpp = 22.490448_dp
         sep%gp2 = 19.599957_dp
         sep%hsp = 0.008058_dp
         sep%f0sd = 5.917404_dp
         sep%g2sd = 5.859738_dp
      CASE (45)
         ! Rhodium
         sep%defined = .TRUE.
         sep%uss = -20.513756_dp
         sep%upp = -40.045431_dp
         sep%udd = -35.818492_dp
         sep%beta(0) = -8.222141_dp
         sep%beta(1) = -15.556691_dp
         sep%beta(2) = -13.396182_dp
         sep%sto_exponents(0) = 1.324919_dp
         sep%sto_exponents(1) = 4.306111_dp
         sep%sto_exponents(2) = 2.901406_dp
         sep%zn(0) = 0.809923_dp
         sep%zn(1) = 6.898259_dp
         sep%zn(2) = 0.643134_dp
         sep%gss = 3.631179_dp
         sep%gsp = 4.407820_dp
         sep%gpp = 33.825599_dp
         sep%gp2 = 29.478305_dp
         sep%hsp = 0.000092_dp
         sep%f0sd = 1.775497_dp
         sep%g2sd = 1.851571_dp
      CASE (46)
         ! Palladium
         sep%defined = .TRUE.
         sep%uss = -76.140196_dp
         sep%upp = -21.073362_dp
         sep%udd = -85.325301_dp
         sep%beta(0) = -8.038245_dp
         sep%beta(1) = 0.740037_dp
         sep%beta(2) = -2.394498_dp
         sep%sto_exponents(0) = 1.658503_dp
         sep%sto_exponents(1) = 1.156718_dp
         sep%sto_exponents(2) = 2.219861_dp
         sep%zn(0) = 1.794085_dp
         sep%zn(1) = 6.158778_dp
         sep%zn(2) = 1.630913_dp
         sep%gss = 8.043535_dp
         sep%gsp = 9.755042_dp
         sep%gpp = 30.199556_dp
         sep%gp2 = 26.318284_dp
         sep%hsp = 0.086121_dp
         sep%f0sd = 8.004447_dp
         sep%g2sd = 2.613148_dp
      CASE (47)
         ! Silver
         sep%defined = .TRUE.
         sep%uss = -25.484137_dp
         sep%upp = -36.116023_dp
         sep%udd = -35.668272_dp
         sep%beta(0) = -6.129623_dp
         sep%beta(1) = 1.004115_dp
         sep%beta(2) = -69.238347_dp
         sep%sto_exponents(0) = 1.994004_dp
         sep%sto_exponents(1) = 0.681817_dp
         sep%sto_exponents(2) = 6.007328_dp
         sep%zn(0) = 0.695514_dp
         sep%zn(1) = 4.729949_dp
         sep%zn(2) = 0.506522_dp
         sep%gss = 3.118242_dp
         sep%gsp = 3.785152_dp
         sep%gpp = 23.193295_dp
         sep%gp2 = 20.212474_dp
         sep%hsp = 0.000432_dp
         sep%f0sd = 1.938327_dp
         sep%g2sd = 1.071901_dp
      CASE (48)
         ! Cadmium
         sep%defined = .TRUE.
         sep%uss = -14.645792_dp
         sep%upp = -9.318664_dp
         sep%beta(0) = -11.613183_dp
         sep%beta(1) = 1.663168_dp
         sep%sto_exponents(0) = 1.384108_dp
         sep%sto_exponents(1) = 1.957413_dp
         sep%gss = 6.677284_dp
         sep%gsp = 5.953373_dp
         sep%gpp = 18.729843_dp
         sep%gp2 = 9.917452_dp
         sep%hsp = 0.825192_dp
      CASE (49)
         ! Indium
         sep%defined = .TRUE.
         sep%uss = -28.339246_dp
         sep%upp = -23.373875_dp
         sep%beta(0) = -1.982376_dp
         sep%beta(1) = -3.330294_dp
         sep%sto_exponents(0) = 2.023087_dp
         sep%sto_exponents(1) = 2.106618_dp
         sep%gss = 9.906091_dp
         sep%gsp = 10.520060_dp
         sep%gpp = 4.826006_dp
         sep%gp2 = 7.906563_dp
         sep%hsp = 3.500299_dp
      CASE (50)
         ! Tin
         sep%defined = .TRUE.
         sep%uss = -29.888217_dp
         sep%upp = -22.156954_dp
         sep%beta(0) = -8.621087_dp
         sep%beta(1) = -4.989752_dp
         sep%sto_exponents(0) = 2.383941_dp
         sep%sto_exponents(1) = 2.057908_dp
         sep%gss = 8.269655_dp
         sep%gsp = 5.013349_dp
         sep%gpp = 6.584874_dp
         sep%gp2 = 5.855159_dp
         sep%hsp = 0.531212_dp
         sep%a = -1.004587_dp
         sep%b = 4.706252_dp
         sep%c = 1.180218_dp
      CASE (51)
         ! Antimony
         sep%defined = .TRUE.
         sep%uss = -41.688879_dp
         sep%upp = -39.541180_dp
         sep%udd = -6.581663_dp
         sep%beta(0) = -7.472322_dp
         sep%beta(1) = -5.940750_dp
         sep%beta(2) = -3.979108_dp
         sep%sto_exponents(0) = 2.391178_dp
         sep%sto_exponents(1) = 1.773006_dp
         sep%sto_exponents(2) = 2.465590_dp
         sep%zn(0) = 5.993591_dp
         sep%zn(1) = 6.145086_dp
         sep%zn(2) = 5.704031_dp
         sep%gss = 10.588832_dp
         sep%gsp = 7.310023_dp
         sep%gpp = 9.281609_dp
         sep%gp2 = 8.954081_dp
         sep%hsp = 0.779112_dp
      CASE (52)
         ! Tellurium
         sep%defined = .TRUE.
         sep%uss = -114.733316_dp
         sep%upp = -50.096389_dp
         sep%beta(0) = -70.001062_dp
         sep%beta(1) = -6.151642_dp
         sep%sto_exponents(0) = 2.769862_dp
         sep%sto_exponents(1) = 1.731319_dp
         sep%gss = 7.030626_dp
         sep%gsp = 12.601389_dp
         sep%gpp = 7.883479_dp
         sep%gp2 = 6.973163_dp
         sep%hsp = 5.000826_dp
      CASE (53)
         ! Iodine
         sep%defined = .TRUE.
         sep%uss = -59.973232_dp
         sep%upp = -56.459835_dp
         sep%udd = -28.822603_dp
         sep%beta(0) = -30.522481_dp
         sep%beta(1) = -5.942120_dp
         sep%beta(2) = -7.676107_dp
         sep%sto_exponents(0) = 4.498653_dp
         sep%sto_exponents(1) = 1.917072_dp
         sep%sto_exponents(2) = 1.875175_dp
         sep%zn(0) = 9.135244_dp
         sep%zn(1) = 6.888191_dp
         sep%zn(2) = 3.791523_dp
         sep%gss = 7.234759_dp
         sep%gsp = 9.154406_dp
         sep%gpp = 9.877466_dp
         sep%gp2 = 8.035916_dp
         sep%hsp = 5.004215_dp
         sep%a = -0.035519_dp
         sep%b = 1.744389_dp
         sep%c = 1.223844_dp
      CASE (54)
         ! Xenon
         sep%defined = .TRUE.
         sep%uss = -18.270227_dp
         sep%upp = -167.163063_dp
         sep%beta(0) = -3.980622_dp
         sep%beta(1) = -38.822792_dp
         sep%sto_exponents(0) = 2.759787_dp
         sep%sto_exponents(1) = 1.977446_dp
         sep%gss = 20.000252_dp
         sep%gsp = 4.175902_dp
         sep%gpp = 2.305787_dp
         sep%gp2 = 4.063220_dp
         sep%hsp = 4.418843_dp
      CASE (55)
         ! Cesium
         sep%defined = .TRUE.
         sep%uss = -3.748609_dp
         sep%upp = -2.348109_dp
         sep%beta(0) = 2.287838_dp
         sep%beta(1) = -5.908071_dp
         sep%sto_exponents(0) = 5.956008_dp
         sep%sto_exponents(1) = 1.619485_dp
         sep%gss = 6.464751_dp
         sep%gsp = 4.004501_dp
         sep%gpp = 13.775390_dp
         sep%gp2 = 12.912537_dp
         sep%hsp = 1.026928_dp
      CASE (56)
         ! Barium
         sep%defined = .TRUE.
         sep%uss = -9.306985_dp
         sep%upp = -8.826713_dp
         sep%beta(0) = 10.003125_dp
         sep%beta(1) = -6.335160_dp
         sep%sto_exponents(0) = 1.395379_dp
         sep%sto_exponents(1) = 1.430139_dp
         sep%gss = 3.600823_dp
         sep%gsp = 4.740579_dp
         sep%gpp = 3.345166_dp
         sep%gp2 = 3.142783_dp
         sep%hsp = 0.929429_dp
      CASE (57)
         ! Lanthanum
         sep%defined = .TRUE.
         sep%uss = -19.641953_dp
         sep%upp = -22.059431_dp
         sep%udd = -22.638986_dp
         sep%beta(0) = 0.796727_dp
         sep%beta(1) = -10.856056_dp
         sep%beta(2) = -0.484922_dp
         sep%sto_exponents(0) = 2.673780_dp
         sep%sto_exponents(1) = 1.248192_dp
         sep%sto_exponents(2) = 1.688562_dp
         sep%zn(0) = 1.617784_dp
         sep%zn(1) = 4.331620_dp
         sep%zn(2) = 2.285738_dp
         sep%gss = 6.154440_dp
         sep%gsp = 7.322704_dp
         sep%gpp = 18.077465_dp
         sep%gp2 = 15.679057_dp
         sep%hsp = 0.138601_dp
         sep%rho = 2.511701_dp
         sep%f0sd = 8.856858_dp
         sep%g2sd = 7.925585_dp
      CASE (58:70)
      CASE (71)
         ! Lutetium
         sep%defined = .TRUE.
         sep%uss = -15.954994_dp
         sep%upp = -11.606213_dp
         sep%udd = -13.050056_dp
         sep%beta(0) = -5.590778_dp
         sep%beta(1) = -0.937679_dp
         sep%beta(2) = -7.737752_dp
         sep%sto_exponents(0) = 5.471741_dp
         sep%sto_exponents(1) = 1.712296_dp
         sep%sto_exponents(2) = 2.225892_dp
         sep%zn(0) = 1.632335_dp
         sep%zn(1) = 4.033128_dp
         sep%zn(2) = 0.921999_dp
         sep%gss = 6.209796_dp
         sep%gsp = 7.379102_dp
         sep%gpp = 16.831746_dp
         sep%gp2 = 14.598613_dp
         sep%hsp = 0.209008_dp
         sep%rho = 2.743262_dp
         sep%f0sd = 3.924927_dp
         sep%g2sd = 1.000946_dp
      CASE (72)
         ! Hafnium
         sep%defined = .TRUE.
         sep%uss = -22.375140_dp
         sep%upp = -13.081670_dp
         sep%udd = -20.637741_dp
         sep%beta(0) = -5.366351_dp
         sep%beta(1) = -21.550119_dp
         sep%beta(2) = -3.884443_dp
         sep%sto_exponents(0) = 3.085344_dp
         sep%sto_exponents(1) = 1.575819_dp
         sep%sto_exponents(2) = 1.840840_dp
         sep%zn(0) = 0.946927_dp
         sep%zn(1) = 3.538911_dp
         sep%zn(2) = 0.940283_dp
         sep%gss = 3.602338_dp
         sep%gsp = 4.293729_dp
         sep%gpp = 14.769194_dp
         sep%gp2 = 12.809708_dp
         sep%hsp = 0.011028_dp
         sep%f0sd = 4.842900_dp
         sep%g2sd = 4.386101_dp
      CASE (73)
         ! Tantalum
         sep%defined = .TRUE.
         sep%uss = -39.009984_dp
         sep%upp = 1.163975_dp
         sep%udd = -43.266315_dp
         sep%beta(0) = -17.199605_dp
         sep%beta(1) = -5.818839_dp
         sep%beta(2) = -9.816794_dp
         sep%sto_exponents(0) = 4.578087_dp
         sep%sto_exponents(1) = 4.841244_dp
         sep%sto_exponents(2) = 1.838249_dp
         sep%zn(0) = 1.741367_dp
         sep%zn(1) = 3.430157_dp
         sep%zn(2) = 2.311198_dp
         sep%gss = 6.624580_dp
         sep%gsp = 7.805321_dp
         sep%gpp = 14.315323_dp
         sep%gp2 = 12.416054_dp
         sep%hsp = 0.577263_dp
         sep%f0sd = 8.544427_dp
         sep%g2sd = 2.074254_dp
      CASE (74)
         ! Tungsten
         sep%defined = .TRUE.
         sep%uss = -44.524950_dp
         sep%upp = -40.011500_dp
         sep%udd = -46.490410_dp
         sep%beta(0) = -16.946460_dp
         sep%beta(1) = 5.623170_dp
         sep%beta(2) = -2.947340_dp
         sep%sto_exponents(0) = 2.664560_dp
         sep%sto_exponents(1) = 1.624010_dp
         sep%sto_exponents(2) = 1.794400_dp
         sep%zn(0) = 1.498860_dp
         sep%zn(1) = 1.965900_dp
         sep%zn(2) = 1.876450_dp
         sep%gss = 5.702025_dp
         sep%gsp = 6.323145_dp
         sep%gpp = 8.204433_dp
         sep%gp2 = 7.115919_dp
         sep%hsp = 1.319912_dp
         sep%f0sd = 7.788180_dp
         sep%g2sd = 1.684940_dp
      CASE (75)
         ! Rhenium
         sep%defined = .TRUE.
         sep%uss = -41.291342_dp
         sep%upp = -35.089592_dp
         sep%udd = -44.178985_dp
         sep%beta(0) = 3.830075_dp
         sep%beta(1) = -1.638530_dp
         sep%beta(2) = -1.414411_dp
         sep%sto_exponents(0) = 2.411839_dp
         sep%sto_exponents(1) = 1.815351_dp
         sep%sto_exponents(2) = 2.522766_dp
         sep%zn(0) = 1.680823_dp
         sep%zn(1) = 1.331218_dp
         sep%zn(2) = 1.490623_dp
         sep%gss = 6.394256_dp
         sep%gsp = 5.555571_dp
         sep%gpp = 5.555669_dp
         sep%gp2 = 4.818577_dp
         sep%hsp = 1.220913_dp
         sep%f0sd = 5.442812_dp
         sep%g2sd = 2.376279_dp
      CASE (76)
         ! Osmium
         sep%defined = .TRUE.
         sep%uss = -26.434080_dp
         sep%upp = -48.739500_dp
         sep%udd = -55.837880_dp
         sep%beta(0) = -12.508730_dp
         sep%beta(1) = 0.846880_dp
         sep%beta(2) = 5.164360_dp
         sep%sto_exponents(0) = 3.031000_dp
         sep%sto_exponents(1) = 1.593960_dp
         sep%sto_exponents(2) = 1.775570_dp
         sep%zn(0) = 1.844700_dp
         sep%zn(1) = 1.564220_dp
         sep%zn(2) = 1.770010_dp
         sep%gss = 7.017683_dp
         sep%gsp = 6.384200_dp
         sep%gpp = 6.528073_dp
         sep%gp2 = 5.661968_dp
         sep%hsp = 1.508926_dp
         sep%f0sd = 2.021170_dp
         sep%g2sd = 1.392130_dp
      CASE (77)
         ! Iridium
         sep%defined = .TRUE.
         sep%uss = -29.703974_dp
         sep%upp = -38.210924_dp
         sep%udd = -32.538202_dp
         sep%beta(0) = -10.943427_dp
         sep%beta(1) = 2.908880_dp
         sep%beta(2) = -3.791731_dp
         sep%sto_exponents(0) = 1.500907_dp
         sep%sto_exponents(1) = 4.106373_dp
         sep%sto_exponents(2) = 2.676047_dp
         sep%zn(0) = 0.927246_dp
         sep%zn(1) = 3.191892_dp
         sep%zn(2) = 0.662007_dp
         sep%gss = 3.527467_dp
         sep%gsp = 4.203820_dp
         sep%gpp = 13.320955_dp
         sep%gp2 = 11.553612_dp
         sep%hsp = 0.018501_dp
         sep%f0sd = 2.627170_dp
         sep%g2sd = 2.996029_dp
      CASE (78)
         ! Platinum
         sep%defined = .TRUE.
         sep%uss = -73.516173_dp
         sep%upp = -68.320056_dp
         sep%udd = -76.598873_dp
         sep%beta(0) = 1.151418_dp
         sep%beta(1) = 3.298694_dp
         sep%beta(2) = -18.044737_dp
         sep%sto_exponents(0) = 2.301264_dp
         sep%sto_exponents(1) = 1.662404_dp
         sep%sto_exponents(2) = 3.168852_dp
         sep%zn(0) = 2.270699_dp
         sep%zn(1) = 1.949896_dp
         sep%zn(2) = 1.713856_dp
         sep%gss = 8.638286_dp
         sep%gsp = 7.922254_dp
         sep%gpp = 8.137643_dp
         sep%gp2 = 7.057990_dp
         sep%hsp = 1.892617_dp
         sep%f0sd = 7.098591_dp
         sep%g2sd = 4.484183_dp
      CASE (79)
         ! Gold
         sep%defined = .TRUE.
         sep%uss = -95.041846_dp
         sep%upp = -63.890158_dp
         sep%udd = -88.066087_dp
         sep%beta(0) = -7.479625_dp
         sep%beta(1) = 3.664356_dp
         sep%beta(2) = -61.715468_dp
         sep%sto_exponents(0) = 1.814169_dp
         sep%sto_exponents(1) = 1.618657_dp
         sep%sto_exponents(2) = 5.053167_dp
         sep%zn(0) = 2.444680_dp
         sep%zn(1) = 7.014990_dp
         sep%zn(2) = 1.777089_dp
         sep%gss = 9.300152_dp
         sep%gsp = 11.073443_dp
         sep%gpp = 29.276168_dp
         sep%gp2 = 25.391984_dp
         sep%hsp = 0.144384_dp
         sep%f0sd = 8.827257_dp
         sep%g2sd = 4.915625_dp
      CASE (80)
         ! Mercury
         sep%defined = .TRUE.
         sep%uss = -17.608732_dp
         sep%upp = -18.369417_dp
         sep%beta(0) = -3.045239_dp
         sep%beta(1) = -5.693556_dp
         sep%sto_exponents(0) = 2.104896_dp
         sep%sto_exponents(1) = 1.516293_dp
         sep%gss = 6.372822_dp
         sep%gsp = 10.143176_dp
         sep%gpp = 10.397393_dp
         sep%gp2 = 14.794056_dp
         sep%hsp = 0.926128_dp
      CASE (81)
         ! Thallium
         sep%defined = .TRUE.
         sep%uss = -29.518621_dp
         sep%upp = -29.826907_dp
         sep%beta(0) = -7.230170_dp
         sep%beta(1) = -7.575544_dp
         sep%sto_exponents(0) = 3.335883_dp
         sep%sto_exponents(1) = 1.766141_dp
         sep%gss = 5.015118_dp
         sep%gsp = 13.932049_dp
         sep%gpp = 10.495551_dp
         sep%gp2 = 10.526198_dp
         sep%hsp = 0.293760_dp
      CASE (82)
         ! Lead
         sep%defined = .TRUE.
         sep%uss = -35.038145_dp
         sep%upp = -25.413401_dp
         sep%beta(0) = -8.323792_dp
         sep%beta(1) = -2.237891_dp
         sep%sto_exponents(0) = 2.368901_dp
         sep%sto_exponents(1) = 1.685246_dp
         sep%gss = 5.254128_dp
         sep%gsp = 7.061016_dp
         sep%gpp = 6.818551_dp
         sep%gp2 = 5.603019_dp
         sep%hsp = 1.018819_dp
         sep%a = -0.239463_dp
         sep%b = 5.444338_dp
         sep%c = 1.613682_dp
      CASE (83)
         ! Bismuth
         sep%defined = .TRUE.
         sep%uss = -42.409177_dp
         sep%upp = -36.393746_dp
         sep%beta(0) = -34.951578_dp
         sep%beta(1) = -7.359060_dp
         sep%sto_exponents(0) = 3.702377_dp
         sep%sto_exponents(1) = 1.872327_dp
         sep%gss = 5.851803_dp
         sep%gsp = 6.790583_dp
         sep%gpp = 8.389442_dp
         sep%gp2 = 7.724219_dp
         sep%hsp = 0.295606_dp
      CASE (84:nelem)
      END SELECT
      ! Finalize parameters for the element
      CALL valence_electrons(sep, extended_basis_set=.TRUE.)
      CALL calpar(z, sep)
      CALL convert_param_to_cp2k(sep)
      CALL init_pm6_pair_params()

      sep%aab(0:nelem) = aab_pm6(z, 0:nelem)
      sep%xab(0:nelem) = xab_pm6(z, 0:nelem)
      !    ! We seem to have also C-H as a special case
      !    IF (z == 1) sep%aab(6)= sep%aab(6)/bohr
      !    IF (z == 1) sep%aab(7)= sep%aab(7)/bohr
      !    IF (z == 1) sep%aab(8)= sep%aab(8)/bohr
      !    IF (z == 6) sep%aab(1)= sep%aab(1)/bohr
      !    IF (z == 7) sep%aab(1)= sep%aab(1)/bohr
      !    IF (z == 8) sep%aab(1)= sep%aab(1)/bohr

   END SUBROUTINE pm6_default_parameter

! **************************************************************************************************
!> \brief Default parameter sets for semi empirical models: PDDG
!> \param sep ...
!> \param z ...
! **************************************************************************************************
   SUBROUTINE pdg_default_parameter(sep, z)
      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER                                            :: z

      CHARACTER(len=*), PARAMETER :: routineN = 'pdg_default_parameter', &
         routineP = moduleN//':'//routineN

      sep%typ = do_method_pdg
      sep%name = ptable(z)%name
      sep%eheat = ptable(z)%heat_of_formation/kcalmol
      sep%z = z

      SELECT CASE (z)
      CASE DEFAULT
         CALL cp__b("semi_empirical_parameters.F",3887,"invalid atomic number")
      CASE (0)
         sep%defined = .TRUE.
      CASE (1)
         !Hydrogen
         sep%defined = .TRUE.
         sep%uss = -12.893272_dp
         sep%beta(0) = -6.152654_dp
         sep%sto_exponents(0) = 0.972786_dp
         sep%alp = 3.381686_dp
         sep%eisol = -13.120566_dp
         sep%gss = 14.794208_dp
         sep%am = 0.5_dp/0.919616_dp
         sep%ad = 0.5_dp/0.919616_dp
         sep%aq = 0.5_dp/0.919616_dp
         sep%fn1(1) = 1.122244_dp
         sep%fn2(1) = 4.707790_dp
         sep%fn3(1) = 1.547099_dp
         sep%fn1(2) = -1.069737_dp
         sep%fn2(2) = 5.857995_dp
         sep%fn3(2) = 1.567893_dp
         sep%pre(1) = 0.057193_dp
         sep%pre(2) = -0.034823_dp
         sep%d(1) = 0.663395_dp
         sep%d(2) = 1.081901_dp
      CASE (2:5)
      CASE (6)
         !Carbon
         sep%defined = .TRUE.
         sep%uss = -48.241241_dp
         sep%upp = -36.461256_dp
         sep%beta(0) = -11.952818_dp
         sep%beta(1) = -9.922411_dp
         sep%sto_exponents(0) = 1.567864_dp
         sep%sto_exponents(1) = 1.846659_dp
         sep%alp = 2.725772_dp
         sep%eisol = -113.428242_dp
         sep%gss = 11.200708_dp
         sep%gsp = 10.265027_dp
         sep%gpp = 10.796292_dp
         sep%gp2 = 9.042566_dp
         sep%hsp = 2.290980_dp
         sep%dd = 0.831413_dp
         sep%qq = 0.663222_dp
         sep%am = 0.5_dp/1.214657_dp
         sep%ad = 0.5_dp/0.848467_dp
         sep%aq = 0.5_dp/0.652785_dp
         sep%fn1(1) = 0.048906_dp
         sep%fn2(1) = 5.765340_dp
         sep%fn3(1) = 1.682232_dp
         sep%fn1(2) = 0.047697_dp
         sep%fn2(2) = 5.973721_dp
         sep%fn3(2) = 0.894406_dp
         sep%pre(1) = -0.000743_dp
         sep%pre(2) = 0.000985_dp
         sep%d(1) = 0.836915_dp
         sep%d(2) = 1.585236_dp
      CASE (7)
         !Nitrogen
         sep%defined = .TRUE.
         sep%uss = -49.454546_dp
         sep%upp = -47.757406_dp
         sep%beta(0) = -14.117230_dp
         sep%beta(1) = -19.938509_dp
         sep%sto_exponents(0) = 2.035807_dp
         sep%sto_exponents(1) = 2.324327_dp
         sep%alp = 2.849124_dp
         sep%eisol = -158.416205_dp
         sep%gss = 11.904787_dp
         sep%gsp = 7.348565_dp
         sep%gpp = 11.754672_dp
         sep%gp2 = 10.807277_dp
         sep%hsp = 1.136713_dp
         sep%dd = 0.654855_dp
         sep%qq = 0.526924_dp
         sep%am = 0.5_dp/1.142818_dp
         sep%ad = 0.5_dp/0.991235_dp
         sep%aq = 0.5_dp/0.676704_dp
         sep%fn1(1) = 1.513320_dp
         sep%fn2(1) = 5.904394_dp
         sep%fn3(1) = 1.728376_dp
         sep%fn1(2) = -1.511892_dp
         sep%fn2(2) = 6.030014_dp
         sep%fn3(2) = 1.734108_dp
         sep%pre(1) = -0.003160_dp
         sep%pre(2) = 0.012501_dp
         sep%d(1) = 1.004172_dp
         sep%d(2) = 1.516336_dp
      CASE (8)
         !Oxygen
         sep%defined = .TRUE.
         sep%uss = -87.412505_dp
         sep%upp = -72.183070_dp
         sep%beta(0) = -44.874553_dp
         sep%beta(1) = -24.601939_dp
         sep%sto_exponents(0) = 3.814565_dp
         sep%sto_exponents(1) = 2.318011_dp
         sep%alp = 3.225309_dp
         sep%eisol = -292.188766_dp
         sep%gss = 15.755760_dp
         sep%gsp = 10.621160_dp
         sep%gpp = 13.654016_dp
         sep%gp2 = 12.406095_dp
         sep%hsp = 0.593883_dp
         sep%dd = 0.403741_dp
         sep%qq = 0.528360_dp
         sep%am = 0.5_dp/0.863494_dp
         sep%ad = 0.5_dp/0.936266_dp
         sep%aq = 0.5_dp/0.624291_dp
         sep%fn1(1) = -1.138455_dp
         sep%fn2(1) = 6.000043_dp
         sep%fn3(1) = 1.622362_dp
         sep%fn1(2) = 1.146007_dp
         sep%fn2(2) = 5.963494_dp
         sep%fn3(2) = 1.614788_dp
         sep%pre(1) = -0.001000_dp
         sep%pre(2) = -0.001522_dp
         sep%d(1) = 1.360685_dp
         sep%d(2) = 1.366407_dp
      CASE (9)
         !Fluorine
         sep%defined = .TRUE.
         sep%uss = -110.400432_dp
         sep%upp = -106.395264_dp
         sep%beta(0) = -50.937301_dp
         sep%beta(1) = -31.636976_dp
         sep%sto_exponents(0) = 5.538033_dp
         sep%sto_exponents(1) = 2.538066_dp
         sep%alp = 3.200571_dp
         sep%eisol = -442.457133_dp
         sep%gss = 10.496667_dp
         sep%gsp = 16.073689_dp
         sep%gpp = 14.817256_dp
         sep%gp2 = 14.418393_dp
         sep%hsp = 0.727763_dp
         sep%dd = 0.246601_dp
         sep%qq = 0.482551_dp
         sep%am = 0.5_dp/1.296126_dp
         sep%ad = 0.5_dp/0.6346333_dp
         sep%aq = 0.5_dp/0.805802_dp
         sep%fn1(1) = -0.008079_dp
         sep%fn2(1) = 5.938969_dp
         sep%fn3(1) = 1.863949_dp
         sep%fn1(2) = -0.002659_dp
         sep%fn2(2) = 5.925105_dp
         sep%fn3(2) = 2.388864_dp
         sep%pre(1) = -0.012866_dp
         sep%pre(2) = -0.007315_dp
         sep%d(1) = 1.305681_dp
         sep%d(2) = 1.842572_dp
      CASE (10:13)
      CASE (14)
         !Silicon
         sep%defined = .TRUE.
         sep%uss = -26.332522_dp
         sep%upp = -22.602540_dp
         sep%beta(0) = -3.376445_dp
         sep%beta(1) = -3.620969_dp
         sep%sto_exponents(0) = 1.586389_dp
         sep%sto_exponents(1) = 1.485958_dp
         sep%alp = 2.215157_dp
         sep%eisol = -66.839000_dp
         sep%gss = 5.047196_dp
         sep%gsp = 5.949057_dp
         sep%gpp = 6.759367_dp
         sep%gp2 = 5.161297_dp
         sep%hsp = 0.919832_dp
         sep%dd = 1.310515_dp
         sep%qq = 1.126089_dp
         sep%am = 0.5_dp/2.695556_dp
         sep%ad = 0.5_dp/1.630757_dp
         sep%aq = 0.5_dp/0.949200_dp
         sep%fn1(1) = -0.071314_dp
         sep%fn2(1) = 6.000000_dp
         sep%fn3(1) = 0.237995_dp
         sep%fn1(2) = 0.089451_dp
         sep%fn2(2) = 6.000000_dp
         sep%fn3(2) = 1.897728_dp
         sep%pre(1) = -0.091928_dp
         sep%pre(2) = -0.040753_dp
         sep%d(1) = 1.163190_dp
         sep%d(2) = 2.190526_dp
      CASE (15)
         !Phospherus
         sep%defined = .TRUE.
         sep%uss = -37.882113_dp
         sep%upp = -30.312979_dp
         sep%beta(0) = -12.676297_dp
         sep%beta(1) = -7.093318_dp
         sep%sto_exponents(0) = 2.395882_dp
         sep%sto_exponents(1) = 1.742213_dp
         sep%alp = 2.005294_dp
         sep%eisol = -117.212854_dp
         sep%gss = 7.801615_dp
         sep%gsp = 5.186949_dp
         sep%gpp = 6.618478_dp
         sep%gp2 = 6.062002_dp
         sep%hsp = 1.542809_dp
         sep%dd = 0.893978_dp
         sep%qq = 0.960457_dp
         sep%am = 0.5_dp/1.743870_dp
         sep%ad = 0.5_dp/1.050851_dp
         sep%aq = 0.5_dp/1.208907_dp
         sep%fn1(1) = -0.398055_dp
         sep%fn2(1) = 1.997272_dp
         sep%fn3(1) = 0.950073_dp
         sep%fn1(2) = -0.079653_dp
         sep%fn2(2) = 1.998360_dp
         sep%fn3(2) = 2.336959_dp
         sep%pre(1) = 0.462741_dp
         sep%pre(2) = -0.020444_dp
         sep%d(1) = 0.714296_dp
         sep%d(2) = 2.041209_dp
      CASE (16)
         !Sulfur
         sep%defined = .TRUE.
         sep%uss = -43.906366_dp
         sep%upp = -43.461348_dp
         sep%beta(0) = -2.953912_dp
         sep%beta(1) = -8.507779_dp
         sep%sto_exponents(0) = 1.012002_dp
         sep%sto_exponents(1) = 1.876999_dp
         sep%alp = 2.539751_dp
         sep%eisol = -166.336554_dp
         sep%gss = 8.964667_dp
         sep%gsp = 6.785936_dp
         sep%gpp = 9.968164_dp
         sep%gp2 = 7.970247_dp
         sep%hsp = 4.041836_dp
         sep%dd = 1.006989_dp
         sep%qq = 0.891487_dp
         sep%am = 0.5_dp/1.517625_dp
         sep%ad = 0.5_dp/0.711672_dp
         sep%aq = 0.5_dp/0.754336_dp
         sep%fn1(1) = -0.330692_dp
         sep%fn2(1) = 6.000000_dp
         sep%fn3(1) = 0.823837_dp
         sep%fn1(2) = 0.024171_dp
         sep%fn2(2) = 6.000000_dp
         sep%fn3(2) = 2.017756_dp
         sep%pre(1) = 0.120434_dp
         sep%pre(2) = -0.002663_dp
         sep%d(1) = 0.672870_dp
         sep%d(2) = 2.032340_dp
      CASE (17)
         !Chlorine
         sep%defined = .TRUE.
         sep%uss = -95.094434_dp
         sep%upp = -53.921651_dp
         sep%beta(0) = -26.913129_dp
         sep%beta(1) = -14.991178_dp
         sep%sto_exponents(0) = 2.548268_dp
         sep%sto_exponents(1) = 2.284624_dp
         sep%sto_exponents(2) = 1.000000_dp
         sep%alp = 2.497617_dp
         sep%eisol = -305.715201_dp
         sep%gss = 16.013601_dp
         sep%gsp = 8.048115_dp
         sep%gpp = 7.522215_dp
         sep%gp2 = 7.504154_dp
         sep%hsp = 3.481153_dp
         sep%dd = 0.827561_dp
         sep%qq = 0.732427_dp
         sep%am = 0.5_dp/0.8495900_dp
         sep%ad = 0.5_dp/0.696164_dp
         sep%aq = 0.5_dp/2.299104_dp
         sep%fn1(1) = -0.112222_dp
         sep%fn2(1) = 5.963719_dp
         sep%fn3(1) = 1.027719_dp
         sep%fn1(2) = -0.013061_dp
         sep%fn2(2) = 1.999556_dp
         sep%fn3(2) = 2.286377_dp
         sep%pre(1) = -0.016552_dp
         sep%pre(2) = -0.016646_dp
         sep%d(1) = 1.727690_dp
         sep%d(2) = 1.784655_dp
      CASE (18:34)
      CASE (35)
         !Bromine
         sep%defined = .TRUE.
         sep%uss = -115.841963_dp
         sep%upp = -74.205146_dp
         sep%beta(0) = -21.538044_dp
         sep%beta(1) = -8.524764_dp
         sep%sto_exponents(0) = 4.345079_dp
         sep%sto_exponents(1) = 2.190961_dp
         sep%sto_exponents(2) = 1.000000_dp
         sep%alp = 2.424673_dp
         sep%eisol = -351.013887_dp
         sep%gss = 15.943425_dp
         sep%gsp = 16.061680_dp
         sep%gpp = 8.282763_dp
         sep%gp2 = 7.816849_dp
         sep%hsp = 0.578869_dp
         sep%dd = 0.473860_dp
         sep%qq = 0.968214_dp
         sep%am = 0.5_dp/0.853330_dp
         sep%ad = 0.5_dp/1.046430_dp
         sep%aq = 0.5_dp/1.280643_dp
         sep%fn1(1) = 0.9613620_dp
         sep%fn2(1) = 6.0136000_dp
         sep%fn3(1) = 2.3404450_dp
         sep%fn1(2) = -0.9548834_dp
         sep%fn2(2) = 5.9763290_dp
         sep%fn3(2) = 2.3487450_dp
         sep%pre(1) = -0.013772_dp
         sep%pre(2) = 0.008849_dp
         sep%d(1) = 1.852030_dp
         sep%d(2) = 2.338958_dp
      CASE (36:52)
      CASE (53)
         !Iodine
         sep%defined = .TRUE.
         sep%uss = -96.664174_dp
         sep%upp = -61.167137_dp
         sep%beta(0) = -16.592621_dp
         sep%beta(1) = -6.599816_dp
         sep%sto_exponents(0) = 5.062801_dp
         sep%sto_exponents(1) = 2.417757_dp
         sep%sto_exponents(2) = 1.000000_dp
         sep%alp = 1.978170_dp
         sep%eisol = -291.537869_dp
         sep%gss = 13.631943_dp
         sep%gsp = 14.990406_dp
         sep%gpp = 7.288330_dp
         sep%gp2 = 5.966407_dp
         sep%hsp = 2.630035_dp
         sep%dd = 0.407261_dp
         sep%qq = 1.062574_dp
         sep%am = 0.5_dp/0.998024_dp
         sep%ad = 0.5_dp/0.532290_dp
         sep%aq = 0.5_dp/0.979783_dp
         sep%fn1(1) = -0.136003_dp
         sep%fn2(1) = 3.852912_dp
         sep%fn3(1) = 1.697455_dp
         sep%fn1(2) = -0.037287_dp
         sep%fn2(2) = 5.229264_dp
         sep%fn3(2) = 2.768669_dp
         sep%pre(1) = 0.012901_dp
         sep%pre(2) = -0.012825_dp
         sep%d(1) = 1.994299_dp
         sep%d(2) = 2.263417_dp
      CASE (54:nelem)
      END SELECT
      ! Finalize parameters for the element
      CALL valence_electrons(sep, extended_basis_set=.FALSE.)
      CALL calpar(z, sep)
      CALL convert_param_to_cp2k(sep)

   END SUBROUTINE pdg_default_parameter

! **************************************************************************************************
!> \brief Default pair parameter for: PM6
! **************************************************************************************************
   SUBROUTINE init_pm6_pair_params()

      CHARACTER(len=*), PARAMETER :: routineN = 'init_pm6_pair_params', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      INTEGER, SAVE                                      :: init = 0
      REAL(dp)                                           :: aab, xab

      IF (init /= 0) RETURN

      xab_pm6 = 0._dp
      aab_pm6 = 0._dp

      xab_pm6(1, 1) = 2.2435870171_dp
      xab_pm6(2, 1) = 2.3711988926_dp
      xab_pm6(2, 2) = 3.4509000778_dp
      xab_pm6(3, 1) = 2.1919848919_dp
      xab_pm6(3, 2) = 9.2736759186_dp
      xab_pm6(3, 3) = 16.1163845062_dp
      xab_pm6(4, 1) = 2.5628309250_dp
      xab_pm6(4, 2) = 12.5448780060_dp
      xab_pm6(4, 3) = 3.2871649265_dp
      xab_pm6(4, 4) = 0.2386330068_dp
      xab_pm6(5, 1) = 1.3213939667_dp
      xab_pm6(5, 2) = 1.9741699696_dp
      xab_pm6(5, 3) = 7.8860177994_dp
      xab_pm6(5, 4) = 1.1517920494_dp
      xab_pm6(5, 5) = 3.5936191082_dp
      xab_pm6(6, 1) = 0.2165060000_dp
      xab_pm6(6, 2) = 3.2139708996_dp
      xab_pm6(6, 3) = 16.1800022125_dp
      xab_pm6(6, 4) = 25.0358791351_dp
      xab_pm6(6, 5) = 1.8748589754_dp
      xab_pm6(6, 6) = 0.8135100007_dp
      xab_pm6(7, 1) = 0.1755059958_dp
      xab_pm6(7, 2) = 1.0778609514_dp
      xab_pm6(7, 3) = 2.8234028816_dp
      xab_pm6(7, 4) = 1.7406049967_dp
      xab_pm6(7, 5) = 0.9528819919_dp
      xab_pm6(7, 6) = 0.8599489927_dp
      xab_pm6(7, 7) = 0.6753129959_dp
      xab_pm6(8, 1) = 0.1922950000_dp
      xab_pm6(8, 2) = 6.6845250130_dp
      xab_pm6(8, 3) = 1.9685980082_dp
      xab_pm6(8, 4) = 3.2181549072_dp
      xab_pm6(8, 5) = 1.2698010206_dp
      xab_pm6(8, 6) = 0.9902110100_dp
      xab_pm6(8, 7) = 0.7647560239_dp
      xab_pm6(8, 8) = 0.5351120234_dp
      xab_pm6(9, 1) = 0.8158019781_dp
      xab_pm6(9, 2) = 0.7451069951_dp
      xab_pm6(9, 3) = 1.9759850502_dp
      xab_pm6(9, 4) = 3.8829929829_dp
      xab_pm6(9, 5) = 0.8627610207_dp
      xab_pm6(9, 6) = 0.7329679728_dp
      xab_pm6(9, 7) = 0.6358540058_dp
      xab_pm6(9, 8) = 0.6742510200_dp
      xab_pm6(9, 9) = 0.6813430190_dp
      xab_pm6(10, 1) = 5.5350208282_dp
      xab_pm6(10, 2) = 1.9609240294_dp
      xab_pm6(10, 3) = 0.7049580216_dp
      xab_pm6(10, 4) = 0.3926280141_dp
      xab_pm6(10, 5) = 2.7641398907_dp
      xab_pm6(10, 6) = 5.4687800407_dp
      xab_pm6(10, 7) = 29.9996089935_dp
      xab_pm6(10, 8) = 0.7638990283_dp
      xab_pm6(10, 9) = 2.7067539692_dp
      xab_pm6(10, 10) = 2.7948300838_dp
      xab_pm6(11, 1) = 0.2078309953_dp
      xab_pm6(11, 2) = 4.2825169563_dp
      xab_pm6(11, 3) = 0.8814820051_dp
      xab_pm6(11, 4) = 3.1216199398_dp
      xab_pm6(11, 5) = 3.1886079311_dp
      xab_pm6(11, 6) = 4.5204291344_dp
      xab_pm6(11, 7) = 8.5863866806_dp
      xab_pm6(11, 8) = 3.2700788975_dp
      xab_pm6(11, 9) = 7.0473508835_dp
      xab_pm6(11, 10) = 1.3430370092_dp
      xab_pm6(11, 11) = 0.2871370018_dp
      xab_pm6(12, 1) = 7.7582368851_dp
      xab_pm6(12, 2) = 3.7258501053_dp
      xab_pm6(12, 3) = 2.4902501106_dp
      xab_pm6(12, 4) = 2.0663919449_dp
      xab_pm6(12, 5) = 6.1467008591_dp
      xab_pm6(12, 6) = 10.5176897049_dp
      xab_pm6(12, 7) = 1.2080750465_dp
      xab_pm6(12, 8) = 1.5357340574_dp
      xab_pm6(12, 9) = 5.8590230942_dp
      xab_pm6(12, 10) = 1.2148590088_dp
      xab_pm6(12, 11) = 8.6756191254_dp
      xab_pm6(12, 12) = 0.4656449854_dp
      xab_pm6(13, 1) = 2.9583790302_dp
      xab_pm6(13, 2) = 2.7014000416_dp
      xab_pm6(13, 3) = 1.1068190336_dp
      xab_pm6(13, 4) = 5.0372138023_dp
      xab_pm6(13, 5) = 2.7414789200_dp
      xab_pm6(13, 6) = 2.9280560017_dp
      xab_pm6(13, 7) = 1.3452019691_dp
      xab_pm6(13, 8) = 2.1313960552_dp
      xab_pm6(13, 9) = 1.9756350517_dp
      xab_pm6(13, 10) = 1.7092000246_dp
      xab_pm6(13, 11) = 2.0718469620_dp
      xab_pm6(13, 12) = 13.4724426270_dp
      xab_pm6(13, 13) = 2.1391999722_dp
      xab_pm6(14, 1) = 0.9241960049_dp
      xab_pm6(14, 2) = 1.8535829782_dp
      xab_pm6(14, 3) = 3.0907909870_dp
      xab_pm6(14, 4) = 0.6234329939_dp
      xab_pm6(14, 5) = 1.0282870531_dp
      xab_pm6(14, 6) = 0.7857450247_dp
      xab_pm6(14, 7) = 0.5929719806_dp
      xab_pm6(14, 8) = 0.7510949969_dp
      xab_pm6(14, 9) = 0.5435159802_dp
      xab_pm6(14, 10) = 14.3786764145_dp
      xab_pm6(14, 11) = 9.2376441956_dp
      xab_pm6(14, 12) = 29.9945201874_dp
      xab_pm6(14, 13) = 2.0000000000_dp
      xab_pm6(14, 14) = 0.2734769881_dp
      xab_pm6(15, 1) = 1.2349859476_dp
      xab_pm6(15, 2) = 1.4902180433_dp
      xab_pm6(15, 3) = 1.1229499578_dp
      xab_pm6(15, 4) = 1.6848310232_dp
      xab_pm6(15, 5) = 1.4508860111_dp
      xab_pm6(15, 6) = 0.9795119762_dp
      xab_pm6(15, 7) = 0.9721540213_dp
      xab_pm6(15, 8) = 0.8787050247_dp
      xab_pm6(15, 9) = 0.5145750046_dp
      xab_pm6(15, 10) = 0.7749540210_dp
      xab_pm6(15, 11) = 2.8370950222_dp
      xab_pm6(15, 12) = 1.1778810024_dp
      xab_pm6(15, 13) = 5.0508160591_dp
      xab_pm6(15, 14) = 13.2391214371_dp
      xab_pm6(15, 15) = 0.9025009871_dp
      xab_pm6(16, 1) = 0.8497120142_dp
      xab_pm6(16, 2) = 0.4376179874_dp
      xab_pm6(16, 3) = 2.6425020695_dp
      xab_pm6(16, 4) = 3.7915649414_dp
      xab_pm6(16, 5) = 1.1253939867_dp
      xab_pm6(16, 6) = 0.6668490171_dp
      xab_pm6(16, 7) = 0.7387099862_dp
      xab_pm6(16, 8) = 0.7472149730_dp
      xab_pm6(16, 9) = 0.3752509952_dp
      xab_pm6(16, 10) = 3.2961599827_dp
      xab_pm6(16, 11) = 0.8524339795_dp
      xab_pm6(16, 12) = 0.5007479787_dp
      xab_pm6(16, 13) = 2.3473839760_dp
      xab_pm6(16, 14) = 0.8766580224_dp
      xab_pm6(16, 15) = 0.5622659922_dp
      xab_pm6(16, 16) = 0.4738560021_dp
      xab_pm6(17, 1) = 0.7548310161_dp
      xab_pm6(17, 2) = 0.2729640007_dp
      xab_pm6(17, 3) = 4.2277941704_dp
      xab_pm6(17, 4) = 2.5072751045_dp
      xab_pm6(17, 5) = 0.8221290112_dp
      xab_pm6(17, 6) = 0.5157870054_dp
      xab_pm6(17, 7) = 0.5207449794_dp
      xab_pm6(17, 8) = 0.5855100155_dp
      xab_pm6(17, 9) = 0.4111239910_dp
      xab_pm6(17, 10) = 0.1251329929_dp
      xab_pm6(17, 11) = 1.3578939438_dp
      xab_pm6(17, 12) = 2.4308559895_dp
      xab_pm6(17, 13) = 2.1534509659_dp
      xab_pm6(17, 14) = 0.5130000114_dp
      xab_pm6(17, 15) = 0.3523609936_dp
      xab_pm6(17, 16) = 0.3569709957_dp
      xab_pm6(17, 17) = 0.3329190016_dp
      xab_pm6(18, 1) = 3.9334449768_dp
      xab_pm6(18, 2) = 1.1772110462_dp
      xab_pm6(18, 3) = 3.3629100323_dp
      xab_pm6(18, 4) = 2.7554919720_dp
      xab_pm6(18, 5) = 1.9315860271_dp
      xab_pm6(18, 6) = 0.1223089993_dp
      xab_pm6(18, 7) = 0.5625810027_dp
      xab_pm6(18, 8) = 0.3557949960_dp
      xab_pm6(18, 9) = 9.2697153091_dp
      xab_pm6(18, 10) = 1.3046970367_dp
      xab_pm6(18, 11) = 3.3981380463_dp
      xab_pm6(18, 12) = 1.9706380367_dp
      xab_pm6(18, 13) = 1.8520090580_dp
      xab_pm6(18, 14) = 1.0678900480_dp
      xab_pm6(18, 15) = 0.1710139960_dp
      xab_pm6(18, 16) = 0.6537690163_dp
      xab_pm6(18, 17) = 2.2560939789_dp
      xab_pm6(18, 18) = 0.9726989865_dp
      xab_pm6(19, 1) = 0.3693400025_dp
      xab_pm6(19, 2) = 2.8950450420_dp
      xab_pm6(19, 3) = 4.3745670319_dp
      xab_pm6(19, 4) = 6.7322211266_dp
      xab_pm6(19, 5) = 8.9005413055_dp
      xab_pm6(19, 6) = 10.3179874420_dp
      xab_pm6(19, 7) = 7.9777069092_dp
      xab_pm6(19, 8) = 1.0122749805_dp
      xab_pm6(19, 9) = 6.5929708481_dp
      xab_pm6(19, 10) = 0.2339950055_dp
      xab_pm6(19, 11) = 5.5630269051_dp
      xab_pm6(19, 12) = 3.2905020714_dp
      xab_pm6(19, 13) = 29.9447078705_dp
      xab_pm6(19, 14) = 8.2791996002_dp
      xab_pm6(19, 15) = 4.4753842354_dp
      xab_pm6(19, 16) = 29.5289516449_dp
      xab_pm6(19, 17) = 1.2314809561_dp
      xab_pm6(19, 18) = 9.7105083466_dp
      xab_pm6(19, 19) = 5.9343290329_dp
      xab_pm6(20, 1) = 7.7286062241_dp
      xab_pm6(20, 2) = 2.9138519764_dp
      xab_pm6(20, 5) = 1.7000099421_dp
      xab_pm6(20, 6) = 0.1484500021_dp
      xab_pm6(20, 7) = 2.9880740643_dp
      xab_pm6(20, 8) = 17.0289459229_dp
      xab_pm6(20, 9) = 3.4828209877_dp
      xab_pm6(20, 10) = 0.3325859904_dp
      xab_pm6(20, 11) = 9.6575088501_dp
      xab_pm6(20, 12) = 8.5998001099_dp
      xab_pm6(20, 13) = 4.1885547638_dp
      xab_pm6(20, 14) = 0.3362329900_dp
      xab_pm6(20, 15) = 0.4108400047_dp
      xab_pm6(20, 16) = 0.3257389963_dp
      xab_pm6(20, 17) = 5.9561438560_dp
      xab_pm6(20, 18) = 0.2910720110_dp
      xab_pm6(20, 19) = 1.2403199673_dp
      xab_pm6(20, 20) = 30.0035915375_dp
      xab_pm6(21, 1) = 0.3511990011_dp
      xab_pm6(21, 6) = 8.6080522537_dp
      xab_pm6(21, 7) = 3.2318809032_dp
      xab_pm6(21, 8) = 3.0586719513_dp
      xab_pm6(21, 9) = 7.2523469925_dp
      xab_pm6(21, 13) = 0.5006200075_dp
      xab_pm6(21, 14) = 3.2190699577_dp
      xab_pm6(21, 15) = 0.6267489791_dp
      xab_pm6(21, 16) = 0.2118500024_dp
      xab_pm6(21, 17) = 2.9961290359_dp
      xab_pm6(21, 21) = 2.5981659889_dp
      xab_pm6(22, 1) = 0.1437219977_dp
      xab_pm6(22, 5) = 0.6493600011_dp
      xab_pm6(22, 6) = 0.4167059958_dp
      xab_pm6(22, 7) = 0.5454609990_dp
      xab_pm6(22, 8) = 0.7994859815_dp
      xab_pm6(22, 9) = 1.0857419968_dp
      xab_pm6(22, 12) = 4.3302397728_dp
      xab_pm6(22, 13) = 2.0918409824_dp
      xab_pm6(22, 14) = 6.7738151550_dp
      xab_pm6(22, 15) = 4.1504998207_dp
      xab_pm6(22, 16) = 0.9437839985_dp
      xab_pm6(22, 17) = 0.3332970142_dp
      xab_pm6(22, 20) = 4.1091408730_dp
      xab_pm6(22, 22) = 2.0000000000_dp
      xab_pm6(23, 1) = 0.1052040011_dp
      xab_pm6(23, 6) = 1.9387600422_dp
      xab_pm6(23, 7) = 0.2767249942_dp
      xab_pm6(23, 8) = 0.4153119922_dp
      xab_pm6(23, 9) = 0.3428150117_dp
      xab_pm6(23, 11) = 8.2760200500_dp
      xab_pm6(23, 15) = 6.2506241798_dp
      xab_pm6(23, 16) = 2.0350389481_dp
      xab_pm6(23, 17) = 0.2436569929_dp
      xab_pm6(23, 19) = 2.0265901089_dp
      xab_pm6(23, 23) = 10.7798919678_dp
      xab_pm6(24, 1) = 0.0444689989_dp
      xab_pm6(24, 6) = 6.1101870537_dp
      xab_pm6(24, 7) = 1.9203239679_dp
      xab_pm6(24, 8) = 1.0555109978_dp
      xab_pm6(24, 9) = 0.7376070023_dp
      xab_pm6(24, 11) = 8.3642740250_dp
      xab_pm6(24, 14) = 1.0291099548_dp
      xab_pm6(24, 15) = 0.6001769900_dp
      xab_pm6(24, 16) = 0.5503339767_dp
      xab_pm6(24, 17) = 0.3690730035_dp
      xab_pm6(24, 19) = 2.0000000000_dp
      xab_pm6(24, 24) = 10.3186073303_dp
      xab_pm6(25, 1) = 1.2692099810_dp
      xab_pm6(25, 6) = 2.5831100941_dp
      xab_pm6(25, 7) = 1.9567500353_dp
      xab_pm6(25, 8) = 1.2856199741_dp
      xab_pm6(25, 9) = 1.1130700111_dp
      xab_pm6(25, 13) = 1.0407899618_dp
      xab_pm6(25, 14) = 0.9505800009_dp
      xab_pm6(25, 15) = 1.1303199530_dp
      xab_pm6(25, 16) = 1.6126500368_dp
      xab_pm6(25, 17) = 0.2018499970_dp
      xab_pm6(25, 20) = 0.6201800108_dp
      xab_pm6(25, 25) = 2.4600400925_dp
      xab_pm6(26, 1) = 0.0251950007_dp
      xab_pm6(26, 6) = 0.3668349981_dp
      xab_pm6(26, 7) = 0.1553419977_dp
      xab_pm6(26, 8) = 0.1364219934_dp
      xab_pm6(26, 9) = 3.6573500633_dp
      xab_pm6(26, 15) = 0.4312910140_dp
      xab_pm6(26, 16) = 0.0334779993_dp
      xab_pm6(26, 17) = 0.0194729995_dp
      xab_pm6(26, 19) = 6.0000000000_dp
      xab_pm6(26, 26) = 1.8468899727_dp
      xab_pm6(27, 1) = 2.4724650383_dp
      xab_pm6(27, 6) = 2.1239299774_dp
      xab_pm6(27, 7) = 2.6538360119_dp
      xab_pm6(27, 8) = 5.2520217896_dp
      xab_pm6(27, 9) = 4.5850300789_dp
      xab_pm6(27, 14) = 1.0902400017_dp
      xab_pm6(27, 15) = 0.1059359983_dp
      xab_pm6(27, 16) = 0.4367069900_dp
      xab_pm6(27, 17) = 1.0334140062_dp
      xab_pm6(27, 27) = 3.9196178913_dp
      xab_pm6(28, 1) = 1.7631239891_dp
      xab_pm6(28, 6) = 7.1333241463_dp
      xab_pm6(28, 7) = 4.2867999077_dp
      xab_pm6(28, 8) = 0.8859419823_dp
      xab_pm6(28, 9) = 1.0882079601_dp
      xab_pm6(28, 14) = 0.9386460185_dp
      xab_pm6(28, 15) = 5.0597271919_dp
      xab_pm6(28, 16) = 0.2748520076_dp
      xab_pm6(28, 17) = 0.2023130059_dp
      xab_pm6(28, 28) = 0.0354739986_dp
      xab_pm6(29, 1) = 0.6035910249_dp
      xab_pm6(29, 6) = 7.0677938461_dp
      xab_pm6(29, 7) = 3.2286670208_dp
      xab_pm6(29, 8) = 2.0000000000_dp
      xab_pm6(29, 9) = 1.2821079493_dp
      xab_pm6(29, 15) = 0.0201260000_dp
      xab_pm6(29, 16) = 0.0052479999_dp
      xab_pm6(29, 17) = 0.1390649974_dp
      xab_pm6(29, 29) = 5.1843757629_dp
      xab_pm6(30, 1) = 3.1091930866_dp
      xab_pm6(30, 6) = 0.9914649725_dp
      xab_pm6(30, 7) = 0.9524760246_dp
      xab_pm6(30, 8) = 2.2653129101_dp
      xab_pm6(30, 9) = 1.2255450487_dp
      xab_pm6(30, 14) = 3.7839050293_dp
      xab_pm6(30, 15) = 0.5815299749_dp
      xab_pm6(30, 16) = 0.6480000019_dp
      xab_pm6(30, 17) = 0.7213510275_dp
      xab_pm6(30, 20) = 1.2402900457_dp
      xab_pm6(30, 30) = 0.4650000036_dp
      xab_pm6(31, 1) = 1.3866519928_dp
      xab_pm6(31, 6) = 1.9629900455_dp
      xab_pm6(31, 7) = 1.1883380413_dp
      xab_pm6(31, 8) = 1.5236439705_dp
      xab_pm6(31, 9) = 1.4169420004_dp
      xab_pm6(31, 14) = 1.0022900105_dp
      xab_pm6(31, 15) = 0.5000000000_dp
      xab_pm6(31, 16) = 2.4562840462_dp
      xab_pm6(31, 17) = 1.1866610050_dp
      xab_pm6(31, 31) = 1.1983939409_dp
      xab_pm6(32, 1) = 1.7332259417_dp
      xab_pm6(32, 6) = 1.2975100279_dp
      xab_pm6(32, 7) = 0.6375060081_dp
      xab_pm6(32, 8) = 0.8269640207_dp
      xab_pm6(32, 9) = 0.6519770026_dp
      xab_pm6(32, 14) = 0.1786800027_dp
      xab_pm6(32, 15) = 5.6163492203_dp
      xab_pm6(32, 16) = 1.1609569788_dp
      xab_pm6(32, 17) = 0.5452389717_dp
      xab_pm6(32, 25) = 2.2551510334_dp
      xab_pm6(32, 27) = 2.1518499851_dp
      xab_pm6(32, 32) = 3.0230000019_dp
      xab_pm6(33, 1) = 1.0905890465_dp
      xab_pm6(33, 6) = 0.5790979862_dp
      xab_pm6(33, 7) = 0.2733370066_dp
      xab_pm6(33, 8) = 0.7016140223_dp
      xab_pm6(33, 9) = 0.4026280046_dp
      xab_pm6(33, 13) = 1.0035799742_dp
      xab_pm6(33, 14) = 1.4307060242_dp
      xab_pm6(33, 16) = 1.0337840319_dp
      xab_pm6(33, 17) = 0.4544329941_dp
      xab_pm6(33, 22) = 1.5813169479_dp
      xab_pm6(33, 27) = 1.6752400398_dp
      xab_pm6(33, 30) = 3.1565709114_dp
      xab_pm6(33, 31) = 1.6862980127_dp
      xab_pm6(33, 33) = 0.7373070121_dp
      xab_pm6(34, 1) = 0.8479980230_dp
      xab_pm6(34, 6) = 1.1147869825_dp
      xab_pm6(34, 7) = 0.4828400016_dp
      xab_pm6(34, 8) = 0.9551609755_dp
      xab_pm6(34, 9) = 0.4448060095_dp
      xab_pm6(34, 14) = 0.5182269812_dp
      xab_pm6(34, 15) = 0.2920520008_dp
      xab_pm6(34, 16) = 0.3917210102_dp
      xab_pm6(34, 17) = 0.9810670018_dp
      xab_pm6(34, 25) = 2.1807200909_dp
      xab_pm6(34, 27) = 2.2024099827_dp
      xab_pm6(34, 30) = 0.5115939975_dp
      xab_pm6(34, 32) = 5.8720507622_dp
      xab_pm6(34, 33) = 0.7112609744_dp
      xab_pm6(34, 34) = 0.8218230009_dp
      xab_pm6(35, 1) = 0.8503779769_dp
      xab_pm6(35, 2) = 1.0620429516_dp
      xab_pm6(35, 3) = 1.8588659763_dp
      xab_pm6(35, 4) = 1.9409329891_dp
      xab_pm6(35, 5) = 1.2264200449_dp
      xab_pm6(35, 6) = 0.5706859827_dp
      xab_pm6(35, 7) = 30.0001335144_dp
      xab_pm6(35, 8) = 0.7065839767_dp
      xab_pm6(35, 9) = 0.2935000062_dp
      xab_pm6(35, 10) = 1.0061589479_dp
      xab_pm6(35, 11) = 1.7529369593_dp
      xab_pm6(35, 12) = 2.9162800312_dp
      xab_pm6(35, 13) = 2.3571300507_dp
      xab_pm6(35, 14) = 0.5895109773_dp
      xab_pm6(35, 15) = 0.4565210044_dp
      xab_pm6(35, 16) = 0.2866880000_dp
      xab_pm6(35, 17) = 0.3892379999_dp
      xab_pm6(35, 18) = 3.2626678944_dp
      xab_pm6(35, 19) = 3.3227949142_dp
      xab_pm6(35, 20) = 4.0529098511_dp
      xab_pm6(35, 21) = 2.0982511044_dp
      xab_pm6(35, 22) = 0.8834339976_dp
      xab_pm6(35, 23) = 0.6126980186_dp
      xab_pm6(35, 24) = 0.2178529948_dp
      xab_pm6(35, 25) = 1.1835800409_dp
      xab_pm6(35, 26) = 6.0619211197_dp
      xab_pm6(35, 27) = 0.4251480103_dp
      xab_pm6(35, 28) = 0.6321449876_dp
      xab_pm6(35, 29) = 0.7685170174_dp
      xab_pm6(35, 30) = 0.7470269799_dp
      xab_pm6(35, 31) = 1.2610360384_dp
      xab_pm6(35, 32) = 0.6277369857_dp
      xab_pm6(35, 33) = 0.5141530037_dp
      xab_pm6(35, 34) = 0.3193419874_dp
      xab_pm6(35, 35) = 0.6153079867_dp
      xab_pm6(36, 1) = 5.1258969307_dp
      xab_pm6(36, 2) = 0.6277009845_dp
      xab_pm6(36, 3) = 8.7586965561_dp
      xab_pm6(36, 4) = 10.2377958298_dp
      xab_pm6(36, 5) = 2.9467809200_dp
      xab_pm6(36, 6) = 0.6526229978_dp
      xab_pm6(36, 7) = 0.1996060014_dp
      xab_pm6(36, 8) = 0.0067329998_dp
      xab_pm6(36, 9) = 4.1344070435_dp
      xab_pm6(36, 10) = 1.4337220192_dp
      xab_pm6(36, 11) = 8.3544483185_dp
      xab_pm6(36, 12) = 0.8884360194_dp
      xab_pm6(36, 13) = 5.0917158127_dp
      xab_pm6(36, 14) = 0.5542500019_dp
      xab_pm6(36, 17) = 0.5202170014_dp
      xab_pm6(36, 18) = 0.5548740029_dp
      xab_pm6(36, 19) = 8.6097822189_dp
      xab_pm6(36, 20) = 0.8788909912_dp
      xab_pm6(36, 35) = 0.3080979884_dp
      xab_pm6(36, 36) = 0.0520990007_dp
      xab_pm6(37, 1) = 29.8616313934_dp
      xab_pm6(37, 2) = 1.8625849485_dp
      xab_pm6(37, 5) = 9.0404930115_dp
      xab_pm6(37, 6) = 29.9740314484_dp
      xab_pm6(37, 7) = 0.0246360004_dp
      xab_pm6(37, 8) = 1.1253499985_dp
      xab_pm6(37, 9) = 28.8152770996_dp
      xab_pm6(37, 10) = 7.7365632057_dp
      xab_pm6(37, 13) = 2.9924569130_dp
      xab_pm6(37, 16) = 0.9644110203_dp
      xab_pm6(37, 17) = 10.3844861984_dp
      xab_pm6(37, 18) = 18.4333286285_dp
      xab_pm6(37, 35) = 5.1762142181_dp
      xab_pm6(37, 36) = 15.3075027466_dp
      xab_pm6(37, 37) = 20.1476097107_dp
      xab_pm6(38, 1) = 12.9733161926_dp
      xab_pm6(38, 6) = 6.6546568871_dp
      xab_pm6(38, 7) = 6.8538661003_dp
      xab_pm6(38, 8) = 3.5613958836_dp
      xab_pm6(38, 9) = 10.9717054367_dp
      xab_pm6(38, 14) = 2.7647500038_dp
      xab_pm6(38, 15) = 2.5520999432_dp
      xab_pm6(38, 16) = 3.1296029091_dp
      xab_pm6(38, 17) = 3.7839550972_dp
      xab_pm6(38, 22) = 2.8172500134_dp
      xab_pm6(38, 35) = 2.7665669918_dp
      xab_pm6(38, 38) = 5.3721199036_dp
      xab_pm6(39, 1) = 0.6123989820_dp
      xab_pm6(39, 6) = 0.5043060184_dp
      xab_pm6(39, 7) = 1.6279029846_dp
      xab_pm6(39, 8) = 1.7429219484_dp
      xab_pm6(39, 9) = 4.4338088036_dp
      xab_pm6(39, 13) = 0.5006700158_dp
      xab_pm6(39, 14) = 3.2190299034_dp
      xab_pm6(39, 15) = 0.5416600108_dp
      xab_pm6(39, 16) = 0.3182219863_dp
      xab_pm6(39, 17) = 1.1549589634_dp
      xab_pm6(39, 35) = 1.0543160439_dp
      xab_pm6(39, 39) = 1.6917250156_dp
      xab_pm6(40, 1) = 0.5937319994_dp
      xab_pm6(40, 6) = 1.9991819859_dp
      xab_pm6(40, 7) = 0.9950450063_dp
      xab_pm6(40, 8) = 1.0575250387_dp
      xab_pm6(40, 9) = 0.8611419797_dp
      xab_pm6(40, 13) = 0.8740599751_dp
      xab_pm6(40, 14) = 1.7233430147_dp
      xab_pm6(40, 15) = 0.7483760118_dp
      xab_pm6(40, 16) = 2.4293239117_dp
      xab_pm6(40, 17) = 0.4430989921_dp
      xab_pm6(40, 35) = 0.8589090109_dp
      xab_pm6(40, 40) = 3.0777730942_dp
      xab_pm6(41, 1) = 3.6037790775_dp
      xab_pm6(41, 6) = 4.5754809380_dp
      xab_pm6(41, 7) = 1.2135870457_dp
      xab_pm6(41, 8) = 1.1847189665_dp
      xab_pm6(41, 9) = 3.6636819839_dp
      xab_pm6(41, 11) = 8.2760200500_dp
      xab_pm6(41, 15) = 6.2015070915_dp
      xab_pm6(41, 16) = 2.4600200653_dp
      xab_pm6(41, 17) = 1.8915569782_dp
      xab_pm6(41, 19) = 2.0265901089_dp
      xab_pm6(41, 35) = 1.9212690592_dp
      xab_pm6(41, 41) = 2.1223878860_dp
      xab_pm6(42, 1) = 0.9346860051_dp
      xab_pm6(42, 6) = 1.1907420158_dp
      xab_pm6(42, 7) = 0.6082680225_dp
      xab_pm6(42, 8) = 0.5112670064_dp
      xab_pm6(42, 9) = 0.6104289889_dp
      xab_pm6(42, 11) = 8.2865495682_dp
      xab_pm6(42, 15) = 1.5228459835_dp
      xab_pm6(42, 16) = 0.8304280043_dp
      xab_pm6(42, 17) = 0.4743250012_dp
      xab_pm6(42, 19) = 2.1423900127_dp
      xab_pm6(42, 24) = 1.7419430017_dp
      xab_pm6(42, 35) = 0.2259179950_dp
      xab_pm6(42, 42) = 0.6264619827_dp
      xab_pm6(43, 1) = 6.3103342056_dp
      xab_pm6(43, 6) = 3.9724390507_dp
      xab_pm6(43, 7) = 0.7271299958_dp
      xab_pm6(43, 8) = 1.0246160030_dp
      xab_pm6(43, 9) = 5.8117837906_dp
      xab_pm6(43, 16) = 1.4965020418_dp
      xab_pm6(43, 17) = 1.6515829563_dp
      xab_pm6(43, 32) = 2.1520600319_dp
      xab_pm6(43, 34) = 2.2026200294_dp
      xab_pm6(43, 35) = 3.8201301098_dp
      xab_pm6(44, 1) = 7.1379761696_dp
      xab_pm6(44, 6) = 1.1349359751_dp
      xab_pm6(44, 7) = 2.3340940475_dp
      xab_pm6(44, 8) = 2.9762790203_dp
      xab_pm6(44, 9) = 6.9471278191_dp
      xab_pm6(44, 14) = 0.8494300246_dp
      xab_pm6(44, 15) = 0.0569740012_dp
      xab_pm6(44, 16) = 1.0066829920_dp
      xab_pm6(44, 17) = 0.1265860051_dp
      xab_pm6(44, 32) = 2.1515600681_dp
      xab_pm6(44, 34) = 2.2021200657_dp
      xab_pm6(44, 35) = 0.6598809958_dp
      xab_pm6(44, 44) = 0.0978050008_dp
      xab_pm6(45, 1) = 2.3061070442_dp
      xab_pm6(45, 6) = 3.4880790710_dp
      xab_pm6(45, 7) = 4.0009469986_dp
      xab_pm6(45, 8) = 10.2986755371_dp
      xab_pm6(45, 9) = 9.0653839111_dp
      xab_pm6(45, 14) = 0.8500099778_dp
      xab_pm6(45, 15) = 1.0381410122_dp
      xab_pm6(45, 16) = 4.8164100647_dp
      xab_pm6(45, 17) = 3.5868649483_dp
      xab_pm6(45, 32) = 2.1521399021_dp
      xab_pm6(45, 34) = 2.2026998997_dp
      xab_pm6(45, 35) = 1.5101490021_dp
      xab_pm6(45, 45) = 2.0701138973_dp
      xab_pm6(46, 1) = 0.4432690144_dp
      xab_pm6(46, 6) = 9.8537149429_dp
      xab_pm6(46, 7) = 0.2497030050_dp
      xab_pm6(46, 8) = 0.2164029926_dp
      xab_pm6(46, 9) = 6.9453120232_dp
      xab_pm6(46, 13) = 1.0572899580_dp
      xab_pm6(46, 14) = 2.2251040936_dp
      xab_pm6(46, 15) = 0.0450170003_dp
      xab_pm6(46, 16) = 0.2552289963_dp
      xab_pm6(46, 17) = 2.9698910713_dp
      xab_pm6(46, 35) = 4.6380510330_dp
      xab_pm6(46, 46) = 0.0519560017_dp
      xab_pm6(47, 1) = 1.9951679707_dp
      xab_pm6(47, 6) = 11.3354558945_dp
      xab_pm6(47, 7) = 19.8037109375_dp
      xab_pm6(47, 8) = 0.1656610072_dp
      xab_pm6(47, 9) = 12.6958837509_dp
      xab_pm6(47, 13) = 0.8965139985_dp
      xab_pm6(47, 15) = 0.0499319993_dp
      xab_pm6(47, 16) = 11.1880216599_dp
      xab_pm6(47, 17) = 23.7654590607_dp
      xab_pm6(47, 35) = 1.7143690586_dp
      xab_pm6(47, 47) = 0.5577419996_dp
      xab_pm6(48, 1) = 11.9142007828_dp
      xab_pm6(48, 6) = 0.6034410000_dp
      xab_pm6(48, 7) = 0.1806630045_dp
      xab_pm6(48, 8) = 0.9261459708_dp
      xab_pm6(48, 9) = 1.3536649942_dp
      xab_pm6(48, 14) = 2.2533459663_dp
      xab_pm6(48, 16) = 0.3613890111_dp
      xab_pm6(48, 17) = 0.1404239982_dp
      xab_pm6(48, 35) = 0.2722670138_dp
      xab_pm6(48, 48) = 18.6179981232_dp
      xab_pm6(49, 1) = 14.9752931595_dp
      xab_pm6(49, 6) = 2.1873850822_dp
      xab_pm6(49, 7) = 3.3699929714_dp
      xab_pm6(49, 8) = 4.1285829544_dp
      xab_pm6(49, 9) = 3.7010159492_dp
      xab_pm6(49, 16) = 6.3411049843_dp
      xab_pm6(49, 17) = 2.3885519505_dp
      xab_pm6(49, 31) = 2.4219870567_dp
      xab_pm6(49, 33) = 6.2083501816_dp
      xab_pm6(49, 34) = 2.3193230629_dp
      xab_pm6(49, 35) = 3.7285981178_dp
      xab_pm6(49, 49) = 8.0634908676_dp
      xab_pm6(50, 1) = 6.5351619720_dp
      xab_pm6(50, 6) = 3.3743550777_dp
      xab_pm6(50, 7) = 1.3918999434_dp
      xab_pm6(50, 8) = 4.3740167618_dp
      xab_pm6(50, 9) = 18.5986633301_dp
      xab_pm6(50, 16) = 2.3148701191_dp
      xab_pm6(50, 17) = 0.8077819943_dp
      xab_pm6(50, 32) = 12.3434114456_dp
      xab_pm6(50, 34) = 3.0618851185_dp
      xab_pm6(50, 35) = 0.6687980294_dp
      xab_pm6(50, 50) = 0.2870000005_dp
      xab_pm6(51, 1) = 0.7953429818_dp
      xab_pm6(51, 6) = 0.5792120099_dp
      xab_pm6(51, 7) = 0.0820650011_dp
      xab_pm6(51, 8) = 0.6342340112_dp
      xab_pm6(51, 9) = 0.6502770185_dp
      xab_pm6(51, 13) = 1.6166900396_dp
      xab_pm6(51, 14) = 8.7137489319_dp
      xab_pm6(51, 16) = 0.3969689906_dp
      xab_pm6(51, 17) = 0.1564749926_dp
      xab_pm6(51, 25) = 2.2367100716_dp
      xab_pm6(51, 27) = 2.2760500908_dp
      xab_pm6(51, 35) = 0.1980440021_dp
      xab_pm6(51, 43) = 2.2762598991_dp
      xab_pm6(51, 44) = 2.2757599354_dp
      xab_pm6(51, 45) = 2.2763400078_dp
      xab_pm6(51, 49) = 6.6608009338_dp
      xab_pm6(51, 51) = 0.7248849869_dp
      xab_pm6(52, 1) = 1.8076790571_dp
      xab_pm6(52, 6) = 0.9704939723_dp
      xab_pm6(52, 7) = 0.3585929871_dp
      xab_pm6(52, 8) = 0.3829259872_dp
      xab_pm6(52, 9) = 0.2008219957_dp
      xab_pm6(52, 13) = 2.1068120003_dp
      xab_pm6(52, 15) = 1.1092890501_dp
      xab_pm6(52, 16) = 0.9439250231_dp
      xab_pm6(52, 17) = 0.2854779959_dp
      xab_pm6(52, 30) = 1.7560700178_dp
      xab_pm6(52, 32) = 7.0190491676_dp
      xab_pm6(52, 33) = 0.6857740283_dp
      xab_pm6(52, 34) = 1.1878260374_dp
      xab_pm6(52, 35) = 0.3942019939_dp
      xab_pm6(52, 48) = 1.0859190226_dp
      xab_pm6(52, 49) = 2.0395820141_dp
      xab_pm6(52, 50) = 2.9519760609_dp
      xab_pm6(52, 52) = 0.6424859762_dp
      xab_pm6(53, 1) = 0.9818980098_dp
      xab_pm6(53, 2) = 1.6307209730_dp
      xab_pm6(53, 3) = 4.1685991287_dp
      xab_pm6(53, 4) = 2.3518979549_dp
      xab_pm6(53, 5) = 3.1613850594_dp
      xab_pm6(53, 6) = 0.8101559877_dp
      xab_pm6(53, 7) = 0.2649030089_dp
      xab_pm6(53, 8) = 0.8662040234_dp
      xab_pm6(53, 9) = 0.3924250007_dp
      xab_pm6(53, 10) = 1.5035680532_dp
      xab_pm6(53, 11) = 1.9861119986_dp
      xab_pm6(53, 12) = 3.2769138813_dp
      xab_pm6(53, 13) = 2.9290800095_dp
      xab_pm6(53, 14) = 0.7002990246_dp
      xab_pm6(53, 15) = 3.0472071171_dp
      xab_pm6(53, 16) = 0.7099289894_dp
      xab_pm6(53, 17) = 0.3104740083_dp
      xab_pm6(53, 18) = 0.3053669930_dp
      xab_pm6(53, 19) = 4.8243532181_dp
      xab_pm6(53, 20) = 7.6899209023_dp
      xab_pm6(53, 21) = 3.1142818928_dp
      xab_pm6(53, 22) = 2.4267470837_dp
      xab_pm6(53, 23) = 6.1981120110_dp
      xab_pm6(53, 24) = 2.5985898972_dp
      xab_pm6(53, 25) = 1.1934100389_dp
      xab_pm6(53, 26) = 0.5326219797_dp
      xab_pm6(53, 27) = 1.1052390337_dp
      xab_pm6(53, 28) = 0.0174589995_dp
      xab_pm6(53, 29) = 0.0067810002_dp
      xab_pm6(53, 30) = 0.9766070247_dp
      xab_pm6(53, 31) = 1.2521680593_dp
      xab_pm6(53, 32) = 1.3232669830_dp
      xab_pm6(53, 33) = 0.3108240068_dp
      xab_pm6(53, 35) = 0.4830540121_dp
      xab_pm6(53, 36) = 0.2011359930_dp
      xab_pm6(53, 37) = 4.0924458504_dp
      xab_pm6(53, 38) = 2.1039409637_dp
      xab_pm6(53, 39) = 1.0214020014_dp
      xab_pm6(53, 40) = 4.5139431953_dp
      xab_pm6(53, 41) = 2.3992979527_dp
      xab_pm6(53, 42) = 0.1246950030_dp
      xab_pm6(53, 43) = 0.1105939969_dp
      xab_pm6(53, 44) = 7.8377099037_dp
      xab_pm6(53, 45) = 2.3579440117_dp
      xab_pm6(53, 46) = 3.5220170021_dp
      xab_pm6(53, 47) = 0.0489040017_dp
      xab_pm6(53, 48) = 0.3967840075_dp
      xab_pm6(53, 49) = 5.9478211403_dp
      xab_pm6(53, 50) = 1.7831629515_dp
      xab_pm6(53, 51) = 0.3181900084_dp
      xab_pm6(53, 52) = 1.1011159420_dp
      xab_pm6(53, 53) = 0.5105419755_dp
      xab_pm6(54, 1) = 0.7010160089_dp
      xab_pm6(54, 2) = 2.5994710922_dp
      xab_pm6(54, 3) = 4.5820808411_dp
      xab_pm6(54, 4) = 0.6605250239_dp
      xab_pm6(54, 5) = 1.1006120443_dp
      xab_pm6(54, 6) = 0.8267269731_dp
      xab_pm6(54, 7) = 0.9256240129_dp
      xab_pm6(54, 8) = 0.0353560001_dp
      xab_pm6(54, 9) = 0.0650110021_dp
      xab_pm6(54, 10) = 0.2938619852_dp
      xab_pm6(54, 11) = 8.3682041168_dp
      xab_pm6(54, 12) = 9.7235717773_dp
      xab_pm6(54, 13) = 7.4044651985_dp
      xab_pm6(54, 14) = 16.0919990540_dp
      xab_pm6(54, 17) = 0.4637579918_dp
      xab_pm6(54, 18) = 0.0492659993_dp
      xab_pm6(54, 19) = 1.2248890400_dp
      xab_pm6(54, 20) = 1.7171210051_dp
      xab_pm6(54, 35) = 0.4751160145_dp
      xab_pm6(54, 36) = 0.0497930013_dp
      xab_pm6(54, 37) = 0.9749649763_dp
      xab_pm6(54, 53) = 0.1120899990_dp
      xab_pm6(54, 54) = 0.3444739878_dp
      xab_pm6(55, 1) = 0.0969009995_dp
      xab_pm6(55, 5) = 10.3926095963_dp
      xab_pm6(55, 6) = 24.5146236420_dp
      xab_pm6(55, 7) = 29.7110767365_dp
      xab_pm6(55, 8) = 8.1768426895_dp
      xab_pm6(55, 9) = 22.2334156036_dp
      xab_pm6(55, 15) = 0.5065119863_dp
      xab_pm6(55, 16) = 0.0917429999_dp
      xab_pm6(55, 17) = 4.5319647789_dp
      xab_pm6(55, 35) = 1.6584270000_dp
      xab_pm6(55, 53) = 1.0721780062_dp
      xab_pm6(55, 55) = 25.3200550079_dp
      xab_pm6(56, 1) = 2.0400040150_dp
      xab_pm6(56, 6) = 0.1197929978_dp
      xab_pm6(56, 7) = 0.2079340070_dp
      xab_pm6(56, 8) = 0.3489449918_dp
      xab_pm6(56, 9) = 5.5752549171_dp
      xab_pm6(56, 13) = 9.5390987396_dp
      xab_pm6(56, 14) = 1.2126599550_dp
      xab_pm6(56, 16) = 0.2153860033_dp
      xab_pm6(56, 17) = 0.1601770073_dp
      xab_pm6(56, 22) = 9.4935302734_dp
      xab_pm6(56, 35) = 0.8287940025_dp
      xab_pm6(56, 53) = 0.8355969787_dp
      xab_pm6(56, 56) = 0.3561860025_dp
      xab_pm6(57, 1) = 0.6235010028_dp
      xab_pm6(57, 6) = 0.1086490005_dp
      xab_pm6(57, 7) = 0.1047779992_dp
      xab_pm6(57, 8) = 0.5579569936_dp
      xab_pm6(57, 9) = 2.4019029140_dp
      xab_pm6(57, 13) = 0.5005400181_dp
      xab_pm6(57, 14) = 3.2190299034_dp
      xab_pm6(57, 15) = 0.5416600108_dp
      xab_pm6(57, 16) = 2.6824119091_dp
      xab_pm6(57, 17) = 0.2302030027_dp
      xab_pm6(57, 35) = 0.2385820001_dp
      xab_pm6(57, 53) = 0.2268829942_dp
      xab_pm6(57, 57) = 5.1751618385_dp
      xab_pm6(64, 1) = 0.1358100027_dp
      xab_pm6(64, 6) = 0.0530400015_dp
      xab_pm6(64, 7) = 0.2050500065_dp
      xab_pm6(64, 8) = 0.1757999957_dp
      xab_pm6(64, 9) = 0.3346300125_dp
      xab_pm6(64, 13) = 0.5005400181_dp
      xab_pm6(64, 14) = 3.2190299034_dp
      xab_pm6(64, 15) = 0.5416600108_dp
      xab_pm6(64, 16) = 2.6554000378_dp
      xab_pm6(64, 17) = 0.0899700001_dp
      xab_pm6(64, 35) = 0.2407400012_dp
      xab_pm6(64, 53) = 0.2782399952_dp
      xab_pm6(64, 64) = 2.6703999043_dp
      xab_pm6(71, 1) = 0.7879199982_dp
      xab_pm6(71, 6) = 4.4538249969_dp
      xab_pm6(71, 7) = 2.8608279228_dp
      xab_pm6(71, 8) = 2.9170761108_dp
      xab_pm6(71, 15) = 0.5000000000_dp
      xab_pm6(71, 17) = 12.7570991516_dp
      xab_pm6(71, 35) = 8.6482744217_dp
      xab_pm6(71, 53) = 10.0823154449_dp
      xab_pm6(72, 1) = 3.4273118973_dp
      xab_pm6(72, 5) = 0.6592699885_dp
      xab_pm6(72, 6) = 0.3785789907_dp
      xab_pm6(72, 7) = 0.6557949781_dp
      xab_pm6(72, 8) = 1.0347180367_dp
      xab_pm6(72, 9) = 1.6793349981_dp
      xab_pm6(72, 12) = 4.3302497864_dp
      xab_pm6(72, 13) = 0.6225200295_dp
      xab_pm6(72, 14) = 3.3822999001_dp
      xab_pm6(72, 15) = 0.5055299997_dp
      xab_pm6(72, 16) = 1.6667599678_dp
      xab_pm6(72, 17) = 0.7064210176_dp
      xab_pm6(72, 20) = 4.3195099831_dp
      xab_pm6(72, 33) = 1.2808200121_dp
      xab_pm6(72, 35) = 0.6924560070_dp
      xab_pm6(72, 53) = 0.8209480047_dp
      xab_pm6(72, 56) = 9.0225200653_dp
      xab_pm6(72, 72) = 1.0589109659_dp
      xab_pm6(73, 1) = 2.8276689053_dp
      xab_pm6(73, 6) = 0.8474389911_dp
      xab_pm6(73, 7) = 1.0154609680_dp
      xab_pm6(73, 8) = 1.7510830164_dp
      xab_pm6(73, 9) = 3.1465198994_dp
      xab_pm6(73, 11) = 8.2761297226_dp
      xab_pm6(73, 15) = 6.2618799210_dp
      xab_pm6(73, 16) = 2.9759800434_dp
      xab_pm6(73, 17) = 0.5164129734_dp
      xab_pm6(73, 19) = 2.0267000198_dp
      xab_pm6(73, 35) = 0.7914450169_dp
      xab_pm6(73, 53) = 6.5515508652_dp
      xab_pm6(73, 73) = 10.9870529175_dp
      xab_pm6(74, 1) = 1.8322700262_dp
      xab_pm6(74, 6) = 1.1607700586_dp
      xab_pm6(74, 7) = 0.4783500135_dp
      xab_pm6(74, 8) = 0.3490099907_dp
      xab_pm6(74, 9) = 0.2138900012_dp
      xab_pm6(74, 11) = 8.2760400772_dp
      xab_pm6(74, 15) = 5.9538598061_dp
      xab_pm6(74, 16) = 0.4886299968_dp
      xab_pm6(74, 17) = 0.2779999971_dp
      xab_pm6(74, 19) = 2.0266098976_dp
      xab_pm6(74, 35) = 0.3723900020_dp
      xab_pm6(74, 53) = 1.0773700476_dp
      xab_pm6(74, 74) = 7.4713897705_dp
      xab_pm6(75, 1) = 0.3458940089_dp
      xab_pm6(75, 6) = 0.6906870008_dp
      xab_pm6(75, 7) = 0.4452129900_dp
      xab_pm6(75, 8) = 0.6359599829_dp
      xab_pm6(75, 9) = 0.5359659791_dp
      xab_pm6(75, 14) = 0.8494499922_dp
      xab_pm6(75, 15) = 0.9669420123_dp
      xab_pm6(75, 16) = 0.0688740015_dp
      xab_pm6(75, 17) = 0.1463190019_dp
      xab_pm6(75, 32) = 2.1515800953_dp
      xab_pm6(75, 34) = 2.2021400928_dp
      xab_pm6(75, 35) = 0.2875280082_dp
      xab_pm6(75, 51) = 2.2757799625_dp
      xab_pm6(75, 53) = 3.5592861176_dp
      xab_pm6(75, 75) = 4.4888520241_dp
      xab_pm6(76, 1) = 4.3938698769_dp
      xab_pm6(76, 6) = 0.4984099865_dp
      xab_pm6(76, 7) = 0.0808700025_dp
      xab_pm6(76, 8) = 0.1843000054_dp
      xab_pm6(76, 9) = 0.1400499940_dp
      xab_pm6(76, 11) = 8.2757501602_dp
      xab_pm6(76, 15) = 6.0583000183_dp
      xab_pm6(76, 16) = 4.1860499382_dp
      xab_pm6(76, 17) = 0.3279199898_dp
      xab_pm6(76, 19) = 2.0263199806_dp
      xab_pm6(76, 35) = 0.3824299872_dp
      xab_pm6(76, 53) = 2.1991899014_dp
      xab_pm6(76, 76) = 0.8304399848_dp
      xab_pm6(77, 1) = 0.0580470003_dp
      xab_pm6(77, 6) = 0.1150470003_dp
      xab_pm6(77, 7) = 8.5186395645_dp
      xab_pm6(77, 8) = 9.6254024506_dp
      xab_pm6(77, 9) = 1.4996390343_dp
      xab_pm6(77, 11) = 8.2758302689_dp
      xab_pm6(77, 15) = 6.2846698761_dp
      xab_pm6(77, 16) = 4.1357321739_dp
      xab_pm6(77, 17) = 0.2589159906_dp
      xab_pm6(77, 19) = 2.0264000893_dp
      xab_pm6(77, 35) = 0.1718789935_dp
      xab_pm6(77, 53) = 1.4971480370_dp
      xab_pm6(77, 77) = 11.1751928329_dp
      xab_pm6(78, 1) = 8.9240150452_dp
      xab_pm6(78, 6) = 3.4934029579_dp
      xab_pm6(78, 7) = 0.5407299995_dp
      xab_pm6(78, 8) = 0.4877560139_dp
      xab_pm6(78, 9) = 5.6170139313_dp
      xab_pm6(78, 13) = 1.0569299459_dp
      xab_pm6(78, 14) = 0.0999900028_dp
      xab_pm6(78, 15) = 0.2337120026_dp
      xab_pm6(78, 16) = 2.2242629528_dp
      xab_pm6(78, 17) = 0.3410010040_dp
      xab_pm6(78, 35) = 0.5203610063_dp
      xab_pm6(78, 53) = 4.6012477875_dp
      xab_pm6(78, 78) = 9.0102519989_dp
      xab_pm6(79, 1) = 2.6052830219_dp
      xab_pm6(79, 6) = 21.4856338501_dp
      xab_pm6(79, 7) = 0.2220589966_dp
      xab_pm6(79, 8) = 0.0771920010_dp
      xab_pm6(79, 9) = 9.5943841934_dp
      xab_pm6(79, 13) = 1.0571399927_dp
      xab_pm6(79, 15) = 0.0670010000_dp
      xab_pm6(79, 16) = 21.6191444397_dp
      xab_pm6(79, 17) = 2.2577021122_dp
      xab_pm6(79, 35) = 0.0042369999_dp
      xab_pm6(79, 53) = 0.0088160001_dp
      xab_pm6(79, 79) = 0.0130909998_dp
      xab_pm6(80, 1) = 0.7993990183_dp
      xab_pm6(80, 6) = 0.1471280009_dp
      xab_pm6(80, 7) = 0.0502399988_dp
      xab_pm6(80, 8) = 0.2407200038_dp
      xab_pm6(80, 9) = 0.1138269976_dp
      xab_pm6(80, 14) = 3.6807401180_dp
      xab_pm6(80, 15) = 0.2149509937_dp
      xab_pm6(80, 16) = 0.3473829925_dp
      xab_pm6(80, 17) = 0.0536600016_dp
      xab_pm6(80, 22) = 2.9572000504_dp
      xab_pm6(80, 35) = 0.1723629981_dp
      xab_pm6(80, 52) = 0.2127320021_dp
      xab_pm6(80, 53) = 0.3420580029_dp
      xab_pm6(80, 80) = 0.4232760072_dp
      xab_pm6(81, 1) = 0.1382050067_dp
      xab_pm6(81, 5) = 10.5043382645_dp
      xab_pm6(81, 6) = 0.5828949809_dp
      xab_pm6(81, 7) = 0.1588120013_dp
      xab_pm6(81, 8) = 0.6369060278_dp
      xab_pm6(81, 9) = 0.2261659950_dp
      xab_pm6(81, 16) = 0.3034259975_dp
      xab_pm6(81, 17) = 0.1620370001_dp
      xab_pm6(81, 35) = 0.2968359888_dp
      xab_pm6(81, 53) = 0.4300329983_dp
      xab_pm6(81, 81) = 9.5351266861_dp
      xab_pm6(82, 1) = 0.8400959969_dp
      xab_pm6(82, 3) = 1.2850639820_dp
      xab_pm6(82, 5) = 1.1381570101_dp
      xab_pm6(82, 6) = 0.4046559930_dp
      xab_pm6(82, 7) = 0.3357869983_dp
      xab_pm6(82, 8) = 0.7825059891_dp
      xab_pm6(82, 9) = 8.3685617447_dp
      xab_pm6(82, 15) = 5.0331997871_dp
      xab_pm6(82, 16) = 0.1751500070_dp
      xab_pm6(82, 17) = 0.1648139954_dp
      xab_pm6(82, 23) = 1.0000000000_dp
      xab_pm6(82, 24) = 1.0291099548_dp
      xab_pm6(82, 30) = 1.0000000000_dp
      xab_pm6(82, 34) = 0.1111949980_dp
      xab_pm6(82, 35) = 0.1482290030_dp
      xab_pm6(82, 41) = 1.0000000000_dp
      xab_pm6(82, 42) = 5.0000000000_dp
      xab_pm6(82, 52) = 0.8090419769_dp
      xab_pm6(82, 53) = 0.2674260139_dp
      xab_pm6(82, 82) = 2.3623430729_dp
      xab_pm6(83, 1) = 1.3974620104_dp
      xab_pm6(83, 3) = 0.6953200102_dp
      xab_pm6(83, 6) = 0.5761790276_dp
      xab_pm6(83, 7) = 0.1527380049_dp
      xab_pm6(83, 8) = 0.3330419958_dp
      xab_pm6(83, 9) = 1.0353239775_dp
      xab_pm6(83, 16) = 0.6209970117_dp
      xab_pm6(83, 17) = 0.3268710077_dp
      xab_pm6(83, 34) = 0.6512079835_dp
      xab_pm6(83, 35) = 0.3811700046_dp
      xab_pm6(83, 53) = 0.8623769879_dp
      xab_pm6(83, 83) = 1.1682139635_dp

      aab_pm6(1, 1) = 3.5409419537_dp
      aab_pm6(2, 1) = 2.9898810387_dp
      aab_pm6(2, 2) = 3.7835590839_dp
      aab_pm6(3, 1) = 2.1362650394_dp
      aab_pm6(3, 2) = 3.1124029160_dp
      aab_pm6(3, 3) = 4.7146739960_dp
      aab_pm6(4, 1) = 2.4754180908_dp
      aab_pm6(4, 2) = 3.3067018986_dp
      aab_pm6(4, 3) = 2.2367279530_dp
      aab_pm6(4, 4) = 1.4999070168_dp
      aab_pm6(5, 1) = 2.6152310371_dp
      aab_pm6(5, 2) = 3.1631400585_dp
      aab_pm6(5, 3) = 3.7593970299_dp
      aab_pm6(5, 4) = 1.8889980316_dp
      aab_pm6(5, 5) = 3.3186240196_dp
      aab_pm6(6, 1) = 1.0278060000_dp
      aab_pm6(6, 2) = 3.0427050591_dp
      aab_pm6(6, 3) = 3.2418739796_dp
      aab_pm6(6, 4) = 4.2128820419_dp
      aab_pm6(6, 5) = 2.9190070629_dp
      aab_pm6(6, 6) = 2.6137130260_dp
      aab_pm6(7, 1) = 0.9694060087_dp
      aab_pm6(7, 2) = 2.8143389225_dp
      aab_pm6(7, 3) = 2.6406230927_dp
      aab_pm6(7, 4) = 2.5808949471_dp
      aab_pm6(7, 5) = 2.4770040512_dp
      aab_pm6(7, 6) = 2.6861081123_dp
      aab_pm6(7, 7) = 2.5745019913_dp
      aab_pm6(8, 1) = 1.2609419823_dp
      aab_pm6(8, 2) = 3.6537749767_dp
      aab_pm6(8, 3) = 2.5844419003_dp
      aab_pm6(8, 4) = 3.0518670082_dp
      aab_pm6(8, 5) = 2.6953508854_dp
      aab_pm6(8, 6) = 2.8896069527_dp
      aab_pm6(8, 7) = 2.7842919827_dp
      aab_pm6(8, 8) = 2.6239979267_dp
      aab_pm6(9, 1) = 3.1367399693_dp
      aab_pm6(9, 2) = 2.8565430641_dp
      aab_pm6(9, 3) = 3.0439009666_dp
      aab_pm6(9, 4) = 3.7269229889_dp
      aab_pm6(9, 5) = 2.8238370419_dp
      aab_pm6(9, 6) = 3.0276000500_dp
      aab_pm6(9, 7) = 2.8566460609_dp
      aab_pm6(9, 8) = 3.0154440403_dp
      aab_pm6(9, 9) = 3.1757590771_dp
      aab_pm6(10, 1) = 5.9996800423_dp
      aab_pm6(10, 2) = 3.6777579784_dp
      aab_pm6(10, 3) = 2.1936659813_dp
      aab_pm6(10, 4) = 1.3165880442_dp
      aab_pm6(10, 5) = 2.7561900616_dp
      aab_pm6(10, 6) = 3.4411880970_dp
      aab_pm6(10, 7) = 4.4263701439_dp
      aab_pm6(10, 8) = 2.8895869255_dp
      aab_pm6(10, 9) = 3.6756110191_dp
      aab_pm6(10, 10) = 3.9745669365_dp
      aab_pm6(11, 1) = 0.5003259778_dp
      aab_pm6(11, 2) = 1.7030290365_dp
      aab_pm6(11, 3) = 1.2672990561_dp
      aab_pm6(11, 4) = 1.2554800510_dp
      aab_pm6(11, 5) = 1.5699609518_dp
      aab_pm6(11, 6) = 2.1960499287_dp
      aab_pm6(11, 7) = 2.4943840504_dp
      aab_pm6(11, 8) = 1.9814490080_dp
      aab_pm6(11, 9) = 2.6195509434_dp
      aab_pm6(11, 10) = 1.7742359638_dp
      aab_pm6(11, 11) = 0.4464350045_dp
      aab_pm6(12, 1) = 2.6515939236_dp
      aab_pm6(12, 2) = 2.2106029987_dp
      aab_pm6(12, 3) = 1.1843800545_dp
      aab_pm6(12, 4) = 1.5575909615_dp
      aab_pm6(12, 5) = 2.5274410248_dp
      aab_pm6(12, 6) = 3.0409460068_dp
      aab_pm6(12, 7) = 2.0791249275_dp
      aab_pm6(12, 8) = 2.2515199184_dp
      aab_pm6(12, 9) = 3.3622078896_dp
      aab_pm6(12, 10) = 2.0316760540_dp
      aab_pm6(12, 11) = 1.5067729950_dp
      aab_pm6(12, 12) = 1.0935729742_dp
      aab_pm6(13, 1) = 2.0259959698_dp
      aab_pm6(13, 2) = 2.2558300495_dp
      aab_pm6(13, 3) = 1.5815930367_dp
      aab_pm6(13, 4) = 1.9382369518_dp
      aab_pm6(13, 5) = 2.0595688820_dp
      aab_pm6(13, 6) = 2.2674400806_dp
      aab_pm6(13, 7) = 2.0097539425_dp
      aab_pm6(13, 8) = 2.4986600876_dp
      aab_pm6(13, 9) = 3.0842580795_dp
      aab_pm6(13, 10) = 2.4478690624_dp
      aab_pm6(13, 11) = 1.2028709650_dp
      aab_pm6(13, 12) = 1.9725300074_dp
      aab_pm6(13, 13) = 1.3877140284_dp
      aab_pm6(14, 1) = 1.8969500065_dp
      aab_pm6(14, 2) = 2.0404980183_dp
      aab_pm6(14, 3) = 1.7896089554_dp
      aab_pm6(14, 4) = 1.2631319761_dp
      aab_pm6(14, 5) = 1.9826530218_dp
      aab_pm6(14, 6) = 1.9844980240_dp
      aab_pm6(14, 7) = 1.8189879656_dp
      aab_pm6(14, 8) = 1.9235999584_dp
      aab_pm6(14, 9) = 2.1310279369_dp
      aab_pm6(14, 10) = 2.8677840233_dp
      aab_pm6(14, 11) = 2.0076150894_dp
      aab_pm6(14, 12) = 3.1397490501_dp
      aab_pm6(14, 13) = 1.8999999762_dp
      aab_pm6(14, 14) = 1.3289999962_dp
      aab_pm6(15, 1) = 1.9265370369_dp
      aab_pm6(15, 2) = 2.0931580067_dp
      aab_pm6(15, 3) = 1.3945440054_dp
      aab_pm6(15, 4) = 1.8000700474_dp
      aab_pm6(15, 5) = 1.9231679440_dp
      aab_pm6(15, 6) = 1.9946529865_dp
      aab_pm6(15, 7) = 2.1470420361_dp
      aab_pm6(15, 8) = 2.2207679749_dp
      aab_pm6(15, 9) = 2.2343559265_dp
      aab_pm6(15, 10) = 2.2190361023_dp
      aab_pm6(15, 11) = 1.5003199577_dp
      aab_pm6(15, 12) = 1.3837729692_dp
      aab_pm6(15, 13) = 1.9807269573_dp
      aab_pm6(15, 14) = 3.3134660721_dp
      aab_pm6(15, 15) = 1.5057920218_dp
      aab_pm6(16, 1) = 2.2159750462_dp
      aab_pm6(16, 2) = 1.9591490030_dp
      aab_pm6(16, 3) = 2.2942750454_dp
      aab_pm6(16, 4) = 2.7817358971_dp
      aab_pm6(16, 5) = 2.4036960602_dp
      aab_pm6(16, 6) = 2.2103049755_dp
      aab_pm6(16, 7) = 2.2899899483_dp
      aab_pm6(16, 8) = 2.3832890987_dp
      aab_pm6(16, 9) = 2.1871860027_dp
      aab_pm6(16, 10) = 2.7870581150_dp
      aab_pm6(16, 11) = 1.4008500576_dp
      aab_pm6(16, 12) = 1.5001629591_dp
      aab_pm6(16, 13) = 1.9767049551_dp
      aab_pm6(16, 14) = 1.8859159946_dp
      aab_pm6(16, 15) = 1.5953249931_dp
      aab_pm6(16, 16) = 1.7945560217_dp
      aab_pm6(17, 1) = 2.4028859138_dp
      aab_pm6(17, 2) = 1.6716769934_dp
      aab_pm6(17, 3) = 2.7830009460_dp
      aab_pm6(17, 4) = 2.8226759434_dp
      aab_pm6(17, 5) = 2.2593228817_dp
      aab_pm6(17, 6) = 2.1621971130_dp
      aab_pm6(17, 7) = 2.1721339226_dp
      aab_pm6(17, 8) = 2.3232359886_dp
      aab_pm6(17, 9) = 2.3132700920_dp
      aab_pm6(17, 10) = 1.7031509876_dp
      aab_pm6(17, 11) = 1.8164290190_dp
      aab_pm6(17, 12) = 2.3918058872_dp
      aab_pm6(17, 13) = 2.1259388924_dp
      aab_pm6(17, 14) = 1.6849780083_dp
      aab_pm6(17, 15) = 1.4683059454_dp
      aab_pm6(17, 16) = 1.7154350281_dp
      aab_pm6(17, 17) = 1.8232389688_dp
      aab_pm6(18, 1) = 4.0561671257_dp
      aab_pm6(18, 2) = 2.7165620327_dp
      aab_pm6(18, 3) = 3.1228950024_dp
      aab_pm6(18, 4) = 3.0440070629_dp
      aab_pm6(18, 5) = 2.4154710770_dp
      aab_pm6(18, 6) = 1.4713089466_dp
      aab_pm6(18, 7) = 2.3268051147_dp
      aab_pm6(18, 8) = 2.2406730652_dp
      aab_pm6(18, 9) = 3.9206581116_dp
      aab_pm6(18, 10) = 2.9637470245_dp
      aab_pm6(18, 11) = 2.1676769257_dp
      aab_pm6(18, 12) = 2.0926640034_dp
      aab_pm6(18, 13) = 2.6451649666_dp
      aab_pm6(18, 14) = 1.7803499699_dp
      aab_pm6(18, 15) = 4.3725161552_dp
      aab_pm6(18, 16) = 2.0493979454_dp
      aab_pm6(18, 17) = 2.5544490814_dp
      aab_pm6(18, 18) = 2.3064320087_dp
      aab_pm6(19, 1) = 0.6481729746_dp
      aab_pm6(19, 2) = 1.4185010195_dp
      aab_pm6(19, 3) = 1.0364869833_dp
      aab_pm6(19, 4) = 1.9318879843_dp
      aab_pm6(19, 5) = 2.0317680836_dp
      aab_pm6(19, 6) = 2.2417569160_dp
      aab_pm6(19, 7) = 2.3258590698_dp
      aab_pm6(19, 8) = 1.5085710287_dp
      aab_pm6(19, 9) = 3.1828169823_dp
      aab_pm6(19, 10) = 1.1380209923_dp
      aab_pm6(19, 11) = 0.8843070269_dp
      aab_pm6(19, 12) = 0.8848099709_dp
      aab_pm6(19, 13) = 1.9760760069_dp
      aab_pm6(19, 14) = 1.6759300232_dp
      aab_pm6(19, 15) = 1.4437379837_dp
      aab_pm6(19, 16) = 2.5121560097_dp
      aab_pm6(19, 17) = 1.6221630573_dp
      aab_pm6(19, 18) = 2.3028030396_dp
      aab_pm6(19, 19) = 1.4355139732_dp
      aab_pm6(20, 1) = 2.1418590546_dp
      aab_pm6(20, 2) = 1.7198469639_dp
      aab_pm6(20, 5) = 1.7000099421_dp
      aab_pm6(20, 6) = 1.0353050232_dp
      aab_pm6(20, 7) = 2.3866000175_dp
      aab_pm6(20, 8) = 3.2638969421_dp
      aab_pm6(20, 9) = 2.6450529099_dp
      aab_pm6(20, 10) = 0.9545300007_dp
      aab_pm6(20, 11) = 3.1071040630_dp
      aab_pm6(20, 12) = 2.2997999191_dp
      aab_pm6(20, 13) = 1.6125650406_dp
      aab_pm6(20, 14) = 1.2187880278_dp
      aab_pm6(20, 15) = 1.0241420269_dp
      aab_pm6(20, 16) = 0.9581710100_dp
      aab_pm6(20, 17) = 2.3833909035_dp
      aab_pm6(20, 18) = 1.0348809958_dp
      aab_pm6(20, 19) = 1.1191999912_dp
      aab_pm6(20, 20) = 1.8896739483_dp
      aab_pm6(21, 1) = 1.1794849634_dp
      aab_pm6(21, 6) = 2.6304900646_dp
      aab_pm6(21, 7) = 2.2700040340_dp
      aab_pm6(21, 8) = 2.2565159798_dp
      aab_pm6(21, 9) = 3.1079850197_dp
      aab_pm6(21, 13) = 1.0035500526_dp
      aab_pm6(21, 14) = 2.0168700218_dp
      aab_pm6(21, 15) = 0.8681650162_dp
      aab_pm6(21, 16) = 0.4229390025_dp
      aab_pm6(21, 17) = 2.1414740086_dp
      aab_pm6(21, 21) = 1.1328380108_dp
      aab_pm6(22, 1) = 0.8326690197_dp
      aab_pm6(22, 5) = 1.6287100315_dp
      aab_pm6(22, 6) = 1.5979729891_dp
      aab_pm6(22, 7) = 1.6786860228_dp
      aab_pm6(22, 8) = 1.7891180515_dp
      aab_pm6(22, 9) = 2.3070869446_dp
      aab_pm6(22, 12) = 1.9113399982_dp
      aab_pm6(22, 13) = 1.3694859743_dp
      aab_pm6(22, 14) = 2.8560380936_dp
      aab_pm6(22, 15) = 2.1519289017_dp
      aab_pm6(22, 16) = 1.8464390039_dp
      aab_pm6(22, 17) = 1.4610340595_dp
      aab_pm6(22, 20) = 2.0000000000_dp
      aab_pm6(22, 22) = 2.6485970020_dp
      aab_pm6(23, 1) = 1.2801330090_dp
      aab_pm6(23, 6) = 2.7898550034_dp
      aab_pm6(23, 7) = 1.6075400114_dp
      aab_pm6(23, 8) = 1.6239730120_dp
      aab_pm6(23, 9) = 1.8251600266_dp
      aab_pm6(23, 11) = 2.5510098934_dp
      aab_pm6(23, 15) = 2.5491540432_dp
      aab_pm6(23, 16) = 2.7041239738_dp
      aab_pm6(23, 17) = 1.6885290146_dp
      aab_pm6(23, 19) = 4.5213599205_dp
      aab_pm6(23, 23) = 4.8323907852_dp
      aab_pm6(24, 1) = 0.8826609850_dp
      aab_pm6(24, 6) = 3.6567540169_dp
      aab_pm6(24, 7) = 3.0291860104_dp
      aab_pm6(24, 8) = 2.5000000000_dp
      aab_pm6(24, 9) = 2.7165210247_dp
      aab_pm6(24, 11) = 2.2950561047_dp
      aab_pm6(24, 14) = 1.8607599735_dp
      aab_pm6(24, 15) = 1.6953829527_dp
      aab_pm6(24, 16) = 2.2609779835_dp
      aab_pm6(24, 17) = 2.1526179314_dp
      aab_pm6(24, 19) = 2.0000000000_dp
      aab_pm6(24, 24) = 4.6554188728_dp
      aab_pm6(25, 1) = 2.3099400997_dp
      aab_pm6(25, 6) = 3.0007500648_dp
      aab_pm6(25, 7) = 2.9214699268_dp
      aab_pm6(25, 8) = 2.5775399208_dp
      aab_pm6(25, 9) = 2.7919499874_dp
      aab_pm6(25, 13) = 1.7683600187_dp
      aab_pm6(25, 14) = 1.9379589558_dp
      aab_pm6(25, 15) = 1.9470200539_dp
      aab_pm6(25, 16) = 2.4825100899_dp
      aab_pm6(25, 17) = 1.6570099592_dp
      aab_pm6(25, 20) = 1.4914400578_dp
      aab_pm6(25, 25) = 2.6654200554_dp
      aab_pm6(26, 1) = 0.8544880152_dp
      aab_pm6(26, 6) = 3.9913430214_dp
      aab_pm6(26, 7) = 2.5004858971_dp
      aab_pm6(26, 8) = 1.7263129950_dp
      aab_pm6(26, 9) = 4.2947068214_dp
      aab_pm6(26, 15) = 2.5675339699_dp
      aab_pm6(26, 16) = 0.9889910221_dp
      aab_pm6(26, 17) = 1.2297929525_dp
      aab_pm6(26, 19) = 2.0000000000_dp
      aab_pm6(26, 26) = 2.7207849026_dp
      aab_pm6(27, 1) = 2.9665179253_dp
      aab_pm6(27, 6) = 3.7162330151_dp
      aab_pm6(27, 7) = 3.6186380386_dp
      aab_pm6(27, 8) = 3.7269110680_dp
      aab_pm6(27, 9) = 3.9563469887_dp
      aab_pm6(27, 14) = 2.4698050022_dp
      aab_pm6(27, 15) = 1.1525050402_dp
      aab_pm6(27, 16) = 2.4292550087_dp
      aab_pm6(27, 17) = 3.2174971104_dp
      aab_pm6(27, 27) = 3.2881660461_dp
      aab_pm6(28, 1) = 2.6352798939_dp
      aab_pm6(28, 6) = 4.2855129242_dp
      aab_pm6(28, 7) = 3.8452150822_dp
      aab_pm6(28, 8) = 2.9372320175_dp
      aab_pm6(28, 9) = 3.4402410984_dp
      aab_pm6(28, 14) = 2.0688810349_dp
      aab_pm6(28, 15) = 3.2602829933_dp
      aab_pm6(28, 16) = 2.0027520657_dp
      aab_pm6(28, 17) = 2.2005119324_dp
      aab_pm6(28, 28) = 1.0979599953_dp
      aab_pm6(29, 1) = 2.3353590965_dp
      aab_pm6(29, 6) = 4.6387729645_dp
      aab_pm6(29, 7) = 4.2143368721_dp
      aab_pm6(29, 8) = 3.9599509239_dp
      aab_pm6(29, 9) = 4.4788317680_dp
      aab_pm6(29, 15) = 0.2106399983_dp
      aab_pm6(29, 16) = 0.2731119990_dp
      aab_pm6(29, 17) = 2.7765309811_dp
      aab_pm6(29, 29) = 3.6168460846_dp
      aab_pm6(30, 1) = 1.9878909588_dp
      aab_pm6(30, 6) = 1.8023270369_dp
      aab_pm6(30, 7) = 1.8445789814_dp
      aab_pm6(30, 8) = 2.3350539207_dp
      aab_pm6(30, 9) = 2.4100210667_dp
      aab_pm6(30, 14) = 1.8320579529_dp
      aab_pm6(30, 15) = 1.2204799652_dp
      aab_pm6(30, 16) = 1.4550000429_dp
      aab_pm6(30, 17) = 1.6251759529_dp
      aab_pm6(30, 20) = 1.1191799641_dp
      aab_pm6(30, 30) = 0.9290000200_dp
      aab_pm6(31, 1) = 1.8473500013_dp
      aab_pm6(31, 6) = 2.3254098892_dp
      aab_pm6(31, 7) = 2.1218199730_dp
      aab_pm6(31, 8) = 2.3483469486_dp
      aab_pm6(31, 9) = 2.6798689365_dp
      aab_pm6(31, 14) = 1.9137799740_dp
      aab_pm6(31, 15) = 2.9796500206_dp
      aab_pm6(31, 16) = 2.2321081161_dp
      aab_pm6(31, 17) = 2.0247099400_dp
      aab_pm6(31, 31) = 1.3346430063_dp
      aab_pm6(32, 1) = 2.2067930698_dp
      aab_pm6(32, 6) = 2.2574689388_dp
      aab_pm6(32, 7) = 1.9882260561_dp
      aab_pm6(32, 8) = 2.1394131184_dp
      aab_pm6(32, 9) = 2.3847770691_dp
      aab_pm6(32, 14) = 0.2997210026_dp
      aab_pm6(32, 15) = 2.4692909718_dp
      aab_pm6(32, 16) = 2.0245881081_dp
      aab_pm6(32, 17) = 1.7712279558_dp
      aab_pm6(32, 25) = 2.3828339577_dp
      aab_pm6(32, 27) = 2.8526101112_dp
      aab_pm6(32, 32) = 2.0190000534_dp
      aab_pm6(33, 1) = 1.9935270548_dp
      aab_pm6(33, 6) = 1.8550690413_dp
      aab_pm6(33, 7) = 1.4965430498_dp
      aab_pm6(33, 8) = 2.0039501190_dp
      aab_pm6(33, 9) = 2.0125830173_dp
      aab_pm6(33, 13) = 1.1527860165_dp
      aab_pm6(33, 14) = 1.9155999422_dp
      aab_pm6(33, 16) = 1.9543679953_dp
      aab_pm6(33, 17) = 1.6910699606_dp
      aab_pm6(33, 22) = 1.9329110384_dp
      aab_pm6(33, 27) = 3.3681399822_dp
      aab_pm6(33, 30) = 1.4591300488_dp
      aab_pm6(33, 31) = 1.7309770584_dp
      aab_pm6(33, 33) = 1.5882639885_dp
      aab_pm6(34, 1) = 2.0350680351_dp
      aab_pm6(34, 6) = 2.3871181011_dp
      aab_pm6(34, 7) = 1.9377640486_dp
      aab_pm6(34, 8) = 2.4842629433_dp
      aab_pm6(34, 9) = 2.3021800518_dp
      aab_pm6(34, 14) = 1.5298169851_dp
      aab_pm6(34, 15) = 1.0481829643_dp
      aab_pm6(34, 16) = 1.4796060324_dp
      aab_pm6(34, 17) = 2.1288609505_dp
      aab_pm6(34, 25) = 2.6480379105_dp
      aab_pm6(34, 27) = 2.5234498978_dp
      aab_pm6(34, 30) = 1.1862419844_dp
      aab_pm6(34, 32) = 2.6690568924_dp
      aab_pm6(34, 33) = 1.6652799845_dp
      aab_pm6(34, 34) = 1.7958940268_dp
      aab_pm6(35, 1) = 2.1928029060_dp
      aab_pm6(35, 2) = 2.1282749176_dp
      aab_pm6(35, 3) = 2.0744409561_dp
      aab_pm6(35, 4) = 2.3671460152_dp
      aab_pm6(35, 5) = 2.3078899384_dp
      aab_pm6(35, 6) = 2.0150859356_dp
      aab_pm6(35, 7) = 4.2249011993_dp
      aab_pm6(35, 8) = 2.2830460072_dp
      aab_pm6(35, 9) = 2.0317649841_dp
      aab_pm6(35, 10) = 2.4641718864_dp
      aab_pm6(35, 11) = 1.6222180128_dp
      aab_pm6(35, 12) = 2.1956970692_dp
      aab_pm6(35, 13) = 1.8941409588_dp
      aab_pm6(35, 14) = 1.5708249807_dp
      aab_pm6(35, 15) = 1.4021389484_dp
      aab_pm6(35, 16) = 1.5098739862_dp
      aab_pm6(35, 17) = 1.7103309631_dp
      aab_pm6(35, 18) = 2.4508008957_dp
      aab_pm6(35, 19) = 1.6160930395_dp
      aab_pm6(35, 20) = 2.0784049034_dp
      aab_pm6(35, 21) = 1.7934859991_dp
      aab_pm6(35, 22) = 1.6748470068_dp
      aab_pm6(35, 23) = 1.9029040337_dp
      aab_pm6(35, 24) = 1.5660279989_dp
      aab_pm6(35, 25) = 2.2838199139_dp
      aab_pm6(35, 26) = 3.6417820454_dp
      aab_pm6(35, 27) = 2.6326880455_dp
      aab_pm6(35, 28) = 2.7721359730_dp
      aab_pm6(35, 29) = 5.8264069557_dp
      aab_pm6(35, 30) = 1.4161200523_dp
      aab_pm6(35, 31) = 1.8191050291_dp
      aab_pm6(35, 32) = 1.6023659706_dp
      aab_pm6(35, 33) = 1.5201699734_dp
      aab_pm6(35, 34) = 1.4837130308_dp
      aab_pm6(35, 35) = 1.7581460476_dp
      aab_pm6(36, 1) = 3.7704529762_dp
      aab_pm6(36, 2) = 1.9969429970_dp
      aab_pm6(36, 3) = 3.3145620823_dp
      aab_pm6(36, 4) = 3.2530479431_dp
      aab_pm6(36, 5) = 2.3631689548_dp
      aab_pm6(36, 6) = 2.0767381191_dp
      aab_pm6(36, 7) = 1.6440520287_dp
      aab_pm6(36, 8) = 0.2922999859_dp
      aab_pm6(36, 9) = 3.4523210526_dp
      aab_pm6(36, 10) = 2.8136789799_dp
      aab_pm6(36, 11) = 2.4805979729_dp
      aab_pm6(36, 12) = 1.3914870024_dp
      aab_pm6(36, 13) = 2.4671308994_dp
      aab_pm6(36, 14) = 1.7640999556_dp
      aab_pm6(36, 17) = 1.8849740028_dp
      aab_pm6(36, 18) = 1.9951250553_dp
      aab_pm6(36, 19) = 2.1824870110_dp
      aab_pm6(36, 20) = 1.3051970005_dp
      aab_pm6(36, 35) = 1.5290060043_dp
      aab_pm6(36, 36) = 1.1353189945_dp
      aab_pm6(37, 1) = 2.4435560703_dp
      aab_pm6(37, 2) = 1.2707409859_dp
      aab_pm6(37, 5) = 5.5322389603_dp
      aab_pm6(37, 6) = 2.7658300400_dp
      aab_pm6(37, 7) = 0.7610470057_dp
      aab_pm6(37, 8) = 1.3349080086_dp
      aab_pm6(37, 9) = 3.6381220818_dp
      aab_pm6(37, 10) = 2.2675909996_dp
      aab_pm6(37, 13) = 0.7987740040_dp
      aab_pm6(37, 16) = 1.3031840324_dp
      aab_pm6(37, 17) = 2.2744109631_dp
      aab_pm6(37, 18) = 2.5109770298_dp
      aab_pm6(37, 35) = 1.7977659702_dp
      aab_pm6(37, 36) = 2.2687530518_dp
      aab_pm6(37, 37) = 1.1808179617_dp
      aab_pm6(38, 1) = 2.1059141159_dp
      aab_pm6(38, 6) = 1.9866880178_dp
      aab_pm6(38, 7) = 2.1836290359_dp
      aab_pm6(38, 8) = 2.1383988857_dp
      aab_pm6(38, 9) = 3.0506660938_dp
      aab_pm6(38, 14) = 2.9697799683_dp
      aab_pm6(38, 15) = 2.7891499996_dp
      aab_pm6(38, 16) = 1.5981060266_dp
      aab_pm6(38, 17) = 1.8541899920_dp
      aab_pm6(38, 22) = 2.8800299168_dp
      aab_pm6(38, 35) = 1.5243159533_dp
      aab_pm6(38, 38) = 1.0000400543_dp
      aab_pm6(39, 1) = 1.1890530586_dp
      aab_pm6(39, 6) = 1.3360940218_dp
      aab_pm6(39, 7) = 1.7787959576_dp
      aab_pm6(39, 8) = 1.8510299921_dp
      aab_pm6(39, 9) = 2.6480460167_dp
      aab_pm6(39, 13) = 1.0034999847_dp
      aab_pm6(39, 14) = 2.0168199539_dp
      aab_pm6(39, 15) = 0.9544500113_dp
      aab_pm6(39, 16) = 0.9716879725_dp
      aab_pm6(39, 17) = 1.6301519871_dp
      aab_pm6(39, 35) = 1.4012080431_dp
      aab_pm6(39, 39) = 1.0126810074_dp
      aab_pm6(40, 1) = 1.3797030449_dp
      aab_pm6(40, 6) = 2.0294270515_dp
      aab_pm6(40, 7) = 1.7070829868_dp
      aab_pm6(40, 8) = 1.7095700502_dp
      aab_pm6(40, 9) = 1.9009250402_dp
      aab_pm6(40, 13) = 1.2706199884_dp
      aab_pm6(40, 14) = 1.7508330345_dp
      aab_pm6(40, 15) = 1.0918580294_dp
      aab_pm6(40, 16) = 2.1297609806_dp
      aab_pm6(40, 17) = 1.3288350105_dp
      aab_pm6(40, 35) = 1.4468679428_dp
      aab_pm6(40, 40) = 3.8659679890_dp
      aab_pm6(41, 1) = 2.5059120655_dp
      aab_pm6(41, 6) = 2.6210119724_dp
      aab_pm6(41, 7) = 2.0238630772_dp
      aab_pm6(41, 8) = 2.0494890213_dp
      aab_pm6(41, 9) = 3.0031569004_dp
      aab_pm6(41, 11) = 2.5510098934_dp
      aab_pm6(41, 15) = 2.2216079235_dp
      aab_pm6(41, 16) = 2.2494819164_dp
      aab_pm6(41, 17) = 2.2152750492_dp
      aab_pm6(41, 19) = 4.5213599205_dp
      aab_pm6(41, 35) = 2.0066781044_dp
      aab_pm6(41, 41) = 1.7279410362_dp
      aab_pm6(42, 1) = 2.0357480049_dp
      aab_pm6(42, 6) = 2.1986720562_dp
      aab_pm6(42, 7) = 1.8694750071_dp
      aab_pm6(42, 8) = 1.7554240227_dp
      aab_pm6(42, 9) = 2.2025930882_dp
      aab_pm6(42, 11) = 2.4407699108_dp
      aab_pm6(42, 15) = 1.8504409790_dp
      aab_pm6(42, 16) = 1.9396580458_dp
      aab_pm6(42, 17) = 1.7833620310_dp
      aab_pm6(42, 19) = 3.9394199848_dp
      aab_pm6(42, 24) = 2.6746160984_dp
      aab_pm6(42, 35) = 1.2833340168_dp
      aab_pm6(42, 42) = 2.0342540741_dp
      aab_pm6(43, 1) = 2.8303449154_dp
      aab_pm6(43, 6) = 3.1983261108_dp
      aab_pm6(43, 7) = 2.3154170513_dp
      aab_pm6(43, 8) = 2.4051899910_dp
      aab_pm6(43, 9) = 3.6048150063_dp
      aab_pm6(43, 16) = 2.4634010792_dp
      aab_pm6(43, 17) = 2.5720429420_dp
      aab_pm6(43, 32) = 2.8528199196_dp
      aab_pm6(43, 34) = 2.5236599445_dp
      aab_pm6(43, 35) = 2.8282639980_dp
      aab_pm6(44, 1) = 2.8928990364_dp
      aab_pm6(44, 6) = 2.7848329544_dp
      aab_pm6(44, 7) = 3.0555040836_dp
      aab_pm6(44, 8) = 3.1349399090_dp
      aab_pm6(44, 9) = 3.8787109852_dp
      aab_pm6(44, 14) = 2.7759099007_dp
      aab_pm6(44, 15) = 0.2989160120_dp
      aab_pm6(44, 16) = 2.5080759525_dp
      aab_pm6(44, 17) = 1.7598830462_dp
      aab_pm6(44, 32) = 2.8523199558_dp
      aab_pm6(44, 34) = 2.5231599808_dp
      aab_pm6(44, 35) = 2.5847349167_dp
      aab_pm6(44, 44) = 0.5720559955_dp
      aab_pm6(45, 1) = 3.1041650772_dp
      aab_pm6(45, 6) = 3.4159910679_dp
      aab_pm6(45, 7) = 3.5854620934_dp
      aab_pm6(45, 8) = 3.9278299809_dp
      aab_pm6(45, 9) = 4.0516538620_dp
      aab_pm6(45, 14) = 2.7764899731_dp
      aab_pm6(45, 15) = 2.3346068859_dp
      aab_pm6(45, 16) = 3.1540060043_dp
      aab_pm6(45, 17) = 3.3001298904_dp
      aab_pm6(45, 32) = 2.8529000282_dp
      aab_pm6(45, 34) = 2.5237400532_dp
      aab_pm6(45, 35) = 2.9280819893_dp
      aab_pm6(45, 45) = 2.4973280430_dp
      aab_pm6(46, 1) = 2.1837608814_dp
      aab_pm6(46, 6) = 4.7771921158_dp
      aab_pm6(46, 7) = 2.3280460835_dp
      aab_pm6(46, 8) = 2.1548669338_dp
      aab_pm6(46, 9) = 4.2373118401_dp
      aab_pm6(46, 13) = 1.5727200508_dp
      aab_pm6(46, 14) = 2.9481999874_dp
      aab_pm6(46, 15) = 0.8036299944_dp
      aab_pm6(46, 16) = 2.1778008938_dp
      aab_pm6(46, 17) = 3.8712430000_dp
      aab_pm6(46, 35) = 5.9948787689_dp
      aab_pm6(46, 46) = 1.0643750429_dp
      aab_pm6(47, 1) = 2.8959360123_dp
      aab_pm6(47, 6) = 4.4043359756_dp
      aab_pm6(47, 7) = 4.6598711014_dp
      aab_pm6(47, 8) = 1.8938740492_dp
      aab_pm6(47, 9) = 4.6284232140_dp
      aab_pm6(47, 13) = 1.9287999868_dp
      aab_pm6(47, 15) = 6.0000061989_dp
      aab_pm6(47, 16) = 3.6531209946_dp
      aab_pm6(47, 17) = 4.4411759377_dp
      aab_pm6(47, 35) = 3.6774909496_dp
      aab_pm6(47, 47) = 2.1276450157_dp
      aab_pm6(48, 1) = 2.6287479401_dp
      aab_pm6(48, 6) = 1.4256780148_dp
      aab_pm6(48, 7) = 0.9704229832_dp
      aab_pm6(48, 8) = 1.6966730356_dp
      aab_pm6(48, 9) = 2.3121349812_dp
      aab_pm6(48, 14) = 1.3712249994_dp
      aab_pm6(48, 16) = 1.1822019815_dp
      aab_pm6(48, 17) = 0.9435470104_dp
      aab_pm6(48, 35) = 1.0014510155_dp
      aab_pm6(48, 48) = 1.5640439987_dp
      aab_pm6(49, 1) = 3.0641438961_dp
      aab_pm6(49, 6) = 2.1892719269_dp
      aab_pm6(49, 7) = 2.4698679447_dp
      aab_pm6(49, 8) = 2.6620950699_dp
      aab_pm6(49, 9) = 2.9487969875_dp
      aab_pm6(49, 16) = 2.5421309471_dp
      aab_pm6(49, 17) = 2.2334051132_dp
      aab_pm6(49, 31) = 1.6288700104_dp
      aab_pm6(49, 33) = 2.2995519638_dp
      aab_pm6(49, 34) = 1.9065719843_dp
      aab_pm6(49, 35) = 2.2579569817_dp
      aab_pm6(49, 49) = 2.0732409954_dp
      aab_pm6(50, 1) = 2.6489100456_dp
      aab_pm6(50, 6) = 2.4405379295_dp
      aab_pm6(50, 7) = 2.0855889320_dp
      aab_pm6(50, 8) = 2.7272601128_dp
      aab_pm6(50, 9) = 3.7242860794_dp
      aab_pm6(50, 16) = 2.1315419674_dp
      aab_pm6(50, 17) = 1.7715220451_dp
      aab_pm6(50, 32) = 2.5246329308_dp
      aab_pm6(50, 34) = 2.1273770332_dp
      aab_pm6(50, 35) = 1.5350890160_dp
      aab_pm6(50, 50) = 0.9210000038_dp
      aab_pm6(51, 1) = 1.5712720156_dp
      aab_pm6(51, 6) = 1.6962059736_dp
      aab_pm6(51, 7) = 0.6761149764_dp
      aab_pm6(51, 8) = 1.8463840485_dp
      aab_pm6(51, 9) = 2.1829218864_dp
      aab_pm6(51, 13) = 1.4226410389_dp
      aab_pm6(51, 14) = 2.6865899563_dp
      aab_pm6(51, 16) = 1.4188369513_dp
      aab_pm6(51, 17) = 1.1172870398_dp
      aab_pm6(51, 25) = 2.4003200531_dp
      aab_pm6(51, 27) = 2.2046298981_dp
      aab_pm6(51, 35) = 1.0639159679_dp
      aab_pm6(51, 43) = 2.2048499584_dp
      aab_pm6(51, 44) = 2.2043499947_dp
      aab_pm6(51, 45) = 2.2049300671_dp
      aab_pm6(51, 49) = 2.1419329643_dp
      aab_pm6(51, 51) = 1.3485349417_dp
      aab_pm6(52, 1) = 2.0391299725_dp
      aab_pm6(52, 6) = 1.9928159714_dp
      aab_pm6(52, 7) = 1.7222690582_dp
      aab_pm6(52, 8) = 1.8530639410_dp
      aab_pm6(52, 9) = 1.9985760450_dp
      aab_pm6(52, 13) = 1.3875410557_dp
      aab_pm6(52, 15) = 1.4537179470_dp
      aab_pm6(52, 16) = 1.8301700354_dp
      aab_pm6(52, 17) = 1.3002599478_dp
      aab_pm6(52, 30) = 1.2189290524_dp
      aab_pm6(52, 32) = 2.3423719406_dp
      aab_pm6(52, 33) = 1.1892529726_dp
      aab_pm6(52, 34) = 1.5660079718_dp
      aab_pm6(52, 35) = 1.2509399652_dp
      aab_pm6(52, 48) = 1.3072619438_dp
      aab_pm6(52, 49) = 1.5409879684_dp
      aab_pm6(52, 50) = 1.7639410496_dp
      aab_pm6(52, 52) = 1.1649780273_dp
      aab_pm6(53, 1) = 2.1399130821_dp
      aab_pm6(53, 2) = 2.1729838848_dp
      aab_pm6(53, 3) = 2.1212511063_dp
      aab_pm6(53, 4) = 2.2880229950_dp
      aab_pm6(53, 5) = 2.6676049232_dp
      aab_pm6(53, 6) = 2.0687100887_dp
      aab_pm6(53, 7) = 1.6775180101_dp
      aab_pm6(53, 8) = 2.2889189720_dp
      aab_pm6(53, 9) = 2.2035799026_dp
      aab_pm6(53, 10) = 2.4144148827_dp
      aab_pm6(53, 11) = 1.4030900002_dp
      aab_pm6(53, 12) = 2.0451369286_dp
      aab_pm6(53, 13) = 1.8160680532_dp
      aab_pm6(53, 14) = 1.5595790148_dp
      aab_pm6(53, 15) = 2.1315929890_dp
      aab_pm6(53, 16) = 1.8551100492_dp
      aab_pm6(53, 17) = 1.5741610527_dp
      aab_pm6(53, 18) = 1.5765869617_dp
      aab_pm6(53, 19) = 1.5397139788_dp
      aab_pm6(53, 20) = 2.1964900494_dp
      aab_pm6(53, 21) = 1.8148839474_dp
      aab_pm6(53, 22) = 1.9334690571_dp
      aab_pm6(53, 23) = 2.6835200787_dp
      aab_pm6(53, 24) = 2.6342239380_dp
      aab_pm6(53, 25) = 2.2665998936_dp
      aab_pm6(53, 26) = 1.9128290415_dp
      aab_pm6(53, 27) = 3.2352039814_dp
      aab_pm6(53, 28) = 1.0853430033_dp
      aab_pm6(53, 29) = 0.8343049884_dp
      aab_pm6(53, 30) = 1.3947620392_dp
      aab_pm6(53, 31) = 1.6717289686_dp
      aab_pm6(53, 32) = 1.8174250126_dp
      aab_pm6(53, 33) = 1.2452620268_dp
      aab_pm6(53, 35) = 1.5793759823_dp
      aab_pm6(53, 36) = 1.2385740280_dp
      aab_pm6(53, 37) = 1.4326750040_dp
      aab_pm6(53, 38) = 1.2620420456_dp
      aab_pm6(53, 39) = 1.2791099548_dp
      aab_pm6(53, 40) = 1.9951820374_dp
      aab_pm6(53, 41) = 1.9672509432_dp
      aab_pm6(53, 42) = 0.9484609962_dp
      aab_pm6(53, 43) = 1.2923120260_dp
      aab_pm6(53, 44) = 3.9532029629_dp
      aab_pm6(53, 45) = 3.7081699371_dp
      aab_pm6(53, 46) = 5.1445441246_dp
      aab_pm6(53, 47) = 2.5931611061_dp
      aab_pm6(53, 48) = 0.9962379932_dp
      aab_pm6(53, 49) = 2.3517580032_dp
      aab_pm6(53, 50) = 1.8556330204_dp
      aab_pm6(53, 51) = 1.1553150415_dp
      aab_pm6(53, 52) = 1.4939509630_dp
      aab_pm6(53, 53) = 1.5199249983_dp
      aab_pm6(54, 1) = 1.3568609953_dp
      aab_pm6(54, 2) = 2.4978320599_dp
      aab_pm6(54, 3) = 2.4668951035_dp
      aab_pm6(54, 4) = 6.0000028610_dp
      aab_pm6(54, 5) = 5.0519571304_dp
      aab_pm6(54, 6) = 1.7044399977_dp
      aab_pm6(54, 7) = 1.9329520464_dp
      aab_pm6(54, 8) = 0.8392329812_dp
      aab_pm6(54, 9) = 1.1288119555_dp
      aab_pm6(54, 10) = 1.3302019835_dp
      aab_pm6(54, 11) = 2.1030030251_dp
      aab_pm6(54, 12) = 2.6984140873_dp
      aab_pm6(54, 13) = 2.4120390415_dp
      aab_pm6(54, 14) = 3.0870599747_dp
      aab_pm6(54, 17) = 1.5463960171_dp
      aab_pm6(54, 18) = 0.5915200114_dp
      aab_pm6(54, 19) = 1.1712499857_dp
      aab_pm6(54, 20) = 1.5106530190_dp
      aab_pm6(54, 35) = 1.4396179914_dp
      aab_pm6(54, 36) = 0.5515609980_dp
      aab_pm6(54, 37) = 1.0878230333_dp
      aab_pm6(54, 53) = 0.7991549969_dp
      aab_pm6(54, 54) = 1.2447619438_dp
      aab_pm6(55, 1) = 0.2648819983_dp
      aab_pm6(55, 5) = 1.4871100187_dp
      aab_pm6(55, 6) = 2.1471040249_dp
      aab_pm6(55, 7) = 2.4465320110_dp
      aab_pm6(55, 8) = 2.0851390362_dp
      aab_pm6(55, 9) = 2.8341000080_dp
      aab_pm6(55, 15) = 2.9249529839_dp
      aab_pm6(55, 16) = 0.2894119918_dp
      aab_pm6(55, 17) = 1.6736630201_dp
      aab_pm6(55, 35) = 1.1671890020_dp
      aab_pm6(55, 53) = 0.9195619822_dp
      aab_pm6(55, 55) = 1.1708430052_dp
      aab_pm6(56, 1) = 6.0001349449_dp
      aab_pm6(56, 6) = 0.7706260085_dp
      aab_pm6(56, 7) = 1.1482330561_dp
      aab_pm6(56, 8) = 1.2830179930_dp
      aab_pm6(56, 9) = 3.0006179810_dp
      aab_pm6(56, 13) = 2.1059238911_dp
      aab_pm6(56, 14) = 1.2404199839_dp
      aab_pm6(56, 16) = 0.7051879764_dp
      aab_pm6(56, 17) = 1.0710439682_dp
      aab_pm6(56, 22) = 2.1760399342_dp
      aab_pm6(56, 35) = 1.1903460026_dp
      aab_pm6(56, 53) = 0.9825279713_dp
      aab_pm6(56, 56) = 0.3392690122_dp
      aab_pm6(57, 1) = 0.8336669803_dp
      aab_pm6(57, 6) = 0.6048690081_dp
      aab_pm6(57, 7) = 0.7588809729_dp
      aab_pm6(57, 8) = 1.3183330297_dp
      aab_pm6(57, 9) = 2.3793349266_dp
      aab_pm6(57, 13) = 1.0035099983_dp
      aab_pm6(57, 14) = 2.0168199539_dp
      aab_pm6(57, 15) = 0.9544500113_dp
      aab_pm6(57, 16) = 1.8341289759_dp
      aab_pm6(57, 17) = 0.9937530160_dp
      aab_pm6(57, 35) = 0.7581840158_dp
      aab_pm6(57, 53) = 0.5926659703_dp
      aab_pm6(57, 57) = 4.2480669022_dp
      aab_pm6(64, 1) = 0.3908700049_dp
      aab_pm6(64, 6) = 0.4468699992_dp
      aab_pm6(64, 7) = 1.1594099998_dp
      aab_pm6(64, 8) = 0.8620399833_dp
      aab_pm6(64, 9) = 1.4979799986_dp
      aab_pm6(64, 13) = 1.0035099983_dp
      aab_pm6(64, 14) = 2.0168199539_dp
      aab_pm6(64, 15) = 0.9544500113_dp
      aab_pm6(64, 16) = 2.0039300919_dp
      aab_pm6(64, 17) = 0.8068100214_dp
      aab_pm6(64, 35) = 0.7158100009_dp
      aab_pm6(64, 53) = 0.5853599906_dp
      aab_pm6(64, 64) = 3.3481800556_dp
      aab_pm6(71, 1) = 1.4157899618_dp
      aab_pm6(71, 6) = 2.3128130436_dp
      aab_pm6(71, 7) = 2.1413021088_dp
      aab_pm6(71, 8) = 2.1924860477_dp
      aab_pm6(71, 15) = 5.6188201904_dp
      aab_pm6(71, 17) = 2.7536358833_dp
      aab_pm6(71, 35) = 2.3226180077_dp
      aab_pm6(71, 53) = 2.2483479977_dp
      aab_pm6(72, 1) = 1.4237879515_dp
      aab_pm6(72, 5) = 1.6334999800_dp
      aab_pm6(72, 6) = 1.0021940470_dp
      aab_pm6(72, 7) = 1.3324099779_dp
      aab_pm6(72, 8) = 1.6332889795_dp
      aab_pm6(72, 9) = 2.2908029556_dp
      aab_pm6(72, 12) = 1.9113500118_dp
      aab_pm6(72, 13) = 0.9491500258_dp
      aab_pm6(72, 14) = 2.1893000603_dp
      aab_pm6(72, 15) = 1.2312200069_dp
      aab_pm6(72, 16) = 2.3271100521_dp
      aab_pm6(72, 17) = 1.2971169949_dp
      aab_pm6(72, 20) = 2.0545001030_dp
      aab_pm6(72, 33) = 1.7994999886_dp
      aab_pm6(72, 35) = 1.0907590389_dp
      aab_pm6(72, 53) = 1.0140960217_dp
      aab_pm6(72, 56) = 2.2648301125_dp
      aab_pm6(72, 72) = 0.5441439748_dp
      aab_pm6(73, 1) = 2.2880139351_dp
      aab_pm6(73, 6) = 1.8389489651_dp
      aab_pm6(73, 7) = 2.0536789894_dp
      aab_pm6(73, 8) = 2.4126288891_dp
      aab_pm6(73, 9) = 3.1073899269_dp
      aab_pm6(73, 11) = 2.5511200428_dp
      aab_pm6(73, 15) = 2.5137999058_dp
      aab_pm6(73, 16) = 2.2467229366_dp
      aab_pm6(73, 17) = 1.6088049412_dp
      aab_pm6(73, 19) = 4.5214700699_dp
      aab_pm6(73, 35) = 1.6403759718_dp
      aab_pm6(73, 53) = 2.4010529518_dp
      aab_pm6(73, 73) = 2.0828630924_dp
      aab_pm6(74, 1) = 2.1308801174_dp
      aab_pm6(74, 6) = 2.0974800587_dp
      aab_pm6(74, 7) = 1.5960400105_dp
      aab_pm6(74, 8) = 1.3590199947_dp
      aab_pm6(74, 9) = 1.4460500479_dp
      aab_pm6(74, 11) = 2.5510299206_dp
      aab_pm6(74, 15) = 2.3380599022_dp
      aab_pm6(74, 16) = 1.5425699949_dp
      aab_pm6(74, 17) = 1.3106900454_dp
      aab_pm6(74, 19) = 4.5213799477_dp
      aab_pm6(74, 35) = 1.2932599783_dp
      aab_pm6(74, 53) = 1.5735700130_dp
      aab_pm6(74, 74) = 2.9408700466_dp
      aab_pm6(75, 1) = 1.6345000267_dp
      aab_pm6(75, 6) = 2.3062849045_dp
      aab_pm6(75, 7) = 1.9183319807_dp
      aab_pm6(75, 8) = 1.9677469730_dp
      aab_pm6(75, 9) = 2.1542189121_dp
      aab_pm6(75, 14) = 2.7759299278_dp
      aab_pm6(75, 15) = 1.8041679859_dp
      aab_pm6(75, 16) = 1.0839190483_dp
      aab_pm6(75, 17) = 1.4338749647_dp
      aab_pm6(75, 32) = 2.8523399830_dp
      aab_pm6(75, 34) = 2.5231699944_dp
      aab_pm6(75, 35) = 1.6030600071_dp
      aab_pm6(75, 51) = 2.2043600082_dp
      aab_pm6(75, 53) = 2.6101191044_dp
      aab_pm6(75, 75) = 6.0002579689_dp
      aab_pm6(76, 1) = 3.4041800499_dp
      aab_pm6(76, 6) = 2.3364999294_dp
      aab_pm6(76, 7) = 1.1430900097_dp
      aab_pm6(76, 8) = 1.3503600359_dp
      aab_pm6(76, 9) = 1.5076199770_dp
      aab_pm6(76, 11) = 2.5507400036_dp
      aab_pm6(76, 15) = 2.8360900879_dp
      aab_pm6(76, 16) = 2.8094999790_dp
      aab_pm6(76, 17) = 1.8330700397_dp
      aab_pm6(76, 19) = 4.5210900307_dp
      aab_pm6(76, 35) = 1.7668800354_dp
      aab_pm6(76, 53) = 2.2037599087_dp
      aab_pm6(76, 76) = 2.0216300488_dp
      aab_pm6(77, 1) = 1.0339000225_dp
      aab_pm6(77, 6) = 1.6902949810_dp
      aab_pm6(77, 7) = 3.9345080853_dp
      aab_pm6(77, 8) = 3.7482719421_dp
      aab_pm6(77, 9) = 2.9827990532_dp
      aab_pm6(77, 11) = 2.5508201122_dp
      aab_pm6(77, 15) = 2.7140600681_dp
      aab_pm6(77, 16) = 3.2048339844_dp
      aab_pm6(77, 17) = 2.0097699165_dp
      aab_pm6(77, 19) = 4.5211701393_dp
      aab_pm6(77, 35) = 2.0381419659_dp
      aab_pm6(77, 53) = 3.4109139442_dp
      aab_pm6(77, 77) = 5.7716631889_dp
      aab_pm6(78, 1) = 4.0011978149_dp
      aab_pm6(78, 6) = 3.3067219257_dp
      aab_pm6(78, 7) = 2.3079230785_dp
      aab_pm6(78, 8) = 2.1105630398_dp
      aab_pm6(78, 9) = 3.7144410610_dp
      aab_pm6(78, 13) = 1.5723600388_dp
      aab_pm6(78, 14) = 0.9999899864_dp
      aab_pm6(78, 15) = 1.4032390118_dp
      aab_pm6(78, 16) = 2.7915000916_dp
      aab_pm6(78, 17) = 2.1085259914_dp
      aab_pm6(78, 35) = 2.1853070259_dp
      aab_pm6(78, 53) = 3.0773379803_dp
      aab_pm6(78, 78) = 3.4042758942_dp
      aab_pm6(79, 1) = 3.3690409660_dp
      aab_pm6(79, 6) = 4.5800161362_dp
      aab_pm6(79, 7) = 2.1380949020_dp
      aab_pm6(79, 8) = 1.5487630367_dp
      aab_pm6(79, 9) = 4.4531450272_dp
      aab_pm6(79, 13) = 1.5725699663_dp
      aab_pm6(79, 15) = 1.6187130213_dp
      aab_pm6(79, 16) = 4.3062381744_dp
      aab_pm6(79, 17) = 3.5394139290_dp
      aab_pm6(79, 35) = 0.5819110274_dp
      aab_pm6(79, 53) = 0.5779160261_dp
      aab_pm6(79, 79) = 0.9031620026_dp
      aab_pm6(80, 1) = 1.1365870237_dp
      aab_pm6(80, 6) = 0.7958160043_dp
      aab_pm6(80, 7) = 0.3321520090_dp
      aab_pm6(80, 8) = 1.0521450043_dp
      aab_pm6(80, 9) = 1.2405719757_dp
      aab_pm6(80, 14) = 2.7708599567_dp
      aab_pm6(80, 15) = 0.6086040139_dp
      aab_pm6(80, 16) = 1.0416820049_dp
      aab_pm6(80, 17) = 0.4307309985_dp
      aab_pm6(80, 22) = 3.4146299362_dp
      aab_pm6(80, 35) = 0.6387169957_dp
      aab_pm6(80, 52) = 0.2915000021_dp
      aab_pm6(80, 53) = 0.7581620216_dp
      aab_pm6(80, 80) = 0.4744130075_dp
      aab_pm6(81, 1) = 0.6736580133_dp
      aab_pm6(81, 5) = 1.5283470154_dp
      aab_pm6(81, 6) = 1.3903449774_dp
      aab_pm6(81, 7) = 0.9823349714_dp
      aab_pm6(81, 8) = 1.5500680208_dp
      aab_pm6(81, 9) = 1.4695160389_dp
      aab_pm6(81, 16) = 0.9948509932_dp
      aab_pm6(81, 17) = 0.8461930156_dp
      aab_pm6(81, 35) = 0.8744189739_dp
      aab_pm6(81, 53) = 0.9020119905_dp
      aab_pm6(81, 81) = 1.1916840076_dp
      aab_pm6(82, 1) = 1.5226759911_dp
      aab_pm6(82, 3) = 1.0018099546_dp
      aab_pm6(82, 5) = 0.9111970067_dp
      aab_pm6(82, 6) = 1.5255930424_dp
      aab_pm6(82, 7) = 1.3173940182_dp
      aab_pm6(82, 8) = 1.7632100582_dp
      aab_pm6(82, 9) = 3.2889020443_dp
      aab_pm6(82, 15) = 4.5167999268_dp
      aab_pm6(82, 16) = 1.0275189877_dp
      aab_pm6(82, 17) = 1.0941230059_dp
      aab_pm6(82, 23) = 1.5000000000_dp
      aab_pm6(82, 24) = 1.8607599735_dp
      aab_pm6(82, 30) = 1.5000000000_dp
      aab_pm6(82, 34) = 2.0000000000_dp
      aab_pm6(82, 35) = 0.8655499816_dp
      aab_pm6(82, 41) = 1.5000000000_dp
      aab_pm6(82, 42) = 2.0000000000_dp
      aab_pm6(82, 52) = 1.0025589466_dp
      aab_pm6(82, 53) = 0.9834740162_dp
      aab_pm6(82, 82) = 1.8817640543_dp
      aab_pm6(83, 1) = 1.6799050570_dp
      aab_pm6(83, 3) = 0.3401400149_dp
      aab_pm6(83, 6) = 1.5340249538_dp
      aab_pm6(83, 7) = 1.1438759565_dp
      aab_pm6(83, 8) = 1.5532970428_dp
      aab_pm6(83, 9) = 2.3554000854_dp
      aab_pm6(83, 16) = 1.4668790102_dp
      aab_pm6(83, 17) = 1.2729749680_dp
      aab_pm6(83, 34) = 1.3447459936_dp
      aab_pm6(83, 35) = 1.1462329626_dp
      aab_pm6(83, 53) = 1.3021709919_dp
      aab_pm6(83, 83) = 1.0740640163_dp

      DO i = 0, nelem
         DO j = i+1, nelem
            xab = xab_pm6(i, j)+xab_pm6(j, i)
            xab_pm6(i, j) = xab
            xab_pm6(j, i) = xab
            aab = aab_pm6(i, j)+aab_pm6(j, i)
            aab_pm6(i, j) = aab
            aab_pm6(j, i) = aab
         END DO
      END DO

      init = 1

   END SUBROUTINE init_pm6_pair_params

END MODULE semi_empirical_parameters

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Working with the  semi empirical parameter types.
!> \author JGH (14.08.2004)
! **************************************************************************************************
MODULE semi_empirical_utils
   USE basis_set_types,                 ONLY: allocate_sto_basis_set,&
                                              create_gto_from_sto_basis,&
                                              gto_basis_set_type,&
                                              set_sto_basis_set
   USE cell_types,                      ONLY: cell_type,&
                                              plane_distance
   USE cp_control_types,                ONLY: semi_empirical_control_type
   USE input_constants,                 ONLY: &
        do_method_am1, do_method_mndo, do_method_mndod, do_method_pchg, do_method_pdg, &
        do_method_pm3, do_method_pm6, do_method_pnnl, do_method_rm1
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE semi_empirical_mpole_methods,    ONLY: semi_empirical_mpole_p_setup
   USE semi_empirical_par_utils,        ONLY: get_se_basis,&
                                              setup_1c_2el_int
   USE semi_empirical_parameters,       ONLY: am1_default_parameter,&
                                              mndo_default_parameter,&
                                              pcharge_default_parameter,&
                                              pdg_default_parameter,&
                                              pm3_default_parameter,&
                                              pm6_default_parameter,&
                                              pnnl_default_parameter,&
                                              rm1_default_parameter
   USE semi_empirical_types,            ONLY: se_taper_type,&
                                              semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_utils'

   PUBLIC :: init_se_param, se_param_set_default, get_se_type, &
             initialize_se_taper, finalize_se_taper, se_cutoff_compatible

CONTAINS
! **************************************************************************************************
!> \brief  Reset cutoffs trying to be somehow a bit smarter
!> \param se_control ...
!> \param se_section ...
!> \param cell ...
!> \param output_unit ...
!> \author Teodoro Laino [tlaino] - 03.2009
! **************************************************************************************************
   SUBROUTINE se_cutoff_compatible(se_control, se_section, cell, output_unit)
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(section_vals_type), POINTER                   :: se_section
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'se_cutoff_compatible', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: explicit1, explicit2
      REAL(KIND=dp)                                      :: rc

! Coulomb Cutoff Taper

      CALL section_vals_val_get(se_section, "COULOMB%CUTOFF", explicit=explicit1)
      CALL section_vals_val_get(se_section, "COULOMB%RC_TAPER", explicit=explicit2)
      IF ((.NOT. explicit1) .AND. se_control%do_ewald_gks) THEN
         rc = MAX(0.5*plane_distance(1, 0, 0, cell), &
                  0.5*plane_distance(0, 1, 0, cell), &
                  0.5*plane_distance(0, 0, 1, cell))
         IF (rc /= se_control%cutoff_cou) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, *)
               WRITE (output_unit, '(A,T37,A)') " SEMIEMPIRICAL|", &
                  " Coulomb Integral cutoff/taper was redefined"
               WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Old value [a.u.]", &
                  se_control%cutoff_cou
               WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| New value [a.u.]", rc
               WRITE (output_unit, *)
            END IF
         END IF
         se_control%cutoff_cou = rc
         IF (.NOT. explicit2) se_control%taper_cou = rc
      ELSE IF ((.NOT. explicit1) .AND. (ALL(cell%perd == 0))) THEN
         rc = MAX(plane_distance(1, 0, 0, cell), &
                  plane_distance(0, 1, 0, cell), &
                  plane_distance(0, 0, 1, cell))
         IF (rc /= se_control%cutoff_cou) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, *)
               WRITE (output_unit, '(A,T37,A)') " SEMIEMPIRICAL|", &
                  " Coulomb Integral cutoff/taper was redefined"
               WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Old value [a.u.]", &
                  se_control%cutoff_cou
               WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| New value [a.u.]", rc
               WRITE (output_unit, *)
            END IF
         END IF
         se_control%cutoff_cou = rc
         IF (.NOT. explicit2) se_control%taper_cou = rc
      END IF
      IF (output_unit > 0) THEN
         WRITE (output_unit, *)
         WRITE (output_unit, '(A,T44,A)') " SEMIEMPIRICAL|", &
            " Coulomb Integral cutoff/taper values"
         WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Cutoff [a.u.]", &
            se_control%cutoff_cou
         WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Taper  [a.u.]", &
            se_control%taper_cou
         WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Range  [a.u.]", &
            se_control%range_cou
         WRITE (output_unit, *)
      END IF
      ! Exchange Cutoff Taper
      CALL section_vals_val_get(se_section, "EXCHANGE%CUTOFF", explicit=explicit1)
      CALL section_vals_val_get(se_section, "EXCHANGE%RC_TAPER", explicit=explicit2)
      rc = se_control%cutoff_exc
      IF (.NOT. explicit1) THEN
         rc = MIN(rc, MAX(0.25_dp*plane_distance(1, 0, 0, cell), &
                          0.25_dp*plane_distance(0, 1, 0, cell), &
                          0.25_dp*plane_distance(0, 0, 1, cell)))

         IF (rc /= se_control%cutoff_exc) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, *)
               WRITE (output_unit, '(A,T36,A)') " SEMIEMPIRICAL|", &
                  " Exchange Integral cutoff/taper was redefined"
               WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Default value [a.u.]", &
                  se_control%cutoff_exc
               WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| New value [a.u.]", rc
               WRITE (output_unit, *)
            END IF
         END IF
      END IF
      se_control%cutoff_exc = rc
      IF (.NOT. explicit2) se_control%taper_exc = rc

      IF (output_unit > 0) THEN
         WRITE (output_unit, *)
         WRITE (output_unit, '(A,T43,A)') " SEMIEMPIRICAL|", &
            " Exchange Integral cutoff/taper values"
         WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Cutoff [a.u.]", &
            se_control%cutoff_exc
         WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Taper  [a.u.]", &
            se_control%taper_exc
         WRITE (output_unit, '(A,T71,F10.3)') " SEMIEMPIRICAL| Range  [a.u.]", &
            se_control%range_exc
         WRITE (output_unit, *)
      END IF

   END SUBROUTINE se_cutoff_compatible

! **************************************************************************************************
!> \brief  Initializes the semi-empirical taper for a chunk calculation
!> \param se_taper ...
!> \param coulomb ...
!> \param exchange ...
!> \param lr_corr ...
!> \author Teodoro Laino [tlaino] - 03.2009
! **************************************************************************************************
   SUBROUTINE initialize_se_taper(se_taper, coulomb, exchange, lr_corr)
      TYPE(se_taper_type), POINTER                       :: se_taper
      LOGICAL, INTENT(IN), OPTIONAL                      :: coulomb, exchange, lr_corr

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_se_taper', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check, l_coulomb, l_exchange, l_lrc

      check = .NOT. ASSOCIATED(se_taper%taper)
      IF(.NOT.(check))CALL cp__a("semi_empirical_utils.F",177)
      l_coulomb = .FALSE.
      l_exchange = .FALSE.
      l_lrc = .FALSE.
      IF (PRESENT(coulomb)) l_coulomb = coulomb
      IF (PRESENT(exchange)) l_exchange = exchange
      IF (PRESENT(lr_corr)) l_lrc = lr_corr
      IF (l_coulomb) THEN
         check = (.NOT. l_exchange) .AND. (.NOT. l_lrc)
         IF(.NOT.(check))CALL cp__a("semi_empirical_utils.F",186)
         se_taper%taper => se_taper%taper_cou
      END IF
      IF (l_exchange) THEN
         check = (.NOT. l_coulomb) .AND. (.NOT. l_lrc)
         IF(.NOT.(check))CALL cp__a("semi_empirical_utils.F",191)
         se_taper%taper => se_taper%taper_exc
      END IF
      IF (l_lrc) THEN
         check = (.NOT. l_coulomb) .AND. (.NOT. l_exchange)
         IF(.NOT.(check))CALL cp__a("semi_empirical_utils.F",196)
         se_taper%taper => se_taper%taper_lrc
      END IF
   END SUBROUTINE initialize_se_taper

! **************************************************************************************************
!> \brief  Finalizes the semi-empirical taper for a chunk calculation
!> \param se_taper ...
!> \author Teodoro Laino [tlaino] - 03.2009
! **************************************************************************************************
   SUBROUTINE finalize_se_taper(se_taper)
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'finalize_se_taper', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      check = ASSOCIATED(se_taper%taper)
      IF(.NOT.(check))CALL cp__a("semi_empirical_utils.F",215)
      NULLIFY (se_taper%taper)
   END SUBROUTINE finalize_se_taper

! **************************************************************************************************
!> \brief Initialize semi_empirical type
!> \param sep ...
!> \param orb_basis_set ...
!> \param ngauss ...
! **************************************************************************************************
   SUBROUTINE init_se_param(sep, orb_basis_set, ngauss)
      TYPE(semi_empirical_type), POINTER                 :: sep
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      INTEGER, INTENT(IN)                                :: ngauss

      CHARACTER(len=*), PARAMETER :: routineN = 'init_se_param', routineP = moduleN//':'//routineN

      CHARACTER(LEN=6), DIMENSION(:), POINTER            :: symbol
      INTEGER                                            :: l, nshell
      INTEGER, DIMENSION(:), POINTER                     :: lq, nq
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet

      IF (ASSOCIATED(sep)) THEN
         CALL allocate_sto_basis_set(sep%basis)
         nshell = 0
         IF (sep%natorb == 1) nshell = 1
         IF (sep%natorb == 4) nshell = 2
         IF (sep%natorb == 9) nshell = 3
         ALLOCATE (nq(0:3), lq(0:3), zet(0:3))

         ALLOCATE (symbol(0:3))

         symbol = ""
         nq = 0
         lq = 0
         zet = 0._dp
         DO l = 0, nshell-1
            nq(l) = get_se_basis(sep, l)
            lq(l) = l
            zet(l) = sep%sto_exponents(l)
            IF (l == 0) WRITE (symbol(0), '(I1,A1)') nq(l), "S"
            IF (l == 1) WRITE (symbol(1), '(I1,A1)') nq(l), "P"
            IF (l == 2) WRITE (symbol(2), '(I1,A1)') nq(l), "D"
         END DO

         IF (nshell > 0) THEN
            sep%ngauss = ngauss
            CALL set_sto_basis_set(sep%basis, name=sep%name, nshell=nshell, symbol=symbol, &
                                   nq=nq, lq=lq, zet=zet)
            CALL create_gto_from_sto_basis(sep%basis, orb_basis_set, sep%ngauss)
         END IF

         DEALLOCATE (nq)
         DEALLOCATE (lq)
         DEALLOCATE (zet)
         DEALLOCATE (symbol)
      ELSE
         CALL cp__b("semi_empirical_utils.F",272,"The pointer sep is not associated")
      END IF

   END SUBROUTINE init_se_param

! **************************************************************************************************
!> \brief Initialize parameter for a semi_empirival type
!> \param sep ...
!> \param z ...
!> \param method ...
! **************************************************************************************************
   SUBROUTINE se_param_set_default(sep, z, method)

      TYPE(semi_empirical_type), POINTER                 :: sep
      INTEGER, INTENT(IN)                                :: z, method

      CHARACTER(len=*), PARAMETER :: routineN = 'se_param_set_default', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(sep)) THEN
         IF (z < 0) THEN
            CALL cp__b("semi_empirical_utils.F",293,"Atomic number < 0")
         END IF
         SELECT CASE (method)
         CASE (do_method_am1)
            CALL am1_default_parameter(sep, z)
         CASE (do_method_rm1)
            CALL rm1_default_parameter(sep, z)
         CASE (do_method_pm3)
            CALL pm3_default_parameter(sep, z)
         CASE (do_method_pm6)
            CALL pm6_default_parameter(sep, z)
         CASE (do_method_pdg)
            CALL pdg_default_parameter(sep, z)
         CASE (do_method_mndo)
            CALL mndo_default_parameter(sep, z, do_method_mndo)
         CASE (do_method_mndod)
            CALL mndo_default_parameter(sep, z, do_method_mndod)
         CASE (do_method_pnnl)
            CALL pnnl_default_parameter(sep, z)
         CASE (do_method_pchg)
            CALL pcharge_default_parameter(sep, z)
         CASE DEFAULT
            CALL cp__b("semi_empirical_utils.F",315,"Semiempirical method unknown")
         END SELECT
      ELSE
         CALL cp__b("semi_empirical_utils.F",318,"The pointer sep is not associated")
      END IF

      ! Check if the element has been defined..
      IF (.NOT. sep%defined) &
         CALL cp_abort(cp__l("semi_empirical_utils.F",323), &
                       "Semiempirical type ("//TRIM(sep%name)//") cannot be defined for "// &
                       "the requested parameterization.")

      ! Fill 1 center - 2 electron integrals
      CALL setup_1c_2el_int(sep)

      ! Fill multipolar expansion of atomic orbitals charge distributions
      CALL semi_empirical_mpole_p_setup(sep%w_mpole, sep, method)

      ! Get the value of the size of CORE integral array
      sep%core_size = 0
      IF (sep%natorb > 0) sep%core_size = 1
      IF (sep%natorb > 1) sep%core_size = 4
      IF (sep%dorb) sep%core_size = 10

      ! Get size of the all possible combinations of atomic orbitals
      sep%atm_int_size = (sep%natorb+1)*sep%natorb/2

   END SUBROUTINE se_param_set_default

! **************************************************************************************************
!> \brief Gives back the unique semi_empirical METHOD type
!> \param se_method ...
!> \retval se_type ...
! **************************************************************************************************
   FUNCTION get_se_type(se_method) RESULT(se_type)

      INTEGER, INTENT(IN)                                :: se_method
      INTEGER                                            :: se_type

      CHARACTER(len=*), PARAMETER :: routineN = 'get_se_type', routineP = moduleN//':'//routineN

      SELECT CASE (se_method)
      CASE DEFAULT
         se_type = se_method
      CASE (do_method_am1, do_method_rm1)
         se_type = do_method_am1
      END SELECT

   END FUNCTION get_se_type

END MODULE semi_empirical_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Working with the DFTB parameter types.
!> \author JGH (24.02.2007)
! **************************************************************************************************
MODULE qs_dftb_utils

   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dftb_utils'

   PUBLIC :: allocate_dftb_atom_param, &
             deallocate_dftb_atom_param, &
             get_dftb_atom_param, &
             set_dftb_atom_param, &
             write_dftb_atom_param

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param dftb_parameter ...
! **************************************************************************************************
   SUBROUTINE allocate_dftb_atom_param(dftb_parameter)

      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_dftb_atom_param', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dftb_parameter)) &
         CALL deallocate_dftb_atom_param(dftb_parameter)

      ALLOCATE (dftb_parameter)

      dftb_parameter%defined = .FALSE.
      dftb_parameter%name = ""
      dftb_parameter%typ = "NONE"
      dftb_parameter%z = -1
      dftb_parameter%zeff = -1.0_dp
      dftb_parameter%natorb = 0
      dftb_parameter%lmax = -1
      dftb_parameter%skself = 0.0_dp
      dftb_parameter%occupation = 0.0_dp
      dftb_parameter%eta = 0.0_dp
      dftb_parameter%energy = 0.0_dp
      dftb_parameter%xi = 0.0_dp
      dftb_parameter%di = 0.0_dp
      dftb_parameter%rcdisp = 0.0_dp
      dftb_parameter%dudq = 0.0_dp

   END SUBROUTINE allocate_dftb_atom_param

! **************************************************************************************************
!> \brief ...
!> \param dftb_parameter ...
! **************************************************************************************************
   SUBROUTINE deallocate_dftb_atom_param(dftb_parameter)

      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_dftb_atom_param', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(dftb_parameter)))CALL cp__a("qs_dftb_utils.F",83)
      DEALLOCATE (dftb_parameter)

   END SUBROUTINE deallocate_dftb_atom_param

! **************************************************************************************************
!> \brief ...
!> \param dftb_parameter ...
!> \param name ...
!> \param typ ...
!> \param defined ...
!> \param z ...
!> \param zeff ...
!> \param natorb ...
!> \param lmax ...
!> \param skself ...
!> \param occupation ...
!> \param eta ...
!> \param energy ...
!> \param cutoff ...
!> \param xi ...
!> \param di ...
!> \param rcdisp ...
!> \param dudq ...
! **************************************************************************************************
   SUBROUTINE get_dftb_atom_param(dftb_parameter, name, typ, defined, z, zeff, natorb, &
                                  lmax, skself, occupation, eta, energy, cutoff, xi, di, rcdisp, dudq)

      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name, typ
      LOGICAL, INTENT(OUT), OPTIONAL                     :: defined
      INTEGER, INTENT(OUT), OPTIONAL                     :: z
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: zeff
      INTEGER, INTENT(OUT), OPTIONAL                     :: natorb, lmax
      REAL(KIND=dp), DIMENSION(0:3), OPTIONAL            :: skself, occupation, eta
      REAL(KIND=dp), OPTIONAL                            :: energy, cutoff, xi, di, rcdisp, dudq

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_dftb_atom_param', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(dftb_parameter)))CALL cp__a("qs_dftb_utils.F",124)

      IF (PRESENT(name)) name = dftb_parameter%name
      IF (PRESENT(typ)) typ = dftb_parameter%typ
      IF (PRESENT(defined)) defined = dftb_parameter%defined
      IF (PRESENT(z)) z = dftb_parameter%z
      IF (PRESENT(zeff)) zeff = dftb_parameter%zeff
      IF (PRESENT(natorb)) natorb = dftb_parameter%natorb
      IF (PRESENT(lmax)) lmax = dftb_parameter%lmax
      IF (PRESENT(skself)) skself = dftb_parameter%skself
      IF (PRESENT(eta)) eta = dftb_parameter%eta
      IF (PRESENT(energy)) energy = dftb_parameter%energy
      IF (PRESENT(cutoff)) cutoff = dftb_parameter%cutoff
      IF (PRESENT(occupation)) occupation = dftb_parameter%occupation
      IF (PRESENT(xi)) xi = dftb_parameter%xi
      IF (PRESENT(di)) di = dftb_parameter%di
      IF (PRESENT(rcdisp)) rcdisp = dftb_parameter%rcdisp
      IF (PRESENT(dudq)) dudq = dftb_parameter%dudq

   END SUBROUTINE get_dftb_atom_param

! **************************************************************************************************
!> \brief ...
!> \param dftb_parameter ...
!> \param name ...
!> \param typ ...
!> \param defined ...
!> \param z ...
!> \param zeff ...
!> \param natorb ...
!> \param lmax ...
!> \param skself ...
!> \param occupation ...
!> \param eta ...
!> \param energy ...
!> \param cutoff ...
!> \param xi ...
!> \param di ...
!> \param rcdisp ...
!> \param dudq ...
! **************************************************************************************************
   SUBROUTINE set_dftb_atom_param(dftb_parameter, name, typ, defined, z, zeff, natorb, &
                                  lmax, skself, occupation, eta, energy, cutoff, xi, di, rcdisp, dudq)

      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter
      CHARACTER(LEN=default_string_length), INTENT(IN), &
         OPTIONAL                                        :: name, typ
      LOGICAL, INTENT(IN), OPTIONAL                      :: defined
      INTEGER, INTENT(IN), OPTIONAL                      :: z
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: zeff
      INTEGER, INTENT(IN), OPTIONAL                      :: natorb, lmax
      REAL(KIND=dp), DIMENSION(0:3), OPTIONAL            :: skself, occupation, eta
      REAL(KIND=dp), OPTIONAL                            :: energy, cutoff, xi, di, rcdisp, dudq

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_dftb_atom_param', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(dftb_parameter)))CALL cp__a("qs_dftb_utils.F",181)

      IF (PRESENT(name)) dftb_parameter%name = name
      IF (PRESENT(typ)) dftb_parameter%typ = typ
      IF (PRESENT(defined)) dftb_parameter%defined = defined
      IF (PRESENT(z)) dftb_parameter%z = z
      IF (PRESENT(zeff)) dftb_parameter%zeff = zeff
      IF (PRESENT(natorb)) dftb_parameter%natorb = natorb
      IF (PRESENT(lmax)) dftb_parameter%lmax = lmax
      IF (PRESENT(skself)) dftb_parameter%skself = skself
      IF (PRESENT(eta)) dftb_parameter%eta = eta
      IF (PRESENT(occupation)) dftb_parameter%occupation = occupation
      IF (PRESENT(energy)) dftb_parameter%energy = energy
      IF (PRESENT(cutoff)) dftb_parameter%cutoff = cutoff
      IF (PRESENT(xi)) dftb_parameter%xi = xi
      IF (PRESENT(di)) dftb_parameter%di = di
      IF (PRESENT(rcdisp)) dftb_parameter%rcdisp = rcdisp
      IF (PRESENT(dudq)) dftb_parameter%dudq = dudq

   END SUBROUTINE set_dftb_atom_param

! **************************************************************************************************
!> \brief ...
!> \param dftb_parameter ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE write_dftb_atom_param(dftb_parameter, subsys_section)

      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_dftb_atom_param', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name, typ
      INTEGER                                            :: lmax, natorb, output_unit, z
      LOGICAL                                            :: defined
      REAL(dp)                                           :: zeff
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      IF (ASSOCIATED(dftb_parameter) .AND. &
          BTEST(cp_print_key_should_output(logger%iter_info, subsys_section, &
                                           "PRINT%KINDS/POTENTIAL"), cp_p_file)) THEN

         output_unit = cp_print_key_unit_nr(logger, subsys_section, "PRINT%KINDS", &
                                            extension=".Log")

         IF (output_unit > 0) THEN
            CALL get_dftb_atom_param(dftb_parameter, name=name, typ=typ, defined=defined, &
                                     z=z, zeff=zeff, natorb=natorb, lmax=lmax)

            WRITE (UNIT=output_unit, FMT="(/,A,T67,A14)") &
               " DFTB  parameters: ", TRIM(name)
            IF (defined) THEN
               WRITE (UNIT=output_unit, FMT="(T16,A,T71,F10.2)") &
                  "Effective core charge:", zeff
               WRITE (UNIT=output_unit, FMT="(T16,A,T71,I10)") &
                  "Number of orbitals:", natorb
            ELSE
               WRITE (UNIT=output_unit, FMT="(T55,A)") &
                  "Parameters are not defined"
            END IF
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                           "PRINT%KINDS")
      END IF

   END SUBROUTINE write_dftb_atom_param

END MODULE qs_dftb_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate spherical harmonics
!> \note
!>      Spherical Harmonics
!>          Numerical Stability up to L=15
!>          Accuracy > 1.E-12 up to L=15 tested
!>          Definition is consistent with orbital_transformation_matrices
!>      Clebsch-Gordon Coefficients
!>          Tested up to l=7 (i.e. L=14)
!> \todo
!>      1) Check if this definition is consistent with the
!>         Slater-Koster module
!> \par History
!>      JGH 28-Feb-2002 : Change of sign convention (-1^m)
!>      JGH  1-Mar-2002 : Clebsch-Gordon Coefficients
!> \author JGH 6-Oct-2000
! **************************************************************************************************
MODULE spherical_harmonics

   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              maxfac,&
                                              pi
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: y_lm
   PUBLIC :: clebsch_gordon_deallocate, clebsch_gordon_init, &
             clebsch_gordon
   PUBLIC :: legendre, dlegendre

   INTERFACE y_lm
      MODULE PROCEDURE rvy_lm, rry_lm, cvy_lm, ccy_lm
   END INTERFACE

   INTERFACE clebsch_gordon
      MODULE PROCEDURE clebsch_gordon_real, clebsch_gordon_complex
   END INTERFACE

   INTERFACE clebsch_gordon_getm
      MODULE PROCEDURE getm
   END INTERFACE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'spherical_harmonics'

   REAL(KIND=dp), DIMENSION(:, :, :), ALLOCATABLE :: cg_table
   INTEGER :: lmax = -1
   REAL(KIND=dp) :: osq2, sq2

CONTAINS

! Clebsch-Gordon Coefficients

! **************************************************************************************************
!> \brief ...
!> \param l1 ...
!> \param m1 ...
!> \param l2 ...
!> \param m2 ...
!> \param clm ...
! **************************************************************************************************
   SUBROUTINE clebsch_gordon_complex(l1, m1, l2, m2, clm)
      INTEGER, INTENT(IN)                                :: l1, m1, l2, m2
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: clm

      CHARACTER(len=*), PARAMETER :: routineN = 'clebsch_gordon_complex', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icase, ind, l, lm, lp, n

      l = l1+l2
      IF (l > lmax) CALL clebsch_gordon_init(l)
      n = l/2+1
      IF (n > SIZE(clm)) CALL cp__b("common/spherical_harmonics.F",85,"Array too small")

      IF ((m1 >= 0 .AND. m2 >= 0) .OR. (m1 < 0 .AND. m2 < 0)) THEN
         icase = 1
      ELSE
         icase = 2
      END IF
      ind = order(l1, m1, l2, m2)

      DO lp = MOD(l, 2), l, 2
         lm = lp/2+1
         clm(lm) = cg_table(ind, lm, icase)
      END DO

   END SUBROUTINE clebsch_gordon_complex

! **************************************************************************************************
!> \brief ...
!> \param l1 ...
!> \param m1 ...
!> \param l2 ...
!> \param m2 ...
!> \param rlm ...
! **************************************************************************************************
   SUBROUTINE clebsch_gordon_real(l1, m1, l2, m2, rlm)
      INTEGER, INTENT(IN)                                :: l1, m1, l2, m2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: rlm

      CHARACTER(len=*), PARAMETER :: routineN = 'clebsch_gordon_real', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icase1, icase2, ind, l, lm, lp, mm(2), n
      REAL(KIND=dp)                                      :: xsi

      l = l1+l2
      IF (l > lmax) CALL clebsch_gordon_init(l)
      n = l/2+1
      IF (n > SIZE(rlm, 1)) CALL cp__b("common/spherical_harmonics.F",122,"Array too small")

      ind = order(l1, m1, l2, m2)
      mm = getm(m1, m2)
      IF ((m1 >= 0 .AND. m2 >= 0) .OR. (m1 < 0 .AND. m2 < 0)) THEN
         icase1 = 1
         icase2 = 2
      ELSE
         icase1 = 2
         icase2 = 1
      END IF

      DO lp = MOD(l, 2), l, 2
         lm = lp/2+1
         xsi = get_factor(m1, m2, mm(1))
         rlm(lm, 1) = xsi*cg_table(ind, lm, icase1)
         xsi = get_factor(m1, m2, mm(2))
         rlm(lm, 2) = xsi*cg_table(ind, lm, icase2)
      END DO

   END SUBROUTINE clebsch_gordon_real

! **************************************************************************************************
!> \brief ...
!> \param m1 ...
!> \param m2 ...
!> \retval m ...
! **************************************************************************************************
   FUNCTION getm(m1, m2) RESULT(m)
      INTEGER, INTENT(IN)                                :: m1, m2
      INTEGER, DIMENSION(2)                              :: m

      INTEGER                                            :: mm, mp

      mp = m1+m2
      mm = m1-m2
      IF (m1*m2 < 0 .OR. (m1*m2 == 0 .AND. (m1 < 0 .OR. m2 < 0))) THEN
         mp = -ABS(mp)
         mm = -ABS(mm)
      ELSE
         mp = ABS(mp)
         mm = ABS(mm)
      END IF
      m(1) = mp
      m(2) = mm
   END FUNCTION getm

! **************************************************************************************************
!> \brief ...
!> \param m1 ...
!> \param m2 ...
!> \param m ...
!> \retval f ...
! **************************************************************************************************
   FUNCTION get_factor(m1, m2, m) RESULT(f)
      INTEGER, INTENT(IN)                                :: m1, m2, m
      REAL(KIND=dp)                                      :: f

      INTEGER                                            :: mx, my

      f = 1.0_dp
      IF (ABS(m1) >= ABS(m2)) THEN
         mx = m1
         my = m2
      ELSE
         mx = m2
         my = m1
      ENDIF
      IF (mx*my == 0) THEN
         f = 1.0_dp
      ELSE IF (m == 0) THEN
         IF (ABS(mx) /= ABS(my)) WRITE (*, '(A,3I6)') " 1) Illegal Case ", m1, m2, m
         IF (mx*my > 0) THEN
            f = 1.0_dp
         ELSE
            f = 0.0_dp
         END IF
      ELSE IF (ABS(mx)+ABS(my) == m) THEN
         f = osq2
         IF (mx < 0) f = -osq2
      ELSE IF (ABS(mx)+ABS(my) == -m) THEN
         f = osq2
      ELSE IF (ABS(mx)-ABS(my) == -m) THEN
         IF (mx*my > 0) WRITE (*, '(A,3I6)') " 2) Illegal Case ", m1, m2, m
         IF (mx > 0) f = -osq2
         IF (mx < 0) f = osq2
      ELSE IF (ABS(mx)-ABS(my) == m) THEN
         IF (mx*my < 0) WRITE (*, '(A,3I6)') " 3) Illegal Case ", m1, m2, m
         f = osq2
      ELSE
         WRITE (*, '(A,3I6)') " 4) Illegal Case ", m1, m2, m
      END IF
   END FUNCTION get_factor

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE clebsch_gordon_deallocate()
      CHARACTER(len=*), PARAMETER :: routineN = 'clebsch_gordon_deallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ALLOCATED(cg_table)) THEN
         DEALLOCATE (cg_table)
      END IF
      CALL timestop(handle)

   END SUBROUTINE clebsch_gordon_deallocate

! **************************************************************************************************
!> \brief ...
!> \param l ...
! **************************************************************************************************
   SUBROUTINE clebsch_gordon_init(l)
      INTEGER, INTENT(IN)                                :: l

      CHARACTER(len=*), PARAMETER :: routineN = 'clebsch_gordon_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i1, i2, ierr, ix, iy, l1, l2, &
                                                            lp, m, m1, m2, ml, mp, n

      CALL timeset(routineN, handle)

      sq2 = SQRT(2.0_dp)
      osq2 = 1.0_dp/sq2

      IF (l < 0) CALL cp__b("common/spherical_harmonics.F",252,"l < 0")
      IF (ALLOCATED(cg_table)) THEN
         DEALLOCATE (cg_table)
      END IF
      ! maximum size of table
      n = (l**4+6*l**3+15*l**2+18*l+8)/8
      m = l+1
      ALLOCATE (cg_table(n, m, 2), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 260, &
                                      "cg_table", dp_size*2*n*m)
      lmax = l

      DO l1 = 0, lmax
         DO m1 = 0, l1
            iy = (l1*(l1+1))/2+m1+1
            DO l2 = l1, lmax
               ml = 0
               IF (l1 == l2) ml = m1
               DO m2 = ml, l2
                  ix = (l2*(l2+1))/2+m2+1
                  i1 = (ix*(ix-1))/2+iy
                  DO lp = MOD(l1+l2, 2), l1+l2, 2
                     i2 = lp/2+1
                     mp = m2+m1
                     cg_table(i1, i2, 1) = cgc(l1, m1, l2, m2, lp, mp)
                     mp = ABS(m2-m1)
                     IF (m2 >= m1) THEN
                        cg_table(i1, i2, 2) = cgc(l1, m1, lp, mp, l2, m2)
                     ELSE
                        cg_table(i1, i2, 2) = cgc(l2, m2, lp, mp, l1, m1)
                     END IF
                  END DO
               END DO
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE clebsch_gordon_init

! **************************************************************************************************
!> \brief ...
!> \param l1 ...
!> \param m1 ...
!> \param l2 ...
!> \param m2 ...
!> \param lp ...
!> \param mp ...
!> \retval cgc ...
! **************************************************************************************************
   FUNCTION cgc(l1, m1, l2, m2, lp, mp)
      INTEGER, INTENT(IN)                                :: l1, m1, l2, m2, lp, mp
      REAL(KIND=dp)                                      :: cgc

      CHARACTER(len=*), PARAMETER :: routineN = 'cgc', routineP = moduleN//':'//routineN

      INTEGER                                            :: la, lb, ll, ma, mb, s, t, tmax, tmin, &
                                                            z1, z2, z3, z4, z5
      REAL(KIND=dp)                                      :: f1, f2, pref

! m1 >= 0; m2 >= 0; mp >= 0

      IF (m1 < 0 .OR. m2 < 0 .OR. mp < 0) THEN
         WRITE (*, *) l1, l2, lp
         WRITE (*, *) m1, m2, mp
         CALL cp__b("common/spherical_harmonics.F",318,"Illegal input values")
      END IF
      IF (l2 < l1) THEN
         la = l2
         ma = m2
         lb = l1
         mb = m1
      ELSE
         la = l1
         ma = m1
         lb = l2
         mb = m2
      END IF

      IF (MOD(la+lb+lp, 2) == 0 .AND. la+lb >= lp .AND. lp >= lb-la &
          .AND. lb-mb >= 0) THEN
         ll = (2*lp+1)*(2*la+1)*(2*lb+1)
         pref = 1.0_dp/SQRT(4.0_dp*pi)*0.5_dp*SQRT(REAL(ll, dp)* &
                                                   (sfac(lp-mp)/sfac(lp+mp))* &
                                                   (sfac(la-ma)/sfac(la+ma))*(sfac(lb-mb)/sfac(lb+mb)))
         s = (la+lb+lp)/2
         tmin = MAX(0, -lb+la-mp)
         tmax = MIN(lb+la-mp, lp-mp, la-ma)
         f1 = REAL(2*(-1)**(s-lb-ma), KIND=dp)*(sfac(lb+mb)/sfac(lb-mb))* &
              sfac(la+ma)/(sfac(s-lp)*sfac(s-lb))*sfac(2*s-2*la)/sfac(s-la)* &
              (sfac(s)/sfac(2*s+1))
         f2 = 0.0_dp
         DO t = tmin, tmax
            z1 = lp+mp+t
            z2 = la+lb-mp-t
            z3 = lp-mp-t
            z4 = lb-la+mp+t
            z5 = la-ma-t
            f2 = f2+(-1)**t*(sfac(z1)/(sfac(t)*sfac(z3)))*(sfac(z2)/(sfac(z4)*sfac(z5)))
         END DO
         cgc = pref*f1*f2
      ELSE
         cgc = 0.0_dp
      END IF

   END FUNCTION cgc

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \retval fval ...
! **************************************************************************************************
   FUNCTION sfac(n) RESULT(fval)
      INTEGER                                            :: n
      REAL(KIND=dp)                                      :: fval

      INTEGER                                            :: i

      IF (n > maxfac) THEN
         fval = fac(maxfac)
         DO i = maxfac+1, n
            fval = REAL(i, dp)*fval
         END DO
      ELSE IF (n >= 0) THEN
         fval = fac(n)
      ELSE
         fval = 1.0_dp
      END IF
   END FUNCTION sfac

! **************************************************************************************************
!> \brief ...
!> \param l1 ...
!> \param m1 ...
!> \param l2 ...
!> \param m2 ...
!> \retval ind ...
! **************************************************************************************************
   FUNCTION order(l1, m1, l2, m2) RESULT(ind)
      INTEGER, INTENT(IN)                                :: l1, m1, l2, m2
      INTEGER                                            :: ind

      INTEGER                                            :: i1, i2, ix, iy

      i1 = (l1*(l1+1))/2+ABS(m1)+1
      i2 = (l2*(l2+1))/2+ABS(m2)+1
      ix = MAX(i1, i2)
      iy = MIN(i1, i2)
      ind = (ix*(ix-1))/2+iy
   END FUNCTION order

! Calculation of Spherical Harmonics

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param y ...
!> \param l ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE rvy_lm(r, y, l, m)
!
! Real Spherical Harmonics
!                   _                   _
!                  |  [(2l+1)(l-|m|)!]   |1/2 m         cos(m p)   m>=0
!  Y_lm ( t, p ) = |---------------------|   P_l(cos(t))
!                  |[2Pi(1+d_m0)(l+|m|)!]|              sin(|m| p) m<0
!
! Input: r == (x,y,z) : normalised    x^2 + y^2 + z^2 = 1
!
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: y
      INTEGER, INTENT(IN)                                :: l, m

      CHARACTER(len=*), PARAMETER :: routineN = 'rvy_lm', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: cp, lmm, lpm, pf, plm, rxy, sp, t, z

      SELECT CASE (l)
      CASE (:-1)
         CALL cp__b("common/spherical_harmonics.F",434,"Negative l value")
      CASE (0)
         pf = SQRT(1.0_dp/(4.0_dp*pi))
         IF (m /= 0) CALL cp__b("common/spherical_harmonics.F",437,"l = 0 and m value out of bounds")
         y(:) = pf
      CASE (1)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",442,"l = 1 and m value out of bounds")
         CASE (1)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y(:) = pf*r(1, :)
         CASE (0)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y(:) = pf*r(3, :)
         CASE (-1)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y(:) = pf*r(2, :)
         END SELECT
      CASE (2)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",456,"l = 2 and m value out of bounds")
         CASE (2)
            pf = SQRT(15.0_dp/(16.0_dp*pi))
            y(:) = pf*(r(1, :)*r(1, :)-r(2, :)*r(2, :))
         CASE (1)
            pf = SQRT(15.0_dp/(4.0_dp*pi))
            y(:) = pf*r(3, :)*r(1, :)
         CASE (0)
            pf = SQRT(5.0_dp/(16.0_dp*pi))
            y(:) = pf*(3.0_dp*r(3, :)*r(3, :)-1.0_dp)
         CASE (-1)
            pf = SQRT(15.0_dp/(4.0_dp*pi))
            y(:) = pf*r(3, :)*r(2, :)
         CASE (-2)
            pf = SQRT(15.0_dp/(16.0_dp*pi))
            y(:) = pf*2.0_dp*r(1, :)*r(2, :)
         END SELECT
      CASE (3)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",476,"l = 3 and m value out of bounds")
         CASE (3)
            pf = SQRT(35.0_dp/(32.0_dp*pi))
            y(:) = pf*r(1, :)*(r(1, :)**2-3.0_dp*r(2, :)**2)
         CASE (2)
            pf = SQRT(105.0_dp/(16.0_dp*pi))
            y(:) = pf*r(3, :)*(r(1, :)**2-r(2, :)**2)
         CASE (1)
            pf = SQRT(21.0_dp/(32.0_dp*pi))
            y(:) = pf*r(1, :)*(5.0_dp*r(3, :)*r(3, :)-1.0_dp)
         CASE (0)
            pf = SQRT(7.0_dp/(16.0_dp*pi))
            y(:) = pf*r(3, :)*(5.0_dp*r(3, :)*r(3, :)-3.0_dp)
         CASE (-1)
            pf = SQRT(21.0_dp/(32.0_dp*pi))
            y(:) = pf*r(2, :)*(5.0_dp*r(3, :)*r(3, :)-1.0_dp)
         CASE (-2)
            pf = SQRT(105.0_dp/(16.0_dp*pi))
            y(:) = pf*2.0_dp*r(1, :)*r(2, :)*r(3, :)
         CASE (-3)
            pf = SQRT(35.0_dp/(32.0_dp*pi))
            y(:) = pf*r(2, :)*(3.0_dp*r(1, :)**2-r(2, :)**2)
         END SELECT
      CASE DEFAULT
         IF (m < -l .OR. m > l) CALL cp__b("common/spherical_harmonics.F",500,"m value out of bounds")
         lpm = fac(l+ABS(m))
         lmm = fac(l-ABS(m))
         IF (m == 0) THEN
            t = 4.0_dp*pi
         ELSE
            t = 2.0_dp*pi
         END IF
         IF (ABS(lpm) < EPSILON(1.0_dp)) THEN
            pf = REAL(2*l+1, KIND=dp)/t
         ELSE
            pf = (REAL(2*l+1, KIND=dp)*lmm)/(t*lpm)
         ENDIF
         pf = SQRT(pf)
         DO i = 1, SIZE(r, 2)
            z = r(3, i)
!      plm = legendre ( z, l, m )
            plm = legendre(z, l, ABS(m))
            IF (m == 0) THEN
               y(i) = pf*plm
            ELSE
               rxy = SQRT(r(1, i)**2+r(2, i)**2)
               IF (rxy < EPSILON(1.0_dp)) THEN
                  y(i) = 0.0_dp
               ELSE
                  cp = r(1, i)/rxy
                  sp = r(2, i)/rxy
                  IF (m > 0) THEN
                     y(i) = pf*plm*cosn(m, cp, sp)
                  ELSE
                     y(i) = pf*plm*sinn(-m, cp, sp)
                  END IF
               END IF
            END IF
         END DO
      END SELECT

   END SUBROUTINE rvy_lm

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param y ...
!> \param l ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE rry_lm(r, y, l, m)
!
! Real Spherical Harmonics
!                   _                   _
!                  |  [(2l+1)(l-|m|)!]   |1/2 m         cos(m p)   m>=0
!  Y_lm ( t, p ) = |---------------------|   P_l(cos(t))
!                  |[2Pi(1+d_m0)(l+|m|)!]|              sin(|m| p) m<0
!
! Input: r == (x,y,z) : normalised    x^2 + y^2 + z^2 = 1
!
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r
      REAL(KIND=dp), INTENT(OUT)                         :: y
      INTEGER, INTENT(IN)                                :: l, m

      CHARACTER(len=*), PARAMETER :: routineN = 'rry_lm', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: cp, lmm, lpm, pf, plm, rxy, sp, t, z

      SELECT CASE (l)
      CASE (:-1)
         CALL cp__b("common/spherical_harmonics.F",566,"Negative l value")
      CASE (0)
         pf = SQRT(1.0_dp/(4.0_dp*pi))
         IF (m /= 0) CALL cp__b("common/spherical_harmonics.F",569,"l = 0 and m value out of bounds")
         y = pf
      CASE (1)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",574,"l = 1 and m value out of bounds")
         CASE (1)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y = pf*r(1)
         CASE (0)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y = pf*r(3)
         CASE (-1)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y = pf*r(2)
         END SELECT
      CASE (2)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",588,"l = 2 and m value out of bounds")
         CASE (2)
            pf = SQRT(15.0_dp/(16.0_dp*pi))
            y = pf*(r(1)*r(1)-r(2)*r(2))
         CASE (1)
            pf = SQRT(15.0_dp/(4.0_dp*pi))
            y = pf*r(3)*r(1)
         CASE (0)
            pf = SQRT(5.0_dp/(16.0_dp*pi))
            y = pf*(3.0_dp*r(3)*r(3)-1.0_dp)
         CASE (-1)
            pf = SQRT(15.0_dp/(4.0_dp*pi))
            y = pf*r(3)*r(2)
         CASE (-2)
            pf = SQRT(15.0_dp/(16.0_dp*pi))
            y = pf*2.0_dp*r(1)*r(2)
         END SELECT
      CASE (3)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",608,"l = 3 and m value out of bounds")
         CASE (3)
            pf = SQRT(35.0_dp/(32.0_dp*pi))
            y = pf*r(1)*(r(1)**2-3.0_dp*r(2)**2)
         CASE (2)
            pf = SQRT(105.0_dp/(16.0_dp*pi))
            y = pf*r(3)*(r(1)**2-r(2)**2)
         CASE (1)
            pf = SQRT(21.0_dp/(32.0_dp*pi))
            y = pf*r(1)*(5.0_dp*r(3)*r(3)-1.0_dp)
         CASE (0)
            pf = SQRT(7.0_dp/(16.0_dp*pi))
            y = pf*r(3)*(5.0_dp*r(3)*r(3)-3.0_dp)
         CASE (-1)
            pf = SQRT(21.0_dp/(32.0_dp*pi))
            y = pf*r(2)*(5.0_dp*r(3)*r(3)-1.0_dp)
         CASE (-2)
            pf = SQRT(105.0_dp/(16.0_dp*pi))
            y = pf*2.0_dp*r(1)*r(2)*r(3)
         CASE (-3)
            pf = SQRT(35.0_dp/(32.0_dp*pi))
            y = pf*r(2)*(3.0_dp*r(1)**2-r(2)**2)
         END SELECT
      CASE DEFAULT
         IF (m < -l .OR. m > l) CALL cp__b("common/spherical_harmonics.F",632,"m value out of bounds")
         lpm = fac(l+ABS(m))
         lmm = fac(l-ABS(m))
         IF (m == 0) THEN
            t = 4.0_dp*pi
         ELSE
            t = 2.0_dp*pi
         END IF
         IF (ABS(lpm) < EPSILON(1.0_dp)) THEN
            pf = REAL(2*l+1, KIND=dp)/t
         ELSE
            pf = (REAL(2*l+1, KIND=dp)*lmm)/(t*lpm)
         ENDIF
         pf = SQRT(pf)
         z = r(3)
         plm = legendre(z, l, m)
         IF (m == 0) THEN
            y = pf*plm
         ELSE
            rxy = SQRT(r(1)**2+r(2)**2)
            IF (rxy < EPSILON(1.0_dp)) THEN
               y = 0.0_dp
            ELSE
               cp = r(1)/rxy
               sp = r(2)/rxy
               IF (m > 0) THEN
                  y = pf*plm*cosn(m, cp, sp)
               ELSE
                  y = pf*plm*sinn(-m, cp, sp)
               END IF
            END IF
         END IF
      END SELECT

   END SUBROUTINE rry_lm

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param y ...
!> \param l ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE cvy_lm(r, y, l, m)
!
! Complex Spherical Harmonics
!                   _                _
!                  | [(2l+1)(l-|m|)!] |1/2 m
!  Y_lm ( t, p ) = |------------------|   P_l(cos(t)) Exp[ i m p ]
!                  |  [4Pi(l+|m|)!]|  |
!
! Input: r == (x,y,z) : normalised    x^2 + y^2 + z^2 = 1
!
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r
      COMPLEX(KIND=dp), DIMENSION(:), INTENT(OUT)        :: y
      INTEGER, INTENT(IN)                                :: l, m

      CHARACTER(len=*), PARAMETER :: routineN = 'cvy_lm', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      REAL(KIND=dp)                                      :: cp, lmm, lpm, pf, plm, rxy, sp, t, ym, &
                                                            yp, z

      n = SIZE(r, 2)
      SELECT CASE (l)
      CASE (:-1)
         CALL cp__b("common/spherical_harmonics.F",698,"Negative l value")
      CASE (0)
         pf = SQRT(1.0_dp/(4.0_dp*pi))
         IF (m /= 0) CALL cp__b("common/spherical_harmonics.F",701,"l = 0 and m value out of bounds")
         y(:) = pf
      CASE (1)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",706,"l = 1 and m value out of bounds")
         CASE (1)
            pf = SQRT(3.0_dp/(8.0_dp*pi))
            DO i = 1, n
               yp = r(1, i)
               ym = r(2, i)
               y(i) = pf*CMPLX(yp, ym, KIND=dp)
            END DO
         CASE (0)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y(:) = pf*r(3, :)
         CASE (-1)
            pf = SQRT(3.0_dp/(8.0_dp*pi))
            DO i = 1, n
               yp = r(1, i)
               ym = r(2, i)
               y(i) = pf*CMPLX(yp, -ym, KIND=dp)
            END DO
         END SELECT
      CASE (2)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",728,"l = 2 and m value out of bounds")
         CASE (2)
            pf = SQRT(15.0_dp/(32.0_dp*pi))
            DO i = 1, n
               yp = (r(1, i)*r(1, i)-r(2, i)*r(2, i))
               ym = 2.0_dp*r(1, i)*r(2, i)
               y(i) = pf*CMPLX(yp, ym, KIND=dp)
            END DO
         CASE (1)
            pf = SQRT(15.0_dp/(8.0_dp*pi))
            DO i = 1, n
               yp = r(3, i)*r(1, i)
               ym = r(3, i)*r(2, i)
               y(i) = pf*CMPLX(yp, ym, KIND=dp)
            END DO
         CASE (0)
            pf = SQRT(5.0_dp/(16.0_dp*pi))
            y(:) = pf*(3.0_dp*r(3, :)*r(3, :)-1.0_dp)
         CASE (-1)
            pf = SQRT(15.0_dp/(8.0_dp*pi))
            DO i = 1, n
               yp = r(3, i)*r(1, i)
               ym = r(3, i)*r(2, i)
               y(i) = pf*CMPLX(yp, -ym, KIND=dp)
            END DO
         CASE (-2)
            pf = SQRT(15.0_dp/(32.0_dp*pi))
            DO i = 1, n
               yp = (r(1, i)*r(1, i)-r(2, i)*r(2, i))
               ym = 2.0_dp*r(1, i)*r(2, i)
               y(i) = pf*CMPLX(yp, -ym, KIND=dp)
            END DO
         END SELECT
      CASE (3)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",764,"l = 3 and m value out of bounds")
         CASE (3)
            pf = SQRT(35.0_dp/(64.0_dp*pi))
            DO i = 1, n
               yp = r(1, i)*(r(1, i)**2-3.0_dp*r(2, i)**2)
               ym = r(2, i)*(3.0_dp*r(1, i)**2-r(2, i)**2)
               y(i) = pf*CMPLX(yp, ym, KIND=dp)
            END DO
         CASE (2)
            pf = SQRT(105.0_dp/(32.0_dp*pi))
            DO i = 1, n
               yp = r(3, i)*(r(1, i)**2-r(2, i)**2)
               ym = 2.0_dp*r(1, i)*r(2, i)*r(3, i)
               y(i) = pf*CMPLX(yp, ym, KIND=dp)
            END DO
         CASE (1)
            pf = SQRT(21.0_dp/(64.0_dp*pi))
            DO i = 1, n
               yp = r(1, i)*(5.0_dp*r(3, i)*r(3, i)-1.0_dp)
               ym = r(2, i)*(5.0_dp*r(3, i)*r(3, i)-1.0_dp)
               y(i) = pf*CMPLX(yp, ym, KIND=dp)
            END DO
         CASE (0)
            pf = SQRT(7.0_dp/(16.0_dp*pi))
            y(:) = pf*r(3, :)*(5.0_dp*r(3, :)*r(3, :)-3.0_dp)
         CASE (-1)
            pf = SQRT(21.0_dp/(64.0_dp*pi))
            DO i = 1, n
               yp = r(1, i)*(5.0_dp*r(3, i)*r(3, i)-1.0_dp)
               ym = r(2, i)*(5.0_dp*r(3, i)*r(3, i)-1.0_dp)
               y(i) = pf*CMPLX(yp, -ym, KIND=dp)
            END DO
         CASE (-2)
            pf = SQRT(105.0_dp/(32.0_dp*pi))
            DO i = 1, n
               yp = r(3, i)*(r(1, i)**2-r(2, i)**2)
               ym = 2.0_dp*r(1, i)*r(2, i)*r(3, i)
               y(i) = pf*CMPLX(yp, -ym, KIND=dp)
            END DO
         CASE (-3)
            pf = SQRT(35.0_dp/(64.0_dp*pi))
            DO i = 1, n
               yp = r(1, i)*(r(1, i)**2-3.0_dp*r(2, i)**2)
               ym = r(2, i)*(3.0_dp*r(1, i)**2-r(2, i)**2)
               y(i) = pf*CMPLX(yp, -ym, KIND=dp)
            END DO
         END SELECT
      CASE DEFAULT
         IF (m < -l .OR. m > l) CALL cp__b("common/spherical_harmonics.F",812,"m value out of bounds")
         lpm = fac(l+ABS(m))
         lmm = fac(l-ABS(m))
         t = 4.0_dp*pi
         IF (ABS(lpm) < EPSILON(1.0_dp)) THEN
            pf = REAL(2*l+1, KIND=dp)/t
         ELSE
            pf = (REAL(2*l+1, KIND=dp)*lmm)/(t*lpm)
         ENDIF
         pf = SQRT(pf)
         DO i = 1, n
            z = r(3, i)
            plm = legendre(z, l, m)
            IF (m == 0) THEN
               y(i) = pf*plm
            ELSE
               rxy = SQRT(r(1, i)**2+r(2, i)**2)
               IF (rxy < EPSILON(1.0_dp)) THEN
                  y(i) = 0.0_dp
               ELSE
                  cp = r(1, i)/rxy
                  sp = r(2, i)/rxy
                  IF (m > 0) THEN
                     yp = cosn(m, cp, sp)
                     ym = sinn(m, cp, sp)
                     y(i) = pf*plm*CMPLX(yp, ym, KIND=dp)
                  ELSE
                     yp = cosn(-m, cp, sp)
                     ym = sinn(-m, cp, sp)
                     y(i) = pf*plm*CMPLX(yp, -ym, KIND=dp)
                  END IF
               END IF
            END IF
         END DO
      END SELECT

   END SUBROUTINE cvy_lm

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param y ...
!> \param l ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE ccy_lm(r, y, l, m)
!
! Complex Spherical Harmonics
!                   _                _
!                  | [(2l+1)(l-|m|)!] |1/2 m
!  Y_lm ( t, p ) = |------------------|   P_l(cos(t)) Exp[ i m p ]
!                  |  [4Pi(l+|m|)!]|  |
!
! Input: r == (x,y,z) : normalised    x^2 + y^2 + z^2 = 1
!
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r
      COMPLEX(KIND=dp), INTENT(OUT)                      :: y
      INTEGER, INTENT(IN)                                :: l, m

      CHARACTER(len=*), PARAMETER :: routineN = 'ccy_lm', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: cp, lmm, lpm, pf, plm, rxy, sp, t, ym, &
                                                            yp, z

      SELECT CASE (l)
      CASE (:-1)
         CALL cp__b("common/spherical_harmonics.F",878,"Negative l value")
      CASE (0)
         pf = SQRT(1.0_dp/(4.0_dp*pi))
         IF (m /= 0) CALL cp__b("common/spherical_harmonics.F",881,"l = 0 and m value out of bounds")
         y = pf
      CASE (1)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",886,"l = 1 and m value out of bounds")
         CASE (1)
            pf = SQRT(3.0_dp/(8.0_dp*pi))
            yp = r(1)
            ym = r(2)
            y = pf*CMPLX(yp, ym, KIND=dp)
         CASE (0)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            y = pf*r(3)
         CASE (-1)
            pf = SQRT(3.0_dp/(8.0_dp*pi))
            yp = r(1)
            ym = r(2)
            y = pf*CMPLX(yp, -ym, KIND=dp)
         END SELECT
      CASE (2)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",904,"l = 2 and m value out of bounds")
         CASE (2)
            pf = SQRT(15.0_dp/(32.0_dp*pi))
            yp = (r(1)*r(1)-r(2)*r(2))
            ym = 2.0_dp*r(1)*r(2)
            y = pf*CMPLX(yp, ym, KIND=dp)
         CASE (1)
            pf = SQRT(15.0_dp/(8.0_dp*pi))
            yp = r(3)*r(1)
            ym = r(3)*r(2)
            y = pf*CMPLX(yp, ym, KIND=dp)
         CASE (0)
            pf = SQRT(5.0_dp/(16.0_dp*pi))
            y = pf*(3.0_dp*r(3)*r(3)-1.0_dp)
         CASE (-1)
            pf = SQRT(15.0_dp/(8.0_dp*pi))
            yp = r(3)*r(1)
            ym = r(3)*r(2)
            y = pf*CMPLX(yp, -ym, KIND=dp)
         CASE (-2)
            pf = SQRT(15.0_dp/(32.0_dp*pi))
            yp = (r(1)*r(1)-r(2)*r(2))
            ym = 2.0_dp*r(1)*r(2)
            y = pf*CMPLX(yp, -ym, KIND=dp)
         END SELECT
      CASE (3)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",932,"l = 3 and m value out of bounds")
         CASE (3)
            pf = SQRT(35.0_dp/(64.0_dp*pi))
            yp = r(1)*(r(1)**2-3.0_dp*r(2)**2)
            ym = r(2)*(3.0_dp*r(1)**2-r(2)**2)
            y = pf*CMPLX(yp, ym, KIND=dp)
         CASE (2)
            pf = SQRT(105.0_dp/(32.0_dp*pi))
            yp = r(3)*(r(1)**2-r(2)**2)
            ym = 2.0_dp*r(1)*r(2)*r(3)
            y = pf*CMPLX(yp, ym, KIND=dp)
         CASE (1)
            pf = SQRT(21.0_dp/(64.0_dp*pi))
            yp = r(1)*(5.0_dp*r(3)*r(3)-1.0_dp)
            ym = r(2)*(5.0_dp*r(3)*r(3)-1.0_dp)
            y = pf*CMPLX(yp, ym, KIND=dp)
         CASE (0)
            pf = SQRT(7.0_dp/(16.0_dp*pi))
            y = pf*r(3)*(5.0_dp*r(3)*r(3)-3.0_dp)
         CASE (-1)
            pf = SQRT(21.0_dp/(64.0_dp*pi))
            yp = r(1)*(5.0_dp*r(3)*r(3)-1.0_dp)
            ym = r(2)*(5.0_dp*r(3)*r(3)-1.0_dp)
            y = pf*CMPLX(yp, -ym, KIND=dp)
         CASE (-2)
            pf = SQRT(105.0_dp/(32.0_dp*pi))
            yp = r(3)*(r(1)**2-r(2)**2)
            ym = 2.0_dp*r(1)*r(2)*r(3)
            y = pf*CMPLX(yp, -ym, KIND=dp)
         CASE (-3)
            pf = SQRT(35.0_dp/(64.0_dp*pi))
            yp = r(1)*(r(1)**2-3.0_dp*r(2)**2)
            ym = r(2)*(3.0_dp*r(1)**2-r(2)**2)
            y = pf*CMPLX(yp, -ym, KIND=dp)
         END SELECT
      CASE DEFAULT
         IF (m < -l .OR. m > l) CALL cp__b("common/spherical_harmonics.F",968,"m value out of bounds")
         lpm = fac(l+ABS(m))
         lmm = fac(l-ABS(m))
         t = 4.0_dp*pi
         IF (ABS(lpm) < EPSILON(1.0_dp)) THEN
            pf = REAL(2*l+1, KIND=dp)/t
         ELSE
            pf = (REAL(2*l+1, KIND=dp)*lmm)/(t*lpm)
         ENDIF
         pf = SQRT(pf)
         z = r(3)
         plm = legendre(z, l, m)
         IF (m == 0) THEN
            y = pf*plm
         ELSE
            rxy = SQRT(r(1)**2+r(2)**2)
            IF (rxy < EPSILON(1.0_dp)) THEN
               y = 0.0_dp
            ELSE
               cp = r(1)/rxy
               sp = r(2)/rxy
               IF (m > 0) THEN
                  yp = cosn(m, cp, sp)
                  ym = sinn(m, cp, sp)
                  y = pf*plm*CMPLX(yp, ym, KIND=dp)
               ELSE
                  yp = cosn(-m, cp, sp)
                  ym = sinn(-m, cp, sp)
                  y = pf*plm*CMPLX(yp, -ym, KIND=dp)
               END IF
            END IF
         END IF
      END SELECT

   END SUBROUTINE ccy_lm

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param l ...
!> \param m ...
!> \retval plm ...
! **************************************************************************************************
   FUNCTION legendre(x, l, m) RESULT(plm)

      REAL(KIND=dp), INTENT(IN)                          :: x
      INTEGER, INTENT(IN)                                :: l, m
      REAL(KIND=dp)                                      :: plm

      CHARACTER(len=*), PARAMETER :: routineN = 'legendre', routineP = moduleN//':'//routineN

      INTEGER                                            :: il, im, mm
      REAL(KIND=dp)                                      :: fact, pll, pmm, pmmp1, somx2

      IF (ABS(x) > 1.0_dp) CALL cp__b("common/spherical_harmonics.F",1022,"x value > 1")
      SELECT CASE (l)
      CASE (:-1)
         CALL cp__b("common/spherical_harmonics.F",1025,"Negative l value")
      CASE (0)
         plm = 1.0_dp
      CASE (1)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1031,"l = 1 and m value out of bounds")
         CASE (1)
            plm = SQRT(1.0_dp-x*x)
         CASE (0)
            plm = x
         END SELECT
      CASE (2)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1040,"l = 2 and m value out of bounds")
         CASE (2)
            plm = 3.0_dp*(1.0_dp-x*x)
         CASE (1)
            plm = 3.0_dp*x*SQRT(1.0_dp-x*x)
         CASE (0)
            plm = 1.5_dp*x*x-0.5_dp
         END SELECT
      CASE (3)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1051,"l = 3 and m value out of bounds")
         CASE (3)
            plm = 15.0_dp*(1.0_dp-x*x)**1.5_dp
         CASE (2)
            plm = 15.0_dp*x*(1.0_dp-x*x)
         CASE (1)
            plm = (7.5_dp*x*x-1.5_dp)*SQRT(1.0_dp-x*x)
         CASE (0)
            plm = 2.5_dp*x**3-1.5_dp*x
         END SELECT
      CASE (4)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1064,"l = 4 and m value out of bounds")
         CASE (4)
            plm = 105.0_dp*(1.0_dp-x*x)**2
         CASE (3)
            plm = 105.0_dp*x*(1.0_dp-x*x)**1.5_dp
         CASE (2)
            plm = (52.5_dp*x*x-7.5_dp)*(1.0_dp-x*x)
         CASE (1)
            plm = (17.5_dp*x**3-7.5_dp*x)*SQRT(1.0_dp-x*x)
         CASE (0)
            plm = 4.375_dp*x**4-3.75_dp*x**2+0.375_dp
         END SELECT
      CASE (5)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1079,"l = 5 and m value out of bounds")
         CASE (5)
            plm = 945.0_dp*(1.0_dp-x*x)**2.5_dp
         CASE (4)
            plm = 945.0_dp*x*(1.0_dp-x*x)**2
         CASE (3)
            plm = -(-472.5_dp*x*x+52.5_dp)*(1.0_dp-x*x)**1.5_dp
         CASE (2)
            plm = (157.5_dp*x**3-52.5_dp*x)*(1.0_dp-x*x)
         CASE (1)
            plm = -(-39.375_dp*x**4+26.25_dp*x*x- &
                    1.875_dp)*SQRT(1.0_dp-x*x)
         CASE (0)
            plm = 7.875_dp*x**5-8.75_dp*x**3+1.875_dp*x
         END SELECT
      CASE (6)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1097,"l = 6 and m value out of bounds")
         CASE (6)
            plm = 10395.0_dp*(1.0_dp-x*x)**3
         CASE (5)
            plm = 10395.0_dp*x*(1.0_dp-x*x)**2.5_dp
         CASE (4)
            plm = (5197.5_dp*x*x-472.5_dp)*(1.0_dp-x*x)**2
         CASE (3)
            plm = -(-1732.5_dp*x**3+472.5_dp*x)* &
                  (1.0_dp-x*x)**1.5_dp
         CASE (2)
            plm = (433.125_dp*x**4-236.25_dp*x**2+ &
                   13.125_dp)*(1.0_dp-x*x)
         CASE (1)
            plm = -(-86.625_dp*x**5+78.75_dp*x**3- &
                    13.125_dp*x)*SQRT(1.0_dp-x*x)
         CASE (0)
            plm = 14.4375_dp*x**6-19.6875_dp*x**4+ &
                  6.5625_dp*x**2-0.3125_dp
         END SELECT
      CASE DEFAULT
         mm = ABS(m)
         IF (mm > l) CALL cp__b("common/spherical_harmonics.F",1119,"m out of bounds")
! use recurence from numerical recipies
         pmm = 1.0_dp
         IF (mm > 0) THEN
            somx2 = SQRT((1.0_dp-x)*(1.0_dp+x))
            fact = 1.0_dp
            DO im = 1, mm
               pmm = pmm*fact*somx2
               fact = fact+2.0_dp
            END DO
         END IF
         IF (l == mm) THEN
            plm = pmm
         ELSE
            pmmp1 = x*REAL(2*mm+1, KIND=dp)*pmm
            IF (l == mm+1) THEN
               plm = pmmp1
            ELSE
               DO il = mm+2, l
                  pll = (x*REAL(2*il-1, KIND=dp)*pmmp1- &
                         REAL(il+mm-1, KIND=dp)*pmm)/REAL(il-mm, KIND=dp)
                  pmm = pmmp1
                  pmmp1 = pll
               END DO
               plm = pll
            END IF
         END IF
      END SELECT

   END FUNCTION legendre

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param l ...
!> \param m ...
!> \retval dplm ...
! **************************************************************************************************
   FUNCTION dlegendre(x, l, m) RESULT(dplm)
      REAL(KIND=dp), INTENT(IN)                          :: x
      INTEGER, INTENT(IN)                                :: l, m
      REAL(KIND=dp)                                      :: dplm

      CHARACTER(len=*), PARAMETER :: routineN = 'dlegendre', routineP = moduleN//':'//routineN

      INTEGER                                            :: il, im, mm
      REAL(KIND=dp)                                      :: dpll, dpmm, dpmmp1, dsomx2, fact, pll, &
                                                            plm, pmm, pmmp1, somx2

      IF (ABS(x) > 1.0_dp) CALL cp__b("common/spherical_harmonics.F",1168,"x value > 1")
      SELECT CASE (l)
      CASE (0)
         dplm = 0.0_dp
      CASE (1)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1175,"l = 1 and m value out of bounds")
         CASE (1)
            dplm = -x/SQRT(1.0_dp-x*x)
         CASE (0)
            dplm = 1.0_dp
         END SELECT
      CASE (2)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1184,"l = 2 and m value out of bounds")
         CASE (2)
            dplm = -6.0_dp*x
         CASE (1)
            dplm = 3.0_dp*SQRT(1.0_dp-x*x)-3.0_dp*x*x/SQRT(1.0_dp-x*x)
         CASE (0)
            dplm = 3.0_dp*x
         END SELECT
      CASE (3)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1195,"l = 3 and m value out of bounds")
         CASE (3)
            dplm = -45.0_dp*SQRT(1.0_dp-x*x)*x
         CASE (2)
            dplm = 15.0_dp*(1.0_dp-x*x)-30.0_dp*x*x
         CASE (1)
            dplm = 15.0_dp*x*SQRT(1.0_dp-x*x)-(x*(7.5_dp*x*x-1.5_dp))/SQRT(1.0_dp-x*x)
         CASE (0)
            dplm = 7.5_dp*x*x-1.5_dp
         END SELECT
      CASE (4)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1208,"l = 4 and m value out of bounds")
         CASE (4)
            dplm = -420*x*(1-x*x)
         CASE (3)
            dplm = 105.0_dp*((1.0_dp-x*x)**1.5_dp-3.0_dp*x*x*(1.0_dp-x*x)**0.5_dp)
         CASE (2)
            dplm = 105.0_dp*x*(1.0_dp-x*x)-2.0_dp*x*(52.5_dp*x*x-7.5_dp)
         CASE (1)
            IF (ABS(x)-1.0_dp < EPSILON(1.0_dp)) THEN
               dplm = 0.0_dp
            ELSE
               dplm = (17.5_dp*3.0_dp*x**2-7.5_dp)*SQRT(1.0_dp-x*x)- &
                      x*(17.5_dp*x**3-7.5_dp*x)/SQRT(1.0_dp-x*x)
            END IF
         CASE (0)
            dplm = 4.375_dp*4.0_dp*x**3-3.75_dp*2.0_dp*x
         END SELECT
      CASE (5)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1228,"l = 5 and m value out of bounds")
         CASE (5)
            dplm = -945.0_dp*5.0_dp*x*(1.0_dp-x*x)**1.5_dp
         CASE (4)
            dplm = 945.0_dp*((1.0_dp-x*x)**2-2.0_dp*x*x*(1.0_dp-x*x))
         CASE (3)
            dplm = 945.0_dp*x*(1.0_dp-x*x)**1.5_dp- &
                   3.0_dp*x*(472.5_dp*x*x-52.5_dp)*(1.0_dp-x*x)**0.5_dp
         CASE (2)
            dplm = (3.0_dp*157.5_dp*x**2-52.5_dp)*(1.0_dp-x*x)- &
                   (157.5_dp*x**3-52.5_dp*x)*(-2.0_dp*x)
         CASE (1)
            IF (ABS(x)-1.0_dp < EPSILON(1.0_dp)) THEN
               dplm = 0.0_dp
            ELSE
               dplm = -(-39.375_dp*4.0_dp*x*x*x+2.0_dp*26.25_dp*x)*SQRT(1.0_dp-x*x)+ &
                      x*(-39.375_dp*x**4+26.25_dp*x*x-1.875_dp)/SQRT(1.0_dp-x*x)
            END IF
         CASE (0)
            dplm = 5.0_dp*7.875_dp*x**4-3.0_dp*8.75_dp*x**2+1.875_dp
         END SELECT
      CASE (6)
         SELECT CASE (ABS (m))
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1252,"l = 6 and m value out of bounds")
         CASE (6)
            dplm = -10395.0_dp*6.0_dp*x*(1.0_dp-x*x)**2
         CASE (5)
            dplm = 10395.0_dp*((1.0_dp-x*x)**2.5_dp-5.0_dp*x*x*(1.0_dp-x*x)**1.5_dp)
         CASE (4)
            dplm = 2.0_dp*5197.5_dp*x*(1.0_dp-x*x)**2- &
                   4.0_dp*x*(5197.5_dp*x*x-472.5_dp)*(1.0_dp-x*x)
         CASE (3)
            dplm = -(-3.0_dp*1732.5_dp*x*x+472.5_dp)*(1.0_dp-x*x)**1.5_dp+ &
                   (-1732.5_dp*x**3+472.5_dp*x)*3.0_dp*x*(1.0_dp-x*x)**0.5_dp
         CASE (2)
            dplm = (433.125_dp*4.0_dp*x**3-2.0_dp*236.25_dp*x)*(1.0_dp-x*x)- &
                   2.0_dp*x*(433.125_dp*x**4-236.25_dp*x**2+13.125_dp)
         CASE (1)
            IF (ABS(x)-1.0_dp < EPSILON(1.0_dp)) THEN
               dplm = 0.0_dp
            ELSE
               dplm = -(-5.0_dp*86.625_dp*x**4+3.0_dp*78.75_dp**2-13.125_dp)*SQRT(1.0_dp-x*x)+ &
                      x*(-86.625_dp*x**5+78.75_dp*x**3-13.125_dp*x)/SQRT(1.0_dp-x*x)
            END IF
         CASE (0)
            dplm = 14.4375_dp*6.0_dp*x**5-19.6875_dp*4.0_dp*x**3+ &
                   6.5625_dp*2.0_dp*x
         END SELECT
      CASE DEFAULT
!    CALL cp__b("common/spherical_harmonics.F",1278,"l > 6 dplm not implemented")
         mm = ABS(m)
         IF (mm > l) CALL cp__b("common/spherical_harmonics.F",1280,"m out of bounds")
! use recurence from numerical recipies
         pmm = 1.0_dp
         dpmm = 0.0_dp
         IF (mm > 0) THEN
            somx2 = SQRT((1.0_dp-x)*(1.0_dp+x))
            dsomx2 = ((1e0_dp-x)*(1e0_dp+x))**(0.5e0_dp*m-1)
            dpmm = -m*x*dsomx2*dfac(2*m-1)
            fact = 1.0_dp
            DO im = 1, mm
               pmm = -pmm*fact*somx2
               dpmm = -dpmm
               fact = fact+2.0_dp
            END DO
         END IF
         IF (l == mm) THEN
            plm = pmm
            dplm = dpmm
         ELSE
            pmmp1 = x*REAL(2*mm+1, KIND=dp)*pmm
            dpmmp1 = (2*mm+1)*pmm+x*(2*mm+1)*dpmm
            IF (l == mm+1) THEN
               plm = pmmp1
               dplm = dpmmp1
            ELSE
               DO il = mm+2, l
                  pll = (x*REAL(2*il-1, KIND=dp)*pmmp1- &
                         REAL(il+mm-1, KIND=dp)*pmm)/REAL(il-mm, KIND=dp)
                  dpll = ((2*il-1)*pmmp1+x*(2*il-1)*dpmmp1-(il+mm-1)*dpmm)/(il-mm)
                  pmm = pmmp1
                  pmmp1 = pll
                  dpmm = dpmmp1
                  dpmmp1 = dpll
               END DO
               plm = pll
               dplm = dpll
            END IF
         END IF

      END SELECT

   END FUNCTION dlegendre

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \param l ...
!> \retval dPof1 ...
! **************************************************************************************************
   FUNCTION dPof1(x, l)

      REAL(KIND=dp), INTENT(IN)                          :: x
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp)                                      :: dPof1

      CHARACTER(len=*), PARAMETER :: routineN = 'dPof1', routineP = moduleN//':'//routineN

      IF (ABS(x)-1.0_dp > EPSILON(1.0_dp)) THEN
         CALL cp__b("common/spherical_harmonics.F",1338,"|x| is not 1")
      END IF
      IF (x > 0.0_dp) THEN
         SELECT CASE (l)
         CASE (:-1)
            CALL cp__b("common/spherical_harmonics.F",1343,"Negative l value")
         CASE (0)
            dPof1 = 0.0_dp
         CASE (1)
            dPof1 = 1.0_dp
         CASE (2)
            dPof1 = 3.0_dp
         CASE (3)
            dPof1 = 6.0_dp
         CASE (4)
            dPof1 = 10.0_dp
         CASE (5)
            dPof1 = 15.0_dp
         CASE (6)
            dPof1 = 21.0_dp
         CASE (7:)
            CALL cp__b("common/spherical_harmonics.F",1359,"Not implemented")
         END SELECT
      ELSE
         SELECT CASE (l)
         CASE (:-1)
            CALL cp__b("common/spherical_harmonics.F",1364,"Negative l value")
         CASE (0)
            dPof1 = 0.0_dp
         CASE (1)
            dPof1 = 1.0_dp
         CASE (2)
            dPof1 = -3.0_dp
         CASE (3)
            dPof1 = 6.0_dp
         CASE (4)
            dPof1 = -10.0_dp
         CASE (5)
            dPof1 = 15.0_dp
         CASE (6)
            dPof1 = -21.0_dp
         CASE (7:)
            CALL cp__b("common/spherical_harmonics.F",1380,"Not implemented")
         END SELECT
      END IF

   END FUNCTION dPof1

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param k ...
!> \retval choose ...
! **************************************************************************************************
   FUNCTION choose(n, k)

      INTEGER, INTENT(IN)                                :: n, k
      REAL(KIND=dp)                                      :: choose

      IF (n >= k) THEN
         choose = REAL(NINT(fac(n)/(fac(k)*fac(n-k))), KIND=dp)
      ELSE
         choose = 0.0_dp
      ENDIF

   END FUNCTION choose

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param c ...
!> \param s ...
!> \retval cosn ...
! **************************************************************************************************
   FUNCTION cosn(n, c, s)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: c, s
      REAL(KIND=dp)                                      :: cosn

      INTEGER                                            :: i, j

      cosn = 0.0_dp
      IF (ABS(c) < EPSILON(1.0_dp) .OR. n == 0) THEN
         IF (MOD(n, 2) == 0) THEN
            cosn = (-1.0_dp)**INT(n/2)
         ELSE
            cosn = 0.0_dp
         END IF
      ELSE
         DO i = n, 0, -2
            IF (i == 0) THEN
               j = n
               IF (j == 0) THEN
                  cosn = cosn+choose(n, j)
               ELSE IF (MOD(j/2, 2) == 0) THEN
                  cosn = cosn+choose(n, j)*s**j
               ELSE
                  cosn = cosn-choose(n, j)*s**j
               END IF
            ELSE
               j = n-i
               IF (j == 0) THEN
                  cosn = cosn+choose(n, j)*c**i
               ELSE IF (MOD(j/2, 2) == 0) THEN
                  cosn = cosn+choose(n, j)*c**i*s**j
               ELSE
                  cosn = cosn-choose(n, j)*c**i*s**j
               END IF
            END IF
         END DO
      END IF

   END FUNCTION cosn

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param c ...
!> \param s ...
!> \retval dcosn_dcp ...
! **************************************************************************************************
   FUNCTION dcosn_dcp(n, c, s)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: c, s
      REAL(KIND=dp)                                      :: dcosn_dcp

      INTEGER                                            :: i, j

      dcosn_dcp = 0.0_dp

      IF (s < EPSILON(1.0_dp)) THEN
         dcosn_dcp = 0.0_dp
      ELSE
         DO i = n, 0, -2
            IF (i == 0) THEN
               dcosn_dcp = dcosn_dcp
            ELSE
               j = n-i
               IF (MOD(j/2, 2) == 0) THEN
                  dcosn_dcp = dcosn_dcp+choose(n, j)*REAL(i, dp)*c**(i-1)*s**j
               ELSE
                  dcosn_dcp = dcosn_dcp-choose(n, j)*REAL(i, dp)*c**(i-1)*s**j
               END IF
            END IF
         END DO
      END IF

   END FUNCTION dcosn_dcp

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param c ...
!> \param s ...
!> \retval dcosn_dsp ...
! **************************************************************************************************
   FUNCTION dcosn_dsp(n, c, s)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: c, s
      REAL(KIND=dp)                                      :: dcosn_dsp

      INTEGER                                            :: i, j

      dcosn_dsp = 0.0_dp
      IF (c < EPSILON(1.0_dp) .OR. s < EPSILON(1.0_dp)) THEN
         dcosn_dsp = 0.0_dp
      ELSE
         DO i = n, 0, -2
            j = n-i
            IF (j == 0) THEN
               dcosn_dsp = dcosn_dsp
            ELSE
               IF (MOD(j/2, 2) == 0) THEN
                  dcosn_dsp = dcosn_dsp+choose(n, j)*REAL(j, dp)*s**(j-1)*c**i
               ELSE
                  dcosn_dsp = dcosn_dsp-choose(n, j)*REAL(j, dp)*s**(j-1)*c**i
               END IF
            END IF
         END DO
      END IF

   END FUNCTION dcosn_dsp

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param c ...
!> \param s ...
!> \retval sinn ...
! **************************************************************************************************
   FUNCTION sinn(n, c, s)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: c, s
      REAL(KIND=dp)                                      :: sinn

      INTEGER                                            :: i, j

      sinn = 0.0_dp

      IF (ABS(s) < EPSILON(1.0_dp) .OR. n == 0) THEN
         sinn = 0.0_dp
      ELSE IF (ABS(c) < EPSILON(1.0_dp)) THEN
         IF (MOD(n, 2) == 0) THEN
            sinn = 0.0_dp
         ELSE
            sinn = s*(-1.0_dp)**INT((n-1)/2)
         END IF
      ELSE
         DO i = n-1, 0, -2
            IF (i == 0) THEN
               j = n
               IF (MOD(j/2, 2) == 0) THEN
                  sinn = sinn+choose(n, j)*s**j
               ELSE
                  sinn = sinn-choose(n, j)*s**j
               END IF
            ELSE
               j = n-i
               IF (MOD(j/2, 2) == 0) THEN
                  sinn = sinn+choose(n, j)*c**i*s**j
               ELSE
                  sinn = sinn-choose(n, j)*c**i*s**j
               END IF
            END IF
         END DO
      END IF

   END FUNCTION sinn

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param c ...
!> \param s ...
!> \retval dsinn_dcp ...
! **************************************************************************************************
   FUNCTION dsinn_dcp(n, c, s)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: c, s
      REAL(KIND=dp)                                      :: dsinn_dcp

      INTEGER                                            :: i, j

      dsinn_dcp = 0.0_dp

      IF (c < EPSILON(1.0_dp) .OR. s < EPSILON(1.0_dp)) THEN
         dsinn_dcp = 0.0_dp
      ELSE
         DO i = n-1, 0, -2
            IF (i == 0) THEN
               dsinn_dcp = dsinn_dcp
            ELSE
               j = n-i
               IF (MOD(j/2, 2) == 0) THEN
                  dsinn_dcp = dsinn_dcp+choose(n, j)*REAL(i, dp)*c**(i-1)*s**j
               ELSE
                  dsinn_dcp = dsinn_dcp-choose(n, j)*REAL(i, dp)*c**(i-1)*s**j
               END IF
            END IF
         END DO
      END IF

   END FUNCTION dsinn_dcp

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param c ...
!> \param s ...
!> \retval dsinn_dsp ...
! **************************************************************************************************
   FUNCTION dsinn_dsp(n, c, s)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: c, s
      REAL(KIND=dp)                                      :: dsinn_dsp

      INTEGER                                            :: i, j

      dsinn_dsp = 0.0_dp

      IF (c < EPSILON(1.0_dp) .OR. s < EPSILON(1.0_dp)) THEN
         dsinn_dsp = 0.0_dp
      ELSE
         DO i = n-1, 0, -2
            j = n-i
            IF (j == 0) THEN
               dsinn_dsp = dsinn_dsp
            ELSE
               IF (MOD(j/2, 2) == 0) THEN
                  dsinn_dsp = dsinn_dsp+choose(n, j)*c**i*REAL(j, dp)*s**(j-1)
               ELSE
                  dsinn_dsp = dsinn_dsp-choose(n, j)*c**i*REAL(j, dp)*s**(j-1)
               END IF
            END IF
         END DO
      END IF

   END FUNCTION dsinn_dsp

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param dy ...
!> \param l ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE drvy_lm(r, dy, l, m)
!
! Real Spherical Harmonics
!                   _                   _
!                  |  [(2l+1)(l-|m|)!]   |1/2 m         cos(m p)   m>=0
!  Y_lm ( t, p ) = |---------------------|   P_l(cos(t))
!                  |[2Pi(1+d_m0)(l+|m|)!]|              sin(|m| p) m<0
!
! Input: r == (x,y,z) : normalised    x^2 + y^2 + z^2 = 1
!
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: dy
      INTEGER, INTENT(IN)                                :: l, m

      CHARACTER(len=*), PARAMETER :: routineN = 'drvy_lm', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: cp, dcp_dx, dcp_dy, dplm, dsp_dx, &
                                                            dsp_dy, lmm, lpm, pf, plm, rxy, sp, t, &
                                                            z

      SELECT CASE (l)
      CASE (:-1)
         CALL cp__b("common/spherical_harmonics.F",1673,"Negative l value")
      CASE (0)
         pf = SQRT(1.0_dp/(4.0_dp*pi))
         IF (m /= 0) CALL cp__b("common/spherical_harmonics.F",1676,"l = 0 and m value out of bounds")
         dy(1:3, :) = 0.0_dp
      CASE (1)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1681,"l = 1 and m value out of bounds")
         CASE (1)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            dy(1, :) = pf
            dy(2, :) = 0.0_dp
            dy(3, :) = 0.0_dp
         CASE (0)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            dy(1, :) = 0.0_dp
            dy(2, :) = 0.0_dp
            dy(3, :) = pf
         CASE (-1)
            pf = SQRT(3.0_dp/(4.0_dp*pi))
            dy(1, :) = 0.0_dp
            dy(2, :) = pf
            dy(3, :) = 0.0_dp
         END SELECT
      CASE (2)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1701,"l = 2 and m value out of bounds")
         CASE (2)
            pf = SQRT(15.0_dp/(16.0_dp*pi))
            dy(1, :) = pf*2.0_dp*r(1, :)
            dy(2, :) = -pf*2.0_dp*r(2, :)
            dy(3, :) = 0.0_dp
         CASE (1)
            pf = SQRT(15.0_dp/(4.0_dp*pi))
            dy(1, :) = pf*r(3, :)
            dy(2, :) = 0.0_dp
            dy(3, :) = pf*r(1, :)
         CASE (0)
            pf = SQRT(5.0_dp/(16.0_dp*pi))
            dy(1, :) = 0.0_dp
            dy(2, :) = 0.0_dp
            dy(3, :) = pf*6.0_dp*r(3, :)
         CASE (-1)
            pf = SQRT(15.0_dp/(4.0_dp*pi))
            dy(1, :) = 0.0_dp
            dy(2, :) = pf*r(3, :)
            dy(3, :) = pf*r(2, :)
         CASE (-2)
            pf = SQRT(15.0_dp/(16.0_dp*pi))
            dy(1, :) = pf*2.0_dp*r(2, :)
            dy(2, :) = pf*2.0_dp*r(1, :)
            dy(3, :) = 0.0_dp
         END SELECT
      CASE (3)
         SELECT CASE (m)
         CASE DEFAULT
            CALL cp__b("common/spherical_harmonics.F",1731,"l = 3 and m value out of bounds")
         CASE (3)
            pf = SQRT(35.0_dp/(32.0_dp*pi))
            dy(1, :) = pf*3.0_dp*(r(1, :)*r(1, :)-r(2, :)*r(2, :))
            dy(2, :) = -pf*6.0_dp*r(1, :)*r(2, :)
            dy(3, :) = 0.0_dp
         CASE (2)
            pf = SQRT(105.0_dp/(16.0_dp*pi))
            dy(1, :) = pf*2.0_dp*r(3, :)*r(1, :)
            dy(2, :) = -pf*2.0_dp*r(3, :)*r(2, :)
            dy(3, :) = pf*(r(1, :)**2-r(2, :)**2)
         CASE (1)
            pf = SQRT(21.0_dp/(32.0_dp*pi))
            dy(1, :) = pf*(5.0_dp*r(3, :)*r(3, :)-1.0_dp)
            dy(2, :) = 0.0_dp
            dy(3, :) = pf*10.0_dp*r(1, :)*r(3, :)
         CASE (0)
            pf = SQRT(7.0_dp/(16.0_dp*pi))
            dy(1, :) = 0.0_dp
            dy(2, :) = 0.0_dp
            dy(3, :) = pf*(15.0_dp*r(3, :)*r(3, :)-3.0_dp)
         CASE (-1)
            pf = SQRT(21.0_dp/(32.0_dp*pi))
            dy(1, :) = 0.0_dp
            dy(2, :) = pf*(5.0_dp*r(3, :)*r(3, :)-1.0_dp)
            dy(3, :) = pf*10.0_dp*r(2, :)*r(3, :)
         CASE (-2)
            pf = SQRT(105.0_dp/(16.0_dp*pi))
            dy(1, :) = pf*2.0_dp*r(2, :)*r(3, :)
            dy(2, :) = pf*2.0_dp*r(1, :)*r(3, :)
            dy(3, :) = pf*2.0_dp*r(1, :)*r(2, :)
         CASE (-3)
            pf = SQRT(35.0_dp/(32.0_dp*pi))
            dy(1, :) = pf*6.0_dp*r(1, :)*r(2, :)
            dy(2, :) = pf*3.0_dp*(r(1, :)*r(1, :)-r(2, :)*r(2, :))
            dy(3, :) = 0.0_dp
         END SELECT
      CASE DEFAULT
         IF (m < -l .OR. m > l) CALL cp__b("common/spherical_harmonics.F",1769,"m value out of bounds")
         lpm = fac(l+ABS(m))
         lmm = fac(l-ABS(m))
         IF (m == 0) THEN
            t = 4.0_dp*pi
         ELSE
            t = 2.0_dp*pi
         END IF
         IF (ABS(lpm) < EPSILON(1.0_dp)) THEN
            pf = REAL(2*l+1, KIND=dp)/t
         ELSE
            pf = (REAL(2*l+1, KIND=dp)*lmm)/(t*lpm)
         ENDIF
         pf = SQRT(pf)
         DO i = 1, SIZE(r, 2)
            z = r(3, i)
            dplm = dlegendre(z, l, ABS(m))
            plm = legendre(z, l, m)

            IF (m == 0) THEN
               dy(1, i) = 0.0_dp
               dy(2, i) = 0.0_dp
               dy(3, i) = pf*dplm
            ELSE
               rxy = SQRT(r(1, i)**2+r(2, i)**2)
               IF (rxy < EPSILON(1.0_dp)) THEN
                  dy(1:3, i) = 0.0_dp
               ELSE
                  cp = r(1, i)/rxy
                  sp = r(2, i)/rxy
                  dcp_dx = (rxy-2.0_dp*r(1, i)**2)/(rxy*rxy)
                  dcp_dy = -2.0_dp*r(1, i)*r(2, i)/(rxy*rxy)
                  dsp_dy = (rxy-2.0_dp*r(2, i)**2)/(rxy*rxy)
                  dsp_dx = -2.0_dp*r(1, i)*r(2, i)/(rxy*rxy)
                  IF (m > 0) THEN
                     dy(1, i) = pf*plm*(dcosn_dcp(m, cp, sp)*dcp_dx+ &
                                        dcosn_dsp(m, cp, sp)*dsp_dx)
                     dy(2, i) = pf*plm*(dcosn_dcp(m, cp, sp)*dcp_dy+ &
                                        dcosn_dsp(m, cp, sp)*dsp_dy)
                     dy(3, i) = pf*dplm*cosn(m, cp, sp)
                  ELSE
                     dy(1, i) = pf*plm*(dsinn_dcp(-m, cp, sp)*dcp_dx+ &
                                        dsinn_dsp(-m, cp, sp)*dsp_dx)
                     dy(2, i) = pf*plm*(dsinn_dcp(-m, cp, sp)*dcp_dy+ &
                                        dsinn_dsp(-m, cp, sp)*dsp_dy)
                     dy(3, i) = pf*dplm*sinn(-m, cp, sp)
                  END IF
               END IF
            END IF
         END DO
      END SELECT

   END SUBROUTINE drvy_lm

END MODULE spherical_harmonics
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE qs_harmonics_atom

   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE kinds,                           ONLY: dp
   USE lebedev,                         ONLY: lebedev_grid
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: indco,&
                                              indso,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset,&
                                              soset
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE spherical_harmonics,             ONLY: y_lm
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_harmonics_atom'

   TYPE harmonics_atom_type
      INTEGER                                :: max_s_harm, llmax, &
                                                max_iso_not0, &
                                                dmax_iso_not0, &
                                                damax_iso_not0, &
                                                ngrid
      REAL(dp), DIMENSION(:, :), POINTER   :: a, slm, dmslm
      REAL(dp), DIMENSION(:, :, :), POINTER   :: dslm, dslm_dxyz
      REAL(dp), DIMENSION(:, :, :), POINTER   :: my_CG, my_dCG
      REAL(dp), DIMENSION(:, :, :, :), POINTER :: my_CG_dxyz
      REAL(dp), DIMENSION(:, :, :, :), POINTER :: my_CG_dxyz_asym
      REAL(dp), DIMENSION(:), POINTER        :: slm_int

   END TYPE harmonics_atom_type

   PUBLIC :: allocate_harmonics_atom, &
             create_harmonics_atom, &
             deallocate_harmonics_atom, &
             get_none0_cg_list

   PUBLIC :: harmonics_atom_type, get_maxl_CG

   INTERFACE get_none0_cg_list
      MODULE PROCEDURE get_none0_cg_list3, get_none0_cg_list4
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief   Allocate a spherical harmonics set for the atom grid.
!> \param harmonics ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_harmonics_atom(harmonics)

      TYPE(harmonics_atom_type), POINTER                 :: harmonics

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_harmonics_atom', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(harmonics)) CALL deallocate_harmonics_atom(harmonics)

      ALLOCATE (harmonics)

      harmonics%max_s_harm = 0
      harmonics%llmax = 0
      harmonics%max_iso_not0 = 0
      harmonics%dmax_iso_not0 = 0
      harmonics%damax_iso_not0 = 0
      harmonics%ngrid = 0

      NULLIFY (harmonics%slm)
      NULLIFY (harmonics%dmslm)
      NULLIFY (harmonics%dslm)
      NULLIFY (harmonics%dslm_dxyz)
      NULLIFY (harmonics%slm_int)
      NULLIFY (harmonics%my_CG)
      NULLIFY (harmonics%my_dCG)
      NULLIFY (harmonics%my_CG_dxyz)
      NULLIFY (harmonics%my_CG_dxyz_asym)
      NULLIFY (harmonics%a)

   END SUBROUTINE allocate_harmonics_atom

! **************************************************************************************************
!> \brief   Deallocate the spherical harmonics set for the atom grid.
!> \param harmonics ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_harmonics_atom(harmonics)

      TYPE(harmonics_atom_type), POINTER                 :: harmonics

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_harmonics_atom', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(harmonics)) THEN

         IF (ASSOCIATED(harmonics%slm)) THEN
            DEALLOCATE (harmonics%slm)
         ENDIF

         IF (ASSOCIATED(harmonics%dmslm)) THEN
            DEALLOCATE (harmonics%dmslm)
         ENDIF

         IF (ASSOCIATED(harmonics%dslm)) THEN
            DEALLOCATE (harmonics%dslm)
         ENDIF

         IF (ASSOCIATED(harmonics%dslm_dxyz)) THEN
            DEALLOCATE (harmonics%dslm_dxyz)
         ENDIF

         IF (ASSOCIATED(harmonics%slm_int)) THEN
            DEALLOCATE (harmonics%slm_int)
         ENDIF

         IF (ASSOCIATED(harmonics%my_CG)) THEN
            DEALLOCATE (harmonics%my_CG)
         ENDIF

         IF (ASSOCIATED(harmonics%my_dCG)) THEN
            DEALLOCATE (harmonics%my_dCG)
         ENDIF

         IF (ASSOCIATED(harmonics%my_CG_dxyz)) THEN
            DEALLOCATE (harmonics%my_CG_dxyz)
         ENDIF

         IF (ASSOCIATED(harmonics%my_CG_dxyz_asym)) THEN
            DEALLOCATE (harmonics%my_CG_dxyz_asym)
         ENDIF

         IF (ASSOCIATED(harmonics%a)) THEN
            DEALLOCATE (harmonics%a)
         ENDIF

         DEALLOCATE (harmonics)
      ELSE
         CALL cp_abort(cp__l("qs_harmonics_atom.F",150), &
                       "The pointer harmonics is not associated and "// &
                       "cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_harmonics_atom

! **************************************************************************************************
!> \brief ...
!> \param harmonics ...
!> \param my_CG ...
!> \param na ...
!> \param llmax ...
!> \param maxs ...
!> \param max_s_harm ...
!> \param ll ...
!> \param wa ...
!> \param pol ...
!> \param cos_pol ...
!> \param cotan_pol ...
! **************************************************************************************************
   SUBROUTINE create_harmonics_atom(harmonics, my_CG, na, &
                                    llmax, maxs, max_s_harm, ll, wa, pol, cos_pol, cotan_pol)

      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG
      INTEGER, INTENT(IN)                                :: na, llmax, maxs, max_s_harm, ll
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: wa, pol, cos_pol, cotan_pol

      CHARACTER(len=*), PARAMETER :: routineN = 'create_harmonics_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ia, ic, is, is1, is2, iso, &
                                                            iso1, iso2, iso_mm, iso_mm1, iso_mp1, &
                                                            l, l1, l2, lx, ly, lz, m, m1, m2, n
      REAL(dp)                                           :: abs_m, cosmp1, drx, dry, drz, int1, &
                                                            int2, int3, l_r, m_r, rx, ry, rz, &
                                                            sinmp1
      REAL(dp), DIMENSION(:), POINTER                    :: slm_int, y
      REAL(dp), DIMENSION(:, :), POINTER                 :: dc, dy, slm
      REAL(dp), DIMENSION(:, :, :), POINTER              :: dslm_dxyz

      CALL timeset(routineN, handle)

      NULLIFY (y, dy, slm, dslm_dxyz, dc)

      IF(.NOT.(ASSOCIATED(harmonics)))CALL cp__a("qs_harmonics_atom.F",196)

      harmonics%max_s_harm = max_s_harm
      harmonics%llmax = llmax
      harmonics%ngrid = na

      NULLIFY (harmonics%my_CG, harmonics%my_CG_dxyz)
      CALL reallocate(harmonics%my_CG, 1, maxs, 1, maxs, 1, max_s_harm)
      CALL reallocate(harmonics%my_CG_dxyz, 1, 3, 1, maxs, 1, maxs, 1, max_s_harm)
      CALL reallocate(harmonics%my_CG_dxyz_asym, 1, 3, 1, maxs, 1, maxs, 1, max_s_harm)

      DO i = 1, max_s_harm
         DO is1 = 1, maxs
            harmonics%my_CG(1:maxs, is1, i) = my_CG(1:maxs, is1, i)
         END DO
      END DO

      !   *** allocate and calculate the spherical harmonics LM for this grid ***
      !       and their derivatives
      CALL reallocate(harmonics%slm, 1, na, 1, max_s_harm)
      CALL reallocate(harmonics%dmslm, 1, na, 1, max_s_harm)
      CALL reallocate(harmonics%dslm, 1, 2, 1, na, 1, max_s_harm)
      CALL reallocate(harmonics%dslm_dxyz, 1, 3, 1, na, 1, max_s_harm)
      CALL reallocate(harmonics%a, 1, 3, 1, na)
      CALL reallocate(y, 1, na)
      CALL reallocate(dy, 1, 3, 1, na)
      CALL reallocate(dc, 1, nco(llmax), 1, 3)
      CALL reallocate(harmonics%slm_int, 1, max_s_harm)

      slm => harmonics%slm
      dslm_dxyz => harmonics%dslm_dxyz
      slm_int => harmonics%slm_int
      slm_int = 0.0_dp

      DO iso = 1, max_s_harm

         l = indso(1, iso)
         m = indso(2, iso)

         CALL y_lm(lebedev_grid(ll)%r, y, l, m)
         slm(1:na, iso) = y(1:na)

         DO i = 1, na
            slm_int(iso) = slm_int(iso)+slm(i, iso)*wa(i)
         END DO ! i

      END DO ! iso

      DO i = 1, 3
         harmonics%a(i, 1:na) = lebedev_grid(ll)%r(i, 1:na)
      END DO

      DO ia = 1, na
         DO l = 0, indso(1, max_s_harm)
            DO ic = 1, nco(l)
               lx = indco(1, ic+ncoset(l-1))
               ly = indco(2, ic+ncoset(l-1))
               lz = indco(3, ic+ncoset(l-1))

               IF (lx == 0) THEN
                  rx = 1.0_dp
                  drx = 0.0_dp
               ELSE IF (lx == 1) THEN
                  rx = lebedev_grid(ll)%r(1, ia)
                  drx = 1.0_dp
               ELSE
                  rx = lebedev_grid(ll)%r(1, ia)**lx
                  drx = REAL(lx, dp)*lebedev_grid(ll)%r(1, ia)**(lx-1)
               END IF
               IF (ly == 0) THEN
                  ry = 1.0_dp
                  dry = 0.0_dp
               ELSE IF (ly == 1) THEN
                  ry = lebedev_grid(ll)%r(2, ia)
                  dry = 1.0_dp
               ELSE
                  ry = lebedev_grid(ll)%r(2, ia)**ly
                  dry = REAL(ly, dp)*lebedev_grid(ll)%r(2, ia)**(ly-1)
               END IF
               IF (lz == 0) THEN
                  rz = 1.0_dp
                  drz = 0.0_dp
               ELSE IF (lz == 1) THEN
                  rz = lebedev_grid(ll)%r(3, ia)
                  drz = 1.0_dp
               ELSE
                  rz = lebedev_grid(ll)%r(3, ia)**lz
                  drz = REAL(lz, dp)*lebedev_grid(ll)%r(3, ia)**(lz-1)
               END IF
               dc(ic, 1) = drx*ry*rz
               dc(ic, 2) = rx*dry*rz
               dc(ic, 3) = rx*ry*drz
            END DO
            n = nsoset(l-1)
            DO is = 1, nso(l)
               iso = is+n
               dslm_dxyz(1:3, ia, iso) = 0.0_dp
               DO ic = 1, nco(l)
                  dslm_dxyz(1, ia, iso) = dslm_dxyz(1, ia, iso)+ &
                                          orbtramat(l)%slm(is, ic)*dc(ic, 1)
                  dslm_dxyz(2, ia, iso) = dslm_dxyz(2, ia, iso)+ &
                                          orbtramat(l)%slm(is, ic)*dc(ic, 2)
                  dslm_dxyz(3, ia, iso) = dslm_dxyz(3, ia, iso)+ &
                                          orbtramat(l)%slm(is, ic)*dc(ic, 3)
               END DO
            END DO
         END DO ! l
      END DO

      ! Expansion coefficients of the cartesian derivatives
      ! of the product of two harmonics :
      ! d(Y(l1m1) * Y(l2m2))/dx ; d(Y(l1m1) * Y(l2m2))/dy ; d(Y(l1m1) * Y(l2m2))/dz
      harmonics%my_CG_dxyz = 0.0_dp
      DO iso1 = 1, maxs
         DO iso2 = 1, maxs
            int1 = 0.0_dp
            int2 = 0.0_dp
            int3 = 0.0_dp
            DO ia = 1, na
               int1 = int1+wa(ia)* &
                      (dslm_dxyz(1, ia, iso1)*slm(ia, iso2)+slm(ia, iso1)*dslm_dxyz(1, ia, iso2))
               int2 = int2+wa(ia)* &
                      (dslm_dxyz(2, ia, iso1)*slm(ia, iso2)+slm(ia, iso1)*dslm_dxyz(2, ia, iso2))
               int3 = int3+wa(ia)* &
                      (dslm_dxyz(3, ia, iso1)*slm(ia, iso2)+slm(ia, iso1)*dslm_dxyz(3, ia, iso2))
            END DO
            DO iso = 1, max_s_harm
               rx = 0.0_dp
               drx = 0.0_dp
               ry = 0.0_dp
               dry = 0.0_dp
               rz = 0.0_dp
               drz = 0.0_dp
               DO ia = 1, na
                  rx = rx+wa(ia)*slm(ia, iso)* &
                       (dslm_dxyz(1, ia, iso1)*slm(ia, iso2)+slm(ia, iso1)*dslm_dxyz(1, ia, iso2))
                  drx = drx+wa(ia)*slm(ia, iso)
                  ry = ry+wa(ia)*slm(ia, iso)* &
                       (dslm_dxyz(2, ia, iso1)*slm(ia, iso2)+slm(ia, iso1)*dslm_dxyz(2, ia, iso2))
                  dry = dry+wa(ia)*slm(ia, iso)
                  rz = rz+wa(ia)*slm(ia, iso)* &
                       (dslm_dxyz(3, ia, iso1)*slm(ia, iso2)+slm(ia, iso1)*dslm_dxyz(3, ia, iso2))
                  drz = drz+wa(ia)*slm(ia, iso)
               END DO

               harmonics%my_CG_dxyz(1, iso1, iso2, iso) = rx

               harmonics%my_CG_dxyz(2, iso1, iso2, iso) = ry

               harmonics%my_CG_dxyz(3, iso1, iso2, iso) = rz

            END DO
         END DO
      END DO

      ! Expansion coefficients of the cartesian of the combinations
      ! Y(l1m1) * d(Y(l2m2))/dx -  d(Y(l1m1))/dx * Y(l2m2)
      ! Y(l1m1) * d(Y(l2m2))/dy -  d(Y(l1m1))/dy * Y(l2m2)
      ! Y(l1m1) * d(Y(l2m2))/dz -  d(Y(l1m1))/dz * Y(l2m2)
      harmonics%my_CG_dxyz_asym = 0.0_dp
      DO iso1 = 1, maxs
         DO iso2 = 1, maxs
            DO iso = 1, max_s_harm
               rx = 0.0_dp
               drx = 0.0_dp
               ry = 0.0_dp
               dry = 0.0_dp
               rz = 0.0_dp
               drz = 0.0_dp
               DO ia = 1, na
                  rx = rx+wa(ia)*slm(ia, iso)* &
                       (-dslm_dxyz(1, ia, iso1)*slm(ia, iso2)+ &
                        slm(ia, iso1)*dslm_dxyz(1, ia, iso2))
                  drx = drx+wa(ia)*slm(ia, iso)
                  ry = ry+wa(ia)*slm(ia, iso)* &
                       (-dslm_dxyz(2, ia, iso1)*slm(ia, iso2)+ &
                        slm(ia, iso1)*dslm_dxyz(2, ia, iso2))
                  dry = dry+wa(ia)*slm(ia, iso)
                  rz = rz+wa(ia)*slm(ia, iso)* &
                       (-dslm_dxyz(3, ia, iso1)*slm(ia, iso2)+ &
                        slm(ia, iso1)*dslm_dxyz(3, ia, iso2))
                  drz = drz+wa(ia)*slm(ia, iso)
               END DO

               harmonics%my_CG_dxyz_asym(1, iso1, iso2, iso) = rx

               harmonics%my_CG_dxyz_asym(2, iso1, iso2, iso) = ry

               harmonics%my_CG_dxyz_asym(3, iso1, iso2, iso) = rz

            END DO ! iso
         END DO ! iso2
      END DO ! iso1

      ! Calculate the derivatives of the harmonics with respect of the 2 angles
      ! the first angle (polar) is acos(lebedev_grid(ll)%r(3))
      ! the second angle (azimutal) is atan(lebedev_grid(ll)%r(2)/lebedev_grid(ll)%r(1))
      DO iso = 1, max_s_harm

         l = indso(1, iso)
         m = indso(2, iso)
         l_r = REAL(l, dp)
         m_r = REAL(m, dp)

         ! Derivative with respect to the azimuthal angle
         IF (m == 0) THEN
            harmonics%dslm(2, 1:na, iso) = 0.0_dp
         ELSE
            iso_mm = soset(l, -m)
            harmonics%dslm(2, 1:na, iso) = -m_r*harmonics%slm(1:na, iso_mm)
         END IF

         !  Derivative with respect to the polar angle
         IF (m > 0) THEN

            DO ia = 1, na

               harmonics%dslm(1, ia, iso) = m_r*cotan_pol(ia)*harmonics%slm(ia, iso)

               IF (m /= l) THEN
                  iso_mp1 = soset(l, m+1)

                  cosmp1 = COS((m_r+1)*pol(ia))
                  IF (ABS(cosmp1) > EPSILON(1.0_dp)) THEN
                     harmonics%dslm(1, ia, iso) = harmonics%dslm(1, ia, iso)+ &
                                                  COS(m_r*pol(ia))*harmonics%slm(ia, iso_mp1)/ &
                                                  cosmp1*SQRT((l_r-m_r)*(l_r+m_r+1._dp))
                  END IF

               END IF
            END DO

         ELSEIF (m == 0) THEN
            IF (l == 0) THEN
               harmonics%dslm(1, 1:na, iso) = 0.0_dp
            ELSE
               iso_mp1 = soset(l, 1)
               iso_mm1 = soset(l, -1)
               DO ia = 1, na

                  IF (ABS(cos_pol(ia)) > EPSILON(1.0_dp)) THEN
                     harmonics%dslm(1, ia, iso) = SQRT(0.5_dp*l_r*(l_r+1._dp))* &
                                                  harmonics%slm(ia, iso_mp1)/cos_pol(ia)
                  END IF
               END DO
            END IF
         ELSE

            abs_m = ABS(m_r)
            DO ia = 1, na
               harmonics%dslm(1, ia, iso) = abs_m*cotan_pol(ia)* &
                                            harmonics%slm(ia, iso)
               IF (m /= -l) THEN
                  iso_mm1 = soset(l, m-1)

                  sinmp1 = SIN((abs_m+1._dp)*pol(ia))
                  IF (ABS(sinmp1) > EPSILON(1.0_dp)) THEN
                     harmonics%dslm(1, ia, iso) = harmonics%dslm(1, ia, iso)+ &
                                                  SIN(abs_m*pol(ia))*harmonics%slm(ia, iso_mm1)/ &
                                                  sinmp1*SQRT((l_r-abs_m)*(l_r+abs_m+1._dp))
                  END IF
               END IF
            END DO

         END IF

      END DO

      ! expansion coefficients of product of polar angle derivatives (dslm(1...)) in
      ! spherical harmonics (used for tau functionals)
      CALL reallocate(harmonics%my_dCG, 1, maxs, 1, maxs, 1, max_s_harm)
      DO is1 = 1, maxs
         l1 = indso(1, is1)
         m1 = indso(2, is1)
         DO is2 = 1, maxs
            l2 = indso(1, is2)
            m2 = indso(2, is2)
            DO iso = 1, max_s_harm
               l = indso(1, iso)
               m = indso(2, iso)
               CALL y_lm(lebedev_grid(ll)%r, y, l, m)
               y(1:na) = y(1:na)*lebedev_grid(ll)%w(1:na)*harmonics%dslm(1, 1:na, is1)*harmonics%dslm(1, 1:na, is2)
               harmonics%my_dCG(is1, is2, iso) = SUM(y(1:na))
            END DO
         END DO
      END DO

      DEALLOCATE (y, dy, dc)

      CALL timestop(handle)

   END SUBROUTINE create_harmonics_atom

! **************************************************************************************************
!> \brief ...
!> \param harmonics ...
!> \param orb_basis ...
!> \param llmax ...
!> \param max_s_harm ...
! **************************************************************************************************
   SUBROUTINE get_maxl_CG(harmonics, orb_basis, llmax, max_s_harm)

      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      INTEGER, INTENT(IN)                                :: llmax, max_s_harm

      CHARACTER(len=*), PARAMETER :: routineN = 'get_maxl_CG', routineP = moduleN//':'//routineN

      INTEGER                                            :: damax_iso_not0, dmax_iso_not0, handle, &
                                                            is1, is2, itmp, max_iso_not0, nset
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(harmonics)))CALL cp__a("qs_harmonics_atom.F",510)

      CALL get_gto_basis_set(gto_basis_set=orb_basis, lmax=lmax, lmin=lmin, nset=nset)

      !   *** Assign indexes for the non null CG coefficients ***
      max_iso_not0 = 0
      dmax_iso_not0 = 0
      damax_iso_not0 = 0
      DO is1 = 1, nset
         DO is2 = 1, nset
            CALL get_none0_cg_list(harmonics%my_CG, &
                                   lmin(is1), lmax(is1), lmin(is2), lmax(is2), &
                                   max_s_harm, llmax, max_iso_not0=itmp)
            max_iso_not0 = MAX(max_iso_not0, itmp)
            CALL get_none0_cg_list(harmonics%my_dCG, &
                                   lmin(is1), lmax(is1), lmin(is2), lmax(is2), &
                                   max_s_harm, llmax, max_iso_not0=itmp)
            dmax_iso_not0 = MAX(dmax_iso_not0, itmp)
            CALL get_none0_cg_list(harmonics%my_CG_dxyz_asym, &
                                   lmin(is1), lmax(is1), lmin(is2), lmax(is2), &
                                   max_s_harm, llmax, max_iso_not0=itmp)
            damax_iso_not0 = MAX(damax_iso_not0, itmp)
         END DO ! is2
      END DO ! is1
      harmonics%max_iso_not0 = max_iso_not0
      harmonics%dmax_iso_not0 = dmax_iso_not0
      harmonics%damax_iso_not0 = damax_iso_not0

      CALL timestop(handle)

   END SUBROUTINE get_maxl_CG

! **************************************************************************************************
!> \brief ...
!> \param cgc ...
!> \param lmin1 ...
!> \param lmax1 ...
!> \param lmin2 ...
!> \param lmax2 ...
!> \param max_s_harm ...
!> \param llmax ...
!> \param list ...
!> \param n_list ...
!> \param max_iso_not0 ...
! **************************************************************************************************
   SUBROUTINE get_none0_cg_list4(cgc, lmin1, lmax1, lmin2, lmax2, max_s_harm, llmax, &
                                 list, n_list, max_iso_not0)

      REAL(dp), DIMENSION(:, :, :, :), INTENT(IN)        :: cgc
      INTEGER, INTENT(IN)                                :: lmin1, lmax1, lmin2, lmax2, max_s_harm, &
                                                            llmax
      INTEGER, DIMENSION(:, :, :), INTENT(OUT), OPTIONAL :: list
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL       :: n_list
      INTEGER, INTENT(OUT)                               :: max_iso_not0

      CHARACTER(len=*), PARAMETER :: routineN = 'get_none0_cg_list4', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iso, iso1, iso2, l1, l2, nlist

      IF(.NOT.(nsoset(lmax1) .LE. SIZE(cgc, 2)))CALL cp__a("qs_harmonics_atom.F",570)
      IF(.NOT.(nsoset(lmax2) .LE. SIZE(cgc, 3)))CALL cp__a("qs_harmonics_atom.F",571)
      IF(.NOT.(max_s_harm .LE. SIZE(cgc, 4)))CALL cp__a("qs_harmonics_atom.F",572)
      IF (PRESENT(n_list) .AND. PRESENT(list)) THEN
         IF(.NOT.(max_s_harm .LE. SIZE(list, 3)))CALL cp__a("qs_harmonics_atom.F",574)
      ENDIF
      max_iso_not0 = 0
      IF (PRESENT(n_list) .AND. PRESENT(list)) n_list = 0
      DO iso = 1, max_s_harm
         nlist = 0
         DO l1 = lmin1, lmax1
            DO iso1 = nsoset(l1-1)+1, nsoset(l1)
               DO l2 = lmin2, lmax2
                  IF (l1+l2 > llmax) CYCLE
                  DO iso2 = nsoset(l2-1)+1, nsoset(l2)
                     IF (ABS(cgc(1, iso1, iso2, iso))+ &
                         ABS(cgc(2, iso1, iso2, iso))+ &
                         ABS(cgc(3, iso1, iso2, iso)) > 1.E-8_dp) THEN
                        nlist = nlist+1
                        IF (PRESENT(n_list) .AND. PRESENT(list)) THEN
                           list(1, nlist, iso) = iso1
                           list(2, nlist, iso) = iso2
                        ENDIF
                        max_iso_not0 = MAX(max_iso_not0, iso)
                     ENDIF
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
         IF (PRESENT(n_list) .AND. PRESENT(list)) n_list(iso) = nlist
      ENDDO
   END SUBROUTINE get_none0_cg_list4

! **************************************************************************************************
!> \brief ...
!> \param cgc ...
!> \param lmin1 ...
!> \param lmax1 ...
!> \param lmin2 ...
!> \param lmax2 ...
!> \param max_s_harm ...
!> \param llmax ...
!> \param list ...
!> \param n_list ...
!> \param max_iso_not0 ...
! **************************************************************************************************
   SUBROUTINE get_none0_cg_list3(cgc, lmin1, lmax1, lmin2, lmax2, max_s_harm, llmax, &
                                 list, n_list, max_iso_not0)

      REAL(dp), DIMENSION(:, :, :), INTENT(IN)           :: cgc
      INTEGER, INTENT(IN)                                :: lmin1, lmax1, lmin2, lmax2, max_s_harm, &
                                                            llmax
      INTEGER, DIMENSION(:, :, :), INTENT(OUT), OPTIONAL :: list
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL       :: n_list
      INTEGER, INTENT(OUT)                               :: max_iso_not0

      CHARACTER(len=*), PARAMETER :: routineN = 'get_none0_cg_list3', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iso, iso1, iso2, l1, l2, nlist

      IF(.NOT.(nsoset(lmax1) .LE. SIZE(cgc, 1)))CALL cp__a("qs_harmonics_atom.F",631)
      IF(.NOT.(nsoset(lmax2) .LE. SIZE(cgc, 2)))CALL cp__a("qs_harmonics_atom.F",632)
      IF(.NOT.(max_s_harm .LE. SIZE(cgc, 3)))CALL cp__a("qs_harmonics_atom.F",633)
      IF (PRESENT(n_list) .AND. PRESENT(list)) THEN
         IF(.NOT.(max_s_harm .LE. SIZE(list, 3)))CALL cp__a("qs_harmonics_atom.F",635)
      ENDIF
      max_iso_not0 = 0
      IF (PRESENT(n_list) .AND. PRESENT(list)) n_list = 0
      DO iso = 1, max_s_harm
         nlist = 0
         DO l1 = lmin1, lmax1
            DO iso1 = nsoset(l1-1)+1, nsoset(l1)
               DO l2 = lmin2, lmax2
                  IF (l1+l2 > llmax) CYCLE
                  DO iso2 = nsoset(l2-1)+1, nsoset(l2)
                     IF (ABS(cgc(iso1, iso2, iso)) > 1.E-8_dp) THEN
                        nlist = nlist+1
                        IF (PRESENT(n_list) .AND. PRESENT(list)) THEN
                           list(1, nlist, iso) = iso1
                           list(2, nlist, iso) = iso2
                        ENDIF
                        max_iso_not0 = MAX(max_iso_not0, iso)
                     ENDIF
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
         IF (PRESENT(n_list) .AND. PRESENT(list)) n_list(iso) = nlist
      ENDDO
   END SUBROUTINE get_none0_cg_list3

END MODULE qs_harmonics_atom
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
!> \author MI (08.01.2004)
! **************************************************************************************************
MODULE soft_basis_set

   USE ao_util,                         ONLY: exp_radius
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type,&
                                              init_cphi_and_sphi
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: indco,&
                                              nco,&
                                              ncoset,&
                                              nso
   USE orbital_symbols,                 ONLY: cgf_symbol,&
                                              sgf_symbol
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'soft_basis_set'

   INTEGER, PARAMETER :: max_name_length = 60

! *** Public subroutines ***

   PUBLIC :: create_soft_basis

CONTAINS

! **************************************************************************************************
!> \brief   create the soft basis from a GTO basis
!> \param orb_basis ...
!> \param soft_basis ...
!> \param eps_fit ...
!> \param rc ...
!> \param paw_atom ...
!> \param paw_type_forced ...
!> \param gpw_type_forced ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE create_soft_basis(orb_basis, soft_basis, eps_fit, rc, paw_atom, &
                                paw_type_forced, gpw_type_forced)

      TYPE(gto_basis_set_type), POINTER                  :: orb_basis, soft_basis
      REAL(dp), INTENT(IN)                               :: eps_fit, rc
      LOGICAL, INTENT(OUT)                               :: paw_atom
      LOGICAL, INTENT(IN)                                :: paw_type_forced, gpw_type_forced

      CHARACTER(len=*), PARAMETER :: routineN = 'create_soft_basis', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: bsname
      INTEGER :: ico, ipgf, ipgf_s, iset, iset_s, ishell, lshell, lshell_old, m, maxco, maxpgf, &
         maxpgf_s, maxshell, maxshell_s, ncgf, nset, nset_s, nsgf
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iset_s2h
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l, n
      LOGICAL                                            :: my_gpw_type_forced
      REAL(KIND=dp)                                      :: minzet, radius
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc

      NULLIFY (gcc, l, lmax, lmin, n, npgf, nshell, zet)
      paw_atom = .FALSE.
      my_gpw_type_forced = gpw_type_forced
      IF (paw_type_forced) THEN
         paw_atom = .TRUE.
         my_gpw_type_forced = .FALSE.
      END IF
      IF (.NOT. my_gpw_type_forced) THEN
         CALL get_gto_basis_set(gto_basis_set=orb_basis, name=bsname, &
                                maxpgf=maxpgf, maxshell=maxshell, nset=nset)

         soft_basis%name = TRIM(bsname)//"_soft"

         CALL reallocate(npgf, 1, nset)
         CALL reallocate(nshell, 1, nset)
         CALL reallocate(lmax, 1, nset)
         CALL reallocate(lmin, 1, nset)

         CALL reallocate(n, 1, maxshell, 1, nset)
         CALL reallocate(l, 1, maxshell, 1, nset)

         CALL reallocate(zet, 1, maxpgf, 1, nset)
         CALL reallocate(gcc, 1, maxpgf, 1, maxshell, 1, nset)

         ALLOCATE (iset_s2h(nset))

         iset_s2h = 0
         iset_s = 0
         maxpgf_s = 0
         maxshell_s = 0

         DO iset = 1, nset ! iset
            minzet = orb_basis%zet(orb_basis%npgf(iset), iset)
            DO ipgf = orb_basis%npgf(iset)-1, 1, -1
               IF (orb_basis%zet(ipgf, iset) < minzet) THEN
                  minzet = orb_basis%zet(ipgf, iset)
               ENDIF
            ENDDO
            radius = exp_radius(orb_basis%lmax(iset), minzet, eps_fit, 1.0_dp)

            !      The soft basis contains this set
            iset_s = iset_s+1
            nshell(iset_s) = orb_basis%nshell(iset)
            lmax(iset_s) = orb_basis%lmax(iset)
            lmin(iset_s) = orb_basis%lmin(iset)

            iset_s2h(iset_s) = iset

            DO ishell = 1, nshell(iset_s)
               n(ishell, iset_s) = orb_basis%n(ishell, iset)
               l(ishell, iset_s) = orb_basis%l(ishell, iset)
            ENDDO

            IF (nshell(iset_s) > maxshell_s) THEN
               maxshell_s = nshell(iset_s)
            ENDIF

            IF (radius < rc) THEN
               ! The soft basis does not contain this set
               ! For the moment I keep the set as a dummy set
               ! with no exponents, in order to have the right number of contractions
               ! In a second time it can be taken away, by creating a pointer
               ! which connects the remaining sets to the right contraction index
               paw_atom = .TRUE.
               npgf(iset_s) = 0
               CYCLE
            ENDIF

            ipgf_s = 0
            DO ipgf = 1, orb_basis%npgf(iset) ! ipgf
               IF (orb_basis%zet(ipgf, iset) > 100.0_dp) THEN
                  ! The soft basis does not contain this exponent
                  paw_atom = .TRUE.
                  CYCLE
               ENDIF

               radius = exp_radius(orb_basis%lmax(iset), orb_basis%zet(ipgf, iset), &
                                   eps_fit, 1.0_dp)
               IF (radius < rc) THEN
                  ! The soft basis does not contain this exponent
                  paw_atom = .TRUE.
                  CYCLE
               ENDIF

               ! The soft basis contains this exponent
               ipgf_s = ipgf_s+1
               zet(ipgf_s, iset_s) = orb_basis%zet(ipgf, iset)

               lshell_old = orb_basis%l(1, iset)
               radius = exp_radius(lshell_old, zet(ipgf_s, iset_s), eps_fit, 1.0_dp)

               DO ishell = 1, nshell(iset_s)
                  lshell = orb_basis%l(ishell, iset)
                  IF (lshell == lshell_old) THEN
                  ELSE
                     lshell_old = lshell
                     radius = exp_radius(lshell_old, zet(ipgf_s, iset_s), eps_fit, 1.0_dp)
                  ENDIF
                  IF (radius < rc) THEN
                     gcc(ipgf_s, ishell, iset_s) = 0.0_dp
                     paw_atom = .TRUE.
                  ELSE
                     gcc(ipgf_s, ishell, iset_s) = orb_basis%gcc(ipgf, ishell, iset)
                  ENDIF
               ENDDO
            ENDDO
            npgf(iset_s) = ipgf_s
            IF (ipgf_s > maxpgf_s) THEN
               maxpgf_s = ipgf_s
            ENDIF
         ENDDO
         nset_s = iset_s
         IF (paw_atom) THEN
            soft_basis%nset = nset_s
            CALL reallocate(soft_basis%lmax, 1, nset_s)
            CALL reallocate(soft_basis%lmin, 1, nset_s)
            CALL reallocate(soft_basis%npgf, 1, nset_s)
            CALL reallocate(soft_basis%nshell, 1, nset_s)
            CALL reallocate(soft_basis%n, 1, maxshell_s, 1, nset_s)
            CALL reallocate(soft_basis%l, 1, maxshell_s, 1, nset_s)
            CALL reallocate(soft_basis%zet, 1, maxpgf_s, 1, nset_s)
            CALL reallocate(soft_basis%gcc, 1, maxpgf_s, 1, maxshell_s, 1, nset_s)

            ! *** Copy the basis set information into the data structure ***

            DO iset = 1, nset_s
               soft_basis%lmax(iset) = lmax(iset)
               soft_basis%lmin(iset) = lmin(iset)
               soft_basis%npgf(iset) = npgf(iset)
               soft_basis%nshell(iset) = nshell(iset)
               DO ishell = 1, nshell(iset)
                  soft_basis%n(ishell, iset) = n(ishell, iset)
                  soft_basis%l(ishell, iset) = l(ishell, iset)
                  DO ipgf = 1, npgf(iset)
                     soft_basis%gcc(ipgf, ishell, iset) = gcc(ipgf, ishell, iset)
                  END DO
               END DO
               DO ipgf = 1, npgf(iset)
                  soft_basis%zet(ipgf, iset) = zet(ipgf, iset)
               END DO
            END DO

            ! *** Initialise the depending soft_basis pointers ***
            CALL reallocate(soft_basis%set_radius, 1, nset_s)
            CALL reallocate(soft_basis%pgf_radius, 1, maxpgf_s, 1, nset_s)
            CALL reallocate(soft_basis%first_cgf, 1, maxshell_s, 1, nset_s)
            CALL reallocate(soft_basis%first_sgf, 1, maxshell_s, 1, nset_s)
            CALL reallocate(soft_basis%last_cgf, 1, maxshell_s, 1, nset_s)
            CALL reallocate(soft_basis%last_sgf, 1, maxshell_s, 1, nset_s)
            CALL reallocate(soft_basis%ncgf_set, 1, nset_s)
            CALL reallocate(soft_basis%nsgf_set, 1, nset_s)

            maxco = 0
            ncgf = 0
            nsgf = 0

            DO iset = 1, nset_s
               soft_basis%ncgf_set(iset) = 0
               soft_basis%nsgf_set(iset) = 0
               DO ishell = 1, nshell(iset)
                  lshell = soft_basis%l(ishell, iset)
                  soft_basis%first_cgf(ishell, iset) = ncgf+1
                  ncgf = ncgf+nco(lshell)
                  soft_basis%last_cgf(ishell, iset) = ncgf
                  soft_basis%ncgf_set(iset) = &
                     soft_basis%ncgf_set(iset)+nco(lshell)
                  soft_basis%first_sgf(ishell, iset) = nsgf+1
                  nsgf = nsgf+nso(lshell)
                  soft_basis%last_sgf(ishell, iset) = nsgf
                  soft_basis%nsgf_set(iset) = &
                     soft_basis%nsgf_set(iset)+nso(lshell)
               END DO
               maxco = MAX(maxco, npgf(iset)*ncoset(lmax(iset)))
            END DO
            soft_basis%ncgf = ncgf
            soft_basis%nsgf = nsgf

            CALL reallocate(soft_basis%cphi, 1, maxco, 1, ncgf)
            CALL reallocate(soft_basis%sphi, 1, maxco, 1, nsgf)
            CALL reallocate(soft_basis%scon, 1, maxco, 1, nsgf)
            CALL reallocate(soft_basis%lx, 1, ncgf)
            CALL reallocate(soft_basis%ly, 1, ncgf)
            CALL reallocate(soft_basis%lz, 1, ncgf)
            CALL reallocate(soft_basis%m, 1, nsgf)
            CALL reallocate(soft_basis%norm_cgf, 1, ncgf)
            ALLOCATE (soft_basis%cgf_symbol(ncgf))
            ALLOCATE (soft_basis%sgf_symbol(nsgf))

            ncgf = 0
            nsgf = 0

            DO iset = 1, nset_s
               DO ishell = 1, nshell(iset)
                  lshell = soft_basis%l(ishell, iset)
                  DO ico = ncoset(lshell-1)+1, ncoset(lshell)
                     ncgf = ncgf+1
                     soft_basis%lx(ncgf) = indco(1, ico)
                     soft_basis%ly(ncgf) = indco(2, ico)
                     soft_basis%lz(ncgf) = indco(3, ico)
                     soft_basis%cgf_symbol(ncgf) = &
                        cgf_symbol(n(ishell, iset), (/soft_basis%lx(ncgf), &
                                                      soft_basis%ly(ncgf), &
                                                      soft_basis%lz(ncgf)/))
                  END DO
                  DO m = -lshell, lshell
                     nsgf = nsgf+1
                     soft_basis%m(nsgf) = m
                     soft_basis%sgf_symbol(nsgf) = &
                        sgf_symbol(n(ishell, iset), lshell, m)
                  END DO
               END DO
            ENDDO

            ! *** Normalization factor of the contracted Gaussians ***
            soft_basis%norm_type = orb_basis%norm_type
            soft_basis%norm_cgf = orb_basis%norm_cgf
            ! *** Initialize the transformation matrices ***
            CALL init_cphi_and_sphi(soft_basis)
         ENDIF

         DEALLOCATE (gcc, l, lmax, lmin, n, npgf, nshell, zet, iset_s2h)
      END IF

      IF (.NOT. paw_atom) THEN
         DEALLOCATE (soft_basis)
         soft_basis => orb_basis
      END IF

   END SUBROUTINE create_soft_basis

END MODULE soft_basis_set
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Define the quickstep kind type and their sub types
!> \author  Ole Schuett
!>
!> <b>Modification history:</b>
!> - 01.2002 creation [MK]
!> - 04.2002 added pao [fawzi]
!> - 09.2002 adapted for POL/KG use [GT]
!> - 02.2004 flexible normalization of basis sets [jgh]
!> - 03.2004 attach/detach routines [jgh]
!> - 10.2004 removed pao [fawzi]
!> - 08.2014 separated qs-related stuff from atomic_kind_types.F [Ole Schuett]
!> - 07.2015 new container for basis sets [jgh]
! **************************************************************************************************
MODULE qs_kind_types
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_container_types,       ONLY: add_basis_set_to_container,&
                                              basis_set_container_type,&
                                              get_basis_from_container,&
                                              remove_basis_from_container,&
                                              remove_basis_set_container
   USE basis_set_types,                 ONLY: &
        allocate_gto_basis_set, combine_basis_sets, deallocate_gto_basis_set, get_gto_basis_set, &
        gto_basis_set_type, init_aux_basis_set, init_orb_basis_set, read_gto_basis_set, &
        write_gto_basis_set, write_orb_basis_set
   USE cp_control_types,                ONLY: dft_control_type,&
                                              qs_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE external_potential_types,        ONLY: &
        all_potential_type, allocate_potential, deallocate_potential, get_potential, &
        gth_potential_type, init_potential, local_potential_type, read_potential, &
        set_default_all_potential, set_potential, write_potential
   USE input_constants,                 ONLY: &
        do_method_am1, do_method_dftb, do_method_mndo, do_method_mndod, do_method_pdg, &
        do_method_pm3, do_method_pm6, do_method_pnnl, do_method_rm1, do_method_scptb, do_qs
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
   USE paw_proj_set_types,              ONLY: allocate_paw_proj_set,&
                                              deallocate_paw_proj_set,&
                                              get_paw_proj_set,&
                                              paw_proj_set_type,&
                                              projectors
   USE periodic_table,                  ONLY: get_ptable_info,&
                                              ptable
   USE physcon,                         ONLY: bohr,&
                                              evolt
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
   USE qs_dftb_utils,                   ONLY: deallocate_dftb_atom_param,&
                                              get_dftb_atom_param,&
                                              write_dftb_atom_param
   USE qs_dispersion_types,             ONLY: qs_atom_dispersion_type
   USE qs_grid_atom,                    ONLY: allocate_grid_atom,&
                                              deallocate_grid_atom,&
                                              grid_atom_type
   USE qs_harmonics_atom,               ONLY: allocate_harmonics_atom,&
                                              deallocate_harmonics_atom,&
                                              harmonics_atom_type
   USE scptb_types,                     ONLY: deallocate_scptb_parameter,&
                                              get_scptb_parameter,&
                                              scptb_parameter_type,&
                                              write_scptb_parameter
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              semi_empirical_create,&
                                              semi_empirical_release,&
                                              semi_empirical_type,&
                                              write_se_param
   USE semi_empirical_utils,            ONLY: init_se_param,&
                                              se_param_set_default
   USE soft_basis_set,                  ONLY: create_soft_basis
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_kind_types'

!> Input parameters for the DFT+U method
! **************************************************************************************************
   TYPE dft_plus_u_type
      INTEGER                                :: l, max_scf
      REAL(KIND=dp)                          :: eps_u_ramping, &
                                                eps_scf, &
                                                u_minus_j_target, &
                                                u_minus_j, &
                                                u_ramping
      INTEGER, DIMENSION(:), POINTER         :: orbitals
      LOGICAL                                :: init_u_ramping_each_scf, &
                                                smear
   END TYPE dft_plus_u_type

!> Provides all information about a quickstep kind
! **************************************************************************************************
   TYPE qs_kind_type
      CHARACTER(LEN=default_string_length)   :: name = ""
      CHARACTER(LEN=2)                       :: element_symbol = ""
      INTEGER                                :: natom = -1
      TYPE(all_potential_type), POINTER      :: all_potential => Null()
      TYPE(local_potential_type), POINTER    :: tnadd_potential => Null()
      TYPE(gth_potential_type), POINTER      :: gth_potential => Null()
      TYPE(semi_empirical_type), POINTER     :: se_parameter => Null()
      TYPE(qs_dftb_atom_type), POINTER       :: dftb_parameter => Null()
      TYPE(scptb_parameter_type), POINTER    :: scptb_parameter => Null()
      !
      TYPE(gto_basis_set_type), POINTER      :: soft_basis_set => Null()
      TYPE(gto_basis_set_type), POINTER      :: hard_basis_set => Null()
      !
      TYPE(basis_set_container_type), &
         DIMENSION(20)                        :: basis_sets
      !
      TYPE(paw_proj_set_type), POINTER      :: paw_proj_set => Null()
      !
      REAL(dp)                               :: hard_radius = 0.8_dp*bohr ! for hard and soft exp
      REAL(dp)                               :: hard0_radius = 0.8_dp*bohr ! for hard exp of rho0
      REAL(dp)                               :: max_rad_local = 13.2_dp*bohr ! max GTO radius used in GAPW
      LOGICAL                                :: paw_atom = .FALSE. ! needs atomic rho1
      LOGICAL                                :: gpw_type_forced = .FALSE. ! gpw atom even if with hard exponents
      LOGICAL                                :: ghost = .FALSE.
      INTEGER                                :: lmax_dftb = -1
      REAL(KIND=dp)                        :: dudq_dftb3 = 0.0_dp
      INTEGER, DIMENSION(:, :), POINTER       :: addel => Null()
      INTEGER, DIMENSION(:, :), POINTER       :: laddel => Null()
      INTEGER, DIMENSION(:, :), POINTER       :: naddel => Null()
      TYPE(harmonics_atom_type), POINTER     :: harmonics => Null()
      TYPE(grid_atom_type), POINTER          :: grid_atom => Null()
      INTEGER                                :: ngrid_rad = 50
      INTEGER                                :: ngrid_ang = 50
      INTEGER                                :: lmax_rho0 = 0
      INTEGER                                :: mao = -1
      INTEGER, DIMENSION(:), POINTER         :: elec_conf => Null() ! used to set up the initial atomic guess
      LOGICAL                                :: bs_occupation = .FALSE.
      TYPE(dft_plus_u_type), POINTER         :: dft_plus_u => Null()
      LOGICAL                                :: no_optimize = .TRUE.
      !
      TYPE(qs_atom_dispersion_type), POINTER :: dispersion => Null()
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: reltmat => Null()
      INTEGER                                :: pao_basis_size = -1
      INTEGER                                :: pao_pot_maxl = -1
      REAL(KIND=dp)                          :: pao_pot_beta = 0.0_dp
      REAL(KIND=dp)                          :: pao_pot_weight = 0.0_dp
      INTEGER                                :: pao_pot_max_projector = -1
      INTEGER                                :: pao_pot_neighbors = -1
   END TYPE qs_kind_type

!> Provides a vector of pointers of type qs_kind_type
! **************************************************************************************************
   TYPE qs_kind_p_type
      TYPE(qs_kind_type), DIMENSION(:), &
         POINTER                             :: qs_kind_set
   END TYPE qs_kind_p_type

   ! Public subroutines

   PUBLIC :: check_qs_kind_set, &
             deallocate_qs_kind_set, &
             get_qs_kind, &
             get_qs_kind_set, &
             init_qs_kind_set, &
             init_gapw_basis_set, &
             create_qs_kind_set, &
             set_qs_kind, &
             write_qs_kind_set, &
             write_gto_basis_sets, &
             init_atom_electronic_state

   ! Public data types
   PUBLIC :: qs_kind_type

CONTAINS

! **************************************************************************************************
!> \brief   Destructor routine for a set of qs kinds
!> \param qs_kind_set ...
!> \date    02.01.2002
!> \author  Matthias Krack (MK)
!> \version 2.0
! **************************************************************************************************
   SUBROUTINE deallocate_qs_kind_set(qs_kind_set)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_qs_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nkind

      IF (ASSOCIATED(qs_kind_set)) THEN

         nkind = SIZE(qs_kind_set)

         DO ikind = 1, nkind
            IF (ASSOCIATED(qs_kind_set(ikind)%all_potential)) THEN
               CALL deallocate_potential(qs_kind_set(ikind)%all_potential)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%tnadd_potential)) THEN
               CALL deallocate_potential(qs_kind_set(ikind)%tnadd_potential)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%gth_potential)) THEN
               CALL deallocate_potential(qs_kind_set(ikind)%gth_potential)
            ENDIF
            IF (ASSOCIATED(qs_kind_set(ikind)%se_parameter)) THEN
               CALL semi_empirical_release(qs_kind_set(ikind)%se_parameter)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%dftb_parameter)) THEN
               CALL deallocate_dftb_atom_param(qs_kind_set(ikind)%dftb_parameter)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%scptb_parameter)) THEN
               CALL deallocate_scptb_parameter(qs_kind_set(ikind)%scptb_parameter)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%soft_basis_set) .AND. &
                qs_kind_set(ikind)%paw_atom) THEN
               CALL deallocate_gto_basis_set(qs_kind_set(ikind)%soft_basis_set)
            ELSEIF (ASSOCIATED(qs_kind_set(ikind)%soft_basis_set) .AND. &
                    (.NOT. qs_kind_set(ikind)%paw_atom)) THEN
               NULLIFY (qs_kind_set(ikind)%soft_basis_set)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%paw_proj_set)) THEN
               CALL deallocate_paw_proj_set(qs_kind_set(ikind)%paw_proj_set)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%harmonics)) THEN
               CALL deallocate_harmonics_atom(qs_kind_set(ikind)%harmonics)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%grid_atom)) THEN
               CALL deallocate_grid_atom(qs_kind_set(ikind)%grid_atom)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%elec_conf)) THEN
               DEALLOCATE (qs_kind_set(ikind)%elec_conf)
            END IF

            IF (ASSOCIATED(qs_kind_set(ikind)%dft_plus_u)) THEN
               IF (ASSOCIATED(qs_kind_set(ikind)%dft_plus_u%orbitals)) THEN
                  DEALLOCATE (qs_kind_set(ikind)%dft_plus_u%orbitals)
               END IF
               DEALLOCATE (qs_kind_set(ikind)%dft_plus_u)
            END IF

            IF (ASSOCIATED(qs_kind_set(ikind)%dispersion)) THEN
               DEALLOCATE (qs_kind_set(ikind)%dispersion)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%addel)) THEN
               DEALLOCATE (qs_kind_set(ikind)%addel)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%naddel)) THEN
               DEALLOCATE (qs_kind_set(ikind)%naddel)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%laddel)) THEN
               DEALLOCATE (qs_kind_set(ikind)%laddel)
            END IF
            IF (ASSOCIATED(qs_kind_set(ikind)%reltmat)) THEN
               DEALLOCATE (qs_kind_set(ikind)%reltmat)
            END IF

            CALL remove_basis_set_container(qs_kind_set(ikind)%basis_sets)

         END DO
         DEALLOCATE (qs_kind_set)
      ELSE
         CALL cp_abort(cp__l("qs_kind_types.F",277), &
                       "The pointer qs_kind_set is not associated and "// &
                       "cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_qs_kind_set

! **************************************************************************************************
!> \brief Get attributes of an atomic kind.
!> \param qs_kind ...
!> \param basis_set ...
!> \param basis_type ...
!> \param ncgf ...
!> \param nsgf ...
!> \param all_potential ...
!> \param tnadd_potential ...
!> \param gth_potential ...
!> \param se_parameter ...
!> \param dftb_parameter ...
!> \param scptb_parameter ...
!> \param dftb3_param ...
!> \param zeff ...
!> \param elec_conf ...
!> \param mao ...
!> \param lmax_dftb ...
!> \param alpha_core_charge ...
!> \param ccore_charge ...
!> \param core_charge ...
!> \param core_charge_radius ...
!> \param soft_basis_set ...
!> \param hard_basis_set ...
!> \param paw_proj_set ...
!> \param softb ...
!> \param paw_atom ...
!> \param hard_radius ...
!> \param hard0_radius ...
!> \param max_rad_local ...
!> \param gpw_type_forced ...
!> \param harmonics ...
!> \param max_iso_not0 ...
!> \param max_s_harm ...
!> \param grid_atom ...
!> \param ngrid_ang ...
!> \param ngrid_rad ...
!> \param lmax_rho0 ...
!> \param dft_plus_u_atom ...
!> \param l_of_dft_plus_u ...
!> \param u_minus_j ...
!> \param dispersion ...
!> \param bs_occupation ...
!> \param no_optimize ...
!> \param addel ...
!> \param laddel ...
!> \param naddel ...
!> \param orbitals ...
!> \param max_scf ...
!> \param eps_scf ...
!> \param smear ...
!> \param u_ramping ...
!> \param u_minus_j_target ...
!> \param eps_u_ramping ...
!> \param init_u_ramping_each_scf ...
!> \param reltmat ...
!> \param ghost ...
!> \param name ...
!> \param element_symbol ...
!> \param pao_basis_size ...
!> \param pao_pot_maxl ...
!> \param pao_pot_max_projector ...
!> \param pao_pot_beta ...
!> \param pao_pot_weight ...
!> \param pao_pot_neighbors ...
! **************************************************************************************************
   SUBROUTINE get_qs_kind(qs_kind, &
                          basis_set, basis_type, ncgf, nsgf, &
                          all_potential, tnadd_potential, gth_potential, &
                          se_parameter, dftb_parameter, scptb_parameter, &
                          dftb3_param, zeff, elec_conf, mao, lmax_dftb, &
                          alpha_core_charge, ccore_charge, core_charge, core_charge_radius, &
                          soft_basis_set, hard_basis_set, paw_proj_set, softb, &
                          paw_atom, hard_radius, hard0_radius, max_rad_local, &
                          gpw_type_forced, harmonics, max_iso_not0, max_s_harm, grid_atom, &
                          ngrid_ang, ngrid_rad, lmax_rho0, &
                          dft_plus_u_atom, l_of_dft_plus_u, u_minus_j, dispersion, &
                          bs_occupation, no_optimize, addel, laddel, naddel, orbitals, &
                          max_scf, eps_scf, smear, u_ramping, u_minus_j_target, eps_u_ramping, &
                          init_u_ramping_each_scf, reltmat, ghost, name, element_symbol, &
                          pao_basis_size, pao_pot_maxl, pao_pot_max_projector, pao_pot_beta, pao_pot_weight, &
                          pao_pot_neighbors)

      TYPE(qs_kind_type)                                 :: qs_kind
      TYPE(gto_basis_set_type), OPTIONAL, POINTER        :: basis_set
      CHARACTER(len=*), OPTIONAL                         :: basis_type
      INTEGER, INTENT(OUT), OPTIONAL                     :: ncgf, nsgf
      TYPE(all_potential_type), OPTIONAL, POINTER        :: all_potential
      TYPE(local_potential_type), OPTIONAL, POINTER      :: tnadd_potential
      TYPE(gth_potential_type), OPTIONAL, POINTER        :: gth_potential
      TYPE(semi_empirical_type), OPTIONAL, POINTER       :: se_parameter
      TYPE(qs_dftb_atom_type), OPTIONAL, POINTER         :: dftb_parameter
      TYPE(scptb_parameter_type), OPTIONAL, POINTER      :: scptb_parameter
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: dftb3_param, zeff
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: elec_conf
      INTEGER, INTENT(OUT), OPTIONAL                     :: mao, lmax_dftb
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: alpha_core_charge, ccore_charge, &
                                                            core_charge, core_charge_radius
      TYPE(gto_basis_set_type), OPTIONAL, POINTER        :: soft_basis_set, hard_basis_set
      TYPE(paw_proj_set_type), OPTIONAL, POINTER         :: paw_proj_set
      LOGICAL, INTENT(IN), OPTIONAL                      :: softb
      LOGICAL, INTENT(OUT), OPTIONAL                     :: paw_atom
      REAL(dp), INTENT(OUT), OPTIONAL                    :: hard_radius, hard0_radius, max_rad_local
      LOGICAL, INTENT(OUT), OPTIONAL                     :: gpw_type_forced
      TYPE(harmonics_atom_type), OPTIONAL, POINTER       :: harmonics
      INTEGER, INTENT(OUT), OPTIONAL                     :: max_iso_not0, max_s_harm
      TYPE(grid_atom_type), OPTIONAL, POINTER            :: grid_atom
      INTEGER, INTENT(OUT), OPTIONAL                     :: ngrid_ang, ngrid_rad, lmax_rho0
      LOGICAL, INTENT(OUT), OPTIONAL                     :: dft_plus_u_atom
      INTEGER, INTENT(OUT), OPTIONAL                     :: l_of_dft_plus_u
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: u_minus_j
      TYPE(qs_atom_dispersion_type), OPTIONAL, POINTER   :: dispersion
      LOGICAL, INTENT(OUT), OPTIONAL                     :: bs_occupation, no_optimize
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: addel, laddel, naddel
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: orbitals
      INTEGER, OPTIONAL                                  :: max_scf
      REAL(KIND=dp), OPTIONAL                            :: eps_scf
      LOGICAL, OPTIONAL                                  :: smear
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: u_ramping, u_minus_j_target, &
                                                            eps_u_ramping
      LOGICAL, OPTIONAL                                  :: init_u_ramping_each_scf
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: reltmat
      LOGICAL, OPTIONAL                                  :: ghost
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      CHARACTER(LEN=2), INTENT(OUT), OPTIONAL            :: element_symbol
      INTEGER, INTENT(OUT), OPTIONAL                     :: pao_basis_size, pao_pot_maxl, &
                                                            pao_pot_max_projector
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: pao_pot_beta, pao_pot_weight
      INTEGER, INTENT(OUT), OPTIONAL                     :: pao_pot_neighbors

      CHARACTER(len=*), PARAMETER :: routineN = 'get_qs_kind', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_basis_type
      INTEGER                                            :: l
      LOGICAL                                            :: softb_local
      TYPE(gto_basis_set_type), POINTER                  :: tmp_basis_set

! Retrieve basis set from the kind container

      softb_local = .FALSE.
      IF (PRESENT(softb)) softb_local = softb

      IF (PRESENT(basis_type)) THEN
         my_basis_type = basis_type
      ELSE
         my_basis_type = "ORB"
      END IF

      IF (PRESENT(basis_set)) THEN
         IF (softb_local) THEN
            basis_set => qs_kind%soft_basis_set
         ELSE
            CALL get_basis_from_container(qs_kind%basis_sets, basis_set=basis_set, &
                                          basis_type=my_basis_type)
         END IF
      END IF

      IF (PRESENT(ncgf)) THEN
         CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis_set, &
                                       basis_type=my_basis_type)
         IF (ASSOCIATED(tmp_basis_set)) THEN
            CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, ncgf=ncgf)
         ELSE IF (ASSOCIATED(qs_kind%dftb_parameter)) THEN
            l = qs_kind%dftb_parameter%lmax
            ncgf = ((l+1)*(l+2)*(l+3))/6
         ELSE
            ncgf = 0
         END IF
      END IF

      IF (PRESENT(nsgf)) THEN
         CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis_set, &
                                       basis_type=my_basis_type)
         IF (ASSOCIATED(tmp_basis_set)) THEN
            CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, nsgf=nsgf)
         ELSE IF (ASSOCIATED(qs_kind%dftb_parameter)) THEN
            nsgf = qs_kind%dftb_parameter%natorb
         ELSE
            nsgf = 0
         END IF
      END IF

      IF (PRESENT(all_potential)) all_potential => qs_kind%all_potential
      IF (PRESENT(tnadd_potential)) tnadd_potential => qs_kind%tnadd_potential
      IF (PRESENT(gth_potential)) gth_potential => qs_kind%gth_potential
      IF (PRESENT(se_parameter)) se_parameter => qs_kind%se_parameter
      IF (PRESENT(dftb_parameter)) dftb_parameter => qs_kind%dftb_parameter
      IF (PRESENT(scptb_parameter)) scptb_parameter => qs_kind%scptb_parameter

      IF (PRESENT(element_symbol)) element_symbol = qs_kind%element_symbol
      IF (PRESENT(name)) name = qs_kind%name
      IF (PRESENT(dftb3_param)) dftb3_param = qs_kind%dudq_dftb3
      IF (PRESENT(elec_conf)) elec_conf => qs_kind%elec_conf
      IF (PRESENT(alpha_core_charge)) THEN
         IF (ASSOCIATED(qs_kind%all_potential)) THEN
            CALL get_potential(potential=qs_kind%all_potential, &
                               alpha_core_charge=alpha_core_charge)
         ELSE IF (ASSOCIATED(qs_kind%gth_potential)) THEN
            CALL get_potential(potential=qs_kind%gth_potential, &
                               alpha_core_charge=alpha_core_charge)
         ELSE
            alpha_core_charge = 1.0_dp
         END IF
      END IF
      IF (PRESENT(ccore_charge)) THEN
         IF (ASSOCIATED(qs_kind%all_potential)) THEN
            CALL get_potential(potential=qs_kind%all_potential, &
                               ccore_charge=ccore_charge)
         ELSE IF (ASSOCIATED(qs_kind%gth_potential)) THEN
            CALL get_potential(potential=qs_kind%gth_potential, &
                               ccore_charge=ccore_charge)
         ELSE
            ccore_charge = 0.0_dp
         END IF
      END IF
      IF (PRESENT(core_charge_radius)) THEN
         IF (ASSOCIATED(qs_kind%all_potential)) THEN
            CALL get_potential(potential=qs_kind%all_potential, &
                               core_charge_radius=core_charge_radius)
         ELSE IF (ASSOCIATED(qs_kind%gth_potential)) THEN
            CALL get_potential(potential=qs_kind%gth_potential, &
                               core_charge_radius=core_charge_radius)
         ELSE
            core_charge_radius = 0.0_dp
         END IF
      END IF
      IF (PRESENT(core_charge)) THEN
         IF (ASSOCIATED(qs_kind%all_potential)) THEN
            CALL get_potential(potential=qs_kind%all_potential, &
                               zeff=core_charge)
         ELSE IF (ASSOCIATED(qs_kind%gth_potential)) THEN
            CALL get_potential(potential=qs_kind%gth_potential, &
                               zeff=core_charge)
         ELSE
            core_charge = 0.0_dp
         END IF
      END IF

      IF (PRESENT(zeff)) THEN
         IF (ASSOCIATED(qs_kind%all_potential)) THEN
            CALL get_potential(potential=qs_kind%all_potential, zeff=zeff)
         ELSE IF (ASSOCIATED(qs_kind%gth_potential)) THEN
            CALL get_potential(potential=qs_kind%gth_potential, zeff=zeff)
         ELSE
            zeff = 0.0_dp
         END IF
      END IF

      IF (PRESENT(soft_basis_set)) soft_basis_set => qs_kind%soft_basis_set
      IF (PRESENT(hard_basis_set)) hard_basis_set => qs_kind%hard_basis_set

      IF (PRESENT(paw_proj_set)) paw_proj_set => qs_kind%paw_proj_set
      IF (PRESENT(paw_atom)) paw_atom = qs_kind%paw_atom
      IF (PRESENT(gpw_type_forced)) gpw_type_forced = qs_kind%gpw_type_forced
      IF (PRESENT(hard_radius)) hard_radius = qs_kind%hard_radius
      IF (PRESENT(hard0_radius)) hard0_radius = qs_kind%hard0_radius
      IF (PRESENT(max_rad_local)) max_rad_local = qs_kind%max_rad_local
      IF (PRESENT(harmonics)) harmonics => qs_kind%harmonics
      IF (PRESENT(max_s_harm)) THEN
         IF (ASSOCIATED(qs_kind%harmonics)) THEN
            max_s_harm = qs_kind%harmonics%max_s_harm
         ELSE
            max_s_harm = 0
         END IF
      END IF
      IF (PRESENT(max_iso_not0)) THEN
         IF (ASSOCIATED(qs_kind%harmonics)) THEN
            max_iso_not0 = qs_kind%harmonics%max_iso_not0
         ELSE
            max_iso_not0 = 0
         END IF
      END IF
      IF (PRESENT(grid_atom)) grid_atom => qs_kind%grid_atom
      IF (PRESENT(ngrid_ang)) ngrid_ang = qs_kind%ngrid_ang
      IF (PRESENT(ngrid_rad)) ngrid_rad = qs_kind%ngrid_rad
      IF (PRESENT(lmax_rho0)) lmax_rho0 = qs_kind%lmax_rho0
      IF (PRESENT(ghost)) ghost = qs_kind%ghost
      IF (PRESENT(dft_plus_u_atom)) dft_plus_u_atom = ASSOCIATED(qs_kind%dft_plus_u)
      IF (PRESENT(l_of_dft_plus_u)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            l_of_dft_plus_u = qs_kind%dft_plus_u%l
         ELSE
            l_of_dft_plus_u = -1
         END IF
      END IF
      IF (PRESENT(u_minus_j)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            u_minus_j = qs_kind%dft_plus_u%u_minus_j
         ELSE
            u_minus_j = 0.0_dp
         END IF
      END IF
      IF (PRESENT(u_minus_j_target)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            u_minus_j_target = qs_kind%dft_plus_u%u_minus_j_target
         ELSE
            u_minus_j_target = 0.0_dp
         END IF
      END IF
      IF (PRESENT(init_u_ramping_each_scf)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            init_u_ramping_each_scf = qs_kind%dft_plus_u%init_u_ramping_each_scf
         ELSE
            init_u_ramping_each_scf = .FALSE.
         END IF
      END IF
      IF (PRESENT(u_ramping)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            u_ramping = qs_kind%dft_plus_u%u_ramping
         ELSE
            u_ramping = 0.0_dp
         END IF
      END IF
      IF (PRESENT(eps_u_ramping)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            eps_u_ramping = qs_kind%dft_plus_u%eps_u_ramping
         ELSE
            eps_u_ramping = 1.0E-5_dp
         END IF
      END IF
      IF (PRESENT(orbitals)) THEN
         NULLIFY (orbitals)
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            IF (ASSOCIATED(qs_kind%dft_plus_u%orbitals)) THEN
               orbitals => qs_kind%dft_plus_u%orbitals
            END IF
         END IF
      END IF
      IF (PRESENT(eps_scf)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            eps_scf = qs_kind%dft_plus_u%eps_scf
         ELSE
            eps_scf = 1.0E30_dp
         END IF
      END IF
      IF (PRESENT(max_scf)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            max_scf = qs_kind%dft_plus_u%max_scf
         ELSE
            max_scf = -1
         END IF
      END IF
      IF (PRESENT(smear)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            smear = qs_kind%dft_plus_u%smear
         ELSE
            smear = .FALSE.
         END IF
      END IF
      IF (PRESENT(dispersion)) dispersion => qs_kind%dispersion
      IF (PRESENT(bs_occupation)) bs_occupation = qs_kind%bs_occupation
      IF (PRESENT(addel)) addel => qs_kind%addel
      IF (PRESENT(laddel)) laddel => qs_kind%laddel
      IF (PRESENT(naddel)) naddel => qs_kind%naddel

      IF (PRESENT(no_optimize)) no_optimize = qs_kind%no_optimize

      IF (PRESENT(reltmat)) reltmat => qs_kind%reltmat

      IF (PRESENT(mao)) mao = qs_kind%mao

      IF (PRESENT(lmax_dftb)) lmax_dftb = qs_kind%lmax_dftb

      IF (PRESENT(pao_basis_size)) pao_basis_size = qs_kind%pao_basis_size
      IF (PRESENT(pao_pot_maxl)) pao_pot_maxl = qs_kind%pao_pot_maxl
      IF (PRESENT(pao_pot_max_projector)) pao_pot_max_projector = qs_kind%pao_pot_max_projector
      IF (PRESENT(pao_pot_neighbors)) pao_pot_neighbors = qs_kind%pao_pot_neighbors
      IF (PRESENT(pao_pot_beta)) pao_pot_beta = qs_kind%pao_pot_beta
      IF (PRESENT(pao_pot_weight)) pao_pot_weight = qs_kind%pao_pot_weight
   END SUBROUTINE get_qs_kind

! **************************************************************************************************
!> \brief Get attributes of an atomic kind set.
!> \param qs_kind_set ...
!> \param all_potential_present ...
!> \param tnadd_potential_present ...
!> \param gth_potential_present ...
!> \param paw_atom_present ...
!> \param dft_plus_u_atom_present ...
!> \param maxcgf ...
!> \param maxsgf ...
!> \param maxco ...
!> \param maxco_proj ...
!> \param maxgtops ...
!> \param maxlgto ...
!> \param maxlprj ...
!> \param maxnset ...
!> \param maxsgf_set ...
!> \param ncgf ...
!> \param npgf ...
!> \param nset ...
!> \param nsgf ...
!> \param nshell ...
!> \param maxpol ...
!> \param maxlppl ...
!> \param maxlppnl ...
!> \param maxppnl ...
!> \param nelectron ...
!> \param maxder ...
!> \param max_ngrid_rad ...
!> \param max_sph_harm ...
!> \param maxg_iso_not0 ...
!> \param lmax_rho0 ...
!> \param basis_type ...
! **************************************************************************************************
   SUBROUTINE get_qs_kind_set(qs_kind_set, &
                              all_potential_present, tnadd_potential_present, gth_potential_present, &
                              paw_atom_present, dft_plus_u_atom_present, &
                              maxcgf, maxsgf, maxco, maxco_proj, maxgtops, maxlgto, maxlprj, maxnset, maxsgf_set, &
                              ncgf, npgf, nset, nsgf, nshell, maxpol, maxlppl, maxlppnl, maxppnl, &
                              nelectron, maxder, max_ngrid_rad, max_sph_harm, maxg_iso_not0, lmax_rho0, &
                              basis_type)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      LOGICAL, INTENT(OUT), OPTIONAL :: all_potential_present, tnadd_potential_present, &
         gth_potential_present, paw_atom_present, dft_plus_u_atom_present
      INTEGER, INTENT(OUT), OPTIONAL :: maxcgf, maxsgf, maxco, maxco_proj, maxgtops, maxlgto, &
         maxlprj, maxnset, maxsgf_set, ncgf, npgf, nset, nsgf, nshell, maxpol, maxlppl, maxlppnl, &
         maxppnl, nelectron
      INTEGER, INTENT(IN), OPTIONAL                      :: maxder
      INTEGER, INTENT(OUT), OPTIONAL                     :: max_ngrid_rad, max_sph_harm, &
                                                            maxg_iso_not0, lmax_rho0
      CHARACTER(len=*), OPTIONAL                         :: basis_type

      CHARACTER(len=*), PARAMETER :: routineN = 'get_qs_kind_set', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: my_basis_type
      INTEGER                                            :: ikind, imax, lmax_rho0_kind, &
                                                            max_iso_not0, max_s_harm, n, &
                                                            ngrid_rad, nkind
      LOGICAL                                            :: dft_plus_u_atom, paw_atom
      REAL(KIND=dp)                                      :: zeff, zeff_correction
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_type), POINTER                  :: tmp_basis_set
      TYPE(local_potential_type), POINTER                :: tnadd_potential
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      IF (PRESENT(basis_type)) THEN
         my_basis_type = basis_type
      ELSE
         my_basis_type = "ORB"
      END IF

      IF (ASSOCIATED(qs_kind_set)) THEN

         IF (PRESENT(maxcgf)) maxcgf = 0
         IF (PRESENT(maxco)) maxco = 0
         IF (PRESENT(maxco_proj)) maxco_proj = 0
         IF (PRESENT(maxg_iso_not0)) maxg_iso_not0 = 0
         IF (PRESENT(maxgtops)) maxgtops = 0
         IF (PRESENT(maxlgto)) maxlgto = -1
         IF (PRESENT(maxlppl)) maxlppl = -1
         IF (PRESENT(maxlppnl)) maxlppnl = -1
         IF (PRESENT(maxpol)) maxpol = -1
         IF (PRESENT(maxlprj)) maxlprj = -1
         IF (PRESENT(maxnset)) maxnset = 0
         IF (PRESENT(maxppnl)) maxppnl = 0
         IF (PRESENT(maxsgf)) maxsgf = 0
         IF (PRESENT(maxsgf_set)) maxsgf_set = 0
         IF (PRESENT(ncgf)) ncgf = 0
         IF (PRESENT(nelectron)) nelectron = 0
         IF (PRESENT(npgf)) npgf = 0
         IF (PRESENT(nset)) nset = 0
         IF (PRESENT(nsgf)) nsgf = 0
         IF (PRESENT(nshell)) nshell = 0
         IF (PRESENT(all_potential_present)) all_potential_present = .FALSE.
         IF (PRESENT(tnadd_potential_present)) tnadd_potential_present = .FALSE.
         IF (PRESENT(gth_potential_present)) gth_potential_present = .FALSE.
         IF (PRESENT(paw_atom_present)) paw_atom_present = .FALSE.
         IF (PRESENT(max_ngrid_rad)) max_ngrid_rad = 0
         IF (PRESENT(max_sph_harm)) max_sph_harm = 0
         IF (PRESENT(lmax_rho0)) lmax_rho0 = 0

         nkind = SIZE(qs_kind_set)
         DO ikind = 1, nkind
            qs_kind => qs_kind_set(ikind)
            CALL get_qs_kind(qs_kind=qs_kind, &
                             all_potential=all_potential, &
                             tnadd_potential=tnadd_potential, &
                             gth_potential=gth_potential, &
                             paw_proj_set=paw_proj_set, &
                             dftb_parameter=dftb_parameter, &
                             ngrid_rad=ngrid_rad, &
                             max_s_harm=max_s_harm, &
                             max_iso_not0=max_iso_not0, &
                             paw_atom=paw_atom, &
                             dft_plus_u_atom=dft_plus_u_atom, &
                             lmax_rho0=lmax_rho0_kind)

            IF (PRESENT(maxlppl) .AND. ASSOCIATED(gth_potential)) THEN
               CALL get_potential(potential=gth_potential, nexp_ppl=n)
               maxlppl = MAX(maxlppl, 2*(n-1))
            END IF

            IF (PRESENT(maxlppnl) .AND. ASSOCIATED(gth_potential)) THEN
               CALL get_potential(potential=gth_potential, lprj_ppnl_max=imax)
               maxlppnl = MAX(maxlppnl, imax)
            END IF

            IF (PRESENT(maxpol) .AND. ASSOCIATED(tnadd_potential)) THEN
               CALL get_potential(potential=tnadd_potential, npol=n)
               maxpol = MAX(maxpol, 2*(n-1))
            END IF

            IF (PRESENT(maxco_proj) .AND. ASSOCIATED(paw_proj_set)) THEN
               CALL get_paw_proj_set(paw_proj_set=paw_proj_set, ncgauprj=imax)
               maxco_proj = MAX(maxco_proj, imax)
            ENDIF

            IF (PRESENT(maxlprj) .AND. ASSOCIATED(paw_proj_set)) THEN
               CALL get_paw_proj_set(paw_proj_set=paw_proj_set, maxl=imax)
               maxlprj = MAX(maxlprj, imax)
            ENDIF

            IF (PRESENT(maxppnl) .AND. ASSOCIATED(gth_potential)) THEN
               CALL get_potential(potential=gth_potential, nppnl=imax)
               maxppnl = MAX(maxppnl, imax)
            END IF

            CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis_set, &
                                          basis_type=my_basis_type)

            IF (PRESENT(maxcgf)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, ncgf=imax)
                  maxcgf = MAX(maxcgf, imax)
               ELSE IF (ASSOCIATED(qs_kind%dftb_parameter)) THEN
                  CALL get_dftb_atom_param(dftb_parameter=dftb_parameter, lmax=imax)
                  imax = ((imax+1)*(imax+2)*(imax+3))/6
                  maxcgf = MAX(maxcgf, imax)
               END IF
            END IF

            IF (PRESENT(maxco)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  IF (PRESENT(maxder)) THEN
                     CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, &
                                            maxco=imax, maxder=maxder)
                  ELSE
                     CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, maxco=imax)
                  END IF
                  maxco = MAX(maxco, imax)
               END IF
               IF (ASSOCIATED(gth_potential)) THEN
                  CALL get_potential(potential=gth_potential, lprj_ppnl_max=imax)
                  maxco = MAX(maxco, ncoset(imax))
               END IF
            END IF

            IF (PRESENT(maxgtops)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, maxso=imax, nset=n)
                  maxgtops = MAX(maxgtops, n*imax)
               END IF
            END IF

            IF (PRESENT(maxlgto)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, maxl=imax)
                  maxlgto = MAX(maxlgto, imax)
               ELSE IF (ASSOCIATED(qs_kind%dftb_parameter)) THEN
                  CALL get_dftb_atom_param(dftb_parameter=dftb_parameter, lmax=imax)
                  maxlgto = MAX(maxlgto, imax)
               END IF
            END IF

            IF (PRESENT(maxnset)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, nset=n)
                  maxnset = MAX(maxnset, n)
               END IF
            END IF

            IF (PRESENT(maxsgf)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, nsgf=imax)
                  maxsgf = MAX(maxsgf, imax)
               END IF
            END IF

            IF (PRESENT(maxsgf_set)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, maxsgf_set=imax)
                  maxsgf_set = MAX(maxsgf_set, imax)
               END IF
            END IF

            IF (PRESENT(ncgf)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, ncgf=n)
                  ncgf = ncgf+n*qs_kind_set(ikind)%natom
               ELSE IF (ASSOCIATED(qs_kind%dftb_parameter)) THEN
                  CALL get_dftb_atom_param(dftb_parameter=dftb_parameter, lmax=imax)
                  n = ((imax+1)*(imax+2)*(imax+3))/6
                  ncgf = ncgf+n*qs_kind_set(ikind)%natom
               END IF
            END IF

            IF (PRESENT(npgf)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, npgf_sum=n)
                  npgf = npgf+n*qs_kind_set(ikind)%natom
               END IF
            END IF

            IF (PRESENT(nset)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, nset=n)
                  nset = nset+n*qs_kind_set(ikind)%natom
               END IF
            END IF

            IF (PRESENT(nsgf)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, nsgf=n)
                  nsgf = nsgf+n*qs_kind_set(ikind)%natom
               ELSE IF (ASSOCIATED(qs_kind%dftb_parameter)) THEN
                  CALL get_dftb_atom_param(dftb_parameter=dftb_parameter, natorb=n)
                  nsgf = nsgf+n*qs_kind_set(ikind)%natom
               END IF
            END IF

            IF (PRESENT(nshell)) THEN
               IF (ASSOCIATED(tmp_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=tmp_basis_set, nshell_sum=n)
                  nshell = nshell+n*qs_kind_set(ikind)%natom
               ELSE IF (ASSOCIATED(qs_kind%dftb_parameter)) THEN
                  CALL get_dftb_atom_param(dftb_parameter=dftb_parameter, lmax=n)
                  nshell = nshell+(n+1)*qs_kind_set(ikind)%natom
               END IF
            END IF

            IF (PRESENT(nelectron)) THEN
               IF (ASSOCIATED(qs_kind%all_potential)) THEN
                  CALL get_potential(potential=qs_kind%all_potential, &
                                     zeff=zeff, zeff_correction=zeff_correction)
               ELSE IF (ASSOCIATED(qs_kind%gth_potential)) THEN
                  CALL get_potential(potential=qs_kind%gth_potential, &
                                     zeff=zeff, zeff_correction=zeff_correction)
               ELSE
                  zeff = 0.0_dp
                  zeff_correction = 0.0_dp
               END IF
               nelectron = nelectron+qs_kind_set(ikind)%natom*NINT(zeff-zeff_correction)
            END IF

            IF (PRESENT(all_potential_present)) THEN
               IF (ASSOCIATED(all_potential)) THEN
                  all_potential_present = .TRUE.
               END IF
            END IF

            IF (PRESENT(tnadd_potential_present)) THEN
               IF (ASSOCIATED(tnadd_potential)) THEN
                  tnadd_potential_present = .TRUE.
               END IF
            END IF

            IF (PRESENT(gth_potential_present)) THEN
               IF (ASSOCIATED(gth_potential)) THEN
                  gth_potential_present = .TRUE.
               END IF
            END IF

            IF (PRESENT(paw_atom_present)) THEN
               IF (paw_atom) THEN
                  paw_atom_present = .TRUE.
               END IF
            END IF

            IF (PRESENT(dft_plus_u_atom_present)) THEN
               IF (dft_plus_u_atom) THEN
                  dft_plus_u_atom_present = .TRUE.
               END IF
            END IF

            IF (PRESENT(max_ngrid_rad)) THEN
               max_ngrid_rad = MAX(max_ngrid_rad, ngrid_rad)
            ENDIF

            IF (PRESENT(max_sph_harm)) THEN
               max_sph_harm = MAX(max_sph_harm, max_s_harm)
            ENDIF

            IF (PRESENT(maxg_iso_not0)) THEN
               maxg_iso_not0 = MAX(maxg_iso_not0, max_iso_not0)
            END IF

            IF (PRESENT(lmax_rho0)) THEN
               lmax_rho0 = MAX(lmax_rho0, lmax_rho0_kind)
            END IF

         END DO
      ELSE
         CALL cp__b("qs_kind_types.F",983,"The pointer qs_kind_set is not associated")
      END IF

   END SUBROUTINE get_qs_kind_set

! **************************************************************************************************
!> \brief Initialise an atomic kind data set.
!> \param qs_kind ...
!> \author Creation (11.01.2002,MK)
!>                20.09.2002 adapted for pol/kg use, gtb
! **************************************************************************************************
   SUBROUTINE init_qs_kind(qs_kind)
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CHARACTER(len=*), PARAMETER :: routineN = 'init_qs_kind', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: basis_type
      INTEGER                                            :: handle, i
      TYPE(gto_basis_set_type), POINTER                  :: tmp_basis_set

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(qs_kind)))CALL cp__a("qs_kind_types.F",1005)

      IF (ASSOCIATED(qs_kind%gth_potential)) THEN
         CALL init_potential(qs_kind%gth_potential)
      END IF

      DO i = 1, SIZE(qs_kind%basis_sets, 1)
         NULLIFY (tmp_basis_set)
         CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis_set, &
                                       inumbas=i, basis_type=basis_type)
         IF (basis_type == "") CYCLE
         IF (basis_type == "AUX") THEN
            IF (tmp_basis_set%norm_type < 0) tmp_basis_set%norm_type = 1
            CALL init_aux_basis_set(tmp_basis_set)
         ELSE
            IF (tmp_basis_set%norm_type < 0) tmp_basis_set%norm_type = 2
            CALL init_orb_basis_set(tmp_basis_set)
         END IF
      END DO

      CALL timestop(handle)

   END SUBROUTINE init_qs_kind

! **************************************************************************************************
!> \brief Initialise an atomic kind set data set.
!> \param qs_kind_set ...
!> \author - Creation (17.01.2002,MK)
!>      - 20.09.2002 para_env passed (gt)
! **************************************************************************************************
   SUBROUTINE init_qs_kind_set(qs_kind_set)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'init_qs_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(qs_kind_set)) THEN
         CALL cp__b("qs_kind_types.F",1048,"init_qs_kind_set: The pointer qs_kind_set is not associated")
      ENDIF

      DO ikind = 1, SIZE(qs_kind_set)
         qs_kind => qs_kind_set(ikind)
         CALL init_qs_kind(qs_kind)
      END DO

      CALL timestop(handle)

   END SUBROUTINE init_qs_kind_set

! **************************************************************************************************
!> \brief ...
!> \param qs_kind_set ...
!> \param qs_control ...
!> \param force_env_section ...
! **************************************************************************************************
   SUBROUTINE init_gapw_basis_set(qs_kind_set, qs_control, force_env_section)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_control_type), POINTER                     :: qs_control
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'init_gapw_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nkind
      LOGICAL                                            :: gpw, paw_atom
      REAL(dp)                                           :: max_rad_local_type, rc
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      IF (ASSOCIATED(qs_kind_set)) THEN

         qs_control%gapw_control%non_paw_atoms = .FALSE.
         nkind = SIZE(qs_kind_set)

         DO ikind = 1, nkind

            qs_kind => qs_kind_set(ikind)

            CALL allocate_gto_basis_set(qs_kind%soft_basis_set)

            ! The hard_basis points to the orb_basis
            CALL get_qs_kind(qs_kind=qs_kind, basis_set=orb_basis)
            qs_kind%hard_basis_set => orb_basis

            CALL get_qs_kind(qs_kind=qs_kind, hard_radius=rc, &
                             max_rad_local=max_rad_local_type, gpw_type_forced=gpw)

            CALL create_soft_basis(orb_basis, qs_kind%soft_basis_set, &
                                   qs_control%gapw_control%eps_fit, rc, paw_atom, &
                                   qs_control%gapw_control%force_paw, gpw)

            CALL set_qs_kind(qs_kind=qs_kind, paw_atom=paw_atom)

            IF (paw_atom) THEN
               CALL allocate_paw_proj_set(qs_kind%paw_proj_set)
               CALL get_qs_kind(qs_kind=qs_kind, paw_proj_set=paw_proj)

               CALL projectors(paw_proj, orb_basis, rc, qs_control, max_rad_local_type, &
                               force_env_section)
            ELSE
               qs_control%gapw_control%non_paw_atoms = .TRUE.
            END IF

            ! grid_atom and harmonics are allocated even if NOT PAW_ATOM
            NULLIFY (qs_kind%grid_atom, qs_kind%harmonics)
            CALL allocate_grid_atom(qs_kind%grid_atom)
            CALL allocate_harmonics_atom(qs_kind%harmonics)
         END DO

         IF (qs_control%gapw_control%non_paw_atoms) THEN
            qs_control%gapw_control%nopaw_as_gpw = .TRUE.
         ELSE
            qs_control%gapw_control%nopaw_as_gpw = .FALSE.
         END IF
      ELSE
         CALL cp__b("qs_kind_types.F",1128,"The pointer qs_kind_set is not associated")
      END IF

   END SUBROUTINE init_gapw_basis_set

! **************************************************************************************************
!> \brief Read an atomic kind data set from the input file.
!> \param qs_kind ...
!> \param kind_section ...
!> \param para_env ...
!> \param force_env_section ...
!> \param no_fail ...
!> \param method_id ...
!> \par History
!>      - Creation (09.02.2002,MK)
!>      - 20.09.2002,gt: adapted for POL/KG use (elp_potential)
!>      - 05.03.2010: split elp_potential into fist_potential and kg_potential
! **************************************************************************************************
   SUBROUTINE read_qs_kind(qs_kind, kind_section, para_env, force_env_section, &
                           no_fail, method_id)

      TYPE(qs_kind_type), INTENT(INOUT)                  :: qs_kind
      TYPE(section_vals_type), POINTER                   :: kind_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: force_env_section
      LOGICAL, INTENT(IN)                                :: no_fail
      INTEGER, INTENT(IN)                                :: method_id

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_qs_kind', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: maxbas = 20

      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=default_string_length) :: akind_name, aux_basis_set_name, &
         aux_fit_basis_set_name, basis_type, keyword, kgpot_name, kgpot_type, lri_basis_set_name, &
         orb_basis_set_name, potential_name, potential_type, ri_aux_basis_set_name, tmp
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      CHARACTER(LEN=default_string_length), &
         DIMENSION(maxbas)                               :: basis_set_name, basis_set_type
      INTEGER                                            :: handle, i, i_rep, ipos, j, k_rep, l, m, &
                                                            n_rep, nb_rep, nexp, ngauss, &
                                                            norbitals, z
      INTEGER, DIMENSION(:), POINTER                     :: add_el, elec_conf, orbitals
      LOGICAL :: check, explicit, explicit_basis, explicit_kgpot, explicit_potential, nobasis, &
         section_enabled, subsection_enabled
      REAL(KIND=dp)                                      :: zeff_correction
      TYPE(gto_basis_set_type), POINTER                  :: sup_basis_set, tmp_basis_set
      TYPE(section_vals_type), POINTER :: basis_section, bs_section, dft_plus_u_section, &
         dft_section, enforce_occupation_section, kgpot_section, pao_pot_section, &
         potential_section, spin_section

      CALL timeset(routineN, handle)

      NULLIFY (elec_conf)

      basis_set_name(:) = ""
      basis_set_type(:) = ""
      aux_basis_set_name = ""
      ri_aux_basis_set_name = ""
      orb_basis_set_name = ""
      potential_name = ""
      potential_type = ""
      kgpot_name = ""
      kgpot_type = ""
      z = -1
      zeff_correction = 0.0_dp
      dft_section => section_vals_get_subs_vals(force_env_section, "DFT")
      CALL section_vals_get(kind_section, n_repetition=n_rep)
      k_rep = -1
      akind_name = qs_kind%name
      CALL uppercase(akind_name)
      ! First we use the atom_name to find out the proper KIND section
      DO i_rep = 1, n_rep
         CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                   c_val=keyword, i_rep_section=i_rep)
         CALL uppercase(keyword)
         IF (keyword == akind_name) THEN
            k_rep = i_rep
            EXIT
         END IF
      END DO
      ! The search for the KIND section failed.. check for a QM/MM link atom
      IF (k_rep < 1) THEN
         ipos = INDEX(qs_kind%name, "_")
         IF (((ipos == 2) .OR. (ipos == 3)) .AND. (INDEX(qs_kind%name, "_ghost") == 0)) THEN
            ! If the atm_name could not match any KIND section it maybe be a QM/MM link atom.
            ! ghost atoms will be treated differently.
            akind_name = qs_kind%name(1:ipos-1)
            CALL uppercase(akind_name)
            DO i_rep = 1, n_rep
               CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                         c_val=keyword, i_rep_section=i_rep)
               CALL uppercase(keyword)
               IF (keyword == akind_name) THEN
                  k_rep = i_rep
                  EXIT
               END IF
            END DO
         END IF
      END IF
      ! The search for the KIND section failed.. check element_symbol
      IF (k_rep < 1) THEN
         ! If it's not a link atom let's check for the element and map
         ! the KIND section to the element.
         element_symbol = qs_kind%element_symbol(1:2)
         CALL uppercase(element_symbol)
         DO i_rep = 1, n_rep
            CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                      c_val=keyword, i_rep_section=i_rep)
            CALL uppercase(keyword)
            IF (keyword == element_symbol) THEN
               k_rep = i_rep
               EXIT
            END IF
         END DO
      END IF
      ! In case it should not really match any possible KIND section
      ! let's look if a default one is defined..
      IF (k_rep < 1) THEN
         DO i_rep = 1, n_rep
            CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                      c_val=keyword, i_rep_section=i_rep)
            CALL uppercase(keyword)
            IF (keyword == "DEFAULT") THEN
               k_rep = i_rep
               EXIT
            END IF
         END DO
      END IF
      IF (k_rep < 0 .AND. (.NOT. no_fail)) THEN
         CALL cp_abort(cp__l("qs_kind_types.F",1258), &
                       "No &KIND section was possible to associate to the atomic kind <"// &
                       TRIM(akind_name)//">. The KIND section were also scanned for the"// &
                       " corresponding element <"//TRIM(qs_kind%element_symbol)//"> "// &
                       " and for the DEFAULT section but no match was found. Check your input file!")
      END IF
      ! Retrieve information on element
      CALL get_ptable_info(qs_kind%element_symbol, ielement=z)

      ! Normal parsing of the KIND section
      IF (k_rep > 0) THEN
         ! new style basis set input
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="BASIS_SET", &
                                   explicit=explicit, &
                                   n_rep_val=nb_rep)
         IF (.NOT. explicit) nb_rep = 0
         IF(.NOT.(nb_rep <= maxbas))CALL cp__a("qs_kind_types.F",1275)
         DO i = 1, nb_rep
            CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                      keyword_name="BASIS_SET", i_rep_val=i, c_vals=tmpstringlist)
            IF (SIZE(tmpstringlist) == 1) THEN
               ! default is orbital type
               basis_set_type(i) = "ORB"
               basis_set_name(i) = tmpstringlist(1)
            ELSEIF (SIZE(tmpstringlist) == 2) THEN
               basis_set_type(i) = tmpstringlist(1)
               basis_set_name(i) = tmpstringlist(2)
            ELSE
               CALL cp__b("qs_kind_types.F",1287,"")
            END IF
         END DO

         ! Deprecated keywords
         DO i = 1, nb_rep
            IF (basis_set_type(i) == "ORB") THEN
               orb_basis_set_name = basis_set_name(i)
               EXIT
            END IF
         END DO
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="AUX_FIT_BASIS_SET", c_val=aux_fit_basis_set_name)
         IF (aux_fit_basis_set_name /= "") THEN
            nb_rep = nb_rep+1
            IF(.NOT.(nb_rep <= maxbas))CALL cp__a("qs_kind_types.F",1302)
            basis_set_type(nb_rep) = "AUX_FIT"
            basis_set_name(nb_rep) = aux_fit_basis_set_name
         END IF
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="AUX_BASIS_SET", c_val=aux_basis_set_name)
         IF (aux_basis_set_name /= "") THEN
            nb_rep = nb_rep+1
            IF(.NOT.(nb_rep <= maxbas))CALL cp__a("qs_kind_types.F",1310)
            basis_set_type(nb_rep) = "AUX"
            basis_set_name(nb_rep) = aux_basis_set_name
         END IF
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="LRI_BASIS_SET", c_val=lri_basis_set_name)
         IF (lri_basis_set_name /= "") THEN
            nb_rep = nb_rep+1
            IF(.NOT.(nb_rep <= maxbas))CALL cp__a("qs_kind_types.F",1318)
            basis_set_type(nb_rep) = "LRI"
            basis_set_name(nb_rep) = lri_basis_set_name
         END IF
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="RI_AUX_BASIS_SET", c_val=ri_aux_basis_set_name)
         IF (ri_aux_basis_set_name /= "") THEN
            nb_rep = nb_rep+1
            IF(.NOT.(nb_rep <= maxbas))CALL cp__a("qs_kind_types.F",1326)
            basis_set_type(nb_rep) = "RI_AUX"
            basis_set_name(nb_rep) = ri_aux_basis_set_name
         END IF
         ! end of deprecated input

         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="PAO_BASIS_SIZE", i_val=qs_kind%pao_basis_size)
         pao_pot_section => section_vals_get_subs_vals(kind_section, "PAO_POTENTIAL", i_rep_section=k_rep)
         CALL section_vals_val_get(pao_pot_section, keyword_name="MAXL", i_val=qs_kind%pao_pot_maxl)
         CALL section_vals_val_get(pao_pot_section, keyword_name="MAX_PROJECTOR", i_val=qs_kind%pao_pot_max_projector)
         CALL section_vals_val_get(pao_pot_section, keyword_name="NEIGHBORS", i_val=qs_kind%pao_pot_neighbors)
         CALL section_vals_val_get(pao_pot_section, keyword_name="BETA", r_val=qs_kind%pao_pot_beta)
         CALL section_vals_val_get(pao_pot_section, keyword_name="WEIGHT", r_val=qs_kind%pao_pot_weight)

         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="ELEC_CONF", n_rep_val=i)
         IF (i > 0) THEN
            CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                      keyword_name="ELEC_CONF", i_vals=elec_conf)
            CALL set_qs_kind(qs_kind, elec_conf=elec_conf)
         ENDIF
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="CORE_CORRECTION", r_val=zeff_correction)
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="POTENTIAL", c_val=potential_name)
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="KG_POTENTIAL", c_val=kgpot_name)
         ! assign atom dependent defaults, only H special case
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, n_rep_val=i, &
                                   keyword_name="HARD_EXP_RADIUS")
         IF (i == 0) THEN
            IF (z == 1) THEN
               qs_kind%hard_radius = 1.2_dp
            ELSE
               qs_kind%hard_radius = 0.8_dp*bohr
            ENDIF
         ELSE
            CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                      keyword_name="HARD_EXP_RADIUS", r_val=qs_kind%hard_radius)
         ENDIF

         ! assign atom dependent defaults, only H special case
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, n_rep_val=i, &
                                   keyword_name="RHO0_EXP_RADIUS")
         IF (i == 0) THEN
            qs_kind%hard0_radius = qs_kind%hard_radius
         ELSE
            CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                      keyword_name="RHO0_EXP_RADIUS", r_val=qs_kind%hard0_radius)
         ENDIF
         IF (qs_kind%hard_radius < qs_kind%hard0_radius) &
            CALL cp__b("qs_kind_types.F",1378,"rc0 should be <= rc")

         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="MAX_RAD_LOCAL", r_val=qs_kind%max_rad_local)
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="LEBEDEV_GRID", i_val=qs_kind%ngrid_ang)
         IF (qs_kind%ngrid_ang <= 0) &
            CALL cp__b("qs_kind_types.F",1385,"# point lebedev grid < 0")
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="RADIAL_GRID", i_val=qs_kind%ngrid_rad)
         IF (qs_kind%ngrid_rad <= 0) &
            CALL cp__b("qs_kind_types.F",1389,"# point radial grid < 0")
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="GPW_TYPE", l_val=qs_kind%gpw_type_forced)
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="GHOST", l_val=qs_kind%ghost)
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="NO_OPTIMIZE", l_val=qs_kind%no_optimize)

         ! DFTB3 param
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="DFTB3_PARAM", r_val=qs_kind%dudq_dftb3)
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="LMAX_DFTB", i_val=qs_kind%lmax_dftb)

         ! MAOS
         CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                   keyword_name="MAO", i_val=qs_kind%mao)

         ! Read the BS subsection of the current atomic kind, if enabled
         NULLIFY (bs_section)
         bs_section => section_vals_get_subs_vals(kind_section, "BS", &
                                                  i_rep_section=k_rep)
         section_enabled = .FALSE.
         CALL section_vals_val_get(bs_section, "_SECTION_PARAMETERS_", &
                                   l_val=section_enabled)
         IF (section_enabled) THEN
            qs_kind%bs_occupation = .TRUE.
            !Alpha spin
            NULLIFY (spin_section)
            spin_section => section_vals_get_subs_vals(bs_section, "ALPHA")
            CALL section_vals_get(spin_section, explicit=explicit)
            IF (explicit) THEN
               NULLIFY (add_el)
               CALL section_vals_val_get(spin_section, &
                                         keyword_name="NEL", i_vals=add_el)
               IF(.NOT.(ASSOCIATED(add_el)))CALL cp__a("qs_kind_types.F",1424)
               ALLOCATE (qs_kind%addel(SIZE(add_el), 2))
               qs_kind%addel = 0
               qs_kind%addel(1:SIZE(add_el), 1) = add_el(1:SIZE(add_el))
               NULLIFY (add_el)
               CALL section_vals_val_get(spin_section, &
                                         keyword_name="L", i_vals=add_el)
               IF(.NOT.(ASSOCIATED(add_el)))CALL cp__a("qs_kind_types.F",1431)
               IF(.NOT.(SIZE(add_el) == SIZE(qs_kind%addel, 1)))CALL cp__a("qs_kind_types.F",1432)
               ALLOCATE (qs_kind%laddel(SIZE(add_el), 2))
               qs_kind%laddel = 0
               qs_kind%laddel(1:SIZE(add_el), 1) = add_el(1:SIZE(add_el))
               ALLOCATE (qs_kind%naddel(SIZE(add_el), 2))
               qs_kind%naddel = 0
               NULLIFY (add_el)
               CALL section_vals_val_get(spin_section, &
                                         keyword_name="N", n_rep_val=i)
               IF (i > 0) THEN
                  CALL section_vals_val_get(spin_section, &
                                            keyword_name="N", i_vals=add_el)
                  IF (SIZE(add_el) == SIZE(qs_kind%addel, 1)) THEN
                     qs_kind%naddel(1:SIZE(add_el), 1) = add_el(1:SIZE(add_el))
                  END IF
               END IF
            END IF
            !Beta spin
            NULLIFY (spin_section)
            spin_section => section_vals_get_subs_vals(bs_section, "BETA")
            CALL section_vals_get(spin_section, explicit=explicit)
            IF (explicit) THEN

               NULLIFY (add_el)
               CALL section_vals_val_get(spin_section, &
                                         keyword_name="NEL", i_vals=add_el)
               IF(.NOT.(SIZE(add_el) == SIZE(qs_kind%addel, 1)))CALL cp__a("qs_kind_types.F",1458)
               qs_kind%addel(1:SIZE(add_el), 2) = add_el(1:SIZE(add_el))
               qs_kind%addel(:, :) = qs_kind%addel(:, :)
               NULLIFY (add_el)
               CALL section_vals_val_get(spin_section, &
                                         keyword_name="L", i_vals=add_el)
               IF(.NOT.(SIZE(add_el) == SIZE(qs_kind%addel, 1)))CALL cp__a("qs_kind_types.F",1464)
               qs_kind%laddel(1:SIZE(add_el), 2) = add_el(1:SIZE(add_el))

               CALL section_vals_val_get(spin_section, &
                                         keyword_name="N", n_rep_val=i)
               IF (i > 0) THEN
                  NULLIFY (add_el)
                  CALL section_vals_val_get(spin_section, &
                                            keyword_name="N", i_vals=add_el)
                  IF (SIZE(add_el) == SIZE(qs_kind%addel, 1)) THEN
                     qs_kind%naddel(1:SIZE(add_el), 2) = add_el(1:SIZE(add_el))
                  END IF
               END IF
            END IF
         END IF

         ! Read the DFT+U subsection of the current atomic kind, if enabled

         NULLIFY (dft_plus_u_section)
         dft_plus_u_section => section_vals_get_subs_vals(kind_section, &
                                                          subsection_name="DFT_PLUS_U", &
                                                          i_rep_section=k_rep)
         section_enabled = .FALSE.
         CALL section_vals_val_get(dft_plus_u_section, &
                                   keyword_name="_SECTION_PARAMETERS_", &
                                   l_val=section_enabled)
         IF (section_enabled) THEN
            ALLOCATE (qs_kind%dft_plus_u)
            NULLIFY (qs_kind%dft_plus_u%orbitals)
            CALL section_vals_val_get(dft_plus_u_section, &
                                      keyword_name="L", &
                                      i_val=l)
            qs_kind%dft_plus_u%l = l
            IF (l < 0) &
               CALL cp__b("qs_kind_types.F",1498,"DFT+U| Invalid orbital angular momentum quantum number specified: l < 0")
            CALL section_vals_val_get(dft_plus_u_section, &
                                      keyword_name="U_MINUS_J", &
                                      r_val=qs_kind%dft_plus_u%u_minus_j_target)
            CALL section_vals_val_get(dft_plus_u_section, &
                                      keyword_name="U_RAMPING", &
                                      r_val=qs_kind%dft_plus_u%u_ramping)
            CALL section_vals_val_get(dft_plus_u_section, &
                                      keyword_name="INIT_U_RAMPING_EACH_SCF", &
                                      l_val=qs_kind%dft_plus_u%init_u_ramping_each_scf)
            IF (qs_kind%dft_plus_u%u_ramping > 0.0_dp) THEN
               qs_kind%dft_plus_u%u_minus_j = 0.0_dp
            ELSE
               qs_kind%dft_plus_u%u_minus_j = qs_kind%dft_plus_u%u_minus_j_target
            END IF
            CALL section_vals_val_get(dft_plus_u_section, &
                                      keyword_name="EPS_U_RAMPING", &
                                      r_val=qs_kind%dft_plus_u%eps_u_ramping)

            NULLIFY (enforce_occupation_section)
            enforce_occupation_section => section_vals_get_subs_vals(dft_plus_u_section, &
                                                                     subsection_name="ENFORCE_OCCUPATION")
            subsection_enabled = .FALSE.
            CALL section_vals_val_get(enforce_occupation_section, &
                                      keyword_name="_SECTION_PARAMETERS_", &
                                      l_val=subsection_enabled)
            IF (subsection_enabled) THEN
               NULLIFY (orbitals)
               CALL section_vals_val_get(enforce_occupation_section, &
                                         keyword_name="ORBITALS", &
                                         i_vals=orbitals)
               norbitals = SIZE(orbitals)
               IF (norbitals <= 0 .OR. norbitals > 2*l+1) &
                  CALL cp_abort(cp__l("qs_kind_types.F",1531), "DFT+U| Invalid number of ORBITALS specified: "// &
                                "1 to 2*L+1 integer numbers are expected")
               ALLOCATE (qs_kind%dft_plus_u%orbitals(norbitals))
               qs_kind%dft_plus_u%orbitals(:) = orbitals(:)
               NULLIFY (orbitals)
               DO m = 1, norbitals
                  IF (qs_kind%dft_plus_u%orbitals(m) > l) &
                     CALL cp__b("qs_kind_types.F",1538,"DFT+U| Invalid orbital magnetic quantum number specified: m > l")
                  IF (qs_kind%dft_plus_u%orbitals(m) < -l) &
                     CALL cp__b("qs_kind_types.F",1540,"DFT+U| Invalid orbital magnetic quantum number specified: m < -l")
                  DO j = 1, norbitals
                     IF (j /= m) THEN
                        IF (qs_kind%dft_plus_u%orbitals(j) == qs_kind%dft_plus_u%orbitals(m)) &
                           CALL cp__b("qs_kind_types.F",1544,"DFT+U| An orbital magnetic quantum number was specified twice")
                     END IF
                  END DO
               END DO
               CALL section_vals_val_get(enforce_occupation_section, &
                                         keyword_name="EPS_SCF", &
                                         r_val=qs_kind%dft_plus_u%eps_scf)
               CALL section_vals_val_get(enforce_occupation_section, &
                                         keyword_name="MAX_SCF", &
                                         i_val=i)
               qs_kind%dft_plus_u%max_scf = MAX(-1, i)
               CALL section_vals_val_get(enforce_occupation_section, &
                                         keyword_name="SMEAR", &
                                         l_val=qs_kind%dft_plus_u%smear)
            END IF ! subsection enabled
         END IF ! section enabled

      END IF

      ! Allocate and initialise the orbital basis set data set structure
      CALL init_orbital_pointers(5) ! debug the SUN optimizer

      ! BASIS  and POTENTIAL read only when strictly necessary otherwise, even if not used
      ! we just print misleading informations
      explicit_basis = .FALSE.
      IF (k_rep > 0) THEN
         basis_section => section_vals_get_subs_vals(kind_section, "BASIS", i_rep_section=k_rep, &
                                                     can_return_null=.TRUE.)
         CALL section_vals_get(basis_section, explicit=explicit_basis)
      END IF

      explicit_potential = .FALSE.
      IF (k_rep > 0) THEN
         potential_section => section_vals_get_subs_vals(kind_section, "POTENTIAL", &
                                                         i_rep_section=k_rep, can_return_null=.TRUE.)
         CALL section_vals_get(potential_section, explicit=explicit_potential)
      END IF

      explicit_kgpot = .FALSE.
      IF (k_rep > 0) THEN
         kgpot_section => section_vals_get_subs_vals(kind_section, "KG_POTENTIAL", &
                                                     i_rep_section=k_rep, can_return_null=.TRUE.)
         CALL section_vals_get(kgpot_section, explicit=explicit_kgpot)
      END IF

      SELECT CASE (method_id)
      CASE (do_method_dftb, do_method_scptb, do_method_rm1, do_method_am1, do_method_mndo, &
            do_method_pdg, do_method_pm3, do_method_pm6, do_method_mndod, &
            do_method_pnnl)
         ! Allocate all_potential
         CALL allocate_potential(qs_kind%all_potential)
         CALL set_default_all_potential(qs_kind%all_potential, z, zeff_correction)
         CALL get_qs_kind(qs_kind, elec_conf=elec_conf)
         IF (.NOT. ASSOCIATED(elec_conf)) THEN
            CALL get_potential(potential=qs_kind%all_potential, elec_conf=elec_conf)
            CALL set_qs_kind(qs_kind, elec_conf=elec_conf)
         END IF
         IF (qs_kind%ghost) THEN
            CALL get_qs_kind(qs_kind=qs_kind, elec_conf=elec_conf)
            elec_conf(:) = 0
            CALL get_potential(potential=qs_kind%all_potential, &
                               elec_conf=elec_conf)
            elec_conf(:) = 0
            CALL set_potential(potential=qs_kind%all_potential, &
                               zeff=0.0_dp, &
                               zeff_correction=0.0_dp)
         END IF

         ! Basis set (Parameters)
         SELECT CASE (method_id)
         CASE (do_method_rm1, do_method_am1, do_method_mndo, do_method_pdg, &
               do_method_pm3, do_method_pm6, do_method_mndod, do_method_pnnl)
            ! Setup proper semiempirical parameters
            check = .NOT. ASSOCIATED(qs_kind%se_parameter)
            IF(.NOT.(check))CALL cp__a("qs_kind_types.F",1618)
            CALL semi_empirical_create(qs_kind%se_parameter)
            ! Check if we allow p-orbitals on H
            SELECT CASE (z)
            CASE (1)
               IF (k_rep > 0) THEN
                  CALL section_vals_val_get(kind_section, i_rep_section=k_rep, &
                                            keyword_name="SE_P_ORBITALS_ON_H", l_val=qs_kind%se_parameter%p_orbitals_on_h)
               END IF
            CASE DEFAULT
               ! No special cases for other elements..
            END SELECT
            ! Set default parameters
            CALL section_vals_val_get(dft_section, "QS%SE%STO_NG", i_val=ngauss)
            CALL se_param_set_default(qs_kind%se_parameter, z, method_id)
            NULLIFY (tmp_basis_set)
            CALL init_se_param(qs_kind%se_parameter, tmp_basis_set, ngauss)
            CALL add_basis_set_to_container(qs_kind%basis_sets, tmp_basis_set, "ORB")
            CALL init_potential(qs_kind%all_potential, itype="BARE", &
                                zeff=qs_kind%se_parameter%zeff, zeff_correction=zeff_correction)
            qs_kind%se_parameter%zeff = qs_kind%se_parameter%zeff-zeff_correction
         CASE (do_method_dftb, do_method_scptb)
            ! Do nothing at this stage.. The initialization of the parameters will be done
            ! later
         CASE DEFAULT
            CALL cp__b("qs_kind_types.F",1643,"")
         END SELECT

         check = ((potential_name /= '') .OR. explicit_potential)
         IF (check) &
            CALL cp_warn(cp__l("qs_kind_types.F",1648), &
                         "Information provided in the input file regarding POTENTIAL for KIND <"// &
                         TRIM(qs_kind%name)//"> will be ignored!")

         check = ((orb_basis_set_name /= '') .OR. explicit_basis)
         IF (check) &
            CALL cp_warn(cp__l("qs_kind_types.F",1654), &
                         "Information provided in the input file regarding BASIS for KIND <"// &
                         TRIM(qs_kind%name)//"> will be ignored!")

      CASE DEFAULT

         ! Allocate and initialise the basis set data set structure
         ! first external basis sets
         DO i = 1, nb_rep
            NULLIFY (tmp_basis_set)
            CALL allocate_gto_basis_set(tmp_basis_set)
            CALL read_gto_basis_set(qs_kind%element_symbol, basis_set_name(i), &
                                    tmp_basis_set, para_env, dft_section)
            tmp = basis_set_type(i)
            CALL uppercase(tmp)
            CALL add_basis_set_to_container(qs_kind%basis_sets, tmp_basis_set, tmp)
         END DO
         ! now explicit basis sets
         IF (explicit_basis) THEN
            CALL section_vals_get(basis_section, n_repetition=nexp)
            DO i = 1, nexp
               NULLIFY (tmp_basis_set)
               CALL allocate_gto_basis_set(tmp_basis_set)
               CALL read_gto_basis_set(qs_kind%element_symbol, basis_type, &
                                       tmp_basis_set, basis_section, i)
               tmp = basis_type
               CALL uppercase(tmp)
               CALL add_basis_set_to_container(qs_kind%basis_sets, tmp_basis_set, tmp)
            END DO
         END IF
         ! combine multiple basis sets
         DO i = 1, SIZE(qs_kind%basis_sets)
            NULLIFY (tmp_basis_set)
            CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis_set, &
                                          inumbas=i, basis_type=basis_type)
            IF (basis_type == "") CYCLE
            DO j = i+1, SIZE(qs_kind%basis_sets)
               NULLIFY (sup_basis_set)
               CALL get_basis_from_container(qs_kind%basis_sets, basis_set=sup_basis_set, &
                                             inumbas=j, basis_type=tmp)
               IF (basis_type == tmp) THEN
                  ! we found a match, combine the basis sets and delete the second
                  CALL combine_basis_sets(tmp_basis_set, sup_basis_set)
                  CALL remove_basis_from_container(qs_kind%basis_sets, j)
               END IF
            END DO
            NULLIFY (sup_basis_set)
         END DO
         NULLIFY (tmp_basis_set)

         ! check that we have an orbital basis set
         nobasis = .TRUE.
         DO i = 1, SIZE(qs_kind%basis_sets)
            NULLIFY (tmp_basis_set)
            CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis_set, &
                                          inumbas=i, basis_type=basis_type)
            IF (basis_type == "ORB") nobasis = .FALSE.
         END DO
         IF (nobasis) THEN
            CALL cp_abort(cp__l("qs_kind_types.F",1713), &
                          "No basis set type was defined for the "// &
                          "atomic kind <"//TRIM(qs_kind%name)//">")
         END IF

         ! If Ghost atom we don't need to allocate/initialize anything connected to POTENTIAL
         IF (qs_kind%ghost) THEN
            IF (ASSOCIATED(qs_kind%elec_conf)) qs_kind%elec_conf = 0
         ELSE
            ! Allocate and initialise the potential data set structure
            IF ((potential_name /= '') .OR. explicit_potential) THEN
               ipos = INDEX(potential_name, "-")
               IF (ipos > 1) THEN
                  potential_type = potential_name(:ipos-1)
               ELSE
                  potential_type = potential_name
               END IF
               CALL uppercase(potential_type)

               SELECT CASE (TRIM (potential_type))
               CASE ("ALL")
                  CALL allocate_potential(qs_kind%all_potential)
                  CALL read_potential(qs_kind%element_symbol, potential_name, &
                                      qs_kind%all_potential, zeff_correction, para_env, &
                                      dft_section, potential_section)
                  CALL set_potential(qs_kind%all_potential, z=z)
                  CALL get_qs_kind(qs_kind, elec_conf=elec_conf)
                  IF (.NOT. ASSOCIATED(elec_conf)) THEN
                     CALL get_potential(potential=qs_kind%all_potential, elec_conf=elec_conf)
                     CALL set_qs_kind(qs_kind, elec_conf=elec_conf)
                  END IF
               CASE ("GTH")
                  CALL allocate_potential(qs_kind%gth_potential)
                  CALL read_potential(qs_kind%element_symbol, potential_name, &
                                      qs_kind%gth_potential, zeff_correction, para_env, &
                                      dft_section, potential_section)
                  CALL set_potential(qs_kind%gth_potential, z=z)
                  CALL get_qs_kind(qs_kind, elec_conf=elec_conf)
                  IF (.NOT. ASSOCIATED(elec_conf)) THEN
                     CALL get_potential(potential=qs_kind%gth_potential, elec_conf=elec_conf)
                     CALL set_qs_kind(qs_kind, elec_conf=elec_conf)
                  END IF
               CASE DEFAULT
                  CALL cp_abort(cp__l("qs_kind_types.F",1756), &
                                "An invalid potential type <"// &
                                TRIM(potential_name)//"> was specified "// &
                                "for the atomic kind <"// &
                                TRIM(qs_kind%name))
               END SELECT
            ELSE
               CALL cp_abort(cp__l("qs_kind_types.F",1763), &
                             "No potential type was defined for the "// &
                             "atomic kind <"//TRIM(qs_kind%name)//">")
            END IF
            ! Allocate and initialise the potential data set structure
            IF ((kgpot_name /= '') .OR. explicit_kgpot) THEN
               ipos = INDEX(kgpot_name, "-")
               IF (ipos > 1) THEN
                  kgpot_type = kgpot_name(:ipos-1)
               ELSE
                  kgpot_type = kgpot_name
               END IF
               CALL uppercase(kgpot_type)

               SELECT CASE (TRIM (kgpot_type))
               CASE ("TNADD")
                  CALL allocate_potential(qs_kind%tnadd_potential)
                  CALL read_potential(qs_kind%element_symbol, kgpot_name, &
                                      qs_kind%tnadd_potential, para_env, &
                                      dft_section, kgpot_section)
               CASE ("NONE")
                  NULLIFY (qs_kind%tnadd_potential)
               CASE DEFAULT
                  CALL cp_abort(cp__l("qs_kind_types.F",1786), &
                                "An invalid kg_potential type <"// &
                                TRIM(potential_name)//"> was specified "// &
                                "for the atomic kind <"// &
                                TRIM(qs_kind%name))
               END SELECT
            END IF
         END IF
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE read_qs_kind

! **************************************************************************************************
!> \brief Read an atomic kind set data set from the input file.
!> \param qs_kind_set ...
!> \param atomic_kind_set ...
!> \param kind_section ...
!> \param para_env ...
!> \param force_env_section ...
! **************************************************************************************************
   SUBROUTINE create_qs_kind_set(qs_kind_set, atomic_kind_set, kind_section, para_env, force_env_section)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(section_vals_type), POINTER                   :: kind_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qs_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, method, nkind, qs_method
      LOGICAL                                            :: no_fail

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(qs_kind_set)) CALL cp__b("qs_kind_types.F",1824,"create_qs_kind_set: qs_kind_set already associated")
      IF (.NOT. ASSOCIATED(atomic_kind_set)) CALL cp__b("qs_kind_types.F",1825,"create_qs_kind_set: atomic_kind_set not associated")

      no_fail = .FALSE.

      ! Between all methods only SE and DFTB may not need a KIND section.
      CALL section_vals_val_get(force_env_section, "METHOD", i_val=method)
      IF (method == do_qs) THEN
         CALL section_vals_val_get(force_env_section, "DFT%QS%METHOD", i_val=qs_method)
         SELECT CASE (qs_method)
         CASE (do_method_mndo, do_method_am1, do_method_pm3, do_method_pm6, &
               do_method_pdg, do_method_rm1, do_method_mndod, do_method_pnnl)
            no_fail = .TRUE.
         CASE (do_method_dftb, do_method_scptb)
            no_fail = .TRUE.
         END SELECT
      ELSE
         qs_method = method
      END IF

      nkind = SIZE(atomic_kind_set)
      ALLOCATE (qs_kind_set(nkind))

      DO ikind = 1, nkind
         qs_kind_set(ikind)%name = atomic_kind_set(ikind)%name
         qs_kind_set(ikind)%element_symbol = atomic_kind_set(ikind)%element_symbol
         qs_kind_set(ikind)%natom = atomic_kind_set(ikind)%natom
         CALL read_qs_kind(qs_kind_set(ikind), kind_section, para_env, force_env_section, no_fail, qs_method)
      END DO

      CALL timestop(handle)

   END SUBROUTINE create_qs_kind_set

! **************************************************************************************************
!> \brief This routines should perform only checks. no settings are allowed at
!>     this level anymore..
!> \param qs_kind ...
!> \param dft_control ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE check_qs_kind(qs_kind, dft_control, subsys_section)

      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(dft_control_type), INTENT(IN)                 :: dft_control
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_qs_kind', routineP = moduleN//':'//routineN

      LOGICAL                                            :: defined
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_parameter
      TYPE(scptb_parameter_type), POINTER                :: scptb_parameter
      TYPE(semi_empirical_type), POINTER                 :: se_parameter

      IF (dft_control%qs_control%semi_empirical) THEN
         CALL get_qs_kind(qs_kind, se_parameter=se_parameter)
         IF(.NOT.(ASSOCIATED(se_parameter)))CALL cp__a("qs_kind_types.F",1880)
         CALL get_se_param(se_parameter, defined=defined)
         IF(.NOT.(defined))CALL cp__a("qs_kind_types.F",1882)
         CALL write_se_param(se_parameter, subsys_section)
      ELSE IF (dft_control%qs_control%dftb) THEN
         CALL get_qs_kind(qs_kind, dftb_parameter=dftb_parameter)
         IF(.NOT.(ASSOCIATED(dftb_parameter)))CALL cp__a("qs_kind_types.F",1886)
         CALL get_dftb_atom_param(dftb_parameter, defined=defined)
         IF(.NOT.(defined))CALL cp__a("qs_kind_types.F",1888)
         CALL write_dftb_atom_param(dftb_parameter, subsys_section)
      ELSE IF (dft_control%qs_control%scptb) THEN
         CALL get_qs_kind(qs_kind, scptb_parameter=scptb_parameter)
         IF(.NOT.(ASSOCIATED(scptb_parameter)))CALL cp__a("qs_kind_types.F",1892)
         CALL get_scptb_parameter(scptb_parameter, defined=defined)
         IF(.NOT.(defined))CALL cp__a("qs_kind_types.F",1894)
         CALL write_scptb_parameter(scptb_parameter, subsys_section)
      END IF

   END SUBROUTINE check_qs_kind

! **************************************************************************************************
!> \brief ...
!> \param qs_kind_set ...
!> \param dft_control ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE check_qs_kind_set(qs_kind_set, dft_control, subsys_section)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(dft_control_type), INTENT(IN)                 :: dft_control
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'check_qs_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, nkind
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(qs_kind_set)) THEN
         nkind = SIZE(qs_kind_set)
         DO ikind = 1, nkind
            qs_kind => qs_kind_set(ikind)
            CALL check_qs_kind(qs_kind, dft_control, subsys_section)
         END DO
      ELSE
         CALL cp__b("qs_kind_types.F",1926,"The pointer qs_kind_set is not associated")
      END IF
      CALL timestop(handle)
   END SUBROUTINE check_qs_kind_set

! **************************************************************************************************
!> \brief Set the components of an atomic kind data set.
!> \param qs_kind ...
!> \param paw_atom ...
!> \param ghost ...
!> \param hard_radius ...
!> \param hard0_radius ...
!> \param soft_basis_set ...
!> \param hard_basis_set ...
!> \param lmax_rho0 ...
!> \param zeff ...
!> \param no_optimize ...
!> \param dispersion ...
!> \param u_minus_j ...
!> \param reltmat ...
!> \param dftb_parameter ...
!> \param scptb_parameter ...
!> \param elec_conf ...
! **************************************************************************************************
   SUBROUTINE set_qs_kind(qs_kind, paw_atom, ghost, hard_radius, hard0_radius, &
                          soft_basis_set, hard_basis_set, lmax_rho0, zeff, &
                          no_optimize, dispersion, u_minus_j, reltmat, &
                          dftb_parameter, scptb_parameter, &
                          elec_conf)

      TYPE(qs_kind_type), INTENT(INOUT)                  :: qs_kind
      LOGICAL, INTENT(IN), OPTIONAL                      :: paw_atom, ghost
      REAL(dp), INTENT(IN), OPTIONAL                     :: hard_radius, hard0_radius
      TYPE(gto_basis_set_type), OPTIONAL, POINTER        :: soft_basis_set, hard_basis_set
      INTEGER, INTENT(IN), OPTIONAL                      :: lmax_rho0
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: zeff
      LOGICAL, INTENT(IN), OPTIONAL                      :: no_optimize
      TYPE(qs_atom_dispersion_type), OPTIONAL, POINTER   :: dispersion
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: u_minus_j
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: reltmat
      TYPE(qs_dftb_atom_type), OPTIONAL, POINTER         :: dftb_parameter
      TYPE(scptb_parameter_type), OPTIONAL, POINTER      :: scptb_parameter
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: elec_conf

      CHARACTER(len=*), PARAMETER :: routineN = 'set_qs_kind', routineP = moduleN//':'//routineN

      IF (PRESENT(dftb_parameter)) qs_kind%dftb_parameter => dftb_parameter
      IF (PRESENT(scptb_parameter)) qs_kind%scptb_parameter => scptb_parameter
      IF (PRESENT(elec_conf)) THEN
         IF (ASSOCIATED(qs_kind%elec_conf)) THEN
            DEALLOCATE (qs_kind%elec_conf)
         ENDIF
         ALLOCATE (qs_kind%elec_conf(0:SIZE(elec_conf)-1))
         qs_kind%elec_conf(:) = elec_conf(:)
      ENDIF
      IF (PRESENT(paw_atom)) qs_kind%paw_atom = paw_atom
      IF (PRESENT(hard_radius)) qs_kind%hard_radius = hard_radius
      IF (PRESENT(hard0_radius)) qs_kind%hard0_radius = hard0_radius
      IF (PRESENT(soft_basis_set)) qs_kind%soft_basis_set => soft_basis_set
      IF (PRESENT(hard_basis_set)) qs_kind%hard_basis_set => hard_basis_set
      IF (PRESENT(lmax_rho0)) qs_kind%lmax_rho0 = lmax_rho0
      IF (PRESENT(zeff)) THEN
         IF (ASSOCIATED(qs_kind%all_potential)) THEN
            CALL set_potential(potential=qs_kind%all_potential, zeff=zeff)
         ELSE IF (ASSOCIATED(qs_kind%gth_potential)) THEN
            CALL set_potential(potential=qs_kind%gth_potential, zeff=zeff)
         END IF
      END IF
      IF (PRESENT(ghost)) qs_kind%ghost = ghost

      IF (PRESENT(no_optimize)) qs_kind%no_optimize = no_optimize

      IF (PRESENT(dispersion)) qs_kind%dispersion => dispersion

      IF (PRESENT(u_minus_j)) THEN
         IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
            qs_kind%dft_plus_u%u_minus_j = u_minus_j
         END IF
      END IF

      IF (PRESENT(reltmat)) qs_kind%reltmat => reltmat

   END SUBROUTINE set_qs_kind

! **************************************************************************************************
!> \brief Write an atomic kind data set to the output unit.
!> \param qs_kind ...
!> \param kind_number ...
!> \param output_unit ...
!> \par History
!>      Creation (09.02.2002,MK)
! **************************************************************************************************
   SUBROUTINE write_qs_kind(qs_kind, kind_number, output_unit)

      TYPE(qs_kind_type), POINTER                        :: qs_kind
      INTEGER, INTENT(in)                                :: kind_number, output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'write_qs_kind', routineP = moduleN//':'//routineN

      CHARACTER(LEN=3)                                   :: yon
      CHARACTER(LEN=default_string_length)               :: basis_type, bstring
      INTEGER                                            :: ibas
      TYPE(gto_basis_set_type), POINTER                  :: tmp_basis

      IF (output_unit > 0) THEN

         IF (ASSOCIATED(qs_kind)) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,I2,A,T57,A,T75,I6)") &
               kind_number, ". Atomic kind: "//TRIM(qs_kind%name), &
               "Number of atoms: ", qs_kind%natom

            DO ibas = 1, SIZE(qs_kind%basis_sets, 1)
               NULLIFY (tmp_basis)
               CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis, &
                                             inumbas=ibas, basis_type=basis_type)
               IF (basis_type == "") CYCLE
               SELECT CASE (basis_type)
               CASE DEFAULT
                  bstring = "Basis Set"
               CASE ("ORB")
                  bstring = "Orbital Basis Set"
               CASE ("AUX")
                  bstring = "Auxiliary Basis Set"
               CASE ("RI_AUX")
                  bstring = "RI Auxiliary Basis Set"
               CASE ("AUX_FIT")
                  bstring = "Auxiliary Fit Basis Set"
               CASE ("LRI")
                  bstring = "LRI Basis Set"
               END SELECT

               CALL write_orb_basis_set(tmp_basis, output_unit, bstring)

            END DO

            IF (qs_kind%ghost) THEN
               WRITE (UNIT=output_unit, FMT="(/,T6,A)") &
                  "The atoms of this atomic kind are GHOST atoms!"
            END IF
            IF (qs_kind%paw_atom) THEN
               WRITE (UNIT=output_unit, FMT="(/,T6,A)") &
                  "The atoms of this atomic kind are PAW atoms (GAPW):"
               WRITE (UNIT=output_unit, FMT="(T8,A,T71,F10.3)") &
                  "Hard Gaussian function radius:", qs_kind%hard_radius, &
                  "Rho0 radius:", qs_kind%hard0_radius, &
                  "Maximum GTO radius used for PAW projector construction:", &
                  qs_kind%max_rad_local
               CALL write_orb_basis_set(qs_kind%soft_basis_set, output_unit, &
                                        "GAPW Soft Basis Set")
            END IF
            ! Potentials
            CALL write_potential(qs_kind%all_potential, output_unit)
            CALL write_potential(qs_kind%gth_potential, output_unit)
            CALL write_potential(qs_kind%tnadd_potential, output_unit)
            IF (ASSOCIATED(qs_kind%dft_plus_u)) THEN
               WRITE (UNIT=output_unit, FMT="(/,T6,A,/,T8,A,T76,I5,/,T8,A,T73,F8.3)") &
                  "A DFT+U correction is applied to atoms of this atomic kind:", &
                  "Angular quantum momentum number L:", qs_kind%dft_plus_u%l, &
                  "U(eff) = (U - J) value in [eV]:", qs_kind%dft_plus_u%u_minus_j_target*evolt
               IF (qs_kind%dft_plus_u%u_ramping > 0.0_dp) THEN
                  IF (qs_kind%dft_plus_u%init_u_ramping_each_scf) THEN
                     yon = "YES"
                  ELSE
                     yon = " NO"
                  END IF
                  WRITE (UNIT=output_unit, FMT="(T8,A,T73,F8.3,/,T8,A,T73,ES8.1,/,T8,A,T78,A3)") &
                     "Increment for U ramping in [eV]:", qs_kind%dft_plus_u%u_ramping*evolt, &
                     "SCF threshold value for U ramping:", qs_kind%dft_plus_u%eps_u_ramping, &
                     "Set U ramping value to zero before each wavefunction optimisation:", yon
               END IF
               IF (ASSOCIATED(qs_kind%dft_plus_u%orbitals)) THEN
                  WRITE (UNIT=output_unit, FMT="(T8,A)") &
                     "An initial orbital occupation is requested:"
                  WRITE (UNIT=output_unit, FMT="(T9,A,(T78,I3))") &
                     "Preferred (initial) orbital occupation order (orbital M values):", &
                     qs_kind%dft_plus_u%orbitals(:)
                  WRITE (UNIT=output_unit, FMT="(T9,A,T71,ES10.3,/,T9,A,T76,I5)") &
                     "Threshold value for the SCF convergence criterion:", &
                     qs_kind%dft_plus_u%eps_scf, &
                     "Number of initial SCF iterations:", &
                     qs_kind%dft_plus_u%max_scf
                  IF (qs_kind%dft_plus_u%smear) THEN
                     WRITE (UNIT=output_unit, FMT="(T9,A)") &
                        "A smearing of the orbital occupations will be performed"
                  END IF
               END IF
            END IF
         ELSE
            CALL cp__b("qs_kind_types.F",2114,"")
         END IF

      END IF

   END SUBROUTINE write_qs_kind

! **************************************************************************************************
!> \brief Write an atomic kind set data set to the output unit.
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \par History
!>      Creation (09.02.2002,MK)
! **************************************************************************************************
   SUBROUTINE write_qs_kind_set(qs_kind_set, subsys_section)
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_qs_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, nkind, output_unit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%KINDS", extension=".Log")
      IF (output_unit > 0) THEN
         IF (ASSOCIATED(qs_kind_set)) THEN
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") "ATOMIC KIND INFORMATION"
            nkind = SIZE(qs_kind_set)
            DO ikind = 1, nkind
               qs_kind => qs_kind_set(ikind)
               CALL write_qs_kind(qs_kind, ikind, output_unit)
            END DO
         ELSE
            CALL cp__b("qs_kind_types.F",2154,"")
         END IF
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%KINDS")

      CALL timestop(handle)

   END SUBROUTINE write_qs_kind_set

! **************************************************************************************************
!> \brief Write all the GTO basis sets of an atomic kind set to the output
!>     unit (for the printing of the unnormalized basis sets as read from
!>           database).
!> \param qs_kind_set ...
!> \param subsys_section ...
!> \par History
!>      Creation (17.01.2002,MK)
! **************************************************************************************************
   SUBROUTINE write_gto_basis_sets(qs_kind_set, subsys_section)

      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_gto_basis_sets', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: basis_type, bstring
      INTEGER                                            :: handle, ibas, ikind, nkind, output_unit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: tmp_basis
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%KINDS/BASIS_SET", &
                                         extension=".Log")
      IF (output_unit > 0) THEN
         IF (ASSOCIATED(qs_kind_set)) THEN
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") &
               "BASIS SET INFORMATION (Unnormalised Gaussian-type functions)"
            nkind = SIZE(qs_kind_set)
            DO ikind = 1, nkind
               qs_kind => qs_kind_set(ikind)
               WRITE (UNIT=output_unit, FMT="(/,T2,I2,A)") &
                  ikind, ". Atomic kind: "//TRIM(qs_kind%name)

               DO ibas = 1, SIZE(qs_kind%basis_sets, 1)
                  NULLIFY (tmp_basis)
                  CALL get_basis_from_container(qs_kind%basis_sets, basis_set=tmp_basis, &
                                                inumbas=ibas, basis_type=basis_type)
                  IF (basis_type == "") CYCLE
                  SELECT CASE (basis_type)
                  CASE DEFAULT
                     bstring = "Basis Set"
                  CASE ("ORB")
                     bstring = "Orbital Basis Set"
                  CASE ("AUX")
                     bstring = "Auxiliary Basis Set"
                  CASE ("RI_AUX")
                     bstring = "RI Auxiliary Basis Set"
                  CASE ("AUX_FIT")
                     bstring = "Auxiliary Fit Basis Set"
                  CASE ("LRI")
                     bstring = "LRI Basis Set"
                  END SELECT

                  CALL write_gto_basis_set(tmp_basis, output_unit, bstring)

               END DO

               CALL write_gto_basis_set(qs_kind%soft_basis_set, &
                                        output_unit, "GAPW Soft Basis Set")

            END DO
         ELSE
            CALL cp__b("qs_kind_types.F",2234,"")
         END IF
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%KINDS/BASIS_SET")

      CALL timestop(handle)

   END SUBROUTINE write_gto_basis_sets

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind ...
!> \param qs_kind ...
!> \param ispin ...
!> \param ncalc ...
!> \param ncore ...
!> \param nelem ...
! **************************************************************************************************
   SUBROUTINE init_atom_electronic_state(atomic_kind, qs_kind, ispin, ncalc, ncore, nelem)

      TYPE(atomic_kind_type), INTENT(IN)                 :: atomic_kind
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      INTEGER, INTENT(IN)                                :: ispin
      INTEGER, DIMENSION(0:3, 10), INTENT(OUT)           :: ncalc, ncore, nelem

      CHARACTER(len=*), PARAMETER :: routineN = 'init_atom_electronic_state', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, l, ll, ne, nn, z
      INTEGER, DIMENSION(0:3)                            :: econfx
      INTEGER, DIMENSION(:), POINTER                     :: econf
      INTEGER, DIMENSION(:, :), POINTER                  :: addel, laddel, naddel
      LOGICAL                                            :: bs_occupation
      TYPE(gth_potential_type), POINTER                  :: gth_potential

      CALL get_atomic_kind(atomic_kind, z=z)
      NULLIFY (gth_potential)
      CALL get_qs_kind(qs_kind, &
                       gth_potential=gth_potential, &
                       bs_occupation=bs_occupation, &
                       addel=addel, laddel=laddel, naddel=naddel)

      ! electronic state
      nelem = 0
      ncore = 0
      ncalc = 0
      NULLIFY (econf)
      IF (ASSOCIATED(gth_potential)) THEN
         CALL get_potential(gth_potential, elec_conf=econf)
         econfx = 0
         econfx(0:SIZE(econf)-1) = econf
         IF (SUM(econf) >= 0) THEN
            DO l = 0, 3
               ll = 2*(2*l+1)
               nn = ptable(z)%e_conv(l)-econfx(l)
               ii = 0
               DO
                  ii = ii+1
                  IF (nn <= ll) THEN
                     ncore(l, ii) = nn
                     EXIT
                  ELSE
                     ncore(l, ii) = ll
                     nn = nn-ll
                  END IF
               END DO
            END DO
            DO l = 0, 3
               ll = 2*(2*l+1)
               nn = ptable(z)%e_conv(l)
               ii = 0
               DO
                  ii = ii+1
                  IF (nn <= ll) THEN
                     nelem(l, ii) = nn
                     EXIT
                  ELSE
                     nelem(l, ii) = ll
                     nn = nn-ll
                  END IF
               END DO
            END DO
            ncalc = nelem-ncore
         ELSE
            ncore = 0
            ncalc = 0
            DO l = 0, 3
               ll = 2*(2*l+1)
               nn = ABS(econfx(l))
               ii = 0
               DO
                  ii = ii+1
                  IF (nn <= ll) THEN
                     ncalc(l, ii) = -nn
                     EXIT
                  ELSE
                     ncalc(l, ii) = -ll
                     nn = nn-ll
                  END IF
               END DO
            END DO
         END IF
      ELSE
         DO l = 0, 3
            ll = 2*(2*l+1)
            nn = ptable(z)%e_conv(l)
            ii = 0
            DO
               ii = ii+1
               IF (nn <= ll) THEN
                  nelem(l, ii) = nn
                  EXIT
               ELSE
                  nelem(l, ii) = ll
                  nn = nn-ll
               END IF
            END DO
         END DO
         ncalc = nelem-ncore
      END IF

      ! readjust the occupation number of the orbitals as requested by user
      ! this is done to break symmetry (bs) and bias the intial guess
      ! to the pre-defined multiplicity/charge state of the atom
      IF (bs_occupation) THEN
         DO i = 1, SIZE(addel, 1)
            ne = addel(i, ispin)
            l = laddel(i, ispin)
            nn = naddel(i, ispin)-l
            IF (ne /= 0) THEN
               IF (nn == 0) THEN
                  DO ii = SIZE(nelem, 2), 1, -1
                     IF (ncalc(l, ii) > 0) THEN
                        IF ((ncalc(l, ii)+ne) < 2*(2*l+1)+1) THEN
                           ncalc(l, ii) = ncalc(l, ii)+ne
                           nn = ii
                        ELSE
                           ncalc(l, ii+1) = ncalc(l, ii+1)+ne
                           nn = ii+1
                        END IF
                        EXIT
                     ELSE IF (ii == 1) THEN
                        ncalc(l, ii) = ncalc(l, ii)+ne
                        nn = ii
                     END IF
                  END DO
               ELSE
                  ncalc(l, nn) = ncalc(l, nn)+ne
               END IF
               IF (ncalc(l, nn) < 0) THEN
                  ncalc(l, nn) = 0
               END IF
            END IF
         END DO
      END IF

      IF (qs_kind%ghost) THEN
         nelem = 0
         ncore = 0
         ncalc = 0
      END IF

   END SUBROUTINE init_atom_electronic_state

! **************************************************************************************************

END MODULE qs_kind_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Add CP2K error reporting, new add_force routine [07.2014,JGH]
!> \author MK (03.06.2002)
! **************************************************************************************************
MODULE qs_force_types

   !USE cp_control_types,                ONLY: qs_control_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_force_types'
   PRIVATE

   TYPE qs_force_type
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: all_potential, &
                                                 core_overlap, &
                                                 gth_ppl, &
                                                 gth_nlcc, &
                                                 gth_ppnl, &
                                                 kinetic, &
                                                 overlap, &
                                                 overlap_admm, &
                                                 rho_core, &
                                                 rho_elec, &
                                                 rho_lri_elec, &
                                                 vhxc_atom, &
                                                 g0s_Vh_elec, &
                                                 repulsive, &
                                                 dispersion, &
                                                 other, &
                                                 ch_pulay, &
                                                 fock_4c, &
                                                 ehrenfest, &
                                                 efield, &
                                                 eev, &
                                                 mp2_sep, &
                                                 mp2_non_sep, &
                                                 total
   END TYPE qs_force_type

   PUBLIC :: qs_force_type

   PUBLIC :: allocate_qs_force, &
             add_qs_force, &
             deallocate_qs_force, &
             zero_qs_force

CONTAINS

! **************************************************************************************************
!> \brief   Allocate a Quickstep force data structure.
!> \param qs_force ...
!> \param natom_of_kind ...
!> \date    05.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_qs_force(qs_force, natom_of_kind)

      TYPE(qs_force_type), DIMENSION(:), POINTER         :: qs_force
      INTEGER, DIMENSION(:), INTENT(IN)                  :: natom_of_kind

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_qs_force', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, n, nkind

      IF (ASSOCIATED(qs_force)) CALL deallocate_qs_force(qs_force)

      nkind = SIZE(natom_of_kind)

      ALLOCATE (qs_force(nkind))

      DO ikind = 1, nkind
         n = natom_of_kind(ikind)
         ALLOCATE (qs_force(ikind)%all_potential(3, n))
         ALLOCATE (qs_force(ikind)%core_overlap(3, n))
         ALLOCATE (qs_force(ikind)%gth_ppl(3, n))
         ALLOCATE (qs_force(ikind)%gth_nlcc(3, n))
         ALLOCATE (qs_force(ikind)%gth_ppnl(3, n))
         ALLOCATE (qs_force(ikind)%kinetic(3, n))
         ALLOCATE (qs_force(ikind)%overlap(3, n))
         ALLOCATE (qs_force(ikind)%overlap_admm(3, n))
         ALLOCATE (qs_force(ikind)%rho_core(3, n))
         ALLOCATE (qs_force(ikind)%rho_elec(3, n))
         ALLOCATE (qs_force(ikind)%rho_lri_elec(3, n))
         ALLOCATE (qs_force(ikind)%vhxc_atom(3, n))
         ALLOCATE (qs_force(ikind)%g0s_Vh_elec(3, n))
         ALLOCATE (qs_force(ikind)%repulsive(3, n))
         ALLOCATE (qs_force(ikind)%dispersion(3, n))
         ALLOCATE (qs_force(ikind)%other(3, n))
         ALLOCATE (qs_force(ikind)%ch_pulay(3, n))
         ALLOCATE (qs_force(ikind)%ehrenfest(3, n))
         ALLOCATE (qs_force(ikind)%efield(3, n))
         ALLOCATE (qs_force(ikind)%eev(3, n))
         ! Always initialize ch_pulay to zero..
         qs_force(ikind)%ch_pulay = 0.0_dp
         ALLOCATE (qs_force(ikind)%fock_4c(3, n))
         ALLOCATE (qs_force(ikind)%mp2_sep(3, n))
         ALLOCATE (qs_force(ikind)%mp2_non_sep(3, n))
         ALLOCATE (qs_force(ikind)%total(3, n))
      END DO

   END SUBROUTINE allocate_qs_force

! **************************************************************************************************
!> \brief   Deallocate a Quickstep force data structure.
!> \param qs_force ...
!> \date    05.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_qs_force(qs_force)

      TYPE(qs_force_type), DIMENSION(:), POINTER         :: qs_force

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_qs_force', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nkind

      IF(.NOT.(ASSOCIATED(qs_force)))CALL cp__a("qs_force_types.F",131)

      nkind = SIZE(qs_force)

      DO ikind = 1, nkind

         IF (ASSOCIATED(qs_force(ikind)%all_potential)) THEN
            DEALLOCATE (qs_force(ikind)%all_potential)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%core_overlap)) THEN
            DEALLOCATE (qs_force(ikind)%core_overlap)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%gth_ppl)) THEN
            DEALLOCATE (qs_force(ikind)%gth_ppl)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%gth_nlcc)) THEN
            DEALLOCATE (qs_force(ikind)%gth_nlcc)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%gth_ppnl)) THEN
            DEALLOCATE (qs_force(ikind)%gth_ppnl)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%kinetic)) THEN
            DEALLOCATE (qs_force(ikind)%kinetic)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%overlap)) THEN
            DEALLOCATE (qs_force(ikind)%overlap)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%overlap_admm)) THEN
            DEALLOCATE (qs_force(ikind)%overlap_admm)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%rho_core)) THEN
            DEALLOCATE (qs_force(ikind)%rho_core)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%rho_elec)) THEN
            DEALLOCATE (qs_force(ikind)%rho_elec)
         END IF
         IF (ASSOCIATED(qs_force(ikind)%rho_lri_elec)) THEN
            DEALLOCATE (qs_force(ikind)%rho_lri_elec)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%vhxc_atom)) THEN
            DEALLOCATE (qs_force(ikind)%vhxc_atom)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%g0s_Vh_elec)) THEN
            DEALLOCATE (qs_force(ikind)%g0s_Vh_elec)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%repulsive)) THEN
            DEALLOCATE (qs_force(ikind)%repulsive)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%dispersion)) THEN
            DEALLOCATE (qs_force(ikind)%dispersion)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%other)) THEN
            DEALLOCATE (qs_force(ikind)%other)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%total)) THEN
            DEALLOCATE (qs_force(ikind)%total)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%ch_pulay)) THEN
            DEALLOCATE (qs_force(ikind)%ch_pulay)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%fock_4c)) THEN
            DEALLOCATE (qs_force(ikind)%fock_4c)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%mp2_sep)) THEN
            DEALLOCATE (qs_force(ikind)%mp2_sep)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%mp2_non_sep)) THEN
            DEALLOCATE (qs_force(ikind)%mp2_non_sep)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%ehrenfest)) THEN
            DEALLOCATE (qs_force(ikind)%ehrenfest)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%efield)) THEN
            DEALLOCATE (qs_force(ikind)%efield)
         END IF

         IF (ASSOCIATED(qs_force(ikind)%eev)) THEN
            DEALLOCATE (qs_force(ikind)%eev)
         END IF
      END DO

      DEALLOCATE (qs_force)

   END SUBROUTINE deallocate_qs_force

! **************************************************************************************************
!> \brief    Initialize a Quickstep force data structure.
!> \param qs_force ...
!> \date    15.07.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE zero_qs_force(qs_force)

      TYPE(qs_force_type), DIMENSION(:), POINTER         :: qs_force

      CHARACTER(len=*), PARAMETER :: routineN = 'zero_qs_force', routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      IF(.NOT.(ASSOCIATED(qs_force)))CALL cp__a("qs_force_types.F",252)

      DO ikind = 1, SIZE(qs_force)
         qs_force(ikind)%all_potential(:, :) = 0.0_dp
         qs_force(ikind)%core_overlap(:, :) = 0.0_dp
         qs_force(ikind)%gth_ppl(:, :) = 0.0_dp
         qs_force(ikind)%gth_nlcc(:, :) = 0.0_dp
         qs_force(ikind)%gth_ppnl(:, :) = 0.0_dp
         qs_force(ikind)%kinetic(:, :) = 0.0_dp
         qs_force(ikind)%overlap(:, :) = 0.0_dp
         qs_force(ikind)%overlap_admm(:, :) = 0.0_dp
         qs_force(ikind)%rho_core(:, :) = 0.0_dp
         qs_force(ikind)%rho_elec(:, :) = 0.0_dp
         qs_force(ikind)%rho_lri_elec(:, :) = 0.0_dp
         qs_force(ikind)%vhxc_atom(:, :) = 0.0_dp
         qs_force(ikind)%g0s_Vh_elec(:, :) = 0.0_dp
         qs_force(ikind)%repulsive(:, :) = 0.0_dp
         qs_force(ikind)%dispersion(:, :) = 0.0_dp
         qs_force(ikind)%other(:, :) = 0.0_dp
         qs_force(ikind)%fock_4c(:, :) = 0.0_dp
         qs_force(ikind)%ehrenfest(:, :) = 0.0_dp
         qs_force(ikind)%efield(:, :) = 0.0_dp
         qs_force(ikind)%eev(:, :) = 0.0_dp
         qs_force(ikind)%mp2_non_sep(:, :) = 0.0_dp
         qs_force(ikind)%mp2_sep(:, :) = 0.0_dp
         qs_force(ikind)%total(:, :) = 0.0_dp
      END DO

   END SUBROUTINE zero_qs_force

! **************************************************************************************************
!> \brief Add force to a force_type  variable.
!> \param force Input force, dimension (3,natom)
!> \param qs_force The force type variable to be used
!> \param forcetype ...
!> \param atomic_kind_set ...
!> \par History
!>      07.2014 JGH
!> \author JGH
! **************************************************************************************************
   SUBROUTINE add_qs_force(force, qs_force, forcetype, atomic_kind_set)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: force
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: qs_force
      CHARACTER(LEN=*), INTENT(IN)                       :: forcetype
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'add_qs_force', routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, iatom, ikind, natom_kind
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

!   ------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(qs_force)))CALL cp__a("qs_force_types.F",306)

      SELECT CASE (forcetype)
      CASE ("overlap_admm")
         DO ikind = 1, SIZE(atomic_kind_set, 1)
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, natom=natom_kind)
            DO ia = 1, natom_kind
               iatom = atomic_kind%atom_list(ia)
               qs_force(ikind)%overlap_admm(:, ia) = qs_force(ikind)%overlap_admm(:, ia)+force(:, iatom)
            END DO
         END DO
      CASE DEFAULT
         CALL cp__b("qs_force_types.F",319,"")
      END SELECT

   END SUBROUTINE add_qs_force

END MODULE qs_force_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize the collective variables types
!> \par History
!>      5.2004 created [fawzi and alessandro]
!>      1.2009 Fabio Sterpone : added the population COLVAR
!> \author Teodoro Laino
! **************************************************************************************************
MODULE colvar_types

   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'colvar_types'

   INTEGER, PARAMETER, PUBLIC               :: plane_def_atoms = 0, &
                                               plane_def_vec = 1

   INTEGER, PARAMETER, PUBLIC               :: do_clv_geo_center = 0, &
                                               do_clv_fix_point = 1, &
                                               do_clv_xyz = 0, &
                                               do_clv_x = 1, &
                                               do_clv_y = 2, &
                                               do_clv_z = 3, &
                                               do_clv_xy = 4, &
                                               do_clv_xz = 5, &
                                               do_clv_yz = 6
   PUBLIC :: colvar_type, &
             colvar_p_type, &
             colvar_p_reallocate, &
             colvar_p_release, &
             colvar_create, &
             colvar_clone, &
             colvar_setup, &
             colvar_release, &
             colvar_counters, &
             eval_point_der, &
             eval_point_pos, &
             eval_point_mass, &
             diff_colvar

   INTEGER, PARAMETER, PUBLIC :: no_colvar_id = -2, &
                                 dist_colvar_id = 1, &
                                 coord_colvar_id = 2, &
                                 torsion_colvar_id = 3, &
                                 angle_colvar_id = 4, &
                                 plane_distance_colvar_id = 5, &
                                 rotation_colvar_id = 6, &
                                 dfunct_colvar_id = 7, &
                                 qparm_colvar_id = 8, &
                                 hydronium_colvar_id = 9, &
                                 reaction_path_colvar_id = 10, &
                                 combine_colvar_id = 11, &
                                 population_colvar_id = 12, &
                                 plane_plane_angle_colvar_id = 13, &
                                 gyration_colvar_id = 14, &
                                 rmsd_colvar_id = 15, &
                                 distance_from_path_colvar_id = 16, &
                                 xyz_diag_colvar_id = 17, &
                                 xyz_outerdiag_colvar_id = 18, &
                                 u_colvar_id = 19, &
                                 Wc_colvar_id = 20, &
                                 hbp_colvar_id = 21, &
                                 ring_puckering_colvar_id = 22, &
                                 mindist_colvar_id = 23, &
                                 acid_hyd_dist_colvar_id = 24, &
                                 acid_hyd_shell_colvar_id = 25

! **************************************************************************************************
!> \brief parameters for the distance collective variable
!> \param i_at ,j_at: indexes of the two atoms between which you calculate
!>        the distance
!> \author alessandro laio and fawzi mohamed
! **************************************************************************************************
   TYPE dist_colvar_type
      INTEGER       :: i_at, j_at, axis_id
   END TYPE dist_colvar_type

! **************************************************************************************************
   TYPE coord_colvar_type
      LOGICAL                    :: do_chain, use_kinds_from, use_kinds_to, &
                                    use_kinds_to_b
      INTEGER                         :: n_atoms_to, &
                                         n_atoms_from, &
                                         nncrd, &
                                         ndcrd, &
                                         n_atoms_to_b, &
                                         nncrd_b, &
                                         ndcrd_b
      INTEGER, POINTER, DIMENSION(:) :: i_at_from, &
                                        i_at_to, &
                                        i_at_to_b
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: c_kinds_from, &
                                                                     c_kinds_to, &
                                                                     c_kinds_to_b
      REAL(KIND=dp)                   :: r_0, r_0_b
   END TYPE coord_colvar_type

! **************************************************************************************************
   TYPE population_colvar_type
      LOGICAL                         :: use_kinds_from, use_kinds_to
      INTEGER                         :: n_atoms_to, &
                                         n_atoms_from, &
                                         nncrd, &
                                         ndcrd, &
                                         n0
      INTEGER, POINTER, DIMENSION(:) :: i_at_from, &
                                        i_at_to
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: c_kinds_from, &
                                                                     c_kinds_to
      REAL(KIND=dp)                   :: r_0, sigma
   END TYPE population_colvar_type

! **************************************************************************************************
   TYPE gyration_colvar_type
      LOGICAL                         :: use_kinds
      INTEGER                         :: n_atoms
      INTEGER, POINTER, DIMENSION(:) :: i_at
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: c_kinds
   END TYPE gyration_colvar_type

! **************************************************************************************************
   TYPE torsion_colvar_type
      REAL(KIND=dp)                   :: o0
      INTEGER, DIMENSION(4)          :: i_at_tors
   END TYPE torsion_colvar_type

! **************************************************************************************************
   TYPE plane_distance_colvar_type
      LOGICAL               :: use_pbc
      INTEGER, DIMENSION(3) :: plane
      INTEGER               :: point
   END TYPE plane_distance_colvar_type

! **************************************************************************************************
   TYPE plane_def_type
      INTEGER                     :: type_of_def
      INTEGER, DIMENSION(3)       :: points
      REAL(KIND=dp), DIMENSION(3) :: normal_vec
   END TYPE plane_def_type

   TYPE plane_plane_angle_colvar_type
      TYPE(plane_def_type)        :: plane1, plane2
   END TYPE plane_plane_angle_colvar_type

! **************************************************************************************************
   TYPE angle_colvar_type
      INTEGER, DIMENSION(3) :: i_at_angle
   END TYPE angle_colvar_type

! **************************************************************************************************
   TYPE rotation_colvar_type
      INTEGER :: i_at1_bond1, &
                 i_at2_bond1, &
                 i_at1_bond2, &
                 i_at2_bond2
   END TYPE rotation_colvar_type

! **************************************************************************************************
   TYPE dfunct_colvar_type
      INTEGER, DIMENSION(4)         :: i_at_dfunct
      LOGICAL                       :: use_pbc
      REAL(KIND=dp)                 :: coeff
   END TYPE dfunct_colvar_type

! **************************************************************************************************
   TYPE qparm_colvar_type
      INTEGER                         :: l
      INTEGER                         :: n_atoms_to, &
                                         n_atoms_from
      INTEGER, POINTER, DIMENSION(:) :: i_at_from, &
                                        i_at_to
      REAL(KIND=dp)                   :: rcut, alpha
   END TYPE qparm_colvar_type

! **************************************************************************************************
   TYPE hydronium_colvar_type
      INTEGER                         :: n_oxygens, &
                                         n_hydrogens, &
                                         poh, qoh, poo, qoo, &
                                         pm, qm
      INTEGER, POINTER, DIMENSION(:)  :: i_oxygens, &
                                         i_hydrogens
      REAL(KIND=dp)                   :: roo, roh, lambda, nh
   END TYPE hydronium_colvar_type
! **************************************************************************************************
   TYPE acid_hyd_dist_colvar_type
      INTEGER                         :: n_oxygens_water, &
                                         n_oxygens_acid, &
                                         n_hydrogens, &
                                         pwoh, qwoh, paoh, qaoh, pcut, qcut
      INTEGER, POINTER, DIMENSION(:)  :: i_oxygens_water, i_oxygens_acid, &
                                         i_hydrogens
      REAL(KIND=dp)                   :: rwoh, raoh, lambda, nc
   END TYPE acid_hyd_dist_colvar_type

! **************************************************************************************************
   TYPE acid_hyd_shell_colvar_type
      INTEGER                         :: n_oxygens_water, &
                                         n_oxygens_acid, &
                                         n_hydrogens, &
                                         pwoh, qwoh, paoh, qaoh, &
                                         poo, qoo, pcut, qcut, pm, qm
      INTEGER, POINTER, DIMENSION(:)  :: i_oxygens_water, i_oxygens_acid, &
                                         i_hydrogens
      REAL(KIND=dp)                   :: rwoh, raoh, roo, lambda, nc, nh
   END TYPE acid_hyd_shell_colvar_type

! **************************************************************************************************
   TYPE reaction_path_colvar_type
      INTEGER                                  :: type_id
      INTEGER                                  :: n_components, nr_frames, subset
      INTEGER, DIMENSION(2)                     :: function_bounds
      INTEGER, POINTER, DIMENSION(:)          :: i_rmsd
      LOGICAL                                  :: align_frames, dist_rmsd, rmsd
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: f_vals
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: r_ref
      REAL(KIND=dp)                            :: lambda
      REAL(KIND=dp)                            :: step_size
      TYPE(colvar_p_type), POINTER, DIMENSION(:) :: colvar_p
   END TYPE reaction_path_colvar_type

! **************************************************************************************************
   TYPE combine_colvar_type
      INTEGER                                  :: type_id
      TYPE(colvar_p_type), POINTER, DIMENSION(:) :: colvar_p
      REAL(KIND=dp)                            :: lerr, dx
      CHARACTER(LEN=default_path_length)       :: FUNCTION
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER               :: c_parameters, variables
      REAL(KIND=dp), DIMENSION(:), POINTER     :: v_parameters
   END TYPE combine_colvar_type
! **************************************************************************************************
   TYPE rmsd_colvar_type
      INTEGER                                  :: n_atoms, nr_frames, subset
      INTEGER, POINTER, DIMENSION(:)          :: i_rmsd
      LOGICAL                                  :: align_frames
      REAL(KIND=dp), DIMENSION(:), POINTER     :: weights
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: r_ref
   END TYPE rmsd_colvar_type

! **************************************************************************************************
   TYPE point_type
      INTEGER :: type_id
      INTEGER, DIMENSION(:), POINTER          :: atoms
      REAL(KIND=dp), DIMENSION(:), POINTER    :: weights
      REAL(KIND=dp), DIMENSION(3)              :: r
   END TYPE point_type

! **************************************************************************************************
   TYPE xyz_diag_colvar_type
      LOGICAL                                  :: use_pbc
      LOGICAL                                  :: use_absolute_position
      INTEGER                                  :: i_atom
      INTEGER                                  :: component
      REAL(KIND=dp), DIMENSION(3)              :: r0
   END TYPE xyz_diag_colvar_type

! **************************************************************************************************
   TYPE xyz_outerdiag_colvar_type
      LOGICAL                                  :: use_pbc
      INTEGER, DIMENSION(2)                    :: i_atoms
      INTEGER, DIMENSION(2)                    :: components
      REAL(KIND=dp), DIMENSION(3, 2)            :: r0
   END TYPE xyz_outerdiag_colvar_type

! **************************************************************************************************
   TYPE u_colvar_type
      TYPE(section_vals_type), POINTER         :: mixed_energy_section
      INTEGER       :: natom
   END TYPE u_colvar_type

! **************************************************************************************************
   TYPE Wc_colvar_type
      INTEGER                         :: ids(3) ! first is the Od, second the H, third the Oa
      REAL(KIND=dp)                   :: ewc
      REAL(KIND=dp)                   :: rcut
   END TYPE Wc_colvar_type

! **************************************************************************************************
   TYPE HBP_colvar_type
      INTEGER                         :: nPoints ! number of the points in the path
      INTEGER, POINTER                :: ids(:, :) ! first is the Od, second the H,
      ! third the Oa and contains a row for each intermediate point in the path
      REAL(KIND=dp), POINTER           :: ewc(:) ! one for each point in the path
      REAL(KIND=dp)                   :: rcut
      REAL(KIND=dp)                   :: shift ! shift applied for each term in the collective variable
   END TYPE HBP_colvar_type

! **************************************************************************************************
   TYPE ring_puckering_colvar_type
      INTEGER                         :: nring
      INTEGER, POINTER, DIMENSION(:) :: atoms
      INTEGER                         :: iq
   END TYPE ring_puckering_colvar_type

! **************************************************************************************************
   TYPE mindist_colvar_type
      LOGICAL                    :: use_kinds_from, use_kinds_to
      INTEGER                         :: n_coord_to, &
                                         n_coord_from, &
                                         n_dist_from, &
                                         p_exp, &
                                         q_exp
      INTEGER, POINTER, DIMENSION(:) :: i_coord_from, &
                                        i_coord_to, &
                                        i_dist_from
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: k_coord_from, &
                                                                     k_coord_to
      REAL(KIND=dp)                   :: lambda, r_cut
   END TYPE mindist_colvar_type

! **************************************************************************************************
!> \brief parameters for a collective variable
!> \author alessandro laio and fawzi mohamed
! **************************************************************************************************
   TYPE colvar_type
      INTEGER                                     :: type_id
      LOGICAL                                     :: use_points
      REAL(kind=dp)                               :: ss ! Value of the colvar
      REAL(kind=dp), DIMENSION(:, :), POINTER     :: dsdr ! Derivative of colvar (3,:)
      INTEGER, DIMENSION(:), POINTER             :: i_atom ! Mapping of dsdr
      INTEGER                                     :: n_atom_s
      ! Available COLVAR types
      TYPE(dist_colvar_type), POINTER              :: dist_param
      TYPE(coord_colvar_type), POINTER             :: coord_param
      TYPE(population_colvar_type), POINTER        :: population_param
      TYPE(gyration_colvar_type), POINTER          :: gyration_param
      TYPE(torsion_colvar_type), POINTER           :: torsion_param
      TYPE(angle_colvar_type), POINTER             :: angle_param
      TYPE(plane_distance_colvar_type), POINTER    :: plane_distance_param
      TYPE(plane_plane_angle_colvar_type), POINTER :: plane_plane_angle_param
      TYPE(rotation_colvar_type), POINTER         :: rotation_param
      TYPE(dfunct_colvar_type), POINTER           :: dfunct_param
      TYPE(qparm_colvar_type), POINTER            :: qparm_param
      TYPE(hydronium_colvar_type), POINTER        :: hydronium_param
      TYPE(acid_hyd_dist_colvar_type), POINTER    :: acid_hyd_dist_param
      TYPE(acid_hyd_shell_colvar_type), POINTER   :: acid_hyd_shell_param
      TYPE(reaction_path_colvar_type), POINTER     :: reaction_path_param
      TYPE(combine_colvar_type), POINTER          :: combine_cvs_param
      TYPE(rmsd_colvar_type), POINTER             :: rmsd_param
      TYPE(xyz_diag_colvar_type), POINTER         :: xyz_diag_param
      TYPE(xyz_outerdiag_colvar_type), POINTER    :: xyz_outerdiag_param
      TYPE(u_colvar_type), POINTER                 :: u_param
      TYPE(point_type), DIMENSION(:), POINTER     :: points
      TYPE(Wc_colvar_type), POINTER               :: Wc
      TYPE(HBP_colvar_type), POINTER              :: HBP
      TYPE(ring_puckering_colvar_type), POINTER   :: ring_puckering_param
      TYPE(mindist_colvar_type), POINTER          :: mindist_param
   END TYPE colvar_type

! **************************************************************************************************
   TYPE colvar_p_type
      TYPE(colvar_type), POINTER :: colvar
   END TYPE colvar_p_type

! **************************************************************************************************
   TYPE colvar_counters
      INTEGER :: ndist
      INTEGER :: nangle
      INTEGER :: ntorsion
      INTEGER :: ncoord
      INTEGER :: nplane_dist
      INTEGER :: nplane_angle
      INTEGER :: nrot
      INTEGER :: ndfunct
      INTEGER :: nqparm
      INTEGER :: nhydronium
      INTEGER :: nacid_hyd_dist
      INTEGER :: nacid_hyd_shell
      INTEGER :: nreactionpath
      INTEGER :: ncombinecvs
      INTEGER :: nrestraint
      INTEGER :: npopulation
      INTEGER :: ngyration
      INTEGER :: nrmsd
      INTEGER :: nxyz_diag
      INTEGER :: nxyz_outerdiag
      INTEGER :: ntot
      INTEGER :: nWc
      INTEGER :: nHBP
      INTEGER :: nring_puckering
   END TYPE colvar_counters

CONTAINS

! **************************************************************************************************
!> \brief initializes a colvar_param type
!> \param colvar the colvat to initialize
!> \param colvar_id ...
!> \author alessandro laio and fawzi mohamed
! **************************************************************************************************
   SUBROUTINE colvar_create(colvar, colvar_id)
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER, INTENT(in)                                :: colvar_id

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(colvar)))CALL cp__a("subsys/colvar_types.F",411)
      ALLOCATE (colvar)
      ! Nullify all available COLVAR types
      NULLIFY (colvar%dist_param)
      NULLIFY (colvar%angle_param)
      NULLIFY (colvar%torsion_param)
      NULLIFY (colvar%plane_distance_param)
      NULLIFY (colvar%plane_plane_angle_param)
      NULLIFY (colvar%rotation_param)
      NULLIFY (colvar%dfunct_param)
      NULLIFY (colvar%coord_param)
      NULLIFY (colvar%population_param)
      NULLIFY (colvar%gyration_param)
      NULLIFY (colvar%qparm_param)
      NULLIFY (colvar%xyz_diag_param)
      NULLIFY (colvar%xyz_outerdiag_param)
      NULLIFY (colvar%u_param)
      NULLIFY (colvar%hydronium_param)
      NULLIFY (colvar%acid_hyd_dist_param)
      NULLIFY (colvar%acid_hyd_shell_param)
      NULLIFY (colvar%reaction_path_param)
      NULLIFY (colvar%combine_cvs_param)
      NULLIFY (colvar%rmsd_param)
      ! More types to nullify
      NULLIFY (colvar%points)
      NULLIFY (colvar%dsdr)
      NULLIFY (colvar%i_atom)
      NULLIFY (colvar%Wc)
      NULLIFY (colvar%ring_puckering_param)
      NULLIFY (colvar%mindist_param)
      colvar%type_id = colvar_id
      colvar%use_points = .FALSE.
      SELECT CASE (colvar_id)
      CASE (dist_colvar_id)
         ALLOCATE (colvar%dist_param)
         colvar%dist_param%axis_id = do_clv_xyz
      CASE (coord_colvar_id)
         ALLOCATE (colvar%coord_param)
      CASE (population_colvar_id)
         ALLOCATE (colvar%population_param)
      CASE (gyration_colvar_id)
         ALLOCATE (colvar%gyration_param)
      CASE (angle_colvar_id)
         ALLOCATE (colvar%angle_param)
      CASE (torsion_colvar_id)
         ALLOCATE (colvar%torsion_param)
      CASE (plane_distance_colvar_id)
         ALLOCATE (colvar%plane_distance_param)
      CASE (plane_plane_angle_colvar_id)
         ALLOCATE (colvar%plane_plane_angle_param)
      CASE (rotation_colvar_id)
         ALLOCATE (colvar%rotation_param)
      CASE (dfunct_colvar_id)
         ALLOCATE (colvar%dfunct_param)
      CASE (qparm_colvar_id)
         ALLOCATE (colvar%qparm_param)
      CASE (xyz_diag_colvar_id)
         ALLOCATE (colvar%xyz_diag_param)
         ! Initialize r0 with dummy..
         colvar%xyz_diag_param%r0 = HUGE(0.0_dp)
      CASE (xyz_outerdiag_colvar_id)
         ALLOCATE (colvar%xyz_outerdiag_param)
         ! Initialize r0 with dummy..
         colvar%xyz_outerdiag_param%r0 = HUGE(0.0_dp)
      CASE (u_colvar_id)
         ALLOCATE (colvar%u_param)
         NULLIFY (colvar%u_param%mixed_energy_section)
      CASE (hydronium_colvar_id)
         ALLOCATE (colvar%hydronium_param)
      CASE (acid_hyd_dist_colvar_id)
         ALLOCATE (colvar%acid_hyd_dist_param)
      CASE (acid_hyd_shell_colvar_id)
         ALLOCATE (colvar%acid_hyd_shell_param)
      CASE (reaction_path_colvar_id)
         ALLOCATE (colvar%reaction_path_param)
      CASE (distance_from_path_colvar_id)
         ALLOCATE (colvar%reaction_path_param)
      CASE (combine_colvar_id)
         ALLOCATE (colvar%combine_cvs_param)
      CASE (rmsd_colvar_id)
         ALLOCATE (colvar%rmsd_param)
      CASE (Wc_colvar_id)
         ALLOCATE (colvar%Wc)
      CASE (HBP_colvar_id)
         ALLOCATE (colvar%HBP)
      CASE (ring_puckering_colvar_id)
         ALLOCATE (colvar%ring_puckering_param)
      CASE (mindist_colvar_id)
         ALLOCATE (colvar%mindist_param)
      CASE (no_colvar_id)
         ! Do nothing
      CASE DEFAULT
         CALL cp__b("subsys/colvar_types.F",503,"")
      END SELECT

   END SUBROUTINE colvar_create

! **************************************************************************************************
!> \brief Finalize the setup of the collective variable
!> \param colvar the colvar to initialize
!> \author Teodoro Laino, [teo] 09.03.2006
! **************************************************************************************************
   SUBROUTINE colvar_setup(colvar)
      TYPE(colvar_type), POINTER                         :: colvar

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idum, iend, ii, istart, j, np, stat
      INTEGER, DIMENSION(:), POINTER                     :: list

      IF(.NOT.(ASSOCIATED(colvar)))CALL cp__a("subsys/colvar_types.F",521)
      SELECT CASE (colvar%type_id)
      CASE (dist_colvar_id)
         np = 2
         i = colvar%dist_param%i_at
         j = colvar%dist_param%j_at
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = COLV_SIZE(colvar, i)+ &
                           COLV_SIZE(colvar, j)
         ! Create a List of points...
         ALLOCATE (list(np))
         list(1) = colvar%dist_param%i_at
         list(2) = colvar%dist_param%j_at
      CASE (coord_colvar_id)
         np = colvar%coord_param%n_atoms_from+colvar%coord_param%n_atoms_to &
              +colvar%coord_param%n_atoms_to_b
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%coord_param%n_atoms_from
            i = colvar%coord_param%i_at_from(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         DO ii = 1, colvar%coord_param%n_atoms_to
            i = colvar%coord_param%i_at_to(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         IF (colvar%coord_param%n_atoms_to_b /= 0) THEN
            DO ii = 1, colvar%coord_param%n_atoms_to_b
               i = colvar%coord_param%i_at_to_b(ii)
               colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
            ENDDO
         END IF
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%coord_param%n_atoms_from
            idum = idum+1
            i = colvar%coord_param%i_at_from(ii)
            list(idum) = i
         ENDDO
         DO ii = 1, colvar%coord_param%n_atoms_to
            idum = idum+1
            i = colvar%coord_param%i_at_to(ii)
            list(idum) = i
         ENDDO
         IF (colvar%coord_param%n_atoms_to_b /= 0) THEN
            DO ii = 1, colvar%coord_param%n_atoms_to_b
               idum = idum+1
               i = colvar%coord_param%i_at_to_b(ii)
               list(idum) = i
            ENDDO
         END IF
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",573)
      CASE (population_colvar_id)
         np = colvar%population_param%n_atoms_from+colvar%population_param%n_atoms_to
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%population_param%n_atoms_from
            i = colvar%population_param%i_at_from(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         DO ii = 1, colvar%population_param%n_atoms_to
            i = colvar%population_param%i_at_to(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%population_param%n_atoms_from
            idum = idum+1
            i = colvar%population_param%i_at_from(ii)
            list(idum) = i
         ENDDO
         DO ii = 1, colvar%population_param%n_atoms_to
            idum = idum+1
            i = colvar%population_param%i_at_to(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",599)
      CASE (gyration_colvar_id)
         np = colvar%gyration_param%n_atoms
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%gyration_param%n_atoms
            i = colvar%gyration_param%i_at(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%gyration_param%n_atoms
            idum = idum+1
            i = colvar%gyration_param%i_at(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",616)
      CASE (angle_colvar_id)
         np = 3
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, 3
            i = colvar%angle_param%i_at_angle(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, 3
            idum = idum+1
            i = colvar%angle_param%i_at_angle(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",633)
      CASE (torsion_colvar_id)
         np = 4
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, 4
            i = colvar%torsion_param%i_at_tors(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, 4
            idum = idum+1
            i = colvar%torsion_param%i_at_tors(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",650)
      CASE (plane_distance_colvar_id)
         np = 4
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, 3
            i = colvar%plane_distance_param%plane(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         i = colvar%plane_distance_param%point
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, 3
            idum = idum+1
            i = colvar%plane_distance_param%plane(ii)
            list(idum) = i
         ENDDO
         i = colvar%plane_distance_param%point
         list(4) = i
         idum = idum+1
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",672)
      CASE (plane_plane_angle_colvar_id)
         np = 0
         IF (colvar%plane_plane_angle_param%plane1%type_of_def == plane_def_atoms) np = np+3
         IF (colvar%plane_plane_angle_param%plane2%type_of_def == plane_def_atoms) np = np+3
         ! if np is equal to zero this means that this is not a COLLECTIVE variable..
         IF (np == 0) &
            CALL cp_abort(cp__l("subsys/colvar_types.F",679), &
                          "PLANE_PLANE_ANGLE Colvar defined using two normal vectors! This is "// &
                          "not a COLLECTIVE VARIABLE! One of the two planes must be defined "// &
                          "using atomic positions.")

         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         IF (colvar%plane_plane_angle_param%plane1%type_of_def == plane_def_atoms) THEN
            DO ii = 1, 3
               i = colvar%plane_plane_angle_param%plane1%points(ii)
               colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
            ENDDO
         END IF
         IF (colvar%plane_plane_angle_param%plane2%type_of_def == plane_def_atoms) THEN
            DO ii = 1, 3
               i = colvar%plane_plane_angle_param%plane2%points(ii)
               colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
            ENDDO
         END IF

         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         IF (colvar%plane_plane_angle_param%plane1%type_of_def == plane_def_atoms) THEN
            DO ii = 1, 3
               idum = idum+1
               i = colvar%plane_plane_angle_param%plane1%points(ii)
               list(idum) = i
            ENDDO
         END IF
         IF (colvar%plane_plane_angle_param%plane2%type_of_def == plane_def_atoms) THEN
            DO ii = 1, 3
               idum = idum+1
               i = colvar%plane_plane_angle_param%plane2%points(ii)
               list(idum) = i
            ENDDO
         END IF
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",716)
      CASE (dfunct_colvar_id)
         np = 4
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, 4
            i = colvar%dfunct_param%i_at_dfunct(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, 4
            idum = idum+1
            i = colvar%dfunct_param%i_at_dfunct(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",733)
      CASE (rotation_colvar_id)
         np = 4
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         i = colvar%rotation_param%i_at1_bond1
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         i = colvar%rotation_param%i_at2_bond1
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         i = colvar%rotation_param%i_at1_bond2
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         i = colvar%rotation_param%i_at2_bond2
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ! Create a List of points...
         ALLOCATE (list(np))
         i = colvar%rotation_param%i_at1_bond1
         list(1) = i
         i = colvar%rotation_param%i_at2_bond1
         list(2) = i
         i = colvar%rotation_param%i_at1_bond2
         list(3) = i
         i = colvar%rotation_param%i_at2_bond2
         list(4) = i
      CASE (qparm_colvar_id)
         np = colvar%qparm_param%n_atoms_from+colvar%qparm_param%n_atoms_to
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%qparm_param%n_atoms_from
            i = colvar%qparm_param%i_at_from(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         DO ii = 1, colvar%qparm_param%n_atoms_to
            i = colvar%qparm_param%i_at_to(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%qparm_param%n_atoms_from
            idum = idum+1
            i = colvar%qparm_param%i_at_from(ii)
            list(idum) = i
         ENDDO
         DO ii = 1, colvar%qparm_param%n_atoms_to
            idum = idum+1
            i = colvar%qparm_param%i_at_to(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",781)
      CASE (hydronium_colvar_id)
         np = colvar%hydronium_param%n_oxygens+colvar%hydronium_param%n_hydrogens
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%hydronium_param%n_oxygens
            i = colvar%hydronium_param%i_oxygens(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         DO ii = 1, colvar%hydronium_param%n_hydrogens
            i = colvar%hydronium_param%i_hydrogens(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%hydronium_param%n_oxygens
            idum = idum+1
            i = colvar%hydronium_param%i_oxygens(ii)
            list(idum) = i
            IF (ANY(colvar%hydronium_param%i_hydrogens == i)) &
               CALL cp__b("subsys/colvar_types.F",802,"COLVAR: atoms doubled in OXYGENS and HYDROGENS list")
         ENDDO
         DO ii = 1, colvar%hydronium_param%n_hydrogens
            idum = idum+1
            i = colvar%hydronium_param%i_hydrogens(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",809)
         DO i = 1, np
            DO ii = i+1, np
               IF (list(i) == list(ii)) THEN
                  IF (i <= colvar%hydronium_param%n_oxygens) &
                     CALL cp__b("subsys/colvar_types.F",814,"atoms doubled in OXYGENS list")
                  IF (i > colvar%hydronium_param%n_oxygens) &
                     CALL cp__b("subsys/colvar_types.F",816,"atoms doubled in HYDROGENS list")
               ENDIF
            ENDDO
         ENDDO
      CASE (acid_hyd_dist_colvar_id)
         np = colvar%acid_hyd_dist_param%n_oxygens_water &
              +colvar%acid_hyd_dist_param%n_oxygens_acid &
              +colvar%acid_hyd_dist_param%n_hydrogens
         ALLOCATE (list(np))
         CALL setup_acid_hydronium_colvars(colvar, acid_hyd_dist_colvar_id, list)
      CASE (acid_hyd_shell_colvar_id)
         np = colvar%acid_hyd_shell_param%n_oxygens_water &
              +colvar%acid_hyd_shell_param%n_oxygens_acid &
              +colvar%acid_hyd_shell_param%n_hydrogens
         ALLOCATE (list(np))
         CALL setup_acid_hydronium_colvars(colvar, acid_hyd_shell_colvar_id, list)
      CASE (rmsd_colvar_id)
         np = colvar%rmsd_param%n_atoms
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%rmsd_param%n_atoms
            i = colvar%rmsd_param%i_rmsd(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%rmsd_param%n_atoms
            idum = idum+1
            i = colvar%rmsd_param%i_rmsd(ii)
            list(idum) = i
         END DO
      CASE (reaction_path_colvar_id, distance_from_path_colvar_id)
         colvar%n_atom_s = 0
         IF (colvar%reaction_path_param%dist_rmsd .OR. colvar%reaction_path_param%rmsd) THEN
            colvar%n_atom_s = colvar%reaction_path_param%n_components
         ELSE
            DO ii = 1, SIZE(colvar%reaction_path_param%colvar_p)
               colvar%n_atom_s = colvar%n_atom_s+colvar%reaction_path_param%colvar_p(ii)%colvar%n_atom_s
            END DO
         END IF
         ALLOCATE (list(colvar%n_atom_s))
         idum = 0
         IF (colvar%reaction_path_param%dist_rmsd .OR. colvar%reaction_path_param%rmsd) THEN
            DO ii = 1, SIZE(colvar%reaction_path_param%i_rmsd)
               idum = idum+1
               i = colvar%reaction_path_param%i_rmsd(ii)
               list(idum) = i
            END DO
         ELSE
            DO ii = 1, SIZE(colvar%reaction_path_param%colvar_p)
               DO j = 1, colvar%reaction_path_param%colvar_p(ii)%colvar%n_atom_s
                  idum = idum+1
                  list(idum) = colvar%reaction_path_param%colvar_p(ii)%colvar%i_atom(j)
               END DO
            END DO
         END IF
      CASE (xyz_diag_colvar_id)
         np = 1
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         i = colvar%xyz_diag_param%i_atom
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ! Create a List of points...
         ALLOCATE (list(np))
         i = colvar%xyz_diag_param%i_atom
         list(1) = i
      CASE (xyz_outerdiag_colvar_id)
         np = 2
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         i = colvar%xyz_outerdiag_param%i_atoms(1)
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         i = colvar%xyz_outerdiag_param%i_atoms(2)
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ! Create a List of points...
         ALLOCATE (list(np))
         i = colvar%xyz_outerdiag_param%i_atoms(1)
         list(1) = i
         i = colvar%xyz_outerdiag_param%i_atoms(2)
         list(2) = i
      CASE (u_colvar_id)
         np = 1; ALLOCATE (list(np), stat=stat)
         IF(.NOT.(stat == 0))CALL cp__a("subsys/colvar_types.F",899)
         colvar%n_atom_s = np; list(1) = 1
      CASE (Wc_colvar_id)
         np = 3
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, 3
            i = colvar%Wc%ids(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, 3
            idum = idum+1
            i = colvar%Wc%ids(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",917)
      CASE (HBP_colvar_id)
         np = 3*colvar%HBP%nPoints
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO j = 1, colvar%HBP%nPoints
            DO ii = 1, 3
               i = colvar%HBP%ids(j, ii)
               colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
            ENDDO
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO j = 1, colvar%HBP%nPoints
            DO ii = 1, 3
               idum = idum+1
               i = colvar%HBP%ids(j, ii)
               list(idum) = i
            ENDDO
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",938)
      CASE (ring_puckering_colvar_id)
         np = colvar%ring_puckering_param%nring
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%ring_puckering_param%nring
            i = colvar%ring_puckering_param%atoms(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%ring_puckering_param%nring
            idum = idum+1
            i = colvar%ring_puckering_param%atoms(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",955)
      CASE (mindist_colvar_id)
         np = colvar%mindist_param%n_dist_from+ &
              colvar%mindist_param%n_coord_from+colvar%mindist_param%n_coord_to
         ! Number of real atoms involved in the colvar
         colvar%n_atom_s = 0
         DO ii = 1, colvar%mindist_param%n_dist_from
            i = colvar%mindist_param%i_dist_from(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         DO ii = 1, colvar%mindist_param%n_coord_from
            i = colvar%mindist_param%i_coord_from(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         DO ii = 1, colvar%mindist_param%n_coord_to
            i = colvar%mindist_param%i_coord_to(ii)
            colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
         ENDDO
         ! Create a List of points...
         ALLOCATE (list(np))
         idum = 0
         DO ii = 1, colvar%mindist_param%n_dist_from
            idum = idum+1
            i = colvar%mindist_param%i_dist_from(ii)
            list(idum) = i
         ENDDO
         DO ii = 1, colvar%mindist_param%n_coord_from
            idum = idum+1
            i = colvar%mindist_param%i_coord_from(ii)
            list(idum) = i
         ENDDO
         DO ii = 1, colvar%mindist_param%n_coord_to
            idum = idum+1
            i = colvar%mindist_param%i_coord_to(ii)
            list(idum) = i
         ENDDO
         IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",991)
      CASE (combine_colvar_id)
         colvar%n_atom_s = 0
         DO ii = 1, SIZE(colvar%combine_cvs_param%colvar_p)
            colvar%n_atom_s = colvar%n_atom_s+colvar%combine_cvs_param%colvar_p(ii)%colvar%n_atom_s
         END DO
         ALLOCATE (list(colvar%n_atom_s))
         idum = 0
         DO ii = 1, SIZE(colvar%combine_cvs_param%colvar_p)
            DO j = 1, colvar%combine_cvs_param%colvar_p(ii)%colvar%n_atom_s
               idum = idum+1
               list(idum) = colvar%combine_cvs_param%colvar_p(ii)%colvar%i_atom(j)
            END DO
         END DO
      END SELECT

      IF (ASSOCIATED(colvar%dsdr)) THEN
         DEALLOCATE (colvar%dsdr)
      END IF
      IF (ASSOCIATED(colvar%i_atom)) THEN
         DEALLOCATE (colvar%i_atom)
      END IF
      ALLOCATE (colvar%dsdr(3, colvar%n_atom_s))
      ALLOCATE (colvar%i_atom(colvar%n_atom_s))
      ! And now map real atoms
      istart = 0
      iend = 0
      DO i = 1, SIZE(list)
         IF (.NOT. colvar%use_points) THEN
            ! No point centers
            colvar%i_atom(i) = list(i)
            iend = iend+1
         ELSE
            IF (ASSOCIATED(colvar%points(list(i))%atoms)) THEN
               iend = istart+SIZE(colvar%points(list(i))%atoms)
               colvar%i_atom(istart+1:iend) = colvar%points(list(i))%atoms
               istart = iend
            END IF
         END IF
      END DO
      IF(.NOT.(iend == colvar%n_atom_s))CALL cp__a("subsys/colvar_types.F",1031)
      DEALLOCATE (list)

   END SUBROUTINE colvar_setup

! **************************************************************************************************
!> \brief Finalize the setup of the collective variable for the dissociation
!>        of a carboxylic acid in water
!> \param colvar the colvar to initialize
!> \param colvar_id ...
!> \param list ...
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE setup_acid_hydronium_colvars(colvar, colvar_id, list)
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER, INTENT(IN)                                :: colvar_id
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_acid_hydronium_colvars', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idum, ii, n_hydrogens, &
                                                            n_oxygens_acid, n_oxygens_water, np
      INTEGER, DIMENSION(:), POINTER                     :: i_hydrogens, i_oxygens_acid, &
                                                            i_oxygens_water

      NULLIFY (i_oxygens_water, i_oxygens_acid, i_hydrogens)

      SELECT CASE (colvar_id)
      CASE (acid_hyd_dist_colvar_id)
         n_oxygens_water = colvar%acid_hyd_dist_param%n_oxygens_water
         n_oxygens_acid = colvar%acid_hyd_dist_param%n_oxygens_acid
         n_hydrogens = colvar%acid_hyd_dist_param%n_hydrogens
         i_oxygens_water => colvar%acid_hyd_dist_param%i_oxygens_water
         i_oxygens_acid => colvar%acid_hyd_dist_param%i_oxygens_acid
         i_hydrogens => colvar%acid_hyd_dist_param%i_hydrogens
      CASE (acid_hyd_shell_colvar_id)
         n_oxygens_water = colvar%acid_hyd_shell_param%n_oxygens_water
         n_oxygens_acid = colvar%acid_hyd_shell_param%n_oxygens_acid
         n_hydrogens = colvar%acid_hyd_shell_param%n_hydrogens
         i_oxygens_water => colvar%acid_hyd_shell_param%i_oxygens_water
         i_oxygens_acid => colvar%acid_hyd_shell_param%i_oxygens_acid
         i_hydrogens => colvar%acid_hyd_shell_param%i_hydrogens
      END SELECT

      np = n_oxygens_water+n_oxygens_acid+n_hydrogens
      ! Number of real atoms involved in the colvar
      colvar%n_atom_s = 0
      DO ii = 1, n_oxygens_water
         i = i_oxygens_water(ii)
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
      ENDDO
      DO ii = 1, n_oxygens_acid
         i = i_oxygens_acid(ii)
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
      ENDDO
      DO ii = 1, n_hydrogens
         i = i_hydrogens(ii)
         colvar%n_atom_s = colvar%n_atom_s+COLV_SIZE(colvar, i)
      ENDDO
      idum = 0
      DO ii = 1, n_oxygens_water
         idum = idum+1
         i = i_oxygens_water(ii)
         list(idum) = i
         IF (ANY(i_hydrogens == i)) &
            CALL cp__b("subsys/colvar_types.F",1097,"COLVAR: atoms doubled in OXYGENS_WATER and HYDROGENS list")
         IF (ANY(i_oxygens_acid == i)) &
            CALL cp__b("subsys/colvar_types.F",1099,"COLVAR: atoms doubled in OXYGENS_WATER and OXYGENS_ACID list")
      ENDDO
      DO ii = 1, n_oxygens_acid
         idum = idum+1
         i = i_oxygens_acid(ii)
         list(idum) = i
         IF (ANY(i_hydrogens == i)) &
            CALL cp__b("subsys/colvar_types.F",1106,"COLVAR: atoms doubled in OXYGENS_ACID and HYDROGENS list")
      ENDDO
      DO ii = 1, n_hydrogens
         idum = idum+1
         i = i_hydrogens(ii)
         list(idum) = i
      ENDDO
      IF(.NOT.(idum == np))CALL cp__a("subsys/colvar_types.F",1113)
      DO i = 1, np
         DO ii = i+1, np
            IF (list(i) == list(ii)) THEN
               IF (i <= n_oxygens_water) &
                  CALL cp__b("subsys/colvar_types.F",1118,"atoms doubled in OXYGENS_WATER list")
               IF (i > n_oxygens_water .AND. i <= n_oxygens_water+n_oxygens_acid) &
                  CALL cp__b("subsys/colvar_types.F",1120,"atoms doubled in OXYGENS_ACID list")
               IF (i > n_oxygens_water+n_oxygens_acid) &
                  CALL cp__b("subsys/colvar_types.F",1122,"atoms doubled in HYDROGENS list")
            ENDIF
         ENDDO
      ENDDO

   END SUBROUTINE setup_acid_hydronium_colvars

! **************************************************************************************************
!> \brief Gives back the size of an array of integer. If not associated gives back 1
!> \param colvar ...
!> \param i ...
!> \retval my_size ...
!> \author Teodoro Laino - 03.2007
! **************************************************************************************************
   FUNCTION colv_size(colvar, i) RESULT(my_size)
      TYPE(colvar_type), POINTER                         :: colvar
      INTEGER                                            :: i, my_size

      my_size = 1
      IF (ASSOCIATED(colvar%points)) THEN
         IF (ASSOCIATED(colvar%points(i)%atoms)) THEN
            my_size = SIZE(colvar%points(i)%atoms)
         ELSE
            my_size = 0
         END IF
      END IF
   END FUNCTION colv_size

! **************************************************************************************************
!> \brief releases the memory that might have been allocated by the colvar
!> \param colvar the colvar to deallocate
!> \author alessandro laio and fawzi mohamed
! **************************************************************************************************
   RECURSIVE SUBROUTINE colvar_release(colvar)
      TYPE(colvar_type), POINTER                         :: colvar

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(colvar)))CALL cp__a("subsys/colvar_types.F",1162)
      IF (ASSOCIATED(colvar%dsdr)) THEN
         DEALLOCATE (colvar%dsdr)
      END IF
      IF (ASSOCIATED(colvar%i_atom)) THEN
         DEALLOCATE (colvar%i_atom)
      END IF
      IF (ASSOCIATED(colvar%points)) THEN
         DO i = 1, SIZE(colvar%points)
            IF (ASSOCIATED(colvar%points(i)%atoms)) THEN
               DEALLOCATE (colvar%points(i)%atoms)
            END IF
            IF (ASSOCIATED(colvar%points(i)%weights)) THEN
               DEALLOCATE (colvar%points(i)%weights)
            END IF
         END DO
         DEALLOCATE (colvar%points)
      END IF
      SELECT CASE (colvar%type_id)
      CASE (dist_colvar_id)
         DEALLOCATE (colvar%dist_param)
      CASE (coord_colvar_id)
         IF (ASSOCIATED(colvar%coord_param%i_at_from)) THEN
            DEALLOCATE (colvar%coord_param%i_at_from)
         END IF
         IF (ASSOCIATED(colvar%coord_param%i_at_to)) THEN
            DEALLOCATE (colvar%coord_param%i_at_to)
         END IF
         IF (ASSOCIATED(colvar%coord_param%c_kinds_from)) THEN
            DEALLOCATE (colvar%coord_param%c_kinds_from)
         END IF
         IF (ASSOCIATED(colvar%coord_param%c_kinds_to)) THEN
            DEALLOCATE (colvar%coord_param%c_kinds_to)
         END IF
         IF (ASSOCIATED(colvar%coord_param%i_at_to_b)) THEN
            DEALLOCATE (colvar%coord_param%i_at_to_b)
         END IF
         IF (ASSOCIATED(colvar%coord_param%c_kinds_to_b)) THEN
            DEALLOCATE (colvar%coord_param%c_kinds_to_b)
         END IF
         DEALLOCATE (colvar%coord_param)
      CASE (population_colvar_id)
         IF (ASSOCIATED(colvar%population_param%i_at_from)) THEN
            DEALLOCATE (colvar%population_param%i_at_from)
         END IF
         IF (ASSOCIATED(colvar%population_param%i_at_to)) THEN
            DEALLOCATE (colvar%population_param%i_at_to)
         END IF
         IF (ASSOCIATED(colvar%population_param%c_kinds_from)) THEN
            DEALLOCATE (colvar%population_param%c_kinds_from)
         END IF
         IF (ASSOCIATED(colvar%population_param%c_kinds_to)) THEN
            DEALLOCATE (colvar%population_param%c_kinds_to)
         END IF
         DEALLOCATE (colvar%population_param)
      CASE (gyration_colvar_id)
         IF (ASSOCIATED(colvar%gyration_param%i_at)) THEN
            DEALLOCATE (colvar%gyration_param%i_at)
         END IF
         IF (ASSOCIATED(colvar%gyration_param%c_kinds)) THEN
            DEALLOCATE (colvar%gyration_param%c_kinds)
         END IF
         DEALLOCATE (colvar%gyration_param)
      CASE (angle_colvar_id)
         DEALLOCATE (colvar%angle_param)
      CASE (torsion_colvar_id)
         DEALLOCATE (colvar%torsion_param)
      CASE (plane_distance_colvar_id)
         DEALLOCATE (colvar%plane_distance_param)
      CASE (plane_plane_angle_colvar_id)
         DEALLOCATE (colvar%plane_plane_angle_param)
      CASE (dfunct_colvar_id)
         DEALLOCATE (colvar%dfunct_param)
      CASE (rotation_colvar_id)
         DEALLOCATE (colvar%rotation_param)
      CASE (qparm_colvar_id)
         DEALLOCATE (colvar%qparm_param%i_at_from)
         DEALLOCATE (colvar%qparm_param%i_at_to)
         DEALLOCATE (colvar%qparm_param)
      CASE (xyz_diag_colvar_id)
         DEALLOCATE (colvar%xyz_diag_param)
      CASE (xyz_outerdiag_colvar_id)
         DEALLOCATE (colvar%xyz_outerdiag_param)
      CASE (u_colvar_id)
         NULLIFY (colvar%u_param%mixed_energy_section)
         DEALLOCATE (colvar%u_param)
      CASE (hydronium_colvar_id)
         DEALLOCATE (colvar%hydronium_param%i_oxygens)
         DEALLOCATE (colvar%hydronium_param%i_hydrogens)
         DEALLOCATE (colvar%hydronium_param)
      CASE (acid_hyd_dist_colvar_id)
         DEALLOCATE (colvar%acid_hyd_dist_param%i_oxygens_water)
         DEALLOCATE (colvar%acid_hyd_dist_param%i_oxygens_acid)
         DEALLOCATE (colvar%acid_hyd_dist_param%i_hydrogens)
         DEALLOCATE (colvar%acid_hyd_dist_param)
      CASE (acid_hyd_shell_colvar_id)
         DEALLOCATE (colvar%acid_hyd_shell_param%i_oxygens_water)
         DEALLOCATE (colvar%acid_hyd_shell_param%i_oxygens_acid)
         DEALLOCATE (colvar%acid_hyd_shell_param%i_hydrogens)
         DEALLOCATE (colvar%acid_hyd_shell_param)
      CASE (reaction_path_colvar_id, distance_from_path_colvar_id)
         IF (colvar%reaction_path_param%dist_rmsd .OR. colvar%reaction_path_param%rmsd) THEN
            DEALLOCATE (colvar%reaction_path_param%r_ref)
            DEALLOCATE (colvar%reaction_path_param%i_rmsd)
         ELSE
            DO i = 1, SIZE(colvar%reaction_path_param%colvar_p)
               CALL colvar_release(colvar%reaction_path_param%colvar_p(i)%colvar)
            END DO
            DEALLOCATE (colvar%reaction_path_param%colvar_p)
            DEALLOCATE (colvar%reaction_path_param%f_vals)
         END IF
         DEALLOCATE (colvar%reaction_path_param)
      CASE (combine_colvar_id)
         DO i = 1, SIZE(colvar%combine_cvs_param%colvar_p)
            CALL colvar_release(colvar%combine_cvs_param%colvar_p(i)%colvar)
         END DO
         DEALLOCATE (colvar%combine_cvs_param%colvar_p)
         DEALLOCATE (colvar%combine_cvs_param%c_parameters)
         DEALLOCATE (colvar%combine_cvs_param%v_parameters)
         DEALLOCATE (colvar%combine_cvs_param%variables)
         DEALLOCATE (colvar%combine_cvs_param)
      CASE (rmsd_colvar_id)
         DEALLOCATE (colvar%rmsd_param%weights)
         DEALLOCATE (colvar%rmsd_param%r_ref)
         DEALLOCATE (colvar%rmsd_param%i_rmsd)
         DEALLOCATE (colvar%rmsd_param)
      CASE (Wc_colvar_id)
         DEALLOCATE (colvar%Wc)
      CASE (HBP_colvar_id)
         DEALLOCATE (colvar%HBP%ewc)
         DEALLOCATE (colvar%HBP%ids)
         DEALLOCATE (colvar%HBP)
      CASE (ring_puckering_colvar_id)
         DEALLOCATE (colvar%ring_puckering_param%atoms)
         DEALLOCATE (colvar%ring_puckering_param)
      CASE (mindist_colvar_id)
         IF (ASSOCIATED(colvar%mindist_param%i_dist_from)) THEN
            DEALLOCATE (colvar%mindist_param%i_dist_from)
         END IF
         IF (ASSOCIATED(colvar%mindist_param%i_coord_from)) THEN
            DEALLOCATE (colvar%mindist_param%i_coord_from)
         END IF
         IF (ASSOCIATED(colvar%mindist_param%i_coord_to)) THEN
            DEALLOCATE (colvar%mindist_param%i_coord_to)
         END IF
         IF (ASSOCIATED(colvar%mindist_param%k_coord_from)) THEN
            DEALLOCATE (colvar%mindist_param%k_coord_from)
         END IF
         IF (ASSOCIATED(colvar%mindist_param%k_coord_to)) THEN
            DEALLOCATE (colvar%mindist_param%k_coord_to)
         END IF
         DEALLOCATE (colvar%mindist_param)
      CASE (no_colvar_id)
         ! Do nothing
      CASE default
         CALL cp__b("subsys/colvar_types.F",1317,"")
      END SELECT
      DEALLOCATE (colvar)

   END SUBROUTINE colvar_release

! **************************************************************************************************
!> \brief Clone a colvar type
!> \param colvar_out ...
!> \param colvar_in the colvar to deallocate
!> \param i_atom_offset ...
!> \author Teodoro Laino [tlaino] 04.2006
! **************************************************************************************************
   RECURSIVE SUBROUTINE colvar_clone(colvar_out, colvar_in, i_atom_offset)
      TYPE(colvar_type), POINTER                         :: colvar_out, colvar_in
      INTEGER, INTENT(IN), OPTIONAL                      :: i_atom_offset

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_clone', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, my_offset, ndim, ndim2, stat

      my_offset = 0
      IF(.NOT.(ASSOCIATED(colvar_in)))CALL cp__a("subsys/colvar_types.F",1339)
      IF(.NOT.(.NOT. ASSOCIATED(colvar_out)))CALL cp__a("subsys/colvar_types.F",1340)
      IF (PRESENT(i_atom_offset)) my_offset = i_atom_offset
      CALL colvar_create(colvar_out, colvar_in%type_id)
      CALL colvar_clone_points(colvar_out, colvar_in, my_offset)
      IF (colvar_in%use_points) my_offset = 0
      SELECT CASE (colvar_out%type_id)
      CASE (dist_colvar_id)
         colvar_out%dist_param%i_at = colvar_in%dist_param%i_at+my_offset
         colvar_out%dist_param%j_at = colvar_in%dist_param%j_at+my_offset
         colvar_out%dist_param%axis_id = colvar_in%dist_param%axis_id
      CASE (coord_colvar_id)
         colvar_out%coord_param%n_atoms_to = colvar_in%coord_param%n_atoms_to
         colvar_out%coord_param%n_atoms_to_b = colvar_in%coord_param%n_atoms_to_b
         colvar_out%coord_param%n_atoms_from = colvar_in%coord_param%n_atoms_from
         colvar_out%coord_param%nncrd = colvar_in%coord_param%nncrd
         colvar_out%coord_param%ndcrd = colvar_in%coord_param%ndcrd
         colvar_out%coord_param%r_0 = colvar_in%coord_param%r_0
         colvar_out%coord_param%nncrd_b = colvar_in%coord_param%nncrd_b
         colvar_out%coord_param%ndcrd_b = colvar_in%coord_param%ndcrd_b
         colvar_out%coord_param%r_0_b = colvar_in%coord_param%r_0_b
         colvar_out%coord_param%use_kinds_from = colvar_in%coord_param%use_kinds_from
         colvar_out%coord_param%use_kinds_to = colvar_in%coord_param%use_kinds_to
         colvar_out%coord_param%use_kinds_to_b = colvar_in%coord_param%use_kinds_to_b
         NULLIFY (colvar_out%coord_param%c_kinds_from, colvar_out%coord_param%i_at_from)
         NULLIFY (colvar_out%coord_param%c_kinds_to, colvar_out%coord_param%i_at_to)
         NULLIFY (colvar_out%coord_param%c_kinds_to_b, colvar_out%coord_param%i_at_to_b)
         IF (colvar_in%coord_param%use_kinds_from) THEN
            ! KINDS
            ndim = SIZE(colvar_in%coord_param%c_kinds_from)
            ALLOCATE (colvar_out%coord_param%c_kinds_from(ndim))
            colvar_out%coord_param%c_kinds_from = colvar_in%coord_param%c_kinds_from
         ELSE
            ! INDEX
            ndim = SIZE(colvar_in%coord_param%i_at_from)
            ALLOCATE (colvar_out%coord_param%i_at_from(ndim))
            colvar_out%coord_param%i_at_from = colvar_in%coord_param%i_at_from+my_offset
         END IF
         IF (colvar_in%coord_param%use_kinds_to) THEN
            ! KINDS
            ndim = SIZE(colvar_in%coord_param%c_kinds_to)
            ALLOCATE (colvar_out%coord_param%c_kinds_to(ndim))
            colvar_out%coord_param%c_kinds_to = colvar_in%coord_param%c_kinds_to
         ELSE
            ! INDEX
            ndim = SIZE(colvar_in%coord_param%i_at_to)
            ALLOCATE (colvar_out%coord_param%i_at_to(ndim))
            colvar_out%coord_param%i_at_to = colvar_in%coord_param%i_at_to+my_offset
         END IF
         IF (colvar_in%coord_param%use_kinds_to_b) THEN
            ! KINDS
            ndim = SIZE(colvar_in%coord_param%c_kinds_to_b)
            ALLOCATE (colvar_out%coord_param%c_kinds_to_b(ndim))
            colvar_out%coord_param%c_kinds_to_b = colvar_in%coord_param%c_kinds_to_b
         ELSEIF (ASSOCIATED(colvar_in%coord_param%i_at_to_b)) THEN
            ! INDEX
            ndim = SIZE(colvar_in%coord_param%i_at_to_b)
            ALLOCATE (colvar_out%coord_param%i_at_to_b(ndim))
            colvar_out%coord_param%i_at_to_b = colvar_in%coord_param%i_at_to_b+my_offset
         END IF

      CASE (population_colvar_id)
         colvar_out%population_param%n_atoms_to = colvar_in%population_param%n_atoms_to
         colvar_out%population_param%n_atoms_from = colvar_in%population_param%n_atoms_from
         colvar_out%population_param%nncrd = colvar_in%population_param%nncrd
         colvar_out%population_param%ndcrd = colvar_in%population_param%ndcrd
         colvar_out%population_param%r_0 = colvar_in%population_param%r_0
         colvar_out%population_param%use_kinds_from = colvar_in%population_param%use_kinds_from
         colvar_out%population_param%use_kinds_to = colvar_in%population_param%use_kinds_to
         NULLIFY (colvar_out%population_param%c_kinds_from, colvar_out%population_param%i_at_from)
         NULLIFY (colvar_out%population_param%c_kinds_to, colvar_out%population_param%i_at_to)
         IF (colvar_in%population_param%use_kinds_from) THEN
            ! KINDS
            ndim = SIZE(colvar_in%population_param%c_kinds_from)
            ALLOCATE (colvar_out%population_param%c_kinds_from(ndim))
            colvar_out%population_param%c_kinds_from = colvar_in%population_param%c_kinds_from
         ELSE
            ! INDEX
            ndim = SIZE(colvar_in%population_param%i_at_from)
            ALLOCATE (colvar_out%population_param%i_at_from(ndim))
            colvar_out%population_param%i_at_from = colvar_in%population_param%i_at_from+my_offset
         END IF
         IF (colvar_in%population_param%use_kinds_to) THEN
            ! KINDS
            ndim = SIZE(colvar_in%population_param%c_kinds_to)
            ALLOCATE (colvar_out%population_param%c_kinds_to(ndim))
            colvar_out%population_param%c_kinds_to = colvar_in%population_param%c_kinds_to
         ELSE
            ! INDEX
            ndim = SIZE(colvar_in%population_param%i_at_to)
            ALLOCATE (colvar_out%population_param%i_at_to(ndim))
            colvar_out%population_param%i_at_to = colvar_in%population_param%i_at_to+my_offset
         END IF

      CASE (gyration_colvar_id)
         colvar_out%gyration_param%n_atoms = colvar_in%gyration_param%n_atoms
         colvar_out%gyration_param%use_kinds = colvar_in%gyration_param%use_kinds
         NULLIFY (colvar_out%gyration_param%c_kinds, colvar_out%gyration_param%i_at)
         IF (colvar_in%gyration_param%use_kinds) THEN
            ! KINDS
            ndim = SIZE(colvar_in%gyration_param%c_kinds)
            ALLOCATE (colvar_out%gyration_param%c_kinds(ndim))
            colvar_out%gyration_param%c_kinds = colvar_in%gyration_param%c_kinds
         ELSE
            ! INDEX
            ndim = SIZE(colvar_in%gyration_param%i_at)
            ALLOCATE (colvar_out%gyration_param%i_at(ndim))
            colvar_out%gyration_param%i_at = colvar_in%gyration_param%i_at+my_offset
         END IF
      CASE (angle_colvar_id)
         colvar_out%angle_param%i_at_angle = colvar_in%angle_param%i_at_angle+my_offset
      CASE (torsion_colvar_id)
         colvar_out%torsion_param%i_at_tors = colvar_in%torsion_param%i_at_tors+my_offset
         colvar_out%torsion_param%o0 = colvar_in%torsion_param%o0
      CASE (plane_distance_colvar_id)
         colvar_out%plane_distance_param%use_pbc = colvar_in%plane_distance_param%use_pbc
         colvar_out%plane_distance_param%plane = colvar_in%plane_distance_param%plane+my_offset
         colvar_out%plane_distance_param%point = colvar_in%plane_distance_param%point+my_offset
      CASE (plane_plane_angle_colvar_id)
         colvar_out%plane_plane_angle_param%plane1%type_of_def = colvar_in%plane_plane_angle_param%plane1%type_of_def
         IF (colvar_out%plane_plane_angle_param%plane1%type_of_def == plane_def_vec) THEN
            colvar_out%plane_plane_angle_param%plane1%normal_vec = colvar_in%plane_plane_angle_param%plane1%normal_vec
         ELSE
            colvar_out%plane_plane_angle_param%plane1%points = colvar_in%plane_plane_angle_param%plane1%points+my_offset
         END IF

         colvar_out%plane_plane_angle_param%plane2%type_of_def = colvar_in%plane_plane_angle_param%plane2%type_of_def
         IF (colvar_out%plane_plane_angle_param%plane2%type_of_def == plane_def_vec) THEN
            colvar_out%plane_plane_angle_param%plane2%normal_vec = colvar_in%plane_plane_angle_param%plane2%normal_vec
         ELSE
            colvar_out%plane_plane_angle_param%plane2%points = colvar_in%plane_plane_angle_param%plane2%points+my_offset
         END IF
      CASE (rotation_colvar_id)
         colvar_out%rotation_param%i_at1_bond1 = colvar_in%rotation_param%i_at1_bond1+my_offset
         colvar_out%rotation_param%i_at2_bond1 = colvar_in%rotation_param%i_at2_bond1+my_offset
         colvar_out%rotation_param%i_at1_bond2 = colvar_in%rotation_param%i_at1_bond2+my_offset
         colvar_out%rotation_param%i_at2_bond2 = colvar_in%rotation_param%i_at2_bond2+my_offset
      CASE (dfunct_colvar_id)
         colvar_out%dfunct_param%i_at_dfunct = colvar_in%dfunct_param%i_at_dfunct+my_offset
         colvar_out%dfunct_param%coeff = colvar_in%dfunct_param%coeff
         colvar_out%dfunct_param%use_pbc = colvar_in%dfunct_param%use_pbc
      CASE (qparm_colvar_id)
         colvar_out%qparm_param%n_atoms_to = colvar_in%qparm_param%n_atoms_to
         colvar_out%qparm_param%n_atoms_from = colvar_in%qparm_param%n_atoms_from
         colvar_out%qparm_param%rcut = colvar_in%qparm_param%rcut
         colvar_out%qparm_param%l = colvar_in%qparm_param%l
         colvar_out%qparm_param%alpha = colvar_in%qparm_param%alpha
         ndim = SIZE(colvar_in%qparm_param%i_at_from)
         ALLOCATE (colvar_out%qparm_param%i_at_from(ndim))
         ndim = SIZE(colvar_in%qparm_param%i_at_to)
         ALLOCATE (colvar_out%qparm_param%i_at_to(ndim))
         colvar_out%qparm_param%i_at_from = colvar_in%qparm_param%i_at_from+my_offset
         colvar_out%qparm_param%i_at_to = colvar_in%qparm_param%i_at_from+my_offset
      CASE (xyz_diag_colvar_id)
         colvar_out%xyz_diag_param%i_atom = colvar_in%xyz_diag_param%i_atom+my_offset
         colvar_out%xyz_diag_param%component = colvar_in%xyz_diag_param%component
         colvar_out%xyz_diag_param%r0 = colvar_in%xyz_diag_param%r0
         colvar_out%xyz_diag_param%use_pbc = colvar_in%xyz_diag_param%use_pbc
         colvar_out%xyz_diag_param%use_absolute_position = colvar_in%xyz_diag_param%use_absolute_position
      CASE (xyz_outerdiag_colvar_id)
         colvar_out%xyz_outerdiag_param%i_atoms = colvar_in%xyz_outerdiag_param%i_atoms+my_offset
         colvar_out%xyz_outerdiag_param%components = colvar_in%xyz_outerdiag_param%components
         colvar_out%xyz_outerdiag_param%r0 = colvar_in%xyz_outerdiag_param%r0
         colvar_out%xyz_outerdiag_param%use_pbc = colvar_in%xyz_outerdiag_param%use_pbc
      CASE (u_colvar_id)
         colvar_out%u_param%natom = colvar_in%u_param%natom
      CASE (hydronium_colvar_id)
         colvar_out%hydronium_param%n_hydrogens = colvar_in%hydronium_param%n_hydrogens
         colvar_out%hydronium_param%n_oxygens = colvar_in%hydronium_param%n_oxygens
         colvar_out%hydronium_param%nh = colvar_in%hydronium_param%nh
         colvar_out%hydronium_param%poh = colvar_in%hydronium_param%poh
         colvar_out%hydronium_param%poo = colvar_in%hydronium_param%poo
         colvar_out%hydronium_param%qoh = colvar_in%hydronium_param%qoh
         colvar_out%hydronium_param%qoo = colvar_in%hydronium_param%qoo
         colvar_out%hydronium_param%pm = colvar_in%hydronium_param%pm
         colvar_out%hydronium_param%qm = colvar_in%hydronium_param%qm
         colvar_out%hydronium_param%roo = colvar_in%hydronium_param%roo
         colvar_out%hydronium_param%roh = colvar_in%hydronium_param%roh
         ndim = SIZE(colvar_in%hydronium_param%i_oxygens)
         ALLOCATE (colvar_out%hydronium_param%i_oxygens(ndim))
         ndim = SIZE(colvar_in%hydronium_param%i_hydrogens)
         ALLOCATE (colvar_out%hydronium_param%i_hydrogens(ndim))
         colvar_out%hydronium_param%i_oxygens = colvar_in%hydronium_param%i_oxygens+my_offset
         colvar_out%hydronium_param%i_hydrogens = colvar_in%hydronium_param%i_hydrogens+my_offset
      CASE (acid_hyd_dist_colvar_id)
         colvar_out%acid_hyd_dist_param%n_hydrogens = colvar_in%acid_hyd_dist_param%n_hydrogens
         colvar_out%acid_hyd_dist_param%n_oxygens_water = colvar_in%acid_hyd_dist_param%n_oxygens_water
         colvar_out%acid_hyd_dist_param%n_oxygens_acid = colvar_in%acid_hyd_dist_param%n_oxygens_acid
         colvar_out%acid_hyd_dist_param%nc = colvar_in%acid_hyd_dist_param%nc
         colvar_out%acid_hyd_dist_param%pwoh = colvar_in%acid_hyd_dist_param%pwoh
         colvar_out%acid_hyd_dist_param%qwoh = colvar_in%acid_hyd_dist_param%qwoh
         colvar_out%acid_hyd_dist_param%paoh = colvar_in%acid_hyd_dist_param%paoh
         colvar_out%acid_hyd_dist_param%qaoh = colvar_in%acid_hyd_dist_param%qaoh
         colvar_out%acid_hyd_dist_param%pcut = colvar_in%acid_hyd_dist_param%pcut
         colvar_out%acid_hyd_dist_param%qcut = colvar_in%acid_hyd_dist_param%qcut
         colvar_out%acid_hyd_dist_param%rwoh = colvar_in%acid_hyd_dist_param%rwoh
         colvar_out%acid_hyd_dist_param%raoh = colvar_in%acid_hyd_dist_param%raoh
         ndim = SIZE(colvar_in%acid_hyd_dist_param%i_oxygens_water)
         ALLOCATE (colvar_out%acid_hyd_dist_param%i_oxygens_water(ndim))
         ndim = SIZE(colvar_in%acid_hyd_dist_param%i_oxygens_acid)
         ALLOCATE (colvar_out%acid_hyd_dist_param%i_oxygens_acid(ndim))
         ndim = SIZE(colvar_in%acid_hyd_dist_param%i_hydrogens)
         ALLOCATE (colvar_out%acid_hyd_dist_param%i_hydrogens(ndim))
         colvar_out%acid_hyd_dist_param%i_oxygens_water = colvar_in%acid_hyd_dist_param%i_oxygens_water+my_offset
         colvar_out%acid_hyd_dist_param%i_oxygens_acid = colvar_in%acid_hyd_dist_param%i_oxygens_acid+my_offset
         colvar_out%acid_hyd_dist_param%i_hydrogens = colvar_in%acid_hyd_dist_param%i_hydrogens+my_offset
      CASE (acid_hyd_shell_colvar_id)
         colvar_out%acid_hyd_shell_param%n_hydrogens = colvar_in%acid_hyd_shell_param%n_hydrogens
         colvar_out%acid_hyd_shell_param%n_oxygens_water = colvar_in%acid_hyd_shell_param%n_oxygens_water
         colvar_out%acid_hyd_shell_param%n_oxygens_acid = colvar_in%acid_hyd_shell_param%n_oxygens_acid
         colvar_out%acid_hyd_shell_param%nc = colvar_in%acid_hyd_shell_param%nc
         colvar_out%acid_hyd_shell_param%nh = colvar_in%acid_hyd_shell_param%nh
         colvar_out%acid_hyd_shell_param%pwoh = colvar_in%acid_hyd_shell_param%pwoh
         colvar_out%acid_hyd_shell_param%qwoh = colvar_in%acid_hyd_shell_param%qwoh
         colvar_out%acid_hyd_shell_param%paoh = colvar_in%acid_hyd_shell_param%paoh
         colvar_out%acid_hyd_shell_param%qaoh = colvar_in%acid_hyd_shell_param%qaoh
         colvar_out%acid_hyd_shell_param%poo = colvar_in%acid_hyd_shell_param%poo
         colvar_out%acid_hyd_shell_param%qoo = colvar_in%acid_hyd_shell_param%qoo
         colvar_out%acid_hyd_shell_param%pm = colvar_in%acid_hyd_shell_param%pm
         colvar_out%acid_hyd_shell_param%qm = colvar_in%acid_hyd_shell_param%qm
         colvar_out%acid_hyd_shell_param%pcut = colvar_in%acid_hyd_shell_param%pcut
         colvar_out%acid_hyd_shell_param%qcut = colvar_in%acid_hyd_shell_param%qcut
         colvar_out%acid_hyd_shell_param%rwoh = colvar_in%acid_hyd_shell_param%rwoh
         colvar_out%acid_hyd_shell_param%raoh = colvar_in%acid_hyd_shell_param%raoh
         colvar_out%acid_hyd_shell_param%roo = colvar_in%acid_hyd_shell_param%roo
         ndim = SIZE(colvar_in%acid_hyd_shell_param%i_oxygens_water)
         ALLOCATE (colvar_out%acid_hyd_shell_param%i_oxygens_water(ndim))
         ndim = SIZE(colvar_in%acid_hyd_shell_param%i_oxygens_acid)
         ALLOCATE (colvar_out%acid_hyd_shell_param%i_oxygens_acid(ndim))
         ndim = SIZE(colvar_in%acid_hyd_shell_param%i_hydrogens)
         ALLOCATE (colvar_out%acid_hyd_shell_param%i_hydrogens(ndim))
         colvar_out%acid_hyd_shell_param%i_oxygens_water = colvar_in%acid_hyd_shell_param%i_oxygens_water+my_offset
         colvar_out%acid_hyd_shell_param%i_oxygens_acid = colvar_in%acid_hyd_shell_param%i_oxygens_acid+my_offset
         colvar_out%acid_hyd_shell_param%i_hydrogens = colvar_in%acid_hyd_shell_param%i_hydrogens+my_offset
      CASE (reaction_path_colvar_id, distance_from_path_colvar_id)
         colvar_out%reaction_path_param%dist_rmsd = colvar_in%reaction_path_param%dist_rmsd
         colvar_out%reaction_path_param%rmsd = colvar_in%reaction_path_param%rmsd
         colvar_out%reaction_path_param%nr_frames = colvar_in%reaction_path_param%nr_frames
         IF (colvar_in%reaction_path_param%dist_rmsd .OR. colvar_in%reaction_path_param%rmsd) THEN
            colvar_out%reaction_path_param%align_frames = colvar_in%reaction_path_param%align_frames
            colvar_out%reaction_path_param%subset = colvar_in%reaction_path_param%subset
            NULLIFY (colvar_out%reaction_path_param%i_rmsd)
            NULLIFY (colvar_out%reaction_path_param%r_ref)
            ndim = SIZE(colvar_in%reaction_path_param%i_rmsd)
            ALLOCATE (colvar_out%reaction_path_param%i_rmsd(ndim), stat=stat)
            colvar_out%reaction_path_param%i_rmsd = colvar_in%reaction_path_param%i_rmsd
            ndim = SIZE(colvar_in%reaction_path_param%r_ref, 1)
            ndim2 = SIZE(colvar_in%reaction_path_param%r_ref, 2)
            ALLOCATE (colvar_out%reaction_path_param%r_ref(ndim, ndim2), stat=stat)
            colvar_out%reaction_path_param%r_ref = colvar_in%reaction_path_param%r_ref
         ELSE
            ndim = SIZE(colvar_in%reaction_path_param%colvar_p)
            ALLOCATE (colvar_out%reaction_path_param%colvar_p(ndim))
            DO i = 1, ndim
               NULLIFY (colvar_out%reaction_path_param%colvar_p(i)%colvar)
               CALL colvar_clone(colvar_out%reaction_path_param%colvar_p(i)%colvar, &
                                 colvar_in%reaction_path_param%colvar_p(i)%colvar, &
                                 my_offset)
            END DO
            colvar_out%reaction_path_param%function_bounds = colvar_in%reaction_path_param%function_bounds
            NULLIFY (colvar_out%reaction_path_param%f_vals)
            ndim = SIZE(colvar_in%reaction_path_param%f_vals, 1)
            ndim2 = SIZE(colvar_in%reaction_path_param%f_vals, 2)
            ALLOCATE (colvar_out%reaction_path_param%f_vals(ndim, ndim2))
            colvar_out%reaction_path_param%f_vals = colvar_in%reaction_path_param%f_vals
         END IF
         colvar_out%reaction_path_param%step_size = colvar_in%reaction_path_param%step_size
         colvar_out%reaction_path_param%n_components = colvar_in%reaction_path_param%n_components
         colvar_out%reaction_path_param%lambda = colvar_in%reaction_path_param%lambda
      CASE (combine_colvar_id)
         ndim = SIZE(colvar_in%combine_cvs_param%colvar_p)
         ALLOCATE (colvar_out%combine_cvs_param%colvar_p(ndim))
         DO i = 1, ndim
            NULLIFY (colvar_out%combine_cvs_param%colvar_p(i)%colvar)
            CALL colvar_clone(colvar_out%combine_cvs_param%colvar_p(i)%colvar, &
                              colvar_in%combine_cvs_param%colvar_p(i)%colvar, &
                              my_offset)
         END DO
         colvar_out%combine_cvs_param%lerr = colvar_in%combine_cvs_param%lerr
         colvar_out%combine_cvs_param%dx = colvar_in%combine_cvs_param%dx
         colvar_out%combine_cvs_param%function = colvar_in%combine_cvs_param%function
         !
         ndim = SIZE(colvar_in%combine_cvs_param%c_parameters)
         ALLOCATE (colvar_out%combine_cvs_param%c_parameters(ndim))
         colvar_out%combine_cvs_param%c_parameters = colvar_in%combine_cvs_param%c_parameters
         !
         ndim = SIZE(colvar_in%combine_cvs_param%v_parameters)
         ALLOCATE (colvar_out%combine_cvs_param%v_parameters(ndim))
         colvar_out%combine_cvs_param%v_parameters = colvar_in%combine_cvs_param%v_parameters
         !
         ndim = SIZE(colvar_in%combine_cvs_param%variables)
         ALLOCATE (colvar_out%combine_cvs_param%variables(ndim))
         colvar_out%combine_cvs_param%variables = colvar_in%combine_cvs_param%variables
      CASE (rmsd_colvar_id)
         colvar_out%rmsd_param%n_atoms = colvar_in%rmsd_param%n_atoms
         colvar_out%rmsd_param%align_frames = colvar_in%rmsd_param%align_frames
         colvar_out%rmsd_param%nr_frames = colvar_in%rmsd_param%nr_frames
         colvar_out%rmsd_param%subset = colvar_in%rmsd_param%subset
         NULLIFY (colvar_out%rmsd_param%i_rmsd)
         NULLIFY (colvar_out%rmsd_param%weights)
         NULLIFY (colvar_out%rmsd_param%r_ref)
         ! INDEX
         ndim = SIZE(colvar_in%rmsd_param%i_rmsd)
         ALLOCATE (colvar_out%rmsd_param%i_rmsd(ndim))
         colvar_out%rmsd_param%i_rmsd = colvar_in%rmsd_param%i_rmsd+my_offset
         ! A and Bconfigurations and weights
         ndim = SIZE(colvar_in%rmsd_param%weights)
         ALLOCATE (colvar_out%rmsd_param%weights(ndim))
         colvar_out%rmsd_param%weights = colvar_in%rmsd_param%weights
         ndim = SIZE(colvar_in%rmsd_param%r_ref, 1)
         ndim2 = SIZE(colvar_in%rmsd_param%r_ref, 2)
         ALLOCATE (colvar_out%rmsd_param%r_ref(ndim, ndim2))
         colvar_out%rmsd_param%r_ref = colvar_in%rmsd_param%r_ref
      CASE (Wc_colvar_id)
         colvar_out%Wc%ids = colvar_in%Wc%ids+my_offset
         colvar_out%Wc%rcut = colvar_in%Wc%rcut
      CASE (HBP_colvar_id)
         ndim = colvar_out%HBP%nPoints
         ALLOCATE (colvar_out%HBP%ids(ndim, 3))
         ALLOCATE (colvar_out%HBP%ewc(ndim))
         colvar_out%HBP%ids = colvar_in%HBP%ids+my_offset
         colvar_out%HBP%ewc = colvar_in%HBP%ewc+my_offset
         colvar_out%HBP%nPoints = colvar_in%HBP%nPoints
         colvar_out%HBP%rcut = colvar_in%HBP%rcut
         colvar_out%HBP%shift = colvar_in%HBP%shift
      CASE (ring_puckering_colvar_id)
         ndim = colvar_in%ring_puckering_param%nring
         colvar_out%ring_puckering_param%nring = colvar_in%ring_puckering_param%nring
         colvar_out%ring_puckering_param%iq = colvar_in%ring_puckering_param%iq
         ALLOCATE (colvar_out%ring_puckering_param%atoms(ndim))
         colvar_out%ring_puckering_param%atoms = colvar_in%ring_puckering_param%atoms+my_offset
      CASE (mindist_colvar_id)
         colvar_out%mindist_param%n_dist_from = colvar_in%mindist_param%n_dist_from
         colvar_out%mindist_param%n_coord_to = colvar_in%mindist_param%n_coord_to
         colvar_out%mindist_param%n_coord_from = colvar_in%mindist_param%n_coord_from
         colvar_out%mindist_param%p_exp = colvar_in%mindist_param%p_exp
         colvar_out%mindist_param%q_exp = colvar_in%mindist_param%q_exp
         colvar_out%mindist_param%r_cut = colvar_in%mindist_param%r_cut
         colvar_out%mindist_param%lambda = colvar_in%mindist_param%lambda
         colvar_out%mindist_param%use_kinds_from = colvar_in%mindist_param%use_kinds_from
         colvar_out%mindist_param%use_kinds_to = colvar_in%mindist_param%use_kinds_to
         NULLIFY (colvar_out%mindist_param%k_coord_from, colvar_out%mindist_param%i_coord_from)
         NULLIFY (colvar_out%mindist_param%k_coord_to, colvar_out%mindist_param%i_coord_to)
         NULLIFY (colvar_out%mindist_param%i_dist_from)
         ! INDEX
         ndim = SIZE(colvar_in%mindist_param%i_dist_from)
         ALLOCATE (colvar_out%mindist_param%i_dist_from(ndim))
         colvar_out%mindist_param%i_dist_from = colvar_in%mindist_param%i_dist_from+my_offset
         IF (colvar_in%mindist_param%use_kinds_from) THEN
            ! KINDS
            ndim = SIZE(colvar_in%mindist_param%k_coord_from)
            ALLOCATE (colvar_out%mindist_param%k_coord_from(ndim))
            colvar_out%mindist_param%k_coord_from = colvar_in%mindist_param%k_coord_from
         ELSE
            ! INDEX
            ndim = SIZE(colvar_in%mindist_param%i_coord_from)
            ALLOCATE (colvar_out%mindist_param%i_coord_from(ndim))
            colvar_out%mindist_param%i_coord_from = colvar_in%mindist_param%i_coord_from+my_offset
         END IF
         IF (colvar_in%mindist_param%use_kinds_to) THEN
            ! KINDS
            ndim = SIZE(colvar_in%mindist_param%k_coord_to)
            ALLOCATE (colvar_out%mindist_param%k_coord_to(ndim))
            colvar_out%mindist_param%k_coord_to = colvar_in%mindist_param%k_coord_to
         ELSE
            ! INDEX
            ndim = SIZE(colvar_in%mindist_param%i_coord_to)
            ALLOCATE (colvar_out%mindist_param%i_coord_to(ndim))
            colvar_out%mindist_param%i_coord_to = colvar_in%mindist_param%i_coord_to+my_offset
         END IF

      END SELECT
      CALL colvar_setup(colvar_out)
   END SUBROUTINE colvar_clone

! **************************************************************************************************
!> \brief Clone points type of a colvar type
!> \param colvar_out ...
!> \param colvar_in the colvar to deallocate
!> \param offset ...
!> \author Teodoro Laino [tlaino] 03.2007
! **************************************************************************************************
   SUBROUTINE colvar_clone_points(colvar_out, colvar_in, offset)
      TYPE(colvar_type), POINTER                         :: colvar_out, colvar_in
      INTEGER, INTENT(IN)                                :: offset

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_clone_points', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, natoms, npoints

      colvar_out%use_points = colvar_in%use_points
      IF (colvar_in%use_points) THEN
         IF(.NOT.(ASSOCIATED(colvar_in%points)))CALL cp__a("subsys/colvar_types.F",1732)
         npoints = SIZE(colvar_in%points)
         ALLOCATE (colvar_out%points(npoints))
         DO i = 1, npoints
            IF (ASSOCIATED(colvar_in%points(i)%atoms)) THEN
               natoms = SIZE(colvar_in%points(i)%atoms)
               ALLOCATE (colvar_out%points(i)%atoms(natoms))
               colvar_out%points(i)%atoms = colvar_in%points(i)%atoms+offset
            ELSE
               NULLIFY (colvar_out%points(i)%atoms)
            END IF
            IF (ASSOCIATED(colvar_in%points(i)%weights)) THEN
               natoms = SIZE(colvar_in%points(i)%weights)
               ALLOCATE (colvar_out%points(i)%weights(natoms))
               colvar_out%points(i)%weights = colvar_in%points(i)%weights
            ELSE
               NULLIFY (colvar_out%points(i)%weights)
            END IF
            colvar_out%points(i)%type_id = colvar_in%points(i)%type_id
            colvar_out%points(i)%r = colvar_in%points(i)%r
         END DO
      ELSE
         NULLIFY (colvar_out%points)
      END IF

   END SUBROUTINE colvar_clone_points

! **************************************************************************************************
!> \brief Change the dimension of a colvar_p_type
!> \param colvar_set ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \author Teodoro Laino [tlaino] 04.2006
! **************************************************************************************************
   SUBROUTINE colvar_p_reallocate(colvar_set, lb1_new, ub1_new)
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_set
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(len=*), PARAMETER :: routineN = 'colvar_p_reallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j, lb1, lb1_old, ub1, ub1_old
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: work

      NULLIFY (work)
      IF (ASSOCIATED(colvar_set)) THEN
         lb1_old = LBOUND(colvar_set, 1)
         ub1_old = UBOUND(colvar_set, 1)
         lb1 = MAX(lb1_new, lb1_old)
         ub1 = MIN(ub1_new, ub1_old)
         ALLOCATE (work(lb1:ub1))
         DO j = lb1, ub1
            NULLIFY (work(j)%colvar)
            CALL colvar_clone(work(j)%colvar, colvar_set(j)%colvar)
         END DO
         DO j = lb1, ub1
            CALL colvar_release(colvar_set(j)%colvar)
            NULLIFY (colvar_set(j)%colvar)
         END DO
         DEALLOCATE (colvar_set)
      END IF

      ALLOCATE (colvar_set(lb1_new:ub1_new))

      DO j = lb1_new, ub1_new
         NULLIFY (colvar_set(j)%colvar)
      END DO
      IF (ASSOCIATED(work)) THEN
         lb1 = MAX(lb1_new, lb1_old)
         ub1 = MIN(ub1_new, ub1_old)
         DO j = lb1, ub1
            CALL colvar_clone(colvar_set(j)%colvar, work(j)%colvar)
         END DO
         DO j = lb1, ub1
            CALL colvar_release(work(j)%colvar)
            NULLIFY (work(j)%colvar)
         END DO
         DEALLOCATE (work)
      END IF
   END SUBROUTINE colvar_p_reallocate

! **************************************************************************************************
!> \brief Deallocate a set of colvar_p_type
!> \param colvar_p ...
!> \par History
!>      07.2003 created [fawzi]
!>      01.2014 moved from cp_subsys_release() into separate routine.
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE colvar_p_release(colvar_p)
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_p

      INTEGER                                            :: i

! Colvar info

      IF (ASSOCIATED(colvar_p)) THEN
         DO i = 1, SIZE(colvar_p)
            IF (ASSOCIATED(colvar_p(i)%colvar)) &
               CALL colvar_release(colvar_p(i)%colvar)
         ENDDO
         DEALLOCATE (colvar_p)
      END IF
   END SUBROUTINE colvar_p_release

! **************************************************************************************************
!> \brief Evaluate the position of the geometrical point
!> \param point ...
!> \param particles ...
!> \param r ...
!> \author Teodoro Laino - 03.2007
! **************************************************************************************************
   SUBROUTINE eval_point_pos(point, particles, r)
      TYPE(point_type)                                   :: point
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: r

      INTEGER                                            :: i

      SELECT CASE (point%type_id)
      CASE (do_clv_geo_center)
         r = 0.0_dp
         DO i = 1, SIZE(point%atoms)
            r = r+particles(point%atoms(i))%r*point%weights(i)
         END DO
      CASE (do_clv_fix_point)
         r = point%r
      END SELECT

   END SUBROUTINE eval_point_pos

! **************************************************************************************************
!> \brief ...
!> \param point ...
!> \param particles ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE eval_point_mass(point, particles, m)
      TYPE(point_type)                                   :: point
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      REAL(KIND=dp), INTENT(OUT)                         :: m

      INTEGER                                            :: i

      SELECT CASE (point%type_id)
      CASE (do_clv_geo_center)
         m = 0.0_dp
         DO i = 1, SIZE(point%atoms)
            m = m+particles(point%atoms(i))%atomic_kind%mass*point%weights(i)
         END DO
      CASE (do_clv_fix_point)
         m = 0.0_dp
      END SELECT

   END SUBROUTINE eval_point_mass

! **************************************************************************************************
!> \brief Evaluate the position of the geometrical point
!> \param points ...
!> \param i ...
!> \param dsdr ...
!> \param f ...
!> \author Teodoro Laino - 03.2007
! **************************************************************************************************
   SUBROUTINE eval_point_der(points, i, dsdr, f)
      TYPE(point_type), DIMENSION(:), POINTER            :: points
      INTEGER, INTENT(IN)                                :: i
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dsdr
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: f

      INTEGER                                            :: ind, j
      REAL(KIND=dp)                                      :: fac

      SELECT CASE (points (i)%type_id)
      CASE (do_clv_geo_center)
         ind = 0
         DO j = 1, i-1
            IF (ASSOCIATED(points(j)%atoms)) THEN
               ind = ind+SIZE(points(j)%atoms)
            END IF
         END DO
         DO j = 1, SIZE(points(i)%atoms)
            fac = points(i)%weights(j)
            dsdr(:, ind+j) = dsdr(:, ind+j)+f*fac
         END DO
      CASE (do_clv_fix_point)
         ! Do nothing if it's a fixed point in space
      END SELECT

   END SUBROUTINE eval_point_der

! **************************************************************************************************
!> \brief subtract b from the ss value of a colvar: general function for handling
!>        periodic/non-periodic colvar
!> \param colvar ...
!> \param b ...
!> \retval diff ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   FUNCTION diff_colvar(colvar, b) RESULT(diff)
      TYPE(colvar_type), POINTER                         :: colvar
      REAL(KIND=dp), INTENT(IN)                          :: b
      REAL(KIND=dp)                                      :: diff

      CHARACTER(len=*), PARAMETER :: routineN = 'diff_colvar', routineP = moduleN//':'//routineN

      diff = colvar%ss-b
      IF (colvar%type_id == torsion_colvar_id) THEN
         ! The difference of a periodic COLVAR is always within [-pi,pi]
         diff = SIGN(1.0_dp, ASIN(SIN(diff)))*ACOS(COS(diff))
      END IF
   END FUNCTION diff_colvar

END MODULE colvar_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define all structure types related to force field kinds
!> \par History
!>      10.2014 Moved kind types out of force_field_types.F [Ole Schuett]
!> \author Ole Schuett
! **************************************************************************************************
MODULE force_field_kind_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_field_kind_types'

   INTEGER, PARAMETER, PUBLIC               :: do_ff_undef = 0, &
                                               do_ff_quartic = 1, &
                                               do_ff_g96 = 2, &
                                               do_ff_charmm = 3, &
                                               do_ff_harmonic = 4, &
                                               do_ff_g87 = 5, &
                                               do_ff_morse = 6, &
                                               do_ff_cubic = 7, &
                                               do_ff_mixed_bend_stretch = 8, &
                                               do_ff_amber = 9, &
                                               do_ff_mm2 = 10, &
                                               do_ff_mm3 = 11, &
                                               do_ff_mm4 = 12, &
                                               do_ff_fues = 13, &
                                               do_ff_legendre = 14, &
                                               do_ff_opls = 15

! *** Define the derived structure types ***

! **************************************************************************************************
   TYPE legendre_data_type
      INTEGER :: order
      REAL(KIND=dp), DIMENSION(:), POINTER :: coeffs
   END TYPE legendre_data_type

! **************************************************************************************************
   TYPE bond_kind_type
      INTEGER                        :: id_type
      REAL(KIND=dp)                :: k(3), r0, cs
      INTEGER                        :: kind_number
   END TYPE bond_kind_type

! **************************************************************************************************
   TYPE bend_kind_type
      INTEGER                        :: id_type
      REAL(KIND=dp)                :: k, theta0, cb
      REAL(KIND=dp)                :: r012, r032, kbs12, kbs32, kss
      TYPE(legendre_data_type)       :: legendre
      INTEGER                        :: kind_number
   END TYPE bend_kind_type

! **************************************************************************************************
   TYPE ub_kind_type
      INTEGER                        :: id_type
      REAL(KIND=dp)                :: k(3), r0
      INTEGER                        :: kind_number
   END TYPE ub_kind_type

! **************************************************************************************************
   TYPE torsion_kind_type
      INTEGER                        :: id_type
      INTEGER                        :: nmul
      INTEGER, POINTER               :: m(:)
      REAL(KIND=dp), POINTER       :: k(:), phi0(:)
      INTEGER                        :: kind_number
   END TYPE torsion_kind_type

! **************************************************************************************************
   TYPE impr_kind_type
      INTEGER                        :: id_type
      REAL(KIND=dp)                :: k, phi0
      INTEGER                        :: kind_number
   END TYPE impr_kind_type

! **************************************************************************************************
   TYPE opbend_kind_type
      INTEGER                        :: id_type
      REAL(KIND=dp)                :: k, phi0
      INTEGER                        :: kind_number
   END TYPE opbend_kind_type

! *** Public subroutines ***

   PUBLIC :: allocate_bend_kind_set, &
             allocate_bond_kind_set, &
             allocate_ub_kind_set, &
             allocate_torsion_kind_set, &
             allocate_impr_kind_set, &
             allocate_opbend_kind_set, &
             deallocate_bend_kind_set, &
             deallocate_bond_kind_set, &
             torsion_kind_dealloc_ref, &
             impr_kind_dealloc_ref

! *** Public data types ***

   PUBLIC :: bend_kind_type, &
             bond_kind_type, &
             impr_kind_type, &
             torsion_kind_type, &
             opbend_kind_type, &
             ub_kind_type, &
             ub_kind_dealloc_ref, &
             legendre_data_type
CONTAINS

! **************************************************************************************************
!> \brief Allocate and initialize a bend kind set.
!> \param bend_kind_set ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_bend_kind_set(bend_kind_set, nkind)

      TYPE(bend_kind_type), DIMENSION(:), POINTER        :: bend_kind_set
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_bend_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      NULLIFY (bend_kind_set)
      ALLOCATE (bend_kind_set(nkind))
      DO ikind = 1, nkind
         bend_kind_set(ikind)%id_type = do_ff_undef
         bend_kind_set(ikind)%k = 0.0_dp
         bend_kind_set(ikind)%theta0 = 0.0_dp
         bend_kind_set(ikind)%cb = 0.0_dp
         bend_kind_set(ikind)%r012 = 0.0_dp
         bend_kind_set(ikind)%r032 = 0.0_dp
         bend_kind_set(ikind)%kbs12 = 0.0_dp
         bend_kind_set(ikind)%kbs32 = 0.0_dp
         bend_kind_set(ikind)%kss = 0.0_dp
         bend_kind_set(ikind)%legendre%order = 0
         NULLIFY (bend_kind_set(ikind)%legendre%coeffs)
         bend_kind_set(ikind)%kind_number = ikind
      END DO
   END SUBROUTINE allocate_bend_kind_set

! **************************************************************************************************
!> \brief Allocate and initialize a bond kind set.
!> \param bond_kind_set ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_bond_kind_set(bond_kind_set, nkind)

      TYPE(bond_kind_type), DIMENSION(:), POINTER        :: bond_kind_set
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_bond_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      NULLIFY (bond_kind_set)
      ALLOCATE (bond_kind_set(nkind))
      DO ikind = 1, nkind
         bond_kind_set(ikind)%id_type = do_ff_undef
         bond_kind_set(ikind)%k(:) = 0.0_dp
         bond_kind_set(ikind)%r0 = 0.0_dp
         bond_kind_set(ikind)%cs = 0.0_dp
         bond_kind_set(ikind)%kind_number = ikind
      END DO
   END SUBROUTINE allocate_bond_kind_set

! **************************************************************************************************
!> \brief Allocate and initialize a torsion kind set.
!> \param torsion_kind_set ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_torsion_kind_set(torsion_kind_set, nkind)

      TYPE(torsion_kind_type), DIMENSION(:), POINTER     :: torsion_kind_set
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_torsion_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      NULLIFY (torsion_kind_set)
      ALLOCATE (torsion_kind_set(nkind))

      DO ikind = 1, nkind
         torsion_kind_set(ikind)%id_type = do_ff_undef
         torsion_kind_set(ikind)%nmul = 0
         NULLIFY (torsion_kind_set(ikind)%k)
         NULLIFY (torsion_kind_set(ikind)%m)
         NULLIFY (torsion_kind_set(ikind)%phi0)
         torsion_kind_set(ikind)%kind_number = ikind
      END DO
   END SUBROUTINE allocate_torsion_kind_set

! **************************************************************************************************
!> \brief Allocate and initialize a ub kind set.
!> \param ub_kind_set ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_ub_kind_set(ub_kind_set, nkind)

      TYPE(ub_kind_type), DIMENSION(:), POINTER          :: ub_kind_set
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_ub_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      NULLIFY (ub_kind_set)
      ALLOCATE (ub_kind_set(nkind))
      DO ikind = 1, nkind
         ub_kind_set(ikind)%id_type = do_ff_undef
         ub_kind_set(ikind)%k = 0.0_dp
         ub_kind_set(ikind)%r0 = 0.0_dp
         ub_kind_set(ikind)%kind_number = ikind
      END DO
   END SUBROUTINE allocate_ub_kind_set

! **************************************************************************************************
!> \brief Allocate and initialize a impr kind set.
!> \param impr_kind_set ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_impr_kind_set(impr_kind_set, nkind)

      TYPE(impr_kind_type), DIMENSION(:), POINTER        :: impr_kind_set
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_impr_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      NULLIFY (impr_kind_set)
      ALLOCATE (impr_kind_set(nkind))
      DO ikind = 1, nkind
         impr_kind_set(ikind)%id_type = do_ff_undef
         impr_kind_set(ikind)%k = 0.0_dp
         impr_kind_set(ikind)%phi0 = 0.0_dp
         impr_kind_set(ikind)%kind_number = ikind
      END DO
   END SUBROUTINE allocate_impr_kind_set

! **************************************************************************************************
!> \brief Allocate and initialize a opbend kind set.
!> \param opbend_kind_set ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_opbend_kind_set(opbend_kind_set, nkind)

      TYPE(opbend_kind_type), DIMENSION(:), POINTER      :: opbend_kind_set
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_opbend_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind

      NULLIFY (opbend_kind_set)
      ALLOCATE (opbend_kind_set(nkind))
      DO ikind = 1, nkind
         opbend_kind_set(ikind)%id_type = do_ff_undef
         opbend_kind_set(ikind)%k = 0.0_dp
         opbend_kind_set(ikind)%phi0 = 0.0_dp
         opbend_kind_set(ikind)%kind_number = ikind
      END DO
   END SUBROUTINE allocate_opbend_kind_set

! **************************************************************************************************
!> \brief Deallocate a bend kind set.
!> \param bend_kind_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_bend_kind_set(bend_kind_set)

      TYPE(bend_kind_type), DIMENSION(:), POINTER        :: bend_kind_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_bend_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(bend_kind_set)) RETURN
      DO i = 1, SIZE(bend_kind_set)
         IF (ASSOCIATED(bend_kind_set(i)%legendre%coeffs)) THEN
            DEALLOCATE (bend_kind_set(i)%legendre%coeffs)
            NULLIFY (bend_kind_set(i)%legendre%coeffs)
         END IF
      END DO
      DEALLOCATE (bend_kind_set)
   END SUBROUTINE deallocate_bend_kind_set

! **************************************************************************************************
!> \brief Deallocate a bond kind set.
!> \param bond_kind_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_bond_kind_set(bond_kind_set)

      TYPE(bond_kind_type), DIMENSION(:), POINTER        :: bond_kind_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_bond_kind_set', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (bond_kind_set)

   END SUBROUTINE deallocate_bond_kind_set

! **************************************************************************************************
!> \brief Deallocate a torsion kind element
!> \param torsion_kind ...
! **************************************************************************************************
   SUBROUTINE torsion_kind_dealloc_ref(torsion_kind)

      TYPE(torsion_kind_type), INTENT(INOUT)             :: torsion_kind

      CHARACTER(len=*), PARAMETER :: routineN = 'torsion_kind_dealloc_ref', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(torsion_kind%k)) THEN
         DEALLOCATE (torsion_kind%k)
      END IF
      IF (ASSOCIATED(torsion_kind%m)) THEN
         DEALLOCATE (torsion_kind%m)
      END IF
      IF (ASSOCIATED(torsion_kind%phi0)) THEN
         DEALLOCATE (torsion_kind%phi0)
      END IF

   END SUBROUTINE torsion_kind_dealloc_ref

! **************************************************************************************************
!> \brief Deallocate a ub kind set.
!> \param ub_kind_set ...
! **************************************************************************************************
   SUBROUTINE ub_kind_dealloc_ref(ub_kind_set)
      TYPE(ub_kind_type), DIMENSION(:), POINTER          :: ub_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'ub_kind_dealloc_ref', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (ub_kind_set)

   END SUBROUTINE ub_kind_dealloc_ref

! **************************************************************************************************
!> \brief Deallocate a impr kind element
! **************************************************************************************************
   SUBROUTINE impr_kind_dealloc_ref()

      CHARACTER(len=*), PARAMETER :: routineN = 'impr_kind_dealloc_ref', &
         routineP = moduleN//':'//routineN

!
! Questa e' la migliore routine che mente umana abbia concepito! ;-)
! Translation to english: This is the best subroutine that humanity can imagine! ;-)
!

   END SUBROUTINE impr_kind_dealloc_ref

END MODULE force_field_kind_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define the molecule kind structure types and the corresponding
!>      functionality
!> \par History
!>      Teodoro Laino [tlaino] 12.2008 - Preparing for VIRTUAL SITE constraints
!>                                       (patch by Marcel Baer)
!> \author MK (22.08.2003)
! **************************************************************************************************
MODULE molecule_kind_types
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE colvar_types,                    ONLY: &
        acid_hyd_dist_colvar_id, acid_hyd_shell_colvar_id, angle_colvar_id, colvar_counters, &
        combine_colvar_id, coord_colvar_id, dfunct_colvar_id, dist_colvar_id, gyration_colvar_id, &
        hydronium_colvar_id, plane_distance_colvar_id, plane_plane_angle_colvar_id, &
        population_colvar_id, qparm_colvar_id, reaction_path_colvar_id, rotation_colvar_id, &
        torsion_colvar_id, xyz_diag_colvar_id, xyz_outerdiag_colvar_id
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE force_field_kind_types,          ONLY: &
        bend_kind_type, bond_kind_type, impr_kind_dealloc_ref, impr_kind_type, opbend_kind_type, &
        torsion_kind_dealloc_ref, torsion_kind_type, ub_kind_dealloc_ref, ub_kind_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE shell_potential_types,           ONLY: shell_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'molecule_kind_types'

! *** Define the derived structure types ***

! **************************************************************************************************
   TYPE atom_type
      TYPE(atomic_kind_type), POINTER :: atomic_kind
      INTEGER :: id_name
   END TYPE atom_type

! **************************************************************************************************
   TYPE shell_type
      INTEGER :: a
      CHARACTER(LEN=default_string_length)  :: name
      TYPE(shell_kind_type), POINTER :: shell_kind
   END TYPE shell_type

! **************************************************************************************************
   TYPE bond_type
      INTEGER :: a, b
      INTEGER :: id_type, itype
      TYPE(bond_kind_type), POINTER :: bond_kind
   END TYPE bond_type

! **************************************************************************************************
   TYPE bend_type
      INTEGER :: a, b, c
      INTEGER :: id_type, itype
      TYPE(bend_kind_type), POINTER :: bend_kind
   END TYPE bend_type

! **************************************************************************************************
   TYPE ub_type
      INTEGER :: a, b, c
      INTEGER :: id_type, itype
      TYPE(ub_kind_type), POINTER :: ub_kind
   END TYPE ub_type

! **************************************************************************************************
   TYPE torsion_type
      INTEGER :: a, b, c, d
      INTEGER :: id_type, itype
      TYPE(torsion_kind_type), POINTER :: torsion_kind
   END TYPE torsion_type

! **************************************************************************************************
   TYPE impr_type
      INTEGER :: a, b, c, d
      INTEGER :: id_type, itype
      TYPE(impr_kind_type), POINTER :: impr_kind
   END TYPE impr_type

! **************************************************************************************************
   TYPE opbend_type
      INTEGER :: a, b, c, d
      INTEGER :: id_type, itype
      TYPE(opbend_kind_type), POINTER :: opbend_kind
   END TYPE opbend_type

! **************************************************************************************************
   TYPE restraint_type
      LOGICAL       :: active
      REAL(KIND=dp) :: k0
   END TYPE restraint_type

! **************************************************************************************************
   TYPE colvar_constraint_type
      INTEGER                        :: type_id
      INTEGER                        :: inp_seq_num
      LOGICAL                        :: use_points
      REAL(KIND=dp)                :: expected_value
      REAL(KIND=dp)                :: expected_value_growth_speed
      INTEGER, POINTER, DIMENSION(:) :: i_atoms
      TYPE(restraint_type)           :: restraint
   END TYPE colvar_constraint_type

! **************************************************************************************************
   TYPE g3x3_constraint_type
      INTEGER                        :: a, b, c
      REAL(KIND=dp)                :: dab, dac, dbc
      TYPE(restraint_type)           :: restraint
   END TYPE g3x3_constraint_type

! **************************************************************************************************
   TYPE g4x6_constraint_type
      INTEGER                        :: a, b, c, d
      REAL(KIND=dp)                :: dab, dac, dbc, dad, dbd, dcd
      TYPE(restraint_type)           :: restraint
   END TYPE g4x6_constraint_type

! **************************************************************************************************
   TYPE vsite_constraint_type
      INTEGER                        :: a, b, c, d
      REAL(KIND=dp)                :: wbc, wdc
      TYPE(restraint_type)           :: restraint
   END TYPE vsite_constraint_type

! **************************************************************************************************
   TYPE fixd_constraint_type
      TYPE(restraint_type)           :: restraint
      INTEGER                        :: fixd, itype
      REAL(KIND=dp), DIMENSION(3)    :: coord
   END TYPE fixd_constraint_type

! **************************************************************************************************
   TYPE local_fixd_constraint_type
      INTEGER                        :: ifixd_index, ikind
   END TYPE local_fixd_constraint_type

! **************************************************************************************************
   TYPE molecule_kind_type
      TYPE(atom_type), DIMENSION(:), POINTER            :: atom_list
      TYPE(bond_kind_type), DIMENSION(:), POINTER       :: bond_kind_set
      TYPE(bond_type), DIMENSION(:), POINTER            :: bond_list
      TYPE(bend_kind_type), DIMENSION(:), POINTER       :: bend_kind_set
      TYPE(bend_type), DIMENSION(:), POINTER            :: bend_list
      TYPE(ub_kind_type), DIMENSION(:), POINTER         :: ub_kind_set
      TYPE(ub_type), DIMENSION(:), POINTER              :: ub_list
      TYPE(torsion_kind_type), DIMENSION(:), POINTER    :: torsion_kind_set
      TYPE(torsion_type), DIMENSION(:), POINTER         :: torsion_list
      TYPE(impr_kind_type), DIMENSION(:), POINTER       :: impr_kind_set
      TYPE(impr_type), DIMENSION(:), POINTER            :: impr_list
      TYPE(opbend_kind_type), DIMENSION(:), POINTER     :: opbend_kind_set
      TYPE(opbend_type), DIMENSION(:), POINTER          :: opbend_list
      TYPE(colvar_constraint_type), DIMENSION(:), POINTER :: colv_list
      TYPE(g3x3_constraint_type), DIMENSION(:), POINTER   :: g3x3_list
      TYPE(g4x6_constraint_type), DIMENSION(:), POINTER   :: g4x6_list
      TYPE(vsite_constraint_type), DIMENSION(:), POINTER  :: vsite_list
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER   :: fixd_list
      TYPE(shell_type), DIMENSION(:), POINTER           :: shell_list
      CHARACTER(LEN=default_string_length)              :: name
      REAL(KIND=dp)                                   :: charge, &
                                                         mass
      INTEGER                                           :: kind_number, &
                                                           natom, &
                                                           nbond, &
                                                           nbend, &
                                                           nimpr, &
                                                           nopbend, &
                                                           ntorsion, &
                                                           nub, &
                                                           ng3x3, ng3x3_restraint, &
                                                           ng4x6, ng4x6_restraint, &
                                                           nvsite, nvsite_restraint, &
                                                           nfixd, nfixd_restraint, &
                                                           nmolecule, nshell
      TYPE(colvar_counters)                             :: ncolv
      INTEGER                                           :: nsgf, nelectron, &
                                                           nelectron_alpha, &
                                                           nelectron_beta
      INTEGER, DIMENSION(:), POINTER                    :: molecule_list
      LOGICAL                                           :: molname_generated
   END TYPE molecule_kind_type

   ! *** Public subroutines ***
   PUBLIC :: allocate_molecule_kind_set, &
             deallocate_molecule_kind_set, &
             get_molecule_kind, &
             get_molecule_kind_set, &
             set_molecule_kind, &
             write_molecule_kind_set, &
             setup_colvar_counters

   ! *** Public data types ***
   PUBLIC :: atom_type, &
             bend_type, &
             bond_type, &
             ub_type, &
             torsion_type, &
             impr_type, &
             opbend_type, &
             colvar_constraint_type, &
             g3x3_constraint_type, &
             g4x6_constraint_type, &
             vsite_constraint_type, &
             fixd_constraint_type, &
             local_fixd_constraint_type, &
             molecule_kind_type, &
             shell_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param colv_list ...
!> \param ncolv ...
! **************************************************************************************************
   SUBROUTINE setup_colvar_counters(colv_list, ncolv)
      TYPE(colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: colv_list
      TYPE(colvar_counters)                              :: ncolv

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_colvar_counters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k

      ncolv%ndist = 0
      ncolv%nangle = 0
      ncolv%ndfunct = 0
      ncolv%ntorsion = 0
      ncolv%ncoord = 0
      ncolv%nplane_dist = 0
      ncolv%nplane_angle = 0
      ncolv%nrot = 0
      ncolv%nqparm = 0
      ncolv%nxyz_diag = 0
      ncolv%nxyz_outerdiag = 0
      ncolv%nhydronium = 0
      ncolv%nacid_hyd_dist = 0
      ncolv%nacid_hyd_shell = 0
      ncolv%nreactionpath = 0
      ncolv%ncombinecvs = 0
      ncolv%nrestraint = 0
      ncolv%npopulation = 0
      ncolv%ngyration = 0

      IF (ASSOCIATED(colv_list)) THEN
         DO k = 1, SIZE(colv_list)
            IF (colv_list(k)%restraint%active) ncolv%nrestraint = ncolv%nrestraint+1
            SELECT CASE (colv_list (k)%type_id)
            CASE (angle_colvar_id)
               ncolv%nangle = ncolv%nangle+1
            CASE (coord_colvar_id)
               ncolv%ncoord = ncolv%ncoord+1
            CASE (population_colvar_id)
               ncolv%npopulation = ncolv%npopulation+1
            CASE (gyration_colvar_id)
               ncolv%ngyration = ncolv%ngyration+1
            CASE (rotation_colvar_id)
               ncolv%nrot = ncolv%nrot+1
            CASE (dist_colvar_id)
               ncolv%ndist = ncolv%ndist+1
            CASE (dfunct_colvar_id)
               ncolv%ndfunct = ncolv%ndfunct+1
            CASE (plane_distance_colvar_id)
               ncolv%nplane_dist = ncolv%nplane_dist+1
            CASE (plane_plane_angle_colvar_id)
               ncolv%nplane_angle = ncolv%nplane_angle+1
            CASE (torsion_colvar_id)
               ncolv%ntorsion = ncolv%ntorsion+1
            CASE (qparm_colvar_id)
               ncolv%nqparm = ncolv%nqparm+1
            CASE (xyz_diag_colvar_id)
               ncolv%nxyz_diag = ncolv%nxyz_diag+1
            CASE (xyz_outerdiag_colvar_id)
               ncolv%nxyz_outerdiag = ncolv%nxyz_outerdiag+1
            CASE (hydronium_colvar_id)
               ncolv%nhydronium = ncolv%nhydronium+1
            CASE (acid_hyd_dist_colvar_id)
               ncolv%nacid_hyd_dist = ncolv%nacid_hyd_dist+1
            CASE (acid_hyd_shell_colvar_id)
               ncolv%nacid_hyd_shell = ncolv%nacid_hyd_shell+1
            CASE (reaction_path_colvar_id)
               ncolv%nreactionpath = ncolv%nreactionpath+1
            CASE (combine_colvar_id)
               ncolv%ncombinecvs = ncolv%ncombinecvs+1
            CASE DEFAULT
               CALL cp__b("subsys/molecule_kind_types.F",305,"")
            END SELECT
         END DO
      END IF
      ncolv%ntot = ncolv%ndist+ &
                   ncolv%nangle+ &
                   ncolv%ntorsion+ &
                   ncolv%ncoord+ &
                   ncolv%nplane_dist+ &
                   ncolv%nplane_angle+ &
                   ncolv%ndfunct+ &
                   ncolv%nrot+ &
                   ncolv%nqparm+ &
                   ncolv%nxyz_diag+ &
                   ncolv%nxyz_outerdiag+ &
                   ncolv%nhydronium+ &
                   ncolv%nacid_hyd_dist+ &
                   ncolv%nacid_hyd_shell+ &
                   ncolv%nreactionpath+ &
                   ncolv%ncombinecvs+ &
                   ncolv%npopulation+ &
                   ncolv%ngyration

   END SUBROUTINE setup_colvar_counters

! **************************************************************************************************
!> \brief   Allocate and initialize a molecule kind set.
!> \param molecule_kind_set ...
!> \param nmolecule_kind ...
!> \date    22.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_molecule_kind_set(molecule_kind_set, nmolecule_kind)
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      INTEGER, INTENT(IN)                                :: nmolecule_kind

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_molecule_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imolecule_kind

      IF (ASSOCIATED(molecule_kind_set)) THEN
         CALL deallocate_molecule_kind_set(molecule_kind_set)
      END IF

      ALLOCATE (molecule_kind_set(nmolecule_kind))

      DO imolecule_kind = 1, nmolecule_kind
         NULLIFY (molecule_kind_set(imolecule_kind)%atom_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%bond_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%bend_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%colv_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%ub_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%ub_kind_set)
         NULLIFY (molecule_kind_set(imolecule_kind)%impr_kind_set)
         NULLIFY (molecule_kind_set(imolecule_kind)%impr_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%opbend_kind_set)
         NULLIFY (molecule_kind_set(imolecule_kind)%opbend_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%g3x3_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%g4x6_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%vsite_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%fixd_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%shell_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%torsion_list)
         NULLIFY (molecule_kind_set(imolecule_kind)%bond_kind_set)
         NULLIFY (molecule_kind_set(imolecule_kind)%bend_kind_set)
         NULLIFY (molecule_kind_set(imolecule_kind)%torsion_kind_set)
         molecule_kind_set(imolecule_kind)%charge = 0.0_dp
         molecule_kind_set(imolecule_kind)%mass = 0.0_dp
         molecule_kind_set(imolecule_kind)%name = ""
         molecule_kind_set(imolecule_kind)%molname_generated = .FALSE.
         molecule_kind_set(imolecule_kind)%kind_number = imolecule_kind
         molecule_kind_set(imolecule_kind)%natom = 0
         molecule_kind_set(imolecule_kind)%nbend = 0
         molecule_kind_set(imolecule_kind)%nbond = 0
         molecule_kind_set(imolecule_kind)%nimpr = 0
         molecule_kind_set(imolecule_kind)%nopbend = 0
         molecule_kind_set(imolecule_kind)%nub = 0
         CALL setup_colvar_counters(molecule_kind_set(imolecule_kind)%colv_list, &
                                    molecule_kind_set(imolecule_kind)%ncolv)
         molecule_kind_set(imolecule_kind)%ng3x3 = 0
         molecule_kind_set(imolecule_kind)%ng4x6 = 0
         molecule_kind_set(imolecule_kind)%nvsite = 0
         molecule_kind_set(imolecule_kind)%nfixd = 0
         molecule_kind_set(imolecule_kind)%ng3x3_restraint = 0
         molecule_kind_set(imolecule_kind)%ng4x6_restraint = 0
         molecule_kind_set(imolecule_kind)%nvsite_restraint = 0
         molecule_kind_set(imolecule_kind)%nfixd_restraint = 0
         molecule_kind_set(imolecule_kind)%nmolecule = 0
         molecule_kind_set(imolecule_kind)%ntorsion = 0
         molecule_kind_set(imolecule_kind)%nshell = 0
         NULLIFY (molecule_kind_set(imolecule_kind)%molecule_list)
      END DO

   END SUBROUTINE allocate_molecule_kind_set

! **************************************************************************************************
!> \brief   Deallocate a molecule kind set.
!> \param molecule_kind_set ...
!> \date    22.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_molecule_kind_set(molecule_kind_set)

      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_molecule_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imolecule_kind, j, nmolecule_kind

      IF (ASSOCIATED(molecule_kind_set)) THEN

         nmolecule_kind = SIZE(molecule_kind_set)

         DO imolecule_kind = 1, nmolecule_kind

            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%atom_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%atom_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%bend_kind_set)) THEN
               DO i = 1, SIZE(molecule_kind_set(imolecule_kind)%bend_kind_set)
                  IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%bend_kind_set(i)%legendre%coeffs)) &
                     DEALLOCATE (molecule_kind_set(imolecule_kind)%bend_kind_set(i)%legendre%coeffs)
                  NULLIFY (molecule_kind_set(imolecule_kind)%bend_kind_set(i)%legendre%coeffs)
               END DO
               DEALLOCATE (molecule_kind_set(imolecule_kind)%bend_kind_set)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%bend_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%bend_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%ub_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%ub_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%ub_kind_set)) THEN
               CALL ub_kind_dealloc_ref(molecule_kind_set(imolecule_kind)%ub_kind_set)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%impr_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%impr_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%impr_kind_set)) THEN
               DO i = 1, SIZE(molecule_kind_set(imolecule_kind)%impr_kind_set)
                  CALL impr_kind_dealloc_ref() !This Subroutine doesn't deallocate anything, maybe needs to be implemented
               END DO
               DEALLOCATE (molecule_kind_set(imolecule_kind)%impr_kind_set)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%opbend_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%opbend_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%opbend_kind_set)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%opbend_kind_set)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%bond_kind_set)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%bond_kind_set)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%bond_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%bond_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%colv_list)) THEN
               DO j = 1, SIZE(molecule_kind_set(imolecule_kind)%colv_list)
                  DEALLOCATE (molecule_kind_set(imolecule_kind)%colv_list(j)%i_atoms)
               END DO
               DEALLOCATE (molecule_kind_set(imolecule_kind)%colv_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%g3x3_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%g3x3_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%g4x6_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%g4x6_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%vsite_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%vsite_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%fixd_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%fixd_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%torsion_kind_set)) THEN
               DO i = 1, SIZE(molecule_kind_set(imolecule_kind)%torsion_kind_set)
                  CALL torsion_kind_dealloc_ref(molecule_kind_set(imolecule_kind)%torsion_kind_set(i))
               END DO
               DEALLOCATE (molecule_kind_set(imolecule_kind)%torsion_kind_set)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%shell_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%shell_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%torsion_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%torsion_list)
            END IF
            IF (ASSOCIATED(molecule_kind_set(imolecule_kind)%molecule_list)) THEN
               DEALLOCATE (molecule_kind_set(imolecule_kind)%molecule_list)
            ENDIF
         END DO

         DEALLOCATE (molecule_kind_set)
      ELSE
         CALL cp__b("subsys/molecule_kind_types.F",502,"The pointer molecule_kind_set is not associated and cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_molecule_kind_set

! **************************************************************************************************
!> \brief   Get informations about a molecule kind.
!> \param molecule_kind ...
!> \param atom_list ...
!> \param bond_list ...
!> \param bend_list ...
!> \param ub_list ...
!> \param impr_list ...
!> \param opbend_list ...
!> \param colv_list ...
!> \param fixd_list ...
!> \param g3x3_list ...
!> \param g4x6_list ...
!> \param vsite_list ...
!> \param torsion_list ...
!> \param shell_list ...
!> \param name ...
!> \param mass ...
!> \param charge ...
!> \param kind_number ...
!> \param natom ...
!> \param nbend ...
!> \param nbond ...
!> \param nub ...
!> \param nimpr ...
!> \param nopbend ...
!> \param nconstraint ...
!> \param nconstraint_fixd ...
!> \param nfixd ...
!> \param ncolv ...
!> \param ng3x3 ...
!> \param ng4x6 ...
!> \param nvsite ...
!> \param nfixd_restraint ...
!> \param ng3x3_restraint ...
!> \param ng4x6_restraint ...
!> \param nvsite_restraint ...
!> \param nrestraints ...
!> \param nmolecule ...
!> \param nsgf ...
!> \param nshell ...
!> \param ntorsion ...
!> \param molecule_list ...
!> \param nelectron ...
!> \param nelectron_alpha ...
!> \param nelectron_beta ...
!> \param bond_kind_set ...
!> \param bend_kind_set ...
!> \param ub_kind_set ...
!> \param impr_kind_set ...
!> \param opbend_kind_set ...
!> \param torsion_kind_set ...
!> \param molname_generated ...
!> \date    27.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_molecule_kind(molecule_kind, atom_list, bond_list, bend_list, &
                                ub_list, impr_list, opbend_list, colv_list, fixd_list, &
                                g3x3_list, g4x6_list, vsite_list, torsion_list, shell_list, &
                                name, mass, charge, kind_number, natom, nbend, nbond, nub, &
                                nimpr, nopbend, nconstraint, nconstraint_fixd, nfixd, ncolv, ng3x3, ng4x6, &
                                nvsite, nfixd_restraint, ng3x3_restraint, ng4x6_restraint, &
                                nvsite_restraint, nrestraints, nmolecule, nsgf, nshell, ntorsion, &
                                molecule_list, nelectron, nelectron_alpha, nelectron_beta, &
                                bond_kind_set, bend_kind_set, &
                                ub_kind_set, impr_kind_set, opbend_kind_set, torsion_kind_set, &
                                molname_generated)

      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(atom_type), DIMENSION(:), OPTIONAL, POINTER   :: atom_list
      TYPE(bond_type), DIMENSION(:), OPTIONAL, POINTER   :: bond_list
      TYPE(bend_type), DIMENSION(:), OPTIONAL, POINTER   :: bend_list
      TYPE(ub_type), DIMENSION(:), OPTIONAL, POINTER     :: ub_list
      TYPE(impr_type), DIMENSION(:), OPTIONAL, POINTER   :: impr_list
      TYPE(opbend_type), DIMENSION(:), OPTIONAL, POINTER :: opbend_list
      TYPE(colvar_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: colv_list
      TYPE(fixd_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: fixd_list
      TYPE(g3x3_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: g3x3_list
      TYPE(g4x6_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: g4x6_list
      TYPE(vsite_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: vsite_list
      TYPE(torsion_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: torsion_list
      TYPE(shell_type), DIMENSION(:), OPTIONAL, POINTER  :: shell_list
      CHARACTER(LEN=default_string_length), &
         INTENT(OUT), OPTIONAL                           :: name
      REAL(KIND=dp), OPTIONAL                            :: mass, charge
      INTEGER, INTENT(OUT), OPTIONAL                     :: kind_number, natom, nbend, nbond, nub, &
                                                            nimpr, nopbend, nconstraint, &
                                                            nconstraint_fixd, nfixd
      TYPE(colvar_counters), INTENT(out), OPTIONAL       :: ncolv
      INTEGER, INTENT(OUT), OPTIONAL :: ng3x3, ng4x6, nvsite, nfixd_restraint, ng3x3_restraint, &
         ng4x6_restraint, nvsite_restraint, nrestraints, nmolecule, nsgf, nshell, ntorsion
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: molecule_list
      INTEGER, INTENT(OUT), OPTIONAL                     :: nelectron, nelectron_alpha, &
                                                            nelectron_beta
      TYPE(bond_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: bond_kind_set
      TYPE(bend_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: bend_kind_set
      TYPE(ub_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ub_kind_set
      TYPE(impr_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: impr_kind_set
      TYPE(opbend_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: opbend_kind_set
      TYPE(torsion_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: torsion_kind_set
      LOGICAL, INTENT(OUT), OPTIONAL                     :: molname_generated

      CHARACTER(len=*), PARAMETER :: routineN = 'get_molecule_kind', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(molecule_kind)) THEN

         IF (PRESENT(atom_list)) atom_list => molecule_kind%atom_list
         IF (PRESENT(bend_list)) bend_list => molecule_kind%bend_list
         IF (PRESENT(bond_list)) bond_list => molecule_kind%bond_list
         IF (PRESENT(impr_list)) impr_list => molecule_kind%impr_list
         IF (PRESENT(opbend_list)) opbend_list => molecule_kind%opbend_list
         IF (PRESENT(ub_list)) ub_list => molecule_kind%ub_list
         IF (PRESENT(bond_kind_set)) bond_kind_set => molecule_kind%bond_kind_set
         IF (PRESENT(bend_kind_set)) bend_kind_set => molecule_kind%bend_kind_set
         IF (PRESENT(ub_kind_set)) ub_kind_set => molecule_kind%ub_kind_set
         IF (PRESENT(impr_kind_set)) impr_kind_set => molecule_kind%impr_kind_set
         IF (PRESENT(opbend_kind_set)) opbend_kind_set => molecule_kind%opbend_kind_set
         IF (PRESENT(torsion_kind_set)) torsion_kind_set => molecule_kind%torsion_kind_set
         IF (PRESENT(colv_list)) colv_list => molecule_kind%colv_list
         IF (PRESENT(g3x3_list)) g3x3_list => molecule_kind%g3x3_list
         IF (PRESENT(g4x6_list)) g4x6_list => molecule_kind%g4x6_list
         IF (PRESENT(vsite_list)) vsite_list => molecule_kind%vsite_list
         IF (PRESENT(fixd_list)) fixd_list => molecule_kind%fixd_list
         IF (PRESENT(torsion_list)) torsion_list => molecule_kind%torsion_list
         IF (PRESENT(shell_list)) shell_list => molecule_kind%shell_list
         IF (PRESENT(name)) name = molecule_kind%name
         IF (PRESENT(molname_generated)) molname_generated = molecule_kind%molname_generated
         IF (PRESENT(mass)) mass = molecule_kind%mass
         IF (PRESENT(charge)) charge = molecule_kind%charge
         IF (PRESENT(kind_number)) kind_number = molecule_kind%kind_number
         IF (PRESENT(natom)) natom = molecule_kind%natom
         IF (PRESENT(nbend)) nbend = molecule_kind%nbend
         IF (PRESENT(nbond)) nbond = molecule_kind%nbond
         IF (PRESENT(nub)) nub = molecule_kind%nub
         IF (PRESENT(nimpr)) nimpr = molecule_kind%nimpr
         IF (PRESENT(nopbend)) nopbend = molecule_kind%nopbend
         IF (PRESENT(nconstraint)) nconstraint = (molecule_kind%ncolv%ntot-molecule_kind%ncolv%nrestraint)+ &
                                                 3*(molecule_kind%ng3x3-molecule_kind%ng3x3_restraint)+ &
                                                 6*(molecule_kind%ng4x6-molecule_kind%ng4x6_restraint)+ &
                                                 3*(molecule_kind%nvsite-molecule_kind%nvsite_restraint)
         IF (PRESENT(ncolv)) ncolv = molecule_kind%ncolv
         IF (PRESENT(ng3x3)) ng3x3 = molecule_kind%ng3x3
         IF (PRESENT(ng4x6)) ng4x6 = molecule_kind%ng4x6
         IF (PRESENT(nvsite)) nvsite = molecule_kind%nvsite
         ! Number of atoms that have one or more components fixed
         IF (PRESENT(nfixd)) nfixd = molecule_kind%nfixd
         ! Number of degrees of freedom fixed
         IF (PRESENT(nconstraint_fixd)) THEN
            nconstraint_fixd = 0
            IF (molecule_kind%nfixd /= 0) THEN
               DO i = 1, SIZE(molecule_kind%fixd_list)
                  IF (molecule_kind%fixd_list(i)%restraint%active) CYCLE
                  SELECT CASE (molecule_kind%fixd_list (i)%itype)
                  CASE (use_perd_x, use_perd_y, use_perd_z)
                     nconstraint_fixd = nconstraint_fixd+1
                  CASE (use_perd_xy, use_perd_xz, use_perd_yz)
                     nconstraint_fixd = nconstraint_fixd+2
                  CASE (use_perd_xyz)
                     nconstraint_fixd = nconstraint_fixd+3
                  END SELECT
               END DO
            END IF
         END IF
         IF (PRESENT(ng3x3_restraint)) ng3x3_restraint = molecule_kind%ng3x3_restraint
         IF (PRESENT(ng4x6_restraint)) ng4x6_restraint = molecule_kind%ng4x6_restraint
         IF (PRESENT(nvsite_restraint)) nvsite_restraint = molecule_kind%nvsite_restraint
         IF (PRESENT(nfixd_restraint)) nfixd_restraint = molecule_kind%nfixd_restraint
         IF (PRESENT(nrestraints)) nrestraints = molecule_kind%ncolv%nrestraint+ &
                                                 molecule_kind%ng3x3_restraint+ &
                                                 molecule_kind%ng4x6_restraint+ &
                                                 molecule_kind%nvsite_restraint
         IF (PRESENT(nmolecule)) nmolecule = molecule_kind%nmolecule
         IF (PRESENT(nshell)) nshell = molecule_kind%nshell
         IF (PRESENT(ntorsion)) ntorsion = molecule_kind%ntorsion
         IF (PRESENT(nsgf)) nsgf = molecule_kind%nsgf
         IF (PRESENT(nelectron)) nelectron = molecule_kind%nelectron
         IF (PRESENT(nelectron_alpha)) nelectron_alpha = molecule_kind%nelectron_beta
         IF (PRESENT(nelectron_beta)) nelectron_beta = molecule_kind%nelectron_alpha
         IF (PRESENT(molecule_list)) molecule_list => molecule_kind%molecule_list

      ELSE

         CALL cp__b("subsys/molecule_kind_types.F",705,"The pointer molecule_kind is not associated")

      END IF

   END SUBROUTINE get_molecule_kind

! **************************************************************************************************
!> \brief   Get informations about a molecule kind set.
!> \param molecule_kind_set ...
!> \param maxatom ...
!> \param natom ...
!> \param nbond ...
!> \param nbend ...
!> \param nub ...
!> \param ntorsion ...
!> \param nimpr ...
!> \param nopbend ...
!> \param nconstraint ...
!> \param nconstraint_fixd ...
!> \param nmolecule ...
!> \param nrestraints ...
!> \date    27.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_molecule_kind_set(molecule_kind_set, maxatom, natom, &
                                    nbond, nbend, nub, ntorsion, nimpr, nopbend, &
                                    nconstraint, nconstraint_fixd, nmolecule, &
                                    nrestraints)

      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      INTEGER, INTENT(OUT), OPTIONAL                     :: maxatom, natom, nbond, nbend, nub, &
                                                            ntorsion, nimpr, nopbend, nconstraint, &
                                                            nconstraint_fixd, nmolecule, &
                                                            nrestraints

      CHARACTER(len=*), PARAMETER :: routineN = 'get_molecule_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER :: ibend, ibond, iimpr, imolecule_kind, iopbend, itorsion, iub, na, nc, nc_fixd, &
         nfixd_restraint, nm, nmolecule_kind, nrestraints_tot
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      IF (ASSOCIATED(molecule_kind_set)) THEN

         IF (PRESENT(maxatom)) maxatom = 0
         IF (PRESENT(natom)) natom = 0
         IF (PRESENT(nbond)) nbond = 0
         IF (PRESENT(nbend)) nbend = 0
         IF (PRESENT(nub)) nub = 0
         IF (PRESENT(ntorsion)) ntorsion = 0
         IF (PRESENT(nimpr)) nimpr = 0
         IF (PRESENT(nopbend)) nopbend = 0
         IF (PRESENT(nconstraint)) nconstraint = 0
         IF (PRESENT(nconstraint_fixd)) nconstraint_fixd = 0
         IF (PRESENT(nrestraints)) nrestraints = 0
         IF (PRESENT(nmolecule)) nmolecule = 0

         nmolecule_kind = SIZE(molecule_kind_set)

         DO imolecule_kind = 1, nmolecule_kind

            molecule_kind => molecule_kind_set(imolecule_kind)

            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   natom=na, &
                                   nbond=ibond, &
                                   nbend=ibend, &
                                   nub=iub, &
                                   ntorsion=itorsion, &
                                   nimpr=iimpr, &
                                   nopbend=iopbend, &
                                   nconstraint=nc, &
                                   nconstraint_fixd=nc_fixd, &
                                   nfixd_restraint=nfixd_restraint, &
                                   nrestraints=nrestraints_tot, &
                                   nmolecule=nm)
            IF (PRESENT(maxatom)) maxatom = MAX(maxatom, na)
            IF (PRESENT(natom)) natom = natom+na*nm
            IF (PRESENT(nbond)) nbond = nbond+ibond*nm
            IF (PRESENT(nbend)) nbend = nbend+ibend*nm
            IF (PRESENT(nub)) nub = nub+iub*nm
            IF (PRESENT(ntorsion)) ntorsion = ntorsion+itorsion*nm
            IF (PRESENT(nimpr)) nimpr = nimpr+iimpr*nm
            IF (PRESENT(nopbend)) nopbend = nopbend+iopbend*nm
            IF (PRESENT(nconstraint)) nconstraint = nconstraint+nc*nm+nc_fixd
            IF (PRESENT(nconstraint_fixd)) nconstraint_fixd = nconstraint_fixd+nc_fixd
            IF (PRESENT(nmolecule)) nmolecule = nmolecule+nm
            IF (PRESENT(nrestraints)) nrestraints = nrestraints+nm*nrestraints_tot+nfixd_restraint

         END DO

      ELSE

         CALL cp__b("subsys/molecule_kind_types.F",799,"The pointer molecule_kind_set is not associated")

      END IF

   END SUBROUTINE get_molecule_kind_set

! **************************************************************************************************
!> \brief   Set the components of a molecule kind.
!> \param molecule_kind ...
!> \param name ...
!> \param mass ...
!> \param charge ...
!> \param kind_number ...
!> \param molecule_list ...
!> \param atom_list ...
!> \param nbond ...
!> \param bond_list ...
!> \param nbend ...
!> \param bend_list ...
!> \param nub ...
!> \param ub_list ...
!> \param nimpr ...
!> \param impr_list ...
!> \param nopbend ...
!> \param opbend_list ...
!> \param ntorsion ...
!> \param torsion_list ...
!> \param fixd_list ...
!> \param ncolv ...
!> \param colv_list ...
!> \param ng3x3 ...
!> \param g3x3_list ...
!> \param ng4x6 ...
!> \param nfixd ...
!> \param g4x6_list ...
!> \param nvsite ...
!> \param vsite_list ...
!> \param ng3x3_restraint ...
!> \param ng4x6_restraint ...
!> \param nfixd_restraint ...
!> \param nshell ...
!> \param shell_list ...
!> \param nvsite_restraint ...
!> \param bond_kind_set ...
!> \param bend_kind_set ...
!> \param ub_kind_set ...
!> \param torsion_kind_set ...
!> \param impr_kind_set ...
!> \param opbend_kind_set ...
!> \param nelectron ...
!> \param nsgf ...
!> \param molname_generated ...
!> \date    27.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_molecule_kind(molecule_kind, name, mass, charge, kind_number, &
                                molecule_list, atom_list, nbond, bond_list, &
                                nbend, bend_list, nub, ub_list, nimpr, impr_list, &
                                nopbend, opbend_list, ntorsion, &
                                torsion_list, fixd_list, ncolv, colv_list, ng3x3, &
                                g3x3_list, ng4x6, nfixd, g4x6_list, nvsite, &
                                vsite_list, ng3x3_restraint, ng4x6_restraint, &
                                nfixd_restraint, nshell, shell_list, &
                                nvsite_restraint, bond_kind_set, bend_kind_set, &
                                ub_kind_set, torsion_kind_set, impr_kind_set, &
                                opbend_kind_set, nelectron, nsgf, &
                                molname_generated)

      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: name
      REAL(KIND=dp), OPTIONAL                            :: mass, charge
      INTEGER, INTENT(IN), OPTIONAL                      :: kind_number
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: molecule_list
      TYPE(atom_type), DIMENSION(:), OPTIONAL, POINTER   :: atom_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nbond
      TYPE(bond_type), DIMENSION(:), OPTIONAL, POINTER   :: bond_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nbend
      TYPE(bend_type), DIMENSION(:), OPTIONAL, POINTER   :: bend_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nub
      TYPE(ub_type), DIMENSION(:), OPTIONAL, POINTER     :: ub_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nimpr
      TYPE(impr_type), DIMENSION(:), OPTIONAL, POINTER   :: impr_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nopbend
      TYPE(opbend_type), DIMENSION(:), OPTIONAL, POINTER :: opbend_list
      INTEGER, INTENT(IN), OPTIONAL                      :: ntorsion
      TYPE(torsion_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: torsion_list
      TYPE(fixd_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: fixd_list
      TYPE(colvar_counters), INTENT(IN), OPTIONAL        :: ncolv
      TYPE(colvar_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: colv_list
      INTEGER, INTENT(IN), OPTIONAL                      :: ng3x3
      TYPE(g3x3_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: g3x3_list
      INTEGER, INTENT(IN), OPTIONAL                      :: ng4x6, nfixd
      TYPE(g4x6_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: g4x6_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nvsite
      TYPE(vsite_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: vsite_list
      INTEGER, INTENT(IN), OPTIONAL                      :: ng3x3_restraint, ng4x6_restraint, &
                                                            nfixd_restraint, nshell
      TYPE(shell_type), DIMENSION(:), OPTIONAL, POINTER  :: shell_list
      INTEGER, INTENT(IN), OPTIONAL                      :: nvsite_restraint
      TYPE(bond_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: bond_kind_set
      TYPE(bend_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: bend_kind_set
      TYPE(ub_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ub_kind_set
      TYPE(torsion_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: torsion_kind_set
      TYPE(impr_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: impr_kind_set
      TYPE(opbend_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: opbend_kind_set
      INTEGER, INTENT(IN), OPTIONAL                      :: nelectron, nsgf
      LOGICAL, INTENT(IN), OPTIONAL                      :: molname_generated

      CHARACTER(len=*), PARAMETER :: routineN = 'set_molecule_kind', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n

      IF (ASSOCIATED(molecule_kind)) THEN

         IF (PRESENT(atom_list)) THEN
            n = SIZE(atom_list)
            molecule_kind%natom = n
            molecule_kind%atom_list => atom_list
         END IF
         IF (PRESENT(molname_generated)) molecule_kind%molname_generated = molname_generated
         IF (PRESENT(name)) molecule_kind%name = name
         IF (PRESENT(mass)) molecule_kind%mass = mass
         IF (PRESENT(charge)) molecule_kind%charge = charge
         IF (PRESENT(kind_number)) molecule_kind%kind_number = kind_number
         IF (PRESENT(nbond)) molecule_kind%nbond = nbond
         IF (PRESENT(bond_list)) molecule_kind%bond_list => bond_list
         IF (PRESENT(nbend)) molecule_kind%nbend = nbend
         IF (PRESENT(nelectron)) molecule_kind%nelectron = nelectron
         IF (PRESENT(nsgf)) molecule_kind%nsgf = nsgf
         IF (PRESENT(bend_list)) molecule_kind%bend_list => bend_list
         IF (PRESENT(nub)) molecule_kind%nub = nub
         IF (PRESENT(ub_list)) molecule_kind%ub_list => ub_list
         IF (PRESENT(ntorsion)) molecule_kind%ntorsion = ntorsion
         IF (PRESENT(torsion_list)) molecule_kind%torsion_list => torsion_list
         IF (PRESENT(nimpr)) molecule_kind%nimpr = nimpr
         IF (PRESENT(impr_list)) molecule_kind%impr_list => impr_list
         IF (PRESENT(nopbend)) molecule_kind%nopbend = nopbend
         IF (PRESENT(opbend_list)) molecule_kind%opbend_list => opbend_list
         IF (PRESENT(ncolv)) molecule_kind%ncolv = ncolv
         IF (PRESENT(colv_list)) molecule_kind%colv_list => colv_list
         IF (PRESENT(ng3x3)) molecule_kind%ng3x3 = ng3x3
         IF (PRESENT(g3x3_list)) molecule_kind%g3x3_list => g3x3_list
         IF (PRESENT(ng4x6)) molecule_kind%ng4x6 = ng4x6
         IF (PRESENT(nvsite)) molecule_kind%nvsite = nvsite
         IF (PRESENT(nfixd)) molecule_kind%nfixd = nfixd
         IF (PRESENT(nfixd_restraint)) molecule_kind%nfixd_restraint = nfixd_restraint
         IF (PRESENT(ng3x3_restraint)) molecule_kind%ng3x3_restraint = ng3x3_restraint
         IF (PRESENT(ng4x6_restraint)) molecule_kind%ng4x6_restraint = ng4x6_restraint
         IF (PRESENT(nvsite_restraint)) molecule_kind%nvsite_restraint = nvsite_restraint
         IF (PRESENT(g4x6_list)) molecule_kind%g4x6_list => g4x6_list
         IF (PRESENT(vsite_list)) molecule_kind%vsite_list => vsite_list
         IF (PRESENT(fixd_list)) molecule_kind%fixd_list => fixd_list
         IF (PRESENT(bond_kind_set)) molecule_kind%bond_kind_set => bond_kind_set
         IF (PRESENT(bend_kind_set)) molecule_kind%bend_kind_set => bend_kind_set
         IF (PRESENT(ub_kind_set)) molecule_kind%ub_kind_set => ub_kind_set
         IF (PRESENT(torsion_kind_set)) molecule_kind%torsion_kind_set => torsion_kind_set
         IF (PRESENT(impr_kind_set)) molecule_kind%impr_kind_set => impr_kind_set
         IF (PRESENT(opbend_kind_set)) molecule_kind%opbend_kind_set => opbend_kind_set
         IF (PRESENT(nshell)) molecule_kind%nshell = nshell
         IF (PRESENT(shell_list)) molecule_kind%shell_list => shell_list
         IF (PRESENT(molecule_list)) THEN
            n = SIZE(molecule_list)
            molecule_kind%nmolecule = n
            molecule_kind%molecule_list => molecule_list
         END IF

      ELSE

         CALL cp__b("subsys/molecule_kind_types.F",981,"The pointer molecule_kind is not associated")

      END IF

   END SUBROUTINE set_molecule_kind

! **************************************************************************************************
!> \brief   Write a molecule kind data set to the output unit.
!> \param molecule_kind ...
!> \param output_unit ...
!> \date    24.09.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_molecule_kind(molecule_kind, output_unit)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      INTEGER, INTENT(in)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'write_molecule_kind', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name
      INTEGER                                            :: iatom, imolecule, natom, nmolecule
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      IF (output_unit > 0) THEN
         IF (ASSOCIATED(molecule_kind)) THEN
            natom = SIZE(molecule_kind%atom_list)
            nmolecule = SIZE(molecule_kind%molecule_list)

            IF (natom == 1) THEN
               atomic_kind => molecule_kind%atom_list(1)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, name=name)
               WRITE (UNIT=output_unit, FMT="(/,T2,I5,A,T36,A,A,T64,A)") &
                  molecule_kind%kind_number, &
                  ". Molecule kind: "//TRIM(molecule_kind%name), &
                  "Atomic kind name:   ", TRIM(name)
               WRITE (UNIT=output_unit, FMT="(T9,A,L1,T55,A,T75,I6)") &
                  "Automatic name: ", molecule_kind%molname_generated, &
                  "Number of molecules:", nmolecule
            ELSE
               WRITE (UNIT=output_unit, FMT="(/,T2,I5,A,T50,A,T75,I6,/,T22,A)") &
                  molecule_kind%kind_number, &
                  ". Molecule kind: "//TRIM(molecule_kind%name), &
                  "Number of atoms:    ", natom, &
                  "Atom         Atomic kind name"
               DO iatom = 1, natom
                  atomic_kind => molecule_kind%atom_list(iatom)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, name=name)
                  WRITE (UNIT=output_unit, FMT="(T20,I6,(7X,A18))") &
                     iatom, TRIM(name)
               END DO
               WRITE (UNIT=output_unit, FMT="(/,T9,A,L1)") &
                  "The name was automatically generated: ", &
                  molecule_kind%molname_generated
               WRITE (UNIT=output_unit, FMT="(T9,A,I6,/,T9,A,(T30,5I10))") &
                  "Number of molecules: ", nmolecule, "Molecule list:", &
                  (molecule_kind%molecule_list(imolecule), imolecule=1, nmolecule)
               IF (molecule_kind%nbond > 0) &
                  WRITE (UNIT=output_unit, FMT="(1X,A30,I6)") &
                  "Number of bonds:       ", molecule_kind%nbond
               IF (molecule_kind%nbend > 0) &
                  WRITE (UNIT=output_unit, FMT="(1X,A30,I6)") &
                  "Number of bends:       ", molecule_kind%nbend
               IF (molecule_kind%nub > 0) &
                  WRITE (UNIT=output_unit, FMT="(1X,A30,I6)") &
                  "Number of Urey-Bradley:", molecule_kind%nub
               IF (molecule_kind%ntorsion > 0) &
                  WRITE (UNIT=output_unit, FMT="(1X,A30,I6)") &
                  "Number of torsions:    ", molecule_kind%ntorsion
               IF (molecule_kind%nimpr > 0) &
                  WRITE (UNIT=output_unit, FMT="(1X,A30,I6)") &
                  "Number of improper:    ", molecule_kind%nimpr
               IF (molecule_kind%nopbend > 0) &
                  WRITE (UNIT=output_unit, FMT="(1X,A30,I6)") &
                  "Number of out opbends:    ", molecule_kind%nopbend
            END IF

         ELSE
            CALL cp__b("subsys/molecule_kind_types.F",1060,"")
         END IF

      END IF
   END SUBROUTINE write_molecule_kind

! **************************************************************************************************
!> \brief   Write a moleculeatomic kind set data set to the output unit.
!> \param molecule_kind_set ...
!> \param subsys_section ...
!> \date    24.09.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_molecule_kind_set(molecule_kind_set, subsys_section)
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_molecule_kind_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, imolecule_kind, natom, nbend, &
                                                            nbond, nimpr, nmolecule, &
                                                            nmolecule_kind, nopbend, ntors, &
                                                            ntotal, nub, output_unit
      LOGICAL                                            :: all_single_atoms
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, subsys_section, &
                                         "PRINT%MOLECULES", extension=".Log")
      IF (output_unit > 0) THEN
         IF (ASSOCIATED(molecule_kind_set)) THEN
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") "MOLECULE KIND INFORMATION"

            nmolecule_kind = SIZE(molecule_kind_set)

            all_single_atoms = .TRUE.
            DO imolecule_kind = 1, nmolecule_kind
               natom = SIZE(molecule_kind_set(imolecule_kind)%atom_list)
               nmolecule = SIZE(molecule_kind_set(imolecule_kind)%molecule_list)
               IF (natom*nmolecule > 1) all_single_atoms = .FALSE.
            ENDDO

            IF (all_single_atoms) THEN
               WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") &
                  "All atoms are their own molecule, skipping detailed information"
            ELSE
               DO imolecule_kind = 1, nmolecule_kind
                  molecule_kind => molecule_kind_set(imolecule_kind)
                  CALL write_molecule_kind(molecule_kind, output_unit)
               END DO
            ENDIF

            CALL get_molecule_kind_set(molecule_kind_set=molecule_kind_set, &
                                       nbond=nbond, &
                                       nbend=nbend, &
                                       nub=nub, &
                                       ntorsion=ntors, &
                                       nimpr=nimpr, &
                                       nopbend=nopbend)
            ntotal = nbond+nbend+nub+ntors+nimpr+nopbend
            IF (ntotal > 0) THEN
               WRITE (UNIT=output_unit, FMT="(/,/,T2,A,T45,A30,I6)") &
                  "MOLECULE KIND SET INFORMATION", &
                  "Total Number of bonds:       ", nbond
               WRITE (UNIT=output_unit, FMT="(T45,A30,I6)") &
                  "Total Number of bends:       ", nbend
               WRITE (UNIT=output_unit, FMT="(T45,A30,I6)") &
                  "Total Number of Urey-Bradley:", nub
               WRITE (UNIT=output_unit, FMT="(T45,A30,I6)") &
                  "Total Number of torsions:    ", ntors
               WRITE (UNIT=output_unit, FMT="(T45,A30,I6)") &
                  "Total Number of improper:    ", nimpr
               WRITE (UNIT=output_unit, FMT="(T45,A30,I6)") &
                  "Total Number of opbends:    ", nopbend
            END IF
         ELSE
            CALL cp__b("subsys/molecule_kind_types.F",1142,"")
         END IF

      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%MOLECULES")

      CALL timestop(handle)

   END SUBROUTINE write_molecule_kind_set

END MODULE molecule_kind_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define the data structure for the molecule information.
!> \par History
!>      JGH (22.05.2004) add last_atom information
!>      Teodoro Laino [tlaino] 12.2008 - Preparing for VIRTUAL SITE constraints
!>                                       (patch by Marcel Baer)
!> \author MK (29.08.2003)
! **************************************************************************************************
MODULE molecule_types_new

   USE colvar_types,                    ONLY: colvar_counters,&
                                              colvar_release,&
                                              colvar_type
   USE kinds,                           ONLY: dp,&
                                              int_size
   USE molecule_kind_types,             ONLY: colvar_constraint_type,&
                                              fixd_constraint_type,&
                                              g3x3_constraint_type,&
                                              g4x6_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type,&
                                              vsite_constraint_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (in this module) ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'molecule_types_new'

! *** Data types ***
! **************************************************************************************************
   TYPE local_colvar_constraint_type
      LOGICAL                       :: init
      TYPE(colvar_type), POINTER     :: colvar
      TYPE(colvar_type), POINTER     :: colvar_old
      REAL(KIND=dp)               :: lambda, sigma
   END TYPE local_colvar_constraint_type

! **************************************************************************************************
   TYPE local_g3x3_constraint_type
      LOGICAL                       :: init
      REAL(KIND=dp)               :: scale, scale_old, imass1, imass2, imass3
      REAL(KIND=dp), DIMENSION(3) :: fa, fb, fc, f_roll1, f_roll2, f_roll3, &
                                     ra_old, rb_old, rc_old, &
                                     va, vb, vc, lambda, del_lambda, lambda_old, &
                                     r0_12, r0_13, r0_23
      REAL(KIND=dp), DIMENSION(3, 3) :: amat
   END TYPE local_g3x3_constraint_type

! **************************************************************************************************
   TYPE local_g4x6_constraint_type
      LOGICAL                       :: init
      REAL(KIND=dp)               :: scale, scale_old, imass1, imass2, imass3, imass4
      REAL(KIND=dp), DIMENSION(3) :: fa, fb, fc, fd, fe, ff, &
                                     f_roll1, f_roll2, f_roll3, f_roll4, f_roll5, f_roll6, &
                                     ra_old, rb_old, rc_old, rd_old, re_old, rf_old, &
                                     va, vb, vc, vd, ve, vf, &
                                     r0_12, r0_13, r0_14, r0_23, r0_24, r0_34
      REAL(KIND=dp), DIMENSION(6)   :: lambda, del_lambda, lambda_old
      REAL(KIND=dp), DIMENSION(6, 6) :: amat
   END TYPE local_g4x6_constraint_type

! **************************************************************************************************
   TYPE local_molecule_type
      INTEGER, DIMENSION(:), POINTER :: states ! indices of Kohn-Sham states for molecule
      INTEGER                        :: nstates ! Kohn-Sham states for molecule
   END TYPE local_molecule_type

! **************************************************************************************************
   TYPE local_constraint_type
      TYPE(local_colvar_constraint_type), DIMENSION(:), POINTER :: lcolv
      TYPE(local_g3x3_constraint_type), DIMENSION(:), POINTER :: lg3x3
      TYPE(local_g4x6_constraint_type), DIMENSION(:), POINTER :: lg4x6
   END TYPE local_constraint_type

! **************************************************************************************************
   TYPE global_constraint_type
      TYPE(colvar_counters)                    :: ncolv
      INTEGER                                  :: ntot, nrestraint
      INTEGER                                  :: ng3x3, ng3x3_restraint
      INTEGER                                  :: ng4x6, ng4x6_restraint
      INTEGER                                  :: nvsite, nvsite_restraint
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER   :: fixd_list
      TYPE(colvar_constraint_type), DIMENSION(:), POINTER :: colv_list
      TYPE(g3x3_constraint_type), DIMENSION(:), POINTER   :: g3x3_list
      TYPE(g4x6_constraint_type), DIMENSION(:), POINTER   :: g4x6_list
      TYPE(vsite_constraint_type), DIMENSION(:), POINTER  :: vsite_list
      TYPE(local_colvar_constraint_type), DIMENSION(:), POINTER :: lcolv
      TYPE(local_g3x3_constraint_type), DIMENSION(:), POINTER :: lg3x3
      TYPE(local_g4x6_constraint_type), DIMENSION(:), POINTER :: lg4x6
   END TYPE global_constraint_type

! **************************************************************************************************
   TYPE molecule_type
      TYPE(molecule_kind_type), POINTER    :: molecule_kind ! pointer to molecule kind information
      TYPE(local_molecule_type), POINTER   :: lmi ! local molecule information
      TYPE(local_constraint_type), POINTER :: lci ! local molecule constraint info
      INTEGER                              :: first_atom ! global index of first atom in molecule
      INTEGER                              :: last_atom ! global index of last atom in molecule
      INTEGER                              :: first_shell ! global index of first shell atom in molecule
      INTEGER                              :: last_shell ! global index of last shell atom in molecule
   END TYPE molecule_type

! *** Public data types ***

   PUBLIC :: local_colvar_constraint_type, &
             local_g3x3_constraint_type, &
             local_g4x6_constraint_type, &
             local_constraint_type, &
             local_molecule_type, &
             global_constraint_type, &
             molecule_type

! *** Public subroutines ***

   PUBLIC :: deallocate_global_constraint, &
             allocate_molecule_set, &
             deallocate_molecule_set, &
             get_molecule, &
             set_molecule, &
             set_molecule_set, &
             molecule_of_atom, &
             get_molecule_set_info

CONTAINS

! **************************************************************************************************
!> \brief   Deallocate a global constraint.
!> \param gci ...
!> \par History
!>      07.2003 created [fawzi]
!>      01.2014 moved from cp_subsys_release() into separate routine.
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE deallocate_global_constraint(gci)
      TYPE(global_constraint_type), POINTER              :: gci

      INTEGER                                            :: i

      IF (ASSOCIATED(gci)) THEN
         ! List of constraints
         IF (ASSOCIATED(gci%colv_list)) THEN
            DO i = 1, SIZE(gci%colv_list)
               DEALLOCATE (gci%colv_list(i)%i_atoms)
            END DO
            DEALLOCATE (gci%colv_list)
         END IF

         IF (ASSOCIATED(gci%g3x3_list)) &
            DEALLOCATE (gci%g3x3_list)

         IF (ASSOCIATED(gci%g4x6_list)) &
            DEALLOCATE (gci%g4x6_list)

         ! Local information
         IF (ASSOCIATED(gci%lcolv)) THEN
            DO i = 1, SIZE(gci%lcolv)
               CALL colvar_release(gci%lcolv(i)%colvar)
               CALL colvar_release(gci%lcolv(i)%colvar_old)
               NULLIFY (gci%lcolv(i)%colvar, gci%lcolv(i)%colvar_old)
            END DO
            DEALLOCATE (gci%lcolv)
         ENDIF

         IF (ASSOCIATED(gci%lg3x3)) &
            DEALLOCATE (gci%lg3x3)

         IF (ASSOCIATED(gci%lg4x6)) &
            DEALLOCATE (gci%lg4x6)

         IF (ASSOCIATED(gci%fixd_list)) &
            DEALLOCATE (gci%fixd_list)

         DEALLOCATE (gci)
      ENDIF
   END SUBROUTINE deallocate_global_constraint

! **************************************************************************************************
!> \brief   Allocate a molecule set.
!> \param molecule_set ...
!> \param nmolecule ...
!> \date    29.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_molecule_set(molecule_set, nmolecule)
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      INTEGER, INTENT(IN)                                :: nmolecule

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_molecule_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imolecule, istat

      IF (ASSOCIATED(molecule_set)) CALL deallocate_molecule_set(molecule_set)

      ALLOCATE (molecule_set(nmolecule), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 207, &
                                       "molecule_set", nmolecule*int_size)

      DO imolecule = 1, nmolecule
         NULLIFY (molecule_set(imolecule)%molecule_kind)
         NULLIFY (molecule_set(imolecule)%lmi)
         NULLIFY (molecule_set(imolecule)%lci)

         molecule_set(imolecule)%first_atom = 0
         molecule_set(imolecule)%last_atom = 0
         molecule_set(imolecule)%first_shell = 0
         molecule_set(imolecule)%last_shell = 0
      END DO

   END SUBROUTINE allocate_molecule_set

! **************************************************************************************************
!> \brief   Deallocate a molecule set.
!> \param molecule_set ...
!> \date    29.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_molecule_set(molecule_set)
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_molecule_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imolecule, j

      IF (ASSOCIATED(molecule_set)) THEN

         DO imolecule = 1, SIZE(molecule_set)
            IF (ASSOCIATED(molecule_set(imolecule)%lmi)) THEN
               IF (ASSOCIATED(molecule_set(imolecule)%lmi%states)) THEN
                  DEALLOCATE (molecule_set(imolecule)%lmi%states)
               ENDIF
               DEALLOCATE (molecule_set(imolecule)%lmi)
            ENDIF
            IF (ASSOCIATED(molecule_set(imolecule)%lci)) THEN
               IF (ASSOCIATED(molecule_set(imolecule)%lci%lcolv)) THEN
                  DO j = 1, SIZE(molecule_set(imolecule)%lci%lcolv)
                     CALL colvar_release(molecule_set(imolecule)%lci%lcolv(j)%colvar)
                     CALL colvar_release(molecule_set(imolecule)%lci%lcolv(j)%colvar_old)
                     NULLIFY (molecule_set(imolecule)%lci%lcolv(j)%colvar)
                     NULLIFY (molecule_set(imolecule)%lci%lcolv(j)%colvar_old)
                  END DO
                  DEALLOCATE (molecule_set(imolecule)%lci%lcolv)
               ENDIF
               IF (ASSOCIATED(molecule_set(imolecule)%lci%lg3x3)) THEN
                  DEALLOCATE (molecule_set(imolecule)%lci%lg3x3)
               ENDIF
               IF (ASSOCIATED(molecule_set(imolecule)%lci%lg4x6)) THEN
                  DEALLOCATE (molecule_set(imolecule)%lci%lg4x6)
               ENDIF
               DEALLOCATE (molecule_set(imolecule)%lci)
            ENDIF
         ENDDO
         DEALLOCATE (molecule_set)

      ELSE

         CALL cp_abort(cp__l("subsys/molecule_types_new.F",270), &
                       "The pointer molecule_set is not associated and "// &
                       "cannot be deallocated")

      END IF

   END SUBROUTINE deallocate_molecule_set

! **************************************************************************************************
!> \brief   Get components from a molecule data set.
!> \param molecule ...
!> \param molecule_kind ...
!> \param lmi ...
!> \param lci ...
!> \param lg3x3 ...
!> \param lg4x6 ...
!> \param lcolv ...
!> \param first_atom ...
!> \param last_atom ...
!> \param first_shell ...
!> \param last_shell ...
!> \date    29.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_molecule(molecule, molecule_kind, lmi, lci, lg3x3, lg4x6, lcolv, &
                           first_atom, last_atom, first_shell, last_shell)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(molecule_kind_type), OPTIONAL, POINTER        :: molecule_kind
      TYPE(local_molecule_type), OPTIONAL, POINTER       :: lmi
      TYPE(local_constraint_type), OPTIONAL, POINTER     :: lci
      TYPE(local_g3x3_constraint_type), OPTIONAL, &
         POINTER                                         :: lg3x3(:)
      TYPE(local_g4x6_constraint_type), OPTIONAL, &
         POINTER                                         :: lg4x6(:)
      TYPE(local_colvar_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: lcolv
      INTEGER, OPTIONAL                                  :: first_atom, last_atom, first_shell, &
                                                            last_shell

      CHARACTER(len=*), PARAMETER :: routineN = 'get_molecule', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(molecule)) THEN

         IF (PRESENT(first_atom)) first_atom = molecule%first_atom
         IF (PRESENT(last_atom)) last_atom = molecule%last_atom
         IF (PRESENT(first_shell)) first_shell = molecule%first_shell
         IF (PRESENT(last_shell)) last_shell = molecule%last_shell
         IF (PRESENT(molecule_kind)) molecule_kind => molecule%molecule_kind
         IF (PRESENT(lmi)) lmi => molecule%lmi
         IF (PRESENT(lci)) lci => molecule%lci
         IF (PRESENT(lcolv)) THEN
            IF (ASSOCIATED(molecule%lci)) THEN
               lcolv => molecule%lci%lcolv
            ELSE
               CALL cp__b("subsys/molecule_types_new.F",326,"The pointer lci is not associated")
            ENDIF
         ENDIF
         IF (PRESENT(lg3x3)) THEN
            IF (ASSOCIATED(molecule%lci)) THEN
               lg3x3 => molecule%lci%lg3x3
            ELSE
               CALL cp__b("subsys/molecule_types_new.F",333,"The pointer lci is not associated")
            ENDIF
         ENDIF
         IF (PRESENT(lg4x6)) THEN
            IF (ASSOCIATED(molecule%lci)) THEN
               lg4x6 => molecule%lci%lg4x6
            ELSE
               CALL cp__b("subsys/molecule_types_new.F",340,"The pointer lci is not associated")
            ENDIF
         ENDIF

      ELSE

         CALL cp__b("subsys/molecule_types_new.F",346,"The pointer lci is not associated")

      END IF

   END SUBROUTINE get_molecule

! **************************************************************************************************
!> \brief   Set a molecule data set.
!> \param molecule ...
!> \param molecule_kind ...
!> \param lmi ...
!> \param lci ...
!> \param lcolv ...
!> \param lg3x3 ...
!> \param lg4x6 ...
!> \date    29.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_molecule(molecule, molecule_kind, lmi, lci, lcolv, lg3x3, lg4x6)
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(molecule_kind_type), OPTIONAL, POINTER        :: molecule_kind
      TYPE(local_molecule_type), OPTIONAL, POINTER       :: lmi
      TYPE(local_constraint_type), OPTIONAL, POINTER     :: lci
      TYPE(local_colvar_constraint_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: lcolv
      TYPE(local_g3x3_constraint_type), OPTIONAL, &
         POINTER                                         :: lg3x3(:)
      TYPE(local_g4x6_constraint_type), OPTIONAL, &
         POINTER                                         :: lg4x6(:)

      CHARACTER(len=*), PARAMETER :: routineN = 'set_molecule', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(molecule)) THEN

         IF (PRESENT(molecule_kind)) molecule%molecule_kind => molecule_kind
         IF (PRESENT(lmi)) molecule%lmi => lmi
         IF (PRESENT(lci)) molecule%lci => lci
         IF (PRESENT(lcolv)) THEN
            IF (ASSOCIATED(molecule%lci)) THEN
               molecule%lci%lcolv => lcolv
            ELSE
               CALL cp__b("subsys/molecule_types_new.F",388,"The pointer lci is not associated")
            ENDIF
         ENDIF
         IF (PRESENT(lg3x3)) THEN
            IF (ASSOCIATED(molecule%lci)) THEN
               molecule%lci%lg3x3 => lg3x3
            ELSE
               CALL cp__b("subsys/molecule_types_new.F",395,"The pointer lci is not associated")
            ENDIF
         ENDIF
         IF (PRESENT(lg4x6)) THEN
            IF (ASSOCIATED(molecule%lci)) THEN
               molecule%lci%lg4x6 => lg4x6
            ELSE
               CALL cp__b("subsys/molecule_types_new.F",402,"The pointer lci is not associated")
            ENDIF
         ENDIF
      ELSE

         CALL cp__b("subsys/molecule_types_new.F",407,"The pointer molecule is not associated")

      END IF

   END SUBROUTINE set_molecule

! **************************************************************************************************
!> \brief   Set a molecule data set.
!> \param molecule_set ...
!> \param first_atom ...
!> \param last_atom ...
!> \date    29.08.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_molecule_set(molecule_set, first_atom, last_atom)
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: first_atom, last_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'set_molecule_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imolecule

      IF (ASSOCIATED(molecule_set)) THEN

         IF (PRESENT(first_atom)) THEN

            IF (SIZE(first_atom) /= SIZE(molecule_set)) THEN
               CALL cp_abort(cp__l("subsys/molecule_types_new.F",436), &
                             "The sizes of first_atom and molecule_set "// &
                             "are different")
            END IF

            DO imolecule = 1, SIZE(molecule_set)
               molecule_set(imolecule)%first_atom = first_atom(imolecule)
            END DO

         END IF

         IF (PRESENT(last_atom)) THEN

            IF (SIZE(last_atom) /= SIZE(molecule_set)) THEN
               CALL cp_abort(cp__l("subsys/molecule_types_new.F",450), &
                             "The sizes of last_atom and molecule_set "// &
                             "are different")
            END IF

            DO imolecule = 1, SIZE(molecule_set)
               molecule_set(imolecule)%last_atom = last_atom(imolecule)
            END DO

         END IF

      ELSE

         CALL cp__b("subsys/molecule_types_new.F",463,"The pointer molecule_set is not associated")

      END IF

   END SUBROUTINE set_molecule_set

! **************************************************************************************************
!> \brief   finds for each atom the molecule it belongs to
!> \param molecule_set ...
!> \param atom_to_mol ...
! **************************************************************************************************
   SUBROUTINE molecule_of_atom(molecule_set, atom_to_mol)
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: atom_to_mol

      CHARACTER(len=*), PARAMETER :: routineN = 'molecule_of_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, iatom, imol, last_atom
      TYPE(molecule_type), POINTER                       :: molecule

      DO imol = 1, SIZE(molecule_set)
         molecule => molecule_set(imol)
         CALL get_molecule(molecule=molecule, first_atom=first_atom, last_atom=last_atom)
         DO iatom = first_atom, last_atom
            atom_to_mol(iatom) = imol
         ENDDO ! iatom
      END DO ! imol

   END SUBROUTINE molecule_of_atom

! **************************************************************************************************
!> \brief returns information about molecules in the set.
!> \param molecule_set ...
!> \param atom_to_mol ...
!> \param mol_to_first_atom ...
!> \param mol_to_last_atom ...
!> \param mol_to_nelectrons ...
!> \param mol_to_nbasis ...
!> \param mol_to_charge ...
!> \param mol_to_multiplicity ...
!> \par History
!>       2011.06 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE get_molecule_set_info(molecule_set, atom_to_mol, mol_to_first_atom, &
                                    mol_to_last_atom, mol_to_nelectrons, mol_to_nbasis, mol_to_charge, &
                                    mol_to_multiplicity)

      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL :: atom_to_mol, mol_to_first_atom, &
         mol_to_last_atom, mol_to_nelectrons, mol_to_nbasis, mol_to_charge, mol_to_multiplicity

      CHARACTER(len=*), PARAMETER :: routineN = 'get_molecule_set_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, iatom, imol, last_atom, &
                                                            nbasis, nelec
      REAL(KIND=dp)                                      :: charge
      TYPE(molecule_kind_type), POINTER                  :: imol_kind
      TYPE(molecule_type), POINTER                       :: molecule

      DO imol = 1, SIZE(molecule_set)

         molecule => molecule_set(imol)
         CALL get_molecule(molecule=molecule, molecule_kind=imol_kind, &
                           first_atom=first_atom, last_atom=last_atom)

         IF (PRESENT(mol_to_nelectrons)) THEN
            CALL get_molecule_kind(imol_kind, nelectron=nelec)
            mol_to_nelectrons(imol) = nelec
         ENDIF

         IF (PRESENT(mol_to_multiplicity)) THEN
            ! RZK-warning: At the moment we can only get the total number
            !  of electrons (alpha+beta) and we do not have a way to get the multiplicity of mols.
            !  Therefore, the best we can do is to assume the singlet state for even number of electrons
            !  and doublet state for odd number of electrons (assume ne_alpha > ne_beta).
            !  The best way to implement a correct multiplicity subroutine in the future is to get
            !  the number of alpha and beta e- for each atom from init_atom_electronic_state. This way (as opposed to
            !  reading the multiplicities from file) the number of occupied and virtual orbitals
            !  will be consistent with atomic guess. A guess with broken symmetry will be easy to
            !  implement as well.
            CALL get_molecule_kind(imol_kind, nelectron=nelec)
            IF (MOD(nelec, 2) .EQ. 0) THEN
               mol_to_multiplicity(imol) = 1
            ELSE
               mol_to_multiplicity(imol) = 2
            ENDIF
         ENDIF

         IF (PRESENT(mol_to_charge)) THEN
            CALL get_molecule_kind(imol_kind, charge=charge)
            mol_to_charge(imol) = NINT(charge)
         ENDIF

         IF (PRESENT(mol_to_nbasis)) THEN
            CALL get_molecule_kind(imol_kind, nsgf=nbasis)
            mol_to_nbasis(imol) = nbasis
         ENDIF

         IF (PRESENT(mol_to_first_atom)) THEN
            mol_to_first_atom(imol) = first_atom
         ENDIF

         IF (PRESENT(mol_to_last_atom)) THEN
            mol_to_last_atom(imol) = last_atom
         ENDIF

         IF (PRESENT(atom_to_mol)) THEN
            DO iatom = first_atom, last_atom
               atom_to_mol(iatom) = imol
            ENDDO ! iatom
         ENDIF

      END DO ! imol

   END SUBROUTINE get_molecule_set_info

END MODULE molecule_types_new
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH [04042007] refactoring
! **************************************************************************************************
MODULE virial_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'virial_types'
   INTEGER, PRIVATE, SAVE :: last_virial_id_nr = 0

   PUBLIC :: virial_type, virial_p_type, virial_set, cp_virial, &
             virial_create, virial_release, zero_virial, sym_virial

! **************************************************************************************************
   TYPE virial_type
      INTEGER                            :: ref_count, id_nr
      REAL(KIND=dp), DIMENSION(3, 3) :: pv_total, &
                                        pv_virial, &
                                        pv_kinetic, &
                                        pv_constraint, &
                                        pv_xc, &
                                        pv_fock_4c
      LOGICAL                            :: pv_availability, &
                                            pv_calculate, &
                                            pv_numer, &
                                            pv_diagonal
   END TYPE virial_type

! **************************************************************************************************
   TYPE virial_p_type
      TYPE(virial_type), POINTER        :: virial
   END TYPE virial_p_type

CONTAINS

! **************************************************************************************************
!> \brief   copy virial_in into virial_out
!> \param virial_in ...
!> \param virial_out ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_virial(virial_in, virial_out)
      TYPE(virial_type), POINTER                         :: virial_in, virial_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_virial', routineP = moduleN//':'//routineN

      virial_out%pv_total = virial_in%pv_total
      virial_out%pv_kinetic = virial_in%pv_kinetic
      virial_out%pv_virial = virial_in%pv_virial
      virial_out%pv_xc = virial_in%pv_xc
      virial_out%pv_fock_4c = virial_in%pv_fock_4c
      virial_out%pv_constraint = virial_in%pv_constraint
      virial_out%pv_availability = virial_in%pv_availability
      virial_out%pv_calculate = virial_in%pv_calculate
      virial_out%pv_numer = virial_in%pv_numer
      virial_out%pv_diagonal = virial_in%pv_diagonal

   END SUBROUTINE cp_virial

! **************************************************************************************************
!> \brief   symmetrize the virial components
!> \param virial ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE sym_virial(virial)
      TYPE(virial_type), POINTER                         :: virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sym_virial', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      DO i = 1, 3
         DO j = 1, i-1
            ! Symmetrize total
            virial%pv_total(j, i) = (virial%pv_total(i, j)+virial%pv_total(j, i))*0.5_dp
            virial%pv_total(i, j) = virial%pv_total(j, i)
            ! Symmetrize Kinetic
            virial%pv_kinetic(j, i) = (virial%pv_kinetic(i, j)+virial%pv_kinetic(j, i))*0.5_dp
            virial%pv_kinetic(i, j) = virial%pv_kinetic(j, i)
            ! Symmetrize Virial
            virial%pv_virial(j, i) = (virial%pv_virial(i, j)+virial%pv_virial(j, i))*0.5_dp
            virial%pv_virial(i, j) = virial%pv_virial(j, i)
            ! Symmetrize XC
            virial%pv_xc(j, i) = (virial%pv_xc(i, j)+virial%pv_xc(j, i))*0.5_dp
            virial%pv_xc(i, j) = virial%pv_xc(j, i)
            ! Symmetrize HFX
            virial%pv_fock_4c(j, i) = (virial%pv_fock_4c(i, j)+virial%pv_fock_4c(j, i))*0.5_dp
            virial%pv_fock_4c(i, j) = virial%pv_fock_4c(j, i)

            ! Symmetrize constraints
            virial%pv_constraint(j, i) = (virial%pv_constraint(i, j)+virial%pv_constraint(j, i))*0.5_dp
            virial%pv_constraint(i, j) = virial%pv_constraint(j, i)
         END DO
      END DO

   END SUBROUTINE sym_virial

! **************************************************************************************************
!> \brief ...
!> \param virial ...
!> \param reset ...
! **************************************************************************************************
   SUBROUTINE zero_virial(virial, reset)

      TYPE(virial_type), INTENT(OUT)                     :: virial
      LOGICAL, INTENT(IN), OPTIONAL                      :: reset

      CHARACTER(LEN=*), PARAMETER :: routineN = 'zero_virial', routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_reset

      my_reset = .TRUE.
      IF (PRESENT(reset)) my_reset = reset

      virial%pv_total = 0.0_dp
      virial%pv_kinetic = 0.0_dp
      virial%pv_virial = 0.0_dp
      virial%pv_xc = 0.0_dp
      virial%pv_fock_4c = 0.0_dp
      virial%pv_constraint = 0.0_dp
      IF (my_reset) THEN
         virial%pv_availability = .FALSE.
         virial%pv_numer = .FALSE.
         virial%pv_calculate = .FALSE.
         virial%pv_diagonal = .FALSE.
      END IF
   END SUBROUTINE zero_virial

! **************************************************************************************************
!> \brief ...
!> \param virial ...
!> \param pv_total ...
!> \param pv_virial ...
!> \param pv_xc ...
!> \param pv_fock_4c ...
!> \param pv_constraint ...
!> \param pv_kinetic ...
!> \param pv_availability ...
!> \param pv_calculate ...
!> \param pv_numer ...
!> \param pv_diagonal ...
! **************************************************************************************************
   SUBROUTINE virial_set(virial, pv_total, pv_virial, pv_xc, pv_fock_4c, pv_constraint, &
                         pv_kinetic, pv_availability, pv_calculate, pv_numer, pv_diagonal)

      TYPE(virial_type), POINTER                         :: virial
      REAL(KIND=dp), DIMENSION(3, 3), OPTIONAL           :: pv_total, pv_virial, pv_xc, pv_fock_4c, &
                                                            pv_constraint, pv_kinetic
      LOGICAL, OPTIONAL                                  :: pv_availability, pv_calculate, pv_numer, &
                                                            pv_diagonal

      CHARACTER(LEN=*), PARAMETER :: routineN = 'virial_set', routineP = moduleN//':'//routineN

      IF (PRESENT(pv_total)) virial%pv_total = pv_total
      IF (PRESENT(pv_virial)) virial%pv_virial = pv_virial
      IF (PRESENT(pv_xc)) virial%pv_xc = pv_xc
      IF (PRESENT(pv_fock_4c)) virial%pv_fock_4c = pv_fock_4c
      IF (PRESENT(pv_constraint)) virial%pv_constraint = pv_constraint
      IF (PRESENT(pv_kinetic)) virial%pv_kinetic = pv_kinetic
      IF (PRESENT(pv_availability)) virial%pv_availability = pv_availability
      IF (PRESENT(pv_calculate)) virial%pv_calculate = pv_calculate
      IF (PRESENT(pv_numer)) virial%pv_numer = pv_numer
      IF (PRESENT(pv_diagonal)) virial%pv_diagonal = pv_diagonal

   END SUBROUTINE virial_set

! **************************************************************************************************
!> \brief ...
!> \param virial ...
! **************************************************************************************************
   SUBROUTINE virial_create(virial)

      TYPE(virial_type), POINTER                         :: virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'virial_create', routineP = moduleN//':'//routineN

      ALLOCATE (virial)

      CALL zero_virial(virial)
      last_virial_id_nr = last_virial_id_nr+1
      virial%id_nr = last_virial_id_nr
      virial%ref_count = 1
   END SUBROUTINE virial_create

! **************************************************************************************************
!> \brief releases the given virial_type
!> \param virial the virial_type to release
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE virial_release(virial)
      TYPE(virial_type), POINTER                         :: virial

      CHARACTER(len=*), PARAMETER :: routineN = 'virial_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(virial)) THEN
         IF(.NOT.(virial%ref_count > 0))CALL cp__a("subsys/virial_types.F",209)
         virial%ref_count = virial%ref_count-1
         IF (virial%ref_count .EQ. 0) THEN
            DEALLOCATE (virial)
         ENDIF
         NULLIFY (virial)
      END IF
   END SUBROUTINE virial_release

END MODULE virial_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a simple array based list of the given type
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE mol_new_list_types
   USE molecule_types_new,              ONLY: deallocate_molecule_set,&
                                              molecule_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mol_new_list_types'
   INTEGER, PRIVATE, SAVE :: last_mol_new_list_id = 0

   !API
   PUBLIC :: mol_new_list_type, mol_new_list_p_type, &
             mol_new_list_create, mol_new_list_retain, &
             mol_new_list_release

!***

! **************************************************************************************************
!> \brief represent a list of objects
!> \param id_nr identification number of this list
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param n_el the number of elements in the list
!> \param owns_list if the els are owned by this structure, and
!>        should be deallocated by it
!> \param list the array of object, might be oversized,
!>        only the fist n_el have some meaning
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE mol_new_list_type
      INTEGER :: id_nr, ref_count, n_els
      LOGICAL :: owns_els
      TYPE(molecule_type), DIMENSION(:), POINTER :: els
   END TYPE mol_new_list_type

! **************************************************************************************************
!> \brief represents a pointer to a list
!> \param list the pointer to the list
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE mol_new_list_p_type
      TYPE(mol_new_list_type), POINTER :: list
   END TYPE mol_new_list_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a list
!> \param list the list to allocate and initialize
!> \param els_ptr the elements to store in the list (the array is only,
!>        referenced, not copied!)
!> \param owns_els if the list takes the ownership of els_ptr and
!>        will deallocate it (defaults to true)
!> \param n_els number of elements in the list (at least one els_ptr or
!>        n_els should be given)
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE mol_new_list_create(list, els_ptr, &
                                  owns_els, n_els)
      TYPE(mol_new_list_type), OPTIONAL, POINTER         :: list
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: els_ptr
      LOGICAL, INTENT(in), OPTIONAL                      :: owns_els
      INTEGER, INTENT(in), OPTIONAL                      :: n_els

      CHARACTER(len=*), PARAMETER :: routineN = 'mol_new_list_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(PRESENT(els_ptr) .OR. PRESENT(n_els)))CALL cp__a("subsys/mol_new_list_types.F",110)

      ALLOCATE (list)
      last_mol_new_list_id = last_mol_new_list_id+1
      list%id_nr = last_mol_new_list_id
      list%ref_count = 1
      list%owns_els = .TRUE.
      list%n_els = 0
      IF (PRESENT(owns_els)) list%owns_els = owns_els
      NULLIFY (list%els)
      IF (PRESENT(els_ptr)) THEN
         list%els => els_ptr
         IF (ASSOCIATED(els_ptr)) THEN
            list%n_els = SIZE(els_ptr)
         END IF
      END IF
      IF (PRESENT(n_els)) list%n_els = n_els
      IF (.NOT. ASSOCIATED(list%els)) THEN
         ALLOCATE (list%els(list%n_els))
         IF(.NOT.(list%owns_els))CALL cp__a("subsys/mol_new_list_types.F",129)
      END IF
   END SUBROUTINE mol_new_list_create

! **************************************************************************************************
!> \brief retains a list (see doc/ReferenceCounting.html)
!> \param list the list to retain
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE mol_new_list_retain(list)
      TYPE(mol_new_list_type), POINTER                   :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'mol_new_list_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(list)))CALL cp__a("subsys/mol_new_list_types.F",146)
      IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/mol_new_list_types.F",147)
      list%ref_count = list%ref_count+1
   END SUBROUTINE mol_new_list_retain

! **************************************************************************************************
!> \brief releases a  list (see doc/ReferenceCounting.html)
!> \param list the list to release
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE mol_new_list_release(list)
      TYPE(mol_new_list_type), POINTER                   :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'mol_new_list_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(list)) THEN
         IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/mol_new_list_types.F",165)
         list%ref_count = list%ref_count-1
         IF (list%ref_count == 0) THEN
            IF (list%owns_els) THEN
               IF (ASSOCIATED(list%els)) THEN
                  CALL deallocate_molecule_set(list%els)
               END IF
            END IF
            NULLIFY (list%els)
            DEALLOCATE (list)
         END IF
      END IF
      NULLIFY (list)
   END SUBROUTINE mol_new_list_release

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(el_typename,el_type,USE,deallocate_els_code)]
! ARGS:
!  USE = "use molecule_types_new, only: molecule_type, deallocate_molecule_set"
!  deallocate_els_code = "call deallocate_molecule_set(list%els)"
!  el_type = "type(molecule_type)"
!  el_typename = "mol_new"

END MODULE mol_new_list_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a simple array based list of the given type
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE mol_kind_new_list_types
   USE molecule_kind_types,             ONLY: deallocate_molecule_kind_set,&
                                              molecule_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mol_kind_new_list_types'
   INTEGER, PRIVATE, SAVE :: last_mol_kind_new_list_id = 0

   !API
   PUBLIC :: mol_kind_new_list_type, mol_kind_new_list_p_type, &
             mol_kind_new_list_create, mol_kind_new_list_retain, &
             mol_kind_new_list_release

!***

! **************************************************************************************************
!> \brief represent a list of objects
!> \param id_nr identification number of this list
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param n_el the number of elements in the list
!> \param owns_list if the els are owned by this structure, and
!>        should be deallocated by it
!> \param list the array of object, might be oversized,
!>        only the fist n_el have some meaning
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE mol_kind_new_list_type
      INTEGER :: id_nr, ref_count, n_els
      LOGICAL :: owns_els
      TYPE(molecule_kind_type), DIMENSION(:), POINTER :: els
   END TYPE mol_kind_new_list_type

! **************************************************************************************************
!> \brief represents a pointer to a list
!> \param list the pointer to the list
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE mol_kind_new_list_p_type
      TYPE(mol_kind_new_list_type), POINTER :: list
   END TYPE mol_kind_new_list_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a list
!> \param list the list to allocate and initialize
!> \param els_ptr the elements to store in the list (the array is only,
!>        referenced, not copied!)
!> \param owns_els if the list takes the ownership of els_ptr and
!>        will deallocate it (defaults to true)
!> \param n_els number of elements in the list (at least one els_ptr or
!>        n_els should be given)
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE mol_kind_new_list_create(list, els_ptr, &
                                       owns_els, n_els)
      TYPE(mol_kind_new_list_type), OPTIONAL, POINTER    :: list
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: els_ptr
      LOGICAL, INTENT(in), OPTIONAL                      :: owns_els
      INTEGER, INTENT(in), OPTIONAL                      :: n_els

      CHARACTER(len=*), PARAMETER :: routineN = 'mol_kind_new_list_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(PRESENT(els_ptr) .OR. PRESENT(n_els)))CALL cp__a("subsys/mol_kind_new_list_types.F",110)

      ALLOCATE (list)
      last_mol_kind_new_list_id = last_mol_kind_new_list_id+1
      list%id_nr = last_mol_kind_new_list_id
      list%ref_count = 1
      list%owns_els = .TRUE.
      list%n_els = 0
      IF (PRESENT(owns_els)) list%owns_els = owns_els
      NULLIFY (list%els)
      IF (PRESENT(els_ptr)) THEN
         list%els => els_ptr
         IF (ASSOCIATED(els_ptr)) THEN
            list%n_els = SIZE(els_ptr)
         END IF
      END IF
      IF (PRESENT(n_els)) list%n_els = n_els
      IF (.NOT. ASSOCIATED(list%els)) THEN
         ALLOCATE (list%els(list%n_els))
         IF(.NOT.(list%owns_els))CALL cp__a("subsys/mol_kind_new_list_types.F",129)
      END IF
   END SUBROUTINE mol_kind_new_list_create

! **************************************************************************************************
!> \brief retains a list (see doc/ReferenceCounting.html)
!> \param list the list to retain
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE mol_kind_new_list_retain(list)
      TYPE(mol_kind_new_list_type), POINTER              :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'mol_kind_new_list_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(list)))CALL cp__a("subsys/mol_kind_new_list_types.F",146)
      IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/mol_kind_new_list_types.F",147)
      list%ref_count = list%ref_count+1
   END SUBROUTINE mol_kind_new_list_retain

! **************************************************************************************************
!> \brief releases a  list (see doc/ReferenceCounting.html)
!> \param list the list to release
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE mol_kind_new_list_release(list)
      TYPE(mol_kind_new_list_type), POINTER              :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'mol_kind_new_list_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(list)) THEN
         IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/mol_kind_new_list_types.F",165)
         list%ref_count = list%ref_count-1
         IF (list%ref_count == 0) THEN
            IF (list%owns_els) THEN
               IF (ASSOCIATED(list%els)) THEN
                  CALL deallocate_molecule_kind_set(list%els)
               END IF
            END IF
            NULLIFY (list%els)
            DEALLOCATE (list)
         END IF
      END IF
      NULLIFY (list)
   END SUBROUTINE mol_kind_new_list_release

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(el_typename,el_type,USE,deallocate_els_code)]
! ARGS:
!  USE = "use molecule_kind_types, only: molecule_kind_type, deallocate_molecule_kind_set"
!  deallocate_els_code = "call deallocate_molecule_kind_set(list%els)"
!  el_type = "type(molecule_kind_type)"
!  el_typename = "mol_kind_new"

END MODULE mol_kind_new_list_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a simple array based list of the given type
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE atomic_kind_list_types
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              deallocate_atomic_kind_set
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atomic_kind_list_types'
   INTEGER, PRIVATE, SAVE :: last_atomic_kind_list_id = 0

   !API
   PUBLIC :: atomic_kind_list_type, atomic_kind_list_p_type, &
             atomic_kind_list_create, atomic_kind_list_retain, &
             atomic_kind_list_release

!***

! **************************************************************************************************
!> \brief represent a list of objects
!> \param id_nr identification number of this list
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param n_el the number of elements in the list
!> \param owns_list if the els are owned by this structure, and
!>        should be deallocated by it
!> \param list the array of object, might be oversized,
!>        only the fist n_el have some meaning
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE atomic_kind_list_type
      INTEGER :: id_nr, ref_count, n_els
      LOGICAL :: owns_els
      TYPE(atomic_kind_type), DIMENSION(:), POINTER :: els
   END TYPE atomic_kind_list_type

! **************************************************************************************************
!> \brief represents a pointer to a list
!> \param list the pointer to the list
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE atomic_kind_list_p_type
      TYPE(atomic_kind_list_type), POINTER :: list
   END TYPE atomic_kind_list_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a list
!> \param list the list to allocate and initialize
!> \param els_ptr the elements to store in the list (the array is only,
!>        referenced, not copied!)
!> \param owns_els if the list takes the ownership of els_ptr and
!>        will deallocate it (defaults to true)
!> \param n_els number of elements in the list (at least one els_ptr or
!>        n_els should be given)
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE atomic_kind_list_create(list, els_ptr, &
                                      owns_els, n_els)
      TYPE(atomic_kind_list_type), OPTIONAL, POINTER     :: list
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: els_ptr
      LOGICAL, INTENT(in), OPTIONAL                      :: owns_els
      INTEGER, INTENT(in), OPTIONAL                      :: n_els

      CHARACTER(len=*), PARAMETER :: routineN = 'atomic_kind_list_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(PRESENT(els_ptr) .OR. PRESENT(n_els)))CALL cp__a("subsys/atomic_kind_list_types.F",110)

      ALLOCATE (list)
      last_atomic_kind_list_id = last_atomic_kind_list_id+1
      list%id_nr = last_atomic_kind_list_id
      list%ref_count = 1
      list%owns_els = .TRUE.
      list%n_els = 0
      IF (PRESENT(owns_els)) list%owns_els = owns_els
      NULLIFY (list%els)
      IF (PRESENT(els_ptr)) THEN
         list%els => els_ptr
         IF (ASSOCIATED(els_ptr)) THEN
            list%n_els = SIZE(els_ptr)
         END IF
      END IF
      IF (PRESENT(n_els)) list%n_els = n_els
      IF (.NOT. ASSOCIATED(list%els)) THEN
         ALLOCATE (list%els(list%n_els))
         IF(.NOT.(list%owns_els))CALL cp__a("subsys/atomic_kind_list_types.F",129)
      END IF
   END SUBROUTINE atomic_kind_list_create

! **************************************************************************************************
!> \brief retains a list (see doc/ReferenceCounting.html)
!> \param list the list to retain
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE atomic_kind_list_retain(list)
      TYPE(atomic_kind_list_type), POINTER               :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'atomic_kind_list_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(list)))CALL cp__a("subsys/atomic_kind_list_types.F",146)
      IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/atomic_kind_list_types.F",147)
      list%ref_count = list%ref_count+1
   END SUBROUTINE atomic_kind_list_retain

! **************************************************************************************************
!> \brief releases a  list (see doc/ReferenceCounting.html)
!> \param list the list to release
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE atomic_kind_list_release(list)
      TYPE(atomic_kind_list_type), POINTER               :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'atomic_kind_list_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(list)) THEN
         IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/atomic_kind_list_types.F",165)
         list%ref_count = list%ref_count-1
         IF (list%ref_count == 0) THEN
            IF (list%owns_els) THEN
               IF (ASSOCIATED(list%els)) THEN
                  CALL deallocate_atomic_kind_set(list%els)
               END IF
            END IF
            NULLIFY (list%els)
            DEALLOCATE (list)
         END IF
      END IF
      NULLIFY (list)
   END SUBROUTINE atomic_kind_list_release

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(el_typename,el_type,USE,deallocate_els_code)]
! ARGS:
!  USE = "use atomic_kind_types, only: atomic_kind_type,deallocate_atomic_kind_set"
!  deallocate_els_code = "call deallocate_atomic_kind_set(list%els)"
!  el_type = "type(atomic_kind_type)"
!  el_typename = "atomic_kind"

END MODULE atomic_kind_list_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  set of type/routines to handle the storage of results in force_envs
!> \author fschiff (12.2007)
!> \par    History
!>         - 10.2008 Teodoro Laino [tlaino] - University of Zurich
!>                   major rewriting:
!>                   - information stored in a proper type (not in a character!)
!>                   - module more lean
!>                   - splitting types and creating methods for cp_results
! **************************************************************************************************
MODULE cp_result_types

   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_result_types'

   INTEGER, PARAMETER, PUBLIC :: result_type_logical = 1, &
                                 result_type_integer = 2, &
                                 result_type_real = 3

! *** Public data types ***
   PUBLIC :: cp_result_type, &
             cp_result_p_type

! *** Public subroutines ***
   PUBLIC :: cp_result_create, &
             cp_result_release, &
             cp_result_retain, &
             cp_result_clean, &
             cp_result_copy, &
             cp_result_value_create, &
             cp_result_value_copy, &
             cp_result_value_p_reallocate, &
             cp_result_value_init

! **************************************************************************************************
!> \brief low level type for storing real informations
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   TYPE cp_result_value_type
      INTEGER                                              :: type_in_use
      LOGICAL, DIMENSION(:), POINTER                       :: logical_type
      INTEGER, DIMENSION(:), POINTER                       :: integer_type
      REAL(KIND=dp), DIMENSION(:), POINTER                 :: real_type
   END TYPE cp_result_value_type

! **************************************************************************************************
   TYPE cp_result_value_p_type
      TYPE(cp_result_value_type), POINTER                  :: value
   END TYPE cp_result_value_p_type

! **************************************************************************************************
!> \brief contains arbitrary information which need to be stored
!> \note
!>      result_list is a character list, in which everthing can be stored
!>      before passing any variable just name the variable like '[NAME]'
!>      brackets will be used to identify the start of a new set
!> \author fschiff (12.2007)
! **************************************************************************************************
   TYPE cp_result_type
      INTEGER                                              :: ref_count
      TYPE(cp_result_value_p_type), POINTER, DIMENSION(:)  :: result_value
      CHARACTER(LEN=default_string_length), DIMENSION(:), &
         POINTER                                         :: result_label
   END TYPE cp_result_type

! **************************************************************************************************
   TYPE cp_result_p_type
      TYPE(cp_result_type), POINTER                        :: results
   END TYPE cp_result_p_type

CONTAINS

! **************************************************************************************************
!> \brief Allocates and intitializes the cp_result
!> \param results ...
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE cp_result_create(results)
      TYPE(cp_result_type), POINTER                      :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      ALLOCATE (results)
      NULLIFY (results%result_value, results%result_label)
      results%ref_count = 1
      ALLOCATE (results%result_label(0))
      ALLOCATE (results%result_value(0))
      CALL timestop(handle)
   END SUBROUTINE cp_result_create

! **************************************************************************************************
!> \brief Releases cp_result type
!> \param results ...
!> \par History
!>      12.2007 created
!>      10.2008 Teodoro Laino [tlaino] - major rewriting
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE cp_result_release(results)
      TYPE(cp_result_type), POINTER                      :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(results)) THEN
         IF(.NOT.(results%ref_count > 0))CALL cp__a("common/cp_result_types.F",128)
         results%ref_count = results%ref_count-1
         IF (results%ref_count == 0) THEN
            ! Description
            IF (ASSOCIATED(results%result_label)) THEN
               DEALLOCATE (results%result_label)
            END IF
            ! Values
            IF (ASSOCIATED(results%result_value)) THEN
               DO i = 1, SIZE(results%result_value)
                  CALL cp_result_value_release(results%result_value(i)%value)
               END DO
               DEALLOCATE (results%result_value)
            END IF
            DEALLOCATE (results)
         END IF
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_result_release

! **************************************************************************************************
!> \brief Releases cp_result clean
!> \param results ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
! **************************************************************************************************
   SUBROUTINE cp_result_clean(results)
      TYPE(cp_result_type), POINTER                      :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_clean', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(results)) THEN
         ! Description
         IF (ASSOCIATED(results%result_label)) THEN
            DEALLOCATE (results%result_label)
         END IF
         ! Values
         IF (ASSOCIATED(results%result_value)) THEN
            DO i = 1, SIZE(results%result_value)
               CALL cp_result_value_release(results%result_value(i)%value)
            END DO
            DEALLOCATE (results%result_value)
         END IF
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_result_clean

! **************************************************************************************************
!> \brief Retains cp_result type
!> \param results ...
!> \par History
!>      12.2007 created
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE cp_result_retain(results)
      TYPE(cp_result_type), POINTER                      :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(results)))CALL cp__a("common/cp_result_types.F",191)
      IF(.NOT.(results%ref_count > 0))CALL cp__a("common/cp_result_types.F",192)
      results%ref_count = results%ref_count+1
   END SUBROUTINE cp_result_retain

! **************************************************************************************************
!> \brief Allocates and intitializes the cp_result_value type
!> \param value ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE cp_result_value_create(value)
      TYPE(cp_result_value_type), POINTER                :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_value_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      ALLOCATE (value)
      value%type_in_use = -1
      NULLIFY (value%real_type)
      NULLIFY (value%logical_type)
      NULLIFY (value%integer_type)
      CALL timestop(handle)
   END SUBROUTINE cp_result_value_create

! **************************************************************************************************
!> \brief Setup of the cp_result_value type
!> \param value ...
!> \param type_in_use ...
!> \param size_value ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE cp_result_value_init(value, type_in_use, size_value)
      TYPE(cp_result_value_type), POINTER                :: value
      INTEGER, INTENT(IN)                                :: type_in_use, size_value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_value_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(value)))CALL cp__a("common/cp_result_types.F",235)
      value%type_in_use = type_in_use
      SELECT CASE (value%type_in_use)
      CASE (result_type_real)
         ALLOCATE (value%real_type(size_value))
      CASE (result_type_integer)
         ALLOCATE (value%integer_type(size_value))
      CASE (result_type_logical)
         ALLOCATE (value%logical_type(size_value))
      CASE DEFAULT
         ! Type not implemented in cp_result_type
         CALL cp__b("common/cp_result_types.F",246,"")
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE cp_result_value_init

! **************************************************************************************************
!> \brief Releases the cp_result_value type
!> \param value ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE cp_result_value_release(value)
      TYPE(cp_result_value_type), POINTER                :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_value_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(value)) THEN
         SELECT CASE (value%type_in_use)
         CASE (result_type_real)
            IF (ASSOCIATED(value%real_type)) THEN
               DEALLOCATE (value%real_type)
            END IF
            IF(.NOT.(.NOT. ASSOCIATED(value%integer_type)))CALL cp__a("common/cp_result_types.F",271)
            IF(.NOT.(.NOT. ASSOCIATED(value%logical_type)))CALL cp__a("common/cp_result_types.F",272)
         CASE (result_type_integer)
            IF (ASSOCIATED(value%integer_type)) THEN
               DEALLOCATE (value%integer_type)
            END IF
            IF(.NOT.(.NOT. ASSOCIATED(value%real_type)))CALL cp__a("common/cp_result_types.F",277)
            IF(.NOT.(.NOT. ASSOCIATED(value%logical_type)))CALL cp__a("common/cp_result_types.F",278)
         CASE (result_type_logical)
            IF (ASSOCIATED(value%logical_type)) THEN
               DEALLOCATE (value%logical_type)
            END IF
            IF(.NOT.(.NOT. ASSOCIATED(value%integer_type)))CALL cp__a("common/cp_result_types.F",283)
            IF(.NOT.(.NOT. ASSOCIATED(value%real_type)))CALL cp__a("common/cp_result_types.F",284)
         CASE DEFAULT
            ! Type not implemented in cp_result_type
            CALL cp__b("common/cp_result_types.F",287,"")
         END SELECT
         DEALLOCATE (value)
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_result_value_release

! **************************************************************************************************
!> \brief Copies the cp_result type
!> \param results_in ...
!> \param results_out ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE cp_result_copy(results_in, results_out)
      TYPE(cp_result_type), POINTER                      :: results_in, results_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_copy', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ndim
      LOGICAL                                            :: check

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(results_in)))CALL cp__a("common/cp_result_types.F",309)
      IF(.NOT.(ASSOCIATED(results_out)))CALL cp__a("common/cp_result_types.F",310)
      CALL cp_result_clean(results_out)

      check = SIZE(results_in%result_label) == SIZE(results_in%result_value)
      IF(.NOT.(check))CALL cp__a("common/cp_result_types.F",314)
      ndim = SIZE(results_in%result_value)
      ALLOCATE (results_out%result_label(ndim))
      ALLOCATE (results_out%result_value(ndim))
      DO i = 1, ndim
         results_out%result_label(i) = results_in%result_label(i)
         CALL cp_result_value_create(results_out%result_value(i)%value)
         CALL cp_result_value_copy(results_out%result_value(i)%value, &
                                   results_in%result_value(i)%value)
      END DO
      CALL timestop(handle)
   END SUBROUTINE cp_result_copy

! **************************************************************************************************
!> \brief Copies the cp_result_value type
!> \param value_out ...
!> \param value_in ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE cp_result_value_copy(value_out, value_in)
      TYPE(cp_result_value_type), POINTER                :: value_out, value_in

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_value_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, isize

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(value_in)))CALL cp__a("common/cp_result_types.F",342)
      IF(.NOT.(ASSOCIATED(value_out)))CALL cp__a("common/cp_result_types.F",343)
      value_out%type_in_use = value_in%type_in_use
      SELECT CASE (value_out%type_in_use)
      CASE (result_type_real)
         isize = SIZE(value_in%real_type)
         ALLOCATE (value_out%real_type(isize))
         value_out%real_type = value_in%real_type
      CASE (result_type_integer)
         isize = SIZE(value_in%integer_type)
         ALLOCATE (value_out%integer_type(isize))
         value_out%integer_type = value_in%integer_type
      CASE (result_type_logical)
         isize = SIZE(value_in%logical_type)
         ALLOCATE (value_out%logical_type(isize))
         value_out%logical_type = value_in%logical_type
      CASE DEFAULT
         ! Type not implemented in cp_result_type
         CALL cp__b("common/cp_result_types.F",360,"")
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE cp_result_value_copy

! **************************************************************************************************
!> \brief Reallocates the cp_result_value type
!> \param result_value ...
!> \param istart ...
!> \param iend ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE cp_result_value_p_reallocate(result_value, istart, iend)
      TYPE(cp_result_value_p_type), DIMENSION(:), &
         POINTER                                         :: result_value
      INTEGER, INTENT(in)                                :: istart, iend

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_result_value_p_reallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, lb_size, ub_size
      TYPE(cp_result_value_p_type), DIMENSION(:), &
         POINTER                                         :: tmp_value

      CALL timeset(routineN, handle)
      ub_size = 0
      lb_size = 0
      IF (ASSOCIATED(result_value)) THEN
         ub_size = UBOUND(result_value, 1)
         lb_size = LBOUND(result_value, 1)
      END IF
      ! Allocate and copy new values while releases old
      ALLOCATE (tmp_value(istart:iend))
      DO i = istart, iend
         NULLIFY (tmp_value(i)%value)
         CALL cp_result_value_create(tmp_value(i)%value)
         IF ((i <= ub_size) .AND. (i >= lb_size)) THEN
            CALL cp_result_value_copy(tmp_value(i)%value, result_value(i)%value)
            CALL cp_result_value_release(result_value(i)%value)
         END IF
      END DO
      DEALLOCATE (result_value)
      result_value => tmp_value
      CALL timestop(handle)
   END SUBROUTINE cp_result_value_p_reallocate

END MODULE cp_result_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (11.08.2002) exchange and correlation energy now in exc
!>      TL  (25.05.2004) qmmm energy
!> \author MK (13.06.2002)
! **************************************************************************************************
MODULE qs_energy_types

   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_energy_types'

   TYPE qs_energy_type
      REAL(KIND=dp) :: core, &
                       core_overlap, &
                       core_overlap0, &
                       core_self, &
                       repulsive, &
                       dispersion, &
                       ex, &
                       exc, &
                       exc_aux_fit, &
                       exc1, &
                       hartree, &
                       hartree_1c, &
                       e_hartree, & ! contains the hartree energy of electrons only
                       ! computed if requested
                       image_charge, &
                       qmmm_el, &
                       qmmm_nu, &
                       mulliken, &
                       becke, &
                       ee, &
                       ee_core, &
                       efield, &
                       efield_core, &
                       s2_restraint, &
                       dft_plus_u, & ! DFT+U energy contribution
                       sccs_hartree, & ! SCCS Hartree energy of solute + solvent
                       sccs_pol, & ! SCCS polarisation energy
                       sccs_mpc, & ! Makov-Payne energy correction for charged supercells
                       sccs_cav, & ! SCCS cavitation energy
                       sccs_dis, & ! SCCS dispersion free energy
                       sccs_rep, & ! SCCS repulsion free energy
                       ktS, & ! electronic entropic contribution
                       efermi, & ! Fermi energy
                       dftb3, & ! DFTB 3rd order correction
                       mp2, &
                       total, &
                       tot_old, &
                       kinetic, & !total kinetic energy [rk]
                       s_square, &
                       surf_dipole
      REAL(KIND=dp), DIMENSION(:), POINTER :: ddapc_restraint
   END TYPE qs_energy_type

   ! Public data types

   PUBLIC :: qs_energy_type

   ! Public subroutines

   PUBLIC :: allocate_qs_energy, &
             deallocate_qs_energy, &
             init_qs_energy

CONTAINS

! **************************************************************************************************
!> \brief   Allocate and/or initialise a Quickstep energy data structure.
!> \param qs_energy ...
!> \date    13.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_qs_energy(qs_energy)

      TYPE(qs_energy_type), POINTER                      :: qs_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_qs_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, istat

      CALL timeset(routineN, handle)
      IF (.NOT. ASSOCIATED(qs_energy)) THEN
         ALLOCATE (qs_energy, STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 101, &
                                          "qs_energy", int_size)
         NULLIFY (qs_energy%ddapc_restraint)
      END IF

      CALL init_qs_energy(qs_energy)
      CALL timestop(handle)

   END SUBROUTINE allocate_qs_energy

! **************************************************************************************************
!> \brief   Deallocate a Quickstep energy data structure.
!> \param qs_energy ...
!> \date    13.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_qs_energy(qs_energy)
      TYPE(qs_energy_type), POINTER                      :: qs_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_qs_energy', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(qs_energy)) THEN
         IF (ASSOCIATED(qs_energy%ddapc_restraint)) THEN
            DEALLOCATE (qs_energy%ddapc_restraint)
         END IF
         DEALLOCATE (qs_energy)
      ELSE
         CALL cp_abort(cp__l("qs_energy_types.F",130), &
                       "The qs_energy pointer is not associated "// &
                       "and cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_qs_energy

! **************************************************************************************************
!> \brief   Initialise a Quickstep energy data structure.
!> \param qs_energy ...
!> \date    13.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_qs_energy(qs_energy)

      TYPE(qs_energy_type), POINTER                      :: qs_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'init_qs_energy', routineP = moduleN//':'//routineN

      INTEGER                                            :: istat

      IF (ASSOCIATED(qs_energy)) THEN
         qs_energy%core = 0.0_dp
         qs_energy%core_overlap = 0.0_dp
         qs_energy%core_overlap0 = 0.0_dp
         qs_energy%core_self = 0.0_dp
         qs_energy%repulsive = 0.0_dp
         qs_energy%dispersion = 0.0_dp
         qs_energy%qmmm_el = 0.0_dp
         qs_energy%qmmm_nu = 0.0_dp
         qs_energy%ex = 0.0_dp
         qs_energy%exc = 0.0_dp
         qs_energy%exc_aux_fit = 0.0_dp
         qs_energy%exc1 = 0.0_dp
         qs_energy%e_hartree = 0.0_dp
         qs_energy%hartree = 0.0_dp
         qs_energy%hartree_1c = 0.0_dp
         qs_energy%image_charge = 0.0_dp
         qs_energy%mulliken = 0.0_dp
         qs_energy%becke = 0.0_dp
         qs_energy%efield = 0.0_dp
         qs_energy%efield_core = 0.0_dp
         qs_energy%ee = 0.0_dp
         qs_energy%ee_core = 0.0_dp
         qs_energy%s2_restraint = 0.0_dp
         qs_energy%dft_plus_u = 0.0_dp
         qs_energy%sccs_hartree = 0.0_dp
         qs_energy%sccs_pol = 0.0_dp
         qs_energy%sccs_mpc = 0.0_dp
         qs_energy%sccs_cav = 0.0_dp
         qs_energy%sccs_dis = 0.0_dp
         qs_energy%sccs_rep = 0.0_dp
         qs_energy%dftb3 = 0.0_dp
         qs_energy%kTS = 0.0_dp
         qs_energy%mp2 = 0.0_dp
         qs_energy%efermi = 0.0_dp
         qs_energy%kinetic = 0.0_dp
         qs_energy%surf_dipole = 0.0_dp
         qs_energy%total = 0.0_dp
         IF (.NOT. ASSOCIATED(qs_energy%ddapc_restraint)) THEN
            ALLOCATE (qs_energy%ddapc_restraint(1), STAT=istat)
            IF (istat /= 0) THEN
               CALL stop_memory(routineN, moduleN, 193, &
                                "qs_energy%ddapc_restraint", dp_size)
            END IF
         END IF
         qs_energy%ddapc_restraint(:) = 0.0_dp
      ELSE
         CALL cp_abort(cp__l("qs_energy_types.F",199), &
                       "The qs_energy pointer is not associated "// &
                       "and cannot be initialised")
      END IF

   END SUBROUTINE init_qs_energy

END MODULE qs_energy_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief stores a lists of integer that are local to a processor.
!>      The idea is that these integers represent objects that are distributed
!>      between the different processors.
!>      The ordering is just to make some operation more efficent, logically
!>      these lists are like sets.
!>      Some operations assume that the integers represent a range of values
!>      from 1 to a (not too big) maxval, and that an element is present just
!>      once, and only on a processor (these assumption are marked in the
!>      documentation of such operations).
!>      The concrete task for which this structure was developed was
!>      distributing atoms between the processors.
!> \par History
!>      05.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE distribution_1d_types

   USE cp_array_utils_i,                ONLY: cp_1d_i_p_type
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE parallel_rng_types,              ONLY: delete_rng_stream,&
                                              rng_stream_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'distribution_1d_types'
   INTEGER, SAVE, PRIVATE :: distribution_1d_last_id_nr = 0

   PUBLIC :: distribution_1d_type
   PUBLIC :: distribution_1d_create, distribution_1d_retain, distribution_1d_release

! **************************************************************************************************
   TYPE local_particle_type
      TYPE(rng_stream_p_type), DIMENSION(:), POINTER :: rng
   END TYPE local_particle_type
!***

! **************************************************************************************************
!> \brief structure to store local (to a processor) ordered lists of integers.
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param n_el n_el(i) is number of elements of list(i)
!> \param list list(i) contains an ordered list of integer (the array
!>        might be bigger than n_el(i), but the extra elements should be
!>        ignored)
!> \param para_env the parallel environment underlying the distribution
!> \param listbased_distribution true if each list has its own
!>        distribution
!> \par History
!>      06.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE distribution_1d_type
      INTEGER :: ref_count, id_nr
      LOGICAL :: listbased_distribution
      INTEGER, DIMENSION(:), POINTER :: n_el
      TYPE(cp_1d_i_p_type), DIMENSION(:), POINTER :: list
      TYPE(cp_para_env_type), POINTER :: para_env
      TYPE(local_particle_type), DIMENSION(:), POINTER :: local_particle_set
   END TYPE distribution_1d_type

! **************************************************************************************************
!> \brief pointer to a distribution_1d_type
!> \param distribution_1d: the pointer to the distribution_1d
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE distribution_1d_p_type
      TYPE(distribution_1d_type), POINTER :: distribution_1d
   END TYPE distribution_1d_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a local list
!> \param distribution_1d the lists to create
!> \param para_env parallel environement to be used
!> \param listbased_distribution if each list has its own distribution
!>        (defaults to false)
!> \param n_el number of elements in each list (defaults to 0)
!> \param n_lists number of lists to create (defaults to 1, or size(n_el))
!> \par History
!>      05.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE distribution_1d_create(distribution_1d, para_env, listbased_distribution, &
                                     n_el, n_lists)
      TYPE(distribution_1d_type), POINTER                :: distribution_1d
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(in), OPTIONAL                      :: listbased_distribution
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: n_el
      INTEGER, INTENT(in), OPTIONAL                      :: n_lists

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_1d_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ilist, my_n_lists, stat

      my_n_lists = 1
      IF (PRESENT(n_el)) my_n_lists = SIZE(n_el)
      IF (PRESENT(n_lists)) my_n_lists = n_lists

      ALLOCATE (distribution_1d)
      distribution_1d_last_id_nr = distribution_1d_last_id_nr+1
      distribution_1d%id_nr = distribution_1d_last_id_nr

      distribution_1d%ref_count = 1

      distribution_1d%para_env => para_env
      CALL cp_para_env_retain(para_env)

      NULLIFY (distribution_1d%list, distribution_1d%n_el)

      distribution_1d%listbased_distribution = .FALSE.
      IF (PRESENT(listbased_distribution)) &
         distribution_1d%listbased_distribution = listbased_distribution

      ALLOCATE (distribution_1d%n_el(my_n_lists), distribution_1d%list(my_n_lists), &
                stat=stat)
      IF(.NOT.(stat == 0))CALL cp__a("common/distribution_1d_types.F",131)

      IF (PRESENT(n_el)) THEN
         distribution_1d%n_el(1:my_n_lists) = n_el(1:my_n_lists)
         DO ilist = 1, my_n_lists
            ALLOCATE (distribution_1d%list(ilist)%array(n_el(ilist)))
            distribution_1d%list(ilist)%array = -1
         END DO
      ELSE
         distribution_1d%n_el(1:my_n_lists) = 0
         DO ilist = 1, my_n_lists
            NULLIFY (distribution_1d%list(ilist)%array)
         END DO
      END IF

      NULLIFY (distribution_1d%local_particle_set)

   END SUBROUTINE distribution_1d_create

! **************************************************************************************************
!> \brief retains a distribution_1d
!> \param distribution_1d  the object to retain
!> \par History
!>      05.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE distribution_1d_retain(distribution_1d)
      TYPE(distribution_1d_type), POINTER                :: distribution_1d

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_1d_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(distribution_1d)))CALL cp__a("common/distribution_1d_types.F",163)
      IF(.NOT.(distribution_1d%ref_count > 0))CALL cp__a("common/distribution_1d_types.F",164)
      distribution_1d%ref_count = distribution_1d%ref_count+1
   END SUBROUTINE distribution_1d_retain

! **************************************************************************************************
!> \brief releases the given distribution_1d
!> \param distribution_1d the object to release
!> \par History
!>      05.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE distribution_1d_release(distribution_1d)
      TYPE(distribution_1d_type), POINTER                :: distribution_1d

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_1d_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ilist, iparticle_kind, iparticle_local, &
                                                            nparticle_kind, nparticle_local
      TYPE(local_particle_type), DIMENSION(:), POINTER   :: local_particle_set

      IF (ASSOCIATED(distribution_1d)) THEN
         IF(.NOT.(distribution_1d%ref_count > 0))CALL cp__a("common/distribution_1d_types.F",186)
         distribution_1d%ref_count = distribution_1d%ref_count-1
         IF (distribution_1d%ref_count == 0) THEN
            DEALLOCATE (distribution_1d%n_el)

            DO ilist = 1, SIZE(distribution_1d%list)
               DEALLOCATE (distribution_1d%list(ilist)%array)
            END DO
            DEALLOCATE (distribution_1d%list)

            !MK Delete Wiener process

            local_particle_set => distribution_1d%local_particle_set

            IF (ASSOCIATED(local_particle_set)) THEN
               nparticle_kind = SIZE(local_particle_set)
               DO iparticle_kind = 1, nparticle_kind
                  IF (ASSOCIATED(local_particle_set(iparticle_kind)%rng)) THEN
                     nparticle_local = SIZE(local_particle_set(iparticle_kind)%rng)
                     DO iparticle_local = 1, nparticle_local
                        IF (ASSOCIATED(local_particle_set(iparticle_kind)% &
                                       rng(iparticle_local)%stream)) THEN
                           CALL delete_rng_stream(local_particle_set(iparticle_kind)% &
                                                  rng(iparticle_local)%stream)
                        END IF
                     END DO
                     DEALLOCATE (local_particle_set(iparticle_kind)%rng)
                  END IF
               END DO
               DEALLOCATE (local_particle_set)
            END IF

            CALL cp_para_env_release(distribution_1d%para_env)

            DEALLOCATE (distribution_1d)
         END IF
      END IF

   END SUBROUTINE distribution_1d_release

END MODULE distribution_1d_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Holds information on atomic properties
!> \par History
!>      07.2011 created
!> \author JHU
! **************************************************************************************************
MODULE atprop_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atprop_types'

   PUBLIC :: atprop_type, atprop_create, atprop_release, atprop_init
   PUBLIC :: atprop_array_init, atprop_array_add

! **************************************************************************************************
!> \brief type for the atomic properties
!> \author fawzi
! **************************************************************************************************
   TYPE atprop_type
      LOGICAL                                   :: energy
      LOGICAL                                   :: stress
      REAL(KIND=dp), DIMENSION(:), POINTER      :: atener
      REAL(KIND=dp), DIMENSION(:), POINTER      :: ateb
      REAL(KIND=dp), DIMENSION(:), POINTER      :: atexc
      REAL(KIND=dp), DIMENSION(:), POINTER      :: ateself
      REAL(KIND=dp), DIMENSION(:), POINTER      :: atevdw
      REAL(KIND=dp), DIMENSION(:), POINTER      :: atecc
      REAL(KIND=dp), DIMENSION(:), POINTER      :: ate1c
      REAL(KIND=dp), DIMENSION(:), POINTER      :: atecoul
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER  :: atstress
   END TYPE atprop_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atprop_env ...
! **************************************************************************************************
   SUBROUTINE atprop_create(atprop_env)
      TYPE(atprop_type), POINTER                         :: atprop_env

      CHARACTER(len=*), PARAMETER :: routineN = 'atprop_create', routineP = moduleN//':'//routineN

      CALL atprop_release(atprop_env)
      ALLOCATE (atprop_env)
      NULLIFY (atprop_env%atener, atprop_env%atstress)
      NULLIFY (atprop_env%ateb, atprop_env%atevdw, atprop_env%atecc, atprop_env%atecoul)
      NULLIFY (atprop_env%ateself, atprop_env%atexc, atprop_env%ate1c)
      atprop_env%energy = .FALSE.
      atprop_env%stress = .FALSE.

   END SUBROUTINE atprop_create

! **************************************************************************************************
!> \brief ...
!> \param atprop_env ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE atprop_init(atprop_env, natom)
      TYPE(atprop_type), POINTER                         :: atprop_env
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'atprop_init', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atprop_env)))CALL cp__a("subsys/atprop_types.F",75)

      IF (atprop_env%energy) THEN
         CALL atprop_array_init(atprop_env%atener, natom)
         CALL atprop_array_release(atprop_env%ateb)
         CALL atprop_array_release(atprop_env%atevdw)
         CALL atprop_array_release(atprop_env%atecc)
         CALL atprop_array_release(atprop_env%atecoul)
         CALL atprop_array_release(atprop_env%ateself)
         CALL atprop_array_release(atprop_env%atexc)
         CALL atprop_array_release(atprop_env%ate1c)
      END IF

      IF (atprop_env%stress) THEN
         IF (ASSOCIATED(atprop_env%atstress)) THEN
            IF(.NOT.(SIZE(atprop_env%atstress, 3) == natom))CALL cp__a("subsys/atprop_types.F",90)
         ELSE
            ALLOCATE (atprop_env%atstress(3, 3, natom))
         END IF
         atprop_env%atstress = 0._dp
      END IF

   END SUBROUTINE atprop_init

! **************************************************************************************************
!> \brief ...
!> \param atarray ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE atprop_array_init(atarray, natom)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: atarray
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'atprop_array_init', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(atarray)) THEN
         IF(.NOT.(SIZE(atarray) == natom))CALL cp__a("subsys/atprop_types.F",112)
      ELSE
         ALLOCATE (atarray(natom))
      END IF
      atarray = 0._dp

   END SUBROUTINE atprop_array_init

! **************************************************************************************************
!> \brief ...
!> \param atarray ...
! **************************************************************************************************
   SUBROUTINE atprop_array_release(atarray)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: atarray

      CHARACTER(len=*), PARAMETER :: routineN = 'atprop_array_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(atarray)) THEN
         DEALLOCATE (atarray)
      END IF

   END SUBROUTINE atprop_array_release

! **************************************************************************************************
!> \brief ...
!> \param array_a ...
!> \param array_b ...
! **************************************************************************************************
   SUBROUTINE atprop_array_add(array_a, array_b)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: array_a, array_b

      CHARACTER(len=*), PARAMETER :: routineN = 'atprop_array_add', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(array_b)) THEN
         IF(.NOT.(ASSOCIATED(array_a)))CALL cp__a("subsys/atprop_types.F",148)
         array_a = array_a+array_b
      END IF

   END SUBROUTINE atprop_array_add

! **************************************************************************************************
!> \brief releases the atprop
!> \param atprop_env the object to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE atprop_release(atprop_env)
      TYPE(atprop_type), POINTER                         :: atprop_env

      CHARACTER(len=*), PARAMETER :: routineN = 'atprop_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(atprop_env)) THEN
         ! energy
         CALL atprop_array_release(atprop_env%atener)
         CALL atprop_array_release(atprop_env%ateb)
         CALL atprop_array_release(atprop_env%ateself)
         CALL atprop_array_release(atprop_env%atexc)
         CALL atprop_array_release(atprop_env%atevdw)
         CALL atprop_array_release(atprop_env%atecc)
         CALL atprop_array_release(atprop_env%ate1c)
         CALL atprop_array_release(atprop_env%atecoul)
         ! stress
         IF (ASSOCIATED(atprop_env%atstress)) THEN
            DEALLOCATE (atprop_env%atstress)
         END IF
         ! atprop type
         DEALLOCATE (atprop_env)
      END IF
      NULLIFY (atprop_env)
   END SUBROUTINE atprop_release

END MODULE atprop_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a simple array based list of the given type
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE particle_list_types
   USE particle_types,                  ONLY: deallocate_particle_set,&
                                              particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'particle_list_types'
   INTEGER, PRIVATE, SAVE :: last_particle_list_id = 0

   !API
   PUBLIC :: particle_list_type, particle_list_p_type, &
             particle_list_create, particle_list_retain, &
             particle_list_release

!***

! **************************************************************************************************
!> \brief represent a list of objects
!> \param id_nr identification number of this list
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param n_el the number of elements in the list
!> \param owns_list if the els are owned by this structure, and
!>        should be deallocated by it
!> \param list the array of object, might be oversized,
!>        only the fist n_el have some meaning
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE particle_list_type
      INTEGER :: id_nr, ref_count, n_els
      LOGICAL :: owns_els
      TYPE(particle_type), DIMENSION(:), POINTER :: els
   END TYPE particle_list_type

! **************************************************************************************************
!> \brief represents a pointer to a list
!> \param list the pointer to the list
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE particle_list_p_type
      TYPE(particle_list_type), POINTER :: list
   END TYPE particle_list_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a list
!> \param list the list to allocate and initialize
!> \param els_ptr the elements to store in the list (the array is only,
!>        referenced, not copied!)
!> \param owns_els if the list takes the ownership of els_ptr and
!>        will deallocate it (defaults to true)
!> \param n_els number of elements in the list (at least one els_ptr or
!>        n_els should be given)
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE particle_list_create(list, els_ptr, &
                                   owns_els, n_els)
      TYPE(particle_list_type), OPTIONAL, POINTER        :: list
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: els_ptr
      LOGICAL, INTENT(in), OPTIONAL                      :: owns_els
      INTEGER, INTENT(in), OPTIONAL                      :: n_els

      CHARACTER(len=*), PARAMETER :: routineN = 'particle_list_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(PRESENT(els_ptr) .OR. PRESENT(n_els)))CALL cp__a("subsys/particle_list_types.F",110)

      ALLOCATE (list)
      last_particle_list_id = last_particle_list_id+1
      list%id_nr = last_particle_list_id
      list%ref_count = 1
      list%owns_els = .TRUE.
      list%n_els = 0
      IF (PRESENT(owns_els)) list%owns_els = owns_els
      NULLIFY (list%els)
      IF (PRESENT(els_ptr)) THEN
         list%els => els_ptr
         IF (ASSOCIATED(els_ptr)) THEN
            list%n_els = SIZE(els_ptr)
         END IF
      END IF
      IF (PRESENT(n_els)) list%n_els = n_els
      IF (.NOT. ASSOCIATED(list%els)) THEN
         ALLOCATE (list%els(list%n_els))
         IF(.NOT.(list%owns_els))CALL cp__a("subsys/particle_list_types.F",129)
      END IF
   END SUBROUTINE particle_list_create

! **************************************************************************************************
!> \brief retains a list (see doc/ReferenceCounting.html)
!> \param list the list to retain
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE particle_list_retain(list)
      TYPE(particle_list_type), POINTER                  :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'particle_list_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(list)))CALL cp__a("subsys/particle_list_types.F",146)
      IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/particle_list_types.F",147)
      list%ref_count = list%ref_count+1
   END SUBROUTINE particle_list_retain

! **************************************************************************************************
!> \brief releases a  list (see doc/ReferenceCounting.html)
!> \param list the list to release
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE particle_list_release(list)
      TYPE(particle_list_type), POINTER                  :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'particle_list_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(list)) THEN
         IF(.NOT.(list%ref_count > 0))CALL cp__a("subsys/particle_list_types.F",165)
         list%ref_count = list%ref_count-1
         IF (list%ref_count == 0) THEN
            IF (list%owns_els) THEN
               IF (ASSOCIATED(list%els)) THEN
                  CALL deallocate_particle_set(list%els)
               END IF
            END IF
            NULLIFY (list%els)
            DEALLOCATE (list)
         END IF
      END IF
      NULLIFY (list)
   END SUBROUTINE particle_list_release

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(el_typename,el_type,USE,deallocate_els_code)]
! ARGS:
!  USE = "use particle_types, only: particle_type, deallocate_particle_set"
!  deallocate_els_code = "call deallocate_particle_set(list%els)"
!  el_type = "type(particle_type)"
!  el_typename = "particle"

END MODULE particle_list_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief types that represent a subsys, i.e. a part of the system
!> \par History
!>      07.2003 created [fawzi]
!>      09.2007 cleaned [tlaino] - University of Zurich
!>      22.11.2010 pack/unpack particle routines added (MK)
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_subsys_types
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_release,&
                                              atomic_kind_list_retain,&
                                              atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE atprop_types,                    ONLY: atprop_release,&
                                              atprop_type
   USE cell_types,                      ONLY: cell_release,&
                                              cell_retain,&
                                              cell_type,&
                                              real_to_scaled,&
                                              scaled_to_real
   USE colvar_types,                    ONLY: colvar_p_release,&
                                              colvar_p_type
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_release,&
                                              cp_result_retain,&
                                              cp_result_type
   USE distribution_1d_types,           ONLY: distribution_1d_release,&
                                              distribution_1d_retain,&
                                              distribution_1d_type
   USE kinds,                           ONLY: dp
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_release,&
                                              mol_kind_new_list_retain,&
                                              mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_release,&
                                              mol_new_list_retain,&
                                              mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: deallocate_global_constraint,&
                                              global_constraint_type,&
                                              molecule_type
   USE multipole_types,                 ONLY: multipole_type,&
                                              release_multipole_type,&
                                              retain_multipole_type
   USE particle_list_types,             ONLY: particle_list_release,&
                                              particle_list_retain,&
                                              particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE virial_types,                    ONLY: virial_release,&
                                              virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_subsys_types'
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   PUBLIC :: cp_subsys_type, &
             cp_subsys_p_type

   PUBLIC :: cp_subsys_retain, &
             cp_subsys_release, &
             cp_subsys_get, &
             cp_subsys_set, &
             pack_subsys_particles, &
             unpack_subsys_particles

! **************************************************************************************************
!> \brief represents a system: atoms, molecules, their pos,vel,...
!> \param atomic_kinds list with all the kinds in the actual subsys
!> \param particles list with the particles of the actual subsys
!> \param local_particles the particles that are local to the actual processor
!> \param molecule_kinds list with the molecule kinds
!> \param local_molecules the molecule structures of the actual subsys
!>        that are local to this processor
!> \param para_env the parallel environment of the actual subsys
!> \param shell_particles list with the shells of the actual subsys if shell-model is used
!> \param core_particles list with the shells of the actual subsys if shell-model is used
!> \par History
!>      07.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_subsys_type
      INTEGER :: ref_count = 1
      TYPE(atomic_kind_list_type), POINTER       :: atomic_kinds => Null()
      TYPE(particle_list_type), POINTER          :: particles => Null()
      TYPE(particle_list_type), POINTER          :: shell_particles => Null()
      TYPE(particle_list_type), POINTER          :: core_particles => Null()
      TYPE(distribution_1d_type), POINTER        :: local_particles => Null()
      TYPE(cp_para_env_type), POINTER            :: para_env => Null()
      ! New molecules kinds
      TYPE(mol_new_list_type), POINTER           :: molecules_new => Null()
      TYPE(mol_kind_new_list_type), POINTER      :: molecule_kinds_new => Null()
      TYPE(distribution_1d_type), POINTER        :: local_molecules_new => Null()
      ! Definitions of the collective variables
      TYPE(colvar_p_type), DIMENSION(:), POINTER :: colvar_p => Null()
      ! Intermolecular constraints
      TYPE(global_constraint_type), POINTER      :: gci => Null()
      ! Multipoles
      TYPE(multipole_type), POINTER              :: multipoles => Null()
      TYPE(atprop_type), POINTER                 :: atprop => Null()
      TYPE(virial_type), POINTER                 :: virial => Null()
      TYPE(cp_result_type), POINTER              :: results => Null()
      TYPE(cell_type), POINTER                    :: cell => Null()
   END TYPE cp_subsys_type

! **************************************************************************************************
!> \brief represent a pointer to a subsys, to be able to create arrays
!>      of pointers
!> \param subsys the pointer to the subsys
!> \par History
!>      07.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_subsys_p_type
      TYPE(cp_subsys_type), POINTER :: subsys
   END TYPE cp_subsys_p_type

CONTAINS

! **************************************************************************************************
!> \brief retains a subsys (see doc/ReferenceCounting.html)
!> \param subsys the subsys to retain
!> \par History
!>      07.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_subsys_retain(subsys)
      TYPE(cp_subsys_type), POINTER                      :: subsys

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_subsys_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(subsys)))CALL cp__a("subsys/cp_subsys_types.F",141)
      IF(.NOT.(subsys%ref_count > 0))CALL cp__a("subsys/cp_subsys_types.F",142)
      subsys%ref_count = subsys%ref_count+1
   END SUBROUTINE cp_subsys_retain

! **************************************************************************************************
!> \brief releases a subsys (see doc/ReferenceCounting.html)
!> \param subsys the subsys to release
!> \par History
!>      07.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_subsys_release(subsys)
      TYPE(cp_subsys_type), POINTER                      :: subsys

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_subsys_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(subsys)) THEN
         IF(.NOT.(subsys%ref_count > 0))CALL cp__a("subsys/cp_subsys_types.F",160)
         subsys%ref_count = subsys%ref_count-1
         IF (subsys%ref_count == 0) THEN
            CALL atomic_kind_list_release(subsys%atomic_kinds)
            CALL particle_list_release(subsys%particles)
            CALL particle_list_release(subsys%shell_particles)
            CALL particle_list_release(subsys%core_particles)
            CALL distribution_1d_release(subsys%local_particles)
            CALL mol_kind_new_list_release(subsys%molecule_kinds_new)
            CALL mol_new_list_release(subsys%molecules_new)
            CALL distribution_1d_release(subsys%local_molecules_new)
            CALL cp_para_env_release(subsys%para_env)
            CALL release_multipole_type(subsys%multipoles)
            CALL colvar_p_release(subsys%colvar_p)
            CALL deallocate_global_constraint(subsys%gci)
            CALL atprop_release(subsys%atprop)
            CALL virial_release(subsys%virial)
            CALL cp_result_release(subsys%results)
            CALL cell_release(subsys%cell)
            DEALLOCATE (subsys)
         END IF
         NULLIFY (subsys)
      END IF
   END SUBROUTINE cp_subsys_release

! **************************************************************************************************
!> \brief sets various propreties of the subsys
!> \param subsys the subsys you want to modify
!> \param atomic_kinds ...
!> \param particles ...
!> \param local_particles ...
!> \param molecules_new ...
!> \param molecule_kinds_new ...
!> \param local_molecules_new ...
!> \param para_env ...
!> \param colvar_p ...
!> \param shell_particles ...
!> \param core_particles ...
!> \param gci ...
!> \param multipoles ...
!> \param results ...
!> \param cell ...
!> \par History
!>      08.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_subsys_set(subsys, atomic_kinds, particles, local_particles, &
                            molecules_new, molecule_kinds_new, local_molecules_new, para_env, &
                            colvar_p, shell_particles, core_particles, gci, multipoles, results, cell)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(atomic_kind_list_type), OPTIONAL, POINTER     :: atomic_kinds
      TYPE(particle_list_type), OPTIONAL, POINTER        :: particles
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles
      TYPE(mol_new_list_type), OPTIONAL, POINTER         :: molecules_new
      TYPE(mol_kind_new_list_type), OPTIONAL, POINTER    :: molecule_kinds_new
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_molecules_new
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(colvar_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: colvar_p
      TYPE(particle_list_type), OPTIONAL, POINTER        :: shell_particles, core_particles
      TYPE(global_constraint_type), OPTIONAL, POINTER    :: gci
      TYPE(multipole_type), OPTIONAL, POINTER            :: multipoles
      TYPE(cp_result_type), OPTIONAL, POINTER            :: results
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_subsys_set', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(subsys)))CALL cp__a("subsys/cp_subsys_types.F",227)
      IF(.NOT.(subsys%ref_count > 0))CALL cp__a("subsys/cp_subsys_types.F",228)
      IF (PRESENT(multipoles)) THEN
         CALL retain_multipole_type(multipoles)
         CALL release_multipole_type(subsys%multipoles)
         subsys%multipoles => multipoles
      END IF
      IF (PRESENT(atomic_kinds)) THEN
         CALL atomic_kind_list_retain(atomic_kinds)
         CALL atomic_kind_list_release(subsys%atomic_kinds)
         subsys%atomic_kinds => atomic_kinds
      END IF
      IF (PRESENT(particles)) THEN
         CALL particle_list_retain(particles)
         CALL particle_list_release(subsys%particles)
         subsys%particles => particles
      END IF
      IF (PRESENT(local_particles)) THEN
         CALL distribution_1d_retain(local_particles)
         CALL distribution_1d_release(subsys%local_particles)
         subsys%local_particles => local_particles
      END IF
      IF (PRESENT(local_molecules_new)) THEN
         CALL distribution_1d_retain(local_molecules_new)
         CALL distribution_1d_release(subsys%local_molecules_new)
         subsys%local_molecules_new => local_molecules_new
      END IF
      IF (PRESENT(molecule_kinds_new)) THEN
         CALL mol_kind_new_list_retain(molecule_kinds_new)
         CALL mol_kind_new_list_release(subsys%molecule_kinds_new)
         subsys%molecule_kinds_new => molecule_kinds_new
      END IF
      IF (PRESENT(molecules_new)) THEN
         CALL mol_new_list_retain(molecules_new)
         CALL mol_new_list_release(subsys%molecules_new)
         subsys%molecules_new => molecules_new
      END IF
      IF (PRESENT(para_env)) THEN
         CALL cp_para_env_retain(para_env)
         CALL cp_para_env_release(subsys%para_env)
         subsys%para_env => para_env
      END IF
      IF (PRESENT(colvar_p)) THEN
         IF(.NOT.(.NOT. ASSOCIATED(subsys%colvar_p)))CALL cp__a("subsys/cp_subsys_types.F",270)
         subsys%colvar_p => colvar_p
      ENDIF
      IF (PRESENT(shell_particles)) THEN
         IF (ASSOCIATED(shell_particles)) THEN
            CALL particle_list_retain(shell_particles)
            CALL particle_list_release(subsys%shell_particles)
            subsys%shell_particles => shell_particles
         END IF
      END IF
      IF (PRESENT(core_particles)) THEN
         IF (ASSOCIATED(core_particles)) THEN
            CALL particle_list_retain(core_particles)
            CALL particle_list_release(subsys%core_particles)
            subsys%core_particles => core_particles
         END IF
      END IF
      IF (PRESENT(gci)) THEN
         IF(.NOT.(.NOT. ASSOCIATED(subsys%gci)))CALL cp__a("subsys/cp_subsys_types.F",288)
         subsys%gci => gci
      ENDIF
      IF (PRESENT(results)) THEN
         IF (ASSOCIATED(results)) THEN
            CALL cp_result_retain(results)
            CALL cp_result_release(subsys%results)
            subsys%results => results
         END IF
      END IF
      IF (PRESENT(cell)) THEN
         IF (ASSOCIATED(cell)) THEN
            CALL cell_retain(cell)
            CALL cell_release(subsys%cell)
            subsys%cell => cell
         END IF
      END IF
   END SUBROUTINE cp_subsys_set

! **************************************************************************************************
!> \brief returns information about various attributes of the given subsys
!> \param subsys the subsys you want info about
!> \param ref_count ...
!> \param atomic_kinds ...
!> \param atomic_kind_set ...
!> \param particles ...
!> \param particle_set ...
!> \param local_particles ...
!> \param molecules_new ...
!> \param molecule_set ...
!> \param molecule_kinds_new ...
!> \param molecule_kind_set ...
!> \param local_molecules_new ...
!> \param para_env ...
!> \param colvar_p ...
!> \param shell_particles ...
!> \param core_particles ...
!> \param gci ...
!> \param multipoles ...
!> \param natom ...
!> \param nparticle ...
!> \param ncore ...
!> \param nshell ...
!> \param nkind ...
!> \param atprop ...
!> \param virial ...
!> \param results ...
!> \param cell ...
!> \par History
!>      08.2003 created [fawzi]
!>      22.11.2010 (MK)
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_subsys_get(subsys, ref_count, atomic_kinds, atomic_kind_set, &
                            particles, particle_set, &
                            local_particles, molecules_new, molecule_set, molecule_kinds_new, &
                            molecule_kind_set, local_molecules_new, para_env, colvar_p, &
                            shell_particles, core_particles, gci, multipoles, &
                            natom, nparticle, ncore, nshell, nkind, atprop, virial, &
                            results, cell)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      INTEGER, INTENT(out), OPTIONAL                     :: ref_count
      TYPE(atomic_kind_list_type), OPTIONAL, POINTER     :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(particle_list_type), OPTIONAL, POINTER        :: particles
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles
      TYPE(mol_new_list_type), OPTIONAL, POINTER         :: molecules_new
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_set
      TYPE(mol_kind_new_list_type), OPTIONAL, POINTER    :: molecule_kinds_new
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_kind_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_molecules_new
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(colvar_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: colvar_p
      TYPE(particle_list_type), OPTIONAL, POINTER        :: shell_particles, core_particles
      TYPE(global_constraint_type), OPTIONAL, POINTER    :: gci
      TYPE(multipole_type), OPTIONAL, POINTER            :: multipoles
      INTEGER, INTENT(out), OPTIONAL                     :: natom, nparticle, ncore, nshell, nkind
      TYPE(atprop_type), OPTIONAL, POINTER               :: atprop
      TYPE(virial_type), OPTIONAL, POINTER               :: virial
      TYPE(cp_result_type), OPTIONAL, POINTER            :: results
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_subsys_get', routineP = moduleN//':'//routineN

      INTEGER                                            :: n_atom, n_core, n_shell

      n_atom = 0
      n_core = 0
      n_shell = 0

      IF(.NOT.(ASSOCIATED(subsys)))CALL cp__a("subsys/cp_subsys_types.F",384)
      IF(.NOT.(subsys%ref_count > 0))CALL cp__a("subsys/cp_subsys_types.F",385)

      IF (PRESENT(ref_count)) ref_count = subsys%ref_count
      IF (PRESENT(atomic_kinds)) atomic_kinds => subsys%atomic_kinds
      IF (PRESENT(atomic_kind_set)) atomic_kind_set => subsys%atomic_kinds%els
      IF (PRESENT(particles)) particles => subsys%particles
      IF (PRESENT(particle_set)) particle_set => subsys%particles%els
      IF (PRESENT(local_particles)) local_particles => subsys%local_particles
      IF (PRESENT(molecules_new)) molecules_new => subsys%molecules_new
      IF (PRESENT(molecule_set)) molecule_set => subsys%molecules_new%els
      IF (PRESENT(molecule_kinds_new)) molecule_kinds_new => subsys%molecule_kinds_new
      IF (PRESENT(molecule_kind_set)) molecule_kind_set => subsys%molecule_kinds_new%els
      IF (PRESENT(local_molecules_new)) local_molecules_new => subsys%local_molecules_new
      IF (PRESENT(para_env)) para_env => subsys%para_env
      IF (PRESENT(colvar_p)) colvar_p => subsys%colvar_p
      IF (PRESENT(shell_particles)) shell_particles => subsys%shell_particles
      IF (PRESENT(core_particles)) core_particles => subsys%core_particles
      IF (PRESENT(gci)) gci => subsys%gci
      IF (PRESENT(multipoles)) multipoles => subsys%multipoles
      IF (PRESENT(virial)) virial => subsys%virial
      IF (PRESENT(atprop)) atprop => subsys%atprop
      IF (PRESENT(results)) results => subsys%results
      IF (PRESENT(cell)) cell => subsys%cell
      IF (PRESENT(nkind)) nkind = SIZE(subsys%atomic_kinds%els)

      IF (PRESENT(natom) .OR. PRESENT(nparticle) .OR. PRESENT(nshell)) THEN
         ! An atomic particle set should be present in each subsystem at the moment
         IF(.NOT.(ASSOCIATED(subsys%particles)))CALL cp__a("subsys/cp_subsys_types.F",412)
         n_atom = subsys%particles%n_els
         ! Check if we have other kinds of particles in this subsystem
         IF (ASSOCIATED(subsys%shell_particles)) THEN
            n_shell = subsys%shell_particles%n_els
            IF(.NOT.(ASSOCIATED(subsys%core_particles)))CALL cp__a("subsys/cp_subsys_types.F",417)
            n_core = subsys%core_particles%n_els
            ! The same number of shell and core particles is assumed
            IF(.NOT.(n_core == n_shell))CALL cp__a("subsys/cp_subsys_types.F",420)
         ELSE IF (ASSOCIATED(subsys%core_particles)) THEN
            ! This case should not occur at the moment
            IF(.NOT.(ASSOCIATED(subsys%shell_particles)))CALL cp__a("subsys/cp_subsys_types.F",423)
         ELSE
            n_core = 0
            n_shell = 0
         END IF
         IF (PRESENT(natom)) natom = n_atom
         IF (PRESENT(nparticle)) nparticle = n_atom+n_shell
         IF (PRESENT(ncore)) ncore = n_core
         IF (PRESENT(nshell)) nshell = n_shell
      END IF

   END SUBROUTINE cp_subsys_get

! **************************************************************************************************
!> \brief   Pack components of a subsystem particle sets into a single vector
!> \param subsys ...
!> \param f ...
!> \param r ...
!> \param s ...
!> \param v ...
!> \param fscale ...
!> \param cell ...
!> \date    19.11.10
!> \author  Matthias Krack (MK)
!> \version 1.0
!> \note    It is assumed that f, r, s, or v are properly allocated already
! **************************************************************************************************
   SUBROUTINE pack_subsys_particles(subsys, f, r, s, v, fscale, cell)

      TYPE(cp_subsys_type), POINTER                      :: subsys
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT), OPTIONAL :: f, r, s, v
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: fscale
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pack_subsys_particles', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, j, k, natom, nparticle, nsize, &
                                                            shell_index
      REAL(KIND=dp), DIMENSION(3)                        :: rs
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles

      IF(.NOT.(ASSOCIATED(subsys)))CALL cp__a("subsys/cp_subsys_types.F",466)

      IF (PRESENT(s)) THEN
         IF(.NOT.(PRESENT(cell)))CALL cp__a("subsys/cp_subsys_types.F",469)
      END IF

      NULLIFY (core_particles)
      NULLIFY (particles)
      NULLIFY (shell_particles)

      CALL cp_subsys_get(subsys, &
                         core_particles=core_particles, &
                         natom=natom, &
                         nparticle=nparticle, &
                         particles=particles, &
                         shell_particles=shell_particles)

      nsize = 3*nparticle

      ! Pack forces

      IF (PRESENT(f)) THEN
         IF(.NOT.((SIZE(f) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",488)
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               DO i = 1, 3
                  j = j+1
                  f(j) = particles%els(iatom)%f(i)
               END DO
            ELSE
               DO i = 1, 3
                  j = j+1
                  f(j) = core_particles%els(shell_index)%f(i)+particles%els(iatom)%f(i)
               END DO
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  f(k+i) = shell_particles%els(shell_index)%f(i)
               END DO
            END IF
         END DO
         IF (PRESENT(fscale)) f(1:nsize) = fscale*f(1:nsize)
      END IF

      ! Pack coordinates

      IF (PRESENT(r)) THEN
         IF(.NOT.((SIZE(r) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",514)
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               DO i = 1, 3
                  j = j+1
                  r(j) = particles%els(iatom)%r(i)
               END DO
            ELSE
               DO i = 1, 3
                  j = j+1
                  r(j) = core_particles%els(shell_index)%r(i)
               END DO
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  r(k+i) = shell_particles%els(shell_index)%r(i)
               END DO
            END IF
         END DO
      END IF

      ! Pack as scaled coordinates

      IF (PRESENT(s)) THEN
         IF(.NOT.((SIZE(s) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",539)
         IF(.NOT.(PRESENT(cell)))CALL cp__a("subsys/cp_subsys_types.F",540)
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               CALL real_to_scaled(rs, particles%els(iatom)%r, cell)
               DO i = 1, 3
                  j = j+1
                  s(j) = rs(i)
               END DO
            ELSE
               CALL real_to_scaled(rs, core_particles%els(shell_index)%r, cell)
               DO i = 1, 3
                  j = j+1
                  s(j) = rs(i)
               END DO
               CALL real_to_scaled(rs, shell_particles%els(shell_index)%r, cell)
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  s(k+i) = rs(i)
               END DO
            END IF
         END DO
      END IF

      ! Pack velocities

      IF (PRESENT(v)) THEN
         IF(.NOT.((SIZE(v) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",568)
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               DO i = 1, 3
                  j = j+1
                  v(j) = particles%els(iatom)%v(i)
               END DO
            ELSE
               DO i = 1, 3
                  j = j+1
                  v(j) = core_particles%els(shell_index)%v(i)
               END DO
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  v(k+i) = shell_particles%els(shell_index)%v(i)
               END DO
            END IF
         END DO
      END IF

   END SUBROUTINE pack_subsys_particles

! **************************************************************************************************
!> \brief   Unpack components of a subsystem particle sets into a single vector
!> \param subsys ...
!> \param f ...
!> \param r ...
!> \param s ...
!> \param v ...
!> \param fscale ...
!> \param cell ...
!> \date    19.11.10
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE unpack_subsys_particles(subsys, f, r, s, v, fscale, cell)

      TYPE(cp_subsys_type), POINTER                      :: subsys
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: f, r, s, v
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: fscale
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'unpack_subsys_particles', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, j, k, natom, nparticle, nsize, &
                                                            shell_index
      REAL(KIND=dp)                                      :: fc, fs, mass, my_fscale
      REAL(KIND=dp), DIMENSION(3)                        :: rs
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles

      IF(.NOT.(ASSOCIATED(subsys)))CALL cp__a("subsys/cp_subsys_types.F",622)

      NULLIFY (core_particles)
      NULLIFY (particles)
      NULLIFY (shell_particles)

      CALL cp_subsys_get(subsys, &
                         core_particles=core_particles, &
                         natom=natom, &
                         nparticle=nparticle, &
                         particles=particles, &
                         shell_particles=shell_particles)

      nsize = 3*nparticle

      ! Unpack forces

      IF (PRESENT(f)) THEN
         IF(.NOT.((SIZE(f) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",640)
         IF (PRESENT(fscale)) THEN
            my_fscale = fscale
         ELSE
            my_fscale = 1.0_dp
         END IF
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               DO i = 1, 3
                  j = j+1
                  particles%els(iatom)%f(i) = my_fscale*f(j)
               END DO
            ELSE
               DO i = 1, 3
                  j = j+1
                  core_particles%els(shell_index)%f(i) = my_fscale*f(j)
               END DO
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  shell_particles%els(shell_index)%f(i) = my_fscale*f(k+i)
               END DO
            END IF
         END DO
      END IF

      ! Unpack coordinates

      IF (PRESENT(r)) THEN
         IF(.NOT.((SIZE(r) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",670)
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               DO i = 1, 3
                  j = j+1
                  particles%els(iatom)%r(i) = r(j)
               END DO
            ELSE
               DO i = 1, 3
                  j = j+1
                  core_particles%els(shell_index)%r(i) = r(j)
               END DO
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  shell_particles%els(shell_index)%r(i) = r(k+i)
               END DO
               ! Update atomic position due to core and shell motion
               mass = particles%els(iatom)%atomic_kind%mass
               fc = core_particles%els(shell_index)%atomic_kind%shell%mass_core/mass
               fs = shell_particles%els(shell_index)%atomic_kind%shell%mass_shell/mass
               particles%els(iatom)%r(1:3) = fc*core_particles%els(shell_index)%r(1:3)+ &
                                             fs*shell_particles%els(shell_index)%r(1:3)
            END IF
         END DO
      END IF

      ! Unpack scaled coordinates

      IF (PRESENT(s)) THEN
         IF(.NOT.((SIZE(s) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",701)
         IF(.NOT.(PRESENT(cell)))CALL cp__a("subsys/cp_subsys_types.F",702)
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               DO i = 1, 3
                  j = j+1
                  rs(i) = s(j)
               END DO
               CALL scaled_to_real(particles%els(iatom)%r, rs, cell)
            ELSE
               DO i = 1, 3
                  j = j+1
                  rs(i) = s(j)
               END DO
               CALL scaled_to_real(core_particles%els(shell_index)%r, rs, cell)
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  rs(i) = s(k+i)
               END DO
               CALL scaled_to_real(shell_particles%els(shell_index)%r, rs, cell)
               ! Update atomic position due to core and shell motion
               mass = particles%els(iatom)%atomic_kind%mass
               fc = core_particles%els(shell_index)%atomic_kind%shell%mass_core/mass
               fs = shell_particles%els(shell_index)%atomic_kind%shell%mass_shell/mass
               particles%els(iatom)%r(1:3) = fc*core_particles%els(shell_index)%r(1:3)+ &
                                             fs*shell_particles%els(shell_index)%r(1:3)
            END IF
         END DO
      END IF

      ! Unpack velocities

      IF (PRESENT(v)) THEN
         IF(.NOT.((SIZE(v) >= nsize)))CALL cp__a("subsys/cp_subsys_types.F",736)
         j = 0
         DO iatom = 1, natom
            shell_index = particles%els(iatom)%shell_index
            IF (shell_index == 0) THEN
               DO i = 1, 3
                  j = j+1
                  particles%els(iatom)%v(i) = v(j)
               END DO
            ELSE
               DO i = 1, 3
                  j = j+1
                  core_particles%els(shell_index)%v(i) = v(j)
               END DO
               k = 3*(natom+shell_index-1)
               DO i = 1, 3
                  shell_particles%els(shell_index)%v(i) = v(k+i)
               END DO
               ! Update atomic velocity due to core and shell motion
               mass = particles%els(iatom)%atomic_kind%mass
               fc = core_particles%els(shell_index)%atomic_kind%shell%mass_core/mass
               fs = shell_particles%els(shell_index)%atomic_kind%shell%mass_shell/mass
               particles%els(iatom)%v(1:3) = fc*core_particles%els(shell_index)%v(1:3)+ &
                                             fs*shell_particles%els(shell_index)%v(1:3)
            END IF
         END DO
      END IF

   END SUBROUTINE unpack_subsys_particles

END MODULE cp_subsys_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief types that represent a quickstep subsys
!> \author Ole Schuett
! **************************************************************************************************
MODULE qs_subsys_types
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_release,&
                                              cell_retain,&
                                              cell_type
   USE colvar_types,                    ONLY: colvar_p_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_release,&
                                              cp_subsys_retain,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE multipole_types,                 ONLY: multipole_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_types,                 ONLY: deallocate_qs_energy,&
                                              qs_energy_type
   USE qs_force_types,                  ONLY: deallocate_qs_force,&
                                              qs_force_type
   USE qs_kind_types,                   ONLY: deallocate_qs_kind_set,&
                                              qs_kind_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_subsys_types'

   PUBLIC :: qs_subsys_type

   PUBLIC :: qs_subsys_retain, &
             qs_subsys_release, &
             qs_subsys_get, &
             qs_subsys_set

   TYPE qs_subsys_type
      PRIVATE
      INTEGER                                               :: ref_count = 1
      INTEGER                                               :: nelectron_total = -1
      INTEGER                                               :: nelectron_spin(2) = -1
      TYPE(cp_subsys_type), POINTER                         :: cp_subsys => Null()
      TYPE(qs_kind_type), DIMENSION(:), POINTER             :: qs_kind_set => Null()
      TYPE(cell_type), POINTER                              :: cell_ref => Null()
      LOGICAL                                               :: use_ref_cell = .FALSE.
      TYPE(qs_energy_type), POINTER                         :: energy => Null()
      TYPE(qs_force_type), DIMENSION(:), POINTER            :: force => Null()
   END TYPE qs_subsys_type

CONTAINS

! **************************************************************************************************
!> \brief retains a subsys (see doc/ReferenceCounting.html)
!> \param subsys the subsys to retain
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qs_subsys_retain(subsys)
      TYPE(qs_subsys_type), POINTER                      :: subsys

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_subsys_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(subsys)))CALL cp__a("qs_subsys_types.F",81)
      IF(.NOT.(subsys%ref_count > 0))CALL cp__a("qs_subsys_types.F",82)
      subsys%ref_count = subsys%ref_count+1
   END SUBROUTINE qs_subsys_retain

! **************************************************************************************************
!> \brief releases a subsys (see doc/ReferenceCounting.html)
!> \param subsys the subsys to release
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qs_subsys_release(subsys)
      TYPE(qs_subsys_type), POINTER                      :: subsys

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_subsys_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(subsys)) THEN
         IF(.NOT.(subsys%ref_count > 0))CALL cp__a("qs_subsys_types.F",98)
         subsys%ref_count = subsys%ref_count-1
         IF (subsys%ref_count == 0) THEN
            CALL cp_subsys_release(subsys%cp_subsys)
            CALL cell_release(subsys%cell_ref)
            IF (ASSOCIATED(subsys%qs_kind_set)) &
               CALL deallocate_qs_kind_set(subsys%qs_kind_set)
            IF (ASSOCIATED(subsys%energy)) &
               CALL deallocate_qs_energy(subsys%energy)
            IF (ASSOCIATED(subsys%force)) &
               CALL deallocate_qs_force(subsys%force)
            DEALLOCATE (subsys)
         END IF
      END IF
   END SUBROUTINE qs_subsys_release

! **************************************************************************************************
!> \brief ...
!> \param subsys ...
!> \param atomic_kinds ...
!> \param atomic_kind_set ...
!> \param particles ...
!> \param particle_set ...
!> \param local_particles ...
!> \param molecules_new ...
!> \param molecule_set ...
!> \param molecule_kinds_new ...
!> \param molecule_kind_set ...
!> \param local_molecules_new ...
!> \param para_env ...
!> \param colvar_p ...
!> \param shell_particles ...
!> \param core_particles ...
!> \param gci ...
!> \param multipoles ...
!> \param natom ...
!> \param nparticle ...
!> \param ncore ...
!> \param nshell ...
!> \param nkind ...
!> \param atprop ...
!> \param virial ...
!> \param results ...
!> \param cell ...
!> \param cell_ref ...
!> \param use_ref_cell ...
!> \param energy ...
!> \param force ...
!> \param qs_kind_set ...
!> \param cp_subsys ...
!> \param nelectron_total ...
!> \param nelectron_spin ...
! **************************************************************************************************
   SUBROUTINE qs_subsys_get(subsys, atomic_kinds, atomic_kind_set, particles, particle_set, &
                            local_particles, molecules_new, molecule_set, &
                            molecule_kinds_new, molecule_kind_set, &
                            local_molecules_new, para_env, colvar_p, &
                            shell_particles, core_particles, gci, multipoles, &
                            natom, nparticle, ncore, nshell, nkind, atprop, virial, &
                            results, cell, cell_ref, use_ref_cell, energy, force, &
                            qs_kind_set, cp_subsys, nelectron_total, nelectron_spin)
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(atomic_kind_list_type), OPTIONAL, POINTER     :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(particle_list_type), OPTIONAL, POINTER        :: particles
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles
      TYPE(mol_new_list_type), OPTIONAL, POINTER         :: molecules_new
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_set
      TYPE(mol_kind_new_list_type), OPTIONAL, POINTER    :: molecule_kinds_new
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_kind_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_molecules_new
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(colvar_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: colvar_p
      TYPE(particle_list_type), OPTIONAL, POINTER        :: shell_particles, core_particles
      TYPE(global_constraint_type), OPTIONAL, POINTER    :: gci
      TYPE(multipole_type), OPTIONAL, POINTER            :: multipoles
      INTEGER, INTENT(out), OPTIONAL                     :: natom, nparticle, ncore, nshell, nkind
      TYPE(atprop_type), OPTIONAL, POINTER               :: atprop
      TYPE(virial_type), OPTIONAL, POINTER               :: virial
      TYPE(cp_result_type), OPTIONAL, POINTER            :: results
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref
      LOGICAL, OPTIONAL                                  :: use_ref_cell
      TYPE(qs_energy_type), OPTIONAL, POINTER            :: energy
      TYPE(qs_force_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: force
      TYPE(qs_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: qs_kind_set
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: cp_subsys
      INTEGER, OPTIONAL                                  :: nelectron_total
      INTEGER, DIMENSION(2), OPTIONAL                    :: nelectron_spin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_subsys_get', routineP = moduleN//':'//routineN

      CALL cp_subsys_get(subsys%cp_subsys, &
                         atomic_kinds=atomic_kinds, &
                         atomic_kind_set=atomic_kind_set, &
                         particles=particles, &
                         particle_set=particle_set, &
                         local_particles=local_particles, &
                         molecules_new=molecules_new, &
                         molecule_set=molecule_set, &
                         molecule_kinds_new=molecule_kinds_new, &
                         molecule_kind_set=molecule_kind_set, &
                         local_molecules_new=local_molecules_new, &
                         para_env=para_env, &
                         colvar_p=colvar_p, &
                         shell_particles=shell_particles, &
                         core_particles=core_particles, &
                         gci=gci, &
                         multipoles=multipoles, &
                         natom=natom, &
                         nkind=nkind, &
                         nparticle=nparticle, &
                         ncore=ncore, &
                         nshell=nshell, &
                         atprop=atprop, &
                         virial=virial, &
                         results=results, &
                         cell=cell)

      IF (PRESENT(cell_ref)) cell_ref => subsys%cell_ref
      IF (PRESENT(use_ref_cell)) use_ref_cell = subsys%use_ref_cell
      IF (PRESENT(energy)) energy => subsys%energy
      IF (PRESENT(force)) force => subsys%force
      IF (PRESENT(qs_kind_set)) qs_kind_set => subsys%qs_kind_set
      IF (PRESENT(cp_subsys)) cp_subsys => subsys%cp_subsys
      IF (PRESENT(nelectron_total)) nelectron_total = subsys%nelectron_total
      IF (PRESENT(nelectron_spin)) nelectron_spin = subsys%nelectron_spin
   END SUBROUTINE qs_subsys_get

! **************************************************************************************************
!> \brief ...
!> \param subsys ...
!> \param cp_subsys ...
!> \param local_particles ...
!> \param local_molecules_new ...
!> \param cell ...
!> \param cell_ref ...
!> \param use_ref_cell ...
!> \param energy ...
!> \param force ...
!> \param qs_kind_set ...
!> \param nelectron_total ...
!> \param nelectron_spin ...
! **************************************************************************************************
   SUBROUTINE qs_subsys_set(subsys, cp_subsys, &
                            local_particles, local_molecules_new, cell, &
                            cell_ref, use_ref_cell, energy, force, &
                            qs_kind_set, nelectron_total, nelectron_spin)
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: cp_subsys
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles, local_molecules_new
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref
      LOGICAL, OPTIONAL                                  :: use_ref_cell
      TYPE(qs_energy_type), OPTIONAL, POINTER            :: energy
      TYPE(qs_force_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: force
      TYPE(qs_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: qs_kind_set
      INTEGER, OPTIONAL                                  :: nelectron_total
      INTEGER, DIMENSION(2), OPTIONAL                    :: nelectron_spin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_subsys_set', routineP = moduleN//':'//routineN

      IF (PRESENT(cp_subsys)) THEN
         CALL cp_subsys_retain(cp_subsys)
         CALL cp_subsys_release(subsys%cp_subsys)
         subsys%cp_subsys => cp_subsys
      END IF

      CALL cp_subsys_set(subsys%cp_subsys, &
                         local_particles=local_particles, &
                         local_molecules_new=local_molecules_new, &
                         cell=cell)

      IF (PRESENT(cell_ref)) THEN
         CALL cell_retain(cell_ref)
         CALL cell_release(subsys%cell_ref)
         subsys%cell_ref => cell_ref
      END IF

      IF (PRESENT(use_ref_cell)) subsys%use_ref_cell = use_ref_cell
      IF (PRESENT(energy)) subsys%energy => energy
      ! if intels checking (-C) complains here, you have rediscovered a bug in the intel
      ! compiler (present in at least 10.0.025). A testcase has been submitted to intel.
      IF (PRESENT(force)) subsys%force => force
      IF (PRESENT(qs_kind_set)) subsys%qs_kind_set => qs_kind_set
      IF (PRESENT(nelectron_total)) subsys%nelectron_total = nelectron_total
      IF (PRESENT(nelectron_spin)) subsys%nelectron_spin = nelectron_spin
   END SUBROUTINE qs_subsys_set

END MODULE qs_subsys_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Scratch space for multiple threads writing to rs grids (see
!>        qs_collocate_density.F for an example
!> \par History
!>      IAB 26-Apr-2010 : initial version - moved out of qs_collocate_density.F
!>                        (c) The Numerical Algorithms Group (NAG) Ltd, 2010 on behalf of the HECToR project
!> \author IAB
! **************************************************************************************************

MODULE lgrid_types

   USE kinds,                           ONLY: dp
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              rs_grid_max_ngpts
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: lgrid_type, lgrid_release, lgrid_create, lgrid_allocate_grid

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lgrid_types'

   TYPE lgrid_type
      INTEGER :: ldim, ref_count
      REAL(dp), DIMENSION(:), POINTER :: r
   END TYPE lgrid_type

   TYPE lgrid_p_type
      TYPE(lgrid_type), POINTER :: l
   END TYPE lgrid_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates an lgrid, ldim set based on the rs_grid_descriptors.
!>        The grid is not allocated
!> \param lgrid the lgrid that gets created
!> \param rs_descs the rs grid descriptors used to set the lgrid size
!> \par History
!>      10.2011 created [IAB]
!> \author Iain Bethune
! **************************************************************************************************
   SUBROUTINE lgrid_create(lgrid, rs_descs)
      TYPE(lgrid_type), POINTER                          :: lgrid
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs

      CHARACTER(len=*), PARAMETER :: routineN = 'lgrid_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ngpts

      IF(.NOT.(.NOT. ASSOCIATED(lgrid)))CALL cp__a("pw/lgrid_types.F",59)
      ALLOCATE (lgrid)
      NULLIFY (lgrid%r)
      lgrid%ref_count = 1
      ! Find the maximum number of grid points needed
      ngpts = 0
      DO i = 1, SIZE(rs_descs)
         ngpts = MAX(ngpts, rs_grid_max_ngpts(rs_descs(i)%rs_desc))
      END DO
      lgrid%ldim = ngpts
   END SUBROUTINE lgrid_create

! **************************************************************************************************
!> \brief retains the lgrid (see doc/ReferenceCounting.html)
!> \param lgrid the lgrid_type to retain
!> \par History
!>      10.2011 created [IAB]
!> \author Iain Bethune
! **************************************************************************************************
   SUBROUTINE lgrid_retain(lgrid)
      TYPE(lgrid_type), POINTER                          :: lgrid

      CHARACTER(len=*), PARAMETER :: routineN = 'lgrid_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(lgrid)))CALL cp__a("pw/lgrid_types.F",83)
      IF(.NOT.(lgrid%ref_count > 0))CALL cp__a("pw/lgrid_types.F",84)
      lgrid%ref_count = lgrid%ref_count+1
   END SUBROUTINE lgrid_retain

! **************************************************************************************************
!> \brief releases the given lgrid (see doc/ReferenceCounting.html)
!> \param lgrid the lgrid_type to release
!> \par History
!>      10.2011 created [IAB]
!> \author Iain Bethune
! **************************************************************************************************
   SUBROUTINE lgrid_release(lgrid)
      TYPE(lgrid_type), POINTER                          :: lgrid

      CHARACTER(len=*), PARAMETER :: routineN = 'lgrid_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(lgrid)) THEN
         IF(.NOT.(lgrid%ref_count > 0))CALL cp__a("pw/lgrid_types.F",101)
         lgrid%ref_count = lgrid%ref_count-1
         IF (lgrid%ref_count < 1) THEN
            IF (ASSOCIATED(lgrid%r)) THEN
               DEALLOCATE (lgrid%r)
            END IF
            DEALLOCATE (lgrid)
         END IF
      END IF
   END SUBROUTINE

! **************************************************************************************************
!> \brief allocates the lgrid for a given number of threads
!> \param lgrid the lgrid_type for which the grid will be allocated
!> \param nthreads how many threads to allocate for
!> \par History
!>      10.2011 created [IAB]
!> \author Iain Bethune
! **************************************************************************************************
   SUBROUTINE lgrid_allocate_grid(lgrid, nthreads)
      TYPE(lgrid_type), POINTER                          :: lgrid
      INTEGER, INTENT(in)                                :: nthreads

      CHARACTER(len=*), PARAMETER :: routineN = 'lgrid_allocate_grid', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(lgrid)))CALL cp__a("pw/lgrid_types.F",127)
      IF(.NOT.(.NOT. ASSOCIATED(lgrid%r)))CALL cp__a("pw/lgrid_types.F",128)
      ALLOCATE (lgrid%r(lgrid%ldim*nthreads))
   END SUBROUTINE

END MODULE

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Code to return a gridlevel associated with a given gaussian exponent
!> \author Joost VandeVondele (27.02.02)
! **************************************************************************************************
MODULE gaussian_gridlevels
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'gaussian_gridlevels'

! **************************************************************************************************
   TYPE gridlevel_info_type
      INTEGER                         :: ngrid_levels
      REAL(KIND=dp), POINTER, DIMENSION(:) :: cutoff
      INTEGER(KIND=int_8), POINTER, DIMENSION(:) :: count
      INTEGER(KIND=int_8)                         :: total_count
      REAL(KIND=dp)                  :: rel_cutoff
      TYPE(section_vals_type), POINTER :: print_section
   END TYPE gridlevel_info_type

   PUBLIC :: gridlevel_info_type
   PUBLIC :: gaussian_gridlevel
   PUBLIC :: init_gaussian_gridlevel
   PUBLIC :: destroy_gaussian_gridlevel

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param gridlevel_info ...
!> \param ngrid_levels ...
!> \param cutoff ...
!> \param rel_cutoff ...
!> \param print_section ...
! **************************************************************************************************
   SUBROUTINE init_gaussian_gridlevel(gridlevel_info, ngrid_levels, cutoff, rel_cutoff, print_section)
      TYPE(gridlevel_info_type)                          :: gridlevel_info
      INTEGER                                            :: ngrid_levels
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cutoff
      REAL(KIND=dp)                                      :: rel_cutoff
      TYPE(section_vals_type), POINTER                   :: print_section

      INTEGER                                            :: i

      ALLOCATE (gridlevel_info%cutoff(ngrid_levels))
      ALLOCATE (gridlevel_info%count(ngrid_levels))
      gridlevel_info%ngrid_levels = ngrid_levels
      gridlevel_info%rel_cutoff = rel_cutoff
      gridlevel_info%total_count = 0
      DO i = 1, ngrid_levels
         gridlevel_info%cutoff(i) = cutoff(i)
         gridlevel_info%count(i) = 0
      ENDDO
      gridlevel_info%print_section => print_section
      CALL section_vals_retain(print_section)
   END SUBROUTINE init_gaussian_gridlevel

! **************************************************************************************************
!> \brief ...
!> \param gridlevel_info ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE destroy_gaussian_gridlevel(gridlevel_info, para_env)
      TYPE(gridlevel_info_type)                          :: gridlevel_info
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'destroy_gaussian_gridlevel', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, i, output_unit
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      IF (PRESENT(para_env)) THEN
         group = para_env%group
      ELSE
         group = logger%para_env%group
      END IF
      output_unit = cp_print_key_unit_nr(logger, gridlevel_info%print_section, &
                                         "", extension=".Log")

      CALL mp_sum(gridlevel_info%total_count, group)
      CALL mp_sum(gridlevel_info%count, group)
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(/,T2,A,A)') "----------------------------------------", &
            "---------------------------------------"
         WRITE (output_unit, '(T2,A,T35,A,T77,A)') "----", "MULTIGRID INFO", "----"
         WRITE (output_unit, '(T2,A,A)') "----------------------------------------", &
            "---------------------------------------"
         IF (gridlevel_info%ngrid_levels > 1) THEN
            DO i = 1, gridlevel_info%ngrid_levels
               WRITE (output_unit, '(T2,A,I4,A,I14,9x,A,F12.2)') "count for grid     ", i, ": ", &
                  gridlevel_info%count(i), " cutoff [a.u.]    ", gridlevel_info%cutoff(i)
            END DO
            WRITE (output_unit, '(T2,A,I14)') "total gridlevel count  : ", &
               gridlevel_info%total_count
         ELSE
            WRITE (output_unit, '(T2,A,I14,T51,A,F12.2)') "total grid count     :", &
               gridlevel_info%count(1), " cutoff [a.u.]    ", gridlevel_info%cutoff(1)
         END IF
      END IF

      DEALLOCATE (gridlevel_info%cutoff)

      CALL cp_print_key_finished_output(output_unit, logger, gridlevel_info%print_section, &
                                        "")

      CALL section_vals_release(gridlevel_info%print_section)

      DEALLOCATE (gridlevel_info%count)

   END SUBROUTINE destroy_gaussian_gridlevel

! **************************************************************************************************
!> \brief ...
!> \param gridlevel_info ...
!> \param exponent ...
!> \retval gridlevel ...
! **************************************************************************************************
   FUNCTION gaussian_gridlevel(gridlevel_info, exponent) RESULT(gridlevel)
      TYPE(gridlevel_info_type), INTENT(INOUT)           :: gridlevel_info
      REAL(KIND=dp)                                      :: exponent
      INTEGER                                            :: gridlevel

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: needed_cutoff

      gridlevel = 1
      needed_cutoff = ABS(exponent)*gridlevel_info%rel_cutoff
      DO i = 1, gridlevel_info%ngrid_levels
         IF ((gridlevel_info%cutoff(i)+1E-6_dp) .GE. needed_cutoff) THEN
            gridlevel = i
         ENDIF
      ENDDO
      gridlevel_info%total_count = gridlevel_info%total_count+1
      gridlevel_info%count(gridlevel) = gridlevel_info%count(gridlevel)+1

   END FUNCTION gaussian_gridlevel

END MODULE gaussian_gridlevels
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief for a given dr()/dh(r) this will provide the bounds to be used if
!>      one wants to go over a sphere-subregion of given radius
!> \note
!>      the computation of the exact sphere radius is sensitive to roundoff (e.g.
!>      compiler optimization level) and hence this small roundoff can result in
!>      energy difference of about EPS_DEFAULT in QS energies (one gridpoint more or
!>      less in the density mapping)
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE cube_utils

   USE kinds,                           ONLY: dp,&
                                              int_size
   USE mathlib,                         ONLY: matvec_3x3
   USE realspace_grid_types,            ONLY: realspace_grid_desc_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cube_utils'

   PUBLIC :: cube_info_type

   PUBLIC :: init_cube_info, destroy_cube_info, &
             return_cube, return_cube_max_iradius, return_cube_nonortho, &
             compute_cube_center

   TYPE :: cube_ptr
      INTEGER, POINTER, DIMENSION(:) :: p
   END TYPE cube_ptr

   TYPE :: cube_info_type
      INTEGER                      :: max_radius
      REAL(KIND=dp)              :: dr(3), drmin
      REAL(KIND=dp)              :: dh(3, 3)
      REAL(KIND=dp)              :: dh_inv(3, 3)
      LOGICAL                      :: orthorhombic
      INTEGER, POINTER             :: lb_cube(:, :)
      INTEGER, POINTER             :: ub_cube(:, :)
      TYPE(cube_ptr), POINTER, DIMENSION(:)  :: sphere_bounds
      INTEGER, POINTER             :: sphere_bounds_count(:)
      REAL(KIND=dp)              :: max_rad_ga
   END TYPE cube_info_type

CONTAINS
! **************************************************************************************************
!> \brief unifies the computation of the cube center, so that differences in
!>        implementation, and thus roundoff and numerics can not lead to
!>        off-by-one errors (which can lead to out-of-bounds access with distributed grids).
!>        in principle, something similar would be needed for the computation of the cube bounds
!>
!> \param cube_center ...
!> \param rs_desc ...
!> \param zeta ...
!> \param zetb ...
!> \param ra ...
!> \param rab ...
!> \par History
!>      11.2008 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE compute_cube_center(cube_center, rs_desc, zeta, zetb, ra, rab)

      INTEGER, DIMENSION(3), INTENT(OUT)                 :: cube_center
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      REAL(KIND=dp), INTENT(IN)                          :: zeta, zetb, ra(3), rab(3)

      REAL(KIND=dp)                                      :: zetp
      REAL(KIND=dp), DIMENSION(3)                        :: rp

      zetp = zeta+zetb
      rp(:) = ra(:)+zetb/zetp*rab(:)
      cube_center(:) = FLOOR(MATMUL(rs_desc%dh_inv, rp))

   END SUBROUTINE compute_cube_center

! **************************************************************************************************
!> \brief ...
!> \param info ...
!> \param radius ...
!> \param lb ...
!> \param ub ...
!> \param rp ...
! **************************************************************************************************
   SUBROUTINE return_cube_nonortho(info, radius, lb, ub, rp)

      TYPE(cube_info_type)                               :: info
      REAL(KIND=dp)                                      :: radius
      INTEGER                                            :: lb(3), ub(3)
      REAL(KIND=dp)                                      :: rp(3)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'return_cube_nonortho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      REAL(KIND=dp)                                      :: point(3), res(3)

      IF (radius > info%max_rad_ga) THEN
         !
         ! This is an important check. If the required radius for mapping the density is different
         ! from the actual computed one, (significant) errors can occur.
         ! This error can invariably be fixed by improving the computation of maxradius
         ! in the call to init_cube_info
         !
         WRITE (*, *) info%max_rad_ga, radius
         CALL cp__b("pw/cube_utils.F",112,"Called with too large radius.")
      ENDIF

      ! get the min max indices that contain at least the cube that contains a sphere around rp of radius radius
      ! if the cell is very non-orthogonal this implies that many useless points are included
      ! this estimate can be improved (i.e. not box but sphere should be used)
      lb = HUGE(lb)
      ub = -HUGE(ub)
      DO i = -1, 1
      DO j = -1, 1
      DO k = -1, 1
         point(1) = rp(1)+i*radius
         point(2) = rp(2)+j*radius
         point(3) = rp(3)+k*radius
         CALL matvec_3x3(res, info%dh_inv, point)
         lb = MIN(lb, FLOOR(res))
         ub = MAX(ub, CEILING(res))
      ENDDO
      ENDDO
      ENDDO

   END SUBROUTINE return_cube_nonortho

! **************************************************************************************************
!> \brief ...
!> \param info ...
!> \param radius ...
!> \param lb_cube ...
!> \param ub_cube ...
!> \param sphere_bounds ...
! **************************************************************************************************
   SUBROUTINE return_cube(info, radius, lb_cube, ub_cube, sphere_bounds)

      TYPE(cube_info_type)                               :: info
      REAL(KIND=dp)                                      :: radius
      INTEGER                                            :: lb_cube(3), ub_cube(3)
      INTEGER, DIMENSION(:), POINTER                     :: sphere_bounds

      CHARACTER(LEN=*), PARAMETER :: routineN = 'return_cube', routineP = moduleN//':'//routineN

      INTEGER                                            :: imr

      IF (info%orthorhombic) THEN
         imr = MAX(1, CEILING(radius/info%drmin))
         IF (imr .GT. info%max_radius) THEN
            !
            ! This is an important check. If the required radius for mapping the density is different
            ! from the actual computed one, (significant) errors can occur.
            ! This error can invariably be fixed by improving the computation of maxradius
            ! in the call to init_cube_info
            !
            CALL cp__b("pw/cube_utils.F",163,"Called with too large radius.")
         ENDIF
         lb_cube(:) = info%lb_cube(:, imr)
         ub_cube(:) = info%ub_cube(:, imr)
         sphere_bounds => info%sphere_bounds(imr)%p
      ELSE
         ! nothing yet, we should check the radius
      ENDIF

   END SUBROUTINE return_cube

   ! this is the integer max radius of the cube
! **************************************************************************************************
!> \brief ...
!> \param info ...
!> \retval return_cube_max_iradius ...
! **************************************************************************************************
   INTEGER FUNCTION return_cube_max_iradius(info)
      TYPE(cube_info_type)                     :: info
      return_cube_max_iradius = info%max_radius
   END FUNCTION return_cube_max_iradius

! **************************************************************************************************
!> \brief ...
!> \param info ...
! **************************************************************************************************
   SUBROUTINE destroy_cube_info(info)
      TYPE(cube_info_type)                               :: info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'destroy_cube_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (info%orthorhombic) THEN
         DEALLOCATE (info%lb_cube)
         DEALLOCATE (info%ub_cube)
         DEALLOCATE (info%sphere_bounds_count)
         DO i = 1, info%max_radius
            DEALLOCATE (info%sphere_bounds(i)%p)
         END DO
         DEALLOCATE (info%sphere_bounds)
      ELSE
         ! no info to be deallocated
      ENDIF
   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param info ...
!> \param dr ...
!> \param dh ...
!> \param dh_inv ...
!> \param ortho ...
!> \param max_radius ...
! **************************************************************************************************
   SUBROUTINE init_cube_info(info, dr, dh, dh_inv, ortho, max_radius)
      TYPE(cube_info_type), INTENT(OUT)                  :: info
      REAL(KIND=dp), INTENT(IN)                          :: dr(3), dh(3, 3), dh_inv(3, 3)
      LOGICAL, INTENT(IN)                                :: ortho
      REAL(KIND=dp), INTENT(IN)                          :: max_radius

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_cube_info', routineP = moduleN//':'//routineN

      INTEGER                                            :: check_1, check_2, handle, i, igmin, imr, &
                                                            istat, jg, jg2, jgmin, k, kg, kg2, &
                                                            kgmin, lb(3), ub(3)
      REAL(KIND=dp)                                      :: drmin, dxi, dy2, dyi, dz2, dzi, radius, &
                                                            radius2, rp(3)

      CALL timeset(routineN, handle)
      info%dr = dr
      info%dh = dh
      info%dh_inv = dh_inv
      info%orthorhombic = ortho
      info%max_rad_ga = max_radius
      drmin = MINVAL(dr)
      info%drmin = drmin

      NULLIFY (info%lb_cube, info%ub_cube, &
               info%sphere_bounds_count, info%sphere_bounds)

      IF (.NOT. info%orthorhombic) THEN

         rp = 0.0_dp
         !
         ! could still be wrong (maybe needs an additional +1 to account for off-gridpoint rp's)
         !
         CALL return_cube_nonortho(info, max_radius, lb, ub, rp)
         info%max_radius = MAX(MAXVAL(ABS(lb)), MAXVAL(ABS(ub)))

      ELSE

         ! this info is specialized to orthogonal grids
         imr = CEILING((max_radius)/drmin)
         info%max_radius = imr
         dzi = 1.0_dp/dr(3)
         dyi = 1.0_dp/dr(2)
         dxi = 1.0_dp/dr(1)
         dz2 = (dr(3))**2
         dy2 = (dr(2))**2

         ALLOCATE (info%lb_cube(3, imr), info%ub_cube(3, imr), &
                   info%sphere_bounds_count(imr), info%sphere_bounds(imr))
         check_1 = 0
         check_2 = 0
!       count and allocate

         DO i = 1, imr
            k = 1
            radius = i*drmin
            radius2 = radius**2
            kgmin = do_and_hide_it_1(dzi, i, drmin, 0.0_dp, 0.0_dp, 0, 0)
            k = k+1
            DO kg = kgmin, 0
               kg2 = kg*kg
               jgmin = do_and_hide_it_1(dyi, i, drmin, dz2, 0.0_dp, kg2, 0)
               k = k+1
               DO jg = jgmin, 0
                  jg2 = jg*jg
                  igmin = do_and_hide_it_1(dxi, i, drmin, dz2, dy2, kg2, jg2)
                  check_1 = MODULO((kgmin*97+jgmin*37+igmin*113)*check_1+1277, 9343)
                  k = k+1
               ENDDO
            ENDDO
            info%sphere_bounds_count(i) = k-1
            ALLOCATE (info%sphere_bounds(i)%p(info%sphere_bounds_count(i)), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 290, &
                                             "info%sphere_bounds(i)%p", &
                                             int_size*info%sphere_bounds_count(i))
         ENDDO

!       init sphere_bounds array
         ! notice : as many points in lb_cube..0 as 1..ub_cube
         DO i = 1, imr
            k = 1
            radius = i*drmin
            info%lb_cube(:, i) = -1
            radius2 = radius**2
            kgmin = do_and_hide_it_1(dzi, i, drmin, 0.0_dp, 0.0_dp, 0, 0)
            info%lb_cube(3, i) = MIN(kgmin, info%lb_cube(3, i))
            info%sphere_bounds(i)%p(k) = kgmin
            k = k+1
            DO kg = kgmin, 0
               kg2 = kg*kg
               jgmin = do_and_hide_it_1(dyi, i, drmin, dz2, 0.0_dp, kg2, 0)
               info%lb_cube(2, i) = MIN(jgmin, info%lb_cube(2, i))
               info%sphere_bounds(i)%p(k) = jgmin
               k = k+1
               DO jg = jgmin, 0
                  jg2 = jg*jg
                  igmin = do_and_hide_it_1(dxi, i, drmin, dz2, dy2, kg2, jg2)
                  check_2 = MODULO((kgmin*97+jgmin*37+igmin*113)*check_2+1277, 9343)
                  info%lb_cube(1, i) = MIN(igmin, info%lb_cube(1, i))
                  info%sphere_bounds(i)%p(k) = igmin
                  k = k+1
               ENDDO
            ENDDO
            info%ub_cube(:, i) = 1-info%lb_cube(:, i)
         ENDDO
         IF (check_1 .NE. check_2) THEN
            CALL cp__b("pw/cube_utils.F",324,"Irreproducible fp math caused memory corruption")
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE

   ! try to hide things from the optimizer, so that we get the same numbers,
   ! always (this solves the optimisation problems with the intel and nag compiler
   ! in which the counting loops and execution loops above are executed a different
   ! number of times, even at -O1
! **************************************************************************************************
!> \brief ...
!> \param prefactor ...
!> \param i ...
!> \param drmin ...
!> \param dz2 ...
!> \param dy2 ...
!> \param kg2 ...
!> \param jg2 ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION do_and_hide_it_1(prefactor, i, drmin, dz2, dy2, kg2, jg2) RESULT(res)
      REAL(KIND=dp)                                      :: prefactor
      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: drmin, dz2, dy2
      INTEGER                                            :: kg2, jg2, res

      REAL(KIND=dp), DIMENSION(:), POINTER               :: buf

      ALLOCATE (buf(4))
      buf(1) = prefactor
      buf(2) = drmin
      buf(3) = dz2
      buf(4) = dy2
      res = do_and_hide_it_2(buf, i, jg2, kg2)
      DEALLOCATE (buf)
   END FUNCTION do_and_hide_it_1

! **************************************************************************************************
!> \brief ...
!> \param buf ...
!> \param i ...
!> \param jg2 ...
!> \param kg2 ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION do_and_hide_it_2(buf, i, jg2, kg2) RESULT(res)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: buf
      INTEGER                                            :: i, jg2, kg2, res

      buf(2) = (i*buf(2))**2
      res = CEILING(-0.1E-7_dp-buf(1)*SQRT(MAX(buf(2)-kg2*buf(3)-jg2*buf(4), 0.0_dp)))
   END FUNCTION do_and_hide_it_2

END MODULE cube_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief container for various plainwaves related things
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE pw_env_types
   USE cube_utils,                      ONLY: cube_info_type,&
                                              destroy_cube_info
   USE gaussian_gridlevels,             ONLY: destroy_gaussian_gridlevel,&
                                              gridlevel_info_type
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE lgrid_types,                     ONLY: lgrid_release,&
                                              lgrid_type
   USE pw_poisson_types,                ONLY: pw_poisson_release,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_p_type,&
                                              pw_pool_release,&
                                              pw_pool_type,&
                                              pw_pools_dealloc
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_desc_type,&
                                              realspace_grid_p_type,&
                                              realspace_grid_type,&
                                              rs_grid_release,&
                                              rs_grid_release_descriptor
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_env_types'

   PUBLIC :: pw_env_type
   PUBLIC :: pw_env_retain, pw_env_release, pw_env_get
!***

! **************************************************************************************************
!> \brief contained for different pw related things
!> \param pw_pools pools for multigrids
!> \param rs_descs descriptors for the realspace grids
!> \param rs_grids the realspace grids
!> \param lgrid scratch space for OpenMP threads rs_grid operations
!> \param xc_pw_pool the pool used for xc calculation on grid (might be the
!>        same as the finest, or finer still).
!> \param vdw_pw_pool the pool used for vdw-nl calculation on grid (might be the
!>        same or a grid with a new cutoff
!> \param gridlevel_info gaussian gridlevel info
!> \param cube_info info on cube and max sphere size for multigrids
!> \param aux_bas which multigrid is the auxiliar basis
!>
!>      readonly attributes
!> \param auxbas_pw_pool a pool that allocates grids in the auxiliary basis
!> \param auxbas_rs_desc real space grid in the auxiliary basis, be careful
!>        in parallel nsmax is choosen with multigrid in mind!
!> \note
!>      be careful in parallel nsmax for the rs_grids is choosen with
!>      multigrid in mind! (well not as of this writing but...)
!> \par History
!>      10.2002 created [fawzi]
!>      04.2003 added rs grid pools [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE pw_env_type
      INTEGER                                     :: ref_count
      INTEGER                                     :: auxbas_grid
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER :: pw_pools
      TYPE(pw_pool_type), POINTER                 :: xc_pw_pool
      TYPE(pw_pool_type), POINTER                 :: vdw_pw_pool
      TYPE(realspace_grid_desc_p_type), &
         DIMENSION(:), POINTER                    :: rs_descs
      TYPE(realspace_grid_p_type), &
         DIMENSION(:), POINTER                    :: rs_grids
      TYPE(lgrid_type), POINTER                   :: lgrid
      TYPE(gridlevel_info_type), POINTER          :: gridlevel_info
      TYPE(cube_info_type), DIMENSION(:), POINTER :: cube_info
      TYPE(pw_poisson_type), POINTER              :: poisson_env
      TYPE(section_vals_type), POINTER            :: interp_section
      ! store current cell, used to rebuild lazily.
      REAL(KIND=dp), DIMENSION(3, 3)          :: cell_hmat = -1.0_dp
   END TYPE pw_env_type

CONTAINS

! **************************************************************************************************
!> \brief returns the various attributes of the pw env
!> \param pw_env the einvironment you want info about
!> \param pw_pools ...
!> \param cube_info ...
!> \param gridlevel_info ...
!> \param auxbas_pw_pool ...
!> \param auxbas_grid ...
!> \param auxbas_rs_desc ...
!> \param auxbas_rs_grid ...
!> \param rs_descs ...
!> \param rs_grids ...
!> \param lgrid ...
!> \param xc_pw_pool ...
!> \param vdw_pw_pool ...
!> \param poisson_env ...
!> \param interp_section ...
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_env_get(pw_env, pw_pools, cube_info, gridlevel_info, &
                         auxbas_pw_pool, auxbas_grid, auxbas_rs_desc, auxbas_rs_grid, rs_descs, rs_grids, lgrid, &
                         xc_pw_pool, vdw_pw_pool, poisson_env, interp_section)
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_pool_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: pw_pools
      TYPE(cube_info_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: cube_info
      TYPE(gridlevel_info_type), OPTIONAL, POINTER       :: gridlevel_info
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: auxbas_pw_pool
      INTEGER, INTENT(out), OPTIONAL                     :: auxbas_grid
      TYPE(realspace_grid_desc_type), OPTIONAL, POINTER  :: auxbas_rs_desc
      TYPE(realspace_grid_type), OPTIONAL, POINTER       :: auxbas_rs_grid
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: rs_grids
      TYPE(lgrid_type), OPTIONAL, POINTER                :: lgrid
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: xc_pw_pool, vdw_pw_pool
      TYPE(pw_poisson_type), OPTIONAL, POINTER           :: poisson_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: interp_section

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_env_get', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("pw_env_types.F",138)
      IF(.NOT.(pw_env%ref_count > 0))CALL cp__a("pw_env_types.F",139)
      IF (PRESENT(pw_pools)) pw_pools => pw_env%pw_pools
      IF (PRESENT(rs_descs)) rs_descs => pw_env%rs_descs
      IF (PRESENT(rs_grids)) rs_grids => pw_env%rs_grids
      IF (PRESENT(lgrid)) lgrid => pw_env%lgrid
      IF (PRESENT(cube_info)) cube_info => pw_env%cube_info
      IF (PRESENT(gridlevel_info)) gridlevel_info => pw_env%gridlevel_info
      IF (PRESENT(auxbas_pw_pool)) THEN
         auxbas_pw_pool => pw_env%pw_pools(pw_env%auxbas_grid)%pool
      END IF
      IF (PRESENT(auxbas_rs_desc)) THEN
         auxbas_rs_desc => pw_env%rs_descs(pw_env%auxbas_grid)%rs_desc
      END IF
      IF (PRESENT(auxbas_rs_grid)) THEN
         auxbas_rs_grid => pw_env%rs_grids(pw_env%auxbas_grid)%rs_grid
      END IF
      IF (PRESENT(auxbas_grid)) auxbas_grid = pw_env%auxbas_grid
      IF (PRESENT(xc_pw_pool)) xc_pw_pool => pw_env%xc_pw_pool
      IF (PRESENT(vdw_pw_pool)) vdw_pw_pool => pw_env%vdw_pw_pool
      IF (PRESENT(poisson_env)) poisson_env => pw_env%poisson_env
      IF (PRESENT(interp_section)) interp_section => pw_env%interp_section
   END SUBROUTINE pw_env_get

! **************************************************************************************************
!> \brief retains the pw_env (see doc/ReferenceCounting.html)
!> \param pw_env the pw_env to retain
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_env_retain(pw_env)
      TYPE(pw_env_type), POINTER                         :: pw_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_env_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("pw_env_types.F",174)
      IF(.NOT.(pw_env%ref_count > 0))CALL cp__a("pw_env_types.F",175)
      pw_env%ref_count = pw_env%ref_count+1
   END SUBROUTINE pw_env_retain

! **************************************************************************************************
!> \brief releases the given pw_env (see doc/ReferenceCounting.html)
!> \param pw_env the pw_env to release
!> \param kg ...
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_env_release(pw_env, kg)
      TYPE(pw_env_type), POINTER                         :: pw_env
      LOGICAL, OPTIONAL                                  :: kg

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_env_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, igrid_level
      LOGICAL                                            :: my_kg

      my_kg = .FALSE.
      IF (PRESENT(kg)) my_kg = kg
      IF (ASSOCIATED(pw_env)) THEN
         IF(.NOT.(pw_env%ref_count > 0))CALL cp__a("pw_env_types.F",199)
         pw_env%ref_count = pw_env%ref_count-1
         IF (pw_env%ref_count < 1) THEN
            CALL pw_poisson_release(pw_env%poisson_env)
            CALL pw_pools_dealloc(pw_env%pw_pools)
            IF (ASSOCIATED(pw_env%gridlevel_info)) THEN
               IF (my_kg) THEN
                  DEALLOCATE (pw_env%gridlevel_info%cutoff)
                  CALL section_vals_release(pw_env%gridlevel_info%print_section)
                  DEALLOCATE (pw_env%gridlevel_info%count)
               ELSE
                  CALL destroy_gaussian_gridlevel(pw_env%gridlevel_info)
               END IF
               DEALLOCATE (pw_env%gridlevel_info)
            END IF
            IF (ASSOCIATED(pw_env%cube_info)) THEN
               DO igrid_level = 1, SIZE(pw_env%cube_info)
                  CALL destroy_cube_info(pw_env%cube_info(igrid_level))
               END DO
               DEALLOCATE (pw_env%cube_info)
            END IF
            CALL pw_pool_release(pw_env%xc_pw_pool)
            CALL pw_pool_release(pw_env%vdw_pw_pool)
            IF (ASSOCIATED(pw_env%rs_descs)) THEN
               DO i = 1, SIZE(pw_env%rs_descs)
                  CALL rs_grid_release_descriptor(pw_env%rs_descs(i)%rs_desc)
               END DO
               DEALLOCATE (pw_env%rs_descs)
            END IF
            IF (ASSOCIATED(pw_env%rs_grids)) THEN
               DO i = 1, SIZE(pw_env%rs_grids)
                  CALL rs_grid_release(pw_env%rs_grids(i)%rs_grid)
               END DO
               DEALLOCATE (pw_env%rs_grids)
            END IF
            IF (ASSOCIATED(pw_env%lgrid)) THEN
               CALL lgrid_release(pw_env%lgrid)
            END IF
            DEALLOCATE (pw_env)
         END IF
      END IF
      NULLIFY (pw_env)
   END SUBROUTINE pw_env_release

END MODULE pw_env_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief types for task lists
!> \par History
!>      01.2008 [Joost VandeVondele] refactered out of qs_collocate / qs_integrate
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE task_list_types

   USE kinds,                           ONLY: dp,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'task_list_types'

! **************************************************************************************************
   TYPE task_list_type
      INTEGER(kind=int_8), DIMENSION(:, :), POINTER :: tasks
      REAL(KIND=dp), DIMENSION(:, :), POINTER       :: dist_ab
      INTEGER(kind=int_8), DIMENSION(:), POINTER    :: atom_pair_send, atom_pair_recv
      INTEGER                                       :: ntasks
      INTEGER, DIMENSION(:, :), POINTER               :: taskstart, taskstop
      INTEGER, DIMENSION(:), POINTER                 :: npairs
   END TYPE task_list_type

   PUBLIC :: task_list_type

   PUBLIC :: allocate_task_list, &
             deallocate_task_list

CONTAINS

! **************************************************************************************************
!> \brief allocates and initialised the components of the task_list_type
!> \param task_list ...
!> \par History
!>      01.2008 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE allocate_task_list(task_list)
      TYPE(task_list_type), POINTER                      :: task_list

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_task_list', &
         routineP = moduleN//':'//routineN

      ALLOCATE (task_list)

      NULLIFY (task_list%tasks)
      NULLIFY (task_list%dist_ab)
      NULLIFY (task_list%atom_pair_send)
      NULLIFY (task_list%atom_pair_recv)
      NULLIFY (task_list%taskstart)
      NULLIFY (task_list%taskstop)
      NULLIFY (task_list%npairs)
      task_list%ntasks = 0
   END SUBROUTINE allocate_task_list

! **************************************************************************************************
!> \brief deallocates the components and the object itself
!> \param task_list ...
!> \par History
!>      01.2008 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE deallocate_task_list(task_list)
      TYPE(task_list_type), POINTER                      :: task_list

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_task_list', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(task_list%tasks)) THEN
         DEALLOCATE (task_list%tasks)
      ENDIF
      IF (ASSOCIATED(task_list%dist_ab)) THEN
         DEALLOCATE (task_list%dist_ab)
      ENDIF
      IF (ASSOCIATED(task_list%atom_pair_send)) THEN
         DEALLOCATE (task_list%atom_pair_send)
      ENDIF
      IF (ASSOCIATED(task_list%atom_pair_recv)) THEN
         DEALLOCATE (task_list%atom_pair_recv)
      ENDIF
      IF (ASSOCIATED(task_list%taskstart)) THEN
         DEALLOCATE (task_list%taskstart)
      ENDIF
      IF (ASSOCIATED(task_list%taskstop)) THEN
         DEALLOCATE (task_list%taskstop)
      ENDIF
      IF (ASSOCIATED(task_list%npairs)) THEN
         DEALLOCATE (task_list%npairs)
      ENDIF

      DEALLOCATE (task_list)
   END SUBROUTINE deallocate_task_list
END MODULE task_list_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! less not much meningful...

! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_fm
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_fm'

!API type
   PUBLIC :: cp_sll_fm_type, cp_sll_fm_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_fm_create, cp_sll_fm_dealloc, &
             cp_sll_fm_next, &
             cp_sll_fm_get_length, &
             cp_sll_fm_get_el_at, cp_sll_fm_set_el_at, &
             cp_sll_fm_get_first_el, cp_sll_fm_get_rest, &
             cp_sll_fm_insert_el, cp_sll_fm_insert_el_at, &
             cp_sll_fm_rm_first_el, cp_sll_fm_rm_el_at, &
             cp_sll_fm_rm_all_el, &
             cp_sll_fm_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_fm_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_fm_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_fm_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_fm_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_fm_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_fm_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_fm_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_fm_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_fm_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_fm_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_fm_rm_first_el, &
         cp_sll_fm_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_fm_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_fm_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_fm_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_fm_type
      TYPE(cp_fm_type), POINTER :: first_el
      TYPE(cp_sll_fm_type), POINTER :: rest
   END TYPE cp_sll_fm_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_fm_p_type
      TYPE(cp_sll_fm_type), POINTER :: list
   END TYPE cp_sll_fm_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_create(sll, first_el, rest)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: first_el
      TYPE(cp_sll_fm_type), OPTIONAL, POINTER            :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el => first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_fm_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_dealloc(sll)
      TYPE(cp_sll_fm_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_fm_rm_all_el(sll)
   END SUBROUTINE cp_sll_fm_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_dealloc_node(sll)
      TYPE(cp_sll_fm_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_fm_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_fm_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      TYPE(cp_fm_type), POINTER                          :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_get_first_el', &
         routineP = moduleN//':'//routineN

      res => sll%first_el
   END FUNCTION cp_sll_fm_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_fm_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_fm_type), POINTER                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("fm/cp_linked_list_fm.F",295,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_fm_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_fm_get_length(sll) RESULT(res)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_fm_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_fm_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(cp_fm_type), POINTER                          :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: pos

      IF (index == -1) THEN
         pos => cp_sll_fm_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_fm_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("fm/cp_linked_list_fm.F",367)

      res => pos%first_el
   END FUNCTION cp_sll_fm_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_set_el_at(sll, index, value)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(cp_fm_type), POINTER                          :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: pos

      IF (index == -1) THEN
         pos => cp_sll_fm_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_fm_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("fm/cp_linked_list_fm.F",399)

      pos%first_el => value
   END SUBROUTINE cp_sll_fm_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_fm_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_fm_type), POINTER                      :: iterator
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_next', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att => iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_fm_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_insert_el(sll, el)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      TYPE(cp_fm_type), POINTER                          :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_fm_create(newSlot, first_el=el, &
                            rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_fm_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_rm_first_el(sll)
      TYPE(cp_sll_fm_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_fm_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("fm/cp_linked_list_fm.F",484,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_fm_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_insert_el_at(sll, el, index)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      TYPE(cp_fm_type), POINTER                          :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: pos

      IF (index == 1) THEN
         CALL cp_sll_fm_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_fm_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_fm_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("fm/cp_linked_list_fm.F",518)
         CALL cp_sll_fm_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_fm_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_rm_el_at(sll, index)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: pos

      IF (index == 1) THEN
         CALL cp_sll_fm_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_fm_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_fm_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("fm/cp_linked_list_fm.F",550)
         CALL cp_sll_fm_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_fm_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_fm_rm_all_el(sll)
      TYPE(cp_sll_fm_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_fm_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_fm_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_fm_to_array(sll) RESULT(res)
      TYPE(cp_sll_fm_type), POINTER                      :: sll
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_fm_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_fm_type), POINTER                      :: iter

      len = cp_sll_fm_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i)%matrix => iter%first_el
         IF(.NOT.(cp_sll_fm_next(iter) .OR. i == len))CALL cp__a("fm/cp_linked_list_fm.F",606)
      END DO
   END FUNCTION cp_sll_fm_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "=>"
!  USE = "USE cp_fm_types, ONLY: cp_fm_type, cp_fm_p_type"
!  array= = "=>"
!  arrayEl = "%matrix"
!  common_dir = "../common"
!  defines =
!    "! less not much meningful...
!     #define ( el1 %id_nr < el2 %id_nr ) ( el1 %id_nr < el2 %id_nr )
!     #define ( el1 %id_nr == el2 %id_nr ) ( el1 %id_nr == el2 %id_nr )
!     "
!  equalQ = "CP_SLL_FM_EQUAL_Q"
!  lessQ = "CP_SLL_FM_LESS_Q"
!  nametype1 = "fm"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "type(cp_fm_type),pointer"
!  type1arrayEl = "type(cp_fm_p_type)"
!  type1in = "type(cp_fm_type), pointer"
!  type1out = "type(cp_fm_type), pointer"
!  write_el = ""

END MODULE cp_linked_list_fm

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief pool for for elements that are retained and released
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_fm_pool_types
   USE cp_fm_struct,                    ONLY: cp_fm_struct_release,&
                                              cp_fm_struct_retain,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_linked_list_fm,               ONLY: cp_sll_fm_dealloc,&
                                              cp_sll_fm_get_first_el,&
                                              cp_sll_fm_insert_el,&
                                              cp_sll_fm_next,&
                                              cp_sll_fm_rm_first_el,&
                                              cp_sll_fm_type
   USE cp_log_handling,                 ONLY: cp_to_string
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_fm_pool_types'
   INTEGER, SAVE, PRIVATE :: last_fm_pool_id_nr = 0

   PUBLIC :: cp_fm_pool_type, cp_fm_pool_p_type
   PUBLIC :: fm_pool_create, fm_pool_retain, &
             fm_pool_release, &
             fm_pool_create_fm, fm_pool_give_back_fm, &
             fm_pool_get_el_struct
   PUBLIC :: fm_pools_dealloc, &
             fm_pools_create_fm_vect, &
             fm_pools_give_back_fm_vect
!***

! **************************************************************************************************
!> \brief represent a pool of elements with the same structure
!> \param ref_count reference count (see /cp2k/doc/ReferenceCounting.html)
!> \param el_struct the structure of the elements stored in this pool
!> \param cache linked list with the elements in the pool
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_fm_pool_type
      PRIVATE
      INTEGER :: ref_count, id_nr
      TYPE(cp_fm_struct_type), POINTER :: el_struct

      TYPE(cp_sll_fm_type), POINTER :: cache
   END TYPE cp_fm_pool_type

! **************************************************************************************************
!> \brief to create arrays of pools
!> \param pool the pool
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_fm_pool_p_type
      TYPE(cp_fm_pool_type), POINTER :: pool
   END TYPE cp_fm_pool_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a pool of elements
!> \param pool the pool to create
!> \param el_struct the structure of the elements that are stored in
!>        this pool
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pool_create(pool, el_struct)
      TYPE(cp_fm_pool_type), POINTER                     :: pool
      TYPE(cp_fm_struct_type), POINTER                   :: el_struct

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pool_create', routineP = moduleN//':'//routineN

      ALLOCATE (pool)
      pool%el_struct => el_struct
      CALL cp_fm_struct_retain(pool%el_struct)
      last_fm_pool_id_nr = last_fm_pool_id_nr+1
      pool%id_nr = last_fm_pool_id_nr
      pool%ref_count = 1
      NULLIFY (pool%cache)

   END SUBROUTINE fm_pool_create

! **************************************************************************************************
!> \brief retains the pool (see cp2k/doc/ReferenceCounting.html)
!> \param pool the pool to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pool_retain(pool)
      TYPE(cp_fm_pool_type), POINTER                     :: pool

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pool_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("fm/cp_fm_pool_types.F",113)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("fm/cp_fm_pool_types.F",114)

      pool%ref_count = pool%ref_count+1
   END SUBROUTINE fm_pool_retain

! **************************************************************************************************
!> \brief deallocates all the cached elements
!> \param pool the pool to flush
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pool_flush_cache(pool)
      TYPE(cp_fm_pool_type), POINTER                     :: pool

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pool_flush_cache', &
         routineP = moduleN//':'//routineN

      TYPE(cp_fm_type), POINTER                          :: el_att
      TYPE(cp_sll_fm_type), POINTER                      :: iterator

      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("fm/cp_fm_pool_types.F",135)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("fm/cp_fm_pool_types.F",136)
      iterator => pool%cache
      DO
         IF (.NOT. cp_sll_fm_next(iterator, el_att=el_att)) EXIT
         CALL cp_fm_release(el_att)
      END DO
      CALL cp_sll_fm_dealloc(pool%cache)
   END SUBROUTINE fm_pool_flush_cache

! **************************************************************************************************
!> \brief releases the given pool (see cp2k/doc/ReferenceCounting.html)
!> \param pool the pool to release
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pool_release(pool)
      TYPE(cp_fm_pool_type), POINTER                     :: pool

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pool_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(pool)) THEN
         IF(.NOT.(pool%ref_count > 0))CALL cp__a("fm/cp_fm_pool_types.F",159)
         pool%ref_count = pool%ref_count-1
         IF (pool%ref_count == 0) THEN
            pool%ref_count = 1
            CALL fm_pool_flush_cache(pool)
            CALL cp_fm_struct_release(pool%el_struct)
            pool%ref_count = 0

            DEALLOCATE (pool)
         END IF
      END IF
      NULLIFY (pool)
   END SUBROUTINE fm_pool_release

! **************************************************************************************************
!> \brief returns an element, allocating it if none is in the pool
!> \param pool the pool from where you get the element
!> \param element will contain the new element
!>\param name the name for the new matrix (optional)
!> \param name ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pool_create_fm(pool, element, &
                                name)
      TYPE(cp_fm_pool_type), POINTER                     :: pool
      TYPE(cp_fm_type), POINTER                          :: element
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pool_create_fm', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("fm/cp_fm_pool_types.F",192)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("fm/cp_fm_pool_types.F",193)
      IF (ASSOCIATED(pool%cache)) THEN
         element => cp_sll_fm_get_first_el(pool%cache)
         CALL cp_sll_fm_rm_first_el(pool%cache)

      ELSE
         NULLIFY (element)
         CALL cp_fm_create(element, matrix_struct=pool%el_struct)
      END IF

      IF (PRESENT(name)) THEN
         element%name = name
         element%print_count = 0
      ELSE
         element%name = "tmp-"//TRIM(ADJUSTL(cp_to_string(element%id_nr)))
         element%print_count = 0
         ! guarantee output unicity?
      END IF

      IF(.NOT.(ASSOCIATED(element)))CALL cp__a("fm/cp_fm_pool_types.F",212)
      IF(.NOT.(element%ref_count == 1))CALL cp__a("fm/cp_fm_pool_types.F",213)
   END SUBROUTINE fm_pool_create_fm

! **************************************************************************************************
!> \brief returns the element to the pool
!> \param pool the pool where to cache the element
!> \param element the element to give back
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      transfers the ownership of the element to the pool
!>      (it is as if you had called cp_fm_release)
!>      Accept give_backs of non associated elements?
! **************************************************************************************************
   SUBROUTINE fm_pool_give_back_fm(pool, element)
      TYPE(cp_fm_pool_type), POINTER                     :: pool
      TYPE(cp_fm_type), POINTER                          :: element

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pool_give_back_fm', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("fm/cp_fm_pool_types.F",235)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("fm/cp_fm_pool_types.F",236)
      IF(.NOT.(ASSOCIATED(element)))CALL cp__a("fm/cp_fm_pool_types.F",237)
      IF (pool%el_struct%id_nr /= element%matrix_struct%id_nr) &
         CALL cp__b("fm/cp_fm_pool_types.F",239,"pool cannot reuse matrixes with another structure")

      IF(.NOT.(element%ref_count == 1))CALL cp__a("fm/cp_fm_pool_types.F",241)
      CALL cp_sll_fm_insert_el(pool%cache, el=element)
      NULLIFY (element)
   END SUBROUTINE fm_pool_give_back_fm

! **************************************************************************************************
!> \brief returns the structure of the elements in this pool
!> \param pool the pool you are interested in
!> \retval res ...
!> \par History
!>      05.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION fm_pool_get_el_struct(pool) RESULT(res)
      TYPE(cp_fm_pool_type), POINTER                     :: pool
      TYPE(cp_fm_struct_type), POINTER                   :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pool_get_el_struct', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pool)))CALL cp__a("fm/cp_fm_pool_types.F",261)
      IF(.NOT.(pool%ref_count > 0))CALL cp__a("fm/cp_fm_pool_types.F",262)
      res => pool%el_struct
   END FUNCTION fm_pool_get_el_struct

!================== pools ================

! **************************************************************************************************
!> \brief shallow copy of an array of pools (retains each pool)
!> \param source_pools the pools to copy
!> \param target_pools will contains the new pools
!> \par History
!>      11.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pools_copy(source_pools, target_pools)
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: source_pools, target_pools

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pools_copy', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(source_pools)))CALL cp__a("fm/cp_fm_pool_types.F",283)
      ALLOCATE (target_pools(SIZE(source_pools)))
      DO i = 1, SIZE(source_pools)
         target_pools(i)%pool => source_pools(i)%pool
         CALL fm_pool_retain(source_pools(i)%pool)
      END DO
   END SUBROUTINE fm_pools_copy

! **************************************************************************************************
!> \brief deallocate an array of pools (releasing each pool)
!> \param pools the pools to release
!> \par History
!>      11.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pools_dealloc(pools)
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: pools

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pools_dealloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(pools)) THEN
         DO i = 1, SIZE(pools)
            CALL fm_pool_release(pools(i)%pool)
         END DO
         DEALLOCATE (pools)
      END IF
   END SUBROUTINE fm_pools_dealloc

! **************************************************************************************************
!> \brief Returns a vector with an element from each pool
!> \param pools the pools to create the elements from
!> \param elements will contain the vector of elements
!> \param name the name for the new matrixes (optional)
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fm_pools_create_fm_vect(pools, elements, &
                                      name)
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: pools
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: elements
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pools_create_fm_vect', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(cp_fm_pool_type), POINTER                     :: pool

      NULLIFY (pool)

      IF(.NOT.(ASSOCIATED(pools)))CALL cp__a("fm/cp_fm_pool_types.F",337)
      ALLOCATE (elements(SIZE(pools)))
      DO i = 1, SIZE(pools)
         NULLIFY (elements(i)%matrix)
         pool => pools(i)%pool
         IF (PRESENT(name)) THEN
            CALL fm_pool_create_fm(pool, elements(i)%matrix, &
                                   name=name//"-"//ADJUSTL(cp_to_string(i)))
         ELSE
            CALL fm_pool_create_fm(pool, elements(i)%matrix)
         END IF

      END DO

   END SUBROUTINE fm_pools_create_fm_vect

! **************************************************************************************************
!> \brief returns a vector to the pools. The vector is deallocated
!>      (like cp_fm_vect_dealloc)
!> \param pools the pool where to give back the vector
!> \param elements the vector of elements to give back
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      accept unassociated vect?
! **************************************************************************************************
   SUBROUTINE fm_pools_give_back_fm_vect(pools, elements)
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: pools
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: elements

      CHARACTER(len=*), PARAMETER :: routineN = 'fm_pools_give_back_fm_vect', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(pools)))CALL cp__a("fm/cp_fm_pool_types.F",373)
      IF(.NOT.(ASSOCIATED(elements)))CALL cp__a("fm/cp_fm_pool_types.F",374)
      IF(.NOT.(SIZE(pools) == SIZE(elements)))CALL cp__a("fm/cp_fm_pool_types.F",375)
      DO i = 1, SIZE(pools)
         CALL fm_pool_give_back_fm(pools(i)%pool, &
                                   elements(i)%matrix)
      END DO
      DEALLOCATE (elements)
   END SUBROUTINE fm_pools_give_back_fm_vect

END MODULE cp_fm_pool_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a complex full matrix distribued on many processors
!> \author Joost VandeVondele, based on Fawzi's cp_fm_* routines
! **************************************************************************************************
MODULE cp_cfm_types

   USE cp_blacs_calls,                  ONLY: cp_blacs_zgebr2d,&
                                              cp_blacs_zgebs2d
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_fm_struct,                    ONLY: cp_fm_struct_equivalent,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_retain,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: cp2k_is_parallel,&
                                              mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_cfm_types'

   INTEGER, SAVE, PRIVATE :: last_cfm_id_nr = 0

   PUBLIC :: cp_cfm_type, cp_cfm_p_type
   PUBLIC :: cp_cfm_create, cp_cfm_release
   PUBLIC :: cp_cfm_set_all, cp_cfm_get_info, cp_cfm_get_element
   PUBLIC :: cp_cfm_set_element, cp_cfm_to_cfm, cp_cfm_to_fm
   PUBLIC :: cp_cfm_set_submatrix, cp_cfm_to_cfm_triangular
   PUBLIC :: cp_cfm_get_submatrix

   INTERFACE cp_cfm_to_cfm
      MODULE PROCEDURE cp_cfm_to_cfm_matrix, & ! a full matrix
         cp_cfm_to_cfm_columns ! just a number of columns
   END INTERFACE

! **************************************************************************************************
!> \brief represent a complex full matrix
!> \param name the name of the matrix, used for printing
!> \param id_nr identification number that makes this matrix unique
!> \param ref_cont reference count (see doc/ReferenceCounting.html)
!> \param print_count sequential number to make the printed files unique
!> \param matrix_struct structure of this matrix
!> \param local_data array with the data of the matrix (its contents
!>        depend on the matrix type used: in parallel runs it will be
!>        in scalapack format, in sequential, it will simply contain
!>        the matrix)
! **************************************************************************************************
   TYPE cp_cfm_type
!    PRIVATE
      CHARACTER(LEN=60) :: name
      INTEGER :: id_nr, ref_count, print_count
      TYPE(cp_fm_struct_type), POINTER :: matrix_struct
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: local_data
   END TYPE cp_cfm_type

! **************************************************************************************************
!> \brief just to build arrays of pointers to matrixes
!> \param matrix the pointer to the matrix
! **************************************************************************************************
   TYPE cp_cfm_p_type
      TYPE(cp_cfm_type), POINTER :: matrix
   END TYPE cp_cfm_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a new full matrix with the given structure
!> \param matrix the matrix to be created
!> \param matrix_struct the structure of matrix
!> \param name ...
!> \note
!>      preferred allocation routine
! **************************************************************************************************
   SUBROUTINE cp_cfm_create(matrix, matrix_struct, name)
      TYPE(cp_cfm_type), POINTER               :: matrix
      TYPE(cp_fm_struct_type), POINTER         :: matrix_struct
      CHARACTER(len=*), INTENT(in), OPTIONAL   :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_create', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: ncol_local, npcol, nprow, &
                                                  nrow_local
      TYPE(cp_blacs_env_type), POINTER         :: context


      IF(.NOT.(ASSOCIATED(matrix_struct)))CALL cp__a("fm/cp_cfm_types.F",101)
      ALLOCATE (matrix)
      context => matrix_struct%context
      matrix%matrix_struct => matrix_struct
      CALL cp_fm_struct_retain(matrix%matrix_struct)
      last_cfm_id_nr = last_cfm_id_nr+1
      matrix%id_nr = last_cfm_id_nr
      matrix%ref_count = 1
      matrix%print_count = 0

      nprow = context%num_pe(1)
      npcol = context%num_pe(2)
      NULLIFY (matrix%local_data)

      nrow_local = matrix_struct%local_leading_dimension
      ncol_local = MAX(1, matrix_struct%ncol_locals(context%mepos(2)))
      ALLOCATE (matrix%local_data(nrow_local, ncol_local))

      CALL dcopy(2*nrow_local*ncol_local, 0.0_dp, 0, matrix%local_data, 1)
! matrix%local_data(:,:) = 0.0_dp

      IF (PRESENT(name)) THEN
         matrix%name = name
      ELSE
         matrix%name = 'full matrix'//cp_to_string(matrix%id_nr)
      END IF
   END SUBROUTINE cp_cfm_create

! **************************************************************************************************
!> \brief retains a full matrix
!> \param matrix the matrix to retain
! **************************************************************************************************
   SUBROUTINE cp_cfm_retain(matrix)
      TYPE(cp_cfm_type), POINTER                         :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_cfm_types.F",138)
      IF(.NOT.(matrix%ref_count > 0))CALL cp__a("fm/cp_cfm_types.F",139)
      matrix%ref_count = matrix%ref_count+1
   END SUBROUTINE cp_cfm_retain

! **************************************************************************************************
!> \brief releases a full matrix
!> \param matrix the matrix to release
! **************************************************************************************************
   SUBROUTINE cp_cfm_release(matrix)
      TYPE(cp_cfm_type), POINTER                         :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(matrix)) THEN
         IF(.NOT.(matrix%ref_count > 0))CALL cp__a("fm/cp_cfm_types.F",153)
         matrix%ref_count = matrix%ref_count-1
         IF (matrix%ref_count < 1) THEN
            IF (ASSOCIATED(matrix%local_data)) THEN
               DEALLOCATE (matrix%local_data)
            END IF
            matrix%name = ""
            CALL cp_fm_struct_release(matrix%matrix_struct)
            DEALLOCATE (matrix)
         END IF
      END IF
      NULLIFY (matrix)
   END SUBROUTINE cp_cfm_release

! **************************************************************************************************
!> \brief   Set the FULL matrix elements to alpha.
!> \param matrix ...
!> \param alpha ...
!> \param beta ...
!> \date    12.06.2001
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_cfm_set_all(matrix, alpha, beta)
      TYPE(cp_cfm_type), POINTER                         :: matrix
      COMPLEX(KIND=dp), INTENT(IN)                       :: alpha
      COMPLEX(KIND=dp), INTENT(IN), OPTIONAL             :: beta

      INTEGER                                            :: i

      matrix%local_data(:, :) = alpha
      IF (PRESENT(beta)) THEN
         DO i = 1, matrix%matrix_struct%nrow_global
            CALL cp_cfm_set_element(matrix, i, i, beta)
         ENDDO
      ENDIF

   END SUBROUTINE cp_cfm_set_all

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param irow_global ...
!> \param icol_global ...
!> \param alpha ...
!> \par History
!>      , TCH, created
!>      always return the answer
! **************************************************************************************************
   SUBROUTINE cp_cfm_get_element(matrix, irow_global, icol_global, alpha)

      TYPE(cp_cfm_type), POINTER               :: matrix
      INTEGER, INTENT(IN)                      :: irow_global, icol_global
      COMPLEX(KIND=dp), INTENT(OUT)            :: alpha

      INTEGER                                  :: mypcol, myprow, npcol, nprow
      TYPE(cp_blacs_env_type), POINTER         :: context

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)

      alpha = matrix%local_data(irow_global, icol_global)

   END SUBROUTINE cp_cfm_get_element

! **************************************************************************************************
!> \brief gets a submatrix of a full matrix
!>       op(target_m)(1:n_rows,1:n_cols)
!>       =fm(start_row:start_row+n_rows,start_col:start_col+n_cols)
!>      target_m is replicated on all cpus
!>      using this call is expensive
!> \param fm        : the full you want to get the info from
!> \param target_m  : a replicated full matrix that will contain the result
!> \param start_row : the starting row of b_matrix (defaults to 1)
!> \param start_col : the starting col of b_matrix (defaults to 1)
!> \param n_rows    : the number of row to change in b (defaults to
!>                    size(op(new_values),1))
!> \param n_cols    : the number of columns to change in b (defaults to
!>                    size(op(new_values),2))
!> \param transpose : if target_m should be transposed: if true
!>                    op(target_m)=target_m^T, else op(target_m)=target_m
!>                    (defaults to false)
!> \par History
!>      2016/04/05 created borrowing from Fawzi's cp_fm_get_submatrix
!> \author Lianheng Tong
!> \note
!>      optimized for full column updates. Zeros out a little too much
!>      of target_m
!>      the target_m is replicated and valid on all cpus
! **************************************************************************************************
   SUBROUTINE cp_cfm_get_submatrix(fm, &
                                   target_m, &
                                   start_row, &
                                   start_col, &
                                   n_rows, &
                                   n_cols, &
                                   transpose)
      TYPE(cp_cfm_type), POINTER                         :: fm
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(out)     :: target_m
      INTEGER, INTENT(in), OPTIONAL                      :: start_row, start_col, n_rows, n_cols
      LOGICAL, INTENT(in), OPTIONAL                      :: transpose

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_get_submatrix', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: full_block
      INTEGER                                            :: handle, i, i0, j, j0, ncol, ncol_global, &
                                                            ncol_local, nrow, nrow_global, &
                                                            nrow_local, this_col, this_row
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: tr_a
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      i0 = 1
      j0 = 1
      tr_a = .FALSE.

      IF(.NOT.(ASSOCIATED(fm)))CALL cp__a("fm/cp_cfm_types.F",297)
      IF(.NOT.(fm%ref_count .GT. 0))CALL cp__a("fm/cp_cfm_types.F",298)

      IF (PRESENT(start_row)) i0 = start_row
      IF (PRESENT(start_col)) j0 = start_col
      IF (PRESENT(transpose)) tr_a = transpose
      IF (tr_a) THEN
         nrow = SIZE(target_m, 2)
         ncol = SIZE(target_m, 1)
      ELSE
         nrow = SIZE(target_m, 1)
         ncol = SIZE(target_m, 2)
      END IF
      IF (PRESENT(n_rows)) nrow = n_rows
      IF (PRESENT(n_cols)) ncol = n_cols

      para_env => fm%matrix_struct%para_env
      full_block => fm%local_data

      CALL cp_cfm_get_info(matrix=fm, &
                           nrow_global=nrow_global, ncol_global=ncol_global, &
                           nrow_local=nrow_local, ncol_local=ncol_local, &
                           row_indices=row_indices, col_indices=col_indices)
      DO j = 1, ncol_local
         this_col = col_indices(j)-j0+1
         IF (this_col .GE. 1 .AND. this_col .LE. ncol) THEN
            IF (tr_a) THEN
               IF (i0 .EQ. 1 .AND. nrow_global .EQ. nrow) THEN
                  DO i = 1, nrow_local
                     target_m(this_col, row_indices(i)) = full_block(i, j)
                  END DO
               ELSE
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row .GE. 1 .AND. this_row .LE. nrow) THEN
                        target_m(this_col, this_row) = full_block(i, j)
                     END IF
                  END DO
               END IF
            ELSE
               IF (i0 .EQ. 1 .AND. nrow_global .EQ. nrow) THEN
                  DO i = 1, nrow_local
                     target_m(row_indices(i), this_col) = full_block(i, j)
                  END DO
               ELSE
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row .GE. 1 .AND. this_row .LE. nrow) THEN
                        target_m(this_row, this_col) = full_block(i, j)
                     END IF
                  END DO
               END IF
            END IF
         END IF
      END DO

      CALL mp_sum(target_m, para_env%group)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_get_submatrix

! **************************************************************************************************
!> \brief   Set the FULL matrix element (irow_global,icol_global) to alpha.
!> \param matrix ...
!> \param irow_global ...
!> \param icol_global ...
!> \param alpha ...
!> \date    12.06.2001
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_cfm_set_element(matrix, irow_global, icol_global, alpha)

      TYPE(cp_cfm_type), POINTER               :: matrix
      INTEGER, INTENT(IN)                      :: irow_global, icol_global
      COMPLEX(KIND=dp), INTENT(IN)             :: alpha

      INTEGER                                  :: mypcol, myprow, npcol, nprow
      TYPE(cp_blacs_env_type), POINTER         :: context

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)


      matrix%local_data(irow_global, icol_global) = alpha

   END SUBROUTINE cp_cfm_set_element

! **************************************************************************************************
!> \brief sets a submatrix of a full matrix
!>       matrix(start_row:start_row+n_rows,start_col:start_col+n_cols)
!>       = alpha*op(new_values)(1:n_rows,1:n_cols)+ beta
!>       * matrix(start_row:start_row+n_rows,start_col:start_col+n_cols)
!> \param matrix the full to change
!> \param new_values a replicated full matrix with the new values
!> \param start_row the starting row of b_matrix (defaults to 1)
!> \param start_col the starting col of b_matrix (defaults to 1)
!> \param n_rows the number of row to change in b (defaults to
!>        size(op(new_values),1))
!> \param n_cols the number of columns to change in b (defaults to
!>        size(op(new_values),2))
!> \param alpha rescaling factor for the new values (defaults to 1.0)
!> \param beta rescaling factor for the old values (defaults to 0.0)
!> \param transpose if new_values should be transposed: if true
!>        op(new_values)=new_values^T, else op(new_values)=new_values
!>        (defaults to false)
!> \par History
!>      07.2002 created borrowing from Fawzi's cp_fm_set_submatrix
!> \author Lianheng Tong
!> \note
!>      optimized for full column updates and alpha=1.0, beta=0.0
!>      the new_values need to be valid on all cpus
! **************************************************************************************************
   SUBROUTINE cp_cfm_set_submatrix(matrix, new_values, start_row, &
                                   start_col, n_rows, n_cols, alpha, beta, transpose)
      TYPE(cp_cfm_type), POINTER                         :: matrix
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(in)      :: new_values
      INTEGER, INTENT(in), OPTIONAL                      :: start_row, start_col, n_rows, n_cols
      COMPLEX(KIND=dp), INTENT(in), OPTIONAL             :: alpha, beta
      LOGICAL, INTENT(in), OPTIONAL                      :: transpose

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_set_submatrix', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: al, be
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: full_block
      INTEGER                                            :: i, i0, j, j0, ncol, ncol_block, &
                                                            ncol_global, ncol_local, nrow, &
                                                            nrow_block, nrow_global, nrow_local, &
                                                            this_col, this_row
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: tr_a

      al = (1.0_dp, 0.0_dp)
      be = (0.0_dp, 0.0_dp)
      i0 = 1
      j0 = 1
      tr_a = .FALSE.
      ! can be called too many times, making it a bit useless
      ! CALL timeset(routineN//','//moduleN,handle)

      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_cfm_types.F",470)
      IF(.NOT.(matrix%ref_count > 0))CALL cp__a("fm/cp_cfm_types.F",471)

      IF (PRESENT(alpha)) al = alpha
      IF (PRESENT(beta)) be = beta
      IF (PRESENT(start_row)) i0 = start_row
      IF (PRESENT(start_col)) j0 = start_col
      IF (PRESENT(transpose)) tr_a = transpose
      IF (tr_a) THEN
         nrow = SIZE(new_values, 2)
         ncol = SIZE(new_values, 1)
      ELSE
         nrow = SIZE(new_values, 1)
         ncol = SIZE(new_values, 2)
      END IF
      IF (PRESENT(n_rows)) nrow = n_rows
      IF (PRESENT(n_cols)) ncol = n_cols

      full_block => matrix%local_data

      CALL cp_cfm_get_info(matrix=matrix, &
                           nrow_global=nrow_global, &
                           ncol_global=ncol_global, &
                           nrow_block=nrow_block, &
                           ncol_block=ncol_block, &
                           nrow_local=nrow_local, &
                           ncol_local=ncol_local, &
                           row_indices=row_indices, &
                           col_indices=col_indices)

      IF (al .EQ. (1.0_dp, 0.0_dp) .AND. be .EQ. (0.0_dp, 0.0_dp)) THEN
         DO j = 1, ncol_local
            this_col = col_indices(j)-j0+1
            IF (this_col .GE. 1 .AND. this_col .LE. ncol) THEN
               IF (tr_a) THEN
                  IF (i0 .EQ. 1 .AND. nrow_global .EQ. nrow) THEN
                     DO i = 1, nrow_local
                        full_block(i, j) = new_values(this_col, row_indices(i))
                     END DO
                  ELSE
                     DO i = 1, nrow_local
                        this_row = row_indices(i)-i0+1
                        IF (this_row .GE. 1 .AND. this_row .LE. nrow) THEN
                           full_block(i, j) = new_values(this_col, this_row)
                        END IF
                     END DO
                  END IF
               ELSE
                  IF (i0 .EQ. 1 .AND. nrow_global .EQ. nrow) THEN
                     DO i = 1, nrow_local
                        full_block(i, j) = new_values(row_indices(i), this_col)
                     END DO
                  ELSE
                     DO i = 1, nrow_local
                        this_row = row_indices(i)-i0+1
                        IF (this_row .GE. 1 .AND. this_row .LE. nrow) THEN
                           full_block(i, j) = new_values(this_row, this_col)
                        END IF
                     END DO
                  END IF
               END IF
            END IF
         END DO
      ELSE
         DO j = 1, ncol_local
            this_col = col_indices(j)-j0+1
            IF (this_col .GE. 1 .AND. this_col .LE. ncol) THEN
               IF (tr_a) THEN
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row .GE. 1 .AND. this_row .LE. nrow) THEN
                        full_block(i, j) = al*new_values(this_col, this_row)+ &
                                           be*full_block(i, j)
                     END IF
                  END DO
               ELSE
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row .GE. 1 .AND. this_row .LE. nrow) THEN
                        full_block(i, j) = al*new_values(this_row, this_col)+ &
                                           be*full_block(i, j)
                     END IF
                  END DO
               END IF
            END IF
         END DO
      END IF

      ! CALL timestop(handle)

   END SUBROUTINE cp_cfm_set_submatrix

! **************************************************************************************************
!> \brief   Return informations about the specified FULL matrix.
!>          the indices vectors are cached.
!>          indices is an array that maps the local indices to the global ones
!>          local data is a little hole to access the blacs block, needed in some rare cases
!>          for efficient use
!> \param matrix ...
!> \param name ...
!> \param nrow_global ...
!> \param ncol_global ...
!> \param nrow_block ...
!> \param ncol_block ...
!> \param nrow_local ...
!> \param ncol_local ...
!> \param row_indices ...
!> \param col_indices ...
!> \param local_data ...
!> \param context ...
!> \param matrix_struct ...
!> \param para_env ...
!> \date    12.06.2001
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE cp_cfm_get_info(matrix, name, nrow_global, ncol_global, &
                              nrow_block, ncol_block, nrow_local, ncol_local, &
                              row_indices, col_indices, local_data, context, &
                              matrix_struct, para_env)
      TYPE(cp_cfm_type), POINTER        :: matrix
      CHARACTER(LEN=*), OPTIONAL, INTENT(OUT) :: name
      INTEGER, OPTIONAL, INTENT(OUT)          :: ncol_block, ncol_global, &
                                                 nrow_block, nrow_global, &
                                                 nrow_local, ncol_local
      INTEGER, OPTIONAL, DIMENSION(:), POINTER   :: row_indices, col_indices
      TYPE(cp_para_env_type), POINTER, OPTIONAL :: para_env
      TYPE(cp_blacs_env_type), POINTER, OPTIONAL :: context
      TYPE(cp_fm_struct_type), POINTER, OPTIONAL :: matrix_struct
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER, OPTIONAL :: local_data

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_get_info', &
                                     routineP = moduleN//':'//routineN
      INTEGER i, nprow, npcol, myprow, mypcol
      TYPE(cp_blacs_env_type), POINTER :: ctxt

      IF (PRESENT(name)) name = matrix%name
      IF (PRESENT(nrow_global)) nrow_global = matrix%matrix_struct%nrow_global
      IF (PRESENT(ncol_global)) ncol_global = matrix%matrix_struct%ncol_global
      IF (PRESENT(nrow_block)) nrow_block = matrix%matrix_struct%nrow_block
      IF (PRESENT(ncol_block)) ncol_block = matrix%matrix_struct%ncol_block
      IF (PRESENT(context)) context => matrix%matrix_struct%context
      IF (PRESENT(matrix_struct)) matrix_struct => matrix%matrix_struct
      IF (PRESENT(local_data)) local_data => matrix%local_data ! not hiding things anymore :-(

      ctxt => matrix%matrix_struct%context
      myprow = ctxt%mepos(1)
      mypcol = ctxt%mepos(2)
      nprow = ctxt%num_pe(1)
      npcol = ctxt%num_pe(2)

      IF (PRESENT(nrow_local)) THEN
         nrow_local = matrix%matrix_struct%nrow_locals(myprow)
      ENDIF

      IF (PRESENT(ncol_local)) THEN
         ncol_local = matrix%matrix_struct%ncol_locals(mypcol)
      ENDIF

      IF (PRESENT(row_indices)) THEN
         row_indices => matrix%matrix_struct%row_indices
         IF (.NOT. ASSOCIATED(row_indices)) THEN
            ALLOCATE (matrix%matrix_struct%row_indices &
                      (matrix%matrix_struct%nrow_locals(myprow)))
            row_indices => matrix%matrix_struct%row_indices
            DO i = 1, SIZE(row_indices)
               row_indices(i) = i
            ENDDO
         END IF
      ENDIF

      IF (PRESENT(col_indices)) THEN
         col_indices => matrix%matrix_struct%col_indices
         IF (.NOT. ASSOCIATED(col_indices)) THEN
            ALLOCATE (matrix%matrix_struct%col_indices &
                      (matrix%matrix_struct%ncol_locals(mypcol)))
            col_indices => matrix%matrix_struct%col_indices
            DO i = 1, SIZE(col_indices)
               col_indices(i) = i
            ENDDO
         END IF
      ENDIF

      IF (PRESENT(para_env)) para_env => matrix%matrix_struct%para_env

   END SUBROUTINE cp_cfm_get_info

! **************************************************************************************************
!> \brief copy one identically sized matrix in the other
!> \param source ...
!> \param destination ...
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_cfm_to_cfm_matrix(source, destination)
      TYPE(cp_cfm_type), POINTER                         :: source, destination

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cfm_to_cfm_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: npcol, nprow

      nprow = source%matrix_struct%context%num_pe(1)
      npcol = source%matrix_struct%context%num_pe(2)

      IF (.NOT. cp2k_is_parallel .OR. &
          cp_fm_struct_equivalent(source%matrix_struct, &
                                  destination%matrix_struct)) THEN
         IF (SIZE(source%local_data, 1) /= SIZE(destination%local_data, 1) .OR. &
             SIZE(source%local_data, 2) /= SIZE(destination%local_data, 2)) &
            CALL cp__b("fm/cp_cfm_types.F",698,"internal local_data has different sizes")
         CALL dcopy(SIZE(source%local_data, 1)*SIZE(source%local_data, 2)*2, &
                    source%local_data(1, 1), 1, destination%local_data(1, 1), 1)
      ELSE
         IF (source%matrix_struct%nrow_global /= destination%matrix_struct%nrow_global) &
            CALL cp__b("fm/cp_cfm_types.F",703,"cannot copy between full matrixes of differen sizes")
         IF (source%matrix_struct%ncol_global /= destination%matrix_struct%ncol_global) &
            CALL cp__b("fm/cp_cfm_types.F",705,"cannot copy between full matrixes of differen sizes")
         CALL cp__b("fm/cp_cfm_types.F",712,"")
      END IF
   END SUBROUTINE cp_cfm_to_cfm_matrix

! **************************************************************************************************
!> \brief copy just a subset of columns of a cfm to a cfm
!> \param msource ...
!> \param mtarget ...
!> \param ncol ...
!> \param source_start ...
!> \param target_start ...
! **************************************************************************************************
   SUBROUTINE cp_cfm_to_cfm_columns(msource, mtarget, ncol, source_start, &
                                    target_start)

      TYPE(cp_cfm_type), POINTER               :: msource, mtarget
      INTEGER, INTENT(IN)                      :: ncol
      INTEGER, INTENT(IN), OPTIONAL            :: source_start, target_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_cfm_to_cfm_columns', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, n, ss, ts
      COMPLEX(KIND=dp), DIMENSION(:, :), &
         POINTER                                :: a, b

      CALL timeset(routineN, handle)

      ss = 1
      ts = 1

      IF (PRESENT(source_start)) ss = source_start
      IF (PRESENT(target_start)) ts = target_start

      n = msource%matrix_struct%nrow_global

      a => msource%local_data
      b => mtarget%local_data

      CALL zcopy(ncol*n, a(1, ss), 1, b(1, ts), 1)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_to_cfm_columns

! **************************************************************************************************
!> \brief copy just a triangular matrix
!> \param msource ...
!> \param mtarget ...
!> \param uplo : 'U' for upper triangular, 'L' for lower triangular
! **************************************************************************************************
   SUBROUTINE cp_cfm_to_cfm_triangular(msource, mtarget, uplo)
      TYPE(cp_cfm_type), POINTER               :: msource, mtarget
      CHARACTER(LEN=*), INTENT(IN)             :: uplo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_cfm_to_cfm_triangular', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ncol, nrow
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER  :: aa, bb

      CALL timeset(routineN, handle)

      nrow = msource%matrix_struct%nrow_global
      ncol = msource%matrix_struct%ncol_global

      aa => msource%local_data
      bb => mtarget%local_data

      CALL zlacpy(uplo, nrow, ncol, aa(1, 1), nrow, bb(1, 1), nrow)

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_to_cfm_triangular

! **************************************************************************************************
!> \brief copy real and imaginary parts of a cfm to fm
!>        works only for absolutely equivalent structures
!> \param msource  : complex matrix
!> \param mtargetr : optional, if present outputs real part of msource
!> \param mtargeti : optional, if present outputs imag part of msource
! **************************************************************************************************
   SUBROUTINE cp_cfm_to_fm(msource, mtargetr, mtargeti)

      TYPE(cp_cfm_type), POINTER                         :: msource
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: mtargetr, mtargeti

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_cfm_to_fm', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: zmat
      INTEGER                                            :: handle
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: imat, rmat

      CALL timeset(routineN, handle)

      zmat => msource%local_data
      IF (PRESENT(mtargetr)) THEN
         rmat => mtargetr%local_data
         IF ((.NOT. cp_fm_struct_equivalent(mtargetr%matrix_struct, msource%matrix_struct)) .OR. &
             (SIZE(rmat, 1) .NE. SIZE(zmat, 1)) .OR. &
             (SIZE(rmat, 2) .NE. SIZE(zmat, 2))) THEN
            CALL cp__b("fm/cp_cfm_types.F",835,"size of local_data of mtargetr differ to msource")
         END IF
         ! copy local data
         rmat = REAL(zmat, KIND=dp)
      ELSE
         NULLIFY (rmat)
      END IF
      IF (PRESENT(mtargeti)) THEN
         imat => mtargeti%local_data
         IF ((.NOT. cp_fm_struct_equivalent(mtargeti%matrix_struct, msource%matrix_struct)) .OR. &
             (SIZE(imat, 1) .NE. SIZE(zmat, 1)) .OR. &
             (SIZE(imat, 2) .NE. SIZE(zmat, 2))) THEN
            CALL cp__b("fm/cp_cfm_types.F",847,"size of local_data of mtargeti differ to msource")
         END IF
         ! copy local data
         imat = REAL(AIMAG(zmat), KIND=dp)
      ELSE
         NULLIFY (imat)
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_cfm_to_fm

! **************************************************************************************************
!> \brief copy real and imaginary in forms of fms to a cfm
!>        works only for absolutely equivalent structures
!> \param msourcer : optional, if present input for real part.
!>                   if not present, treat as zero
!> \param msourcei : optional, if present input for imag part.
!>                   if not present, treat as zero
!> \param mtarget  : outputs complex matrix
! **************************************************************************************************
   SUBROUTINE cp_fm_to_cfm(msourcer, msourcei, mtarget)
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: msourcer, msourcei
      TYPE(cp_cfm_type), POINTER                         :: mtarget

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_to_cfm', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: zmat
      INTEGER                                            :: handle, mode
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: imat, rmat

      CALL timeset(routineN, handle)

      mode = 0
      zmat => mtarget%local_data
      IF (PRESENT(msourcer)) THEN
         rmat => msourcer%local_data
         IF ((.NOT. cp_fm_struct_equivalent(msourcer%matrix_struct, mtarget%matrix_struct)) .OR. &
             (SIZE(rmat, 1) .NE. SIZE(zmat, 1)) .OR. &
             (SIZE(rmat, 2) .NE. SIZE(zmat, 2))) THEN
            CALL cp__b("fm/cp_cfm_types.F",887,"size of local_data of msourcer differ to mtarget")
         END IF
         mode = mode+1
      ELSE
         NULLIFY (rmat)
      END IF
      IF (PRESENT(msourcei)) THEN
         imat => msourcei%local_data
         IF ((.NOT. cp_fm_struct_equivalent(msourcei%matrix_struct, mtarget%matrix_struct)) .OR. &
             (SIZE(imat, 1) .NE. SIZE(zmat, 1)) .OR. &
             (SIZE(imat, 2) .NE. SIZE(zmat, 2))) THEN
            CALL cp__b("fm/cp_cfm_types.F",898,"size of local_data of msourcei differ to mtarget")
         END IF
         mode = mode+2
      ELSE
         NULLIFY (imat)
      END IF
      ! copy local data
      SELECT CASE (mode)
      CASE (0)
         zmat(:, :) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      CASE (1)
         zmat(:, :) = CMPLX(rmat(:, :), 0.0_dp, KIND=dp)
      CASE (2)
         zmat(:, :) = CMPLX(0.0_dp, imat(:, :), KIND=dp)
      CASE (3)
         zmat = CMPLX(rmat, imat, KIND=dp)
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE cp_fm_to_cfm

END MODULE cp_cfm_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief basic linear algebra operations for full matrixes
!> \par History
!>      08.2002 splitted out of qs_blacs [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_fm_basic_linalg
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_fm_struct,                    ONLY: cp_fm_struct_equivalent
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_element, cp_fm_get_info, cp_fm_get_submatrix, cp_fm_release, &
        cp_fm_set_all, cp_fm_set_submatrix, cp_fm_to_fm, cp_fm_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp,&
                                              sp
   USE machine,                         ONLY: m_memory
   USE mathlib,                         ONLY: invert_matrix
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_fm_basic_linalg'

   PUBLIC :: cp_fm_scale, & ! scale a matrix
             cp_fm_scale_and_add, & ! scale and add two matrices
             cp_fm_geadd, & ! similar to cp_fm_scale_and_add, but allows transpose
             cp_fm_column_scale, & ! scale colummns of a matrix
             cp_fm_trace, & ! trace of the transpose(A)*B
             cp_fm_schur_product, & ! schur product
             cp_fm_transpose, & ! transpose a matrix
             cp_fm_upper_to_full, & ! symmetrise an upper symmetric matrix
             cp_fm_syrk, & ! rank k update
             cp_fm_triangular_multiply, & ! triangular matrix multiply / solve
             cp_fm_symm, & ! multiply a symmetric with a non-symmetric matrix
             cp_fm_gemm, & ! multiply two matrices
             cp_complex_fm_gemm, & ! multiply two complex matrices, represented by non_complex fm matrices
             cp_fm_invert, & ! computes the inverse and determinant
             cp_fm_frobenius_norm, & ! frobenius norm
             cp_fm_triangular_invert, & ! compute the reciprocal of a tirangular matrix
             cp_fm_qr_factorization, & ! compute the QR factorization of a rectangular matrix
             cp_fm_solve, & ! solves the equation  A*B=C A and C are input
             cp_fm_lu_invert, & ! inverts a general matrix using LU factorisation
             cp_fm_norm, & ! calculates norm of a matrix
             cp_fm_latra ! calculates trace of a square matrix using (p)dlatra

   REAL(KIND=dp), EXTERNAL :: dlange, pdlange, pdlatra
   REAL(KIND=sp), EXTERNAL :: slange, pslange, pslatra

CONTAINS

! **************************************************************************************************
!> \brief calc A <- alpha*A + beta*B
!>      optimized for alpha == 1.0 (just add beta*B) and beta == 0.0 (just
!>      scale A)
!> \param alpha ...
!> \param matrix_a ...
!> \param beta ...
!> \param matrix_b ...
! **************************************************************************************************
   SUBROUTINE cp_fm_scale_and_add(alpha, matrix_a, beta, matrix_b)

      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(cp_fm_type), POINTER                          :: matrix_a
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: matrix_b

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_scale_and_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, size_a
      REAL(KIND=dp)                                      :: my_beta
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a, b
      REAL(KIND=sp), DIMENSION(:, :), POINTER            :: a_sp, b_sp

      CALL timeset(routineN, handle)

      IF (PRESENT(matrix_b)) THEN
         my_beta = 1.0_dp
      ELSE
         my_beta = 0.0_dp
      ENDIF
      IF (PRESENT(beta)) my_beta = beta
      NULLIFY (a, b)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_fm_basic_linalg.F",94)
      IF(.NOT.(matrix_a%ref_count > 0))CALL cp__a("fm/cp_fm_basic_linalg.F",95)

      IF (PRESENT(beta)) THEN
         IF(.NOT.(PRESENT(matrix_b)))CALL cp__a("fm/cp_fm_basic_linalg.F",98)
         IF(.NOT.(ASSOCIATED(matrix_b)))CALL cp__a("fm/cp_fm_basic_linalg.F",99)
         IF(.NOT.(matrix_b%ref_count > 0))CALL cp__a("fm/cp_fm_basic_linalg.F",100)
         IF (matrix_a%id_nr == matrix_b%id_nr) THEN
            IF (matrix_a%id_nr == matrix_b%id_nr) &
               CALL cp__w("fm/cp_fm_basic_linalg.F",103,"Bad use of routine. Call cp_fm_scale instead")
            CALL cp_fm_scale(alpha+beta, matrix_a)
            CALL timestop(handle)
            RETURN
         END IF
      END IF

      a => matrix_a%local_data
      a_sp => matrix_a%local_data_sp

      IF (matrix_a%use_sp) THEN
         size_a = SIZE(a_sp, 1)*SIZE(a_sp, 2)
      ELSE
         size_a = SIZE(a, 1)*SIZE(a, 2)
      ENDIF

      IF (alpha /= 1.0_dp) THEN
         IF (matrix_a%use_sp) THEN
            CALL sscal(size_a, REAL(alpha, sp), a_sp, 1)
         ELSE
            CALL dscal(size_a, alpha, a, 1)
         ENDIF
      ENDIF
      IF (my_beta .NE. 0.0_dp) THEN
         IF (matrix_a%matrix_struct%context%group /= matrix_b%matrix_struct%context%group) &
            CALL cp__b("fm/cp_fm_basic_linalg.F",128,"matrixes must be in the same blacs context")

         IF (cp_fm_struct_equivalent(matrix_a%matrix_struct, &
                                     matrix_b%matrix_struct)) THEN

            b => matrix_b%local_data
            b_sp => matrix_b%local_data_sp

            IF (matrix_a%use_sp .AND. matrix_b%use_sp) THEN
               CALL saxpy(size_a, REAL(my_beta, sp), b_sp, 1, a_sp, 1)
            ELSEIF (matrix_a%use_sp .AND. .NOT. matrix_b%use_sp) THEN
               CALL saxpy(size_a, REAL(my_beta, sp), REAL(b, sp), 1, a_sp, 1)
            ELSEIF (.NOT. matrix_a%use_sp .AND. matrix_b%use_sp) THEN
               CALL daxpy(size_a, my_beta, REAL(b_sp, dp), 1, a, 1)
            ELSE
               CALL daxpy(size_a, my_beta, b, 1, a, 1)
            ENDIF

         ELSE
            CALL cp__b("fm/cp_fm_basic_linalg.F",150,"")
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_scale_and_add

! **************************************************************************************************
!> \brief interface to BLACS geadd:
!>                matrix_b = beta*matrix_b + alpha*opt(matrix_a)
!>        where opt(matrix_a) can be either:
!>              'N':  matrix_a
!>              'T':  matrix_a^T
!>              'C':  matrix_a^H (Hermitian conjugate)
!>        note that this is a level three routine, use cp_fm_scale_and_add if that
!>        is sufficient for your needs
!> \param alpha  : complex scalar
!> \param trans  : 'N' normal, 'T' transposed
!> \param matrix_a : input matrix_a
!> \param beta   : complex scalar
!> \param matrix_b : input matrix_b, upon out put the updated matrix_b
!> \author  Lianheng Tong
! **************************************************************************************************
   SUBROUTINE cp_fm_geadd(alpha, trans, matrix_a, beta, matrix_b)
      REAL(KIND=dp), INTENT(IN) :: alpha, beta
      CHARACTER, INTENT(IN) :: trans
      TYPE(cp_fm_type), POINTER :: matrix_a, matrix_b

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_geadd', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrow_global, ncol_global, handle
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: aa, bb
      INTEGER :: ii, jj

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_fm_basic_linalg.F",194)
      IF(.NOT.(ASSOCIATED(matrix_b)))CALL cp__a("fm/cp_fm_basic_linalg.F",195)
      nrow_global = matrix_a%matrix_struct%nrow_global
      ncol_global = matrix_a%matrix_struct%ncol_global
      IF(.NOT.(nrow_global .EQ. matrix_b%matrix_struct%nrow_global))CALL cp__a("fm/cp_fm_basic_linalg.F",198)
      IF(.NOT.(ncol_global .EQ. matrix_b%matrix_struct%ncol_global))CALL cp__a("fm/cp_fm_basic_linalg.F",199)

      aa => matrix_a%local_data
      bb => matrix_b%local_data

      ! dgeadd is not a standard BLAS function, although is implemented
      ! in some libraries like OpenBLAS, so not going to use it here
      SELECT CASE (trans)
      CASE ('T')
         DO jj = 1, ncol_global
            DO ii = 1, nrow_global
               bb(ii, jj) = beta*bb(ii, jj)+alpha*aa(jj, ii)
            END DO
         END DO
      CASE DEFAULT
         DO jj = 1, ncol_global
            DO ii = 1, nrow_global
               bb(ii, jj) = beta*bb(ii, jj)+alpha*aa(ii, jj)
            END DO
         END DO
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE cp_fm_geadd

! **************************************************************************************************
!> \brief Computes the LU-decomposition of the matrix, and the determinant of the matrix
!>      IMPORTANT : the sign of the determinant is not defined correctly yet ....
!> \param matrix_a ...
!> \param almost_determinant ...
!> \param correct_sign ...
!> \par History
!>      added correct_sign 02.07 (fschiff)
!> \author Joost VandeVondele
!> \note
!>      - matrix_a is overwritten
!>      - the sign of the determinant might be wrong
!>      - SERIOUS WARNING (KNOWN BUG) : the sign of the determinant depends on ipivot
!>      - one should be able to find out if ipivot is an even or an odd permutation...
!>        if you need the correct sign, just add correct_sign==.TRUE. (fschiff)
! **************************************************************************************************
   SUBROUTINE cp_fm_lu_decompose(matrix_a, almost_determinant, correct_sign)
      TYPE(cp_fm_type), POINTER                :: matrix_a
      REAL(KIND=dp), INTENT(OUT)               :: almost_determinant
      LOGICAL, INTENT(IN), OPTIONAL            :: correct_sign

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_lu_decompose', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, i, info, n
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: ipivot
      REAL(KIND=dp)                            :: determinant
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      INTEGER                                  :: lda

      CALL timeset(routineN, handle)

      a => matrix_a%local_data
      n = matrix_a%matrix_struct%nrow_global
      ALLOCATE (ipivot(n+matrix_a%matrix_struct%nrow_block))

      lda = SIZE(a, 1)
      CALL dgetrf(n, n, a(1, 1), lda, ipivot, info)
      determinant = 1.0_dp
      IF (correct_sign) THEN
         DO i = 1, n
            IF (ipivot(i) .NE. i) THEN
               determinant = -determinant*a(i, i)
            ELSE
               determinant = determinant*a(i, i)
            END IF
         END DO
      ELSE
         DO i = 1, n
            determinant = determinant*a(i, i)
         ENDDO
      END IF
      ! info is allowed to be zero
      ! this does just signal a zero diagonal element
      DEALLOCATE (ipivot)
      almost_determinant = determinant ! notice that the sign is random
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief computes matrix_c = beta * matrix_c + alpha * ( matrix_a  ** transa ) * ( matrix_b ** transb )
!> \param transa : 'N' -> normal   'T' -> transpose
!>      alpha,beta :: can be 0.0_dp and 1.0_dp
!> \param transb ...
!> \param m ...
!> \param n ...
!> \param k ...
!> \param alpha ...
!> \param matrix_a : m x k matrix ( ! for transa = 'N')
!> \param matrix_b : k x n matrix ( ! for transb = 'N')
!> \param beta ...
!> \param matrix_c : m x n matrix
!> \param a_first_col ...
!> \param a_first_row ...
!> \param b_first_col : the k x n matrix starts at col b_first_col of matrix_b (avoid usage)
!> \param b_first_row ...
!> \param c_first_col ...
!> \param c_first_row ...
!> \author Matthias Krack
!> \note
!>      matrix_c should have no overlap with matrix_a, matrix_b
! **************************************************************************************************
   SUBROUTINE cp_fm_gemm(transa, transb, m, n, k, alpha, matrix_a, matrix_b, beta, &
                         matrix_c, a_first_col, a_first_row, b_first_col, b_first_row, &
                         c_first_col, c_first_row)

      CHARACTER(LEN=1), INTENT(IN)             :: transa, transb
      INTEGER, INTENT(IN)                      :: m, n, k
      REAL(KIND=dp), INTENT(IN)                :: alpha
      TYPE(cp_fm_type), POINTER                :: matrix_a, matrix_b
      REAL(KIND=dp), INTENT(IN)                :: beta
      TYPE(cp_fm_type), POINTER                :: matrix_c
      INTEGER, INTENT(IN), OPTIONAL            :: a_first_col, a_first_row, &
                                                  b_first_col, b_first_row, &
                                                  c_first_col, c_first_row

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_gemm', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, i_a, i_b, i_c, j_a, &
                                                  j_b, j_c
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, b, c
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: a_sp, b_sp, c_sp
      INTEGER                                  :: lda, ldb, ldc

      CALL timeset(routineN, handle)

      !sample peak memory
      CALL m_memory()

      a => matrix_a%local_data
      b => matrix_b%local_data
      c => matrix_c%local_data

      a_sp => matrix_a%local_data_sp
      b_sp => matrix_b%local_data_sp
      c_sp => matrix_c%local_data_sp

      IF (PRESENT(a_first_row)) THEN
         i_a = a_first_row
      ELSE
         i_a = 1
      END IF
      IF (PRESENT(a_first_col)) THEN
         j_a = a_first_col
      ELSE
         j_a = 1
      END IF
      IF (PRESENT(b_first_row)) THEN
         i_b = b_first_row
      ELSE
         i_b = 1
      END IF
      IF (PRESENT(b_first_col)) THEN
         j_b = b_first_col
      ELSE
         j_b = 1
      END IF
      IF (PRESENT(c_first_row)) THEN
         i_c = c_first_row
      ELSE
         i_c = 1
      END IF
      IF (PRESENT(c_first_col)) THEN
         j_c = c_first_col
      ELSE
         j_c = 1
      END IF


      IF (matrix_a%use_sp .AND. matrix_b%use_sp .AND. matrix_c%use_sp) THEN

         lda = SIZE(a_sp, 1)
         ldb = SIZE(b_sp, 1)
         ldc = SIZE(c_sp, 1)

         CALL sgemm(transa, transb, m, n, k, REAL(alpha, sp), a_sp(i_a, j_a), lda, b_sp(i_b, j_b), ldb, &
                    REAL(beta, sp), c_sp(i_c, j_c), ldc)

      ELSEIF ((.NOT. matrix_a%use_sp) .AND. (.NOT. matrix_b%use_sp) .AND. (.NOT. matrix_c%use_sp)) THEN

         lda = SIZE(a, 1)
         ldb = SIZE(b, 1)
         ldc = SIZE(c, 1)

         CALL dgemm(transa, transb, m, n, k, alpha, a(i_a, j_a), lda, b(i_b, j_b), ldb, beta, c(i_c, j_c), ldc)

      ELSE
         CALL cp__b("fm/cp_fm_basic_linalg.F",455,"Mixed precision gemm NYI")
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_gemm

! **************************************************************************************************
!> \brief computes matrix_c = beta * matrix_c + alpha *  matrix_a  *  matrix_b
!>      computes matrix_c = beta * matrix_c + alpha *  matrix_b  *  matrix_a
!>      where matrix_a is symmetric
!> \param side : 'L' -> matrix_a is on the left 'R' -> matrix_a is on the right
!>      alpha,beta :: can be 0.0_dp and 1.0_dp
!> \param uplo ...
!> \param m ...
!> \param n ...
!> \param alpha ...
!> \param matrix_a : m x m matrix
!> \param matrix_b : m x n matrix
!> \param beta ...
!> \param matrix_c : m x n matrix
!> \author Matthias Krack
!> \note
!>      matrix_c should have no overlap with matrix_a, matrix_b
!>      all matrices in QS are upper triangular, so uplo should be 'U' always
!>      matrix_a is always an m x m matrix
!>      it is typically slower to do cp_fm_symm than cp_fm_gemm (especially in parallel easily 50 percent !)
! **************************************************************************************************
   SUBROUTINE cp_fm_symm(side, uplo, m, n, alpha, matrix_a, matrix_b, beta, matrix_c)

      CHARACTER(LEN=1), INTENT(IN)             :: side, uplo
      INTEGER, INTENT(IN)                      :: m, n
      REAL(KIND=dp), INTENT(IN)                :: alpha
      TYPE(cp_fm_type), POINTER                :: matrix_a, matrix_b
      REAL(KIND=dp), INTENT(IN)                :: beta
      TYPE(cp_fm_type), POINTER                :: matrix_c

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_symm', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, b, c
      INTEGER                                  :: lda, ldb, ldc

      CALL timeset(routineN, handle)

      a => matrix_a%local_data
      b => matrix_b%local_data
      c => matrix_c%local_data


      lda = matrix_a%matrix_struct%local_leading_dimension
      ldb = matrix_b%matrix_struct%local_leading_dimension
      ldc = matrix_c%matrix_struct%local_leading_dimension

      CALL dsymm(side, uplo, m, n, alpha, a(1, 1), lda, b(1, 1), ldb, beta, c(1, 1), ldc)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_symm

! **************************************************************************************************
!> \brief computes the Frobenius norm of matrix_a
!> \param matrix_a : m x n matrix
!> \param norm ...
!> \author VW
! **************************************************************************************************
   SUBROUTINE cp_fm_frobenius_norm(matrix_a, norm)
      TYPE(cp_fm_type), POINTER                :: matrix_a
      REAL(KIND=dp), INTENT(inout)             :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_frobenius_norm', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, size_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      REAL(KIND=dp), EXTERNAL                  :: DDOT

      CALL timeset(routineN, handle)

      norm = 0.0_dp
      a => matrix_a%local_data
      size_a = SIZE(a, 1)*SIZE(a, 2)
      norm = DDOT(size_a, a(1, 1), 1, a(1, 1), 1)
      norm = SQRT(norm)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_frobenius_norm

! **************************************************************************************************
!> \brief performs a rank-k update of a symmetric matrix_c
!>         matrix_c = beta * matrix_c + alpha * matrix_a * transpose ( matrix_a )
!> \param uplo : 'U'   ('L')
!> \param trans : 'N'  ('T')
!> \param k : number of cols to use in matrix_a
!>      ia,ja ::  1,1 (could be used for selecting subblock of a)
!> \param alpha ...
!> \param matrix_a ...
!> \param ia ...
!> \param ja ...
!> \param beta ...
!> \param matrix_c ...
!> \author Matthias Krack
!> \note
!>      In QS uplo should 'U' (upper part updated)
! **************************************************************************************************
   SUBROUTINE cp_fm_syrk(uplo, trans, k, alpha, matrix_a, ia, ja, beta, matrix_c)
      CHARACTER(LEN=1), INTENT(IN)             :: uplo, trans
      INTEGER, INTENT(IN)                      :: k
      REAL(KIND=dp), INTENT(IN)                :: alpha
      TYPE(cp_fm_type), POINTER                :: matrix_a
      INTEGER, INTENT(IN)                      :: ia, ja
      REAL(KIND=dp), INTENT(IN)                :: beta
      TYPE(cp_fm_type), POINTER                :: matrix_c

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_syrk', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, n
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, c
      INTEGER                                  :: lda, ldc

      CALL timeset(routineN, handle)

      n = matrix_c%matrix_struct%nrow_global

      a => matrix_a%local_data
      c => matrix_c%local_data


      lda = SIZE(a, 1)
      ldc = SIZE(c, 1)

      CALL dsyrk(uplo, trans, n, k, alpha, a(ia, ja), lda, beta, c(1, 1), ldc)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_syrk

! **************************************************************************************************
!> \brief computes the schur product of two matrices
!>       c_ij = a_ij * b_ij
!> \param matrix_a ...
!> \param matrix_b ...
!> \param matrix_c ...
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_fm_schur_product(matrix_a, matrix_b, matrix_c)

      TYPE(cp_fm_type), POINTER                          :: matrix_a, matrix_b, matrix_c

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_schur_product', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icol_local, irow_local, mypcol, &
                                                            myprow, ncol_local, npcol, nprow, &
                                                            nrow_local
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a, b, c
      TYPE(cp_blacs_env_type), POINTER                   :: context

      CALL timeset(routineN, handle)

      context => matrix_a%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)

      a => matrix_a%local_data
      b => matrix_b%local_data
      c => matrix_c%local_data

      nrow_local = matrix_a%matrix_struct%nrow_locals(myprow)
      ncol_local = matrix_a%matrix_struct%ncol_locals(mypcol)

      DO icol_local = 1, ncol_local
         DO irow_local = 1, nrow_local
            c(irow_local, icol_local) = a(irow_local, icol_local)*b(irow_local, icol_local)
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE cp_fm_schur_product

! **************************************************************************************************
!> \brief returns the trace of matrix_a^T matrix_b, i.e
!>      sum_{i,j}(matrix_a(i,j)*matrix_b(i,j))
!> \param matrix_a a matrix
!> \param matrix_b another matrix
!> \param trace ...
!> \par History
!>      11.06.2001 Creation (Matthias Krack)
!>      12.2002 added doc [fawzi]
!> \author Matthias Krack
!> \note
!>      note the transposition of matrix_a!
! **************************************************************************************************
   SUBROUTINE cp_fm_trace(matrix_a, matrix_b, trace)

      TYPE(cp_fm_type), POINTER                          :: matrix_a, matrix_b
      REAL(KIND=dp), INTENT(OUT)                         :: trace

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_trace', routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, mypcol, myprow, &
                                                            ncol_local, npcol, nprow, nrow_local
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a, b
      REAL(KIND=sp), DIMENSION(:, :), POINTER            :: a_sp, b_sp
      TYPE(cp_blacs_env_type), POINTER                   :: context

      CALL timeset(routineN, handle)

      context => matrix_a%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)

      group = matrix_a%matrix_struct%para_env%group

      a => matrix_a%local_data
      b => matrix_b%local_data

      a_sp => matrix_a%local_data_sp
      b_sp => matrix_b%local_data_sp

      nrow_local = MIN(matrix_a%matrix_struct%nrow_locals(myprow), matrix_b%matrix_struct%nrow_locals(myprow))
      ncol_local = MIN(matrix_a%matrix_struct%ncol_locals(mypcol), matrix_b%matrix_struct%ncol_locals(mypcol))

      ! cries for an accurate_dot_product
      IF (matrix_a%use_sp .AND. matrix_b%use_sp) THEN
         trace = accurate_sum(REAL(a_sp(1:nrow_local, 1:ncol_local)* &
                                   b_sp(1:nrow_local, 1:ncol_local), dp))
      ELSEIF (matrix_a%use_sp .AND. .NOT. matrix_b%use_sp) THEN
         trace = accurate_sum(REAL(a_sp(1:nrow_local, 1:ncol_local), dp)* &
                              b(1:nrow_local, 1:ncol_local))
      ELSEIF (.NOT. matrix_a%use_sp .AND. matrix_b%use_sp) THEN
         trace = accurate_sum(a(1:nrow_local, 1:ncol_local)* &
                              REAL(b_sp(1:nrow_local, 1:ncol_local), dp))
      ELSE
         trace = accurate_sum(a(1:nrow_local, 1:ncol_local)* &
                              b(1:nrow_local, 1:ncol_local))
      ENDIF

      CALL mp_sum(trace, group)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_trace

! **************************************************************************************************
!> \brief multiplies in place by a triangular matrix:
!>       matrix_b = alpha op(triangular_matrix) matrix_b
!>      or (if side='R')
!>       matrix_b = alpha matrix_b op(triangular_matrix)
!>      op(triangular_matrix) is:
!>       triangular_matrix (if transpose_tr=.false. and invert_tr=.false.)
!>       triangular_matrix^T (if transpose_tr=.true. and invert_tr=.false.)
!>       triangular_matrix^(-1) (if transpose_tr=.false. and invert_tr=.true.)
!>       triangular_matrix^(-T) (if transpose_tr=.true. and invert_tr=.true.)
!> \param triangular_matrix the triangular matrix that multiplies the other
!> \param matrix_b the matrix that gets multiplied and stores the result
!> \param side on which side of matrix_b stays op(triangular_matrix)
!>        (defaults to 'L')
!> \param transpose_tr if the triangular matrix should be transposed
!>        (defaults to false)
!> \param invert_tr if the triangular matrix should be inverted
!>        (defaults to false)
!> \param uplo_tr if triangular_matrix is stored in the upper ('U') or
!>        lower ('L') triangle (defaults to 'U')
!> \param unit_diag_tr if the diagonal elements of triangular_matrix should
!>        be assumed to be 1 (defaults to false)
!> \param n_rows the number of rows of the result (defaults to
!>        size(matrix_b,1))
!> \param n_cols the number of columns of the result (defaults to
!>        size(matrix_b,2))
!> \param alpha ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      needs an mpi env
! **************************************************************************************************
   SUBROUTINE cp_fm_triangular_multiply(triangular_matrix, matrix_b, side, &
                                        transpose_tr, invert_tr, uplo_tr, unit_diag_tr, n_rows, n_cols, &
                                        alpha)
      TYPE(cp_fm_type), POINTER                          :: triangular_matrix, matrix_b
      CHARACTER, INTENT(in), OPTIONAL                    :: side
      LOGICAL, INTENT(in), OPTIONAL                      :: transpose_tr, invert_tr
      CHARACTER, INTENT(in), OPTIONAL                    :: uplo_tr
      LOGICAL, INTENT(in), OPTIONAL                      :: unit_diag_tr
      INTEGER, INTENT(in), OPTIONAL                      :: n_rows, n_cols
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_triangular_multiply', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: side_char, transa, unit_diag, uplo
      INTEGER                                            :: handle, m, n
      LOGICAL                                            :: invert
      REAL(KIND=dp)                                      :: al

      CALL timeset(routineN, handle)
      side_char = 'L'
      unit_diag = 'N'
      uplo = 'U'
      transa = 'N'
      invert = .FALSE.
      al = 1.0_dp
      CALL cp_fm_get_info(matrix_b, nrow_global=m, ncol_global=n)
      IF (PRESENT(side)) side_char = side
      IF (PRESENT(invert_tr)) invert = invert_tr
      IF (PRESENT(uplo_tr)) uplo = uplo_tr
      IF (PRESENT(unit_diag_tr)) THEN
         IF (unit_diag_tr) THEN
            unit_diag = 'U'
         ELSE
            unit_diag = 'N'
         END IF
      END IF
      IF (PRESENT(transpose_tr)) THEN
         IF (transpose_tr) THEN
            transa = 'T'
         ELSE
            transa = 'N'
         END IF
      END IF
      IF (PRESENT(alpha)) al = alpha
      IF (PRESENT(n_rows)) m = n_rows
      IF (PRESENT(n_cols)) n = n_cols

      IF (invert) THEN

         CALL dtrsm(side_char, uplo, transa, unit_diag, m, n, al, &
                    triangular_matrix%local_data(1, 1), &
                    SIZE(triangular_matrix%local_data, 1), &
                    matrix_b%local_data(1, 1), SIZE(matrix_b%local_data, 1))

      ELSE

         CALL dtrmm(side_char, uplo, transa, unit_diag, m, n, al, &
                    triangular_matrix%local_data(1, 1), &
                    SIZE(triangular_matrix%local_data, 1), &
                    matrix_b%local_data(1, 1), SIZE(matrix_b%local_data, 1))

      END IF

      CALL timestop(handle)
   END SUBROUTINE cp_fm_triangular_multiply

! **************************************************************************************************
!> \brief scales a matrix
!>      matrix_a = alpha * matrix_b
!> \param alpha ...
!> \param matrix_a ...
!> \note
!>      use cp_fm_set_all to zero (avoids problems with nan)
! **************************************************************************************************
   SUBROUTINE cp_fm_scale(alpha, matrix_a)
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(cp_fm_type), POINTER                          :: matrix_a

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_scale', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, size_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a

      CALL timeset(routineN, handle)

      NULLIFY (a)

      IF(.NOT.(ASSOCIATED(matrix_a)))CALL cp__a("fm/cp_fm_basic_linalg.F",879)
      IF(.NOT.(matrix_a%ref_count > 0))CALL cp__a("fm/cp_fm_basic_linalg.F",880)

      a => matrix_a%local_data
      size_a = SIZE(a, 1)*SIZE(a, 2)

      CALL DSCAL(size_a, alpha, a, 1)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_scale

! **************************************************************************************************
!> \brief transposes a matrix
!>      matrixt = matrix ^ T
!> \param matrix ...
!> \param matrixt ...
!> \note
!>      all matrix elements are transpose (see cp_fm_upper_to_half to symmetrise a matrix)
!>      all matrix elements are transpose (see cp_fm_upper_to_half to symmetrize a matrix)
! **************************************************************************************************
   SUBROUTINE cp_fm_transpose(matrix, matrixt)
      TYPE(cp_fm_type), POINTER                :: matrix, matrixt

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_transpose', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ncol_global, &
                                                  nrow_global
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, c
      INTEGER                                  :: i, j

      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_fm_basic_linalg.F",915)
      IF(.NOT.(ASSOCIATED(matrixt)))CALL cp__a("fm/cp_fm_basic_linalg.F",916)
      nrow_global = matrix%matrix_struct%nrow_global
      ncol_global = matrix%matrix_struct%ncol_global
      IF(.NOT.(nrow_global == ncol_global))CALL cp__a("fm/cp_fm_basic_linalg.F",919)

      CALL timeset(routineN, handle)

      a => matrix%local_data
      c => matrixt%local_data

      DO j = 1, ncol_global
         DO i = 1, nrow_global
            c(j, i) = a(i, j)
         ENDDO
      ENDDO
      CALL timestop(handle)

   END SUBROUTINE cp_fm_transpose

! **************************************************************************************************
!> \brief given an upper triangular matrix computes the corresponding full matrix
!> \param matrix the upper triangular matrix as input, the full matrix as output
!> \param work a matrix of the same size as matrix
!> \author Matthias Krack
!> \note
!>       the lower triangular part is irrelevant
! **************************************************************************************************
   SUBROUTINE cp_fm_upper_to_full(matrix, work)

      TYPE(cp_fm_type), POINTER                :: matrix, work

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_upper_to_full', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, icol_global, irow_global, &
                                                  mypcol, myprow, ncol_global, &
                                                  npcol, nprow, nrow_global
      REAL(KIND=dp), DIMENSION(:, :), POINTER   :: a
      REAL(KIND=sp), DIMENSION(:, :), POINTER   :: a_sp
      TYPE(cp_blacs_env_type), POINTER         :: context


      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_fm_basic_linalg.F",973)
      IF(.NOT.(ASSOCIATED(work)))CALL cp__a("fm/cp_fm_basic_linalg.F",974)
      nrow_global = matrix%matrix_struct%nrow_global
      ncol_global = matrix%matrix_struct%ncol_global
      IF(.NOT.(nrow_global == ncol_global))CALL cp__a("fm/cp_fm_basic_linalg.F",977)
      nrow_global = work%matrix_struct%nrow_global
      ncol_global = work%matrix_struct%ncol_global
      IF(.NOT.(nrow_global == ncol_global))CALL cp__a("fm/cp_fm_basic_linalg.F",980)
      IF(.NOT.(matrix%use_sp .EQV. work%use_sp))CALL cp__a("fm/cp_fm_basic_linalg.F",981)

      CALL timeset(routineN, handle)

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)


      a => matrix%local_data
      a_sp => matrix%local_data_sp
      DO irow_global = 1, nrow_global
         DO icol_global = irow_global+1, ncol_global
            IF (matrix%use_sp) THEN
               a_sp(icol_global, irow_global) = a_sp(irow_global, icol_global)
            ELSE
               a(icol_global, irow_global) = a(irow_global, icol_global)
            ENDIF
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE cp_fm_upper_to_full

! **************************************************************************************************
!> \brief scales column i of matrix a with scaling(i)
!> \param matrixa ...
!> \param scaling : an array used for scaling the columns, SIZE(scaling) determines the number of columns to be scaled
!> \author Joost VandeVondele
!> \note
!>      this is very useful as a first step in the computation of C = sum_i alpha_i A_i transpose (A_i)
!>      that is a rank-k update (cp_fm_syrk , cp_sm_plus_fm_fm_t) where every vector has a different prefactor
!>      this procedure can be up to 20 times faster than calling cp_fm_syrk n times
! **************************************************************************************************
   SUBROUTINE cp_fm_column_scale(matrixa, scaling)
      TYPE(cp_fm_type), POINTER                :: matrixa
      REAL(KIND=dp), DIMENSION(:), INTENT(in)  :: scaling

      INTEGER                                  :: k, mypcol, myprow, n, ncol_global, &
                                                  npcol, nprow
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: a_sp
      INTEGER                                  :: i

      myprow = matrixa%matrix_struct%context%mepos(1)
      mypcol = matrixa%matrix_struct%context%mepos(2)
      nprow = matrixa%matrix_struct%context%num_pe(1)
      npcol = matrixa%matrix_struct%context%num_pe(2)

      ncol_global = matrixa%matrix_struct%ncol_global

      a => matrixa%local_data
      a_sp => matrixa%local_data_sp
      IF (matrixa%use_sp) THEN
         n = SIZE(a_sp, 1)
      ELSE
         n = SIZE(a, 1)
      ENDIF
      k = MIN(SIZE(scaling), ncol_global)

      DO i = 1, k
         IF (matrixa%use_sp) THEN
            CALL SSCAL(n, REAL(scaling(i), sp), a_sp(1, i), 1)
         ELSE
            CALL DSCAL(n, scaling(i), a(1, i), 1)
         ENDIF
      ENDDO
   END SUBROUTINE cp_fm_column_scale

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_inverse ...
!> \param det_a ...
!> \author Florian Schiffmann(02.2007)
!> \note of Jan Wilhelm (12.2015)
!>      - computation of determinant corrected
!>      - determinant only computed if det_a is present
! **************************************************************************************************
   SUBROUTINE cp_fm_invert(matrix_a, matrix_inverse, det_a)

      TYPE(cp_fm_type), POINTER                :: matrix_a, matrix_inverse
      REAL(KIND=dp), INTENT(OUT), OPTIONAL     :: det_a

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_invert', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: n
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: ipivot
      REAL(KIND=dp)                            :: determinant
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      TYPE(cp_fm_type), POINTER                :: matrix_B, matrix_lu
      LOGICAL                                  :: sign
      REAL(KIND=dp)                            :: eps1

      CALL cp_fm_create(matrix=matrix_lu, &
                        matrix_struct=matrix_a%matrix_struct, &
                        name="A_lu"//TRIM(ADJUSTL(cp_to_string(1)))//"MATRIX")
      CALL cp_fm_to_fm(matrix_a, matrix_lu)

      CALL cp_fm_create(matrix=matrix_B, &
                        matrix_struct=matrix_a%matrix_struct, &
                        name="B_mat"//TRIM(ADJUSTL(cp_to_string(1)))//"MATRIX")
      a => matrix_lu%local_data
      n = matrix_lu%matrix_struct%nrow_global
      ALLOCATE (ipivot(n+matrix_a%matrix_struct%nrow_block))
      ipivot(:) = 0
      sign = .TRUE.
      CALL invert_matrix(matrix_a%local_data, matrix_inverse%local_data, &
                         eval_error=eps1)
      CALL cp_fm_lu_decompose(matrix_lu, determinant, correct_sign=sign)
      CALL cp_fm_release(matrix_lu)
      CALL cp_fm_release(matrix_B)
      DEALLOCATE (ipivot)
      IF (PRESENT(det_a)) det_a = determinant
   END SUBROUTINE cp_fm_invert

! **************************************************************************************************
!> \brief inverts a triangular matrix
!> \param matrix_a ...
!> \param uplo_tr ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE cp_fm_triangular_invert(matrix_a, uplo_tr)

      TYPE(cp_fm_type), POINTER                :: matrix_a
      CHARACTER, INTENT(IN), OPTIONAL          :: uplo_tr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_triangular_invert', &
                                     routineP = moduleN//':'//routineN

      CHARACTER                                :: unit_diag, uplo
      INTEGER                                  :: handle, info, ncol_global
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a

      CALL timeset(routineN, handle)

      unit_diag = 'N'
      uplo = 'U'
      IF (PRESENT(uplo_tr)) uplo = uplo_tr

      ncol_global = matrix_a%matrix_struct%ncol_global

      a => matrix_a%local_data

      CALL dtrtri(uplo, unit_diag, ncol_global, a(1, 1), ncol_global, info)

      CALL timestop(handle)
   END SUBROUTINE cp_fm_triangular_invert

! **************************************************************************************************
!> \brief  perfoms a QR factorization of the input rectangular matrix A or of a submatrix of A
!>         the computed upper triangular matrix R is in output in the submatrix sub(A) of size NxN
!>         M and M give the dimention of the submatrix that has to be factorized (MxN) with M>N
!> \param matrix_a ...
!> \param matrix_r ...
!> \param nrow_fact ...
!> \param ncol_fact ...
!> \param first_row ...
!> \param first_col ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE cp_fm_qr_factorization(matrix_a, matrix_r, nrow_fact, ncol_fact, first_row, first_col)
      TYPE(cp_fm_type), POINTER                :: matrix_a, matrix_r
      INTEGER, INTENT(IN), OPTIONAL            :: nrow_fact, ncol_fact, &
                                                  first_row, first_col

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_qr_factorization', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, i, icol, info, irow, &
                                                  j, lda, lwork, ncol, &
                                                  ndim, nrow
      REAL(dp), ALLOCATABLE, DIMENSION(:)      :: tau, work
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)   :: r_mat
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a

      CALL timeset(routineN, handle)

      ncol = matrix_a%matrix_struct%ncol_global
      nrow = matrix_a%matrix_struct%nrow_global
      lda = nrow

      a => matrix_a%local_data

      IF (PRESENT(nrow_fact)) nrow = nrow_fact
      IF (PRESENT(ncol_fact)) ncol = ncol_fact
      irow = 1
      IF (PRESENT(first_row)) irow = first_row
      icol = 1
      IF (PRESENT(first_col)) icol = first_col

      IF(.NOT.(nrow >= ncol))CALL cp__a("fm/cp_fm_basic_linalg.F",1333)
      ndim = SIZE(a, 2)
!    ALLOCATE(ipiv(ndim))
      ALLOCATE (tau(ndim))

      lwork = -1
      ALLOCATE (work(2*ndim))
      CALL dgeqrf(nrow, ncol, a, lda, tau, work, lwork, info)
      lwork = INT(work(1))
      DEALLOCATE (work)
      ALLOCATE (work(lwork))
      CALL dgeqrf(nrow, ncol, a, lda, tau, work, lwork, info)


      ALLOCATE (r_mat(ncol, ncol))
      CALL cp_fm_get_submatrix(matrix_a, r_mat, 1, 1, ncol, ncol)
      DO i = 1, ncol
         DO j = i+1, ncol
            r_mat(j, i) = 0.0_dp
         END DO
      END DO
      CALL cp_fm_set_submatrix(matrix_r, r_mat, 1, 1, ncol, ncol)

      DEALLOCATE (tau, work, r_mat)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_qr_factorization

! **************************************************************************************************
!> \brief computs the the solution to A*b=A_general using lu decomposition
!>        pay attention, both matrices are overwritten, a_general contais the result
!> \param matrix_a ...
!> \param general_a ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE cp_fm_solve(matrix_a, general_a)
      TYPE(cp_fm_type), POINTER                :: matrix_a, general_a

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_solve', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, info, n
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: ipivot
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, a_general
      INTEGER                                  :: lda, ldb

      CALL timeset(routineN, handle)

      a => matrix_a%local_data
      a_general => general_a%local_data
      n = matrix_a%matrix_struct%nrow_global
      ALLOCATE (ipivot(n+matrix_a%matrix_struct%nrow_block))

      lda = SIZE(a, 1)
      ldb = SIZE(a_general, 1)
      CALL dgetrf(n, n, a(1, 1), lda, ipivot, info)
      CALL dgetrs("N", n, n, a(1, 1), lda, ipivot, a_general, ldb, info)

      ! info is allowed to be zero
      ! this does just signal a zero diagonal element
      DEALLOCATE (ipivot)
      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief Convenience function. Computes the matrix multiplications needed
!>        for the multiplication of complex matrices.
!>        C = beta * C + alpha * ( A  ** transa ) * ( B ** transb )
!> \param transa : 'N' -> normal   'T' -> transpose
!>      alpha,beta :: can be 0.0_dp and 1.0_dp
!> \param transb ...
!> \param m ...
!> \param n ...
!> \param k ...
!> \param alpha ...
!> \param A_re m x k matrix ( ! for transa = 'N'), real part
!> \param A_im m x k matrix ( ! for transa = 'N'), imaginary part
!> \param B_re k x n matrix ( ! for transa = 'N'), real part
!> \param B_im k x n matrix ( ! for transa = 'N'), imaginary part
!> \param beta ...
!> \param C_re m x n matrix, real part
!> \param C_im m x n matrix, imaginary part
!> \param a_first_col ...
!> \param a_first_row ...
!> \param b_first_col : the k x n matrix starts at col b_first_col of matrix_b (avoid usage)
!> \param b_first_row ...
!> \param c_first_col ...
!> \param c_first_row ...
!> \author Samuel Andermatt
!> \note
!>      C should have no overlap with A, B
! **************************************************************************************************
   SUBROUTINE cp_complex_fm_gemm(transa, transb, m, n, k, alpha, A_re, A_im, B_re, B_im, beta, &
                                 C_re, C_im, a_first_col, a_first_row, b_first_col, b_first_row, c_first_col, &
                                 c_first_row)
      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      INTEGER, INTENT(IN)                                :: m, n, k
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      TYPE(cp_fm_type), POINTER                          :: A_re, A_im, B_re, B_im
      REAL(KIND=dp), INTENT(IN)                          :: beta
      TYPE(cp_fm_type), POINTER                          :: C_re, C_im
      INTEGER, INTENT(IN), OPTIONAL                      :: a_first_col, a_first_row, b_first_col, &
                                                            b_first_row, c_first_col, c_first_row

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_complex_fm_gemm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      CALL cp_fm_gemm(transa, transb, m, n, k, alpha, A_re, B_re, beta, C_re, &
                      a_first_col=a_first_col, &
                      a_first_row=a_first_row, &
                      b_first_col=b_first_col, &
                      b_first_row=b_first_row, &
                      c_first_col=c_first_col, &
                      c_first_row=c_first_row)
      CALL cp_fm_gemm(transa, transb, m, n, k, -alpha, A_im, B_im, 1.0_dp, C_re, &
                      a_first_col=a_first_col, &
                      a_first_row=a_first_row, &
                      b_first_col=b_first_col, &
                      b_first_row=b_first_row, &
                      c_first_col=c_first_col, &
                      c_first_row=c_first_row)
      CALL cp_fm_gemm(transa, transb, m, n, k, alpha, A_re, B_im, beta, C_im, &
                      a_first_col=a_first_col, &
                      a_first_row=a_first_row, &
                      b_first_col=b_first_col, &
                      b_first_row=b_first_row, &
                      c_first_col=c_first_col, &
                      c_first_row=c_first_row)
      CALL cp_fm_gemm(transa, transb, m, n, k, alpha, A_im, B_re, 1.0_dp, C_im, &
                      a_first_col=a_first_col, &
                      a_first_row=a_first_row, &
                      b_first_col=b_first_col, &
                      b_first_row=b_first_row, &
                      c_first_col=c_first_col, &
                      c_first_row=c_first_row)

      CALL timestop(handle)

   END SUBROUTINE cp_complex_fm_gemm

! **************************************************************************************************
!> \brief inverts a matrix using LU decomposition
!>        the input matrix will be overwritten
!> \param matrix   : input a general square non-singular matrix, outputs its inverse
!> \param info_out : optional, if present outputs the info from (p)zgetri
!> \author Lianheng Tong
! **************************************************************************************************
   SUBROUTINE cp_fm_lu_invert(matrix, info_out)
      TYPE(cp_fm_type), POINTER                :: matrix
      INTEGER, INTENT(OUT), OPTIONAL           :: info_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_lu_invert', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrows_global, handle, info, lwork
      INTEGER, DIMENSION(:), ALLOCATABLE       :: ipivot
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: mat
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: mat_sp
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: work
      REAL(KIND=sp), DIMENSION(:), ALLOCATABLE :: work_sp
      INTEGER                                  :: lda

      CALL timeset(routineN, handle)

      mat => matrix%local_data
      mat_sp => matrix%local_data_sp
      nrows_global = matrix%matrix_struct%nrow_global
      IF(.NOT.(nrows_global .EQ. matrix%matrix_struct%ncol_global))CALL cp__a("fm/cp_fm_basic_linalg.F",1538)
      ALLOCATE (ipivot(nrows_global))
      ! do LU decomposition
      lda = SIZE(mat, 1)
      IF (matrix%use_sp) THEN
         CALL sgetrf(nrows_global, nrows_global, &
                     mat_sp, lda, ipivot, info)
      ELSE
         CALL dgetrf(nrows_global, nrows_global, &
                     mat, lda, ipivot, info)
      END IF
      IF (info /= 0) THEN
         CALL cp_abort(cp__l("fm/cp_fm_basic_linalg.F",1561), "LU decomposition has failed")
      END IF
      ! do inversion
      IF (matrix%use_sp) THEN
         ALLOCATE (work(1))
      ELSE
         ALLOCATE (work_sp(1))
      END IF
      IF (matrix%use_sp) THEN
         CALL sgetri(nrows_global, mat_sp, lda, &
                     ipivot, work_sp, -1, info)
         lwork = INT(work_sp(1))
         DEALLOCATE (work_sp)
         ALLOCATE (work_sp(lwork))
         CALL sgetri(nrows_global, mat_sp, lda, &
                     ipivot, work_sp, lwork, info)
      ELSE
         CALL dgetri(nrows_global, mat, lda, &
                     ipivot, work, -1, info)
         lwork = INT(work(1))
         DEALLOCATE (work)
         ALLOCATE (work(lwork))
         CALL dgetri(nrows_global, mat, lda, &
                     ipivot, work, lwork, info)
      END IF
      IF (matrix%use_sp) THEN
         DEALLOCATE (work_sp)
      ELSE
         DEALLOCATE (work)
      END IF
      DEALLOCATE (ipivot)

      IF (PRESENT(info_out)) THEN
         info = info_out
      ELSE
         IF (info /= 0) &
            CALL cp_abort(cp__l("fm/cp_fm_basic_linalg.F",1625), "LU inversion has failed")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_lu_invert

! **************************************************************************************************
!> \brief norm of matrix using (p)dlange
!> \param matrix   : input a general matrix
!> \param mode     : 'M' max abs element value,
!>                   '1' or 'O' one norm, i.e. maximum column sum
!>                   'I' infinity norm, i.e. maximum row sum
!>                   'F' or 'E' Frobenius norm, i.e. sqrt of sum of all squares of elements
!> \retval res     : the norm according to mode
!> \author Lianheng Tong
! **************************************************************************************************
   FUNCTION cp_fm_norm(matrix, mode) RESULT(res)
      TYPE(cp_fm_type), POINTER :: matrix
      CHARACTER, INTENT(IN) :: mode
      REAL(KIND=dp) :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_norm', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrows, ncols, handle, lwork, nrows_local, ncols_local
      REAL(KIND=sp) :: res_sp
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: aa
      REAL(KIND=sp), DIMENSION(:, :), POINTER :: aa_sp
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: work
      REAL(KIND=sp), DIMENSION(:), ALLOCATABLE :: work_sp
      INTEGER :: lda

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(matrix=matrix, &
                          nrow_global=nrows, &
                          ncol_global=ncols, &
                          nrow_local=nrows_local, &
                          ncol_local=ncols_local)
      aa => matrix%local_data
      aa_sp => matrix%local_data_sp

      SELECT CASE (mode)
      CASE ('M', 'm')
         lwork = 1
      CASE ('1', 'O', 'o')
         lwork = 1
      CASE ('I', 'i')
         lwork = nrows
      CASE ('F', 'f', 'E', 'e')
         lwork = 1
      CASE DEFAULT
         CALL cp__b("fm/cp_fm_basic_linalg.F",1707,"mode input is not valid")
      END SELECT
      IF (matrix%use_sp) THEN
         ALLOCATE (work_sp(lwork))
         lda = SIZE(aa_sp, 1)
         res_sp = slange(mode, nrows, ncols, aa_sp, lda, work_sp)
         DEALLOCATE (work_sp)
         res = REAL(res_sp, KIND=dp)
      ELSE
         ALLOCATE (work(lwork))
         lda = SIZE(aa, 1)
         res = dlange(mode, nrows, ncols, aa, lda, work)
         DEALLOCATE (work)
      END IF

      CALL timestop(handle)

   END FUNCTION cp_fm_norm

! **************************************************************************************************
!> \brief trace of a matrix using pdlatra
!> \param matrix   : input a square matrix
!> \retval res     : the trace
!> \author Lianheng Tong
! **************************************************************************************************
   FUNCTION cp_fm_latra(matrix) RESULT(res)
      TYPE(cp_fm_type), POINTER :: matrix
      REAL(KIND=dp) :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_latra', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: nrows, ncols, handle
      REAL(KIND=sp) :: res_sp
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: aa
      REAL(KIND=sp), DIMENSION(:, :), POINTER :: aa_sp
      INTEGER :: ii

      CALL timeset(routineN, handle)

      nrows = matrix%matrix_struct%nrow_global
      ncols = matrix%matrix_struct%ncol_global
      IF(.NOT.(nrows .EQ. ncols))CALL cp__a("fm/cp_fm_basic_linalg.F",1754)
      aa => matrix%local_data
      aa_sp => matrix%local_data_sp

      IF (matrix%use_sp) THEN
         res_sp = 0.0_sp
         DO ii = 1, nrows
            res_sp = res_sp+aa_sp(ii, ii)
         END DO
         res = REAL(res_sp, KIND=dp)
      ELSE
         res = 0.0_dp
         DO ii = 1, nrows
            res = res+aa(ii, ii)
         END DO
      END IF

      CALL timestop(handle)

   END FUNCTION cp_fm_latra

END MODULE cp_fm_basic_linalg
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR operations in CP2K
!> \author  Urban Borstnik
!> \date    2009-05-12
!> \version 0.8
!>
!> <b>Modification history:</b>
!> - Created 2009-05-12
! **************************************************************************************************
MODULE cp_dbcsr_operations
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type,&
                                              get_blacs_info
   USE cp_cfm_types,                    ONLY: cp_cfm_type
   USE cp_dbcsr_interface,              ONLY: &
        convert_sizes_to_offsets, cp_add_work_coordinate, cp_create_bl_distribution, cp_dbcsr_add, &
        cp_dbcsr_col_block_sizes, cp_dbcsr_complete_redistribute, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_distribution, cp_dbcsr_distribution_new, cp_dbcsr_finalize, cp_dbcsr_get_data_p, &
        cp_dbcsr_get_data_size, cp_dbcsr_get_data_type, cp_dbcsr_get_info, &
        cp_dbcsr_get_matrix_type, cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_mp_new, cp_dbcsr_mp_release, cp_dbcsr_multiply, cp_dbcsr_name, cp_dbcsr_norm, &
        cp_dbcsr_release, cp_dbcsr_row_block_sizes, cp_dbcsr_scale, cp_dbcsr_type, &
        cp_dbcsr_valid_index, cp_dbcsr_verify_matrix, cp_dbcsr_work_create, cp_set_work_size, &
        dbcsr_create_dist_block_cyclic, dbcsr_create_dist_r_unrot, &
        dbcsr_distribution_col_clusters, dbcsr_distribution_col_dist, dbcsr_distribution_init, &
        dbcsr_distribution_local_cols, dbcsr_distribution_local_rows, dbcsr_distribution_mp, &
        dbcsr_distribution_ncols, dbcsr_distribution_nlocal_cols, dbcsr_distribution_nlocal_rows, &
        dbcsr_distribution_nrows, dbcsr_distribution_obj, dbcsr_distribution_release, &
        dbcsr_distribution_row_clusters, dbcsr_distribution_row_dist, dbcsr_mp_hold, &
        dbcsr_mp_npcols, dbcsr_mp_nprows, dbcsr_mp_obj, dbcsr_norm_frobenius, &
        dbcsr_type_antisymmetric, dbcsr_type_complex_8, dbcsr_type_no_symmetry, dbcsr_type_real_4, &
        dbcsr_type_real_8, dbcsr_type_symmetric
   USE cp_fm_basic_linalg,              ONLY: cp_fm_gemm
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_2d_types,           ONLY: distribution_2d_get,&
                                              distribution_2d_type
   USE kinds,                           ONLY: dp,&
                                              real_4,&
                                              sp

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_dbcsr_operations'

   PUBLIC :: cp_dbcsr_multiply_local

   INTERFACE cp_dbcsr_multiply_local
      MODULE PROCEDURE cp_dbcsr_multiply_local_d, &
         cp_dbcsr_multiply_local_s
   END INTERFACE

   ! CP2K API emulation
   PUBLIC :: copy_fm_to_dbcsr, copy_dbcsr_to_fm, &
             copy_dbcsr_to_cfm, copy_cfm_to_dbcsr, &
             cp_dbcsr_sm_fm_multiply, cp_dbcsr_plus_fm_fm_t, &
             copy_dbcsr_to_fm_bc, copy_fm_to_dbcsr_bc, cp_fm_to_dbcsr_row_template, &
             cp_dbcsr_m_by_n_from_template, cp_dbcsr_m_by_n_from_row_template

   ! distribution_2d_type compatibility
   PUBLIC :: cp_dbcsr_dist2d_to_dist

   PUBLIC :: cp_dbcsr_copy_columns_hack

   INTERFACE cp_dbcsr_plus_fm_fm_t
      MODULE PROCEDURE cp_dbcsr_plus_fm_fm_t_native
   END INTERFACE

   PRIVATE

CONTAINS

! **************************************************************************************************
!> \brief hack for dbcsr_copy_columns
!> \param matrix_b ...
!> \param matrix_a ...
!> \param ncol ...
!> \param source_start ...
!> \param target_start ...
!> \param para_env ...
!> \param blacs_env ...
!> \author vw
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_copy_columns_hack(matrix_b, matrix_a, &
                                         ncol, source_start, target_start, para_env, blacs_env)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_b
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_a
      INTEGER, INTENT(IN)                                :: ncol, source_start, target_start
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_copy_columns_hack', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nfullcols_total, nfullrows_total
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_a, fm_matrix_b

      NULLIFY (fm_matrix_a, fm_matrix_b, fm_struct)
      CALL cp_dbcsr_get_info(matrix_a, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env)
      CALL cp_fm_create(fm_matrix_a, fm_struct, name="fm_matrix_a")
      CALL cp_fm_struct_release(fm_struct)

      CALL cp_dbcsr_get_info(matrix_b, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env)
      CALL cp_fm_create(fm_matrix_b, fm_struct, name="fm_matrix_b")
      CALL cp_fm_struct_release(fm_struct)

      CALL copy_dbcsr_to_fm(matrix_a, fm_matrix_a)
      CALL copy_dbcsr_to_fm(matrix_b, fm_matrix_b)

      CALL cp_fm_to_fm(fm_matrix_a, fm_matrix_b, ncol, source_start, target_start)

      CALL copy_fm_to_dbcsr(fm_matrix_b, matrix_b)

      CALL cp_fm_release(fm_matrix_a)
      CALL cp_fm_release(fm_matrix_b)

   END SUBROUTINE cp_dbcsr_copy_columns_hack

! **************************************************************************************************
!> \brief Creates a DBCSR distribution from a distribution_2d
!> \param[in] dist2d          distribution_2d
!> \param[out] dist           DBCSR distribution
!> \param unit_nr ...
!> \param mp_obj ...
!> \par History
!>    move form dbcsr_operation 01.2010
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_dist2d_to_dist(dist2d, dist, unit_nr, mp_obj)
      TYPE(distribution_2d_type), INTENT(IN), TARGET     :: dist2d
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist
      INTEGER, INTENT(IN)                                :: unit_nr
      TYPE(dbcsr_mp_obj), INTENT(IN), OPTIONAL           :: mp_obj

      INTEGER                                            :: mypcol, myproc, myprow, numproc
      INTEGER, DIMENSION(:, :), POINTER                  :: col_dist_data, pgrid, row_dist_data
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_mp_obj)                                 :: mp_env
      TYPE(distribution_2d_type), POINTER                :: dist2d_p

!
!   ---------------------------------------------------------------------------

      dist2d_p => dist2d
      CALL distribution_2d_get(dist2d_p, &
                               row_distribution=row_dist_data, col_distribution=col_dist_data, &
                               blacs_env=blacs_env)
      CALL get_blacs_info(blacs_env, para_env=para_env, &
                          my_process_row=myprow, my_process_column=mypcol, &
                          blacs2mpi=pgrid)
      myproc = para_env%mepos
      numproc = para_env%num_pe
      IF (PRESENT(mp_obj)) THEN
         mp_env = mp_obj
         CALL dbcsr_mp_hold(mp_env)
      ELSE
         CALL cp_dbcsr_mp_new(mp_env, pgrid, para_env%group, myproc, numproc, &
                              myprow, mypcol)
      ENDIF
      CALL cp_dbcsr_distribution_new(dist, mp_env, row_dist_data, col_dist_data, unit_nr=unit_nr)
      CALL cp_dbcsr_mp_release(mp_env)
   END SUBROUTINE cp_dbcsr_dist2d_to_dist

! **************************************************************************************************
!> \brief multiply a dbcsr with a replicated array
!>        c = alpha_scalar * A (dbscr) * b + c
!> \param[in] matrix_a DBSCR matrxx
!> \param[in]  vec_b        vectors b
!> \param[inout] vec_c      vectors c
!> \param[in]  ncol         nbr of columns
!> \param[in]  alpha        alpha
!>
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_multiply_local_d(matrix_a, vec_b, vec_c, ncol, alpha)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_a
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: vec_b
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: vec_c
      INTEGER, INTENT(in), OPTIONAL                      :: ncol
      REAL(dp), INTENT(IN), OPTIONAL                     :: alpha

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_multiply_local_d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, coloff, my_ncol, row, rowoff, &
                                                            timing_handle
      LOGICAL                                            :: has_symm
      REAL(dp)                                           :: my_alpha, my_alpha2
      REAL(dp), DIMENSION(:, :), POINTER                 :: data_d
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, timing_handle)

      my_alpha = 1.0_dp
      IF (PRESENT(alpha)) my_alpha = alpha

      my_ncol = SIZE(vec_b, 2)
      IF (PRESENT(ncol)) my_ncol = ncol

      my_alpha2 = 0.0_dp
      IF (cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_symmetric) my_alpha2 = my_alpha
      IF (cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_antisymmetric) my_alpha2 = -my_alpha

      has_symm = (cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_symmetric .OR. &
                  cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_antisymmetric)

!$OMP     PARALLEL DEFAULT(NONE) SHARED(matrix_a,vec_b,vec_c,ncol,my_alpha2,my_alpha,my_ncol,has_symm) &
!$OMP              PRIVATE(iter,row,col,data_d,blk,rowoff,coloff)
      CALL cp_dbcsr_iterator_start(iter, matrix_a, read_only=.TRUE., dynamic=.TRUE., dynamic_byrows=.TRUE.)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, data_d, blk, row_offset=rowoff, col_offset=coloff)
         IF (my_ncol .NE. 1) THEN
            CALL dgemm('N', 'N', &
                       SIZE(data_d, 1), my_ncol, SIZE(data_d, 2), &
                       my_alpha, data_d(1, 1), SIZE(data_d, 1), &
                       vec_b(coloff, 1), SIZE(vec_b, 1), &
                       1.0_dp, vec_c(rowoff, 1), SIZE(vec_c, 1))
         ELSE
            CALL dgemv('N', SIZE(data_d, 1), SIZE(data_d, 2), &
                       my_alpha, data_d(1, 1), SIZE(data_d, 1), &
                       vec_b(coloff, 1), 1, &
                       1.0_dp, vec_c(rowoff, 1), 1)
         ENDIF
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)
!$OMP     END PARALLEL

      ! FIXME ... in the symmetric case, the writes to vec_c depend on the column, not the row. This makes OMP-ing more difficult
      ! needs e.g. a buffer for vec_c and a reduction of that buffer.
      IF (has_symm) THEN
         CALL cp_dbcsr_iterator_start(iter, matrix_a)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_d, blk, row_offset=rowoff, col_offset=coloff)
            IF (row .NE. col) THEN
               IF (my_ncol .NE. 1) THEN
                  CALL dgemm('T', 'N', &
                             SIZE(data_d, 2), my_ncol, SIZE(data_d, 1), &
                             my_alpha2, data_d(1, 1), SIZE(data_d, 1), &
                             vec_b(rowoff, 1), SIZE(vec_b, 1), &
                             1.0_dp, vec_c(coloff, 1), SIZE(vec_c, 1))
               ELSE
                  CALL dgemv('T', SIZE(data_d, 1), SIZE(data_d, 2), &
                             my_alpha2, data_d(1, 1), SIZE(data_d, 1), &
                             vec_b(rowoff, 1), 1, &
                             1.0_dp, vec_c(coloff, 1), 1)
               ENDIF
            ENDIF
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)
      ENDIF

      CALL timestop(timing_handle)
   END SUBROUTINE cp_dbcsr_multiply_local_d

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param vec_b ...
!> \param vec_c ...
!> \param ncol ...
!> \param alpha ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_multiply_local_s(matrix_a, vec_b, vec_c, ncol, alpha)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_a
      REAL(KIND=real_4), DIMENSION(:, :), INTENT(IN)     :: vec_b
      REAL(KIND=real_4), DIMENSION(:, :), INTENT(INOUT)  :: vec_c
      INTEGER, INTENT(in), OPTIONAL                      :: ncol
      REAL(KIND=real_4), INTENT(IN), OPTIONAL            :: alpha

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_multiply_local_s', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, coloff, my_ncol, row, rowoff, &
                                                            timing_handle
      REAL(KIND=real_4)                                  :: my_alpha, my_alpha2
      REAL(KIND=real_4), DIMENSION(:, :), POINTER        :: data_d
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, timing_handle)

      my_alpha = 1.0
      IF (PRESENT(alpha)) my_alpha = alpha

      my_ncol = SIZE(vec_b, 2)
      IF (PRESENT(ncol)) my_ncol = ncol

      my_alpha2 = 0.0
      IF (cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_symmetric) my_alpha2 = my_alpha
      IF (cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_antisymmetric) my_alpha2 = -my_alpha

      CALL cp_dbcsr_iterator_start(iter, matrix_a)

      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

         CALL cp_dbcsr_iterator_next_block(iter, row, col, data_d, blk, row_offset=rowoff, col_offset=coloff)

         CALL sgemm('N', 'N', &
                    SIZE(data_d, 1), my_ncol, SIZE(data_d, 2), &
                    my_alpha, data_d(1, 1), SIZE(data_d, 1), &
                    vec_b(coloff, 1), SIZE(vec_b, 1), &
                    1.0, vec_c(rowoff, 1), SIZE(vec_c, 1))

         IF ((cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_symmetric .OR. &
              cp_dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_antisymmetric)) THEN
            IF (row .NE. col) THEN
               CALL sgemm('T', 'N', &
                          SIZE(data_d, 2), my_ncol, SIZE(data_d, 1), &
                          my_alpha2, data_d(1, 1), SIZE(data_d, 1), &
                          vec_b(rowoff, 1), SIZE(vec_b, 1), &
                          1.0, vec_c(coloff, 1), SIZE(vec_c, 1))
            ENDIF
         ENDIF
      ENDDO

      CALL cp_dbcsr_iterator_stop(iter)

      CALL timestop(timing_handle)
   END SUBROUTINE cp_dbcsr_multiply_local_s

! **************************************************************************************************
!> \brief multiply a dbcsr with a fm matrix
!>
!> For backwards compatibility with BLAS XGEMM, this routine supports
!> the multiplication of matrices with incompatible dimensions.
!>
!> \param[in]  matrix         DBCSR matrix
!> \param fm_in full matrix
!> \param fm_out full matrix
!> \param[in]  ncol           nbr of columns
!> \param[in]  alpha          alpha
!> \param[in]  beta           beta
!>
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_sm_fm_multiply(matrix, fm_in, fm_out, ncol, alpha, beta)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(cp_fm_type), POINTER                          :: fm_in, fm_out
      INTEGER, INTENT(IN)                                :: ncol
      REAL(dp), INTENT(IN), OPTIONAL                     :: alpha, beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_sm_fm_multiply', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k_in, k_out, timing_handle, &
                                                            timing_handle_mult
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size_right_in, &
                                                            col_blk_size_right_out
      TYPE(cp_dbcsr_type)                                :: in, out
      REAL(dp)                                           :: my_alpha, my_beta
      TYPE(dbcsr_distribution_obj)                       :: dist_right_in, product_dist

      CALL timeset(routineN, timing_handle)

      my_alpha = 1.0_dp
      my_beta = 0.0_dp
      IF (PRESENT(alpha)) my_alpha = alpha
      IF (PRESENT(beta)) my_beta = beta

      CALL cp_fm_get_info(fm_out, ncol_global=k_out)

      CALL cp_fm_get_info(fm_in, ncol_global=k_in)
      !write(*,*)routineN//" -----------------------------------"
      !IF (k_in .NE. k_out) &
      !   WRITE(*,'(3(A,I5,1X),2(A,F5.2,1X))')&
      !   routineN//" ncol", ncol,'k_in',k_in,'k_out',k_out,&
      !   'alpha',my_alpha,'beta',my_beta

      IF (ncol .GT. 0 .AND. k_out .GT. 0 .AND. k_in .GT. 0) THEN
         CALL dbcsr_create_dist_r_unrot(dist_right_in, cp_dbcsr_distribution(matrix), k_in, &
                                        col_blk_size_right_in)

         CALL cp_dbcsr_init(in)
         CALL cp_dbcsr_create(in, "D", dist_right_in, dbcsr_type_no_symmetry, &
                              cp_dbcsr_row_block_sizes(matrix), col_blk_size_right_in, &
                              nze=0)

         CALL cp_dbcsr_init(out)
         CALL cp_dbcsr_distribution_new(product_dist, &
                                        dbcsr_distribution_mp(cp_dbcsr_distribution(matrix)), &
                                        dbcsr_distribution_row_dist(cp_dbcsr_distribution(matrix)), &
                                        dbcsr_distribution_col_dist(dist_right_in), &
                                        dbcsr_distribution_row_clusters(cp_dbcsr_distribution(matrix)), &
                                        dbcsr_distribution_col_clusters(dist_right_in))
         ALLOCATE (col_blk_size_right_out(SIZE(col_blk_size_right_in)))
         col_blk_size_right_out = col_blk_size_right_in
         CALL match_col_sizes(col_blk_size_right_out, col_blk_size_right_in, k_out)

         !if (k_in .ne. k_out) then
         !   write(*,*)routineN//" in cs", col_blk_size_right_in
         !   write(*,*)routineN//" out cs", col_blk_size_right_out
         !endif

         CALL cp_dbcsr_create(out, "D", product_dist, dbcsr_type_no_symmetry, &
                              cp_dbcsr_row_block_sizes(matrix), col_blk_size_right_out, &
                              nze=0)

         CALL copy_fm_to_dbcsr(fm_in, in)
         IF (ncol .NE. k_out .OR. my_beta .NE. 0.0_dp) &
            CALL copy_fm_to_dbcsr(fm_out, out)

         CALL timeset(routineN//'_core', timing_handle_mult)
         CALL cp_dbcsr_multiply("N", "N", my_alpha, matrix, in, my_beta, out, &
                                last_column=ncol, match_matrix_sizes=.TRUE.)
         CALL timestop(timing_handle_mult)

         CALL copy_dbcsr_to_fm(out, fm_out)

         CALL cp_dbcsr_release(in)
         CALL cp_dbcsr_release(out)
         DEALLOCATE (col_blk_size_right_in, col_blk_size_right_out)
         CALL dbcsr_distribution_release(dist_right_in)
         CALL dbcsr_distribution_release(product_dist)

      ENDIF

      CALL timestop(timing_handle)

   END SUBROUTINE cp_dbcsr_sm_fm_multiply

! **************************************************************************************************
!> \brief ...
!> \param sizes1 ...
!> \param sizes2 ...
!> \param full_num ...
! **************************************************************************************************
   SUBROUTINE match_col_sizes(sizes1, sizes2, full_num)
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: sizes1
      INTEGER, DIMENSION(:), INTENT(IN)                  :: sizes2
      INTEGER, INTENT(IN)                                :: full_num

      INTEGER                                            :: left, n1, n2, p, rm, used

      n1 = SIZE(sizes1)
      n2 = SIZE(sizes2)
      IF (n1 .NE. n2) &
         CALL cp__b("cp_dbcsr_operations.F",455,"distributions must be equal!")
      sizes1(1:n1) = sizes2(1:n1)
      used = SUM(sizes1(1:n1))
      ! If sizes1 does not cover everything, then we increase the
      ! size of the last block; otherwise we reduce the blocks
      ! (from the end) until it is small enough.
      IF (used .LT. full_num) THEN
         sizes1(n1) = sizes1(n1)+full_num-used
      ELSE
         left = used-full_num
         p = n1
         DO WHILE (left .GT. 0 .AND. p .GT. 0)
            rm = MIN(left, sizes1(p))
            sizes1(p) = sizes1(p)-rm
            left = left-rm
            p = p-1
         ENDDO
      ENDIF
   END SUBROUTINE match_col_sizes
! **************************************************************************************************
!> \brief performs the multiplication sparse_matrix+dense_mat*dens_mat^T
!>        if matrix_g is not explicitly given, matrix_v^T will be used
!>        this can be important to save the necessary redistribute for a
!>        different matrix_g and increase performance.
!> \param sparse_matrix ...
!> \param matrix_v ...
!> \param matrix_g ...
!> \param ncol ...
!> \param alpha ...
!> \param keep_sparsity Determines if the sparsity of sparse_matrix is retained
!>        by default it is TRUE
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_plus_fm_fm_t_native(sparse_matrix, matrix_v, matrix_g, ncol, &
                                           alpha, keep_sparsity)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: sparse_matrix
      TYPE(cp_fm_type), POINTER                          :: matrix_v
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: matrix_g
      INTEGER, INTENT(IN)                                :: ncol
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: alpha
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_sparsity

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_plus_fm_fm_t_native', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k, nao, timing_handle
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size_left, col_dist_cluster, &
                                                            col_dist_left
      LOGICAL                                            :: check_product, my_keep_sparsity
      REAL(KIND=dp)                                      :: my_alpha, norm
      TYPE(cp_dbcsr_type)                                :: mat_g, mat_v, sparse_matrix2, &
                                                            sparse_matrix3
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: fm_matrix
      TYPE(dbcsr_distribution_obj)                       :: dist_left
      TYPE(dbcsr_mp_obj)                                 :: mp

      check_product = .FALSE.

      CALL timeset(routineN, timing_handle)

      my_keep_sparsity = .TRUE.
      IF (PRESENT(keep_sparsity)) my_keep_sparsity = keep_sparsity
      NULLIFY (col_dist_left)

      IF (ncol .GT. 0) THEN
         IF (.NOT. cp_dbcsr_valid_index(sparse_matrix)) &
            CALL cp__b("cp_dbcsr_operations.F",521,"sparse_matrix must pre-exist")
         !
         ! Setup matrix_v
         CALL cp_fm_get_info(matrix_v, ncol_global=k)
         !WRITE(*,*)routineN//'truncated mult k, ncol',k,ncol,' PRESENT (matrix_g)',PRESENT (matrix_g)
         mp = dbcsr_distribution_mp(cp_dbcsr_distribution(sparse_matrix))
         CALL cp_create_bl_distribution(col_dist_left, col_blk_size_left, &
                                        k, dbcsr_mp_npcols(mp))
         NULLIFY (col_dist_cluster)
         CALL cp_dbcsr_distribution_new(dist_left, mp, &
                                        dbcsr_distribution_row_dist(cp_dbcsr_distribution(sparse_matrix)), &
                                        col_dist_left, &
                                        dbcsr_distribution_row_clusters(cp_dbcsr_distribution(sparse_matrix)), &
                                        col_dist_cluster)
         DEALLOCATE (col_dist_left)
         CALL cp_dbcsr_init(mat_v)
         CALL cp_dbcsr_create(mat_v, "DBCSR matrix_v", dist_left, dbcsr_type_no_symmetry, &
                              cp_dbcsr_row_block_sizes(sparse_matrix), col_blk_size_left, nze=0, &
                              data_type=cp_dbcsr_get_data_type(sparse_matrix))
         CALL copy_fm_to_dbcsr(matrix_v, mat_v)
         CALL cp_dbcsr_verify_matrix(mat_v)
         !
         ! Setup matrix_g
         IF (PRESENT(matrix_g)) THEN
            CALL cp_dbcsr_init(mat_g)
            CALL cp_dbcsr_create(mat_g, "DBCSR matrix_g", dist_left, &
                                 dbcsr_type_no_symmetry, &
                                 cp_dbcsr_row_block_sizes(sparse_matrix), &
                                 cp_dbcsr_col_block_sizes(mat_v), &
                                 data_type=cp_dbcsr_get_data_type(sparse_matrix))
            CALL copy_fm_to_dbcsr(matrix_g, mat_g)
         ENDIF
         !
         DEALLOCATE (col_blk_size_left)
         CALL dbcsr_distribution_release(dist_left)
         !
         !
         IF (check_product) THEN
            NULLIFY (fm_matrix)
            CALL cp_fm_get_info(matrix_v, nrow_global=nao)
            CALL cp_fm_struct_create(fm_struct_tmp, context=matrix_v%matrix_struct%context, nrow_global=nao, &
                                     ncol_global=nao, para_env=matrix_v%matrix_struct%para_env)
            CALL cp_fm_create(fm_matrix, fm_struct_tmp, name="fm matrix")
            CALL cp_fm_struct_release(fm_struct_tmp)
            CALL copy_dbcsr_to_fm(sparse_matrix, fm_matrix)
            CALL cp_dbcsr_init(sparse_matrix3)
            CALL cp_dbcsr_copy(sparse_matrix3, sparse_matrix)
         ENDIF
         !
         my_alpha = 1.0_dp
         IF (PRESENT(alpha)) my_alpha = alpha
         IF (PRESENT(matrix_g)) THEN
            CALL cp_dbcsr_multiply("N", "T", my_alpha, mat_v, mat_g, &
                                   1.0_dp, sparse_matrix, &
                                   retain_sparsity=my_keep_sparsity, &
                                   last_k=ncol)
         ELSE
            CALL cp_dbcsr_multiply("N", "T", my_alpha, mat_v, mat_v, &
                                   1.0_dp, sparse_matrix, &
                                   retain_sparsity=my_keep_sparsity, &
                                   last_k=ncol)
         ENDIF

         IF (check_product) THEN
            IF (PRESENT(matrix_g)) THEN
               CALL cp_fm_gemm("N", "T", nao, nao, ncol, my_alpha, matrix_v, matrix_g, &
                               1.0_dp, fm_matrix)
            ELSE
               CALL cp_fm_gemm("N", "T", nao, nao, ncol, my_alpha, matrix_v, matrix_v, &
                               1.0_dp, fm_matrix)
            ENDIF

            CALL cp_dbcsr_init(sparse_matrix2)
            CALL cp_dbcsr_copy(sparse_matrix2, sparse_matrix)
            CALL cp_dbcsr_scale(sparse_matrix2, alpha_scalar=0.0_dp)
            CALL copy_fm_to_dbcsr(fm_matrix, sparse_matrix2, keep_sparsity=my_keep_sparsity)
            CALL cp_dbcsr_add(sparse_matrix2, sparse_matrix, alpha_scalar=1.0_dp, &
                              beta_scalar=-1.0_dp)
            CALL cp_dbcsr_norm(sparse_matrix2, which_norm=dbcsr_norm_frobenius, &
                               norm_scalar=norm)
            WRITE (*, *) 'nao=', nao, ' k=', k, ' ncol=', ncol, ' my_alpha=', my_alpha
            WRITE (*, *) 'PRESENT (matrix_g)', PRESENT(matrix_g)
            WRITE (*, *) 'matrix_type=', cp_dbcsr_get_matrix_type(sparse_matrix)
            WRITE (*, *) 'norm(sm+alpha*v*g^t - fm+alpha*v*g^t)/n=', norm/REAL(nao, dp)
            IF (norm/REAL(nao, dp) .GT. 1e-12_dp) THEN
               !WRITE(*,*) 'fm_matrix'
               !DO j=1,SIZE(fm_matrix%local_data,2)
               !   DO i=1,SIZE(fm_matrix%local_data,1)
               !      WRITE(*,'(A,I3,A,I3,A,E26.16,A)') 'a(',i,',',j,')=',fm_matrix%local_data(i,j),';'
               !   ENDDO
               !ENDDO
               !WRITE(*,*) 'mat_v'
               !CALL cp_dbcsr_print(mat_v,matlab_format=.TRUE.)
               !WRITE(*,*) 'mat_g'
               !CALL cp_dbcsr_print(mat_g,matlab_format=.TRUE.)
               !WRITE(*,*) 'sparse_matrix'
               !CALL cp_dbcsr_print(sparse_matrix,matlab_format=.TRUE.)
               !WRITE(*,*) 'sparse_matrix2 (-sm + sparse(fm))'
               !CALL cp_dbcsr_print(sparse_matrix2,matlab_format=.TRUE.)
               !WRITE(*,*) 'sparse_matrix3 (copy of sm input)'
               !CALL cp_dbcsr_print(sparse_matrix3,matlab_format=.TRUE.)
               !stop
            ENDIF
            CALL cp_dbcsr_release(sparse_matrix2)
            CALL cp_dbcsr_release(sparse_matrix3)
            CALL cp_fm_release(fm_matrix)
         ENDIF
         CALL cp_dbcsr_release(mat_v)
         IF (PRESENT(matrix_g)) CALL cp_dbcsr_release(mat_g)
      ENDIF
      CALL timestop(timing_handle)

   END SUBROUTINE cp_dbcsr_plus_fm_fm_t_native

! **************************************************************************************************
!> \brief   Copy a BLACS matrix to a dbcsr matrix.
!>
!>          real_matrix=beta*real_matrix+alpha*fm
!>          beta defaults to 0, alpha to 1
!> \param[in] fm              full matrix
!> \param[out] matrix         DBCSR matrix
!> \param[in] alpha           (optional) scaling of FM
!> \param[in] beta            (optional) scaling of existing SM
!> \param[in] keep_sparsity   (optional) retains the sparsity of the input
!>                            matrix
!> \date    2009-10-13
!> \par History
!>          2009-10-13 rewritten based on copy_dbcsr_to_fm
!> \author  Urban Borstnik
!> \version 2.0
! **************************************************************************************************
   SUBROUTINE copy_fm_to_dbcsr(fm, matrix, alpha, beta, keep_sparsity)
      TYPE(cp_fm_type), POINTER                          :: fm
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      REAL(kind=dp), INTENT(IN), OPTIONAL                :: alpha, beta
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_sparsity

      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_fm_to_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk_p, col, col_l, col_size, group, handle, nblkcols_local, nblkcols_total, &
         nblkrows_local, nblkrows_total, ncol_block, ncol_global, nfullcols_local, &
         nfullcols_total, nfullrows_local, nfullrows_total, nrow_block, nrow_global, nze, row, &
         row_l, row_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_col, first_row, last_col, &
                                                            last_row, local_col_sizes, &
                                                            local_row_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, local_cols, local_rows, &
                                                            row_blk_size
      REAL(kind=dp)                                      :: my_beta
      REAL(KIND=dp), DIMENSION(:), POINTER               :: blk_1d_dp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fm_block
      REAL(KIND=sp), DIMENSION(:), POINTER               :: blk_1d_sp
      REAL(kind=sp), DIMENSION(:, :), POINTER            :: fm_block_sp
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_dbcsr_type)                                :: bc_mat
      TYPE(dbcsr_distribution_obj)                       :: bc_dist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (col_blk_size, row_blk_size)

      my_beta = 0._dp
      IF (PRESENT(beta)) THEN
         IF (beta .NE. my_beta) &
            CALL cp__b("cp_dbcsr_operations.F",688,"beta not supported, use matrix addition instead")
         my_beta = beta
      ENDIF
      group = fm%matrix_struct%para_env%group
      context => fm%matrix_struct%context
      nrow_block = fm%matrix_struct%nrow_block
      ncol_block = fm%matrix_struct%ncol_block
      nrow_global = fm%matrix_struct%nrow_global
      ncol_global = fm%matrix_struct%ncol_global

      CALL cp_dbcsr_get_info(matrix, &
                             nfullrows_total=nfullrows_total, &
                             nfullcols_total=nfullcols_total)

      IF (nrow_global .NE. nfullrows_total) &
         CALL cp__b("cp_dbcsr_operations.F",703,"FM and DBCSR matrix sizes do not match in rows")
      IF (ncol_global .NE. nfullcols_total) &
         CALL cp__b("cp_dbcsr_operations.F",705,"FM and DBCSR matrix sizes do not match in columns")

      ! Create a block-cyclic distribution compatible with the FM matrix.
      CALL dbcsr_distribution_init(bc_dist)
      CALL dbcsr_create_dist_block_cyclic(bc_dist, &
                                          nfullrows_total, nfullcols_total, & ! Actual full matrix size
                                          nrow_block, ncol_block, & ! BLACS parameters
                                          dbcsr_distribution_mp(cp_dbcsr_distribution(matrix)), &
                                          row_blk_size, col_blk_size) ! block-cyclic row/col sizes

      ! Create the block-cyclic DBCSR matrix
      CALL cp_dbcsr_init(bc_mat)
      CALL cp_dbcsr_create(bc_mat, "Block-cyclic "//cp_dbcsr_name(matrix), bc_dist, &
                           cp_dbcsr_get_matrix_type(matrix), row_blk_size, col_blk_size, nze=0, &
                           reuse_arrays=.TRUE., data_type=cp_dbcsr_get_data_type(matrix))

      !call dbcsr_finalize (bc_mat)
      CALL dbcsr_distribution_release(bc_dist)
      NULLIFY (col_blk_size, row_blk_size)

      CALL cp_dbcsr_get_info(bc_mat, &
                             nblkrows_total=nblkrows_total, &
                             nblkcols_total=nblkcols_total, &
                             nblkrows_local=nblkrows_local, &
                             nblkcols_local=nblkcols_local, &
                             nfullrows_local=nfullrows_local, &
                             nfullcols_local=nfullcols_local, &
                             nfullrows_total=nfullrows_total, &
                             nfullcols_total=nfullcols_total, &
                             local_rows=local_rows, &
                             local_cols=local_cols, &
                             row_blk_size=row_blk_size, &
                             col_blk_size=col_blk_size)

      ALLOCATE (local_row_sizes(nblkrows_total))
      ALLOCATE (local_col_sizes(nblkcols_total))
      local_row_sizes(:) = 0
      IF (nblkrows_local .GE. 1) THEN
         DO row = 1, nblkrows_local
            local_row_sizes(local_rows(row)) = row_blk_size(local_rows(row))
         END DO
      ENDIF
      local_col_sizes(:) = 0
      IF (nblkcols_local .GE. 1) THEN
         DO col = 1, nblkcols_local
            local_col_sizes(local_cols(col)) = col_blk_size(local_cols(col))
         END DO
      ENDIF

      ALLOCATE (first_row(nblkrows_total), last_row(nblkrows_total))
      ALLOCATE (first_col(nblkcols_total), last_col(nblkcols_total))
      CALL convert_sizes_to_offsets(local_row_sizes, first_row, last_row)
      CALL convert_sizes_to_offsets(local_col_sizes, first_col, last_col)

      ! Copy the FM data to the block-cyclic DBCSR matrix.  This step
      ! could be skipped with appropriate DBCSR index manipulation.
      fm_block => fm%local_data
      fm_block_sp => fm%local_data_sp

      CALL cp_dbcsr_work_create(bc_mat, nblks_guess=nblkrows_local*nblkcols_local, &
                                sizedata_guess=nfullrows_local*nfullcols_local, work_mutable=.FALSE., &
                                n=1)
      blk_p = 1
      bc_rows: DO row_l = 1, nblkrows_local
         row = local_rows(row_l)
         row_size = row_blk_size(row)
         bc_cols: DO col_l = 1, nblkcols_local
            col = local_cols(col_l)
            col_size = col_blk_size(col)
            nze = row_size*col_size
            !WRITE(*,*)routineN//" Adding block",row,col,"size",nze
            CALL cp_add_work_coordinate(bc_mat, 1, row, col, blk_p)
            IF (fm%use_sp) THEN
               !blk_1d_sp => bc_mat%m%wms(1)%data_area%d%r_sp(blk_p:blk_p+nze-1)
               blk_1d_sp => cp_dbcsr_get_data_p(bc_mat, 1, &
                                                select_data_type=REAL(0.0, KIND=sp), lb=blk_p, ub=blk_p+nze-1)
            ELSE
               !blk_1d_dp => bc_mat%m%wms(1)%data_area%d%r_dp(blk_p:blk_p+nze-1)
               blk_1d_dp => cp_dbcsr_get_data_p(bc_mat, 1, &
                                                select_data_type=REAL(0.0, KIND=dp), lb=blk_p, ub=blk_p+nze-1)
            ENDIF
            IF (nze .NE. (last_row(row)-first_row(row)+1)*(last_col(col)-first_col(col)+1)) &
               CALL cp__b("cp_dbcsr_operations.F",787,"Block size does not match block row/col sizes")
            IF (fm%use_sp) THEN
               blk_1d_sp(1:nze) = RESHAPE( &
                                  fm_block_sp( &
                                  first_row(row):last_row(row), first_col(col):last_col(col) &
                                  ), (/nze/))
            ELSE
               blk_1d_dp(1:nze) = RESHAPE( &
                                  fm_block( &
                                  first_row(row):last_row(row), first_col(col):last_col(col) &
                                  ), (/nze/))
            ENDIF
            blk_p = blk_p+nze
         ENDDO bc_cols
      ENDDO bc_rows
      CALL cp_set_work_size(bc_mat, 1, blk_p-1)
      CALL cp_dbcsr_finalize(bc_mat, reshuffle=.FALSE.)

      ! Now convert to the desired matrix distribution
      IF (PRESENT(alpha)) THEN
         CALL cp__b("cp_dbcsr_operations.F",807,'no more alpha... clean me')
      ELSE
         CALL cp_dbcsr_complete_redistribute(bc_mat, matrix, &
                                             keep_sparsity=keep_sparsity)
      ENDIF
      CALL cp_dbcsr_release(bc_mat)

      CALL timestop(handle)
   END SUBROUTINE copy_fm_to_dbcsr

! **************************************************************************************************
!> \brief ...
!> \param fm ...
!> \param bc_mat ...
!> \param beta ...
! **************************************************************************************************
   SUBROUTINE copy_fm_to_dbcsr_bc(fm, bc_mat, beta)
      TYPE(cp_fm_type), POINTER                          :: fm
      TYPE(cp_dbcsr_type)                                :: bc_mat
      REAL(kind=dp), INTENT(IN), OPTIONAL                :: beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_fm_to_dbcsr_bc', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk_p, col, col_l, col_size, data_type, group, handle, my_blk, mypcol, myproc, &
         myprow, nblkcols_local, nblkcols_total, nblkrows_local, nblkrows_total, ncol_block, &
         ncol_global, nfullcols_local, nfullcols_total, nfullrows_local, nfullrows_total, &
         nrow_block, nrow_global, numproc, nze, row, row_l, row_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_col, first_row, last_col, &
                                                            last_row, local_col_sizes, &
                                                            local_row_sizes
      INTEGER, DIMENSION(:), POINTER                     :: blk_map, col_blk_size, local_cols, &
                                                            local_rows, row_blk_size
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid
      REAL(kind=dp)                                      :: my_beta
      REAL(KIND=dp), DIMENSION(:), POINTER               :: blk_1d_dp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fm_block
      REAL(KIND=sp), DIMENSION(:), POINTER               :: blk_1d_sp
      REAL(kind=sp), DIMENSION(:, :), POINTER            :: fm_block_sp
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(dbcsr_distribution_obj)                       :: bc_dist
      TYPE(dbcsr_mp_obj)                                 :: mp_env

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (col_blk_size, row_blk_size)

      my_beta = 0._dp
      IF (PRESENT(beta)) THEN
         IF (beta .NE. my_beta) &
            CALL cp__b("cp_dbcsr_operations.F",859,"beta not supported, use matrix addition instead")
         my_beta = beta
      ENDIF
      group = fm%matrix_struct%para_env%group
      context => fm%matrix_struct%context
      nrow_block = fm%matrix_struct%nrow_block
      ncol_block = fm%matrix_struct%ncol_block
      nrow_global = fm%matrix_struct%nrow_global
      ncol_global = fm%matrix_struct%ncol_global
      myproc = fm%matrix_struct%para_env%mepos
      numproc = fm%matrix_struct%para_env%num_pe
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      pgrid => context%blacs2mpi

      nfullrows_total = nrow_global
      nfullcols_total = ncol_global
      CALL cp_dbcsr_mp_new(mp_env, pgrid, group, myproc, numproc, &
                           myprow, mypcol)

      ! Create a block-cyclic distribution compatible with the FM matrix.
      CALL dbcsr_distribution_init(bc_dist)
      CALL dbcsr_create_dist_block_cyclic(bc_dist, &
                                          nfullrows_total, nfullcols_total, & ! Actual full matrix size
                                          nrow_block, ncol_block, & ! BLACS parameters
                                          mp_env, &
                                          row_blk_size, col_blk_size) ! block-cyclic row/col sizes

      CALL cp_dbcsr_mp_release(mp_env)

      ! Create the block-cyclic DBCSR matrix
      data_type = dbcsr_type_real_8
      IF (fm%use_sp) data_type = dbcsr_type_real_4
      CALL cp_dbcsr_init(bc_mat)
      CALL cp_dbcsr_create(bc_mat, "Block-cyclic ", bc_dist, &
                           dbcsr_type_no_symmetry, row_blk_size, col_blk_size, nze=0, &
                           reuse_arrays=.TRUE., data_type=data_type)

      !call dbcsr_finalize (bc_mat)
      CALL dbcsr_distribution_release(bc_dist)
      NULLIFY (col_blk_size, row_blk_size)

      CALL cp_dbcsr_get_info(bc_mat, &
                             nblkrows_total=nblkrows_total, &
                             nblkcols_total=nblkcols_total, &
                             nblkrows_local=nblkrows_local, &
                             nblkcols_local=nblkcols_local, &
                             nfullrows_local=nfullrows_local, &
                             nfullcols_local=nfullcols_local, &
                             nfullrows_total=nfullrows_total, &
                             nfullcols_total=nfullcols_total, &
                             local_rows=local_rows, &
                             local_cols=local_cols, &
                             row_blk_size=row_blk_size, &
                             col_blk_size=col_blk_size)

      ALLOCATE (local_row_sizes(nblkrows_total))
      ALLOCATE (local_col_sizes(nblkcols_total))
      local_row_sizes(:) = 0
      IF (nblkrows_local .GE. 1) THEN
         DO row = 1, nblkrows_local
            local_row_sizes(local_rows(row)) = row_blk_size(local_rows(row))
         END DO
      ENDIF
      local_col_sizes(:) = 0
      IF (nblkcols_local .GE. 1) THEN
         DO col = 1, nblkcols_local
            local_col_sizes(local_cols(col)) = col_blk_size(local_cols(col))
         END DO
      ENDIF

      ALLOCATE (first_row(nblkrows_total), last_row(nblkrows_total))
      ALLOCATE (first_col(nblkcols_total), last_col(nblkcols_total))
      ALLOCATE (blk_map(0:nblkrows_total*nblkcols_total))
      blk_map(0) = 1
      CALL convert_sizes_to_offsets(local_row_sizes, first_row, last_row)
      CALL convert_sizes_to_offsets(local_col_sizes, first_col, last_col)

      ! Copy the FM data to the block-cyclic DBCSR matrix.  This step
      ! could be skipped with appropriate DBCSR index manipulation.
      fm_block => fm%local_data
      fm_block_sp => fm%local_data_sp

      CALL cp_dbcsr_work_create(bc_mat, nblks_guess=nblkrows_local*nblkcols_local, &
                                sizedata_guess=nfullrows_local*nfullcols_local, work_mutable=.FALSE., &
                                n=1)
      blk_p = 1
      bc_rows: DO row_l = 1, nblkrows_local
         row = local_rows(row_l)
         row_size = row_blk_size(row)
         bc_cols: DO col_l = 1, nblkcols_local
            col = local_cols(col_l)
            col_size = col_blk_size(col)
            my_blk = (row_l-1)*nblkcols_local+col_l-1
            nze = row_size*col_size
            CALL cp_add_work_coordinate(bc_mat, 1, row, col, blk_map(my_blk))
            blk_map(my_blk+1) = blk_map(my_blk)+nze
         ENDDO bc_cols
      ENDDO bc_rows

!$OMP PARALLEL DO DEFAULT(NONE) PRIVATE(my_blk,nze,row_l,col_l,row,col,blk_1d_dp,blk_1d_sp) &
!$OMP SHARED(local_rows, nblkrows_local, nblkcols_local, local_cols, blk_map, fm, bc_mat,&
!$OMP        last_row, first_row, last_col, first_col, fm_block, fm_block_sp)
      bc_rows1: DO row_l = 1, nblkrows_local
         row = local_rows(row_l)
         bc_cols1: DO col_l = 1, nblkcols_local
            col = local_cols(col_l)
            my_blk = (row_l-1)*nblkcols_local+col_l-1
            nze = blk_map(my_blk+1)-blk_map(my_blk)
            IF (fm%use_sp) THEN
               !blk_1d_sp => bc_mat%m%wms(1)%data_area%d%r_sp(blk_p:blk_p+nze-1)
               blk_1d_sp => cp_dbcsr_get_data_p(bc_mat, 1, &
                                                select_data_type=REAL(0.0, KIND=sp), lb=blk_map(my_blk), ub=blk_map(my_blk)+nze-1)
            ELSE
               !blk_1d_dp => bc_mat%m%wms(1)%data_area%d%r_dp(blk_p:blk_p+nze-1)
               blk_1d_dp => cp_dbcsr_get_data_p(bc_mat, 1, &
                                                select_data_type=REAL(0.0, KIND=dp), lb=blk_map(my_blk), ub=blk_map(my_blk)+nze-1)
            ENDIF
            IF (nze .NE. (last_row(row)-first_row(row)+1)*(last_col(col)-first_col(col)+1)) &
               CALL cp__b("cp_dbcsr_operations.F",978,"Block size does not match block row/col sizes")
            IF (fm%use_sp) THEN
               blk_1d_sp(1:nze) = RESHAPE( &
                                  fm_block_sp( &
                                  first_row(row):last_row(row), first_col(col):last_col(col) &
                                  ), (/nze/))
            ELSE
               blk_1d_dp(1:nze) = RESHAPE( &
                                  fm_block( &
                                  first_row(row):last_row(row), first_col(col):last_col(col) &
                                  ), (/nze/))
            ENDIF
         ENDDO bc_cols1
      ENDDO bc_rows1
!$OMP END PARALLEL DO
      CALL cp_set_work_size(bc_mat, 1, blk_map(nblkrows_local*nblkcols_local)-1)
      CALL cp_dbcsr_finalize(bc_mat, reshuffle=.FALSE.)
      DEALLOCATE (blk_map)

      CALL timestop(handle)
   END SUBROUTINE copy_fm_to_dbcsr_bc
! **************************************************************************************************
!> \brief   Copy a BLACS matrix to a dbcsr matrix.
!>
!>          real_matrix=beta*real_matrix+alpha*fm
!>          beta defaults to 0, alpha to 1
!> \param[in] fm              full matrix
!> \param[out] matrix         DBCSR matrix
!> \param[in] keep_sparsity   (optional) retains the sparsity of the input
!>                            matrix
!> \date    2010
!> \par History
!>          2010  copied from copy_dbcsr_to_fm
!> \author  VW
!> \version 2.0
! **************************************************************************************************
   SUBROUTINE copy_cfm_to_dbcsr(fm, matrix, keep_sparsity)
      TYPE(cp_cfm_type), POINTER                         :: fm
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_sparsity

      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_cfm_to_dbcsr', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: blk_1d
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: fm_block
      INTEGER :: blk_p, col, col_l, col_size, group, handle, nblkcols_local, nblkcols_total, &
         nblkrows_local, nblkrows_total, ncol_block, ncol_global, nfullcols_local, &
         nfullcols_total, nfullrows_local, nfullrows_total, nrow_block, nrow_global, nze, row, &
         row_l, row_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_col, first_row, last_col, &
                                                            last_row, local_col_sizes, &
                                                            local_row_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, local_cols, local_rows, &
                                                            row_blk_size
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_dbcsr_type)                                :: bc_mat
      TYPE(dbcsr_distribution_obj)                       :: bc_dist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (col_blk_size, row_blk_size)

      group = fm%matrix_struct%para_env%group
      context => fm%matrix_struct%context
      nrow_block = fm%matrix_struct%nrow_block
      ncol_block = fm%matrix_struct%ncol_block
      nrow_global = fm%matrix_struct%nrow_global
      ncol_global = fm%matrix_struct%ncol_global

      CALL cp_dbcsr_get_info(matrix, &
                             nfullrows_total=nfullrows_total, &
                             nfullcols_total=nfullcols_total)

      IF (nrow_global .NE. nfullrows_total) &
         CALL cp__b("cp_dbcsr_operations.F",1055,"FM and DBCSR matrix sizes do not match in rows")
      IF (ncol_global .NE. nfullcols_total) &
         CALL cp__b("cp_dbcsr_operations.F",1057,"FM and DBCSR matrix sizes do not match in columns")

      ! Create a block-cyclic distribution compatible with the FM matrix.
      CALL dbcsr_distribution_init(bc_dist)
      CALL dbcsr_create_dist_block_cyclic(bc_dist, &
                                          nfullrows_total, nfullcols_total, & ! Actual full matrix size
                                          nrow_block, ncol_block, & ! BLACS parameters
                                          dbcsr_distribution_mp(cp_dbcsr_distribution(matrix)), &
                                          row_blk_size, col_blk_size) ! block-cyclic row/col sizes

      ! Create the block-cyclic DBCSR matrix
      CALL cp_dbcsr_init(bc_mat)
      CALL cp_dbcsr_create(bc_mat, "Block-cyclic "//cp_dbcsr_name(matrix), bc_dist, &
                           cp_dbcsr_get_matrix_type(matrix), row_blk_size, col_blk_size, nze=0, &
                           data_type=dbcsr_type_complex_8, reuse_arrays=.TRUE.) ! type hard coded !
      !call dbcsr_finalize (bc_mat)
      CALL dbcsr_distribution_release(bc_dist)
      NULLIFY (col_blk_size, row_blk_size)

      CALL cp_dbcsr_get_info(bc_mat, &
                             nblkrows_total=nblkrows_total, &
                             nblkcols_total=nblkcols_total, &
                             nblkrows_local=nblkrows_local, &
                             nblkcols_local=nblkcols_local, &
                             nfullrows_local=nfullrows_local, &
                             nfullcols_local=nfullcols_local, &
                             nfullrows_total=nfullrows_total, &
                             nfullcols_total=nfullcols_total, &
                             local_rows=local_rows, &
                             local_cols=local_cols, &
                             row_blk_size=row_blk_size, &
                             col_blk_size=col_blk_size)

      ALLOCATE (local_row_sizes(nblkrows_total))
      ALLOCATE (local_col_sizes(nblkcols_total))
      local_row_sizes(:) = 0
      IF (nblkrows_local .GE. 1) THEN
         DO row = 1, nblkrows_local
            local_row_sizes(local_rows(row)) = row_blk_size(local_rows(row))
         END DO
      ENDIF
      local_col_sizes(:) = 0
      IF (nblkcols_local .GE. 1) THEN
         DO col = 1, nblkcols_local
            local_col_sizes(local_cols(col)) = col_blk_size(local_cols(col))
         END DO
      ENDIF

      ALLOCATE (first_row(nblkrows_total), last_row(nblkrows_total))
      ALLOCATE (first_col(nblkcols_total), last_col(nblkcols_total))
      CALL convert_sizes_to_offsets(local_row_sizes, first_row, last_row)
      CALL convert_sizes_to_offsets(local_col_sizes, first_col, last_col)

      ! Copy the FM data to the block-cyclic DBCSR matrix.  This step
      ! could be skipped with appropriate DBCSR index manipulation.
      fm_block => fm%local_data
      CALL cp_dbcsr_work_create(bc_mat, nblks_guess=nblkrows_local*nblkcols_local, &
                                sizedata_guess=nfullrows_local*nfullcols_local, work_mutable=.FALSE., &
                                n=1)
      blk_p = 1
      bc_rows: DO row_l = 1, nblkrows_local
         row = local_rows(row_l)
         row_size = row_blk_size(row)
         bc_cols: DO col_l = 1, nblkcols_local
            col = local_cols(col_l)
            col_size = col_blk_size(col)
            nze = row_size*col_size
            !WRITE(*,*)routineN//" Adding block",row,col,"size",nze
            CALL cp_add_work_coordinate(bc_mat, 1, row, col, blk_p)
            !blk_1d => bc_mat%m%wms(1)%data_area%d%c_dp(blk_p:blk_p+nze-1)
            blk_1d => cp_dbcsr_get_data_p(bc_mat, 1, &
                                          select_data_type=CMPLX(0.0, KIND=dp), lb=blk_p, ub=blk_p+nze-1)
            IF (nze .NE. (last_row(row)-first_row(row)+1)*(last_col(col)-first_col(col)+1)) &
               CALL cp__b("cp_dbcsr_operations.F",1130,"Block size does not match block row/col sizes")
            blk_1d(1:nze) = RESHAPE( &
                            fm_block( &
                            first_row(row):last_row(row), first_col(col):last_col(col) &
                            ), (/nze/))
            blk_p = blk_p+nze
         ENDDO bc_cols
      ENDDO bc_rows
      CALL cp_set_work_size(bc_mat, 1, blk_p-1)
      CALL cp_dbcsr_finalize(bc_mat, reshuffle=.FALSE.)

      ! Now convert to the desired matrix distribution
      CALL cp_dbcsr_complete_redistribute(bc_mat, matrix, keep_sparsity=keep_sparsity)
      CALL cp_dbcsr_release(bc_mat)

      CALL timestop(handle)

   END SUBROUTINE copy_cfm_to_dbcsr
! **************************************************************************************************
!> \brief   Copy a DBCSR matrix to a BLACS matrix
!> \param[in] matrix          DBCSR matrix
!> \param[out] fm             full matrix
! **************************************************************************************************
   SUBROUTINE copy_dbcsr_to_fm(matrix, fm)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(cp_fm_type), POINTER                          :: fm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_dbcsr_to_fm', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER :: col, group, handle, mypcol, mype, myprow, nblkcols_local, nblkcols_total, &
         nblkrows_local, nblkrows_total, ncol_block, ncol_global, nfullcols_total, &
         nfullrows_total, npcol, npe, nprow, nrow_block, nrow_global, row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_col, first_row, last_col, &
                                                            last_row, local_col_sizes, &
                                                            local_row_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, col_blk_size_tmp, &
                                                            local_cols, local_rows, row_blk_size, &
                                                            row_blk_size_tmp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: blk_2d, fm_block
      REAL(KIND=sp), DIMENSION(:, :), POINTER            :: fm_block_sp
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: bc_mat
      TYPE(dbcsr_distribution_obj)                       :: bc_dist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (col_blk_size, row_blk_size)

      ! info about the full matrix
      group = fm%matrix_struct%para_env%group
      context => fm%matrix_struct%context
      mype = context%my_pid
      npe = context%n_pid
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)
      nrow_block = fm%matrix_struct%nrow_block
      ncol_block = fm%matrix_struct%ncol_block
      nrow_global = fm%matrix_struct%nrow_global
      ncol_global = fm%matrix_struct%ncol_global

      CALL cp_dbcsr_get_info(matrix, &
                             nfullrows_total=nfullrows_total, &
                             nfullcols_total=nfullcols_total)

      ! Convert DBCSR to a block-cyclic one
      CALL dbcsr_distribution_init(bc_dist)
      CALL dbcsr_create_dist_block_cyclic(bc_dist, &
                                          nfullrows_total, nfullcols_total, &
                                          nrow_block, ncol_block, &
                                          dbcsr_distribution_mp(cp_dbcsr_distribution(matrix)), &
                                          row_blk_size, col_blk_size)

      CALL cp_dbcsr_init(bc_mat)
      row_blk_size_tmp => row_blk_size
      col_blk_size_tmp => col_blk_size
      CALL cp_dbcsr_create(bc_mat, "Block-cyclic"//cp_dbcsr_name(matrix), bc_dist, &
                           dbcsr_type_no_symmetry, row_blk_size_tmp, col_blk_size_tmp, nze=0, &
                           reuse_arrays=.TRUE.)
      CALL dbcsr_distribution_release(bc_dist)
      CALL cp_dbcsr_complete_redistribute(matrix, bc_mat)

      ! Find the local extents of the local blocked rows so that index lookups
      ! into the FM matrix work correctly.
      local_rows => dbcsr_distribution_local_rows(bc_dist)
      local_cols => dbcsr_distribution_local_cols(bc_dist)
      ALLOCATE (local_row_sizes(dbcsr_distribution_nrows(bc_dist)))
      ALLOCATE (local_col_sizes(dbcsr_distribution_ncols(bc_dist)))
      nblkrows_local = dbcsr_distribution_nlocal_rows(bc_dist)
      nblkcols_local = dbcsr_distribution_nlocal_cols(bc_dist)
      local_row_sizes(:) = 0
      IF (nblkrows_local .GE. 1) THEN
         DO row = 1, nblkrows_local
            local_row_sizes(local_rows(row)) = row_blk_size(local_rows(row))
         END DO
      ENDIF
      local_col_sizes(:) = 0
      IF (nblkcols_local .GE. 1) THEN
         DO col = 1, nblkcols_local
            local_col_sizes(local_cols(col)) = col_blk_size(local_cols(col))
         END DO
      ENDIF
      nblkrows_total = dbcsr_distribution_nrows(bc_dist)
      nblkcols_total = dbcsr_distribution_ncols(bc_dist)
      ALLOCATE (first_row(nblkrows_total), last_row(nblkrows_total))
      ALLOCATE (first_col(nblkcols_total), last_col(nblkcols_total))
      CALL convert_sizes_to_offsets(local_row_sizes, first_row, last_row)
      CALL convert_sizes_to_offsets(local_col_sizes, first_col, last_col)
      !
      ! Now copy data to the FM matrix
      fm_block => fm%local_data
      fm_block_sp => fm%local_data_sp
      IF (fm%use_sp) THEN
         fm_block_sp = 0.0_sp
      ELSE
         fm_block = 0.0_dp
      ENDIF

      IF (dbg) THEN
         WRITE (*, *) routineN//" FM data size is", UBOUND(fm_block)
         WRITE (*, *) routineN//" dbcsr data size is", cp_dbcsr_get_data_size(bc_mat)
         WRITE (*, *) routineN//" FM block sizes are", nrow_block, '/', nfullrows_total
         WRITE (*, *) routineN//" FM block sizes are", ncol_block, '/', nfullcols_total
         WRITE (*, *) routineN//" dbcsr row sizes are", cp_dbcsr_row_block_sizes(bc_mat)
         WRITE (*, *) routineN//" dbcsr col sizes are", cp_dbcsr_col_block_sizes(bc_mat)
      ENDIF
      !
      CALL cp_dbcsr_iterator_start(iter, bc_mat)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, blk_2d)
         ! Convert absolute coordinates to FM-local coordinates
         IF (fm%use_sp) THEN
            fm_block_sp(first_row(row):last_row(row), first_col(col):last_col(col)) &
               = REAL(blk_2d(:, :), sp)
         ELSE
            IF (dbg) THEN
               WRITE (*, *) routineN//" blk2d size", UBOUND(blk_2d)
               WRITE (*, *) routineN//" want to set coor.", row, col
               WRITE (*, *) routineN//" local extents", &
                  first_row(row), last_row(row), first_col(col), last_col(col)
            ENDIF
            fm_block(first_row(row):last_row(row), first_col(col):last_col(col)) &
               = blk_2d(:, :)
         ENDIF
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL cp_dbcsr_release(bc_mat)

      CALL timestop(handle)
   END SUBROUTINE copy_dbcsr_to_fm

! **************************************************************************************************
!> \brief   Copy a DBCSR matrix to a BLACS matrix
!> \param[in] matrix          DBCSR matrix
!> \param[out] fm             full matrix
! **************************************************************************************************
   SUBROUTINE copy_dbcsr_to_cfm(matrix, fm)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix
      TYPE(cp_cfm_type), POINTER                         :: fm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_dbcsr_to_cfm', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: blk_2d, fm_block
      INTEGER :: col, group, handle, mypcol, mype, myprow, nblkcols_local, nblkcols_total, &
         nblkrows_local, nblkrows_total, ncol_block, ncol_global, nfullcols_total, &
         nfullrows_total, npcol, npe, nprow, nrow_block, nrow_global, row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_col, first_row, last_col, &
                                                            last_row, local_col_sizes, &
                                                            local_row_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, col_blk_size_tmp, &
                                                            local_cols, local_rows, row_blk_size, &
                                                            row_blk_size_tmp
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: bc_mat
      TYPE(dbcsr_distribution_obj)                       :: bc_dist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (col_blk_size, row_blk_size)

      ! info about the full matrix
      group = fm%matrix_struct%para_env%group
      context => fm%matrix_struct%context
      mype = context%my_pid
      npe = context%n_pid
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)
      nrow_block = fm%matrix_struct%nrow_block
      ncol_block = fm%matrix_struct%ncol_block
      nrow_global = fm%matrix_struct%nrow_global
      ncol_global = fm%matrix_struct%ncol_global

      CALL cp_dbcsr_get_info(matrix, &
                             nfullrows_total=nfullrows_total, &
                             nfullcols_total=nfullcols_total)

      ! Convert DBCSR to a block-cyclic one
      CALL dbcsr_distribution_init(bc_dist)
      CALL dbcsr_create_dist_block_cyclic(bc_dist, &
                                          nfullrows_total, nfullcols_total, &
                                          nrow_block, ncol_block, &
                                          dbcsr_distribution_mp(cp_dbcsr_distribution(matrix)), &
                                          row_blk_size, col_blk_size)

      CALL cp_dbcsr_init(bc_mat)
      row_blk_size_tmp => row_blk_size
      col_blk_size_tmp => col_blk_size
      CALL cp_dbcsr_create(bc_mat, "Block-cyclic"//cp_dbcsr_name(matrix), bc_dist, &
                           dbcsr_type_no_symmetry, row_blk_size_tmp, col_blk_size_tmp, &
                           nze=0, data_type=cp_dbcsr_get_data_type(matrix), &
                           reuse_arrays=.TRUE.)
      CALL dbcsr_distribution_release(bc_dist)
      CALL cp_dbcsr_complete_redistribute(matrix, bc_mat)

      ! Find the local extents of the local blocked rows so that index lookups
      ! into the FM matrix work correctly.
      local_rows => dbcsr_distribution_local_rows(bc_dist)
      local_cols => dbcsr_distribution_local_cols(bc_dist)
      ALLOCATE (local_row_sizes(dbcsr_distribution_nrows(bc_dist)))
      ALLOCATE (local_col_sizes(dbcsr_distribution_ncols(bc_dist)))
      nblkrows_local = dbcsr_distribution_nlocal_rows(bc_dist)
      nblkcols_local = dbcsr_distribution_nlocal_cols(bc_dist)
      local_row_sizes(:) = 0
      IF (nblkrows_local .GE. 1) THEN
         DO row = 1, nblkrows_local
            local_row_sizes(local_rows(row)) = row_blk_size(local_rows(row))
         END DO
      ENDIF
      local_col_sizes(:) = 0
      IF (nblkcols_local .GE. 1) THEN
         DO col = 1, nblkcols_local
            local_col_sizes(local_cols(col)) = col_blk_size(local_cols(col))
         END DO
      ENDIF
      nblkrows_total = dbcsr_distribution_nrows(bc_dist)
      nblkcols_total = dbcsr_distribution_ncols(bc_dist)
      ALLOCATE (first_row(nblkrows_total), last_row(nblkrows_total))
      ALLOCATE (first_col(nblkcols_total), last_col(nblkcols_total))
      CALL convert_sizes_to_offsets(local_row_sizes, first_row, last_row)
      CALL convert_sizes_to_offsets(local_col_sizes, first_col, last_col)
      !
      ! Now copy data to the FM matrix
      fm_block => fm%local_data
      fm_block = (0.0_dp, 0.0_dp)

      IF (dbg) THEN
         WRITE (*, *) routineN//" FM data size is", UBOUND(fm_block)
         WRITE (*, *) routineN//" dbcsr data size is", cp_dbcsr_get_data_size(bc_mat)
         WRITE (*, *) routineN//" FM block sizes are", nrow_block, '/', nfullrows_total
         WRITE (*, *) routineN//" FM block sizes are", ncol_block, '/', nfullcols_total
         WRITE (*, *) routineN//" dbcsr row sizes are", cp_dbcsr_row_block_sizes(bc_mat)
         WRITE (*, *) routineN//" dbcsr col sizes are", cp_dbcsr_col_block_sizes(bc_mat)
      ENDIF
      !
      CALL cp_dbcsr_iterator_start(iter, bc_mat)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, blk_2d)
         ! Convert absolute coordinates to FM-local coordinates
         IF (dbg) THEN
            WRITE (*, *) routineN//" blk2d size", UBOUND(blk_2d)
            WRITE (*, *) routineN//" want to set coor.", row, col
            WRITE (*, *) routineN//" local extents", &
               first_row(row), last_row(row), first_col(col), last_col(col)
         ENDIF
         fm_block(first_row(row):last_row(row), first_col(col):last_col(col)) &
            = blk_2d(:, :)
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL cp_dbcsr_release(bc_mat)

      CALL timestop(handle)
   END SUBROUTINE copy_dbcsr_to_cfm
! **************************************************************************************************
!> \brief   Copy a DBCSR_BLACS matrix to a BLACS matrix
!> \param bc_mat DBCSR matrix
!> \param[out] fm             full matrix
! **************************************************************************************************
   SUBROUTINE copy_dbcsr_to_fm_bc(bc_mat, fm)
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: bc_mat
      TYPE(cp_fm_type), POINTER                          :: fm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_dbcsr_to_fm_bc', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER :: col, group, handle, mypcol, mype, myprow, nblkcols_local, nblkcols_total, &
         nblkrows_local, nblkrows_total, ncol_block, ncol_global, nfullcols_total, &
         nfullrows_total, npcol, npe, nprow, nrow_block, nrow_global, row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_col, first_row, last_col, &
                                                            last_row, local_col_sizes, &
                                                            local_row_sizes
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, local_cols, local_rows, &
                                                            row_blk_size
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: blk_2d, fm_block
      REAL(KIND=sp), DIMENSION(:, :), POINTER            :: blk_2d_sp, fm_block_sp
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(dbcsr_distribution_obj)                       :: bc_dist

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (col_blk_size, row_blk_size)

      ! info about the full matrix
      group = fm%matrix_struct%para_env%group
      context => fm%matrix_struct%context
      mype = context%my_pid
      npe = context%n_pid
      myprow = context%mepos(1); mypcol = context%mepos(2)
      nprow = context%num_pe(1); npcol = context%num_pe(2)
      nrow_block = fm%matrix_struct%nrow_block; ncol_block = fm%matrix_struct%ncol_block
      nrow_global = fm%matrix_struct%nrow_global; ncol_global = fm%matrix_struct%ncol_global
      nfullrows_total = nrow_global; nfullcols_total = ncol_global

      ! Convert DBCSR to a block-cyclic one
      CALL dbcsr_distribution_init(bc_dist)
      CALL dbcsr_create_dist_block_cyclic(bc_dist, &
                                          nfullrows_total, nfullcols_total, &
                                          nrow_block, ncol_block, &
                                          dbcsr_distribution_mp(cp_dbcsr_distribution(bc_mat)), &
                                          row_blk_size, col_blk_size)

      ! Find the local extents of the local blocked rows so that index lookups
      ! into the FM matrix work correctly.
      local_rows => dbcsr_distribution_local_rows(bc_dist)
      local_cols => dbcsr_distribution_local_cols(bc_dist)
      ALLOCATE (local_row_sizes(dbcsr_distribution_nrows(bc_dist)))
      ALLOCATE (local_col_sizes(dbcsr_distribution_ncols(bc_dist)))
      nblkrows_local = dbcsr_distribution_nlocal_rows(bc_dist)
      nblkcols_local = dbcsr_distribution_nlocal_cols(bc_dist)
      local_row_sizes(:) = 0
      IF (nblkrows_local .GE. 1) THEN
         DO row = 1, nblkrows_local
            local_row_sizes(local_rows(row)) = row_blk_size(local_rows(row))
         END DO
      ENDIF
      local_col_sizes(:) = 0
      IF (nblkcols_local .GE. 1) THEN
         DO col = 1, nblkcols_local
            local_col_sizes(local_cols(col)) = col_blk_size(local_cols(col))
         END DO
      ENDIF
      nblkrows_total = dbcsr_distribution_nrows(bc_dist)
      nblkcols_total = dbcsr_distribution_ncols(bc_dist)
      ALLOCATE (first_row(nblkrows_total), last_row(nblkrows_total))
      ALLOCATE (first_col(nblkcols_total), last_col(nblkcols_total))
      CALL convert_sizes_to_offsets(local_row_sizes, first_row, last_row)
      CALL convert_sizes_to_offsets(local_col_sizes, first_col, last_col)
      !
      ! Now copy data to the FM matrix
      fm_block => fm%local_data
      fm_block_sp => fm%local_data_sp
      IF (fm%use_sp) THEN
         fm_block_sp = 0.0_sp
      ELSE
         fm_block = 0.0_dp
      ENDIF

      IF (dbg) THEN
         WRITE (*, *) routineN//" FM data size is", UBOUND(fm_block_sp), cp_dbcsr_get_data_type(bc_mat)
         WRITE (*, *) routineN//" dbcsr data size is", cp_dbcsr_get_data_size(bc_mat)
         WRITE (*, *) routineN//" FM block sizes are", nrow_block, '/', nfullrows_total
         WRITE (*, *) routineN//" FM block sizes are", ncol_block, '/', nfullcols_total
         WRITE (*, *) routineN//" dbcsr row sizes are", cp_dbcsr_row_block_sizes(bc_mat)
         WRITE (*, *) routineN//" dbcsr col sizes are", cp_dbcsr_col_block_sizes(bc_mat)
      ENDIF
      !
      CALL cp_dbcsr_iterator_start(iter, bc_mat)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         IF (cp_dbcsr_get_data_type(bc_mat) == dbcsr_type_real_4) THEN
            CALL cp_dbcsr_iterator_next_block(iter, row, col, blk_2d_sp)
            ! Convert absolute coordinates to FM-local coordinates
            IF (fm%use_sp) THEN
               fm_block_sp(first_row(row):last_row(row), first_col(col):last_col(col)) &
                  = blk_2d_sp(:, :)
            ELSE
               fm_block(first_row(row):last_row(row), first_col(col):last_col(col)) &
                  = REAL(blk_2d_sp(:, :), dp)
            END IF
         ELSE
            CALL cp_dbcsr_iterator_next_block(iter, row, col, blk_2d)
            IF (fm%use_sp) THEN
               fm_block_sp(first_row(row):last_row(row), first_col(col):last_col(col)) &
                  = REAL(blk_2d_sp(:, :), sp)
            ELSE
               fm_block(first_row(row):last_row(row), first_col(col):last_col(col)) &
                  = blk_2d(:, :)
            END IF
         ENDIF
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)
      DEALLOCATE (row_blk_size, col_blk_size)

      CALL dbcsr_distribution_release(bc_dist)

      CALL timestop(handle)
   END SUBROUTINE copy_dbcsr_to_fm_bc

! **************************************************************************************************
!> \brief Utility function to copy a specially shaped fm to dbcsr_matrix
!>        The result matrix will be the matrix in dbcsr format
!>        with the row blocks sizes according to the block_sizes of the template
!>        and the col blocks sizes evenly blocked with the internal dbcsr conversion
!>        size (32 is the current default)
!> \param matrix ...
!> \param fm_in ...
!> \param template ...
! **************************************************************************************************
   SUBROUTINE cp_fm_to_dbcsr_row_template(matrix, fm_in, template)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix
      TYPE(cp_fm_type), POINTER                          :: fm_in
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: template

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_to_dbcsr_row_template', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k_in
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size_right_in
      TYPE(dbcsr_distribution_obj)                       :: dist_right_in

      CALL cp_fm_get_info(fm_in, ncol_global=k_in)

      CALL dbcsr_create_dist_r_unrot(dist_right_in, cp_dbcsr_distribution(template), k_in, &
                                     col_blk_size_right_in)
      CALL cp_dbcsr_init(matrix)
      CALL cp_dbcsr_create(matrix, "D", dist_right_in, dbcsr_type_no_symmetry, &
                           cp_dbcsr_row_block_sizes(template), col_blk_size_right_in, &
                           nze=0, data_type=cp_dbcsr_get_data_type(template))

      CALL copy_fm_to_dbcsr(fm_in, matrix)
      DEALLOCATE (col_blk_size_right_in)
      CALL dbcsr_distribution_release(dist_right_in)

   END SUBROUTINE cp_fm_to_dbcsr_row_template

! **************************************************************************************************
!> \brief Utility function to create an arbitrary shaped dbcsr matrix
!>        with the same processor grid as the template matrix
!>        both row sizes and col sizes are evenly blocked with the internal
!>        dbcsr_conversion size (32 is the current default)
!> \param matrix dbcsr matrix to be created
!> \param template template dbcsr matrix giving its mp_env
!> \param m global row size of output matrix
!> \param n global col size of output matrix
!> \param sym ...
!> \param data_type ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_m_by_n_from_template(matrix, template, m, n, sym, data_type)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix, template
      INTEGER                                            :: m, n
      CHARACTER, OPTIONAL                                :: sym
      INTEGER, OPTIONAL                                  :: data_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_m_by_n_from_template', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: mysym
      INTEGER                                            :: my_data_type
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, col_cluster_dist, &
                                                            col_dist, row_blk_size, &
                                                            row_cluster_dist, row_dist
      TYPE(dbcsr_distribution_obj)                       :: dist_m_n

      mysym = cp_dbcsr_get_matrix_type(template)
      IF (PRESENT(sym)) mysym = sym
      my_data_type = cp_dbcsr_get_data_type(template)
      IF (PRESENT(data_type)) my_data_type = data_type

      NULLIFY (row_dist, col_dist)
      NULLIFY (row_blk_size, col_blk_size)
      NULLIFY (row_cluster_dist, col_cluster_dist)

      CALL cp_create_bl_distribution(row_dist, row_blk_size, m, &
                                     dbcsr_mp_nprows(dbcsr_distribution_mp(cp_dbcsr_distribution(template))))
      CALL cp_create_bl_distribution(col_dist, col_blk_size, n, &
                                     dbcsr_mp_npcols(dbcsr_distribution_mp(cp_dbcsr_distribution(template))))
      CALL cp_dbcsr_distribution_new(dist_m_n, dbcsr_distribution_mp(cp_dbcsr_distribution(template)), &
                                     row_dist, col_dist, &
                                     row_cluster_dist, col_cluster_dist, &
                                     reuse_arrays=.TRUE.)

      CALL cp_dbcsr_create(matrix, "m_n_template", dist_m_n, mysym, &
                           row_blk_size, col_blk_size, nze=0, data_type=my_data_type, &
                           reuse_arrays=.TRUE.)
      CALL dbcsr_distribution_release(dist_m_n)

   END SUBROUTINE cp_dbcsr_m_by_n_from_template

! **************************************************************************************************
!> \brief Utility function to create dbcsr matrix, m x n matrix (n arbitrary)
!>        with the same processor grid and row distribution  as the template matrix
!>        col sizes are evenly blocked with the internal
!>        dbcsr_conversion size (32 is the current default)
!> \param matrix dbcsr matrix to be created
!> \param template template dbcsr matrix giving its mp_env
!> \param n global col size of output matrix
!> \param sym ...
!> \param data_type ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_m_by_n_from_row_template(matrix, template, n, sym, data_type)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix, template
      INTEGER                                            :: n
      CHARACTER, OPTIONAL                                :: sym
      INTEGER, OPTIONAL                                  :: data_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_dbcsr_m_by_n_from_row_template', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: mysym
      INTEGER                                            :: my_data_type
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, col_cluster_dist, col_dist
      TYPE(dbcsr_distribution_obj)                       :: dist_m_n

      mysym = cp_dbcsr_get_matrix_type(template)
      IF (PRESENT(sym)) mysym = sym
      my_data_type = cp_dbcsr_get_data_type(template)
      IF (PRESENT(data_type)) my_data_type = data_type

      NULLIFY (col_dist, col_blk_size)

      CALL cp_create_bl_distribution(col_dist, col_blk_size, n, &
                                     dbcsr_mp_npcols(dbcsr_distribution_mp(cp_dbcsr_distribution(template))))
      NULLIFY (col_cluster_dist)
      CALL cp_dbcsr_distribution_new(dist_m_n, dbcsr_distribution_mp(cp_dbcsr_distribution(template)), &
                                     dbcsr_distribution_row_dist(cp_dbcsr_distribution(template)), col_dist, &
                                     dbcsr_distribution_row_clusters(cp_dbcsr_distribution(template)), &
                                     col_cluster_dist)

      CALL cp_dbcsr_create(matrix, "m_n_template", dist_m_n, mysym, &
                           cp_dbcsr_row_block_sizes(template), col_blk_size, nze=0, data_type=my_data_type)

      DEALLOCATE (col_dist, col_blk_size)
      CALL dbcsr_distribution_release(dist_m_n)

   END SUBROUTINE cp_dbcsr_m_by_n_from_row_template

END MODULE cp_dbcsr_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition and initialisation of the mo data type.
!> \par History
!>      - adapted to the new QS environment data structure (02.04.2002,MK)
!>      - set_mo_occupation added (17.04.02,MK)
!>      - correct_mo_eigenvalues added (18.04.02,MK)
!>      - calculate_density_matrix moved from qs_scf to here (22.04.02,MK)
!>      - mo_set_p_type added (23.04.02,MK)
!>      - PRIVATE attribute set for TYPE mo_set_type (23.04.02,MK)
!>      - started conversion to LSD (1.2003, Joost VandeVondele)
!>      - set_mo_occupation moved to qs_mo_occupation (11.12.14 MI)
!>      - correct_mo_eigenvalues moved to qs_scf_methods (03.2016, Sergey Chulkov)
!> \author Matthias Krack (09.05.2001,MK)
! **************************************************************************************************
MODULE qs_mo_types

   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_init_p,&
                                              cp_dbcsr_release_p,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_copy_columns_hack
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_type,&
                                              fm_pool_create_fm
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_mo_types'

   TYPE mo_set_type
      ! the actual MO coefficients as a matrix
      TYPE(cp_fm_type), POINTER                 :: mo_coeff
      TYPE(cp_dbcsr_type), POINTER              :: mo_coeff_b
      ! we are using the dbcsr mo_coeff_b
      LOGICAL                                   :: use_mo_coeff_b
      ! number of molecular orbitals (# cols in mo_coeff)
      INTEGER                                   :: nmo
      ! number of atomic orbitals (# rows in mo_coeff)
      INTEGER                                   :: nao
      ! occupation - eigenvalues  of the nmo states (if eigenstates)
      REAL(KIND=dp), DIMENSION(:), POINTER    :: eigenvalues, occupation_numbers
      ! maximum allowed occupation number of an MO (1 or 2)
      REAL(KIND=dp)                           :: maxocc
      ! number of electrons (taking occupation into account)
      INTEGER                                   :: nelectron
      REAL(KIND=dp)                             :: n_el_f
      ! highest non-zero occupied orbital
      INTEGER                                   :: homo
      ! lowest non maxocc occupied orbital (e.g. fractional or zero)
      INTEGER                                   :: lfomo
      ! flag that indicates if the MOS have the same occupation number
      LOGICAL                                   :: uniform_occupation
      ! the entropic energy contribution
      REAL(KIND=dp)                             :: kTS
      ! Fermi energy level
      REAL(KIND=dp)                             :: mu
      ! Threshold value for multiplicity change
      REAL(KIND=dp)                             :: flexible_electron_count
   END TYPE mo_set_type

   TYPE mo_set_p_type
      TYPE(mo_set_type), POINTER :: mo_set
   END TYPE mo_set_p_type

   PUBLIC :: mo_set_p_type, &
             mo_set_type

   PUBLIC :: allocate_mo_set, &
             deallocate_mo_set, &
             get_mo_set, &
             init_mo_set, &
             set_mo_set, &
             mo_set_restrict, &
             duplicate_mo_set

CONTAINS

! **************************************************************************************************
!> \brief allocate a new mo_set, and copy the old data
!> \param mo_set_new ...
!> \param mo_set_old ...
!> \date 2009-7-19
!> \par History
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE duplicate_mo_set(mo_set_new, mo_set_old)
      TYPE(mo_set_type), POINTER                         :: mo_set_new, mo_set_old

      CHARACTER(LEN=*), PARAMETER :: routineN = 'duplicate_mo_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nmo

      ALLOCATE (mo_set_new)

      mo_set_new%maxocc = mo_set_old%maxocc
      mo_set_new%nelectron = mo_set_old%nelectron
      mo_set_new%n_el_f = mo_set_old%n_el_f
      mo_set_new%nao = mo_set_old%nao
      mo_set_new%nmo = mo_set_old%nmo
      mo_set_new%homo = mo_set_old%homo
      mo_set_new%lfomo = mo_set_old%lfomo
      mo_set_new%uniform_occupation = mo_set_old%uniform_occupation
      mo_set_new%kTS = mo_set_old%kTS
      mo_set_new%mu = mo_set_old%mu
      mo_set_new%flexible_electron_count = mo_set_old%flexible_electron_count

      nmo = mo_set_new%nmo

      NULLIFY (mo_set_new%mo_coeff)
      CALL cp_fm_create(mo_set_new%mo_coeff, mo_set_old%mo_coeff%matrix_struct)
      CALL cp_fm_to_fm(mo_set_old%mo_coeff, mo_set_new%mo_coeff)

      NULLIFY (mo_set_new%mo_coeff_b)
      IF (ASSOCIATED(mo_set_old%mo_coeff_b)) THEN
         CALL cp_dbcsr_init_p(mo_set_new%mo_coeff_b)
         CALL cp_dbcsr_copy(mo_set_new%mo_coeff_b, mo_set_old%mo_coeff_b)
      ENDIF
      mo_set_new%use_mo_coeff_b = mo_set_old%use_mo_coeff_b

      ALLOCATE (mo_set_new%eigenvalues(nmo))
      mo_set_new%eigenvalues = mo_set_old%eigenvalues

      ALLOCATE (mo_set_new%occupation_numbers(nmo))
      mo_set_new%occupation_numbers = mo_set_old%occupation_numbers

   END SUBROUTINE duplicate_mo_set

! **************************************************************************************************
!> \brief Allocates a mo set and partially initializes it (nao,nmo,nelectron,
!>        and flexible_electron_count are vaild).
!>        For the full initialization you need to call init_mo_set
!> \param mo_set the mo_set to allocate
!> \param nao number of atom orbitals
!> \param nmo number of molecular orbitals
!> \param nelectron number of electrons
!> \param n_el_f ...
!> \param maxocc maximum occupation of an orbital (LDA: 2, LSD:1)
!> \param flexible_electron_count the number of electrons can be changed
!> \date 15.05.2001
!> \par History
!>      11.2002 splitted initialization in two phases [fawzi]
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE allocate_mo_set(mo_set, nao, nmo, nelectron, n_el_f, maxocc, &
                              flexible_electron_count)

      TYPE(mo_set_type), POINTER                         :: mo_set
      INTEGER, INTENT(IN)                                :: nao, nmo, nelectron
      REAL(KIND=dp), INTENT(IN)                          :: n_el_f, maxocc, flexible_electron_count

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_mo_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mo_set)) CALL deallocate_mo_set(mo_set)

      ALLOCATE (mo_set)

      mo_set%maxocc = maxocc
      mo_set%nelectron = nelectron
      mo_set%n_el_f = n_el_f
      mo_set%nao = nao
      mo_set%nmo = nmo
      mo_set%homo = 0
      mo_set%lfomo = 0
      mo_set%uniform_occupation = .TRUE.
      mo_set%kTS = 0.0_dp
      mo_set%mu = 0.0_dp
      mo_set%flexible_electron_count = flexible_electron_count

      NULLIFY (mo_set%eigenvalues)
      NULLIFY (mo_set%occupation_numbers)
      NULLIFY (mo_set%mo_coeff)
      NULLIFY (mo_set%mo_coeff_b)
      mo_set%use_mo_coeff_b = .FALSE.

   END SUBROUTINE allocate_mo_set

! **************************************************************************************************
!> \brief initializes an allocated mo_set.
!>      eigenvalues, mo_coeff, occupation_numbers are valid only
!>      after this call.
!> \param mo_set the mo_set to initialize
!> \param fm_pool a pool out which you initialize the mo_set
!> \param fm_ref  a reference  matrix from which you initialize the mo_set
!> \param name ...
!> \par History
!>      11.2002 rewamped [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE init_mo_set(mo_set, fm_pool, fm_ref, name)

      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_fm_pool_type), OPTIONAL, POINTER           :: fm_pool
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_ref
      CHARACTER(LEN=*), INTENT(in)                       :: name

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_mo_set', routineP = moduleN//':'//routineN

      INTEGER                                            :: nao, nmo, nomo

      IF(.NOT.(ASSOCIATED(mo_set)))CALL cp__a("qs_mo_types.F",215)
      IF(.NOT.(.NOT. ASSOCIATED(mo_set%eigenvalues)))CALL cp__a("qs_mo_types.F",216)
      IF(.NOT.(.NOT. ASSOCIATED(mo_set%occupation_numbers)))CALL cp__a("qs_mo_types.F",217)
      IF(.NOT.(.NOT. ASSOCIATED(mo_set%mo_coeff)))CALL cp__a("qs_mo_types.F",218)

      IF(.NOT.(PRESENT(fm_pool) .OR. PRESENT(fm_ref)))CALL cp__a("qs_mo_types.F",220)
      IF (PRESENT(fm_pool)) THEN
         IF(.NOT.(ASSOCIATED(fm_pool)))CALL cp__a("qs_mo_types.F",222)
         CALL fm_pool_create_fm(fm_pool, mo_set%mo_coeff, name=name)
      ELSE IF (PRESENT(fm_ref)) THEN
         IF(.NOT.(ASSOCIATED(fm_ref)))CALL cp__a("qs_mo_types.F",225)
         CALL cp_fm_create(mo_set%mo_coeff, fm_ref%matrix_struct, name=name)
      END IF
      CALL cp_fm_get_info(mo_set%mo_coeff, nrow_global=nao, ncol_global=nmo)
      IF(.NOT.(nao >= mo_set%nao))CALL cp__a("qs_mo_types.F",229)
      IF(.NOT.(nmo >= mo_set%nmo))CALL cp__a("qs_mo_types.F",230)

      ALLOCATE (mo_set%eigenvalues(nmo))
      mo_set%eigenvalues(:) = 0.0_dp

      ALLOCATE (mo_set%occupation_numbers(nmo))
      ! Initialize MO occupations
      mo_set%occupation_numbers(:) = 0.0_dp
      ! Quick return, if no electrons are available
      IF (mo_set%nelectron == 0) THEN
         RETURN
      END IF

      IF (MODULO(mo_set%nelectron, INT(mo_set%maxocc)) == 0) THEN
         nomo = NINT(mo_set%nelectron/mo_set%maxocc)
         mo_set%occupation_numbers(1:nomo) = mo_set%maxocc
      ELSE
         nomo = INT(mo_set%nelectron/mo_set%maxocc)+1
         ! Initialize MO occupations
         mo_set%occupation_numbers(1:nomo-1) = mo_set%maxocc
         mo_set%occupation_numbers(nomo) = mo_set%nelectron-(nomo-1)*mo_set%maxocc
      END IF

      IF(.NOT.(nmo >= nomo))CALL cp__a("qs_mo_types.F",253)
      IF(.NOT.((SIZE(mo_set%occupation_numbers) == nmo)))CALL cp__a("qs_mo_types.F",254)

      mo_set%homo = nomo
      mo_set%lfomo = nomo+1
      mo_set%mu = mo_set%eigenvalues(nomo)

   END SUBROUTINE init_mo_set

! **************************************************************************************************
!> \brief make the beta orbitals explicitly equal to the alpha orbitals
!>       effectively copying the orbital data
!> \param mo_array ...
!> \param convert_dbcsr ...
!> \par History
!>      10.2004 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE mo_set_restrict(mo_array, convert_dbcsr)
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      LOGICAL, INTENT(in), OPTIONAL                      :: convert_dbcsr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mo_set_restrict', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_convert_dbcsr

      CALL timeset(routineN, handle)

      my_convert_dbcsr = .FALSE.
      IF (PRESENT(convert_dbcsr)) my_convert_dbcsr = convert_dbcsr

      IF(.NOT.(ASSOCIATED(mo_array)))CALL cp__a("qs_mo_types.F",285)
      IF(.NOT.(SIZE(mo_array) .EQ. 2))CALL cp__a("qs_mo_types.F",286)
      IF(.NOT.(mo_array(1)%mo_set%nmo >= mo_array(2)%mo_set%nmo))CALL cp__a("qs_mo_types.F",287)

      ! first nmo_beta orbitals are copied from alpha to beta
      IF (my_convert_dbcsr) THEN !fm->dbcsr
         CALL cp_dbcsr_copy_columns_hack(mo_array(2)%mo_set%mo_coeff_b, mo_array(1)%mo_set%mo_coeff_b, & !fm->dbcsr
                                         mo_array(2)%mo_set%nmo, 1, 1, & !fm->dbcsr
                                         para_env=mo_array(1)%mo_set%mo_coeff%matrix_struct%para_env, & !fm->dbcsr
                                         blacs_env=mo_array(1)%mo_set%mo_coeff%matrix_struct%context) !fm->dbcsr
      ELSE !fm->dbcsr
         CALL cp_fm_to_fm(mo_array(1)%mo_set%mo_coeff, mo_array(2)%mo_set%mo_coeff, mo_array(2)%mo_set%nmo)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE mo_set_restrict

! **************************************************************************************************
!> \brief   Deallocate a wavefunction data structure.
!> \param mo_set ...
!> \date    15.05.2001
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_mo_set(mo_set)

      TYPE(mo_set_type), POINTER                         :: mo_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_mo_set', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mo_set)) THEN
         IF (ASSOCIATED(mo_set%eigenvalues)) THEN
            DEALLOCATE (mo_set%eigenvalues)
         END IF
         IF (ASSOCIATED(mo_set%occupation_numbers)) THEN
            DEALLOCATE (mo_set%occupation_numbers)
         END IF
         CALL cp_fm_release(mo_set%mo_coeff)
         IF (ASSOCIATED(mo_set%mo_coeff_b)) CALL cp_dbcsr_release_p(mo_set%mo_coeff_b)
         DEALLOCATE (mo_set)
      END IF

   END SUBROUTINE deallocate_mo_set

! **************************************************************************************************
!> \brief   Get the components of a MO set data structure.
!> \param mo_set ...
!> \param maxocc ...
!> \param homo ...
!> \param lfomo ...
!> \param nao ...
!> \param nelectron ...
!> \param n_el_f ...
!> \param nmo ...
!> \param eigenvalues ...
!> \param occupation_numbers ...
!> \param mo_coeff ...
!> \param mo_coeff_b ...
!> \param uniform_occupation ...
!> \param kTS ...
!> \param mu ...
!> \param flexible_electron_count ...
!> \date    22.04.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_mo_set(mo_set, maxocc, homo, lfomo, nao, nelectron, n_el_f, nmo, &
                         eigenvalues, occupation_numbers, mo_coeff, mo_coeff_b, &
                         uniform_occupation, kTS, mu, flexible_electron_count)

      TYPE(mo_set_type), POINTER                         :: mo_set
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: maxocc
      INTEGER, INTENT(OUT), OPTIONAL                     :: homo, lfomo, nao, nelectron
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: n_el_f
      INTEGER, INTENT(OUT), OPTIONAL                     :: nmo
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: eigenvalues, occupation_numbers
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: mo_coeff
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: mo_coeff_b
      LOGICAL, INTENT(OUT), OPTIONAL                     :: uniform_occupation
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: kTS, mu, flexible_electron_count

      IF (PRESENT(maxocc)) maxocc = mo_set%maxocc
      IF (PRESENT(homo)) homo = mo_set%homo
      IF (PRESENT(lfomo)) lfomo = mo_set%lfomo
      IF (PRESENT(nao)) nao = mo_set%nao
      IF (PRESENT(nelectron)) nelectron = mo_set%nelectron
      IF (PRESENT(n_el_f)) n_el_f = mo_set%n_el_f
      IF (PRESENT(nmo)) nmo = mo_set%nmo
      IF (PRESENT(eigenvalues)) eigenvalues => mo_set%eigenvalues
      IF (PRESENT(occupation_numbers)) THEN
         occupation_numbers => mo_set%occupation_numbers
      END IF
      IF (PRESENT(mo_coeff)) mo_coeff => mo_set%mo_coeff
      IF (PRESENT(mo_coeff_b)) mo_coeff_b => mo_set%mo_coeff_b
      IF (PRESENT(uniform_occupation)) uniform_occupation = mo_set%uniform_occupation
      IF (PRESENT(kTS)) kTS = mo_set%kTS
      IF (PRESENT(mu)) mu = mo_set%mu
      IF (PRESENT(flexible_electron_count)) flexible_electron_count = mo_set%flexible_electron_count

   END SUBROUTINE get_mo_set

! **************************************************************************************************
!> \brief   Set the components of a MO set data structure.
!> \param mo_set ...
!> \param maxocc ...
!> \param homo ...
!> \param lfomo ...
!> \param nao ...
!> \param nelectron ...
!> \param n_el_f ...
!> \param nmo ...
!> \param eigenvalues ...
!> \param occupation_numbers ...
!> \param uniform_occupation ...
!> \param kTS ...
!> \param mu ...
!> \param flexible_electron_count ...
!> \date    22.04.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_mo_set(mo_set, maxocc, homo, lfomo, nao, nelectron, n_el_f, nmo, &
                         eigenvalues, occupation_numbers, uniform_occupation, &
                         kTS, mu, flexible_electron_count)

      TYPE(mo_set_type), POINTER                         :: mo_set
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: maxocc
      INTEGER, INTENT(IN), OPTIONAL                      :: homo, lfomo, nao, nelectron
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: n_el_f
      INTEGER, INTENT(IN), OPTIONAL                      :: nmo
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: eigenvalues, occupation_numbers
      LOGICAL, INTENT(IN), OPTIONAL                      :: uniform_occupation
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: kTS, mu, flexible_electron_count

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_mo_set', routineP = moduleN//':'//routineN

      IF (PRESENT(maxocc)) mo_set%maxocc = maxocc
      IF (PRESENT(homo)) mo_set%homo = homo
      IF (PRESENT(lfomo)) mo_set%lfomo = lfomo
      IF (PRESENT(nao)) mo_set%nao = nao
      IF (PRESENT(nelectron)) mo_set%nelectron = nelectron
      IF (PRESENT(n_el_f)) mo_set%n_el_f = n_el_f
      IF (PRESENT(nmo)) mo_set%nmo = nmo
      IF (PRESENT(eigenvalues)) THEN
         IF (ASSOCIATED(mo_set%eigenvalues)) THEN
            DEALLOCATE (mo_set%eigenvalues)
         END IF
         mo_set%eigenvalues => eigenvalues
      END IF
      IF (PRESENT(occupation_numbers)) THEN
         IF (ASSOCIATED(mo_set%occupation_numbers)) THEN
            DEALLOCATE (mo_set%occupation_numbers)
         END IF
         mo_set%occupation_numbers => occupation_numbers
      END IF
      IF (PRESENT(uniform_occupation)) mo_set%uniform_occupation = uniform_occupation
      IF (PRESENT(kTS)) mo_set%kTS = kTS
      IF (PRESENT(mu)) mo_set%mu = mu
      IF (PRESENT(flexible_electron_count)) mo_set%flexible_electron_count = flexible_electron_count

   END SUBROUTINE set_mo_set

END MODULE qs_mo_types
